Package: aqp
Version: 2.3.1
Title: Algorithms for Quantitative Pedology
Authors@R: c(person(given="Dylan", family="Beaudette", role = c("aut", "cre"), email = "dylan.beaudette@usda.gov", comment = c(ORCID="0009-0008-2780-4785")), 
             person(given="Pierre", family="Roudier", email="roudierp@landcareresearch.co.nz", role = c("aut", "ctb")), 
             person(given="Andrew", family="Brown", email="andrew.g.brown@usda.gov", role = c("aut", "ctb")), 
             person(given="Stephen", family="Roecker", role = c("aut", "ctb")),
             person(given = "David", family = "Rossiter", role = c("ctb")))
Author: Dylan Beaudette [aut, cre] (ORCID:
    <https://orcid.org/0009-0008-2780-4785>),
  Pierre Roudier [aut, ctb],
  Andrew Brown [aut, ctb],
  Stephen Roecker [aut, ctb],
  David Rossiter [ctb]
Maintainer: Dylan Beaudette <dylan.beaudette@usda.gov>
Depends: R (>= 4.1.0)
Imports: grDevices, graphics, stats, utils, methods, grid, lattice,
        cluster, data.table, farver, digest, colorspace, ape
Suggests: mvtnorm, soilDB, sp, sf, latticeExtra, tactile, compositions,
        markovchain, xtable, testthat, Gmedian, Hmisc, tibble,
        RColorBrewer, scales, mpspline2 (>= 0.1.9), soiltexture, gower,
        knitr, rmarkdown, dendextend
Description: The Algorithms for Quantitative Pedology (AQP) project was started in 2009 to organize a loosely-related set of concepts and source code on the topic of soil profile visualization, aggregation, and classification into this package (aqp). Over the past 8 years, the project has grown into a suite of related R packages that enhance and simplify the quantitative analysis of soil profile data. Central to the AQP project is a new vocabulary of specialized functions and data structures that can accommodate the inherent complexity of soil profile information; freeing the scientist to focus on ideas rather than boilerplate data processing tasks <doi:10.1016/j.cageo.2012.10.020>. These functions and data structures have been extensively tested and documented, applied to projects involving hundreds of thousands of soil profiles, and deeply integrated into widely used tools such as SoilWeb <https://casoilresource.lawr.ucdavis.edu/soilweb-apps>. Components of the AQP project (aqp, soilDB, sharpshootR, soilReports packages) serve an important role in routine data analysis within the USDA-NRCS Soil Science Division. The AQP suite of R packages offer a convenient platform for bridging the gap between pedometric theory and practice.
License: GPL (>= 3)
LazyLoad: yes
Repository: CRAN
URL: https://ncss-tech.github.io/aqp/, https://ncss-tech.github.io/AQP/
BugReports: https://github.com/ncss-tech/aqp/issues
Language: en-US
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Config/Needs/website: rmarkdown
NeedsCompilation: no
Packaged: 2026-03-19 17:12:25 UTC; Dylan.Beaudette
Date/Publication: 2026-03-20 06:11:09 UTC

# Generated by roxygen2: do not edit by hand

S3method(.DollarNames,SoilProfileCollection)
export(HzDepthLogicSubset)
export(L1_profiles)
export(NCSP)
export(ReactionClassLevels)
export(SANN_1D)
export(SoilProfileCollection)
export(SoilTextureLevels)
export(accumulateDepths)
export(addBracket)
export(addDiagnosticBracket)
export(addVolumeFraction)
export(aggregateColor)
export(aggregateColorPlot)
export(aggregateSoilDepth)
export(alignTransect)
export(allocate)
export(aqp.env)
export(argillic.clay.increase.depth)
export(barron.torrent.redness.LAB)
export(bootstrapSoilTexture)
export(brierScore)
export(buntley.westin.index)
export(checkHzDepthLogic)
export(checkSPC)
export(col2Munsell)
export(collapseHz)
export(colorChart)
export(colorContrast)
export(colorContrastPlot)
export(colorQuantiles)
export(colorVariation)
export(compareSites)
export(compositeSPC)
export(confusionIndex)
export(contrastChart)
export(contrastClass)
export(correctAWC)
export(crit.clay.argillic)
export(denormalize)
export(depthOf)
export(depthWeights)
export(dissolve_hz)
export(duplicate)
export(electroStatics_1D)
export(equivalentMunsellChips)
export(estimateAWC)
export(estimatePSCS)
export(estimateSoilColor)
export(estimateSoilDepth)
export(evalGenHZ)
export(evalMissingData)
export(explainPlotSPC)
export(fillHzGaps)
export(findOverlap)
export(fixOverlap)
export(flagOverlappingHz)
export(formatMunsell)
export(fragmentClasses)
export(fragmentSieve)
export(fragvol_to_texmod)
export(genSlabLabels)
export(generalize.hz)
export(genhzTableToAdjMat)
export(get.increase.depths)
export(get.increase.matrix)
export(get.ml.hz)
export(get.slice)
export(getArgillicBounds)
export(getCambicBounds)
export(getClosestMunsellChip)
export(getLastHorizonID)
export(getMineralSoilSurfaceDepth)
export(getPlowLayerDepth)
export(getSoilDepthClass)
export(getSurfaceHorizonDepth)
export(glom)
export(glomApply)
export(grepSPC)
export(groupSPC)
export(groupedProfilePlot)
export(guessGenHzLevels)
export(guessHzAttrName)
export(guessHzDesgnName)
export(guessHzTexClName)
export(harden.melanization)
export(harden.rubification)
export(hasDarkColors)
export(horizonColorIndices)
export(huePosition)
export(huePositionCircle)
export(hurst.redness)
export(hzAbove)
export(hzBelow)
export(hzDepthTests)
export(hzDistinctnessCodeToOffset)
export(hzOffset)
export(hzTopographyCodeToLineType)
export(hzTopographyCodeToOffset)
export(hzTransitionProbabilities)
export(hz_dissolve)
export(hz_intersect)
export(hz_lag)
export(hz_segment)
export(hz_to_taxpartsize)
export(invertLabelColor)
export(launderMunsell)
export(lookup_taxpartsize)
export(lunique)
export(maxDepthOf)
export(minDepthOf)
export(missingDataGrid)
export(mixMunsell)
export(mollic.thickness.requirement)
export(mostLikelyHzSequence)
export(mu_confusion_matrix)
export(munsell2rgb)
export(mutate_profile)
export(mutate_profile_raw)
export(overlapMetrics)
export(panel.depth_function)
export(parseMunsell)
export(pbindlist)
export(perturb)
export(ph_to_rxnclass)
export(plotColorMixture)
export(plotColorQuantiles)
export(plotMultipleSPC)
export(plotProfileDendrogram)
export(plotSPC)
export(plot_distance_graph)
export(prepanel.depth_function)
export(previewColors)
export(profileGroupLabels)
export(profileInformationIndex)
export(quickSPC)
export(random_profile)
export(rebuildSPC)
export(reduceSPC)
export(repairMissingHzDepths)
export(rgb2munsell)
export(rp)
export(rxnclass_to_ph)
export(segment)
export(shannonEntropy)
export(shuffle)
export(sim)
export(simulateColor)
export(slab_function)
export(slice.fast)
export(slicedHSD)
export(soilColorSignature)
export(soilPalette)
export(soilTextureColorPal)
export(spc_in_sync)
export(spec2Munsell)
export(splitLogicErrors)
export(ssc_to_texcl)
export(subApply)
export(summarizeSPC)
export(summaryTauW)
export(tauW)
export(texcl_to_ssc)
export(texmod_to_fragvoltot)
export(textureTriangleSummary)
export(texture_to_taxpartsize)
export(texture_to_texmod)
export(thicknessOf)
export(thompson.bell.darkness)
export(validateMunsell)
export(warpHorizons)
export(xtableTauW)
exportClasses(SoilProfileCollection)
exportMethods("GHL<-")
exportMethods("aqp_df_class<-")
exportMethods("coordinates<-")
exportMethods("depth_units<-")
exportMethods("depths<-")
exportMethods("diagnostic_hz<-")
exportMethods("horizonDepths<-")
exportMethods("horizonNames<-")
exportMethods("horizons<-")
exportMethods("hzID<-")
exportMethods("hzdesgnname<-")
exportMethods("hzidname<-")
exportMethods("hzmetaname<-")
exportMethods("hztexclname<-")
exportMethods("initSpatial<-")
exportMethods("metadata<-")
exportMethods("prj<-")
exportMethods("profile_id<-")
exportMethods("proj4string<-")
exportMethods("replaceHorizons<-")
exportMethods("restrictions<-")
exportMethods("site<-")
exportMethods("siteNames<-")
exportMethods(GHL)
exportMethods(aqp_df_class)
exportMethods(as.character)
exportMethods(as.data.frame)
exportMethods(c)
exportMethods(combine)
exportMethods(coordinates)
exportMethods(depth_units)
exportMethods(depths)
exportMethods(diagnostic_hz)
exportMethods(dice)
exportMethods(generalizeHz)
exportMethods(getSpatial)
exportMethods(harmonize)
exportMethods(horizonDepths)
exportMethods(horizonNames)
exportMethods(horizons)
exportMethods(hzDesgn)
exportMethods(hzID)
exportMethods(hzMetadata)
exportMethods(hzdesgnname)
exportMethods(hzidname)
exportMethods(hzmetaname)
exportMethods(hztexclname)
exportMethods(idname)
exportMethods(isEmpty)
exportMethods(length)
exportMethods(max)
exportMethods(metadata)
exportMethods(min)
exportMethods(munsell2spc)
exportMethods(names)
exportMethods(nrow)
exportMethods(plot)
exportMethods(prj)
exportMethods(profileApply)
exportMethods(profile_id)
exportMethods(proj4string)
exportMethods(reorderHorizons)
exportMethods(restrictions)
exportMethods(site)
exportMethods(siteNames)
exportMethods(slab)
exportMethods(slice)
exportMethods(spc2mpspline)
exportMethods(split)
exportMethods(subset)
exportMethods(subsetHz)
exportMethods(subsetProfiles)
exportMethods(transform)
exportMethods(trunc)
exportMethods(unique)
exportMethods(validSpatialData)
import(data.table)
importFrom(ape,as.phylo)
importFrom(ape,plot.phylo)
importFrom(ape,tiplabels)
importFrom(cluster,daisy)
importFrom(cluster,pam)
importFrom(cluster,silhouette)
importFrom(colorspace,deutan)
importFrom(colorspace,protan)
importFrom(colorspace,tritan)
importFrom(digest,digest)
importFrom(farver,compare_colour)
importFrom(farver,convert_colour)
importFrom(grDevices,chull)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRamp)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,colors)
importFrom(grDevices,convertColor)
importFrom(grDevices,grey)
importFrom(grDevices,hsv)
importFrom(grDevices,rgb)
importFrom(grDevices,rgb2hsv)
importFrom(graphics,abline)
importFrom(graphics,arrows)
importFrom(graphics,axis)
importFrom(graphics,box)
importFrom(graphics,grconvertX)
importFrom(graphics,grid)
importFrom(graphics,image)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(graphics,rect)
importFrom(graphics,segments)
importFrom(graphics,strheight)
importFrom(graphics,strwidth)
importFrom(graphics,text)
importFrom(grid,gpar)
importFrom(grid,grid.text)
importFrom(grid,unit)
importFrom(lattice,levelplot)
importFrom(lattice,panel.abline)
importFrom(lattice,panel.grid)
importFrom(lattice,panel.lines)
importFrom(lattice,panel.points)
importFrom(lattice,panel.polygon)
importFrom(lattice,panel.rect)
importFrom(lattice,panel.segments)
importFrom(lattice,panel.text)
importFrom(lattice,strip.custom)
importFrom(lattice,trellis.par.get)
importFrom(lattice,xyplot)
importFrom(methods,"slot<-")
importFrom(methods,.hasSlot)
importFrom(methods,as)
importFrom(methods,isGeneric)
importFrom(methods,new)
importFrom(methods,prototype)
importFrom(methods,representation)
importFrom(methods,setClass)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(methods,setOldClass)
importFrom(methods,setReplaceMethod)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(stats,TukeyHSD)
importFrom(stats,aggregate)
importFrom(stats,aov)
importFrom(stats,approxfun)
importFrom(stats,as.dist)
importFrom(stats,as.formula)
importFrom(stats,as.hclust)
importFrom(stats,cmdscale)
importFrom(stats,complete.cases)
importFrom(stats,cov)
importFrom(stats,dist)
importFrom(stats,formula)
importFrom(stats,median)
importFrom(stats,model.frame)
importFrom(stats,na.omit)
importFrom(stats,na.pass)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,splinefun)
importFrom(stats,terms)
importFrom(stats,update)
importFrom(stats,weighted.mean)
importFrom(stats,xtabs)
importFrom(utils,.DollarNames)
importFrom(utils,capture.output)
importFrom(utils,combn)
importFrom(utils,object.size)
importFrom(utils,packageVersion)
importFrom(utils,str)

## ----setup, echo=FALSE, results='hide', warning=FALSE---------------------------------------------
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = as.integer(Sys.getenv("R_AQP_VIGNETTE_IMAGE_DPI", unset = 32)),
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)

## ----setup, echo=FALSE, results='hide', warning=FALSE---------------------------------------------
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = 96,
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)

suppressMessages(library(aqp, quietly = TRUE))

## ----echo = FALSE, fig.width=8, fig.height=4------------------------------------------------------
.m <- '10YR 6/6'

data("munsell.spectra.wide")
w <- munsell.spectra.wide[, 1]
s <- munsell.spectra.wide[, .m]

par(mar = c(4.5, 4.5, 2, 1), cex.axis = 0.75, lend = 2)
plot(w, s, xlab = 'Wavelength (nm)', ylab = 'Reflectance', type = 'b', main = .m, cex = 0.8, pch = 16, axes = FALSE)
rect(xleft = 400, ybottom = 0.35, xright = 450, ytop = 0.4, col = parseMunsell(.m), border = 1, lwd = 1)
axis(side = 1, at = seq(380, 730, by = 20))
axis(side = 2, las = 1)

## -------------------------------------------------------------------------------------------------
# Munsell -> hex color
parseMunsell('5PB 4/6')

# Munsell -> sRGB
parseMunsell('5PB 4/6',  return_triplets = TRUE)

# Munsell -> CIELAB
parseMunsell('5PB 4/6',  returnLAB = TRUE)

# hex color -> Munsell
col2Munsell('#476189FF')

# neutral color
parseMunsell('N 5/')

# non-standard notation
getClosestMunsellChip('3.3YR 4.4/6.1', convertColors = FALSE)

## ----fig.width=10, fig.height=4-------------------------------------------------------------------
# example moist soil colors from the Musick soil series
m <- c("7.5YR 2.5/1", "10YR 3/2", "7.5YR 4/3", "2.5YR 3/6", "2.5YR 3/6", 
       "2.5YR 4/6", "5YR 4/6", "7.5YR 5/4")

mm <- parseMunsell(m, convertColors = FALSE)

d.p <- estimateSoilColor(
  hue = mm$hue, 
  value = mm$value, 
  chroma = mm$chroma, 
  method = 'procrustes', 
  sourceMoistureState = 'moist'
)

d.ols <- estimateSoilColor(
  hue = mm$hue, 
  value = mm$value, 
  chroma = mm$chroma, 
  method = 'ols', 
  sourceMoistureState = 'moist'
)

d.p <- formatMunsell(d.p$hue, d.p$value, d.p$chroma)
d.ols <- formatMunsell(d.ols$hue, d.ols$value, d.ols$chroma)

colorContrastPlot(m, d.p, labels = c('Moist', 'Estimated\nDry'), d.cex = 0.9)

# it is the same
# colorContrastPlot(m, d.ols, labels = c('Moist', 'Estimated\nDry'), d.cex = 0.9)

## ----fig.width=6, fig.height=3--------------------------------------------------------------------
data("Ohz.colors")

Ohz.colors$col <- parseMunsell(Ohz.colors$L1.munsell)

op <- par(mfrow = c(2, 1), mar = c(0.5, 0.5, 1.5, 0))

with(
  Ohz.colors[Ohz.colors$state == 'dry', ],
  soilPalette(colors = col, lab = sprintf("%s\n%s", genhz, L1.munsell), lab.cex = 1)
)
title(main = 'Dry Colors')

with(
  Ohz.colors[Ohz.colors$state == 'moist', ],
  soilPalette(colors = col, lab = sprintf("%s\n%s", genhz, L1.munsell), lab.cex = 1)
)
title(main = 'Moist Colors')

# restore original base graphics state
par(op)

## ----setup, echo=FALSE, results='hide', warning=FALSE---------------------------------------------
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = 96,
  optipng = knitr::hook_optipng,
  comment = "#>"
)
options(width = 100, stringsAsFactors = FALSE, timeout = 600)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

## ----install, eval=FALSE--------------------------------------------------------------------------
# # install CRAN release + dependencies
# install.packages('aqp', dependencies = TRUE)
# install.packages('remotes', dependencies = TRUE)
# 
# # install latest version from GitHub
# remotes::install_github("ncss-tech/aqp", dependencies = FALSE)

## ----load-packages--------------------------------------------------------------------------------
library(aqp)
library(lattice)

## ----creation-------------------------------------------------------------------------------------
# load sample data set, a data.frame object with horizon-level data from 10 profiles
data(sp4)
str(sp4)

# optionally read about it...
# ?sp4

# upgrade to SoilProfileCollection
# 'id' is the name of the column containing the profile ID
# 'top' is the name of the column containing horizon upper boundaries
# 'bottom' is the name of the column containing horizon lower boundaries
depths(sp4) <- id ~ top + bottom

# define "horizon designation" column name for the collection
hzdesgnname(sp4) <- 'name'

# check it out:
class(sp4)
print(sp4)

## ----promote_to_spc, eval=FALSE-------------------------------------------------------------------
# idcolumn ~ hz_top_column + hz_bottom_column

## ----fig.width = 6--------------------------------------------------------------------------------
# character vector of horizon templates
# must all use the same formatting
x <- c(
  'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
  'P2:ApAp|AE|BhsBhs|Bw1Bw1|Bw2|CCCCCCC',
  'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRRR'
)

# each horizon label is '10' depth-units (default)
s <- quickSPC(x)

# sketch profiles
par(mar = c(0, 0, 0, 0))
plotSPC(s, name.style = 'center-center', 
        cex.names = 1, depth.axis = FALSE, 
        hz.depths = TRUE
)

## ----hz_and_site_data-----------------------------------------------------------------------------
# site-level (based on length of assigned data == number of profiles)
sp4$elevation <- rnorm(n = length(sp4), mean = 1000, sd = 150) 

# horizon-level (calculated from two horizon-level columns)
sp4$thickness <- sp4$bottom - sp4$top 

# extraction of specific attributes by name
sp4$clay # vector of clay content (horizon data)
sp4$elevation # vector of simulated elevation (site data)

# unit-length value explicitly targeting site data
site(sp4)$collection_id <- 1

# assign a single single value into horizon-level attributes
sp4$constant <- rep(1, times = nrow(sp4))

# unit-length value explicitly targeting horizon data
horizons(sp4)$analysis_group <- "SERP"

# _moves_ the named column from horizon to site
site(sp4) <- ~ constant 

## ----hz_and_site_data_2---------------------------------------------------------------------------
# extract horizon data to data.frame
h <- horizons(sp4)

# add a new column and save back to original object
h$random.numbers <- rnorm(n = nrow(h), mean = 0, sd = 1)

# _replace_ original horizon data with modified version
replaceHorizons(sp4) <- h

# extract site data to data.frame
s <- site(sp4)

# add a fake group to the site data
s$group <- factor(rep(c('A', 'B'), length.out = nrow(s)))

# join new site data with previous data: old data are _not_ replaced
site(sp4) <- s

# check
sp4

## ----diagnostic-----------------------------------------------------------------------------------
dh <- data.frame(
  id = 'colusa',
  kind = 'argillic',
  top = 8,
  bottom = 42,
  stringsAsFactors = FALSE
)

# overwrite any existing diagnostic horizon data
diagnostic_hz(sp4) <- dh

# append to diagnostic horizon data
dh <- diagnostic_hz(sp4)
dh.new <- data.frame(
  id = 'napa',
  kind = 'argillic',
  top = 6,
  bottom = 20,
  stringsAsFactors = FALSE
)

# overwrite existing diagnostic horizon data with appended data
diagnostic_hz(sp4) <- rbind(dh, dh.new)

## ----restrictions---------------------------------------------------------------------------------
# get the depth of each profile
rf.top <- profileApply(sp4, max)
rf.bottom <- rf.top + 20

# the profile IDs can be extracted from the names attribute
pIDs <- names(rf.top)

# note: profile IDs must be stored in a column named for idname(sp4) -> 'id'
rf <- data.frame(
  id = pIDs, 
  top = rf.top, 
  bottom = rf.bottom, 
  kind = 'fake',
  stringsAsFactors = FALSE
)

# overwrite any existing diagnostic horizon data
restrictions(sp4) <- rf

# check
restrictions(sp4)

## ----metadata-------------------------------------------------------------------------------------
# metadata structure
str(metadata(sp4))

# alter the depth unit metadata attribute
depth_units(sp4) <- 'inches' # units are really 'cm'

# add or replace custom metadata
metadata(sp4)$describer <- 'DGM'
metadata(sp4)$date <- as.Date('2009-01-01')
metadata(sp4)$citation <- 'McGahan, D.G., Southard, R.J, Claassen, V.P. 2009. Plant-Available Calcium Varies Widely in Soils on Serpentinite Landscapes. Soil Sci. Soc. Am. J. 73: 2087-2095.'

# check new values have been added
str(metadata(sp4))

# fix depth units, set back to 'cm'
depth_units(sp4) <- 'cm'

## -------------------------------------------------------------------------------------------------
# generate some fake coordinates as site level attributes
sp4$x <- rnorm(n = length(sp4), mean = 354000, sd = 100)
sp4$y <- rnorm(n = length(sp4), mean = 4109533, sd = 100)

# initialize spatial coordinates (CRS optional)
initSpatial(sp4, crs = "EPSG:26911") <- ~ x + y

# extract coordinates as matrix
getSpatial(sp4)

# get/set spatial reference system using prj()<-
prj(sp4) <- '+proj=utm +zone=11 +datum=NAD83'

# return CRS information
prj(sp4)

if (requireNamespace("sf", quietly = TRUE)) {
  # extract spatial data + site level attributes in new spatial object
  sp4.sp <- as(sp4, 'SpatialPointsDataFrame')
  sp4.sf <- as(sp4, 'sf')
}

## ----eval=FALSE-----------------------------------------------------------------------------------
# checkSPC(sp4)

## -------------------------------------------------------------------------------------------------
spc_in_sync(sp4)

## -------------------------------------------------------------------------------------------------
z <- rebuildSPC(sp4)

## -------------------------------------------------------------------------------------------------
checkHzDepthLogic(sp4)
checkHzDepthLogic(sp4, byhz = TRUE)

## ----coercion, eval=FALSE-------------------------------------------------------------------------
# # check our work by viewing the internal structure
# str(sp4)
# 
# # create a data.frame from horizon+site data
# as(sp4, 'data.frame')
# 
# # or, equivalently:
# as.data.frame(sp4)
# 
# # convert SoilProfileCollection to a named list containing all slots
# as(sp4, 'list')
# 
# # extraction of site + spatial data as SpatialPointsDataFrame
# as(sp4, 'SpatialPointsDataFrame')

## ----eval=FALSE-----------------------------------------------------------------------------------
# # explicit string matching
# idx <- which(sp4$group == 'A')
# 
# # numerical expressions
# idx <- which(sp4$elevation < 1000)
# 
# # regular expression, matches any profile ID containing 'shasta'
# idx <- grep('shasta', profile_id(sp4), ignore.case = TRUE)
# 
# # perform subset based on index
# sp4[idx, ]

## ----eval=FALSE-----------------------------------------------------------------------------------
# subset(sp4, group == 'A')
# subset(sp4, elevation < 1000)
# subset(sp4, grepl('shasta', id, ignore.case = TRUE))

## ----eval = FALSE---------------------------------------------------------------------------------
# sp4[, 2]

## ----eval = FALSE---------------------------------------------------------------------------------
# sp4[, , .FIRST]
# sp4[, , .LAST]

## ----eval = FALSE---------------------------------------------------------------------------------
# sp4[, , .FIRST, .TOP]
# sp4[, , .LAST, .HZID]

## ----concatenation, eval=FALSE--------------------------------------------------------------------
# # subset data into chunks
# s1 <- sp4[1:2, ]
# s2 <- sp4[4, ]
# s3 <- sp4[c(6, 8, 9), ]
# 
# # combine subsets
# s <- combine(list(s1, s2, s3))
# 
# # double-check result
# plotSPC(s)

## ----eval=FALSE-----------------------------------------------------------------------------------
# # sample data as data.frame objects
# data(sp1)
# data(sp3)
# 
# # rename IDs horizon top / bottom columns
# sp3$newid <- sp3$id
# sp3$hztop <- sp3$top
# sp3$hzbottom <- sp3$bottom
# 
# # remove originals
# sp3$id <- NULL
# sp3$top <- NULL
# sp3$bottom <- NULL
# 
# # promote to SoilProfileCollection
# depths(sp1) <- id ~ top + bottom
# depths(sp3) <- newid ~ hztop + hzbottom
# 
# # label each group via site-level attribute
# site(sp1)$g <- 'sp1'
# site(sp3)$g <- 'sp3'
# 
# # combine
# x <- combine(list(sp1, sp3))
# 
# # make grouping variable into a factor for groupedProfilePlot
# x$g <- factor(x$g)
# 
# # check results
# str(x)
# 
# # graphical check
# # convert character horizon IDs into numeric
# x$.horizon_ids_numeric <- as.numeric(hzID(x))
# 
# par(mar = c(0, 0, 3, 1))
# plotSPC(x, color='.horizon_ids_numeric', col.label = 'Horizon ID')
# groupedProfilePlot(x, 'g', color='.horizon_ids_numeric', col.label = 'Horizon ID', group.name.offset = -15)

## ----eval=FALSE-----------------------------------------------------------------------------------
# # continuing from above
# # split subsets of x into a list of SoilProfileCollection objects using site-level attribute 'g'
# res <- split(x, 'g')
# str(res, 1)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
d <- duplicate(sp4[1, ], times = 8)
par(mar = c(0, 2, 0, 1))
plotSPC(d, color = 'ex_Ca_to_Mg')

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
# an example soil profile  
x <- data.frame(
  id = 'A',
  name = c('A', 'E', 'Bhs', 'Bt1', 'Bt2', 'BC', 'C'),
  top = c(0, 10, 20, 30, 40, 50, 100),
  bottom = c(10, 20, 30, 40, 50, 100, 125),
  z = c(8, 5, 3, 7, 10, 2, 12)
)

# init SPC
depths(x) <- id ~ top + bottom
hzdesgnname(x) <- 'name'

# horizon depth variability for simulation
horizons(x)$.sd <- 2

# duplicate several times
x.dupes <- duplicate(x, times = 5)

# simulate some new profiles based on example
# 2cm constant standard deviation of transition between horizons assumed
x.sim <- perturb(x, n = 5, thickness.attr = '.sd')

# graphical check
par(mar = c(0, 2, 0, 4))

# inspect unique results
plotSPC(unique(x.dupes, vars = c('top', 'bottom')),
        name.style = 'center-center',
        width = 0.15)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
# "uniqueness" is a function of variables selected to consider
plotSPC(unique(x.sim, vars = c('top', 'bottom')),
        name.style = 'center-center')

plotSPC(unique(x.sim, vars = c('name')),
        name.style = 'center-center',
        width = 0.15)

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
explainPlotSPC(sp4, name = 'name')

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
explainPlotSPC(sp4, name = 'name', width = 0.3)

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
explainPlotSPC(sp4, name = 'name', y.offset = 5)

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
explainPlotSPC(sp4, name = 'name', y.offset = -10)

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
explainPlotSPC(sp4, name = 'name', scaling.factor = 0.5)

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
explainPlotSPC(sp4, name = 'name', plot.order = length(sp4):1)

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
explainPlotSPC(sp4, name = 'name', n = length(sp4) + 2)

## ----fig.width=5, fig.height=6--------------------------------------------------------------------
data(osd)
x <- osd

## ----fig.width=5, fig.height=6--------------------------------------------------------------------
par(mar = c(0, 2, 0, 4), xpd = NA)
plotSPC(x[1, ], cex.names = 1)

## ----fig.width=7, fig.height=6--------------------------------------------------------------------
# set margins and turn off clipping
par(mar = c(0, 2, 0, 4), xpd = NA)
plotSPC(x[1:2, ], cex.names = 1, width = 0.25)

## ----fig.width=8, fig.height=6--------------------------------------------------------------------
par(mar = c(0, 2, 0, 4), xpd = NA)
plotSPC(x, cex.names = 1, depth.axis = list(line = -0.1), width = 0.3)

## ----fig.width=8, fig.height=6--------------------------------------------------------------------
par(mar = c(0, 0, 1, 1))
plotSPC(
  x,
  cex.names = 1,
  name.style = 'center-center',
  width = 0.3,
  depth.axis = FALSE,
  hz.depths = TRUE,
  hz.depths.offset = 0.08
)

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
pos <- jitter(1:length(sp4))
explainPlotSPC(sp4, name = 'name', relative.pos = pos)

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
pos <- c(1, 1.2, 3, 4, 5, 5.2, 7, 8, 9, 10)
explainPlotSPC(sp4, name = 'name', relative.pos = pos)

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
new.pos <- fixOverlap(pos)
explainPlotSPC(sp4, name = 'name', relative.pos = new.pos)

## ----fig.height=5, fig.width=9--------------------------------------------------------------------
par(mar = c(4, 3, 2, 2))
new.pos <- fixOverlap(pos, thresh = 0.7)
explainPlotSPC(sp4, name = 'name', relative.pos = new.pos)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
par(mar = c(0, 0, 3, 0))
plotSPC(sp4,
        name = 'name',
        color = 'clay',
        col.label = 'Clay Content (%)')

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
par(mar = c(0, 0, 3, 0)) 
plotSPC(
  sp4,
  name = 'name',
  color = 'clay',
  col.label = 'Clay Content (%)'
)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
par(mar = c(0, 0, 3, 0))
plotSPC(
  sp4,
  name = 'name',
  color = 'name',
  col.label = 'Original Horizon Name'
)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
par(mar = c(0, 0, 3, 0)) 

# generalize horizon names into 3 groups
sp4$genhz <- generalize.hz(sp4$name, new = c('A', 'AB', 'Bt'), pat = c('A[0-9]?', 'AB', '^Bt'))

plotSPC(
  sp4,
  name = 'name',
  color = 'genhz',
  col.label = 'Generalized Horizon Name'
)

## ----plotting-vol-fraction, fig.height=4, fig.width=8---------------------------------------------
par(mar = c(0, 0, 3, 0)) 
# convert coarse rock fragment proportion to percentage
sp4$frag_pct <- sp4$CF * 100

# label horizons with fragment percent
plotSPC(sp4, name = 'frag_pct', color = 'frag_pct')

# symbolize volume fraction data
addVolumeFraction(sp4, colname = 'frag_pct')

## ----plotting-depth-interval----------------------------------------------------------------------
# extract top/bottom depths associated with all A horizons
tops <- minDepthOf(sp4, pattern = '^A', hzdesgn = 'name', top = TRUE)
bottoms <- maxDepthOf(sp4, pattern = '^A', hzdesgn = 'name', top = FALSE)

IDs <- profile_id(sp4)

# assemble data.frame
a <- data.frame(id = IDs, top = tops$top, bottom = bottoms$bottom)

# check
a

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
par(mar = c(0, 0, 0, 0))
plotSPC(sp4)

# annotate A horizon depth interval with brackets
addBracket(a, col = 'red', offset = -0.4)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
par(mar = c(0, 0, 0, 0))
plotSPC(sp4, name = 'name')

# label
a$label <- site(sp4)$id

# note that depth brackets "know which profiles to use" via profile ID
addBracket(
  a,
  labcol = 'label',
  col = 'red',
  label.cex = 0.75,
  missing.bottom.depth = 25,
  offset = -0.4
)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
par(mar = c(0, 0, 0, 0)) 
groupedProfilePlot(sp4, groups = 'group')
addBracket(a, col = 'red', offset = -0.4)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
par(mar = c(0, 0, 0, 0))
a.sub <- a[1:4,]
groupedProfilePlot(sp4, groups = 'group')
addBracket(a.sub, col = 'red', offset = -0.4)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
a$bottom <- NA
par(mar = c(0, 0, 0, 0))
groupedProfilePlot(sp4, groups = 'group')
addBracket(a, col = 'red', offset = -0.4)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
par(mar = c(0, 0, 0, 0))
groupedProfilePlot(sp4, groups = 'group')
addBracket(
  a,
  col = 'red',
  label.cex = 0.75,
  missing.bottom.depth = 25,
  offset = -0.4
)

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
par(mar = c(0, 0, 0, 0))
plotSPC(sp4, max.depth = 75)

# copy root-restricting features
a <- restrictions(sp4)

# add a label: restrictive feature 'kind'
a$label <- a$kind

# add restrictions using vertical bars
addBracket(
  a,
  labcol = 'label',
  col = 'red',
  label.cex = 0.75,
  tick.length = 0,
  lwd = 3,
  offset = -0.4
)

## ----eval=FALSE-----------------------------------------------------------------------------------
# # library(svglite)
# # svglite(filename = 'e:/temp/fig.svg', width = 7, height = 6, pointsize = 12)
# #
# # par(mar = c(0,2,0,4), xpd = NA)
# # plotSPC(x, cex.names=1, depth.axis = list(line = -0.2), width=0.3)
# #
# # dev.off()

## ----profileApply-1-------------------------------------------------------------------------------
# max() returns the depth of a soil profile
sp4$soil.depth <- profileApply(sp4, FUN = max)

# max() with additional argument give max depth to non-missing 'clay'
sp4$soil.depth.clay <- profileApply(sp4, FUN = max, v = 'clay')

# nrow() returns the number of horizons
sp4$n.hz <- profileApply(sp4, FUN = nrow)

# compute the mean clay content by profile using an inline function
sp4$mean.clay <- profileApply(sp4, FUN = function(i) mean(i$clay))

# estimate soil depth based on horizon designation
sp4$soil.depth <- profileApply(sp4, estimateSoilDepth, name = 'name')

## ----profileApply-2-------------------------------------------------------------------------------
# save as horizon-level attribute
sp4$delta.clay <- profileApply(sp4, FUN = function(i) c(NA, diff(i$clay)))

# check results:
horizons(sp4)[1:6, c('id', 'top', 'bottom', 'clay', 'delta.clay')]

## ----profileApply-3-------------------------------------------------------------------------------
# compute hz-thickness weighted mean exchangeable-Ca:Mg
wt.mean.ca.mg <- function(i) {
  # use horizon thickness as a weight
  thick <- i$bottom - i$top
  
  # compute the thickness weighted mean, ignoring missing values
  weighted.mean(i$ex_Ca_to_Mg, w = thick, na.rm = TRUE)
}

# apply our custom function and save results as a site-level attribute
sp4$wt.mean.ca.to.mg <- profileApply(sp4, wt.mean.ca.mg)

## ----profileApply-4, fig.height=5, fig.width=8----------------------------------------------------
# plot the data using our new order based on Ca:Mg weighted average
# the result is an index of rank 
new.order <- order(sp4$wt.mean.ca.to.mg)

par(mar = c(4, 0, 3, 0)) # tighten figure margins
plotSPC(sp4,
        name = 'name',
        color = 'ex_Ca_to_Mg',
        plot.order = new.order)

# add an axis labeled with the sorting criteria
axis(1, at = 1:length(sp4), labels = round(sp4$wt.mean.ca.to.mg, 3), cex.axis = 0.75)
mtext(1, line = 2.25, text = 'Horizon Thickness Weighted Mean Ex. Ca:Mg', cex = 0.75)

## ----slice-formula, eval=FALSE--------------------------------------------------------------------
# # slice select horizon-level attributes
# seq ~ var.1 + var.2 + var.3 + ...
# # slice all horizon-level attributes
# seq ~ .

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
# resample to 1cm slices
s <- dice(sp4, fm = 0:15 ~ sand + silt + clay + name + ex_Ca_to_Mg)

# check the result
class(s)

# plot sliced data
par(mar = c(0, 0, 3, 0)) # tighten figure margins
plotSPC(s, color = 'ex_Ca_to_Mg')

## ----eval=FALSE-----------------------------------------------------------------------------------
# # slice from 0 to max depth in the collection
# s <- dice(sp4, fm= 0:max(sp4) ~ sand + silt + clay + name + ex_Ca_to_Mg)
# 
# # extract all data over the range of 5--10 cm:
# plotSPC(s[, 5:10])
# 
# # extract all data over the range of 25--50 cm:
# plotSPC(s[, 25:50])
# 
# # extract all data over the range of 10--20 and 40--50 cm:
# plotSPC(s[, c(10:20, 40:50)])

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
# truncate to the interval 5-15cm
clods <- glom(sp4, z1 = 5, z2 = 15)

# plot outlines of original profiles
par(mar = c(0, 0, 3, 1.5)) 
plotSPC(sp4, color = NA, name = NA, print.id = FALSE, depth.axis = FALSE, lwd = 0.5)

# overlay glom() depth criteria
rect(xleft = 0.5, ybottom = 15, xright = length(sp4) + 0.5, ytop = 5, lty = 2)

# add SoilProfileCollection with selected horizons
plotSPC(clods, add = TRUE, cex.names = 0.6, name = 'name', color = 'ex_Ca_to_Mg', name.style = 'center-center')

## ----fig.height=4, fig.width=8--------------------------------------------------------------------
# truncate to the interval 5-15cm
sp4.truncated <- trunc(sp4, 5, 15)

# plot outlines of original profiles
par(mar = c(0, 0, 3, 1.5)) 
plotSPC(sp4, color = NA, name = NA, print.id = FALSE, lwd = 0.5)

# overlay truncation criteria
rect(xleft = 0.5, ybottom = 15, xright = length(sp4) + 0.5, ytop = 5, lty = 2)

# add truncated SoilProfileCollection
plotSPC(sp4.truncated, depth.axis = FALSE, add = TRUE, cex.names = 0.6, name = 'name', color = 'ex_Ca_to_Mg', name.style = 'center-center')

## ----slab, fig.height=4, fig.width=8--------------------------------------------------------------
# aggregate a couple of the horizon-level attributes, 
# across the entire collection, 
# from 0--10 cm
# computing the mean value ignoring missing data
slab(
  sp4,
  fm = ~ sand + silt + clay,
  slab.structure = c(0, 10),
  slab.fun = mean,
  na.rm = TRUE
)

# again, this time within groups defined by a site-level attribute:
slab(
  sp4,
  fm = group ~ sand + silt + clay,
  slab.structure = c(0,  10),
  slab.fun = mean,
  na.rm = TRUE
)

# again, this time over several depth ranges
slab(
  sp4,
  fm = ~ sand + silt + clay,
  slab.structure = c(0, 10, 25, 40),
  slab.fun = mean,
  na.rm = TRUE
)

# again, this time along 1-cm slices, computing quantiles
agg <- slab(sp4, fm = ~ Mg + Ca + ex_Ca_to_Mg + CEC_7 + clay)

# see ?slab for details on the default aggregate function
head(agg)

# plot median +/i bounds defined by the 25th and 75th percentiles
# this is lattice graphics, syntax is a little rough
xyplot(top ~ p.q50 | variable, data = agg, ylab = 'Depth', 
       xlab = 'median bounded by 25th and 75th percentiles',
       lower = agg$p.q25, upper = agg$p.q75, ylim = c(42, -2),
       panel = panel.depth_function,
       alpha = 0.25, sync.colors = TRUE,
       par.settings = list(superpose.line = list(col = 'RoyalBlue', lwd = 2)), 
       prepanel = prepanel.depth_function,
       cf = agg$contributing_fraction, cf.col = 'black', cf.interval = 5,
       layout = c(5, 1), strip = strip.custom(bg = grey(0.8)),
       scales = list(x = list(
         tick.number = 4,
         alternating = 3,
         relation = 'free'
       ))
)

## ----slab-2, fig.height=4, fig.width=8------------------------------------------------------------
# processing the "napa" and tehama profiles
idx <- which(profile_id(sp4) %in% c('napa', 'tehama'))
napa.and.tehama <- slab(sp4[idx,], fm = ~ Mg + Ca + ex_Ca_to_Mg + CEC_7 + clay)

# combine with the collection-wide aggregate data
g <- make.groups(collection = agg, napa.and.tehama = napa.and.tehama)

# compare graphically:
xyplot(top ~ p.q50 | variable, groups = which, data = g, ylab = 'Depth',
       xlab = 'median bounded by 25th and 75th percentiles',
       lower = g$p.q25, upper = g$p.q75, ylim = c(42, -2),
       panel = panel.depth_function,
       alpha = 0.25, sync.colors = TRUE, cf = g$contributing_fraction, cf.interval = 10,
       par.settings = list(superpose.line = list(
         col = c('RoyalBlue', 'Red4'),
         lwd = 2,
         lty = c(1, 2)
       )),
       prepanel = prepanel.depth_function,
       layout = c(5, 1), 
       strip = strip.custom(bg = grey(0.8)),
       scales = list(x = list(
         tick.number = 4,
         alternating = 3,
         relation = 'free'
       )),
       auto.key = list(columns = 2,
                       lines = TRUE,
                       points = FALSE)
)

## ----fig.height=4.5, fig.width=9------------------------------------------------------------------
library(data.table)

# 9 random profiles
# 1 simulated property via logistic power peak (LPP) function
# 6, 7, or 8 horizons per profile
# result is a list of single-profile SPC
d <- lapply(
  as.character(1:9), 
  random_profile, 
  n = c(6, 7, 8), 
  n_prop = 1, 
  method = 'LPP',
  SPC = TRUE
)

# combine into single SPC
d <- combine(d)

# GSM depths
gsm.depths <- c(0, 5, 15, 30, 60, 100, 200)

# aggregate using mean: wt.mean within slabs
# see ?slab for ideas on how to parameterize slab.fun
d.gsm <- slab(d, fm = id ~ p1, slab.structure = gsm.depths, slab.fun = mean, na.rm = TRUE)

# note: result is in long-format
# note: horizon names are lost due to aggregation
head(d.gsm, 7)

## ----fig.height=7, fig.width=9--------------------------------------------------------------------
# reshape to wide format
# this scales to > 1 aggregated variables
d.gsm.pedons <- data.table::dcast(
  data.table(d.gsm), 
  formula = id + top + bottom ~ variable, 
  value.var = 'value'
)

# init SPC
depths(d.gsm.pedons) <- id ~ top + bottom

# iterate over aggregated profiles and make new hz names
d.gsm.pedons$hzname <- profileApply(d.gsm.pedons, function(i) {
  paste0('GSM-', 1:nrow(i))
})

# compare original and aggregated
par(mar = c(1, 0, 3, 3), mfrow = c(2, 1))
plotSPC(d, color = 'p1')
mtext('original depths', side = 2, line = -1.5)
plotSPC(d.gsm.pedons, color = 'p1')
mtext('GSM depths', side = 2, line = -1.5)

## ----fig.height=7, fig.width=9--------------------------------------------------------------------
# reshape to wide format
d.gsm.pedons.2 <- data.table::dcast(
  data.table(d.gsm), 
  formula = id + top + bottom ~ variable, 
  value.var = 'contributing_fraction'
)

# init SPC
depths(d.gsm.pedons.2) <- id ~ top + bottom

# compare original and aggregated
par(mar = c(1, 0, 3, 3), mfrow = c(2, 1))
plotSPC(d.gsm.pedons, name = '', color = 'p1')
mtext('GSM depths', side = 2, line = -1.5)

plotSPC(
  d.gsm.pedons.2,
  name = '',
  color = 'p1',
  col.label = 'Contributing Fraction',
)
mtext('GSM depths', side = 2, line = -1.5)

## ----fig.height=6, fig.width=10-------------------------------------------------------------------
library(cluster)
library(ape)

# start fresh
data(sp4)
depths(sp4) <- id ~ top + bottom
hzdesgnname(sp4) <- 'name'

# eval dissimilarity:
# using Ex-Ca:Mg ratio and CEC at pH 7
# no depth-weighting (k = 0)
# to a maximum depth of 40 cm
d <- NCSP(sp4, vars = c('ex_Ca_to_Mg', 'CEC_7'), k = 0, maxDepth = 40)

# check distance matrix:
round(d, 1)

# visualize dissimilarity matrix via divisive hierarchical clustering
d.diana <- diana(d)

# may require some margin adjustments
par(mar = c(0, 0, 4, 0))

plotProfileDendrogram(
  sp4,
  d.diana,
  scaling.factor = 0.9,
  y.offset = 5,
  cex.names = 0.7,
  width = 0.3,
  color = 'ex_Ca_to_Mg',
  name.style = 'center-center',
  hz.depths = TRUE,
  depth.axis = FALSE
)

## ----setup, echo=FALSE, results='hide', warning=FALSE---------------------------------------------
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = as.integer(Sys.getenv("R_AQP_VIGNETTE_IMAGE_DPI", unset = 32)),
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)

## -------------------------------------------------------------------------------------------------
library(aqp)
library(cluster)

# load data and make a copy
data("osd")
x <- osd

## -------------------------------------------------------------------------------------------------
# assume a standard deviation of 10cm for horizon boundary depths
# far too large for most horizons, but helps to make a point
x$hzd <- 10

# generate 4 realizations of each soil profile in `x`
# limit the minimum horizon thickness to 5cm
set.seed(10101)
s <- perturb(x, id = sprintf("sim-%02d", 1:4), boundary.attr = 'hzd', min.thickness = 5)

# combine source + simulated data into a single SoilProfileCollection
z <- combine(x, s)

## ----fig.width=8.5, fig.height=5------------------------------------------------------------------
# set plotSPC argument defaults
options(.aqp.plotSPC.args = list(name.style = 'center-center', depth.axis = list(style = 'compact', line = -2.5), width = 0.33, cex.names = 0.75, cex.id = 0.66, max.depth = 185))

par(mar = c(0, 0, 0, 1))
plotSPC(z)

## ----fig.width=8.5, fig.height=5.25---------------------------------------------------------------
# encode as a factor for distance calculation
z$subgroup <- factor(z$subgroup)

par(mar = c(0, 0, 1, 1))
groupedProfilePlot(z, groups = 'subgroup', group.name.offset = -10, break.style = 'arrow', group.line.lty = 1, group.line.lwd = 1)

## ----fig.width=8.5, fig.height=5.25---------------------------------------------------------------
# assign GHL
z$genhz <- generalize.hz(
  z$hzname, new = c('A', 'E', 'Bt', 'C'), 
  pattern = c('A', 'E', 'Bt', 'C|Bt4')
)

# check GHL
par(mar = c(0, 0, 3, 1))
groupedProfilePlot(z, groups = 'subgroup', group.name.offset = -10, break.style = 'arrow', group.line.lty = 1, group.line.lwd = 1, color = 'genhz')

## -------------------------------------------------------------------------------------------------
# horizon-level distance matrix weight
w1 <- 1
# perform NCSP using only the GHL (ordered factors) to a depth of 185cm
d1 <- NCSP(z, vars = c('genhz'), maxDepth = 185, k = 0, rescaleResult = TRUE)

# site-level distance matrix weight
w2 <- 2
# Gower's distance metric applied to subgroup classification (nominal factor)
d2 <- compareSites(z, 'subgroup')

# perform weighted average of distance matrices
D <- Reduce(
  `+`, 
  list(d1 * w1, d2 * w2)
) / sum(c(w1, w2))

## ----fig.width=8.5, fig.height=6------------------------------------------------------------------
library(ape)

# divisive hierarchical clustering
h <- as.hclust(diana(D))

# hang soil profile sketches from resulting dendrogram
par(mar = c(1, 0, 0, 1))
plotProfileDendrogram(z, clust = h, scaling.factor = 0.0075, y.offset = 0.15, width = 0.33, color = 'genhz', name = NA)

# annotate dendrogram with subgroup classification
# this handy function provided by the ape package
tiplabels(pch = 15, col = c(2, 3)[z$subgroup], cex = 1.5, offset = 0.05)

# helpful legend
legend('topleft', legend = levels(z$subgroup), pch = 15, col = c(2, 3), bty = 'n')

## ----setup, echo=FALSE, results='hide', warning=FALSE---------------------------------------------
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = as.integer(Sys.getenv("R_AQP_VIGNETTE_IMAGE_DPI", unset = 32)),
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)

## -------------------------------------------------------------------------------------------------
library(aqp)

## -------------------------------------------------------------------------------------------------
# example data
data("jacobs2000")

# fully populated
plotSPC(jacobs2000, name.style = 'center-center', 
        cex.names = 0.8, color = 'time_saturated')

# missing some data
plotSPC(jacobs2000, name.style = 'center-center', 
        cex.names = 0.8, color = 'concentration_color')

# very nearly complete
plotSPC(jacobs2000, name.style = 'center-center', 
        cex.names = 0.8, color = 'matrix_color')


# variables to consider
v <- c('time_saturated', 'concentration_color', 'matrix_color')

# compute data completeness by profile
# ignore 2C horizons
jacobs2000$data.complete <- evalMissingData(
  jacobs2000, 
  vars = v, 
  method = 'relative',
  p = '2C'
)

jacobs2000$data.complete.abs <- evalMissingData(
  jacobs2000, 
  vars = v, 
  method = 'absolute',
  p = '2C'
)

# compute data completeness by horizon
# ignore 2C horizons
jacobs2000$hz.data.complete <- evalMissingData(
  jacobs2000, 
  vars = v, 
  method = 'horizon',
  p = '2C'
)


# "fraction complete" by horizon
plotSPC(
  jacobs2000, name.style = 'center-center', 
  cex.names = 0.8, color = 'hz.data.complete'
)


# rank on profile completeness
new.order <- order(jacobs2000$data.complete)

# plot along data completeness ranking
plotSPC(
  jacobs2000, name.style = 'center-center', 
  cex.names = 0.8, color = 'hz.data.complete', 
  plot.order = new.order
)

# add relative completeness axis
# note re-ordering of axis labels
axis(
  side = 1, at = 1:length(jacobs2000), 
  labels = round(jacobs2000$data.complete[new.order], 2),
  line = 0, cex.axis = 0.75
)

# add absolute completeness (cm)
axis(
  side = 1, at = 1:length(jacobs2000), 
  labels = jacobs2000$data.complete.abs[new.order],
  line = 2.5, cex.axis=0.75
)

# label axes
mtext('Relative\nCompleteness', side = 1, at = 0.25, line = 0.25, cex = 0.8)
mtext('Absolute\nCompleteness (cm)', side = 1, at = 0.25, line = 2.75, cex = 0.8)

## ----eval=requireNamespace("soilDB", quietly=TRUE)------------------------------------------------
library(soilDB)
x <- fetchKSSL(series = 'pierre')
par(mar = c(0, 0, 3, 2))

plotSPC(x, color = 'clay', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')
plotSPC(x, color = 'cec7', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')
plotSPC(x, color = 'estimated_oc', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')
plotSPC(x, color = 'ph_h2o', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')
plotSPC(x, color = 'db_13b', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')


par(mar = c(1, 0, 3, 2))
plotSPC(x, color = 'ph_h2o', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')

.b <- x[, , .LAST, .BOTTOM]
text(x = 1:length(x), y = .b, labels = x$pi, cex = 0.85, pos = 1)
mtext('Profile Information Index (bytes)', side = 1, line = -0.5)

v <- c('clay', 'db_13b', 'cec7', 'ph_h2o')
x$rel.not.missing <- evalMissingData(x, vars = v, method = 'relative')
x$abs.not.missing <- evalMissingData(x, vars = v, method = 'absolute')
x$hz.not.missing <- evalMissingData(x, vars = v, method = 'horizon')

o <- order(x$rel.not.missing)
plotSPC(x, color = 'hz.not.missing', width = 0.33, name.style = 'center-center', label = 'pedon_completeness_index', plot.order = o)
text(x = 1:length(x), y = .b[o], labels = round(x$rel.not.missing[o], 2), cex = 0.85, pos = 1)
mtext('Relative Non-Missing Fraction', side = 1, line = -0.5)

o <- order(x$abs.not.missing)
plotSPC(x, color = 'hz.not.missing', width = 0.33, name.style = 'center-center', label = 'pedon_completeness_index', plot.order = o)
text(x = 1:length(x), y = .b[o], labels = x$abs.not.missing[o], cex = 0.85, pos = 1)
mtext('Absolute Non-Missing (cm)', side = 1, line = -0.5)

##
## main demo for AQP package
##

# load packages
library(aqp)
library(ape)
library(cluster)
library(lattice)
library(data.table)

# Example 1: sp1: 9 soil profiles from Pinnacles National Monument, CA.
data(sp1)

# 
# 1. basic profile aggregation and plotting
# 
depths(sp1) <- id ~ top + bottom

# aggregate all profiles into 1,5,10,20 cm thick slabs, computing mean values by slab
s1 <- slab(sp1, ~ prop, slab.fun=mean, na.rm=TRUE)
# 5cm
s5 <- slab(sp1, ~ prop, slab.fun=mean, na.rm=TRUE, slab.structure=5)
# 10cm segments:
s10 <- slab(sp1, ~ prop, slab.fun=mean, na.rm=TRUE, slab.structure=10)
# 20cm
s20 <- slab(sp1, ~ prop, slab.fun=mean, na.rm=TRUE, slab.structure=20)

# check results
head(s1)

# variation in segment-weighted mean property: very little
round(sapply(list(s1, s5, s10, s20),
             function(i) {
               with(i, sum((bottom - top) * value) / sum(bottom - top))
             }), 1)

# combined viz
g2 <- make.groups(
  "1cm interval" = s1,
  "5cm interval" = s5,
  "10cm interval" = s10,
  "20cm interval" = s20
)

# note special syntax for plotting step function
xyplot(
  cbind(top, bottom) ~ value,
  groups = which,
  data = g2,
  id = g2$which,
  panel = panel.depth_function,
  ylim = c(250, -10),
  scales = list(y = list(tick.number = 10)),
  xlab = 'Property',
  ylab = 'Depth (cm)',
  main = 'Soil Profile Aggregation by Regular Depth-slice',
  auto.key = list(
    columns = 2,
    points = FALSE,
    lines = TRUE
  )
)

# Example 2: sp3: 10 soil profiles from the Sierra Nevada Foothills Region of California.
data(sp3)

# 
# 2. investigate the concept of a 'median profile'
# note that this involves aggregation between two dissimilar groups of soils
# 

# generate a RGB version of soil colors
# and convert to HSV for aggregation
sp3$h <- NA
sp3$s <- NA
sp3$v <- NA
sp3.rgb <- with(sp3, munsell2rgb(hue, value, chroma, return_triplets = TRUE))
sp3[, c('h', 's', 'v')] <- t(with(sp3.rgb, rgb2hsv(r, g, b, maxColorValue = 1)))

# promote to SoilProfileCollection
depths(sp3) <- id ~ top + bottom

# aggregate across entire collection
a <- slab(sp3,
          fm = ~ clay + cec + ph + h + s + v,
          slab.structure = 10)

# check
str(a)

# convert back to wide format
a.wide.q25 <- dcast(as.data.table(a), top + bottom ~ variable, value.var = c('p.q25'))
a.wide.q50 <- dcast(as.data.table(a), top + bottom ~ variable, value.var = c('p.q50'))
a.wide.q75 <- dcast(as.data.table(a), top + bottom ~ variable, value.var = c('p.q75'))

# add a new id for the 25th, 50th, and 75th percentile pedons
a.wide.q25$id <- 'Q25'
a.wide.q50$id <- 'Q50'
a.wide.q75$id <- 'Q75'

# combine original data with "mean profile"
vars <- c('id', 'top', 'bottom', 'clay', 'cec', 'ph', 'h', 's', 'v')

# make data.frame version of sp3
sp3.grouped <- as.data.frame(rbind(as.data.table(horizons(sp3))[, .SD, .SDcol = vars], 
                     a.wide.q25[, .SD, .SDcol = vars],
                     a.wide.q50[, .SD, .SDcol = vars], 
                     a.wide.q75[, .SD, .SDcol = vars]))

# re-constitute the soil color from HSV triplets
# convert HSV back to standard R colors
sp3.grouped$soil_color <- with(sp3.grouped, hsv(h, s, v))

# give each horizon a name
sp3.grouped$name <- paste(round(sp3.grouped$clay), '/' ,
                          round(sp3.grouped$cec), '/',
                          round(sp3.grouped$ph, 1))

# first promote to SoilProfileCollection
depths(sp3.grouped) <- id ~ top + bottom

plot(sp3.grouped)

## perform comparison, and convert to phylo class object
## D is rescaled to [0,]
d <- NCSP(
  sp3.grouped,
  vars = c('clay', 'cec', 'ph'),
  maxDepth = 100,
  k = 0.01
)

h <- agnes(d, method = 'ward')
p <- ladderize(as.phylo(as.hclust(h)))

# look at distance plot-- just the median profile
plot_distance_graph(d, 12)

# similarity relative to median profile (profile #12)
round(1 - (as.matrix(d)[12, ] / max(as.matrix(d)[12, ])), 2)

## make dendrogram + soil profiles

# setup plot: note that D has a scale of [0,1]
par(mar = c(1, 1, 1, 1))
p.plot <- plot(p,
               cex = 0.8,
               label.offset = 3,
               direction = 'up',
               y.lim = c(200, 0),
               x.lim = c(1.25, length(sp3.grouped) + 1),
               show.tip.label = FALSE)

# get the last plot geometry
lastPP <- get("last_plot.phylo", envir = ape::.PlotPhyloEnv)

# the original labels, and new (indexed) order of pedons in dendrogram
d.labels <- attr(d, 'Labels')

new_order <- sapply(1:lastPP$Ntip,
                    function(i)
                      which(as.integer(lastPP$xx[1:lastPP$Ntip]) == i))

# plot the profiles, in the ordering defined by the dendrogram
# with a couple fudge factors to make them fit
plotSPC(
  sp3.grouped,
  color = "soil_color",
  plot.order = new_order,
  y.offset = max(lastPP$yy) + 10,
  width = 0.1,
  cex.names = 0.5,
  add = TRUE
)

library(lattice)
library(Hmisc)
library(aqp)

# convert angle in degrees to percent
deg2pct <- function(theta) {
	tan(theta * pi/180) * 100
}

# convert angle in percent to radians
pct2rad <- function(pct) {
	atan(pct / 100)
}

# compute difference in thickness as a function of slope pct
thickDiff <- function(pct, Tv) {
	Tp <- Tv * cos(pct2rad(pct))
	return(Tv - Tp)
}


# make a grid of horizon thickness and slope percents
d <- expand.grid(thick=c(5,10,20,50,100), slope=1:70)
d$thickDiff <- thickDiff(d$slope, d$thick)

# plot effects for different horizon thickness
xYplot(thickDiff ~ slope, data=d, groups=factor(thick, labels=paste(c(5,10,20,50,100), 'cm depth')), type='l', ylab='(Vertical - Perpendicular) Hz Thickness', xlab='Slope (%)', main='Difference in Apparent Horizon Thickness as a Function of Slope', scales=list(alternating=3, tick.number=10), par.settings=list(superpose.line=list(col='black', lwd=1.5)), label.curves=list(cex=1), panel=function(...) {
	panel.abline(v=seq(0, 70, by=10), h=seq(0, 20, by=2), col=grey(0.75), lty=3)
	panel.xYplot(...)
})



## plot the effects via 10 pedons

# generate a profile, keep only the relevant stuff
p <- random_profile(1)[, c('id', 'top', 'bottom', 'name')]

# stack 10 of these up
pp <- rbind(p, p, p, p, p, p, p, p, p, p)

# fix IDs
pp$id <- rep(1:10, each=nrow(p))

# upgrade to SPC
depths(pp) <- id ~ top + bottom

# add fake slope to site
s <- data.frame(id=1:10, slope=seq(from=1, to=70, length.out=10))
site(pp) <- s


# function for transforming vertical to perpendicular depths
# operates on a single SPC object
VtoP <- function(i, which) {
	# get vector of bottom depths
	b <- horizons(i)[[which]]
	# convert to perpendicular depths and round to an integer
	b.p <- round(b * cos(pct2rad(i$slope)))
	return(b.p)
}


# add perpendicular depths to original data
pp$Tp <- profileApply(pp, VtoP, which='bottom')


# basic plot
par(mar=c(2,0,0.75,0))
plot(pp, n.depth.ticks=10)

# title
title('Vertical Horizon Depths (black) vs. Perpendicular Horizon Depths (blue)', cex.main=0.75)

# add axis with slope values
axis(side=1, cex.axis=0.75, line=-1.25, at=1:10, labels=round(s$slope))
mtext(side=1, 'Slope (%)', line=0.5, cex=0.75)

# iterate over profiles and add offset hz boundaries
for(i in 1:length(pp)) {
	pp.i <- pp[i, ]
	# use color to show only those offsets that are > 1 cm
	cols <- ifelse(abs(pp.i$bottom - pp.i$Tp) > 1, 'RoyalBlue', NA)
	arrows(x0=i, x1=i, y0=pp.i$bottom, y1=pp.i$Tp, length=0.08, col=cols)
	segments(x0=i-0.2, x1=i+0.2, y0=pp.i$Tp, y1=pp.i$Tp, col=cols, lty=2)
}


## another approach- change actual hz depths
pp.perp <- pp

pp.perp$top <- profileApply(pp.perp, VtoP, which='top')
pp.perp$bottom <- profileApply(pp.perp, VtoP, which='bottom')

par(mar=c(2,0,0.75,0))
plot(pp.perp, n.depth.ticks=10)
title('Changes in Horizon Depth as a Function of Surface Slope', cex.main=0.75)
axis(side=1, cex.axis=0.75, line=-1.25, at=1:10, labels=round(s$slope))
mtext(side=1, 'Slope (%)', line=0.5, cex=0.75)
abline(h=seq(from=5, to=75, by=5), lty=3, col='grey')


## compute mean and total difference in horizon thickness:
# compute difference in horizon thickness
diff.thick <- (pp$bottom - pp$top) - (pp.perp$bottom - pp.perp$top)
theID <- horizons(pp)$id # extract IDs
# compute mean and total difference by pedon
diff.by.pedon <- by(diff.thick, theID, function(i) cbind(mean.diff.cm=mean(i), total.diff.cm=sum(i)))
# composite into a data.frame and add slope values
data.frame(do.call('rbind', diff.by.pedon), slope=round(s$slope))

## ----setup, echo=FALSE, results='hide', warning=FALSE, message=FALSE------------------------------
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = as.integer(Sys.getenv("R_AQP_VIGNETTE_IMAGE_DPI", unset = 32)),
  optipng = knitr::hook_optipng,
  comment = "#>"
)
options(width = 100, stringsAsFactors = FALSE, timeout = 600)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

library(aqp, quietly = TRUE)
.bluecolors <- hcl.colors(n = 25, palette = 'Blues')[-25]

## ----fig.width=6, fig.height=5, echo = FALSE------------------------------------------------------
data("osd")
par(mar = c(0, 0, 0, 0))
plotSPC(osd, name = NA, fixLabelCollisions = FALSE, cex.names = 1.75, hz.depths = TRUE, depth.axis = FALSE, cex.id = 0.66, name.style = 'center-center')

## ----fig.width=8, fig.height=5--------------------------------------------------------------------
library(aqp)

# soil profile template
x <- quickSPC(
  "SPC:Oi|Oe|AAAAA|E1|E2|Bhs1Bhs1Bhs1|Bhs2|CCCCCC|RRRRRRRRRR", 
  interval = 3
)

# set variability in horizon thickness
horizons(x)$thick.sd <- 3
# setup horizon designations
horizons(x)$nm <- factor(x$name, levels = c('Oi', 'Oe', 'A', 'E1', 'E2', 'Bhs1', 'Bhs2', 'C', 'R'))

# simulate 3 realizations of original profile
set.seed(10101)
s <- perturb(x, n = 3, thickness.attr = 'thick.sd', min.thickness = 2)

# create 3 copies of each simulation
s <- duplicate(s, times = 3)

## ----fig.width=8, fig.height=5, echo = FALSE------------------------------------------------------
.a <- list(cex.names = 0.8, color = 'nm', name = NA, show.legend = FALSE, hz.depths = TRUE, depth.axis = FALSE, cex.id = 0.66, col.palette = .bluecolors)

options(.aqp.plotSPC.args = .a)

par(mar = c(0, 0, 0, 0))
plotSPC(s, fixLabelCollisions = FALSE)
title('Example Data', line = -1.5)

## ----fig.width=8, fig.height=5, echo = FALSE------------------------------------------------------
par(mar = c(0, 0, 0, 0))
# select "Electrostatic" label placement method
# adjust default charge density to 0.5
plotSPC(s, fixOverlapArgs = list(method = 'E', q = 0.5))
title('Label Adjustment by Electrostatic Simulation', line = -1.5)

## ----fig.width=8, fig.height=5, echo = FALSE------------------------------------------------------
par(mar = c(0, 0, 0, 0))
plotSPC(s, fixOverlapArgs = list(method = 'S'))
title('Label Adjustment by Simulated Annealing', line = -1.5)

options(.aqp.plotSPC.args = NULL)

## ----fig.width=8, fig.height=5--------------------------------------------------------------------
x <- quickSPC(
  "SPC:Oi|Oe|AAA|E1|E2|E3|BhsBhsBhsBhs|Bt1|Bt2|Bt3Bt3|CCCCCC|Ab1|Ab2|2C2C2C2C2C2C|2Cr|2R2R2R2R2R2R2R2R", 
  interval = 1
)

x$z <- as.numeric(x$hzID)

## ----fig.width=8, fig.height=5--------------------------------------------------------------------
# pretty colors
.bluecolors <- hcl.colors(n = 25, palette = 'Blues')[-25]

# plotSPC arguments
.a <- list(
  width = 0.2, 
  max.depth = 40,
  hz.depths = TRUE, 
  name.style = 'center-center', 
  cex.names = 1.5,
  name = NA,
  depth.axis = FALSE, 
  color = 'z',
  show.legend = FALSE,
  print.id = FALSE,
  col.palette = .bluecolors
)

# set plotSPC default arguments
options(.aqp.plotSPC.args = .a)

# wrapper function to test label collision solutions
testIt <- function(x, ...) {
  
  # make sketches
  plotSPC(x, ...)
  
  # a normalized index of label adjustment
  .LAI <- get('last_spc_plot', envir = aqp.env)$hz.depth.LAI
  .LAI <- ifelse(is.na(.LAI), 0, .LAI)
  
  # annotate with label adjustment index
  .txt <- sprintf("LAI: %0.3f", .LAI)
  mtext(.txt, side = 1, at = 1, line = -1.5, cex = 0.8)
}

## ----fig.width=8, fig.height=5--------------------------------------------------------------------
par(mar = c(1, 0, 0, 0), mfcol = c(1, 8))

testIt(x, fixLabelCollisions = FALSE)
title('No\nAdjustment', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 1', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 2', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 3', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 1.5))
title('Electrostatic\nq = 1.5', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 1))
title('Electrostatic\nq = 1', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 0.5))
title('Electrostatic\nq = 0.5', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 0.25))
title('Electrostatic\nq = 0.25', line = -3.5, adj = 0.5)

## ----eval=requireNamespace("soilDB", quietly=TRUE), fig.width=8, fig.height=5---------------------
library(aqp)
library(soilDB)

s <- c('inks' , 'pardee', 'clarksville', 'palau', 'hao', 'inks', 'eheuiki', 'puaulu', 'zook', 'cecil')
x <- fetchOSD(s)

par(mar = c(0, 0, 0, 3))

.args <- list(width = 0.3, name.style = 'center-center', hz.depths = TRUE, cex.names = 1)
options(.aqp.plotSPC.args = .args)

plotSPC(x, fixOverlapArgs = list(method = 'E', q = 1.25), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 1), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.75), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.5), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.25), max.depth = 151)

plotSPC(x, fixOverlapArgs = list(method = 'S'), max.depth = 151)

## -------------------------------------------------------------------------------------------------
x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)

overlapMetrics(x, thresh = 0.5)

## ----message=TRUE---------------------------------------------------------------------------------
# vector of positions, typically labels but could be profile sketch alignment on the x-axis
s <- c(1, 2, 2.3, 4, 5, 5.5, 7)

# simulated annealing, solution is non-deterministic
fixOverlap(s, thresh = 0.5, method = 'S')

# electrostatics-inspired simulation of particles
# solution is deterministic
fixOverlap(s, thresh = 0.5, method = 'E')

## ----fig.width=8.5, fig.height=5------------------------------------------------------------------
evalMethods <- function(x, thresh, q, ...) {
  
  cols <- hcl.colors(n = 9, palette = 'Zissou 1', rev = TRUE)
  cols <- colorRampPalette(cols)(length(x))
  
  z <- fixOverlap(x, thresh = thresh, method = 'E', maxIter = 100, trace = TRUE, q = q)
  .n <- nrow(z$states)
  
  op <- par(mar = c(0, 2, 2, 0.5), bg = 'black', fg = 'white')
  layout(matrix(c(1, 2, 3, 4), ncol = 2, nrow = 2), heights = c(0.33, 0.66))
  
  plot(seq_along(z$cost), z$cost, las = 1, type = 'b', axes = FALSE, cex = 0.66, xlim = c(1, .n))
  mtext(text = sprintf("Converged (%s): %s", .n, z$converged), at = 0, side = 3, line = 0, cex = 0.75, font = 3, adj = 0)
  matplot(rbind(x, z$states), type = 'l', lty = 1, las = 1, axes = FALSE, col = cols, lwd = 1)
  
  points(x = rep(1, times = length(x)), y = x, cex = 0.66, pch = 16, col = cols)
  points(x = rep(.n + 1, times = length(x)), y = z$x, cex = 0.66, pch = 16, col = cols)
  
  text(x = 1, y = x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 2)
  text(x = .n + 1, y = z$x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 4)
  
  axis(side = 2, at = unique(x), labels = round(unique(x), 1), col.axis = par('fg'), las = 1, cex.axis = 0.6)
  title(main = 'Electrostatic Simulation', line = 1, col.main = 'white')
  
  ## SANN_1D doesn't always preserve rank ordering
  ##  ->> not designed to use unsorted input
  ##  ->> maybe impossible with ties in x?
  
  z <- fixOverlap(x, thresh = thresh, method = 'S', trace = TRUE, maxIter = 1000)
  .n <- nrow(z$states)
  
  plot(seq_along(z$stats), z$stats, las = 1, type = 'b', axes = FALSE, cex = 0.66, xlim = c(1, .n))
  mtext(text = sprintf("Converged (%s): %s", .n, z$converged), at = 0, side = 3, line = 0, cex = 0.75, font = 3, adj = 0)
  
  matplot(z$states, type = 'l', lty = 1, las = 1, axes = FALSE, col = cols)
  
  points(x = rep(1, times = length(x)), y = z$states[1, ], cex = 0.66, pch = 16, col = cols)
  points(x = rep(.n, times = length(x)), y = z$x, cex = 0.66, pch = 16, col = cols)
  
  text(x = 1, y = z$states[1, ], col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 2)
  text(x = .n, y = z$x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 4)
  
  axis(side = 2, at = unique(x), labels = round(unique(x), 1), col.axis = par('fg'), las = 1, cex.axis = 0.6)
  
  title(main = 'Simulated Annealing', line = 1, col.main = 'white')
  
  # reset graphics state
  par(op)
  layout(1)
  
}

## ----fig.width=8.5, fig.height=5------------------------------------------------------------------
# explore effect of charge (q)
# too large -> chaos
x <- c(0, 2, 5, 12, 18, 20, 35, 40, 50, 56, 90, 120, 145, 150)

# just about right, very few perturbations required
evalMethods(x, thresh = 5, q = 1.1)

# ok, but now most label positions are affected
evalMethods(x, thresh = 5, q = 1.8)

# too high, wasting time on more iterations
evalMethods(x, thresh = 5, q = 3)

# far too high, wasting more time with little gain
evalMethods(x, thresh = 5, q = 4)

# chaos and failure to converge
evalMethods(x, thresh = 5, q = 5)

## ----eval = FALSE---------------------------------------------------------------------------------
# # threshold too large
# evalMethods(x, thresh = 10, q = 3)
# 
# 
# # large threshold
# x <- c(0, 5, 12, 18, 20, 35, 40, 55, 90, 120, 145, 150)
# evalMethods(x, thresh = 9, q = 2)
# 
# # single iteration enough
# x <- c(0, 3, 20, 35, 40, 55, 90, 120, 145, 150)
# evalMethods(x, thresh = 6, q = 1)
# 
# # clusters
# x <- sort(c(0, jitter(rep(10, 3)), jitter(rep(25, 3)), jitter(rep(90, 3)), 150))
# evalMethods(x, thresh = 6, q = 3)
# evalMethods(x, thresh = 6, q = 2)
# 
# 
# ## impact of scale / offset
# x <- c(0, 5, 12, 18, 20, 35, 40, 50, 120, 145, 150)
# 
# # works as expected
# evalMethods(x, thresh = 5, q = 1.1)
# 
# # works as expected, as long as threshold is scaled
# evalMethods(x / 10, thresh = 5 / 10, q = 1.1)
# 
# # works as expected, as long as threshold is scaled
# evalMethods(x * 10, thresh = 5 * 10, q = 1.1)
# 
# 
# # all work as expected, threshold not modified
# evalMethods(x + 10, thresh = 5, q = 1.1)
# evalMethods(x + 100, thresh = 5, q = 1.1)
# evalMethods(x + 1000, thresh = 5, q = 1.1)
# 
# # works as expected
# x <- c(315, 325, 341, 353, 366, 374, 422)
# fixOverlap(x, thresh = 9.7, q = 1, method = 'E')
# evalMethods(x, thresh = 9.7, q = 1)
# 
# 
# x <- c(1.0075, 1.1200, 1.3450, 1.6450, 1.8700, 1.8825)
# fixOverlap(x, thresh = 0.05442329, q = 1)
# evalMethods(x, thresh = 0.05442329, q = 1)

## ----fig.width=8, fig.height=6--------------------------------------------------------------------
tracePlot <- function(x, z, cex.axis.labels = 0.85) {
  # setup plot device
  op <- par(mar = c(4, 4, 1, 1), bg = 'black', fg = 'white')
  layout(matrix(c(1,2,3)), widths = 1, heights = c(1,1,2))
  
  # order:
  # B: boundary condition violation
  # O: rank (order) violation
  # +: accepted perturbation
  # -: rejected perturbation
  cols <- c(grey(0.5), grey(0.85), 'royalblue', 'firebrick')
  
  cols.lines <- hcl.colors(9, 'Zissou 1', rev = TRUE)
  cols.lines <- colorRampPalette(cols.lines)(length(x))
  
  # total overlap (objective function) progress
  plot(
    seq_along(z$stats), z$stats, 
    type = 'h', las = 1,
    xlab = 'Iteration', ylab = 'Total Overlap',
    axes = FALSE,
    col = cols[as.numeric(z$log)]
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, line = -2)
  mtext('Overlap', side = 2, line = 2, cex = cex.axis.labels, font = 2)
  
  # deviation from original configuration
  plot(
    seq_along(z$stats), z$ssd, 
    type = 'h', las = 1,
    xlab = 'Iteration', ylab = 'Deviation',
    axes = FALSE,
    col = cols[as.numeric(z$log)]
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, line = -2)
  mtext('Deviation', side = 2, line = 2, cex = cex.axis.labels, font = 2)
  legend('top', legend = c('boundary\nviolation', 'rank\nviolation', 'accepted\nperturbation', 'rejected\nperturbation'), col = cols, bty = 'n', horiz = TRUE, inset = -0.5, lty = 1, lwd = 2, xpd = NA)
  
  # adjustments at each iteration
  matplot(
    z$states, type = 'l', 
    lty = 1, las = 1, 
    xlab = 'Iteration', ylab = 'x-position',
    axes = FALSE,
    col = cols.lines
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, at = x, labels = round(x, 1))
  axis(side = 4, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, at = z$x, labels = round(z$x, 1), line = -2)
  mtext('Position', side = 2, line = 2.5, cex = cex.axis.labels, font = 2)
  
  axis(side = 1, cex.axis = 1, col.axis = 'white', line = 0)
  mtext('Iteration', side = 1, line = 2.5, cex = cex.axis.labels, font = 2)
  
  par(op)
  layout(1)
}

## ----fig.width=8, fig.height=6, message=TRUE------------------------------------------------------
x <- c(0, 1, 2, 2.2, 2.8, 3.5, 6, 8, 10, 10.1, 12.8, 13, 14.8, 15, 15.5)

# fix overlap, return debugging information
set.seed(10101)
z <- fixOverlap(x, thresh = 0.73, method = 'S', trace = TRUE)

# check convergence
z$converged

# inspect algorithm trace
tracePlot(x, z)

# trace log
# B: boundary condition violation
# O: rank (order) violation
# +: accepted perturbation
# -: rejected perturbation
table(z$log)

## ----fig.width=8, fig.height=6, message=TRUE------------------------------------------------------
# fix overlap, return debugging information
set.seed(101010)
x <- sort(runif(10, min = 2.5, max = 3.5))

# widen boundary conditions
z <- fixOverlap(x, thresh = 0.2, trace = TRUE, min.x = 0, max.x = 10, maxIter = 2000, adj = 0.05)

# check convergence
z$converged

# inspect algorithm trace
tracePlot(x, z)

## -------------------------------------------------------------------------------------------------
# reset plotSPC() options
options(.aqp.plotSPC.args = NULL)

# split standard Munsell hue into character | numeric parts
# function is vectorized

# additional ideas from munsellinterpol::HueNumberFromString

# TODO: consider a strict argument for only allowing standard hues

# output:
# hue.numeric hue.character
#     2.3            YR
.parseMunsellHue <- function(hue) {
  
  # NA not permitted, convert to ''
  hue <- ifelse(is.na(hue), '', hue)
  
  # extract numeric part from hue
  # danger! empty strings will result in an empty list element
  nm.part <- strsplit(hue, split='[^0-9.]+', )
  
  # replace empty list elements with ''
  nm.part[which(sapply(nm.part, length) < 1)] <- ''
  
  # the numeric part is always the first returned match
  # solves bogus rows associated with invalid hues
  # # https://github.com/ncss-tech/aqp/issues/66
  nm.part <- lapply(nm.part, '[', 1)
  
  # convert to vector
  hue.numeric <- unlist(nm.part)
  
  # extract character part from hue
  hue.character <- vector(mode='character', length = length(hue))
  for(i in seq_along(hue)){
    # check for NA
    if(is.na(hue.numeric[i])) {
      hue.character[i] <- NA
      next
    }
    # check for empty string
    if(hue.numeric[i] == '') {
      hue.character[i] <- NA
      next
    }
    # otherwise continue processing
    # remove numeric portion
    hue.character[i] <- gsub(hue.numeric[i], '', hue[i], fixed = TRUE)
  }
  
  # convert numeric part to numbers
  hue.numeric <- as.numeric(hue.numeric)
  
  # strip white space form character portion
  hue.character <- trimws(hue.character, which = 'both')
  
  res <- data.frame(hue.numeric, hue.character, stringsAsFactors = FALSE)
  return(res)
}




#' @title Parse Munsell Color Notation
#' 
#' @description Split Munsell color notation into "hue", "value", and "chroma", with optional conversion to sRGB hex notation, sRGB coordinates, and CIELAB coordinates. Conversion is performed by [munsell2rgb()].
#'
#' @note This function should be able to parse and convert 100k Munsell colors in 0.5 seconds (single-threaded, 2.5GHz i7).
#'
#' @param munsellColor character vector of Munsell colors (e.g. `c('10YR 3/4', '5YR 4/6')`)
#' @param convertColors logical, convert colors to sRGB hex notation, sRGB coordinates, CIELAB coordinates
#' @param delim optional, specify the type of delimiter used between value and chroma parts of the Munsell code. By default ":", ",:, "'", and "/" are supported.
#' @param ... additional arguments to [munsell2rgb()]
#'
#' @return a `data.frame` object
#' 
#' @export
#'
#' @author P. Roudier and D.E. Beaudette
#'
#' @examples
#' 
#' # just sRGB
#' parseMunsell("10YR 3/5", return_triplets = TRUE)
#' 
#' # sRGB + CIELAB (D65 illuminant)
#' parseMunsell("10YR 3/5", return_triplets = TRUE, returnLAB = TRUE)
#' 
#' # CIELAB only
#' parseMunsell("10YR 3/5", return_triplets = FALSE, returnLAB = TRUE)
#' 
#' # neutral hue
#' # note chroma encoded as '0'
#' parseMunsell('N 3/', convertColors = FALSE)
#' 
parseMunsell <- function(munsellColor, convertColors = TRUE, delim = NA, ...) {
  
  # now fully vectorized, no wrangling of tiny data.frames
  # https://github.com/ncss-tech/aqp/issues/230
  
  # empty result set for convertColors = FALSE
  empty <- data.frame(
    hue = NA_character_, 
    value = NA_real_, 
    chroma = NA_real_, 
    stringsAsFactors = FALSE
  )
  
  # sanity check: all NA
  if(all(is.na(munsellColor)) | all(is.null(munsellColor)) | all(munsellColor == '')) {
    # return empty data.frame for each entry
    if(convertColors) {
      res <- empty[rep(1, times = length(munsellColor)), ]
      row.names(res) <- NULL
      return(res)
      
    } else {
      # return NA for each entry
      res <- rep(NA, times = length(munsellColor))
      return()
    }
  }
  
  
  # ensure colours are all upper case
  munsellColor <- toupper(munsellColor)
  
  # total number of records, used to NA-padding
  n <- length(munsellColor)
  
  # Munsell notation validation
  
  # If the very first character of the Munsell string is not numeric OR 'N'
  not.numeric.idx <- grep('^[0-9|N]', munsellColor, invert = TRUE)
  if(length(not.numeric.idx) > 0) {
    # can't do anything with this value
    munsellColor[not.numeric.idx] <- NA
  }
  
  # index non-NA values
  not.na.idx <- which(!is.na(munsellColor))
  mn <- munsellColor[not.na.idx]
  
  
  ## split pieces
  
  # Extract hue number
  hue_number <- trimws(sub("[A-Z].*", "", mn), which = "both")
  remaining <- substr(mn, nchar(hue_number) + 1, nchar(mn))
  hue_letter <- trimws(sub("[0-9].*", "", remaining), which = "both") 
  remaining <- substr(remaining, nchar(hue_letter) + 1, nchar(remaining))
  remaining <- trimws(remaining, which = "both")
  
  # re-constitute hue
  hue <- paste0(hue_number, hue_letter)
  
  # list, of typically 2-element vectors
  if (is.na(delim)) {
    value_chroma <- strsplit(remaining, "[:,'/_]")
  } else {
    value_chroma <- strsplit(remaining, delim)
  }
  
  # attempt to extract value and chroma
  value <- sapply(value_chroma, '[', 1)
  chroma <- sapply(value_chroma, '[', 2)
  
  # clean any remaining whitespace
  value <- trimws(value, which = "both")
  chroma <- trimws(chroma, which = "both")
  
  # convert NA chroma -> 0 for N hues
  chroma <- ifelse(is.na(chroma) & hue == 'N', 0, chroma)
  
  # init empty data.frame
  res <- empty[rep(1, times = length(munsellColor)), ]
  row.names(res) <- NULL
  
  # insert non-NA values
  # always character
  res$hue[not.na.idx] <- hue
  # always numeric
  res$value[not.na.idx] <- as.numeric(value)
  # always numeric
  res$chroma[not.na.idx] <- as.numeric(chroma)
  
  # parsed into columns within a data.frame, but not converted to colors
  if(convertColors == FALSE) {
    return(res)
  }
  
  # otherwise convert using NA-padded data.frame
  res <- munsell2rgb(res$hue, res$value, res$chroma, ...)
  
  return(res)
}

context("invertLabelColor")


test_that("works as expected", {
  
  cols <- parseMunsell(c('10YR 2/1', '10YR 6/8', NA, NA))
  icols <- invertLabelColor(cols)
  
  # white
  expect_equal(icols[1], '#FFFFFF')
  
  # black
  expect_equal(icols[2], '#000000')
  
  # NA input -> black
  expect_equal(icols[3], '#000000')
  expect_equal(icols[4], '#000000')
  
})

context("estimateSoilColor")

# note, not symmetric

test_that("works as expected", {
  
  # moist -> dry
  x <- estimateSoilColor(hue = '10YR', value = 3, chroma = 3, sourceMoistureState = 'moist')
  
  expect_equal(x$hue, '10YR')
  expect_equal(x$value, 4)
  expect_equal(x$chroma, 3)
  
  # dry -> moist
  x <- estimateSoilColor(hue = '10YR', value = 5, chroma = 3, sourceMoistureState = 'dry')
  
  expect_equal(x$hue, '10YR')
  expect_equal(x$value, 4)
  expect_equal(x$chroma, 3)
  
})

test_that("unusual colors", {
  
  # moist -> dry
  x <- estimateSoilColor(hue = '10YR', value = 3, chroma = 3, sourceMoistureState = 'moist')
  
  expect_equal(x$hue, '10YR')
  expect_equal(x$value, 4)
  expect_equal(x$chroma, 3)
  
  # dry -> moist
  x <- estimateSoilColor(hue = '10YR', value = 5, chroma = 3, sourceMoistureState = 'dry')
  
  expect_equal(x$hue, '10YR')
  expect_equal(x$value, 4)
  expect_equal(x$chroma, 3)
  
})


test_that("OLS", {
  
  # not the same as method = 'procrustes'
  # moist -> dry
  x <- estimateSoilColor(hue = '10YR', value = 3, chroma = 3, sourceMoistureState = 'moist', method = 'ols')
  
  expect_equal(x$hue, '10YR')
  expect_equal(x$value, 5)
  expect_equal(x$chroma, 3)
  
  # dry -> moist
  x <- estimateSoilColor(hue = '10YR', value = 5, chroma = 3, sourceMoistureState = 'dry')
  
  expect_equal(x$hue, '10YR')
  expect_equal(x$value, 4)
  expect_equal(x$chroma, 3)
  
})






# test_that("more complex input", {
#   
#   # data("jacobs2000")
#   # m <- parseMunsell(jacobs2000$matrix_color_munsell, convertColors = FALSE)
#   
#   # m <- structure(list(
#   #   hue = c("10YR", "2.5Y", "10YR", "10YR", "10YR", 
#   #           NA, "2.5YR", "10YR", "10YR", "10YR", "10YR", "10YR", "2.5YR", 
#   #           "10YR", "10YR", "10YR", "2.5Y", "10YR", "10YR", "10YR", "10YR", 
#   #           "10YR", "10YR", "10YR", "10YR", "2.5Y", "10YR", "10YR", "2.5Y", 
#   #           "2.5Y", "10YR", "10YR", "10YR", "2.5Y", "2.5Y", "2.5Y", "10YR", 
#   #           "10YR", "10YR", "10YR", "7.5YR", "10YR", "10YR", "10YR", "10YR", 
#   #           "10YR"), 
#   #   value = c(4, 6, 5, 6, 6, NA, 4, 2, 6, 5, 6, 6, 4, 4, 
#   #             5, 7, 6, 6, 6, 7, 5, 6, 7, 7, 8, 8, 7, 5, 7, 7, 6, 4, 5, 6, 7, 
#   #             6, 7, 7, 4, 8, 3, 3, 6, 7, 2, 6), 
#   #   chroma = c(1, 6, 8, 8, 8, NA, 
#   #              8, 2, 4, 8, 8, 8, 8, 2, 2, 4, 6, 4, 4, 2, 1, 4, 4, 4, 1, 2, 2, 
#   #              2, 4, 3, 6, 6, 1, 3, 3, 3, 2, 1, 1, 1, 2, 2, 3, 2, 1, 1)
#   # ), row.names = c(NA, -46L), class = "data.frame")
#   # 
#   # x <- estimateSoilColor(m$hue, m$value, m$chroma, sourceMoistureState = 'moist')
#   # 
#   # # should be a data.frame of same length
#   # expect_true(inherits(x, 'data.frame'))
#   # expect_equal(nrow(m), nrow(x))
#   
#   # m <- structure(
#   #   list(
#   #     dry = c("10YR 4/1", "10YR 5/1", "10YR 6/2", "10Y 8/0", "10YR 6/0", "10YR 4/2", "10YR 5/2", "2.5Y 5/1", "10YR 4/0", "10YR 5/1"), 
#   #     moist = c("10YR 3/1", "10YR 3/1", "10YR 5/1", "10Y 8/0", "10YR 4/0", "10YR 2.5/1", "10YR 4/2", "2.5Y 4/1", "10YR 2.5/0", "10YR 3/1")
#   #   ), row.names = c(NA, -10L), class = "data.frame"
#   # )
#   # 
#   # m.dry <- parseMunsell(m$dry, convertColors = FALSE)
#   # m.moist <- parseMunsell(m$moist, convertColors = FALSE)
#   # 
#   # e.d <- estimateSoilColor(m.moist$hue, m.moist$value, m.moist$chroma, sourceMoistureState = 'moist')
#   # e.m <- estimateSoilColor(m.dry$hue, m.dry$value, m.dry$chroma, sourceMoistureState = 'dry')
#   # 
#   # od <- sprintf("%s %s/%s", m.dry$hue, m.dry$value, m.dry$chroma)
#   # ed <- sprintf("%s %s/%s", e.d$hue, e.d$value, e.d$chroma)
#   # 
#   # colorContrast(od, ed)
#   
# })

context("color conversion")

## sample data
x <- '10YR 3/4'
x.p <- parseMunsell(x, convertColors = FALSE)

# Munsell --> sRGB conversion
m <- munsell2rgb(x.p$hue, x.p$value, x.p$chroma)
m.rgb <- munsell2rgb(x.p$hue, x.p$value, x.p$chroma, return_triplets = TRUE)

# sRGB --> Munsell
x.back <- col2Munsell(col = m.rgb, nClosest = 1)
# using truncated sRGB values
x.back.trunc <- col2Munsell(col = data.frame(r = 0.36, g = 0.26, b = 0.13))

# neutral colors map to shades of gray
x.neutral <- parseMunsell('N 2/', return_triplets = TRUE)

## tests


# internally-used function for detecting color specification
test_that(".detectColorSpec", {
  
  expect_equal(
    .detectColorSpec(matrix(c(1, 1, 1), nrow = 1)),
    'color-coordinate-matrix'
  )
  
  expect_equal(
    .detectColorSpec(data.frame(r = 1, b = 1, g = 1)),
    'color-coordinate-data.frame'
  )
  
  expect_equal(
    .detectColorSpec(c('red', 'blue', 'white')),
    'named-colors'
  )
  
  expect_equal(
    .detectColorSpec(c('#AABBCC', '#EEFF22')),
    'hex-sRGB'
  )
  
  expect_equal(
    .detectColorSpec(c('10YR 3/3', '5G 4/4', 'N 3/')),
    'munsell'
  )
  
  # partial Munsell
  expect_equal(
    .detectColorSpec(c('10YR ', '5G 4')),
    'munsell'
  )
  
  # unknown conditions
  expect_equal(
    .detectColorSpec(c('red', 'blue', 'dd')),
    'unknown'
  )
  
  expect_equal(
    .detectColorSpec(1),
    'unknown'
  )
  
  expect_equal(
    .detectColorSpec(character(0)),
    'unknown'
  )
  
})


test_that("col2Munsell works as expected", {
  
  # 10YR 3/4 as hex notation sRGB 
  .res <- col2Munsell(m)
  expect_equal(.res$hue, '10YR')
  expect_equal(.res$value, 3L)
  expect_equal(.res$chroma, 4L)
  
  
  # N 2/ as sRGB [0,1]
  .res <- col2Munsell(x.neutral, space = 'sRGB')
  expect_equal(.res$hue, 'N')
  expect_equal(.res$value, 2L)
  expect_equal(.res$chroma, 0L)
  
})



test_that("parseMunsell()", {
  
  # parsing bogus notation generates NA
  # will also generate a warning from munsell2rgb()
  expect_equal(suppressWarnings(parseMunsell('10YZ 4/5')), NA_character_)
  expect_equal(suppressWarnings(parseMunsell('10YR /5')), NA_character_)
  expect_equal(suppressWarnings(parseMunsell('10YR ')), NA_character_)
  expect_equal(suppressWarnings(parseMunsell('10YR 4/')), NA_character_)
  expect_equal(suppressWarnings(parseMunsell('G1 6/N')), NA_character_)
  
  # parsing bogus notation without conversion
  bogus <- parseMunsell('G1 3/X', convertColors = FALSE)
  expect_equal(bogus$hue, NA_character_)
  expect_equal(bogus$value, NA_real_)
  expect_equal(bogus$chroma, NA_real_)
  
  # test NA
  some.NA <- parseMunsell(c(NA, '10YR 3/3'))
  expect_true(inherits(some.NA, 'character'))
  expect_true(length(some.NA) == 2)
  
  some.NA <- parseMunsell(c(NA, '10YR 3/3'), convertColors = FALSE)
  expect_true(inherits(some.NA, 'data.frame'))
  expect_true(nrow(some.NA) == 2)
  
  # splitting of text into columns within data.frame
  expect_identical(x.p, data.frame(hue = "10YR", value = 3, chroma = 4, stringsAsFactors = FALSE))
  
  # Test not using spaces
  expect_equal(suppressWarnings(parseMunsell('2.5YR 3/4')), suppressWarnings(parseMunsell('2.5YR3/4')))
  
  # Test different delimiters
  expect_equal(suppressWarnings(parseMunsell('2.5YR 3/4')), suppressWarnings(parseMunsell('2.5YR 3_4', delim = "_")))
  expect_equal(suppressWarnings(parseMunsell('2.5YR 3/4')), suppressWarnings(parseMunsell('2.5YR 3_4')))
})


# addresses #66 (https://github.com/ncss-tech/aqp/issues/66)
test_that("Munsell hue parsing", {
  
  # normal operation
  res <- .parseMunsellHue('10YR')
  expect_true(inherits(res, 'data.frame'))
  expect_equal(res$hue.numeric, 10L)
  expect_equal(res$hue.character, 'YR')
  expect_equal(nrow(res), 1)
  
  # white space is trimmed
  res <- .parseMunsellHue('10 YR')
  expect_true(inherits(res, 'data.frame'))
  expect_equal(res$hue.numeric, 10L)
  expect_equal(res$hue.character, 'YR')
  expect_equal(nrow(res), 1)
  
  # decimal, won't convert correctly, but should be split
  res <- .parseMunsellHue('10.1YR')
  expect_true(inherits(res, 'data.frame'))
  expect_equal(res$hue.numeric, 10.1)
  expect_equal(res$hue.character, 'YR')
  expect_equal(nrow(res), 1)
  
  # bogus hue
  res <- .parseMunsellHue('G1 ')
  expect_true(inherits(res, 'data.frame'))
  expect_true(is.na(res$hue.numeric))
  expect_true(is.na(res$hue.character))
  expect_equal(nrow(res), 1)
})


test_that("non-integer value and chroma are snapped to valid possibilities", {
  
  # rounding of value, throws warning
  expect_warning(res <- parseMunsell('10YR 3.3/4'), regexp = 'non-standard notation')
  
  # this will not throw a warning
  res <- parseMunsell('10YR 3.3/4', convertColors = FALSE)
  
  # results should be the same
  expect_equal(
    suppressWarnings(parseMunsell('10YR 3.3/4')),
    parseMunsell('10YR 3/4')
  )
  
  # 2.5,  8.5, 9.5 are valid Munsell value
  # snap nearby to these
  expect_equal(
    suppressWarnings(parseMunsell('10YR 2.6/4')),
    parseMunsell('10YR 2.5/4')
  )
  
  expect_equal(
    suppressWarnings(parseMunsell('10YR 8.4/4')),
    parseMunsell('10YR 8.5/4')
  )
  
  expect_equal(
    suppressWarnings(parseMunsell('10YR 9.55/4')),
    parseMunsell('10YR 9.5/4')
  )
  
  # rounding of chroma, throws warning
  expect_warning(res <- parseMunsell('10YR 3/4.6'), regexp = 'non-standard notation')
  # this will not throw a warning
  res <- parseMunsell('10YR 3/4.6', convertColors = FALSE)
  
  # results should be the same
  expect_equal(
    suppressWarnings(parseMunsell('10YR 3/4.6')),
    parseMunsell('10YR 3/5')
  )
  
  # double-check no snapping of 2.5 values
  res <- parseMunsell('10YR 2.5/2')
  res.test <- col2Munsell(col = '#493A2BFF')
  
  expect_true(res.test$value == 2.5)
  
  
  # test .snapValid() directly
  .valid <- c(2, 2.5, 3, 4, 5, 6, 7, 8, 8.5, 9, 9.5, 10)
  v <- c(2, 4, 6, 8, NA, 2.5, 2.6, 8.1, 9.6, 12)
  res <- aqp:::.snapValid(v, .valid)
  
  # ensure snapping worked as expected
  expect_equal(
    res, 
    c(2, 4, 6, 8, NA, 2.5, 2.5, 8, 9.5, 10)
  )
  
})


test_that("Munsell <--> sRGB and back again", {
  
  # sRGB in hex notation
  expect_equal(m, '#5C4222FF')
  expect_equal(parseMunsell(x), m)
  
  # sRGB triplets
  expect_equal(m.rgb$r, 0.3618738, tolerance=0.0001)
  expect_equal(m.rgb$g, 0.2598939, tolerance=0.0001)
  expect_equal(m.rgb$b, 0.1337521, tolerance=0.0001)
  
  # neutral colors
  expect_equal(x.neutral$r, 0.1996, tolerance=0.001)
  expect_equal(x.neutral$g, 0.1996, tolerance=0.001)
  expect_equal(x.neutral$b, 0.1996, tolerance=0.001)
  
  # sRGB --> Munsell
  expect_equal(x.back$hue, '10YR')
  expect_equal(x.back$value, 3)
  expect_equal(x.back$chroma, 4)
  expect_equal(x.back$sigma, 0)
  
  expect_equal(x.back.trunc$hue, '10YR')
  expect_equal(x.back.trunc$value, 3)
  expect_equal(x.back.trunc$chroma, 4)
})


test_that("missing data", {
  
  # data with missing sRGB coordinates
  color <- rbind(
    cbind(NA, NA, NA),
    cbind(0.5, 0.2, 0.2),
    cbind(1, 1, 1),
    cbind(NA, NA, NA)
  )
  
  # conversion should work without error
  res <- col2Munsell(color, space = 'sRGB')
  
  # same number of rows in / out
  expect_true(nrow(res) == nrow(color))
  
  # row order preserved
  expect_true(is.na(res$hue[1]) & is.na(res$hue[4]))
  
  # data with missing character representation
  color <- c('red', NA, NA, 'blue')
  res <- col2Munsell(color, space = 'sRGB')
  expect_true(nrow(res) == length(color))
  
  
  # all NA
  color <- c(NA, NA, NA, NA)
  res <- col2Munsell(color, space = 'sRGB')
  expect_true(nrow(res) == length(color))
  
  color <- rbind(
    cbind(NA, NA, NA),
    cbind(NA, NA, NA)
  )
  
  res <- col2Munsell(color, space = 'sRGB')
  expect_true(nrow(res) == nrow(color))
  
})

test_that("neutral hues", {
  
  # neutral colors
  res <- parseMunsell('N 2/', convertColors = FALSE)
  expect_true(inherits(res, 'data.frame'))
  
  # chroma is arbitrarily set to 0
  expect_true(res$chroma == 0)
  
  N2 <- parseMunsell('N 2/')
  N6 <- parseMunsell('N 6/')
  expect_equal(N2, '#333333FF')
  expect_equal(N6, '#8F8F8FFF')
  
  # alternative notation
  N2 <- parseMunsell('N 2')
  N6 <- parseMunsell('N 6/0')
  expect_equal(N2, '#333333FF')
  expect_equal(N6, '#8F8F8FFF')
  
  # missing, 0, or bogus chroma are all ignored
  N2 <- munsell2rgb('N', 2, NA)
  N4 <- munsell2rgb('N', 4, 0)
  N6 <- munsell2rgb('N', 6, 4)
  expect_equal(N2, '#333333FF')
  expect_equal(N4, '#5C5C5CFF')
  expect_equal(N6, '#8F8F8FFF')
  
  # 0 chroma with any hue -> neutral
  N <- munsell2rgb('10YR', 4, 0)
  expect_equal(N, '#5C5C5CFF')
})


test_that("closest Munsell chip based on sRGB coordinates", {
  
  # closest chip in aqp LUT
  expect_equal(getClosestMunsellChip('10YR 3.3/5', convertColors = FALSE), '10YR 3/5')
  expect_equal(getClosestMunsellChip('9YR 3.8/3', convertColors = FALSE), '10YR 4/3')
  
  # 2022-03-30: now we get correct selection of 2.5 value
  expect_equal(getClosestMunsellChip('7.9YR 2.7/2.0', convertColors = FALSE), '7.5YR 2.5/2')
})


# https://github.com/ncss-tech/aqp/issues/69
test_that("Munsell --> LAB + sRGB coordinates", {
  
  # sRGB
  test.1 <- parseMunsell("10YR 3/5", return_triplets=TRUE)
  expect_equal(names(test.1), c('r', 'g', 'b'))
  
  
  # sRGB and LAB
  test.2 <- parseMunsell("10YR 3/5", return_triplets=TRUE, returnLAB=TRUE)
  expect_equal(names(test.2), c('r', 'g', 'b', 'L', 'A', 'B'))
  
  # LAB
  test.3 <- parseMunsell("10YR 3/5", return_triplets=FALSE, returnLAB=TRUE)
  expect_equal(names(test.3), c('L', 'A', 'B'))
  
  # test the LAB ---> sRGB is close
  test.4 <- grDevices::convertColor(test.3, from = 'Lab', to='sRGB')
  
  # sRGB (r)
  expect_equal(test.1[, 1], test.4[, 1], tolerance=0.1)
  # sRGB (g)
  expect_equal(test.1[, 2], test.4[, 2], tolerance=0.1)
  # sRGB (b)
  expect_equal(test.1[, 3], test.4[, 3], tolerance=0.1)
})


test_that("similar colors result in same, closest chip", {
  
  cols <- t(col2rgb(c('#5F5335', '#5F5236'))) / 255
  res <-  col2Munsell(cols)
  
  expect_equal(res$hue[1], res$hue[2])
  expect_equal(res$value[1], res$value[2])
  expect_equal(res$chroma[1], res$chroma[2])
})


test_that("munsell2spc wrapper method works as expected", {
  
  data(sp3)
  depths(sp3) <- id ~ top + bottom
  
  # inspect input data
  # horizons(sp3)[,c("hue","value","chroma")]
  
  # do color conversions to sRGB and LAB, join into horizon data
  expect_silent( {sp3 <- munsell2spc(sp3)})
  expect_true(inherits(sp3, 'SoilProfileCollection'))
  
  # # plot rgb "R" coordinate by horizon
  # plot(sp3, color = "rgb_R")
  #
  # # plot lab "A" coordinate by horizon
  # plot(sp3, color = "lab_A")
  
  # test returning profile+horizon ID data.frame with results
  expect_silent( { dftest <- munsell2spc(sp3, as.spc = FALSE) } )
  expect_true(inherits(dftest, 'data.frame'))
  
  # foo is not a column in horizons()
  expect_error( { err1 <- munsell2spc(sp3, hue = "foo") } )
  
  # chip is not a column in horizons
  expect_error( { d1 <- horizons(munsell2spc(sp3, .data = "chip")) } )
  
  # create chip as a combination of hue value/chroma
  sp3$chip <- with(horizons(sp3), sprintf("%s %s/%s", hue, value, chroma))
  
  # calculate from: column name, vector, data.frame
  expect_silent( { d1 <- horizons(munsell2spc(sp3, .data = "chip")) } )
  expect_silent( { d2 <- horizons(munsell2spc(sp3, .data = sp3$chip)) } )
  expect_silent( { d3 <- horizons(munsell2spc(sp3, .data = parseMunsell(sp3$chip, convertColors = FALSE))) } )
  expect_silent( { d4 <- horizons(munsell2spc(sp3, .data = data.frame(foo = sp3$chip))) } )
})

context("aggregateColor")

## sample data
data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

sp1$soil_color <- munsell2rgb(sp1$hue, sp1$value, sp1$chroma)

## tests

test_that("basic functionality", {
  
  skip_if_not_installed('gower')
  
  # group all horizons
  x <- sp1
  x$genhz <- rep('A', times=nrow(x))
  a <- aggregateColor(x, groups='genhz', col='soil_color')

  # object of correct structure
  expect_true(inherits(a,'list'))
  expect_true(length(a) == 2)
  expect_true(all(names(a) == c('scaled.data', 'aggregate.data')))
  expect_true(names(a$scaled.data) == 'A')
  expect_true(a$aggregate.data$genhz == 'A')

  # number of colors in result should be the number of unique colors (after removing NA)
  expect_true(nrow(a$scaled.data$A) == length(na.omit(unique(x$soil_color))))
  # number of horizons in aggregate.data should be the same
  expect_true(a$aggregate.data$n == length(na.omit(unique(x$soil_color))))

  # colors / Munsell representations match
  test <- parseMunsell(a$scaled.data$A$munsell) == a$scaled.data$A$soil_color
  expect_true(all(test))
})

## TODO: test for expected error conditions
test_that("expected error conditions", {
  skip_if_not_installed('gower')
  expect_error(aggregateColor(x, groups='foo', col='soil_color'))
  expect_error(aggregateColor(x, groups='genhz', col='foo'))
  expect_error(aggregateColor(x, groups='genhz', col='soil_color', k=NA))
})

## TODO: add a couple more of these
test_that("manual calculation using CIE2000, single profile", {
  skip_if_not_installed('gower')

  x <- sp1[1, ]
  x$genhz <- rep('A', times=nrow(x))
  a <- aggregateColor(x, groups='genhz', col='soil_color')
  a2 <- aggregateColor(x, groups='genhz', col='soil_color')
  a3 <- aggregateColor(x, groups='genhz', col='soil_color', k=1)

  # known number of horizons / color
  # table(x$soil_color)
  expect_equal(a$scaled.data$A$n.hz, c(2,1,1,1))
  expect_equal(a2$scaled.data$A$n.hz, c(2,1,1,1))
  expect_equal(a3$scaled.data$A$n.hz, 5)

  expect_equal(round(a$scaled.data$A$weight, 3), c(0.342, 0.270, 0.258, 0.129))

  test <- with(a$aggregate.data, paste0(hue, ' ', value, '/', chroma))
  test2 <- with(a2$aggregate.data, paste0(hue, ' ', value, '/', chroma))

  if(packageVersion("farver") >= '2.0.2') {
    expect_equal(test, '7.5YR 3/2')
    expect_equal(test2, '7.5YR 3/2')
  } else {
    expect_equal(test, '10YR 3/2')
    expect_equal(test2, '7.5YR 3/2')
  }
})


test_that("manual calculation, single profile", {
  skip_if_not_installed('gower')
  
  data(sp3)
  depths(sp3) <- id ~ top + bottom
  
  # single profile
  x <- sp3[4, ]
  # group all horizons together
  site(x)$group <- 'A'
  
  # do the work
  a <- aggregateColor(x, groups = 'group', col = 'soil_color')
  
  # manually verified
  expect_true(a$aggregate.data$hue == '5YR')
  expect_true(a$aggregate.data$value == '5')
  expect_true(a$aggregate.data$chroma == '5')
  
  # TODO: verify weights
  
})

#' @title Subset an SPC by applying glom to each profile
#' @name glomApply
#' @aliases glomApply,SoilProfileCollection-method
#'
#' @description \code{glomApply()} is a function used for subsetting SoilProfileCollection objects by depth. It is a wrapper around \code{glom} which is intended to subset single-profile SPCs based on depth intervals/intersection.
#'
#' \code{glomApply} works by accepting a function \code{.fun} as argument. This function is used on each profile to process a multi-profile SPC for input to \code{glom} (via \code{profileApply}). For each profile, \code{.fun} returns a 2-length numeric vector of top and bottom boundaries \code{glom} arguments: \code{z1}, \code{z2}.
#'
#' \code{glomApply} provides the option to generate profile-specific glom depths for a large SPC and handles iteration and rebuilding of a subset SPC object. Optional arguments include: \code{truncate} to cut the boundaries to specified \code{[z1, z2]}; \code{invert} to the portion outside \code{[z1, z2]}, \code{modality} to either \code{"all"} horizons or \code{"thickest"} horizon in the \code{glom} interval. \code{...} are various expressions you can run on the individual profiles using NSE, similar to \code{mutate}.
#'
#' @param object A SoilProfileCollection
#'
#' @param .fun A function that returns vector with top and bottom depth (\code{z1} and \code{z2} arguments to \code{glom}) for a single profile \code{p} (as passed by \code{profileApply})
#'
#' @param truncate Truncate horizon top and bottom depths to \code{[z1, z2]}
#'
#' @param invert Truncate horizon top and bottom depths to \code{[z1, z2]} and then invert result?
#'
#' @param modality Aggregation method for glom result. Default \code{"all"}: return all horizons; \code{"thickest"}: return (shallowest) thickest horizon
#'
#' @param ... A set of comma-delimited R expressions that resolve to a transformation to be applied to a single profile e.g \code{glomApply(hzdept = max(hzdept) - hzdept)} like \code{aqp::mutate}
#'
#' @param chunk.size Chunk size parameter for \code{profileApply}
#'
#' @seealso \code{\link{glom}}  \code{\link{trunc}}
#'
#' @return A SoilProfileCollection.
#' @author Andrew G. Brown.
#'
#' @rdname glomApply
#' @export glomApply
#'
#' @seealso \code{\link{glom}}  \code{\link{glomApply}}
#' @examples
#'
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#'
#' data(sp3)
#' depths(sp3) <- id ~ top + bottom
#' 
#' # init horizon designation column in metadata, used by estimateSoilDepth
#' hzdesgnname(sp3) <- 'name'
#'
#' # constant depths, whole horizon returns by default
#' plot(glomApply(sp3, function(p) c(25,100)))
#'
#' # constant depths, truncated
#' #(see aqp::trunc for helper function)
#' plot(glomApply(sp3, function(p) c(25,30), truncate = TRUE))
#'
#' # constant depths, inverted
#' plot(glomApply(sp3, function(p) c(25,100), invert = TRUE))
#'
#' # constant depths, inverted + truncated (same as above)
#' plot(glomApply(sp3, function(p) c(25,30), invert = TRUE, truncate=TRUE))
#'
#' # random boundaries in each profile
#' plot(glomApply(sp3, function(p) round(sort(runif(2, 0, max(sp3))))))
#'
#' # random boundaries in each profile (truncated)
#' plot(glomApply(sp3, function(p) round(sort(runif(2, 0, max(sp3)))), truncate = TRUE))
#'
#' # calculate some boundaries as site level attribtes
#' sp3$glom_top <- profileApply(sp3, getMineralSoilSurfaceDepth)
#' sp3$glom_bottom <- profileApply(sp3, estimateSoilDepth)
#'
#' # use site level attributes for glom intervals for each profile
#' plot(glomApply(sp3, function(p) return(c(p$glom_top, p$glom_bottom))))
#'
glomApply <- function(object, .fun = NULL, truncate = FALSE, invert = FALSE,
                      modality = "all", ..., chunk.size = 100) {
  if (is.null(.fun) | !inherits(.fun, 'function'))
    stop("function `.fun`` to return glom boundaries for profiles is missing", call. = FALSE)
  pbindlist(profileApply(object, function(p, ...) {
    dep <- .fun(p, ...)
    return(glom(p, dep[1], dep[2], truncate = truncate, modality = modality, invert = invert))
  }, simplify = FALSE, chunk.size = chunk.size))
}

context("glomming clods - ragged horizon intersection for SPCs")

data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

p <- sp1[6]
attr <- 'prop' # clay contents %

test_that("intersection of horizons by depth", {
    # intersection at a single depth should return only one horizon
    expect_equal(sum(glom(p, 50, ids = TRUE) %in% hzID(p)), 1)

    # intersection from 25 to 100 should return four horizons
    expect_equal(sum(glom(p, 25, 100, ids = TRUE) %in% hzID(p)), 4)
})

test_that("depthWeights parity", {
  # depthWeights calculates contributing fractions within glom intervals for a profile p
  wts <- depthWeights(p$top, p$bottom, 25, 100)

  # wrong length errors
  expect_error(depthWeights(p$top, p$bottom, c(25,50), 100))
  expect_error(depthWeights(p$top[1], p$bottom, 25, 100))

  expect_equal(wts, c(0, 0, 0.32, 0.186666666666667, 0.36, 0.133333333333333, 0, 0, 0, 0))
  expect_equal(sum(wts), 1)
  glmp <- glom(p, 25, 100, truncate = TRUE)
  glmp$thk <- glmp$bottom - glmp$top
  glmp$relthk <- glmp$thk / sum(glmp$thk)
  expect_equal(glmp$relthk, wts[wts > 0])
})

test_that("degenerate single horizon case", {
    # glom tests with a degenerate case
    test <- data.frame(id = 1, top = 0, bottom = 100)
    depths(test) <- id ~ top + bottom

    # default glom
    expect_silent(t1 <- glom(test, 25, 50))
    expect_equal(t1$bottom - t1$top, 100)

    # thickest-modality default glom
    expect_silent(t1 <- glom(test, 25, 50, modality = "thickest"))
    expect_equal(t1$bottom - t1$top, 100)

    # truncate glom
    expect_silent(t1 <- glom(test, 25, 50, truncate = TRUE))
    expect_equal(t1$bottom - t1$top, 25)

    # thickest-modality truncate glom
    expect_silent(t1 <- glom(test, 25, 50, truncate = TRUE, modality = "thickest"))
    expect_equal(t1$bottom - t1$top, 25)

    # invert glom
    expect_silent(t1 <- glom(test, 25, 50, invert = TRUE))
    expect_equal(t1$bottom - t1$top, 100)

    # invert + truncate glom
    expect_silent(t1 <- glom(test, 25, 50, invert = TRUE, truncate = TRUE))
    expect_equal(t1$bottom - t1$top, c(25, 50))

    # thickest-modality invert + truncate glom
    expect_silent(t1 <- glom(test, 25, 50, invert = TRUE, truncate = TRUE, modality = "thickest"))
    expect_equal(t1$bottom - t1$top, 50)
})

test_that("glom by depth returns a SPC clod", {
  # glom 'gloms' your input SPC `p`'s horizons (by depths specified) into a 'clod'

  # currently "clods" can be either represented as an SPC, or a data.frame with just
  # the horizons that are contained within the "clod".
  foo <- glom(p, 25, 100)
  # and returns an SPC
  expect_true(inherits(foo, 'SoilProfileCollection'))

  # within that SPC there should be only one profile
  expect_equal(length(foo), 1)

  # and that profile should have 4 horizons in 25-100cm
  expect_equal(nrow(foo), 4)
})

test_that("glom by depth returns a data.frame clod", {
  # glom 'gloms' your input SPC `p`'s horizons (by depths specified) into a 'clod'

  # currently "clods" can be either represented as an SPC, or a data.frame with just
  # the horizons that are contained within the "clod".
  foo <- glom(p, 25, 100, df = TRUE)

  # and returns an data.frame
  expect_true(inherits(foo, 'data.frame'))

  # within that data.frame, length() returns 18
  expect_equal(length(foo), 18)

  # and that data.frame should have 4 horizons (rows) in 25-100cm
  expect_equal(nrow(foo), 4)
})

test_that("glom truncate = TRUE works as expected", {
  # glom 'gloms' your input SPC `p`'s horizons (by depths specified) into a 'clod'

  # get truncated clod
  foo <- glom(p, 25, 100, truncate=TRUE)

  # and returns an data.frame
  expect_true(inherits(foo, 'SoilProfileCollection'))

  ## test that:
  # shallowest top truncated to z1
  # deepest bottom truncated to z2
  expect_equal(min(foo$top), 25)
  expect_equal(max(foo$bottom), 100)

})

test_that("glomApply works as expected", {

  # using same depth returns a single horizon from every profile
  # glomApply with same z1 as z2
  expect_silent(res <- glomApply(sp1, function(p) return(c(25,25))))

  # above is the same as calling glom with just z1 specified (for first profile in sp1)
  expect_equivalent(res[1,], glom(sp1[1,], 25))

  # after pbindlist hzID 3 becomes 1 (since sp1 does not have a true hzID specified)
  tdepths <- horizonDepths(sp1)
  expect_equal(horizons(res[1,])[,tdepths], horizons(glom(sp1[1,], 25))[,tdepths])

  # every profile returns one horizon (all profiles at least 25cm deep)
  expect_true(all(profileApply(res, nrow) == 1))

  # 7 profiles have 200 as an invalid upper bound
  ## note that P006 and P008 end EXACTLY at 200 and are NOT included
  # plot(sp1)
  # abline(h = 200)
  expect_silent(res2 <- glomApply(sp1, function(p) return(c(200,200))))

  expect_equal(names(profileApply(res2, nrow)), c("P007", "P009"))

  test_that("realistic glomApply scenarios", {
    data(sp3)
    depths(sp3) <- id ~ top + bottom
    hzdesgnname(sp3) <- 'name'

    # constant depths, whole horizon returns by default
    expect_silent(glomApply(sp3, function(p) c(25,100)))

    # constant depths, truncated
    #(see aqp::trunc for helper function)
    expect_silent(glomApply(sp3, function(p) c(25,30), truncate = TRUE))

    # constant depths, inverted
    expect_silent(glomApply(sp3, function(p) c(25,100), invert = TRUE))

    # constant depths, inverted + truncated (same as above)
    expect_silent(res <- glomApply(sp3, function(p) c(25,30), invert = TRUE, truncate = TRUE))

    # before invert, 46 horizons
    expect_equal(nrow(sp3), 46)

    # after glom invert, some horizons are split, increasing the total number
    expect_equal(nrow(res), 52)

    # random boundaries in each profile the specific warnings are a product of pseudorandom numbers
    set.seed(100)
    expect_silent(glomApply(sp3, function(p) round(sort(runif(2, 0, max(sp3))))))

    # random boundaries in each profile (truncated)
    expect_silent(glomApply(sp3, function(p) round(sort(runif(2, 0, max(sp3)))), truncate = TRUE))

    # calculate some boundaries as site level attribtes
    expect_warning(sp3$glom_top <- profileApply(sp3, getMineralSoilSurfaceDepth))
    expect_silent(sp3$glom_bottom <- profileApply(sp3, estimateSoilDepth))

    # use site level attributes for glom intervals for each profile
    expect_silent(glomApply(sp3, function(p) return(c(p$glom_top, p$glom_bottom))))
  })

  # trunc using vectorized glom v.s. iterating with constant depths works as expected
  expect_equivalent({
      glomApply(sp1, function(p) return(c(25,36)), truncate = TRUE)
    },{
      trunc(sp1, 25, 36)
    })
})

test_that("glom vectorization", {
  # 4 of 9 profiles dropped
  expect_equal(length(glom(sp1, 75, 100)), 5)

  # truncated max SPC depth is 100cm
  expect_equal(max(glom(sp1, 75, 100, truncate = TRUE)), 100)

  # truncated and inverted
  t3 <- glom(sp1, 75, 100, truncate = TRUE, invert = TRUE)

  # all of profiles in invert result
  expect_equal(length(t3), 9)

  # max SPC depth is 240cm
  expect_equal(max(t3), 240)
  
  z1 <- c(rep(c(0,55), 5))[1:9]
  z2 <- c(rep(c(10,60), 5))[1:9]
  
  # profile specific glom boundaries (alternating [0,10] [55,60])
  t4 <- glom(sp1, z1, z2)
  expect_equal(t4$bottom, c(2L, 14L, 59L, 2L, 13L, 62L, 5L,
                            30L, 63L, 4L, 16L, 68L, 3L, 14L))  
 
  # truncate = TRUE using same boundaries 
  t5 <- glom(sp1, z1, z2, truncate = TRUE)
  expect_equal(t5$bottom, c(2L, 10L, 59L, 2L, 10L, 60L, 5L, 
                            10L, 60L, 4L, 10L, 60L, 3L, 10L))  
  
  # truncate twice, initially with drop=FALSE
  expect_equal(length(trunc(trunc(sp1, 75, 100, drop = FALSE), 90, 100)), 4)
  
  # no horizons in glom interval, with missing profiles dropped
  expect_equal(length(trunc(sp1, 250, 300)), 0)
  
  # no horizons in glom interval, with missing profiles filled
  expect_equal(length(trunc(sp1, 250, 300, drop = FALSE)), 9)
})

#' Store groupings within a profile collection.
#' @name groupSPC
#' @param object SoilProfileCollection.
#' @param ... One or more expressions evaluated within the context of \code{object} that resolve to vectors that can be coerced to factor "groups."
#' @export
groupSPC <- function(object, ...) {

    # capture expression(s) at function
    x <- substitute(list(...))

    if (is.null(names(x)))
      names(x) <- as.character(x)

    # create composite object to facilitate eval_tidy
    data <- compositeSPC(object)

    # evaluate n expressions to get n grouping vars
    foo <- NULL

    for(n in 2:length(x)) {

      bar <- as.character(x[[n]])

      if(inherits(foo, 'try-error'))
        stop("groupSPC expects expressions that resolve to one or more site or horizon level column names", call.=FALSE)

      if(any(is.na(bar))) {
        stop(sprintf("groupSPC expects no NA values in '%s'", n), call.=FALSE)
      } else if(length(bar) == length(object)){
        foo <- c(foo, n)
      } else if(bar %in% siteNames(object)) {
        foo <- c(foo, bar)
      } else {
        stop("groupSPC expects expressions that resolve to one or more site or horizon level attributes", call.=FALSE)
      }
    }

    # TODO: safe getter/setter metod
    # set value of metadata "SPC group" - which defines one or more
    #   site (or horizon) level attrs that are used for grouping in certain functions
    #   in order of precedence, pipe-delimited.
    object@metadata$aqp_group_by <- paste0(foo, collapse = "|")

    return(object)
}

data(sp3)
depths(sp3) <- id ~ top + bottom
site(sp3)$group <- "A"
sp3$group[3:6] <- "B"

test_that("groupSPC & summarizeSPC", {
  
  sp3 <- groupSPC(sp3, group)
  expect_equal(metadata(sp3)$aqp_group_by, "group")
  
  # mean for A and B group horizon data
  summa <- summarizeSPC(sp3, round(mean(clay)), round(sd(clay)))
  expect_equal(summa, structure(list(group = c("A", "B"),
                                     `round(mean(clay))` = c(11,29),
                                     `round(sd(clay))` = c(5, 12)),
                                class = "data.frame", row.names = c(NA, -2L)))
})

#' Determine thickness of horizons (continuous from surface) matching a pattern
#'
#' @description Find the thickness of horizon designations, or any other character patterns, that are continuous from the soil surface (depth = 0 or shallowest depth in profile).
#'
#' @details The horizon designation to match is specified with the regular expression pattern 'pattern'. All horizons matching that pattern, that are continuous from the soil surface, count towards the depth / thickness value that is ultimately returned. For instance: horizon designations: A1-A2-A3-C-Ab , would return A3 bottom depth given \code{pattern = "^A[1-9]*$"}.
#'
#' `getSurfaceHorizonDepth` is used by `getPlowLayerDepth` for matching Ap horizons; and, it is used by `getMineralSoilSurfaceDepth` to find the thickness of O horizons in lieu of lab data.
#'
#' @param p a SoilProfileCollection
#' @param pattern a regular expression pattern to match for all horizons to be considered part of the "surface".
#' @param hzdesgn column name containing horizon designation. Default: `hzdesgnname(p, required = TRUE)`.
#' @param simplify logical. Return single profile results as vector (default: `TRUE`) or `data.frame` (`FALSE`)
#' @return a numeric value corresponding to the bottom depth of the last horizon matching 'pattern' that is contiguous with other matching horizons up to the soil surface. If `length(p) > 1` then a _data.frame_ containing profile ID, horizon ID, top or bottom depths, horizon designation and pattern.
#'
#' @author Andrew G. Brown
#'
#' @aliases getMineralSoilSurfaceDepth
#' @aliases getPlowLayerDepth
#'
#' @export 
#' @examples
#' library(aqp)
#' data(sp1, package = 'aqp')
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' p <- sp1[1]
#' q <- sp1[2]
#'
#' # look at horizon designations in p and q
#' p$name
#' q$name
#'
#' # thickness of all surface horizons containing A
#' getSurfaceHorizonDepth(p, pattern = 'A', hzdesgn = 'name')
#'
#' # thickness of all surface horizons that start with A
#' getSurfaceHorizonDepth(p, pattern = '^A', hzdesgn = 'name')
#'
#' # thickness of all surface horizons that start with A, and the A is not followed by B
#' getSurfaceHorizonDepth(p, pattern = '^A[^B]*', hzdesgn = 'name')
#'
#' # thickness of all surface horizons that start with A
#' #  followed by a number from _2_ to 9 (returns ZERO)
#' getSurfaceHorizonDepth(p, pattern = '^A[2-9]*', hzdesgn = 'name')
#'
#' # getPlowLayerDepth matches first two horizons in fake Ap horizon data with "buried Ap"
#' p$aphorizons <- c("Ap1","Ap2","AB", rep('C', nrow(p) - 4), "Apb")
#' getPlowLayerDepth(p, hzdesgn = 'aphorizons')
#'
#' # getMineralSoilSurfaceDepthmatches first 3 horizons in fake O horizon data
#' p$ohorizons <- c("Oi1","Oi2","Oe", rep('C', nrow(p) - 4), "2C")
#' getMineralSoilSurfaceDepth(p, hzdesgn='ohorizons')
#'
#' # matches first Oi horizon with original horizon designations of pedon 2
#' getMineralSoilSurfaceDepth(q, hzdesgn='name')
#'
getSurfaceHorizonDepth <- function(p,
                                   pattern,
                                   hzdesgn = hzdesgnname(p, required = TRUE),
                                   simplify = TRUE) {
  
  
  if (is.null(hzdesgn) || !hzdesgn %in% horizonNames(p)) {
    stop("Horizon designation column (", hzdesgn, ") does not exist.")
  }
  
  hz <- data.table::as.data.table(horizons(p))
  depthz <- horizonDepths(p)
  
  .FIRST <- NULL
  .SD <- NULL
  
  shallowest.depth <- p[, , .FIRST][[depthz[1]]]
  
  .get_contiguous_surface_hz <- function(h, hzdesgn) {
    # identify horizons matching pattern
    match.idx <- grepl(h[[hzdesgn]], pattern = pattern)
    
    emptyres <- h[1, .SD, .SDcols = c(hzidname(p), depthz[2])] 
    emptyres[[idname(p)]] <- emptyres[[idname(p)]][1]
    emptyres[[hzidname(p)]] <- NA_character_
    emptyres[[depthz[2]]] <- as.double(0)
    emptyres$pattern <- pattern
    emptyres$hzdesgn <- hzdesgn
    
    # no match? return zero or shallowest top depth (the minimum depth)
    if (length(which(match.idx)) < 1) {
      return(emptyres)
    }
    
    # identify surface horizon
    mod.idx <- c(1, rep(0, length(match.idx) - 1))
    
    # identify where matches and surface horizon co-occur
    # matching horizons get a 1, matching surface horizon gets a 2,
    #   0s kick us out
    new.idx <- match.idx + mod.idx
    
    who.idx <- numeric(0)
    # we only have a matching surface if the first value is 2
    #  (meets both above crit)
    if (new.idx[1] == 2) {
      # convert that into logical to identify contiguous matches
      contig <- new.idx > 0 & new.idx <= 2
      
      # calculate difference between contiguous matches/nonmatches
      dcontig <- diff(as.integer(contig))
      
      # max depth is, at first, the bottom depth of last contiguous hz
      max.idx <- length(contig)
      
      if(length(max.idx)) {
        # if we have a negative change at any depth,
        #  we have a discontinuity
        
        # adjust max index (depth) accordingly
        if(any(dcontig < 0)) {
          # take first index of negative dcontig
          # add one to correct for indexing offset due to diff()
          max.idx <- which(dcontig < 0)[1] + 1
        }
        
        # return last value from contig
        # (last contiguous horizon with surface)
        who.idx <- rev(which(contig[1:max.idx]))[1]
      }
    }
    
    if (length(who.idx) == 0)
      return(emptyres)
    
    res <- h[who.idx, .SD, .SDcols = c(hzidname(p), depthz[2])] 
    res[[depthz[2]]] <- as.double(res[[depthz[2]]])
    res$pattern <- pattern
    res$hzdesgn <- hzdesgn
    res
  }
  
  hzids <- list(hz[[idname(p)]])
  names(hzids) <- idname(p)
  
  sitetemplate <- data.table::as.data.table(site(p))[,.SD, .SDcols = idname(p)]
  sitetemplate[[idname(p)]] <- as.character(sitetemplate[[idname(p)]])
  hzsub <- hz[, .get_contiguous_surface_hz(.SD, hzdesgn), by = hzids]
  hzsub[[idname(p)]] <- as.character(hzsub[[idname(p)]])
  out <- hzsub[sitetemplate, on = idname(p)]
  
  # not found
  notfound <- sum(shallowest.depth > out[[depthz[2]]], na.rm = TRUE)
  if (notfound > 0) {
    warning(sprintf("found %s profiles where pattern did not match and shallowest depth is greater than 0", notfound), call. = FALSE)
  }
  
  if(length(p) == 1 && simplify) {
    return(out[[depthz[2]]])
  }
  as.data.frame(out)
}

#' @rdname getSurfaceHorizonDepth
#' @export
getMineralSoilSurfaceDepth <-  function(p, hzdesgn = hzdesgnname(p, required = TRUE), pattern = "O", simplify = TRUE) {
  
  if (is.null(hzdesgn) || !hzdesgn %in% horizonNames(p)) {
    stop("Horizon designation column (", hzdesgn, ") does not exist.")
  }

  # assumes OSM is given O horizon designation;
  # TODO: add support for lab-sampled organic measurements
  #       keep organic horizons with andic soil properties
  return(getSurfaceHorizonDepth(p, hzdesgn = hzdesgn, pattern = pattern, simplify = simplify))
}

#' @rdname getSurfaceHorizonDepth
#' @export
getPlowLayerDepth <- function(p, hzdesgn = hzdesgnname(p, required = TRUE), pattern = "^Ap[^b]*", simplify = TRUE) {
  return(getSurfaceHorizonDepth(p, hzdesgn = hzdesgn, pattern = pattern, simplify = simplify))
}

# getSurfaceHorizonDepth / Mineral Soil Surface, Organic Soil Horizons

context("surface horizon thickness, mineral soil surface, organic soil horizon")

data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group
hzdesgnname(sp1) <- "name"

p <- sp1[1]
attr <- 'prop' # clay contents % 

q <- sp1[2]

test_that("getSurfaceHorizonDepth", {
  # all horizons containing A (n=4 from top)
  t1 <- getSurfaceHorizonDepth(p, pattern = 'A', hzdesgn = 'name')
  expect_equal(t1, horizons(p)[4, 'bottom'])
  
  # all horizons that start with A and have a number from _2_ to 9 (n=0)
  t4 <- getSurfaceHorizonDepth(p, pattern = '^A[2-9]', hzdesgn = 'name')
  expect_equal(t4, 0)
  
  # contiguous and non-contiguous matches below a non-match are ignored (n=1)
  p$mollic_color <- c("d", "l", "d", "l", "d", "d")
  t6 <- getSurfaceHorizonDepth(p, pattern = "d", hzdesgn = "mollic_color")
  expect_equal(t6, horizons(p)[1, 'bottom'])
})

test_that("getSurfaceHorizonDepth in multiple profiles", {
  expect_equal(as.numeric(profileApply(sp1, getSurfaceHorizonDepth, pattern = "A")),
               getSurfaceHorizonDepth(sp1, pattern = "A")$bottom)
})

test_that("getPlowLayerDepth()", {
  # matches first two horizons in fake Ap horizon data with "buried Ap"
  p$aphorizons <- c("Ap1","Ap2","AB", rep('C', nrow(p) - 4), "Apb")
  rez <- getPlowLayerDepth(p, hzdesgn = 'aphorizons')
  expect_equivalent(rez, 14)
})

test_that("getMineralSoilSurfaceDepth()", {
  # matches first 3 horizons in fake O horizon data
  p$ohorizons <- c("Oi1","Oi2","Oe", rep('C', nrow(p) - 4), "2C")
  rez <- getMineralSoilSurfaceDepth(p, hzdesgn='ohorizons')
  expect_equivalent(rez, 49)
  
  # matches first horizon with original horizon designations
  rez2 <- getMineralSoilSurfaceDepth(q, hzdesgn='name')
  expect_equivalent(rez2, 5)
})

## ----setup, echo=FALSE, results='hide', warning=FALSE---------------------------------------------
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = 96,
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)

suppressMessages(library(aqp, quietly = TRUE))

## ----echo = FALSE, fig.width=8, fig.height=4------------------------------------------------------
.m <- '10YR 6/6'

data("munsell.spectra.wide")
w <- munsell.spectra.wide[, 1]
s <- munsell.spectra.wide[, .m]

par(mar = c(4.5, 4.5, 2, 1), cex.axis = 0.75, lend = 2)
plot(w, s, xlab = 'Wavelength (nm)', ylab = 'Reflectance', type = 'b', main = .m, cex = 0.8, pch = 16, axes = FALSE)
rect(xleft = 400, ybottom = 0.35, xright = 450, ytop = 0.4, col = parseMunsell(.m), border = 1, lwd = 1)
axis(side = 1, at = seq(380, 730, by = 20))
axis(side = 2, las = 1)

## -------------------------------------------------------------------------------------------------
# Munsell -> hex color
parseMunsell('5PB 4/6')

# Munsell -> sRGB
parseMunsell('5PB 4/6',  return_triplets = TRUE)

# Munsell -> CIELAB
parseMunsell('5PB 4/6',  returnLAB = TRUE)

# hex color -> Munsell
col2Munsell('#476189FF')

# neutral color
parseMunsell('N 5/')

# non-standard notation
getClosestMunsellChip('3.3YR 4.4/6.1', convertColors = FALSE)

## ----fig.width=10, fig.height=4-------------------------------------------------------------------
# example moist soil colors from the Musick soil series
m <- c("7.5YR 2.5/1", "10YR 3/2", "7.5YR 4/3", "2.5YR 3/6", "2.5YR 3/6", 
       "2.5YR 4/6", "5YR 4/6", "7.5YR 5/4")

mm <- parseMunsell(m, convertColors = FALSE)

d.p <- estimateSoilColor(
  hue = mm$hue, 
  value = mm$value, 
  chroma = mm$chroma, 
  method = 'procrustes', 
  sourceMoistureState = 'moist'
)

d.ols <- estimateSoilColor(
  hue = mm$hue, 
  value = mm$value, 
  chroma = mm$chroma, 
  method = 'ols', 
  sourceMoistureState = 'moist'
)

d.p <- sprintf("%s %s/%s", d.p$hue, d.p$value, d.p$chroma)
d.ols <- sprintf("%s %s/%s", d.ols$hue, d.ols$value, d.ols$chroma)

colorContrastPlot(m, d.p, labels = c('Moist', 'Estimated\nDry'), d.cex = 0.9)

# it is the same
# colorContrastPlot(m, d.ols, labels = c('Moist', 'Estimated\nDry'), d.cex = 0.9)

## ----fig.width=6, fig.height=3--------------------------------------------------------------------
data("Ohz.colors")

Ohz.colors$col <- parseMunsell(Ohz.colors$L1.munsell)

op <- par(mfrow = c(2, 1), mar = c(0.5, 0.5, 1.5, 0))

with(
  Ohz.colors[Ohz.colors$state == 'dry', ],
  soilPalette(colors = col, lab = sprintf("%s\n%s", genhz, L1.munsell), lab.cex = 1)
)
title(main = 'Dry Colors')

with(
  Ohz.colors[Ohz.colors$state == 'moist', ],
  soilPalette(colors = col, lab = sprintf("%s\n%s", genhz, L1.munsell), lab.cex = 1)
)
title(main = 'Moist Colors')

# restore original base graphics state
par(op)

#' Split a `SoilProfileCollection` into a list based on types of horizon logic errors
#'
#' Uses `checkHzDepthLogic()` to identify presence of depth logic errors, same depths, missing depths, and overlaps/gaps between the horizons of each profile in a `SoilProfileCollection.`
#' 
#' @param object A `SoilProfileCollection`
#' @param interact Calculate interaction between the four logic errors for groups? Default: \code{FALSE} always returns 4 groups, one for each logic error type.
#' @param ... Additional arguments to \code{split.default}, called when \code{interact = TRUE}
#'
#' @return A named list of `SoilProfileCollections` (or \code{NULL}), with names: "depthLogic", "sameDepth", "missingDepth", "overlapOrGap". If \code{interact = TRUE} then the list elements groups determined by \code{interaction()} of the error types.
#' @export
#'
#' @examples
#' 
#' data(sp4)
#' depths(sp4) <- id ~ top + bottom
#' 
#' # no errors (all four list elements return NULL)
#' splitLogicErrors(sp4)
#' 
#' # NA in top depth triggers depth logic and missing depth errors
#' data(sp4)
#' sp4$top[1] <- NA
#' depths(sp4) <- id ~ top + bottom
#' 
#' splitLogicErrors(sp4)
#' 
#' # interact = TRUE gets errors for profile 1 in same group
#' #  and allows you to pass extra arguments to split.default()
#' splitLogicErrors(sp4, interact = TRUE, sep = "_", drop = TRUE)
#' 
splitLogicErrors <- function(object, interact = FALSE, ...) {
  
  # do check logic on logic for each profile
  f.logic <- checkHzDepthLogic(object)
  logicNames <- c("depthLogic", "sameDepth", "missingDepth", "overlapOrGap")
  names.idx <- match(logicNames, names(f.logic))
  
  # reformat errors
  f.errors <- do.call('rbind', apply(f.logic, 1, function(x) {
    as.data.frame(t(ifelse(as.logical(x[names.idx]), logicNames, "")))
  }))
  f.errors[] <- lapply(f.errors, factor)
  colnames(f.errors) <- logicNames
  
  if (interact == TRUE) {
    # interact == TRUE returns list elements based on interaction of f.errors
    f.goodbad <- split.default(object, f = f.errors, ...)
    f.goodbad <- lapply(f.goodbad, function(x) {
                                      if (length(x) == 0)
                                        return(NULL)
                                      return(x)
                                    })
  } else {
    # iterate over the four error types individually, ensuring no interaction
    f.goodbad <- lapply(f.errors, function(x) {
                                      res <- split(object, f = x)
                                      if (length(res) == 2)
                                        return(res[[2]])
                                      return(NULL)
                                    })
  }
  
  # return result as list of SPCs
  return(f.goodbad)
}

#' 
#' 
#' #' @param sub logical, return class subdivisions
#' #' @param flat logical, use flat coarse fragment classes
#' #' @param .upper_limit numeric, upper limit for particle size
#' #' @return vector of particle classes and associated diameters in mm
#' .particles.usda <- function(sub, flat, .upper_limit = 1e5, ...) {
#'   
#'   if(flat) {
#'     # flat, top-level classes
#'     .p <- c(
#'       clay = 0.002,
#'       silt = 0.05,
#'       sand = 2,
#'       channers = 150,
#'       flagstones = 380,
#'       stones = 600,
#'       boulders = .upper_limit
#'     )
#'   } else {
#'     if(sub) {
#'       # non-flat, subdivisions
#'       .p <- c(
#'         fine_clay = 0.0002,
#'         coarse_clay = 0.002,
#'         fine_silt = 0.02,
#'         coarse_silt = 0.05,
#'         very_fine_sand = 0.1,
#'         fine_sand = 0.25,
#'         medium_sand = 0.5,
#'         coarse_sand = 1,
#'         very_coarse_sand = 2,
#'         fine_gravel = 5,
#'         medium_gravel = 20,
#'         coarse_gravel = 76,
#'         cobbles = 250,
#'         stones = 600,
#'         boulders = .upper_limit
#'       )    
#'     } else {
#'       # non-flat, top-level classes
#'       .p <- c(
#'         clay = 0.002,
#'         silt = 0.05,
#'         sand = 2,
#'         gravel = 76,
#'         cobbles = 250,
#'         stones = 600,
#'         boulders = .upper_limit
#'       )
#'     }
#'   }
#'   
#'   return(.p)
#' }
#' 
#' 
#' 
#' #' @param sub logical, return class subdivisions
#' #' @param .upper_limit numeric, upper limit for particle size
#' #' @return vector of particle classes and associated diameters in mm
#' .particles.international <- function(sub, .upper_limit = 1e5, ...) {
#'   
#'   if(sub) {
#'     .p <- c(
#'       clay = 0.002,
#'       silt = 0.02,
#'       fine_sand = 0.2,
#'       coarse_sand = 2,
#'       gravel = 20,
#'       stones = .upper_limit
#'     )
#'   } else {
#'     .p <- c(
#'       clay = 0.002,
#'       silt = 0.02,
#'       sand = 2,
#'       gravel = 20,
#'       stones = .upper_limit
#'     )
#'   }
#'   
#'   return(.p)
#' }
#' 
#' 
#' 
#' 
#' #' @param sub logical, return class subdivisions
#' #' @param .upper_limit numeric, upper limit for particle size
#' #' @return vector of particle classes and associated diameters in mm
#' .particles.unified <- function(sub, .upper_limit = 1e5, ...) {
#'   
#'   if(sub) {
#'     .p <- c(
#'       silt_or_clay = 0.074,
#'       fine_sand = 0.42,
#'       medium_sand = 2,
#'       coarse_sand = 4.8,
#'       fine_gravel = 19,
#'       coarse_gravel = 76,
#'       cobbles = 300,
#'       boulders = .upper_limit
#'     )
#'   } else {
#'     .p <- c(
#'       silt_or_clay = 0.074,
#'       sand = 4.8,
#'       gravel = 76,
#'       cobbles = 300,
#'       boulders = .upper_limit
#'     )
#'   }
#'   
#'   return(.p)
#' }
#' 
#' 
#' 
#' #' @param sub logical, return class subdivisions
#' #' @param .upper_limit numeric, upper limit for particle size
#' #' @return vector of particle classes and associated diameters in mm
#' .particles.aashto <- function(sub, rounded, .upper_limit = 1e5, ...) {
#'   
#'   if(sub) {
#'     .p <- c(
#'       clay = 0.005,
#'       silt = 0.74,
#'       fine_sand = 0.42,
#'       coarse_sand = 2,
#'       fine_gravel = 9.5,
#'       medium_gravel = 25,
#'       coarse_gravel = 75,
#'       big = .upper_limit
#'     )
#'   } else {
#'     .p <- c(
#'       clay = 0.005,
#'       silt = 0.74,
#'       sand = 2,
#'       gravel = 75,
#'       big = .upper_limit
#'     )
#'   }
#'   
#'   # adjust largest particle size class based on roundness
#'   if(rounded) {
#'     .bigstuff <- 'broken rock'  
#'   } else {
#'     .bigstuff <- 'boulders'
#'   }
#'   
#'   names(.p)[length(.p)] <- 'bigstuff'
#'   
#'   return(.p)
#' }
#' 
#' 
#' 
#' 
#' #' @param sub logical, return class subdivisions
#' #' @param .upper_limit numeric, upper limit for particle size
#' #' @return vector of particle classes and associated diameters in mm
#' .particles.mod.wentworth <- function(.upper_limit = 1e5, ...) {
#'   
#'   if(sub) {
#'     .p <- c(
#'       clay = 0.004,
#'       silt = 0.062,
#'       sand = 2,
#'       pebbles = 64,
#'       cobbles = 256,
#'       boulders = .upper_limit
#'     )
#'   } else {
#'     .p <- c(
#'       clay = 0.004,
#'       silt = 0.062,
#'       sand = 2,
#'       pebbles = 64,
#'       cobbles = 256,
#'       boulders = .upper_limit
#'     )
#'   }
#'   
#'   
#'   return(.p)
#' }
#' 
#' particleClasses <- function(sys = c('usda', 'international', 'unified', 'aashto', 'mod.wentworth'), sub = TRUE, flat = FALSE, rounded = FALSE, .upper_limit = 1e5) {
#'   
#'   # normalize to lower case
#'   sys <- tolower(sys)
#'   
#'   # most frequent option, usually not specified
#'   if(missing(sys)) {
#'     sys <- 'usda'
#'   } else {
#'     sys <- match.arg(sys)
#'   }
#'   
#'   .pc <- switch(sys, 
#'                 'usda' = {
#'                   .particles.usda(sub = sub, flat = flat, .upper_limit = .upper_limit)
#'                 },
#'                 'international' = {
#'                   .particles.international(sub = sub, .upper_limit = .upper_limit)
#'                 },
#'                 'unified' = {
#'                   .particles.unified(sub = sub, .upper_limit = .upper_limit)
#'                 },
#'                 'aashto' = {
#'                   .particles.aashto(sub = sub, rounded = rounded, .upper_limit = .upper_limit)
#'                 },
#'                 'mod.wentworth' = {
#'                   .particles.mod.wentworth(sub = sub, .upper_limit = .upper_limit)
#'                 }
#'   )
#'   
#'   return(.pc)
#'   
#' }
#' 
#' 
#' 
#' 
#' 
#' particleSieve <- function() {
#'   
#' }
#' 
#' 
#' 
#' 
#' 
#' 
#' ## TODO:
#' ##  * NA handling
#' ##  * vectorization
#' ##  * distance-based matching -> is this sufficient
#' ##  * other systems
#' 
#' 
#' convertParticleDiameterUnits <- function(x, from, to) {
#'   
#'   ## enumerate possible from/to conversions
#'   .possibilities <- combn(
#'     x = c('mm', 'sieve', 'phi'),
#'     m = 2
#'   )
#'   
#'   # combine forward and reverse conversions
#'   .possibilities <- t(
#'     cbind(
#'       .possibilities, 
#'       apply(.possibilities, 2, rev)
#'     )
#'   )
#'   
#'   # simplify to data.frame
#'   .possibilities <- data.frame(.possibilities)
#'   names(.possibilities) <- c('from', 'to')
#'   
#'   # setup conversion rules
#'   .possibilities$rule <- paste(.possibilities$from, .possibilities$to, sep = '-')
#'   
#'   # select conversion rule
#'   .rule <- .possibilities$rule[.possibilities$from == from & .possibilities$to == to]
#'   
#'   
#'   ## US standard sieve sizes
#'   # .lut <- read.table(textConnection(readClipboard()))
#'   # names(.lut) <- c('mm', 'in', 'no')
#'   
#'   .us_sieve_size_lut <- structure(
#'     list(
#'       `mm` = c(11.2, 6.35, 5.6, 4.75, 4, 3.35, 2.8, 2.36, 
#'                2, 1.7, 1.4, 1.18, 1, 0.841, 0.71, 0.595, 0.5, 0.4, 0.355, 0.3, 
#'                0.25, 0.21, 0.177, 0.149, 0.125, 0.105, 0.088, 0.074, 0.063, 
#'                0.053, 0.044, 0.037, 0.025, 0.02), 
#'       `in` = c(0.438, 0.25, 0.223, 
#'                0.187, 0.157, 0.132, 0.11, 0.0937, 0.0787, 0.0661, 0.0555, 0.0469, 
#'                0.0394, 0.0331, 0.0278, 0.0232, 0.0197, 0.0165, 0.0139, 0.0117, 
#'                0.0098, 0.0083, 0.007, 0.0059, 0.0049, 0.0041, 0.0035, 0.0029, 
#'                0.0024, 0.0021, 0.0017, 0.0015, 0.001, 8e-04), 
#'       `no` = c("7/16", 
#'                "1/4", "3.5", "4", "5", "6", "7", "8", "10", "12", "14", "16", 
#'                "18", "20", "25", "30", "35", "40", "45", "50", "60", "70", "80", 
#'                "100", "120", "140", "170", "200", "230", "270", "325", "400", 
#'                "500", "632")), 
#'     class = "data.frame", 
#'     row.names = c(NA, -34L)
#'   )
#'   
#'   ## phi numbers, used with modified wentworth
#'   .phi_lut <- data.frame(
#'     phi = c(12:1, 0, -c(1:12)),
#'     mm = c(
#'       1 / 2^c(12:1),
#'       2^c(0:12)
#'     )
#'   )
#'   
#'   
#'   
#'   ## process rules
#'   if(.rule == 'mm-sieve') {
#'     .idx <- which.min(outer(x, .us_sieve_size_lut$mm, FUN = '-')^2)
#'     .us_sieve_size_lut$no[.idx]
#'     
#'   }
#'   
#'   if(.rule == 'sieve-mm') {
#'     
#'   }
#'   
#'   if(.rule == 'mm-phi') {
#'     
#'   }
#'   
#'   if(.rule == 'phi-mm') {
#'     
#'   }
#'   
#'   if(.rule == 'sieve-phi') {
#'     
#'   }
#'   
#'   if(.rule == 'phi-sieve') {
#'     
#'   }
#'   
#'   
#'   return(.res)
#'   
#' }
#' 
#' 
#' 
#' 

#' @title Format Munsell Notation from Hue, Value, and Chroma
#' 
#' @description Format hue, value, and chroma values into valid Munsell notation, or NA when not possible. The following rules are applied to the input:
#' 
#'   * for all hues other than N, there must be a valid hue, value, and chroma
#'   * N hues may specify chroma as 0, NA, or '' (empty string)
#'   * when `standardHues = TRUE`, hue must be a standard Munsell hue, see [huePosition()]
#' 
#' 
#'
#' @param hue character vector of Munsell hue
#' @param value character or numeric vector of Munsell value
#' @param chroma character or numeric vector of Munsell chroma
#' @param neutralConvention character, neutral color encoding convention
#' 
#' Neutral colors are encoded as:
#'   * 'empty': N 3/
#'   * 'zero': N 3/0
#' 
#' @param standardHues logical, when `TRUE` non-standard hues are converted to NA
#'
#' @author D.E. Beaudette
#'
#' @returns character vector of same length as `hue`, `value`, and `chroma`
#' @export
#' 
#' @seealso [launderMunsell()]
#'
#' @examples
#' 
#' d <- data.frame(
#' hue = c('10YR', NA, 'N', 'N', 'N', '5G', '5Z'),
#' value = c(4, 3, 2, 4, 3, NA, 4),
#' chroma = c(4, 3, NA, 0, NA_integer_, 6, 3)
#' )
#' 
#' formatMunsell(d$hue, d$value, d$chroma)
#' formatMunsell(d$hue, d$value, d$chroma, neutralConvention = 'empty')
#' formatMunsell(d$hue, d$value, d$chroma, standardHues = FALSE)
#' 
#' # all result in 'N 6/0' (neutralConvention = 'zero')
#' formatMunsell('N', 6, 0)
#' formatMunsell('N', 6, '')
#' formatMunsell('N', 6, NA)
#' 
formatMunsell <- function(hue, value, chroma, neutralConvention = c('zero', 'empty'), standardHues = TRUE) {
  
  neutralConvention <- match.arg(neutralConvention)
  
  # value and chroma must be safely convertable to numeric
  value <- as.numeric(value)
  chroma <- as.numeric(chroma)
  
  # chroma could be unspecified ('' or NA)
  
  # optionally enforce standard hues
  if(standardHues) {
    .stdhues <- huePosition(returnHues = TRUE, includeNeutral = TRUE)
    
    # this will also match NA
    .badhue <- which(! hue %in% .stdhues)
    
    # set to NA, these colors will not be formatted
    if(length(.badhue) > 0) {
      hue[.badhue] <- NA
      message('some colors have non-standard hue, result is NA')
    }
  }
  
  
  # neutral hues, may have NA in chroma
  .n <- which(hue == 'N')
  
  # normalize neutral chroma according to convention
  if(length(.n) > 0) {
    .replacement <- switch(neutralConvention, 'empty' = '', 'zero' = '0')
    
    chroma[.n] <- .replacement
  }
  
  
  # all other hues
  # NA in hue | value | chroma => formatting not possible
  .naidx <- which(is.na(hue) | is.na(value) | is.na(chroma))
  
  # compile standard notation
  .res <- sprintf("%s %s/%s", hue, value, chroma)
  
  # NA in hue | value => result is NA
  if(length(.naidx) > 0) {
    .res[.naidx] <- NA
    message('some colors missing hue or value, result is NA')
  }
  
  
  return(.res)
  
}

context("Munsell formatting")


test_that("formatMunsell() works as expected", {
  
  
  .res <- formatMunsell('10YR', 3, 4)
  expect_equal(.res, '10YR 3/4')
  
  # no automatic conversion to neutral, use launderMunsell() for that
  .res <- formatMunsell('10YR', 3, 0)
  expect_equal(.res, '10YR 3/0')
  
  .res <- formatMunsell('N', 3, 4, neutralConvention = 'zero')
  expect_equal(.res, 'N 3/0')
  
  # chroma can be an empty string
  .res <- formatMunsell('N', 3, '', neutralConvention = 'zero')
  expect_equal(.res, 'N 3/0')
  
  .res <- formatMunsell('N', 3, 4, neutralConvention = 'empty')
  expect_equal(.res, 'N 3/')
  
  .res <- formatMunsell('N', 3, NA, neutralConvention = 'empty')
  expect_equal(.res, 'N 3/')
  
  .res <- formatMunsell('N', 3, 0, neutralConvention = 'empty')
  expect_equal(.res, 'N 3/')
  
  .res <- formatMunsell('5YR', NA, 4)
  expect_equal(.res, NA_character_)
  
  # missing chroma, not neutral hue => NA
  .res <- formatMunsell('5YR', 3, NA)
  expect_equal(.res, NA_character_)
  
  .res <- formatMunsell(NA, 6, 4)
  expect_equal(.res, NA_character_)
  
  .res <- formatMunsell('5Z', 6, 4)
  expect_equal(.res, NA_character_)
  
})

test_that("launderMunsell() works as expected", {
  
  .res <- launderMunsell('10YR 3/4')
  expect_equal(.res, '10YR 3/4')
  
  .res <- launderMunsell('10YR 3/0')
  expect_equal(.res, 'N 3/0')
  
  # missing chroma, not neutral hue => NA
  .res <- launderMunsell('10YR 3/')
  expect_equal(.res, NA_character_)
  
  .res <- launderMunsell('N 8/')
  expect_equal(.res, 'N 8/0')
  
  .res <- launderMunsell('N 8/0', neutralConvention = 'empty')
  expect_equal(.res, 'N 8/')
  
})

# if (!isGeneric("harmonize"))
  setGeneric("harmonize", function(x, x.names, keep.cols = NULL, grp.name = "hgroup")
    standardGeneric("harmonize"))

#' @title Harmonize a property by profile-level denormalization for convenient visualization or analysis of ranges
#'
#' @description It is sometimes convenient to be able to "denormalize" to a `SoilProfileCollection` with fewer attributes but more profiles. This is helpful wherever calculations are made on a profile basis and ranges or repeated measures are depicted with multiple attributes per soil horizon.
#'
#'  \code{harmonize} is most commonly used for creating "comparison" soil profile sketches with \code{plotSPC}--where the thematic attribute is derived from multiple data sources or summary statistics (such as quantiles of a property for Low-RV-High). However, the method more generally applies wherever one wants to alias between multiple columns containing "similar" data as input to an algorithm.
#'
#' Data are "harmonized" to a common attribute names specified by the names of list elements in \code{x.names}. Profiles are essentially duplicated. In order to satisfy uniqueness constraints of the `SoilProfileCollection`, the label from the sub-elements of \code{x.names} are used to disambiguate profiles. A new column in the site table is calculated to reflect these groupings and facilitate filtering. See examples below.
#'
#' @param x A `SoilProfileCollection`.
#' @param x.names a named list of character vectors specifying target names, profile ID suffixes and source attribute names for harmonization
#' @param keep.cols a character vector of column names to keep unaltered from the horizon data
#' @param grp.name a character vector with column name to store grouping variable in site table (default: "hgroup")
#'
#' @return A (redundant) `SoilProfileCollection`, with one profile for each set of harmonizations specified by \code{x.names}.
#'
#' @details If attributes reflecting the same or similar property within a soil layer have different names (e.g. \code{socQ05}, \code{socQ50}, \code{socQ95}) it is sometimes inconvenient to work with them as multiple attributes within the same profile. These similar attributes may need to be analyzed together, or in sequence by profile, displayed using the same name or using a common scale. It is also useful to be able to alias different data sources that have the same attributes with different names.
#'
#' Each list element in \code{x.names} specifies a single "harmonization," which is comprised of one or more mappings from new to old. Each named "sub-element" of \code{x.names} specifies the name and attribute to use for updating the profile ID and site table of  the duplicated profiles.
#'
#' @author Andrew G. Brown
#' @aliases harmonize
#' @export
#' @examples
#'
#' ### single source "harmonization" of single-profile with range -> single attribute, multi-profile
#'
#' # make some test data
#' spc <- combine(lapply(1:10, random_profile, SPC = TRUE))
#'
#' # assume that p1, p2 and p3 are the low RV and high quantiles for a hypothetical property "foo"
#' h1 <- harmonize(spc, x.names = list(foo = c(q05 = "p1", q50 = "p2", q95 = "p3")))
#'
#' # inspect result
#' plotSPC(h1, color = "foo")
#'
#' # filter with calculated "harmonized group" to get just RV profiles
#' plotSPC(subset(h1, hgroup == "q50"), color="foo")
#'
#' ### single source, two properties at once; with common labels: "method1" "method2"
#'
#' # assume that p1, p2 are measurements by two (!=) methods for a hypothetical property "foo"
#' #             p3, p4 are measurements by same two methods for a hypothetical property "bar"
#' h3 <- harmonize(spc, x.names = list(foo = c(method1 = "p1", method2 = "p2"),
#'                                     bar = c(method1 = "p3", method2 = "p4")))
#' plotSPC(h3, color = "foo")
#' plotSPC(h3, color = "bar")
#' head(horizons(h3))
#'
#' # a slight modification, "method 1" onlyused for "foo" and "method 3" for "bar"
#' h3 <- harmonize(spc, x.names = list(foo = c(method1 = "p1", method2 = "p2"),
#'                                     bar = c(method2 = "p3", method3 = "p4")))
#' plotSPC(h3, color = "foo") # note the pattern of values missing for foo (*_method3)
#' plotSPC(h3, color = "bar") #  likewise for bar (*_method1)
#'
#' #' the new labels need not match across harmonizations -- not sure how useful this is but it works
#' h3 <- harmonize(spc, x.names = list(foo = c(method1 = "p1", method2 = "p2"),
#'                                     bar = c(method3 = "p3", method4 = "p4")))
#' plotSPC(h3, color = "foo") # note the pattern of values missing for foo (*_method 3 + 4)
#' plotSPC(h3, color = "bar") #  likewise for bar (*_method 1 + 2)
#'
#' ### two-source harmonization
#'
#' # make test data
#' spc1 <- combine(lapply(LETTERS[1:5], random_profile, SPC = TRUE))
#' spc2 <- combine(lapply(letters[1:5], random_profile, SPC = TRUE))
#'
#' h4 <- combine(list(harmonize(spc1, list(foo = c(transect1 = "p4"))),   # foo is p4 in dataset 1
#'                       harmonize(spc2, list(foo = c(transect2 = "p2")))))  # foo is p2 in dataset 2
#'
#' # same property with different name in two different datasets
#' plotSPC(h4, color = "foo")
#'
#' ### many source harmonization
#'
#' # make test datasets (n=10); highly redundant IDs (1:3 repeated)
#' spcs <- lapply(1:10, function(x) pbindlist(lapply(1:3, random_profile, SPC = TRUE)))
#'
#' # randomly varying column name for demo (in each dataset, foo could could be p1 thru p5)
#' rcolname <- paste0("p", round(runif(10, 0.5, 5.5)))
#'
#' # iterate over data sources
#' bigspc <- combine(lapply(1:length(spcs), function(i) {
#'
#'   # assume each data source has a unique name for the property "foo"
#'   xn <- rcolname[i]
#'
#'   # set names attribute to be equal to index i [creating unique profile IDs]
#'   #   i.e. 2_10 will be profile ID 2 from 10th dataset
#'   names(xn) <- i
#'
#'   # harmonize each data source, using unique column name and target name "foo"
#'   harmonize(spcs[[i]], x.names = list(foo = xn))
#' }))
#'
#' # inspect a subset
#' plotSPC(bigspc[1:30,], color = "foo")
#'
setMethod("harmonize", signature(x = "SoilProfileCollection"),
          function(x, x.names, keep.cols = NULL, grp.name = "hgroup") {

  if (is.null(names(x.names)) | is.null(x.names) | !length(x.names) | !is.list(x.names))
    stop("argument `x.names` must be a named list with each element specifying a harmonization", call. = FALSE)

  if (missing(keep.cols) | is.null(keep.cols))
    keep.cols <- character(0)

  if (!is.character(grp.name) | (!length(keep.cols) > 0 & any(!keep.cols %in% horizonNames(x))))
    stop("argument `keep.cols` must be a character vector of column names to keep [unaltered] from horizon data", call. = FALSE)

  if (length(grp.name) != 1 | !is.character(grp.name))
    stop("argument `grp.name` must specify a single column name for output groups to be stored in site table", call. = FALSE)

  hname <- names(x.names)

  # required horizon-level attributes
  idn <- idname(x)
  dep <- horizonDepths(x)

  # optional horizon.level attributes
  hzd <- hzdesgnname(x)
  hzt <- hztexclname(x)
  hzcol <- c(idn, dep, hzd, hzt)
  hzcol <- hzcol[hzcol != ""]

  x.sets <- unique(unlist(lapply(x.names, names)))
  h.sets <- lapply(lapply(x.names, names), function(hn) x.sets %in% hn)

  # nested loop -- iterating over all profiles as many unique levels we have in elements of x.names
   pbindlist(profileApply(x, function(p) {
      pbindlist(lapply(1:length(x.sets), function(i) {

          # get horizon data for single profile
          p.out <- p
          hz.data <- horizons(p.out)

          xn <- unlist(lapply(x.names, function(xnn) xnn[x.sets[i] == names(xnn)]))

          # subset
          new.hz.data <- .data.frame.j(hz.data, c(hzcol, xn, keep.cols, hzidname(x)), aqp_df_class(x))

          # rename to "harmonized" column name
          hname.sub <- unlist(lapply(h.sets, function(x) x[i]))
          colnames(new.hz.data) <- c(hzcol, hname[hname.sub], keep.cols, hzidname(x))

          # replace horizon data with subset of required + harmonized
          replaceHorizons(p.out) <- new.hz.data

          # input x.names is a named vector, use those names as suffix on profile ID
          profile_id(p.out) <- paste(profile_id(p), x.sets[i], sep = "_")

          # add to site table also (useful for filtering downstream)
          p.out[[grp.name]] <- x.sets[i]

          # return modified profile
          return(p.out)
        }))
  }))
})

#' @title Soil Profile Group Labels
#'
#' @description Labels groups of soil profiles within soil profile sketches.
#'
#' See examples below for ideas.
#'
#' @param x0 integer indices to the first profile within each group
#' @param x1 integer indices to the last profile within each group
#' @param labels vector of group labels
#' @param y0 baseline depth used for group brackets
#' @param y1 depth used for start and end markers for group brackets (see
#' examples)
#' @param label.offset vertical offset of group labels from baseline
#' @param label.cex label size
#' @note This function is typically called by some other convenience function
#' such as \code{\link{plotMultipleSPC}}.
#' @author D.E. Beaudette
#' @seealso \code{\link{plotMultipleSPC}}
#' @export
#' @examples
#'
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#'
#' # load sample data
#' data(sp3)
#' data(sp4)
#'
#' # convert soil colors
#' sp3$h <- NA ; sp3$s <- NA ; sp3$v <- NA
#' sp3.rgb <- with(sp3, munsell2rgb(hue, value, chroma, return_triplets=TRUE))
#' sp3[, c('h','s','v')] <- t(with(sp3.rgb, rgb2hsv(r, g, b, maxColorValue=1)))
#'
#' # promote to SoilProfileCollection
#' depths(sp3) <- id ~ top + bottom
#' depths(sp4) <- id ~ top + bottom
#'
#' # combine into a list
#' spc.list <- list(sp3, sp4)
#'
#' # compute group lengths and start/stop locations
#' n.groups <- length(spc.list)
#' spc.lengths <- sapply(spc.list, length)
#' n.pedons <- sum(spc.lengths)
#' group.starts <- c(1, 1 + cumsum(spc.lengths[-n.groups]))
#' group.ends <- cumsum(spc.lengths)
#'
#' # determine depths of first / last profile in each group
#' yy <- unlist(sapply(spc.list, function(i) profileApply(i, max)))
#' tick.heights <- yy[c(group.starts, group.ends)] + 2
#'
#' # plot 2 SoilProfileCollection objects on the same axis
#' par(mar=c(1,1,1,1))
#' plotSPC(sp3, n = n.pedons)
#' plotSPC(sp4, add = TRUE, x.idx.offset = group.ends[1], 
#' depth.axis = FALSE, id.style = 'side')
#' 
#' # annotate groups
#' profileGroupLabels(x0 = group.starts, x1 = group.ends,
#' labels=c('Collection 1', 'Collection 2'), y0=120, y1=tick.heights)
#'
profileGroupLabels <- function(x0, x1, labels, y0=100, y1=98, label.offset=2, label.cex=0.75) {
  
  # sanity check: start / stop / label lengths should be equal
  if(! all.equal(length(x0), length(x1), length(labels)) )
    stop('start positions, stop positions, and number of labels must be equal', call. = FALSE)
  
  # pre-compute some elements
  n.groups <- length(x0)
  label.centers <- (x0 + x1) / 2
  
  # add group base lines
  segments(x0=x0, x1=x1, y0=y0, y1=y0)
  # add arrows to first / last group members
  arrows(x0=c(x0, x1), x1=c(x0, x1), y0=c(y0, y0), y1=y1, length=0.1)
  
  # annotate with group names
  text(x=label.centers, y=y0 + label.offset, labels=labels, cex=label.cex)
}


## TODO: simple tests
## TODO: figure out intelligent recycling of arguments
## TODO: no mechanism for merged legends
## TODO: this doesn't take into account non-default figure geometry

## TODO: can we integrate .interpretHorizonColor() vs. overly-simplistic .mapColor()?

#' @title Plot Multiple `SoilProfileCollection` Objects
#'
#' @details Combine multiple `SoilProfileCollection` objects into a single profile sketch,
#' with annotated groups.
#'
#' See examples below for usage.
#'
#' @param spc.list a list of \code{SoilProfileCollection} objects
#' 
#' @param group.labels a vector of group labels, one for each
#' \code{SoilProfileCollection} object
#' 
#' @param args a list of arguments passed to \code{plotSPC}, one for each
#' \code{SoilProfileCollection} object
#' 
#' @param merged.legend name of a horizon level attribute from which to create thematic sketches and merged legend
#' 
#' @param merged.colors vector of colors used to create thematic sketches from a shared horizon level attribute
#' 
#' @param merged.legend.title legend title
#' 
#' @param arrow.offset vertical offset in depth from base of start / end
#' profiles and group bracket arrows
#' 
#' @param bracket.base.depth baseline depth used for group brackets
#' 
#' @param label.offset vertical offset of group labels from baseline
#' 
#' @param label.cex label size
#' 
#' @param \dots additional arguments to the first call to \code{plotSPC}
#' 
#' @note For thematic sketches, use the `merged.legend` argument instead of `color` argument to `plotSPC`
#' 
#' @author D.E. Beaudette and Ben Marshall
#' 
#' @seealso \code{\link{profileGroupLabels}}
#' 
#' @keywords hplots
#' @export
#' @examplesIf requireNamespace("scales", quietly = TRUE)
#'
#' ##
#' ## Simple Example
#' ##
#' 
#' # using default arguments to plotSPC()
#' 
#' # load sample data
#' data(sp3)
#' data(sp4)
#' 
#' # promote to SoilProfileCollection
#' depths(sp3) <- id ~ top + bottom
#' depths(sp4) <- id ~ top + bottom
#' 
#' # combine into a list
#' spc.list <- list(sp3, sp4)
#' 
#' # argument list
#' arg.list <- list(
#'   list(name='name', id.style='top'),
#'   list(name='name', id.style='side')
#' )
#' 
#' # plot multiple SPC objects, 
#' # with list of named arguments for each call to plotSPC
#' par(mar=c(1,1,3,3))
#' plotMultipleSPC(
#'   spc.list, 
#'   group.labels = c('Collection 1', 'Collection 2'),
#'   args = arg.list, 
#'   bracket.base.depth = 120, label.cex = 1
#' )
#' 
#' # specify a different max.depth
#' plotMultipleSPC(
#'   spc.list, 
#'   group.labels = c('Collection 1', 'Collection 2'),
#'   args = arg.list, 
#'   bracket.base.depth = 120, label.cex = 1,
#'   max.depth = 250
#' )
#' 
#' 
#' 
#' ##
#' ## Merged Legend Example
#' ##
#' 
#' # merged legend based on hz attribute 'clay'
#' 
#' # reset sample data
#' data(sp3)
#' data(sp4)
#' 
#' # promote to SoilProfileCollection
#' depths(sp3) <- id ~ top + bottom
#' depths(sp4) <- id ~ top + bottom
#' 
#' # combine into a list
#' spc.list <- list(sp3, sp4)
#' 
#' # argument list
#' arg.list <- list(
#'   list(name='name', id.style='top'),
#'   list(name='name', id.style='side')
#' )
#' 
#' 
#' par(mar=c(1,1,3,3))
#' plotMultipleSPC(
#'   spc.list, 
#'   group.labels = c('Collection 1', 'Collection 2'),
#'   args = arg.list, 
#'   label.cex = 1, 
#'   merged.legend = 'clay', merged.legend.title = 'Clay (%)'
#' )
#' 
#' 
#' ##
#' ## Complex Merged Legend Example
#' ##
#' 
#' # create a merged legend from "clay" in sp4 and jacobs2000
#' # use "soil_color" from sp3
#' 
#' # reset sample data
#' data(sp3)
#' data(sp4)
#' data(jacobs2000)
#' 
#' # promote to SoilProfileCollection
#' depths(sp3) <- id ~ top + bottom
#' depths(sp4) <- id ~ top + bottom
#' 
#' # remove 'clay' column from sp3
#' sp3$clay <- NULL
#' 
#' # combine into a list
#' spc.list <- list(sp3, sp4, jacobs2000)
#' 
#' # try some variations on the default arguments
#' # `clay` is missing in the first SPC, safe to specify another column for colors
#' arg.list <- list(
#'   list(color = 'soil_color', id.style='top', name = NA, width = 0.3, hz.depths = TRUE),
#'   list(name='name', id.style='side', name.style = 'center-center'),
#'   list(name='name', id.style='side', name.style = 'left-center', hz.depths = TRUE)
#' )
#' 
#' par(mar=c(1,1,3,3))
#' plotMultipleSPC(
#'   spc.list, 
#'   group.labels = c('sp3', 'sp4', 'jacobs2000'),
#'   label.offset = 3,
#'   args = arg.list, 
#'   merged.legend = 'clay', merged.legend.title = 'Clay (%)',
#'   depth.axis = list(line = 0)
#' )
plotMultipleSPC <- function(spc.list, group.labels, args = rep(list(NA), times = length(spc.list)), merged.legend = NULL, merged.colors = c("#5E4FA2", "#3288BD", "#66C2A5","#ABDDA4", "#E6F598", "#FEE08B","#FDAE61", "#F46D43", "#D53E4F","#9E0142"), merged.legend.title = merged.legend, arrow.offset = 2, bracket.base.depth = 95, label.offset = 2, label.cex = 0.75, ...) {
  
  # compute group stats
  n.groups <- length(spc.list)
  spc.lengths <- sapply(spc.list, length)
  n.pedons <- sum(spc.lengths)
  group.starts <- c(1, 1 + cumsum(spc.lengths[-n.groups]))
  group.ends <- cumsum(spc.lengths)
  
  # get depths + offset to start / end profiles
  yy <- unlist(sapply(spc.list, function(i) profileApply(i, max)))
  tick.heights <- yy[c(group.starts, group.ends)] + arrow.offset
  
  
  # unique set of arguments specified in args and ...
  unique.args <- unique(
    c(
      names(unlist(args)),
      names(list(...))
    )
  )
  
  # estimate a reasonable max depth (over all SPCs)
  # but only when not specified in any arguments
  if(! 'max.depth' %in% unique.args){
    # max over collections
    max.depth <- max(sapply(spc.list, max), na.rm = TRUE)
    
    # note: adding an extra 5% of max.depth for labels
    max.depth + (max.depth / 5)
    
    # insert into first set of arguments to plotSPC
    args[[1]]$max.depth <- max.depth
  }
  
  
  # extend base depth if not supplied
  if(missing(bracket.base.depth)) {
    bracket.base.depth <- max(sapply(spc.list, max), na.rm = TRUE) + 10
  }
  
  # optionally create a merged set of thematic colors and legend
  if(! is.null(merged.legend)) {
    # color ramp function
    cr <- colorRamp(merged.colors, space = 'Lab', interpolate = 'spline')
    
    ## TODO: .interpretHorizonColor() is much more intelligent, consider using it
    
    # NA-padded value -> color mapping for full range of some horizon attribute
    .mapColor <- function(x, r, col.ramp) {
      # rescale from full range {r} -> {0,1}
      
      # dang it, have to use scales::rescale for this
      # how can we adapt aqp:::.rescaleRange?
      c.rgb <- cr(scales::rescale(x, from = r, to = c(0,1)))
      
      cc <- which(complete.cases(c.rgb))
      cols <- rep(NA, times = nrow(c.rgb))
      cols[cc] <- rgb(c.rgb[cc, ], maxColorValue=255)
      return(cols)
    }
    
    # collect values over list of SPCs
    combined.data <- na.omit(
      unlist(
        lapply(spc.list, function(i) i[[merged.legend]])
      )
    )
    # get the full range
    combined.range <- range(combined.data, na.rm = TRUE)
    
    
    # iterate over list of profiles and arguments
    for(i in 1:length(spc.list)) {
      
      # current SPC
      spc_i <- spc.list[[i]]
      arg_i <- args[[i]]
      
      # map colors if column is present
      if(!is.null(spc_i[[merged.legend]])) {
        
        # convert non-NA values into colors
        horizons(spc_i)[['.color']] <- .mapColor(spc_i[[merged.legend]], combined.range, cr)
        
        # add arguments
        
        # thematic flag
        arg_i$color = '.color'
        # suppress legend
        arg_i$show.legend = FALSE
        
        # modify in place
        spc.list[[i]] <- spc_i
        args[[i]] <- arg_i
        
      } else {
        # do nothing
        
      }
      
    } # done iteration over lists of SPCs and arguments
    
    
    ## TODO: this will not work with categorical variables
    ##       -> abstract code from plotSPC into more general purpose functions
    
    # generate combined range / colors for legend
    pretty.vals <- pretty(combined.data, n = 8)
    
    # create legend object
    legend.data <- list(
      legend = pretty.vals, 
      col = rgb(
        cr(
          # rescale to {0,1}
          .rescaleRange(pretty.vals, x0 = 0, x1 = 1)
        ), 
        maxColorValue=255)
    )
    
    
  }
  
  # setup plot with first SPC in list
  do.call(
    what = plotSPC, 
    args = c(
      x = spc.list[[1]], 
      n = n.pedons, 
      na.omit(args[[1]]),
      ...)
  )
  
  # iterate over remaining SPC objs
  if(n.groups > 1) {
    for(i in 2:n.groups) {
      this.obj <- spc.list[[i]]
      this.args <- na.omit(args[[i]])
      suppressMessages(
        do.call(
          what = plotSPC, 
          args = c(
            x = this.obj, 
            x.idx.offset = group.ends[i-1], 
            add = TRUE, 
            depth.axis = FALSE, 
            this.args
          )
        )
      )
    }
  }
  
  # annotate groups with brackets
  profileGroupLabels(
    x0 = group.starts, 
    x1 = group.ends, 
    labels = group.labels, 
    y0 = bracket.base.depth, 
    y1 = tick.heights, 
    label.offset = label.offset, 
    label.cex = label.cex
  )
  
  # add merged legend
  if(! is.null(merged.legend)) {
    mtext(side=3, text = merged.legend.title, font=2, line=1.6)
    legend('bottom', legend=legend.data$legend, col=legend.data$col, bty='n', pch=15, horiz=TRUE, xpd=TRUE, inset=c(0, 0.99))   
  }
  
}

# default slab function for categorical variables
# returns a named vector of results
# this type of function is compatible with aggregate()
.slab.fun.factor.default <- function(values, cpm, ...) {

  if (cpm == 1) {
	  # probabilities are relative to number of contributing profiles
    tb <- table(values, useNA = 'no')
		# convert to proportions
		pt <- prop.table(tb)
	} else if (cpm == 2) {
	  # probabilities are relative to total number of profiles
	  tb <- table(values, useNA = 'always')
		# convert to proportions,
		# the last column will be named 'NA', and contains the tally of NAs --> remove it
		pt <- prop.table(tb)
		pt <- pt[-length(pt)]
	}

  ## NOTE: this will corrupt hz designations associated with lithologic discontinuities
  ## ---> 2Bt will become X2Bt
	# assign safe names to the vector of probabilities
	names(pt) <- make.names(levels(values))

	return(pt)
}

.slab.fun.factor.weighted <- function(values, w, na.rm = TRUE, ...) {
  # TODO
}
  
# default slab function for continuous variables
# returns a named vector of results
# this type of function is compatible with aggregate()
.slab.fun.numeric.default <- function(values, probs = c(0.05, 0.25, 0.5, 0.75, 0.95), na.rm = TRUE, ...) {
  .slab.fun.numeric.fast(values, probs = probs, na.rm = na.rm, ...)
}

# for a site or horizon level weight column, use Hmisc::wtd.quantile
#   note that "frequency weights" are assumed to be most common use case, so normwt=FALSE by default, 
#   normwt=TRUE can be passed as optional argument for slab.fun from slab() interface
.slab.fun.numeric.weighted <- function(values, w, probs = c(0.05, 0.25, 0.5, 0.75, 0.95), normwt = FALSE, na.rm = TRUE, ...) {
  if (!requireNamespace('Hmisc', quietly = TRUE))
    stop('please install the `Hmisc` package to use `wtd.quantile()` method', call. = FALSE)
  res <- try(Hmisc::wtd.quantile(values, w, probs = probs, na.rm = na.rm, normwt = normwt), silent = TRUE)
  if (inherits(res, 'try-error') && grepl("zero non-NA points", res[1])) {
    res <- rep(NA_real_, length(probs))
  } else {
    names(res) <- paste('p.q', round(probs * 100), sep = '')
  }
  return(res)
}

# easy specification of Hmisc::hdquantile if available
.slab.fun.numeric.HD <- function(values, probs = c(0.05, 0.25, 0.5, 0.75, 0.95), na.rm = TRUE, ...) {
  # sanity check, need this for color distance eval
  if (!requireNamespace('Hmisc', quietly = TRUE))
    stop('please install the `Hmisc` package to use `hdquantile()` method', call. = FALSE)
  
  res <- as.numeric(Hmisc::hdquantile(values, probs = probs, na.rm = na.rm))
  
  names(res) <- paste('p.q', round(probs * 100), sep = '')
  return(res)
}

# basic quantile evaluation, better for large data sets
.slab.fun.numeric.fast <- function(values, probs = c(0.05, 0.25, 0.5, 0.75, 0.95), na.rm = TRUE, ...) {
  res <- quantile(values, probs = probs, na.rm = na.rm)
  names(res) <- paste('p.q', round(probs * 100), sep = '')
  return(res)
}

# SoilProfileCollection method
# object: SoilProfileCollection
# fm: formula defining aggregation
# slab.structure: either regular segment interval, or user-defined segment boundaries {starting from 0, or length of 2}
# slab.fun: aggregate function applied to data chunks (must return a single row / chunk)
# cpm: class probability normalization mode
# weights: character vector naming column in site slot containing weights
# ... : extra arguments passed on to slab.fun
.slab <- function(object,
                  fm,
                  slab.structure = 1,
                  strict = FALSE,
                  byhz = TRUE,
                  slab.fun = slab_function(method = "numeric"),
                  cpm = 1,
                  weights = NULL,
                  ...) {

  # ***********
  # setup -----
  # ***********
  
  # define keywords for data.table
  .SD <- NULL; .N <- NULL; value <- NULL

	# get unique list of names in object
	object.names <- unique(unlist(names(object)))

	# get horizon depth column names
	hzd <- horizonDepths(object)
	
	# get name of ID column in original object for later
	object.ID <- idname(object)

  # if we are aggregating a single categorical variable, we need to keep track of the original levels
	original.levels <- NULL

	# extract components of the formula:
	g <- all.vars(update(fm, . ~ 0)) # left-hand side
	vars <- all.vars(update(fm, 0 ~ .)) # right-hand side
	
	
	# ******************
	# sanity checks ----
	# ******************
	
	# formula
	if (!inherits(fm, "formula")) {
	  stop('must provide a valid formula: groups ~ var1 + var2 + ...', call. = FALSE)
	}
	  
	# check for bogus left/right side problems with the formula
	# bogus grouping column
	if (length(g) > 0 && !any(g %in% object.names) && g != '.') {
	  stop('group name either missing from formula, or does match any columns in data.frame', call. = FALSE)
	}
	  
	# bogus column names in right-hand side
	if (any(vars %in% object.names) == FALSE) {
	  stop('column names in formula do not match column names in data.frame', call. = FALSE)
	}
	
	# non-integer depths in slab.structure
	if(any(round(slab.structure) != slab.structure)) {
	  stop('slab.structure may only contain integer depths', call. = FALSE)
	}
	
	
	# change / reset options ----
	
	# prevent very large integers from using scientific notation
	.op <- options(scipen = 10)
	
	# safely return to previous state
	on.exit(options(.op))
	
	
	# ******************************
	# interpret slab.structure ----
	# ******************************
	
	# make formula for slicing
	if (length(slab.structure) == 2) {
	  
	  # ensure that top < bottom
	  if(slab.structure[1] >= slab.structure[2]) {
	    stop('invalid slab structure', call. = FALSE)
	  }
	  
  	# user-defined single slab
	  fm.slice <- formula(paste(slab.structure[1], ':', slab.structure[2], ' ~ ', paste(vars, collapse = ' + '), sep = ''))
	} else {
	  fm.slice <- formula(paste('0:', max(object), ' ~ ', paste(vars, collapse = ' + '), sep = ''))
	}
	
	# slice into 1cm increments, result is a data.frame
	# NOTE: when byhz = TRUE, overlapping horizons cannot be detected
	data <- dice(x = object, fm = fm.slice, strict = strict, byhz = byhz, SPC = FALSE, pctMissing = TRUE)
  
	# Note: in this case we need to subtract the extra slice included by slice()/dice()
  # do it to sliced result so that the genSlabLabels have the correct length
  ldx <- data[[horizonDepths(object)[2]]] > slab.structure[2]
	if (length(slab.structure) == 2 && any(ldx, na.rm = TRUE)) {
	  data <- data[which(!ldx), ]  
	}
   
	# extract site data
	site.data <- site(object)

	# check groups for NA, if grouping variable was given
	if (g != '.' && any(is.na(site.data[, g]))) {
		stop('grouping variable must not contain NA', call. = FALSE)
	}
	
	# merge site data back into the result, this would include site-level weights
	data <- merge(x = data, y = site.data, by = object.ID, all.x = TRUE, sort = FALSE)

	# check variable classes
	if (length(vars) > 1) {
	  vars.numeric.test <- sapply(data[, vars], is.numeric)
	}	else {
	  vars.numeric.test <- is.numeric(data[[vars]])
	}

	# # sanity check: all numeric, or single character/factor
	if (any(!vars.numeric.test) & length(vars) > 1) {
	  stop('mixed variable types and multiple categorical variables are not currently supported in the same call to slab', call. = FALSE)
	}

	# check for single categorical variable, and convert to factor
	if (length(vars) == 1 & inherits(data[, vars], c('character', 'factor'))) {

		# if we have a character, then convert to factor
		if (inherits(data[[vars]],'character')) {
			message('Note: converting categorical variable to factor.')
			data[[vars]] <- factor(data[[vars]])
		}

		# check for weights
		if (!missing(weights)) {
		  slab.fun <- .slab.fun.factor.weighted
		} else {
  		# re-set default function, currently no user-supply-able option
  		slab.fun <- .slab.fun.factor.default
		}

    # save factor levels for later
    original.levels <- levels(data[[vars]])

    # set a flag for post data.table.melt -> factor level setting
    .factorFlag <- TRUE
	} else {
	  .factorFlag <- FALSE
	}

	# add segmenting label to data
 	data$seg.label <- .genSlabLabels2(object, data, slab.structure = slab.structure)

	# if there is no left-hand component in the formula, we are aggregating all data in the collection
	if (g == '.') {
		g <- 'all.profiles' # add new grouping variable to horizons
		data[, g] <- 1
	}
 	
 	if (length(weights) > 1) {
 	  stop("`weights` argument should be a character vector of length one containing (site-level) weight column name", call. = FALSE)
 	}
 	
 	if (!is.null(weights) && !weights %in% siteNames(object)) {
 	  stop("column '", weights, "' not present in site data", call. = FALSE)
 	}
 	
 	# convert wide -> long format; warnings will occur when not all columns are e.g. double
 	d.long <- suppressWarnings(data.table::melt(
 	  data.table::as.data.table(data[which(!is.na(data$seg.label)), ]),
 	  id.vars = unique(c(object.ID, 'seg.label', g, weights)),
 	  measure.vars = vars
 	))
 	
	# check for weights
	if (!missing(weights)) {
	  if (missing(slab.fun)) {
	    # default _weighted_ aggregation fun is .slab.fun.numeric.weighted
	    FUN <- .slab.fun.numeric.weighted
	    # custom weighted aggregation fun should take first argument as values 
	    # second argument is "weights" argument
	  } else {
	    FUN <- slab.fun
	  }
	} else {
	  if (missing(slab.fun) || !inherits(slab.fun, 'function')) {
	    # default numeric aggregation fun is .slab.fun.numeric.default
	    FUN <- .slab.fun.numeric.default
	  } else {
	    FUN <- slab.fun
	  }
	}
 	
 	.internal_wt <- eval(weights)
 	
 	# calculate summary function FUN for each variable*groups*slablabel
 	#  FUN returns a (named) vector of summary statistics, which are converted to "wide" data.frame
 	
 	if (.factorFlag) {
 	  d.long[['value']] <- factor(d.long[['value']], levels = original.levels)
 	}
 	
 	if (!missing(weights)) {
   	d.slabbed <- as.data.frame(d.long[, as.data.frame(t(FUN(value, .SD[[.internal_wt]], ...))),
   	                                  by = c('variable', g, 'seg.label')])
 	} else {
 	  if (!missing(cpm)) {
 	    d.slabbed <- as.data.frame(d.long[, as.data.frame(t(unclass(FUN(value, cpm = cpm, ...)))),
 	                                      by = c('variable', g, 'seg.label')])
 	  } else {
   	  d.slabbed <- as.data.frame(d.long[, as.data.frame(t(FUN(value, ...))),
   	                                    by = c('variable', g, 'seg.label')])
 	  }
 	}
 	
 	# calculate contributing fraction for each variable*groups*slablabel
 	d.slabbed$contributing_fraction <- d.long[, sum(!is.na(.SD[["value"]])) / .N,
 	                                          by = c('variable', g, 'seg.label')]$V1
 	
	# ensure that the names returned from slab.fun are legal
	names(d.slabbed) <- make.names(names(d.slabbed)) 
	  
 	# rename default data.table value column names
 	unn <- grepl("^V\\d+$", colnames(d.slabbed))
 	if (any(unn)) {
 	  colnames(d.slabbed)[unn] <- paste0("value", ifelse(sum(unn) == 1, "", 1:sum(unn)))
 	}
 	
 	# convert the slab.label column into top/bottom as integers
 	slab.depths <- strsplit(as.character(d.slabbed$seg.label), '-')
 	d.slabbed$top <- as.integer(lapply(slab.depths, function(i) i[1]))
 	d.slabbed$bottom <- as.integer(lapply(slab.depths, function(i) i[2]))
 	
	# remove seg.label from result
	d.slabbed$seg.label <- NULL

  # if categorical data have been aggregated, set an attribute with the original levels
  # this allows one to recover values that are not legal column names
  # e.g. 2Bt is corrupted to X2Bt
  if (!is.null(original.levels)) {
    attr(d.slabbed, 'original.levels') <- original.levels
  }
	
	d.slabbed
}

# setup generic function
setGeneric("slab", function(object,
                              fm,
                              slab.structure = 1,
                              strict = FALSE,
                              byhz = TRUE,
                              slab.fun = slab_function(method = "numeric"),
                              cpm = 1,
                              weights = NULL,
                              ...)
    standardGeneric("slab"))

#' @title Slab-Wise Aggregation of SoilProfileCollection Objects
#'
#' @description
#' Aggregate soil properties along user-defined `slabs`, and optionally within
#' groups.
#'
#' Multiple continuous variables OR a single categorical (factor) variable can
#' be aggregated within a call to `slab()`. Basic error checking is
#' performed to make sure that top and bottom horizon boundaries make sense.
#' User-defined aggregate functions (`slab.fun`) should return a named
#' vector of results. A new, named column will appear in the results of
#' `slab()` for every named element of a vector returned by `slab.fun`.
#' See examples below for a simple example of a slab function that computes
#' mean, mean +/- 1 standard deviation. The default slab function wraps
#' [stats::quantile()]. Note that if `group` is a factor it must not contain NAs.
#' 
#' `slab()` uses `dice()` to "resample" profiles to 1cm slices from depth 0 to `max(x)` (or `slab.structure[2]`, if defined).
#'
#' Sometimes `slab()` is used to conveniently re-arrange data vs. aggregate.
#' This is performed by specifying `identity` in `slab.fun`. See
#' examples below for a demonstration of this functionality.
#'
#' The default \code{slab.fun} was changed 2019-10-30 from a wrapper around
#' [Hmisc::hdquantile()] to a wrapper around [stats::quantile()]. See
#' examples below for a simple way to switch to the HD quantile estimator.
#'
#' Execution time scales linearly (slower) with the total number of profiles in
#' `object`, and exponentially (faster) as the number of profiles / group
#' is increased. `slab()` and [dice()] are much faster and require less
#' memory if input data are either numeric or character.
#'
#' There are several possible ways to define slabs, using `slab.structure`:
#'
#'  * a single integer, e.g. 10: data are aggregated over a regular sequence of 10-unit thickness slabs
#'  * a vector of 2 integers, e.g. c(50, 60): data are aggregated over depths spanning 50-60 depth units
#'  * a vector of 3 or more integers, e.g. c(0, 5, 10, 50, 100): data are aggregated over the depths spanning 0-5, 5-10, 10-50, 50-100 depth units
#'  
#' @name slab
#' @aliases slab slab,SoilProfileCollection-method
#' @docType methods
#' 
#' @param object a SoilProfileCollection
#' 
#' @param fm A formula: either `groups ~ var1 + var2 + var3` where named
#' variables are aggregated within `groups' OR where named variables are
#' aggregated across the entire collection ` ~ var1 + var2 + var3`. If `groups`
#' is a factor it must not contain `NA`
#' 
#' @param slab.structure integer vector: user-defined slab structure. See Details.
#' 
#' @param strict logical: should horizons be strictly checked for self-consistency?
#' 
#' @param byhz logical: should horizons or whole profiles be removed by logic checks in `strict`? Default `TRUE` removes only offending horizons, `FALSE` removes whole profiles with one or more illogical horizons.
#' 
#' @param slab.fun Function used to process each 'slab' of data, ideally returning a vector with names attribute. Defaults to a wrapper function around [stats::quantile()]. See details.
#' 
#' @param cpm Strategy for normalizing slice-wise probabilities, dividing by either: 

#'   * number of profiles with data at the current slice (`cpm = 1`), or 
#'   * by the number of profiles in the collection (`cpm = 2`). 
#'   
#' Mode 1 values will always sum to the contributing fraction, while mode 2 values will always sum to 1.
#' 
#' @param weights Column name containing site-level weights
#' 
#' @param \dots further arguments passed to `slab.fun`
#' 
#' @return Output is returned in long format, such that slice-wise aggregates are returned once for each combination of grouping level (optional), variable described in the `fm` argument, and depth-wise 'slab'.
#'
#' Aggregation of numeric variables, using the default slab function:
#' 
#'   * variable: The names of variables included in `fm`
#'   * groupname: The name of the grouping variable when provided, otherwise a fake grouping variable named 'all.profiles'.
#'   * p.q5: The slice-wise 5th percentile.
#'   * p.q25: The slice-wise 25th percentile
#'   * p.q50: The slice-wise 50th percentile (median)
#'   * p.q75: The slice-wise 75th percentile
#'   * p.q95: The slice-wise 95th percentile
#'   * top: The slab top boundary
#'   * bottom The slab bottom boundary.
#'   * contributing_fraction: The fraction of profiles contributing to the aggregate value, ranges from 1/n_profiles to 1
#'
#' When a single factor variable is used, slice-wise probabilities for each
#' level of that factor are returned as: 
#' 
#'   * variable: The names of factor variable included in `fm`
#'   * groupname: The name of the grouping variable when provided, otherwise a fake grouping variable named 'all.profiles'
#'   * A: The slice-wise probability of level A
#'   * B: The slice-wise probability of level B 
#'   * n: The slice-wise probability of level n
#'   * top: The slab top boundary.
#'   * bottom: The slab bottom boundary
#'   * contributing_fraction: The fraction of profiles contributing to the aggregate value, ranges from 1/n_profiles to 1

#' 
#' @note Arguments to `slab()` have changed with aqp 1.5 (2012-12-29) as part of a code clean-up and optimization. Calculation of
#' weighted-summaries was broken in `aqp 1.2-6 (2012-06-26), and removed as of aqp 1.5 (2012-12-29). `slab()`` replaced the previously defined `soil.slot.multiple` function as of aqp 0.98-8.58 (2011-12-21).
#' 
#' @section Methods: \describe{ \item{data = "SoilProfileCollection"}{Typical
#' usage, where input is a \code{\link{SoilProfileCollection}}.} }
#' 
#' @author D.E. Beaudette
#' 
#' @seealso [dice()]
#' 
#' @references D.E. Beaudette, P. Roudier, A.T. O'Geen, Algorithms for
#' quantitative pedology: A toolkit for soil scientists, Computers &
#' Geosciences, Volume 52, March 2013, Pages 258-268,
#' 10.1016/j.cageo.2012.10.020.
#'
#' Harrell FE, Davis CE (1982): A new distribution-free quantile estimator.
#' Biometrika 69:635-640.
#' 
#' @keywords methods manip
#' @export
#' @examples
#'
#' ##
#' ## basic examples
#' ##
#' library(lattice)
#' library(grid)
#' library(data.table)
#'
#' # load sample data, upgrade to SoilProfileCollection
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#'
#' hzdesgnname(sp1) <- "name"
#'
#' # aggregate entire collection with two different segment sizes
#' a <- slab(sp1, fm = ~ prop)
#' b <- slab(sp1, fm = ~ prop, slab.structure=5)
#'
#' # check output
#' str(a)
#'
#' # stack into long format
#' ab <- make.groups(a, b)
#' ab$which <- factor(ab$which, levels=c('a','b'),
#' labels=c('1-cm Interval', '5-cm Interval'))
#'
#' # plot median and IQR
#' # custom plotting function for uncertainty viz.
#' xyplot(top ~ p.q50 | which, data=ab, ylab='Depth',
#' 			 xlab='median bounded by 25th and 75th percentiles',
#' 			 lower=ab$p.q25, upper=ab$p.q75, ylim=c(250,-5),
#' 			 panel=panel.depth_function,
#' 			 prepanel=prepanel.depth_function,
#' 			 cf=ab$contributing_fraction,
#' 			 alpha=0.5,
#' 			 layout=c(2,1), scales=list(x=list(alternating=1))
#' 			 )
#'
#'
#' ###
#' ### re-arrange data / no aggregation
#' ###
#'
#' # load sample data, upgrade to SoilProfileCollection
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#'
#' # arrange data by ID
#' a <- slab(sp1, fm = id ~ prop, slab.fun=identity)
#'
#' # convert id to a factor for plotting
#' a$id <- factor(a$id)
#'
#' # check output
#' str(a)
#'
#' # plot via step function
#' xyplot(top ~ value | id, data=a, ylab='Depth',
#'        ylim=c(250, -5), as.table=TRUE,
#'        panel=panel.depth_function,
#'        prepanel=prepanel.depth_function,
#'        scales=list(x=list(alternating=1))
#' )
#'
#' ##
#' ## categorical variable example
#' ##
#'
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#'
#' # normalize horizon names: result is a factor
#' sp1$name <- generalize.hz(
#'   sp1$name,
#'   new = c('O','A','B','C'),
#'   pat = c('O', '^A','^B','C')
#'   )
#'
#' # compute slice-wise probability so that it sums to contributing fraction, from 0-150
#' a <- slab(sp1, fm= ~ name, cpm=1, slab.structure=0:150)
#'
#' # convert wide -> long for plotting
#' # result is a data.table
#' # genhz factor levels are set by order in `measure.vars`
#' a.long <- data.table::melt(
#'   data.table::as.data.table(a),
#'   id.vars = c('top','bottom'),
#'   measure.vars = c('O', 'A', 'B', 'C'),
#'   )
#'
#'
#' # plot horizon type proportions using panels
#' xyplot(top ~ value | variable,
#'        data = a.long, subset=value > 0,
#'        col = 1, lwd = 2,
#'        xlab = 'Class Probability',
#'        ylab = 'Depth (cm)',
#'        strip = strip.custom(bg = grey(0.85)),
#'        scales = list(x = list(alternating = FALSE)),
#'        ylim = c(150, -5), type=c('S','g'),
#'        horizontal = TRUE, layout = c(4,1)
#'        )
#'
#' # again, this time using groups
#' xyplot(top ~ value,
#'        data = a.long,
#'        groups = variable,
#'        subset = value > 0,
#'        ylab = 'Depth (cm)',
#'        ylim = c(150, -5),
#'        type = c('S','g'),
#'        horizontal = TRUE,
#'        asp = 2,
#'        lwd = 2,
#'        auto.key = list(
#'          lines = TRUE,
#'          points = FALSE,
#'          cex = 0.8,
#'          columns = 1,
#'          space = 'right'
#'        )
#' )
#'
#' # adjust probability to size of collection, from 0-150
#' a.1 <- slab(sp1, fm= ~ name, cpm = 2, slab.structure = 0:150)
#'
#' # convert wide -> long for plotting
#' # result is a data.table
#' # genhz factor levels are set by order in `measure.vars`
#' a.1.long <- data.table::melt(
#'   data.table::as.data.table(a.1),
#'   id.vars = c('top','bottom'),
#'   measure.vars = c('O','A','B','C')
#' )
#'
#' # combine aggregation from `cpm` modes 1 and 2
#' g <- make.groups(cmp.mode.1 = a.long, cmp.mode.2 = a.1.long)
#'
#' # plot horizon type proportions
#' xyplot(top ~ value | variable,
#'        groups = which,
#'        data = g, subset = value > 0,
#'        ylim = c(240, -5),
#'        type = c('S','g'),
#'        horizontal = TRUE,
#'        layout = c(4,1),
#'        auto.key = list(lines = TRUE, points = FALSE, columns = 2),
#'        par.settings = list(superpose.line = list(col = c(1, 2), lwd = 2)),
#'        scales = list(alternating = 3),
#'        xlab = 'Class Probability',
#'        ylab = 'Depth (cm)',
#'        strip = strip.custom(bg = grey(0.85))
#' )
#'
#'
#' # apply slice-wise evaluation of max probability, and assign ML-horizon at each slice
#' gen.hz.ml <- get.ml.hz(a, c('O','A','B','C'))
#'
#'
#' \dontrun{
#' ##
#' ## HD quantile estimator
#' ##
#'
#' library(soilDB)
#' library(lattice)
#' library(data.table)
#'
#' # sample data
#' data('loafercreek', package = 'soilDB')
#'
#' # defaul slab.fun wraps stats::quantile()
#' a <- slab(loafercreek, fm = ~ total_frags_pct + clay)
#'
#' # use HD quantile estimator from Hmisc package instead
#' a.HD <- slab(loafercreek, fm = ~ total_frags_pct + clay, slab.fun = aqp:::.slab.fun.numeric.HD)
#'
#' # combine
#' g <- make.groups(standard=a, HD=a.HD)
#'
#' # note differences
#' densityplot(~ p.q50 | variable, data=g, groups=which,
#'             scales=list(relation='free', alternating=3, tick.number=10, y=list(rot=0)),
#'             xlab='50th Percentile', pch=NA, main='Loafercreek',
#'             auto.key=list(columns=2, points=FALSE, lines=TRUE),
#'             par.settings=list(superpose.line=list(lwd=2, col=c('RoyalBlue', 'Orange2')))
#' )
#'
#' # differences are slight but important
#' xyplot(
#'   top ~ p.q50 | variable, data=g, groups=which,
#'   xlab='Value', ylab='Depth (cm)',
#'   asp=1.5, main='Loafercreek',
#'   lower=g$p.q25, upper=g$p.q75,
#'   sync.colors=TRUE, alpha=0.25, cf=g$contributing_fraction,
#'   ylim=c(115,-5), layout=c(2,1), scales=list(x=list(relation='free')),
#'   par.settings=list(superpose.line=list(lwd=2, col=c('RoyalBlue', 'Orange2'))),
#'   strip=strip.custom(bg=grey(0.85)),
#'   panel=panel.depth_function,
#'   prepanel=prepanel.depth_function,
#'   auto.key=list(columns=2, lines=TRUE, points=FALSE)
#' )
#'
#' ##
#' ## multivariate examples
#' ##
#' data(sp3)
#'
#' # add new grouping factor
#' sp3$group <- 'group 1'
#' sp3$group[as.numeric(sp3$id) > 5] <- 'group 2'
#' sp3$group <- factor(sp3$group)
#'
#' # upgrade to SPC
#' depths(sp3) <- id ~ top + bottom
#' site(sp3) <- ~ group
#'
#' # custom 'slab' function, returning mean +/- 1SD
#' mean.and.sd <- function(values) {
#'   m <- mean(values, na.rm=TRUE)
#'   s <- sd(values, na.rm=TRUE)
#'   upper <- m + s
#'   lower <- m - s
#'   res <- c(mean=m, lower=lower, upper=upper)
#'   return(res)
#' }
#'
#' # aggregate several variables at once, within 'group'
#' a <- slab(sp3, fm = group ~ L + A + B, slab.fun = mean.and.sd)
#'
#' # check the results:
#' # note that 'group' is the column containing group labels
#' xyplot(
#'   top ~ mean | variable, data=a, groups=group,
#'   lower=a$lower, upper=a$upper,
#'   sync.colors=TRUE, alpha=0.5,
#'   cf = a$contributing_fraction,
#'   xlab = 'Mean Bounded by +/- 1SD',
#'   ylab = 'Depth (cm)',
#'   ylim=c(125,-5), layout=c(3,1),
#'   scales=list(x=list(relation='free')),
#'   par.settings = list(superpose.line=list(lwd=2, col=c('RoyalBlue', 'Orange2'))),
#'   panel = panel.depth_function,
#'   prepanel = prepanel.depth_function,
#'   strip = strip.custom(bg=grey(0.85)),
#'   auto.key = list(columns=2, lines=TRUE, points=FALSE)
#' )
#'
#'
#' # compare a single profile to the group-level aggregate values
#' a.1 <- slab(sp3[1, ], fm = group ~ L + A + B, slab.fun = mean.and.sd)
#'
#' # manually update the group column
#' a.1$group <- 'profile 1'
#'
#' # combine into a single data.frame:
#' g <- rbind(a, a.1)
#'
#' # plot with customized line styles
#' xyplot(
#'   top ~ mean | variable, data=g, groups=group, subscripts=TRUE,
#'   lower=a$lower, upper=a$upper, ylim=c(125,-5),
#'   layout=c(3,1), scales=list(x=list(relation='free')),
#'   xlab = 'Mean Bounded by +/- 1SD',
#'   ylab = 'Depth (cm)',
#'   panel=panel.depth_function,
#'   prepanel=prepanel.depth_function,
#'   sync.colors = TRUE, alpha = 0.25,
#'   par.settings = list(
#'     superpose.line = list(
#'       col = c('orange', 'royalblue', 'black'),
#'       lwd = 2, lty = c(1,1,2)
#'     )
#'   ),
#'   strip = strip.custom(bg=grey(0.85)),
#'   auto.key = list(columns=3, lines=TRUE, points=FALSE)
#' )
#'
#'
#'
#'
#' ## again, this time for a user-defined slab from 40-60 cm
#' a <- slab(sp3,
#'           fm = group ~ L + A + B,
#'           slab.structure = c(40,60),
#'           slab.fun = mean.and.sd
#' )
#'
#' # now we have weighted average properties (within the defined slab)
#' # for each variable, and each group
#' # convert long -> wide
#' data.table::dcast(
#'   data.table::as.data.table(a),
#'   formula = group + top + bottom ~ variable,
#'   value.var = 'mean'
#' )
#'
#' ## this time, compute the weighted mean of selected properties, by profile ID
#' a <- slab(sp3,
#'           fm = id ~ L + A + B,
#'           slab.structure = c(40,60),
#'           slab.fun = mean.and.sd
#' )
#'
#' # convert long -> wide
#' data.table::dcast(
#'   data.table::as.data.table(a),
#'   formula = id + top + bottom ~ variable,
#'   value.var = 'mean'
#' )
#'
#'
#' ## aggregate the entire collection, using default slab function
#' ## note the missing left-hand side of the formula
#' a <- slab(sp3, fm = ~ L + A + B)
#'
#' }
#'
setMethod(f = 'slab',
          signature = 'SoilProfileCollection',
          definition = .slab)


#' @param method one of `"numeric"`, `"factor"`, `"hd"`, `"weighted.numeric"`, `"weighted.factor"`, `"fast"`
#' @details `slab_function()`: The default `"numeric"` aggregation method is the `"fast"` numeric (quantile) method. Additional methods include `"factor"` for categorical data, `"hd"` to use the Harrell-Davis Distribution-Free Quantile Estimator from the Hmisc package, and "`weighted`" to use a weighted quantile method from the Hmisc package
#' @return `slab_function()`: return an aggregation function based on the `method` argument
#' @rdname slab
#' 
#' @export
slab_function <- function(method = c("numeric", "factor", "hd", "weighted.numeric", "weighted.factor", "fast")) {
  switch(method,
         numeric = .slab.fun.numeric.default, # uses "fast" numeric method
         factor = .slab.fun.factor.default, 
         weighted.factor = .slab.fun.factor.weighted, 
         hd = .slab.fun.numeric.HD, 
         weighted.numeric = .slab.fun.numeric.weighted,
         weighted.factor = .slab.fun.numeric.weighted,
         fast = .slab.fun.numeric.fast,
         .slab.fun.numeric.default
  )
}

# for debugging;
.slabinternal <- function(...) {
  .slab(...)
}

## fully vectorized
## determine color contrast class, based on
## Soil Survey Technical Note 2 [wayback machine URL](https://web.archive.org/web/20220704214918/https://www.nrcs.usda.gov/wps/portal/nrcs/detail/soils/ref/?cid=nrcs142p2_053569)
##
## and 2-5 in the field book (version 3.0)
#
# v1: Munsell value of first color
# c1: Munsell chroma of first color
# v2: Munsell value of second color
# c2: Munsell chroma of second color
# dH: delta Hue
# dV: delta Value
# dC: delta Chroma
# verbose: return a list for testing rules/cases
#' Soil Color Contrast
#'
#' Determine soil color contrast class according to methods outlined in the
#' Soil Survey Manual. This function is typically called from
#' \code{colorContrast()} which is simpler to use and provides more
#' information.
#'
#' This function is fully vectorized but expects all inputs have the same
#' length.
#'
#' @param v1 Munsell value of first color
#' @param c1 Munsell chroma of first color
#' @param v2 Munsell value of second color
#' @param c2 Munsell chroma of second color
#' @param dH delta Hue
#' @param dV delta Value
#' @param dC delta Chroma
#' @param verbose return a list for testing rules/cases
#' @return A vector of color contrast classes (ordered factor). A list when
#' \code{verbose} is TRUE.
#' @author D.E. Beaudette
#' @seealso \code{\link{colorContrast}}
#' @references
#'  - Soil Survey Technical Note 2 [wayback machine URL](https://web.archive.org/web/20220704214918/https://www.nrcs.usda.gov/wps/portal/nrcs/detail/soils/ref/?cid=nrcs142p2_053569)
#' @keywords manip
#' @export
#' @examples
#'
#' ## standard use, result is an ordered factor
#' # 10YR 6/3 vs 5YR 3/4
#' contrastClass(v1=6, c1=3, v2=3, c2=4, dH=2, dV=3, dC=1)
#'
#' ## verbose output, useful for testing rules/cases
#' # 10YR 6/3 vs 5YR 3/4
#' contrastClass(v1=6, c1=3, v2=3, c2=4, dH=2, dV=3, dC=1, verbose = TRUE)
#'
contrastClass <- function(v1, c1, v2, c2, dH, dV, dC, verbose=FALSE) {

  # sanity check, all inputs must have the same length
  l <- c(length(v1), length(c1), length(v2), length(c2), length(dH), length(dV), length(dC))
  if(length(unique(l)) > 1) {
    stop('inputs must all have the same length', call. = FALSE)
  }

  # init space for contrast classes
  res <- vector(mode='character', length=length(v1))


  ## Faint
  # case 1: difference in hue = 0, difference in value is <= 2, and difference in chroma is <= 1, or
  f.case1 <- (dH == 0) & (dV <= 2) & (dC <= 1)

  # case 2: difference in hue = 1, difference in value is <= 1, and difference in chroma is <= 1, or
  f.case2 <- (dH == 1) & (dV <= 1) & (dC <= 1)

  # case 3: difference in hue = 2, difference in value = 0, and difference in chroma = 0, or
  f.case3 <- (dH == 2) & (dV == 0) & (dC == 0)
  #
  res[f.case1 | f.case2 | f.case3] <- 'Faint'

  ## Distinct
  # case 1: difference in hue = 0, and
  #           a. difference in value is <= 2 and difference in chroma is > 1 to < 4, or
  #           b. difference in value is > 2 to < 4 and difference in chroma is < 4
  d.case1 <- (dH == 0) & ( ( (dV <= 2) & (dC > 1) & (dC < 4) ) | ( (dV > 2) & (dV < 4) & (dC < 4)) )

  # case 2: difference in hue = 1, and
  #           a. difference in value is <= 1 and difference in chroma is > 1 to < 3, or
  #           b. difference in value is > 1 to < 3, and difference in chroma is < 3
  d.case2 <- (dH == 1) & ( ( (dV <= 1) & (dC > 1) & (dC < 3) ) | ( (dV > 1) & (dV < 3) & (dC < 3)) )

  # case 3: difference in hue = 2, and
  #           a. difference in value = 0 and difference in chroma is > 0 to < 2, or
  #           b. difference in value is > 0 to < 2 and difference in chroma is < 2.
  d.case3 <- (dH == 2) & ( ( (dV == 0) & (dC > 0) & (dC < 2) ) | ( (dV > 0) & (dV < 2) & (dC < 2)) )
  #
  res[d.case1 | d.case2 | d.case3] <- 'Distinct'


  # low vlaue / chroma exception
  # both colors have values of <= 3 and chromas of <= 2
  low.value.chroma <- ((v1 <= 3) & (v2 <= 3)) & ((c1 <= 2) & (c2 <= 2))
  res[low.value.chroma] <- 'Faint'

  # anyhting else is Prominent
  res[res == ''] <- 'Prominent'

  # convert to ordered factor
  res <- factor(res, levels=c('Faint', 'Distinct', 'Prominent'), ordered = TRUE)

  # NA propagation
  idx <- apply(cbind(v1, c1, v2, c2, dH, dV, dC), 1, function(i) any(is.na(i)))
  res[idx] <- NA

  # conversion to factors happens elsewhere

  if(verbose){
    return(list(
      'faint'=data.frame(v1, c1, v2, c2, dH, dV, dC, f.case1, f.case2, f.case3, low.value.chroma, res),
      'distinct'=data.frame(v1, c1, v2, c2, dH, dV, dC, d.case1, d.case2, d.case3, res)
    ))
  } else{
    return(res)
  }


}

context("Metrics of soil color contrast")

## sample data
m1 <- c('10YR 6/3', '7.5YR 3/3', '10YR 2/2', '7.5YR 3/4')
m2 <- c('5YR 3/4', '7.5YR 4/4', '2.5YR 2/2', '7.5YR 6/3')

## tests

test_that("contrastClass works as expected", {

  ## hand-done tests

  # 10YR 6/3 vs 5YR 3/4
  x <- contrastClass(v1=6, c1=3, v2=3, c2=4, dH=2, dV=3, dC=1, verbose = TRUE)
  expect_true(x$faint$res == 'Prominent')
  expect_equivalent(unlist(x$faint[, c('f.case1', 'f.case2', 'f.case3', 'low.value.chroma')]),  c(FALSE, FALSE, FALSE, FALSE))
  expect_equivalent(unlist(x$distinct[, c('d.case1', 'd.case2', 'd.case3')]),  c(FALSE, FALSE, FALSE))

  # 7.5YR 3/3 vs 7.5YR 4/4
  x <- contrastClass(v1=3, c1=3, v2=4, c2=4, dH=0, dV=1, dC=1, verbose = TRUE)
  expect_true(x$faint$res == 'Faint')
  expect_equivalent(unlist(x$faint[, c('f.case1', 'f.case2', 'f.case3', 'low.value.chroma')]),  c(TRUE, FALSE, FALSE, FALSE))
  expect_equivalent(unlist(x$distinct[, c('d.case1', 'd.case2', 'd.case3')]),  c(FALSE, FALSE, FALSE))

  # 10YR 2/2 vs 2.5YR 2/2
  x <- contrastClass(v1=2, c1=2, v2=2, c2=2, dH=0, dV=0, dC=0, verbose = TRUE)
  expect_true(x$faint$res == 'Faint')
  expect_equivalent(unlist(x$faint[, c('f.case1', 'f.case2', 'f.case3', 'low.value.chroma')]),  c(TRUE, FALSE, FALSE, TRUE))
  expect_equivalent(unlist(x$distinct[, c('d.case1', 'd.case2', 'd.case3')]),  c(FALSE, FALSE, FALSE))

  # 7.5YR 3/4 vs 7.5YR 6/3
  x <- contrastClass(v1=3, c1=4, v2=5, c2=3, dH=0, dV=3, dC=1, verbose = TRUE)
  expect_true(x$faint$res == 'Distinct')
  expect_equivalent(unlist(x$faint[, c('f.case1', 'f.case2', 'f.case3', 'low.value.chroma')]),  c(FALSE, FALSE, FALSE, FALSE))
  expect_equivalent(unlist(x$distinct[, c('d.case1', 'd.case2', 'd.case3')]),  c(TRUE, FALSE, FALSE))

  # Error: inputs must all have the same length
  expect_error(x <- contrastClass(v1=3, c1=4, v2=5, c2=3, dH=0, dV=3, dC=numeric(0), verbose = TRUE))

  ## TODO: test entire range of rules/cases

})


test_that("only m1 specified, pair-wise comparisons", {
  
  # contrast metrics
  d <- colorContrast(m1)
  
  # output should be a data.frame
  expect_true(inherits(d, 'data.frame'))
  
  # both m1 and m2 columns should contain original m1
  expect_true(
    length(setdiff(m1, unique(c(d$m1, d$m2)))) == 0
  )
  
  # color contrast should be an ordered factor
  expect_true(is.factor(d$cc))
  expect_true(is.ordered(d$cc))
  
})



test_that("colorContrast works as expected", {

  # contrast metrics
  d <- colorContrast(m1, m2)

  # output should be a data.frame
  expect_true(inherits(d, 'data.frame'))
  expect_true(nrow(d) == length(m1))

  # first two columns should contain original colors
  expect_true(all(d$m1 == m1))
  expect_true(all(d$m2 == m2))

  # color contrast should be an ordered factor
  expect_true(is.factor(d$cc))
  expect_true(is.ordered(d$cc))

})

test_that("colorContrast fails as expected", {

  # m1/m2 not same length ---> error
  expect_error(colorContrast(m1[1], m2))

  # bogus hues -> dH and dE00 are NA
  d <- colorContrast('10FG 2/3', '4ZZ 4/5')
  expect_true(is.na(d$dH))
  expect_true(is.na(d$dE00))

  # bogus Munsell colors, all NA
  d <- colorContrast(m1 = '123sdf', m2 = '345gg')
  expect_true(all(is.na(d[, -c(1:2)])))
  
  # m1 specified with < 2 values
  expect_error(colorContrast('10YR 3/3'))
  
  # do not allow 0-length vectors
  expect_error(colorContrast(character(0), character(0)))
})


test_that("valid results", {

  # contrast metrics
  d <- colorContrast(m1, m2)

  # hand-checked
  expect_equal(d$dH, c(2, 0, 3, 0))
  expect_equal(d$dV, c(3, 1, 0, 3))
  expect_equal(d$dC, c(1, 1, 0, 1))
  expect_equal(as.character(d$cc), c('Prominent', 'Faint', 'Faint', 'Distinct'))

  ## TODO add some less-common colors

})



test_that("neutral colors", {
  
  # contrast metrics
  d <- colorContrast(m1 = 'N 3/', m2 = 'N 6/')
  
  # hand-checked
  expect_equal(d$dH, 0)
  expect_equal(d$dV, 3)
  expect_equal(d$dC, 0)
  expect_equal(as.character(d$cc), 'Distinct')

  # more complex examples
  m1 <- c('10YR 6/3', '7.5YR 3/3', '10YR 2/2', 'N 3/')
  m2 <- c('5YR 3/4', '7.5YR 4/4', 'N 2/', 'N 4/')
  
  d <- colorContrast(m1, m2)
  
  # N vs N    -> dH == 0
  # N vs. HUE -> dH == 1
  expect_true(
    all(
      d$dH == c(2L, 0L, 1L, 0L)
    )
  )
  
  # N vs N    -> dV == standard
  # N vs. HUE -> dV == standard
  expect_true(
    all(
      d$dV == c(3L, 1L, 0L, 1L)
    )
  )
   
  # N vs N    -> dC == 0
  # N vs. HUE -> dC == chroma of real hue
  expect_true(
    all(
      d$dC == c(1L, 1L, 2L, 0L)
    )
  )
   
})

#' 
#' @description internally used function to parse slicedHSD formula notation
#' @param fm to parse, 0:100 ~ variable | group
#' @return list with formula pieces
#' 
#' @keywords internal
#' @noRd
.parseGrouped_formula <- function(fm) {

  ## testing
  # fm <- 0:100 ~ estimated_oc | taxonname

  ## parse formula
  # string processing to get the LHS
  # expression parsing to get the RHS

  # extract components of the formula:
  fm.str <- paste(deparse(fm, 500), collapse = '')
  fm.elements <- strsplit(x = fm.str, split = '~', fixed = )[[1]]
  fm.elements <- lapply(strsplit(x = fm.elements, split = '[+*]'), trimws)

  # sanity check:
  if(length(fm.elements) != 2) {
    stop('error in `fm`, should be top:bottom ~ var | group', call. = FALSE)
  }

  # extract depth sequence
  depth.sequence <- fm.elements[[1]]

  # extract var | group
  var.group <- all.vars(update(fm, 0~.)) # right-hand side

  # sanity check
  if(length(var.group) != 2) {
    stop('error in `fm`, should be top:bottom ~ var | group', call. = FALSE)
  }

  # done
  return(list(fm = fm, depth.sequence = depth.sequence, var = var.group[1], group = var.group[2]))
}


#' .HSD
#' @description Safely compute Tukey's HSD within a given slice
#' @param z data.frame containing basic metadata, horizon top/bottom, variable of interest, and grouping variable
#' @param aov.fm formula suitable for aov(): variable ~ group
#' @param conf confidence level for TukeyHSD
#' 
#' @keywords internal
#' @noRd
.HSD <- function(z, aov.fm, conf = 0.95) {

  # sanity checks

  ## TODO: probably a better way to move names around

  # output template for error conditions
  # use first row of input

  # top depth is the second column
  # bottom depth is the third column
  d <- data.frame(
    top = z[1, 2],
    bottom = z[1, 3],
    diff = NA,
    lwr = NA,
    upr = NA,
    p.adj = NA
  )

  # propogate horizon depth names
  names(d)[1:2] <- names(z)[2:3]


  # fit model and catch errors
  mod <- try(
    aov(aov.fm, data = z),
    silent = TRUE
    )

  # sometimes there aren't enough data
  # return NA results
  if(inherits(mod, 'try-error')) {
    return(d)
  }

  # does this need error trapping?
  res <- TukeyHSD(mod, conf.level = conf)

  # results: only works with A/B style testing of two groups
  d$diff <- res[[1]][1]
  d$lwr <- res[[1]][2]
  d$upr <- res[[1]][3]
  d$p.adj <- res[[1]][4]

  return(d)
}



#' @title Tukey's HSD Over Slices
#' @description Apply Tukey's HSD over 1-unit depth slices.
#'
#' @author D.E. Beaudette and Sharon Perrone
#'
#' @param object \code{SoilProfileCollection} object
#' @param fm a formula describing depth sequence, horizon attribute, and site (grouping) attribute.
#' For example 0:100 ~ estimated_oc | taxonname
#' @param conf confidence applied in \code{TukeyHSD}
#' @export
slicedHSD <- function(object, fm, conf = 0.95) {

  # parse specialized HSD formula
  fm.data <- .parseGrouped_formula(fm)

  # sanity checks: is this an SPC?
  if(! inherits(object, 'SoilProfileCollection')) {
    stop('`object` should be a SoilProfileCollection', call. = FALSE)
  }

  # sanity check: do the variables specified in fm exist?
  # note that there is only a single variable and single group
  if(
    ! fm.data$var %in% horizonNames(object) |
    ! fm.data$group %in% siteNames(object)
  ) {
    stop('`fm` must reference a single horizon-level attribute and a single site-level attribute', call. = FALSE)
  }


  # parse depth sequence
  depth.sequence <- eval(parse(text = fm.data$depth.sequence))

  # create slice formula
  slice.fm <- as.formula(sprintf("%s ~ %s", fm.data$depth.sequence, fm.data$var))

  # create slab formula
  slab.fm <- as.formula(sprintf("%s ~ %s", fm.data$group, fm.data$var))

  # create AOV formula
  aov.fm <- as.formula(sprintf("%s ~ %s", fm.data$var, fm.data$group))

  ## TODO: this would be a heck of a lot simpler with a slab() re-write
  # https://github.com/ncss-tech/aqp/issues/3

  ## for now, using slice + slab

  # slice for HSD iteration
  # use results with caution
  s <- dice(object, fm = slice.fm)

  # aggregate by group over slices
  a <- slab(object, fm = slab.fm, slab.structure = depth.sequence)


  # HSD iteration uses only a single site-level attr defining groups and some horizon data
  # hmmm.. does it make sense to cut-down to just those variables that matter?
  ## TODO: this will need some special treatment for tibble / data.table
  hzd <- horizonDepths(s)
  vars <- c(idname(s), hzd, hzidname(s), fm.data$var)

  # convert to data.frame
  x <- horizons(s)[, vars, drop = FALSE]
  # move group from site -> hz
  x[[fm.data$group]] <- denormalize(s, fm.data$group)

  ## TODO: this needs to be more robust
  ## generate slice / segment labels
  # genSlabLabels(slab.structure = depth.sequence, max.d = max(depth.sequence), n.profiles = length(object))

  ## TODO: this will not work for larger slabs, will have to implement
  # split according to slice top depths
  xx <- split(x, x[[hzd[1]]])
  n.slices <- length(xx)

  # iterate over slice-index and perform Tukey's HSD
  # this likely violates the "multiple comparison" criteria... but I am not a statistician
  HSD <- lapply(1:n.slices, function(i) {
    .HSD(z = xx[[i]], aov.fm = aov.fm, conf = conf)
  })

  # list -> DF
  HSD <- do.call('rbind', HSD)

  # pack everything needed to plot into a list
  res <- list(
    HSD = HSD,
    agg = a,
    min.depth = min(depth.sequence),
    max.depth = max(depth.sequence),
    HSD.conf = conf
  )


  # done
  return(res)

}

#' Algorithms for Quantitative Pedology
#'
#' The aqp (Algorithms for Quantitative Pedology) package for R was developed
#' to address some of the difficulties associated with processing soils
#' information, specifically related to visualization, aggregation, and
#' classification of soil profile data. This package is based on a mix of S3/S4
#' functions and classes, and most functions use basic dataframes as input,
#' where rows represent soil horizons and columns define properties of those
#' horizons. Common to most functions are the requirements that horizon
#' boundaries are defined as depth from 0, and that profiles are uniquely
#' defined by an id column. The aqp package defines an S4 class,
#' "SoilProfileCollection", for storage of profile-level metadata, as well as
#' summary, print, and plotting methods that have been customized for common
#' tasks related to soils data.
#'
#' Demos: \code{demo(aqp)}
#'
#'
#' [Project homepage](http://ncss-tech.github.io/AQP/)
#'
#'
#' @name aqp-package
#' 
#' @aliases aqp-package aqp aqp.env
#' 
#' @author Dylan E. Beaudette <debeaudette@@ucdavis.edu>, Pierre Roudier, Andrew G. Brown
#' 
#' @seealso `depths<-()`, `SoilProfileCollection()`, \code{\link{sp1}}, \code{\link{sp2}}, \code{\link{sp3}}, \code{\link{sp4}}, \code{\link{sp5}}, \code{\link{sp6}}
#' 
#' @keywords package
#' 
#' @import data.table
#' 
#' @importFrom grDevices chull col2rgb colorRamp colorRampPalette colors convertColor grey hsv rgb rgb2hsv
#' 
#' @importFrom graphics abline arrows axis box grid image legend lines mtext par points polygon rect segments strheight strwidth text grconvertX
#' 
#' @importFrom methods setClass setOldClass representation prototype new isGeneric setGeneric setReplaceMethod setMethod .hasSlot as new slot slot<- slotNames
#' 
#' @importFrom stats TukeyHSD aggregate aov approxfun as.dist as.formula cmdscale complete.cases dist formula median model.frame na.omit na.pass quantile rnorm runif sd splinefun terms update weighted.mean cov as.hclust xtabs
#' 
#' @importFrom utils object.size packageVersion combn
#' 
#' @importFrom cluster pam daisy silhouette
#' 
#' @importFrom grid grid.text gpar unit
#' 
#' @importFrom graphics plot
#' 
#' @importFrom lattice levelplot xyplot panel.abline panel.grid panel.lines panel.points panel.polygon panel.rect panel.segments panel.text strip.custom trellis.par.get
#' 
#' @importFrom digest digest
#' 
#' @importFrom farver convert_colour compare_colour
#' 
#' @importFrom colorspace deutan tritan protan
#' 
#' @importFrom ape as.phylo plot.phylo tiplabels
#' 
#' 
"_PACKAGE"

#' @export aqp.env
#' @noRd
"aqp.env"

# create a (redundant) horizon-level attribute from a site-level attribute
# @ authors: andrew brown & dylan beaudette
#' Create a (redundant) horizon-level attribute from a site-level attribute
#'
#' @description Create a (redundant) horizon-level attribute from a site-level attribute. Specify a SoilProfileCollection and a site-level attribute from that SPC (by name) to receive a vector of length equal to the number of horizons containing the site-level values. This vector is directly usable with the SoilProfileCollection horizon setter.
#' 
#' \code{denormalize} is the inverse operation for the formula interface that "normalizes" a horizon level variable to site level:
#' 
#' \code{site(object) <- ~ horizonvar}
#' 
#' @details "Denormalization" is the process of trying to improve the read performance of a database, at the expense of losing some write performance, by adding redundant copies of data or by grouping data. Sometimes it is beneficial to have site-level attributes denormalized for grouping of horizon-level data in analyses. \code{denormalize} achieves this result for SoilProfileCollections. 
#' 
#' @author Andrew G. Brown, Dylan Beaudette
#' 
#' @param object A SoilProfileCollection
#' @param attr Site-level attribute name (character string) to denormalize to horizon.
#' @return A vector of values of equal length to the number of rows in the horizon table of the input SPC.
#' 
#' @export denormalize
#'
#' @examples
#' 
#' data(sp1)
#' 
#' # create a SoilProfileCollection from horizon data
#' depths(sp1) <- id ~ top + bottom
#' 
#' # create random site-level attribute `sitevar` with a binary (0/1) outcome
#' sp1$sitevar <- round(runif(length(sp1)))
#' 
#' # use denormalize() to create a mirror of sitevar in the horizon table
#' # name the attribute something different (e.g. `hz.sitevar`) to 
#' # prevent collision with the site attribute
#' # the attributes can have the same name but you will then need 
#' # site() or horizons() to access explicitly
#' sp1$hz.sitevar <- denormalize(sp1, 'sitevar')
#' 
#' # compare number of profiles to number of sitevar assignments
#' length(sp1)
#' table(sp1$sitevar)
#' 
#' # compare number of horizons to number of horizon-level copies of sitevar `hz.'sitevar`
#' nrow(sp1)
#' table(sp1$hz.sitevar)
denormalize <- function(object, attr) {
  
  # extract relevant pieces
  h <- horizons(object)
  s <- site(object)
  idn <- idname(object)
  if(!attr %in% names(s))
    stop("column name %s not found", call. = FALSE)
  
  # make a lookup table of attr in site
  lut <- s[[attr]]
  names(lut) <- s[[idn]]
  
  # return a horizon level attribute for same site IDs
  return(lut[h[[idn]]])
}

context("denormalize - redundant horizon attributes from site")

data(sp1, package = 'aqp')

# create a SoilProfileCollection from horizon data
depths(sp1) <- id ~ top + bottom

# create random site-level attribute `sitevar` with a binary (0/1) outcome
sp1$sitevar <- round(runif(length(sp1)))

test_that("denormalize result is 1:1 with horizons", {
  # use denormalize() to create a mirror of sitevar in the horizon table
  # name the attribute something different (e.g. `hz.sitevar`) to prevent collision with the site attribute
  # the attributes can have the same name but you will then need site() or horizons() to access explicitly
  sp1.hz.sitevar <- denormalize(sp1, 'sitevar')

  expect_error(sp1.hz.sitevar <- denormalize(sp1, 'foo'))

  # compare number of horizons to number of values in denormalize result
  expect_equal(nrow(sp1), length(sp1.hz.sitevar)) # check that the output is 1:1 with horizon

  sp1$hz.sitevar <- sp1.hz.sitevar
})

test_that("round trip normalize/denormalize", {
  
  data(sp3)
  depths(sp3) <- id ~ top + bottom
  
  # create site var -- unique at site level
  site(sp3)$foo <- profile_id(sp3)
  
  # denormalize site var to horizon var (leaves foo in site)
  expect_error({sp3$foo <- denormalize(sp3, "foo")})
  
  # need to create a new variable for hz-denorm var
  sp3$foo2 <- denormalize(sp3, "foo")
  
  # inspect
  plot(sp3, color="foo2")
  
  # normalize to site (removes foo2 in horizon)
  site(sp3) <- ~ foo2
  
  # expected TRUE
  expect_true(all(sp3$foo == sp3$foo2))
  expect_true(all(sp3$foo2 == profile_id(sp3)))
  
  # commence the breakin'
  
  # make another `foo3`
  sp3$foo3 <- denormalize(sp3, "foo")
  
  # not appropriate for normalization (1:1 with horizon, not site)
  sp3$foo4 <- 1:nrow(sp3)
  
  # do that SPC dirty...
  expect_warning(site(sp3) <- ~ foo3 + foo4)
  
  # still valid
  expect_true(spc_in_sync(sp3)$valid)
  
  # didn't do anything
  expect_equal(length(sp3$foo3), nrow(sp3))
  expect_equal(length(sp3$foo4), nrow(sp3))
  
  # try something less diverse and insidious
  
  # 10 unique levels, equal length to number of horizons
  set.seed(100) 
  superbad <- round(runif(nrow(sp3), 0, length(sp3)))
  
  # expect_equal(length(unique(superbad)), 10)
  
  horizons(sp3)$notgood <- superbad
  expect_warning(site(sp3) <- ~ foo3 + notgood)
})

context("data.table and tbl_df inheritance from data.frame")


test_that("basic coercion", {

  skip_on_cran()

  skip_if_not_installed("tibble")
  
  ###  create empty frames
  empty_df <- data.frame()

  if (requireNamespace("data.table", quietly = TRUE)) {
     empty_dt <- data.table::as.data.table(empty_df)
     expect_equal(dim(empty_dt), c(0, 0))
  }

  if (requireNamespace("tibble", quietly = TRUE)) {
    empty_tb <- tibble::as_tibble(empty_df)
    expect_equal(dim(empty_tb), c(0, 0))
  }

  ### .as.data.frame.aqp wrapper on data.frames
  expect_equal(.as.data.frame.aqp(empty_df, "data.frame"), empty_df)
  ## data frame with 0 columns and 0 rows

  expect_equal(.as.data.frame.aqp(empty_df, "data.table"), empty_dt)
  ## Null data.table (0 rows and 0 cols)

  expect_equal(.as.data.frame.aqp(empty_df, "tbl_df"), empty_tb)
  ## A tibble: 0 x 0


  # as.data.frame wrapper on tibbles
  expect_equal(.as.data.frame.aqp(empty_tb, "data.table"), empty_dt)
  # Null data.table (0 rows and 0 cols)
  #
  expect_equal(.as.data.frame.aqp(empty_tb, "data.frame"), empty_df)
  # data frame with 0 columns and 0 rows
  #
  expect_equal(.as.data.frame.aqp(empty_tb, "tbl_df"), empty_tb)
  # A tibble: 0 x 0

  # as.data.frame wrapper on data.table
  expect_equal(.as.data.frame.aqp(empty_dt, "data.table"), empty_dt)
  # Null data.table (0 rows and 0 cols)

  expect_equal(.as.data.frame.aqp(empty_dt, "data.frame"), empty_df)
  # data frame with 0 columns and 0 rows

  expect_equivalent(.as.data.frame.aqp(empty_dt, "tbl_df"), empty_tb)
  # note: that because of slight internal differences (leftover from data.table?)
  #       cannot use expect_equal -- though they are clearly equivalent
  #
  # > str(.as.data.frame.aqp(empty_dt, "tbl_df"))
  # tibble [0 × 0] (S3: tbl_df/tbl/data.frame)
  # Named list()
  # - attr(*, ".internal.selfref")=<externalptr>
  #
  # > str(empty_tb)
  # tibble [0 × 0] (S3: tbl_df/tbl/data.frame)
  # Named list()

  })

dfclasses <- list("data.frame", "data.table", "tbl_df")

res <- lapply(dfclasses, function(use_class) {

  ##
  ##
  ## debug note -- this is vectorized over the data.frame classes
  ##
  ##

  # debug
  # use_class <- "data.frame"

  test_that(sprintf("basic functionality of %s", use_class), {
    # class to use (data.frame, data.table or tbl_df)

    cc <- function(l) {
      do.call('c', as.list(l))
    }

    # basic function for converting inputs btween tibble and data.frame
    test_object <- function(object, use_class) {
      .as.data.frame.aqp(object, as.class = use_class)
    }

    # make some fake data
    df <- data.frame(
      id = cc(lapply(1:4, function(i)
        rep(i, 10))),
      top = cc(rep(0:9, 4)),
      bottom = cc(rep(1:10, 4)),
      siteprop = 8,
      prop = 18
    )

    # construct a test object of type use_class
    test <- test_object(object = df, use_class = use_class)
    test2 <- test
    
    skip_if(is.null(test))
    
    # promote to SPC -- convert numeric ids to character
    expect_message(depths(test) <- id ~ top + bottom,
                   c("converting profile IDs from integer to character"))

    # add fake coordinates
    crds <- data.frame(id = profile_id(test),
                       y  = rnorm(length(test)),
                       x  = rnorm(length(test)))
    site(test) <- crds

    # promote to spatial
    initSpatial(test, crs = "+proj=longlat +datum=WGS84") <- ~ x + y

    # show method should be produce output without error
    expect_output(show(test))
    expect_output(show(test[0,]))

    # fill in diagnostics and restrictions with fake data
    diagnostic_hz(test) <- data.frame(id = profile_id(test),
                                   featkind = "foo",
                                   featdept = 0, featdepb = 10)
    restrictions(test) <- data.frame(id = profile_id(test),
                                  restrkind = "bar",
                                  restrdept = 0, restrdepb = 10)

    # try the character vector interface too
    expect_message(depths(test2) <- c("id", "top", "bottom"),
                   c("converting profile IDs from integer to character"))

    # test rebuild
    expect_message(rebuildSPC(test2), "using `hzID` as a unique horizon ID")

    # test enforce_df_class
    expect_silent(.enforce_df_class(test2, use_class))

    # "normalize" (horizon -> site) a site-level attribute
    site(test) <- ~ siteprop
    
    # new random XY data
    test$y <- rnorm(length(test))
    test$x <- rnorm(length(test))
    test$newx <- denormalize(test, "x")
    test$newy <- denormalize(test, "y")
    
    # promote to spatial works from horizons
    # formula and character are allowed inputs on RHS
    # LHS specifies optional CRS
    initSpatial(test, crs = "OGC:CRS84") <- ~ newx + newy
    
    # siteprop removed from horizons
    if (use_class == "tbl_df") {
      expect_warning(expect_null(horizons(test)$siteprop))
    } else {
      expect_null(horizons(test)$siteprop)
    }
    
    # newx is still in horizon data (converted to site level only on coercion)
    expect_false(is.null(horizons(test)$newx))
    
    # check that ids are in order
    expect_equal(profile_id(test), site(test)$id)

    # check that siteprop is in site
    expect_equal(site(test)$siteprop, rep(8, 4))

    # check that siteprop is not in horizons (anymore)
    switch(
      use_class,
      # data.frame and data.table
      expect_equal(horizons(test)$siteprop, NULL),

      # note: this is only to cover a warning and to document differences
      #       tbl_df will not return NULL -- warning and length(0)
      "tbl_df" = expect_warning(
        expect_equal(length(horizons(test)$siteprop), 0),
        "Unknown or uninitialised column: `siteprop`."
      )
    )

    # "denormalize" into a different-named horizon attribute
    horizons(test)$siteprop_hz <- denormalize(test, "siteprop")

    # check that siteprop_hz is in horizons
    # print(str(names(horizons(test)$siteprop_hz)))
    if(use_class == "tbl_df") {
      expect_equivalent(horizons(test)$siteprop_hz, rep(8, 40))
      # note: tbl_df will have names attribute attached to result
      #       which precludes expect_equal
    } else {
      expect_equal(horizons(test)$siteprop_hz, rep(8, 40))
    }

    # add some site data, for only two sites
    value <- test_object(data.frame(id = as.character(2:3),
                                   siteclass = state.abb[1:2]),
                        use_class)

    # left join
    site(test) <- value

    # inspect site
    site(test)

    # inspect site table with new variable (first/last NA), sorted correct
    expect_equal(which(is.na(site(test)$siteclass)), c(1, 4))

    # inspect horizon table
    horizons(test)

    # test $ [[ setter/getter
    test$foo <- rep(100, nrow(test))
    expect_true(all(test[['foo']] == horizons(test)$foo))

    test[["foo"]] <- rep(200, nrow(test))
    expect_equal(horizons(test)$foo,  rep(200, nrow(test)))

    expect_equal(horizons(test)[[idname(test)]], as.character(cc(lapply(1:4, rep, 10))))
    expect_equal(horizons(test)[[horizonDepths(test)[2]]], cc(rep(1:10, 4)))

    # creating and destroying site and horizon column vectors
    site(test)$a <- "foo"
    site(test)[["b"]] <- "bar"
    horizons(test)$c <- "baz"
    horizons(test)[["d"]] <- "qux"

    expect_equal(test$a, rep("foo", length(test)))
    expect_equal(test$b, rep("bar", length(test)))
    expect_equal(test$c, rep("baz", nrow(test)))
    expect_equal(test$d, rep("qux", nrow(test)))

    site(test)$a <- NULL
    site(test)[["b"]] <- NULL
    test$c <- NULL
    horizons(test)$d <- NULL

    expect_null(test$a)
    expect_null(test$b)
    expect_null(test$c)
    expect_null(test$d)

    # add some horizon data
    value <- test_object(data.frame(
                          id = as.character(2:3),
                          hzID = as.character(20:21),
                          hzclass = letters[1:2]
                         ), use_class)

    switch(
      use_class,
      # data.frame and tbl_df
      expect_silent({
        horizons(test) <- value
      }),
      # note: data.table sorts correctly without invoking re-order
      "data.table" = {
        expect_silent( { horizons(test) <- value } )
      }
    )

    expect_equal(horizons(test)[[idname(test)]], as.character(cc(lapply(1:4, rep, 10))))

    expect_equal(horizons(test)[[horizonDepths(test)[2]]], cc(rep(1:10, 4)))

    # deliberately mess up order without adding anything
    expect_silent(horizons(test) <- horizons(test)[c(1:10,20:30,11:19,31:40),])

    # horizons<- fixes the order
    expect_true(spc_in_sync(test)$valid)

    # if there is no new data and IDs don't match, a message is issued and object is unchanged
    hzbad <- horizons(test)
    hzbad[[idname(test)]] <- paste0("h",1:nrow(hzbad))
    expect_message(horizons(test) <- hzbad , "Some profile IDs in input data are not present in object and no new columns to merge. Doing nothing.")

    ## make sample data using current class type
    data(sp1, package = 'aqp')
    sp1df <- .as.data.frame.aqp(sp1, use_class, stripFactors = TRUE)
    depths(sp1df) <- id ~ top + bottom
    site(sp1df) <- ~ group

    # add real coordinates
    sp1df$x <- seq(-119, -120, length.out = length(sp1df))
    sp1df$y <- seq(38, 39, length.out = length(sp1df))

    ## tests

    test_that(sprintf("SPC construction from a %s", use_class), {

      # did it work?
      expect_true(inherits(sp1df, 'SoilProfileCollection'))

      # ID correctly initialized?
      expect_equal(idname(sp1df), 'id')
      expect_true(length(profile_id(sp1df)) == length(sp1df))

      # ID in the correct order?
      expect_identical(profile_id(sp1df), site(sp1df)[[idname(sp1df)]])

      # depth names?
      expect_equal(horizonDepths(sp1df), c('top', 'bottom'))

      # site-level attributes correctly initialized?
      expect_true(length(sp1df$group) == length(sp1df))

      # correct number of profiles and horizons?
      expect_equal(length(sp1df), 9)
      expect_equal(nrow(sp1df), 60)

      # this should work
      foo <- explainPlotSPC(sp1df)
      expect_true(is.list(foo))
    })

    test_that(sprintf("SPC diagnostics and restrictions (%s)", use_class), {
      # diagnostic & restriction slot should be initialized as an empty data.frame
      sp1.dh <- diagnostic_hz(sp1df)
      expect_true(inherits(sp1.dh, 'data.frame'))
      expect_equal(nrow(sp1.dh), 0)

      sp1.rh <- restrictions(sp1df)
      expect_true(inherits(sp1.rh, 'data.frame'))
      expect_equal(nrow(sp1.rh), 0)
    })

    test_that(sprintf("SPC data.frame interface (%s)", use_class), {

      # init site-level data
      sp1df$x <- seq(-119, -120, length.out = length(sp1df))
      sp1df$y <- seq(38, 39, length.out = length(sp1df))

      # init hz-level data
      sp1df$z <- runif(n = nrow(sp1df))

      expect_equal(length(sp1df$x), length(sp1df))
      expect_equal(length(sp1df$z), nrow(sp1df))
    })

    test_that(sprintf("SPC deconstruction into a data.frame (%s)", use_class), {

      # do it here
      h <- horizons(sp1df)
      s <- site(sp1df)
      d <- .as.data.frame.aqp(as(sp1df, 'data.frame'), use_class)

      expect_true(inherits(h, use_class))
      expect_true(inherits(s, use_class))
      expect_true(inherits(d, use_class))
    })


    test_that(sprintf("SPC deconstruction into a list (%s)", use_class), {

      # do it here
      l <- as(sp1df, 'list')

      # result should be a list
      expect_true(inherits(l, 'list'))

      # there should be no NULL data, e.g. missing slots
      res <- sapply(l, is.null)
      expect_false(any(res))

      # check internals
      expect_equivalent(l$idcol, idname(sp1df))
      expect_equivalent(l$hzidcol, hzidname(sp1df))
      expect_equivalent(l$depthcols, horizonDepths(sp1df))
      expect_equivalent(l$metadata, metadata(sp1df))

      expect_equivalent(l$horizons, horizons(sp1df))
      expect_equivalent(l$site, site(sp1df))
      expect_equivalent(l$diagnostic, diagnostic_hz(sp1df))
      expect_equivalent(l$restrictions, restrictions(sp1df))

      # check internals after [-subsetting
      sp1.sub <- sp1df[2:3,]
      # none of these slots should change, the others will be subset
      # verifying these are transferred ensures key info slots are handled
      # by the SPC subset method
      expect_equivalent(l$idcol, idname(sp1.sub))
      expect_equivalent(l$hzidcol, hzidname(sp1.sub))
      expect_equivalent(l$depthcols, horizonDepths(sp1.sub))

      # TODO: better checking of metadata consistency
      #       offload more important things to slots
      #       use units package for depth units, and offer conversion?
      #       inches to centimeters could be usefl
      expect_equivalent(names(l$metadata), names(metadata(sp1.sub)))
    })


    test_that(sprintf("SPC subsetting (%s)", use_class), {

      # profile subsets
      expect_true(inherits(sp1df[1, ], 'SoilProfileCollection'))
      expect_true(inherits(sp1df[1:5, ], 'SoilProfileCollection'))

      # profile and horizon subsets
      expect_true(inherits(sp1df[1, 1], 'SoilProfileCollection'))

      # horizon subsets
      expect_true(inherits(sp1df[, 2], 'SoilProfileCollection'))

      # there should only be 1 profile and 1 horizon
      expect_equal(length(sp1df[1, 1]), 1)
      expect_equal(nrow(sp1df[1, 1]), 1)

      # there should be 5 profiles and 1 horizon / profile
      expect_equal(length(sp1df[1:5, 1]), 5)
      expect_equal(nrow(sp1df[1:5, 1]), 5)

    })

    test_that(sprintf("SPC subsetting with tidy verbs (%s)", use_class), {

      # filter works as expected
      expect_equal(length(subset(sp1df, structure_type == "PL")), 1)

      # ensure multiple expressions yields same result as single expression
      l1 <-  subset(sp1df, !is.na(texture),
                    prop > mean(prop, na.rm = TRUE))
      l2 <-  subset(sp1df, !is.na(texture) &
                    prop > mean(prop, na.rm = TRUE))
      expect_equivalent(length(l1), length(l2))

      # mixing of site and horizon level expressions is the intersection
      l1 <- subset(sp1df, group == 2, prop > mean(prop, na.rm = TRUE))
      expect_equivalent(length(l1), 4)

      # grepSPC works as expected
      expect_equal(length(grepSPC(sp1df, texture, "SCL")), 1)

      # subApply works as expected
      expect_equal(length(subApply(sp1df, function(p)  TRUE)), length(sp1df))
    })

    if (use_class == "data.table") {
      test_that("data.table specific", {
        expect_equal(min(sp1df), 59)
        expect_equal(max(sp1df), 240)
      })
    }
  })

})

# TODO: better colors
# take a look at between-profile D
# for the range of indexes

#' Between Individual Distance Plot
#'
#' Plot pair-wise distances between individuals as line segments.
#'
#' By default all individuals are plotting on the same axis. When there are
#' more than about 10 individuals, the plot can become quite messy. See
#' examples below for ideas.
#'
#' @param D distance matrix, should be of class 'dist' or compatible class
#' @param idx an integer sequence defining which individuals should be compared
#' @return No value is returned.
#' @author Dylan E Beaudette
#' @seealso [sp2]
#' @references https://casoilresource.lawr.ucdavis.edu/
#' @keywords hplot
#' @export
#' @examples
#'
#' data(sp2)
#' depths(sp2) <- id ~ top + bottom
#' d <- NCSP(
#'   sp2,
#'   vars = c('prop', 'field_ph', 'hue', 'value'),
#'   maxDepth = 100,
#'   k = 0.01
#' )
#'
#' par(mfcol=c(3,1), mar=c(2.5,4.5,1,1))
#' plot_distance_graph(d, idx=1:6)
#' plot_distance_graph(d, idx=7:12)
#' plot_distance_graph(d, idx=12:18)
plot_distance_graph <- function(D, idx=1:dim(as.matrix((D)))[1]) {
	# convert to full matrix representation
	# careful with very large D
	d.mat <- as.matrix(D)

	# mask out NA
	d.mat[d.mat == 0] <- NA

	# get number of individuals
	n <- dim(d.mat)[1]

	# get the original labels
	labs <- attr(D, 'Labels')
	# abbreviate them
	labs <- abbreviate(labs, minlength=3)

	# get the range in D and rescale to {0,1}
	d.range <- range(d.mat, na.rm=TRUE)
	d.mat <- d.mat / d.range[2]

	# empty plot region
	plot(1,1, type='n', xlim=c(1, n), ylim=c((d.range[1]/d.range[2])-0.1, 1),
	axes=FALSE, xlab='', ylab='Relative Distance')

	# add embellishments
	box()
	axis(2, las=2)
	# use original labels
	axis(1, at=1:n, labels=names(labs), las=2, cex.axis=0.75)
	grid(col=grey(0.25))

	# add lines, labels, and symbols
	for(i in idx) {
		# which individual is the most similar (min distance)
		min.idx <- which.min(d.mat[, i])
		# plot a line for each individual
		lines(1:n, d.mat[, i], type='b', pch=NA, col=(i %% 5) + 1, lty=(i %% 3) + 1)
		# label each line at nodes, except for the node of max similarity (min distance)
		text(c(1:n)[-min.idx], d.mat[-min.idx, i], label=labs[i], font=1, cex=0.75, col=(i %% 5) + 1)

		# mark the pair with min distance
		points(min.idx, d.mat[min.idx, i], cex=2, pch='*', col=(i %% 5) + 1)
		text(min.idx, d.mat[min.idx, i], label=labs[i], cex=0.78, font=2, pos=1, col=(i %% 5) + 1)
		}
	}

#'
#' @title Color Contrast Plot
#'
#' @description A simple display of two sets of colors, NCSS color contrast class and CIE
#' delta-E00.
#' 
#' @note This function requires the `farver` package for calculation of CIE
#' delta-E00.
#'
#' @param m1 first set of Munsell colors for comparison (e.g. '5YR 3/2')
#' @param m2 second set of Munsell colors for comparison
#' @param col.cex scaling factor for color labels
#' @param col.font font for color labels
#' @param d.cex contrast for contrast metric labels
#' @param cc.font font for contrast class
#' @param dE00.font font for delta-E00
#' @param labels labels for compared colors, vector length 2
#' @param label.cex scaling factor for labels
#' @param label.font font for labels
#' @param printMetrics logical, print metrics between color swatches
#' @param \dots further arguments to `colorspace::swatchplot`
#' @author D.E. Beaudette
#' @seealso [colorContrast()]
#' @keywords hplots
#' @export
#' 
#' @examples
#'
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#'
#' # two sets of colors to compare
#' m1 <- c('10YR 6/3', '7.5YR 3/3', '10YR 2/2', '7.5YR 3/4')
#' m2 <- c('5YR 3/4', '7.5YR 4/4', '2.5YR 2/2', '7.5YR 6/3')
#'
#' # contrast metrics
#' colorContrast(m1, m2)
#'
#' # graphical display
#' colorContrastPlot(m1, m2)
#' 
colorContrastPlot <- function(
           m1,
           m2,
           col.cex = 1,
           col.font = 2,
           d.cex = 1,
           cc.font = 3,
           dE00.font = 1,
           labels = c('m1', 'm2'),
           label.cex = 1,
           label.font = 1,
           printMetrics = TRUE,
           ...) {


  # compose list of swatchplot
  colors <- list(
   parseMunsell(m1),
   parseMunsell(m2)
  )
  names(colors) <- labels

  # basic plot
  colorspace::swatchplot(colors, cex = label.cex, font = label.font, ...)

  # eval contrast metrics
  d <- colorContrast(m1, m2)

  ## TODO this could use some more thought
  # helpers for computing position
  nx <- length(m1)
  ny <- 2

  # positions of Munsell color labels
  x.pos <- seq(from = 0, to = 1, by = 1/nx)[1:nx]
  y.pos1 <- rep(0.51, times = nx)
  y.pos2 <- rep(0.01, times = nx)

  # create high-contrast label colors
  m1.label.colors <- invertLabelColor(colors[[1]])
  m2.label.colors <- invertLabelColor(colors[[2]])

  # annotate top row
  text(x.pos, y.pos1, labels = m1, col = m1.label.colors, font = col.font, cex = col.cex, adj = c(-0.125, -0.33))
  # annotate bottom row
  text(x.pos, y.pos2, labels = m2, col = m2.label.colors, font = col.font, cex = col.cex, adj = c(-0.125, -0.33))

  # usually want to add the contrast metrics, but sometimes it is handy to suppress
  if(printMetrics) {
    # positions for color contrast class
    cc.y <- rep(0.46, times = nx)
    cc.lab <- as.character(d$cc)
    text(x.pos, cc.y, labels = cc.lab, adj = 0, font = cc.font, cex = d.cex)

    # positions for dE00
    dE00.y <- rep(0.39, times = nx)
    dE00.lab <- signif(d$dE00, 3)

    # this is clunky but I can't figure out a better solution
    # iterate over delta-E00, format with plotmath and add text
    # note on plotmath formatting and bquote() tricks
    # https://lukemiller.org/index.php/2017/05/r-plotmath-functions-combined-with-variable-values/
    for(i in 1:nx) {
      dE00.lab.expr <- bquote(Delta*E['00']~.(dE00.lab[i]))
      text(x.pos[i], dE00.y[i], labels = dE00.lab.expr, adj = 0, font = 3, cex = d.cex)
    }
  }

}

#' @title Perform summaries on groups (from \code{group_by}) and create new site or horizon level attributes
#'
#' @aliases summarizeSPC,SoilProfileCollection-method, summarize
#'
#' @description \code{summarize()} is a function used for summarizing a `SoilProfileCollection` object. Specify the groups using the group_by verb, and then (named) expressions to evaluate on each group. The result is a data.frame with one row per categorical level in the grouping variable and one column for each summary variable.
#'
#' @param object A `SoilProfileCollection`
#' @param ... A set of (named) comma-delimited R expressions that resolve to a summary value. e.g \code{groupmean = mean(clay, na.rm = TRUE)}
#'
#' @return A `data.frame` with one row per level in the grouping variable, and one column for each summary
#'
#' @author Andrew G. Brown
#' @export
#'
summarizeSPC <-  function(object, ...) {
    # TODO: safe setter and accessor methods for grouping variable
    group.by.col <- object@metadata$aqp_group_by

    if (length(group.by.col) == 0)
      group.by.col <- idname(object)

    groups <- levels(factor(as.character(object[[group.by.col]])))

    # capture expression(s) at function
    .dots <- substitute(list(...))
    .dots <- .dots[2:length(.dots)]
    .names <- names(.dots)

    if (is.null(.names))
      .names <- as.character(.dots)

    dfout <- .as.data.frame.aqp(data.frame(matrix(nrow = 0, ncol = length(.names))), aqp_df_class(object))

    # TODO: generalize split for n site or horizon level attributes

    res <- lapply(aqp::split(object, f = group.by.col), function(obj) {

      # TODO: what else would be needed to do similar splits on the horizon data?
      #       essentially, base::split of horizons(obj) + summary on groups within SPC? pretty simple in principle

      #       summarize could operate on groups at multiple levels -- just need another loop in here for horizons
      #       this depends on aqp::split() being able to make splits based on criteria at multiple levels, then
      #       parsing up of the profiles in resulting list elements based on unique levels of horizon-level groups
      #
      #       new behavior of aqp::split would possibly result in a list with overlap of profiles across elements
      #       also, could use diagnostics and or restrictions -- future compositeSPC will include these

      # apply expressions to each group, frameify results
       .data <- compositeSPC(obj)
       for(i in 1:length(.dots)) {
          output <- .data_dots(.data, eval(.dots[[i]]))[[1]]
          names(output) <- .names[i-1]
          if(length(output) > 1) {
            stop("summary value '%s' has length greater than one", call.=FALSE)
          } else if(length(output) == 0) {
            stop("summary value '%s' has length zero", call.=FALSE)
          }
          if(any(c(nrow(dfout), length(dfout)) == 0)) {
            dfout <- output
          } else {
            dfout <- cbind(dfout, output)
          }
       }
       return(dfout)
    })

    # recombine results for each group into a single data.frame
    summaries <- do.call('rbind', res)
    final <- data.frame(groups, summaries)
    colnames(final) <- c(group.by.col, .names)

    # return result in same class as SPC slots
    return(.as.data.frame.aqp(final, aqp_df_class(object)))
}

## TODO: generalize to a particle size clases


#' @title Coarse Fragment Class Labels and Diameter
#' @description This is a convenience function for accessing coarse fragment class labels and associated diameter (mm), as defined in various classification systems such as USDA, Unified, and AASHTO.
#'
#' @param sys character, length 1. This is an abbreviated name used to select class labels and fragment diameter.
#' @param flat logical. Fragments are flat, only used by USDA systems.
#' @param rounded logical. Fragments are rounded, only used by AASHTO system.
#' 
#' @references 
#' Schoeneberger, P.J., D.A. Wysocki, E.C. Benham, and Soil Survey Staff. 2012. Field book for describing and sampling soils, Version 3.0. Natural Resources Conservation Service, National Soil Survey Center, Lincoln, NE.
#' 
#' @return named vector of fragment diameter in mm
#' @export
#' 
#' @seealso [fragmentSieve()]
#'
#' @examples
#' 
#' # use default system: "usda_simplified"
#' fragmentClasses()
#' fragmentClasses(flat = TRUE)
#' 
#' fragmentClasses(sys = 'usda')
#' fragmentClasses(sys = 'USDA', flat = TRUE)
#' 
#' fragmentClasses(sys = 'international')
#' 
#' fragmentClasses(sys = 'unified')
#' 
#' fragmentClasses(sys = 'aashto')
#' fragmentClasses(sys = 'aashto', rounded = TRUE)
#' 
#' fragmentClasses(sys = 'mod.wentworth')
#' 
fragmentClasses <- function(sys = c('usda_simplified', 'usda', 'international', 'unified', 'aashto', 'mod.wentworth'), flat = FALSE, rounded = FALSE) {
  
  # normalize to lower case
  sys <- tolower(sys)
  
  # most frequent option, usually not specified
  if(missing(sys)) {
    sys <- 'usda_simplified'
  } else {
    sys <- match.arg(sys)
  }
  
  # upper fragment diameter, some large number (mm)
  .upper_limit <- 1e5
  
  # select a specification 
  .spec <- switch(sys, 
                  'usda_simplified' = {
                    
                    if(flat) {
                      c(
                        channers = 150,
                        flagstones = 380,
                        stones = 600,
                        boulders = .upper_limit
                      )
                      
                    } else {
                      c(
                        gravel = 76,
                        cobbles = 250,
                        stones = 600,
                        boulders = .upper_limit
                      )
                    }
                    
                  },
                  
                  'usda' = {
                    
                    if(flat) {
                      c(
                        channers = 150,
                        flagstones = 380,
                        stones = 600,
                        boulders = .upper_limit
                      )
                      
                    } else {
                      c(
                        fine_gravel = 5,
                        medium_gravel = 20,
                        coarse_gravel = 76,
                        cobbles = 250,
                        stones = 600,
                        boulders = .upper_limit
                      )
                    }
                    
                  },
                  
                  'international' = {
                    
                    c(
                      gravel = 20,
                      stones = .upper_limit
                    )
                    
                  },
                  
                  'unified' = {
                    
                    c(
                      fine_gravel = 19,
                      coarse_gravel = 76,
                      cobbles = 300,
                      boulders = .upper_limit
                    )
                    
                  },
                  
                  'aashto' = {
                    
                    if(rounded) {
                      c(
                        fine_gravel = 9.5,
                        medium_gravel = 25,
                        coarse_gravel = 75,
                        boulders = .upper_limit
                      )
                      
                    } else {
                      c(
                        fine_gravel = 9.5,
                        medium_gravel = 25,
                        coarse_gravel = 75,
                        broken_rock = .upper_limit
                      )
                    }
                    
                  },
                  
                  'mod.wentworth' = {
                    
                    c(
                      pebbles = 64,
                      cobbles = 256,
                      boulders = .upper_limit
                    )
                    
                  }
                  
  )
  

  return(.spec)
}

#' @title Visual Explanation for \code{plotSPC}
#' 
#' @description Create a visual explanation for the many arguments to \code{plotSPC}. Call this function instead of \code{plotSPC}, all objects after \code{x} are passed on to \code{plotSPC}. Nearly all of the figures in the \href{https://ncss-tech.github.io/AQP/aqp/aqp-intro.html}{Introduction to SoilProfileCollection Objects tutorial} are created with this function.
#'
#' @author D.E. Beaudette
#' 
#' @seealso \code{\link{plotSPC}}
#' 
#' @keywords manip
#' 
#' @param x a \code{SoilProfileCollection} object
#' 
#' @param \dots arguments passed to \code{\link{plotSPC}}
#'
#' @return a list of internally-used ordering vectors and graphical offsets / scaling factors
#'
#' @keywords manip
#' @export
#' @examples
#'
#' # sample data
#' data(sp4)
#' depths(sp4) <- id ~ top + bottom
#'
#' # proposed vector of relative positions, overlap likely
#' pos <- c(1, 1.1, 3, 4, 5, 5.2, 7, 8, 9, 10)
#'
#' # try it
#' explainPlotSPC(sp4, name = 'name', relative.pos=pos)
#'
#' # attempt to fix using an integer sequence, short-circut will prevent adjustments
#' explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(1:10))
#'
#' # attempt to adjust using defaults
#' explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos))
#'
#' # attempt to adjust and tinker with defaults
#' explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos, adj = 0.2))
#'
#' # enforce larger space between
#' explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos, thresh = 0.7))
#'
#' # more complex adjustments required
#' pos <- c(1, 2, 3, 3.3, 5, 5.1, 5.5, 8, 9, 10)
#'
#' # tinker
#' explainPlotSPC(sp4, name = 'name', relative.pos = pos)
#' explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos))
#' 
#' explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos, 
#' thresh = 0.7))
#' 
#' explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos, 
#' thresh = 0.7, adj = 0.2))
#'
#' # SANN: solution requires many iterations, and will not always converge
#' explainPlotSPC(sp4, name = 'name', 
#' relative.pos = fixOverlap(pos, thresh = 0.85, adj = 0.2)
#' )
#' 
#' # electrostatics: solution requires larger charge (q)
#' explainPlotSPC(sp4, name = 'name', 
#' relative.pos = fixOverlap(pos, thresh = 0.85, method = 'E', q = 2)
#' )
#'
explainPlotSPC <- function(x, ...) {
  plotSPC(x, id.style='side', ...)
  box()
  
  # figure coordinates
  .usr <- par('usr')
  
  # get last plot parameters
  lsp <- get('last_spc_plot', envir = aqp.env)
  
  # get max depth by profile
  max.depths <- profileApply(x, max)
  
  # re-order max depths
  max.depths <- max.depths[lsp$plot.order]
  
  # apply y-offset and scaling factor
  # y-offset is a vector
  scaled.max.depths <- lsp$y.offset + (lsp$scaling.factor * max.depths)
  
  # suitable location for y-space annotation
  y.space.x <- 2.5
  
  # suitable location for x-space annotation
  # index of last profile + some
  x.space.x <- lsp$n + (length(x) * 0.05)
  
  # 95% of lower figure coordinates
  x.space.y <- .usr[3] * 0.95
  
  # original profile index text y-coordinate
  # roughly 10% of the max(transformed depths)
  original.profile.idx.y <- lsp$y.offset + (-max(scaled.max.depths) * 0.08)
  
  # segment depth starting locations
  # 90% of bottom of plotting region ----> scaled max depths 
  # or max(scaled depths), whichever is larger
  .segment.bottom.y <- pmax(.usr[3] * 0.9, max(scaled.max.depths, na.rm = TRUE))
  
  # inspect plotting area, very simple to overlay graphical elements
  segments(x0 = lsp$x0, x1 = lsp$x0, y0 = .segment.bottom.y, y1 = scaled.max.depths, lty=3, lwd=2, col='darkgreen')
  
  # profiles are centered at integers, from 1 to length(obj)
  axis(1, line=0.25, at=round(lsp$x0, 2), cex.axis=0.75, font=4, col='darkgreen', col.axis='darkgreen', lwd=2)
  mtext('canvas x-coordinate', side=1, line=2.25, font=4, col='darkgreen')
  
  # y-axis is based on profile depths
  axis(2, line=0.25, cex.axis=0.75, font=4, las=1, col='blue', col.axis='blue', lwd=2)
  mtext('canvas y-coordinate', side=2, line=2.25, font=4, col='blue')
  
  # show extra y-space
  arrows(x0=y.space.x, x1=y.space.x, y0=0, y1=-lsp$extra_y_space, length = 0.08, code = 3, col='orange', lwd=1)
  text(x=y.space.x, y=-lsp$extra_y_space/2, labels = 'extra y-space', cex=0.65, pos=2, font=3, col='orange')
  text(x=y.space.x, y=-lsp$extra_y_space/2, labels = lsp$extra_y_space, cex=0.85, pos=4, font=2, col='orange')
  
  # show extra x-space
  arrows(x0=lsp$n, x1=lsp$n + lsp$extra_x_space, y0=x.space.y, y1=x.space.y, length = 0.08, code = 3, col='orange', lwd=1)
  text(x=x.space.x, y=x.space.y, labels = 'extra x-space', cex=0.65, pos=3, font=3, col='orange')
  text(x=x.space.x, y=x.space.y, labels = lsp$extra_x_space, cex=0.85, pos=1, font=2, col='orange')
  
  # demonstrate width on first profile
  arrows(x0=lsp$x0[1] - lsp$width, x1=lsp$x0[1] + lsp$width, y0=x.space.y, y1=x.space.y, length = 0.08, code = 3, col='orange', lwd=1)
  text(x=lsp$x0[1], y=x.space.y, labels = 'width', cex=0.65, pos=3, font=3, col='orange')
  text(x=lsp$x0[1], y=x.space.y, labels = lsp$width, cex=0.85, pos=1, font=2, col='orange')
  
  # plotting order
  text(x=lsp$x0, y=original.profile.idx.y, labels=lsp$plot.order, col='darkred', font=4, cex=0.75)
  mtext('original profile index', side=3, line=0, font=4, col='darkred')
  
  # hz depth label adjustment index
  if(! all(is.na(lsp$hz.depth.LAI))) {
    # annotate
    .LAI.txt <- sprintf("LAI\n%s", round(lsp$hz.depth.LAI, 2))
    text(x = lsp$x0 + lsp$width, y = .segment.bottom.y, pos = 4, cex = 0.66, labels = .LAI.txt)
  }
  
  
  invisible(lsp)
}

# https://www.fao.org/3/cb0509en/CB0509EN.pdf


#' @title Inflate / Deflate Horizon Thickness
#' 
#' @description This function applies a warping factor to the horizons of a single-profile `SoilProfileCollection` object. Warping values >1 will inflate horizon thickness, values <1 will deflate horizon thickness.
#'
#' @param x a `SoilProfileCollection` object with a single soil profile
#' 
#' @param fact numeric or character; warping factor specified as a single numeric value, vector of numeric values (length = `nrow(x)`), or column name of a horizon level attribute containing numeric values
#' 
#' @param scaleTo numeric, target depth resulting from rescaling all horizon thickness values, due to rounding the actual depth may be within +/- 1 depth unit
#' 
#' @param soilDepthFun function for computing soil depth, either: 
#' 
#'   * `estimateSoilDepth()`: "soil depth" is defined as depth to contact, requires that `hzdesgnname()` be set, see [estimateSoilDepth()] for details
#'   * `max()`: "soil depth" is defined as the bottom depth of the soil profile
#' 
#' @param updateProfileID logical; modify profile IDs
#' 
#' @param suffix character; suffix added to profile IDs when `updateProfileID = TRUE`
#' 
#' @author D.E. Beaudette and S.W. Salley
#'
#' @return a modified version of `x`, `SoilProfileCollection` object
#' @export
#'
#' @examples 
#' 
#' # create an example profile
#' s <- quickSPC('p1:AA|Bt1Bt1Bt1|Bt2Bt2B|Bt3|Cr|RRRRR')
#' 
#' # warp each horizon
#' # values > 1: inflation
#' # values < 1: deflation (erosion / compaction)
#' s.w <- warpHorizons(s, fact = c(1.3, 0.7, 0.8, 1, 1, 1))
#' 
#' # combine original + warped
#' x <- combine(s, s.w)
#' 
#' # compute profile bottom depths
#' .bottoms <- x[, , .LAST, .BOTTOM]
#' 
#' # change in total depth after warping
#' # used to vertically offset the warped profile
#' .yoff <- c(0, .bottoms[1] - .bottoms[2])
#' 
#' # depths for line segments connecting horizon tops
#' .y1 <- x[1, , .TOP]
#' .y2 <- x[2, , .TOP] + .yoff[2]
#' 
#' # sketches
#' # can't automatically add a depth axis
#' par(mar = c(0.5, 0, 0, 2))
#' plotSPC(
#'   x, 
#'   name.style = 'center-center', 
#'   cex.names = 0.8, 
#'   width = 0.2, 
#'   max.depth = 150, 
#'   depth.axis = FALSE, 
#'   y.offset = .yoff
#' )
#' 
#' # illustrate warping with arrows
#' arrows(x0 = 1 + 0.25, y0 = .y1, x1 = 2 - 0.25, y1 = .y2, len = 0.1, col = 2)
#' 
#' # manually add depth axis
#' axis(side = 4, line = -3.5, las = 1, at = seq(from = 0, to = 150, by = 25))
#' 
#' 
#' # apply to multiple profiles
#' # text-based template
#' .template <- c(
#' 'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
#' 'P2:ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC',
#' 'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR'
#' )
#' 
#' # each horizon label is '10' depth-units (default)
#' s <- quickSPC(.template)
#' 
#' # random warping factor, by horizon
#' s$f <- runif(n = nrow(s), min = 0.8, max = 1.2)
#' 
#' # warp horizons by profile, result is a list of SPCs
#' s.w <- profileApply(s, FUN = warpHorizons, fact = 'f')
#'
#' # flatten list -> SoilProfileCollection
#' s.w <- combine(s.w)
#' 
#' # combine with original SPC
#' x <- combine(s, s.w)
#' 
#' # sketches 
#' par(mar = c(0.5, 0, 0, 2.5))
#' plotSPC(
#'   x,
#'   name.style = 'center-center',
#'   cex.names = 0.8,
#'   width = 0.3,
#'   max.depth = 165,
#'   depth.axis = list(line = -2)
#' )
#' 
#' # rescale all profiles to 100cm soil depth (depth to contact)
#' .template <- c(
#'   'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
#'   'P2:ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC',
#'   'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR',
#'   'P4:AAAAA|CCC|RRRRRR'
#' )
#' 
#' # each horizon label is '10' depth-units (default)
#' s <- quickSPC(.template)
#' 
#' # warp horizons by profile, result is a list of SPCs
#' w <- profileApply(s, FUN = function(i) {
#'   warpHorizons(i, scaleTo = 100, soilDepthFun = estimateSoilDepth)
#' })
#' 
#' # flatten list -> SoilProfileCollection
#' w <- combine(w)
#' 
#' # combine with original SPC
#' x <- combine(s, w)
#' 
#' # highlight "contact"
#' x$color <- rep(grey(0.9), times = nrow(x))
#' x$color[grep('R|Cr|Cd', x$name)] <- 'royalblue'
#' 
#' # sketches
#' par(mar = c(0.5, 0, 0, 2.5))
#' plotSPC(
#'   x,
#'   color = 'color',
#'   name.style = 'center-center',
#'   cex.names = 0.8,
#'   width = 0.3,
#'   max.depth = 165,
#'   depth.axis = list(line = -2)
#' )
#' 
#' abline(h = 100, lty = 3)
#'
warpHorizons <- function(x, fact = NULL, scaleTo = NULL, soilDepthFun = estimateSoilDepth, updateProfileID = TRUE, suffix = '-w') {
  
  ## TODO: 
  ## * vectorize over profiles (internal function via profileApply)
  ## * move argument checking to main function
  ## * make more efficient
  
  # extract parts and pieces of the SPC
  .h <- horizons(x)
  .htb <- horizonDepths(x)
  .n <- nrow(.h)
  
  # sanity check: fact should be:
  # * NULL, not specified in favor of scaleTo
  # * numeric, length == 1, used by all horizons
  # * numeric, length == nrow(x), each horizon has its own factor
  # * hz column name
  
  if(!is.null(fact)) {
    if(inherits(fact, 'character')) {
      fact <- x[[fact]]
      if(is.null(fact)) {
        stop('fact must name a column in horizons', call. = FALSE)
      }
    } else if(inherits(fact, 'numeric')) {
      
      if(length(fact) > 1 && length(fact) != .n) {
        stop('fact must be length 1 or nrow(x)', call. = FALSE)
      }
      
    } else {
      stop('fact must be either character vector or numeric', call. = FALSE)
    }
  } else if(!is.null(scaleTo)) {
    
    if(length(scaleTo) > 1) {
      stop('scaleTo should have length of 1')
    }
    
    # compute fact from scaleTo and soil depth
    # this requires
    fact <- (1 / soilDepthFun(x)) * scaleTo
    
  } else {
    # identity
    message('no scaling factor provided, using `fact = 1`')
    fact <- 1
  }
  
  # warping is applied to horizon thickness
  .thick <- .h[[.htb[2]]] - .h[[.htb[1]]]
  
  # apply inflation/deflation factor to horizon thickness
  # round to integers
  .thick <- round(.thick * fact)
  
  # future: apply offset here
  
  # generate new horizon depth sequence
  # starting from original topmost depth
  .start <- .h[1,  .htb[1]]
  
  # tops / bottoms
  .tops <- c(.start, cumsum(.thick[-.n]))
  .bottoms <- c(cumsum(.thick))
  
  # replace original values
  .h[[.htb[1]]] <- .tops
  .h[[.htb[2]]] <- .bottoms
  
  # re-pack horizons
  replaceHorizons(x) <- .h
  
  # optionally update profile ID
  # this is handy when trying to combine(old, new)
  if(updateProfileID) {
    .pID <- profile_id(x)
    profile_id(x) <- sprintf("%s%s", .pID, suffix)
  }
  
  return(x)
}

context("warpHorizons()")


test_that("warping factor", {
  
  s <- quickSPC('p1:AA|Bt1Bt1Bt1|Bt2Bt2B|Bt3|Cr|RRRRR')
  
  # identity
  expect_message(w <- warpHorizons(s))
  expect_true(inherits(w, 'SoilProfileCollection'))
  expect_equal(s[, , .BOTTOM], w[, , .BOTTOM])
  
  w <- warpHorizons(s, fact = 2)
  expect_true(inherits(w, 'SoilProfileCollection'))
  expect_equal(max(w), 280)
  
  w <- warpHorizons(s, fact = 0.5)
  expect_true(inherits(w, 'SoilProfileCollection'))
  expect_equal(max(w), 70)
  
  w <- warpHorizons(s, fact = c(1.3, 0.7, 0.8, 1, 1, 1))
  expect_true(inherits(w, 'SoilProfileCollection'))
  expect_equal(max(w), 133)
  
})

test_that("scaleTo", {
  
  s <- quickSPC('p1:AA|Bt1Bt1Bt1|Bt2Bt2B|Bt3|Cr|RRRRR')
  
  # exact target depth currently impossible due to rounding
  w <- warpHorizons(s, scaleTo = 100, soilDepthFun = estimateSoilDepth)
  expect_true(inherits(w, 'SoilProfileCollection'))
  expect_true(
    abs(estimateSoilDepth(w) - 100) < 2
  )
  
  # exact target depth currently impossible due to rounding
  w <- warpHorizons(s, scaleTo = 100, soilDepthFun = max)
  expect_true(inherits(w, 'SoilProfileCollection'))
  expect_true(
    abs(max(w) - 100) < 2
  )

})

test_that("warping factor is a horizon level attribute", {
  
  s <- quickSPC('p1:AA|Bt1Bt1Bt1|Bt2Bt2B|Bt3|Cr|RRRRR')
  s$wf <- runif(nrow(s), min = 0.5, max = 1.5)
  
  w <- warpHorizons(s, fact = 'wf')
  expect_true(inherits(w, 'SoilProfileCollection'))
})

#' @title Make High Contrast Label Colors
#' @description Generate a vector of white or black label colors conditioned on a vector of colors to maximize label contrast.
#' 
#' @param colors vector of colors
#' @param threshold black | white threshold
#' 
#' @return vector of label colors
#' 
#' @author D.E. Beaudette
#' @export
#' @examples
#' 
#' # test with shades of grey
#' s <- seq(0, 1, by = 0.05)
#' cols <- grey(s)
#' soilPalette(cols, lab = as.character(s))
#' 
#' # test with 10YR x/3
#' m <- sprintf('10YR %s/3', 1:8)
#' cols <- parseMunsell(m)
#' soilPalette(cols, lab = m)
#' 
#' 
invertLabelColor <- function(colors, threshold = 0.65) {
  
  # NA is converted to white, resulting in black label
  # convert colors -> sRGB -> HSV
  hsv.cols <- t(rgb2hsv(col2rgb(colors)))
  
  # truncate hue and saturation
  hsv.cols[, 1] <- 0
  hsv.cols[, 2] <- 0
  
  # conditionally set value according to threshold
  # V > thresh -> black | else white
  hsv.cols[, 3] <- ifelse(hsv.cols[, 3] > threshold, 0, 1)
  
  # new label color: white | black
  label.color <- hsv(hsv.cols[, 1], hsv.cols[, 2], hsv.cols[, 3])
  
  return(label.color)
}

# this function is not (yet) used anywhere

# goal is to be able to take any SPC and convert to X internals

.enforce_df_class <- function(object, use_class) {
  
  # check class and convert as needed
  object@horizons <- .as.data.frame.aqp(object@horizons, use_class)
  object@site <- .as.data.frame.aqp(object@site, use_class)
  object@diagnostic <- .as.data.frame.aqp(object@diagnostic, use_class)
  object@restrictions <- .as.data.frame.aqp(object@restrictions, use_class)

  if (use_class == "data.table") {
    
    # namespace requirement for setting data.table keys
    if (!requireNamespace("data.table"))
      stop("package `data.table is required", call. = FALSE)
    
    # set keys on data.table objects
    data.table::setkeyv(object@horizons, c(idname(object), horizonDepths(object)[1]), verbose = FALSE)
    
    data.table::setkeyv(object@site, idname(object), verbose = FALSE)

    if (idname(object) %in% names(object@diagnostic))
      data.table::setkeyv(object@diagnostic, idname(object), verbose = FALSE)
    
    if (idname(object) %in% names(object@restrictions))
      data.table::setkeyv(object@restrictions, idname(object), verbose = FALSE)
  }

  return(object)
}

## TODO: think about how to better integrate this into aqp and other packages
## TODO: dump scales import
## TODO: implement scales::rescale 'from' argument

# cheap alternative to {scales} rescale(..., to = c(x0, x1)), defaults to interval of {0,1}
.rescaleRange <- function(x, x0, x1) {
  res <- (x1 - x0) * ( (x - min(x, na.rm = TRUE)) / (max(x, na.rm = TRUE) - min(x, na.rm = TRUE))) + x0
  return(res)
}



#' Calculate Relative Positions from Transect Data
#' 
#' @description This function is used to support relative positioning of soil profiles by `plotSPC`, based on transect or gradient values typically associated with a site level attribute (e.g. elevation). Gradient values specified in `x` are translated to the range used by `plotSPC` (usually `1, length(SPC)`) specified in `x.min` and `x.max`.
#'
#' @details See the [Pair-Wise Distances by Generalized Horizon Labels](http://ncss-tech.github.io/AQP/aqp/genhz-distance-eval.html) tutorial for additional examples.
#'
#' @param x numeric vector, describing values along a transect: distance, elevation, climatic variables, etc.. Typically sourced from the site level attributes of a `SoilProfileCollection` object. Order is not important.
#' 
#' @param x.min numeric, lower boundary to relative position scale
#' @param x.max numeric, upper boundary to relative position scale
#' @param fix logical, attempt fixing overlapping positions with `fixOverlap`
#' @param ... additional arguments to `fixOverlap`
#'
#' @return `list` containing:
#'   * `grad`: values of `x` in ascending order
#'   * `order`: ordering vector of `x`
#'   * `relative.pos`: elements of `x` translated to the new relative scale defined by `x.min` and `x.max`
#'   
#' @export
#'
#' @examples
#' 
#' data("sierraTransect")
#' 
#' # split transects
#' g <- subset(sierraTransect, transect == 'Granite')
#' a <- subset(sierraTransect, transect == 'Andesite')
#' 
#' g.p <- alignTransect(g$elev, x.min = 1, x.max = length(g), fix = FALSE)
#' a.p <- alignTransect(a$elev, x.min = 1, x.max = length(a), fix = FALSE)
#' 
#' op <- par(mar=c(2,0,0,2), mfrow=c(2,1))
#' 
#' plotSPC(g, width=0.25, name.style='center-center', 
#'         cex.names=0.75, 
#'         relative.pos = g.p$relative.pos, plot.order = g.p$order)
#' 
#' axis(1, at = g.p$relative.pos, labels = g.p$grad, line = -1.5)
#' 
#' plotSPC(a, width=0.25, name.style='center-center', 
#'         cex.names=0.75, 
#'         relative.pos = a.p$relative.pos, plot.order = a.p$order)
#' 
#' axis(1, at = a.p$relative.pos, labels = a.p$grad, line = -1.5)
#' 
#' 
#' par(op)
#' 
alignTransect <- function(x, x.min, x.max, fix = TRUE, ...) {
  
  # sanity checks
  # no NA allowed in min/max
  if(is.na(x.min) | is.na(x.max)) {
    stop('NA are not allowed in `x.min` or `x.max`', call. = TRUE)
  }
  
  # TODO: it may be possible to "fill" NA in x
  if(any(is.na(x))) {
    warning('NA in `x` may lead to unexpected results', call. = FALSE)
  }
  
  # monotonic mapping from range(x) -> [x.min, x.max]
  o <- order(x)
  r <- .rescaleRange(x[o], x0 = x.min, x1 = x.max)
  
  # optionally fix overlap
  if(fix) {
    r <- fixOverlap(r, ...)  
  }
  
  res <- list(
    grad = x[o],
    order = o,
    relative.pos = r
  )
  
  return(res)
  
}

context("alignTransect")

data("sierraTransect")

# split transects
g <- subset(sierraTransect, transect == 'Granite')
a <- subset(sierraTransect, transect == 'Andesite')

# basic functionality
test_that("alignTransect works as expected", {
  
  # CRAN safe
  
  # data are not pre-sorted by elevation, alpha order
  p <- alignTransect(g$elev, 1, length(g), fix = FALSE)
  
  # structure
  expect_true(inherits(p, 'list'))
  expect_true(length(p) == 3)
  
  # known output
  expect_true(all(p$order == c(7, 1, 2, 4, 5, 6, 3)))
})

# basic functionality
test_that("more complex input", {
  
  # CRAN safe
  
  # more interesting, data are not pre-sorted by elevation
  p <- alignTransect(a$elev, 1, length(a), fix = FALSE)
  
  # structure
  expect_true(inherits(p, 'list'))
  expect_true(length(p) == 3)
  
  # known output
  expect_true(all(p$order == c(2, 5, 1, 3, 7, 4, 6)))
  
})


## TODO: add a couple more with more complex ordering

# setters -- functions that substantially modify data.frame slot contents

##
## depths<- setter method - to create AQP objects: sorts based on ID and top depth
##
# if (!isGeneric('depths<-'))
  setGeneric('depths<-', function(object, value) standardGeneric('depths<-'))


#' @aliases depths<-,SoilProfileCollection-method
#' @rdname depths
#' @export
setReplaceMethod("depths", signature(object = "SoilProfileCollection"),
	function(object, value) {
		message('This is already a SoilProfileCollection-class object, doing nothing.')
		object
	})

#' Initialize a SoilProfileCollection from data.frame 
#' @description `depths(<data.frame>) <- <formula>`: Initialize SoilProfileCollection
#' @description `depths(<SoilProfileCollection>)`: Extract profile ID and horizon depths from SoilProfileCollection
#' @name depths<-
#' @param object An object to promote to SoilProfileCollection (inherits from data.frame)
#' @param value A formula specifying the unique profile ID, top and bottom depth column names
#' @aliases depths<-,data.frame-method depths depths<-
#' @details The input horizon data, and the resulting profile order, is sorted based on unique profile ID and top depth. ID columns are converted to character, depth columns are converted to integer. If `NA` values exist in all of the top depths, a prototype with 1 horizon per profile ID is returned, with `NA` in all non-essential columns. If the input `object` has 0 rows, a prototype with 0 horizons and 0 rows, but same column names as `object`, is returned. 
#' @rdname depths
#' @export
#' @examples
#' ## init SoilProfileCollection objects from data.frame of horizon data
#'
#' # load demo data
#' data(sp1)
#'
#' # promote to SPC
#' depths(sp1) <- id ~ top + bottom
#'
#' # plot
#' plot(sp1)
#'
#' # number of profiles
#' length(sp1)
#'
#' # number of horizons
#' nrow(sp1)
#'
setReplaceMethod("depths", "data.frame",
                           # or a data.frame like obj
  function(object, value) {
    if (inherits(value, "formula")) { # initialization by formula:  id ~ top + bottom
      mf <- model.frame(value, object)
      res <- .initSPCfromMF(data = object, mf = mf)
    } else {
      if (inherits(value, "character") &
          inherits(object, 'data.frame')) { # initialization by colnames: c("id","top","bottom")
	      mf <- .data.frame.j(object, value, class(object)[1])
	      res <- .initSPCfromMF(data = object, mf = mf)
      } else {
	      stop('invalid initial object for promotion to SoilProfileCollection', call. = FALSE)
      }
    }

    # add default site data: profile IDs in site same order as horizons
    site.temp <- data.frame(id = profile_id(res), stringsAsFactors = FALSE)
    names(site.temp) <- idname(res)
    adc <- aqp_df_class(res)
    res@site <- .as.data.frame.aqp(site.temp, adc)
    res@horizons <- .as.data.frame.aqp(res@horizons, adc)

    # done
    return(res)
  }
)

.checkNAdepths <- function(depth, l) {
  if(is.factor(depth)) {
    warning(sprintf("Horizon %s depth is a factor! This happens with automatic character to factor conversions!", l), call. = FALSE)

    # this ensures numeric conversion uses factor labels, not levels
    depth <- as.character(depth)
  }
  if(!is.numeric(depth)) {
    depth <- suppressWarnings(try(as.numeric(depth)))
    if(!inherits(depth, 'try-error')) {
      message(sprintf("Horizon %s depths converted to numeric.", l))
    } else stop(sprintf("Unable to convert %s depths to numeric!", l))
  }
  if(any(is.na(depth)))
    warning(sprintf("Horizon %s depths contain NA! Check depth logic with aqp::checkHzDepthLogic()", l), call. = FALSE)
  return(depth)
}

.checkDepthOrder <- function(x, depthcols) {
  if (any(x[[depthcols[2]]] < x[[depthcols[1]]], na.rm = TRUE)) {
    warning("One or more horizon bottom depths are shallower than top depth. Check depth logic with aqp::checkHzDepthLogic()", call. = FALSE)
  }
}

.screenDepths <- function(x, depthcols = horizonDepths(x)) {
  .checkNAdepths(x[[depthcols[1]]], "top")
  .checkNAdepths(x[[depthcols[2]]], "bottom")
  .checkDepthOrder(x, depthcols)
}

# create 0-length spc from id and horizon depth columns (`idn`, `hzd`)
#  - allows template horizon (`hz`) and site (`st`) data to be provided (for additional columns)
.prototypeSPC <- function(idn, hzd, 
                          hz = data.frame(), st = data.frame(), 
                          fill_top = NA_integer_, fill_bottom = NA_integer_) {
  
  # create bare minimum data
  id <- hz[[idn]]
  nid <- length(id)
  iid <- seq_along(id)
  nuhz <- data.frame(id = as.character(id), 
                     hzID = iid, 
                     top = fill_top[nid],
                     bottom = fill_bottom[nid], 
                     stringsAsFactors = FALSE)
  # use idname/horizon depths
  colnames(nuhz) <- c(idn, "hzID", hzd)
  
  # dummy site data
  nust <- data.frame(id = as.character(id), 
                     stringsAsFactors = FALSE)
  colnames(nust) <- idn
  
  # add other columns
  if (ncol(hz) > 0) {
    hz$.dummyVar <- ""[nrow(hz)]
    nuhz <- cbind(nuhz, .create_placeholder_df(hz[, !colnames(hz) %in% colnames(nuhz), drop = FALSE], length(iid)))
    nuhz$.dummyVar <- NULL
  }
  
  if (ncol(st) > 0) {
    st$.dummyVar <- ""[nrow(st)]
    nust <- cbind(nust, .create_placeholder_df(st[, !colnames(st) %in% colnames(nust), drop = FALSE], length(iid)))
    nust$.dummyVar <- NULL
  }
    
  # return 0-length or n-length (ID only) SPC
  return(SoilProfileCollection(idcol = idn, 
                               depthcols = hzd,
                               horizons = nuhz,
                               site = nust))
}

##
## initialize SP/SPC objects from a model.frame
## - data is a horizon-level data.frame
## - mf is a model frame expressing the id ~ top + bottom formula
## - use_class is the data.frame class to use ("data.frame", "tbl_df", or "data.table")
## 
# TODO: top and bottom are depth used to fill in an empty horizon for profiles where all top depths are missing. Default `NA_integer_`. Not exposed through depths<- at this time because it only covers the case where every single profile in the collection has all missing top depths. Doing more involved inspection of the SPC inputs to find cases that may have multiple NA-horizons per profile, removing extras, splicing in clean NA-filled data may be more than we want to do in the object constructing method
.initSPCfromMF <- function(data, mf, use_class, top = NA_integer_, bottom = NA_integer_) {
  # get column names containing id, top, bottom
  nm <- names(mf)

  # check for factor-class ID
  if (inherits(data[[nm[1]]], 'factor')) {
    message('converting profile IDs from factor to character')
    data[[nm[1]]] <- as.character(data[[nm[1]]])
  }

  # check for integer IDs
  if (inherits(data[[nm[1]]], 'integer')) {
    message('converting profile IDs from integer to character')
    data[[nm[1]]] <- as.character(data[[nm[1]]])
  }

  # depth column names
  depthcols <- c(nm[2], nm[3])

  # get column containing IDs
  data_id <- data[[nm[1]]]

  # check for all depths NA or 0-length
  if (all(is.na(data[[depthcols[1]]]))) {
    # all top depths missing is intractable, all bottom depths is fixable
    warning("all top depths missing from input data", call. = FALSE)
    # return a filled (n-length) SPC prototype from the horizon data
    return(.prototypeSPC(nm[1], depthcols, hz = data, fill_top = top, fill_bottom = bottom))
  } else if(nrow(data) == 0) {
    warning("input data have 0 rows", call. = FALSE)
    # return a empty (0-length) SPC prototype from the horizon data
    return(.prototypeSPC(nm[1], depthcols, hz = data, fill_top = top, fill_bottom = bottom))
  }
  
  # enforce numeric depths and provide QC warnings as needed
  data[[depthcols[1]]] <- .checkNAdepths(data[[depthcols[1]]], "top")
  data[[depthcols[2]]] <- .checkNAdepths(data[[depthcols[2]]], "bottom")
  
  # warn if bottom depth shallower than top (old style O horizons, data entry issues, etc.)
  .checkDepthOrder(data, depthcols)
  
  tdep <- data[[depthcols[1]]]

  # calculate ID-top depth order, re-order input data
  id_tdep_order <- order(as.character(data_id), tdep)
  data <- data[id_tdep_order,]

  # create a site table with just IDs
  nusite <- .as.data.frame.aqp(data.frame(rle(data[[nm[1]]])$values,
                                          stringsAsFactors = FALSE), class(data)[1])
  names(nusite) <- nm[1]

  # create object
  res <- SoilProfileCollection(idcol = nm[1],
                               hzidcol = 'hzID',
                               depthcols = depthcols,
                               site = nusite,
                               horizons = data)

  
  # check for horizon ID name conflict
  if (hzidname(res) %in% names(data)) {

    # original hz ID
    o.hzid <- hzidname(res)

    # is this a good candidate horizon ID?
    res.status <- try(hzID(res) <- data[[o.hzid]], silent = TRUE)

    # if not, re-make one
    if(inherits(res.status, 'try-error')) {
      # add unique horizon IDs to a new column
      n.hzid <- sprintf("%s_", o.hzid)

      # add non-conflicting hz ID
      res@horizons[[n.hzid]] <- as.character(1:nrow(res))

      # update object
      hzidname(res) <- n.hzid

      # notify
      warning(sprintf('`%s` is not a unique horizon ID, using `%s`', o.hzid, n.hzid), call. = FALSE)
    } else {
      # notify that everything is fine
      message(sprintf('using `%s` as a unique horizon ID', o.hzid))
    }

  } else {
    # no conflict, add a reasonable horizon ID
    hzID(res) <- as.character(1:nrow(res))
  }

  # done
  return(res)
}

setGeneric('site<-', function(object, value)
  standardGeneric('site<-'))

#' Create or Add Data to Site Slot
#'
#' @description
#' There are two options available via the \code{site<-} setter.
#'
#' The first is a "normalization" by formula interface, whereby one specifies an attribute that is constant in horizons within profiles to be promoted to a site-level variable: \code{site(spc) <- ~ horizonvariable}
#'
#' The second is creation of site data from an external \code{data.frame} via merge (LEFT JOIN). There must be one or more same-named columns (with at least some matching data) on the left and right hand side to facilitate the join: \code{site(spc) <- newdata}
#'
#' @param object A SoilProfileCollection
#' @param value A formula or object inheriting \code{data.frame}
#' @rdname site
#' @aliases site<-
#' @export
#' @examples
#'
#' # load test data
#' data(sp2)
#'
#' # promote to SPC
#' depths(sp2) <- id ~ top + bottom
#'
#' # normalize a horizon-level attribute to site
#' site(sp2) <- ~ surface
#'
#' # inspect site table
#' site(sp2)
#'
#' # make some data: classify two geomorphic surfaces with numeric value
#' newdata <- data.frame(surface = c("holocene",
#'                                   "lower riverbank"),
#'                       newvalue = c(1,2))
#'
#' # do left join based on newly-normalized "surface" attribute
#' site(sp2) <- newdata
#'
#' # inspect site table: holocene & lower riverbank have values
#' site(sp2)
#'
setReplaceMethod("site", signature(object = "SoilProfileCollection"),
  function(object, value) {

    # get profile IDs from horizon table
    ids <- as.character(horizons(object)[[idname(object)]])
    pid <- rle(ids)$values
    idn <- idname(object)
    adf <- aqp_df_class(object)
    
  	# creation of site data from horizon data
    if (inherits(value, "formula")) {
      mf <- model.frame(value, object@horizons, na.action = na.pass)
      nm <- names(mf)
      mf <- data.frame(ids, mf, stringsAsFactors = FALSE)
      names(mf) <- c(idn, nm)
      object <- .createSiteFromHorizon(object, mf)
    }

    # creation of site data from an external data.frame via merge (LEFT JOIN)
    if (inherits(value, "data.frame")) {
      # get column names from proposed site, and existing horizons
      new.cols <- names(value)
      hz.names <- horizonNames(object)
      si.names <- siteNames(object)
      
      # short circuit: ensure that site(x)$colname <- NULL works
      if (all(new.cols %in% si.names) &&
               idn %in% new.cols && 
                nrow(value) == length(object)) {
        
          if (!all(value[[idn]] %in% pid)) {
            message("Some profile IDs in input data are not present in object and no new columns to merge. Doing nothing.")
            return(object)
          }
        
          # re-sorts for case when "joining" only ID 
          sort.idx <- match(pid, value[[idn]])
          object@site <- .as.data.frame.aqp(value[sort.idx,, drop = FALSE], adf)
          return(object)
      }
      
      # short circuit: other site(x)$colname cases that do not need a merge
      #   (only one new column, in existing site order)
      difcols <- setdiff(new.cols, si.names)
      if (length(difcols) == 1 &&
          idn %in% new.cols && 
          all(value[[idn]] == pid)) {
        object@site[[difcols]] <- value[[difcols]]
        return(object)
      }
      
      # check there is >1 column shared b/w existing site and value
      if (length(new.cols) > 1 && !any(new.cols %in% si.names)) {
        stop("new data must have one or more column names in common with the site data", call. = FALSE)
      }
      
      ## remove ID column from names(horizons)
      ID.idx <- match(idn, hz.names)

      # check to make sure there is no overlap in proposed site + hz variable names
      if (any(new.cols %in% hz.names[-ID.idx])) {
        stop('duplicate names in new site / existing horizon data not allowed', call. = FALSE)
      }
      
      # existing site data
      s <- object@site
      
      # LEFT JOIN to existing data on all matching columns
      site.new <- merge(s, value, all.x = TRUE, sort = FALSE)
      
      new.id.order <- site.new[[idn]]
      if (length(new.id.order) != length(pid) || any(new.id.order != pid)) {
        # join condition resulted in sorting of sites, re-apply original order
        if (any(is.na(pid))) {
          warning("profile IDs derived from horizon data contain NA!", call. = FALSE)
        }
        site.new <- site.new[match(pid, new.id.order), ]
      }

      # sanity check: site + new data should have same number of rows as original
      if (nrow(s) != nrow(site.new)) {
      	message(paste('original data (', nrow(s), ' rows) new data (',
      	              nrow(site.new), ' rows)', sep = ''))
        stop('invalid join condition, site data not changed', call. = FALSE)
      }

      # 2020-05-30: subclasses of data.frame have more than one class
      object@site <- .as.data.frame.aqp(site.new, adf)
	  }

    # done
    return(object)
  }
)

# update an SPC object:
# add site data
# remove named columns from horizons
# return new SPC object
.createSiteFromHorizon <- function(object, mf){

  # create a numeric index for named site columns, as we will remove them
  # from the horizon data
  names_attr <- names(mf)
  idx <- match(names_attr, horizonNames(object))

  # remove the index to the ID column, as we do not want to remove this from
  # the horizon data !
  idx <- idx[-match(idname(object), names_attr)]

  .SD <- NULL
  
  dth <- data.table::as.data.table(horizons(object))
  
  new_site_data <- .as.data.frame.aqp(unique(dth[, .SD, .SDcols = names_attr]), aqp_df_class(object))
  
  if (nrow(new_site_data) != length(object)) {
    warning("One or more horizon columns cannot be normalized to site. Leaving site data unchanged.", call. = FALSE)
    return(object)
  }
  
  # if site data is already present, we don't overwrite/erase it
  site_data <- merge(object@site, new_site_data, by = idname(object), all.x = TRUE, sort = FALSE)

  # remove the named site data from horizon_data
  h <- object@horizons
  hnames <- colnames(h)
  for(i in idx) {
    h[[hnames[i]]] <- NULL
  }

  object@horizons <- .as.data.frame.aqp(h, aqp_df_class(object))

  # replace existing site data
  object@site <- .as.data.frame.aqp(site_data, aqp_df_class(object))

  # done
  return(object)
}


setGeneric('replaceHorizons<-', function(object, value)
  standardGeneric('replaceHorizons<-'))

#' @title Replace Data in Horizon Slot
#'
#' @name replaceHorizons<-
#'
#' @description Replaces horizon data with new `data.frame` object.
#'
#' @param object A `SoilProfileCollection`
#' 
#' @param value An object inheriting `data.frame`
#' 
#' @aliases replaceHorizons<-,SoilProfileCollection-method
#' @docType methods
#'
#' @rdname replaceHorizons
#' @export
#'
#' @examples
#'
#' # load test data
#' data(sp2)
#'
#' # promote to SPC
#' depths(sp2) <- id ~ top + bottom
#'
#' # one profile
#' p <- sp2[1,]
#'
#' # 23 variables in horizon data
#' length(horizonNames(sp2))
#'
#' # remove all but essential ones
#' replaceHorizons(p) <- horizons(p)[,c(idname(p), hzidname(p), horizonDepths(p))]
#'
#' # inspect result (a clean slate)
#' horizons(p)
#'
setReplaceMethod("replaceHorizons",
                 signature(object = "SoilProfileCollection"),
                 function(object, value) {

  hmn <- .hzMetadataNames(object, depths = TRUE)
  required.columns <-  c(idname(object), horizonDepths(object))
  required.missing <- !required.columns %in% names(value)

  if (any(required.missing)) {
    stop(paste0("required horizon data are missing from replacement: ",
                paste0(required.columns[required.missing], 
                       collapse = ", ")), call. = FALSE)
  }
  
  ids.match1 <- all(profile_id(object) %in% value[[idname(object)]])
  if (!ids.match1) {
    stop("profile IDs in site are missing from replacement horizons!", call. = FALSE)
  }
  
  ids.match2 <- all(value[[idname(object)]] %in% profile_id(object))
  if (!ids.match2) {
    stop("profile IDs in replacement are missing from site!", call. = FALSE)
  }
  
  optional.columns <-  hmn[!hmn %in% required.columns]
  optional.missing <- !optional.columns %in% names(value)

  if (any(optional.missing)) {
    message(paste0("optional columns are missing from replacement: ",
                   paste0(optional.columns[optional.missing],
                          collapse = ", ")))
  }

  # assign hzID if hzidname() is missing
  if (optional.missing[1]) {
    value$hzID <- as.character(1:nrow(value))
    hzidname(object) <- "hzID"
    message("no horizon ID present, defaulting to `hzID`")
  }

  object@horizons <- .as.data.frame.aqp(value, aqp_df_class(object))
  return(object)
})


setGeneric('horizons<-', function(object, value)
  standardGeneric('horizons<-'))

#' Join Data to Horizon Slot
#'
#' @name horizons<-
#'
#' @description
#' Horizon data in an object inheriting from \code{data.frame} can easily be added via merge (LEFT JOIN). There must be one or more same-named columns (with at least some matching data) on the left and right hand side to facilitate the join: \code{horizons(spc) <- newdata}
#'
#' @param object A SoilProfileCollection
#' @param value An object inheriting \code{data.frame}
#' @aliases horizons<-,SoilProfileCollection-method
#' @docType methods
#' @export
#' @rdname horizons
#'
#' @examples
#'
#' # load test data
#' data(sp2)
#'
#' # promote to SPC
#' depths(sp2) <- id ~ top + bottom
#'
#' # assign true to surface horizon
#' newdata <- data.frame(top = 0,
#'                       newvalue = TRUE)
#'
#' # do left join
#' horizons(sp2) <- newdata
#'
#' # inspect site table: newvalue TRUE only for horizons
#' #  with top depth equal to zero
#' horizons(sp2)
#'
setReplaceMethod("horizons", signature(object = "SoilProfileCollection"),
  function(object, value) {

  idn <- idname(object)
  hdn <- hzidname(object)
  hzd <- horizonDepths(object)
  
  if (is.null(value))
    stop("new horizon data must not be NULL; to remove a site or horizon attribute use `spc$attribute <- NULL`", call. = FALSE)

  # testing the class of the horizon data to add to the object
  if (!inherits(value, "data.frame"))
    stop("new horizon data input value must inherit from data.frame", call. = FALSE)

  # allow short-circuit (handles NULL and non-op without going thru merge())
  if ((all(horizonNames(object) %in% colnames(value)) ||
       all(colnames(value) %in% horizonNames(object))) &&
       all(c(idn, hdn, hzd) %in% colnames(value)) &&
      nrow(value) == nrow(object)) {
    if (!all(value[[idn]] %in% profile_id(object))) {
      message("Some profile IDs in input data are not present in object and no new columns to merge. Doing nothing.")
      return(object)
    }
    target.order <- order(object@horizons[[idn]], object@horizons[[hzd[1]]])
    input.order <- order(value[[idn]], value[[hzd[1]]])
    idx.order <- match(input.order, target.order)
    # print(idx.order)
    object@horizons <- .as.data.frame.aqp(value, aqp_df_class(object))[idx.order,]
    return(object)
  }

  # get the corresponding vector of IDs, will be used to compute distinct attributes
  ids <- as.character(horizons(object)[[idn]])

  # get column names from proposed horizons, and existing site
  ns <- names(value)
  nh <- siteNames(object)

  # check there is >1 column shared b/w existing horizon and value
  if (length(ns) > 1 && !any(ns %in% horizonNames(object))) {
    stop("new data must have one or more column names in common with the horizon data", call. = FALSE)
  }
  
  ## remove ID column from names(site)
  ID.idx <- match(idn, nh)

  # check to make sure there is no overlap in proposed site + hz variable names
  if (any(ns %in% nh[-ID.idx])) {
    stop('duplicate names in new horizon / existing site data not allowed', call. = FALSE)
  }
  
  h.id <- as.character(object@horizons[[hdn]])
  original.site.order <- match(rle(object@site[[idn]])$values,
                               object@site[[idn]])

  # in keeping with tradition of site<-, we let the join happen
  # left join to existing data
  suppressMessages({
    horizon.new <- merge(object@horizons,
                         value,
                         #by = c(idname(object), hzidname(object)),
                         all.x = TRUE, sort = FALSE)
  })
  
  # TODO: data.table merge() would fix the need for re-sorting, but would check data types
  #       which might cause some backwards compatibility problems (requiring type conversion)
  #       this is a problem because the join is completely open ended -- not just based on ID
  
  chnew <- rle(horizon.new[[idn]])$values
  if (length(chnew) != length(original.site.order) |
     suppressWarnings(any(original.site.order != chnew))) {
    new.horizon.order <- order(horizon.new[[idn]], 
                               horizon.new[[hzd[1]]])
    # message("join condition resulted in sorting of horizons, re-applying original order")
    horizon.new <- horizon.new[new.horizon.order,]
  }

  # sanity check: horizons + new data should have same number of rows as original
  if (nrow(object@horizons) != nrow(horizon.new)) {
    message(paste('original data (', nrow(object@horizons), ' rows) new data (', nrow(horizon.new), ' rows)', sep = ''))
    stop("invalid horizons left join condition, data not changed", call. = FALSE)
  }

  # 2020-05-30: subclasses of data.frame have more than one class
  object@horizons <- .as.data.frame.aqp(horizon.new, aqp_df_class(object))

  # check to make sure same profile IDs are present
  if (any(!(ids %in% as.character(object@horizons[[idn]])))) {
    print(paste('pedons (', nrow(object), ') rows of horizon data (', nrow(object@horizons), ')', sep = ''))
    stop('profile IDs are missing from join result, data not changed', call. = FALSE)
  }

  # done
  return(object)
})

setGeneric('diagnostic_hz<-', function(object, value)
  standardGeneric('diagnostic_hz<-'))

#' @name diagnostic_hz<-
#'
#' @description 
#' 
#'  - `diagnostic_hz<-` (set method): Set diagnostic feature data for a SoilProfileCollection. The profile ID column from `object` (`idname(object)`) must be present in the replacement `value` object.
#' 
#' @param object A SoilProfileCollection
#' @param value An object inheriting from \code{data.frame}
#'
#' @aliases diagnostic_hz<-,SoilProfileCollection-method
#' @docType methods
#' @export
#' @rdname diagnostic_hz
#'
#' @examples
#'
#' # load test data
#' data(sp2)
#'
#' # promote to SPC
#' depths(sp2) <- id ~ top + bottom
#'
#' # assign two profiles a zone related to the mollic epipedon
#' newdata <- data.frame(id = c("hon-1","hon-17"),
#'                       featkind = "fixed-depth surface sample",
#'                       featdept = 0,
#'                       featdepb = 18)
#'
#' # do left join
#' diagnostic_hz(sp2) <- newdata
#'
#' # inspect site table: newvalue TRUE only for horizons
#' #  with top depth equal to zero
#' diagnostic_hz(sp2)
#'
setReplaceMethod("diagnostic_hz",
                 signature(object = "SoilProfileCollection"),
  function(object, value) {

  # get the initial data
  d <- diagnostic_hz(object)

  # get column and ID names
  nm <- names(value)
  idn <- idname(object)
  pIDs <- profile_id(object)

  # testing the class of the new data
  if (!inherits(value, "data.frame"))
    stop("diagnostic horizon data must be a data.frame", call. = FALSE)

  # test for the special case where internally-used functions
  # are copying over data from one object to another, and diagnostic_hz(obj) is a 0-row data.frame
  # short-circut, and return original object
  if (nrow(d) == 0 && nrow(value) == 0)
  	return(object)

  # test to make sure that our common ID is present in the new data
  if (!idn %in% nm)
    stop(paste("diagnostic horizon data are missing pedon ID column: ", idn), call. = FALSE)

  uidm <- unique(value[[idn]]) %in% pIDs
  # warn user if some of the IDs in the candidate data are missing
  if (any(!uidm)) {
    # test to make sure that at least one of the IDS in candidate data are present within SPC
    if (all(!uidm)) {
      warning('candidate diagnostic horizon data have NO matching IDs in target SoilProfileCollection object!', call. = FALSE)
    } else warning('some records in candidate diagnostic horizon data have no matching IDs in target SoilProfileCollection object', call. = FALSE)
  }

  # if data are already present, warn the user
  if (nrow(d) > 0)
  	warning('overwriting existing diagnostic horizon data!', call. = FALSE)
  
  # convert id column to character to match @site
  value[[idn]] <- as.character(value[[idn]])
  
  # copy data over
  object@diagnostic <- .as.data.frame.aqp(value, metadata(object)$aqp_df_class)

  # done
  return(object)
})

setGeneric('restrictions<-', function(object, value)
  standardGeneric('restrictions<-'))

#' @name restrictions<-
#'
#' @description 
#' 
#'  - `restrictions<-` (set method): Set restriction data for a SoilProfileCollection. The profile ID column from `object` (`idname(object)`) must be present in the replacement `value` object.
#' @param object A SoilProfileCollection
#' @param value An data.frame object containing at least a column with name `idname(object)`
#'
#' @aliases restrictions<-,SoilProfileCollection-method
#' @docType methods
#'
#' @rdname restrictions
#' @export
#' @examples
#'
#' # load test data
#' data(sp2)
#'
#' # promote to SPC
#' depths(sp2) <- id ~ top + bottom
#'
#' # assign abrupt textural change to a profile
#' newdata <- data.frame(id = c("hon-21"),
#'                       restrkind = "abrupt textural change",
#'                       restrdep = 46)
#'
#' # do left join
#' restrictions(sp2) <- newdata
#'
#' # inspect site table: newvalue TRUE only for horizons
#' #  with top depth equal to zero
#' restrictions(sp2)
#'
setReplaceMethod("restrictions", signature(object = "SoilProfileCollection"),
                 function(object, value) {

                   # get the initial data
                   d <- restrictions(object)

                   # get column and ID names
                   nm <- names(value)
                   idn <- idname(object)
                   pIDs <- profile_id(object)

                   # testing the class of the new data
                   if (!inherits(value, "data.frame"))
                     stop("restriction data must be a data.frame", call. = FALSE)

                   # test for the special case where internally-used functions
                   # are copying over data from one object to another, and diagnostic_hz(obj) is a 0-row data.frame
                   # short-circuit, and return original object
                   if (nrow(d) == 0 && nrow(value) == 0)
                     return(object)

                   # test to make sure that our common ID is present in the new data
                   if (!idn %in% nm)
                     stop(paste("restriction data are missing pedon ID column: ", idn), call. = FALSE)
                   
                   uidm <- unique(value[[idn]]) %in% pIDs
                   # warn user if some of the IDs in the candidate data are missing
                   if (any(!uidm)) {
                     # test to make sure that at least one of the IDs in candidate data are present within SPC
                     if (all(!uidm)) {
                       warning('restriction data have no matching IDs in target SoilProfileCollection object!', call. = FALSE)
                     } else warning('some records in restriction data have no matching IDs in target SoilProfileCollection object', call. = FALSE)
                   }

                   # if data are already present, warn the user

                   if (nrow(d) > 0)
                     warning('overwriting existing restriction data!', call.=FALSE)
                   
                   # convert id column to character to match @site
                   value[[idn]] <- as.character(value[[idn]])
                  
                   # copy data over
                   object@restrictions <- .as.data.frame.aqp(value, metadata(object)$aqp_df_class)

                   # done
                   return(object)
                 }
)

test_that(".prototypeSPC works", {
  # basic functionality: create a 0 length SPC with specified id ~ top + bottom
  x <- .prototypeSPC("id", c("top", "bottom"))
  expect_true(inherits(x, 'SoilProfileCollection'))
  expect_equal(length(x), 0)
  expect_equal(names(x), c(horizons1 = "id", horizons2 = "hzID", 
                           horizons3 = "top", horizons4 = "bottom"))
  
  # one additional horizon column works (degenerate case)
  x1 <- .prototypeSPC("id", c("top", "bottom"), 
                      hz = data.frame(id = 1, top = NA, bottom = NA, foo = "bar")[0,])
  x2 <- .prototypeSPC("id", c("top", "bottom"),  
                      hz = data.frame(id = 1:10, top = NA, bottom = NA, foo = "bar"))
  expect_equal(length(x1), 0) # zero length in, zero length out
  expect_true(length(x2) == 10 && nrow(x2) == 10) # all NA depths -> 1 NA row / profile
  expect_true("foo" %in% horizonNames(x1) && "foo" %in% horizonNames(x2))
  
  # two or more additional column (site and horizon) with depth filling works
  x3 <- .prototypeSPC("id", c("top", "bottom"), 
                      hz = data.frame(id = 1, top = NA, bottom = NA, foo = "bar", baz = "qux"),
                      st = data.frame(id = 1, sitevar = "eroded"),
                      fill_top = 0, fill_bottom = 200)
  expect_equal(length(x3), 1)
  expect_true(all(c("foo", "baz") %in% horizonNames(x3)))
  expect_true(all(c("id", "sitevar") %in% siteNames(x3)))
  expect_true(x3$top == 0 & x3$bottom == 200)
})

# setup a new environment to store error messages, etc.
aqp.env <- new.env(hash = TRUE, parent = parent.frame())

# register options for later use
.onLoad <- function(libname, pkgname) {
  options(.aqp.show.n.cols = 10)
}

# no longer needed since it is imported
# # 2020-07-10: allows for data.table @Suggests without importing
# # https://cran.r-project.org/web/packages/data.table/vignettes/datatable-importing.html
# .datatable.aware <-  TRUE

# 2020-05-30: make data.table, tbl_df and data.frame slots "co-exist"
# see: https://stackoverflow.com/questions/35642191/tbl-df-with-s4-object-slots
if(requireNamespace("data.table", quietly = TRUE))
  setOldClass(c("data.table", "data.frame"))

if(requireNamespace("tibble", quietly = TRUE))
  setOldClass(c("tbl_df", "tbl", "data.frame"))

## TODO: documentation needs a lot more detail!


## TODO: relationship to .pctMissing (1-unit slice eval)

## missing data fraction computed on a thickness basis
## this is meant to be run on a single profile at a time
# x: single SPC
# v: variables to consider
# n: column containing horizon designations
# p: pattern matching non-soil horizons
# m: method, 'relative', 'absolute', 'horizon' 
.getMissingDataFraction <- function(x, v, n, p, m) {
  # get horizons
  h <- horizons(x)
  
  # get top/bottom
  dc <- horizonDepths(x)
  
  # get all "soil" horizons, for named variables
  soil.hz <- grep(pattern = p, x = as.character(h[[n]]), invert = TRUE)
  
  # compute fraction of data based on:
  
  # fraction of each horizon
  # this requires a little more work
  if(m == 'horizon') {
    
    # select variables, use all horizons
    v <- as.vector(na.omit(match(v, horizonNames(x))))
    d <- h[, v]
    
    d <- as.matrix(d)
    res <- apply(d, 1, function(i) {
      length(which(complete.cases(i))) / ncol(d)
    })
    
    # set non-soil hz to NA
    res[-soil.hz] <- NA
    
  } else {
    # data fraction based on thickness
    hz.thick <- h[[dc[2]]] - h[[dc[1]]]
    
    # extract named columns, but only those from @horizons
    v <- as.vector(na.omit(match(v, horizonNames(x))))
    d <- h[soil.hz, v]
    
    # mask out non-soil horizons
    hz.thick <- hz.thick[soil.hz]
    
    # index hz with full data population
    hz.with.data <- which(complete.cases(d))
    
    # relative thickness
    # absolute thickness
    total <- switch(m,
                    'relative' = sum(hz.thick, na.rm = TRUE),
                    'absolute' = 1
    )
    
    res <- sum(hz.thick[hz.with.data], na.rm = TRUE) / total
  }
  

  return(res)
}


## iterate over profiles
# x: SPC with >=1 profiles
# vars: variables to consider
# name: horizon designations
# p: inverted pattern matching non-soil horizons
# method: 'relative' or 'absolute' reporting

#' @title Evaluate Missing Data within a SoilProfileCollection
#'
#' @description Evaluate missing data within a `SoilProfileCollection` object
#'
#' Data completeness is evaluated by profile or by horizon. Profile-level evaluation is based on the thickness of horizons (`method = absolute`) with complete horizon-level attributes (`vars`), optionally divided by the total thickness (`method = relative`). The REGEX pattern (`p`) is used to filter non-soil horizons from the calculation.
#'
#' @param x `SoilProfileCollection` object
#' 
#' @param vars character vector, naming horizon-level attributes in `x`
#' 
#' @param name character, the name of a horizon-level attribute where horizon designations are stored, defaults to `hzdesgnname(x)`
#' 
#' @param p character, REGEX pattern used to match non-soil horizons
#' 
#' @param method character, one of: 'relative' (proportion of total) depth, 'absolute' depth, or 'horizon' (fraction not-missing by horizon)
#' 
#' @return A vector values ranging from 0 to 1 (`method = 'relative'`) or 0 to maximum depth in specified depth units (`method = 'absolute'`) representing the quantity of non-missing data (as specified in `vars`) for each profile. When `method = 'horizon'` a non-missing data fraction is returned for each horizon.
#' 
#' @author D.E. Beaudette
#' @keywords manip
#' @export
#' @examples
#' 
#' # example data
#' data("jacobs2000")
#' 
#' # fully populated
#' plotSPC(jacobs2000, name.style = 'center-center', 
#'         cex.names = 0.8, color = 'time_saturated')
#' 
#' # missing some data
#' plotSPC(jacobs2000, name.style = 'center-center', 
#'         cex.names = 0.8, color = 'concentration_color')
#' 
#' # very nearly complete
#' plotSPC(jacobs2000, name.style = 'center-center', 
#'         cex.names = 0.8, color = 'matrix_color')
#' 
#' 
#' # variables to consider
#' v <- c('time_saturated', 'concentration_color', 'matrix_color')
#' 
#' # compute data completeness by profile
#' # ignore 2C horizons
#' jacobs2000$data.complete <- evalMissingData(
#'   jacobs2000, 
#'   vars = v, 
#'   method = 'relative',
#'   p = '2C'
#' )
#' 
#' jacobs2000$data.complete.abs <- evalMissingData(
#'   jacobs2000, 
#'   vars = v, 
#'   method = 'absolute',
#'   p = '2C'
#' )
#' 
#' # compute data completeness by horizon
#' # ignore 2C horizons
#' jacobs2000$hz.data.complete <- evalMissingData(
#'   jacobs2000, 
#'   vars = v, 
#'   method = 'horizon',
#'   p = '2C'
#' )
#' 
#' 
#' # "fraction complete" by horizon
#' plotSPC(
#'   jacobs2000, name.style = 'center-center', 
#'   cex.names = 0.8, color = 'hz.data.complete'
#' )
#' 
#' 
#' # rank on profile completeness
#' new.order <- order(jacobs2000$data.complete)
#' 
#' # plot along data completeness ranking
#' plotSPC(
#'   jacobs2000, name.style = 'center-center', 
#'   cex.names = 0.8, color = 'concentration_color', 
#'   plot.order = new.order
#' )
#' 
#' # add relative completeness axis
#' # note re-ordering of axis labels
#' axis(
#'   side = 1, at = 1:length(jacobs2000), 
#'   labels = round(jacobs2000$data.complete[new.order], 2),
#'   line = 0, cex.axis = 0.75
#' )
#' 
#' # add absolute completeness (cm)
#' axis(
#'   side = 1, at = 1:length(jacobs2000), 
#'   labels = jacobs2000$data.complete.abs[new.order],
#'   line = 2.5, cex.axis=0.75
#' )
#' 



evalMissingData <- function(x, vars, name = hzdesgnname(x), p = 'Cr|R|Cd', method = c('relative', 'absolute', 'horizon')) {
  # sanity checks
  method <- match.arg(method)

  # check for missing horizon name
  if(name == '') {
    stop('horizon name not set, specify with `name` argument or set with `hzdesgnname()`', call. = FALSE)
  }
  
  # warn about empty pattern
  if(p == '') {
    warning('an empty pattern `p` will ignore all horizons')
  }
  
  # extract pieces
  hn <- horizonNames(x)
  sdv <- setdiff(vars, hn)

  # check for bad horizon name
  if(! name %in% hn) {
    stop(sprintf('`%s` not a horizon level attribute', name), call. = FALSE)
  }

  # check for incorrectly specified horizon level data
  if(length(sdv > 1)) {
    stop(sprintf('`%s` not horizon level attribute(s)', paste(sdv, collapse = ',')), call. = FALSE)
  }

  # apply missing data fraction eval to each profile
  md <- profileApply(x, .getMissingDataFraction, v = vars, n = name, p = p, m = method)
  return(md)
}

context("evalMissingData")

# setup horizon-level data: data are from lab sampled pedons
# sprinkle-in some NA
# added fake contact to the bottom via R, Cr, Cd
d <- read.csv(
  textConnection('series,name,top,bottom,clay,frags,ph
auburn,A,0,3,21,6,NA
auburn,AB,3,15,21,13,5.6
auburn,Bw,15,25,20,9,5.8
auburn,R,25,47,21,28,5.8
dunstone,A,0,5,16,13,6
dunstone,AB,5,17,17,19,6.3
dunstone,Bt,17,31,20,6,6.3
dunstone,Cr,31,41,21,15,6.3
sobrante,A,0,5,18,0,5.8
sobrante,Ab,5,10,16,2,5.7
sobrante,Bt1,10,28,15,21,5.8
sobrante,Bt2,28,51,NA,NA,NA
sobrante,Cd,51,74,20,12,6.2'), stringsAsFactors = FALSE)

# establish site-level data
s <- data.frame(
  series = c('auburn', 'dunstone', 'sobrante'), 
  precip = c(24, 30, 32),
  stringsAsFactors = FALSE
)

# upgrade to SoilProfile Collection object
depths(d) <- series ~ top + bottom
site(d) <- s


## tests

test_that("runs as expected", {
  
  # does it run?
  e.rel <- evalMissingData(d, vars = c('clay', 'frags', 'ph'), name = 'name', method = 'relative')
  e.abs <- evalMissingData(d, vars = c('clay', 'frags', 'ph'), name = 'name', method = 'absolute')
  e.hz <- evalMissingData(d, vars = c('clay', 'frags', 'ph'), name = 'name', method = 'horizon')
  
  # result is a numeric vector
  expect_true(inherits(e.rel, 'numeric'))
  expect_true(inherits(e.abs, 'numeric'))
  expect_true(inherits(e.hz, 'numeric'))
  
  # there should be no NA in profile-level summaries
  expect_true(all(is.na(e.rel) == FALSE))
  expect_true(all(is.na(e.abs) == FALSE))
  
  # expected lengths
  expect_true(length(e.rel) == length(d))
  expect_true(length(e.abs) == length(d))
  expect_true(length(e.hz) == nrow(d))
  
  # check against hand-computed results
  # as.vector removes names attr
  # auburn
  expect_equal(as.vector(e.rel[1]), 22 / 25, tolerance = 0.001)
  expect_equal(as.vector(e.abs[1]), 22 , tolerance = 0.001)
  # first hz of auburn is missing pH
  expect_equal(as.vector(e.hz[1]), 2/3, tolerance = 0.001)
  
  
  # dunstone
  expect_equal(as.vector(e.rel[2]), 1.000, tolerance = 0.001)
  expect_equal(as.vector(e.abs[2]), 31, tolerance = 0.001)
  
  # sobrante
  expect_equal(as.vector(e.rel[3]), 28 / 51, tolerance=0.001)
  expect_equal(as.vector(e.abs[3]), 28 , tolerance=0.001)
  
})


test_that("expected errors", {
  
  # bad horizon name
  expect_error(evalMissingData(d, vars = c('clay', 'frags', 'ph'), name = 'namae'))
  
  # bad var spec
  expect_error(evalMissingData(d, vars = c('clay', 'frags', 'a'), name = 'name'))
})

#' @title Quickly Assemble a SoilProfileCollection
#' 
#' @description Quickly assemble a single-profile, `SoilProfileCollection` object from two possible templates. This function is a useful shortcut for creating theoretical `SoilProfileCollection` objects for testing or demonstrative purposes.
#'
#' @param x either a `list` or `character` vector, see Details and Examples
#' @param id character, specified when `x` is a `list`, name of ID list element
#' @param d character, specified when `x` is a `list`, name of depths list element
#' @param n character, specified when `x` is a `list`, name of horizon name list element
#' @param m character, specified when `x` is a `list`, name of list element containing Munsell color notation
#' @param interval, numeric, typically an integer and only specified when using character templates in mode 2. See Details.
#'
#' @return `SoilProfileCollection` object
#' 
#' @export
#' 
#' @details The list template for a single SPC allows for full specification of ID, horizon designation, bottom depths, and an arbitrary number of horizon-level attributes. A compact notation is used for profile ID (single value) and horizon depths (bottom depths, assuming datum of 0). Horizon designation and additional data (e.g. clay content) are specified as vectors all of equal length, matching the number of horizons in the profile.
#' 
#' The character template can be provided in one of several formats: 
#'   1. 'A-Bt1-Bt2-Bt3-Cr-R'
#'   2. 'ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC'
#'   
#' Format 1 is interpreted as a horizon sequence delimited by '-' or newline character (\\n). Random integer thickness are assigned to horizons, and profile ID created via `digest::digest(..., algo = 'xxhash32')`. Iteration over templates in this format is automatic when `x` is a character vector of `length > 1`.
#'   
#' Format 2 is interpreted as a horizon sequence delimited by '|'. Horizon thickness is proportional to replication of horizon designation and scaled by the `interval` argument. Profile ID is created via `digest::digest(..., algo = 'xxhash32')`. Iteration over templates in this format is automatic when `x` is a character vector of `length > 1`.
#' 
#' Explicit naming of profile IDs can be accomplished by specifying an ID via prefix, as in "ID:A-Bt1-Bt2-Cr-R" or "ID:ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC". Labels specified before a ":" will be interpreted as a profile ID. These labels are optional but if specified must be unique within `x`.
#' 
#' Single-horizon profile templates must include a trailing horizon delimiter: '-', '\\n', or '|' depending on the format.
#' 
#' @examples
#' 
#' # list-based template
#' x <- list(
#' id = 'P1',
#' depths = c(25, 33, 100, 150),
#' name = c('A', 'Bw', 'Bt', 'Cr'),
#' clay = c(12, 15, 22, 25),
#' soil_color = c('10YR 3/3', '10YR 4/4', '10YR 4/6', '5G 6/2')
#' )
#' 
#' s <- quickSPC(x)
#' plotSPC(s, name.style = 'center-center', cex.names = 1)
#' 
#' # character template, mode 1
#' # horizon thickness is generated at random (uniform [5,20])
#' x <- 'A-Bt1-Bt2-Bt3-Cr-R'
#' 
#' s <- quickSPC(x)
#' plotSPC(s, name.style = 'center-center', cex.names = 1)
#' 
#' 
#' # multiple templates
#' x <- c(
#' 'A-Bt1-Bt2-Bt3-Cr-R', 
#' 'A-C1-C2-C3-C4-Ab', 
#' 'Ap-A-A/E-E-Bhs-Cr'
#' )
#'
#' # this interface is vectorized 
#' s <- quickSPC(x)
#' plotSPC(s, name.style = 'center-center', cex.names = 1)
#' 
#' 
#' # optionally specify profile IDs using "ID:" prefix
#' x <- c(
#' 'P1:A-Bt1-Bt2-Bt3-Cr-R',
#' 'P2:A-C1-C2-C3-C4-Ab',
#' 'P3:Ap-A-A/E-E-Bhs-Cr'
#' )
#' 
#' s <- quickSPC(x)
#' plotSPC(s, name.style = 'center-center', cex.names = 1)
#' 
#' 
#' # optionally specify:
#' # horizon bottom depths in cm
#' # soil color in Munsell notation
#' x <- c(
#' '1. simple:Oe-A-E-Bhs',
#' '2. full:Oe,10,10YR 2/2-A,20,10YR 3/3-E,30,2.5Y 8/2-Bhs,60,7.5YR 4/6'
#' )
#' 
#' s <- quickSPC(x)
#' plotSPC(s, name.style = 'center-center', cex.names = 1)
#' 
#' # use newline character as delimiter, more compact
#' x <- 'Oe,10,10YR 2/2
#' A,20,10YR 3/3
#' E,30,2.5Y 8/2
#' Bhs,60,7.5YR 4/6
#' BC,125,7.5YR 6/4
#' C,150,10YR 6/2'
#' 
#' plotSPC(quickSPC(x), name.style = 'center-center', cex.names = 1)
#' 
#' 
#' # character template, mode 2
#' # horizon thickness is proportional to replication of 
#' # horizon designation and scaled by 'interval' argument
#' # default of 10 depth units
#' # e.g. A horizon is 3 * 10 = 30 depth units thick.
#' x <- c(
#'   'AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
#'   'ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC',
#'   'A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR'
#'   )
#' 
#' # each horizon label is '10' depth-units (default)
#' s <- quickSPC(x)
#' plotSPC(s, name.style = 'center-center', 
#'         cex.names = 1, depth.axis = FALSE, 
#'         hz.depths = TRUE
#' )
#' 
#' # each horizon label is '5' depth-units
#' s <- quickSPC(x, interval = 5)
#' plotSPC(s, name.style = 'center-center', 
#'         cex.names = 1, depth.axis = FALSE, 
#'         hz.depths = TRUE
#' )
#' 
#' # optionally specify some / all profile IDs with "ID:" prefix
#' x <- c(
#'   'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
#'   'P2:ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC',
#'   'A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR'
#'   )
#' 
#' s <- quickSPC(x)
#' plotSPC(s, name.style = 'center-center', 
#'         cex.names = 1, depth.axis = FALSE, 
#'         hz.depths = TRUE
#' )
#' 
#' 
#' # make a NODATA profile, with a random hash ID
#' #  note the use of trailing horizon delimiter
#' #  note the use of NA soil color field
#' x <- 'NODATA,150,NA-'
#' s <- quickSPC(x)
#' plotSPC(s, name.style = 'center-center', 
#'         cex.names = 1, depth.axis = FALSE, 
#'         hz.depths = TRUE)
#'
#'



## TODO: 
# * add vectorization for list-based template

quickSPC <- function(x, id = 'id', d = 'depths', n = 'name', m = 'soil_color', interval = 10) {
  
  # sanity check
  stopifnot(inherits(x, 'list') || inherits(x, 'character'))
  
  # mode switch
  .res <- switch(
    class(x), 
    # more expressive, list specification
    'list' = {
      .data <- .qSPC.list(x, id, d, n, m)
      .data
    }, 
    
    # simpler, hz designation only
    # template must be consistent across elements of x
    'character' = {
      
      # character template mode, detected from first element of x
      
      # mode 1: "A-Bt-Cr-R" -> random depths
      #         "id:A-Bt-Cr-R" -> random depths
      #         "id:A,10,10YR 4/4" -> depths + colors specified
      #         can also use \n as token delimiter
      #         single horizon template must include a trailing hz delimiter
      .m1 <- grepl(pattern = '-', x = x[1], fixed = TRUE) || grepl(pattern = '\n', x = x[1], fixed = TRUE)
      
      # mode 2: "A|BtBt|Cr|RRRR" -> proportional depths
      #         single horizon pattern must include trailing "|"
      .m2 <- grepl(pattern = '|', x = x[1], fixed = TRUE)
      
      # ... mode N
      
      # assemble bit mask
      # overkill, but seemed like a fun idea
      .b <- paste(as.integer(c(.m1, .m2)), collapse = '')
      
      # parse based on mode encoded in bit mask
      switch(.b,
             # mode 1
             '10' = {
               # vectorization
               if(length(x) < 2) {
                 .data <- .qSPC.char.1(x)  
               } else {
                 .data <- lapply(x, .qSPC.char.1)
                 .data <- aqp::combine(.data)
                 .data
               }
             }, 
             # mode 2
             '01' = {
               # vectorization
               if(length(x) < 2) {
                 .data <- .qSPC.char.2(x, interval = interval)  
               } else {
                 .data <- lapply(x, .qSPC.char.2, interval = interval)
                 .data <- aqp::combine(.data)
                 .data
               }
             },
             # error condition: bad formatting
             stop('incorrect horizon sequence specified', call. = FALSE)
      )
      
      .data
    })
  
  return(.res)
}


.parseID <- function(x, d = ':') {
  # attempt splitting optional ID prefix
  .s <- strsplit(x, ':', fixed = TRUE)[[1]]
  
  # length > 1 indicates an ID prefix
  if(length(.s) >1) {
    .id <- .s[1]
    # ID now removed from hz sequence
    x <- .s[2]
  } else {
    .id <- NULL
  }
  
  return(list(id = .id, x = x))
}

# split extract data from within a token
.parseExtra <- function(x, d = ',') {
  .s <- strsplit(x, split = d, fixed = TRUE)[[1]]
  .res <- data.frame(
    name = .s[1],
    bottom = as.numeric(.s[2]),
    m = .s[3]
  )
  return(.res)
}

## TODO: handle horizon boundary codes: AS, CW, etc.
# handle character-based templates, mode 1
# x <- 'A-C-R'
# x <- 'id:A-C-R'
# x <- 'id:A,10,7.5YR 3/3
# delimiter can be either '-' or '\n'
.qSPC.char.1 <- function(x) {
  
  # detect delimiter
  if(grepl(pattern = '\n', x = x[1], fixed = TRUE)) {
    delim = '\n'
  } else {
    delim = '-'
  }
  
  # detect / extract ID prefix
  .s <- .parseID(x)
  x <- .s$x
  
  # detect extra data
  .extraFlag <- grepl(pattern = ',', x, fixed = TRUE)
  
  # split token sequence into horizons
  .names <- strsplit(x, split = delim, fixed = TRUE)[[1]]
  .nhz <- length(.names)
  
  # sanity check
  if(any(.names == '')) {
    stop('Empty horizon designation not allowed in this template', call. = FALSE) 
  }
  
  if(.extraFlag) {
    # horizon bottoms are specified, along with other possible data
    .data <- do.call('rbind', lapply(.names, .parseExtra))
    .data$top <- c(0, .data$bottom[-.nhz])
    
    # convert colors, if present
    .data$soil_color <- parseMunsell(.data$m)
  } else {
    # random horizon thickness
    .thick <- round(runif(.nhz, min = 5, max = 40))
    
    # convert to top/bottom depths
    .bottom <- cumsum(.thick)
    .top <- c(0, .bottom[-.nhz])
    
    # assemble
    .data <- data.frame(
      top = .top,
      bottom = .bottom,
      name = .names
    )
  }
  
  # IDs
  if(is.null(.s$id)) {
    # generate one via digest
    .data$id <- as.character(
      digest::digest(.data, algo = 'xxhash32')
    )
  } else {
    # use the ID prefix
    .data$id <- .s$id
  }
  
  # init SPC
  depths(.data) <- c('id', 'top', 'bottom')
  
  # attempt setting horizon designation
  hzdesgnname(.data) <- 'name'
  
  return(.data)
}

# handle character-based templates, mode 2
# x <- 'ApAp|AA|Bh|BhsBhs|Bt1|Bt2Bt2|CCCC|Cr'
.qSPC.char.2 <- function(x, delim = '|', interval = 10) {
  
  # detect / extract ID prefix
  .s <- .parseID(x)
  x <- .s$x
  
  # split name sequence into horizons
  .name.thick <- strsplit(x, split = delim, fixed = TRUE)[[1]]
  .nhz <- length(.name.thick)
  
  # extract names
  .names <- lapply(strsplit(.name.thick, ''), unique)
  .names <- sapply(.names, paste, collapse = '')
  
  # sanity check
  if(any(.names == '')) {
    stop('Empty horizon designation not allowed in this template', call. = FALSE) 
  }
  
  # proportional horizon thickness
  # count labels
  .matches <- mapply(gregexpr, pattern = .names, text = .name.thick)
  .counts <- sapply(.matches, length)
  
  
  # interval-unit thickness per label instance
  .thick <- .counts * interval
  
  # convert to top/bottom depths
  .bottom <- cumsum(.thick)
  .top <- c(0, .bottom[-.nhz])
  
  # assemble
  .data <- data.frame(
    top = .top,
    bottom = .bottom,
    name = .names
  )
  
  # IDs
  if(is.null(.s$id)) {
    # generate one via digest
    .data$id <- as.character(
      digest::digest(.data, algo = 'xxhash32')
    )
  } else {
    # use the ID prefix
    .data$id <- .s$id
  }
  
  # init SPC
  depths(.data) <- c('id', 'top', 'bottom')
  
  # attempt setting horizon designation
  hzdesgnname(.data) <- 'name'
  
  return(.data)
}

# list-based template
.qSPC.list <- function(x, id, d, n, m) {
  
  # sanity check on column names
  stopifnot(id %in% names(x))
  stopifnot(d %in% names(x))
  
  # sanity check: accidentally included 0?
  if(0 %in% x[[d]]) {
    stop('depths should be specified as horizon bottom depths, did you accidentally include the top (0) of the profile?', call. = FALSE)
  }
  
  # extract top / bottom depths
  # x$d contains horizon bottom depths
  #     referenced to 0-datum
  .bottom <- x[[d]]
  .top <- c(0, .bottom[-length(.bottom)])
  
  # extract ID
  # single value, interpret as character
  .id <- as.character(x[[id]])
  
  # remove depths / id
  # leaving only hz data
  x[[id]] <- NULL
  x[[d]] <- NULL
  
  # list -> DF
  .data <- data.frame(
    .id, 
    top = .top,
    bottom = .bottom,
    as.data.frame(x)
  )
  
  # sanity check on number of rows vs. depths
  stopifnot(nrow(.data) == length(.top))
  
  # retain original ID name
  names(.data)[1] <- id
  
  # attempt to init soil color from Munsell notation
  if(!is.null(.data[[m]])) {
    .data[[m]] <- parseMunsell(.data[[m]])
  }
  
  # init SPC
  depths(.data) <- c(id, 'top', 'bottom')
  
  # attempt setting horizon designation
  try(hzdesgnname(.data) <- n, silent = TRUE)
  
  return(.data)
}

context("quickSPC")

x.l <- list(
  id = 'A',
  depths = c(25, 33, 100, 150),
  name = c('A', 'Bw', 'Bt', 'Cr'),
  clay = c(12, 15, 22, 25)
)

x.c <- 'A-Bt1-Bt2-Bt3-Cr-R'

x.c.multiple <- c(
  'A-Bt1-Bt2-Bt3-Cr-R', 
  'A-C1-C2-C3-C4-Ab', 
  'Ap-A-A/E-E-Bhs-Cr'
)

x.c.mode2 <- c(
    'AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
    'ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC',
    'A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR'
    )


test_that("list template", {
  s <- quickSPC(x.l)
  
  expect_true(inherits(s, 'SoilProfileCollection'))
  expect_equal(length(s), 1)
  expect_equal(nrow(s), 4)
  expect_equal(hzdesgnname(s), 'name')
  
})

test_that("list template, error condition", {
  
  # ID spec
  expect_error(s <- quickSPC(x.l, id = 'xxx'))
  
  # depths spec
  expect_error(s <- quickSPC(x.l, d = 'xxx'))
  
})


test_that("character template, mode 1", {
  s <- quickSPC(x.c)
  
  expect_true(inherits(s, 'SoilProfileCollection'))
  expect_equal(length(s), 1)
  expect_equal(nrow(s), 6)
  expect_equal(hzdesgnname(s), 'name')
  
  # function is vectorized over these kind of templates
  s <- quickSPC(x.c.multiple)
  
  expect_true(inherits(s, 'SoilProfileCollection'))
  expect_equal(length(s), 3)
  
  # garbage input, but still valid template
  s <- quickSPC('A-4-zz-12')
  expect_true(inherits(s, 'SoilProfileCollection'))
  
})

test_that("character template, mode 1, error condition", {
  
  # invalid hz sequence
  expect_error(quickSPC('A-C--'))
  
})


test_that("character template, mode 2", {
  s <- quickSPC(x.c.mode2[1])
  
  expect_true(inherits(s, 'SoilProfileCollection'))
  expect_equal(length(s), 1)
  expect_equal(nrow(s), 4)
  expect_equal(hzdesgnname(s), 'name')
  
  
  # check total depth
  expect_true(max(s) == 10 * 18)
  
  # function is vectorized over these kind of templates
  s <- quickSPC(x.c.mode2)
  
  expect_true(inherits(s, 'SoilProfileCollection'))
  expect_equal(length(s), 3)
  
  # check all profile depths
  expect_true(all(profileApply(s, max) == c(140, 140, 180)))
  
  
  # garbage input, but still valid template
  s <- quickSPC('A|4|zz|12')
  expect_true(inherits(s, 'SoilProfileCollection'))
  
})

test_that("character template, mode 2, error condition", {
  
  # invalid hz sequence
  expect_error(quickSPC('A|C||'))
  
  # invalid specification / mixing modes
  expect_error(quickSPC('A|C||--'), regexp = 'incorrect')
  
})

test_that("character template, other errors", {

  # invalid specification / mixing modes
  expect_error(quickSPC('A|C||--'), regexp = 'incorrect')
  
})


test_that("ID prefix", {
  
  s <- quickSPC('P1:A-Bt-C-Ab-Cr')
  expect_equal('P1', profile_id(s))
  
  s <- quickSPC('P3:AAAA|BtBtBtBt|Cr|R')
  expect_equal('P3', profile_id(s))
  
})

## TODO: 
# * data.table optimization -> no iteration over profiles

#' @title Estimate Soil Depth
#' @description Estimate the soil depth of a single profile within a `SoilProfileCollection` object. This function would typically be called by [profileApply()].
#' 
#' @param f SoilProfileCollection object of length 1, e.g. a single profile
#' @param name name of the column that contains horizon designations
#' @param p REGEX pattern for determining "contact", or depth to some morphologic feature (e.g. `Bt`)
#' @param selection an R function applied in the presence of multiple matching horizons: \code{min} (default), \code{max}, \code{mean}, etc.
#' @param no.contact.depth in the absence of contact matching \code{p}, a depth at which we can assume a standard depth-to-contact
#' @param no.contact.assigned value assigned when no contact is encountered at or below \code{no.contact.depth}
#' 
#' @return single value representing the depth to `contact` or \code{no.contact.assigned}
#' 
#' @details The choice of a \code{selection} function usually follows:
#' 
#' \code{min}: the top of the first matching horizon, \code{max}: the top bot the last matching horizon, 
#' or possibly \code{mean}: somewhere in-between.
#' 
#' 
#' @author D.E. Beaudette and J.M. Skovlin
#' 
#' @seealso \code{\link{getSoilDepthClass}}, \code{\link{profileApply}}
#' 
#' @keywords manip
#' @export
#' @examples 
#' 
#' 
#' ## consider a situation where there were multiple candidate
#' ## "contacts": 2 Cd horizons over an R
#' 
#' # init hypothetical profile
#' d <- data.frame(
#'   id = '1',
#'   top = c(0, 10, 20, 30, 40, 50, 60),
#'   bottom = c(10, 20, 30, 40, 50, 60, 80),
#'   name = c('A', 'Bt1', 'Bt2', 'BC', 'Cd1', 'Cd2', 'R'),
#'   stringsAsFactors = FALSE
#' )
#' 
#' # upgrade to SPC
#' depths(d) <- id ~ top + bottom
#' 
#' # init horizon designation
#' hzdesgnname(d) <- 'name'
#' 
#' # visual check
#' par(mar = c(0, 0, 0, 2))
#' plotSPC(d, hz.depths = TRUE, name.style = 'center-center', cex.names = 1, width = 0.1)
#' 
#' # top of the first Cd
#' estimateSoilDepth(d, name = 'name')
#' 
#' # top of the first Cd
#' estimateSoilDepth(d, name = 'name', selection = min)
#' 
#' # top of the R
#' estimateSoilDepth(d, name = 'name', selection = max)
#' 
#' # top of the second Cd
#' estimateSoilDepth(d, name = 'name', selection = max, p = 'Cd')
#'
#'
#' ## another example
#' 
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#' 
#' # init horizon designation
#' hzdesgnname(d) <- 'name'
#' 
#' # apply to each profile in a collection, and save as site-level attribute
#' sp1$depth <- profileApply(sp1, estimateSoilDepth, name='name')
#' 
#' # this function can be used to "find" depth to any feature 
#' # that can be defined via REGEX pattern matching on the horizon name
#' # for example, locate the depth to the top "Bt" horizon
#' # returning NA when there is no match
#' sp1$top_Bt <- profileApply(
#'   sp1, estimateSoilDepth, 
#'   name='name', 
#'   p='Bt', 
#'   no.contact.depth=0, 
#'   no.contact.assigned=NA
#' )
#' 
#' # reduced margins
#' par(mar=c(1,1,1,2))
#' # adjust default y-offset and depth scaling for following examples
#' plotSPC(sp1, y.offset=10, scaling.factor=0.5)
#' 
#' # get plotting parameters for profile widths and depth scaling factors
#' lsp <- get("last_spc_plot", envir = aqp.env)
#' 
#' # positions on x-axis, same for both depth and top "Bt" horizon
#' x.positions <- (1:length(sp1)) - lsp$width
#' 
#' # annotate contact with unicode right-arrow
#' # y-position is adjusted based on plot y-offset and scaling factor
#' y.positions <- lsp$y.offset + (sp1$depth * lsp$scaling.factor)
#' text(x.positions, y.positions, '\u2192', col='red', adj=1, cex=1.25, lwd=2)
#' 
#' # annotate top "Bt" depth with unicode right-arrow
#' # y-position is adjusted based on plot y-offset and scaling factor
#' y.positions <- lsp$y.offset + (sp1$top_Bt * lsp$scaling.factor)
#' text(x.positions, y.positions, '\u2192', col='blue', adj=1, cex=1.25, lwd=2)
#' 
#' 
#' \dontrun{
#'   # sample data
#'   data(gopheridge, package='soilDB')
#'   
#'   # run on a single profile
#'   estimateSoilDepth(gopheridge[1, ], name = 'hzname')
#'   
#'   # apply to an entire collection
#'   profileApply(gopheridge, estimateSoilDepth, name = 'hzname')
#' }
estimateSoilDepth <- function(f,
                              name = hzdesgnname(f),
                              p = 'Cr|R|Cd',
                              selection = min,
                              no.contact.depth = NULL,
                              no.contact.assigned = NULL) {
  # TODO: vectorize
  
  # sanity check: this function will only operate on an SPC
  if(! inherits(f, 'SoilProfileCollection')) {
    stop('`f` must be a SoilProfileCollection containing one profile', call. = FALSE)
  }
  
  # sanity check: this function works on SPC with length 1
  if(length(f) > 1) {
    stop('`f` can contain only one profile, see manual page for details')
  }
  
  # must have a valid horizon designation
  if(! name %in% horizonNames(f)) {
    stop("soil depth estimation relies on a column containing horizon designations", call.=FALSE)
  }
  
  # use SPC depth column names
  depthcols <- horizonDepths(f)
  top <- depthcols[1]
  bottom <- depthcols[2]

  # extract horizons
  h <- horizons(f)
  h <- .data.frame.j(h, col.names = c(name, top, bottom))
  
  # omit NA in hzname, top, bottom
  # can't match or use these data
  h <- na.omit(h)

  # extract possible contact
  contact.idx <- grep(p, h[[name]], ignore.case=TRUE)
  # everything else
  no.contact.idx <- grep(p, h[[name]], ignore.case=TRUE, invert=TRUE)
  
  ## TODO: this is really hard to follow, simplify / test
  
  # no contact defined, use deepest hz bottom depth
  if (length(contact.idx) < 1) {
    d <- max(h[[bottom]][no.contact.idx], na.rm = TRUE)
    
    # is there a user-specified depth at which we assume a standard depth?
    if (!is.null(no.contact.depth) &
        !is.null(no.contact.assigned)) {
      if (d >= no.contact.depth & !is.null(no.contact.assigned)) {
        res <- no.contact.assigned
      } else {
        res <- d
      }
    } else {
      # otherwise use depth of deepest horizon
      res <- d
    }
    
  } else {
    # contact pattern matched
    # apply selection function
    
    ## BUG: this is not robust to NA in some cases
    ## https://github.com/ncss-tech/aqp/issues/241
    res <- selection(h[[top]][contact.idx], na.rm = TRUE)
  }
    

  return(res)
}

context("estimateSoilDepth")

## example data
d <- data.frame(
  id = '1',
  top = c(0, 10, 20, 30, 40, 50),
  bottom = c(10, 20, 30, 40, 50, 60),
  name = c('A', 'Bt1', 'Bt2', 'BC', 'Cr1', 'Cr2'),
  stringsAsFactors = FALSE
)

depths(d) <- id ~ top + bottom

# plotSPC(d, hz.depths = TRUE, name.style = 'center-center', cex.names = 1)

test_that("basic functionality", {
  
  # defaults
  e <- estimateSoilDepth(d, name = 'name')
  expect_equal(e, 40)
  
  # manually specify selection function
  # min == top of the top-most match
  e <- estimateSoilDepth(d, name = 'name', selection = min)
  expect_equal(e, 40)
  
  # max = top of the bottom-most match
  e <- estimateSoilDepth(d, name = 'name', selection = max)
  expect_equal(e, 50)
  
  # somewhere in-between
  e <- estimateSoilDepth(d, name = 'name', selection = mean)
  expect_equal(e, 45)
  
})

test_that("no match", {
  
  # no match, use depth of profile
  e <- estimateSoilDepth(d, name = 'name', p = 'xxx')
  expect_equal(e, 60)
  
  # no match, use `no contact` rules
  e <- estimateSoilDepth(d, name = 'name', p = 'xxx', no.contact.depth = 25, no.contact.assigned = 'nothing')
  expect_equal(e, 'nothing')
  
  # `no contact` depth rule not triggered
  e <- estimateSoilDepth(d, name = 'name', p = 'xxx', no.contact.depth = 75, no.contact.assigned = 'nothing')
  expect_equal(e, 60)
})

context("SoilProfileCollection iterator (profileApply")

data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

attr <- 'prop' # clay contents %

test_that("profileApply - basic tests of output datatypes", {
  r1 <- profileApply(sp1, estimateSoilDepth, name="name")
  expect_equal(names(r1),c("P001", "P002", "P003", "P004", "P005", "P006", "P007", "P008", "P009"))
  expect_equal(r1, c(P001 = 89L, P002 = 59L, P003 = 67L, P004 = 62L, P005 = 68L, P006 = 200L, P007 = 233L, P008 = 200L, P009 = 240L))

  r2 <- profileApply(sp1, estimateSoilDepth, name="name", simplify = FALSE)
  expect_equal(r2, list(P001 = 89L, P002 = 59L, P003 = 67L, P004 = 62L, P005 = 68L,
                        P006 = 200L, P007 = 233L, P008 = 200L, P009 = 240L))

  r3 <- profileApply(sp1, function(p) {
      d <- estimateSoilDepth(p, name="name")
      res <- data.frame(profile_id(p), d, stringsAsFactors = FALSE)
      colnames(res) <- c(idname(p), "soildepthest")
      return(res)
    }, frameify = TRUE)
  expect_true(inherits(r3,'data.frame'))
})

test_that("profileApply - frameify option", {

  # 1 row per profile
  expect_silent(r1 <- profileApply(sp1, frameify = TRUE, column.names = c("foo","bar"), function(p) {
    data.frame(id = profile_id(p),
                depth = estimateSoilDepth(p, name = "name"))
  }))
  expect_equal(nrow(r1), 9)

  # 1 row per horizon
  expect_silent(r1 <- profileApply(sp1, frameify = TRUE, column.names = c("id","foo","bar"), function(p) {
    data.frame(id = profile_id(p),
               hzID = hzID(p),
               depth = estimateSoilDepth(p, name = "name"))
  }))
  expect_equal(nrow(r1), 60)

  # some profiles with no result
  expect_silent(r2 <- profileApply(sp1, frameify = TRUE, function(p) {
    res <- data.frame(id = profile_id(p),
               depth = estimateSoilDepth(p, name = "name"))
    if (res$depth < 200)
      return(res)
  }))

  # non-data.frame first result
  expect_warning(r2 <- profileApply(sp1[1,], frameify = TRUE, function(p) {
   estimateSoilDepth(p, name = "name")
  }))
  expect_true(inherits(r2, "list"))
})

#' @title Visual Description of Munsell Hue Ordering
#' 
#' @description Munsell hues are arranged on the unit circle with "neutral" at the center.
#' 
#' @param hues vector of Munsell hues, commonly derived from `huePosition()`
#' @param value single integer, Munsell value used to create an actual color
#' @param chroma single integer, Munsell chroma used to create an actual color 
#' @param chip.cex numeric, scaling for color chips
#' @param label.cex numeric, scaling labels
#' @param seg.adj numeric, scaling for line segment cues
#' @param seg.col single color, color used for line segment cues
#' @param plot logical, generate output on the current graphics device
#' @param simulateCVD simulate color vision deficiencies with the colorspace package, should be the character representation of a function name, one of: 'deutan', 'protan', or 'tritan'.
#' @param CVDseverity numeric value between 0 (none) and 1 (total), describing the severity of the color vision deficiency
#'
#' @note The best results are obtained when setting margins to zero, and inverting foreground / background colors. For example: `par(mar = c(0, 0, 0, 0), fg = 'white', bg = 'black')`.
#' 
#' @references 
#' Munsell book of color. 1976. Macbeth, a Division of Kollmorgen Corp., Baltimore, MD.
#' 
#' 
#' @return an invisible `data.frame` of data used to create the figure
#' 
#' @export
#'
#' @examples
#' 
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' 
#' # better graphics defaults
#' op <- par(
#' mar = c(0, 0, 0, 0), 
#' fg = 'white', 
#' bg = 'black', 
#' xpd = NA
#' )
#' 
#' # full set of hues, as generated by huePosition(returnHues = TRUE)
#' huePositionCircle()
#' 
#' # just a few hues
#' huePositionCircle(hues = c('5R', '5Y', '5G', '5B', '5P'))
#' 
#' # adjust Munsell value and chroma
#' huePositionCircle(value = 3, chroma = 6)
#' 
#' # reset graphics state
#' par(op)
#' 
huePositionCircle <- function(hues = huePosition(returnHues = TRUE), value = 6, chroma = 10, chip.cex = 5.5, label.cex = 0.66, seg.adj = 0.8, seg.col = grey(0.4), plot = TRUE, simulateCVD = NULL, CVDseverity = 1) {
  
  # sacrifice to CRAN deity
  munsellHuePosition <- NULL
  
  # note: this is incompatible with LazyData: true
  # load look-up table from our package
  load(system.file("data/munsellHuePosition.rda", package="aqp")[1])
  
  # re-order LABELS according to vector of presented hues
  idx <- match(hues, munsellHuePosition$hue)
  d <- munsellHuePosition[idx, ]
  
  # if using a subset of hues, do not use TN #2 labels
  if(length(hues) == 40) {
    .addSeqLabels <- TRUE
  } else {
    .addSeqLabels <- FALSE
  }
  
  # retain sequence
  d$s <- 1:nrow(d)
  
  # convert colors
  d$cols <- parseMunsell(sprintf('%s %s/%s', d$hues, value, chroma))
  
  ## optionally simulate color vision deficiency with colorspace package
  
  # something specified
  if(!is.null(simulateCVD)) {
    
    # sanity check
    if(!simulateCVD %in% c('deutan', 'protan', 'tritan')) {
      stop("simulateCVD should be one of c('deutan', 'protan', 'tritan')", call. = FALSE)
    }
    
    # simulate full severity
    d$cols <- switch(
      simulateCVD, 
      deutan = colorspace::deutan(d$cols, severity = CVDseverity),
      protan = colorspace::protan(d$cols, severity = CVDseverity),
      tritan = colorspace::tritan(d$cols, severity = CVDseverity)
    )
  }
  
  # neutral color at center
  n.col <- parseMunsell(sprintf('N %s/', value))
  
  # foreground / background colors
  fg.col <- par()$fg
  bg.col <- par()$bg
  
  if(plot) {
    # init plot area
    plot(x = munsellHuePosition$x, y = munsellHuePosition$y, asp = 1, type = 'n', xlab = '', ylab = '', axes = FALSE)
    
    # hue chips + annotation
    points(x = d$x, y = d$y, pch = 21, cex = chip.cex, bg = d$cols, col = fg.col)
    text(x = d$x, y = d$y, labels = d$hues, col = invertLabelColor(d$cols), cex = label.cex, font = 2)
    
    # ordering / direction sequence
    if(.addSeqLabels) {
      text(x = d$x * 0.88, y = d$y * 0.88, labels = d$s, col = fg.col, cex = label.cex)  
    }
    
    # visual cues
    segments(x0 = 0, x1 = d$x * seg.adj, y0 = 0, y1 = d$y * seg.adj, col = seg.col)
    
    # neutral at the center
    points(x = 0, y = 0, bg = n.col, col = fg.col, pch = 21, cex = chip.cex * 1.25)
    text(x = 0, y = 0, labels = 'N', col = invertLabelColor(n.col), cex = label.cex * 1.5, font = 2)
  }
  
  # remove original row names
  row.names(d) <- NULL
  
  # in case this information is helpful
  invisible(d)
  
}

context("Munsell hue position")

## sample data
m1 <- c('10YR 6/3', '7.5YR 3/3', '10YR 2/2', '7.5YR 3/4')
m2 <- c('5YR 3/4', '7.5YR 4/4', '2.5YR 2/2', '7.5YR 6/3')

## tests


test_that("huePosition works as expected", {
  
  x <- c('2.5YR', '7.5YR', '10YR', '5BG')
  z <- huePosition(x)
  
  ## basic operation / vectorization
  # manually counted on the Munsell wheel
  # https://www.nrcs.usda.gov/wps/portal/nrcs/detail/soils/ref/?cid=nrcs142p2_053569
  # huePositionCircle(chroma = 15)
  expect_equal(z, c(4, 6, 7, 21))
  
  ## bogus input should result in NA
  expect_true(is.na(huePosition('10YR 3/3')))
  expect_true(is.na(huePosition('5RR')))
  
  ## all hues
  h <- huePosition(returnHues = TRUE)
  expect_true(length(h) == 40)
  
  ## neutral hues require a special argument
  expect_true(is.na(huePosition('N')))
  expect_true((huePosition('N', includeNeutral = TRUE)) == 41)
  h <- huePosition(returnHues = TRUE, includeNeutral = TRUE)
  expect_true(length(h) == 41)
  
  ## shift origin
  # huePositionCircle(chroma = 15)
  expect_true(
    huePosition(x = '10GY', origin = '10Y') == 5L
  )
  
  # vectorization
  expect_true(
    all(
      huePosition(x, origin = '5P') == c(12L, 14L, 15L, 29L)  
    )
  )
  
  
  ## search direction
  # CW
  expect_true(
    huePosition(x = '5P', direction = 'cw') == 33L
  )
  
  # CCW
  expect_true(
    huePosition(x = '5P', direction = 'ccw') == 9L
  )
  
  
  ## new origin + search direction
  # huePositionCircle(hues = huePosition(returnHues = TRUE, origin = '10Y', direction = 'ccw'), chroma = 15)
  expect_true(
    all(
      huePosition(x = x, origin = '10Y', direction = 'ccw') == c(8L, 6L, 5L, 31L)
    )
  )
  
  
})


test_that("huePositionCircle works as expected", {
  res <- huePositionCircle(hues = huePosition(returnHues = TRUE, origin = '10Y', direction = 'ccw'), chroma = 15)
  
  expect_true(inherits(res, 'data.frame'))
})

#' Generate Labels for Slabs 
#' 
#' This method is used by [slab()] for generating labels that assign IDs to layers in a SoilProfileCollection
#' 
#' The new routine used in aqp 2.0 requires that, at a minimum, the `spc` and `slab.structure` arguments be specified.
#' 
#' @param slab.structure A user-defined slab thickness (defined by an integer), or user-defined structure (numeric vector). See details for `slab()`.
#' @param max.d Maximum depth
#' @param n.profiles Number of profiles
#' @param spc Optional: A SoilProfileCollection
#' @param diced Optional: The `dice()`-ed horizon-level data.frame corresponding to `spc`
#' @return factor. slab IDs, labels are the segment top and bottom depth separated by `"-"`
#' @param ... Additional arguments passed to `dice()` when `spc` is specified.
#' @seealso [slab()]
#' @export
genSlabLabels <- function(slab.structure = 1, max.d = NULL, n.profiles = NULL, spc = NULL, diced = NULL, ...) {

  # new method in aqp 2.x slab() requires SPC input
  if (!is.null(spc)) {
    if (is.null(diced)) {
      diced <- dice(spc, ..., SPC = FALSE)
    }
    
    # call method now used internally in slab()
    return(.genSlabLabels2(spc = spc, diced = diced, slab.structure = slab.structure))
  }
  
  if (is.null(max.d)) {
    stop("Must specify `max.d` when `spc` is NULL")
  }
  
  if (is.null(n.profiles)) {
    stop("Must specify `n.profiles` when `spc` is NULL")
  }
  
  # legacy method from aqp 1.x slab():
  
  # fixed-size slabs
  if (length(slab.structure) == 1) {
    # generate sequence of segment labels
    seg.label <- rep(1:ceiling(max.d / slab.structure), each = slab.structure, length = max.d)
    
    # general segment labels
    seg.label.levels <- tapply(1:max.d, seg.label, 
                               function(i) {
                                 r <- range(i, na.rm = TRUE)
                                 paste(c(r[1] - 1, r[2]), collapse = '-')
                               })
  }
  
  # user-defined slabs
  if (length(slab.structure) > 1) {
    if (slab.structure[1] == 0 & length(slab.structure) > 2) {
      # trivial case where segments start from 0
      
      seg.label <- rep(slab.structure[-1], times = diff(slab.structure))[1:max.d]
    
    } else {
      # other case: user defines an arbitrary lower and upper limit
      if (length(slab.structure) != 2)
        stop('user-defined slab boundaries must either start from 0, or contain two values between 0 and the max soil depth')
      
      # calculate thickness of slab
      slab.thickness <- diff(slab.structure)
      
      # make a new label for the slab
      new.label <- paste(slab.structure, collapse = '-')
      
      slab.idx <- rep(new.label, times = slab.thickness)
      seg.label <- slab.idx
    }
    
    # generate segment labels	
    seg.label.levels <- sapply(1:(length(slab.structure) - 1), function(i) {
      paste(c(slab.structure[i], slab.structure[i + 1]), collapse = '-')
    })
  }
  
  # covert into a factor that can be used to split profiles into slabs
  res <- factor(rep(seg.label, times = n.profiles), labels = seg.label.levels)
  
  return(res)
}

# does not assume logical horizons (may overlap, resulting in different # of slices per profile)
.genSlabLabels2 <- function(spc, diced, slab.structure = 1) {
  
  # handle various allowed slab.structure specifications
  if (length(slab.structure) == 1) {
    i <- seq(from = 0, length.out = (max(spc) / slab.structure) + 1) * slab.structure
  } else if (length(slab.structure) > 1) {
    i <- slab.structure
  } else {
    stop("empty slab.structure", call. = FALSE)
  }
  
  # calculate a sequence of cumulative depths and corresponding indices for each slab
  j <- diff(i)
  hzd <- horizonDepths(spc)
  if (length(j) == 0) {
    stop("Empty slab.structure", call. = FALSE)
  } else if (length(j) == 1 && j == 0) {
    stop("Invalid slab.structure", call. = FALSE)
  } else {
    idx1 <- cumsum(do.call('c', lapply(seq_along(j), function(x) rep(1, j[x])))) + i[1]
    idx2 <- do.call('c', lapply(seq_along(j), function(x) rep(x, j[x])))
    mt <- data.frame(idx1, slab_id = idx2, slab_label = paste0(i[idx2], "-", i[idx2 + 1]))
  }
  colnames(mt) <- c(hzd[2], "slab_id", "slab_label")
  
  # merge the dice'd data.frame result with the slab ID and labels
  res <- merge(diced, mt, by = hzd[2], all.x = TRUE, sort = FALSE)
  
  # order by horizon ID and depth
  res <- res[order(res[[idname(spc)]], res[[hzd[2]]]),]
  
  # return the slab IDs as a factor, with labels denoting each segment top and bottom
  factor(res$slab_id, labels = na.omit(unique(res$slab_label)))
}

context("genSlabLabels")


test_that("reasonable results", {
  
  x <- genSlabLabels(slab.structure = 2, max.d = 50, n.profiles = 5)  
  
  # should be a factor
  expect_true(is.factor(x))
  
  # length of results (n.profiles * max.depth) 
  expect_true(length(x) == (5 * 50))
  
  # levels
  expect_equal(
    levels(x), 
    c("0-2", "2-4", "4-6", "6-8", "8-10", "10-12", "12-14", "14-16", 
      "16-18", "18-20", "20-22", "22-24", "24-26", "26-28", "28-30", 
      "30-32", "32-34", "34-36", "36-38", "38-40", "40-42", "42-44", 
      "44-46", "46-48", "48-50")
  )
  
  
})

test_that("1cm slices, single profile, variable max.d", {

  for(md in c(10, 25, 31, 55, 70, 150, 300)) {
    x <- genSlabLabels(slab.structure = 1, max.d = md, n.profiles = 1)  
    expect_true(length(x) == md)
  }
  
  
})


test_that("variable intervals", {

  x <- genSlabLabels(slab.structure = 25, max.d = 100, n.profiles = 1)  
  
  expect_equal(
    levels(x),
    c("0-25", "25-50", "50-75", "75-100")
  )
  
  x <- genSlabLabels(slab.structure = 50, max.d = 100, n.profiles = 1)  
  
  expect_equal(
    levels(x),
    c("0-50", "50-100")
  )

  
  x <- genSlabLabels(slab.structure = 33, max.d = 100, n.profiles = 1)  
  
  expect_equal(
    levels(x),
    c("0-33", "33-66", "66-99", "99-100")
  )  
})


test_that("user-defined top/bottom interval", {
  
  x <- genSlabLabels(slab.structure = c(0, 50), max.d = 100, n.profiles = 1)  
  expect_equal(levels(x), '0-50')
  
  x <- genSlabLabels(slab.structure = c(0, 25, 50), max.d = 50, n.profiles = 1)  
  expect_equal(levels(x), c('0-25', '25-50'))
  
  x <- genSlabLabels(slab.structure = c(0, 25, 50, 75, 80), max.d = 80, n.profiles = 1)  
  expect_equal(levels(x), c("0-25", "25-50", "50-75", "75-80"))
  
  
})



test_that("edge cases", {
  
  # bottom slab interval > max.d
  x <- genSlabLabels(slab.structure = c(0, 250), max.d = 100, n.profiles = 1)  
  
  expect_true(length(x) == 250)
  expect_equal(levels(x), '0-250')
  
  # bottom slab interval < max.d
  x <- genSlabLabels(slab.structure = c(0, 99), max.d = 100, n.profiles = 1)  
  
  expect_true(length(x) == 99)
  expect_equal(levels(x), '0-99')
  
})

library(testthat)
library(aqp)

# limit all tests involving data.table 
# to 2 threads for CRAN compliance
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# Sys.setenv("_SP_EVOLUTION_STATUS_"=2)
test_check("aqp")

#' Select a subset of columns from a SoilProfileCollection
#'
#' Reduce the number of columns in a `SoilProfileCollection` to a minimal set plus additional selected columns. Optional metadata columns are included if set. At a minimum the profile ID, horizon top and bottom depths, and horizon ID are included. Horizon designation and horizon texture class column names are included if metadata attributes are set. See details.
#' 
#' @param p a `SoilProfileCollection`
#' @param column_names a set of additional columns to include in the result
#' @details 
#' 
#' Minimum column names included (when `column_names = NULL`)
#'  - `idname(p)`, `horizonDepths(p)`, `hzidname(p)`
#'  
#' Optional column names included (when metadata are set)
#'  - `hzdesgnname(p)`, `hztexclname(p)`, `GHL(p)`
#'  
#' @seealso `hzdesgnname()` `hztexclname()` `GHL()`
#' @return a `SoilProfileCollection`
#' @export
reduceSPC <- function(p, column_names = NULL) {
  
  mn <- .siteMetadataNames(p, idname = FALSE)
  nn <- .hzMetadataNames(p, depths = TRUE)
  ecn <- column_names[!column_names %in% nn]
  sn <- siteNames(p)
  hn <- horizonNames(p)
  
  stn <- sn[sn %in% c(mn, ecn)]
  hzn <- c(nn, hn[hn %in% ecn])
  msn <- ecn[!(ecn %in% stn | ecn %in% hzn)]
  
  if (length(msn) > 0) {
    stop(sprintf("column names (%s) not found in SoilProfileCollection",
                 paste0(msn, collapse = ", ")), call. = FALSE)
  }
  
  replaceHorizons(p) <- .data.frame.j(horizons(p), hzn)
  p@site <- .data.frame.j(site(p), c(idname(p), stn))
  p
}

test_that("reduceSPC() works", {
  data(sp4)
  depths(sp4) <- id ~ top + bottom
  
  sp4$x <- 1:length(sp4)
  sp4$y <- 1:length(sp4)
  
  initSpatial(sp4, crs = "EPSG:4326") <- ~ x + y
  
  sp4$texcl <- ssc_to_texcl(sp4$sand, sp4$clay)
  
  hzdesgnname(sp4) <- "name"
  hztexclname(sp4) <- "texcl"
  
  site(sp4)$newsite <- LETTERS[1:length(sp4)]
  
  x <- reduceSPC(sp4, c("newsite", "CF"))
  
  expect_true(
    all(horizonNames(x) %in% c("id", "top", "bottom", "hzID", "name", "texcl", "CF")) &&
    all(siteNames(x) %in% c("id", "x", "y", "newsite"))
  )
  expect_equal(metadata(sp4)$coordinates, c("x", "y"))
  expect_equal(metadata(sp4)$crs, "EPSG:4326")
})

#' @title Duplicate Profiles of a SoilProfileCollection
#' @description A simple function to duplicate the contents of a \code{SoilProfileCollection} object. Old profile IDs are saved as a site-level attribute (\code{oldID}) and new IDs are generated using a numeric serial number.
#' 
#' @author D.E. Beaudette
#' 
#' 
#' @param x a \code{SoilProfileCollection} object with 1 or more profiles
#' @param times requested number of copies
#' @param oldID site-level attribute used to store the original profile IDs
#' 
#' @return a \code{SoilProfileCollection} object
#' 
#' @keywords manip
#' @export
#' @examples 
#' 
#' # sample data
#' data('sp4')
#' 
#' # promote to SPC
#' depths(sp4) <- id ~ top + bottom
#' 
#' # duplicate each profile 2 times
#' d <- duplicate(sp4, times = 2)
#' 
#' # graphical check
#' par(mar = c(0, 0, 3, 1))
#' plotSPC(d, color = 'Ca', width = 0.25)
#' 
duplicate <- function(x, times = 3, oldID = '.oldID') {
  
  res <- lapply(1:times, function(i) {
    # local copy of original object
    .x <- x
    
    # save old ID just in case
    site(.x)[[oldID]] <- profile_id(.x)
    
    # overwrite with new ID
    profile_id(.x) <- sprintf('%s-%02d', profile_id(.x), i)
    
    # done
    return(.x)
  })
  
  # safely combine list -> SPC
  s <- combine(res)
}

context("SPC duplication")

# sample data
data(sp4)
depths(sp4) <- id ~ top + bottom
x <- duplicate(sp4, times = 2)


test_that("duplicate() functions as expected", {
  
  # new object should be 2x longer
  expect_true(length(x) == 2 * length(sp4))
  
  # all old IDs should be saved in .oldID
  id.setdiff <- setdiff(site(sp4)[['id']] , unique(x[['.oldID']]))
  expect_true(length(id.setdiff) == 0)
  
})

#' @title sRGB to Munsell Color Conversion
#' 
#' @description Convert sRGB color coordinates to the closest `n` Munsell chips in the `munsell` lookup table. This function will be replaced by `col2Munsell()` in **aqp 2.1**.
#'  
#'
#' @param color a `data.frame` or `matrix` object containing sRGB coordinates in the range of (0,1)
#' 
#' @param colorSpace distance metric (colorspace) to use for finding the closest chip: CIE2000 is the most accurate but requires farver >= 2.0.3, Euclidean distance in CIELAB is a close second, while Euclidean distance in sRGB is not at all accurate and should only be used for demonstration purposes.
#' 
#' @param nClosest number of closest Munsell colors to return (valid range is 1-20)
#'
#' @note This function is fully vectorized and will pad output with NA-records when NA are present in \code{color}.
#' 
#' @author D.E. Beaudette
#' 
#' @references 
#'   * http://ncss-tech.github.io/AQP/
#'   * http://www.brucelindbloom.com/index.html?ColorCalcHelp.html
#'   * http://www.munsellcolourscienceforpainters.com/MunsellAndKubelkaMunkToolbox/MunsellAndKubelkaMunkToolbox.html
#'   * http://www.cis.rit.edu/mcsl/online/munsell.php
#'
#' @return an (NA-padded) `data.frame` containing `hue`, `value`, `chroma`, and distance (dE00 when \code{colorSpace = 'CIE2000'}, Euclidean distance otherwise) to nearest matching color.
#' 
#' @export
#'
#' @examples 
#' 
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' # Munsell notation to sRGB triplets [0-1] 
#' color <- munsell2rgb(
#'   the_hue = c('10YR', '2.5YR', '5YR'), 
#'   the_value = c(3, 5, 2.5), 
#'   the_chroma = c(5, 6, 2), 
#'   return_triplets = TRUE
#' )
#' 
#' # result is a data.frame
#' color
#' 
#' # back-transform sRGB -> closest Munsell color
#' # sigma is the dE00 color contrast metric
#' rgb2munsell(color)
#'
rgb2munsell <- function(color, colorSpace = c('CIE2000', 'LAB', 'sRGB'), nClosest = 1) {
  
  ## !! will deprecate in aqp 2.1
  # .Deprecated(new = 'col2Munsell', msg = 'please use col2Munsell() instead.')
  message('rgb2munsell() will be deprecated in aqp 2.1, please use col2Munsell() instead.')
  
  
  # argument check
  colorSpace <- match.arg(colorSpace)
  
  # reasonable constraints on n-closest chips
  if(nClosest < 1) {
    message('setting `nClosest to 1`')
    nClosest <- 1
  }
  
  if(nClosest > 20) {
    message('setting `nClosest to 20`')
    nClosest <- 20
  }
  
  
  ## TODO: detect sRGB values in the range of 0-255 and re-scale accordingly
  
  # vectorize via for-loop
  n <- nrow(color)
  res <- vector(length = n, mode='list')
  
  # This is a hack to avoid munsell2rgb: "no visible binding for global variable munsell" at package R CMD check
  munsell <- NULL
  
  # note: this is incompatible with LazyData: true
  # load look-up table from our package
  # This should be more foolproof than data(munsell) c/o PR
  load(system.file("data/munsell.rda", package="aqp")[1])
  
  
  # CIE2000 requires farver >= 2.0.3
  if(colorSpace == 'CIE2000') {
    if(packageVersion("farver") < '2.0.3' ) {
      message('rgb2munsell: using LAB color space; install farver v2.0.3 or higher for perceptual distance in CIE2000')
      colorSpace <- 'LAB';
    }
  }
  
  # accounting for the possibility of NA
  # result should be an empty record
  not.na.idx <- which(apply(color, 1, function(i) ! any(is.na(i))))
  
  
  # iterate over colors
  for(i in not.na.idx) {
    # convert current color to matrix, this will allow matrix and DF as input
    this.color <- as.matrix(color[i, , drop=FALSE])
    
    # TODO: there isn't any reason to use sRGB other than demonstration
    if(colorSpace == 'sRGB') {
      # euclidean distance (in sRGB space) is our metric for closest-color
      # d = sqrt(r^2 + g^2 + b^2)
      sq.diff <- sweep(munsell[, 4:6], MARGIN=2, STATS=this.color, FUN='-')^2
      sigma <- sqrt(rowSums(sq.diff))
      # rescale distances to 0-1
      sigma <- sigma / max(sigma)
      # return the closest n-matches
      idx <- order(sigma)[1:nClosest]
    }
    
    if(colorSpace == 'LAB') {
      # euclidean distance (in CIELAB space) is our metric for closest-color
      # convert sRGB to LAB
      this.color.lab <- convertColor(this.color, from='sRGB', to='Lab', from.ref.white='D65', to.ref.white = 'D65')
      # d = sqrt(L^2 + A^2 + B^2)
      sq.diff <- sweep(munsell[, 7:9], MARGIN=2, STATS=this.color.lab, FUN='-')^2
      sigma <- sqrt(rowSums(sq.diff))
      # rescale distances to 0-1
      sigma <- sigma / max(sigma)
      # return the closest n-matches
      idx <- order(sigma)[1:nClosest]
    }
    
    # most accurate / efficient method as of farver >= 2.0.3
    if(colorSpace == 'CIE2000') {
      # CIE dE00
      # convert sRGB to LAB
      this.color.lab <- convertColor(
        color = this.color, 
        from = 'sRGB', 
        to = 'Lab', 
        from.ref.white = 'D65', 
        to.ref.white = 'D65'
      )
      
      dimnames(this.color.lab)[[2]] <- c('L', 'A', 'B')
      
      # fully-vectorized
      sigma <- farver::compare_colour(
        from = this.color.lab, 
        to = munsell[, 7:9], 
        from_space = 'lab', 
        method = 'CIE2000', 
        white_from = 'D65'
      )
      
      # return the closest n-matches
      idx <- order(sigma)[1:nClosest]
    }
    
    # pack results, sorted by closest results
    res[[i]] <- data.frame(
      munsell[idx, 1:3], 
      sigma = sigma[idx]
    )
    
  }
  
  # pad records with NA in the sRGB input
  # https://github.com/ncss-tech/aqp/issues/160
  na.idx <- which(sapply(res, is.null))
  if(length(na.idx) > 0) {
    for(i in na.idx){
      res[[i]] <- data.frame(hue=NA, value=NA, chroma=NA, sigma=NA, stringsAsFactors=FALSE)
    }
  }
  
  
  # convert to DF
  res <- do.call('rbind', res)
  row.names(res) <- as.character(1:nrow(res))
  
  # save sigma units
  attr(res, which = 'sigma') <- switch(colorSpace, 'sRGB' = 'distance in sRGB', 'LAB' = 'distance in CIELAB', 'CIE2000' = 'dE00')
  
  return(res)
}



# internally used function for snapping a vector of user-supplied values
# to a known subset via absolute distance
# NA safely propagated
.snapValid <- function(s, v) {
  # rows are distances by the_value
  .dist <- outer(s, v, function(i, j) {abs(i - j)})
  
  # index to snapped values, safely handle NA
  idx <- apply(.dist, 1, function(i) {
    if(all(is.na(i))) {
      return(NA)
    } else {
      return(which.min(i))
    }
  })
  
  # replace with snapped + NA
  return(v[idx])
}



#' @title Convert Munsell Color Notation to other Color Space Coordinates (sRGB and CIELAB)
#' 
#' @description Color conversion based on a look-up table of common soil colors.
#'
#' @param the_hue a vector of one or more more hues, upper-case (e.g. '10YR')
#' 
#' @param the_value a vector of one or more values (e.g. '4')
#' 
#' @param the_chroma a vector of one or more chromas (e.g. '6'), may be NA for neutral hues
#' 
#' @param alpha numeric, transparency setting used when \code{return_triplets = FALSE} and \code{returnLAB = FALSE}
#' 
#' @param maxColorValue maximum sRGB color value, typically `1` (see \code{\link{rgb})}
#' 
#' @param return_triplets logical, return sRGB coordinates (range 0-1) instead of standard hex notation of sRGB (e.g. '#8080B')
#' 
#' @param returnLAB logical, return CIELAB coordinates (D65 illuminant)
#' 
#' @return A vector of R colors is returned that is the same length as the input data. When \code{return_triplets = TRUE} and/or \code{returnLAB = TRUE}, then a \code{data.frame} (of sample length as input) is returned.
#' 
#' 
#' @details This function is vectorized without recycling: i.e. the length of each argument must be the same. Both functions will pad output with NA if there are any NA present in the inputs.
#' 
#' Neutral hues are approximated by greyscale shades ranging from 20\% (darker) to 80\% (lighter). No chroma is required for neutral hues.
#' 
#' Gley soil colors that are missing a chroma will not be correctly interpreted. Consider using a chroma of 1.
# 
# 
#' Non-standard Munsell notation (e.g. '7.9YR 2.7/2.0') can be matched (nearest-neighbor, no interpolation) to the closest color within the \code{munsell} sRGB/CIELAB look-up table via \code{getClosestMunsellChip()}. A more accurate estimate of sRGB values from non-standard notation can be achieved with the \href{https://CRAN.R-project.org/package=munsellinterpol}{munsellinterpol} package.
#' 
#' See examples below.
#' 
#' @note Care should be taken when using the resulting sRGB values; they are close to their Munsell counterparts, but will vary based on your monitor and ambient lighting conditions. Also, the value used for \code{maxColorValue} will affect the brightness of the colors. Th default value (1) will usually give acceptable results, but can be adjusted to force the colors closer to what the user thinks they should look like.
#' 
#' @references 
#'  * http://www.brucelindbloom.com/index.html?ColorCalcHelp.html
#'  * https://www.munsellcolourscienceforpainters.com/MunsellAndKubelkaMunkToolbox/MunsellAndKubelkaMunkToolbox.html
#'  * https://www.rit.edu/science/munsell-color-lab
#' 
#' @author D.E. Beaudette
#' 
#' @export
#'
#' @examples
#' 
#' # neutral hues (N) can be defined with chroma of 0 or NA 
#' g <- expand.grid(hue = 'N', value = 2:8, chroma = 0, stringsAsFactors = FALSE)
#' (m <- munsell2rgb(g$hue, g$value, g$chroma))
#' soilPalette(m)
#' 
#' # back-transform to Munsell notation
#' col2Munsell(t(col2rgb(m)) / 255)
#' 
#' 
#' # basic example
#' d <- expand.grid(hue = '10YR', value = 2:8, chroma = 1:8, stringsAsFactors = FALSE)
#' d$color <- with(d, munsell2rgb(hue, value, chroma))
#' 
#' # similar to the 10YR color book page
#' plot(value ~ chroma, data = d, col = d$color, pch = 15, cex = 3, las = 1)
#' 
#' # multiple pages of hue:
#' hues <- c('2.5YR', '5YR', '7.5YR', '10YR')
#' d <- expand.grid(
#'   hue = hues, 
#'   value = c(2, 2.5, 3:8), 
#'   chroma = seq(2, 8, by = 2), 
#'   stringsAsFactors = FALSE
#' )
#' # convert Munsell -> sRGB
#' d$color <- with(d, munsell2rgb(hue, value, chroma))
#' 
#' # extract CIELAB coordinates
#' with(d, munsell2rgb(hue, value, chroma, returnLAB = TRUE))
#' 
#' # plot: note that we are setting panel order from red --> yellow
#' library(lattice)
#' 
#' xyplot(
#'   value ~ factor(chroma) | factor(hue, levels = hues),
#'   main = "Common Soil Colors", layout = c(4, 1), scales = list(alternating = 1),
#'   strip = strip.custom(bg = grey(0.85)),
#'   data = d, as.table = TRUE, subscripts = TRUE,
#'   xlab = 'Chroma', ylab = 'Value',
#'   panel = function(x, y, subscripts, ...) {
#'     panel.xyplot(x, y, pch = 15, cex = 4, col = d$color[subscripts])
#'   }
#' )
#' 
#' 
#' # convert a non-standard color to closest "chip" in `munsell` look-up table
#' getClosestMunsellChip('7.9YR 2.7/2.0', convertColors = FALSE)
#' 
#' # convert directly to hex notation of sRGB
#' getClosestMunsellChip('7.9YR 2.7/2.0')
#' 
munsell2rgb <- function(the_hue, the_value, the_chroma, alpha = 1, maxColorValue = 1, return_triplets = FALSE, returnLAB = FALSE) {
  ## important: change the default behavior of data.frame and melt
  opt.original <- options(stringsAsFactors = FALSE)
  
  # check for missing data
  if(missing(the_hue) | missing(the_chroma) | missing(the_value))
    stop('Must supply a valid Munsell color.')
  
  # check to make sure that each vector is the same length
  if(length(unique( c(length(the_hue), length(the_value), length(the_chroma)))) != 1)
    stop('All inputs must be vectors of equal length.')
  
  # in case of factors, why would anyone do this?
  if(is.factor(the_hue)) {
    the_hue <- as.character(the_hue)
  }
  if(is.factor(the_value)) {
    the_value <- as.character(the_value)
  }
  if(is.factor(the_chroma)) {
    the_chroma <- as.character(the_chroma)
  }
  
  # expected data types
  # hue: character
  the_hue <- as.character(the_hue)
  
  # value and chroma: numeric
  the_value <- as.numeric(the_value)
  the_chroma <- as.numeric(the_chroma)
  
  # note: this is incompatible with LazyData: true
  # load look-up table from our package
  # This should be moreover more foolproof than data(munsell) c/o PR
  munsell <- NULL
  load(system.file("data/munsell.rda", package="aqp")[1])
  
  ## 2016-03-07: "fix" neutral hues
  ## they will typically be missing chroma or have some arbitrary number
  ## set it to 0 for correct matching
  N.idx <- which(the_hue == 'N')
  if(length(N.idx) > 0) {
    the_chroma[N.idx] <- 0
  }
    
  
  ## any other hue with 0 chroma should be interpreted as N
  idx <- which(the_hue != 'N' & the_chroma == 0)
  if(length(idx > 0)) {
    the_hue[idx] <- 'N'
  }
  
  
  ## value / chroma should be within unique set of allowed chips
  
  # the valid range depends on the latest version of the munsell LUT
  # https://github.com/ncss-tech/aqp/issues/318
  valid.value <- unique(munsell$value)
  valid.chroma <- unique(munsell$chroma)
  
  
  # test Munsell values
  if(any(! na.omit(the_value) %in% valid.value)) {
    warning("non-standard notation in Munsell value, snapping to the nearest available value\n consider using getClosestMunsellChip()", call. = FALSE)
    
    ## TODO: optimize by only computing distances for non-standard Munsell value
    
    # snap supplied Munsell value to valid value via absolute distance
    # NA are safely propagated
    the_value <- .snapValid(the_value, valid.value)
  }
  
  # test Munsell chroma
  if(any(! na.omit(the_chroma) %in% valid.chroma)) {
    warning("non-standard notation in Munsell chroma, snapping to the nearest available chroma\n consider using getClosestMunsellChip()", call. = FALSE)
    
    ## TODO: optimize by only computing distances for non-standard Munsell chroma
    
    # snap supplied Munsell chroma to valid chroma via absolute distance
    # NA are safely propagated
    the_chroma <- .snapValid(the_chroma, valid.chroma)
  }
  
  
  
  ## join new data with look-up table
  # note that value / chroma must be same data type as in `munsell` (numeric)
  d <- data.frame(
    hue = the_hue, 
    value = the_value, 
    chroma = the_chroma, 
    stringsAsFactors = FALSE
  )
  
  ## benchmarks:
  # data.table::merge() (with conversion to/from) 5x faster than base::merge()
  
  ## TODO: maybe more efficient with keys
  # round-trip through data.table is still faster
  d <- data.table::as.data.table(d)
  munsell <- data.table::as.data.table(munsell)
  
  
  ## TODO: optimize by first filtering full lookup table on e.g. hue
  
  
  # join
  res <- merge(d, munsell, by = c('hue','value','chroma'), all.x = TRUE, sort = FALSE)
  
  # back to data.frame
  res <- as.data.frame(res)
  
  # reset options:
  options(opt.original)
  
  # syntax is now a little muddled, test for sRGB and LAB
  if(return_triplets & returnLAB)
    return(res[, c('r','g','b', 'L', 'A', 'B')])
  
  # if the user wants the raw sRGB triplets, give those back
  if(return_triplets)
    return(res[, c('r','g','b')])
  
  if(returnLAB)
    return(res[, c('L','A','B')])
  
  # keep track of NA values
  rgb.na <- which(is.na(res$r))
  
  # truncate alpha at maxColorValue, otherwise error
  if(alpha > maxColorValue) {
    alpha <- maxColorValue
  }
  
  
  ## convert to hex notation
  # init an empty column
  res$soil_color <- NA
  
  # account for missing values if present: we have to do this because rgb() doesn't like NA
  if(length(rgb.na > 0)) {
    res$soil_color[-rgb.na] <- with(
      res[-rgb.na,], 
      rgb(red = r, green = g, blue = b, alpha = alpha, maxColorValue = maxColorValue) 
    )
  }	else {
    # no missing values
    res$soil_color <- with(
      res, 
      rgb(red = r, green = g, blue = b, alpha = alpha, maxColorValue = maxColorValue) 
    )
  }
  
  
  # default behavior, vector of colors is returned
  return(res$soil_color)
}


# if (!isGeneric("munsell2spc"))
setGeneric("munsell2spc", function(object, ...) standardGeneric("munsell2spc"))

#' @title Merge Munsell Hue, Value, Chroma converted to sRGB & CIELAB into a SoilProfileCollection
#'
#' @description Convert Munsell hue, value and chroma into sRGB (\code{rgb_R, rgb_G, rgb_B}) and CIELAB (lab_L, lab_A, lab_B) color coordinates using \code{munsell2rgb}. The converted values are stored in the \code{horizons()} slot unless \code{as.spc} is \code{FALSE}, in which case the results are combined with profile and horizon ID columns and returned as the \code{data.frame} subclass used by the SPC.
#'
#' @param object A SoilProfileCollection
#' @param hue Column name containing numeric hue values. Default: \code{"hue"}
#' @param value Column name containing numeric value values. Default: \code{"value"}
#' @param chroma Column name containing numeric chroma values. Default: \code{"chroma"}
#' @param .data Optional: a character vector of equal length to number of horizons (containing Munsell notation), or a column name in the horizon data OR a data.frame containing three columns (names specified in \code{hue}, \code{value}, \code{chroma})
#'
#' @param as.spc Return a data.frame-like object with ID columns?
#'
#' @return A SoilProfileCollection or \code{data.frame}-like object
#' @seealso \code{\link{parseMunsell}} \code{\link{rgb2munsell}} \code{\link{munsell2rgb}}
#' @export 
#' @aliases munsell2spc
#' @examples
#'
#' data(sp3)
#' depths(sp3) <- id ~ top + bottom
#'
#' # inspect input data
#' horizons(sp3)[,c("hue","value","chroma")]
#'
#' # do color conversions to sRGB and LAB, join into horizon data
#' sp3 <- munsell2spc(sp3)
#'
#' # plot rgb "R" coordinate by horizon
#' plot(sp3, color = "rgb_R")
#'
#' # plot lab "A" coordinate by horizon
#' plot(sp3, color = "lab_A")
#'
#' # note that `lab_A` values do not exactly match the original `A` values
#' # this is because `lab_A` was computed from the (field determined) Munsell color notation,
#' # while `A` was directly measured in the lab by colorimeter
#' plot(sp3$A, sp3$lab_A, xlab = 'Measured', ylab = 'Converted from Field Observed Munsell')
#'
setMethod("munsell2spc", signature(object = "SoilProfileCollection"),
          function(object,
                   hue = "hue", value = "value", chroma = "chroma",
                   .data = NULL,
                   as.spc = TRUE) {
            
            # if .data vector/column/data.frame not specified
            if (is.null(.data)) {
              
              # need hue, value, chroma as existing columns in horizon data
              if (!all(c(hue, value, chroma) %in% horizonNames(object))) {
                stop("arguments `hue` [character], `value` [numeric] and `chroma` [numeric] must specify column names in the horizon data",
                     call. = FALSE)
              } else {
                h <- horizons(object)
              }
              
            } else {
              # .data might be a data.frame, containing hue, value, chroma, or a character vector with munsell notation
              #  (e.g from parseMunsell(..., convertColors=FALSE))
              if (inherits(.data, 'data.frame')) {
                
                if (ncol(.data) == 1 && is.character(.data[[1]])) {
                  
                  h <- parseMunsell(.data[[1]], convertColors = FALSE)
                  
                } else if (!all(c(hue, value, chroma) %in% colnames(.data))) {
                  stop("arguments `hue` [character], `value` [numeric] and `chroma` [numeric] must specify column names in `.data`",
                       call. = FALSE)
                } else {
                  h <- .data
                }
                
              } else {
                
                if (length(.data) == 1 && .data %in% horizonNames(object)) {
                  .data <- object[[.data]]
                  
                  # otherwise need munsell character columnname, or a vector with equal length to horizons
                } else if (length(.data) != nrow(object)) {
                  stop("argument `.data` [character or data.frame], must specify either a character vector of equal length to number of horizons , a column name in the horizon data (containing Munsell notation) or a data.frame with three columns (names specified in `hue`, `value`, `chroma`)",
                       call. = FALSE)
                }
                
                h <- parseMunsell(.data, convertColors = FALSE)
              }
            }
            
            # makes a data.frame
            # return sRGB + CIELAB at the same time, note that hex notation of color is not returned
            drgb <- munsell2rgb(h[[hue]], h[[value]], h[[chroma]], return_triplets = TRUE, returnLAB = TRUE)
            colnames(drgb)[1:3] <- paste0("rgb_", c("R","G","B"))
            colnames(drgb)[4:6] <- paste0("lab_", c("L","A","B"))
            
            # ID management
            idn <- idname(object)
            hidn <- hzidname(object)
            h <- horizons(object)
            
            # munsell2rgb does not return ID names (not inherently aware of the SPC)
            idcol <- data.frame(h[[idn]], h[[hidn]])
            colnames(idcol) <- c(idn, hidn)
            
            if (as.spc) {
              # horizons<- will ensure merge.data.table triggers if @horizons is data.table
              horizons(object) <- cbind(idcol, drgb)
              
              return(object)
            } else {
              return(.as.data.frame.aqp(cbind(idcol, drgb), aqp_df_class(object)))
            }
          })

context("color signature")

## sample data
# trivial example, not very interesting
data(sp1)
depths(sp1) <- id ~ top + bottom

# sRGB color coordinates in [0, 1]
.rgb <- munsell2rgb(sp1$hue, sp1$value, sp1$chroma, return_triplets = TRUE)
sp1$r <- .rgb$r
sp1$g <- .rgb$g
sp1$b <- .rgb$b

# CIELAB color
.lab <- munsell2rgb(sp1$hue, sp1$value, sp1$chroma, returnLAB = TRUE)
sp1$CIE_L <- .lab$L
sp1$CIE_A <- .lab$A
sp1$CIE_B <- .lab$B

# Munsell notation
sp1$m <- sprintf("%s %s/%s", sp1$hue, sp1$value, sp1$chroma)

# hex sRGB
sp1$hex <- munsell2rgb(sp1$hue, sp1$value, sp1$chroma)


## tests


## REMOVE this once arguments have been removed
test_that("deprecation of r, g, b arguments", {
  
  # 2025-12-15
  expect_warning(
    pig <- soilColorSignature(sp1, r = 'r', g = 'g', b = 'b'), 
    regexp = 'deprecated'
  )
  
})



test_that("color specification detection / interpretation", {
  
  # sRGB coordinates
  pig <- soilColorSignature(sp1, color = c('r', 'g', 'b'))
  expect_equal(attr(pig, 'colorspec'), 'color-coordinate-data.frame')
  
  # error condition
  # CIELAB coordinates, without correct space argument
  expect_error(soilColorSignature(sp1, color = c('CIE_L', 'CIE_A', 'CIE_B')))
  
  # CIELAB coordinates
  pig <- soilColorSignature(sp1, color = c('CIE_L', 'CIE_A', 'CIE_B'), space = 'CIELAB')
  expect_equal(attr(pig, 'colorspec'), 'color-coordinate-data.frame')
  
  # sRGB hex
  pig <- soilColorSignature(sp1, color = 'hex')
  expect_equal(attr(pig, 'colorspec'), 'hex-sRGB')
  
  # Munsell notation
  pig <- soilColorSignature(sp1, color = 'm')
  expect_equal(attr(pig, 'colorspec'), 'munsell')
  
})



test_that("colorBucket", {
  
  # extract color signature
  pig <- soilColorSignature(sp1, color = 'm', method = 'colorBucket')
  
  # expected output
  expect_true(inherits(pig, 'data.frame'))
  expect_true(nrow(pig) == length(profile_id(sp1)))
  expect_true(ncol(pig) == 6)
  
  # add more of these
  expect_equal(pig$id[1], 'P001')
  expect_equal(pig$.white.pigment[1], 0.635, tolerance = 0.001)
  expect_equal(pig$.red.pigment[1], 0.103, tolerance = 0.001)
  expect_equal(pig$.green.pigment[1], 0, tolerance = 0.001)
  expect_equal(pig$.yellow.pigment[1], 0.261, tolerance = 0.001)
  expect_equal(pig$.blue.pigment[1], 0, tolerance = 0.001)
  
})

test_that("depthSlices", {
  
  # extract color signature
  pig <- soilColorSignature(sp1, color = 'm', method = 'depthSlices')
  
  # expected output
  expect_true(inherits(pig, 'data.frame'))
  expect_true(nrow(pig) == length(profile_id(sp1)))
  expect_true(ncol(pig) == 10)
  
  # add more of these
  expect_equal(pig$id[1], 'P001')
  expect_equal(pig$A.1[1], 5.87, tolerance = 0.01)
  expect_equal(pig$A.2[1], 5.57, tolerance = 0.01)
  expect_equal(pig$B.1[1], 11.07, tolerance = 0.01)
  expect_equal(pig$B.2[1], 17.86, tolerance = 0.01)
  expect_equal(pig$L.1[1], 30.25, tolerance = 0.01)
  
})


test_that("perceptualDistMat = TRUE", {
  
  # result is a distance matrix
  d <- soilColorSignature(sp1, color = 'm', method = 'depthSlices', perceptualDistMat = TRUE)
  
  # expected output
  expect_true(inherits(d, 'dist'))
  
  # all profiles IDs should be present and in the same order
  expect_equal(attr(d, 'Labels'), profile_id(sp1))
  
})




## TODO: create / use local data to check ordering
# using OSDs right now for a diverse range of colors
test_that("expected order from OSDs, depthSlices", {
  
  # only run this locally
  skip_on_cran()
  
  # curl is required to check if offline
  skip_if_not_installed("curl")
  
  skip_if_offline()
  
  # TODO: consider not using soilDB for testing soilColorSignature()
  skip_if_not_installed("soilDB")
  
  suppressWarnings(library(soilDB, quietly = TRUE))
  suppressWarnings(library(cluster, quietly = TRUE))
  
  s.list <- c('amador', 'redding', 'pentz', 'willows', 'pardee', 'yolo', 'hanford', 'cecil', 'sycamore', 'KLAMATH', 'MOGLIA', 'vleck', 'drummer', 'CANEYHEAD', 'musick', 'sierra', 'HAYNER', 'zook', 'argonaut', 'PALAU')
  
  # get these soil series
  s <- soilDB::fetchOSD(s.list)
  
  if (!is.null(s)) {
    # use hex sRGB, provided by fetchOSD()
    pig <- soilColorSignature(s, color = 'soil_color', method = 'depthSlices')
    row.names(pig) <- pig[, 1]
    d <- daisy(pig[, -1])
    dd <- diana(d)
    
    # expected ordering
    o <- c("AMADOR", "MOGLIA", "VLECK", "HANFORD", "PARDEE", "CANEYHEAD", 
           "HAYNER", "ARGONAUT", "MUSICK", "CECIL", "PALAU", "REDDING", 
           "SIERRA", "DRUMMER", "ZOOK", "PENTZ", "YOLO", "SYCAMORE", "WILLOWS", 
           "KLAMATH")
    
    
    expect_true(
      all(profile_id(s)[dd$order] ==  o)
    )
  }
  
})

context("simulateColor")

# example data
data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

sp1$soil_color <- munsell2rgb(sp1$hue, sp1$value, sp1$chroma)

test_that("aggregate color (proportions method)", {
  
  skip_if_not_installed('gower')
  
  horizons(sp1)$genhz <- 'A'
  a <- aggregateColor(sp1, groups='genhz', col='soil_color')
  
  
  # simulate
  set.seed(54654)
  cols <- simulateColor(method = 'proportions', n = 200, parameters = a)
  
  # expected output
  expect_true(inherits(cols, 'list'))
  
  # flatten
  cols <- unlist(cols)
  
  # check that requested number of samples are returned
  expect_true(length(cols) == 200)
  
  # tabulate, sort: most frequent color should be 7.5YR 3/2
  cols <- sort(table(cols), decreasing = TRUE)
  
  expect_equal(names(cols)[1], '7.5YR 3/2')
})


test_that("dE00 threshold method", {
  
  # params
  p <- list(
    'A' = list(m = '7.5YR 3/3', thresh = 20, hues = c('10YR', '7.5YR', '5YR'))
  )
  
  # simulate
  set.seed(54654)
  cols <- simulateColor(method = 'dE00', n = 200, parameters = p)
  
  # expected output
  expect_true(inherits(cols, 'list'))
  
  # flatten
  cols <- unlist(cols)
  
  # check that requested number of samples are returned
  expect_true(length(cols) == 200)
  
  # tabulate, sort: most frequent color should be 7.5YR 3/3
  cols <- sort(table(cols), decreasing = TRUE)
  
  # in the limit, the most frequent sample should be the same as the rep. color
  expect_equal(names(cols)[1], '7.5YR 3/3')
  
})

test_that("SPC modification", {
  
  # seed profile
  s <- sp1[2, ]
  
  # static hz variability
  horizons(s)$.hd <- 5
  
  n.sim <- 10
  
  # simulate
  ids <- sprintf("%s-%03d", 'sim', 1:n.sim)
  set.seed(10101)
  z <- perturb(s, id = ids, boundary.attr = '.hd', min.thickness = 5)

  p <- list(
    'Oi' = list(m = '7.5YR 2/2', thresh = 5, hues = c('7.5YR')),
    'A' = list(m = '7.5YR 4/3', thresh = 8, hues = c('7.5YR')),
    'C1' = list(m = '5YR 4/2', thresh = 8, hues = c('5YR', '7.5YR')),
    'C2' = list(m = '5YR 4/2', thresh = 20, hues = c('5YR', '7.5YR'))
  )
  
  # result is an SPC
  zz <- simulateColor(method = 'dE00', n = n.sim, parameters = p, SPC = z)
  
  expect_true(inherits(zz, 'SoilProfileCollection'))
  
})

# SoilProfileCollection metadata

# this method throws an error if required metadata aren't present
.require.metadata.aqp <- function(object, attr, attrlabel, message, required = FALSE) {
  xx <- metadata(object)[[attr]]
  if (length(xx) == 0 || nchar(xx) == 0) {
    if (required) {
      stop(attrlabel, " (Metadata element: '", attr, "') is not specified for this SoilProfileCollection. It is required.", message,
           call. = FALSE)
    } else {
      xx <- ""
    }
  }
  xx
}

# this method handles setting of user metadata
#  key features:
#   - NA, "", and NULL `value` are treated the same
#   - If `value` does not exist in horizonNames()
#   - allowed names is either "site" "horizon" or NULL to allow options:
#       siteNames(), horizonNames() and no limit 
#       
.set.metadata.aqp <- function(object, value, required = FALSE,
                              attr, attrlabel, message = "", 
                              allowednames = NULL) {
  # test: does it exist?
  if (length(value) == 0) {
    value <- ""
  }
  
  if(length(value) > 0) {
    # several ways to "reset" the metadata to the safe but invalid value
    if((value == "") | is.na(value) | is.null(value)) {
      value <- ""
    } else if (!is.null(allowednames)) {
      namegrp <- switch(tolower(allowednames), 
                        "site" = siteNames(object),
                        "horizon" = horizonNames(object), NULL)
      
      if (!is.null(namegrp) && !(value %in% namegrp)) {
        stop(paste0(attrlabel, " (", value, ") not in ", allowednames, " names."), call. = FALSE)
      }
    }
  }
  
  # replace
  metadata(object)[[attr]] <- value
  
  # check if required
  .require.metadata.aqp(object,
                        attr = attr,
                        attrlabel = attrlabel,
                        message = message,
                        required = required)
  
  # done
  return(object)
}

# takes two SPC as input, takes metadata other than original order from source
# returns the destination SPC with modified metadata
.transfer.metadata.aqp <- function(src, dest) {
  if (inherits(src, 'SoilProfileCollection')) {
    m <- metadata(src)
  } else {
    m <- src
  }
  stopifnot(inherits(dest, 'SoilProfileCollection'))
  
  # transfer attributes https://github.com/ncss-tech/aqp/issues/204
  customattr <- attributes(src)
  customattr <- customattr[!names(customattr) %in% c("sp", names(attributes(SoilProfileCollection())))]
  attributes(dest)[names(customattr)] <- attributes(src)[names(customattr)]
  
  # original.order metadata no longer created, not transferred
  cols <- names(m)[names(m) != "original.order"]
  
  metadata(dest)[cols] <- m[cols]
  dest
}

## get column containing horizon designations (there is a setter of same name)

setGeneric("hzdesgnname", function(object, required = FALSE)
  standardGeneric("hzdesgnname"))

#' @title Get or Set Horizon Designation Column Name
#' @name hzdesgnname
#' @aliases hzdesgnname hzdesgnname,SoilProfileCollection-method hzdesgnname<- hzdesgnname,SoilProfileCollection-method
#' @details Store the column name containing horizon designations or other identifiers in the metadata slot of the SoilProfileCollection.
#' @description `hzdesgnname()`: Get column name containing horizon designations 
#' @param object a SoilProfileCollection
#' @param required logical, is this attribute required? If it is, set to `TRUE` to trigger error on invalid result
#' @docType methods
#' @rdname hzdesgnname
#' @export
setMethod("hzdesgnname", signature(object = "SoilProfileCollection"),
          function(object, required = FALSE) {
            .require.metadata.aqp(object,
                                  attr = "aqp_hzdesgn",
                                  attrlabel = "Horizon designation",
                                  message ="\nSee ??hzdesgnname",
                                  required = required)
          })

setGeneric('hzdesgnname<-', function(object, required = FALSE, value)
  standardGeneric('hzdesgnname<-'))

#' @description `hzdesgnname<-`: Set horizon designation column name
#' @param object a SoilProfileCollection
#' @param value character, name of column containing horizon designations
#' @param required logical, is this attribute required? If it is, set to `TRUE` to trigger error on invalid `value`.
#' @docType methods
#' @seealso [hzDesgn()]
#' @rdname hzdesgnname
#' @export
#' @examples
#'
#' data(sp1)
#'
#' # promote to SPC
#' depths(sp1) <- id ~ top + bottom
#'
#' # set horizon designation column
#' hzdesgnname(sp1) <- "name"
#'
#' # get horizon designation column
#' hzdesgnname(sp1)
setReplaceMethod("hzdesgnname",
                 signature(object = "SoilProfileCollection"),
                 function(object, required = FALSE, value) {
                   .set.metadata.aqp(
                     object = object,
                     value = value,
                     required =  required,
                     attr = "aqp_hzdesgn",
                     attrlabel = "Horizon designation name",
                     message = "\nSee ??hzdesgnname",
                     allowednames = "horizon"
                   )
                 })


setGeneric("hzDesgn", function(object, ...)
  standardGeneric("hzDesgn"))

#' Get horizon designation column name
#'
#' @description Get horizon designation names
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases hzDesgn
#' @rdname hzDesgn
#' @export
setMethod("hzDesgn", signature(object = "SoilProfileCollection"),
          function(object) {
            
            h <- object@horizons
            hzd <- hzdesgnname(object)
            
            if (length(hzd)) {
              
              if (hzd %in% horizonNames(object)) {
                res <- h[[hzd]]
                return(res)
              }
              
            } else {
              
              stop("horizon designation name (",
                   hzd,
                   ") not in horizonNames().",
                   call. = FALSE)
            }
            
          })

## get column containing horizon designations
setGeneric("hztexclname", function(object, required = FALSE)
  standardGeneric("hztexclname"))

setGeneric('hztexclname<-', function(object, required = FALSE, value)
  standardGeneric('hztexclname<-'))

#' @title Get or Set Horizon Texture Class Column Name
#' @name hztexclname
#' @aliases hztexclname hztexclname,SoilProfileCollection-method hztexclname<- hztexclname<-,SoilProfileCollection-method
#' @description `hztexclname()`: Get column name containing horizon designation name
#' @details Store the column name containing horizon texture classes or other identifiers in the metadata slot of the SoilProfileCollection.
#' @param object a SoilProfileCollection 
#' @param required logical, is this attribute required? If it is, set to `TRUE` to trigger error on invalid result
#' @docType methods
#' @rdname hztexclname
#' @export
setMethod("hztexclname", signature(object = "SoilProfileCollection"),
          function(object, required = FALSE) {
            .require.metadata.aqp(object,
                                  attr = "aqp_hztexcl",
                                  attrlabel = "Horizon texture class",
                                  message = "\nSee ??hztexclname",
                                  required = required)
          })
#' @description `hztexclname<-`: Set horizon texture class column name for a SoilProfileCollection
#' @param object a SoilProfileCollection
#' @param value character, name of column containing horizon texture classes
#' @param required logical, is this attribute required? If it is, set to `TRUE` to trigger error on invalid `value`.
#' @docType methods
#' @rdname hztexclname
#' @export
#' @examples
#'
#' data(sp1)
#'
#' # promote to SPC
#' depths(sp1) <- id ~ top + bottom
#'
#' # set horizon texture class column
#' hztexclname(sp1) <- "texture"
#'
#' # get horizon texture class column
#' hztexclname(sp1)
setReplaceMethod("hztexclname", signature(object = "SoilProfileCollection"),
                 function(object, required = FALSE, value) {
                   .set.metadata.aqp(
                     object = object,
                     value = value,
                     required =  required,
                     attr = "aqp_hztexcl",
                     attrlabel = "Horizon texture class name",
                     message = "\nSee ??hztexclname",
                     allowednames = "horizon"
                   )
                 })

## get column containing horizon designations (there is a setter of same name)

setGeneric("GHL", function(object, required = FALSE)
  standardGeneric("GHL"))

#' @title Get or Set Generalized Horizon Label (GHL) Column Name
#' @name GHL
#' @aliases GHL GHL,SoilProfileCollection-method GHL<- GHL,SoilProfileCollection-method
#' @details Store the column name containing generalized horizon labels in the metadata slot of the SoilProfileCollection.
#' @description `GHL()`: Get column name containing generalized horizon labels 
#' @param object a SoilProfileCollection
#' @param required logical, is this attribute required? If it is, set to `TRUE` to trigger error on invalid result
#' @docType methods
#' @rdname GHL
#' @export
setMethod("GHL", signature(object = "SoilProfileCollection"),
          function(object, required = FALSE) {
            .require.metadata.aqp(object,
                                  attr = "aqp_ghl",
                                  attrlabel = "Generalized Horizon Label",
                                  message = "\nSee ??GHL",
                                  required = required)
          })

setGeneric('GHL<-', function(object, required = FALSE, value)
  standardGeneric('GHL<-'))

#' @description `GHL<-`: Set generalized horizon label column name
#' @param object a SoilProfileCollection
#' @param value character, name of column containing generalized horizon labels
#' @param required logical, is this attribute required? If it is, set to `TRUE` to trigger error on invalid `value`.
#' @docType methods
#' @rdname GHL
#' @export
#' @examples
#'
#' data(sp1)
#'
#' # promote to SPC
#' depths(sp1) <- id ~ top + bottom
#'
#' # set horizon designation column
#' GHL(sp1) <- "name"
#'
#' # get horizon designation column
#' GHL(sp1)
setReplaceMethod("GHL",
                 signature(object = "SoilProfileCollection"),
                 function(object, required = FALSE, value) {
                   .set.metadata.aqp(
                     object = object,
                     value = value,
                     required = required,
                     attr = "aqp_ghl",
                     attrlabel = "Generalized Horizon Label",
                     message = "\nSee ??GHL",
                     allowednames = "horizon"
                   )
                 })

setGeneric("hzmetaname", function(object, attr, required = FALSE)
  standardGeneric("hzmetaname"))

#' @title Get or Set Horizon Metadata Column Name
#' @name hzmetaname
#' @aliases hzmetaname hzmetaname,SoilProfileCollection-method hzmetaname<- hzmetaname,SoilProfileCollection-method
#' @details Store the column name containing a specific type of horizon data in the metadata slot of the SoilProfileCollection.
#' @description `hzmetaname()`: Get column name containing horizon data of interest 
#' @param object a SoilProfileCollection
#' @param attr character. Base name for attribute to be stored in metadata. This is prefixed with `"aqp_hz"` for horizon-level metadata for column attributes. e.g. `attr="clay"` results in metadata value retrieved from `"aqp_hzclay"`.
#' @param required logical, is this attribute required? If it is, set to `TRUE` to trigger error on invalid result
#' @docType methods
#' @rdname hzmetaname
#' @export
setMethod("hzmetaname", signature(object = "SoilProfileCollection"),
          function(object, attr, required = FALSE) {
            .require.metadata.aqp(object,
                                  attr = paste0("aqp_hz", attr),
                                  attrlabel = paste0("Horizon metadata (", attr, ") column"),
                                  message = "\nSee ??hzmetaname",
                                  required = required)
          })

setGeneric('hzmetaname<-', function(object, attr, required = FALSE, value)
  standardGeneric('hzmetaname<-'))

#' @description `hzmetaname<-`: Set horizon designation column name
#' @param object A _SoilProfileCollection_
#' @param attr _character_. Base name for attribute to be stored in metadata. This is prefixed with `"aqp_hz"` for horizon-level metadata for column attributes. e.g. `attr="clay"` results in metadata value retrieved from `"aqp_hzclay"`.
#' @param value _character_. Name of horizon-level column containing  data corresponding to `attr`.
#' @param required _logical_. Is this attribute required? If it is, set to `TRUE` to trigger error on invalid `value`.
#' @docType methods
#' @seealso [guessHzAttrName()]
#' @rdname hzmetaname
#' @export
#' @examples
#'
#' data(sp1)
#'
#' # promote to SPC
#' depths(sp1) <- id ~ top + bottom
#'
#' # set important metadata columns
#' hzdesgnname(sp1) <- "name"
#' hztexclname(sp1) <- "texture"
#' 
#' # set custom horizon property (clay content) column
#' hzmetaname(sp1, "clay") <- "prop"
#' 
#' # inspect metadata list
#' metadata(sp1)
#'
#' # get horizon clay content column
#' hzmetaname(sp1, "clay")
#' 
#' # uses hzdesgname(), hztexclname(), hzmetaname(attr="clay") in function definition
#' estimatePSCS(sp1)
setReplaceMethod("hzmetaname",
                 signature(object = "SoilProfileCollection"),
                 function(object, attr, required = FALSE, value) {
                   .set.metadata.aqp(
                     object = object,
                     value = value,
                     required =  required,
                     attr = paste0("aqp_hz", attr),
                     attrlabel = paste0("Horizon metadata (", attr, ") column"),
                     message = "\nSee ??hzmetaname",
                     allowednames = "horizon"
                   )
                 }
)

setGeneric("mutate_profile", function(object,
                                      ...,
                                      col_names = NULL,
                                      horizon_level = NULL)
  standardGeneric("mutate_profile"))

#' @title Transform a SPC (by profile) with a set of expressions
#' @name mutate_profile
#' @aliases mutate_profile,SoilProfileCollection-method
#' @description \code{mutate_profile()} is a function used for transforming SoilProfileCollections. Each expression is applied to site or horizon level attributes of individual profiles. This distinguishes this function from \code{transform}, which is applied to all values in a collection, regardless of which profile they came from.
#' @param object A SoilProfileCollection
#' @param ... A set of comma-delimited R expressions that resolve to a transformation to be applied to a single profile e.g \code{mutate_profile(hzdept = max(hzdept) - hzdept)}
#' @param col_names character. Optional column names. Should match the number of expressions in `...`.
#' @param horizon_level logical. If `TRUE` results of expressions are added to the SoilProfileCollection's horizon slot, if `FALSE` the results are added to the site slot. If `NULL` (default) the results are stored in the site or horizon slot based on the number of rows in each slot compared to the length of the result calculated from the _first_ and _last_ profile in the collection.
#' 
#' @details If the length an expression's result matches the number of horizons, the result is stored as a horizon-level variable. If the result has length 1, it is stored as a site-level variable. In the ambiguous case where the first and last profile have only _one_ horizon, the results are stored in the horizon slot by default. To force results into site slot use `horizon_level = FALSE`.
#' @return A SoilProfileCollection.
#' @author Andrew G. Brown.
#'
#' @rdname mutate_profile
#' @export mutate_profile
#' @examples
#' 
#' data(sp4)
#' depths(sp4) <- id ~ top + bottom
#' 
#' mutate_profile(sp4, clay_wtd_average = weighted.mean(clay, bottom - top))
#' 
setMethod("mutate_profile", signature(object = "SoilProfileCollection"), 
          function(object, ..., col_names = NULL, horizon_level = NULL) {

    # capture expression(s) at function
    .dots <- substitute(list(...))
    .dots <- .dots[2:length(.dots)]
    .names <- names(.dots)
    
    if (is.null(.names)) {
      if (!is.null(col_names) && length(col_names) == length(.dots)) {
        .names <- col_names
      } else {
        .names <- as.character(.dots)
      }
    }
    
    mutate_profile_raw(object, .dots, .names, horizon_level = horizon_level)
})

#' @param expr A list of expressions in terms of column names in site or horizon table of `object`
#' @rdname mutate_profile
#' @export 
#' @examples
#' data(jacobs2000)
#' 
#' set.seed(123)
#' 
#' ## col_names allows for column names to be calculated
#' x <- mutate_profile(jacobs2000, bottom - top / sum(bottom - top),
#'                     col_names = paste0("relthk", floor(runif(1, 0, 100))))
#' x$relthk28
#' 
#' # mutate_profile_raw allows for lists of expressions to be evaluated
#' master_desgn <- c("O", "A", "E", "B", "C", "R", "L", "M")
#' thk_names <- paste0("thk_", master_desgn)
#' 
#' # calculate thickness for each horizon
#' x$thk <- x$bottom - x$top
#' 
#' ## construct an arbitrary number of expressions using variable inputs
#' ops <- lapply(master_desgn, function(x) {
#'   substitute(sum(thk[grepl(PATTERN, name)], na.rm = TRUE), list(PATTERN = x))
#' })
#' names(ops) <- thk_names
#' 
#' # do mutation
#' y <- mutate_profile_raw(x, ops)
#' 
#' site(y)[c(idname(y), thk_names)]
mutate_profile_raw <- function(object, expr, col_names = NULL, horizon_level = NULL) {
  
  idn <- idname(object)
  hzidn <- hzidname(object)
  
  # preference is use col_names if specified
  
  # if no col_names, use the names of expr list
  if (is.null(col_names)) {
    col_names <- names(expr)
  }
  
  # if expr is unnamed, use character conversion of expression
  if (is.null(col_names)) {
    col_names <- as.character(expr)
  }
  
  # cleaner to have horizon_level be applied to each expression independently
  hzin <- horizon_level
  if (is.null(horizon_level) || !is.logical(horizon_level)) {
    horizon_level <- rep(FALSE, length(expr))
  }
  
  x <- data.table::data.table(object@site)[object@horizons, on = idn]
  o1 <- object[1, ]
  o2 <- object[nrow(object), ]  
  o1c <- compositeSPC(o1)
  o2c <- compositeSPC(o2)
  
  # iterate over expressions left to right
  for (i in 1:length(expr)) {
    
    # default is to create site-level properties unless result matches number of horizons
    # decide whether we are adding/modifying a site or horizon level variable so
    #  that degenerate cases do not create identical columns in site and horizon table or get put in unexpected slot
    #  2021-10-29: updated to use first and last profile, and allowing user override via argument
    di <- expr[[i]]
    res_eval1 <- .data_dots(o1c, eval(di))[[1]]
    res_eval2 <- .data_dots(o2c, eval(di))[[1]]
    
    # allow user to override the determination
    # check length of first/last profile result against number of horizons
    if (length(res_eval1) == nrow(o1) && length(res_eval2) == nrow(o2)) {
      horizon_level[i] <- TRUE
    }
    .SD <- NULL
    
    # remove any existing columnnames before joining in result
    if (any(col_names[i] %in% names(object))) {
      for (n in col_names[i]) {
        object[[n]] <- NULL
      }
    }
    
    if (isFALSE(hzin) || !horizon_level[i]) {
      res <- x[, list(eval(expr[[i]], envir = .SD)), by = c(idn)]
      if (length(res[[2]]) > length(object)) {
        stop("mutate_profile: some profiles returned more than one result and `horizon_level=FALSE`", call. = FALSE)
      }
      colnames(res) <- c(idn, col_names[i])
      site(object) <- res
    } else {
      res <- x[, list(.hzidname = .SD[[hzidn]], eval(expr[[i]], envir = .SD)), by = c(idn)]
      colnames(res) <- c(idn, hzidn, col_names[i])
      horizons(object) <- res
    }
    
  }
  
  return(object)
}

data(sp3)
depths(sp3) <- id ~ top + bottom

test_that("transform & mutate_profile", {
  
  # transform
  res <- transform(sp3, thickness = bottom - top)
  expect_equal(sum(res$thickness), 854)
  
  # transform (existing column)
  res <- transform(sp3, thickness = (bottom - top) / 100)
  expect_equal(sum(res$thickness), 8.54)
  
  # mutate_profile
  res <- mutate_profile(res, relthickness = (bottom - top) / (sum(thickness) * 100))
  expect_equal(sum(res$relthickness), 10)
  
  # mutate_profile (two existing columns)
  res <- mutate_profile(res, thickness = bottom - top,
                             relthickness = (thickness) / sum(thickness),
                             sumrelthickness1 = sum(relthickness))
  res <- mutate_profile(res, sumrelthickness2 = sum(relthickness))
  expect_equal(mean(res$relthickness), 0.2173913)
  expect_equal(length(res$sumrelthickness1), 10)
  expect_equal(length(res$sumrelthickness2), 10)

  # mutate existing column name (using same column as input)
  res <- mutate_profile(res, thickness = thickness / 10,
                             thickness = thickness * 10)
  
  # degenerate case where most profiles have only one horizon
  res2 <- mutate_profile(trunc(res, 0, 5), rt2 = (bottom - top) / sum(thickness))
  expect_true(length(res2$rt2) == nrow(res2))
  
  # forcing horizon level result into site produces an error
  expect_error({res3 <- mutate_profile(trunc(res, 0, 5), rt2 = (bottom - top) / sum(thickness), horizon_level = FALSE)})
  
  # however forcing a site-level result into horizon works (using custom column name)
  res4 <- mutate_profile(trunc(res, 0, 5), sum((bottom - top) / sum(thickness)), col_names = "foo", horizon_level = TRUE)
  expect_equal(length(res4$foo), nrow(res4))
})

test_that("mutate_profile_raw", {
  data(jacobs2000)
  set.seed(123)
  x <- mutate_profile(jacobs2000, bottom - top, 
                      col_names = paste0("thk", floor(runif(1, 0, 100))))
  expect_false(is.null(x$thk28))
  
  # example with dynamic number of columns and names
  master_desgn <- c("O", "A", "E", "B", "C", "R", "L", "M")
  thk_names <- paste0("thk_", master_desgn)
  
  x$thk <- x$bottom - x$top
  
  ## construct an arbitrary number of expressions using variable inputs
  ops <- lapply(master_desgn, function(x) substitute(sum(thk[grepl(VAR, name)], na.rm = TRUE), list(VAR = x)))
  names(ops) <- thk_names
  
  # do mutation
  y <- mutate_profile_raw(x, ops)
  
  expect_true(all(c(idname(y), thk_names) %in% siteNames(y)))
})

# repair an SPC by breaking into pieces and re-assembling
# likely only used to fix outdated SPC objects that are missing slots

#' Rebuild a SoilProfileCollection object
#'
#' Rebuild a SoilProfileCollection object
#'
#' Attempt rebuilding a \code{SoilProfileCollection} object by splitting into
#' components and re-assembling. Likely only used to fix outdated
#' \code{SoilProfileCollection} objects that are missing slots.
#'
#' @param x a \code{SoilProfileCollection} object
#' @return A valid \code{SoilProfileCollection} object.
#' @author D.E. Beaudette
#' @seealso \code{\link{checkSPC}}
#' Rebuild a SoilProfileCollection object
#'
#' Rebuild a SoilProfileCollection object
#'
#' Attempt rebuilding a \code{SoilProfileCollection} object by splitting into
#' components and re-assembling. Likely only used to fix outdated
#' \code{SoilProfileCollection} objects that are missing slots.
#'
#' @param x a \code{SoilProfileCollection} object
#' @return A valid \code{SoilProfileCollection} object.
#' @author D.E. Beaudette, A.G. Brown
#' @seealso \code{\link{checkSPC}}
#' @export
rebuildSPC <- function(x) {
  # break into pieces as list
  x.list <- suppressWarnings(as(x, 'list'))

  if (.hasSlot(x, "sp")) {
    x.list <- x.list[!names(x.list) %in% "sp"]
  }

  # seed object for new SPC
  res <- .as.data.frame.aqp(x.list$horizons, aqp_df_class(x))

  # init SPC from pieces
  # note: using depths<- because it will generate a horizon ID
  fm <- as.formula(sprintf("%s ~ %s + %s", x.list$idcol, x.list$depthcols[1], x.list$depthcols[2]))
  depths(res) <- fm

  # preserve original hzidname
  # this is missing in old SPC objects
  if (!is.null(x.list$hzidcol)) {
    hzidname(res) <- x.list$hzidcol
  }
  
  if (is.null(x.list$hzdesgncol)) {
    x.list$hzdesgncol <- ""
  }
  
  # preserve original hzdesgnname
  # this is missing in old SPC objects
  if (!is.null(x.list$hzdesgncol)) {
    hzdesgnname(res) <- x.list$hzdesgncol
  }
  
  if (is.null(x.list$hztexclcol)) {
    x.list$hztexclcol <- ""
  }
  
  # preserve original hztexclname
  # this is missing in old SPC objects
  if (!is.null(x.list$hztexclcol)) {
    hztexclname(res) <- x.list$hztexclcol
  }
  
  # transfer metadata
  res <- .transfer.metadata.aqp(x, res)
  
  # replace site
  site(res) <- x.list$site
  
  # @diagnostic and may be missing if `x` is very old
  # replace with empty data.frame so that setters do not error
  if (is.null(x.list$diagnostic)) {
    x.list$diagnostic <- data.frame()
  }
  
  # @restructions may be missing if `x` is very old
  # replace with empty data.frame so that setters do not error
  if (is.null(x.list$restrictions)) {
    x.list$restrictions <- data.frame()
  }
  
  # set
  diagnostic_hz(res) <- x.list$diagnostic
  restrictions(res) <- x.list$restrictions

  # 2022-12-21: aqp 2.0; sp slot is no longer transferred by rebuildSPC
  #             but we transfer any existing data back to site 
  if (!is.null(x.list$sp)) {
    if (inherits(x.list$sp, 'SpatialPoints') &&
        ncol(x.list$sp@coords) == 2) {
      newsp <- cbind(data.frame(id = profile_id(x), stringsAsFactors = FALSE), 
                     as.data.frame(x.list$sp, stringsAsFactors = FALSE))
      
      crdnms <- c("x", "y")
      if (any(crdnms %in% names(x))) {
        crdnms <-  paste0("sp.", crdnms)
      } 
      colnames(newsp) <- c(idname(x), crdnms)
      
      site(res) <- newsp
      initSpatial(res) <- colnames(newsp)[2:3]
      try(prj(res) <- x.list$sp@proj4string, silent = TRUE)
    }
  }
  return(res)
}

test_that("metadata and attributes are persistent", {

  data("sp3")
  
  depths(sp3) <- id ~ top + bottom
  
  depth_units(sp3) <- "fathoms"
  metadata(sp3)$something_custom <- "all your profiles are belong to us"
  attr(sp3, 'attribution') <- "the most attributable of all attributes"
  
  # thickness standard deviation for perturb()
  sp3$thickness_attr <- rep(1, nrow(sp3))
  
  cols <- c(
    "aqp_df_class",
    "aqp_group_by",
    "aqp_hzdesgn",
    "aqp_hztexcl",
    "depth_units",
    "stringsAsFactors",
    "something_custom"
  )
  
  .check_metadata_and_attribution <- function(obj, src, check_attr = TRUE) {
    # check metadata() other than original.order
    expect_equal(metadata(src)[cols], metadata(obj)[cols])
    # check attribute persistence via .transfer.attributes.aqp
    if (check_attr) expect_equal(attr(src, 'attribution'),  attr(obj, 'attribution'))
  }
  
  # compare when subsetting
  obj <- sp3[1,]
  .check_metadata_and_attribution(obj, sp3)
  
  # compare when combined
  obj <- combine(sp3[1,], sp3[2,])
  # NOTE: do not expect attributes to be transferred after combination operation
  .check_metadata_and_attribution(obj, sp3, check_attr = FALSE)
  
  # compare when rebuilt
  obj <- rebuildSPC(sp3)
  .check_metadata_and_attribution(obj, sp3)
  
  # compare when perturbed
  obj <- perturb(sp3[1,], thickness.attr = "thickness_attr")
  .check_metadata_and_attribution(obj, sp3)
  
  skip_if_not_installed('Gmedian')
  
  # compare when L1_profile-d
  obj <- L1_profiles(sp3[1:3, ], id ~ clay + ph + tc + cec, method = "simple")
  .check_metadata_and_attribution(obj, sp3)
  
})

## TODO: consider hue-based, angular distance between 2 hues


#' @title Munsell Hue Reference and Position Searching
#'
#' @description The 40 Munsell hues are typically arranged from 5R to 2.5R moving clock wise on the unit circle. This function matches a vector of hues to positions on that circle, with options for setting a custom origin or search direction.
#'
#' This function is fully vectorized.
#'
#' @param x character vector of hues, e.g. c('10YR', '5YR'), optional if `returnHues = TRUE`
#' 
#' @param returnHues logical, should the full set of Munsell hues be returned? See details.
#' 
#' @param includeNeutral logical, add 'N' to the end of the full set of Munsell hues 
#' 
#' @param origin hue to be used as the starting point for position searches (position 1)
#' 
#' @param direction indexing direction, should be `cw` (clock wise) or `ccw` (counter-clock wise)
#' 
#' @return A vector of integer hue positions is returned, of the same
#' length and order as `x`. If `returnHues = TRUE`, then all hue
#' names and ordering are returned and `x` is ignored.
#' 
#' @author D.E. Beaudette
#' 
#' @seealso [colorContrast], [huePositionCircle]
#' 
#' @references
#'  - Soil Survey Technical Note 2 [wayback machine URL](https://web.archive.org/web/20220704214918/https://www.nrcs.usda.gov/wps/portal/nrcs/detail/soils/ref/?cid=nrcs142p2_053569)
#' 
#'  - Munsell book of color. 1976. Macbeth, a Division of Kollmorgen Corp., Baltimore, MD.
#' 
#' @keywords manip
#' @export
#' @examples
#'
#' # get hue ordering for setting levels of a factor
#' huePosition(returnHues = TRUE)
#' 
#' # get hue ordering including N (neutral)
#' huePosition(returnHues = TRUE, includeNeutral = TRUE)
#'
#' # get position of the '10YR' hue, relative to standard origin of '5R'
#' # should be 7
#' huePosition(x = '10YR')
#' 
#' # get position of the '10YR' hue, relative to standard origin of '5YR'
#' # should be 3
#' huePosition(x = '10YR', origin = '5YR')
#' 
#' # visualize
#' op <- par(mar = c(0, 0, 0, 0), fg = 'white', bg = 'black')
#' 
#' huePositionCircle(huePosition(returnHues = TRUE, origin = '5YR'))
#' 
#' par(op)
#'
huePosition <- function(x, returnHues = FALSE, includeNeutral = FALSE, origin = '5R', direction = c('cw', 'ccw')) {
  
  # ordering via Tech Note #2
  # Soil Survey Technical Note 2 [wayback machine URL](https://web.archive.org/web/20220704214918/https://www.nrcs.usda.gov/wps/portal/nrcs/detail/soils/ref/?cid=nrcs142p2_053569)
  
  # sacrifice to CRAN deity
  munsellHuePosition <- NULL
  
  # note: this is incompatible with LazyData: true
  # load look-up table from our package
  load(system.file("data/munsellHuePosition.rda", package = "aqp")[1])
  
  ## basic error checking / argument processing
  
  # returnHues must be logical
  if(! inherits(returnHues, 'logical')) {
    stop('returnHues must be TRUE or FALSE', call. = FALSE)
  }
  
  # origin must be a valid hue
  if(! origin %in% munsellHuePosition$hues) {
    stop('invalid hue', call. = FALSE)
  }
  
  ## indexing direction 
  direction <- match.arg(direction)
  
  ## extract hues
  hues <- munsellHuePosition$hues
  
  ## optionally retain neutral, not typically useful
  if(! includeNeutral) {
    hues <- hues[which(hues != 'N')]
  }
  
  ## adjust origin as needed
  start.idx <- match(origin, hues)
  
  # anything other than 5R (position 1) requires an adjustment
  if(start.idx > 1) {
    # wrap-around
    before.idx <- 1:(start.idx - 1)
    hues <- hues[c(start.idx:length(hues), before.idx)]
  } else {
    # no change
    # hues <- hues[start.idx:length(hues)]
  }
  
  # enforce search direction
  if(direction == 'ccw') {
    hues <- c(hues[1], rev(hues[-1]))
  }

  # just the hues
  if(returnHues) {
    return(hues)
  } else {
    # convert hue into position
    res <- match(x, hues)
    return(res)
  }

}

## Coercion methods: general

#' Coerce SoilProfileCollection with \code{as()}
#' 
#' @description SoilProfileCollections can be coerced to other R object types using \code{as(spc, 'type')}.
#' 
#' Possible endpoints include: \code{list}, \code{data.frame}, \code{SpatialPointsDataFrame} and \code{SpatialPoints}.
#'
#' @name as
#' @return list
#' @aliases as,SoilProfileCollection-method
#' @docType methods
#' @rdname coercion-methods
#'
#' @examples
#' # load example data stored as SoilProfileCollection
#' data(sp5)
#' 
#' # sp5
#' str(sp5)
#' 
#' # list output
#' str(as(sp5, 'list'))
#'
#' # data.frame output
#' str(as(sp5, 'data.frame'))
#' 
#' # Spatial Objects
#' # make some random coordinate data for each profile
#' sp5$x <- sp5$y <- rnorm(length(sp5))
#' initSpatial(sp5, crs = "OGC:CRS84") <- ~ x + y
#' 
#' if (requireNamespace("sf")) {
#' 
#'   # sf output
#'   str(as(sp5, 'sf'))
#' 
#'   # SpatialPointsDataFrame output
#'   str(as(sp5, 'SpatialPointsDataFrame'))
#' 
#'   # SpatialPoints output
#'   str(as(sp5, 'SpatialPoints'))
#'   
#' }
#' 
setAs("SoilProfileCollection", "list", function(from) {
  
  # get slot names from prototype
  sn <- slotNames(from)
  
  # test for presence of all slots
  # copy contents over to list with same name
  # if missing return NULL + warning
  s.list <- lapply(sn, function(i) {
    if(.hasSlot(from, name=i)) {
      res <- slot(from, i)
    } else {
      res <- NULL
    }
    return(res)
  })
  
  # copy slot names
  names(s.list) <- sn
  
  # test for missing slots
  if(any(sapply(s.list, is.null))) {
    warning("some slots were missing, use rebuildSPC to fix", call. = FALSE)
  }
  
  return(s.list)
  
})

#' @name as
#' @return data.frame
#' @aliases as,SoilProfileCollection-method
#' @docType methods
#' @rdname coercion-methods
#'
setAs("SoilProfileCollection", "data.frame", function(from) {
  s <- getSpatial(from)
  
  # horizons + site + coordinates
  if(nrow(site(from)) > 0 & nrow(s) == length(from)) {
    site.coords <- data.frame(site(from), s, stringsAsFactors = FALSE)
    return(merge(horizons(from), site.coords, by = idname(from), sort = FALSE, all.x = TRUE))
  }
  
  # horizons + site
  if(nrow(site(from)) > 0 & ! nrow(s) == length(from))
    return(merge(horizons(from), site(from), by = idname(from), sort = FALSE, all.x = TRUE))
    
  # horizons + coordinates
  if(! nrow(site(from)) > 0 & nrow(s) == length(from)) {
    ids.coords <- data.frame(profile_id(from), s, stringsAsFactors = FALSE)
    return(data.frame(horizons(from), ids.coords, stringsAsFactors = FALSE))
  }
  
  # just horizons
  else {
    return(horizons(from))
  }
})

#' @name as
#' @return tbl_df
#' @aliases as,SoilProfileCollection-method
#' @docType methods
#' @rdname coercion-methods
#'
setAs("SoilProfileCollection", "tbl_df", function(from) {
  .as.data.frame.aqp(as(from,'data.frame'), 'tbl_df')
})

#' @name as
#' @return data.table
#' @aliases as,SoilProfileCollection-method
#' @docType methods
#' @rdname coercion-methods
#'
setAs("SoilProfileCollection", "data.table", function(from) {
  .as.data.frame.aqp(as(from, 'data.frame'), 'data.table')
})

#' @name as
#' @return SpatialPointsDataFrame
#' @aliases as,SoilProfileCollection-method
#' @docType methods
#' @rdname coercion-methods
setAs("SoilProfileCollection", "SpatialPointsDataFrame", function(from) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required to coerce a SoilProfileCollection to SpatialPointsDataFrame object", call. = FALSE)
  }
  sf::as_Spatial(as(from, 'sf'))
}
)

#' @name as
#' @return sf
#' @aliases as,SoilProfileCollection-method
#' @docType methods
#' @rdname coercion-methods
setAs("SoilProfileCollection", "sf", 
  function(from) {

    if (!requireNamespace("sf", quietly = TRUE)) {
      stop("Package 'sf' is required to coerce a SoilProfileCollection to sf object", call. = FALSE)
    }
    
    crd <- metadata(from)$coordinates
    prj <- metadata(from)$crs
    
    if (all(is.null(crd) | is.na(crd)) || length(crd) < 2) {
      stop("Two coordinate (X, Y) column names must be defined to coerce a SoilProfileCollection to sf object", call. = FALSE)
    }
    
    if (is.null(prj)) {
      prj <- NA_character_
    }
    
    # keep empty point geometries, and do not remove original columns
    s <- sf::st_as_sf(site(from), crs = prj, coords = crd, na.fail =  FALSE, remove = FALSE)
    message('only site data are extracted')
    return(s)
  }
)

#' @name as
#' @return SpatialPoints
#' @aliases as,SoilProfileCollection-method
#' @docType methods
#' @rdname coercion-methods
setAs("SoilProfileCollection", "SpatialPoints", function(from) {
  
    if (!requireNamespace("sf", quietly = TRUE)) {
      stop("Package 'sf' is required to coerce a SoilProfileCollection to sf object", call. = FALSE)
    }
  
    sf::as_Spatial(sf::st_as_sfc(as(from, 'sf')))
  }
)

#' @param x a SoilProfileCollection
#' @export 
#' @rdname coercion-methods
setMethod("as.data.frame", "SoilProfileCollection", function(x) {
  as(x, 'data.frame')
})

####
#### single-bracket SoilProfileCollection methods
####

### NOTE: this DOES NOT re-order data, only subsets!
##
## matrix / DF style access: only to horizon data
##
## i = profile index
## j = horizon / slice index
##
##
#' Matrix/data.frame-like access to profiles and horizons in a SoilProfileCollection
#'
#' @aliases  [,SoilProfileCollection-method
#'
#' @description You can access the contents of a SoilProfileCollection by profile and horizon "index", \code{i} and \code{j}, respectively: \code{spc[i, j, ...]}. Subset operations are propagated to other slots (such as diagnostics or spatial) when they result in removal of sites from a collection.
#'
#'  - \code{i} refers to the profile position within the collection. By default the order is based on the C SORT order of the variable that you specified as your unique profile ID at time of object construction. Note that if your ID variable was numeric, then it has been sorted as a character.
#'
#'  - \code{j} refers to the horizon or "slice" index. This index is most useful when either a) working with \code{slice}'d SoilProfileCollection or b) working with single-profile collections. \code{j} returns the layer in the specified index positions for all profiles in a collection.
#'
#'  - `...` is an area to specify an expression that is evaluated in the subset. Currently supported
#'
#'    - `.LAST` (last horizon in each profile): return the last horizon from each profile. This uses `i` but ignores the regular `j` index.
#'    - `.FIRST` (first horizon in each profile): return the last horizon from each profile. This uses `i` but ignores the regular `j` index.
#'    
#'    - `.HZID` (horizon index): return the horizon indices corresponding to `i`+`j`+`...` ("k") constraints
#'    - `.NHZ` (number of horizons): return the number of horizons in the profiles resulting from `i`+`j`+`...` ("k") constraints
#'
#' @param x a SoilProfileCollection
#' @param i a numeric or logical value denoting profile indices to select in a subset
#' @param j a numeric or logical value denoting horizon indices to select in a subset
#' @param ... non-standard expressions to evaluate in a subset
#' @param drop Default: `TRUE`. When `drop=FALSE` placeholder horizons (profile ID with all other values `NA`) are created where the specified `j` index results in removal of all horizons. 
#' @rdname singlebracket
#  SPC extract: "[" '[' single bracket SPC object extract method
setMethod("[", signature(x = "SoilProfileCollection"),
          function(x, i, j, ..., drop = TRUE) {

                           # capture k right away
                           ksubflag <- FALSE
                           
                           # local vars to make R CMD check happy
                           .N <- NULL
                           .SD <- NULL
                           .I <- NULL
                           V1 <- NULL
                           
                           idn <- idname(x)
                           hzd <- horizonDepths(x)
                           
                           # 2nd value is first user-supplied expression
                           kargs <- substitute(list(...))
                           ksub <- as.character(kargs)[2:length(kargs)]

                           # handle special keywords in "k" index
                           for (k in ksub) {
                             if (!is.na(k)) {
                               switch(k,
                                      ".FIRST" = {
                                        j <- 1
                                      },
                                      ".LAST" = {
                                        ksubflag <- TRUE
                                      },
                                      ".HZID" = {
                                        ksubflag <- TRUE
                                      },
                                      ".NHZ" = {
                                        ksubflag <- TRUE
                                      },
                                      ".TOP" = {
                                        ksubflag <- TRUE
                                      },
                                      ".BOTTOM" = {
                                        ksubflag <- TRUE
                                      })
                             }
                           }

                           # convert to integer
                           if (!missing(i)) {
                             if (any(is.na(i))) {
                               stop('NA not permitted in profile index', call. = FALSE)
                             }

                             # convert logical to integer
                             # (thanks Jose Padarian for the suggestion!)
                             if (is.logical(i)) {
                               i <- (1:length(x))[i]
                             }

                             can.cast <- is.numeric(i)
                             if (can.cast) {
                               if (all(abs(i - round(i)) < .Machine$double.eps ^ 0.5, na.rm = TRUE)) {
                                 i <- as.integer(i)
                               } else {
                                 stop("Numeric site index does not contain whole numbers.")
                               }
                             } else {
                               stop("Failed to coerce site index to integer.")
                             }
                           } else {
                             # if no index is provided, the user wants all profiles
                             i <- 1:length(x)
                           }


                           # this block does processing for logical or numeric j
                           # not special symbols like .LAST
                           if (!missing(j) & !ksubflag) {

                             # AGB added logical handling to horizon index
                             if (is.logical(j)) {
                               j <- (1:length(x))[j]
                             }

                             can.cast <- is.numeric(j)
                             if (can.cast) {
                               if (all(abs(j - round(j)) < .Machine$double.eps ^ 0.5)) {
                                 j <- as.integer(j)
                               } else {
                                 stop("Numeric horizon/slice index does not contain whole numbers.")
                               }
                             } else {
                               stop("Failed to coerce horizon/slice index to integer.")
                             }

                             if (any(is.na(j))) {
                               stop('NA not permitted in horizon/slice index', call. = FALSE)
                             }
                           }

                           # extract all site and horizon data
                           h <- x@horizons
                           s.all <- x@site

                           # extract requested profile IDs
                           p.ids <- s.all[[idn]][unique(i)]

                           # keep only the requested horizon data (filtered by profile ID)
                           h <- .as.data.frame.aqp(h, aqp_df_class(x))
                           pidx <- h[[idn]] %in% p.ids
                           h <- h[pidx,]
                            
                           # TODO: combine?
                           # if (ksubflag && ".NHZ" %in% ksub) { 
                           #   
                           #   j.idx.allowed <- seq_len(nrow(x@horizons))[pidx]
                           #   
                           # } else 
                           if (ksubflag && ".HZID" %in% ksub) {

                             j.idx.allowed <- seq_len(nrow(x@horizons))[pidx]
                           
                           } else {
                             j.idx.allowed <- seq_len(nrow(h))

                             # keep only the requested site data, (filtered by profile ID)
                             s.i <- which(s.all[[idn]] %in% p.ids)

                             # need to use drop=FALSE when @site contains only a single column
                             s <- s.all[s.i, , drop = FALSE]
                             
                             # copy empty SpatialPoints object
                             # sp <- x@sp
                             
                             # subset diagnostic data
                             d <- diagnostic_hz(x)
                             if (length(d) > 0) {
                               d <- d[which(d[[idn]] %in% p.ids),]
                             }

                             # subset restriction data
                             r <- restrictions(x)
                             if (length(r) > 0) {
                               r <- r[which(r[[idn]] %in% p.ids),]
                             }
                           }

                           # subset horizons/slices based on j --> only when j is given
                           if (!missing(j) || ksubflag) {

                             # faster replacement of j subsetting of horizon data
                             # if (aqp_df_class(x) == "data.table") {

                             h <- data.table::as.data.table(h)

                             # by list @horizons idname (essentially iterating over profiles)
                             bylist <- list(h[[idn]])
                             names(bylist) <- idn

                             # handle special symbols in j index
                             # currently supported:
                             #  - .FIRST: first horizon index per profile
                             #  - .LAST: last horizon index per profile
                             #  - .HZID: return horizon slot row index (short circuit)
                             #  - .TOP/.BOTTOM: return top/bottom depth (short circuit)
                             # the above can be combined. .LAST takes precedence over .FIRST

                             # determine j indices to KEEP

                             # default is an index to each horizon
                             j.idx <- j.idx.allowed

                             if (ksubflag && ".LAST" %in% ksub) {
                               # trigger special last horizon case
                               j.idx <- h[, .I[.N], by = bylist]$V1
                             } else {
                               if (!missing(j)) {
                                 # get row indices of horizon data corresponding to j within profiles
                                 j.idx <- h[, .I[1:.N %in% j], by = bylist]$V1
                               }
                             }
                             
                             # short circuit for horizon-slot indices
                             if (ksubflag && ".HZID" %in% ksub) {
                               return(j.idx)
                             # short circuit for top and/or bottom depths
                             } else if (ksubflag && any(c(".TOP", ".BOTTOM") %in% ksub)) {
                               res <- h[j.idx, .SD, .SDcols = hzd[c(".TOP", ".BOTTOM") %in% ksub]]
                               # vector default with 1 keyword
                               if (ncol(res) == 1) 
                                 return(res[[1]])
                               # TODO: should names be standardized for data.frame?
                               #       always top, bottom order even if keywords are reversed
                               return(.as.data.frame.aqp(res, aqp_df_class(x)))
                             } else if (ksubflag && ".NHZ" %in% ksub) {
                               # short circuit for number of horizons
                               return(h[j.idx, .N, by = list(bylist[[1]][j.idx])]$N)
                             }
                             
                             # determine which site indices to keep
                             # in case all horizons are removed, remove sites too
                             if (length(j.idx) == 0 && drop) {
                               i.idx <- numeric(0)
                             } else {
                               # determine which profiles to KEEP
                               i.idx <- which(profile_id(x) %in% unique(h[[idn]][j.idx]))
                             }
                             
                             # do horizon subset with j index
                             h <- h[j.idx, ]

                             # }
                             # } else {
                             #   # retain a base R way of doing things (plenty fast with SPCs up to ~100k or so)
                             #   j.res <- as.list(aggregate(
                             #     h[[hzidname(x)]],
                             #     by = list(h[[idname(x)]]),
                             #     FUN = function(hh) {
                             #       list(1:length(hh) %in% j)
                             #     },
                             #     drop = FALSE
                             #   )$x)
                             #
                             #   ##  https://github.com/ncss-tech/aqp/issues/89
                             #   # fix #89, where i with no matching j e.g. @site data returned
                             #   i.idx <- which(as.logical(lapply(j.res, function(jr) { any(jr) })))
                             #
                             #   j.idx <-  which(do.call('c', j.res))
                             # }

                             # find any index out of bounds and ignore them
                             # j.idx.bad <- which(abs(j.idx) > nrow(h))
                             # i.idx.bad <- which(abs(i.idx) > nrow(s))
                             #
                             # if (length(i.idx))
                             #   i.idx <- i.idx[-i.idx.bad]
                             #
                             # if (length(j.idx))
                             #   j.idx <- j.idx[-j.idx.bad]
                             
                             
                             if (length(i.idx) > 0 && drop) {
                               # remove sites that have no matching j
                               i.idx <- intersect(s.i, i.idx)
                             }
                             
                             # if sites have been removed based on the j-index+drop args
                             s <- s.all[i.idx, , drop = FALSE]
                             h.ids <- s[[idn]]
                             
                             if (!drop) {
                               res <- .insert_dropped_horizons(x, h, SPC = FALSE)
                               h <- res$horizons
                               s <- res$sites
                             }
                             
                             # remove also: diagnostics
                             d.idx <- which(d[[idn]] %in% h.ids)
                             d <- d[d.idx, , drop = FALSE]
                             
                             # restrictions
                             r.idx <- which(r[[idn]] %in% h.ids)
                             r <- r[r.idx, , drop = FALSE]
                           }
                           
                           rownames(h) <- NULL
                             
                           # rebuild SPC object from slots
                           res <-  SoilProfileCollection(
                             idcol = idn,
                             hzidcol = hzidname(x),
                             depthcols = hzd,
                             metadata = aqp::metadata(x),
                             horizons = .as.data.frame.aqp(h, aqp_df_class(x)),
                             site = .as.data.frame.aqp(s, aqp_df_class(x)),
                             # sp = sp,
                             diagnostic = .as.data.frame.aqp(d, aqp_df_class(x)),
                             restrictions = .as.data.frame.aqp(r, aqp_df_class(x))
                           )
                           
                           # preserve metadata that may have been customized relative to defaults
                           #  in prototype or resulting from construction of SPC
                           res <- .transfer.metadata.aqp(x, res)
                           
                           # fill in any missing data.frame class or group var
                           o.df.class <- aqp::metadata(x)$aqp_df_class
                           if (length(o.df.class) == 0) {
                             o.df.class <- "data.frame"
                           }

                           o.group.by <- aqp::metadata(x)$aqp_group_by
                           if (length(o.group.by) == 0) {
                             o.group.by <- ""
                           }

                           metadata(res)$aqp_df_class <- o.df.class
                           metadata(res)$aqp_group_by <- o.group.by
                           
                           # there should be as many records in @site as there are profile IDs
                           pid.res <- profile_id(res)
                           site.res <- site(res)[[idname(res)]]

                           if (length(pid.res) != length(site.res)) {
                             message("Some profiles have been removed from the collection.")
                           }

                           # the order of profile_ids should be the same as in @site
                           if (!all(pid.res == site.res)) {
                             warning("profile ID order does not match order in @site",
                                     call. = FALSE)
                           }

                           return(res)
                         })

####
#### double bracket SoilProfileCollection methods
####

#' Get column of horizon or site data in a SoilProfileCollection
#'
#' @name [[
#' @description
#'
#' Get the data from a column accessed by name. Column names other than profile ID are not shared between site and horizons.
#' Bonus: \code{[[} gives access to all site and horizon level variables in tab complete for RStudio using the magrittr pipe operator!
#'
#' @param x a SoilProfileCollection
#' @param i an expression resolving to a single column name in site or horizon table
#' @param j (not used)
#' @aliases [[,SoilProfileCollection-method [[,SoilProfileCollection,ANY-method [[,SoilProfileCollection,ANY,ANY-method 
#' @docType methods
#' @rdname doublebracket
#' @examples
#'
#' data(sp2)
#' depths(sp2) <- id ~ top + bottom
#' site(sp2) <- ~ surface
#'
#' # get with [[
#' sp2[['surface']]
#'
#' # get using "unknown" expression:
#' #  "2nd + 3rd horizon column names"
#' for(i in horizonNames(sp2)[2:3])
#'  print(sp2[[i]])
#'
#' data(sp5)
#'
#' # some column names to work with
#' rgb.columns <- c("R25","G25","B25")
#'
#' res <- lapply(rgb.columns, function(x) {
#'
#'   # [[ allows you to access column names in a loop
#'   round(sp5[[x]] * 255)
#'
#' })
#'
#' # rename scaled results
#' names(res) <- paste0(rgb.columns,"_scl")
#'
#' # add horizon ID to results
#' result <- data.frame(hzID = hzID(sp5), do.call('cbind', res))
#' head(result)
#'
#' # join result back into horizons
#' horizons(sp5) <- result

# accessor for site and horizon names via double bracket
#  site names in horizon names results return from site (idname)
#
# prevents:
#   "Error in object[[i]] : this S4 class is not subsettable"
#   which is an error caused by RStudio? when doing tab completion
#   with %>% operator on a SPC

setMethod("[[", signature(x = "SoilProfileCollection"),
          function(x, i, j) {
            if (length(i) == 1) {
              # site names take precedence for those
              #  shared between @site and @horizons (idname)
              if (i %in% siteNames(x))
                return(x@site[[i]])

              if (i %in% horizonNames(x))
                return(x@horizons[[i]])
            }
          })

#' Add or change column of horizon or site data in a SoilProfileCollection
#' @name [[<-
#' @description
#'
#' Add or change the data from a column accessed by name. Column names other
#' than profile ID are not shared between site and horizons. The benefit of
#' using double bracket setter over \code{$} is that \code{name} can be
#' calculated, whereas with \code{$}, it must be known a priori and hard coded.
#'
#' When using the double bracket setter the length of input and output matching
#' either the number of sites or number of horizons is used to determine which
#' slot new columns are assigned to.
#'
#' @param x a SoilProfileCollection
#' @param i an expression resolving to a single column name in site or horizon
#'   table-
#' @param value New value to replace -- unit length or equal in length to number
#'   of sites or horizons in the collection.
#'
#' @aliases [[<-,SoilProfileCollection-method [[<-,SoilProfileCollection,ANY,ANY-method 
#' @docType methods
#' @rdname doublebracket-set
setReplaceMethod("[[", signature(x = "SoilProfileCollection"),
                 function(x, i, value) {
                   lv <- length(value)
                   lx <- length(x)
                   nx <- nrow(x)

                   hznames <- horizonNames(x)
                   stnames <- siteNames(x)

                   # default to creating site var
                   #  as long as its not in horizon names
                   if ((!i %in% hznames) &
                       (i %in% stnames | lv == lx)) {
                     if (lv == lx | is.null(value)) {
                       s <- data.frame(x@site, stringsAsFactors = FALSE)
                       s[[i]] <- value
                       x@site <- .as.data.frame.aqp(s, aqp_df_class(x))
                     } else {
                       stop("replacement length does not match number of profiles!",
                            call. = FALSE)
                     }
                   } else if (i %in% hznames | lv == nx) {
                     if (lv == nx | is.null(value)) {
                       h <- data.frame(x@horizons, stringsAsFactors = FALSE)
                       h[[i]] <- value
                       x@horizons <- .as.data.frame.aqp(h, aqp_df_class(x))
                     } else {
                       stop("replacement length does not match number of horizons!",
                            call. = FALSE)
                     }
                   } else {
                     if(!is.null(value))
                       stop("new data must match either number of profiles or number of horizons",
                          call. = FALSE)
                   }

                   return(x)

                 })


#' @title Get data from column of horizon or site data in a SoilProfileCollection
#'
#' @name $
#' @description  Get the data from a column accessed by name `x$name`. Column
#'   names other than profile ID are not shared between site and horizons.
#'   Accessing the profile ID column (`idname(x)`) with `$` returns the
#'   horizon-level data with profile ID repeated for each horizon in a profile.
#'
#' @param x a SoilProfileCollection
#' @param name a single column name in site or horizon table
#' @docType methods
#' @rdname dollarsign
#'
#' @aliases $,SoilProfileCollection-method
#' @seealso [site()] [horizons()] [idname()] [profile_id()] 
#' @examples
#'
#' data(sp1)
#'
#' depths(sp1) <- id ~ top + bottom
#'
#' # get data from a column by name (prop)
#' sp1$prop
#' 
setMethod("$", signature(x = "SoilProfileCollection"),
          function(x, name) {
            # get names from site and hz data
            s.names <- siteNames(x)
            h.names <- horizonNames(x)

            # ## note: warnings may be issued when using auto-complete feature in RStudio
            # # when site data are initialized from an external DF, it is possible that
            # # there will be duplicate column names
            # if((name %in% h.names) && (name %in% s.names)) {
            #   warning('column name is present in horizon and site data, extracting from horizon data only', call.=FALSE)
            # }

            # get column from horizon data
            if (name %in% h.names) {
              res <- x@horizons[[name]]
            } else {
              # otherwise check site data
              if (name %in% s.names) {
                res <- x@site[[name]]
              } else {
                # if still missing return NULL
                res <- NULL
              }
            }

            return(res)
          })

#' Set data in column of horizon or site data in a SoilProfileCollection
#' @name $<-
#' @description Set the data in a column accessed by name \code{spc$name}. Column names other than profile ID are not shared between site and horizons.
#'
#' When using \code{$<-}, the length of input and output matching either the number of sites or number of horizons is used to determine which slot new columns are assigned to. Use \code{site(x)$name <- value} or  \code{horizons(x)$name <- value} to be explicit about which slot is being accessed.
#'
#' @param x a SoilProfileCollection
#' @param name a single column name in site or horizon table
#' @docType methods
#' @aliases $<-,SoilProfileCollection-method
#' @param value Replacement values: unit length or equal to number of horizons or sites.
#' @rdname dollarsign-set
setReplaceMethod("$", signature(x = "SoilProfileCollection"),
                 function(x, name, value) {
                   #print(name)

                   # extract hz and site data
                   h <- x@horizons
                   s <- x@site

                   # working with horizon data
                   if (name %in% names(h)) {
                     h[[name]] <- value
                     x@horizons <- h
                     return(x)
                   }

                   # working with site data
                   if (name %in% names(s)) {
                     s[[name]] <- value
                     x@site <- s
                     return(x)
                   }

                   # ambiguous: use length of replacement to determine: horizon / site
                   n.site <- nrow(s)
                   n.hz <- nrow(h)
                   l <- length(value)

                   if (l == n.hz) {
                     h[[name]] <- value
                     x@horizons <- h
                     return(x)
                   }

                   if(l == n.site) {
                     s[[name]] <- value
                     x@site <- s
                     return(x)
                   }

                   # otherwise, there is a problem
                   stop('length of replacement must equal number of sites or number of horizons')
                 }
)

# setReplaceMethod("$", signature(x = "SoilProfileCollection"),
#                  function(x, name, value) {
#                    #
#                    # existing values: replace by name
#                    #
#                    # new values: use length of value to
#                    #             determine site/horizon names
#                    #
#                    # cannot change site->horizon or vice versa
#                    # use NULL to remove a value from site or horizon
#                    x[[name]] <- value
#                    return(x)
#                   })

#' @title Horizon Transition Probabilities
#'
#' @description Functions for creating and working with horizon (sequence) transition
#' probability matrices.
#'
#' See the following tutorials for some ideas: 
#' 
#'   * [horizon designation TP](http://ncss-tech.github.io/AQP/aqp/hz-transition-probabilities.html)
#'   * [soil color TP](http://ncss-tech.github.io/AQP/aqp/series-color-TP-graph.html)
#'
#' @aliases hzTransitionProbabilities mostLikelyHzSequence
#' 
#' @param x a `SoilProfileCollection` object.
#' 
#' @param name A horizon level attribute in `x` that names horizons.
#' 
#' @param loopTerminalStates should terminal states loop back to themselves?
#' 
#' This is useful when the transition probability matrix will be used to
#' initialize a `markovchain` object. See examples below.
#' 
#' @return A square matrix of transition probabilities. See examples.
#'
#' The function [genhzTableToAdjMat()] returns a square adjacency matrix.
#' See examples.
#'
#' The function [mostLikelyHzSequence()] returns the most likely sequence of
#' horizons, given a `markovchain` object initialized from horizon
#' transition probabilities and an initial state, `t0`. See examples.
#' 
#' @note These functions are still experimental and subject to change.
#' 
#' @author D.E. Beaudette
#' @seealso [generalize.hz()]
#' @keywords manip
#' @export
#' @examples
#'
#' data(sp4)
#' depths(sp4) <- id ~ top + bottom
#'
#' # horizon transition probabilities: row -> col transitions
#' (tp <- hzTransitionProbabilities(sp4, 'name'))
#'
#'
#'
hzTransitionProbabilities <- function(x, name = GHL(x, required = TRUE), loopTerminalStates = FALSE) {

  # get all horizons
  h <- horizons(x)

  # sanit checks: no missing or NA horizon designation allowed
  idx <- which(h[[name]] == '' | is.na(h[[name]]) )
  if(length(idx) > 0) {
    message('NA horizon names have been removed from TP matrix')
    h <- h[-idx, ]
  }

  # get all hz names
  hz.names <- sort(unique(h[[name]]))
  n.names <- length(hz.names)
  # get profile IDs and depth column names
  # note that we cannot use the full set of IDs, as NA and "" hz names have been filtered
  # convert to character in case all IDs are integers, this makes it possible to address our list by ID
  id.name <- idname(x)
  pIDs <- as.character(unique(h[[id.name]]))
  dc <- horizonDepths(x)
  # init TP matrix with 0's
  m <- matrix(ncol=n.names, nrow=n.names, data = 0)
  # row / col names are entire set of hz names
  dimnames(m) <- list(hz.names, hz.names)

  # split by profile
  h.l <- split(h, h[[id.name]])

  # iterate over profiles
  for(i in pIDs) {
    # the current profile: hz names and top depths
    this.profile <- h.l[[i]][, c(name, dc[1])]
    # sort names by top depth, ascending order
    z <- this.profile[[name]][order(this.profile[, 2])]

    # transition probabilities require at least 2 horizons
    if(length(z) > 1) {
      # iterate over names
      for(j in 1:(length(z)-1)){
        # increment the current transition by 1
        m[z[j], z[j+1]] <- m[z[j], z[j+1]] + 1
      }
    }
  }

  # convert to probabilities by row
  m <- sweep(m, 1, rowSums(m), '/')

  # rows with all NaN are terminal states: they transition to nothing
  # optionally, create a loop (A -> A)
  # this ensures compatibility with markovchain package
  if(loopTerminalStates) {
    loop.hz <- names(which(apply(m, 1, function(i) all(is.nan(i)))))
    # if there are some loops, then set the diagonal for these states to 1
    if(length(loop.hz) > 0) {
      if(length(loop.hz) < 2)
        m[loop.hz, loop.hz] <- 1 # single replacement, no diagonal
      else
        diag(m[loop.hz, loop.hz]) <- 1 # multiple replacements, use diagonal
    }

  }

  # replace NaN with 0
  m[which(is.nan(m))] <- 0

  # test for ties
  f.ties <- function(i) {
    # 0 ties aren't important
    not.zero <- which(! zapsmall(i) == 0)
    # tabulate ties after rounding to 8 decimal places
    res <- table(round(i[not.zero], 8))
    if(any(res > 1))
      return(TRUE)
    else
      return(FALSE)
  }

  if(any(apply(m, 1, f.ties))) {
    warning('ties in transition probability matrix', call. = FALSE)
    attr(m, 'ties') <- TRUE
  }
  else
    attr(m, 'ties') <- FALSE


  return(m)
}

context("generalize horizon names")

## sample data
x <- c('A', 'AC', 'Bt1', '^AC', 'C', 'BC', 'CB')

## tests

test_that("basic pattern matching", {

  # the third pattern will steal from the second
  n <- c('A', '^AC', 'C')
  p <- c('A', '\\^A', 'C')
  res <- generalize.hz(x, new = n, pat = p, non.matching.code = 'not-used')

  # matching only text, not factor levels
  expect_equal(as.character(res), c('A', 'C', 'not-used', 'C', 'C', 'C', 'C'))

  # check levels: these should match the ording of `n` + non matching code
  expect_equal(levels(res), c('A', '^AC', 'C', 'not-used'))
  
  res2 <- generalizeHz(x, new = n, pat = p, non.matching.code = 'not-used')
  
  # matching only text, not factor levels
  expect_equal(as.character(res2), c('A', 'C', 'not-used', 'C', 'C', 'C', 'C'))
  
  # check levels: these should match the ording of `n` + non matching code
  expect_equal(levels(res2), c('A', '^AC', 'C', 'not-used'))
})


test_that("advanced pattern matching, requires perl", {

  # the third pattern may steal from the second
  n <- c('A', '^AC', 'C')
  # A -- ^A -- C without preceding A
  p <- c('A', '\\^A', '(?<!A)C')

  # the last pattern requires perl-compatible REGEX
  # error without perl=TRUE
  expect_error(suppressWarnings(generalize.hz(x, new = n, pat = p, non.matching.code = 'not-used')))

  # this should work
  res <- generalize.hz(x, new = n, pat = p, non.matching.code = 'not-used', perl = TRUE)

  # matching only text, not factor levels
  expect_equal(as.character(res), c('A', 'A', 'not-used', '^AC', 'C', 'C', 'C'))

  # check levels: these should match the ording of `n` + non matching code
  expect_equal(levels(res), c('A', '^AC', 'C', 'not-used'))
})

## TODO:
# more complex patterns: '/', lithologic discontinuities, look-ahead, anchoring, etc.

data(sp3)
depths(sp3) <- id ~ top + bottom

# crude generalized horizons
horizons(sp3) <- rbind(data.frame(hzID = sp3[,1]$hzID,     genhz = "A", stringsAsFactors = FALSE),
                       data.frame(hzID = sp3[,2:100]$hzID, genhz = "C", stringsAsFactors = FALSE))
# slightly realistic
sp3$genhz[sp3$clay > 16] <- "Bt"

test_that("guessGenHzLevels works as expected", {
  expect_error(as.numeric(guessGenHzLevels(sp3)$median.depths))
  GHL(sp3) <- "genhz"
  expect_equal(as.numeric(guessGenHzLevels(sp3)$median.depths), c(5,40,44))
})

test_that("hzTransitionProbabilities works as expected", {
  res <- hzTransitionProbabilities(sp3, "genhz")
  expect_equal(res, structure(c(0, 0, 0, 0.375, 1, 0.111111111111111,
                                0.625, 0, 0.888888888888889), .Dim = c(3L, 3L),
                              .Dimnames = list(c("A", "Bt", "C"),
                                               c("A", "Bt", "C")), ties = FALSE))
  res <- hzTransitionProbabilities(sp3, "genhz", loopTerminalStates = TRUE)
  expect_equal(res, structure(c(0, 0, 0, 0.375, 1, 0.111111111111111,
                                0.625, 0, 0.888888888888889),
                              .Dim = c(3L, 3L), .Dimnames = list(c("A", "Bt", "C"),
                                                                 c("A", "Bt", "C")), ties = FALSE))

  horizons(sp3)$genhz2 <- NA
  expect_message(hzTransitionProbabilities(sp3, "genhz2"))

  # ties in probability matrix
  dftest <- data.frame(id = c(1,1,1,2,2,2),
                       top = c(0,25,50,0,25,50),
                       bottom = c(25,50,100,25,50,100),
                       genhz = c("A","B","C","A","B","R"))
  depths(dftest) <- id ~ top + bottom
  expect_warning(hzTransitionProbabilities(dftest, "genhz"))
})

test_that("evalGenHZ works as expected", {
  res <- evalGenHZ(sp3, genhz = "genhz", vars = "clay")
  expect_equal(names(res), c("horizons","stats","dist"))
  expect_equal(as.character(res$stats$clay), c("11.26 (3.46)", "28.29 (11.49)", "8.3 (1.74)"))
})

test_that("generalize.hz works as expected", {
  res <- generalize.hz(sp3$name, new = "H", pat = ".*")
  expect_equal(levels(res), c("H","not-used"))
  expect_equal(as.character(res)[10], "H")

  res <- generalize.hz(sp3$genhz, new = c("A","Bt"), pat = c("A","[^A]"))
  expect_equal(levels(res), c("A","Bt","not-used"))
  expect_equal(as.character(res)[10], "Bt")
})

test_that("get.ml.hz works as expected", {
  
  data(sp3)
  depths(sp3) <- id ~ top + bottom
  
  # crude generalized horizons
  horizons(sp3) <- rbind(data.frame(hzID = sp3[,1]$hzID,     genhz = "A", stringsAsFactors = FALSE),
                         data.frame(hzID = sp3[,2:100]$hzID, genhz = "C", stringsAsFactors = FALSE))
  # slightly realistic
  sp3$genhz[sp3$clay > 16] <- "Bt"
  
  # make sure it runs
  res <- get.ml.hz(slab(sp3, fm = ~ genhz, cpm = 1, slab.structure = 0:max(sp3)))
  
  # expected output
  expect_equal(nrow(res), 3)
  expect_equal(res$top, c(0,10,60))
  expect_equal(res$confidence, c(50,49,67))
  
  # rough check on Brier scores / Shannon H
  expect_true(
    all(
      round(res$pseudo.brier, 2) == c(0.35, 0.48, 0.21)
      )
  )
  
  expect_true(
    all(
      round(res$mean.H, 2) == c(1.27, 1.01, 0.54)
    )
  )
  
})

test_that("generalizeHz SPC method", {
  data(sp4)
  depths(sp4) <- id ~ top + bottom
  
  # name is the regular hz designation
  hzdesgnname(sp4) <- "name"
  
  sp4 <- generalizeHz(sp4, c('A','B','C'), c('^A', '^B', '^C'), ghl = "mylabel")
  
  expect_equal(GHL(sp4), "mylabel")
})

#' Get top or bottom depths of horizons matching a regular expression pattern
#'
#' @description The \code{depthOf} family of functions calculate depth of occurrence of a horizon designation pattern, or any other value that can be coerced to character and matched with a regular expression.
#'
#' If you need all depths of occurrence for a particular pattern, \code{depthOf} is what you are looking for. \code{minDepthOf} and \code{maxDepthOf} are wrappers around \code{depthOf} that return the minimum and maximum depth. They are all set up to handle missing values and missing "contacts" with the target pattern.
#'
#' @param p a SoilProfileCollection
#' @param pattern a regular expression to match in the horizon designation column. See:\code{hzdesgn}
#' @param FUN a function that returns a single value, and takes argument `na.rm`
#' @param top should the top (TRUE) or bottom (FALSE) depth be returned for matching horizons? Default: \code{TRUE}.
#' @param hzdesgn column name containing horizon designations. Default: \code{guessHzDesgnName(p)}
#' @param no.contact.depth depth to assume that contact did not occur.
#' @param no.contact.assigned depth to assign when a contact did not occur.
#' @param simplify logical. Return single profile results as vector (default: `TRUE`) or `data.frame` (`FALSE`)
#' @param na.rm logical. Remove `NA`? (default: `TRUE`)
#' @return a numeric vector containing specified depth(s) of horizons matching a pattern. If `length(p) > 1` then a _data.frame_ containing profile ID, horizon ID, top or bottom depths, horizon designation and pattern.
#'
#' @author Andrew G. Brown
#'
#' @export depthOf
#' @export minDepthOf
#' @export maxDepthOf
#' @aliases maxDepthOf minDepthOf
#' @examples
#' # construct a fake profile
#' spc <- data.frame(id=1, taxsubgrp = "Lithic Haploxerepts",
#'                   hzname   = c("A","AB","Bw","BC","R"),
#'                   hzdept   = c(0,  20, 32, 42,  49),
#'                   hzdepb   = c(20, 32, 42, 49, 200),
#'                   clay     = c(19, 22, 22, 21,  NA),
#'                   texcl    = c("l","l","l", "l","br"),
#'                   d_value  = c(5,   5,  5,  6,  NA),
#'                   m_value  = c(2.5, 3,  3,  4,  NA),
#'                   m_chroma = c(2,   3,  4,  4,  NA))
#'
#' # promote to SoilProfileCollection
#' depths(spc) <- id ~ hzdept + hzdepb
#' hzdesgnname(spc) <- 'hzname'
#' hztexclname(spc) <- 'texcl'
#'
#' # multiple horizons contain B
#' depthOf(spc, "B")
#'
#' # deepest top depth of horizon containing B
#' maxDepthOf(spc, "B")
#'
#' # shallowest top depth
#' minDepthOf(spc, "B")
#'
#' # deepest bottom depth
#' maxDepthOf(spc, "B", top = FALSE)
#'
#' # deepest bottom depth above 35cm
#' maxDepthOf(spc, "B", top = FALSE, no.contact.depth = 35)
#'
#' # assign infinity (Inf) if B horizon does not start within 10cm
#' minDepthOf(spc, "B", no.contact.depth = 10, no.contact.assigned = Inf)
#'
depthOf <- function(p, 
                    pattern, 
                    FUN = NULL,
                    top = TRUE, 
                    hzdesgn = hzdesgnname(p, required = TRUE),
                    no.contact.depth = NULL, 
                    no.contact.assigned = NA_real_,
                    na.rm = TRUE,
                    simplify = TRUE) {
  
  if (!inherits(p, 'SoilProfileCollection'))
    stop("`p` must be a SoilProfileCollection")
  
  if (length(pattern) != 1)
    stop("`pattern` must be unit length character vector containing a regular expression")
    
  # pass through FUN argument if specified
  if (!is.null(FUN)) {
    return(.funDepthOf(
      p = p,
      pattern = pattern,
      FUN = FUN,
      top = top,
      hzdesgn = hzdesgn,
      no.contact.depth = no.contact.depth,
      no.contact.assigned = no.contact.assigned,
      na.rm = na.rm,
      simplify = simplify
    ))
  }
  
  id <- idname(p)
  hid <- hzidname(p)

  # if the user has not specified a column containing horizon designations
  if (is.null(hzdesgn) || !hzdesgn %in% horizonNames(p)) {
    stop("Horizon designation column (", hzdesgn, ") does not exist.")
  }

  # get horizons matching designation pattern
  hz.match <- horizons(p)[grepl(pattern, p[[hzdesgn]]),]
  
  subsite <- data.frame(idn = profile_id(p), stringsAsFactors = FALSE) 
  
  # get top or bottom depth, based on `top` argument
  depthcol <- horizonDepths(p)[ifelse(top, 1, 2)]
  res <- hz.match[[depthcol]]
  
  # remove results greater than the cutoff depth: `no.contact.depth`
  if (any(res > no.contact.depth)) {
    hz.match <- hz.match[-which(res > no.contact.depth),]
    res <- hz.match[[depthcol]]
  }
  
  # if no horizons match, return `no.contact.assigned`
  if (nrow(hz.match) == 0) {    
    if (length(p) == 1 && simplify) {
      return(no.contact.assigned)
    }
    
    # if no horizons in any profile match, make a conformal result
    emptyres <- data.frame(
      idn = as.character(subsite[["idn"]]),
      hidn = NA,
      depth = no.contact.assigned,
      hzname = NA
    )
    colnames(emptyres) <- c(id, hid, depthcol, hzdesgn)
    emptyres$pattern <- pattern
    return(emptyres)
  }

  # if there are non-NA results, return all of them
  if (length(res) > 0) {
    
    # backwards compatible: simplify=TRUE and SPC length is 1
    if (length(p) == 1 && simplify) {
      return(res)
    }
    
    dfres <- data.table::data.table(idn = as.character(hz.match[[id]]), 
                                    hidn = hz.match[[hid]],
                                    depth = res, 
                                    hzname = hz.match[[hzdesgn]])[subsite, on = "idn"]
    
    # filter out horizons that are > no.contact.depth, keep NA
    if (!is.null(no.contact.depth)) {
      dfres <- dfres[which(dfres$depth <= no.contact.depth | is.na(dfres$depth)),]
    }
    
    # replace NA with no.contact.assigned
    dfres$depth[is.na(dfres$depth)] <- no.contact.assigned
    
    colnames(dfres) <- c(id, hid, depthcol, hzdesgn)
    dfres$pattern <- pattern
    
    return(as.data.frame(dfres))
  }

  # otherwise:
  return(no.contact.assigned)
}

.funDepthOf <- function(p,
                        pattern,
                        FUN,
                        top = TRUE,
                        hzdesgn = hzdesgnname(p, required = TRUE),
                        no.contact.depth = NULL,
                        no.contact.assigned = NA,
                        na.rm = TRUE,
                        simplify = TRUE) {
    
  
  id <- idname(p)
  depthcol <- horizonDepths(p)[ifelse(top, 1, 2)]
  
  # depthOf returns all top or bottom depths of horizons matching `hzdesgn`
  res <- data.table::as.data.table(depthOf(p = p,
                 pattern = pattern,
                 FUN = NULL,
                 top = top,
                 hzdesgn = hzdesgn,
                 no.contact.depth = no.contact.depth,
                 no.contact.assigned = no.contact.assigned,
                 na.rm = na.rm,
                 simplify = FALSE))
  
  # otherwise, return the FUN value)) {
  idx <- numeric(0)
  naldx <- is.na(res[[depthcol]])
  
  if (all(naldx)) {
    naldx <- logical(0)
    idx <- res[, .I[1], by = id]$V1
  } else {
    # handle warnings about e.g. no non-missing arguments to FUN
    idx <- res[, .I[which(.SD[[depthcol]] %in% suppressWarnings(FUN(.SD[[depthcol]], na.rm = na.rm)))][1], by = id, .SDcols = depthcol]$V1
    idx <- idx[!is.na(idx)]
  }
  
  idx2 <- sort(unique(c(which(naldx), idx)))
  res <- res[idx2, .SD[1, ], by = id, .SDcols = c(hzidname(p), depthcol, hzdesgn, "pattern")]
  
  if (length(p) == 1 && simplify) {  
    return(res[[depthcol]])
  }
  .as.data.frame.aqp(res, aqp_df_class(p))
}
  
# maxDepthOf is a wrapper around depthOf to return a single, maximum value
#' @rdname depthOf
#' @export
maxDepthOf <- function(p,
                       pattern,
                       top = TRUE,
                       hzdesgn = hzdesgnname(p, required = TRUE),
                       no.contact.depth = NULL,
                       no.contact.assigned = NA,
                       na.rm = TRUE,
                       simplify = TRUE) {
   .funDepthOf(
      p = p,
      pattern = pattern,
      FUN = max,
      top = top,
      hzdesgn = hzdesgn,
      no.contact.depth = no.contact.depth,
      no.contact.assigned = no.contact.assigned,
      na.rm = na.rm,
      simplify = simplify
    )
    
  }

# minDepthOf is a wrapper around depthOf to return a single, minimum value\
#' @rdname depthOf
#' @export
minDepthOf <- function(p,
                       pattern,
                       top = TRUE,
                       hzdesgn = hzdesgnname(p, required = TRUE),
                       no.contact.depth = NULL,
                       no.contact.assigned = NA,
                       na.rm = TRUE,
                       simplify = TRUE) {
    .funDepthOf(
      p = p,
      pattern = pattern,
      FUN = min,
      top = top,
      hzdesgn = hzdesgn,
      no.contact.depth = no.contact.depth,
      no.contact.assigned = no.contact.assigned,
      na.rm = na.rm,
      simplify = simplify
    )
    
  }

context("soil depth estimation")

## sample data
d <-
  rbind(
  data.frame(
    id = c(1, 1, 1),
    top = c(0, 20, 35),
    bottom = c(20, 35, 110),
    name = c('A', 'Bt', 'C')
  ),
  data.frame(
    id = c(2, 2, 2),
    top = c(0, 20, 55),
    bottom = c(20, 55, 80),
    name = c('A', 'Bt', 'Cr')
  ),
  data.frame(
    id = c(3, 3, 3),
    top = c(0, 20, 48),
    bottom = c(20, 48, 130),
    name = c('A', 'Bt', 'Cd')
  ),
  data.frame(
    id = c(4, 4),
    top = c(0, 20),
    bottom = c(20, 180),
    name = c('A', 'R')
  ))


depths(d) <- id ~ top + bottom
hzdesgnname(d) <- "name"

## tests

test_that("error conditions", {

  # function will only accept a single profile
  expect_error(estimateSoilDepth(d, name='name'))

  hzdesgnname(d) <- ""
  # not specified -> error 
  expect_error(profileApply(d, estimateSoilDepth))
  
  # not in horizonNames() -> error 
  expect_error(profileApply(d, estimateSoilDepth, name='goo'))
})


test_that("basic soil depth evaluation, based on pattern matching of hz designation", {

  # setting hz desgn by argument works
  res <- profileApply(d, estimateSoilDepth, name = 'name')
  expect_equivalent(res, c(110, 55, 48, 20))

  # use hzdesgnname(x)
  hzdesgnname(d) <- "name"
  res <- estimateSoilDepth(d[1,])
  expect_equivalent(res, 110)
})


test_that("application of reasonable depth assumption of 150, given threshold of 100", {
  
  res <- profileApply(d, estimateSoilDepth, name = 'name', no.contact.depth=100, no.contact.assigned=150)

  expect_equivalent(res, c(150, 55, 48, 20))
})


test_that("depth to feature using REGEX on hzname: [Bt]", {

  # example from manual page, NA used when there is no 'Bt' found
  res <- profileApply(d, estimateSoilDepth, name = 'name', p = 'Bt', no.contact.depth=0, no.contact.assigned=NA)

  expect_equivalent(res, c(20, 20, 20, NA))
})


test_that("depthOf - simple match", {
  expect_equal(depthOf(d[1,], "Cr|R|Cd"), NA_real_)
  expect_equal(depthOf(d[2,], "Cr|R|Cd"), 55)
  expect_equal(minDepthOf(d[2,], "Cr|R|Cd"), 55)
  expect_equal(maxDepthOf(d[2,], "Cr|R|Cd"), 55)
  expect_equal(maxDepthOf(d[2,], "Cr|R|Cd", top = FALSE), 80)
})

test_that("depthOf - multiple match", {
  expect_equal(depthOf(d[1,], "A|B|C"), c(0,20,35))
  expect_equal(depthOf(d[1,], "A|B|C", top = FALSE), c(20,35,110))
  expect_equal(minDepthOf(d[1,], "A|B|C"), 0)
  expect_equal(maxDepthOf(d[1,], "A|B|C"), 35)
  expect_equal(minDepthOf(d[1,], "A|B|C", top = FALSE), 20)
  expect_equal(maxDepthOf(d[1,], "A|B|C", top = FALSE), 110)
})

test_that("depthOf - no match", {
  expect_equal(depthOf(d[1,], "X"), NA_real_)
  expect_equal(depthOf(d[2,], "Cr|R|Cd", no.contact.depth = 50), NA_real_)

  # multiple SPC return data.frame 
  expect_true(inherits(depthOf(d, "X"), 'data.frame'))
  
  # one match -- other pedons no match
  d$name[1] <- "X"
  expect_equal(maxDepthOf(d, "X")$top, c(0, NA_real_, NA_real_, NA_real_))
  
  # two matches, more than one NA depth
  d$name[2] <- "X"
  d$top[1:2] <- NA
  
  # first profile has two matches; all NA
  expect_equal(depthOf(d, "X")$top, c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_))
  
  # using functional (1 value per profile)
  expect_equal(maxDepthOf(d, "X")$top, c(NA_real_, NA_real_, NA_real_, NA_real_))
  
  d2 <- d
  d2$name <- NULL
  expect_error(depthOf(d2[1,], "A|B|C"))
})

test_that("soil depth class assignment, using USDA-NRCS class breaks", {

  res <- getSoilDepthClass(d, name = 'name')

  # result should be a data.frame with as many rows as profiles in input
  expect_true(inherits(res, 'data.frame'))
  expect_equal(nrow(res), length(d))

  # depths, should be the same as prior tests using estimateSoilDepth
  expect_equivalent(res$depth, c(110, 55, 48, 20))

  # depth classes are returned as a factor sorted from shallow -> deep
  dc <- factor(c('deep', 'mod.deep', 'shallow', 'very.shallow'), levels=c('very.shallow', 'shallow', 'mod.deep', 'deep', 'very.deep'))
  expect_equivalent(res$depth.class, dc)
})
  

test_that("data.table safety", {
  
  d <-
    rbind(
      data.frame(
        id = c(1, 1, 1),
        top = c(0, 20, 35),
        bottom = c(20, 35, 110),
        name = c('A', 'Bt', 'C')
      ),
      data.frame(
        id = c(2, 2, 2),
        top = c(0, 20, 55),
        bottom = c(20, 55, 80),
        name = c('A', 'Bt', 'Cr')
      ),
      data.frame(
        id = c(3, 3, 3),
        top = c(0, 20, 48),
        bottom = c(20, 48, 130),
        name = c('A', 'Bt', 'Cd')
      ),
      data.frame(
        id = c(4, 4),
        top = c(0, 20),
        bottom = c(20, 180),
        name = c('A', 'R')
      ))
  
  d <- data.table(d)
  depths(d) <- id ~ top + bottom
  
  res <- profileApply(d, estimateSoilDepth, name = 'name', no.contact.depth=100, no.contact.assigned=150)
  expect_equivalent(res, c(150, 55, 48, 20))
  
  sdc <- getSoilDepthClass(d, name = 'name', no.contact.depth=100, no.contact.assigned=150)
  expect_equivalent(sdc$depth, c(150, 55, 48, 20))
})



test_that("really deep", {
  
  d <-
    rbind(
      data.frame(
        id = c(1, 1, 1),
        top = c(0, 20, 35),
        bottom = c(20, 35, 2000),
        name = c('A', 'Bt', 'C')
      ))
  
  depths(d) <- id ~ top + bottom
  
  res <- profileApply(d, estimateSoilDepth, name = 'name')
  expect_equivalent(res, 2000)
  
  sdc <- getSoilDepthClass(d, name = 'name')
  expect_equivalent(sdc$depth, 2000)
  expect_equivalent(as.character(sdc$depth.class), 'very.deep')
})

# simple wrapper around col2Munsell(..., space = 'CIELAB')
# closest Munsell chip to LAB coordinates and error
# sigma is dE00 
.formatClosestMunsell <- function(lab) {
  res <- col2Munsell(lab, space = 'CIELAB', nClosest = 1)
  res.txt <- sprintf("%s %s/%s\n(%.3f)", res$hue, res$value, res$chroma, res$sigma)
  return(res.txt)
}


## TODO: 
##       * allow Munsell, CIELAB, sRGB, etc.
##       * quantiles should be weighted by thickness, can we do this via dice()?
##       * finish documentation + links
##       * consider a third output: chip with pair-wise min(dE00) within distance matrix
##       * unify with colorVariation()


#' @title Soil Color Range via Quantiles
#'
#' @param soilColors vector of R colors (sRGB colorspace)
#' @param p marginal quantiles of interest
#'
#' @description Estimate central tendency and spread of soil color using marginal quantiles and L1 median of CIELAB coordinates.
#' 
#' @details Colors are converted from sRGB to CIELAB (D65 illuminant), marginal quantiles of (L,A,B) coordinates are estimated, and L1 median (L,A,B) is estimates. The closest Munsell chips (via Munsell/CIELAB lookup table provided by `munsell`) and R colors are determined by locating chips closest to the marginal quantiles and L1 median.
#' 
#' The results can be conveniently inspected using [plotColorQuantiles()].
#' 
#' @author D.E. Beaudette
#' 
#' @return A List containing the following elements:
#' 
#'   * `marginal`: `data.frame` containing marginal quantiles in CIELAB (D65), closest Munsell chips, and dE00
#'   * `L1`: L1 median CIELAB (D65) values, closest Munsell chip, and dE00
#' 
#' @export
#'
#' @examples
#' 
#'\dontrun{
#' # example data, see manual page for details
#' data(sp5)
#' 
#' # slice top 25 cm
#' # 24-25cm is the last slice
#' s <- dice(sp5, 0:24 ~ .)
#' 
#' # check some of the data
#' par(mar=c(0,0,0,0))
#' plotSPC(sample(s, 25), divide.hz = FALSE, name = '', print.id = FALSE, width = 0.5)
#' 
#' # colors
#' previewColors(unique(s$soil_color))
#' 
#' # compute marginal quantiles and L1 median
#' cq <- colorQuantiles(s$soil_color)
#' 
#' # simple graphical display of results
#' plotColorQuantiles(cq)
#' }
#' 
colorQuantiles <- function(soilColors, p = c(0.05, 0.5, 0.95)) {
  
  # sanity check, need this for L1 median
  if(!requireNamespace('Gmedian')) {
    stop('package `Gmedian` is required', call. = FALSE)
  }
  
  
  # hex representation -> sRGB
  soilColors.srgb <- t(col2rgb(soilColors)) / 255
  
  # sRGB -> CIE LAB
  soilColors.lab <- convertColor(soilColors.srgb, from = 'sRGB', to = 'Lab', from.ref.white = 'D65', to.ref.white = 'D65', clip = FALSE)
  
  # convert to DF for use in diana
  soilColors.lab <- as.data.frame(soilColors.lab)
  names(soilColors.lab) <- c('L', 'A', 'B')
  
  # marginal quantiles
  q.L <- quantile(soilColors.lab$L, probs = p, na.rm = TRUE)
  q.A <- quantile(soilColors.lab$A, probs = p, na.rm = TRUE)
  q.B <- quantile(soilColors.lab$B, probs = p, na.rm = TRUE)
  
  # find observed color with the closest L-coordinate, linear distances
  L.q.idx <- c(
    which.min(sqrt((q.L[1] - soilColors.lab$L) ^ 2)),
    which.min(sqrt((q.L[2] - soilColors.lab$L) ^ 2)),
    which.min(sqrt((q.L[3] - soilColors.lab$L) ^ 2))
  )
  
  # find observed color with the closest A-coordinate, linear distances
  A.q.idx <- c(
    which.min(sqrt((q.A[1] - soilColors.lab$A) ^ 2)),
    which.min(sqrt((q.A[2] - soilColors.lab$A) ^ 2)),
    which.min(sqrt((q.A[3] - soilColors.lab$A) ^ 2))
  )
  
  # find observed color with the closest B-coordinate, linear distances
  B.q.idx <- c(
    which.min(sqrt((q.B[1] - soilColors.lab$B) ^ 2)),
    which.min(sqrt((q.B[2] - soilColors.lab$B) ^ 2)),
    which.min(sqrt((q.B[3] - soilColors.lab$B) ^ 2))
  )
  
  # L1 median
  # close to but not actually an obs within original data... why?
  # this is sometimes quite distant from marginal medians... why?
  L1 <- Gmedian::Gmedian(soilColors.lab, nstart = 10)
  
  ## find closest Munsell chips via CIE LAB coordinates
  # this is the closest Munsell chip to the L1 median color
  L1.closest <- .formatClosestMunsell(L1)
  
  # closest munsell chip to marginal L,A,B quantiles
  L.closest <- .formatClosestMunsell(soilColors.lab[L.q.idx, ])
  A.closest <- .formatClosestMunsell(soilColors.lab[A.q.idx, ])
  B.closest <- .formatClosestMunsell(soilColors.lab[B.q.idx, ])
  
  
  ## find closest observed color to L1 median via CIE2000 distance metric
  d <- farver::compare_colour(
    from = L1, 
    to = soilColors.lab, 
    from_space = 'lab', 
    method = 'cie2000'
  )
  
  # assign L1 color
  L1.color <- soilColors[which.min(d)]
  
  # closest observed colors to marginal quantiles
  q.L.colors <- soilColors[L.q.idx]
  q.A.colors <- soilColors[A.q.idx]
  q.B.colors <- soilColors[B.q.idx]
  
  # marginal quantiles
  q.L.values <- soilColors.lab$L[L.q.idx]
  q.A.values <- soilColors.lab$A[A.q.idx]
  q.B.values <- soilColors.lab$B[B.q.idx]
  
  
  # combine into single DF for plotting
  res <- list(
    marginal = data.frame(
      p = p,
      L = q.L.values,
      A = q.A.values,
      B = q.B.values,
      L_colors = q.L.colors,
      A_colors = q.A.colors,
      B_colors = q.B.colors,
      L_chip = L.closest,
      A_chip = A.closest,
      B_chip = B.closest,
      stringsAsFactors = FALSE
    ),
    L1 = data.frame(
      p = 0.5,
      L = L1[, 1],
      A = L1[, 2],
      B = L1[, 3],
      L1_color = L1.color,
      L1_chip = L1.closest,
      stringsAsFactors = FALSE
    )
  )
  
  return(res)
  
}


#' @title Visualize Color Quantiles
#' 
#' @description This function creates a visualization of the output from \code{colorQuantiles} using lattice graphics.
#' 
#' @param res list returned by \code{colorQuantiles}
#' @param pt.cex scaling factor for color chips
#' @param lab.cex chip label scaling factor
#'
#' @return a `lattice` graphics object
#' 
#' @details Marginal percentiles and L1 median CIELAB values from `colorQuantiles()` are combined into a single plot, arranged in panels according to L, A, and B coordinates. Munsell "chips" (colors and labels) are based on the closest Munsell color found via [col2Munsell()]. 
#' 
#' @author D.E. Beaudette
#'
#' @export
#' 
#' 
plotColorQuantiles <- function(res, pt.cex = 7, lab.cex = 0.66) {
  
  # convert wide -> long format for plotting in panels
  # using data.table::melt()
  m.long <- melt(
    data.table::as.data.table(res$marginal), 
    id.var = c('p', 'L_colors', 'A_colors', 'B_colors', 'L_chip', 'A_chip', 'B_chip')
  )
  
  # convert back to data.frame
  m.long <- as.data.frame(m.long)
  
  # fake y-variable for plotting marginal vs. L1
  m.long$y <- 1
  
  # local copy and fake y-variable
  L1 <- res$L1
  L1$y <- 2
  
  # panel labels
  m.long$variable <- factor(m.long$variable, levels = c('L', 'A', 'B'), labels = c('CIELAB L-Coordinate', 'CIELAB A-Coordinate', 'CIELAB B-Coordinate'))
  
  # sub text for marginal qualiles
  p.percentiles <- round(unique(res$marginal$p) * 100)
  p.text <- sprintf("%sth", p.percentiles)
  
  p.sub <- sprintf(
    "%s Percentiles",
    paste(p.text, collapse = ", ")
  )
  
  
  # compose figure
  pp <- lattice::xyplot(
    y ~ value | variable, 
    data = m.long, 
    ylim = c(0.5, 2.5),
    xlab = '',
    ylab = '',
    sub = p.sub,
    layout = c(1, 3),
    scales = list(
      y = list(at = c(1, 2), labels = c('Marginal P', 'L1 Median'), alternating = 1), 
      x = list(relation = 'free', tick.number = 10, alternating = 1)
    ),
    as.table = TRUE,
    subscripts = TRUE,
    strip = lattice::strip.custom(bg = grey(0.85)),
    panel = function(x, y, subscripts = subscripts, ...) {
      # local copy
      d <- m.long[subscripts, ]
      
      # panel number tracks levels of L, A, B
      pn <- lattice::panel.number()
      
      # index current colors
      col.idx <- c('L_colors', 'A_colors', 'B_colors')[pn]
      cols <- d[[col.idx]]
      
      # index current chip labels
      chip.idx <- c('L_chip', 'A_chip', 'B_chip')[pn]
      chip <- d[[chip.idx]]
      
      # horizontal grid lines
      panel.abline(h = c(1, 2), col = grey(0.65))
      
      # marginal data
      lattice::panel.points(x = d$value, y = d$y, pch = 15, cex = pt.cex, col = cols)
      lattice::panel.text(x = d$value, y = d$y, label = chip, col = invertLabelColor(cols), cex = lab.cex, font = 2)
      
      # L1 data: these are constant across panels
      # index specific coordinate
      LAB.idx <- which(names(L1) %in% c('L', 'A', 'B'))
      coord.idx <- LAB.idx[pn]
      
      lattice::panel.points(x = L1[[coord.idx]], y = L1$y, pch = 15, cex = pt.cex, col = L1$L1_color)
      lattice::panel.text(x = L1[[coord.idx]], y = L1$y, label = L1$L1_chip, col = invertLabelColor(L1$L1_color), cex = lab.cex, font = 2)
    }
  )
  
  
  return(pp)
}



## old version, base graphics and kind of a hack
# plotColorQuantiles <- function(res, pt.cex=7, title='', mar=c(2,1.5,1,1)) {
#   par(mar=mar, mfrow=c(3,1))
#   
#   # vertical spacing
#   m.y <- 2
#   L1.y <- 1
#   y.lim <- c(0.5, 2.5)
#   
#   # pre-make axis
#   L.axis <- pretty(zapsmall(res$marginal$L), n = 10)
#   A.axis <- pretty(zapsmall(res$marginal$A), n = 10)
#   B.axis <- pretty(zapsmall(res$marginal$B), n = 10)
#   
#   ## L coordinates
#   plot(res$marginal$L, rep(m.y, times=3), pch=22, bg=res$marginal$L_colors, cex=pt.cex, xlim=range(L.axis), ylim=y.lim, axes=FALSE, xlab='', ylab='', main=title, col.main=par('fg'))
#   points(res$L1$L, L1.y, pch=22, bg=res$L1$L1_color, cex=pt.cex)
#   text(res$marginal$L, rep(m.y, times=3), labels = res$marginal$p, pos = 3, offset = 1.5)
#   text(res$L1$L, L1.y, labels = 'L1', pos = 3, offset = 1.5)
#   text(res$marginal$L, rep(m.y, times=3), res$marginal$L_chip, pos=1, offset = 1.5)
#   text(res$L1$L, L1.y, labels = res$L1$L1_chip, pos = 2, offset = 1.5)
#   axis(1, line = -2, at = L.axis, col.axis=par('fg'))
#   mtext('L', side = 2, line=-0.125, font=2, las=1)
#   # title('CIELAB Color Space', cex=1.5, line=-1)
#   
#   ## A coordinates
#   plot(res$marginal$A, rep(m.y, times=3), pch=22, bg=res$marginal$A_colors, cex=pt.cex, xlim=range(A.axis), ylim=y.lim, axes=FALSE, xlab='', ylab='')
#   points(res$L1$A, L1.y, pch=22, bg=res$L1$L1_color, cex=pt.cex)
#   text(res$marginal$A, rep(m.y, times=3), labels = res$marginal$p, pos = 3, offset = 1.5)
#   text(res$L1$A, L1.y, labels = 'L1', pos = 3, offset = 1.5)
#   text(res$marginal$A, rep(m.y, times=3), res$marginal$A_chip, pos=1, offset = 1.5)
#   text(res$L1$A, L1.y, labels = res$L1$L1_chip, pos = 2, offset = 1.5)
#   axis(1, line = -2, at=A.axis, col.axis=par('fg'))
#   mtext('A', side = 2, line=-0.125, font=2, las=1)
#   
#   plot(res$marginal$B, rep(m.y, times=3), pch=22, bg=res$marginal$B_colors, cex=pt.cex, xlim=range(B.axis), ylim=y.lim, axes=FALSE, xlab='', ylab='')
#   points(res$L1$B, L1.y, pch=22, bg=res$L1$L1_color, cex=pt.cex)
#   text(res$marginal$B, rep(m.y, times=3), labels = res$marginal$p, pos = 3, offset = 1.5)
#   text(res$L1$B, L1.y, labels = 'L1', pos = 3, offset = 1.5)
#   text(res$marginal$B, rep(m.y, times=3), res$marginal$B_chip, pos=1, offset = 1.5)
#   text(res$L1$B, L1.y, labels = res$L1$L1_chip, pos = 2, offset = 1.5)
#   axis(1, line = -2, at=B.axis, col.axis=par('fg'))
#   mtext('B', side = 2, line=-0.125, font=2, las=1)
#   
#   
# }

context("colorQuantiles")

data(sp5)


test_that("colorQuantiles works as expected", {
  
  skip_if_not_installed('Gmedian')
  
  # compute marginal quantiles and L1 median
  cq <- colorQuantiles(sp5$soil_color, p = c(0.05, 0.5, 0.95))
  
  # expected output
  expect_true(inherits(cq, 'list'))
  expect_true(all(names(cq) == c('marginal', 'L1')))
  expect_true(nrow(cq$marginal) == 3)
  
})


test_that("plotColorQuantiles works as expected", {
  
  skip_if_not_installed('Gmedian')
  
  # compute marginal quantiles and L1 median
  cq <- colorQuantiles(sp5$soil_color, p = c(0.05, 0.5, 0.95))
  
  p <- plotColorQuantiles(cq)
  
  # expected output
  expect_true(inherits(p, 'trellis'))
  
})

test_that("output is correct", {
  
  skip_if_not_installed('Gmedian')
  
  # compute marginal quantiles and L1 median
  cq <- colorQuantiles(sp5$soil_color, p = c(0.05, 0.5, 0.95))
  
  # L1 CIELAB coordinates
  expect_equal(cq$L1$p, 0.5, tolerance = 0.001)
  expect_equal(cq$L1$L, 32.96, tolerance = 0.01)
  expect_equal(cq$L1$A, 6.38, tolerance = 0.01)
  expect_equal(cq$L1$B, 9.75, tolerance = 0.01)
  
  # L1 Munsell color
  expect_match(cq$L1$L1_chip, '5YR 3/2', fixed = TRUE)
  
})

#' Perturb soil horizon depths using boundary distinctness
#'  
#' @param p A SoilProfileCollection
#' @param n Number of new profiles to generate (default: `100`) per profile in `p`
#' @param id a vector of profile IDs with length equal to (\code{n}). Overrides use of \code{seq_len(n)} as default profile ID values.
#' @param boundary.attr Horizon variance attribute containing numeric "standard deviations" reflecting boundary transition distinctness
#' @param thickness.attr Horizon variance attribute containing numeric "standard deviations" reflecting horizon thickness  
#' @param max.depth Depth below which horizon depths are not perturbed (default: `NULL`)
#' @param min.thickness Minimum thickness of permuted horizons (default: `1`)
#' @param new.idname New column name to contain unique profile ID (default: `pID`)
#' 
#' @description "Perturbs" the **boundary between horizons** or the **thickness of horizons** using a standard deviation specified as a horizon-level attribute. This is selected using either `boundary.attr` or `thickness.attr` to specify the column name.
#' 
#' The boundary standard deviation corresponds roughly to the concept of "horizon boundary distinctness." In contrast, the _horizon thickness_ standard deviation corresponds roughly to the "variation in horizon thickness" so it may be determined from several similar profiles that have a particular layer "in common." 
#' 
#' @details
#' 
#' Imagine a Normal curve with mean centered on the vertical (depth axis) at a representative value (RV) horizon bottom depth or thickness. By the Empirical Rule for Normal distribution, two "standard deviations" above or below that "central" mean value represent 95% of the "typical volume" of that horizon or boundary.
#' 
#' `perturb()` can leverage semi-quantitative (ordered factor) levels of boundary distinctness/topography for the upper and lower boundary of individual horizons. A handy function for this is [hzDistinctnessCodeToOffset()]. The `boundary.attr` is arguably easier to parameterize from a single profile description or "Form 232" where _horizon boundary distinctness_ classes (based on vertical distance of transition) are conventionally recorded for each layer.
#' 
#' Alternately, `perturb()` can be parameterized using standard deviation in thickness of layers derived from a group. Say, the variance parameters are defined from a set of pedons correlated to a particular series or component, and the template "seed" profile is, for example, the Official Series Description or the Representative Component Pedon.
#'
#' @return a SoilProfileCollection with `n` realizations of each profile in `p`
#' 
#' @seealso [random_profile()] [hzDistinctnessCodeToOffset()]
#' 
#' @export
#' @author D.E. Beaudette, A.G. Brown
#' @aliases permute_profile
#' @examples
#' 
#' ### THICKNESS
#' 
#' # load sample data and convert into SoilProfileCollection
#' data(sp3)
#' depths(sp3) <- id ~ top + bottom
#' 
#' # select a profile to use as the basis for simulation
#' s <- sp3[3,]
#' 
#' # reset horizon names
#' s$name <- paste('H', seq_along(s$name), sep = '')
#' 
#' # simulate 25 new profiles
#' horizons(s)$hz.sd <- 2 # constant standard deviation
#' sim.1 <- perturb(s, n = 25, thickness.attr = "hz.sd")
#' 
#' # simulate 25 new profiles using different SD for each horizon
#' horizons(s)$hz.sd <- c(1, 2, 5, 5, 5, 10, 3)
#' sim.2 <- perturb(s, n = 25, thickness.attr = "hz.sd")
#' 
#' # plot
#' par(mfrow = c(2, 1), mar = c(0, 0, 0, 0))
#' plot(sim.1)
#' mtext(
#'   'SD = 2',
#'   side = 2,
#'   line = -1.5,
#'   font = 2,
#'   cex = 0.75
#' )
#' plot(sim.2)
#' mtext(
#'   'SD = c(1, 2, 5, 5, 5, 10, 3)',
#'   side = 2,
#'   line = -1.5,
#'   font = 2,
#'   cex = 0.75
#' )
#' 
#' # aggregate horizonation of simulated data
#' # note: set class_prob_mode=2 as profiles were not defined to a constant depth
#' sim.2$name <- factor(sim.2$name)
#' a <- slab(sim.2, ~ name, cpm=2)
#' 
#' # convert to long format for plotting simplicity
#' library(data.table)
#' a.long <- data.table::melt(data.table::as.data.table(a),
#'                id.vars = c('top', 'bottom'),
#'                 measure.vars = levels(sim.2$name))
#' 
#' # plot horizon probabilities derived from simulated data
#' # dashed lines are the original horizon boundaries
#' library(lattice)
#' 
#' xyplot(
#'   top ~ value,
#'   groups = variable,
#'   data = a.long,
#'   subset = value > 0,
#'   ylim = c(100,-5),
#'   type = c('l', 'g'),
#'   asp = 1.5,
#'   ylab = 'Depth (cm)',
#'   xlab = 'Probability',
#'   auto.key = list(
#'     columns = 4,
#'     lines = TRUE,
#'     points = FALSE
#'   ),
#'   panel = function(...) {
#'     panel.xyplot(...)
#'     panel.abline(h = s$top, lty = 2, lwd = 2)
#'   }
#' )
#' 
#' ### BOUNDARIES
#' 
#' # example with sp1 (using boundary distinctness)
#' data("sp1")
#' depths(sp1) <- id ~ top + bottom
#'
#' # specify "standard deviation" for boundary thickness
#' #   consider a normal curve centered at boundary RV depth
#' # lookup table: ~maximum thickness of boundary distinctness classes, divided by 3
#' bound.lut <- c('V'=0.5,'A'=2,'C'=5,'G'=15,'D'=45) / 3
#'
#' ## V          A          C          G          D
#' ## 0.1666667  0.6666667  1.6666667  5.0000000 15.0000000
#'
#' sp1$bound_sd <- bound.lut[sp1$bound_distinct]
#'
#' # hold any NA boundary distinctness constant
#' sp1$bound_sd[is.na(sp1$bound_sd)] <- 0
#'
#' quantile(sp1$bound_sd, na.rm = TRUE)
#' p <- sp1[3]
#' 
#' # assume boundary sd is 1/12 midpoint of horizon depth
#' # (i.e. general relationship: SD increases (less well known) with depth)
#' sp1 <- transform(sp1, midpt = (bottom - top) / 2 + top, bound_sd = midpt / 12)
#' quantile(sp1$bound_sd)
#' 
#' perturb(p, boundary.attr = "bound_sd", n = 10)
#' 
#' 
#' ### Custom IDs
#'
#' ids <- sprintf("%s-%03d", profile_id(p), 1:10) 
#' perturb(p, boundary.attr = "bound_sd", id = ids)
#' 
#' 
perturb <- function(p,
                    n = 100,
                    id = NULL,
                    thickness.attr = NULL,
                    boundary.attr = NULL,
                    min.thickness = 1,
                    max.depth = NULL,
                    new.idname = 'pID') {
  
  # aqp and data.table global definitions for keywords
  .FIRST <- NULL; value <- NULL; md <- NULL; .N <- NULL; V1 <- NULL; .GRP <- NULL; gidx <- NULL; pidx <- NULL
  
  custom.ids <- FALSE
  by_thickness <- FALSE
  
  if ((missing(boundary.attr) && missing(thickness.attr)) ||
      !is.null(thickness.attr) && !is.null(boundary.attr) ||
      is.null(thickness.attr) && is.null(boundary.attr)) {
    stop("must provide one column name: thickness `thickness.attr` OR boundary `boundary.attr` containing horizon-level standard deviations", call. = FALSE)
  }
  
  if (!is.null(thickness.attr)) {
    by_thickness <- TRUE
  }
  
  if (!missing(id) && !is.null(id)) {
    custom.ids <- TRUE
    
    # keep track of missing `n` argument before it is set
    missing.n <- missing(n)
    
    n <- length(unique(id))

    if (n != length(id))
      stop("custom profile ID vector `id` contains non-unique values", call. = FALSE)

    if (!missing.n)
      message("if profile ID vector `id` is specified, `n` argument is ignored")
  }

  # calculate some SPC metadata
  hz <- data.table::data.table(horizons(p))
  idn <- idname(p)
  depthz <- horizonDepths(p)
  
  # calculate minimum depth of each profile in p
  mindepth <- p[, , .FIRST][[depthz[1]]]

  # setup for variable being perturbed
  if (by_thickness) {
    varattr <- hz[[thickness.attr]]
  } else {
    varattr <- hz[[boundary.attr]]
  }
  
  # logic checks
  if (!is.numeric(varattr) | !length(varattr) == nrow(p)) {
    stop("variance attribute must be a numeric column  name in `p` containing standard deviations of horizon or boundary thickness")
  }

  if (!all(checkHzDepthLogic(p)$valid)) {
    stop("one or more horizon depth logic tests failed for object `p`")
  }

  if (by_thickness) {
    # aqp::sim() traditionally perturbs horizon thickness
    perturb_var <- hz[[depthz[2]]] - hz[[depthz[1]]]
  } else {
    # perturb bottom depths instead of thickness for boundaries
    perturb_var <- hz[[depthz[2]]]
  }
  
  # do not vary layers below `max.depth` (if not NULL) can be arbitrary depth
  if (!is.null(max.depth) && !is.na(max.depth)) {
    idx <- which(max.depth >= mindepth)
    if (length(idx) > 0) {
      if (by_thickness) {
        d1 <- data.table::data.table(id = hz[[idname(p)]], perturb_var)
        d2 <- data.table::data.table(id = profile_id(p), mindepth)
        ldx <- (d1[d2, on = "id"][, cumsum(perturb_var) + mindepth, by = "id"]$V1) > max.depth
      } else {
        ldx <- perturb_var > max.depth
      }
    }
    varattr[ldx] <- 0
  }

  # in practice, qc warnings for improbably large SD would be useful
  # could be data entry error or improbable class assignment
  #
  # with irregular bounds, high SD could be intentional
  #  - additional random processes: waves for wavy/irregular
  #  - allowing irregular boundaries to eclipse/omit thin layers?
  #  - presence/absence of broken horizons; could this be a separate random process?

  # calculate perturbed variable vector
  # for each horizon bottom depth (boundary) calculate the gaussian offset
  res <- hz[, list(round(rnorm(n, perturb_var[.GRP], varattr[.GRP])), 
                   gidx = seq_len(n), # number of replicates
                   pidx = .SD[[idn]]), # profile ID
            by = list(hidx = seq_len(nrow(hz)))]
  
  # order result by profile*rep
  res <- res[order(pidx, gidx),]
 
  # regardless of simulation method new profiles  are reconstructed from layer thicknesses
  # (allows for handling of min.thickness, minimum depth, etc. consistently)
  if (by_thickness) {
    res <- res$V1
  } else {
    res <- res[, list(V1 = diff(c(0, V1))), by = c("pidx", "gidx")]$V1
  }
  
  # create template SPC/horizon data to insert perturb()-ed depths
  p.sub <- duplicate(p, n)
  h.sub <- horizons(p.sub)
  
  # relate .oldID to idname(p) in horizon data.frame
  idlut <- p.sub[[".oldID"]]
  names(idlut) <- p.sub[[idn]]
  nd <- data.table::data.table(id = h.sub[[idn]],
                               .oldID = idlut[h.sub[[idn]]])
  
  # join in mindepth by .oldID
  nd <- nd[data.frame(.oldID = profile_id(p), 
                      md = mindepth), 
           on = ".oldID"]
  
  # insert values
  nd$V1 <- res
  
  # calculate new top and bottom depths
  nd$.newtop <- nd[, cumsum(c(md[1], pmax(min.thickness, V1)))[1:.N], by = c("id")]$V1
  nd$.newbot <- nd[, md[1] + cumsum(pmax(min.thickness, V1)), by = c("id")]$V1
  
  # replace in template SPC
  p.sub[[depthz[1]]] <- nd$.newtop
  p.sub[[depthz[2]]] <- nd$.newbot
    
  # replace ID values if needed
  if (custom.ids & length(unique(id)) == length(p.sub)) {
    profile_id(p.sub) <- id
  }
  
  # replace ID name if needed
  if (!missing(new.idname)) {
    # these are derived from a valid ID in p.sub
    p.sub[[new.idname]] <- p.sub[[idn]]
    p.sub@horizons[[new.idname]] <- horizons(p.sub)[[idn]]
    
    # TODO: safe idname replacement method
    p.sub@idcol <- new.idname
    p.sub@horizons[[idn]] <- NULL
  }
  
  # transfers @metadata slot p->p.sub
  .transfer.metadata.aqp(p, p.sub)
}

# permute_profile <- function(p,
#                             n = 100,
#                             id = NULL,
#                             boundary.attr = NULL,
#                             min.thickness = 1,
#                             soildepth = NULL,
#                             new.idname = 'pID') {
#   .Deprecated("perturb")
#   
#   perturb(p, n = n, id = id, 
#           boundary.attr = boundary.attr, 
#           min.thickness = min.thickness, 
#           max.depth = soildepth, 
#           new.idname = new.idname)
# }

## compare permute_profile and sim, using same estimated SD

# calculate permuations of bottom depths using boundary deviations
#system.time(res <- perturb(p, n=1000, boundary_attr = "bound_sd",
#                                    min.thickness = 1))
#
# # calculate permuations of horizon thickness using horizon thickness deviation
# system.time(res2 <- perturb(p, n=1000, hz.sd = p$bound_sd, min.thick = 1))
#
# # superficially similar output
# plot(res[1:10,])
# plot(res2[1:10,])
#
# # compare slab'd result
# s.res <- slab(res, ~ prop, slab.structure = 1)
# s.res2 <- slab(res2, ~ prop, slab.structure = 1)

# inspect differences visually -- bigger differences in thicker horizons
#  applying same SD to horizon thickness as bottom depth results in more variation between realizations

# .:. thick horizons can still have abrupt boundaries, e.g. you might have a clear or gradual boundary at upper part, and abrupt at contact -- these should be handled seperately. a single SD would be useful for simulating aggregate data. the gaussian assumption is not as bad when applied to a single boundary. presumably other functions could be used to show skewed distributions.
# plot(x=s.res$p.q50, y=s.res$top, ylim=c(100,0), type="l")
# lines(x=s.res$p.q5, y=s.res$top, ylim=c(100,0), col="blue", lty=2)
# lines(x=s.res$p.q95, y=s.res$top, ylim=c(100,0), col="blue", lty=2)
#
# lines(x=s.res2$p.q50, y=s.res2$top, ylim=c(100,0), type="l", lwd=2)
# lines(x=s.res2$p.q5, y=s.res2$top, ylim=c(100,0), col="green", lty=2)
# lines(x=s.res2$p.q95, y=s.res2$top, ylim=c(100,0), col="green", lty=2)
#
# hzdesgnname(res) <- "name"
# hztexclname(res) <- "texture"
# res$clay <- res$prop
# mtr <- profileApply(res[1:1000], mollic.thickness.requirement)
# plot(hist(mtr, breaks = max(3, max(mtr) - min(mtr))))

context("SoilProfileCollection combination methods")

## make sample data
data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

sp1$x <- seq(-119, -120, length.out = length(sp1))
sp1$y <- seq(38, 39, length.out = length(sp1))

initSpatial(sp1) <- ~ x + y
prj(sp1) <- 'OGC:CRS84'

test_that("basic combination tests", {

  # test data
  x <- sp1
  y <- sp1

  # alter horizon and site data in copy
  y$random <- runif(length(y))
  y$chroma <- NULL

  # add diagnostic hz
  diagnostic_hz(y) <- data.frame(id = 'P001', type = 'pizza')

  # this should not work, IDs aren't unqiue
  expect_error(combine(list(x, y)))

  # fix IDs manually
  profile_id(y) <- sprintf("%s-copy", profile_id(y))

  # this should work
  z <- combine(list(x, y))
  
  # as should this
  z2 <- combine(x, y)

  expect_true(inherits(z, 'SoilProfileCollection'))
  expect_true(inherits(z2, 'SoilProfileCollection'))
  expect_equal(length(z), length(x) + length(y))
  expect_equal(length(z2), length(x) + length(y))

  # full site/hz names
  expect_equal(siteNames(z), unique(c(siteNames(x), siteNames(y))))
  expect_equal(horizonNames(z), unique(c(horizonNames(x), horizonNames(y))))

  # diagnostic features
  expect_equal(diagnostic_hz(z)[[idname(z)]], 'P001-copy')

})

test_that("non-conformal combination tests", {

  # random data
  ids <- sprintf("%02d", 1:5)
  x <- do.call('rbind', lapply(ids, random_profile, n=c(6, 7, 8), n_prop=1, method='LPP',
                   lpp.a=5, lpp.b=15, lpp.d=5, lpp.e=5, lpp.u=25))

  depths(x) <- id ~ top + bottom

  # more random data
  ids <- sprintf("%s", letters[1:5])
  y <- do.call('rbind', lapply(ids, random_profile, n=c(6, 7, 8), n_prop=4, method='LPP',
                   lpp.a=5, lpp.b=15, lpp.d=5, lpp.e=5, lpp.u=25))

  # alter ID, top, bottom column names
  y$pID <- y$id
  y$hztop <- y$top
  y$hzbot <- y$bottom

  y$id <- NULL
  y$top <- NULL
  y$bottom <- NULL

  depths(y) <- pID ~ hztop + hzbot

  # very different data
  data('jacobs2000', package = 'aqp')

  # alter depth units
  depth_units(y) <- 'in'

  # should throw an error
  expect_error(combine(x, y), "inconsistent depth units")

  # reset depth units
  depth_units(y) <- 'cm'

  # attempt combine
  z <- combine(list(x, y, jacobs2000))

  # there should be a total of 17 profiles in the result
  expect_equal(sum(sapply(list(x, y, jacobs2000), length)), 17)
  expect_equal(length(z), sum(sapply(list(x, y, jacobs2000), length)))

  # see https://github.com/ncss-tech/aqp/issues/163
  data(sp4)
  depths(sp4) <- id ~ top + bottom

  # profile to general realizations of
  p.idx <- 1

  # spike profile
  spike.idx <- 6

  horizons(sp4)$bdy <- 4
  p <- perturb(sp4[p.idx, ], n = 10, 
               boundary.attr = 'bdy', min.thickness = 2)

  site(p)$id <- NULL

  # these calls should produce same order result
  #  .:. combine uses depths<- internally
  z.1 <- combine(list(sp4[c(p.idx, spike.idx), ], p))
  z.2 <- combine(list(p, sp4[c(p.idx, spike.idx), ]))

  expect_true(spc_in_sync(z.1)$valid)
  expect_true(spc_in_sync(z.2)$valid)
  expect_true(all(profile_id(z.1) == profile_id(z.2)))
})


test_that("combine with non-conformal spatial data", {

  # test data
  x <- sp1
  y <- sp1
  z <- sp1

  # alter CRS, this generates an sp warning
  # 2020-07-12: now caught with all other rgdal 1.5-8+ warnings in proj4string,SoilProfileCollection-method
  prj(y) <- 'EPSG:26910' # NAD83 UTM Zone 10

  # this should not work, IDs aren't unique
  expect_error(expect_message(combine(list(x, y)),
                              "inconsistent CRS, dropping spatial metadata"),
               "non-unique profile IDs detected")

  # make IDs unique
  profile_id(y) <- sprintf("%s-copy", profile_id(y))
  profile_id(z) <- sprintf("%s-copy-copy", profile_id(z))

  res <- combine(list(x, y, z))
  
  # remove CRS (avoids "inconsistent CRS, dropping spatial metadata")
  # equivalent ways to "unset" CRS: "", NA, NULL
  prj(x) <- ''
  prj(y) <- NULL
  
  expect_message(res <- combine(list(x, y, z)), "inconsistent CRS, dropping spatial metadata")
  expect_true(inherits(res, 'SoilProfileCollection'))

  # coordinates are preserved, but projection is not
  expect_null(metadata(res)$projection)
  expect_null(metadata(res)$coordinates)
  
  ## TODO: different coordinate names
})


test_that("filtering NULL/NA elements", {

  # test data
  x <- sp1
  y <- sp1
  profile_id(y) <- sprintf("%s-copy", profile_id(y))

  # add NULLs
  s <- list(NULL, x, y, NULL)

  # this should work
  res <- combine(s)
  expect_true(inherits(res, 'SoilProfileCollection'))


  # add NULLs, different arrangement
  s <- list(x, NULL)

  # should work
  res <- combine(s)
  expect_true(inherits(res, 'SoilProfileCollection'))


  # add NAs
  s <- list(NA, x, y, NA)

  # this should work
  res <- combine(s)
  expect_true(inherits(res, 'SoilProfileCollection'))

  # this should NOT work
  expect_error(combine(s, na.rm = FALSE))

  # all NA ---> result is NULL
  s <- list(NA, NA, NA)
  expect_null(combine(s))

  # all NULL ---> result is NULL
  s <- list(NULL, NULL, NULL)
  expect_null(combine(s))
})

context("soil profile simulation")

## sample data
data(sp3)
depths(sp3) <- id ~ top + bottom

# select a profile to use as the basis for simulation
s <- sp3[3, ]

# reset horizon names
s$name <- paste('H', seq_along(s$name), sep = '')
test_that("perturb() (by thickness) works as expected (single profile)", {

  # simulate 25 new profiles
  expect_warning({sim.1 <- sim(s, n = 25)}) 
  #deprecated version creates sd column for perturb using hz.sd=2

  # manually create hz.sd for perturb()
  s$hz.sd = c(1, 2, 5, 5, 5, 10, 3)
  expect_silent({sim.2 <- perturb(s, n = 25, thickness.attr = "hz.sd")})

  # result is an SPC
  expect_true(inherits(sim.1, 'SoilProfileCollection'))
  expect_true(inherits(sim.2, 'SoilProfileCollection'))

  # expected lengths
  expect_true(length(sim.1) == 25)
  expect_true(length(sim.2) == 25)
})

test_that("expected errors (single profile)", {

  # only 1 seed can be used
  expect_error(perturb(sp3[1:2, ], n = 25))

  # sd must recycle evenly over number of original horizons

  # NOTE: test removed; this recycling is now handled by horizons()<-
  # in the deprecated version of sim()
})

test_that("perturb (by boundaries) works as expected (single profile)", {
  # simulate 25 new profiles with a sd boundary thickness of 0.5 - 2.5cm
  s$bdy <- round(runif(nrow(s), 1, 5)) / 2
  diagnostic_hz(s) <- data.frame(id = profile_id(s),
                                 featkind = "foo",
                                 featdept = 0, featdepb = 10)
  restrictions(s) <- data.frame(id = profile_id(s),
                                 restrkind = "bar",
                                 restrdept = 0, restrdepb = 10)
  perp <- perturb(s, n = 25, boundary.attr = "bdy")

  # result is an SPC
  expect_true(inherits(perp, 'SoilProfileCollection'))

  # expected lengths
  expect_true(length(perp) == 25)

  perp2 <- perturb(s, id = 26:50, boundary.attr = "bdy", new.idname = "foo")

  # result is an SPC
  expect_true(inherits(perp2, 'SoilProfileCollection'))

  # expected lengths
  expect_true(length(perp2) == 25)

  # custom idname
  expect_equal(idname(perp2), "foo")

  # custom IDs
  expect_equal(profile_id(perp2), as.character(26:50))
})

test_that("perturb() (by thickness) works as expected (multiple profiles)", {
  
  horizons(sp3)$hz.sd <- 2
  expect_silent({ sim.1 <- perturb(sp3, n = 25, thickness.attr = "hz.sd") })
  expect_equal(length(sim.1), 250)
  
  # test min.thickness arg (no horizon thinner than min.thickness)
  thk <- sim.1$bottom - sim.1$top
  expect_true(min(thk) >= 1)
  
  # all results are logical
  expect_true(all(checkHzDepthLogic(sim.1)$valid))
})

test_that("perturb() (by boundary) works as expected (multiple profiles)", {
  
  horizons(sp3)$hz.sd <- 2
  expect_silent({ sim.1 <- perturb(sp3[9:10,], n = 250, boundary.attr = "hz.sd", min.thickness = 3, max.depth = 50) })
  expect_equal(length(sim.1), 500)
  
  # test min.thickness arg (no horizon thinner than min.thickness)
  thk <- sim.1$bottom - sim.1$top
  expect_true(min(thk) >= 3)
  
  # all results are logical
  expect_true(all(checkHzDepthLogic(sim.1)$valid))
  
  # test max.depth (thicknesses of 4th horizon below 50cm are unaltered)
  expect_equal(thk[seq_along(thk) %% 4 == 0], c(rep(34, 250), rep(22, 250)))
})

#' @title Get Approximate Munsell Chip
#' 
#' @description Non-standard Munsell notation ('7.9YR 2.7/2.0') can be matched (nearest-neighbor, no interpolation) to the closest color within the `munsell` sRGB/CIELAB look-up table via `getClosestMunsellChip()`. A more accurate estimate of sRGB values from non-standard notation can be achieved with the \href{https://CRAN.R-project.org/package=munsellinterpol}{munsellinterpol} package. For example, conversion from Munsell to CIELAB, assuming a D65 illuminant via: `MunsellToLab('0.1Y 3.3/4.4', white='D65', adapt='Bradford')`.
#'
#' @param munsellColor character vector of strings containing Munsell notation of color, e.g. '10YR 4/3', not NA-safe
#' @param convertColors logical, should parsed Munsell colors be converted into sRGB values
#' @param ... further arguments to \code{munsell2rgb}
#'
#'
#' @return a \code{data.frame} when \code{convertColors=TRUE}, otherwise character vector
#' @export
#'
#' @examples
#' 
#' # convert a non-standard color to closest "chip" in `munsell` look-up table
#' getClosestMunsellChip('7.9YR 2.7/2.0', convertColors = FALSE)
#' 
#' # convert directly to R color
#' getClosestMunsellChip('7.9YR 2.7/2.0')
#' 
#' # special case for 2.5 value -> no rounding, we have these records in the conversion LUT
#' getClosestMunsellChip('7.5YR 2.5/2', convertColors = FALSE)
#' 
#' 
#' getClosestMunsellChip('7.5YR 6.8/4.4', convertColors = FALSE)
#' 
getClosestMunsellChip <- function(munsellColor, convertColors = TRUE, ...) {
  
  # This is a hack to avoid munsell2rgb: "no visible binding for global variable munsell" at package R CMD check
  munsell <- NULL
  
  # # init working vectors
  # # for NA propagation
  # n <- length(munsellColor)
  # closest.hue <- vector(mode = 'character', length = n)
  # closest.value <- vector(mode = 'numeric', length = n)
  # closest.chroma <- vector(mode = 'numeric', length = n)
  # 
  # # remove NA for now
  # na.idx <- which(is.na(munsellColor))
  # if(length(na.idx) > 0) {
  #   x.na <- x[na.idx]
  #   x <- x[-na.idx]
  # }
  
  # extract hue, value, chroma from single string
  cd <- parseMunsell(munsellColor, convertColors = FALSE)
  
  # extract pieces of hue
  hue.data <- .parseMunsellHue(cd$hue)
  
  ## TODO: evaluate closest standard hue via evaluation of hue positions
  ##       -> huePosition(returnHues = TRUE)
  ##       -> interpreting 10YR as the same as 0Y
  
  
  ## TODO: make NA-safe  
  
  
  # note: this is incompatible with LazyData: true
  # extract pieces from unique Munsell hues
  load(system.file("data/munsell.rda", package = "aqp")[1])
  all.hue.data <- na.omit(.parseMunsellHue(unique(munsell$hue)))
  
  # locate closest chip in `munsell` set of hues
  closest.hue <- vector(mode = 'character', length = nrow(hue.data))
  for(i in 1:nrow(hue.data)) {
    # index possible rows based on character part of hue
    idx <- which(all.hue.data$hue.character == hue.data[i, ]$hue.character)
    
    # compute Euclidean distance to all possible numeric parts of hue
    distances <- abs(hue.data$hue.numeric[i] - all.hue.data$hue.numeric[idx])
    closest.idx <- which.min(distances)
    
    # compile closest hue
    closest.hue[i] <- paste0(all.hue.data[idx, ][closest.idx, ], collapse = '')
    
  }
  
  # valid value / chroma in our LUT
  valid.value <- unique(munsell$value)
  valid.chroma <- unique(munsell$chroma)
  
  # treat as numeric
  cd$value <- as.numeric(cd$value)
  cd$chroma <- as.numeric(cd$chroma)
  
  # locate closest value / chroma
  closest.value <- vector(mode = 'numeric', length = nrow(cd))
  closest.chroma <- vector(mode = 'numeric', length = nrow(cd))
  for(i in 1:nrow(cd)) {
    # search for closest value
    idx <- which.min(abs(cd$value[i] - valid.value))
    closest.value[i] <- valid.value[idx]
    
    # search for closest chroma
    idx <- which.min(abs(cd$chroma[i] - valid.chroma))
    closest.chroma[i] <- valid.chroma[idx]
  }
  
  # convert values and chroma < 1 -> 1
  closest.value <- ifelse(closest.value < 1, 1, closest.value)
  closest.chroma <- ifelse(closest.chroma < 1, 1, closest.chroma)
  
  # optionally convert closest Munsell chips to sRGB
  if(convertColors) {
    res <- munsell2rgb(closest.hue, closest.value, closest.chroma, ...)
  } else {
    # otherwise return closest chip
    res <- paste0(closest.hue, ' ', closest.value, '/', closest.chroma)
  }
    
  return(res)
}

## TODO: 
##       * optionally return wt. mean dH, dV, dC | most frequent contrast class
##         metric = c('dE00', 'dV', 'dH, 'dC', 'CC')
##
##       * allow various specifications for color
##
##       * unify with colorQuantiles()



#' @title Quantitative Description of Color Variation
#' 
#' @description
#' This function computes several measures of "color variation", typically associated with soil colors described in the Munsell system, using the CIE2000 dE (dE00) color contrast metric. The resulting dE00 summaries map closely to color differences as perceived by "average human vision".
#' 
#' @details
#' dE00 values are computed according to `method`:
#' 
#'   * 'frequency': relative to most frequent color in `m`
#'   * 'centroid': relative to centroid (CIELAB coordinates) of colors specified in `m`
#'   * 'L1': relative to L1-median (geometric median) CIELAB coordinates of colors specified in `m`, via `Gmedian::Gmedian()`
#'   * 'reference': relative to color specified in `ref`
#' 
#' 
#' The `L1` method is more robust to outliers in `m` as compared to other methods.
#' 
#' @param m character vector of colors, described using the Munsell system e.g. `c('10YR 3/3', '5YR 4/6')`
#' @param method character, one of `c('frequency', 'centroid', 'reference')`, see Details
#' @param ref character, a reference color specified in the Munsell system when `method = 'reference'`
#'
#' @returns numeric, dE00 summary of color variation along with group centroid for `method = c('frequency', 'centroid', 'L1')`
#' @export
#'
#' @examples
#' 
#' # some brownish colors with a wild outlier
#' m <- c('10YR 3/3', '10YR 4/4', '10YR 4/4', '5GY 6/8')
#' 
#' # useful when there may be a lot of duplicates
#' # error when there is not a single, most-frequent color
#' colorVariation(m, method = 'frequency')
#' 
#' # statistical "centroid" of colors, not robust to outliers
#' # result may not match any color in `m`
#' colorVariation(m, method = 'centroid')
#' 
#' # deviation from a known reference
#' colorVariation(m, method = 'reference', ref = '10YR 2/1')
#' 
#' # L1-median (requires Gmedian package) like 'centroid'
#' # more robust to outliers
#' # result will usually be very close to a color in `m`
#' if(requireNamespace('Gmedian')) {
#'   colorVariation(m, method = 'L1')
#' }
#' 
#' 
#' # compare methods using a range of colors on a 
#' # single hue page
#' x <- expand.grid(
#'   hue = '10YR', 
#'   value = 2:7, 
#'   chroma = 2:7
#' )
#' 
#' x$m <- sprintf("%s %s/%s", x$hue, x$value, x$chroma)
#' 
#' colorChart(x$m)
#' 
#' (v <- colorVariation(x$m, method = 'centroid'))
#' contrastChart(attr(v, 'centroid'), hues = x$hue[1], thresh = v)
#' 
#' if(requireNamespace('Gmedian')) {
#'   (v <- colorVariation(x$m, method = 'L1'))
#'   contrastChart(attr(v, 'L1'), hues = x$hue[1], thresh = v)
#' }
#' 
#' 
#' # attempt to simulate colors from a group centroid
#' \dontrun{
#' v <- colorVariation(x$m, method = 'centroid')
#' 
#' s <- simulateColor(
#' method = 'dE00', 
#' n = 200, 
#' parameters = list(m = attr(v, 'centroid'), thresh = v * 1.96, hues = x$hue[1])
#' )
#' colorChart(s[[1]])
#' }
#' 
#' 
colorVariation <- function(m, method = c('frequency', 'centroid', 'L1', 'reference'), ref = NULL) {
  
  method <- match.arg(method)
  
  # dep check
  if(method == 'L1') {
    if(!requireNamespace('Gmedian')) {
      stop('package `Gmedian` is required', call. = FALSE)
    }
  }
  
  # filter Munsell colors:
  #  * NA in vector, or any position of hue value/chroma
  #  * mis-specified
  #  * non-standard notation
  idx <- which(validateMunsell(m))
  m <- m[idx]
  
  # trap all NA
  if(length(m) < 1) {
    return(NA)
  }
  
  # sorted weights and names
  wt <- sort(table(m), decreasing = TRUE)
  wt.m <- names(wt)
  
  # single color, variation is 0 
  if(length(wt) < 2) {
    message('fewer than 2 unique colors')
    return(0)
  }
  
  
  # D(colors, most frequent color)
  if(method == 'frequency') {
    
    # sanity check: there must be a single most-frequent color
    .maxwt <- wt[which.max(wt)]
    if(length(which(wt == .maxwt)) > 1) {
      stop('no single, most-frequent color: select a different method')
    }
    
    # color contrast vs. most frequent color
    cc <- colorContrast(m1 = wt.m, m2 = rep(wt.m[1], times = length(wt)))
    
    if(length(wt) < 3) {
      # unweighted
      res <- mean(cc$dE00, na.rm = TRUE)
    } else {
      # weighted
      res <- weighted.mean(cc$dE00, w = wt, na.rm = TRUE)
    }
    
    attr(res, 'most frequent') <- wt.m[1]
  }
  
  
  # D(colors, wt. color-centroid)
  if(method == 'centroid') {
    
    # centroid is wt. mean in CIELAB space
    lab <- parseMunsell(wt.m, returnLAB = TRUE)
    
    # account for colors which cannot be converted from Munsell -> CIELAB
    not.na <- which(! apply(apply(lab, 2, is.na), 1, any))
    
    lab.centroid <- apply(lab[not.na, ], 2, function(i) {
      sum(wt[not.na] * i) / sum(wt[not.na])
    })
    
    # convert back to Munsell notation for colorContrast()
    m.centroid <- col2Munsell(t(lab.centroid), space = 'CIELAB')
    m.centroid <- sprintf("%s %s/%s", m.centroid$hue, m.centroid$value, m.centroid$chroma)
    
    # color contrast vs. centroid
    cc <- colorContrast(m1 = wt.m, m2 = rep(m.centroid, times = length(wt)))
    
    if(length(wt) < 3) {
      # unweighted
      res <- mean(cc$dE00, na.rm = TRUE)
    } else {
      # weighted
      res <- weighted.mean(cc$dE00, w = wt, na.rm = TRUE)
    }
    
    attr(res, 'centroid') <- m.centroid
    
  }
  
  # D(colors, L1 median)
  if(method == 'L1') {
    
    # use all colors, not unique colors with weights
    lab <- parseMunsell(m, returnLAB = TRUE)
    
    # automatically removes NA
    lab.centroid <- Gmedian::Gmedian(lab)
    
    # convert back to Munsell notation for colorContrast()
    m.centroid <- col2Munsell(lab.centroid, space = 'CIELAB')
    m.centroid <- sprintf("%s %s/%s", m.centroid$hue, m.centroid$value, m.centroid$chroma)
    
    # color contrast vs. centroid
    cc <- colorContrast(m1 = wt.m, m2 = rep(m.centroid, times = length(wt)))
    
    if(length(wt) < 3) {
      # unweighted
      res <- mean(cc$dE00, na.rm = TRUE)
    } else {
      # weighted
      res <- weighted.mean(cc$dE00, w = wt, na.rm = TRUE)
    }
    
    attr(res, 'L1') <- m.centroid
    
  }
  
  # D(color, reference color)
  if(method == 'reference') {
    
    # color contrast vs. reference color
    cc <- colorContrast(m1 = wt.m, m2 = rep(ref, times = length(wt)))
    
    if(length(wt) < 3) {
      # unweighted
      res <- mean(cc$dE00, na.rm = TRUE)
    } else {
      # weighted
      res <- weighted.mean(cc$dE00, w = wt)
    }
    
  }
  
  return(res)
}

context("colorVariation()")

test_that("works as expected", {
  
  # all NA
  v <- colorVariation(NA)
  expect_equal(v, NA)
  
  # too few obs
  v <- colorVariation(c(NA, '10YR 2/2'))
  expect_equal(v, 0)
  
  # frequency
  m <- c('10YR 3/3', '10YR 4/4', '10YR 4/4', '5GY 6/8')
  v <- colorVariation(m, method = 'frequency')
  
  # strip attributes
  expect_equal(round(as.vector(v)), 10.0)
  
  # check attributes
  expect_equal(attr(v, 'most frequent'), '10YR 4/4')
  

  # centroid
  v <- colorVariation(m, method = 'centroid')
  
  # strip attributes
  expect_equal(round(as.vector(v)), 13.0)
  
  # check attributes
  expect_equal(attr(v, 'centroid'), '5Y 4/4')
  
})

# Here are some ideas I have been considering around indexing horizon data in a SoilProfileCollection based on "spatial predicates" that operate on the "horizon geometry". 
# I want to be able to find the horizons that are above or below arbitrary horizons for the whole collection.

# hzOffset() allows for vertical offset (in terms of j index units) of horizon IDs. Two aliases are `hzAbove()` and `hzBelow()` that just use a different sign on the `offset` argument. 

# I think we could also probably do things like hz_adjacent(), hz_between/within(). Also we could make it more convenient to prepare the vectors of target horizon IDs. 



#' @title Select Horizons Above or Below a Reference Horizon
#' 
#' @description
#' These functions return "selected" horizons offset from "reference" horizon defined by logical expressions which include horizon data elements. Typically, `hzAbove()` and `hzBelow()` are used as a simpler interface to `hzOffset()`. Selected horizons can be returned as a complete `SoilProfileCollection` (`SPC = TRUE`), a list of horizon data row indices (`SPC = FALSE, simplify = FALSE`), or as a numeric vector of horizon data row indices (`SPC = FALSE, simplify = TRUE`).
#' 
#' @details Expressions that match multiple reference horizons will result in the inclusion of these reference horizons in the selected horizons. Interpreting multiple reference horizons as a single horizon (`single = TRUE`), defined by min(top) - max(bottom), will ensure that only selected horizons are returned.
#' 
#' The `offset` argument controls which "horizons above" or "horizons below" reference horizon(s) are selected, using the `SoilProfileCollection` `[, j]` (horizon/slice) notation. Possible values and their interpretation are as follows:
#' 
#'  * `offset = NULL`: select all horizons above or below reference horizon(s)
#'  * `offset = 1`: select the first horizon above or below reference horizon(s)
#'  * `offset = 1:2`: select the first and second horizons above or below reference horizon(s)
#'  * `offset = 1:10`: safely select the 1st -> 10th horizons, even if there are fewer, above or below reference horizon(s)
#' 
#' @param x `SoilProfileCollection`
#' 
#' @param ... comma-separated set of logical expressions, in context of horizon data, length for each expression must match number of horizons in `x`
#' 
#' @param hzidx integer vector of target horizon row indices, typically calculated by `hzAbove()` and `hzBelow()`
#' 
#' @param offset integer vector or NULL, when specified this is the `[, j]` (horizon/slice) index definining the horizons above or below a reference horizon that should be returned. see details
#' 
#' @param SPC logical, return a `SoilProfileCollection` subset to selected horizons
#' 
#' @param simplify logical, flatten list of horizon row indices into a numeric vector
#' 
#' @param single logical, interpret multiple matching reference horizons as a single horizon, defined by max(top) - min(bottom)
#' 
#' @return One of:
#'  * `SoilProfileCollection`: when `SPC = TRUE`
#'  * list of horizon row indices when `SPC = FALSE` and `simplify = FALSE`
#'  * vector of horizon row indices: when `SPC = FALSE` and `simplify = TRUE`
#' 
#' The return value will not include data from profiles which have no matching reference horizons.
#' 
#' @export
#' @rdname hzOffset
#' 
#' @examples
#' 
#' # example data
#' x <- c(
#'   'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
#'   'P2:Ap|AA|E|BhsBhs|Bw1Bw1|CCCCC',
#'   'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR',
#'   'P4:AA|EEE|BhsBhsBhsBhs|BwBw|CCCCC',
#'   'P5:AAAA|ACACACACAC|CCCCCCCCCCC|CdCdCd'
#' )
#' 
#' s <- quickSPC(x)
#' 
#' # single match for most profiles
#' .ex <- grepl('Bt3|Bw', s$name)
#' s$e <- .ex
#' 
#' # default: select all horizons above or below
#' a <- hzAbove(s, .ex, SPC = FALSE, simplify = TRUE)
#' b <- hzBelow(s, .ex, SPC = FALSE, simplify = TRUE)
#' 
#' op <- par(no.readonly = TRUE)
#' par(mar = c(0, 0, 3, 0))
#' plotSPC(
#'   s, color = 'e', col.label = 'reference', col.palette = c('grey', 'royalblue'), 
#'   name = 'name', hz.depths = TRUE, depth.axis = FALSE, 
#'   name.style = 'center-center', cex.names = 0.75
#' )
#' 
#' # highlight selected horizons above and below with brackets
#' addBracket(
#'   depths(s, hzID = FALSE)[a, ], agg = TRUE,
#'   offset = -0.3, col = 'darkgreen', tick.length = 0, lwd = 3
#' )
#' 
#' addBracket(
#'   depths(s, hzID = FALSE)[b, ], agg = TRUE,
#'   offset = -0.35, col = 'firebrick', tick.length = 0, lwd = 3
#' )
#' 
#' 
#' # select only 1st horizon above / below
#' a <- hzAbove(s, .ex, SPC = FALSE, simplify = TRUE, offset = 1)
#' b <- hzBelow(s, .ex, SPC = FALSE, simplify = TRUE, offset = 1)
#' 
#' plotSPC(
#'   s, color = 'e', col.label = 'reference', col.palette = c('grey', 'royalblue'), 
#'   name = 'name', hz.depths = TRUE, depth.axis = FALSE, 
#'   name.style = 'center-center', cex.names = 0.75
#' )
#' 
#' # highlight selected horizons above and below with brackets
#' addBracket(
#'   depths(s, hzID = FALSE)[a, ], agg = TRUE,
#'   offset = -0.3, col = 'darkgreen', tick.length = 0, lwd = 3
#' )
#' 
#' addBracket(
#'   depths(s, hzID = FALSE)[b, ], agg = TRUE,
#'   offset = -0.35, col = 'firebrick', tick.length = 0, lwd = 3
#' )
#' 
#' 
#' # multiple matches
#' .ex <- grepl('B', s$name)
#' s$e <- .ex
#' 
#' # default
#' a <- hzAbove(s, .ex, SPC = FALSE, simplify = TRUE)
#' b <- hzBelow(s, .ex, SPC = FALSE, simplify = TRUE)
#' 
#' par(mfcol = c(1, 2))
#' plotSPC(
#'   s,  col.label = 'reference', color = 'e', 
#'   col.palette = c('grey', 'royalblue'), 
#'   name = 'name', hz.depths = TRUE, depth.axis = FALSE, 
#'   name.style = 'center-center', cex.names = 0.75
#' )
#' 
#' addBracket(
#'   depths(s, hzID = FALSE)[a, ], agg = TRUE,
#'   offset = -0.3, col = 'darkgreen', tick.length = 0, lwd = 3
#' )
#' 
#' addBracket(
#'   depths(s, hzID = FALSE)[b, ], agg = TRUE,
#'   offset = -0.35, col = 'firebrick', tick.length = 0, lwd = 3
#' )
#' 
#' mtext('single = FALSE',  side = 1, line = -1.5, at = 0, adj = -0.5)
#' 
#' 
#' # interpret multiple reference hz as a single reference hz
#' a <- hzAbove(s, .ex, SPC = FALSE, simplify = TRUE, single = TRUE)
#' b <- hzBelow(s, .ex, SPC = FALSE, simplify = TRUE, single = TRUE)
#' 
#' 
#' plotSPC(
#'   s, col.label = 'reference', color = 'e', 
#'   col.palette = c('grey', 'royalblue'), name = 'name', 
#'   hz.depths = TRUE, depth.axis = FALSE, 
#'   name.style = 'center-center', cex.names = 0.75
#' )
#' 
#' addBracket(
#'   depths(s, hzID = FALSE)[a, ], agg = TRUE,
#'   offset = -0.3, col = 'darkgreen', tick.length = 0, lwd = 3
#' )
#' 
#' addBracket(
#'   depths(s, hzID = FALSE)[b, ], agg = TRUE,
#'   offset = -0.35, col = 'firebrick', tick.length = 0, lwd = 3
#' )
#' 
#' mtext('single = TRUE',  side = 1, line = -1.5, at = 0, adj = -0.5)
#' 
#' 
#' # demonstrate SPC = TRUE, single = TRUE
#' plotSPC(
#'   s, col.label = 'reference', color = 'e', 
#'   col.palette = c('grey', 'royalblue'), name = 'name', 
#'   hz.depths = TRUE, depth.axis = FALSE, 
#'   name.style = 'center-center', cex.names = 0.75, max.depth = 250
#' )
#' 
#' a <- hzAbove(s, .ex, SPC = TRUE, single = TRUE)
#' 
#' plotSPC(
#'   a, name = 'name', hz.depths = TRUE, depth.axis = FALSE, 
#'   name.style = 'center-center', cex.names = 0.75, max.depth = 250
#' )
#' title('selected profiles/horizons')
#' 
#' par(op)


#' 
hzAbove <- function(x, ..., offset = NULL, SPC = TRUE, simplify = SPC, single = FALSE) {
  .NHZ <- NULL
  
  # "above" is a negative offset in j index
  
  # missing offset indicates "all horizons above target"
  if(is.null(offset)) {
    # use max number of horizons within x
    offset <- seq(from = 1, to = max(x[, , , .NHZ]))
  }
  
  # capture expression(s) at function
  .dots <- substitute(list(...))
  .dots <- .dots[2:length(.dots)]
  
  # create object to facilitate eval
  .data <- horizons(x)
  
  # loop through list of expressions and evaluate
  res <- vector('list', length(.dots))
  for (i in 1:length(.dots)) {
    res[[i]] <- eval(.dots[[i]], .data, parent.frame(n = 2))
  }
  
  subcrit <- Reduce('&', res)
  
  if (!is.logical(subcrit)) {
    badxpr <- paste0("'",paste0(.dots[sapply(.dots, function(x) !is.logical(x))],
                                collapse = ",'"),"'")
    stop(sprintf("%s is not logical", badxpr), call. = FALSE)
  }
  
  # interpret multiple matches as a single reference horizon
  # offsets are applied relative to top-most reference hz
  if(single) {
    .idx <- which(subcrit)
    
    .d <- depths(x, hzID = FALSE)[.idx, ]
    .d$idx <- .idx
    
    # for each profile
    # find the horizon row index of top-most reference hz
    .topmost <- sapply(split(.d, .d[[1]]), function(i) {
      i$idx[which.min(i[[2]])]
    })
  } else {
    .topmost <- which(subcrit)
  }
  
  hzOffset(x, hzidx = .topmost, offset = -offset, SPC = SPC, simplify = simplify)
}

#' @export
#' @rdname hzOffset
hzBelow <- function(x, ..., offset = NULL, SPC = TRUE, simplify = SPC, single = FALSE) {
  .NHZ <- NULL
  
  # "below" is a positive offset in j index
  
  # missing offset indicates "all horizons below target"
  if(is.null(offset)) {
    # use max number of horizons within x
    offset <- seq(from = 1, to = max(x[, , , .NHZ]))
  }
  
  # capture expression(s) at function
  .dots <- substitute(list(...))
  .dots <- .dots[2:length(.dots)]
  
  # create object to facilitate eval
  .data <- horizons(x)
  
  # loop through list of expressions and evaluate
  res <- vector('list', length(.dots))
  for (i in 1:length(.dots)) {
    res[[i]] <- eval(.dots[[i]], .data, parent.frame(n = 2))
  }
  
  subcrit <- Reduce('&', res)
  
  if (!is.logical(subcrit)) {
    badxpr <- paste0("'",paste0(.dots[sapply(.dots, function(x) !is.logical(x))],
                                collapse=",'"),"'")
    stop(sprintf("%s is not logical", badxpr), call. = FALSE)
  }
  
  # interpret multiple matches as a single reference horizon
  # offsets are applied relative to deepest reference hz
  if(single) {
    .idx <- which(subcrit)
    
    .d <- depths(x, hzID = FALSE)[.idx, ]
    .d$idx <- .idx
    
    # for each profile
    # find the horizon row index of deepest reference hz
    .deepest <- sapply(split(.d, .d[[1]]), function(i) {
      i$idx[which.max(i[[3]])]
    })  
  } else {
    .deepest <- which(subcrit)
  }
  
  hzOffset(x, hzidx = .deepest, offset = offset, SPC = SPC, simplify = simplify)
}

#' @export
#' @rdname hzOffset
hzOffset <- function(x, hzidx, offset, SPC = FALSE, simplify = TRUE) {
  # define SPC k-keywords as local vars for R CMD CHECK
  .LAST <- NULL; .HZID <- NULL
  
  if (SPC) {
    simplify <- TRUE
  }
  
  # calculate first and last horizon indices
  hzidl <- x[,, .LAST, .HZID]
  hzidf <- c(1, hzidl[seq_len(pmax(0, length(hzidl) - 1))] + 1)[seq_along(hzidl)]
  
  # determine intersection between each profile horizon index and the target ID + offset
  idx <- lapply(seq_along(hzidf), function(i) {
    haystack <- unique(do.call('c', lapply(offset, function(o) {
      o + hzidx[hzidx >= hzidf[i] & hzidx <= hzidl[i]]
    })))
    intersect(x = seq(hzidf[i], hzidl[i]), y = haystack)
  })
  
  if (simplify) {
    idx <- do.call('c', idx)
  }
  
  if (!SPC) {
    return(idx)
  }
  
  # assemble reduced SPC here
  # profiles with no matching reference horizon are left out
  
  # TODO: subset NSE needs var in the SPC / "needs" rlang here
  .hzldx <- NULL
  x$.hzldx <- seq_len(nrow(x)) %in% idx
  res <- subsetHz(x, .hzldx) 
  res$.hzldx <- NULL
  
  return(res)
}

test_that("hzOffset works", {
  
  # sample data where two profiles have no bottom depths (but all top depths are present)
  h <- data.frame(
    id = c(1, 1, 1, 2, 2, 2, 2, 3, 3),
    top = c(0:2, 0:3, 0:1) * 10,
    bottom = c(rep(NA_integer_, 7), c(10, 99))
  )
  expect_warning(depths(h) <- id ~ top + bottom)
  
  # index of last horizon in each profile
  bottom.idx <- which(hzID(h) %in% getLastHorizonID(h)) 
  #> [1] 3 7 9
  
  # horizons 1 index above/below bottom
  expect_equal(hzOffset(h, bottom.idx, offset = -1), c(2, 6, 8))
  expect_equal(hzOffset(h, bottom.idx, offset = 1), numeric(0))
  
  # simplify = FALSE (list results)
  expect_equal(hzOffset(h, bottom.idx, offset = -1, simplify = FALSE), list(2, 6, 8))
  expect_equal(hzOffset(h, bottom.idx, offset = 1, simplify = FALSE), 
               list(numeric(0), numeric(0), numeric(0)))
  
  # horizons 2 indices above bottom
  expect_equal(hzOffset(h, bottom.idx, offset = -2), c(1, 5))
  expect_equal(hzOffset(h, bottom.idx, offset = 2), numeric(0))
  
  # horizons 1 + 2 indices above bottom
  expect_equal(hzOffset(h, bottom.idx, offset = -(1:2)), c(1, 2, 5, 6, 8))
  expect_equal(hzOffset(h, bottom.idx, offset = 1:2), numeric(0))
  
  # there are never horizons above the first horizon (in a profile)
  first <- h[, 1, .HZID]
  expect_equal(hzOffset(h, first, offset = -1), numeric(0))
  
  # the second horizon is always right after the first (if profile has more than 1 horizon)
  expect_equal(hzOffset(h, first, offset = 1), first + 1)
  
  # single profile SPC has one value in result
  expect_equal(length(hzOffset(h[1,], first, offset = 1)), 1)
})

test_that("basic functionality of hzAbove() / hzBelow()", {
  
  data(sp4)
  depths(sp4) <- id ~ top + bottom
  
  # all horizons indices above horizons with ex-Ca:Mg ratio less than 1:10
  x <- hzAbove(sp4, ex_Ca_to_Mg < 0.1, SPC = FALSE, simplify = TRUE)
  expect_equal(x, c(1L, 2L, 3L, 10L, 11L, 12L, 23L, 24L, 25L, 26L))
  
  # all horizon indices below any horizon with ex-Ca:Mg ratio less than 1:10
  x <- hzBelow(sp4, ex_Ca_to_Mg < 0.1, SPC = FALSE, simplify = TRUE)
  expect_equal(x, c(4L, 27L))
  
})


test_that("interpret multiple reference horizons as a single span", {
  
  # example data
  x <- c(
    'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
    'P2:Ap|AA|E|BhsBhs|Bw1Bw1|CCCCC',
    'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR',
    'P4:AA|EEE|BhsBhsBhsBhs|BwBw|CCCCC',
    'P5:AAAA|ACACACACAC|CCCCCCCCCCC|CdCdCd'
  )
  
  s <- quickSPC(x)

  # interpret multiple reference hz as a single reference hz
  a <- hzAbove(s, grepl('B', name), SPC = FALSE, simplify = TRUE, single = TRUE)
  b <- hzBelow(s, grepl('B', name), SPC = FALSE, simplify = TRUE, single = TRUE)
  
  # hand-verified
  expect_equal(a, c(1L, 5L, 6L, 7L, 11L, 17L, 18L))
  expect_equal(b, c(3L, 4L, 10L, 15L, 16L, 21L))
    
})

# miscellaneous soil color indices

#' @title Horizon Color Indices
#' @description Calculate basic horizon-level color indices for a SoilProfileCollection. Basic indices do not require aggregation over the whole profile or comparison to a "reference" (e.g. parent material) color. Includes Hurst (1977) Redness Index, Barron-Torrent Redness Index (1986) and Buntley-Westin Index (1965). This is a wrapper method around several horizon-level indices. See the individual functions for more details.
#' @param p A SoilProfileCollection
#' @param hue Column name containing moist hue; default: "m_hue"
#' @param value Column name containing moist value; default: "m_value"
#' @param chroma Column name containing moist chroma; default: "m_chroma"
#' @return A data.frame containing unique pedon and horizon IDs and horizon-level color indices.
#' @author Andrew G. Brown
#' @seealso \code{\link{hurst.redness}} \code{\link{barron.torrent.redness.LAB}} \code{\link{buntley.westin.index}}
#' @examples
#' 
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' data(sp1)
#'
#' # promote sp1 data to SoilProfileCollection
#' depths(sp1) <- id ~ top + bottom
#'
#' # move site data
#' site(sp1) <- ~ group
#' 
#' # use Munsell color notation as horizon name
#' sp1$m <- sprintf("%s %s/%s", sp1$hue, sp1$value, sp1$chroma)
#' 
#' # compute indices
#' # merged into `sp1` with left-join on hzidname(sp1)
#' horizons(sp1) <- horizonColorIndices(sp1, hue="hue", value="value", chroma="chroma")
#'
#' # visualize
#' par(mar = c(0, 1, 3, 1))
#' plotSPC(sp1, color='hurst_redness', name = 'm')
#' plotSPC(sp1, color='barron_torrent_redness', name = 'm')
#' plotSPC(sp1, color='buntley_westin', name = 'm')
#' 
#' @rdname horizonColorIndices
#' @export horizonColorIndices
horizonColorIndices <- function(p, hue="m_hue", value="m_value", chroma="m_chroma") {
  hz <- horizons(p)

  hz$hurst_redness <- hurst.redness(hz[[hue]], hz[[value]], hz[[chroma]])

  hz$barron_torrent_redness <- barron.torrent.redness.LAB(hz[[hue]], hz[[value]], hz[[chroma]])

  hz$buntley_westin <- buntley.westin.index(hz[[hue]], hz[[chroma]])

  return(.data.frame.j(hz, c(idname(p), hzidname(p), names(hz)[!names(hz) %in% names(horizons(p))]), aqp_df_class(p)))
}

#' @title Hurst (1977) Redness Index
#' @description Calculate Redness Index after Hurst (1977) "Visual estimation of iron in saprolite" DOI: 10.1130/0016-7606(1977)88<174:VEOIIS>2.0.CO;2. Accepts vectorized inputs for hue, value and chroma, produces vector output.
#' @param hue A character vector containing Munsell hues (e.g. "7.5YR")
#' @param value A numeric vector containing Munsell values
#' @param chroma A numeric vector containing Munsell chromas
#' @return A numeric vector of horizon redness index (lower values = redder).
#' @author Andrew G. Brown
#' @references Hurst, V.J. (1977) Visual estimation of iron in saprolite. GSA Bulletin. 88(2): 174–176. doi: https://doi.org/10.1130/0016-7606(1977)88<174:VEOIIS>2.0.CO;2
#' @rdname hurst.redness
#' @export hurst.redness
hurst.redness <-  function(hue, value, chroma) {
  # after Hurst (1977) "Visual estimation of iron in saprolite"
  # 10.1130/0016-7606(1977)88<174:VEOIIS>2.0.CO;2
  hue.lookup.table <- seq(5, 22.5, 2.5)
  names(hue.lookup.table) <- c('5R','7.5R','10R','2.5YR',
                               '5YR','7.5YR','10YR','2.5Y')
  hstar <- hue.lookup.table[hue]
  return(hstar * value / chroma)
}


## TODO for some future time: consider ln(R_{Lab} due to the extreme differences in scale as Munsell value varies by even a single chip)

#' @title Barron & Torrent (1986) Redness Index in LAB color space
#' @description Calculate Redness Index after Barron & Torrent (1986) "Use of the Kubelka—Munk Theory to Study the Influence of Iron Oxides on Soil Colour" using Munsell colors converted to LAB. DOI: 10.1111/j.1365-2389.1986.tb00382.x. Accepts vectorized inputs for hue, value and chroma, produces vector output.
#' @param hue A character vector containing Munsell hues (e.g. "7.5YR")
#' @param value A numeric vector containing Munsell values
#' @param chroma A numeric vector containing Munsell chromas
#' @return A numeric vector of horizon redness index (higher values = redder).
#' @author Andrew G. Brown
#' @references Barron, V. and Torrent, J. (1986), Use of the Kubelka-Munk theory to study the influence of iron oxides on soil colour. Journal of Soil Science, 37: 499-510. doi:10.1111/j.1365-2389.1986.tb00382.x
#' @rdname barron.torrent.redness.LAB
#' @export barron.torrent.redness.LAB
barron.torrent.redness.LAB <- function(hue, value, chroma) {
  # after Barron & Torrent (1986) "Use of the Kubelka-Munk Theory to Study the Influence of Iron Oxides on Soil Colour"
  # 10.1111/j.1365-2389.1986.tb00382.x
  lab <- munsell2rgb(hue, value, chroma, returnLAB = TRUE)
  res <- with(lab, (A * sqrt(A^2 + B^2) * 10^10) / (B * L^6))
  return(res)
}

#' @title Harden (1982) Rubification
#' @description Calculate "rubification" component of "Profile Development Index" after Harden (1982) "A quantitative index of soil development from field descriptions: Examples from a chronosequence in central California". Accepts vectorized inputs for hue and chroma to produce vector output.
#'
#' In Harden (1982) "rubification" is calculated relative to a reference parent material. Several other non-color components are normalized relative to a maximum value and summed to obtain the overall Profile Development Index.
#'
#' @param hue A character vector containing Munsell hues (e.g. "7.5YR")
#' @param chroma A numeric vector containing Munsell chromas
#' @param hue_ref A character vector containing Munsell hue(s) (e.g. "10YR") for reference material
#' @param chroma_ref A numeric vector containing Munsell chroma(s) for reference material
#' @return A numeric vector reflecting horizon redness increase relative to a reference (e.g. parent) material.
#' @author Andrew G. Brown
#' @references Harden, J.W. (1982) A quantitative index of soil development from field descriptions: Examples from a chronosequence in central California. Geoderma. 28(1) 1-28. doi:  10.1016/0016-7061(82)90037-4
#' @rdname harden.rubification
#' @export harden.rubification
#' @examples
#' 
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' library(aqp)
#' data("jacobs2000", package="aqp")
#'
#' # LEFT JOIN hue, value, chroma matrix color columns
#' horizons(jacobs2000) <- cbind(horizons(jacobs2000)[,c(idname(jacobs2000), hzidname(jacobs2000))],
#'                               parseMunsell(jacobs2000$matrix_color_munsell, convertColors = FALSE))
#'
#' #' # calculate a mixed 150-200cm color ~"parent material"
#' jacobs2000$c_horizon_color <- profileApply(jacobs2000, function(p) {
#'
#'   # and derive the parent material from the 150-200cm interval
#'   p150_200 <- glom(p, 150, 200, truncate = TRUE)
#'   p150_200$thickness <- p150_200$bottom - p150_200$top
#'   
#'   # subset colors and thickness
#'   clrs <- na.omit(horizons(p150_200)[,c('matrix_color_munsell','thickness')])
#'   
#'   # simulate a subtractive mixture using thickness as weight
#'   mixMunsell(
#'   clrs$matrix_color_munsell, 
#'   w = clrs$thickness, 
#'   mixingMethod = 'exact')$munsell
#' })
#'
#' # segment profile into 1cm slices (for proper depth weighting)
#' jacobs2000$rubif <- profileApply(jacobs2000, function(p) {
#'
#'   # sum the melanization index over the 0-100cm interval
#'   p0_100 <- hz_segment(p, 0:100)
#'
#'   ccol <- parseMunsell(p$c_horizon_color, convertColors = FALSE)
#'
#'   sum(harden.rubification(
#'     hue = p0_100$hue,
#'     chroma = as.numeric(p0_100$chroma),
#'     hue_ref = ccol$hue,
#'     chroma_ref = as.numeric(ccol$chroma)
#'   ), na.rm = TRUE)
#'
#' })
#'
#' jacobs2000$rubiforder <- order(jacobs2000$rubif)
#'
#' # Plot in order of increasing Rubification index
#'
#' plotSPC(jacobs2000,
#' color = "matrix_color",
#' label = "rubif",
#' plot.order = jacobs2000$rubiforder,
#' max.depth = 250
#' )
#' 
#' segments(
#'   x0 = 0.5, 
#'   x1 = length(jacobs2000) + 0.5, 
#'   y0 = c(0,100,150,200), 
#'   y1 = c(0,100,150,200), 
#'   lty = 2
#' )
#' 
#' # Add [estimated] parent material color swatches
#' trash <- sapply(seq_along(jacobs2000$c_horizon_color), function(i) {
#'   rect(i - 0.15, 250, i + 0.15, 225,
#'        col = parseMunsell(jacobs2000$c_horizon_color[jacobs2000$rubiforder[i]]))
#' })
harden.rubification <- function(hue, chroma, hue_ref, chroma_ref) {
  # after Harden (1982)  "A quantitative index of soil development from field descriptions:
  #                       Examples from a chronosequence in central California"
  hue.lookup.table <- 0:8
  names(hue.lookup.table) <- c('5R','7.5R','10R','2.5YR','5YR','7.5YR','10YR','2.5Y')
  dH <- -as.numeric(hue.lookup.table[hue] - hue.lookup.table[hue_ref])
  dC <- chroma - chroma_ref
  return(10 * (dH + dC))
}

#' @title Harden (1982) Melanization
#'
#' @description Calculate "melanization" component of "Profile Development Index" after Harden (1982) "A quantitative index of soil development from field descriptions: Examples from a chronosequence in central California". Accepts vectorized inputs for value and reference value to produce vector output. A convenient use case would be to apply this on a profile-specific basis, where the `value_ref` has a single value, and `value` is a vector of length equal to the number of horizons within the upper 100 cm.
#'
#' @details In Harden (1982), "melanization" is calculated relative to a reference parent material for all horizons within 100cm of the soil surface. In addition, several other non-color components are normalized relative to a maximum value and summed to obtain the overall Profile Development Index.
#'
#' @param value numeric vector containing Munsell values
#' @param value_ref A numeric vector containing Munsell value(s) for reference material
#' @return A numeric vector reflecting horizon darkening relative to a reference (e.g. parent) material.
#' @author Andrew G. Brown
#' @references Harden, J.W. (1982) A quantitative index of soil development from field descriptions: Examples from a chronosequence in central California. Geoderma. 28(1) 1-28. doi:  10.1016/0016-7061(82)90037-4
#' @rdname harden.melanization
#' @export harden.melanization
#' @examples
#' 
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' library(aqp)
#' data("jacobs2000", package="aqp")
#'
#' # LEFT JOIN hue, value, chroma matrix color columns
#' horizons(jacobs2000) <- cbind(horizons(jacobs2000)[,c(idname(jacobs2000), hzidname(jacobs2000))],
#'                               parseMunsell(jacobs2000$matrix_color_munsell, convertColors = FALSE))
#'
#' # calculate a mixed 150-200cm color ~"parent material"
#'
#' jacobs2000$c_horizon_color <- profileApply(jacobs2000, function(p) {
#'
#'   # and derive the parent material from the 150-200cm interval
#'   p150_200 <- glom(p, 150, 200, truncate = TRUE)
#'   p150_200$thickness <- p150_200$bottom - p150_200$top
#'
#'   # mix colors
#'   clrs <- na.omit(horizons(p150_200)[,c('matrix_color_munsell','thickness')])
#'   mixMunsell(clrs$matrix_color_munsell, w = clrs$thickness)$munsell
#'
#' })
#'
#' # segment profile into 1cm slices (for proper depth weighting)
#' jacobs2000$melan <- profileApply(jacobs2000, function(p) {
#'
#'   # sum the melanization index over the 0-100cm interval
#'   p0_100 <- hz_segment(p, 0:100)
#'
#'   ccol <- parseMunsell(p$c_horizon_color, convertColors = FALSE)
#'
#'   sum(harden.melanization(
#'     value = as.numeric(p0_100$value),
#'     value_ref = as.numeric(ccol$value)), na.rm = TRUE)
#'
#' })
#'
#' jacobs2000$melanorder <- order(jacobs2000$melan)
#'
#' # Plot in order of increasing Melanization index
#'
#' plotSPC(jacobs2000, 
#'         color = "matrix_color",
#'         label = "melan",
#'         plot.order = jacobs2000$melanorder,
#'         max.depth = 250
#'         )
#'
#' segments(
#'   x0 = 0.5, 
#'   x1 = length(jacobs2000) + 0.5, 
#'   y0 = c(0,100,150,200), 
#'   y1 = c(0,100,150,200), 
#'   lty = 2
#' )
#'
#' # Add [estimated] parent material color swatches
#' lapply(seq_along(jacobs2000$c_horizon_color), function(i) {
#'   rect(i - 0.15, 250, i + 0.15, 225,
#'        col = parseMunsell(jacobs2000$c_horizon_color[jacobs2000$melanorder[i]]))
#' })
harden.melanization <- function(value, value_ref) {
  # for horizons within 100cm of soil surface
  # after Harden (1982)  "A quantitative index of soil development from field descriptions:
  #                       Examples from a chronosequence in central California"
  # 10.1016/0016-7061(82)90037-4
  return(10 * (value_ref - value))
}

#' @title Buntley-Westin (1965) Index
#' @description Calculate "Color Development Equivalent" by the method of Buntley & Westin (1965) "A Comparative Study of Developmental Color in a Chestnut-Chernozem-Brunizem Soil Climosequence" DOI: 10.2136/sssaj1965.03615995002900050029x. Originally developed for Mollisols, the Buntley-Westin index has been used as a tool to separate soils based on depth to particular colors.
#' @param hue A character vector containing Munsell hues (e.g. "7.5YR")
#' @param chroma A numeric vector containing Munsell chromas
#' @return A numeric vector reflecting horizon color development.
#' @author Andrew G. Brown
#' @references Buntley, G.J. and Westin, F.C. (1965), A Comparative Study of Developmental Color in a Chestnut-Chernozem-Brunizem Soil Climosequence. Soil Science Society of America Journal, 29: 579-582. doi:10.2136/sssaj1965.03615995002900050029x
#' @rdname buntley.westin.index
#' @export buntley.westin.index
buntley.westin.index <- function(hue, chroma) {
  # after Buntley-Westin (1965) "A Comparative Study of Developmental Color in a Chestnut-Chernozem-Brunizem Soil Climosequence"
  # 10.2136/sssaj1965.03615995002900050029x
  hue.lookup.table <- 7:0
  #note: original publication included interpolation of 1.25Y hue @ 2.5
  names(hue.lookup.table) <- c('10R','2.5YR','5YR','7.5YR','10YR','2.5Y','5Y','N')
  return(hue.lookup.table[hue] * chroma)
}

#' @title Thompson-Bell (1996) Index
#' @description Calculate the "Profile Darkness Index" by the method of Thompson & Bell (1996) "Color index for identifying hydric conditions for seasonally saturated mollisols in Minnesota" DOI: 10.2136/sssaj1996.03615995006000060051x. The Thompson-Bell Index has been shown to reflect catenary relationships in some Mollisols of Minnesota (generally: wetter landscape positions = thicker, darker surfaces).
#' @param p A single-profile SoilProfileCollection (e.g. via profileApply())
#' @param name Column name containing horizon designations used to find A horizons (default: first column name containing 'name')
#' @param pattern Regular expression to match A horizons (default: "^A" which means horizon designation _starts with_ A)
#' @param value Column name containing horizon color values (default: "m_value")
#' @param chroma Column name containing horizon color chromas (default: "m_chroma")
#' @return A numeric vector reflecting horizon darkness (lower values = darker).
#' @author Andrew G. Brown
#' @references Thompson, J.A. and Bell, J.C. (1996), Color Index for Identifying Hydric Conditions for Seasonally Saturated Mollisols in Minnesota. Soil Science Society of America Journal, 60: 1979-1988. doi:10.2136/sssaj1996.03615995006000060051x
#' @rdname thompson.bell.darkness
#' @export thompson.bell.darkness
thompson.bell.darkness <- function(p,
                                   name = hzdesgnname(p, required = TRUE),
                                   pattern = "^A",
                                   value = "m_value",
                                   chroma = "m_chroma") {
    
  # after Thompson & Bell (1996) "Color index for identifying hydric conditions for seasonally saturated mollisols in Minnesota"
  # 10.2136/sssaj1996.03615995006000060051x
  
  hz <- horizons(p)
  depthz <- horizonDepths(p)
  
  if (is.null(name) || !name %in% horizonNames(p)) {
    stop("Horizon designation column (", name, ") does not exist.")
  }

  a.hz <- hz[grepl(hz[[name]], pattern = pattern),]
  a.hz$pdi_partial <- (a.hz[[depthz[2]]] - a.hz[[depthz[1]]]) / (a.hz[[value]] * a.hz[[chroma]] + 1)
  return(sum(a.hz$pdi_partial))
}

context("Profile- and Horizon-level Color Indices")

data(sp1)
# convert colors from Munsell to hex-encoded RGB
sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))

# promote to SoilProfileCollection
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

## tests

test_that("horizonColorIndices works as expected", {
  # implicit left-join
  horizons(sp1) <- horizonColorIndices(sp1, hue="hue", value="value", chroma="chroma")
  expect_equivalent(length(sp1$hurst_redness), 60)
})

test_that("harden.rubification works as expected", {
  soil_hue <- "5YR"
  soil_chroma <- 4
  
  pm_hue <- "2.5Y"
  pm_chroma <- 2
  
  expect_equivalent(harden.rubification(soil_hue, soil_chroma, pm_hue, pm_chroma), 50) 
})

test_that("harden.melanization works as expected", {
  soil_value <- 2.5
  pm_value <- 5
  expect_equivalent(harden.melanization(soil_value, pm_value), 25)
})

test_that("thompson.bell.darkness works as expected", {
  expect_equivalent(thompson.bell.darkness(
    sp1[1],
    name = "name",
    value = "value",
    chroma = "chroma"
  ), 5.5)
})

#' .makeEquivalentMunsellLUT
#'
#' @description Makes the look up table based on pair-wise CIE2000 color contrast (\code{dE00}) of LAB colors with D65 illuminant of LAB colors for all whole value/chroma "chips" in the \code{aqp::munsell} data set via \code{farver::compare_colour}. Specify a threshold in terms of a probability level of CIE2000 distance (relative to whole dataset).

#' @param threshold Quantile cutoff (of dE00 color contrast) for "equivalent" colors, default \code{0.001} based on all whole value/chroma "chips" in the \code{munsell} data set.
#' 
#' @seealso \code{\link{equivalentMunsellChips}}
#' 
#' @references
#' Gaurav Sharma, Wencheng Wu, Edul N. Dalal. (2005). The CIEDE2000 Color-Difference Formula: Implementation Notes, Supplementary Test Data, and Mathematical Observations. COLOR research and application. 30(1):21-30. http://www2.ece.rochester.edu/~gsharma/ciede2000/ciede2000noteCRNA.pdf
#' 
#' Thomas Lin Pedersen, Berendea Nicolae and Romain François (2020). farver: High Performance Colour Space Manipulation. R package version 2.0.3. https://CRAN.R-project.org/package=farver
#'  
#' Dong, C.E., Webb, J.B., Bottrell, M.C., Saginor, I., Lee, B.D. and Stern, L.A. (2020). Strengths, Limitations, and Recommendations for Instrumental Color Measurement in Forensic Soil Characterization. J Forensic Sci, 65: 438-449. https://doi.org/10.1111/1556-4029.14193
#' 
#' @return A list with equal length to the number of rows (chips) in \code{munsell} data, each containing a numeric vector of row indices that are equivalent (CIE2000 distance less than \code{threshold}).
#'
#' @examples
#' \dontrun{
#'
#'   # use default threshold
#'   equivalent_chip_lut <- .makeEquivalentMunsellLUT()
#'
#'   # inspect the first 10 chips, it seems to work!
#'   lapply(equivalent_chip_lut[1:10], function(i) munsell[i,])
#'
#'   # lets see some info on the number of chips per chip
#'   nchipsper <- sapply(equivalent_chip_lut, length)
#'
#'   # top 10 are very high chroma chips with over 70 chips "identical"
#'   nchipsper[order(nchipsper, decreasing = TRUE)[1:10]]
#'
#'   # look at distribution
#'   plot(density(nchipsper))
#'
#'   # median is 5 -- Q: is this true of the range of Munsell colors typically used for soils?
#'   quantile(nchipsper)
#'
#'   # double the default threshold
#'   doubletest <- sapply(.makeEquivalentMunsellLUT(threshold = 0.002), length)
#'   lines(density(doubletest), lty=2)
#'
#'   # apprx. doubles the number of chips per chip in IQR
#'   quantile(doubletest)
#' }
#'
#' @keywords internal
#' @noRd
.makeEquivalentMunsellLUT <- function(threshold = 0.001) {
  munsell <- NULL
  load(system.file("data/munsell.rda", package = "aqp")[1])
  
  # 2024-10-04: added 8.5 and 9.5 value chips 
  # 2022-03-31: updated neutral chips and 2.5 value chips now included
  
  
  ## this version creates too-large of an output, will have to investigate that some other time
  # this produces an 9227x9227matrix ~285MB; just want upper triangle no diagonal
  # # thanks to @dylanbeaudette for pointing out better farver syntax
  # system.time(x <- farver::compare_colour(munsell[, c('L', 'A', 'B')],
  #                                         from_space='lab', white_from = 'D65', method='cie2000'))
  # # # takes about a half minute to run
  # # # user  system elapsed
  # # # 89.07    0.43   89.65 
  #
  # x[lower.tri(x, diag = TRUE)] <- NA # convert 0 to NA to ignore in stats
  # xdat <- x

  # TODO: ... some unholy indexry I cant quite figure out; the stats are right but order is wrong
  
  # indexing strategy:
  # x <- as.dist(t(x))
  # x <- as.matrix(x)

  
  
  # this one that takes 2x as long to built the LUT, and is 2x as big in memory
  # user  system elapsed 
  # 190.73    0.73  194.42
  system.time(
    x <- farver::compare_colour(
      from = munsell[, c('L', 'A', 'B')],
      from_space = 'lab',
      to = munsell[, c('L', 'A', 'B')],
      to_space = 'lab',
      method = 'cie2000',
      white_from = 'D65',
      white_to = 'D65'
    )
  )
  
  xdat <- x
  x[lower.tri(x, diag = TRUE)] <- NA 
  # remove lower triangle for statistics (only count each pair distance 1x)

  # dE00 ~2.158 -- this is close to the perceptible limit of average human color vision with "good" lighting
  # calculate quantiles
  xqtl <- quantile(x, p = threshold, na.rm = TRUE)[1]

  # storing just the chips that fall within the closest e.g. tenth of a percentile: ~1.4MB in memory as list
  # 
  # TODO: consider applications of a matrix (rather than vector) at defined dE00 levels
  # 
  xin1 <- apply(xdat, 1, function(xrow) as.integer(which(xrow <= xqtl)))

  # # ALTERNATIVE THRESHOLDS
  # ## calculate unique threshold for each row (results in way too many "similar" chips)
  # rowthresholds <- apply(x, 1, function(xrow) quantile(xrow, p = threshold, na.rm = TRUE)[1])
  # xin2 <- lapply(1:nrow(x), function(i) as.integer(which(xdat[i,] <= rowthresholds[i])))
  #
  # ## mean of row-wise quantiles - slightly higher than the global threshold
  # rowthreshbar <- mean(apply(x, 1, function(xrow) quantile(xrow, p = threshold, na.rm = TRUE)[1]), na.rm=TRUE)
  # xin3 <- apply(xdat, 1, function(xrow) as.integer(which(xrow <= rowthreshbar)))
  #
  # ## DEBUG plots
  # # inspect the distribution of dE00 and threshold values
  # par(mfrow=c(2,1))
  # plot(density(x, na.rm=TRUE, from=0), xlim=c(0,130),
  #      main = "Between-whole-chip dE00 -- aqp::munsell pair-wise differences",
  #      sub = sprintf("Dotted vertical lines denote row-wise quantiles @ prob=%s", threshold))
  #
  # # # inspect visual cutoff by method 1 and 3
  # abline(v = rowthresholds, lty=3, col=rgb(0,0,0,0.1))# rowwise quantile at threshold
  #
  # # # row-wise quantiles
  # plot(density(rowthresholds, na.rm=TRUE, from=0), lty = 3, xlim = c(0,130),
  #      main = sprintf("Threshold for 'equivalent' whole-chip dE00 @ prob=%s", threshold))
  # abline(v = xqtl[1], lty=1, lwd=2, col="red")
  # abline(v = rowthreshbar, lty=3, lwd=2, col="blue")  #
  # legend("topright", legend = c(sprintf("Density plot of row-wise dE00 quantiles (@ prob=%s)",
  #                                       threshold),
  #                               sprintf("Global quantile (dE00=%s @ prob=%s)",
  #                                       round(xqtl[1], 2), threshold),
  #                               sprintf("Global mean of row-wise quantile (dE00=%s @ prob=%s)",
  #                                       round(rowthreshbar,2), threshold)),
  #        lty = c(3,1,3), lwd = c(1,2,2), col = c("BLACK","RED","BLUE"))

  # # number of chips per chip
  # par(mfrow=c(1,1))
  # plot(density(sapply(xin1, length), bw=1))
  # lines(density(sapply(xin2, length), bw=1), lty=2)
  # lines(density(sapply(xin3, length), bw=1), lty=3)

  # create a nice lookup table to add to aqp
  equivalent_munsell <- xin1
  names(equivalent_munsell) <- sprintf("%s %s/%s", munsell$hue, munsell$value, munsell$chroma)

  # this is only 107kB written to Rda
  save(equivalent_munsell, file="data/equivalent_munsell.rda")

  return(equivalent_munsell)
}

#' Identify "equivalent" (whole number value/chroma) Munsell chips
#'
#' @description 
#' 
#' Uses a pre-calculated lookup list (\code{\link{equivalent_munsell}}) based on pair-wise CIE2000 contrast (\code{dE00}) of LAB color with D65 illuminant for all whole value/chroma "chips" in the \code{aqp::munsell} data set. 
#' 
#' The intention is to identify Munsell chips that may be "functionally equivalent" to some other given whole value/chroma chip elsewhere in the Munsell color space -- as discretized in the \code{aqp::munsell} data table. This basic assumption needs to be validated against your end goal: probably by visual inspection of some or all of the resulting sets. See \code{\link{colorContrast}} and \code{\link{colorContrastPlot}}.
#'
#' "Equivalent" chips table are based (fairly arbitrarily) on the 0.001 probability level of dE00 (default Type 7 \code{quantile}) within the upper triangle of the 8467x8467 contrast matrix. This corresponds to a \code{dE00} contrast threshold of approximately 2.16. 

#' @param hue A character vector containing Munsell hues
#' @param value A numeric vector containing Munsell values (integer only)
#' @param chroma A numeric vector containing Munsell chromas (integer only)
#'
#' @return A named list; Each list element contains a data.frame with one or more rows of "equivalent" Munsell, RGB and LAB color coordinates from \code{munsell} data set.
#' 
#' @seealso \code{\link{colorContrast}} \code{\link{colorContrastPlot}} \code{\link{equivalent_munsell}} 
#' @references
#' 
#' Gaurav Sharma, Wencheng Wu, Edul N. Dalal. (2005). The CIEDE2000 Color-Difference Formula: Implementation Notes, Supplementary Test Data, and Mathematical Observations. COLOR research and application. 30(1):21-30. http://www2.ece.rochester.edu/~gsharma/ciede2000/ciede2000noteCRNA.pdf
#' 
#' Thomas Lin Pedersen, Berendea Nicolae and Romain François (2020). farver: High Performance Colour Space Manipulation. R package version 2.0.3. https://CRAN.R-project.org/package=farver
#'  
#' Dong, C.E., Webb, J.B., Bottrell, M.C., Saginor, I., Lee, B.D. and Stern, L.A. (2020). Strengths, Limitations, and Recommendations for Instrumental Color Measurement in Forensic Soil Characterization. J Forensic Sci, 65: 438-449. https://doi.org/10.1111/1556-4029.14193
#'
#' @export
#'
#' @examples
#'
#' # 7.5YR 4/4 (the one and only)
#' 
#' equivalentMunsellChips("7.5YR", 4, 4)
#' #>
#' #> $`7.5YR 4/4`
#' #>        hue value chroma         r        g         b        L       A       B
#' #> 8330 7.5YR     4      4 0.4923909 0.352334 0.2313328 41.26403 10.8689 23.5914
#' 
#' # 7.5YR 1/1 (two chips are equivalent; 3 row result)
#' 
#' equivalentMunsellChips("7.5YR", 1, 1)
#' #>
#' #> $`7.5YR 1/1`
#' #>        hue value chroma         r         g          b        L        A        B
#' #> 1983  10YR     1      1 0.1345633 0.1087014 0.07606787 10.64787 1.621323 6.847629
#' #> 6189   5YR     1      1 0.1330994 0.1076359 0.09450179 10.63901 2.489012 3.515146
#' #> 8303 7.5YR     1      1 0.1329483 0.1082380 0.08862581 10.64210 2.065514 4.623922
#' 
#' # 10YR 6/8 (two chips are equivalent; 3 row result)
#' 
#' equivalentMunsellChips("10YR", 6, 8)
#' #>
#' #> $`10YR 6/8`
#' #>       hue value chroma         r         g         b        L        A        B
#' #> 2039 10YR     6      7 0.7382230 0.5512957 0.2680260 61.76795 10.50886 44.78574
#' #> 2040 10YR     6      8 0.7519872 0.5472116 0.2157209 61.77496 11.83215 51.15496
#' #> 2041 10YR     6      9 0.7642826 0.5433189 0.1559069 61.78085 13.09599 57.49773
#' 
#' # compare visually a very red color
#' 
#' veryred <- equivalentMunsellChips("10R", 6, 28)[[1]]
#' 
#' par(mar=c(0,0,1,1))
#' 
#' pie(rep(1, nrow(veryred)), col = with(veryred, munsell2rgb(hue, value, chroma)),
#'     label = with(veryred, sprintf("%s %s/%s", hue, value, chroma)))
#' 
#' table(veryred$hue) # 2 hues
#' #> 
#' #>  10R 7.5R 
#' #>    8   17
#' 
#' table(veryred$value) # 2 values
#' #> 
#' #>  5  6 
#' #> 11 14
#' 
#' table(veryred$chroma) # 10 chromas
#' #> 
#' #> 21 22 23 24 25 26 27 28 29 30 
#' #>  1  2  2  3  3  4  3  3  2  2
#'
equivalentMunsellChips <- function(hue = NULL, value = NULL, chroma = NULL) {

  equivalent_munsell <- NULL
  load(system.file("data/equivalent_munsell.rda", package = "aqp")[1])

  munsell <- NULL
  load(system.file("data/munsell.rda", package = "aqp")[1])

  chipdata <- data.frame(.id = 1:pmax(length(hue), length(value), length(chroma)),
                         hue = hue, value = value, chroma = chroma)

  lidx <- lapply(chipdata$.id, function(x) {
    # TODO: handle half values? e.g. 7.5YR 2.5/2?
    which(munsell$hue == chipdata$hue[x] &
            munsell$value == chipdata$value[x] &
            munsell$chroma == chipdata$chroma[x])
  })
  res <- lapply(lidx, function(i) munsell[equivalent_munsell[i][[1]],])
  names(res) <- sprintf("%s %s/%s", hue, value, chroma)
  rownames(res) <- NULL
  return(res)
}

test_that("equivalentMunsellChips works", {
  
  # same "chip page position" different hue page; identify all perceptually equivalent chips
  
  # test 1
  a <- as.list(equivalentMunsellChips("7.5YR", 2, 1)[[1]][, c("hue", "value", "chroma")])
  
  b <- list(
    hue = c("10YR", "5YR", "7.5YR"),
    value = c(2, 2, 2),
    chroma = c(1, 1, 1)
  )
  
  expect_equal(a, b)
  
  
  # test 2
  a <- as.list(equivalentMunsellChips("5YR", 2, 1)[[1]][, c("hue", "value", "chroma")])
  
  b <- list(
    hue = c("10R", "2.5YR", "5YR", "7.5YR"),
    value = c(2, 2, 2, 2),
    chroma = c(1, 1, 1, 1)
  )
  
  expect_equal(a, b)
  
  # test 3
  a <- as.list(equivalentMunsellChips("10YR", 2, 1)[[1]][, c("hue", "value", "chroma")])
  
  b <- list(
    hue = c("10YR", "2.5Y", "7.5YR"),
    value = c(2, 2, 2),
    chroma = c(1, 1, 1)
  )
  
  expect_equal(a, b)
  
})

##
## document data here
## don't forget: @usage data(XXX)


#' @title Representative O Horizon Colors
#' @description Representative O horizon colors for the most common types of organic soil material, at dry and moist states. Colors were derived from an analysis of soil morphologic data (5245 horizons) within the USDA-NRCS National Soil Information System. Representative colors are the L1 median colors (by [colorVariation()]) within groups generalized to "Oi", "Oe", "Oa", and all "other". These estimates may be useful place-holder values for soil color in collections where O horizon color was not recorded.
#' @keywords datasets
#' @usage data(Ohz.colors)
#' @format A `data.frame` with the following elements:
#' 
#'  * state: soil moisture state, "dry" or "moist"
#'  * genhz: horizon designation generalized into one of "Oi", "Oe", "Oa", or "other"
#'  * L1.munsell: representative color in Munsell notation for each moisture state / generalized horizon designation
#'  * group.dE00: weighted-mean CIE dE 2000 color contrast metric between all soil colors in this group and representative color
#' 
"Ohz.colors"




#' @title Lookup tables for sand, silt, clay, texture class, and textural modifiers.
#' 
#' @description A list that contains a snapshot of the values generated using the logic from
#' the particle size estimator calculation in NASIS, the average values per
#' texture class, and average rock fragment values by textural modifier.
#' 
#' @usage data(soiltexture)
#' @format A list with 3 data frames. The first named values which contains
#' values for sand, silt and clay by texture class. The second with average
#' values for sand, silt and clay per texture class. The third has fragvoltot
#' low, rv and high values for texmod.  \describe{ \item{list("clay")}{clay
#' percentage of the fine earth fraction, a integer vector}
#' \item{list("sand")}{sand percentage of the fine earth fraction, a integer
#' vector} \item{list("silt")}{silt percentage of the fine earth fraction, a
#' integer vector} \item{list("texcl")}{texture class, a character vector}
#' \item{list("texmod")}{textural modifiers, a character vector} }
#' @keywords datasets
"soiltexture"

#' @title Example Output from soilDB::fetchOSD()
#' @description An example `SoilProfileCollection` object created by `soilDB::fetchOSD()`, derived from the Cecil, Appling, and Bonneau Official Series Descriptions.
#' @keywords datasets
#' @usage data(osd)
#' @format A `SoilProfileCollection`
"osd"



#' @title Example SoilProfileCollection with Overlapping Horizons
#' @description A `SoilProfileCollection` with overlapping horizons, derived from a Dynamic Soil Properties project.
#' @keywords datasets
#' @usage data(SPC.with.overlap)
#' @format A `SoilProfileCollection`
"SPC.with.overlap"


#' @title Example Data from Wilson et al. 2022
#' @description An example `SoilProfileCollection`, derived from Wilson et al., 2022. Select data extracted from Appendix tables.
#' @keywords datasets
#' @usage data(wilson2022)
#' @format A `SoilProfileCollection` with the following elements. Total elemental analysis by lithium borate fusion.
#' 
#' Horizon level attributes:
#'   * name: horizon designation
#'   * Al2O3: total Al (wt %)
#'   * Fe2O3: total Fe (wt %)
#'   * K2O: total K (wt %)
#'   * MgO: total Mg (wt %)
#'   * Na2O: total Na (wt %)
#'   * P2O5: total P (wt %)
#'   * SiO2: total Si (wt %)
#'   * CaO: total Ca(wt %)
#'   * Alo: Oxalate Extractable Al (g/kg)
#'   * Feo: Oxalate Extractable Fe (g/kg)
#'   * Fed: Dithionite extractable Fe (g/kg)
#'   * Fed_minus_Feo: Crystalline Fe (hydr)oxides (g/kg)
#'   * CIA: Chemical Index of Alteration, see original paper (ratio, unitless)
#'   * Fed_div_Fet: (ratio, unitless)
#'   * Fet: Total Fe from lithium borate fusion (g/kg)
#'   * resin_Pi: Hedley phosphorus fractions (mg/kg)
#'   * NaHCO3_Pi: Hedley phosphorus fractions (mg/kg) 
#'   * labile_Pi: Sum of resin Pi and NaHCO3 Pi (mg/kg)
#'   * NaCO3_Po: Hedley phosphorus fractions (mg/kg)
#'   * NaOH_Pi: Hedley phosphorus fractions (mg/kg)
#'   * NaOH_Po: Hedley phosphorus fractions (mg/kg)
#'   * Ca_Pi: Hedley phosphorus fractions (mg/kg)
#'   * organic_P: Sum of NaHCO3 and NaOH Po fractions (mg/kg)
#'   * total_P: Total P from lithium borate fusion (mg/kg)
#'   * occluded_P: Difference between total P and sum of Hedley P fractions (mg/kg)
#'   * top: horizon top depth (cm)
#'   * bottom: horizon bottom depth (cm)
#'   * pedonID: pedon ID (serial number)
#'   
#' Site level attributes:
#'   * pm: parent material group
#'   * biome: biome
#'
#' @references 
#' Stewart G. Wilson, Randy A. Dahlgren, Andrew J. Margenot, Craig Rasmussen, Anthony T. O'Geen. 2022. Expanding the Paradigm: The influence of climate and lithology on soil phosphorus, Geoderma: 421. \doi{10.1016/j.geoderma.2022.115809}
#' 
#' 
#' 
#' @examples 
#' 
#' data(wilson2022)
#' 
#' groupedProfilePlot(wilson2022, groups = 'pm', 
#' group.name.offset = -15, label = 'biome', 
#' name.style = 'center-center', color = 'CIA', 
#' cex.names = 0.66, cex.id = 0.66, width = 0.3)
#' 
#' groupedProfilePlot(wilson2022, groups = 'biome', 
#' group.name.offset = -15, label = 'pm', 
#' name.style = 'center-center', color = 'Fet', 
#' cex.names = 0.66, cex.id = 0.66, width = 0.3)
#' 
"wilson2022"






#' Soil Profile Data Example 1
#'
#' Soil profile data from Pinnacles National Monument, CA.
#'
#'
#' @name sp1
#' @docType data
#' @format A data frame with 60 observations on the following 21 variables.
#' \describe{ \item{group}{a numeric vector} \item{id}{a
#' character vector} \item{top}{a numeric vector}
#' \item{bottom}{a numeric vector} \item{bound_distinct}{a
#' character vector} \item{bound_topography}{a character vector}
#' \item{name}{a character vector} \item{texture}{a character
#' vector} \item{prop}{a numeric vector}
#' \item{structure_grade}{a character vector}
#' \item{structure_size}{a character vector}
#' \item{structure_type}{a character vector}
#' \item{stickiness}{a character vector} \item{plasticity}{a
#' character vector} \item{field_ph}{a numeric vector}
#' \item{hue}{a character vector} \item{value}{a numeric
#' vector} \item{chroma}{a numeric vector} }
#' @references \url{https://casoilresource.lawr.ucdavis.edu/}
#' @keywords datasets
#' @examples
#'
#' data(sp1)
#' # convert colors from Munsell to hex-encoded RGB
#' sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))
#'
#' # promote to SoilProfileCollection
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' # re-sample each profile into 1 cm (thick) depth slices
#' # for the variables 'prop', 'name', 'soil_color'
#' # result is a SoilProfileCollection object
#' s <- dice(sp1, 0:25 ~ prop + name + soil_color)
#'
#' # plot, note slices
#' plot(s)
#'
#'
#' # aggregate all profiles along 1 cm depth slices,
#' # using data from column 'prop'
#' s1 <- slab(sp1, fm= ~ prop)
#'
#' # check median & IQR
#' library(lattice)
#' xyplot(top ~ p.q50 + p.q25 + p.q75,
#' data=s1, type='S', horizontal=TRUE, col=1, lty=c(1,2,2),
#' panel=panel.superpose, ylim=c(110,-5), asp=2)
#'
#'
NULL

#' Honcut Creek Soil Profile Data
#'
#' A collection of 18 soil profiles, consisting of select soil morphologic
#' attributes, associated with a stratigraphic study conducted near Honcut
#' Creek, California.
#'
#'
#' @name sp2
#' @docType data
#' @format A data frame with 154 observations on the following 21 variables.
#' \describe{ \item{id}{profile id} \item{surface}{dated
#' surface} \item{top}{horizon top in cm} \item{bottom}{horizon
#' bottom in cm} \item{bound_distinct}{horizon lower boundary
#' distinctness class} \item{bound_topography}{horizon lower boundary
#' topography class} \item{name}{horizon name}
#' \item{texture}{USDA soil texture class}
#' \item{prop}{field-estimated clay content}
#' \item{structure_grade}{soil structure grade}
#' \item{structure_size}{soil structure size}
#' \item{structure_type}{soil structure type}
#' \item{stickiness}{stickiness} \item{plasticity}{plasticity}
#' \item{field_ph}{field-measured pH} \item{hue}{Munsell hue}
#' \item{value}{Munsell value} \item{chroma}{Munsell chroma}
#' \item{r}{RGB red component} \item{g}{RGB green component}
#' \item{b}{RGB blue component} \item{soil_color}{R-friendly
#' encoding of soil color} }
#' @author Dylan E. Beaudette
#' @references \url{https://casoilresource.lawr.ucdavis.edu/}
#' @source Busacca, Alan J.; Singer, Michael J.; Verosub, Kenneth L. 1989. Late
#' Cenozoic stratigraphy of the Feather and Yuba rivers area, California, with
#' a section on soil development in mixed alluvium at Honcut Creek. USGS
#' Bulletin 1590-G.
#' @keywords datasets
#' @examples
#'
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#'
#' data(sp2)
#'
#' # convert into SoilProfileCollection object
#' depths(sp2) <- id ~ top + bottom
#'
#' # transfer site-level data
#' site(sp2) <- ~ surface
#'
#' # generate a new plotting order, based on the dated surface each soil was described on
#' p.order <- order(sp2$surface)
#'
#' # plot
#' par(mar=c(1,0,3,0))
#' plot(sp2, plot.order=p.order)
#'
#' # setup multi-figure output
#' par(mfrow=c(2,1), mar=c(0,0,1,0))
#'
#' # truncate plot to 200 cm depth
#' plot(sp2, plot.order=p.order, max.depth=200)
#' abline(h=200, lty=2, lwd=2)
#'
#' # compute numerical distances between profiles
#' # based on select horizon-level properties, to a depth of 200 cm
#' d <- NCSP(sp2, vars=c('prop','field_ph','hue'), maxDepth = 100, k = 0)
#'
#' # plot dendrogram with ape package:
#' if(require(ape) & require(cluster)) {
#' h <- diana(d)
#' p <- as.phylo(as.hclust(h))
#' plot(p, cex=0.75, label.offset=0.01, font=1, direct='down', srt=90, adj=0.5, y.lim=c(-0.125, 0.5))
#'
#' # add in the dated surface type via color
#' tiplabels(col=as.numeric(sp2$surface), pch=15)
#'
#' # based on distance matrix values, YMMV
#' legend('topleft', legend=levels(sp2$surface), col=1:6, pch=15, bty='n', bg='white', cex=0.75)
#' }
#'
#'
NULL

#' Soil Profile Data Example 3
#'
#' Soil samples from 10 soil profiles, taken from the Sierra Foothill Region of
#' California.
#'
#' These data were collected to support research funded by the Kearney
#' Foundation of Soil Science.
#'
#' @name sp3
#' @docType data
#' @format A data frame with 46 observations on the following 15 variables.
#' \describe{ \item{id}{soil id} \item{top}{horizon upper
#' boundary (cm)} \item{bottom}{horizon lower boundary (cm)}
#' \item{clay}{clay content} \item{cec}{CEC by amonium acetate
#' at pH 7} \item{ph}{pH in 1:1 water-soil mixture}
#' \item{tc}{total carbon percent} \item{hue}{Munsell hue
#' (dry)} \item{value}{Munsell value (dry)}
#' \item{chroma}{Munsell chroma (dry)} \item{mid}{horizon
#' midpoint (cm)} \item{ln_tc}{natural log of total carbon percent}
#' \item{L}{color: l-coordinate, CIE-LAB colorspace (dry)}
#' \item{A}{color: a-coordinate, CIE-LAB colorspace (dry)}
#' \item{B}{color: b-coordinate, CIE-LAB colorspace (dry)}
#' \item{name}{horizon name} \item{soil_color}{horizon color} }
#' @keywords datasets
#' @references \url{https://casoilresource.lawr.ucdavis.edu/}
#' @examples
#'
#' ## this example investigates the concept of a "median profile"
#'
#' # required packages
#' if (require(ape) & require(cluster)) {
#'   data(sp3)
#'
#'   # generate a RGB version of soil colors
#'   # and convert to HSV for aggregation
#'   sp3$h <- NA
#'   sp3$s <- NA
#'   sp3$v <- NA
#'   sp3.rgb <- with(sp3, munsell2rgb(hue, value, chroma, return_triplets = TRUE))
#'
#'   sp3[, c('h', 's', 'v')] <- t(with(sp3.rgb, rgb2hsv(r, g, b, maxColorValue = 1)))
#'
#'   # promote to SoilProfileCollection
#'   depths(sp3) <- id ~ top + bottom
#'
#'   # aggregate across entire collection
#'   a <- slab(sp3, fm = ~ clay + cec + ph + h + s + v, slab.structure = 10)
#'
#'   # check
#'   str(a)
#'
#'   # convert back to wide format
#'   library(data.table)
#'
#'   a.wide.q25 <- dcast(as.data.table(a), top + bottom ~ variable, value.var = c('p.q25'))
#'   a.wide.q50 <- dcast(as.data.table(a), top + bottom ~ variable, value.var = c('p.q50'))
#'   a.wide.q75 <- dcast(as.data.table(a), top + bottom ~ variable, value.var = c('p.q75'))
#'
#'   # add a new id for the 25th, 50th, and 75th percentile pedons
#'   a.wide.q25$id <- 'Q25'
#'   a.wide.q50$id <- 'Q50'
#'   a.wide.q75$id <- 'Q75'
#'
#'   # combine original data with "mean profile"
#'   vars <- c('top', 'bottom', 'id', 'clay', 'cec', 'ph', 'h', 's', 'v')
#'   # make data.frame version of sp3
#'   sp3.df <- as(sp3, 'data.frame')
#' 
#'   sp3.grouped <- as.data.frame(rbind(as.data.table(horizons(sp3))[, .SD, .SDcol = vars], 
#'                                      a.wide.q25[, .SD, .SDcol = vars],
#'                                      a.wide.q50[, .SD, .SDcol = vars], 
#'                                      a.wide.q75[, .SD, .SDcol = vars]))
#'                                      
#'   # re-constitute the soil color from HSV triplets
#'   # convert HSV back to standard R colors
#'   sp3.grouped$soil_color <- with(sp3.grouped, hsv(h, s, v))
#'
#'   # give each horizon a name
#'   sp3.grouped$name <- paste(
#'     round(sp3.grouped$clay),
#'     '/' ,
#'     round(sp3.grouped$cec),
#'     '/',
#'     round(sp3.grouped$ph, 1)
#'   )
#'
#' 
#' # first promote to SoilProfileCollection
#' depths(sp3.grouped) <- id ~ top + bottom
#' 
#' plot(sp3.grouped)
#' 
#' ## perform comparison, and convert to phylo class object
#' ## D is rescaled to [0,]
#' d <- NCSP(
#'   sp3.grouped,
#'   vars = c('clay', 'cec', 'ph'),
#'   maxDepth = 100,
#'   k = 0.01
#' )
#' 
#' h <- agnes(d, method = 'ward')
#' p <- ladderize(as.phylo(as.hclust(h)))
#' 
#' # look at distance plot-- just the median profile
#' plot_distance_graph(d, 12)
#' 
#' # similarity relative to median profile (profile #12)
#' round(1 - (as.matrix(d)[12, ] / max(as.matrix(d)[12, ])), 2)
#' 
#' ## make dendrogram + soil profiles
#' 
#' # setup plot: note that D has a scale of [0,1]
#' par(mar = c(1, 1, 1, 1))
#' p.plot <- plot(p,
#'                cex = 0.8,
#'                label.offset = 3,
#'                direction = 'up',
#'                y.lim = c(200, 0),
#'                x.lim = c(1.25, length(sp3.grouped) + 1),
#'                show.tip.label = FALSE)
#' 
#' # get the last plot geometry
#' lastPP <- get("last_plot.phylo", envir = ape::.PlotPhyloEnv)
#' 
#' # the original labels, and new (indexed) order of pedons in dendrogram
#' d.labels <- attr(d, 'Labels')
#' 
#' new_order <- sapply(1:lastPP$Ntip,
#'                     function(i)
#'                       which(as.integer(lastPP$xx[1:lastPP$Ntip]) == i))
#' 
#' # plot the profiles, in the ordering defined by the dendrogram
#' # with a couple fudge factors to make them fit
#' plotSPC(
#'   sp3.grouped,
#'   color = "soil_color",
#'   plot.order = new_order,
#'   y.offset = max(lastPP$yy) + 10,
#'   width = 0.1,
#'   cex.names = 0.5,
#'   add = TRUE
#' )
#' }
NULL

#' Soil Chemical Data from Serpentinitic Soils of California
#'
#' Soil Chemical Data from Serpentinitic Soils of California
#'
#' Selected soil physical and chemical data from (McGahan et al., 2009).
#'
#' @name sp4
#' @docType data
#' @format A data frame with 30 observations on the following 13 variables.
#' \describe{ \item{id}{site name} \item{name}{horizon
#' designation} \item{top}{horizon top boundary in cm}
#' \item{bottom}{horizon bottom boundary in cm}
#' \item{K}{exchangeable K in c mol/kg} \item{Mg}{exchangeable
#' Mg in cmol/kg} \item{Ca}{exchangeable Ca in cmol/kg}
#' \item{CEC_7}{cation exchange capacity (NH4OAc at pH 7)}
#' \item{ex_Ca_to_Mg}{extractable Ca:Mg ratio} \item{sand}{sand
#' content by weight percentage} \item{silt}{silt content by weight
#' percentage} \item{clay}{clay content by weight percentage}
#' \item{CF}{>2mm fraction by volume percentage} }
#' @references McGahan, D.G., Southard, R.J, Claassen, V.P. 2009.
#' Plant-Available Calcium Varies Widely in Soils on Serpentinite Landscapes.
#' Soil Sci. Soc. Am. J. 73: 2087-2095.
#' @source https://www.soils.org/publications/sssaj/articles/73/6/2087
#' @keywords datasets
#' @examples
#'
#' # load sample data set, a simple data.frame object with horizon-level data from 10 profiles
#' library(aqp)
#' data(sp4)
#' str(sp4)
#' sp4$idbak <- sp4$id
#'
#' # upgrade to SoilProfileCollection
#' # 'id' is the name of the column containing the profile ID
#' # 'top' is the name of the column containing horizon upper boundaries
#' # 'bottom' is the name of the column containing horizon lower boundaries
#' depths(sp4) <- id ~ top + bottom
#'
#' # check it out
#' class(sp4) # class name
#' str(sp4) # internal structure
#'
#' # check integrity of site:horizon linkage
#' spc_in_sync(sp4)
#'
#' # check horizon depth logic
#' checkHzDepthLogic(sp4)
#'
#' # inspect object properties
#' idname(sp4) # self-explanitory
#' horizonDepths(sp4) # self-explanitory
#'
#' # you can change these:
#' depth_units(sp4) # defaults to 'cm'
#' metadata(sp4) # not much to start with
#'
#' # alter the depth unit metadata
#' depth_units(sp4) <- 'inches' # units are really 'cm'
#'
#' # more generic interface for adjusting metadata
#'
#' # add attributes to metadata list
#' metadata(sp4)$describer <- 'DGM'
#' metadata(sp4)$date <- as.Date('2009-01-01')
#' metadata(sp4)$citation <- 'McGahan, D.G., Southard, R.J, Claassen, V.P.
#' 2009. Plant-Available Calcium Varies Widely in Soils
#' on Serpentinite Landscapes. Soil Sci. Soc. Am. J. 73: 2087-2095.'
#'
#' depth_units(sp4) <- 'cm' # fix depth units, back to 'cm'
#'
#' # further inspection with common function overloads
#' length(sp4) # number of profiles in the collection
#' nrow(sp4) # number of horizons in the collection
#' names(sp4) # column names
#' min(sp4) # shallowest profile depth in collection
#' max(sp4) # deepest profile depth in collection
#'
#' # extraction of soil profile components
#' profile_id(sp4) # vector of profile IDs
#' horizons(sp4) # horizon data
#'
#' # extraction of specific horizon attributes
#' sp4$clay # vector of clay content
#'
#' # subsetting SoilProfileCollection objects
#' sp4[1, ] # first profile in the collection
#' sp4[, 1] # first horizon from each profile
#'
#' # basic plot method, highly customizable: see manual page ?plotSPC
#' plot(sp4)
#' # inspect plotting area, very simple to overlay graphical elements
#' abline(v=1:length(sp4), lty=3, col='blue')
#' # profiles are centered at integers, from 1 to length(obj)
#' axis(1, line=-1.5, at=1:10, cex.axis=0.75, font=4, col='blue', lwd=2)
#' # y-axis is based on profile depths
#' axis(2, line=-1, at=pretty(1:max(sp4)), cex.axis=0.75, font=4, las=1, col='blue', lwd=2)
#'
#'
#' # symbolize soil properties via color
#' par(mar=c(0,0,4,0))
#' plot(sp4, color='clay')
#' plot(sp4, color='CF')
#'
#' # apply a function to each profile, returning a single value per profile,
#' # in the same order as profile_id(sp4)
#' soil.depths <- profileApply(sp4, max) # recall that max() gives the depth of a soil profile
#'
#' # check that the order is correct
#' all.equal(names(soil.depths), profile_id(sp4))
#'
#' # a vector of values that is the same length as the number of profiles
#' # can be stored into site-level data
#' sp4$depth <- soil.depths
#' # check: looks good
#' max(sp4[1, ]) == sp4$depth[1]
#'
#' # extract site-level data
#' site(sp4) # as a data.frame
#' sp4$depth # specific columns as a vector
#'
#' # use site-level data to alter plotting order
#' new.order <- order(sp4$depth) # the result is an index of rank
#' par(mar=c(0,0,0,0))
#' plot(sp4, plot.order=new.order)
#'
#' # deconstruct SoilProfileCollection into a data.frame, with horizon+site data
#' as(sp4, 'data.frame')
#'
NULL

#' Sample Soil Database #5
#'
#' 296 Soil Profiles from the La Rochelle region of France (F. Carre and
#' Girard, 2002)
#'
#' These data are c/o F. Carre (Florence.CARRE@ineris.fr).
#'
#' @name sp5
#' @docType data
#' @format `SoilProfileCollection` object
#' @references F. Carre, M.C. Girard. 2002. Quantitative mapping of soil types
#' based on regression kriging of taxonomic distances with landform and land
#' cover attributes. Geoderma. 110: 241--263.
#' @source 296 Soil Profiles from the La Rochelle region of France (F. Carre
#' and Girard, 2002). These data can be found on the OSACA project page
#' (\url{http://eusoils.jrc.ec.europa.eu/projects/OSACA/}).
#' @keywords datasets
#' @examples
#'
#' \dontrun{
#' library(scales)
#' data(sp5)
#' par(mar=c(1,1,1,1))
#' # plot a random sampling of profiles
#' s <- sample(1:length(sp5), size=25)
#' plot(sp5[s, ], divide.hz=FALSE)
#'
#' # plot the first 100 profiles, as 4 rows of 25, hard-coding the max depth
#' layout(matrix(c(1,2,3,4), ncol=1), height=c(0.25,0.25,0.25,0.25))
#' plot(sp5[1:25, ], max.depth=300)
#' plot(sp5[26:50, ], max.depth=300)
#' plot(sp5[51:75, ], max.depth=300)
#' plot(sp5[76:100, ], max.depth=300)
#'
#'
#' # 4x1 matrix of plotting areas
#' layout(matrix(c(1,2,3,4), ncol=1), height=c(0.25,0.25,0.25,0.25))
#'
#' # plot profiles, with points added to the mid-points of randomly selected horizons
#' sub <- sp5[1:25, ]
#' plot(sub, max.depth=300) ; mtext('Set 1', 2, line=-0.5, font=2)
#' y.p <- profileApply(sub, function(x) {
#'   s <- sample(1:nrow(x), 1)
#'   h <- horizons(x); with(h[s,], (top+bottom)/2)
#'   })
#' points(1:25, y.p, bg='white', pch=21)
#'
#' # plot profiles, with arrows pointing to profile bottoms
#' sub <- sp5[26:50, ]
#' plot(sub, max.depth=300); mtext('Set 2', 2, line=-0.5, font=2)
#' y.a <- profileApply(sub, function(x) max(x))
#' arrows(1:25, y.a-50, 1:25, y.a, len=0.1, col='white')
#'
#' # plot profiles, with points connected by lines: ideally reflecting some kind of measured data
#' sub <- sp5[51:75, ]
#' plot(sub, max.depth=300); mtext('Set 3', 2, line=-0.5, font=2)
#' y.p <- 20*(sin(1:25) + 2*cos(1:25) + 5)
#' points(1:25, y.p, bg='white', pch=21)
#' lines(1:25, y.p, lty=2)
#'
#' # plot profiles, with polygons connecting horizons with max clay content (+/-) 10 cm
#' sub <- sp5[76:100, ]
#' y.clay.max <- profileApply(sub, function(x) {
#'   i <- which.max(x$clay)
#'   h <- horizons(x)
#'   with(h[i, ], (top+bottom)/2)
#'   } )
#'
#' plot(sub, max.depth=300); mtext('Set 4', 2, line=-0.5, font=2)
#' polygon(c(1:25, 25:1), c(y.clay.max-10, rev(y.clay.max+10)),
#' border='black', col=rgb(0,0,0.8, alpha=0.25))
#' points(1:25, y.clay.max, pch=21, bg='white')
#'
#' # close plot
#' dev.off()
#'
#'
#' # plotting parameters
#' yo <- 100 # y-offset
#' sf <- 0.65 # scaling factor
#' # plot profile sketches
#' par(mar=c(0,0,0,0))
#' plot(sp5[1:25, ], max.depth=300, y.offset=yo, scaling.factor=sf)
#' # optionally add describe plotting area above profiles with lines
#' # abline(h=c(0,90,100, (300*sf)+yo), lty=2)
#' # simulate an environmental variable associated with profiles (elevation, etc.)
#' r <- vector(mode='numeric', length=25)
#' r[1] <- -50 ; for(i in 2:25) {r[i] <- r[i-1] + rnorm(mean=-1, sd=25, n=1)}
#' # rescale
#' r <- rescale(r, to=c(80, 0))
#' # illustrate gradient with points/lines/arrows
#' lines(1:25, r)
#' points(1:25, r, pch=16)
#' arrows(1:25, r, 1:25, 95, len=0.1)
#' # add scale for simulated gradient
#' axis(2, at=pretty(0:80), labels=rev(pretty(0:80)), line=-1, cex.axis=0.75, las=2)
#' # depict a secondary environmental gradient with polygons (water table depth, etc.)
#' polygon(c(1:25, 25:1), c((100-r)+150, rep((300*sf)+yo, times=25)),
#' border='black', col=rgb(0,0,0.8, alpha=0.25))
#'
#' }
#'
#'
NULL

#' Soil Physical and Chemical Data from Manganiferous Soils
#'
#' Soil Physical and Chemical Data from Manganiferous Soils (Bourgault and
#' Rabenhorst, 2011)
#'
#' Selected soil physical and chemical data from (Bourgault and Rabenhorst,
#' 2011).
#'
#' @name sp6
#' @docType data
#' @format A data frame with 30 observations on the following 13 variables.
#' \describe{ \item{id}{pedon name} \item{name}{horizon
#' designation} \item{top}{horizon top boundary in cm}
#' \item{bottom}{horizon bottom boundary in cm}
#' \item{color}{moist soil color in Munsell notation}
#' \item{texture}{USDA soil texture class} \item{sand}{sand
#' content by weight percentage} \item{silt}{silt content by weight
#' percentage} \item{clay}{clay content by weight percentage}
#' \item{Fe}{DCB-extracted Fe in g/kg (see citation)}
#' \item{Mn}{DCB-extracted Mn in g/kg (see citation)}
#' \item{C}{total organic carbon as g/kg} \item{pH}{measured in
#' 1:1 H20 slurry} \item{Db}{bulk density (g/cc), clod method} }
#' @references Rebecca R. Bourgault, Martin C. Rabenhorst. 2011.  Genesis and
#' characterization of manganiferous soils in the Eastern Piedmont, USA.
#' Geoderma. 165:84-94.
#' @source http://www.sciencedirect.com/science/article/pii/S0016706111001972
#' @keywords datasets
#' @examples
#'
#'   # setup environment
#'   library(aqp)
#'   data(sp6)
#'
#'   # init SPC
#'   depths(sp6) <- id ~ top + bottom
#'   # convert non-standard Munsell colors
#'   sp6$soil_color <- getClosestMunsellChip(sp6$color)
#'
#'   # profile sketches
#'   par(mar=c(0,0,3,0))
#'   plot(sp6, color='soil_color')
#'   plot(sp6, color='Mn')
#'   plot(sp6, color='Fe')
#'   plot(sp6, color='pH')
#'   plot(sp6, color='texture')
#'
#'
NULL


#' Soil Physical and Chemical Data Related to Studies in the Sierra Nevada
#' Mountains, CA, USA.
#'
#' Soil physical and chemical data associated with two bio-climatic sequences
#' (granitic and andesitic parent material) from the western flank of the
#' Sierra Nevada mountains.
#'
#' These data were assembled from Dahlgren et al. (1997) and Rasmussen et al.
#' (2007), with permission granted by lead authors, by D.E. Beaudette.
#'
#' @name sierraTransect
#' @docType data
#' @usage data(sierraTransect)
#'
#' @references R.A. Dahlgren, J.L. Boettinger, G.L. Huntington, R.G. Amundson.
#' Soil development along an elevational transect in the western Sierra Nevada,
#' California, Geoderma, Volume 78, Issues 3–4, 1997, Pages 207-236.
#'
#' Rasmussen, C., Matsuyama, N., Dahlgren, R.A., Southard, R.J. and Brauer, N.
#' (2007), Soil Genesis and Mineral Transformation Across an Environmental
#' Gradient on Andesitic Lahar. Soil Sci. Soc. Am. J., 71: 225-237.
#' @source Original manuscripts and personal communication with authors.
#' @keywords datasets
#' @examples
#'
#' data(sierraTransect)
#'
#' # tighter margins
#' op <- par(mar=c(0,0,0,0))
#'
#' # quick sketch
#' plotSPC(sierraTransect, name.style = 'center-center', width=0.3)
#'
#' # split by transect
#' par(mar=c(0,0,1,1))
#' groupedProfilePlot(
#' sierraTransect, groups='transect',
#' group.name.offset = -15, width=0.3,
#' name.style='center-center'
#' )
#'
#' # thematic
#' groupedProfilePlot(
#' sierraTransect, groups='transect',
#' group.name.offset = -15, width=0.3,
#' name.style='center-center', color='Fe_o_to_Fe_d'
#' )
#'
#' # horizon boundary viz
#' sierraTransect$hzd <- hzDistinctnessCodeToOffset(substr(sierraTransect$hz_boundary, 0, 1))
#'
#' groupedProfilePlot(
#' sierraTransect, groups='transect', group.name.offset = -15,
#' width=0.3, name.style='center-center', color='Fe_o_to_Fe_d',
#' hz.distinctness.offset='hzd')
#'
#'
#' # split transects
#' g <- subset(sierraTransect, transect == 'Granite')
#' a <- subset(sierraTransect, transect == 'Andesite')
#'
#' g.order <- order(g$elev)
#' a.order <- order(a$elev)
#'
#' # order (left -> right) by elevation
#' par(mar=c(2,0,0,2), mfrow=c(2,1))
#' plot(g, width=0.3, name.style='center-center', cex.names=0.75, plot.order=g.order)
#' axis(1, at=1:length(g), labels=g$elev[g.order], line=-1.5)
#'
#' plot(a, width=0.3, name.style='center-center', cex.names=0.75, plot.order=a.order)
#' axis(1, at=1:length(a), labels=a$elev[a.order], line=-1.5)
#'
#'
#' par(op)
#'
"sierraTransect"

#' Soil Data from the Central Sierra Nevada Region of California
#'
#' Site and laboratory data from soils sampled in the central Sierra Nevada
#' Region of California.
#'
#' These data were extracted from the NSSL database. `ca630` is a list composed
#' of site and lab data, each stored as `data.frame` objects. These data are modeled by a
#' 1:many (site:lab) relation, with the `pedon_id` acting as the primary key in
#' the `site` table and as the foreign key in the `lab` table.
#'
#' @name ca630
#' @docType data
#' @usage data(ca630)
#' @format List containing:
#'
#' $site : A data frame containing site information.  \describe{
#' \item{user_site_id}{national user site id} \item{mlra}{the
#' MLRA} \item{county}{the county} \item{ssa}{soil survey area}
#' \item{lon}{longitude, WGS84} \item{lat}{latitude, WGS84}
#' \item{pedon_key}{national soil profile id}
#' \item{user_pedon_id}{local soil profile id}
#' \item{cntrl_depth_to_top}{control section top depth (cm)}
#' \item{cntrl_depth_to_bot}{control section bottom depth (cm)}
#' \item{sampled_taxon_name}{soil series name} }
#'
#' $lab : A data frame containing horizon information.  \describe{
#' \item{pedon_key}{national soil profile id}
#' \item{layer_key}{national horizon id}
#' \item{layer_sequence}{horizon sequence number}
#' \item{hzn_top}{horizon top (cm)} \item{hzn_bot}{horizon
#' bottom (cm)} \item{hzn_desgn}{horizon name}
#' \item{texture_description}{USDA soil texture}
#' \item{nh4_sum_bases}{sum of bases extracted by ammonium acetate (pH
#' 7)} \item{ex_acid}{exchangeable acidity \[method ?]}
#' \item{CEC8.2}{cation exchange capacity by sum of cations method (pH
#' 8.2)} \item{CEC7}{cation exchange capacity by ammonium acetate (pH
#' 7)} \item{bs_8.2}{base saturation by sum of cations method (pH 8.2)}
#' \item{bs_7}{base saturation by ammonium acetate (pH 7)} }
#' @note These data are out of date. Pending some new data + documentation. Use
#' with caution
#' @source \url{https://ncsslabdatamart.sc.egov.usda.gov/}
#' @keywords datasets
#' @examples
#'
#' \dontrun{
#' library(tactile)
#' library(lattice)
#' library(Hmisc)
#' library(sp)
#'
#' # check the data out:
#' data(ca630)
#' str(ca630)
#'
#' # note that pedon_key is the link between the two tables
#'
#' # make a copy of the horizon data
#' ca <- ca630$lab
#'
#' # promote to a SoilProfileCollection class object
#' depths(ca) <- pedon_key ~ hzn_top + hzn_bot
#'
#' # add site data, based on pedon_key
#' site(ca) <- ca630$site
#'
#' # ID data missing coordinates: '|' is a logical OR
#' (missing.coords.idx <- which(is.na(ca$lat) | is.na(ca$lon)))
#'
#' # remove missing coordinates by safely subsetting
#' if(length(missing.coords.idx) > 0)
#' 	ca <- ca[-missing.coords.idx, ]
#'
#' # register spatial data
#' initSpatial(ca) <- ~ lon + lat
#'
#' # assign a coordinate reference system
#' prj(ca) <- 'EPSG:4269'
#'
#' # check the result
#' print(ca)
#'
#' # aggregate %BS 7 for all profiles into 1 cm slices
#' a <- slab(ca, fm= ~ bs_7)
#'
#' # plot median & IQR by 1 cm slice
#' xyplot(
#' top ~ p.q50, 
#' data = a, 
#' lower=a$p.q25, 
#' upper=a$p.q75,
#' alpha=0.5,
#' ylim=c(160,-5), 
#' scales = list(alternating = 1, y = list(tick.num = 7)),
#' panel = panel.depth_function, 
#' prepanel = prepanel.depth_function,
#' ylab='Depth (cm)', xlab='Base Saturation at pH 7',
#' par.settings = tactile.theme(superpose.line = list(col = 'black', lwd = 2))
#' )
#'
#' # aggregate %BS at pH 8.2 for all profiles by MLRA, along 1 cm slices
#' # note that mlra is stored in @site
#' a <- slab(ca, mlra ~ bs_8.2)
#'
#' # keep only MLRA 18 and 22
#' a <- subset(a, subset=mlra %in% c('18', '22'))
#'
#' # plot median & IQR by 1 cm slice, using different colors for each MLRA
#' xyplot(
#' top ~ p.q50, 
#' groups = factor(mlra), 
#' data = a,
#' lower=a$p.q25, 
#' upper=a$p.q75,
#' alpha=0.25,
#' sync.colors = TRUE,
#' ylim=c(160,-5), 
#' scales = list(alternating = 1, y = list(tick.num = 7)),
#' panel = panel.depth_function, 
#' prepanel = prepanel.depth_function,
#' ylab='Depth (cm)', xlab='Base Saturation at pH 7',
#' par.settings = tactile.theme(superpose.line = list(lwd = 2)),
#' auto.key = list(lines = TRUE, points = FALSE, columns = 2)
#' )
#'
#'
#'
#' # Extract the 2nd horizon from all profiles as SPDF
#' ca.2 <- ca[, 2]
#'
#' # subset profiles 1 through 10
#' ca.1.to.10 <- ca[1:10, ]
#'
#' # basic plot method: profile plot
#' par(mar = c(0, 0, 3, 1))
#' plotSPC(ca.1.to.10, name='hzn_desgn', color = 'CEC7')
#' }
#'
"ca630"

#' Soil Morphologic, Geochemical, and Mineralogy Data from Rowley et al. 2019.
#'
#' Data from Table 1 and Supplementary Tables 1 and 2 from "A cascading
#' influence of calcium carbonate on the biogeochemistry and pedogenic
#' trajectories of subalpine soils, Switzerland".
#'
#'
#' @name rowley2019
#' @docType data
#' @usage data(rowley2019)
#' @format A \code{SoilProfileCollection} object:
#'
#' site-level attributes \describe{ \item{id}{profile ID} \item{group}{profile
#' group} }
#'
#' horizon-level attributes \describe{ \item{sample_id}{sample ID}
#' \item{name}{horizon name} \item{pH}{pH} \item{Al_exch}{cmol(+) / kg,
#' exchangeable Al} \item{Ca_exch}{cmol(+) / kg, exchangeable Ca}
#' \item{CEC_sum}{cmol(+) / kg, cation exchange capacity calculated as the sum
#' of exchangeable cations, not including H+}
#' \item{Ca_exch_saturation}{percent} \item{Al_exch_saturation}{percent}
#' \item{TON}{percent, total nitrogen} \item{SOC}{percent, soil organic carbon}
#' \item{C_to_N}{carbon to nitrogen ratio} \item{Alo}{g/kg, oxalate-extractable
#' Al} \item{Feo}{g/kg, oxalate-extractable Fe} \item{Ald}{g/kg,
#' dithionite-extractable Al} \item{Fed}{g/kg, dithionite-extractable Fe}
#' \item{Feo_Fed}{Fe_o to Fe_d ratio} \item{id}{profile ID} \item{top}{horizon
#' top (cm)} \item{bottom}{horizon bottom (cm)} \item{Al}{g/kg by x-ray
#' fluorescence} \item{Ca}{g/kg by x-ray fluorescence} \item{Cr}{g/kg by x-ray
#' fluorescence} \item{Fe}{g/kg by x-ray fluorescence} \item{K}{g/kg by x-ray
#' fluorescence} \item{Mg}{g/kg by x-ray fluorescence} \item{Mn}{g/kg by x-ray
#' fluorescence} \item{Na}{g/kg by x-ray fluorescence} \item{Ni}{g/kg by x-ray
#' fluorescence} \item{P}{g/kg by x-ray fluorescence} \item{Si}{g/kg by x-ray
#' fluorescence} \item{Ti}{g/kg by x-ray fluorescence}
#' \item{Phyllosilicates}{percent by x-ray diffraction spectra}
#' \item{Quartz}{percent by x-ray diffraction spectra}
#' \item{K_Feldspar}{percent by x-ray diffraction spectra}
#' \item{Na_Plagioclase}{percent by x-ray diffraction spectra}
#' \item{Goethite}{percent by x-ray diffraction spectra}
#' \item{Unidentified}{percent by x-ray diffraction spectra}
#' \item{CCE_Total}{percent} \item{CCE_Reactive}{percent}
#' \item{Reactive_carbonate}{percent} \item{Sand}{percent <2um}
#' \item{Silt}{percent 2-50um} \item{Clay}{percent 50-2000um}
#'
#' \item{CaH2O}{Milliq ex: grams of Ca per kilogram of dry soil (g kg-1)}
#' \item{Ca2MKCl}{2M KCl: grams of Ca per kilogram of dry soil (g kg-1)}
#' \item{CaNa2EDTA}{0.05 M Na2EDTA: grams of Ca per kilogram of dry soil (g
#' kg-1)} \item{CaCuCl2}{0.5 M CuCl2: grams of Ca per kilogram of dry soil (g
#' kg-1)}
#'
#' \item{hzID}{horizon ID} }
#' @references Mike C. Rowley, Stephanie Grand, Thierry Adatte, Eric P.
#' Verrecchia, Cascading influence of calcium carbonate on the biogeochemistry and
#' pedogenic trajectories of subalpine soils), Switzerland, Geoderma, 2019,
#' 114065, ISSN 0016-7061, \doi{10.1016/j.geoderma.2019.114065}.
#' @keywords datasets
#' @examples
#'
#' library(lattice)
#'
#' # load data
#' data('rowley2019')
#'
#' # check first 5 rows and 10 columns of horizon data
#' horizons(rowley2019)[1:5, 1:10]
#'
#' # check site data
#' site(rowley2019)
#'
#' # graphical summary
#' par(mar=c(1,1,3,1))
#' plotSPC(rowley2019, color='Feo_Fed', name='name', cex.names=0.85)
#'
#' plotSPC(rowley2019, color='Ca_exch', name='name', cex.names=0.85)
#'
#' # grouped plot
#' groupedProfilePlot(rowley2019, groups = 'group', color='Ca_exch',
#' name='name', cex.names=0.85, group.name.offset = -10)
#'
#' # aggregate over 1cm slices, for select properties
#' a <- slab(rowley2019, group ~ Reactive_carbonate + Ca_exch + pH + K_Feldspar + Na_Plagioclase + Al)
#'
#' # plot styling
#' tps <- list(superpose.line=list(lwd=2, col=c('royalblue', 'firebrick')))
#'
#' # make the figure
#' xyplot(top ~ p.q50 | variable, data=a, ylab='Depth', groups=group,
#'        main='', as.table=TRUE,
#'        xlab='median bounded by 25th and 75th percentiles',
#'        lower=a$p.q25, upper=a$p.q75, ylim=c(55,-5),
#'        panel=panel.depth_function,
#'        prepanel=prepanel.depth_function,
#'        cf=a$contributing_fraction,
#'        alpha=0.33, sync.colors=TRUE,
#'        scales=list(x=list(relation='free', alternating=1)),
#'        par.settings=tps,
#'        auto.key=list(columns=2, lines=TRUE, points=FALSE),
#'        strip=strip.custom(bg=grey(0.9))
#' )
#'
#'
"rowley2019"

#' Soil Morphologic Data from Jacobs et al. 2002.
#'
#' Select soil morphologic data from "Redoximorphic Features as Indicators of
#' Seasonal Saturation, Lowndes County, Georgia". This is a useful sample
#' dataset for testing the analysis and visualization of redoximorphic
#' features.
#'
#'
#' @name jacobs2000
#' @docType data
#'
#' @format A \code{SoilProfileCollection} object.
#'
#' @references Jacobs, P. M., L. T. West, and J. N. Shaw. 2002. Redoximorphic
#' Features as Indicators of Seasonal Saturation, Lowndes County, Georgia. Soil
#' Sci. Soc. Am. J. 66:315-323. doi:\doi{10.2136/sssaj2002.3150}
#' @usage data(jacobs2000)
#'
#' @keywords datasets
#'
#' @examples
#'
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' # load
#' data(jacobs2000)
#' 
#' # basic plot
#' par(mar = c(0, 1, 3, 1.5))
#' plotSPC(jacobs2000, name='name', color='matrix_color', width=0.3)
#' 
#' # add concentrations
#' addVolumeFraction(jacobs2000, 'concentration_pct',
#'                   col = jacobs2000$concentration_color, pch = 16, cex.max = 0.5)
#' 
#' # add depletions
#' plotSPC(jacobs2000, name='name', color='matrix_color', width=0.3)
#' addVolumeFraction(jacobs2000, 'depletion_pct',
#'                   col = jacobs2000$depletion_color, pch = 16, cex.max = 0.5)
#' 
#' # time saturated
#' plotSPC(jacobs2000, color='time_saturated', cex.names=0.8, col.label = 'Time Saturated')
#' 
#' 
#' # color contrast: matrix vs. concentrations
#' cc <- colorContrast(jacobs2000$matrix_color_munsell, jacobs2000$concentration_munsell)
#' cc <- na.omit(cc)
#' 
#' cc <- cc[order(cc$dE00), ]
#' cc <- unique(cc)
#' 
#' par(bg = 'black', fg = 'white')
#' colorContrastPlot(cc$m1[1:10], cc$m2[1:10], labels = c('matrix', 'concentration'))
#' colorContrastPlot(cc$m1[11:21], cc$m2[11:21], labels = c('matrix', 'concentration'))
#' 
#' 
#' # color contrast: depletion vs. concentrations
#' cc <- colorContrast(jacobs2000$depletion_munsell, jacobs2000$concentration_munsell)
#' cc <- na.omit(cc)
#' 
#' cc <- cc[order(cc$dE00), ]
#' cc <- unique(cc)
#' 
#' par(bg = 'black', fg = 'white')
#' colorContrastPlot(cc$m1, cc$m2, labels = c('depletion', 'concentration'))
#'
"jacobs2000"


#' Average Hydraulic Parameters from the ROSETTA Model by USDA Soil Texture
#' Class
#'
#' Average soil hydraulic parameters generated by the first stage predictions
#' of the ROSETTA model by USDA soil texture class. These data were extracted
#' from [ROSETTA documentation](https://www.ars.usda.gov/pacific-west-area/riverside-ca/agricultural-water-efficiency-and-salinity-research-unit/docs/model/rosetta-class-average-hydraulic-parameters/) and re-formatted for ease of use.
#'
#' Theoretical water-retention parameters for uniform soil material of each
#' texture class have been estimated via van Genuchten model.
#'
#' See [the related tutorial](https://ncss-tech.github.io/AQP/soilDB/fetchKSSL-VG-demo.html)
#'
#' @name ROSETTA.centroids
#' @docType data
#' @usage data(ROSETTA.centroids)
#' @format A data frame: \describe{
#'
#' \item{texture}{soil texture class, ordered from low to high
#' available water holding capacity} \item{theta_r}{average saturated
#' water content} \item{theta_s}{average residual water content}
#' \item{alpha}{average value, related to the inverse of the air entry
#' suction, log10-transformed values with units of cm}
#' \item{npar}{average value, index of pore size distribution,
#' log10-transformed values with units of 1/cm}
#'
#' \item{theta_r_sd}{1 standard deviation of \code{theta_r}}
#' \item{theta_s_sd}{1 standard deviation of \code{theta_s}}
#' \item{alpha_sd}{1 standard deviation of \code{alpha}}
#' \item{npar_sd}{1 standard deviation of \code{npar}}
#'
#' \item{sat}{approximate volumetric water content at which soil
#' material is saturated} \item{fc}{approximate volumetric water
#' content at which matrix potential = -33kPa} \item{pwp}{approximate
#' volumetric water content at which matrix potential = -1500kPa}
#' \item{awc}{approximate available water holding capacity: VWC(-33kPa)
#' - VWC(-1500kPa)} }
#'
#' @references van Genuchten, M.Th. (1980). "A closed-form equation for
#' predicting the hydraulic conductivity of unsaturated soils". Soil Science
#' Society of America Journal. 44 (5): 892-898.
#'
#' Schaap, M.G., F.J. Leij, and M.Th. van Genuchten. 2001. ROSETTA: A computer
#' program for estimating soil hydraulic parameters with hierarchical
#' pedotransfer functions. Journal of Hydrology 251(3–4): 163-176.
#' @source
#' [ROSETTA Class Average Hydraulic Parameters](https://www.ars.usda.gov/pacific-west-area/riverside-ca/agricultural-water-efficiency-and-salinity-research-unit/docs/model/rosetta-class-average-hydraulic-parameters/)
#' @keywords datasets
#' @examples
#'
#'
#' \dontrun{
#'
#' library(aqp)
#' library(soilDB)
#' library(latticeExtra)
#'
#' data("ROSETTA.centroids")
#'
#' # iterate over horizons and generate VG model curve
#' res <- lapply(1:nrow(ROSETTA.centroids), function(i) {
#'   m <- KSSL_VG_model(VG_params = ROSETTA.centroids[i, ], phi_min = 10^-3, phi_max=10^6)$VG_curve
#'   # copy generalized hz label
#'   m$hz <- ROSETTA.centroids$hz[i]
#'   # copy ID
#'   m$texture_class <- ROSETTA.centroids$texture[i]
#'   return(m)
#' })
#'
#' # copy over lab sample number as ID
#' res <- do.call('rbind', res)
#'
#' # check: OK
#' str(res)
#'
#'
#' # visual check: OK
#' xyplot(
#'   phi ~ theta | texture_class, data=res,
#'   type=c('l', 'g'),
#'   scales=list(alternating=3, x=list(tick.number=10), y=list(log=10, tick.number=10)),
#'   yscale.components=yscale.components.logpower,
#'   ylab=expression(Suction~~(kPa)),
#'   xlab=expression(Volumetric~Water~Content~~(cm^3/cm^3)),
#'   par.settings = list(superpose.line=list(col='RoyalBlue', lwd=2)),
#'   strip=strip.custom(bg=grey(0.85)),
#'   as.table=TRUE
#' )
#'
#'
#' }
#'
#'
"ROSETTA.centroids"


#'
#' @title US State Soils
#' @description A listing of the 50 US state soils, along with Puerto Rico and Virgin Islands.
#' @keywords datasets
#' @usage data(us.state.soils)
#' @format
#' \describe{
#'   \item{state}{state name}
#'   \item{abbreviated}{abbreviated state name}
#'   \item{series}{soil series name}
#' }
#'
"us.state.soils"

#' Munsell to sRGB Lookup Table for Common Soil Colors
#'
#' A lookup table of interpolated Munsell color chips for common soil colors.
#'
#' See \code{munsell2rgb} for conversion examples. Values
#' are referenced to the D65 standard illuminant.
#'
#' @name munsell
#' @docType data
#'
#' @format A data.frame with 8825 rows.  
#'  - `hue`: Munsell Hue, upper case
#'  - `value`: Munsell Value
#'  - `chroma`: Munsell Chroma
#'  - `r`: sRGB "red" value (0-1)
#'  - `g`: sRGB "green" value (0-1)
#'  - `b`: sRGB "blue" value (0-1)
#'  - `L`: CIELAB "L" coordinate
#'  - `A`: CIELAB "A" coordinate
#'  - `B`: CIELAB "B" coordinate
#'
#' @usage data(munsell)
#'
#' @references 
#'  - Color conversion equations
#'    + http://www.brucelindbloom.com/index.html?ColorCalcHelp.html
#'  - Methods used to generate this table
#'    + http://dx.doi.org/10.1016/j.cageo.2012.10.020
#' @source Color chip XYZ values: \url{https://www.rit.edu/science/munsell-color-science-lab-educational-resources#munsell-renotation-data}
#' @keywords datasets
#' @examples
#'
#' data(munsell)
#'
"munsell"


#' Munsell Colors of Common Soil Minerals
#'
#' Munsell colors for some common soil minerals.
#'
#' Soil color and other properties including texture, structure, and
#' consistence are used to distinguish and identify soil horizons (layers) and
#' to group soils according to the soil classification system called Soil
#' Taxonomy. Color development and distribution of color within a soil profile
#' are part of weathering. As rocks containing iron or manganese weather, the
#' elements oxidize. Iron forms small crystals with a yellow or red color,
#' organic matter decomposes into black humus, and manganese forms black
#' mineral deposits. These pigments paint the soil (Michigan State Soil). Color
#' is also affected by the environment: aerobic environments produce sweeping
#' vistas of uniform or subtly changing color, and anaerobic (lacking oxygen),
#' wet environments disrupt color flow with complex, often intriguing patterns
#' and points of accent. With depth below the soil surface, colors usually
#' become lighter, yellower, or redder.
#'
#' @name soil_minerals
#' @docType data
#' @format A data frame with 20 observations on the following 5 variables.
#' \describe{ \item{mineral}{mineral name} \item{color}{Munsell
#' color} \item{hue}{Munsell hue} \item{value}{Munsell value}
#' \item{chroma}{Munsell chroma} }
#' @usage data(soil_minerals)
#' @references 1. Lynn, W.C. and Pearson, M.J., The Color of Soil, The Science
#' Teacher, May 2000. 2. Schwertmann, U. 1993. Relations Between Iron Oxides,
#' Soil Color, and Soil Formation. "Soil Color". SSSA Special Publication no.
#' 31, pages 51--69.
#' @keywords datasets
#' @examples
#'
#' \dontrun{
#' 
#' library(aqp)
#' library(ape)
#' library(cluster)
#' library(farver)
#' 
#' # load common soil mineral colors
#' data(soil_minerals)
#' 
#' # convert Munsell to R colors
#' soil_minerals$col <- munsell2rgb(
#'   soil_minerals$hue, 
#'   soil_minerals$value,
#'   soil_minerals$chroma
#' )
#' 
#' # make a grid for plotting
#' n <- ceiling(sqrt(nrow(soil_minerals)))
#' 
#' # read from top-left to bottom-right
#' g <- expand.grid(x = 1:n, y = n:1)[1:nrow(soil_minerals),]
#' 
#' # convert Munsell -> CIELAB
#' col.lab <- parseMunsell(soil_minerals$color, returnLAB = TRUE)
#' 
#' # keep track of soil mineral names
#' # in a way that will persist in a dist obj
#' row.names(col.lab) <- soil_minerals$mineral
#' 
#' # pair-wise perceptual distances via CIE dE00
#' d <- compare_colour(
#'   from = col.lab,
#'   from_space = 'lab',
#'   to_space = 'lab',
#'   white_from = 'D65',
#'   method = 'CIE2000'
#' )
#' 
#' # matrix -> dist
#' # note transpose, 
#' # required when specifying only `from` to compare_colour()
#' d <- as.dist(t(d))
#' 
#' # divisive hierarchical clustering of LAB coordinates
#' h <- as.hclust(diana(d))
#' p <- as.phylo(h)
#' 
#' # colors, in order based on clustering
#' # starting from top-left
#' min.cols <- rev(soil_minerals$col[h$order])
#' 
#' # mineral names, in order based on clustering
#' # starting from top-left
#' min.names <- rev(soil_minerals$mineral[h$order])
#' 
#' min.munsell <- rev(soil_minerals$color[h$order])
#' 
#' # plot grid of mineral names / colors
#' layout(matrix(c(1, 2), nrow = 1), widths = c(1.25, 1))
#' 
#' par(mar = c(1, 0, 0, 1))
#' plot(g$x, g$y, pch = 15, cex = 12, axes = FALSE, xlab = '', ylab = '',
#'      col = min.cols, 
#'      xlim = c(0.5, 5.5), ylim = c(1.5, 5.5)
#' )
#' text(g$x, g$y, min.names, adj = c(0.45, 5.5), cex = 0.75, font = 2)
#' text(g$x, g$y, min.munsell, col = invertLabelColor(min.cols), cex = 0.85, font = 2)
#' 
#' title(main = 'Common Soil Pigments', line = -1.75, cex.main = 2)
#' mtext('U. Schwertmann, 1993. SSSA Special Publication no. 31, pages 51--69', side = 1,
#'       cex = 0.75, line = -1.5)
#' 
#' # dendrogram + tip labels with mineral colors
#' plot(p, cex = 0.85, label.offset = 5, font = 1)
#' tiplabels(pch = 15, cex = 3, offset = 2, col = soil_minerals$col)

#'
#' }
#'
#'
"soil_minerals"

#'
#' @title Spectral Library of Munsell Colors
#'
#' @description
#'
#' The original database "SpectralReflectancesOf2007MunsellBookOfColorGlossy.txt" was provided by Paul Centore and downloaded July, 2020. Reflectance values for odd chroma and 8.5-value chips have been interpolated from adjacent chips. See \code{aqp/misc/utils/Munsell/} for the entire set of processing steps.
#'
#' Munsell value typically ranges from 2-9, and chroma from 1-12. Ranges vary by hue. Run `aqp:::.summarizeMunsellSpectraRanges()` for a detailed listing by hue.
#'
#' The original database contains the following description:
#'
#' This file contains spectral reflectance measurements of X-Rite's 2007 Munsell Book of Color (Glossy Finish).  The measurements were made in 2012 with a ColorMunki spectrophotometer.  The first column is the Munsell name.  The remaining columns give reflectance values for 380 nm to 730 nm, in steps of 10 nm.  The reflectance is a value between 0 (indicating that no light at that wavelength is reflected) and 1 (indicating that all the light at that wavelength is reflected).  Occasionally an entry is slightly greater than 1.  The likely cause is random variability, and those entries can be adjusted to 1 with negligible loss. In all, 1485 colour samples were measured.  Researchers are invited to analyze the data in this file.
#'
#' @usage data(munsell.spectra)
#'
#' @references
#' Centore, Paul. Colour Tools for Painters. http://www.munsellcolourscienceforpainters.com/.
#'
#' @aliases munsell.spectra.wide
#'
#' @keywords datasets
#'
#' @format A data frame with 89496 rows and 10 variables:
#' \describe{
#'   \item{munsell}{munsell color}
#'   \item{hue}{hue component}
#'   \item{value}{value component}
#'   \item{chroma}{chroma component}
#'   \item{wavelength}{wavelength (nm)}
#'   \item{reflectance}{reflectance}
#' }
"munsell.spectra"

#' @title Indices of "equivalent" Munsell chips in the `munsell` data set
#'
#' @description
#' A pre-calculated lookup list (made with `farver::compare_colour`) based on pair-wise color contrast (`CIE2000` or `dE00`) evaluated over all "chips" in the `aqp::munsell` data set.
#'
#' The intention is to identify Munsell chips that may be "functionally equivalent" to some other given whole chip elsewhere in the Munsell color space -- as discretized in the \code{aqp::munsell} lookup table.
#'
#' "Equivalent" chips are based (fairly arbitrarily) on the 0.001 probability level of `dE00` (default Type 7 `quantile`) within the upper triangle of the 8467x8467 contrast matrix. This corresponds to a `dE00` threshold of approximately 2.15.
#'
#' This is a naive (to the subtleties of human color perception, and overall magnitude of contrast between some of the "chips") but computationally consistent approach. Using the lookup list, as opposed to manual contrast via e.g. \code{farver::compare_colour} may have some benefits for efficiency in certain applications where the exact contrast value is not as important as the concept of having some threshold that is non-zero, but very small.
#'
#' @usage data(equivalent_munsell)
#'
#' @aliases equivalent_munsell
#'
#' @seealso \code{\link{equivalentMunsellChips}}
#'
#' @references
#' Gaurav Sharma, Wencheng Wu, Edul N. Dalal. (2005). The CIEDE2000 Color-Difference Formula: Implementation Notes, Supplementary Test Data, and Mathematical Observations. COLOR research and application. 30(1):21-30. http://www2.ece.rochester.edu/~gsharma/ciede2000/ciede2000noteCRNA.pdf
#'
#'  Thomas Lin Pedersen, Berendea Nicolae and Romain Francois (2020). farver: High Performance Colour Space Manipulation. R package version 2.0.3. https://CRAN.R-project.org/package=farver
#'
#' Dong, C.E., Webb, J.B., Bottrell, M.C., Saginor, I., Lee, B.D. and Stern, L.A. (2020). Strengths, Limitations, and Recommendations for Instrumental Color Measurement in Forensic Soil Characterization. J Forensic Sci, 65: 438-449. https://doi.org/10.1111/1556-4029.14193
#'
#' @keywords datasets
#'
#' @format A named list with 8467 elements, each containing a numeric vector of indices corresponding to the \code{munsell} data set, which has 8467 rows (unique, whole-number chips). Names have the format \code{HUE VALUE/CHROMA}, e.g. \code{"7.5YR 4/4"}
#'
#' @examples
#' data(equivalent_munsell)
"equivalent_munsell"



#'
#' @title Standard Illuminants and Observers
#' 
#' @description D65 and F2 standard illuminant spectral power distributions, CIE1931 Standard Observer (2 degree), and CIE1964 Supplemental Standard Observer (10 degree)
#'
#'
#'
#' @keywords datasets
#' @usage data(spectral.reference)
#' 
#' @references 
#' 
#' Marcus, R.T. (1998). The Measurement of Color. In K. Nassau (Ed.), Color for Science, Art, and Technology (pp. 32-96). North-Holland.
#' 
#' CIE Colorimetry – Part 1: CIE standard colorimetric observers. CIES014-1/E:2006 – ISO 11664-1:2007(E)
#' 
#' CIE. (n.d.). CIE 15:2004 Tables Data. Retrieved from https://law.resource.org/pub/us/cfr/ibr/003/cie.15.2004.tables.xls
#' 
#' @examples 
#' 
#' data("spectral.reference")
#' 
#' matplot(
#'   x = spectral.reference[, 1],
#'   y = spectral.reference[, c('xbar_2', 'ybar_2', 'zbar_2')],
#'   ylim = c(0, 2),
#'   type = 'l',
#'   lwd = 2, 
#'   lty = 1,
#'   las = 1,
#'   xlab = 'Wavelength (nm)',
#'   ylab = 'Weight | Intensity',
#'   main = "CIE1931 (2\u00B0) and CIE1964 (10\u00B0) Standard Observers
#'   D65 and F2 Illuminant Power Spectrum (rescaled / offset for clarity)",
#'   cex.main = 0.9
#' )
#' 
#' matlines(
#'   x = spectral.reference[, 1],
#'   y = spectral.reference[, c('xbar_10', 'ybar_10', 'zbar_10')],
#'   type = 'l',
#'   lwd = 2, 
#'   lty = 2,
#'   las = 1,
#'   xlab = 'Wavelength (nm)',
#'   ylab = 'Weight | Intensity',
#'   main = 'CIE1931 Standard Observer Weights\nD65 Standard Illuminant'
#' )
#' 
#' lines(
#'   x = spectral.reference$w,
#'   y = (spectral.reference$D65 / 100) + 0.33,
#'   lty = 1,
#'   col = 'royalblue'
#' )
#' 
#' lines(
#'   x = spectral.reference$w,
#'   y = (spectral.reference$F2 / 25) + 0.4,
#'   lty = 1,
#'   col = 'violet'
#' )
#' 
#' legend(
#'   'topright',
#'   legend = c('X_2', 'Y_2', 'Z_2', 'X_10', 'Y_10', 'Z_10', 'D65', 'F2'),
#'   col = c(1, 2, 3, 1, 2, 3, 'royalblue', 'violet'),
#'   lwd = c(2, 2, 2, 2, 2, 2, 1, 1),
#'   lty = c(1, 1, 1, 2, 2, 2, 1, 1),
#'   bty = 'n',
#'   cex = 0.85
#' )
#'
"spectral.reference"



#'
#' @title Traditional Soil Color Names
#' 
#' @description Traditional soil color names associated with select Munsell colors.
#' 
#' @references 
#' 
#' Sourced from the "colorconverter" NASIS property script.
#'
#' @keywords datasets
#' @usage data(traditionalColorNames)
#' 
"traditionalColorNames"


#' @title Munsell Hue Position Reference
#' 
#' @description Position data for the 40 standard Munsell hues (and neutral). Data include angular positions (compass-style, origin at `[x = 0, y = 1]`, CW rotation) and Cartesian coordinates on the unit circle.
#' 
#' @references 
#' 
#' Munsell book of color. 1976. Macbeth, a Division of Kollmorgen Corp., Baltimore, MD.
#'
#' @keywords datasets
#' @usage data(munsellHuePosition)
#' 
"munsellHuePosition"


#' @title pH Reaction Classes
#' 
#' @description Levels of pH (reaction) classes including descriptive name, and range from low to high pH
#' @references Soil Science Division Staff. (2017) Soil Survey Manual. C. Ditzler, K. Scheffe, and H.C. Monger (eds.). USDA Handbook 18. Government Printing Office, Washington, D.C.
#' @keywords datasets
#' @usage data(reactionclass)
#' 
"reactionclass"

#' @title Generate a `SoilProfileCollection` of random profiles
#' 
#' @description
#' This function provides a convenient abstraction of `lapply()`, [random_profile()], and [combine()] which are typically used together to create a `SoilProfileCollection` object with >1 soil profiles. `rp()` creates zero-padded integer IDs for logical sorting and indexing of profiles. For more complex IDs or additional flexibility, see [random_profile()]. See [random_profile()] for all possible arguments.
#' 
#'
#' @param size integer, number of requested profiles 
#' @param prefix prefix added to zero-padded, integer IDs
#' @param ... additional arguments to [random_profile()]
#' 
#' @returns a `SoilProfileCollection` object
#' @export
#'
#' @examples
#' 
#' # generate a SoilProfileCollection object with 10 profiles
#' # using 0-padded, integer IDs for intuitive sorting
#' spc <- rp(10, method = 'LPP')
#' plotSPC(spc, color = 'p1')
#' 
#' # apply a prefix to the IDs
#' spc <- rp(10, prefix = 'A-', method = 'LPP')
#' plotSPC(spc, color = 'p1')
#' 
rp <- function(size, prefix = NULL, ...) {
  
  # format string for just enough 0-padding 
  .padwidth <- nchar(size) + 1
  
  # optionally add prefix create formatting string
  if(is.null(prefix)) {
    .fmt <- sprintf("%%0%sd", .padwidth)
  } else {
    .fmt <- sprintf("%s%%0%sd", prefix, .padwidth)
  }
  
  # generate IDs
  .s <- sprintf(.fmt, 1:size)
  
  # typical pattern of random_profile() then combine()
  .x <- lapply(.s, random_profile, SPC = TRUE, ...)
  .x <- combine(.x)
  
  return(.x)
}


# logistic power peak function
.lpp <- function(x, a, b, u, d, e) {
  # the exponential term
  f.exp <- exp((x + d * log(e) - u) / d)
  # first part
  f1 <- (b/u) * (1 + f.exp)^((-e - 1) / e)
  # second part
  f2 <- f.exp * (e + 1)^((e+1) / e)
  # combine pieces
  res <- a + f1 * f2
  return(res)
  }


#' @title Random Profile
#'
#' @description Generate a random soil profile according to set criteria, with correlated
#' depth trends.
#'
#' The random walk method produces profiles with considerable variation between
#' horizons and is based on values from the normal distribution seeded with
#' means and standard deviations drawn from the uniform distribution of \[0,
#' 10].
#'
#' The logistic power peak (LPP) function can be used to generate random soil
#' property depth functions that are sharply peaked. LPP parameters can be
#' hard-coded using the optional arguments: "lpp.a", "lpp.b", "lpp.u", "lpp.d",
#' "lpp.e". Amplitude of the peak is controlled by ("lpp.a + "lpp.b"), depth of
#' the peak by "lpp.u", and abruptness by "lpp.d" and "lpp.e". Further
#' description of the method is outlined in (Brenton et al, 2011). Simulated
#' horizon distinctness codes are based on the USDA-NCSS field description
#' methods.
#' 
#' Simulated distinctness codes are constrained according to horizon thickness,
#' i.e. a gradual boundary (+/- 5cm) will not be simulated for horizons that
#' are thinner than 3x this vertical distance
#' 
#' The [rp()] function is a convenient wrapper to this function, when requesting a simulated `SoilProfileCollection` of specified size.
#'
#' @aliases random_profile .lpp
#' @param id a character or numeric id used for this profile
#' @param n vector of possible number of horizons, or the exact number of
#' horizons (see below)
#' @param min_thick minimum thickness criteria for a simulated horizon
#' @param max_thick maximum thickness criteria for a simulated horizon
#' @param n_prop number of simulated soil properties (columns in the returned
#' dataframe)
#' @param exact should the exact number of requested horizons be generated?
#' (defaults to FALSE)
#' @param method named method used to synthesize depth function ('random_walk'
#' or 'LPP'), see details
#' @param HzDistinctSim optionally simulate horizon boundary distinctness codes
#' @param SPC result is a `SoilProfileCollection` object, otherwise a `data.frame` object
#' @param \dots additional parameters passed-in to the LPP `.lpp`)
#' function
#' @return A `data.frame` or `SoilProfileCollection` object.
#' @note See examples for ideas on simulating several profiles at once.
#' @author Dylan E. Beaudette
#' @seealso [rp()], [hzDistinctnessCodeToOffset()]
#' @references Myers, D. B.; Kitchen, N. R.; Sudduth, K. A.; Miles, R. J.;
#' Sadler, E. J. & Grunwald, S. Peak functions for modeling high resolution
#' soil profile data Geoderma, 2011, 166, 74-83.
#' @keywords manip
#' @export
#' @examples
#'
#'
#' # generate 10 random profiles, result is a list of SoilProfileCollection objects
#' d <- lapply(1:10, random_profile, SPC=TRUE)
#'
#' # combine
#' d <- combine(d)
#'
#' # plot
#' opar <- par(mar=c(0,0,3,2))
#' plotSPC(d, color='p1', name='name', cex.names=0.75)
#' par(opar)
#'
#' # simulate horizon boundary distinctness codes:
#' d <- lapply(1:10, random_profile, SPC=TRUE, HzDistinctSim=TRUE)
#' d <- combine(d)
#'
#' d$HzD <- hzDistinctnessCodeToOffset(d$HzDistinctCode)
#'
#' opar <- par(mar=c(0,0,3,2))
#' plotSPC(d, name='name', color='p1', hz.distinctness.offset='HzD')
#' par(opar)
#'
#'
#' # depth functions are generated using the LPP function
#' opar <- par(mfrow=c(2,1), mar=c(0,0,3,0))
#'
#' # generate data
#' d.1 <- lapply(1:10, random_profile, SPC=TRUE, n=c(6, 7, 8), n_prop=1, method='LPP')
#' d.1 <- combine(d.1)
#'
#' # plot
#' plotSPC(d.1, name='name', color='p1', col.label = 'LPP Defaults')
#'
#'
#' # do this again, this time set all of the LPP parameters
#' d.2 <- lapply(1:10, random_profile, SPC=TRUE, n=c(6, 7, 8), n_prop=1, method='LPP',
#'            lpp.a=5, lpp.b=10, lpp.d=5, lpp.e=5, lpp.u=25)
#' d.2 <- combine(d.2)
#'
#' # plot
#' plotSPC(d.2, name='name', color='p1', col.label = 'Custom LPP Parameters')
#'
#'
#' # reset plotting defaults
#' par(opar)
#'
#'
#'
#' # try plotting the LPP-derived simulated data
#' # aggregated over all profiles
#' a <- slab(d.2, fm= ~ p1)
#' a$mid <- with(a, (top + bottom) / 2)
#'
#' library(lattice)
#' (p1 <- xyplot(mid ~ p.q50, data=a,
#'               lower=a$p.q25, upper=a$p.q75, ylim=c(150,-5), alpha=0.5,
#'               panel=panel.depth_function, prepanel=prepanel.depth_function,
#'               cf=a$contributing_fraction, xlab='Simulated Data', ylab='Depth',
#'               main='LPP(a=5, b=10, d=5, e=5, u=25)',
#'               par.settings=list(superpose.line=list(col='black', lwd=2))
#' ))
#'
#' # optionally add original data as step-functions
#' if(require(latticeExtra)) {
#'   h <- horizons(d.2)
#'   p1 + as.layer(xyplot(top ~ p1, groups=id, data=h,
#'                        horizontal=TRUE, type='S',
#'                        par.settings=list(superpose.line=list(col='blue', lwd=1, lty=2))))
#' }
#'
#'
#'
random_profile <- function(id, n = c(3,4,5,6), min_thick = 5, max_thick = 30, n_prop = 5, exact = FALSE, method= 'random_walk', HzDistinctSim = FALSE, SPC = FALSE, ...) {

  # sanity check
  if(missing(id))
	  stop('must specify an id')

  if(max_thick < min_thick)
	  stop('illogical horizon thickness constraints')

  if(! method %in% c('random_walk', 'LPP'))
	stop('invalid method')

  # get extra arguments
  dots <- list(...)

  # if requested, give back the exact number of horizons
  if(length(n) == 1 & exact)
	  n_hz <- n

  # otherwise randomly choose from suggestions
  else
	  n_hz <- sample(n, 1)

  # generate hz top bnd
  tops <- integer(n_hz-1)
  for(i in 1:(n_hz-1))
	  tops[i] <- sample(min_thick:max_thick, 1)

  # add 0, then generate bottom bnd
  tops <- as.integer(c(0, tops))
  bottoms <- as.integer(c(tops[-1], sample(min_thick:max_thick, 1)))

  # combine into a df
  # always treat ID as a character: "solves" some cases of SPC corruption due to re-ordering of integers cast to character:
  # https://github.com/ncss-tech/aqp/issues/90
  d <- data.frame(id=as.character(id), top=cumsum(tops), bottom=cumsum(bottoms), name=paste('H',1:n_hz,sep=''), stringsAsFactors = FALSE)

  # generate several properties
  # with different means / sd
  for(i in 1:n_prop) {
	# init storage
	  p <- numeric(n_hz)

	if(method == 'random_walk') {
		p[1] <- rnorm(1, mean=runif(n=1, min=-10, max=10), sd=runif(n=1, min=1, max=10))
		for(j in 2:n_hz)
			p[j] <- p[j-1] + rnorm(1, mean=runif(n=1, min=-10, max=10), sd=runif(n=1, min=1, max=10))
	  }

	if(method == 'LPP') {
	  # generate synthetic values at horizon mid-points
	  mids <- with(d, (top + bottom)/2)

	  # generate LPP parameters from uniform dist if not given as arguments
	  if(is.null(dots[['lpp.a']]))
			lpp.a <- runif(n=1, min=5, max=25)
	  else
			lpp.a <- dots[['lpp.a']]

	  if(is.null(dots[['lpp.b']]))
			lpp.b <- runif(n=1, min=20, max=60)
	  else
			lpp.b <- dots[['lpp.b']]

	  if(is.null(dots[['lpp.u']]))
			lpp.u <- runif(n=1, min=10, max=90)
	  else
			lpp.u <- dots[['lpp.u']]

	  if(is.null(dots[['lpp.d']]))
			lpp.d <- runif(n=1, min=1, max=10)
	  else
			lpp.d <- dots[['lpp.d']]

	  if(is.null(dots[['lpp.e']]))
			lpp.e <- runif(n=1, min=5, max=20)
	  else
			lpp.e <- dots[['lpp.e']]

	  # generate vector of synthetic values based on LPP
	  p <- .lpp(mids, a=lpp.a, b=lpp.b, u=lpp.u, d=lpp.d, e=lpp.e)
	  }

	  # add generated depth profile to horizons
	  new_col <- paste('p',i, sep='')
	  d[, new_col] <- p
	  }

  # optionally add horizon distinctness codes:
  # these are based on USDA-NCSS codes and approximate vertical offsets
	# codes are constrained to the thickness of the horizon
  if(HzDistinctSim) {
  	# standard codes and offsets
  	codes <- c('A','C','G','D')
  	offsets <- hzDistinctnessCodeToOffset(codes)
  	# compute horizon thickness vector
  	thick <- with(d, bottom-top)

  	# create matrix of distinctness codes based on (1/3) horizon thickness
  	# 1 when possible, 0 when impossible
  	prob.matrix <- t(sapply(thick, function(i) (i/3) >= offsets))
  	prob.matrix[which(prob.matrix)] <- 1

  	d.codes <- vector(mode='character', length=n_hz)
  	for(i in 1:n_hz) {
  		d.codes[i] <- sample(codes, size=1, prob=prob.matrix[i, ])
  	}

  	d$HzDistinctCode <- d.codes
  }

  # note: 3-4x performance hit when calling from lapply(1:big.number, ...)
  # optionally return as SPC
  if(SPC) {
    depths(d) <- id ~ top + bottom
  }

  # all done
  return(d)
}

#' @title Repair Problematic Lower Horizon Depths
#' 
#' @description Attempt a simple repair of horizon bottom depths in the presence of `NA`, or in cases where the horizon shares a common top and bottom depth. Both situations are common in pedon description where "contact" (Cd, Cr, R, etc.) was described without a lower depth. 
#' 
#' @details This repair is applied to the deepest horizon within a profile as identified by [`getLastHorizonID`], as well as to bottom depths of any horizon that has a horizon below it. Horizon bottom depths are adjusted by adding `adj` (if non-NA). If the resulting value exceeds `max.depth`, the `max.depth` value is returned (if not `NA`).
#' 
#' @param x `SoilProfileCollection`
#' 
#' @param adj vertical offset applied to "repair" missing bottom depths when top and bottom depth are equal or bottom depth is missing. (`NA` to use `max.depth`)
#' 
#' @param max.depth If adj is `NA`, or the resulting offset sum exceeds `max.depth`, `max.depth` is used. 
#' 
#' @return `SoilProfileCollection` with a new (logical) horizon-level attribute `.repaired` marking affected horizons
#' @export
#' @examples 
#' 
#' h <- data.frame(
#' id = c(1, 1, 1, 2, 2, 2, 2, 3, 3),
#' top = c(0:2, 0:3, 0:1) * 10,
#' bottom = c(rep(NA_integer_, 7), c(10, 99))
#' )
#' 
#' # NA depths result in warnings
#' suppressWarnings({
#'   depths(h) <- id ~ top + bottom
#' })
#' 
#' # inspect data before repairs
#' plotSPC(h)
#' 
#' g <- repairMissingHzDepths(h)
#' 
#' # all depth logic now valid
#' all(checkHzDepthLogic(g)$valid)
#' 
#' # inspect
#' plotSPC(g)
#' 
#' # no adj, max.depth only
#' f <- repairMissingHzDepths(h, adj = NA, max.depth = 200)
#' all(checkHzDepthLogic(f)$valid)
#' plotSPC(f)
#' 
#' # max.depth defaults to max(x) if too small
#' f$bottom[c(3,7)] <- NA
#' d <- repairMissingHzDepths(f, adj = NA, max.depth = 20)
#' all(checkHzDepthLogic(d)$valid)
#' plotSPC(d)
#' 
repairMissingHzDepths <- function(x, adj = 10, max.depth = 200) {
  # define SPC k-keywords as local vars for R CMD CHECK
  .LAST <- NULL; .HZID <- NULL
  
  # sanity checks
  if(!inherits(x, 'SoilProfileCollection')) {
    stop('`x` should be a SoilProfileCollection', call. = FALSE)
  }
  
  if (!is.na(adj)) {
    # adj must be an integer
    adj <- round(adj)
    
    # reasonable adj values
    if(adj < 1) {
      stop('`adj` should be >= 1', call. = FALSE)
    }
  }
  
  if (!is.na(max.depth) && max.depth < max(x)) {
    max.depth <- max(x)
    message("Using max.depth = ", max.depth)
  }
  
  # hz top, bottom, ID names
  hztb <- horizonDepths(x)
  hzidn <- hzidname(x)
  
  # Setup: get horizon IDs of bottom-most horizons of all profiles
  bottom.idx <- x[,, .LAST, .HZID] # which(hzID(h) %in% getLastHorizonID(h))
  valid.hzIDs <- hzID(x)[bottom.idx] # getLastHorizonID(x)
  
  # find NA bottom depths
  na.bottom <- which(is.na(x[[hztb[2]]]))
  
  # # Step 1: 
  # # find non-NA top AND NA bottom AND (only deepest horizon of each profile)
  # fill in the deepest horizon in each profile if it is missing
  idx1 <- intersect(na.bottom, as.numeric(bottom.idx))
  
  # # make the edit 
  x[[hztb[2]]][idx1] <- pmin(x[[hztb[1]]][idx1] + adj, max.depth, na.rm = TRUE)
  
  # Step 2: 
  # calculate the top depths of the underlying horizon for remaining NA bottom depths
  idx2 <- setdiff(na.bottom, bottom.idx)
  
  # replace the bottom depths
  x[[hztb[2]]][is.na(x[[hztb[2]]])] <-  x[[hztb[1]]][hzOffset(x, hzidx = idx2, offset = 1)]
  #                                     x[[hztb[1]]][which(is.na(x[[hztb[2]]])) + 1]
  
  
  # Step 3:
  # top == bottom AND (only deepest horizon of each profile)
  idx3 <- which((x[[hztb[1]]] == x[[hztb[2]]]) & (x[[hzidn]] %in% valid.hzIDs))
  
  # make the edit
  x[[hztb[2]]][idx3] <- pmin(x[[hztb[2]]][idx3] + adj, max.depth, na.rm = TRUE)
  
  # keep track of "repaired" horizons
  idx <- unique(c(idx1, idx2, idx3))
  horizons(x)$.repaired <- rep(FALSE, nrow(x))
  horizons(x)$.repaired[idx] <- TRUE
  
  # repaired SPC
  return(x)
}

context("repairMissingHzDepths")


data(sp4)
depths(sp4) <- id ~ top + bottom

# copy to break
x <- sp4

## legal (deepest) horizons to repair:
# introduce NA
x$bottom[4] <- NA
# top == bottom
x$bottom[6] <- x$top[6]

## illegal horizons to repair
x$bottom[12] <- NA

test_that("works as expected", {
  
  a <- 10
  z <- repairMissingHzDepths(x, adj = a)
  
  # output is an SPC
  expect_true(inherits(z, 'SoilProfileCollection'))
  
  # bottom depths repaired according to `a` [4,6] or adjacent top depths [12]
  expect_true(z$bottom[4] == z$top[4] + a)
  expect_true(z$bottom[6] == z$top[6] + a)
  expect_true(z$bottom[12] == sp4$bottom[12])
  
  # repair all missing bottom depths, existing data unaltered
  h <- data.frame(
    id = c(1, 1, 1, 2, 2, 2, 2, 3, 3),
    top = c(0:2, 0:3, 0:1) * 10,
    bottom = c(rep(NA_integer_, 7), c(10, 99))
  )
  
  # NA depths result in warnings
  expect_warning({
    depths(h) <- id ~ top + bottom
  })
  
  # all depth logic in resulting SPC is valid, existing data preserved
  g <- repairMissingHzDepths(h)
  expect_true(all(checkHzDepthLogic(g)$valid))
  expect_equal(g[3, ]$bottom[1:2], c(10, 99))
  expect_equal(min(g), 30)
  
  # no adj, max.depth only, existing data preserved
  f <- repairMissingHzDepths(h, adj = NA, max.depth = 200)
  expect_true(all(checkHzDepthLogic(f)$valid))
  expect_equal(f$bottom[c(3, 7, 8, 9)], c(200, 200, 10, 99))
  expect_equal(min(f), 99)
  
  # specifying max.depth too small, existing data preserved
  f$bottom[c(3,7)] <- NA
  d <- repairMissingHzDepths(f, max.depth = 20)
  expect_true(all(checkHzDepthLogic(d)$valid))
  expect_equal(d$bottom[c(3, 7, 8, 9)], c(30, 40, 10, 99))
  expect_equal(min(d), 30)
  
  # specifying max.depth too small with adj=NA gets an override to max(x), existing data preserved
  f$bottom[c(3,7)] <- NA
  d <- repairMissingHzDepths(f, adj = NA, max.depth = 20)
  expect_true(all(checkHzDepthLogic(d)$valid))
  expect_equal(d$bottom[c(3, 7, 8, 9)], c(99, 99, 10, 99))
  expect_equal(min(d), 99)
})

#' @title Return a vector of contributing fractions over a depth interval
#' @name depthWeights
#' @aliases depthWeights,SoilProfileCollection-method
#' @description \code{depthWeights()} calculates the contributing fraction for each pair of horizon top and bottom depths, given an upper and lower boundary. 
#' @param top A numeric vector of horizon top depths.
#' @param bottom A numeric vector of horizon bottom depths.
#' @param upper A unit length numeric vector with upper boundary.
#' @param lower A unit length numeric vector with lower boundary.
#' @return A named list.
#' @author Andrew G. Brown.
#' 
#' @rdname depthWeights
#' @export depthWeights
depthWeights <- function(top, bottom, upper, lower) {
  if(length(upper) > 1 | length(lower) > 1)
    stop("upper and lower boundary must have length of one")
  if(length(top) != length(bottom))
    stop("lengths of top and bottom depth vectors must match")
  top[top < upper] <- bottom[bottom < upper] <- upper
  top[top > lower] <- bottom[bottom > lower] <- lower
  return((bottom - top) / (lower - upper))
}

#' @title Return a list representation of site and horizon level data
#' @name compositeSPC
#' @aliases compositeSPC,SoilProfileCollection-method
#' @description \code{compositeSPC()} is a convenience function that returns a named list representation of the columns from the \code{@site} and \code{@horizons} slots.
#' @param object A SoilProfileCollection
#' @return A list.
#' @author Andrew G. Brown.
#' 
#' @rdname compositeSPC
#' @export compositeSPC
compositeSPC <- function(object) {
  
  # create composite horizon/site list object to facilitate eval_tidy
  h <- object@horizons
  s <- object@site
  h[[idname(object)]] <- NULL
  
  # convert all factors to character for logical comparison/matching?
  .unfactor <-  function(x) {
    if (is.factor(x))
      return(as.character(x))
    return(x)
  }
  h[] <- lapply(h, .unfactor)
  s[] <- lapply(s, .unfactor)
  
  res <- c(as.list(s), as.list(h))
  return(res)
}

#' @title Find Overlap within a Sequence
#' @description Establish which elements within a vector of horizontal positions overlap beyond a given threshold
#'
#' @param x vector of relative horizontal positions, one for each profile
#' @param thresh threshold defining "overlap", typically < 1
#' 
#' @return unique index to affected (overlapping) elements in `x`
#' 
#' @export
#'
#' @examples 
#' 
#' x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)
#' 
#' findOverlap(x, thresh = 0.5)
#' 
#' @rdname overlapMetrics
#' 
findOverlap <- function(x, thresh) {
  
  .Deprecated('overlapMetrics', msg = 'This function is deprecated, please use the `overLapMetrics()` function instead.')
  
  # all pair-wise distance
  d <- dist(x)
  m <- as.matrix(d)
  
  # diagonal isn't used here
  diag(m) <- NA
  
  # find matrix elements
  idx <- which(m < thresh)
  
  # use upper-triangle indexes to find elements in original vector
  # only uniquely affected elements
  col.idx <- unique(col(m)[idx])
  
  return(col.idx)
}



#' @title Find and Quantify Overlap within a 1D Sequence
#' 
#' @description Desc.
#' 
#' @param x vector of relative horizontal positions, one for each profile
#' @param thresh threshold defining "overlap", typically < 1
#' 
#'  @return a `list`:
#'   * `idx`: unique index to overlapping elements in `x`
#'   * `ov`: normalized overlap (see details)
#' 
#' @export
#'
#' @examples 
#' 
#' x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)
#' 
#' overlapMetrics(x, thresh = 0.5)
#' 
#' 
overlapMetrics <- function(x, thresh) {
  
  
  ## TODO: 
  # convert to diff(x) vs. dist(x)
  
  # all pair-wise distance
  d <- dist(x)
  m <- as.matrix(d)
  
  # diagonal isn't used here
  diag(m) <- NA
  
  # find matrix elements
  idx <- which(m < thresh)
  
  # use upper-triangle indexes to find elements in original vector
  # only uniquely affected elements
  col.idx <- unique(col(m)[idx])
  
  # overlap = (thresh - distance[i,j]) when d < thresh, otherwise overlap = 0
  # using full matrix, elements are mirrored over diagonal so divide by 2
  ov <- sum(thresh - m[idx]) / 2
  
  # normalize overlap by dividing by total possible overlap
  # all elements overlapping results in values > 1
  # ov_norm = ov / thresh * length(x)
  ov <- ov / (thresh * length(x))
  
  res <- list(
    idx = col.idx,
    ov = ov
  )
  
  return(res)
}




## TODO: incorporate ideas from N-body simulation
# https://en.wikipedia.org/wiki/N-body_simulation


#' @title Simulation of electrostatic force between two charged particles
#' @description This function computes a "force" (attraction or repulsion) between two charged "particles" (usually labels or other graphical elements), using a modification of the 1D electrostatic force equation. This function is used internally for label placement in the presence of overlap, as in [fixOverlap()].
#'
#' @param Q1 numeric, charge on particle 1 (e.g. a label)
#' @param Q2 numeric, charge on particle 2 (e.g. another label)
#' @param Qk numeric, empirical constant
#' @param d numeric, distance between particles
#' @param tiny numeric, number used to represent very small distances (avoid division by 0)
#' @param ex numeric, exponent used in distance-weighting
#' @param const numeric, constant used in distance weighting function, increase to dampen oscillation 
#' 
#' @return numeric
#'
#' @author D.E. Beaudette and K.C. Thompson
#' 
#' @noRd
#' 
#' @seealso [fixOverlap()]
#' 
#' @examples 
#' 
#' ## d is vector
#' # 0.4
#' aqp:::.electricForce(Q1 = 1, Q2 = 1, Qk = 0.5, d = 1)
#' 
#' # 0.4, 0.118
#' aqp:::.electricForce(Q1 = 1, Q2 = 1, Qk = 0.5, d = c(1, 2))
#' 
#' ## varying charge
#' # 0.2, 0.0588
#' aqp:::.electricForce(Q1 = 1, Q2 = 0.5, Qk = 0.5, d = c(1, 2))
#'
#'
#' # d is a square matrix of pair-wise displacement
#' m <- matrix(
#' c(NA, 1, 1, 
#' 1, NA, 1,
#' 1, 1, NA), 
#' nrow = 3, byrow = TRUE,
#' )
#' 
#' # constant charge
#' aqp:::.electricForce(Q1 = 1, Q2 = 1, Qk = 0.5, d = m)
#' 
#' # varying charge
#' aqp:::.electricForce(Q1 = c(1, 1, 2), Q2 = c(1, 1, 2), Qk = 0.5, d = m)
#'
#' @param Q1 numeric vector of point charges (length of 1, `length(d)`, or `ncol(d)`)
#' @param Q2 numeric vector of point charges (length of 1, `length(d)`, or `ncol(d)`)
#' @param Qk numeric scalar, charge constant
#' @param d numeric vector of length >=1, or numeric, square matrix of displacement values
.electricForce <- function(Q1, Q2, Qk, d, tiny = 0.0001, ex = 2, const = 0.25) {
  
  
  # sanity checks
  if(length(Q1) != length(Q2)) {
    stop('charge vectors must be of the same length', call. = FALSE)
  }
  
  if(is.matrix(d)) {
    
    if(length(Q1) > 1 & (ncol(d) != length(Q1))) {
      stop('charge vectors must be of length ncol(d)', call. = FALSE)
    }
    
    if(ncol(d) != nrow(d)) {
      stop('displacement matrix must be square', call. = FALSE)
    }
  }
  
  
  # fix for 0-distance where force is infinite
  d <- ifelse(d < tiny, tiny, d)
  
  # traditional definition of electrostatic force
  # (Qk * Q1 * Q2) / d^2
  
  # modified version, c/o K.C. Thompson
  # increase const --> dampen chaotic oscillation during simulation 
  # "softening" in N-body simulation
  # (Qk * Q1 * Q2) / (d^ex + const)
  
  
  # when d is a numeric vector, use simple calculation
  if(is.vector(d)) {
    res <- (Qk * Q1 * Q2 ) / (d^ex + const)
    return(res)
  }
  
  # generalize to 
  #  * Q1 and Q2 as scalars or vectors
  #  * d is a square matrix of pair-wise displacement
  
  # factor out point charges
  res.1 <- (Qk) / (d^ex + const)
  
  # sweep product of point charges across lower triangle
  res.2 <- sweep(res.1, MARGIN = 1, STATS = Q1 * Q2, FUN = '*')
  
  # sweep product of point charges across upper triangle
  res.3 <- sweep(res.1, MARGIN = 2, STATS = Q1 * Q2, FUN = '*')
  
  # init new result matrix, diagonals are NA
  res <- res.1
  
  # transfer lower and upper triangles
  res[lower.tri(res, diag = FALSE)] <- res.2[lower.tri(res.2, diag = FALSE)]
  res[upper.tri(res, diag = FALSE)] <- res.3[upper.tri(res.3, diag = FALSE)]
  
  return(res)
}




#' @title Label placement based on a simulation of electrostatic forces
#' 
#' @description This function attempts to move labels along a 1D coordinate system such that overlap (as specified by threshold) is minimized. An electrostatic simulation applies forces of repulsion between labels that are within `thresh` (e.g. overlapping) and forces of attraction to a uniformly spaced sequence to iteratively perturb affected labels until either no overlap is reported, or a maximum number of iterations (`maxIter`) has been reached.
#' 
#' @details 
#' 
#' Difficult overlap problems can be addressed by reducing `thresh` and increasing `q`. Large values of `q` can lead to chaotic results.
#' 
#' This function will generate unpredictable output when `x` contains duplicate values.
#' 
#' This function requires input to be pre-sorted, although interesting "artistic" simulations will often result from unsorted `x`.
#'
#' @param x numeric vector, pre-sorted sorted, without duplication, describing 1D label (particle) configuration
#' 
#' @param thresh numeric, overlap threshold, same as in [fixOverlap()]
#' 
#' @param q numeric, electrical charge (typically between 0.1 and 2)
#' 
#' @param chargeDecayRate numeric, exponential decay rate constant for `q` as a function of iteration `i`
#' 
#' @param QkA_GrowthRate numeric, growth rate constant for `Qk` applied to attraction to uniform spacing of labels, invoked when rank order is violated during the simulation
#' 
#' @param maxIter integer, maximum number of iterations before giving up
#' 
#' @param tiny numeric, 0-values replaced by this number to avoid division by 0 and infinite forces
#' 
#' @param const numeric, empirical constant added to the 1D electrostatic force equation to dampen oscillation: `(Qk * Q1 * Q2) / (d^ex + const)`
#' 
#' @param trace logical, include diagnostic output
#' 
#' @param \dots not used, absorbs additional arguments to [fixOverlap()]
#' 
#' @author D.E. Beaudette and K.C. Thompson
#'
#' @return When `trace = TRUE` a `list`, otherwise numeric vector with `converged` attribute.
#' 
#' @seealso [fixOverlap()], [SANN_1D()]
#' 
#' @export
#' 
#' @examples 
#' 
#' # vector of object locations, with potential overlap
#' x <- c(1, 2, 3, 3.3, 3.8, 5, 6, 7, 8, 9, 10)
#' 
#' # full diagnostic output
#' z <- electroStatics_1D(x, thresh = 0.65, trace = TRUE, q = 1)
#' txt <- sprintf("Converged %s (%s iterations)", z$converged, length(z$cost))
#' 
#' plot(
#' seq_along(z$cost),
#' z$cost, 
#' las = 1, 
#' xlab = 'Iteration', 
#' ylab = 'Overlap Cost', 
#' type = 'b', 
#' main = txt
#' )
#' 
#' abline(h = 0, lty = 2, col = 2)
#' 
#' # final configuration only
#' xnew <- electroStatics_1D(x, thresh = 0.65, q = 1)
#' 
#' # check for convergence
#' attr(xnew, 'converged')
#' 
#' # compare original vs. modified
#' data.frame(orig = x, new = round(xnew, 2))
#' 

## TODO: what is a reasonable starting value for q?
##       -> too low, not enough perturbation, does not converge
##       -> too high, chaos

## TODO: allow for ties / un-sorted input

## TODO: implement domain[min, max] argument for convenient anchors
# https://github.com/ncss-tech/aqp/issues/288


electroStatics_1D <- function(x, thresh, q = 1, chargeDecayRate = 0.01, QkA_GrowthRate = 0.05, maxIter = 100, tiny = 0.0001, const = 0.001, trace = FALSE, ...) {
  
  # keep track of original range
  .original_range <- range(x)
  
  # re-scale to 0,1
  x <- .rescaleRange(x, 0, 1)
  
  # apply scaling factor to threshold
  thresh <- thresh / abs(diff(.original_range))
  
  # original configuration
  x.orig <- x
  x.n <- length(x)
  r <- range(x)
  
  # storage for new configurations / total force / const
  xnew <- list()
  F_total <- rep(NA_real_, times = maxIter)
  cost <- rep(NA_real_, times = maxIter)
  
  ## constants used to control effect of force on charged particles
  # mass
  .m <- 100
  
  # time step
  .t <- 1
  
  # initial attractive force constant (Qk) to uniform spacing
  .Fu <- 1.5e-3
  
  # repulsion force constant
  .Qk <- 1e-2
  
  # exponential decay schedule for particle charge
  # from t = 1 -> maxIter
  # q can be either scalar or vector of length x.n
  # safely compute product by elements of q
  qe <- outer(q, exp(-1 * chargeDecayRate * 1:maxIter), FUN = '*')
  
  # simulation
  for(i in 1:maxIter) {
    
    # compute initial overlap metrics at threshold
    .om <- overlapMetrics(x, thresh = thresh)
    
    # constraints:
    # rank order, by element
    .rank_test <- rank(x) == rank(x.orig)
    # overlap test
    .overlap_test <- length(.om$idx) < 1
    
    # stop simulation if there is nothing left to do
    if(all(.rank_test) & .overlap_test) {
      xnew[[i]] <- x
      break
    }
    
    # if any rank order is violated
    # progressively increase Qk for attractive force to uniform spacing
    # this will progressively pull particles into the "right" order
    if(!any(.rank_test)) {
      .Fu <- .Fu + (.Fu * QkA_GrowthRate)
    }
    
    # pair-wise distances
    m <- as.matrix(dist(x))
    
    # remove distance to self
    diag(m) <- NA
    
    
    ## TODO: consider given higher charge density to left/right anchors
    ## https://github.com/ncss-tech/aqp/issues/293
    ## 2025-12-30: now possible with q specified as a vector
    
    # repelling forces (same charge) between all particles
    .F <- .electricForce(Q1 = q, Q2 = q, Qk = .Qk, d = m, tiny = tiny, const = const)
    
    # negative forces are to the left
    # lower triangle is used for particles to the right of any given position
    .F[lower.tri(.F)] <- - .F[lower.tri(.F)]
    
    # net repelling force on each particle
    # force vector: negative <--- | ---> positive
    .F_repl <- colSums(.F, na.rm = TRUE)
    
    # attractive forces between each particle <--> uniform spacing, rank order
    # weaker than repelling forces
    
    # distance from uniform spacing
    .offset <- x - seq(from = r[1], to = r[2], length.out = x.n)
    
    # note: sign(0) --> 0
    # direction is based on offset vector
    .direction <- sign(.offset)
    
    # force vector: negative <--- | ---> positive
    # weaker than repelling forces, set via Qk arg
    .F_attr <- .electricForce(Q1 = q, Q2 = q, Qk = .Fu, d = abs(.offset), tiny = tiny, const = const)
    
    # sum attractive + repelling forces
    .F_net <- (.direction * .F_attr) + .F_repl
    
    # ## debugging
    # print(
    #   rbind(x, x.orig, .direction, .F_repl, .F_attr, .F_net)
    # )
    
    # displacement vector: d = 1/2 F/m * t
    # negative <--- | ---> positive
    .d <- 1/2 * (.F_net / .m) * .t^1
    
    # displacement is only applied to overlapping AND out-of-order points
    .affected_points <- unique(c(.om$idx, which(!.rank_test)))
    .ignore <- setdiff(seq_along(x), .affected_points)
    .d[.ignore] <- 0
    
    # displacement of boundary points is always 0
    .d[1] <- 0
    .d[x.n] <- 0
    
    # keep track of new locations at time step i
    # displacement can't be outside of bounds
    xnew[[i]] <- pmin(pmax(x + .d, x.orig[1]), x.orig[x.n])
    
    # keep track of total force in system at time step i
    F_total[i] <- sum(abs(.F_net))
    
    # update locations
    x <- xnew[[i]]
    
    # apply exponential decay of all charges with each iteration
    # iterations are stored as columns
    # individual charges, or single charge as rows
    q <- qe[, i]
    
    # compute overlap metrics at threshold, after displacement
    .om <- overlapMetrics(x, thresh = thresh)
    cost[i] <- .om$ov
    
    # stop simulation if there is no overlap in this iteration
    # AND rank order is preserved
    if(all(rank(x) == rank(x.orig)) & length(.om$idx) < 1) {
      break
    }
    
  }
  
  # done with iterations
  message(sprintf("%s iterations", i))
  
  # flatten
  xnew <- do.call('rbind', xnew)
  
  ## TODO: optionally return to lowest cost configuration
  ## TODO: optionally re-run with lower / higher q
  
  # check for convergence
  # 1. no change in rank order
  # 2. no overlap
  .converged <- all(rank(xnew[nrow(xnew), ]) == rank(x.orig) & length(.om$idx) < 1)
  
  if(trace) {
    # solutions(iteration) are re-scaled to original range
    xnew <- t(apply(xnew, 1, .rescaleRange, .original_range[1], .original_range[2]))
    
    # compile full results
    .res <- list(
      x = as.vector(xnew[nrow(xnew), ]),
      F_total = as.vector(na.omit(F_total)), 
      cost = as.vector(na.omit(cost)),
      states = xnew, 
      converged = .converged
    )
  } else {
    # only the final configuration
    # solution is re-scaled to original range
    .res <- .rescaleRange(as.vector(xnew[nrow(xnew), ]), .original_range[1], .original_range[2])
    attr(.res, 'converged') <- .converged
  }
  
  
  return(.res)
}





## background:
# https://en.wikipedia.org/wiki/Simulated_annealing
# https://www.r-bloggers.com/2014/09/the-traveling-salesman-with-simulated-annealing-r-and-shiny/
# http://umsl.edu/~adhikarib/cs4130-fall2017/slides/11%20-%20The%20Simulated%20Annealing%20Algorithm.pdf
#

## TODO:
# * enforce rank in degenerate cases
# * secondary objective function: as close as possible to original configuration
# * cleanup variable names: stats -> overlap, log -> x.log, etc.

#' @title Fix Overlap within a Sequence via Simulated Annealing
#' 
#' @description This function makes small adjustments to elements of `x` until overlap defined by `thresh` is removed, or until `maxIter` is reached. Rank order and boundary conditions (defined by `min.x` and `max.x`) are preserved. The underlying algorithm is based on simulated annealing. The "cooling schedule" parameters `T0` and `k` can be used to tune the algorithm for specific applications.
#' 
#' @details Ideas for solving difficult overlap scenarios:
#'   * widen the boundary conditions by adjusting `min.x` and `max.x` beyond the original scale of `x`
#'   * reduce the allowable overlap threshold `thresh`
#'   
#'   * reduce the magnitude of perturbations (`adj`) and increase `maxIter`
#'   
#'   * increase `k`
#' 
#' @param x vector of horizontal positions, pre-sorted
#' 
#' @param thresh horizontal threshold defining "overlap" or distance between elements of `x`. For adjusting soil profile sketches values are typically < 1 and likely in (0.3, 0.8).
#' 
#' @param adj specifies the size of perturbations within `runif(min = adj * -1, max = adj)`. Larger values will sometimes reduce the number of iterations required to solve particularly difficult overlap conditions. See `coolingRate` argument when `adj` is large
#' 
#' @param min.x left-side boundary condition, consider expanding if a solution cannot be found within `maxIter`.
#' 
#' @param max.x right-side boundary condition, consider expanding if a solution cannot be found within `maxIter`.
#' 
#' @param maxIter maximum number of iterations to attempt before giving up and returning a regularly-spaced sequence
#' 
#' @param trace print diagnostics, result is a `list` vs `vector`
#' 
#' @param tiny the smallest allowable overlap
#' 
#' @param T0 starting temperature
#' 
#' @param k cooling constant
#' 
#' @param \dots not used, absorbs additional arguments to [fixOverlap()]
#' 
#' @return When `trace = FALSE`, a vector of the same length as `x`, preserving rank-ordering and boundary conditions. When `trace = TRUE` a list containing the new sequence along with information about objective functions and decisions made during iteration.
#' 
#' @author D.E. Beaudette and K.C. Thompson
#' @export
#'
#' @seealso [electroStatics_1D()], [fixOverlap()]
#'
#' @examples 
#' 
#' x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)
#' 
#' # easy
#' z <- fixOverlap(x, thresh = 0.2, trace = TRUE)
#'
#' # harder
#' z <- fixOverlap(x, thresh = 0.6, trace = TRUE)
#' 
#' # much harder
#' z <- fixOverlap(x, thresh = 0.9, trace = TRUE)
#'
#'
#' # interpret `trace` output
#' 
#' # relatively challenging
#' x <- c(1, 2, 3.4, 3.4, 3.4, 3.4, 6, 8, 10, 12, 13, 13, 15, 15.5)
#' 
#' # fix overlap, return debugging information
#' set.seed(10101)
#' z <- fixOverlap(x, thresh = 0.8, trace = TRUE)
#' 
#' # setup plot device
#' par(mar = c(4, 4, 1, 1))
#' layout(matrix(c(1,2,3)), widths = 1, heights = c(1,1,2))
#' 
#' # objective function = overlap + SSD
#' plot(
#'   seq_along(z$stats), z$stats, 
#'   type = 'h', las = 1,
#'   xlab = 'Iteration', ylab = 'Overlap',
#'   cex.axis = 0.8
#' )
#' 
#' # SSD: deviation from original configuration 
#' plot(
#'   seq_along(z$ssd), z$ssd, 
#'   type = 'h', las = 1,
#'   xlab = 'Iteration', ylab = 'Deviation',
#'   cex.axis = 0.8
#' )
#' # adjustments at each iteration
#' matplot(
#'   z$states, type = 'l', 
#'   lty = 1, las = 1, 
#'   xlab = 'Iteration', ylab = 'x-position'
#' )
#' 
#' # trace log
#' # B: boundary condition violation
#' # O: rank (order) violation
#' # +: accepted perturbation
#' # -: rejected perturbation
#' table(z$log)
#' 
SANN_1D <- function(x, thresh = 0.6, adj = thresh * 2/3, min.x = min(x) - 0.2, max.x = max(x) + 0.2, maxIter = 1000, trace = FALSE, tiny = 0.0001, T0 = 500, k = 10, ...) {
  
  
  ## TODO: convert min.x and max.x to domain
  
  
  # system energy ~ probability ~ metropolis step
  # energy / cost function
  # these are all length-1 vectors
  # n0: starting cost 
  # n1: resulting cost adjustment i
  # Te: temperature 
  # k: cooling constant (empirically determined)
  .P <- function(n0, n1, Te, k = 1) {
    if(n1 < n0) {
      return(1)
    } else {
      # delta-E: n1 - n0
      return(exp(-(n1 - n0) / Te * k))
    }
  }
  
  
  ## TODO: this will violate original ranks... don't do this
  
  # sanity check: cannot have perfect overlap (duplicates) in the initial configuration
  # jitter usually will resolve the problem
  if(any(table(x) > 1)) {
    x <- jitter(x)
    if(trace) {
      message('duplicates in `x`, applying jitter')
    }
  }
  
  ## TODO: consider a cost related to the preservation of "character", relative clustering
  # initial value = 0
  # deviation <- 1 - cor(dist(x), dist(x.test), method = 'spearman')
  
  ## TODO: this may not actually be a good "cost" metric
  # SSD: sum squared differences between two configurations (via distance matrix)
  # worst-possible SSD due to regular sequence
  ssd.max <- sum((dist(x) - dist(seq(from = 1, to = length(x), by = 1)))^2)
  
  # initial configuration
  m <- overlapMetrics(x, thresh)
  
  # initial cost
  # at this point SSD = 0
  cost <- m$ov
  
  # save original for testing rank order
  x.orig <- x
  
  # original adjustment value
  adj.orig <- adj
  
  # counter to prevent run-away while-loop
  i <- 1
  
  ## trace details
  # overlap cost (total overlap)
  stats <- rep(NA, times = maxIter)
  
  # deviation from original configuration
  # sum of squared differences between dist(x.orig), dist(x.new)
  ssd <- rep(NA, times = maxIter)
  
  # algorithm adjustment steps:
  # B: boundary violation
  # O: ordering (rank) violation
  # +: accept adjustments
  # -: reject adjustments
  log <- rep(NA, times = maxIter)
  
  # states
  states <- matrix(data = NA, nrow = maxIter, ncol = length(x))
  
  # short-circuit: only proceed if there is overlap
  if(m$ov <  tiny) {
    return(x)
  }
  
  # continue while total overlap > small number
  while(m$ov > tiny) {
    
    # fail-safe
    if(i > maxIter) {
      message('maximum number of iterations reached, using regular sequence')
      s <- seq(from = min(x.orig), to = max(x.orig), length.out = length(x.orig))
      
      if(trace) {
        
        log <- factor(as.vector(na.omit(log)), levels = c('B', 'O', '+', '-'))
        stats <- as.vector(na.omit(stats))
        ssd <- as.vector(na.omit(ssd))
        
        states <- na.omit(states)
        attr(states, "na.action") <- NULL
        
        return(list(
          x = s,
          stats = stats,
          ssd = ssd,
          log = log,
          converged = FALSE,
          states = states
        ))
      }
      attr(s, 'converged') <- FALSE
      return(s)
    }
    
    # generate random perturbations to affected indices
    perturb <- runif(n = length(m$idx), min = adj * -1, max = adj)
    
    # attempt perturbation
    x.test <- x
    x.test[m$idx] <- x.test[m$idx] + perturb
    
    # re-evaluate metrics
    m.test <- overlapMetrics(x.test, thresh)
    
    ## TODO: consider correlation vs. SSD
    # deviation <- 1 - cor(dist(x), dist(x.test), method = 'spearman')
    
    # SSD
    ssd.test <- sum((dist(x) - dist(x.test))^2)
    # normalize
    ssd.test <- ssd.test / ssd.max
    
    # combined cost: overlap + SSD
    cost.test <- m.test$ov + ssd.test
    
    # enforce boundary conditions
    if(any(x.test < min.x) | any(x.test > max.x)) {
      # print('boundary condition')
      log[i] <- 'B'
      stats[i] <- m.test$ov
      ssd[i] <- ssd.test
      states[i, ] <- x.test
      i <- i + 1
      next
    }
    
    # enforce rank ordering
    if(any(rank(x.orig) != rank(x.test))) {
      # print('rank violation')
      log[i] <- 'O'
      stats[i] <- m.test$ov
      ssd[i] <- ssd.test
      states[i, ] <- x.test
      i <- i + 1
      next
    }
    
    
    # compute current temperature
    # differs from literature where i starts from 0
    Temp <- T0 / i
    
    # acceptance probability
    # n0 = previous cost
    # n1 = current cost
    # Te = current temperature
    # k = cooling constant
    p <- .P(n0 = cost, n1 = cost.test, Te = Temp, k = k)
    
    # accept a more costly proposition if randomly selected
    p.acc <- p > runif(n = 1, min = 0, max = 1)
    
    if( (cost.test < cost) | p.acc) {
      # keep new state
      log[i] <- '+'
      
      # apply perturbation to working copy
      x <- x.test
      
      # save state
      states[i, ] <- x
      
      # keep track of overlap cost
      stats[i] <- m.test$ov
      
      # SSD
      ssd[i] <- ssd.test
      
      # re-evaluate overlap for while() loop
      m <- overlapMetrics(x, thresh)
      
      # re-compute total cost
      cost <- m$ov + ssd.test
      
      # increment iteration counter
      i <- i + 1
    } else {
      # reject proposed state
      log[i] <- '-'
      
      # save state
      states[i, ] <- x.test
      
      # keep track of overlap cost
      stats[i] <- m.test$ov
      
      # SSD
      ssd[i] <- ssd.test
      
      i <- i + 1
      next
    }
    
    
  }
  
  
  # done with iterations
  message(sprintf("%s iterations", i))
  
  # full output
  if(trace) {
    
    log <- factor(as.vector(na.omit(log)), levels = c('B', 'O', '+', '-'))
    stats <- as.vector(na.omit(stats))
    ssd <- as.vector(na.omit(ssd))
    
    states <- na.omit(states)
    attr(states, "na.action") <- NULL
    
    return(list(
      x = x,
      stats = stats,
      ssd = ssd,
      log = log,
      converged = TRUE,
      states = states
    ))
  } else {
    # best configuration only
    attr(x, 'converged') <- TRUE
    return(x)
  }
  
}





#' @title Fix Overlap within a Sequence
#'
#' @param x vector of initial positions, pre-sorted
#' @param thresh numeric, overlap threshold defined on the same scale as `x`
#' @param method character vector, 'S' for simulated annealing via [SANN_1D()] or 'E' for electrostatic simulation via [electroStatics_1D()]
#' @param trace logical, return full output
#' @param \dots additional arguments to [SANN_1D()] or [electroStatics_1D()]
#'
#' @return When `trace = FALSE`, a vector of the same length as `x`, preserving rank-ordering and boundary conditions. When `trace = TRUE` a list containing the new sequence along with information about objective functions and decisions made during adjustment of `x`.
#' 
#' @export
#' 
#' @seealso [electroStatics_1D()], [SANN_1D()]
#' 
#' @examples
#' 
#' s <- c(1, 2, 2.3, 4, 5, 5, 7)
#' 
#' # simulated annealing, solution is non-deterministic
#' fixOverlap(s, thresh = 0.6, method = 'S')
#' 
#' # electrostatics-inspired simulation of particles
#' # solution is deterministic
#' fixOverlap(s, thresh = 0.6, method = 'E')
#' 
#'
#' # create a very busy profile with lots of possible overlapping
#' # depth annotation
#' x <- quickSPC(
#'   "SPC:AAA|BBB|CCC|D|EEEEE|FF|GG|HH|I|I|JJ|KK|LL|M|N|O|P|QQQQ|RR|S|TTTTTT|U", 
#'   interval = 1
#' )
#' 
#' # convert horizon ID to numeric
#' x$z <- as.numeric(x$hzID)
#' 
#' # plotSPC arguments
#' .a <- list(
#'   width = 0.2, 
#'   hz.depths = TRUE, 
#'   name.style = 'center-center', 
#'   cex.names = 1.5, 
#'   depth.axis = FALSE, 
#'   name = NA,
#'   color = 'z',
#'   show.legend = FALSE,
#'   print.id = FALSE,
#'   col.palette = hcl.colors(n = 25, palette = 'Spectral', rev = TRUE)
#' )
#' 
#' # set plotSPC default arguments
#' options(.aqp.plotSPC.args = .a)
#' 
#' # wrapper function to test label collision solutions
#' testIt <- function(x, ...) {
#'   
#'   plotSPC(x, ...)
#'   
#'   # a normalized index of label adjustment
#'   .txt <- sprintf(
#'     "LAI: %0.3f", 
#'     get('last_spc_plot', envir = aqp.env)$hz.depth.LAI
#'   )
#'   mtext(.txt, side = 1, at = 1, line = -2, cex = 0.8)
#'   
#' }
#' 
#' 
#' # compare and contrast
#' op <- par(mar = c(0, 0, 0, 0), mfcol = c(1, 6))
#' 
#' testIt(x)
#' title('ES (defaults)', line = -3)
#' 
#' testIt(x, fixOverlapArgs = list(method = 'S'))
#' title('SANN (defaults)', line = -3)
#' 
#' testIt(x, fixOverlapArgs = list(method = 'E', q = 1.5))
#' title('ES (q = 1.5)', line = -3)
#' 
#' testIt(x, fixOverlapArgs = list(method = 'E', q = 1))
#' title('ES (q = 1)', line = -3)
#' 
#' testIt(x, fixOverlapArgs = list(method = 'E', q = 0.5))
#' title('ES (q = 0.5)', line = -3)
#' 
#' testIt(x, fixOverlapArgs = list(method = 'E', q = 0.1))
#' title('ES (q = 0.1)', line = -3)
#' 
#' par(op)
#' 
fixOverlap <- function(x, thresh = 0.6, method = c('S', 'E'), trace = FALSE, ...) {
  
  # sanity checks on method
  method <- tolower(match.arg(method))
  
  # check for un-sorted input
  if(!all(x == sort(x))) {
    warning('data should be pre-sorted', call. = FALSE)
  }
  
  
  .res <- switch(method, 
                 'e' = {
                   electroStatics_1D(x = x, thresh = thresh, trace = trace, ...) 
                 },
                 
                 's' = {
                   SANN_1D(x = x, thresh = thresh, trace = trace, ...) 
                 }
  )
  
  return(.res)
  
}

context("find/fixOverlap")

# example sequence with "overlap"
x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)


## TODO: more tests for electrostatic simulation

test_that(".electricForce() works as expected", {
  
  
  
  # vector displacement
  e <- aqp:::.electricForce(Q1 = 1, Q2 = 1, Qk = 0.5, d = 5)
  expect_equal(length(e), 1)
  expect_true(is.vector(e))
  expect_equal(e, 0.01980198)
  
  e <- aqp:::.electricForce(Q1 = 1, Q2 = 1, Qk = 0.5, d = c(2, 5))
  expect_equal(length(e), 2)
  expect_true(is.vector(e))
  expect_equal(e, c(0.11764706, 0.01980198))
  
  e <- aqp:::.electricForce(Q1 = c(1, 1), Q2 = c(1, 1), Qk = 0.5, d = c(2, 5))
  expect_equal(length(e), 2)
  expect_true(is.vector(e))
  expect_equal(e, c(0.11764706, 0.01980198))
  
  
  # pair-wise matrix displacement
  e <- aqp:::.electricForce(Q1 = 1, Q2 = 1, Qk = 0.5, d = matrix(c(NA, 1, 1, NA), nrow = 2))
  expect_true(is.matrix(e))
  expect_equal(dim(e), c(2, 2))
  expect_equivalent(na.omit(unique(as.vector(e))), 0.4)
  
  # equal charges
  e <- aqp:::.electricForce(Q1 = c(1, 1), Q2 = c(1, 1), Qk = 0.5, d = matrix(c(NA, 1, 1, NA), nrow = 2))
  expect_true(is.matrix(e))
  expect_equal(dim(e), c(2, 2))
  expect_equivalent(na.omit(unique(as.vector(e))), 0.4)
  
  
  # pair-wise matrix displacement
  # unequal charges
  m <- matrix(
  c(NA, 1, 1,
  1, NA, 1,
  1, 1, NA),
  nrow = 3, byrow = TRUE,
  )
  
  e <- aqp:::.electricForce(Q1 = c(1, 1, 2), Q2 = c(1, 1, 2), Qk = 0.5, d = m)
  expect_true(is.matrix(e))
  expect_equal(dim(e), c(3, 3))
  expect_equivalent(na.omit(unique(as.vector(e))), c(0.4, 1.6))
  
  
  ## errors
  
  # charge vectors not the same length
  expect_error(
    aqp:::.electricForce(Q1 = 1, Q2 = c(1, 2), Qk = 0.5, d = matrix(c(NA, 1, 1, NA), nrow = 2)), 
    regexp = 'charge vectors must be of the same length'
  )
  
  # charge vectors wrong size for displacement matrix
  expect_error(
    aqp:::.electricForce(Q1 = c(1, 1, 1) , Q2 = c(1, 1, 1), Qk = 0.5, d = matrix(c(NA, 1, 1, NA), nrow = 2)),
    regexp = 'charge vectors must be of length ncol\\(d\\)'
  )
  
  # not square
  expect_error(
    aqp:::.electricForce(Q1 = c(1, 1, 1) , Q2 = c(1, 1, 1), Qk = 0.5, d = matrix(c(NA, 1, 1, 1, 1, NA), nrow = 2)),
    regexp = 'displacement matrix must be square'  
  )
  
})




test_that("overlapMetrics", {
  
  x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)
  
  # no overlap with 0 threshold
  z <- overlapMetrics(x, thresh = 0)
  expect_true(length(z$idx) == 0)
  
  # none yet
  z <- overlapMetrics(x, thresh = 0.1)
  expect_true(length(z$idx) == 0)
  
  # a little
  z <- overlapMetrics(x, thresh = 0.3)
  expect_true(length(z$idx) == 2)
  expect_equal(x[z$idx], c(3.4, 3.5))
  
  # more
  z <- overlapMetrics(x, thresh = 1)
  expect_true(length(z$idx) == 3)
  expect_equal(x[z$idx], c(3, 3.4, 3.5))
  
  # all
  z <- overlapMetrics(x, thresh = 2)
  expect_true(length(z$idx) == 7)
})


test_that("fixOverlap, simple cases", {
  
  # stochastic element, don't test on CRAN
  skip_on_cran()
  
  # basic interface
  z <- fixOverlap(x, thresh = 0.3, min.x = 1, max.x = 10)
  
  # numeric vector
  expect_true(inherits(z, 'numeric'))
  
  # same length
  expect_true(length(z) == length(x))
  
  # row order preserved
  expect_true(all(rank(x) == rank(z)))
  
  # boundary conditions
  expect_true(min(z) >= 1)
  expect_true(max(z) <= 10)
  
})


test_that("electrostatic simulation", {
  
  # simple
  x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)
  z <- fixOverlap(x, thresh = 0.3, method = 'E', q = 1, trace = TRUE)
  
  # list
  expect_true(inherits(z, 'list'))
  
  # should converge
  expect_true(z$converged)
  
  # more difficult
  z <- fixOverlap(x, thresh = 0.7, method = 'E', q = 1, trace = TRUE)
  
  # list
  expect_true(inherits(z, 'list'))
  
  # should converge
  expect_true(z$converged)
  
})



test_that("fixOverlap, trace = TRUE", {
  
  # stochastic element, don't test on CRAN
  skip_on_cran()
  
  # verbose
  z <- fixOverlap(x, thresh = 0.3, trace = TRUE)
  
  # list
  expect_true(inherits(z, 'list'))
  
})

#' @title Shuffle a single soil profile (internal function)
#'
#' @param i `SoilProfileCollection` of length 1
#' @param mode character, shuffling mode
#' @param replace logical, replacement argument to `base::sample()`
#' @param size integer, size argument to `base::sample()`
#'
#' @returns `SoilProfileCollection`, modified from original
#' @noRd
#' 
.shuffleProfile <- function(i, mode, replace, size) {
  
  # horizon metadata
  .nm <- horizonNames(i)
  .hzd <- horizonDepths(i)
  .id <- idname(i)
  
  # this function only needs access to horizon-level data
  .h <- horizons(i)
  
  # original horizon depths / thickness preserved
  # associated data are shuffled
  if(mode == 'data') {
    # static horizon data
    .hs <- .h[, c(.id, .hzd)]
    
    # dynamic horizon data -> these will be shuffled
    .hd <- .h[, which(! .nm %in% c(.hzd, .id))]
    
    # shuffle 
    # ignore: replace and size arguments
    .idx <- sample(1:nrow(.hd))
    .hd <- .hd[.idx, ]
    
    # combine static + shuffled
    .hnew <- cbind(.hs, .hd)
    
    # replace
    replaceHorizons(i) <- .hnew
  }
  
  # physical horizons (thickness + data) are shuffled
  if(mode == 'horizon') {
    
    # interpret size argument to sample
    if(is.null(size)) {
      size <- nrow(.h)
    }
    
    # shuffle
    # replacement and size arguments subject to constraints on base::sample()
    .idx <- sample(1:nrow(.h), replace = replace, size = size)
    .hnew <- .h[.idx, ]
    
    # initial conditions:
    # thickness
    .thick <- .hnew[[.hzd[2]]] - .hnew[[.hzd[1]]]
    # top-most depth from original data
    .topmost <- min(.h[[.hzd[1]]])
    
    # compute new depths
    .bottom <- cumsum(.thick)
    .top <- c(.topmost, .bottom[-length(.bottom)])
    
    # replace depths
    .hnew[[.hzd[1]]] <- .top
    .hnew[[.hzd[2]]] <- .bottom
    
    replaceHorizons(i) <- .hnew
  }
  
  return(i)
}



#' @title Shuffle Horizons of a SoilProfileCollection
#' 
#' @description
#' This function shuffles the horizon data or physical ordering of horizons within profiles of a `SoilProfileCollection` object.
#' 
#' @param x `SoilProfileCollection`
#' @param mode character, one of 'data' or 'horizon'
#' 
#'  * 'data': shuffle the _data_ associated with physical horizons, making no change to the original horizon thickness and horizon IDs
#'  
#'  * 'horizon': shuffle physical horizons, horizon top and bottom depths are re-calculated
#' 
#' @param replace logical, replacement argument to `base::sample()`, only used when `mode = 'horizon'`
#' 
#' @param size integer, size argument to `base::sample()`, only used when `mode = 'horizon'`
#'
#' @returns `SoilProfileCollection`
#' 
#' @export
#'
#' @examples
#' 
#' data('osd', package = 'aqp')
#' o <- osd
#' 
#' # shuffling of data only
#' o.d <- shuffle(o, mode = 'data')
#' 
#' # shuffling of horizons
#' o.h <- shuffle(o, mode = 'horizon')
#' 
#' # shuffling / sampling with replacement
#' o.h2 <- shuffle(o, mode = 'horizon', replace = TRUE)
#' 
#' # add method to IDs
#' profile_id(o.d) <- sprintf("%s\ndata", profile_id(o.d))
#' profile_id(o.h) <- sprintf("%s\nhz", profile_id(o.h))
#' profile_id(o.h2) <- sprintf("%s\nhz R", profile_id(o.h2))
#' 
#' # combine into single SPC
#' g <- combine(o, o.d, o.h, o.h2)
#' 
#' # graphical comparison
#' op <- par(mar = c(0, 0, 0.5, 2.5))
#' plotSPC(g, name.style = 'center-center', cex.names = 0.66, width = 0.3, cex.id = 0.75)
#' 
#' par(op)
#' 
shuffle <- function(x, mode = c('data', 'horizon'), replace = FALSE, size = NULL) {
  
  # check arguments
  mode <- match.arg(mode)
  
  # TODO: implement via data.table
  
  # apply to all profiles
  .res <- profileApply(x, .shuffleProfile, mode = mode, replace = replace, size = size)
  
  # list -> SPC
  .res <- combine(.res)
  
  return(.res)
}

context("shuffle method for SoilProfileCollection objects")


test_that("basic functionality", {
  
  data('osd', package = 'aqp')
  o <- osd
  
  o.d <- shuffle(o, mode = 'data')
  o.h <- shuffle(o, mode = 'horizon')
  o.h2 <- shuffle(o, mode = 'horizon', replace = TRUE)
  
  expect_true(inherits(o.d, 'SoilProfileCollection'))
  expect_true(inherits(o.h, 'SoilProfileCollection'))
  expect_true(inherits(o.h2, 'SoilProfileCollection'))
  
})

# TODO: incorrect interpretation: .detectColorSpec(c('red', 1, NA))

#' @title Detect color specification from a vector of values, or a matrix of color coordinates
#' @param col character vector, numeric matrix with 3 columns, or data.frame with 3 columns
.detectColorSpec <- function(col) {
  
  # filter NA
  col <- na.omit(col)
  
  # matrix
  if(inherits(col, 'matrix')) {
    
    # trap 0-length input
    if(nrow(col) < 1) {
      return('unknown')
    }
    
    if(ncol(col) == 3) {
      return('color-coordinate-matrix')
    }
  }
  
  # data.frame
  if(inherits(col, 'data.frame')) {
    
    # trap 0-length input
    if(nrow(col) < 1) {
      return('unknown')
    }
    
    if(ncol(col) == 3) {
      return('color-coordinate-data.frame')
    }
  }
  
  # character vectors
  if(inherits(col, 'character')) {
    
    # trap 0-length input
    if(length(col) < 1) {
      return('unknown')
    }
    
    # hex notation of sRGB
    if(all(grepl('#', fixed = TRUE, x = col))) {
      return('hex-sRGB')
    }
    
    # Munsell notation
    # partial notation will return TRUE
    if(! all(is.na(unlist(parseMunsell(col, convertColors = FALSE))))) {
      return('munsell')
    }
    
    # named colors: error on invalid color names
    if(! inherits(try(col2rgb(col), silent = TRUE), 'try-error')) {
      return('named-colors')
    }
    
  }
  
  # otherwise, inconsistent / illegal color specification
  return('unknown')
}




#' @title Convert colors into Munsell Notation
#' 
#' @description Lookup the `n` closest Munsell chips from the `munsell` lookup table from various color notations. This function replaces `rgb2munsell()`.
#'
#' @param col character vector of colors, `data.frame` or `matrix` of color coordinates in sRGB or CIELAB color space
#' 
#' @param space character, one of `sRGB` or `CIELAB`, defines the input color system
#' 
#' @param nClosest integer, number of closest Munsell colors to return (valid range is 1-20)
#'

#'
#' @note This function is fully vectorized and will pad output with NA-records when NA are present in \code{color}.
#' 
#' @author D.E. Beaudette
#' 
#' @references 
#'   * http://ncss-tech.github.io/AQP/
#'   * http://www.brucelindbloom.com/index.html?ColorCalcHelp.html
#'   * http://www.munsellcolourscienceforpainters.com/MunsellAndKubelkaMunkToolbox/MunsellAndKubelkaMunkToolbox.html
#'   * http://www.cis.rit.edu/mcsl/online/munsell.php


#' @return an (NA-padded) `data.frame` containing `hue`, `value`, `chroma`, and CIE delta-E 2000 color contrast metric between source and nearest matching color(s).
#'
#' @export
#'
#' @examples
#' 
#' # vector of named R colors
#' col2Munsell(c('red', 'green', 'blue'))
#' 
#' # sRGB matrix in the range of 0-255
#' col2Munsell(cbind(255, 0, 0))
#' 
#' # sRGB matrix in the range of 0-1
#' col2Munsell(cbind(1, 0, 0))
#' 
#' # 10YR 5/6 in CIELAB
#' col2Munsell(
#'   cbind(51.4337, 9.917916, 38.6889), 
#'   space = 'CIELAB'
#' )
#' 
#' # 2.5YR 6/8 in hex notation
#' col2Munsell("#D18158FF")
#' 
#' # 7.5YR 8/1 in sRGB {0, 1}
#' col2Munsell(
#'   cbind(0.8240707, 0.7856834, 0.7541048)
#' )
#' 
#' # 7.5YR 8/1 in sRGB {0, 255}
#' col2Munsell(
#'   cbind(0.8240707, 0.7856834, 0.7541048) * 255
#' )
#' 
#' # multple colors in CIELAB
#' col2Munsell(
#'   parseMunsell(c('10BG 6/6', '2.5YR 4/6'), returnLAB = TRUE),
#'   space = 'CIELAB'
#' )
#' 
#' # data.frame input
#' col2Munsell(
#'   data.frame(r = 1, g = 0, b = 0),
#'   space = 'sRGB'
#' )
#'
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' # Munsell notation to sRGB triplets {0, 1} 
#' color <- munsell2rgb(
#'   the_hue = c('10YR', '2.5YR', '5YR'), 
#'   the_value = c(3, 5, 2.5), 
#'   the_chroma = c(5, 6, 2), 
#'   return_triplets = TRUE
#' )
#' 
#' # result is a data.frame of sRGB {0, 1}
#' color
#' 
#' # back-transform sRGB -> closest Munsell color
#' # sigma is the dE00 color contrast metric
#' col2Munsell(color, space = 'sRGB')
#'
col2Munsell <- function(col, space = c('sRGB', 'CIELAB'), nClosest = 1) {
  
  # reasonable constraints on n-closest chips
  if(nClosest < 1) {
    message('setting `nClosest to 1`')
    nClosest <- 1
  }
  
  if(nClosest > 20) {
    message('setting `nClosest to 20`')
    nClosest <- 20
  }
  
  # empty DF for NA-padding
  .empty <- data.frame(
    hue = NA, 
    value = NA, 
    chroma = NA, 
    sigma = NA
  )
  
  # sacrifice to CRAN gods
  munsell <- NULL
  
  # note: this is incompatible with LazyData: true
  # load look-up table from our package
  # This should be more foolproof than data(munsell) c/o PR
  load(system.file("data/munsell.rda", package="aqp")[1])
  
  # col can be either character vector or numeric matrix
  
  # a vector implies that it should be a character vector
  if(is.vector(col)) {
    col <- as.character(col)
  }
  
  # convert character vector of colors -> sRGB -> CIELAB
  if(inherits(col, 'character')) {
    
    # empty string should be interpreted as NA
    empty.idx <- which(col == '')
    col[empty.idx] <- NA
    
    # generate index to NA and not-NA
    na.idx <- which(is.na(col))
    not.na.idx <- setdiff(1:length(col), na.idx)
    
    # all NA short-circuit
    if(length(not.na.idx) < 1) {
      res <- .empty[na.idx, ]
      row.names(res) <- NULL
      return(res)
    }
    
    # sRGB matrix [0,1]
    # not NA-safe: NA are converted to [1,1,1]
    col <- t(col2rgb(col) / 255)
    
    # sRGB [0,1] -> CIELAB
    col <- convertColor(col, from = 'sRGB', to = 'Lab', from.ref.white = 'D65', to.ref.white = 'D65')
    
  } else {
    
    # convert to matrix as needed
    if(inherits(col, 'data.frame')) {
      col <- as.matrix(col)
    }
    
    # generate index to NA and not-NA
    na.idx <- which(apply(col, 1, function(i) any(is.na(i))))
    not.na.idx <- setdiff(1:nrow(col), na.idx)
    
    # all NA short-circuit
    if(length(not.na.idx) < 1) {
      res <- .empty[na.idx, ]
      row.names(res) <- NULL
      return(res)
    }
    
    # interpret space argument
    space <- match.arg(space)
    
    # sRGB input
    # convert to CIELAB
    if(space == 'sRGB') {
      
      # detect 0-1 vs. 0-255 sRGB range
      # not NA-safe
      if(range(col[not.na.idx])[2] > 2) {
        col <- col / 255
      }
      
      # sRGB [0,1] -> CIELAB
      col <- convertColor(col, from = 'sRGB', to = 'Lab', from.ref.white = 'D65', to.ref.white = 'D65')
    }
    
    # CIELAB input
    # nothing left to do
  }
  
  
  ## TODO: implement cache, we are often converting from a fixed set of possible colors
  
  # vectorize via for-loop
  n <- nrow(col)
  res <- vector(length = n, mode = 'list')
  
  # iterate over non-NA colors
  # this will leave NULL "gaps"
  for(i in not.na.idx) {
    # convert current color to matrix, this will allow matrix and DF as input
    this.color <- as.matrix(col[i, , drop = FALSE])
    dimnames(this.color)[[2]] <- c('L', 'A', 'B')
    
    # CIE dE00
    # fully-vectorized
    sigma <- farver::compare_colour(
      from = this.color, 
      to = munsell[, 7:9], 
      from_space = 'lab', 
      method = 'CIE2000', 
      white_from = 'D65'
    )
    
    # return the closest n-matches
    idx <- order(sigma)[1:nClosest]
    
    # pack results, sorted by closest results
    res[[i]] <- data.frame(
      munsell[idx, 1:3], 
      sigma = sigma[idx]
    )
    
  }
  
  # fill NULL gaps with empty DF, all NA
  null.idx <- which(sapply(res, is.null))
  if(length(null.idx) > 0) {
    for(i in null.idx){
      res[[i]] <- .empty
    }
  }
  
  
  # convert to DF
  res <- do.call('rbind', res)
  row.names(res) <- NULL
  
  # save sigma units
  attr(res, which = 'sigma') <- 'CIE delta-E 2000'
  
  return(res)
}

## fairly generic, will be required by (at least):"
# * dice() [x]
# * NCSP()
# * slab()
# * spc2mpspline()



#'
#' @title Subset `SoilProfileCollection` Objects or Horizons via `checkHzDepthLogic`
#' 
#' @description This function removes profiles or horizons from a `SoilProfileCollection` that are flagged as having invalid horizon depth logic by [`checkHzDepthLogic`]. Invalid profiles may be created when setting `byhz = TRUE`; use caution as some functions may not work properly in the presence of gaps. Consider using [`fillHzGaps`] to fill these gaps.
#' 
#' @note This function cannot identify (and remove) overlapping horizons when `byhz = TRUE`.
#' 
#' @param x a `SoilProfileCollection` object
#' 
#' @param byhz logical, evaluate horizon depth logic at the horizon level (profile level if `FALSE`)
#' 
#' @return a `SoilProfileCollection` object
#' 
#' @export
#' 
HzDepthLogicSubset <- function(x, byhz = FALSE) {
  
  # additional arguments?
  hz.tests <- checkHzDepthLogic(x, fast = TRUE, byhz = byhz)
  
  # short-circuit: no invalid records, stop here
  if(all(hz.tests$valid)) {
    return(x)
  }
  
  # invalid data filtering modes:
  if(byhz) {
    # profile-level
    message("dropping horizons with invalid depth logic, see `metadata(x)$removed.horizons`")
    
    # locate horizons to keep
    idx <- which(hz.tests$valid)
    
    # test for empty SPC
    if(length(idx) < 1) {
      stop('there are no valid profiles in this collection', call. = FALSE)
    }
    
    # keep track of invalid horizon IDs in @metadata
    bad.ids <- hz.tests[[hzidname(x)]][-idx]
    metadata(x)$removed.horizons <- bad.ids
    
    # perform drop
    # this will trigger an error if SPC is corrupted (site w/o horizons)
    res <- try(
      replaceHorizons(x) <- horizons(x)[idx, ], 
      silent = TRUE
    )
    
    if(inherits(res, 'try-error')) {
      stop('removing horizons with invalid depth logic would corrupt `x`, use `byhz = FALSE`', call. = FALSE)
    }
    
    
  } else {
    # profile-level
    message("dropping profiles with invalid depth logic, see `metadata(x)$removed.profiles`")
    
    # locate profiles to keep
    idx <- which(hz.tests$valid)
    
    # test for empty SPC
    if(length(idx) < 1) {
      stop('there are no valid profiles in this collection', call. = FALSE)
    }
    
    # keep track of invalid profile IDs in @metadata
    bad.ids <- hz.tests[[idname(x)]][-idx]
    metadata(x)$removed.profiles <- bad.ids
    
    # perform drop
    x <- x[idx, ]
  }
  
  return(x)
  
}

context("fillHzGaps")

# sample data
data(sp4)
depths(sp4) <- id ~ top + bottom


# basic functionality
test_that("fillHzGaps", {
  
  # CRAN safe
  
  # introduce logic errors
  idx <- c(2, 6:7, 8, 12)
  sp4$top[idx] <- NA
  
  # remove logic errors / create gaps
  x <- HzDepthLogicSubset(sp4, byhz = TRUE)
  
  # # gaps and now problematic profiles
  # par(mar = c(0, 0, 0, 1))
  # plotSPC(x, width = 0.3, default.color = 'royalblue')
  
  # fill / noflag
  z <- fillHzGaps(x, flag = FALSE)
  
  # fill / flag
  z <- fillHzGaps(x, flag = TRUE)
  
  # result is always this
  expect_true(inherits(z, 'SoilProfileCollection'))
  
  # flag column present
  expect_true('.filledGap' %in% horizonNames(z))
  
  # correct gaps have been filled
  expect_true(all(which(z[['.filledGap']]) == c(2L, 5L, 7L, 8L, 10L, 
                                                13L, 18L, 21L, 24L, 
                                                27L, 33L, 37L)))
  
  # filled horizons IDs start from the original max (30)
  expect_equal(hzID(z), c("1", "31", "3", "4", "32", "5", "33", "34", "9", "35", "10", 
                          "11", "36", "13", "14", "15", "16", "37", "17", "18", "38", "19", 
                          "20", "39", "21", "22", "40", "23", "24", "25", "26", "27", "41", 
                          "28", "29", "30", "42"))
  
  # fill just top to 10 cm
  z2 <- fillHzGaps(x, flag = TRUE, to_top = 10, to_bottom = NULL)
  expect_equal(which(z2[['.filledGap']]), c(2L, 6L, 10L))
  expect_equal(z2[3, 1]$top, 10L)
  
  # fill just bottom to 200 cm
  z3 <- fillHzGaps(x, flag = TRUE, to_top = NULL, to_bottom = 200)
  expect_equal(z3[, , .LAST]$bottom, rep(200L, length(z3)))
  
  # just fill the gaps (no top or bottom)
  z4 <- fillHzGaps(x, flag = TRUE, to_top = NULL, to_bottom = NULL)
  expect_equal(which(z4[['.filledGap']]), c(2L, 9L))

})


test_that("multiple simultaneous arguments", {
  
  # remove 1st horizon for profiles 1:4
  y <- sp4
  idx <- y[,, .FIRST, .HZID]
  replaceHorizons(y) <- horizons(y)[-idx[1:4], ]
  
  # fill gaps AND NA-pad to top/bottom anchors
  z <- fillHzGaps(y, to_top = 0, to_bottom = 75, flag = TRUE)
  
  # result is always this
  expect_true(inherits(z, 'SoilProfileCollection'))
  
  # flag column present
  expect_true('.filledGap' %in% horizonNames(z))
  
  # test padding "up to" 0
  expect_true(min(z$top) == 0)
  
  # test padding "down to" 75
  expect_true(max(z) == 75)
  
})


test_that("ignore perfectly overlapping horizons", {
  
  y <- data.frame(
    id = 'SPC',
    top = c(0, 25, 25, 50, 75, 100, 100),
    bottom = c(25, 50, 50, 75, 100, 125, 125)
  )
  
  depths(y) <- id ~ top + bottom
  
  # fill gaps AND NA-pad to top/bottom anchors
  z <- fillHzGaps(y)
  
  # result is always this
  expect_true(inherits(z, 'SoilProfileCollection'))
  
  # flag column present
  expect_true('.filledGap' %in% horizonNames(z))
  
  # no filling required
  expect_true(all(!z$.filledGap))
  
})

test_that("gap + perfectly overlapping horizons", {
  
  y <- data.frame(
    id = 'SPC',
    top = c(0, 15, 25, 25, 50, 75, 100, 100),
    bottom = c(10, 25, 50, 50, 75, 100, 125, 125)
  )
  
  depths(y) <- id ~ top + bottom
  
  # fill gaps AND NA-pad to top/bottom anchors
  z <- fillHzGaps(y)
  
  # result is always this
  expect_true(inherits(z, 'SoilProfileCollection'))
  
  # flag column present
  expect_true('.filledGap' %in% horizonNames(z))
  
  # single filled horizons
  expect_true(z$.filledGap[2])
  
})

#' Textural conversions
#'
#' These functions consist of several conversions between sand, silt and clay
#' to texture class and visa versa, textural modifiers to rock fragments, and
#' grain size composition to the family particle size class.
#'
#' These functions are intended to estimate missing values or allocate particle
#' size fractions to classes. The \code{ssc_to_texcl()} function uses the same 
#' logic as the particle size estimator calculation in NASIS to classify sand 
#' and clay into texture class. The results are stored in \code{soiltexture} 
#' and used by \code{texcl_to_ssc()} as a lookup table to convert texture class 
#' to sand, silt and clay. The function \code{texcl_to_ssc()} replicates the
#' functionality described by Levi (2017). The \code{texmod_to_fragvol()} 
#' function similarly uses the logical from the 
#' Exhibit618-11_texture_modifier.xls spreadsheet to determine the textural 
#' modifier from the various combinations of rock and pararock fragments 
#' (e.g. GR and PGR).
#'
#' When \code{sample = TRUE}, the results can be used to estimate within-class,
#' marginal distributions of sand, silt, and clay fractions. It is recommended
#' that at least 10 samples be drawn for reasonable estimates.
#' 
#' The function \code{texmod_to_fragvoltot} returns a data.frame with multiple  
#' fragvoltot columns differentiated by tailing abbreviations (e.g. _r) which 
#' refer to the following: 
#' 1. l = low
#' 2. r = representative
#' 3. h = high
#' 4. nopf = no pararock fragments (i.e. total fragments - pararock fragments)
#'
#' The function \code{texture_to_texmod()} parses texture (e.g. GR-CL) to extract the texmod values from it in the scenario where it is missing from
#' texmod column. If multiple texmod values are present (for example in the 
#' case of stratified textures) and \code{duplicates = "combine"} they will be combined in the output (e.g. GR & CBV). Otherwise if \code{duplicates = "max"} 
#' the texmod with the highest rock fragment (e.g. CBV) will be returned.
#'
#' Unlike the other functions, \code{texture_to_taxpartsize()} is intended to
#' be computed on weighted averages within the family particle size control
#' section. Also recall from the criteria that carbonate clay should be
#' subtracted from clay content and added to silt content. Similarly, if the
#' percent of very fine sand is known it should be subtracted from the sand,
#' and added to the silt content. Unlike the other functions,
#' \code{texture_to_taxpartsize()} is intended to be computed on weighted
#' averages within the family particle size control section. Also recall from
#' the criteria that carbonate clay should be subtracted from clay content and
#' added to silt content. Similarly, if the percent of very fine sand is known
#' it should be subtracted from the sand, and added to the silt content.

#' @param texcl vector of texture classes than conform to the USDA code
#' conventions (e.g. c|C, sil|SIL, sl|SL, cos|COS)
#' @param texmod vector of textural modifiers that conform to the USDA code
#' conventions (e.g. gr|GR, grv|GRV)
#' @param lieutex vector of in lieu of texture terms that conform to the USDA
#' code convenctions (e.g. gr|GR, pg|PG), only used when fragments or artifacts
#' are > 90 percent by volume (default: NULL))
#' @param texture vector of combinations of texcl, texmod, and lieutex (e.g. CL, GR-CL, CBV-S, GR)
#'
#' @param clay vector of clay percentages
#' @param sand vector of sand percentages
#' @param fragvoltot vector of rock fragment percentages
#'
#' @param as.is logical: should character vectors be converted to factors?
#' (default: TRUE)
#'
#' @param droplevels logical: indicating whether to drop unused levels in
#' factors. This is useful when the results have a large number of unused
#' classes, which can waste space in tables and figures.

#' @param sample logical: should ssc be random sampled from the lookup table?
#' (default: FALSE)
#' 
#' @return - `texcl_to_ssc`: A `data.frame` containing columns `"sand"`,`"silt"`, `"clay"`
#'
#' @seealso \code{\link{SoilTextureLevels}}
#' 
#' @author Stephen Roecker
#' 
#' @references Matthew R. Levi, Modified Centroid for Estimating Sand, Silt, and Clay from Soil Texture Class, Soil Science Society of America Journal, 2017, 81(3):578-588, ISSN 1435-0661, \doi{10.2136/sssaj2016.09.0301}.
#' 
#' @rdname texture
#' @keywords manip
#' @export
#' @examples
#' \donttest{
#' # example of ssc_to_texcl()
#' tex <- expand.grid(sand = 0:100, clay = 0:100)
#' tex <- subset(tex, (sand + clay) < 101)
#' tex$texcl <- ssc_to_texcl(sand = tex$sand, clay = tex$clay)
#' head(tex)
#'
#' # example of texcl_to_ssc(texcl)
#' texcl <- c("cos", "s", "fs", "vfs", "lcos", "ls",
#'           "lfs", "lvfs", "cosl", "sl", "fsl", "vfsl", "l",
#'           "sil", "si", "scl", "cl", "sicl", "sc", "sic", "c"
#'           )
#' test <- texcl_to_ssc(texcl)
#' head(test <- cbind(texcl, test), 10)
#'
#'
#' # example of texcl_to_ssc(texcl, clay)
#' data(soiltexture)
#' st <- soiltexture$values
#' idx <- sample(1:length(st$texcl), 10)
#' st <- st[idx, ]
#' ssc <- texcl_to_ssc(texcl = st$texcl)
#' head(cbind(texcl = st$texcl, clay = ssc$clay))
#'
#'
#' # example of texmod_to_fragvoltol
#' frags <- c("gr", "grv", "grx", "pgr", "pgrv", "pgrx")
#' head(texmod_to_fragvoltot(frags))
#'
#'
#' # example of texture_to_taxpartsize()
#' tex <- data.frame(texcl = c("c", "cl", "l", "ls", "s"),
#'                   clay  = c(55, 33, 18, 6, 3),
#'                   sand  = c(20, 33, 42, 82, 93),
#'                   fragvoltot = c(35, 15, 34, 60, 91))
#' tex$fpsc <- texture_to_taxpartsize(texcl = tex$texcl,
#'                                    clay = tex$clay,
#'                                    sand = tex$sand,
#'                                    fragvoltot = tex$fragvoltot)
#' head(tex)
#'
#'
#' # example of texture_to_taxpartsize() with carbonate clay and very fine sand
#' carbclay <- rnorm(5, 2, 3)
#' vfs <- rnorm(5, 10, 3)
#' st$fpsc <- texture_to_taxpartsize(texcl = tex$texcl,
#'                                   clay = tex$clay - carbclay,
#'                                   sand = tex$sand - vfs,
#'                                   fragvoltot = tex$fragvoltot)
#' head(tex)
#'
#'
#' # example of sample = TRUE
#' texcl <- rep(c("cl", "sil", "sl"), 10)
#' ssc1 <- cbind(texcl, texcl_to_ssc(texcl = texcl, sample = FALSE))
#' ssc2 <- cbind(texcl, texcl_to_ssc(texcl = texcl, sample = TRUE))
#' ssc1$sample <- FALSE
#' ssc2$sample <- TRUE
#' ssc  <- rbind(ssc1, ssc2)
#' aggregate(clay ~ sample + texcl, data = ssc, summary)
#' }
texcl_to_ssc <- function(texcl = NULL, clay = NULL, sample = FALSE) {
  # fix for R CMD check
  #  texcl_to_ssc: no visible binding for global variable ‘soiltexture’
  soiltexture <- NULL
  
  # clay is not NULL
  clay_not_null <- all(!is.null(clay))
  clay_is_null  <- !clay_not_null
  
  # standardize the inputs
  df <- data.frame(texcl = tolower(as.character(texcl)),
                   stringsAsFactors = FALSE
  )
  if (clay_not_null) {
    df$clay <- as.integer(round(clay))
  }
  df$rn <- row.names(df)
  
  
  load(system.file("data/soiltexture.rda", package="aqp")[1])
  
  
  # convert fine sand classes to their generic counterparts
  df <- within(df, {
    texcl = ifelse(texcl %in% c("cos",  "fs", "vfs"),   "s",  texcl)
    texcl = ifelse(texcl %in% c("lcos", "lfs", "lvfs"), "ls", texcl)
    texcl = ifelse(texcl %in% c("cosl", "fsl", "vfsl"), "sl", texcl)
  })
  
  
  
  # check for texcl that don't match
  idx <- ! df$texcl %in% unique(soiltexture$averages$texcl)
  if (any(idx)) {
    warning("not all the user supplied texcl values match the lookup table")
  }
  
  
  # check clay values within texcl
  if (clay_not_null) {
    clay_not_na <- !is.na(df$clay)
    
    idx <- paste(df$texcl[clay_not_na], df$clay[clay_not_na]) %in% paste(soiltexture$values$texcl, soiltexture$values$clay)
    
    if (any(!idx)) {
      warning("not all the user supplied clay values fall within the texcl, so they will be set to NA")
      
      df$clay[which(!idx)] <- NA
      
      clay_not_null <- all(!is.na(df$clay))
      clay_is_null  <- !clay_not_null
    }
  }
  
  
  # check clay ranges 0-100
  idx <- clay_not_null & any(clay < 0, na.rm = TRUE) & any(clay > 100, na.rm = TRUE)
  if (idx) {
    warning("some clay records < 0 or > 100%")
  }
  
  
  # if clay is present
  if (clay_not_null & sample == FALSE) {
    
    st <- aggregate(sand ~ texcl + clay, data = soiltexture$values, function(x) as.integer(round(mean(x))))
    st$silt <- 100 - st$clay - st$sand
    
    # some clay present (compute clay weighted averages)
    idx <- is.na(df$clay)
    if (any(idx)) {
      df_na <- merge(df[idx, c("texcl", "rn")], soiltexture$averages, by = "texcl", all.x = TRUE, sort = FALSE)[c("texcl", "clay", "rn")]
      df[idx, ] <- df_na
    }
    
    df <- merge(df[c("texcl", "clay", "rn")], st, by = c("texcl", "clay"), all.x = TRUE, sort = FALSE)
  } else {
    # clay missing (use average)
    df <- merge(df[c("texcl", "rn")], soiltexture$averages, by = "texcl", all.x = TRUE, sort = FALSE)
  }
  
  
  # randomly sample ssc from texcl lookup table
  if (sample == TRUE) {
    
    if (clay_is_null) df$clay <- NA
    
    split(df, df$texcl, drop = TRUE) ->.;
    lapply(., function(x) {
      
      st    <- soiltexture$values
      
      # clay present
      x$idx <- is.na(x$clay)
      x1 <- x[x$idx == FALSE, ]
      x2 <- x[x$idx == TRUE,  ]
      
      if (clay_not_null) {
        temp1 <- st[st$texcl == x1$texcl[1] &
                      st$clay %in% unique(x1$clay),
        ]
        temp1 <- temp1[sample(1:nrow(temp1), size = nrow(x1), replace = TRUE), ]
        temp1 <- cbind(x1[x1$idx == FALSE, c("rn", "texcl")], temp1[c("clay", "sand")])
      } else temp1 <- NULL
      
      # clay missing
      temp2 <- st[st$texcl == x2$texcl[1], ]
      temp2 <- temp2[sample(1:nrow(temp2), size = nrow(x2), replace = TRUE), ]
      temp2 <- cbind(x2[x2$idx == TRUE, c("rn", "texcl")], temp2[c("clay", "sand")])
      
      return(rbind(temp1, temp2))
    }) ->.;
    do.call("rbind", .) -> df
    df$silt <- 100 - df$clay - df$sand
  }
  
  
  # standardize outputs
  vars <- c("sand", "silt", "clay")
  df <- df[(order(as.integer(df$rn))), vars]
  df$rn    <- NULL
  df$texcl <- NULL
  rownames(df) <- NULL
  return(df)
}

#' Convert sand, silt and clay to texture class
#' @param simplify Passed to `SoilTextureLevels()` to set the number of possible texture classes. If `TRUE`, the ordered factor has a maximum of 12 levels, if `FALSE` (default) the ordered factor has a maximum of 21 levels (including e.g. very fine/fine/coarse variants)
#' @rdname texture
#' @return  - `ssc_to_texcl`: A `character` vector containing texture class
#' @export
#'
ssc_to_texcl <- function(sand = NULL, clay = NULL, simplify = FALSE, as.is = FALSE, droplevels = TRUE) {
  # fix for R CMD check:
  #  ssc_to_texcl: no visible binding for global variable ‘silt’
  silt <- NULL
  
  # check lengths
  idx <- length(clay) != length(sand)
  if (idx) {
    stop("length of inputs do not match")
  }
  
  
  # standardize inputs
  df <- data.frame(sand = as.integer(round(sand)),
                   clay = as.integer(round(clay)),
                   stringsAsFactors = FALSE
  )
  df$silt <- 100 - df$clay - df$sand
  
  
  ## TODO: this needs some more work: sum will always be 100, but silt-by-difference may be illogical
  
  # # check sand, silt and clay sum to 100
  # idx <- (df$sand + df$silt + df$clay) > 100 | (df$sand + df$silt + df$clay) < 100
  # if (any(idx) & any(complete.cases(df[c("sand", "clay")]))) {
  #   warning("some records sand, silt, and clay do not sum to 100 %")
  #   }
  
  
  # logic from the particle size estimator calculation from NASIS
  df <- within(df, {
    texcl = NA
    texcl[silt >= 79.99 & clay <  11.99] = "si"
    texcl[silt >= 49.99 & clay <  26.99 & (silt < 79.99 | clay >= 11.99)] = "sil"
    texcl[clay >= 26.99 & clay <  39.99 & sand <= 20.01] = "sicl"
    texcl[clay >= 39.99 & silt >= 39.99] = "sic"
    texcl[clay >= 39.99 & sand <= 45.01 & silt <  39.99] = "c"
    texcl[clay >= 26.99 & clay <  39.99 & sand >  20.01 & sand <= 45.01] = "cl"
    texcl[clay >=  6.99 & clay <  26.99 & silt >= 27.99 & silt < 49.99 & sand <= 52.01] = "l"
    texcl[clay >= 19.99 & clay <  34.99 & silt <  27.99 & sand > 45.01] = "scl"
    texcl[clay >= 34.99 & sand >  45.01] = "sc"
    texcl[(silt + 1.5 * clay) < 15] = "s"
    texcl[(silt + 1.5 * clay) >= 15 & (silt + 2 * clay) < 29.99] = "ls"
    texcl[!is.na(sand) & !is.na(clay) & is.na(texcl)] = "sl"
  })
  
  # encoding according to approximate AWC, from Red Book version 3.0
  if (!as.is) {
    df$texcl <- factor(df$texcl, levels = SoilTextureLevels(which = 'codes', simplify = simplify), ordered = TRUE)
    
    if (droplevels) {
      df$texcl <- droplevels(df$texcl)
    }
  }
  
  return(df$texcl)
}

#' Convert from fragment modifier to estimated total volume
#'
#' @param texmod vector of textural modifiers that conform to the USDA code
#' conventions (e.g. gr|GR, grv|GRV)
#'
#' @param lieutex vector of in lieu of texture terms that conform to the USDA
#' code conventions (e.g. gr|GR, pg|PG), only used when fragments or artifacts
#' are > 90 percent by volume (default: NULL))
#' @rdname texture
#' 
#' @return  - `texmod_to_fragvoltot`: A `data.frame` containing columns `"fragvoltot_l"`,
#' `"fragvoltot_r"`, `"fragvoltot_h"`, `"fragvoltot_l_nopf"`, `"fragvoltot_r_nopf"`, `"fragvoltot_h_nopf"`
#' 
#' @export
texmod_to_fragvoltot <- function(texmod = NULL, lieutex = NULL) {
  
  # standardize inputs ----
  vars_l <- list(texmod = texmod, lieutex = lieutex)
  # vars_l <- list(texmod = c("GR", "CBV", NA), lieutex = c(NA, NA, "BY"))
  df <- NULL
  df <- .format_inputs(vars_l, names(vars_l), NA_character_, "character")
  df$texmod  <- tolower(df$texmod)
  df$lieutex <- toupper(df$lieutex)
  df$rn <- 1:nrow(df)
  
  
  # load lookup table
  # fix for R CMD check
  #  texcl_to_ssc: no visible binding for global variable ‘soiltexture’
  soiltexture <- NULL
  load(system.file("data/soiltexture.rda", package="aqp")[1])
  
  
  # check inputs ----
  ## both present?
  idx <- any(complete.cases(df))
  if (idx) {
    warning("texmod and lieutex should not both be present, they are mutually exclusive, only the results for texmod will be returned")
  }
  
  # texmod and lieutex don't match
  idx <- ! df$texmod %in% soiltexture$texmod$texmod
  if (any(idx & !is.na(df$texmod))) {
    message("not all the texmod supplied match the lookup table")
  }
  
  idx <-  ! df$lieutex %in% c("GR", "CB", "ST", "BY", "CN", "FL", "PG", "PCB", "PST", "PBY", "PCN", "PFL", "BR", "HMM", "MPM", "SPM", "MUCK", "PEAT", "ART", "CGM", "FGM", "ICE", "MAT", "W")
  if (any(idx & !is.na(df$lieutex))) {
    message("not all the lieutex supplied match the lookup table")
  }
  
  
  # merge
  df <- merge(df, soiltexture$texmod, by = "texmod", all.x = TRUE, sort = FALSE)
  df <- df[order(df$rn), ]
  df$rn <- NULL
  
  
  # lieutex
  if (any(!is.na(df$lieutex))) {
    
    idx1 <- !is.na(df$lieutex) & df$lieutex %in% c("GR", "CB",  "ST",  "BY",  "CN", "FL")
    idx2 <- !is.na(df$lieutex) & df$lieutex %in% c("PG", "PCB", "PST", "PBY", "PCN", "PFL")
    
    # df$lieutex <- toupper(lieutex)
    
    df <- within(df, {
      fragvoltot_l = ifelse(idx1, 90, fragvoltot_l)
      fragvoltot_r = ifelse(idx1, 95, fragvoltot_l)
      fragvoltot_h = ifelse(idx1, 100, fragvoltot_l)
      
      fragvoltot_l_nopf = ifelse(idx2, 0,  fragvoltot_l)
      fragvoltot_r_nopf = ifelse(idx2, 0,  fragvoltot_l)
      fragvoltot_h_nopf = ifelse(idx2, 0, fragvoltot_l)
    })
    # df$lieutex <- lieutex
  }
  
  df[c("texmod", "lieutex", "texmod_label")] <- NULL
  
  return(df)
}

#' Convert sand, silt and clay to the family particle size class
#'
#' @param texcl vector of texture classes than conform to the USDA code
#' conventions (e.g. c|C, sil|SIL, sl|SL, cos|COS)
#'
#' @param clay vector of clay percentages
#' @param sand vector of sand percentages
#' @param sandvf vector of very fine sand percentages
#'
#' @param fragvoltot vector of total rock fragment percentages
#'
#' @return - `texture_to_taxpartsize`: a character vector containing `"taxpartsize"` classes
#' 
#' @seealso [hz_to_taxpartsize()], [lookup_taxpartsize()]
#' 
#' @rdname texture
#'
#' @export
#'
texture_to_taxpartsize <- function(texcl = NULL, clay = NULL, sand = NULL, sandvf = NULL, fragvoltot = NULL) {
  
  # check lengths
  idx <- length(texcl) == length(clay) & length(clay) == length(sand) & length(sand) == length(fragvoltot)
  if (!idx) {
    stop("length of inputs do not match")
  }
  
  
  # standarize inputs
  if (is.null(sandvf)) sandvf <- NA
  
  df <- data.frame(texcl      = tolower(texcl),
                   clay       = as.integer(round(clay)),
                   sand       = as.integer(round(sand)),
                   sandvf     = as.integer(round(sandvf)),
                   fragvoltot = as.integer(round(fragvoltot)),
                   fpsc       = as.character(NA),
                   stringsAsFactors = FALSE
  )
  df$silt <- 100 - df$sand - df$clay
  
  sandytextures <- c("cos", "s", "fs", "lcos", "ls", "lfs")
  
  
  # check texcl lookup
  idx <- any(! df$texcl[!is.na(df$texcl)] %in% SoilTextureLevels(which = 'codes'))
  if (idx) {
    warning("not all the texcl supplied match the lookup table")
  }
  
  
  # check percentages
  idx <- df$silt > 100 | df$silt < 0 | df$clay > 100 | df$clay < 0 | df$sand > 100 | df$sand < 0 | df$fragvoltot > 100 | df$fragvoltot < 0
  if (any(idx, na.rm = TRUE)) {
    warning("some records are > 100% or < 0%, or the calcuated silt fraction is > 100% or < 0%")
  }
  
  
  if (any(sandvf > sand & all(!is.na(sandvf)))) {
    warning("the sandvf values should not be greater than the sand values")
  }
  
  
  # check ssc_to_texcl() vs texcl
  df$texcl_calc <- suppressMessages(ssc_to_texcl(sand = df$sand, clay = df$clay, as.is = TRUE))
  
  df <- within(df, {
    texcl_calc = ifelse(texcl_calc == "s"  & grepl("^cos$|^fs$|^vfs$",    texcl), texcl, texcl_calc)
    texcl_calc = ifelse(texcl_calc == "ls" & grepl("^lcos$|^lfs$|^lvfs$", texcl), texcl, texcl_calc)
    texcl_calc = ifelse(texcl_calc == "sl" & grepl("^cosl$|^fsl$|^vfsl$", texcl), texcl, texcl_calc)
    
    sandvf = ifelse(is.na(sandvf) & texcl %in% c("vfs", "lvfs"),    50, sandvf)
    sandvf = ifelse(is.na(sandvf) & texcl %in% c("vfsl"),           40, sandvf)
    sandvf = ifelse(is.na(sandvf) & texcl %in% c("fsl"),            15, sandvf)
    sandvf = ifelse(is.na(sandvf) & texcl %in% c("sl", "l", "scl"), 10, sandvf)
    sandvf = ifelse(is.na(sandvf) & texcl %in% c("fsl"),            7,  sandvf)
    
    sand = ifelse(!is.na(sandvf), sand - sandvf,  sand)
    silt = ifelse(!is.na(sandvf), silt + sandvf,  silt)
  })
  
  idx <- any(df$texcl != df$texcl_calc, na.rm = TRUE)
  if (idx) {
    warning("some of the texcl records don't match the calculated texcl via ssc_to_texcl()")
  }
  
  
  # calculate family particle size control section
  
  idx <- df$fragvoltot >= 35
  if (any(idx)) {
    df[idx,] <- within(df[idx,], {
      fpsc[clay >= 35]               = "clayey-skeletal"
      fpsc[clay <  35]               = "loamy-skeletal"
      fpsc[texcl %in% sandytextures] = "sandy-skeletal"
    })
  }
  
  idx <- df$fragvoltot < 35 & df$texcl %in% sandytextures
  if (any(idx)) {
    df[idx, ]$fpsc <- "sandy"
  }
  
  idx <- df$fragvoltot < 35 & ! df$texcl %in% sandytextures
  if (any(idx)) {
    df[idx, ] <- within(df[idx,], {
      fpsc[clay < 18 & sand >= 15] = "coarse-loamy"
      fpsc[clay < 18 & sand <  15] = "coarse-silty"
      fpsc[clay >= 18 & clay < 35] = "fine-loamy"
      fpsc[clay >= 18 & clay < 35 & sand < 15] = "fine-silty"
      fpsc[clay >= 35 & clay < 60] = "fine"
      fpsc[clay > 60] = "very-fine"
    })
  }
  
  df$fpsc <- ifelse(df$fragvoltot > 90, "fragmental", df$fpsc)
  
  return(df$fpsc)
}



#' Parse texmod from texture
#'
#' @param texmod vector of textural modifiers that conform to the USDA code
#' conventions (e.g. gr|GR, grv|GRV)
#' @param texture vector of combinations of texcl, texmod, and lieutex (e.g. CL, GR-CL, CBV-S, GR)
#'
#' @param duplicates character: specifying how multiple values should be handled, options are `"combined"` (e.g. 'GR & GRV) or `"max"`(e.g. 'GRV') 
#'
#' @return - `texture_to_texmod`: a character vector containing `"texmod"` classes
#' 
#' @rdname texture
#' 
#' @export
#'
#' @examples
#' \donttest{
#' # example of texture_to_texmod()
#' tex <- c("SL", "GR-SL", "CBV-L", "SR- GR-FS GRX-COS")
#' texture_to_texmod(tex)
#' texture_to_texmod(tex, duplicates = "max")
#' }
texture_to_texmod <- function(texture, duplicates = "combine") {
  
  # test
  is_char <- is.character(texture)
  is_fact <- is.factor(texture)
  
  if (!is_char & !is_fact) stop("texture must be a character or a factor")
  
  if (is_fact) texture <- as.character(texture)
  
  
  # load lookup tables
  soiltexture <- NULL
  load(system.file("data/soiltexture.rda", package = "aqp")[1])
  texmod_df <- soiltexture$texmod
  texmod    <- texmod_df$texmod[grepl("gravel|cobbl|ston|boulder|channer|flag", texmod_df$texmod_label)]
  
  # lieutex <- metadata[metadata$ColumnPhysicalName == "lieutex", ]$ChoiceName
  
  tex   <- tolower(texture)
  tex_l <- strsplit(tex, "-| |_|\\/|&|,|;|\\.|\\*|\\+")
  
  
  # find which records have a texmod
  idx_mod  <- which(sapply(tex_l, function(x) any(x[-length(x)] %in% texmod)))
  # idx_lieu <- which(sapply(tex_l, function(x) any(x %in% lieutex)))
  
  
  # create a logical matrix of texmod presence/absence
  if (length(idx_mod > 0)) {
    tex_m  <- lapply(tex_l[idx_mod], function(x)
      texmod %in% x)
    tex_df <- as.data.frame(do.call("rbind", tex_m))
    names(tex_df) <- texmod
    
    # slightly slower simpler(/better?) alternative
    # not better, this mistakenly matches words entered into this field
    # I would rather have false negatives than false positives
    # tex_m  <- sapply(texmod, function(x) grepl(x, tex))
    
    
    # duplicates ----
    ## combine texmod (if multiple exist)
    if (duplicates == "combine") {
      texmod_parse <- apply(tex_df, 1, function(x) {
        paste0(names(x)[which(x)], collapse = " & ")
      })
    }
    
    
    ## find texmod with the highest rock fragment content (if multiple exist)
    if (duplicates == "max") {
      tex_m <- sapply(texmod, function(x) {
        vals <- suppressMessages(texmod_to_fragvoltot(x))
        rf   <- ifelse(tex_df[x] == TRUE, vals$fragvoltot_r, 0)
      })
      idx_col <- max.col(tex_m)
      texmod_parse <- names(tex_df)[idx_col]
    }
  }
  
  # results ----
  n  <- length(texture)
  df <- data.frame(texmod = rep(NA_character_, n)) #, lieutex = rep(NA, n))
  if (length(idx_mod) > 0) df[idx_mod, "texmod"] <- texmod_parse
  
  return(df$texmod)
}


# add feature to uncode
# texture_to_texcl <- function(texclLabel, droplevels = FALSE 
#                              # , stringsAsFactors = FALSE
#                              ) {
#   
#   # standardize inputs
#   texclLabel <- tolower(texclLabel)
#   
#   # load lookup table
#   load(system.file("data/soiltexture.rda", package = "aqp")[1])
#   sub <- soiltexture$averages
#   
#   # check
#   idx <- ! texclLabel %in% sub$texcl_label
#   if (any(idx)) {
#     warning("not all the user supplied texture values match the lookup table")
#   }
#   
#   # convert
#   texcl <- factor(texclLabel, levels = tolower(sub$texcl_label), label = sub$texcl, ordered = TRUE)
#   
#   if (droplevels == TRUE) {
#     texcl <- droplevels(texcl)
#   }
#   
#   # if (stringsAsFactors == TRUE) {
#   #   texcl <- as.character(texcl)
#   # }
#   
#   return(texcl)
# }


#' Convert frags to texmod
#'
#' @param object data.frame: containing the following column names: gravel, cobbles,
#'  stones, boulders, channers, flagstones, paragravel, paracobbles, parastones,
#'   paraboulders, parachanners, paraflagstones
#' @param gravel numeric: gravel volume % 
#' @param cobbles numeric: cobble volume % 
#' @param stones numeric: stone volume % 
#' @param boulders numeric: boulder volume % 
#' @param channers numeric: channer volume % 
#' @param flagstones numeric: flagstone volume % 
#' @param paragravel numeric: para gravel volume % 
#' @param paracobbles numeric: para cobble volume % 
#' @param parastones numeric: para stone volume % 
#' @param paraboulders numeric: para boulder volume % 
#' @param parachanners numeric: para channer volume % 
#' @param paraflagstones numeric: para flagstone volume % 
#'
#' @return - `texmod_to_fragvol`: a data.frame containing `"texmod"` and `"lieutex"` classes
#' 
#' @rdname texture
#' 
#' @export
#'
#' @examples
#' \donttest{
#' # example of fragvol_to_texmod()
#' df <- expand.grid(
#'   gravel  = seq(0, 100, 5), 
#'   cobbles = seq(0, 100, 5), 
#'   stones  = seq(0, 100, 5), 
#'   boulders = seq(0, 100, 5)
#'   )
#' df <- df[rowSums(df) < 100, ]
#' 
#' # data.frame input
#' test <- fragvol_to_texmod(df)
#' table(test$texmod)
#' table(test$lieutex)
#' 
#' # vector inputs
#' fragvol_to_texmod(gravel = 10, cobbles = 10)
#' 
#' }
fragvol_to_texmod <- function(
    object   = NULL,
    gravel   = NULL, 
    cobbles  = NULL,
    stones   = NULL, 
    boulders = NULL, 
    channers = NULL, 
    flagstones = NULL, 
    paragravel     = NULL, 
    paracobbles    = NULL, 
    parastones     = NULL,
    paraboulders   = NULL, 
    parachanners   = NULL, 
    paraflagstones = NULL,
    as.is      = TRUE,
    droplevels = TRUE
) {
  
  # standardize inputs ----
  var_cols <- c("gravel", "cobbles", "stones", "boulders", "channers", "flagstones", "paragravel", "paracobbles", "parastones", "paraboulders", "parachanners", "paraflagstones")
  var_mods <- c("gr", "cb", "st", "by", "cn", "fl", "pgr", "pcb", "pst", "pby", "pcn", "pfl")
  
  # object = NULL; gravel =  10; cobbles =  10; stones =  NULL; boulders =  NULL; channers =  NULL; flagstones =  NULL; paragravel =  NULL; paracobbles =  NULL; parastones =  NULL; paraboulders =  NULL; parachanners =  NULL; paraflagstones =  NULL
  # object <- expand.grid(gravel = seq(0, 105, 5), cobbles = seq(0, 100, 5), stones = seq(0, 100, 5), boulders = seq(0, 100, 5))
  # object <- object[rowSums(object) <= 105, ]
  # object <- rbind(object, rep(NA, ncol(object)))
  vars_l <- list(gravel, cobbles, stones, boulders, channers, flagstones, paragravel, paracobbles, parastones, paraboulders, parachanners, paraflagstones)
  names(vars_l) <- var_cols
  
  
  # check inputs----
  vars_null <- sapply(vars_l, is.null)
  df_null   <- is.null(object)
  vars_len  <- sapply(vars_l, length)
  
  
  ## overlapping inputs
  if (!df_null & any(!vars_null)) {
    warning("if object and any other rock fragment arguments are both not null, only object will be used")
  }
  # ## object matching column
  # if (!df_null & all(!var_cols %in% names(object))) {
  #   stop(paste("object is missing columns matching any of the following columns", 
  #              paste0(var_cols, collapse = ", "))
  #   )
  # }
  # ## vars_l length of inputs
  # if (all(vars_len == max(vars_len)) & df_null) {stop("length of inputs do not match")}
  
  
  # standardize inputs again ----
  ## subset and select correct inputs
  df <- NULL
  if (any(!vars_null) & is.null(object)) {
    # df <- as.data.frame(vars_l[!vars_null])
    df <- .format_inputs(vars_l, var_cols, 0, "numeric")
  } else df <- .format_inputs(as.list(object), var_cols, 0, "numeric")
  
  # idx <- var_cols %in% names(df)
  # var_cols_sub <- var_cols[idx]
  # # var_mods_sub <- var_mods[idx]
  # df <- df[var_cols_sub]
  # 
  # 
  # ## append missing columns
  # df_mis <- as.data.frame(matrix(data = 0, nrow = nrow(df), ncol = sum(!idx)))
  # names(df_mis) <- var_cols[!idx]
  # df <- cbind(df_mis, df)[var_cols]
  names(df) <- var_mods
  # df[is.na(df)] <- 0
  # 
  # 
  # # check inputs again ----
  # ## correct datatype
  # if (!all(sapply(df, function(x)  inherits(x, "integer") | inherits(x, "numeric")))) {
  #   stop("all fragments must be numeric or integer")
  # }
  
  
  ## calculate sums ----
  gr  <- NULL;   cb <- NULL;  st <- NULL;  by <- NULL;  cn <- NULL;  fl <- NULL
  pgr  <- NULL; pcb <- NULL; pst <- NULL; pby <- NULL; pcn <- NULL; pfl <- NULL
  gr_cn <- NULL; cb_fl <- NULL; pgr_pcn <- NULL; pcb_pfl <- NULL;
  sum_nopf <- NULL; sum_pf <- NULL;
  fgr  <- NULL; fcb <- NULL; fst <- NULL; fby <- NULL; fcn <- NULL; ffl <- NULL
  fgr_fcn <- NULL; fcb_ffl <- NULL;
  sum_f <- NULL
  
  df$sum_nopf <- rowSums(df[, 1:6],  na.rm = TRUE)
  df$sum_pf   <- rowSums(df[, 7:12], na.rm = TRUE)
  df$sum_f    <- rowSums(df[, 1:12], na.rm = TRUE)
  
  df$gr_cn    <- rowSums(df[c("gr",  "cn")], na.rm = TRUE)
  df$cb_fl    <- rowSums(df[c("cb",  "fl")], na.rm = TRUE)
  
  df$pgr_pcn  <- rowSums(df[c("pgr", "pcn")], na.rm = TRUE)
  df$pcb_pfl  <- rowSums(df[c("pcb", "pfl")], na.rm = TRUE)
  
  df$fgr_fcn  <- rowSums(df[c("gr",  "cn", "pgr", "pcn")], na.rm = TRUE)
  df$fcb_ffl  <- rowSums(df[c("cb",  "fl", "pcb", "pfl")], na.rm = TRUE)
  
  df$fgr <- rowSums(df[c("gr", "pgr")], na.rm = TRUE)
  df$fcb <- rowSums(df[c("cb", "pcb")], na.rm = TRUE)
  df$fst <- rowSums(df[c("st", "pst")], na.rm = TRUE)
  df$fby <- rowSums(df[c("by", "pby")], na.rm = TRUE)
  df$fcn <- rowSums(df[c("cn", "pcn")], na.rm = TRUE)
  df$ffl <- rowSums(df[c("fl", "pfl")], na.rm = TRUE)
  
  
  # # load lookup table
  # soiltexture <- NULL
  # load(system.file("data/soiltexture.rda", package="aqp")[1])
  # texmod <- soiltexture$texmod
  
  
  ## filter rows with no rock fragments 
  idx_sum <- df$sum_nopf < 15 & df$sum_pf < 15 & df$sum_f < 15
  df_sub  <- df[!idx_sum, ]
  
  
  # check inputs again ----
  if (any(df$sum_nopf > 100)) {
    warning("some rows total rock fragments > 100")
  }
  if (any(df$sum_pf > 100)) {
    warning("some rows total pararock fragments > 100")
  }
  
  
  # calculate texmod & lieutex---- 
  df_sub <- within(df_sub, {
    
    # nopf
    texmod  = NA_character_
    lieutex = NA_character_
    x_gr_by = gr_cn >= ((1.5 * cb_fl) + (2 * st) + (2.5 * by))
    x_cb_by = cb_fl >= ((1.5 * st)    + (2 * by))
    # 15-34%
    x1534 = sum_nopf >= 15 & sum_nopf <  35
    texmod  = ifelse(x1534 & x_gr_by           & gr >= cn & gr > 0                , "gr", texmod)
    texmod  = ifelse(x1534 & x_gr_by           & gr <  cn & cn > 0 & is.na(texmod), "cn", texmod)
    texmod  = ifelse(x1534 & x_cb_by           & cb >= fl & cb > 0 & is.na(texmod), "cb", texmod)
    texmod  = ifelse(x1534 & x_cb_by           & cb <  fl & fl > 0 & is.na(texmod), "fl", texmod)
    texmod  = ifelse(x1534 & st    >= 1.5 * by            & st > 0 & is.na(texmod), "st", texmod)
    texmod  = ifelse(x1534 & st    <  1.5 * by            & by > 0 & is.na(texmod), "by", texmod)
    # 35-59%
    x3559 = sum_nopf >= 35 & sum_nopf <  60
    texmod  = ifelse(x3559 & x_gr_by           & gr >= cn & gr > 0,                 "grv", texmod)
    texmod  = ifelse(x3559 & x_gr_by           & gr <  cn & cn > 0 & is.na(texmod), "cnv", texmod)
    texmod  = ifelse(x3559 & x_cb_by           & cb >= fl & cb > 0 & is.na(texmod), "cbv", texmod)
    texmod  = ifelse(x3559 & x_cb_by           & cb <  fl & fl > 0 & is.na(texmod), "flv", texmod)
    texmod  = ifelse(x3559 & st    >= 1.5 * by            & st > 0 & is.na(texmod), "stv", texmod)
    texmod  = ifelse(x3559 & st    <  1.5 * by            & by > 0 & is.na(texmod), "byv", texmod)
    # 60-89%
    x6089 = sum_nopf >= 60 & sum_nopf <  90
    texmod  = ifelse(x6089 & x_gr_by           & gr >= cn & gr > 0                , "grx", texmod)
    texmod  = ifelse(x6089 & x_gr_by           & gr <  cn & cn > 0 & is.na(texmod), "cnx", texmod)
    texmod  = ifelse(x6089 & x_cb_by           & cb >= fl & cb > 0 & is.na(texmod), "cbx", texmod)
    texmod  = ifelse(x6089 & x_cb_by           & cb <  fl & fl > 0 & is.na(texmod), "flx", texmod)
    texmod  = ifelse(x6089 & st    >= 1.5 * by            & st > 0 & is.na(texmod), "stx", texmod)
    texmod  = ifelse(x6089 & st    <  1.5 * by            & by > 0 & is.na(texmod), "byx", texmod)
    # 90-100%
    x90 = sum_nopf >= 90
    lieutex = ifelse(x90 & x_gr_by           & gr >= cn & gr > 0,                 "gr", lieutex)
    lieutex = ifelse(x90 & x_gr_by           & gr <  cn & cn > 0 & is.na(texmod), "cn", lieutex)
    lieutex = ifelse(x90 & x_cb_by           & cb >= fl & cb > 0 & is.na(texmod), "cb", lieutex)
    lieutex = ifelse(x90 & x_cb_by           & cb <  fl & fl > 0 & is.na(texmod), "fl", lieutex)
    lieutex = ifelse(x90 & st    >= 1.5 * by            & st > 0 & is.na(texmod), "st", lieutex)
    lieutex = ifelse(x90 & st    <  1.5 * by            & by > 0 & is.na(texmod), "by", lieutex)
    
    
    # pf
    texmod_pf  = NA_character_
    lieutex_pf = NA_character_
    x_pgr_pby  = pgr_pcn >= ((1.5 * pcb_pfl) + (2 * pst) + (2.5 * pby))
    x_pcb_pby  = pcb_pfl >= ((1.5 * pst)     + (2 * pby))
    # 15-34%
    x1534 = sum_pf >= 15 & sum_pf <  35 & sum_nopf <= 0
    texmod_pf  = ifelse(x1534 & x_pgr_pby        & pgr >= pcn & pgr > 0,                    "pgr", texmod_pf)
    texmod_pf  = ifelse(x1534 & x_pgr_pby        & pgr <  pcn & pcn > 0 & is.na(texmod_pf), "pcn", texmod_pf)
    texmod_pf  = ifelse(x1534 & x_pcb_pby        & pcb >= pfl & pcb > 0 & is.na(texmod_pf), "pcb", texmod_pf)
    texmod_pf  = ifelse(x1534 & x_pcb_pby        & pcb <  pfl & pfl > 0 & is.na(texmod_pf), "pfl", texmod_pf)
    texmod_pf  = ifelse(x1534 & pst >= 1.5 * pby              & pst > 0 & is.na(texmod_pf), "pst", texmod_pf)
    texmod_pf  = ifelse(x1534 & pst <  1.5 * pby              & pby > 0 & is.na(texmod_pf), "pby", texmod_pf)
    # 35-59%
    x3559 = sum_pf >= 35 & sum_pf <  60 & sum_nopf <= 0
    texmod_pf  = ifelse(x3559 & x_pgr_pby        & pgr >= pcn & pgr > 0,                    "pgrv", texmod_pf)
    texmod_pf  = ifelse(x3559 & x_pgr_pby        & pgr <  pcn & pcn > 0 & is.na(texmod_pf), "pcnv", texmod_pf)
    texmod_pf  = ifelse(x3559 & x_pcb_pby        & pcb >= pfl & pcb > 0 & is.na(texmod_pf), "pcbv", texmod_pf)
    texmod_pf  = ifelse(x3559 & x_pcb_pby        & pcb <  pfl & pfl > 0 & is.na(texmod_pf), "pflv", texmod_pf)
    texmod_pf  = ifelse(x3559 & pst >= 1.5 * pby              & pst > 0 & is.na(texmod_pf), "pstv", texmod_pf)
    texmod_pf  = ifelse(x3559 & pst <  1.5 * pby              & pby > 0 & is.na(texmod_pf), "pbyv", texmod_pf)
    # 60-89%
    x6089 = sum_pf >= 60 & sum_pf <  90 & sum_nopf <= 0
    texmod_pf  = ifelse(x6089 & x_pgr_pby        & pgr >= pcn & pgr > 0,                    "pgrx", texmod_pf)
    texmod_pf  = ifelse(x6089 & x_pgr_pby        & pgr <  pcn & pcn > 0 & is.na(texmod_pf), "pcnx", texmod_pf)
    texmod_pf  = ifelse(x6089 & x_pcb_pby        & pcb >= pfl & pcb > 0 & is.na(texmod_pf), "pcbx", texmod_pf)
    texmod_pf  = ifelse(x6089 & x_pcb_pby        & pcb <  pfl & pfl > 0 & is.na(texmod_pf), "pflx", texmod_pf)
    texmod_pf  = ifelse(x6089 & pst >= 1.5 * pby              & pst > 0 & is.na(texmod_pf), "pstx", texmod_pf)
    texmod_pf  = ifelse(x6089 & pst <  1.5 * pby              & pby > 0 & is.na(texmod_pf), "pbyx", texmod_pf)
    # 90-100%
    x90 = sum_pf >= 90                  & sum_nopf <= 0
    lieutex_pf = ifelse(x90 & x_pgr_pby           & pgr >= pcn & pgr > 0,                    "pgr", lieutex_pf)
    lieutex_pf = ifelse(x90 & x_pgr_pby           & pgr <  pcn & pcn > 0 & is.na(texmod_pf), "pcn", lieutex_pf)
    lieutex_pf = ifelse(x90 & x_pcb_pby           & pcb >= pfl & pcb > 0 & is.na(texmod_pf), "pcb", lieutex_pf)
    lieutex_pf = ifelse(x90 & x_pcb_pby           & pcb <  pfl & pfl > 0 & is.na(texmod_pf), "pfl", lieutex_pf)
    lieutex_pf = ifelse(x90 & pst    >= 1.5 * pby              & pst > 0 & is.na(texmod_pf), "pst", lieutex_pf)
    lieutex_pf = ifelse(x90 & pst    <  1.5 * pby              & pby > 0 & is.na(texmod_pf), "pby", lieutex_pf)
    
    
    # f
    texmod_f = NA_character_
    lieutex_f = NA_character_
    x_fgr_fby = fgr_fcn >= ((1.5 * fcb_ffl) + (2 * fst) + (2.5 * fby))
    x_fcb_fby = fcb_ffl >= ((1.5 * fst)     + (2 * fby))
    x1534 = sum_f >= 15 & sum_f <  35 & sum_nopf < 15
    texmod_f  = ifelse(x1534 & x_fgr_fby        & fgr >= fcn & fgr > 0,                   "pgr", texmod_f)
    texmod_f  = ifelse(x1534 & x_fgr_fby        & fgr <  fcn & fcn > 0 & is.na(texmod_f), "pcn", texmod_f)
    texmod_f  = ifelse(x1534 & x_fcb_fby        & fcb >= ffl & fcb > 0 & is.na(texmod_f), "pcb", texmod_f)
    texmod_f  = ifelse(x1534 & x_fcb_fby        & fcb <  ffl & ffl > 0 & is.na(texmod_f), "pfl", texmod_f)
    texmod_f  = ifelse(x1534 & fst >= 1.5 * fby              & fst > 0 & is.na(texmod_f), "pst", texmod_f)
    texmod_f  = ifelse(x1534 & fst <  1.5 * fby              & fby > 0 & is.na(texmod_f), "pby", texmod_f)
    # 35-59%
    x3559 = sum_f >= 35 & sum_f <  60 & sum_nopf < 15
    texmod_f  = ifelse(x3559 & x_fgr_fby        & fgr >= fcn & fgr > 0,                   "pgrv", texmod_f)
    texmod_f  = ifelse(x3559 & x_fgr_fby        & fgr <  fcn & fcn > 0 & is.na(texmod_f), "pcnv", texmod_f)
    texmod_f  = ifelse(x3559 & x_fcb_fby        & fcb >= ffl & fcb > 0 & is.na(texmod_f), "pcbv", texmod_f)
    texmod_f  = ifelse(x3559 & x_fcb_fby        & fcb <  ffl & ffl > 0 & is.na(texmod_f), "pflv", texmod_f)
    texmod_f  = ifelse(x3559 & fst >= 1.5 * fby              & fst > 0 & is.na(texmod_f), "pstv", texmod_f)
    texmod_f  = ifelse(x3559 & fst <  1.5 * fby              & fby > 0 & is.na(texmod_f), "pbyv", texmod_f)
    # 60-89%
    x6089 = sum_f >= 60 & sum_f <  90 & sum_nopf < 15
    texmod_f  = ifelse(x6089 & x_fgr_fby        & fgr >= fcn & fgr > 0,                   "pgrx", texmod_f)
    texmod_f  = ifelse(x6089 & x_fgr_fby        & fgr <  fcn & fcn > 0 & is.na(texmod_f), "pcnx", texmod_f)
    texmod_f  = ifelse(x6089 & x_fcb_fby        & fcb >= ffl & fcb > 0 & is.na(texmod_f), "pcbx", texmod_f)
    texmod_f  = ifelse(x6089 & x_fcb_fby        & fcb <  ffl & ffl > 0 & is.na(texmod_f), "pflx", texmod_f)
    texmod_f  = ifelse(x6089 & fst >= 1.5 * fby              & fst > 0 & is.na(texmod_f), "pstx", texmod_f)
    texmod_f  = ifelse(x6089 & fst <  1.5 * fby              & fby > 0 & is.na(texmod_f), "pbyx", texmod_f)
    # 90-100%
    x90 = sum_f >= 90                 & sum_nopf < 15
    lieutex_f = ifelse(x90 & x_fgr_fby           & fgr >= fcn & fgr > 0,                   "pgr", lieutex_f)
    lieutex_f = ifelse(x90 & x_fgr_fby           & fgr <  fcn & fcn > 0 & is.na(texmod_f), "pcn", lieutex_f)
    lieutex_f = ifelse(x90 & x_fcb_fby           & fcb >= ffl & fcb > 0 & is.na(texmod_f), "pcb", lieutex_f)
    lieutex_f = ifelse(x90 & x_fcb_fby           & fcb <  ffl & ffl > 0 & is.na(texmod_f), "pfl", lieutex_f)
    lieutex_f = ifelse(x90 & fst    >= 1.5 * fby              & fst > 0 & is.na(texmod_f), "pst", lieutex_f)
    lieutex_f = ifelse(x90 & fst    <  1.5 * fby              & fby > 0 & is.na(texmod_f), "pby", lieutex_f)
    
    
    # combine texmods and lieutex ----
    texmod  = ifelse(is.na(texmod),  texmod_f,   texmod)
    texmod  = ifelse(is.na(texmod),  texmod_pf,  texmod)
    lieutex = ifelse(is.na(lieutex), lieutex_f,  lieutex)
    lieutex = ifelse(is.na(lieutex), lieutex_pf, lieutex)
  })
  
  # idx <- 1:ncol(df_sub)
  # df_mod <- df_sub
  # df_mod[idx] <- lapply(idx, function(i) {
  #   texmod <- rep(NA_character_, nrow(df_sub))
  #   texmod[df_sub[, i] >= 15 & df_sub[, i] < 35] <- var_mods_sub[i]
  #   texmod[df_sub[, i] >= 35 & df_sub[, i] < 60] <- paste0(var_mods_sub[i], "v")
  #   texmod[df_sub[, i] >= 60 & df_sub[, i] < 90] <- paste0(var_mods_sub[i], "x")
  #   return(texmod)
  # })
  # 
  # df_lieu <- df_sub
  # df_lieu[idx] <- lapply(idx, function(i) {
  #   lieutex <- rep(NA_character_, nrow(df_sub))
  #   lieutex[df_sub[i] >= 90] <- var_mods_sub[i]
  #   return(lieutex)
  # })
  
  
  # standardize output----
  if (nrow(df_sub) > 0) {
    df$texmod[!idx_sum]  <- df_sub$texmod
    df$lieutex[!idx_sum] <- df_sub$lieutex
  } else df[c("texmod", "lieutex")] <- NA_character_
  df <- df[c("texmod", "lieutex")]
  
  
  if (as.is == FALSE) {
    lv <- paste0(
      c(rep(var_mods[(7:12)], times = 3,  length.out = 18),
        rep(var_mods[(1:6)],  times = 3,  length.out = 18)
      ),
      rep(c("", "v", "x"),    each  = 6,  length.out = 36)
    )
    tn <- names(sort(table(df$texmod), decreasing = TRUE))
    lv <- c(lv, tn[! tn %in% lv])
    
    df$texmod  <- factor(df$texmod,  levels = lv)
    df$lieutex <- factor(df$lieutex, levels = var_mods[1:6])
  }
  
  if (as.is == FALSE & droplevels == TRUE) {
    df <- droplevels.data.frame(df)
  }
  
  
  return(df)
}


# format function inputs ----
# format vector argument inputs into a data.frame 
.format_inputs <- function(vars_l = NULL, var_cols, mis_value = NA, datatype = NULL) {
  
  # check inputs----
  vars_null <- sapply(vars_l, is.null)
  vars_len  <- sapply(vars_l[!vars_null], length)
  nm        <- var_cols
  n         <- length(var_cols)
  df <- NULL
  
  ## all inputs are NULL
  if (all(vars_null)) {
    stop("all inputs are NULL")
  }
  
  ## length of inputs
  if (!all(vars_len == max(vars_len))) {
    stop("length of inputs do not match")
  }
  
  # standardize inputs
  # create data.frame
  df <- as.data.frame(vars_l[!vars_null])
  
  # append missing columns
  idx <- nm %in% names(df)
  df_mis <- as.data.frame(matrix(data = mis_value, nrow = nrow(df), ncol = sum(!idx)))
  names(df_mis) <- nm[!idx]
  df <- cbind(df_mis, df)[nm]
  
  
  ## correct datatype
  df[1:n] <- lapply(df, function(x) if (is.factor(x))  as.character(x) else x)
  df[1:n] <- lapply(df, function(x) if (is.integer(x)) as.numeric(x)   else x)
  dt <- unlist(lapply(df, class))
  
  ## check datatype
  idx <- dt == datatype | dt == "logical"
  if (!all(idx)) {
    stop(paste("unexpected input data types, must be", datatype))
  }
  
  return(df)
}


#' @title Ranking Systems for USDA Taxonomic Particle-Size and Substitute Classes of Mineral Soils
#' 
#' @description Generate a lookup table of USDA Particle-Size and Substitute Classes  names, ranked according to approximate particle size
#'
#' @references \href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}
#' 
#' @return A data.frame with a rank column, taxonomic family particle size class, and a flag for contrasting.
#' 
#' @author Stephen Roecker
#' 
#' @seealso [hz_to_taxpartsize()], [texture_to_taxpartsize()], [SoilTextureLevels()]
#' 
#' @export
#' @examples
#' 
#' # class codes
#' lu <- lookup_taxpartsize()
#' 
#' idx <- lu$contrasting == FALSE
#' 
#' lu$taxpartsize[idx]
#' 
#' lu$rank[as.integer(lu$taxpartsize)[idx]]
#' 

lookup_taxpartsize <- function() {
  fe <- c("diatomaceous", "very-fine", "clayey", "fine", "hydrous", "fine-silty", 
          "fine-gypseous", "fine-loamy", "medial", "loamy", "coarse-loamy", 
          "coarse-silty", "coarse-gypseous", "ashy", "sandy", "hydrous-pumiceous", 
          "medial-pumiceous", "ashy-pumiceous", "clayey-skeletal", "hydrous-skeletal", 
          "medial-skeletal", "loamy-skeletal", "gypseous-skeletal", "ashy-skeletal", 
          "sandy-skeletal", "pumiceous", "cindery", "fragmental")
  
  rank <- c(84, 74, 60.02, 46.04, 44.04, 26, 25.8, 25.6, 24, 17.24, 8.88, 
            8.5, 7.5, 6.5, 4.67, -55.96, -76, -93.5, -43.33, -55.96, -76, 
            -83.23, -83.35, -93.5, -95.33, -95.83, -96.33, -98.94)
  names(rank) <- fe
  
  # cf <- c("fragmental", "sandy-skeletal", "loamy-skeletal", "clay-skeletal")
  
  test <- strsplit(.pscs_sc, " over | or ")
  names(test) <- .pscs_sc
  
  idx <- lapply(
    test, function(x) {
      idx <- unlist(sapply(x, function(y) rank[which(fe == y)]))
      
      # select the 3rd value when "or" results in 3 values
      if (length(idx) > 2) idx <- c(idx[1], idx[3])
      
      dif <- diff(idx)
      idx <- idx[1] + sqrt(abs(dif)) * sign(dif)
      
      return(idx)
    }
  )
  
  idx <- round(unlist(idx), 1)
  
  # fe <- data.frame(rn = 1:length(fe), fe = fe)
  fe_df <- data.frame(rank = unname(rank), taxpartsize = fe)
  sc_df <- data.frame(rank = unname(idx),  taxpartsize = .pscs_sc)
  lu <- rbind(fe_df, sc_df)
  lu <- lu[order(-lu$rank), ]
  # lu$rn <- 1:nrow(lu)
  lu$contrasting <- grepl(" over ", lu$taxpartsize)
  lu$taxpartsize <- factor(lu$taxpartsize, levels = lu$taxpartsize, ordered = TRUE)
  row.names(lu) <- NULL
  
  return(lu)
}


.pscs_sc <- tolower(c("Ashy over clayey", "Ashy over clayey-skeletal", "Ashy over loamy", "Ashy over loamy-skeletal", "Ashy over medial", "Ashy over medial-skeletal", "Ashy over pumiceous or cindery", "Ashy over sandy or sandy-skeletal", "Ashy-skeletal over clayey", "Ashy-skeletal over fragmental or cindery", "Ashy-skeletal over loamy-skeletal", "Ashy-skeletal over sandy or sandy-skeletal", "Cindery over loamy", "Cindery over medial", "Cindery over medial-skeletal", "Clayey over coarse-gypseous", "Clayey over fine-gypseous", "Clayey over fragmental", "Clayey over gypseous-skeletal", "Clayey over loamy", "Clayey over loamy-skeletal", "Clayey over sandy or sandy-skeletal", "Clayey-skeletal over sandy or sandy-skeletal", "Coarse-loamy over clayey", "Coarse-loamy over fragmental", "Coarse-loamy over sandy or sandy-skeletal", "Coarse-silty over clayey", "Coarse-silty over sandy or sandy-skeletal", "Fine-loamy over clayey", "Fine-loamy over fragmental", "Fine-loamy over sandy or sandy-skeletal", "Fine-silty over clayey", "Fine-silty over fragmental", "Fine-silty over sandy or sandy-skeletal", "Hydrous over clayey", "Hydrous over clayey-skeletal", "Hydrous over fragmental", "Hydrous over loamy", "Hydrous over loamy-skeletal", "Hydrous over sandy or sandy-skeletal", "Loamy over ashy or ashy-pumiceous", "Loamy over coarse-gypseous", "Loamy over fine-gypseous", "Loamy over pumiceous or cindery", "Loamy over sandy or sandy-skeletal", "Loamy-skeletal over cindery", "Loamy-skeletal over clayey", "Loamy-skeletal over fragmental", "Loamy-skeletal over gypseous-skeletal", "Loamy-skeletal over sandy or sandy-skeletal", "Medial over ashy", "Medial over ashy-pumiceous or ashy-skeletal", "Medial over clayey", "Medial over clayey-skeletal", "Medial over fragmental", "Medial over hydrous", "Medial over loamy", "Medial over loamy-skeletal", "Medial over pumiceous or cindery", "Medial over sandy or sandy-skeletal", "Medial-skeletal over fragmental or cindery", "Medial-skeletal over loamy-skeletal", "Medial-skeletal over sandy or sandy-skeletal", "Pumiceous or ashy-pumiceous over loamy", "Pumiceous or ashy-pumiceous over loamy-skeletal", "Pumiceous or ashy-pumiceous over medial", "Pumiceous or ashy-pumiceous over medial-skeletal", "Pumiceous or ashy-pumiceous over sandy or sandy-skeletal", "Sandy over clayey", "Sandy over loamy", "Sandy-skeletal over loamy"))

context("fragvol_to_texmod")


# fragvol_to_texmod ----
test_that("increasing volume of one size class", {
  expect_equal(c("by", "byv", "byx", NA), fragvol_to_texmod(data.frame(boulders = c(25, 48, 75, 95)))$texmod)
  expect_equal(c(NA, NA, NA, "by"), fragvol_to_texmod(data.frame(boulders = c(25, 48, 75, 95)))$lieutex)
})


test_that("2:1 rules for dominant/largest size class", {
  expect_equal("cbx", fragvol_to_texmod(data.frame(gravel = 20, cobbles = 40))$texmod)
  expect_equal("grx", fragvol_to_texmod(data.frame(gravel = 40, cobbles = 20))$texmod)
})


test_that("no fragments > 15%", {
  expect_equal(NA_character_, fragvol_to_texmod(data.frame(gravel = 1, cobbles = 2, flagstones = 10))$texmod)
})
  

test_that("sum of fragments plus parafragments >15%, but nopf<15 and pf<15", {
  expect_equal("pgr", fragvol_to_texmod(data.frame(gravel = 14, paragravel = 2))$texmod)
})


## texture_to_texmod ----

test_that("very cobbly loam (works)", {
  expect_equal(texture_to_texmod("CBV-L"), "cbv")
})

test_that("ashy cobbly sandy loam (works)", {
  expect_equal(texture_to_texmod("ASHY-CB-SL"), "cb")
})

test_that("ashy boulders (works)", {
  expect_equal(texture_to_texmod("ASHY-BY"), NA_character_)
})

## TODO:
# * introduce a `jitter` argument to account for the possibility of varying geometry / class


## helper function for matching either the full code, or first-letter abbreviation to a value
.matchCodeOrTerm <- function(x, codes, values) {
  # just in case single letter codes are used
  codes.single <- substr(codes, 1, 1)
  
  # no factors here
  x <- as.character(x)
  
  # convert to lower case
  x <- tolower(x)
  
  # index to single-letter codes vs. terms
  idx <- which(sapply(x, function(i) nchar(i) == 1))
  
  if(length(idx) > 0) {
    # look-up via single letter codes
    x.code.single <- match(x, codes.single)
    
    # look-up terms just in case
    x.code <- match(x, codes)
    
    # if any single letter code results in NA, attempt full term match
    x.code <- ifelse(is.na(x.code.single), x.code, x.code.single)
  } else {
    # look-up terms only
    x.code <- match(x, codes) 
  }
  
  # look-up associated value
  res <- values[x.code]
  
  return(res)
}



#' @title Convert Horizon Boundary Distinctness to Vertical Offset
#' 
#' @description This function will convert USDA-NCSS horizon boundary distinctness codes into vertical (+/-) offsets in cm, based on the \href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}.
#' 
#' @author D.E. Beaudette
#' 
#' @param x vector of boundary distinctness codes to be converted
#' 
#' @param codes character vector of distinctness terms ('clear') or codes ('C'), case insensitive, see details
#' 
#' @param offset vertical offset factors (cm), approximating 1/2 of the transitional zone thickness, see details
#' 
#' @return vector of offsets with same length as \code{x}
#' 
#' @details The default offsets are based on the high-end of ranges presented in "transitional zone thickness criteria" from the Field Book version 3.0 (page 2-6). Offsets are returned as 1/2 of the transitional zone thickness so that horizon boundaries can be adjusted up/down from horizon depths. See \code{\link{plotSPC}}, specifically the \code{hz.distinctness.offset} argument for visualization ideas. Missing data in \code{x} (NA) or codes that are not defined in \code{codes} are returned as 0 offsets.
#' 
#' Either format (or mixture) are accepted, case insensitive:
#' 
#'   * terms: `c('very abrupt', 'abrupt', 'clear', 'gradual', 'diffuse')`
#'   * coded values: `c('v', 'a', 'c', 'g', d')`
#' 
#' Additional examples are available in the \href{https://ncss-tech.github.io/AQP/aqp/hz-boundaries.html}{Visualization of Horizon Boundaries tutorial}.
#' 
#' @seealso \code{\link{plotSPC}}
#' 
#' @keywords manip
#' 
#' @references \href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}
#' 
#' @export
#' @examples 
#' # example data
#' data(sp1)
#' 
#' # compute 1/2 transitional zone thickness from distinctness codes
#' sp1$hzdo <- hzDistinctnessCodeToOffset(sp1$bound_distinct)
#' 
#' # convert colors from Munsell to hex-encoded RGB
#' sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))
#' 
#' # promote to SoilProfileCollection
#' depths(sp1) <- id ~ top + bottom
#' hzdesgnname(sp1) <- 'name'
#' 
#' # adjust margins
#' op <- par(mar=c(0,0,0,1.5))
#' 
#' # sketches, adjust width, adjust text size, include coded hz distinctness offsets
#' plotSPC(sp1, width=0.3, cex.names=0.75, hz.distinctness.offset = 'hzdo')
#' 
#' # clean-up
#' par(op)
hzDistinctnessCodeToOffset <- function(
  x, 
  codes =  c('very abrupt', 'abrupt', 'clear', 'gradual', 'diffuse'), 
  offset = c(0.5, 2, 5, 15, 20) / 2
) {	
  
  # match term or first letter code
  res <- .matchCodeOrTerm(x, codes, offset)
  
  # interpret NA as 0
  res <- ifelse(is.na(res), 0, res)
  
  return(res)
}



#' @title Convert Horizon Boundary Topography to Line Type
#' 
#' @description This function will convert USDA-NCSS horizon boundary topography codes into line types, based on the \href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}.
#' 
#' @author D.E. Beaudette
#' 
#' @details Visualization of horizon boundary topography can be difficult, line type offers an additional visual cue. See \code{hzTopographyCodeToOffset} for an offset-based approach. Additional examples are available in the \href{https://ncss-tech.github.io/AQP/aqp/hz-boundaries.html}{Visualization of Horizon Boundaries tutorial}. Missing data in \code{x} (NA) or codes that are not defined in \code{codes} are returned as line type 1.
#' 
#' Either format (or mixture) are accepted, case insensitive:
#' 
#'   * terms: `c('smooth', 'wavy', 'irregular', 'broken')`
#'   * coded values: `c('s', 'w', 'i', 'b')`
#' 
#' @param x vector of boundary topography codes to be converted
#' 
#' @param codes character vector of topography terms ('smooth') or codes ('S'), case insensitive, see details
#' 
#' @param lty line types
#' 
#' @return vector of line types with same length as \code{x}
#' 
#' @seealso \code{\link{plotSPC}, \link{hzTopographyCodeToOffset}}
#' 
#' @keywords manip
#' 
#' @references \href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}
#' @export
hzTopographyCodeToLineType <- function(
  x, 
  codes =  c('smooth', 'wavy', 'irregular', 'broken'), 
  lty = c(1, 2, 3, 4)
) {	
  
  # match term or first letter code
  res <- .matchCodeOrTerm(x, codes, lty)
  
  # interpret NA as 1
  res <- ifelse(is.na(res), 1, res)
  
  return(res)
}


#' @title Convert Horizon Boundary Topography to Vertical Offset
#' 
#' @description This function will convert USDA-NCSS horizon boundary topography codes into a vertical offset, suitable for use in \code{plotSPC}. Default values are reasonable starting points for encoding smooth, wavy, irregular, or broken style horizon boundary topography as defined in \href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}.
#' 
#' @details Additional examples are available in the \href{https://ncss-tech.github.io/AQP/aqp/hz-boundaries.html}{Visualization of Horizon Boundaries tutorial}. Missing data in \code{x} (NA) or codes that are not defined in \code{codes} are returned with an offset of 0.
#' 
#' Either format (or mixture) are accepted, case insensitive:
#' 
#'   * terms: `c('smooth', 'wavy', 'irregular', 'broken')`
#'   * coded values: `c('s', 'w', 'i', 'b')`
#' 
#' @author D.E. Beaudette
#' 
#' @param x vector of boundary topography codes to be converted
#' 
#' @param codes character vector of topography terms ('smooth') or codes ('S'), case insensitive, see details
#' 
#' @param offset vertical offset (depth units) used to create "chevron" effect
#' 
#' @return vector of vertical offsets with same length as \code{x}
#' 
#' @seealso \code{\link{plotSPC}}
#' 
#' @keywords manip
#' @export
#' @references \href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}
#' 
hzTopographyCodeToOffset <- function(
  x, 
  codes =  c('smooth', 'wavy', 'irregular', 'broken'), 
  offset = c(0, 4, 8, 12)
) {	
  
  # match term or first letter code
  res <- .matchCodeOrTerm(x, codes, offset)
  
  # interpret NA as 0
  res <- ifelse(is.na(res), 0, res)
  
  return(res)
}

context("horizon boundary interpretation")

test_that("hzDistinctnessCodeToOffset", {
  
  # terms
  expect_equal(
    hzDistinctnessCodeToOffset(c('very abrupt', 'abrupt', 'clear', 'gradual', 'diffuse')),
    c(0.25, 1.00, 2.50, 7.50, 10.00)
  )
  
  # case insensitivity
  expect_equal(
    hzDistinctnessCodeToOffset(c('Very Abrupt', 'Abrupt', 'Clear', 'Gradual', 'Diffuse')),
    c(0.25, 1.00, 2.50, 7.50, 10.00)
  )
  
  # codes
  expect_equal(
    hzDistinctnessCodeToOffset(c('V', 'A', 'C', 'G', 'D')),
    c(0.25, 1.00, 2.50, 7.50, 10.00)
  )
  
  # bogus -> 0
  expect_equal(
    hzDistinctnessCodeToOffset(c('very very', 'A', 'C', 'G', 'zzz')),
    c(0, 1.00, 2.50, 7.50, 0)
  )
  
  # NA -> 0
  expect_equal(
    hzDistinctnessCodeToOffset(NA),
    0
  )
  
  
})


test_that("hzTopographyCodeToLineType", {
  
  # terms
  expect_equal(
    hzTopographyCodeToLineType(c('smooth', 'wavy', 'irregular', 'broken')),
    c(1, 2, 3, 4)
  )
  
  # case insensitivity
  expect_equal(
    hzTopographyCodeToLineType(c('Smooth', 'Wavy', 'Irregular', 'Broken')),
    c(1, 2, 3, 4)
  )
  
  # codes
  expect_equal(
    hzTopographyCodeToLineType(c('B', 'I', 'W', 'S')),
    c(4, 3, 2, 1)
  )
  
  # bogus -> 1
  expect_equal(
    hzTopographyCodeToLineType(c('clowns')),
    1
  )
  
  # NA -> 1
  expect_equal(
    hzTopographyCodeToLineType(NA),
    1
  )
  
})

test_that("hzTopographyCodeToOffset", {
  
  # terms
  expect_equal(
    hzTopographyCodeToOffset(c('smooth', 'wavy', 'irregular', 'broken')),
    c(0, 4, 8, 12)
  )
  
  # case insensitivity
  expect_equal(
    hzTopographyCodeToOffset(c('Smooth', 'Wavy', 'Irregular', 'Broken')),
    c(0, 4, 8, 12)
  )
  
  # codes
  expect_equal(
    hzTopographyCodeToOffset(c('B', 'I', 'W', 'S')),
    c(12, 8, 4, 0)
  )
  
  # bogus -> 0
  expect_equal(
    hzTopographyCodeToOffset(c('clowns')),
    0
  )
  
  # NA -> 0
  expect_equal(
    hzTopographyCodeToOffset(NA),
    0
  )
  
})

## ----setup, echo=FALSE, results='hide', warning=FALSE, message=FALSE------------------------------
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = 96,
  optipng = knitr::hook_optipng,
  comment = "#>"
)
options(width = 100, stringsAsFactors = FALSE, timeout = 600)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

library(aqp, quietly = TRUE)
.bluecolors <- hcl.colors(n = 25, palette = 'Blues')[-25]

## ----fig.width=6, fig.height=5, echo = FALSE------------------------------------------------------
data("osd")
par(mar = c(0, 0, 0, 0))
plotSPC(osd, name = NA, fixLabelCollisions = FALSE, cex.names = 1.75, hz.depths = TRUE, depth.axis = FALSE, cex.id = 0.66, name.style = 'center-center')

## ----fig.width=8, fig.height=5--------------------------------------------------------------------
library(aqp)

# soil profile template
x <- quickSPC(
  "SPC:Oi|Oe|AAAAA|E1|E2|Bhs1Bhs1Bhs1|Bhs2|CCCCCC|RRRRRRRRRR", 
  interval = 3
)

# set variability in horizon thickness
horizons(x)$thick.sd <- 3
# setup horizon designations
horizons(x)$nm <- factor(x$name, levels = c('Oi', 'Oe', 'A', 'E1', 'E2', 'Bhs1', 'Bhs2', 'C', 'R'))

# simulate 3 realizations of original profile
set.seed(10101)
s <- perturb(x, n = 3, thickness.attr = 'thick.sd', min.thickness = 2)

# create 3 copies of each simulation
s <- duplicate(s, times = 3)

## ----fig.width=8, fig.height=5, echo = FALSE------------------------------------------------------
.a <- list(cex.names = 0.8, color = 'nm', name = NA, show.legend = FALSE, hz.depths = TRUE, depth.axis = FALSE, cex.id = 0.66, col.palette = .bluecolors)

options(.aqp.plotSPC.args = .a)

par(mar = c(0, 0, 0, 0))
plotSPC(s, fixLabelCollisions = FALSE)
title('Example Data', line = -1.5)

## ----fig.width=8, fig.height=5, echo = FALSE------------------------------------------------------
par(mar = c(0, 0, 0, 0))
# select "Electrostatic" label placement method
# adjust default charge density to 0.5
plotSPC(s, fixOverlapArgs = list(method = 'E', q = 0.5))
title('Label Adjustment by Electrostatic Simulation', line = -1.5)

## ----fig.width=8, fig.height=5, echo = FALSE------------------------------------------------------
par(mar = c(0, 0, 0, 0))
plotSPC(s, fixOverlapArgs = list(method = 'S'))
title('Label Adjustment by Simulated Annealing', line = -1.5)

options(.aqp.plotSPC.args = NULL)

## ----fig.width=8, fig.height=5--------------------------------------------------------------------
x <- quickSPC(
  "SPC:Oi|Oe|AAA|E1|E2|E3|BhsBhsBhsBhs|Bt1|Bt2|Bt3Bt3|CCCCCC|Ab1|Ab2|2C2C2C2C2C2C|2Cr|2R2R2R2R2R2R2R2R", 
  interval = 1
)

x$z <- as.numeric(x$hzID)

## ----fig.width=8, fig.height=5--------------------------------------------------------------------
# pretty colors
.bluecolors <- hcl.colors(n = 25, palette = 'Blues')[-25]

# plotSPC arguments
.a <- list(
  width = 0.2, 
  max.depth = 40,
  hz.depths = TRUE, 
  name.style = 'center-center', 
  cex.names = 1.5,
  name = NA,
  depth.axis = FALSE, 
  color = 'z',
  show.legend = FALSE,
  print.id = FALSE,
  col.palette = .bluecolors
)

# set plotSPC default arguments
options(.aqp.plotSPC.args = .a)

# wrapper function to test label collision solutions
testIt <- function(x, ...) {
  
  # make sketches
  plotSPC(x, ...)
  
  # a normalized index of label adjustment
  .LAI <- get('last_spc_plot', envir = aqp.env)$hz.depth.LAI
  .LAI <- ifelse(is.na(.LAI), 0, .LAI)
  
  # annotate with label adjustment index
  .txt <- sprintf("LAI: %0.3f", .LAI)
  mtext(.txt, side = 1, at = 1, line = -1.5, cex = 0.8)
}

## ----fig.width=8, fig.height=5--------------------------------------------------------------------
par(mar = c(1, 0, 0, 0), mfcol = c(1, 8))

testIt(x, fixLabelCollisions = FALSE)
title('No\nAdjustment', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 1', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 2', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 3', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 1.5))
title('Electrostatic\nq = 1.5', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 1))
title('Electrostatic\nq = 1', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 0.5))
title('Electrostatic\nq = 0.5', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 0.25))
title('Electrostatic\nq = 0.25', line = -3.5, adj = 0.5)

## ----eval=requireNamespace("soilDB", quietly=TRUE), fig.width=9, fig.height=5---------------------
library(aqp)
library(soilDB)

s <- c('inks' , 'pardee', 'clarksville', 'palau', 'hao', 'inks', 'eheuiki', 'puaulu', 'zook', 'cecil')
x <- fetchOSD(s)

par(mar = c(0, 0, 0, 3))

.args <- list(width = 0.3, name.style = 'center-center', hz.depths = TRUE, cex.names = 0.8)
options(.aqp.plotSPC.args = .args)

plotSPC(x, fixOverlapArgs = list(method = 'E', q = 1.25), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 1), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.75), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.5), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.25), max.depth = 151)

plotSPC(x, fixOverlapArgs = list(method = 'S'), max.depth = 151)

## -------------------------------------------------------------------------------------------------
x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)

overlapMetrics(x, thresh = 0.5)

## ----message=TRUE---------------------------------------------------------------------------------
# vector of positions, typically labels but could be profile sketch alignment on the x-axis
s <- c(1, 2, 2.3, 4, 5, 5.5, 7)

# simulated annealing, solution is non-deterministic
fixOverlap(s, thresh = 0.5, method = 'S')

# electrostatics-inspired simulation of particles
# solution is deterministic
fixOverlap(s, thresh = 0.5, method = 'E')

## ----fig.width=8.5, fig.height=5------------------------------------------------------------------
evalMethods <- function(x, thresh, q, ...) {
  
  cols <- hcl.colors(n = 9, palette = 'Zissou 1', rev = TRUE)
  cols <- colorRampPalette(cols)(length(x))
  
  z <- fixOverlap(x, thresh = thresh, method = 'E', maxIter = 250, trace = TRUE, q = q)
  .n <- nrow(z$states)
  
  op <- par(mar = c(0, 2, 2, 0.5), bg = 'black', fg = 'white')
  layout(matrix(c(1, 2, 3, 4), ncol = 2, nrow = 2), heights = c(0.33, 0.66))
  
  plot(seq_along(z$cost), z$cost, las = 1, type = 'b', axes = FALSE, cex = 0.66, xlim = c(1, .n))
  mtext(text = sprintf("Converged (%s): %s", .n, z$converged), at = 1, side = 3, line = 0, cex = 0.75, font = 3, adj = 0)
  matplot(rbind(x, z$states), type = 'l', lty = 1, las = 1, axes = FALSE, col = cols, lwd = 1)
  
  points(x = rep(1, times = length(x)), y = x, cex = 0.66, pch = 16, col = cols)
  points(x = rep(.n + 1, times = length(x)), y = z$x, cex = 0.66, pch = 16, col = cols)
  
  text(x = 1, y = x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 2)
  text(x = .n + 1, y = z$x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 4)
  
  axis(side = 2, at = unique(x), labels = round(unique(x), 1), col.axis = par('fg'), las = 1, cex.axis = 0.6)
  title(main = 'Electrostatic Simulation', line = 1, col.main = 'white')
  
  ## SANN_1D doesn't always preserve rank ordering
  ##  ->> not designed to use unsorted input
  ##  ->> maybe impossible with ties in x?
  
  z <- fixOverlap(x, thresh = thresh, method = 'S', trace = TRUE, maxIter = 1000)
  .n <- nrow(z$states)
  
  plot(seq_along(z$stats), z$stats, las = 1, type = 'b', axes = FALSE, cex = 0.66, xlim = c(1, .n))
  mtext(text = sprintf("Converged (%s): %s", .n, z$converged), at = 1, side = 3, line = 0, cex = 0.75, font = 3, adj = 0)
  
  matplot(z$states, type = 'l', lty = 1, las = 1, axes = FALSE, col = cols)
  
  points(x = rep(1, times = length(x)), y = z$states[1, ], cex = 0.66, pch = 16, col = cols)
  points(x = rep(.n, times = length(x)), y = z$x, cex = 0.66, pch = 16, col = cols)
  
  text(x = 1, y = z$states[1, ], col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 2)
  text(x = .n, y = z$x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 4)
  
  axis(side = 2, at = unique(x), labels = round(unique(x), 1), col.axis = par('fg'), las = 1, cex.axis = 0.6)
  
  title(main = 'Simulated Annealing', line = 1, col.main = 'white')
  
  # reset graphics state
  par(op)
  layout(1)
  
}

## ----fig.width=9.5, fig.height=5------------------------------------------------------------------
# explore effect of charge (q)
# too large -> chaos
x <- c(0, 2, 5, 12, 18, 20, 35, 40, 50, 56, 90, 120, 145, 150)

# just about right, very few perturbations required
evalMethods(x, thresh = 5, q = 1.1)

# ok, but now most label positions are affected
evalMethods(x, thresh = 5, q = 1.8)

# too high, wasting time on more iterations
evalMethods(x, thresh = 5, q = 3)

# far too high, wasting more time with little gain
evalMethods(x, thresh = 5, q = 4)

# chaos and eventually convergence
evalMethods(x, thresh = 5, q = 5)

## ----eval = FALSE---------------------------------------------------------------------------------
# # threshold too large
# evalMethods(x, thresh = 10, q = 3)
# 
# # large threshold
# x <- c(0, 5, 12, 18, 20, 35, 40, 55, 90, 120, 145, 150)
# evalMethods(x, thresh = 9, q = 2)
# 
# # single iteration enough
# x <- c(0, 3, 20, 35, 40, 55, 90, 120, 145, 150)
# evalMethods(x, thresh = 6, q = 1)
# 
# # clusters
# x <- sort(c(0, jitter(rep(10, 3)), jitter(rep(25, 3)), jitter(rep(90, 3)), 150))
# evalMethods(x, thresh = 6, q = 3)
# evalMethods(x, thresh = 6, q = 2)
# 
# 
# ## impact of scale / offset
# x <- c(0, 5, 12, 18, 20, 35, 40, 50, 120, 145, 150)
# 
# # works as expected
# evalMethods(x, thresh = 5, q = 1.1)
# 
# # works as expected, as long as threshold is scaled
# evalMethods(x / 10, thresh = 5 / 10, q = 1.1)
# 
# # works as expected, as long as threshold is scaled
# evalMethods(x * 10, thresh = 5 * 10, q = 1.1)
# 
# 
# # all work as expected, threshold not modified
# evalMethods(x + 10, thresh = 5, q = 1.1)
# evalMethods(x + 100, thresh = 5, q = 1.1)
# evalMethods(x + 1000, thresh = 5, q = 1.1)
# 
# # works as expected
# x <- c(315, 325, 341, 353, 366, 374, 422)
# fixOverlap(x, thresh = 9.7, q = 1, method = 'E')
# evalMethods(x, thresh = 9.7, q = 1)
# 
# 
# x <- c(1.0075, 1.1200, 1.3450, 1.6450, 1.8700, 1.8825)
# fixOverlap(x, thresh = 0.05442329, q = 1)
# evalMethods(x, thresh = 0.05442329, q = 1)

## ----fig.width=8, fig.height=6--------------------------------------------------------------------
tracePlot <- function(x, z, cex.axis.labels = 0.85) {
  # setup plot device
  op <- par(mar = c(4, 4, 1, 1), bg = 'black', fg = 'white')
  layout(matrix(c(1,2,3)), widths = 1, heights = c(1,1,2))
  
  # order:
  # B: boundary condition violation
  # O: rank (order) violation
  # +: accepted perturbation
  # -: rejected perturbation
  cols <- c(grey(0.85), 'yellow', 'royalblue', 'firebrick')
  
  cols.lines <- hcl.colors(9, 'Zissou 1', rev = TRUE)
  cols.lines <- colorRampPalette(cols.lines)(length(x))
  
  # total overlap (objective function) progress
  plot(
    seq_along(z$stats), z$stats, 
    type = 'h', las = 1,
    xlab = 'Iteration', ylab = 'Total Overlap',
    axes = FALSE,
    col = cols[as.numeric(z$log)]
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, line = -2)
  mtext('Overlap', side = 2, line = 2, cex = cex.axis.labels, font = 2)
  
  # deviation from original configuration
  plot(
    seq_along(z$stats), z$ssd, 
    type = 'h', las = 1,
    xlab = 'Iteration', ylab = 'Deviation',
    axes = FALSE,
    col = cols[as.numeric(z$log)]
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, line = -2)
  mtext('Deviation', side = 2, line = 2, cex = cex.axis.labels, font = 2)
  legend('top', legend = c('boundary\nviolation', 'rank\nviolation', 'accepted\nperturbation', 'rejected\nperturbation'), col = cols, bty = 'n', horiz = TRUE, inset = -0.5, lty = 1, lwd = 2, xpd = NA)
  
  # adjustments at each iteration
  matplot(
    z$states, type = 'l', 
    lty = 1, las = 1, 
    xlab = 'Iteration', ylab = 'x-position',
    axes = FALSE,
    col = cols.lines
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, at = x, labels = round(x, 1))
  axis(side = 4, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, at = z$x, labels = round(z$x, 1), line = -2)
  mtext('Position', side = 2, line = 2.5, cex = cex.axis.labels, font = 2)
  
  axis(side = 1, cex.axis = 1, col.axis = 'white', line = 0)
  mtext('Iteration', side = 1, line = 2.5, cex = cex.axis.labels, font = 2)
  
  par(op)
  layout(1)
}

## ----fig.width=8, fig.height=6, message=TRUE------------------------------------------------------
x <- c(0, 1, 2, 2.2, 2.8, 3.5, 6, 8, 10, 10.1, 12.8, 13, 14.8, 15, 15.5)

# fix overlap, return debugging information
set.seed(10101)
z <- fixOverlap(x, thresh = 0.73, method = 'S', trace = TRUE)

# check convergence
z$converged

# inspect algorithm trace
tracePlot(x, z)

# trace log
# B: boundary condition violation
# O: rank (order) violation
# +: accepted perturbation
# -: rejected perturbation
table(z$log)

## ----fig.width=8, fig.height=6, message=TRUE------------------------------------------------------
# fix overlap, return debugging information
set.seed(101010)
x <- sort(runif(10, min = 2.5, max = 3.5))

# widen boundary conditions
z <- fixOverlap(x, thresh = 0.2, trace = TRUE, min.x = 0, max.x = 10, maxIter = 2000, adj = 0.05)

# check convergence
z$converged

# inspect algorithm trace
tracePlot(x, z)

## -------------------------------------------------------------------------------------------------
# reset plotSPC() options
options(.aqp.plotSPC.args = NULL)

#' Collapse Horizons within Profiles Based on Pattern Matching
#'
#' Combines layers and aggregates data by grouping adjacent horizons which match `pattern` in
#' `hzdesgn` or, alternately, share a common value in `by` argument. Numeric properties are combined
#' using the weighted average, and other properties are derived from the dominant condition based on
#' thickness of layers and values in each group.
#'
#' @param x A _SoilProfileCollection_
#' @param pattern _character_. A regular expression pattern to match in `hzdesgn` column. Default:
#'   `NULL`.
#' @param by _character_. A column name specifying horizons that should be combined. Aggregation
#'   will be applied to adjacent groups of layers within profiles that have the same value in `by`.
#'   Used in lieu of `pattern` and `hzdesgn`. Default: `NULL`.
#' @param hzdesgn _character_. Any character column containing horizon-level identifiers. Default:
#'   `hzdesgnname(x, required = TRUE)`.
#' @param FUN _function_. A function that returns a _logical_ vector equal in length to the number
#'   of horizons in `x`. Used only when `pattern` is specified. See details.
#' @param ... Additional arguments passed to the matching function `FUN`.
#' @param AGGFUN _list_. A _named_ list containing custom aggregation functions. List element names
#'   should match the column name that they transform. The functions defined should take three
#'   arguments: `x` (a vector of horizon property values), `top` (a vector of top depths), and
#'   `bottom` (a vector of bottom depths). Default: `NULL` applies `weighted.mean()` to all numeric
#'   columns not listed in `ignore_numerics` and takes the dominant condition (value with greatest
#'   aggregate thickness sum) for all other columns. See details.
#' @param ignore_numerics _character_. Vector of column names that contain numeric values which
#'   should _not_ be aggregated using `weighted.mean()`. For example, soil color "value" and
#'   "chroma".
#' @param na.rm _logical_. If `TRUE` `NA` values are ignored when calculating min/max boundaries for
#'   each group and in weighted averages. If `FALSE` `NA` values are propagated to the result.
#'   Default: `FALSE`.
#'
#' @details
#'
#' If a custom matching function (`FUN`) is used, it should accept arbitrary additional arguments
#' via an ellipsis (`...`). It is not necessary to do anything with arguments, but the result should
#' match the number of horizons found in the input SoilProfileCollection `x`.
#'
#' Custom aggregation functions defined in the `AGGFUN` argument should either return a single
#' vector value for each group*column combination, or should return a _data.frame_ object with named
#' columns. If the input column name is used as a column name in the result _data.frame_, then the
#' values of that column name in the result _SoilProfileCollection_ will be replaced by the output
#' of the aggregation function. See examples.
#' 
#' @return A _SoilProfileCollection_
#' 
#' @author Andrew G. Brown
#' 
#' @seealso `hz_dissolve()`
#' 
#' @export
#'
#' @examples
#' data(jacobs2000)
#' 
#' # calculate a new SPC with genhz column based on patterns
#' new_labels <- c("A", "E", "Bt", "Bh", "C")
#' patterns <- c("A", "E", "B.*t", "B.*h", "C")
#' jacobs2000_gen <- generalizeHz(jacobs2000, new = new_labels, pattern = patterns)
#' 
#' # use existing generalized horizon labels
#' i <- collapseHz(jacobs2000_gen, by = "genhz")
#' 
#' profile_id(i) <- paste0(profile_id(i), "_collapse")
#' 
#' plot(
#'   c(i, jacobs2000),
#'   color = "genhz",
#'   name = "name",
#'   name.style = "center-center",
#'   cex.names = 1
#' )
#'  
#' # custom pattern argument  
#' j <- collapseHz(jacobs2000,
#'                 c(
#'                   `A` = "^A",
#'                   `E` = "E",
#'                   `Bt` = "[ABC]+t",
#'                   `C` = "^C",
#'                   `foo` = "bar"
#'                 ))
#' profile_id(j) <- paste0(profile_id(j), "_collapse")
#' plot(c(j, jacobs2000), color = "clay")
#' 
#' # custom aggregation function for matrix_color_munsell
#' k <- collapseHz(jacobs2000,
#'                 pattern = c(
#'                   `A` = "^A",
#'                   `E` = "E",
#'                   `Bt` = "[ABC]+t",
#'                   `C` = "^C",
#'                   `foo` = "bar"
#'                 ),
#'                 AGGFUN = list(
#'                   matrix_color_munsell = function(x, top, bottom) {
#'                     thk <- bottom - top
#'                     if (length(x) > 1) {
#'                       xord <- order(thk, decreasing = TRUE)
#'                       paste0(paste0(x[xord], " (t=", thk[xord], ")"), collapse = ", ")
#'                     } else
#'                       x
#'                   }
#'                 )
#'               )
#' profile_id(k) <- paste0(profile_id(k), "_collapse_custom")
#' 
#' unique(k$matrix_color_munsell)
#' 
#' # custom aggregation function for matrix_color_munsell (returns data.frame)
#' m <- collapseHz(jacobs2000,
#'                 pattern = c(
#'                   `A` = "^A",
#'                   `E` = "E",
#'                   `Bt` = "[ABC]+t",
#'                   `C` = "^C",
#'                   `foo` = "bar"
#'                 ),
#'                 AGGFUN = list(
#'                   matrix_color_munsell = function(x, top, bottom) {
#'                     thk <- bottom - top
#'                     if (length(x) > 1) {
#'                       xord <- order(thk, decreasing = TRUE)
#'                       data.frame(matrix_color_munsell = paste0(x, collapse = ";"),
#'                                  n_matrix_color = length(x))
#'                     } else {
#'                       data.frame(matrix_color_munsell = x, 
#'                                  n_matrix_color = length(x))
#'                     }
#'                   }
#'                 )
#'               )
#' profile_id(m) <- paste0(profile_id(m), "_collapse_custom")
#' 
#' m$matrix_color_munsell.n_matrix_color
collapseHz <- function(x,
                       pattern = NULL,
                       by = NULL,
                       hzdesgn = hzdesgnname(x, required = TRUE),
                       FUN = function(x, pattern, hzdesgn, ...) grepl(pattern, x[[hzdesgn]], ignore.case = FALSE),
                       ...,
                       AGGFUN = NULL,
                       ignore_numerics = NULL,
                       na.rm = FALSE) {
  idn <- idname(x)
  hzd <- horizonDepths(x)
  
  .screenDepths(x, hzd)
  
  # use exact match of existing genhz labels as default in lieu of pattern
  if (is.null(pattern) & missing(by)) {
    by <- GHL(x, required = TRUE)
  }
  
  if (length(pattern) == 0) {
    pattern <- NA
  }
  
  # if a named vector of patterns is given, use the names as new labels
  if (!is.null(names(pattern))) {
    labels <- names(pattern)
    pattern <- as.character(pattern)
  } else {
    # otherwise, the patterns and labels are the same
    pattern <- as.character(pattern)
    labels <- pattern
  }
  
  h <- data.table::data.table(horizons(x))
  
  # iterate over patterns
  for (p in seq(pattern)) {
    
    # calculate matches
    if (!is.null(by) && length(pattern) == 1 && is.na(pattern)) {
      
      if (!by %in% horizonNames(x)) {
        stop("Column name `by` (\"", by, ") is not a horizon-level variable.", call. = FALSE) 
      }
      
      labels <- h[[by]]
      
      if (any(is.na(labels))) {
        stop("Missing values are not allowed in `by` column argument", call. = FALSE)
      }
      
      r <- rle(paste0(h[[idn]], "-", as.character(labels)))
      l <- rep(TRUE, nrow(h))
    } else {
      l <- FUN(x, pattern = pattern[p], hzdesgn = hzdesgn, na.rm = na.rm, ...)
      r <- rle(l)
    }
    
    # only apply aggregation if there are adjacent horizons that match the target criteria
    if (any(r$lengths > 1)) {
      g <- unlist(lapply(seq_along(r$lengths), function(i) rep(i, r$lengths[i])))
      hidx <- unlist(lapply(seq_along(r$lengths), function(i) if (r$lengths[i] == 1) TRUE else rep(FALSE, r$lengths[i]))) & l
      gidx <- g %in% unique(g[l]) & !hidx
      naf <- names(AGGFUN)
      
      # iterate over sets of layers needing aggregation within each matching group
      if (sum(gidx) > 0){
        res <- h[gidx, c(list(hzdeptnew = suppressWarnings(min(.SD[[hzd[1]]], na.rm = na.rm)), 
                              hzdepbnew = suppressWarnings(max(.SD[[hzd[2]]], na.rm = na.rm))),
                         
                              # process numeric depth weighted averages w/ dominant condition otherwise                       
                              sapply(colnames(.SD)[!colnames(.SD) %in% c(hzd, naf)],
                                       function(n, top, bottom) {
                                        v <- .SD[[n]]
                                        if (length(v) > 1) {
                                          if (!n %in% ignore_numerics && is.numeric(v)) {
    
                                            # weighted average by thickness (numerics not in exclusion list)
                                            v <- weighted.mean(v, bottom - top, na.rm = na.rm)
    
                                          } else {
                                            # take thickest value
                                            # v[which.max(bottom - top)[1]]
    
                                            # convert factors etc to character
                                            # results may not conform with existing factor levels
                                            v <- as.character(v)
                                            
                                            # replace NA values for use in aggregate()
                                            if (!na.rm) {
                                              v[is.na(v)] <- "<collapseHZ-category-missing>"
                                            } 
                                            
                                            # take dominant condition (based on sum of thickness)
                                            cond <- aggregate(bottom - top, by = list(v), sum, na.rm = na.rm)
                                            v <- cond[[1]][which.max(cond[[2]])[1]]
                                            
                                            if (!na.rm) {
                                              v[v == "<collapseHZ-category-missing>"] <- NA
                                            }
                                          }
                                        }
                                        out <- data.frame(v)
                                        colnames(out) <- n
                                        out
                                      },
                                      top = .SD[[hzd[1]]],
                                      bottom = .SD[[hzd[2]]]),
                         
                             # process custom aggregation functions (may return data.frames)
                             do.call('c', lapply(colnames(.SD)[colnames(.SD) %in% naf], 
                                                 function(n, top, bottom) {
                                                           out <- AGGFUN[[n]](.SD[[n]], top, bottom)
                                                           if (!is.data.frame(out)) {
                                                             out <- data.frame(out)
                                                             colnames(out) <- n
                                                           } else {
                                                             colnames(out) <- paste0(n, ".", colnames(out))
                                                           }
                                                           out
                                                                             },
                                                 top = .SD[[hzd[1]]], 
                                                 bottom = .SD[[hzd[2]]]))), 
                 by = g[gidx]]
        # remove grouping ID
        res$g <- NULL
      } else {
        res <- h[0, ]
      }
      
      # allow for replacing values as well as adding new values with data.frame AGGFUN
      test1.idx <- na.omit(match(colnames(res), paste0(colnames(h), ".", colnames(h)))) 
      test2.idx <- na.omit(match(paste0(colnames(h), ".", colnames(h)), colnames(res)))
      colnames(res)[test2.idx] <- colnames(h)[test1.idx]
      
      # determine matches that are only a single layer (no aggregation applied)      
      res2 <- h[hidx & l, ]
      res2$hzdeptnew <- res2[[hzd[1]]]
      res2$hzdepbnew <- res2[[hzd[2]]]
      res2[[hzd[1]]] <- NULL
      res2[[hzd[2]]] <- NULL
      
      # combine matches
      res3 <- data.table::rbindlist(list(res, res2), fill = TRUE)
      if (missing(by) && nrow(res3) > 0){
        res3[[hzdesgn]] <- labels[p]
      }
      
      # combine matches with horizons that did not match
      agg.idx <- which(g %in% unique(g[l]) | hidx)
      if (length(agg.idx) > 0) {
        h <- h[-agg.idx, ]
      }
      h <- data.table::rbindlist(list(h, res3), fill = TRUE)
      
      # replace depths
      hn <- !is.na(h$hzdeptnew) & !is.na(h$hzdepbnew)
      h[[hzd[1]]][hn] <- h$hzdeptnew[hn]
      h[[hzd[2]]][hn] <- h$hzdepbnew[hn]
      h$hzdeptnew <- NULL
      h$hzdepbnew <- NULL
      
      # sort horizons by id name and top depth
      h <- h[order(h[[idn]], h[[hzd[1]]]),]
      
    }
    
    # replace horizons in parent SPC
    replaceHorizons(x) <- h
  }
  x
}

context("collapseHz()")

test_that("collapseHz works", {
  data("jacobs2000", package = "aqp")
  .BOTTOM <- NULL
  
  # use existing generalized horizon labels
  new_labels <- c("A", "E", "Bt", "Bh", "C")
  patterns <- c("A", "E", "B.*t", "B.*h", "C")

  # calculate a new SPC with genhz column based on patterns
  jacobs2000_gen <- generalizeHz(jacobs2000, new = new_labels, pattern = patterns)

  # create a missing value
  jacobs2000_gen$clay[19] <- NA
  
  # collapse that SPC based on genhz
  i <- collapseHz(jacobs2000_gen, hzdesgn = "genhz")
  expect_equal(length(jacobs2000), length(i))
  expect_equal(nrow(i), 26)
  expect_equal(i[7, , .BOTTOM], c(15, 41, 61, 132, 140, 152))

  # collapses adjacent horizons with same label
  i <- collapseHz(jacobs2000_gen, by = "genhz")
  ii <- collapseHz(jacobs2000_gen, by = "genhz", na.rm = TRUE)
  
  # no effect, horizon designations are unique within profiles
  j <- collapseHz(jacobs2000_gen, by = "name")
  
  expect_equal(nrow(j), 46)
  expect_equal(j[7, , .BOTTOM], jacobs2000[7, , .BOTTOM])
  
  # if using `by` argument, all values must not be NA
  expect_error(collapseHz(jacobs2000_gen, by = "matrix_color_munsell"),
               "Missing values are not allowed")
  
  # `by` column must also be a horizon-level variable
  expect_error(collapseHz(jacobs2000, by = "genhz"), "not a horizon-level variable")
  
  # matches input number of profiles
  expect_equal(length(jacobs2000), length(i))
  
  # horizons have been collapsed
  expect_equal(nrow(i), 26)
  
  # weighted mean (no NA values) works as expected (clay=47.15)
  expect_equal(i$clay[4],
               weighted.mean(jacobs2000_gen$clay[6:7], (jacobs2000_gen$bottom - jacobs2000_gen$top)[6:7]))
  
  # weighted mean (contains NA values, na.rm=FALSE) (clay is NA)
  expect_true(is.na(i$clay[11]))
  
  # weighted mean (contains NA values, na.rm=TRUE, clay=18.72414)
  expect_equal(ii$clay[11],
               weighted.mean(jacobs2000_gen$clay[17:20], (jacobs2000_gen$bottom - jacobs2000_gen$top)[17:20], na.rm = TRUE))
  
  # dominant condition (NA values retained)
  expect_true(is.na(i$depletion_munsell[13]))
  
  # dominant condition (NA values removed)
  expect_equal(ii$depletion_munsell[13], "10YR 8/2")
  
  plot(jacobs2000_gen, color = "concentration_pct")
  
  expect_equal(i[7, , .BOTTOM], c(15, 41, 61, 132, 140, 152))
  expect_true(is.numeric(i$clay))
  expect_true(is.numeric(j$clay))
  
  # "works" on empty SPC ()
  expect_equal(nrow(collapseHz(jacobs2000_gen[0,], by = "genhz")), 0)
                
  # works on SPC with filled profile (1 horizon with NA depths)
  all_na <- subsetHz(jacobs2000_gen[1,], TRUE)
  all_na$top <- NA_real_
  all_na$bottom <- NA_real_
  expect_warning(na_nonna <- c(all_na, jacobs2000_gen[2:5,]))
  expect_warning(f <- collapseHz(all_na, by = "genhz"), "contain NA")
  na_nonna$top[2] <- 19
  expect_warning(n <- collapseHz(na_nonna, by = "genhz"), "bottom depths are shallower than top")
  expect_equal(nrow(n), 14)

  
  a_pattern <- c(`A` = "^A",
                 `E` = "E", 
                 `Bt` = "[ABC]+t", 
                 `C` = "^C", 
                 `foo` = "bar")
  x <- collapseHz(jacobs2000, a_pattern)
  expect_equal(length(jacobs2000), length(x))
  expect_equal(nrow(x), 29)
  expect_true(is.numeric(x$clay))
  
  m <- collapseHz(jacobs2000,
                  pattern = a_pattern,
                  AGGFUN = list(
                    matrix_color_munsell = function(x, top, bottom) {
                      thk <- bottom - top
                      if (length(x) > 1) {
                        xord <- order(thk, decreasing = TRUE)
                        data.frame(matrix_color_munsell = paste0(x, collapse = ";"),
                                   n_matrix_color = length(x))
                      } else {
                        data.frame(matrix_color_munsell = x,
                                   n_matrix_color = length(x))
                      }
                    }
                  )
                )
  profile_id(m) <- paste0(profile_id(m), "_collapse_custom")

  expect_true(all(c("matrix_color_munsell", "matrix_color_munsell.n_matrix_color") %in% names(m)))
  expect_equal(nrow(m), 29)
})

#' @title Metrics of Contrast Suitable for Comparing Soil Colors
#'
#' @description Pair-wise comparisons of Munsell color specifications, based on the NCSS color contrast classes (Soil Survey Technical Note 2) and CIE delta-E 2000 metric.
#'
#' @param m1 vector of Munsell colors ('10YR 3/3'), must be >1 color if `m2` is not specified
#'
#' @param m2 optional vector of Munsell colors ('10YR 3/6')
#'
#' @details This function is fully vectorized over elements of `m1` and `m2` when both are of the same length. Recycling is not implemented. The minimum set of all pair-wise comparisons are performed within elements of `m1` when `m2` is not specified. For example, comparisons between A, B, and C are limited to *A*x*B*, *A*x*C*, and *B*x*C*. Use `expand.grid()` to generate suitable input from 1:many or many:1 type comparisons. See \href{http://ncss-tech.github.io/AQP/aqp/color-contrast.html}{this tutorial} for an expanded discussion and more examples. Neutral colors are not mentioned in SSTN2: in this function any comparison to a neutral color (e.g. 'N 3/') are assigned a delta-hue of 1. Since SSTN2 expects hues to be counted clock wise from 5R, it possible to get very large delta-hue values for otherwise adjacent colors: '5R' vs. '2.5R'. This will be addressed in an update to the standards.
#' 
#' The most meaningful representation of color contrast is the CIE2000 (dE00) metric.
#'
#' @return `data.frame` with the following columns:
#'
#'   * m1: Munsell color 1
#'   * m2: Munsell color 2
#'   * dH: delta-hue, as computed by `huePosition`
#'   * dV: delta-value, absolute value of difference in Munsell value (m1 vs. m2)
#'   * dc: delta-chroma, absolute value of difference in Munsell chroma (m1 vs. m2)
#'   * dE00: delta-E00, e.g. the [CIE delta-E as refined in 2000](https://en.wikipedia.org/wiki/Color_difference#CIEDE2000)
#'   * cc: soil color contrast class, as specified in Soil Survey Technical Note 2.
#'
#' @references
#'
#'  1. https://en.wikipedia.org/wiki/Color_difference
#'
#' @author D.E. Beaudette
#'
#' @note delta-E00 is computed by the \href{https://CRAN.R-project.org/package=farver}{farver package}.
#'
#' @seealso [colorContrastPlot], [huePosition], [huePositionCircle]
#'
#' @keywords manip
#' @export
#' 
#' @examples
#'
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#'
#' # two sets of colors to compare
#' m1 <- c('10YR 6/3', '7.5YR 3/3', '10YR 2/2', '7.5YR 3/4')
#' m2 <- c('5YR 3/4', '7.5YR 4/4', '2.5YR 2/2', '7.5YR 6/3')
#'
#' # contrast metrics
#' colorContrast(m1, m2)
#'
#' # adjacent chips
#' colorContrast('10YR 3/3', '10YR 3/4')
#' colorContrast('10YR 3/3', '7.5YR 3/3')
#'
#' # highly contrasting colors
#' # http://colour.granjow.net/fabercastell-polychromos.html
#' colorContrastPlot('10B 4/13', '10YR 10/15',
#' labels = c('helioblue-reddish', 'light cadmium yellow')
#' )
#' 
#' 
#' ## Note: neutral hues aren't defined in TN2
#' # approximation / extension of the concept
#' colorContrast(m1 = 'N 3/', m2 = 'N 6/')
#' 
#' colorContrast(m1 = '10YR 3/3', m2 = 'N 3/')
#' 
#' # pair-wise comparisons
#' m1 <- c('10YR 6/3', '7.5YR 3/3', '10YR 2/2', 'N 3/')
#' colorContrast(m1)
#' 
colorContrast <- function(m1, m2 = NULL) {
  
  # if m2 is unspecified, re-create m1 and m2 from all combinations (direction doesn't matter)
  if(is.null(m2)) {
    
    # m1 must be >1 color
    if(length(m1) < 2) {
      stop('m1 must contain >1 value if m2 is not specified')
    }
    
    # in case colors are encoded as factors
    m1 <- as.character(m1)
    
    # expand all combinations
    .combinations <- t(combn(m1, m = 2))
    
    m1 <- .combinations[, 1]
    m2 <- .combinations[, 2]
    
  } else {
    # must be more than 0 entries
    if(length(m1) < 1 | length(m2) < 1) {
      stop('both m1 and m2 must have length > 0')
    }
    
    # length of colors to compare should be equal if both are specified
    if(length(m1) != length(m2)) {
      stop('m1 and m2 must be the same length', call. = FALSE)
    }
    
    # in case colors are encoded as factors
    m1 <- as.character(m1)
    m2 <- as.character(m2)
  }
  
  
  # split into data.frame of hue/value/chroma
  m1.pieces <- parseMunsell(m1, convertColors = FALSE)
  m2.pieces <- parseMunsell(m2, convertColors = FALSE)
  
  # convert to value and chroma to numeric
  m1.pieces[[2]] <- as.numeric(m1.pieces[[2]])
  m1.pieces[[3]] <- as.numeric(m1.pieces[[3]])
  m2.pieces[[2]] <- as.numeric(m2.pieces[[2]])
  m2.pieces[[3]] <- as.numeric(m2.pieces[[3]])
  
  ## TODO: contrast with neutral hues are currently undefined in TN 2
  ## TODO: incorporate new huePosition features for locating angular proximity CW or CCW
  ## TODO: add a notes / flag field in the results
  
  # difference in number of hue chips, clock-wise, as specified in:
  # Soil Survey Technical Note 2 [wayback machine URL](https://web.archive.org/web/20220704214918/https://www.nrcs.usda.gov/wps/portal/nrcs/detail/soils/ref/?cid=nrcs142p2_053569)
  # including neutral hues in position 1
  dH <- abs(huePosition(m1.pieces[[1]], includeNeutral = TRUE) - huePosition(m2.pieces[[1]], includeNeutral = TRUE))
  
  ## TODO: not in standards -> propose changes
  # correction for neutral hues, enforce with XOR
  #   no N: dH is left alone
  #   single N: dH is always 1
  #   double N: dH is 0
  N.idx <- which(xor(m1.pieces[[1]] == 'N', m2.pieces[[1]] == 'N'))
  
  if(length(N.idx) > 0) {
    dH[N.idx] <- 1
  }
  
  
  # difference in number of value chips
  dV <- abs(m1.pieces[[2]] - m2.pieces[[2]])
  
  # difference in number of chroma chips
  dC <- abs(m1.pieces[[3]] - m2.pieces[[3]])
  
  # get CIE LAB representation
  m1.lab <- parseMunsell(m1, convertColors = TRUE, returnLAB = TRUE)
  m2.lab <- parseMunsell(m2, convertColors = TRUE, returnLAB = TRUE)
  
  
  ## TODO: there is likely room for improvement here:
  #         * parallel evaluation when length(m1) > 1000 (?) (simple via furrr)
  #         * vectorized call to compare_colour() -> reshaping of result
  
  # delta E00
  #
  # we don't need the full distance matrix,
  # iterate over rows, likely scales better
  d <- list()
  for(i in 1:nrow(m1.lab)){
    d[i] <- farver::compare_colour(m1.lab[i, ], m2.lab[i, ], from_space = 'lab', method = 'CIE2000', white_from = 'D65')
  }
  dE00 <- unlist(d)
  
  # NCSS color contrast classes
  # Soil Survey Technical Note 2 [wayback machine URL](https://web.archive.org/web/20220704214918/https://www.nrcs.usda.gov/wps/portal/nrcs/detail/soils/ref/?cid=nrcs142p2_053569)
  #
  # value1, chroma1, value2, chroma2, dH, dV, dC
  cc <- contrastClass(
    v1 = m1.pieces[[2]], 
    c1 = m1.pieces[[3]], 
    v2 = m2.pieces[[2]], 
    c2 = m2.pieces[[3]], 
    dH = dH, 
    dV = dV, 
    dC = dC
  )
  
  # combine into DF and return
  res <- data.frame(m1, m2, dH, dV, dC, dE00, cc)
  return(res)
}

# functions for argillic horizon detection in SPCs
# getArgillicBounds - upper and lower boundary of argillic horizon
# crit.clay.argillic - argillic horizon threshold function
# argillic.clay.increase.depth() - top depth of argillic

#' Estimate upper and lower boundary of argillic diagnostic subsurface horizon
#'
#' @param p A SoilProfileCollection
#' @param hzdesgn the name of the column/attribute containing the horizon designation; default="hzname"
#' @param clay.attr the name of the column/attribute containing the clay content; default="clay"
#' @param texcl.attr the name of the column/attribute containing the textural class (used for finding sandy horizons); default="texcl"
#' @param require_t require a "t" subscript for positive identification of upper and lower bound of argillic? default: TRUE
#' @param bottom.pattern regular expression passed to \code{estimateSoilDepth} to match the lower boundary of the soil. default is "Cr|R|Cd" which approximately matches paralithic, lithic and densic contacts.
#' @param lower.grad.pattern this is a pattern for adjusting the bottom depth of the argillic horizon upwards from the bottom depth of the soil. The absence of illuviation is used as a final control on horizon pattern matching.
#' @param sandy.texture.pattern this is a pattern for matching sandy textural classes: `-S$|^S$|COS$|L[^V]FS$|[^L]VFS$|LS$|LFS$`
#' @param verbose Print out information about 't' subscripts, sandy textures, plow layers and lower gradational horizons?
#' @param vertical.distance Vertical distance in which clay increase must be met. Default `30` cm
#' @param simplify Return a length 2 vector with upper and lower boundary when `p` has length 1? Default `TRUE`.
#'
#' @description \code{getArgillicBounds} estimates the upper and lower boundary of argillic diagnostic subsurface horizon for a profile in a single-profile SoilProfileCollection object (`p`).
#'
#' The upper boundary is where the clay increase threshold is met. The function uses \code{crit.clay.argillic} as the threshold function for determining whether a clay increase occurs and \code{get.increase.matrix} to determine whether the increase is met, whether vertical distance of increase is sufficiently small, and in which horizon.
#'
#'@details The lower boundary is first approximated as the depth to a lithic/paralithic/densic contact, or some other horizon matchable by a custom regular expression pattern. Subsequently, that boundary is extended upwards to the end of "evidence of illuviation."
#'
#' The depth to contact is estimated using 'bottom.pattern' "Cr|R|Cd" by default. It matches anything containing Cr, R or Cd.
#'
#' The lower gradational horizon regular expression ‘lower.grad.pattern' default is `^[2-9]*B*CB*[^rtd]*[1-9]*$}`. It matches anything that starts with a lithologic discontinuity (or none) and a C master horizon designation. May contain B as second horizon designation in transitional horizon. May not contain 'r' or 't' subscript.
#'
#' The minimum thickness of the argillic horizon is dependent on whether all subhorizons are "sandy" or not. The \code{sandy.texture.pattern} default `-S$|^S$|COS$|L[^V]FS$|[^L]VFS$|LS$|LFS$` captures USDA textural class fine earth fractions that meet "sandy" particle size class criteria.
#'
#' There also is an option ‘require_t' to omit the requirement for evidence of eluviation in form of 't' subscript in 'hzdesgn'. Even if "t" subscript is not required for positive identification, the presence of lower gradational C horizons lacking 't' will still be used to modify the lower boundary upward from a detected contact, if needed. If this behavior is not desired, just set 'lower.grad.pattern' to something that will not match any horizons in your data.
#'
#' @author Andrew G. Brown
#'
#' @return Returns a numeric vector; first value is top depth, second value is bottom depth. If as.list is TRUE, returns a list with top depth named "ubound" and bottom depth named "lbound". If `p` has more than one profile or if `simplify = FALSE` the result is a data.frame containing profile ID, upper and lower boundary columns.
#'
#' @export
#'
#' @examples
#' data(sp1, package = 'aqp')
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' # set required metadata
#' hzdesgnname(sp1) <- 'name'
#' hztexclname(sp1) <- 'texture'
#' hzmetaname(sp1, 'clay') <- 'prop'
#' 
#' x <- getArgillicBounds(sp1)
#' x
#'
getArgillicBounds <- function(p,
                              hzdesgn = hzdesgnname(p, required = TRUE),
                              clay.attr = hzmetaname(p, 'clay', required = TRUE),
                              texcl.attr = hztexclname(p, required = TRUE),
                              require_t = TRUE,
                              bottom.pattern = "Cr|R|Cd",
                              lower.grad.pattern = "^[2-9]*B*CB*[^rtd]*[1-9]*$",
                              sandy.texture.pattern = "-S$|^S$|COS$|L[^V]FS$|[^L]VFS$|LS$|LFS$",
                              vertical.distance = 30,
                              simplify = TRUE,
                              verbose = FALSE) {
  hz <- horizons(p)
  hzd <- horizonDepths(p)

  if (is.null(hzdesgn) || !hzdesgn %in% horizonNames(p)) {
    stop("Horizon designation column (", hzdesgn, ") does not exist.")
  }
  
  if (is.null(clay.attr) | (!clay.attr %in% horizonNames(p))) {
    stop("Horizon clay content column (", texcl.attr, ") does not exist.")
  }
  
  if (is.null(texcl.attr) || !texcl.attr %in% horizonNames(p)) {
    stop("Horizon texture class column (", texcl.attr, ") does not exist.")
  }
  
  # get upper bound...
  mss <- getMineralSoilSurfaceDepth(p, hzdesgn = hzdesgn, simplify = FALSE)
  pld <- getPlowLayerDepth(p, hzdesgn = hzdesgn, simplify = FALSE)

  # estimate the thickness of the soil profile
  # (you will need to specify alternate pattern if Cr|R|Cd
  # doesn't match your contacts)
  soil.depth <- minDepthOf(p, pattern = bottom.pattern, hzdesgn = hzdesgn, simplify = FALSE)

  # find all horizons with t subscripts; some old/converted horizons have all capital letters
  has_t <- grepl(as.character(hz[[hzdesgn]]), pattern = "[Tt]")
  
  # clay attribute and clay thresholds
  ivar <- hz[[clay.attr]]
  thresh <- crit.clay.argillic(ivar) # threshold.fun()
  
  # in lieu of plow layer and LD, the clay increase depth determines upper bound
  # upper.bound <- argillic.clay.increase.depth(p, clay.attr)
  .N <- NULL; .SD <- NULL; increase.var <- NULL; threshold.vector <- NULL; vdist <- NULL; middepth <- NULL; .hzID <- NULL
  dt <- data.table::data.table(
    id = hz[[idname(p)]],
    middepth = hz[[hzd[1]]] + (hz[[hzd[2]]] - hz[[hzd[1]]]) / 2, # TODO: evaluate midpoint
    increase.var = ivar,
    threshold.vector = thresh,
    .hzID = 1:nrow(hz)
  )
  
  .dmax <- function(x, n) {
    res <- c(0, diff(c(0, x[-n])))
    if(length(res) == 0) 
      return(NA_real_)
    res
  }
  
  dt$vdist <- dt[, list(vdist = .dmax(middepth, .N)), by = "id"]$vdist
  increase.met <- dt[, list(increase.met = .hzID[which(increase.var[2:.N] > threshold.vector[1:(.N - 1)] &
                                                         vdist[2:.N] <= vertical.distance)[1]]), 
                     by = "id"]$increase.met
  upper.bound <- hz[increase.met + 1, ][[hzd[1]]]
  lower.bound <- rep(NA, length(upper.bound))
  
  # handle case where Ap disturbs upper bound of argillic
  shallowest_t <- minDepthOf(p, pattern = "[Tt]", hzdesgn = hzdesgn, simplify = FALSE)
  idx2 <- which(pld[[hzd[2]]] == shallowest_t[[hzd[1]]] & pld[[hzd[2]]] != 0)
  upper.bound[idx2] <- shallowest_t[[hzd[1]]][idx2]
  
  # or a lithologic discontinuity overrides the clay increase req
  depth_ld <- depthOf(p, pattern = "^[2-9].*[Tt]", hzdesgn = hzdesgn, simplify = FALSE)
  idx3 <- which(shallowest_t[[hzd[1]]] %in% depth_ld[[hzd[1]]])
  upper.bound[idx3] <- shallowest_t[[hzd[1]]][idx3]

  deepest_t <- maxDepthOf(p, pattern = "[Tt]", hzdesgn = hzdesgn, top = FALSE, simplify = FALSE)
  c_idx <- minDepthOf(p, pattern = lower.grad.pattern, hzdesgn = hzdesgn, simplify = FALSE)
  
  idx4 <- which(!is.na(upper.bound))
  lower.bound[idx4] <- pmin(deepest_t[[hzd[2]]], c_idx[[hzd[1]]], na.rm = TRUE)[idx4]
  
  # calculate baseline minimum rhickness
  min.thickness <- pmax(7.5, soil.depth[[hzd[1]]] / 10)
  textures <- data.table::data.table(glom(p, upper.bound, lower.bound, df = TRUE))
  
  # sandy textures require 15cm minimum thickness
  idx5 <- which(textures[, list(is_sandy = all(grepl(sandy.texture.pattern, .SD[[texcl.attr]]))),
                   by = eval(idname(p))]$is_sandy)
  min.thickness[idx5] <- 15
  
  idx6 <- which((lower.bound - upper.bound ) < min.thickness)
  upper.bound[idx6] <- NA
  lower.bound[idx6] <- NA
  
  if (length(p) == 1 && simplify) {
    return(c(ubound = upper.bound, lbound = lower.bound))
  }
  res <- data.frame(id = profile_id(p), ubound = upper.bound, lbound = lower.bound)
  colnames(res)[1] <- idname(p)
  .as.data.frame.aqp(res, aqp_df_class(p))
}
#' Determines threshold (minimum) clay content for argillic upper bound
#'
#' Given a vector or matrix of "eluvial" horizon clay contents (\%),
#' \code{crit.clay.argillic()} returns a vector or matrix of minimum clay
#' contents (thresholds) that must be met for an argillic horizon clay
#' increase.
#'
#' Uses the standard equations for clay contents less than 15 \%, between 15
#' and 40 \%, and greater than 40 \%. Based on the clay increase criteria in
#' the definition of the argillic horizon from 12th Edition Keys to Soil
#' Taxonomy (Soil Survey Staff, 2014).
#'
#'
#' @param eluvial_clay_content A numeric vector or matrix containing clay
#' contents of potential "eluvial" horizons. May contain \code{NA}.
#' @return A vector or matrix (input-dependent) containing minimum "illuvial"
#' horizon clay contents (thresholds) to be met for argillic horizon clay
#' increase.
#' @note This function is intended for identifying clay content threshold
#' required for an argillic horizon. These thresholds may not apply depending
#' on the specifics of your soil. E.g. if the upper part of argillic has been
#' plowed (has Ap immediately over upper boundary) the clay increase
#' requirement can be waived (Soil Survey Staff, 2014).
#' @author Andrew Gene Brown
#' @seealso \code{\link{getArgillicBounds}}, \code{\link{get.increase.matrix}}
#' @references Soil Survey Staff. 2014. Keys to Soil Taxonomy, 12th ed.
#' USDA-Natural Resources Conservation Service, Washington, DC.
#' @keywords manip
#' @export
#' @examples
#'
#' # crit.clay.argillic uses different equations for clay content
#' # less than 15 %, between 15 and 40 %, and >40 %
#'
#' crit.clay.argillic(eluvial_clay_content=c(5, 20, 45))
#'
crit.clay.argillic <- function(eluvial_clay_content) {
  # eluvial clay content is a numeric vector or matrix subsettable with logical vectors based on clay (NA omitted)
  buf <- eluvial_clay_content
  idx.mask <- is.na(buf)
  buf[idx.mask] <- 0

  idx.lt15 <- buf < 15
  idx.lt15[idx.mask] <- FALSE

  idx.gt40 <- buf >= 40
  idx.gt40[idx.mask] <- FALSE

  idx.other <- (buf >= 15) & (buf < 40)
  idx.other[idx.mask] <- FALSE

  buf[idx.lt15] <- eluvial_clay_content[idx.lt15] + 3
  buf[idx.gt40] <- eluvial_clay_content[idx.gt40] + 8
  buf[idx.other] <- 1.2*eluvial_clay_content[idx.other]
  return(round(buf))
}

#' Return upper boundary of argillic horizon
#' 
#' Returns the top depth of the argillic horizon as a numeric vector.
#' 
#' Uses \code{crit.clay.argillic} to determine threshold clay increase, and
#' \code{get.increase.matrix} to determine where increase is met within a
#' vertical distance of 30 cm.
#' 
#' 
#' @param p A single-profile \code{SoilProfileCollection} object.
#' @param clay.attr OPTIONAL: horizon attribute name referring to clay content.
#' default: `clay`
#' @return A numeric vector containing top depth of argillic horizon, if
#' present, or NA.
#' @author Andrew Gene Brown
#' @seealso \code{getArgillicBounds}, \code{get.increase.matrix},
#' \code{crit.clay.argillic}
#' @keywords manip
#' @export
#' @examples
#' 
#' data(sp1, package = 'aqp')
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#' 
#' p <- sp1[1]
#' attr <- 'prop' # clay contents 
#' foo <- argillic.clay.increase.depth(p, clay.attr = attr)
#' foo
#' 
argillic.clay.increase.depth <- function(p, clay.attr = 'clay') {
  vd <- 30
  return(get.increase.depths(p, attr = clay.attr,
                             threshold.fun = crit.clay.argillic,
                             vertical.distance = vd)[1])
}

context("estimate cambic boundaries")

# construct a fake profile
# one dark colored horizon (AB) and two cambic subhorizons Bw + BC
spc <- data.frame(id = 1, taxsubgrp = "Lithic Haploxerepts",
                  hzname   = c("A","AB","Bw","BC","R"),
                  hzdept   = c(0,  20, 32, 42,  49),
                  hzdepb   = c(20, 32, 42, 49, 200),
                  clay     = c(19, 22, 22, 21,  NA),
                  texcl    = c("l","l","l","l","br"),
                  d_value  = c(5,   5,  5,  6,  NA),
                  m_value  = c(2.5, 3,  3,  4,  NA),
                  m_chroma = c(2,   3,  4,  4,  NA))

# promote to SoilProfileCollection
depths(spc) <- id ~ hzdept + hzdepb
hzdesgnname(spc) <- 'hzname'
hztexclname(spc) <- 'texcl'
hzmetaname(spc, "clay") <- 'clay'

test_that("getCambicBounds - basic functionality", {
  dfbound <- getCambicBounds(spc)
  expect_equal(nrow(dfbound), 1)
  expect_equal(as.numeric(dfbound[,c("cambic_top","cambic_bottom")]), c(32,49))

  # exclude by entry of non-cambic bounds
  expect_true(is.na(getCambicBounds(spc, argi_bounds = c(32,49))$cambic_index))

  # empty spc input
  expect_error(getCambicBounds(spc[0,]))
})

test_that("getCambicBounds - special cases", {
  spc2 <- spc
  spc2$texcl <- "S" # all sandy textures
  dfbound <- getCambicBounds(spc2)
  expect_true(is.na(dfbound$cambic_index))

  spc2 <- spc
  spc2$hzname <- c("A","Bt","BE","Bhs","Bt'")
  spc2$clay <- c(11,18,16,17,22)
  dfbound <- getCambicBounds(spc2)
  expect_true(is.na(dfbound$cambic_index))
})

test_that("getArgillicBounds - error conditions", {
  expect_error(getArgillicBounds(spc2, hzdesgn = "foo"))
  expect_error(getArgillicBounds(spc2, texcl.attr = "foo"))
  expect_error(getArgillicBounds(spc2, clay.attr = "foo"))
})

context("argillic bounds and clay increase")

data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

p <- sp1[1]
hzdesgnname(p) <- "name"
hztexclname(p) <- "texture"
clay.attr <- 'prop' # clay contents %
texcl.attr <- 'texture' # class containing textural class (for finding sandy textures)

#implicitly testing the basic logic of this function :)
threshold.fun <- crit.clay.argillic

#standard for argillic
vertical.distance <- 30

test_that("get.increase.matrix() (used for getArgillicBounds())", {
  m <- get.increase.matrix(p, clay.attr, threshold.fun, vertical.distance)

  # correct data type?
  expect_true(inherits(m, 'matrix'))

  # is square matrix?
  expect_true(nrow(m) == ncol(m))

  # no TRUE in lower triangle?
  expect_true(all(!(m * lower.tri(m))))

  # correct number of eluvial horizons identified in first illuvial column?
  # this is also a test of the underlying crit.clay.argillic
  expect_equal(colSums(m)[4], 2)
})

test_that("argillic.clay.increase() (used for getArgillicBounds())", {
  #make sure the main wrapper method for the increase functions works
  d <- argillic.clay.increase.depth(p, clay.attr = 'prop')

  # this is also a test of the underlying crit.clay.argillic
  expect_equal(d, 49)
})

test_that("getArgillicBounds()", {

  # name and texture class are determined from metadata
  d <- getArgillicBounds(p, clay.attr = 'prop')
  
  # this makes sure estimateSoilDepth() isn't broken...
  expect_equivalent(d, c(49, 89))
  
  # error when hzdesgn and texcl.attr are unknown; no guessing
  expect_error(getArgillicBounds(p, hzdesgn = 'foo', clay.attr = 'prop', texcl.attr =  'bar'))
})

test_that("getArgillicBounds - error conditions", {
  expect_error(getArgillicBounds(p, hzdesgn = "foo"))
  expect_error(getArgillicBounds(p, texcl.attr = "foo"))
  expect_error(getArgillicBounds(p, clay.attr = "foo"))
})

#estimatePSCS()
#' Estimate boundaries of the U.S Soil Taxonomy Particle Size Control Section
#' 
#' Estimates the upper and lower boundary of the particle size control section
#' for Mineral or Organic soilsby applying a programmatic version of the 
#' particle size control section key from the Keys to Soil Taxonomy (13th edition).
#' See details for assumptions and required data elements.
#' 
#' @details
#'
#' Requires information to identify argillic horizons (clay contents, horizon
#' designations) with \code{getArgillicBounds()} as well as the presence of
#' plow layers and surface organic soil material. Any
#' \code{getArgillicBounds()} arguments may be passed to \code{estimatePSCS}.
#'
#' Also, requires information on taxonomic order to handle Andisols.
#' 
#' In aqp 2.1, particle size control sections of organic soils Histosols and
#' Histels are supported. This requires setting the `"lieutex"` horizon metadata
#' column using `hzmetaname<-()` Horizon designations containing `"f"` or `"W"`
#' are  recognized as permafrost and water layers, respectively, for application 
#' of the organic soils key to control sections. In lieu textures `"SPM"` and
#' `"PEAT"` are used to identify low density organic materials used for surface
#' tier thickness. To avoid using the 160 cm surface tier, set the `"lieutex"` 
#' column to any column that does not contain `"SPM"` or `"PEAT"` values.
#'
#' WARNING: Soils in arenic or grossarenic subgroups, with fragipans, or with
#' strongly contrasting PSCs may not be classified correctly. The author would
#' welcome a dataset to develop this functionality for.
#' 
#' @param p A SoilProfileCollection 
#' @param hzdesgn Name of the horizon attribute containing the horizon
#' designation. Default `hzdesgnname(p, required = TRUE)`
#' @param clay.attr Name of the horizon attribute containing clay contents.
#' Default `hzmetaname(p, "clay", required = TRUE)`
#' @param texcl.attr Name of the horizon attribute containing textural class
#' (used for finding sandy textures). Default `hztexclname(p, required = TRUE)`
#' @param tax_order_field Name of the site attribute containing taxonomic
#' order; for handling PSCS rules for Andisols in lieu of lab data. May be NA
#' or column missing altogether, in which case Andisol PSC possibility is
#' ignored.
#' @param lieutex Optional data element used in addition to the horizon
#' designation to identify kinds of organic soil material for soils with organic 
#' surfaces. Default: `hzmetaname(p, "lieutex")`
#' @param bottom.pattern Regular expression pattern to match a root-restrictive
#' contact. Default matches Cr, R, Cd or m. This argument is passed to both
#' `minDepthOf()` and `getArgillicBounds()`.
#' @param simplify Return a length 2 vector with upper and lower boundary when 
#' p has length 1? Default TRUE.
#' @param ...  additional arguments are passed to getArgillicBounds()
#' @return A numeric vector (when `simplify=TRUE`) containing the top and bottom 
#' depth of the particle
#' size control section. First value is top, second value is bottom. 
#' If `p` contains more than one profile, the result is a data.frame with 
#' profile ID plus PSCS top and bottom depths.
#' @author Andrew Gene Brown
#' @seealso [`getArgillicBounds()`], [`getSurfaceHorizonDepth()`]
#' @references Soil Survey Staff. 2014. Keys to Soil Taxonomy, 12th ed.
#' USDA-Natural Resources Conservation Service, Washington, DC.
#' @keywords manip
#' @export 
#' @examples
#'
#' data(sp1, package = 'aqp')
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' # set required metadata
#' hzdesgnname(sp1) <- 'name'
#' hztexclname(sp1) <- 'texture'
#' hzmetaname(sp1, 'clay') <- 'prop'
#' 
#' x <- estimatePSCS(sp1)
#' x
#' 
#' # change horizon texture and set inlieu texture column to turn
#' # first profile into an organic soil
#' sp1$name[1:6] <- c("Oi1", "Oi2", "Oi3", "Oaf", "Cf1", "Cf2")
#' sp1$texture <- as.character(sp1$texture)
#' sp1$texture[1:6] <- c("PEAT", "PEAT", "PEAT", "MUCK", "GRVLS", "GRVLS")
#' sp1$bottom[6] <- 200
#' hzmetaname(sp1, 'lieutex') <- 'texture'
#' 
#' y <- estimatePSCS(sp1[1, ], simplify = FALSE)
#' 
#' # 74cm lower boundary is 25cm past the upper boundary of permafrost (49cm)
#' # but minimum depth is 100cm unless there is a root-limiting layer
#' y
#' 
estimatePSCS <- function(
    p,
    hzdesgn = hzdesgnname(p, required = TRUE),
    clay.attr = hzmetaname(p, "clay", required = TRUE),
    texcl.attr = hztexclname(p, required = TRUE),
    lieutex = hzmetaname(p, "lieutex"),
    tax_order_field = "tax_order",
    bottom.pattern = 'Cr|R|Cd|m',
    simplify = TRUE,
    ...
) {
  
  .LAST <- NULL
  hz.depths <- horizonDepths(p)

  attr.len <- unlist(lapply(c(hzdesgn, clay.attr, texcl.attr), length))
  if (any(attr.len > 1))
    stop("horizon designation, clay attribute or texture class attribute must have length 1")
  
  if (is.null(hzdesgn) || !hzdesgn %in% horizonNames(p)) {
    stop("Horizon designation column (", hzdesgn, ") does not exist.")
  }
  
  if (is.null(texcl.attr) || !texcl.attr %in% horizonNames(p)) {
    stop("Horizon texture class column (", texcl.attr, ") does not exist.")
  }
  
  if (is.null(clay.attr) | (!clay.attr %in% horizonNames(p))) {
    stop("Horizon clay content column (", clay.attr, ") does not exist.")
  }

  soildepth <- minDepthOf(p, hzdesgn = hzdesgn, pattern = bottom.pattern, simplify = FALSE)[[hz.depths[1]]]
  nobottomidx <- is.na(soildepth)
  if (any(nobottomidx)) {
    soildepth[nobottomidx] <- p[which(nobottomidx), , .LAST][[hz.depths[2]]]
  }
  
  andisols_flag <- rep(FALSE, length(soildepth))
  shallow_flag <- rep(FALSE, length(soildepth))

  # Parts D (argillic starts >100cm  depth) and F (all other mineral soils)
  default_t <- rep(25, length(soildepth))
  default_b <- rep(100, length(soildepth))
  
  odepth <- getMineralSoilSurfaceDepth(p, hzdesgn = hzdesgn, simplify = FALSE)[[hz.depths[2]]]
  othick <- thicknessOf(p, pattern = "O", hzdesgn = hzdesgn)$thickness
  ohzidx <- which(odepth > 0)
  
  # Key part A (soils with restriction in shallow depth)
  lt36idx <- which(soildepth - odepth <= 36)
  if (length(lt36idx) > 0) {
    default_t[lt36idx] <- 0 # O horizon correction applied below
    default_b[lt36idx] <- soildepth[lt36idx]
    shallow_flag[lt36idx] <- TRUE
  }
  
  # Key part B (Andisols)
  if (tax_order_field %in% siteNames(p)) {
    andidx <- grep("[Aa]ndisols", p[[tax_order_field]])
    if (length(andidx) > 0) {
      default_t[andidx] <- 0
      default_b[andidx] <- 100
      andisols_flag[andidx] <- TRUE
    }
  }

  # Adjust PSCS range downward if organic soil material is present at surface (i.e. mineral soil surface depth > 0)
  if (length(ohzidx) > 0) {
    
    default_t[ohzidx] <- default_t[ohzidx] + odepth[ohzidx]
    maxdtruncidx <- which(default_b != soildepth)
    
    if (length(maxdtruncidx) > 0) {
      default_b[maxdtruncidx] <- default_b[maxdtruncidx] + odepth[maxdtruncidx]
    }
  }

  # Key parts C and E (has argillic/kandic/natric WITHIN 100CM)

  if (any(!andisols_flag)) {
    argillic_bounds <- getArgillicBounds(
      p,
      clay.attr = clay.attr,
      texcl.attr = texcl.attr,
      hzdesgn = hzdesgn,
      bottom.pattern = bottom.pattern,
      ...,
      simplify = FALSE # always return data.frame
    )
    
    lt100argidx <- which(argillic_bounds$ubound < 100)
    thinargidx <- argillic_bounds$lbound - argillic_bounds$ubound <= 50
    shallowargidx <- which(argillic_bounds$lbound <= 25)
    
    if (length(lt100argidx) > 0) {
      default_t[lt100argidx] <- argillic_bounds$ubound[lt100argidx]
    }
    
    if (length(which(thinargidx)) > 0) {
      default_b[which(thinargidx)] <- argillic_bounds$lbound[which(thinargidx)]
    }
    
    if (length(which(!thinargidx)) > 0) {
      default_b[which(!thinargidx)] <- argillic_bounds$ubound[which(!thinargidx)] + 50
    }
    
    if (length(shallowargidx) > 0) {
      default_b[shallowargidx] <- 100
    }
  }

  # Adjust PSCS top depth to bottom of plow layer (if appropriate)
  plow_layer_depth <- getPlowLayerDepth(p, hzdesgn, simplify = FALSE)[[hz.depths[2]]]
  plowidx <- which(plow_layer_depth >= 25 + odepth)
  if (length(plowidx) > 0) {
    default_t[plowidx] <- plow_layer_depth[plowidx]
  }
  
  # Adjust PSCS top depth to mineral soil surface for soils <36cm to restriction
  shallowoidx <- which(shallow_flag & default_t != 0)
  if (length(shallowoidx) > 0) {
    default_t[shallowoidx] <- odepth[shallowoidx]
  }
  
  # handle thick o horizons that cause the mineral PSCS to be greater than soil depth
  organic.idx <- which(default_t >= soildepth | odepth >= 40 |
                       (othick >= (soildepth * 2 / 3) & (soildepth - othick) <= 10))
  if (length(organic.idx) > 0) {
    organic_cs <- histosol.control.section(p[organic.idx,], hzdesgn = hzdesgn, lieutex = lieutex)
    default_t[organic.idx] <- organic_cs[[2]]
    default_b[organic.idx] <- organic_cs[[3]]
  }
  
  # Adjust PSCS bottom depth to restriction depth, if appropriate
  restridx <- which(soildepth < default_b)
  if (length(restridx) > 0) { #truncate to restriction
    default_b[restridx] <- soildepth[restridx]
  }
  
  if (length(default_t) == 1 && simplify) {
    return(as.numeric(c(default_t, default_b)))
  }
  
  res <- data.frame(id = profile_id(p), pscs_top = default_t, pscs_bottom = default_b)
  names(res)[1] <- idname(p)
  res
}

# calculate control section based on materials in surface tier, root limiting layers, permafrost and water
histosol.control.section <- function(p,
                                     hzdesgn = hzdesgnname(p, required = TRUE),
                                     lieutex = hzmetaname(p, "lieutex", required = TRUE)) {
  depthz <- horizonDepths(p)
  
  # control section stops at Cr|R|Cd
  rll.depth <- minDepthOf(p, pattern = "Cr|R|Cd|m", hzdesgn = hzdesgn, no.contact.depth = 0, no.contact.assigned = 200, simplify = FALSE)[[depthz[1]]]
  
  # control section stops at permafrost + 25cm (must be at least 100cm)
  permafrost.depth <- minDepthOf(p, "f|ff", hzdesgn = hzdesgn, no.contact.assigned = Inf, simplify = FALSE)[[depthz[1]]]
  
  # control section stops at water bottom depth  greater than either 130 or 160 cm, as applicable
  water.depth <- minDepthOf(p, "W", hzdesgn = hzdesgn, no.contact.assigned = Inf, simplify = FALSE)[[depthz[1]]]
  
  # get surface tier thickness for p
  hst <- histosol.surface.tier(p, hzdesgn = hzdesgn, lieutex = lieutex, rll = rll.depth)$hst_b
  
  data.frame(
    peiid = profile_id(p),
    hcs_t = 0,
    hcs_b = pmin(rll.depth, pmax(permafrost.depth + 25, 100), pmin(water.depth, ifelse(hst == 30, 130, 160)))
  )
  
}

histosol.surface.tier <- function(p,
                                  hzdesgn = hzdesgnname(p, required = TRUE),
                                  lieutex = hzmetaname(p, "lieutex", required = TRUE),
                                  rll = NULL) {
    
  .SD <- NULL; .LAST <- NULL; .BOTTOM <- NULL; .TOP <- NULL; .thk <- NULL
  depthz <- horizonDepths(p)
  
  res <- data.frame(
    peiid = profile_id(p),
    hst_t = 0,
    hst_b = 30
  )
  
  # column "lieutex" used in lieu of texture, and in lieu of horizon designations (Oa, Oe, Oi)
  if (!lieutex %in% names(p)) {
    # this will only trigger if one or pedons appears to meet the organic layer thickness requirements
    stop("The in lieu texture ('lieutex') is required to apply rules for organic soils. 
    Set the 'lieutex' horizon designation metadata with `hzmetaname(p, 'lieutex') <- 'lieutex'`
    Soils with low density fibric materials are indicated using lieutex column values `'PEAT'` or `'SPM'`.")
  }
  
  mss <- getMineralSoilSurfaceDepth(p, hzdesgn = hzdesgn, simplify = FALSE)
  
  # if there is a mineral surface,  no surface tier
  noo.idx <- which(mss[[depthz[2]]] == 0)
  
  psub <- p
  if (length(noo.idx) > 0) {
    psub <- p[-noo.idx, ]
  }
  
  if (is.null(rll)) {
    rll <- minDepthOf(psub, hzdesgn = hzdesgn, pattern = "Cr|Cd|R|m", no.contact.depth = 0, no.contact.assigned = 200, simplify = FALSE)[[depthz[1]]]
  }
  
  p60 <- glom(psub, 0, pmin(rll, 60), truncate = TRUE, drop = FALSE)
  
  # allow for multiple lines of evidence for fibric/low density OM
  #  either lieutex of peat or "i" suffix throughout all horizons in upper 60
  
  hz <- data.table::data.table(horizons(p60))
  idx <- which(grepl("i", hz[[hzdesgn]]) | tolower(hz[[lieutex]]) %in% c("peat", "spm"))
  hz$.thk <- 0
  hz$.thk[idx] <- (hz[[depthz[2]]] - hz[[depthz[1]]])[idx]
  
  # calculate sum of fibric material thickness within upper 60cm, compare to total hz thickness
  fibric.thickness <- hz[, sum(.thk, na.rm = TRUE), by = c(idname(p))]$V1
  total.hz.thickness <- p60[, , .LAST, .BOTTOM] - p60[, 1, .TOP]
  
  # if all fibric in upper 60, the limit is 60
  res$hst_b[fibric.thickness >= 60] <- 60
  
  return(res)
}

# pscs unit tests

context("particle size control section estimator")

data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

p <- sp1[1,]
attr <- 'prop' # clay contents %

q <- sp1[2,]

x <- data.frame(
  peiid = 706300,
  taxsubgrp = "Lithic Humicryods",
  top = c(0, 13, 16, 18, 24, 40),
  bottom = c(13, 16, 18, 24, 40, 65),
  name = c("Oi", "A", "E", "Bhs", "2C", "2R"),
  texture = c("SPM", "SIL", "SIL", "SIL", "SIL", "BR"),
  prop = c(0, 6, 6, 6, 6, 6)
)
depths(x) <- peiid ~ top + bottom
site(x) <- ~ taxsubgrp

test_that("estimatePSCS()", {

  # this soil has a clay decrease then a clay increase and an argillic horizon
  # the argillic horizon ends at a bedrock contact
  e <- estimatePSCS(p, clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'name')
  expect_equivalent(e, c(49, 89))

  # this soil does not have an argillic, so it is 25-100 but has 5cm thick O horizon
  # and is moderately deep to bedrock contact
  g <- estimatePSCS(q, clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'name')
  expect_equivalent(g, c(30, 59))

  ## special cases

  # thick (>50cm) Bt
  qbigbt <- sp1[3]
  qbigbt$name <- c("A","Bt1","Bt2","2Bt3", "2Bt4")
  g <- estimatePSCS(qbigbt, clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'name')
  expect_equivalent(g, c(2, 52))

  # thick Ap (>25cm)
  qbigap <- sp1[3]
  qbigap$name <- c("Ap1","Ap2","Ap3","C1", "C2")
  g <- estimatePSCS(qbigap, clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'name')
  expect_equivalent(g, c(35, 67))

  # soil less than 36cm deep
  qshallow <- trunc(q, 0, 27)
  g <- estimatePSCS(qshallow, clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'name')
  expect_equivalent(g, c(5, 27))

  # andisols
  qandisol <- q
  qandisol$tax_order <- "Andisols"
  g <- estimatePSCS(qandisol, clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'name')
  expect_equivalent(g, c(5, 59)) # note: starts at bottom of OSM

  # very shallow (bottom depth <25cm) argillic?
  qminiargi <- sp1[3]
  qminiargi$name <- c("A","Bt","C1","2C2", "2C3") #idk...
  g <- estimatePSCS(qminiargi, clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'name')
  expect_equivalent(g, c(2, 67)) # NOT 2, 13; would be 2,100 without limiting layer

  # error conditions
  q2 <- q
  expect_error(estimatePSCS(q2, clay.attr = 'foo', texcl.attr = "texture", hzdesgn = 'name'))
  q2$texture <- NULL
  expect_error(estimatePSCS(q2, clay.attr = 'prop', texcl.attr = "foo", hzdesgn = 'name'))
  q2 <- q
  q2$name <- NULL
  expect_error(estimatePSCS(q2, clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'foo'))
})

test_that("estimatePSCS() thin soil profile with O horizon", {
  expect_equal(estimatePSCS(x, clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'name'), c(13, 40))
  expect_equal(estimatePSCS(c(q,x), clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'name'),
               data.frame(id = c("706300", "P002"), 
                          pscs_top = c(13, 30),
                          pscs_bottom = c(40, 59)))
})

test_that("estimatePSCS() multiple profiles",{
  e <- estimatePSCS(sp1, clay.attr = 'prop', texcl.attr = "texture", hzdesgn = 'name')
  expect_equal(e$pscs_top, c(49, 30, 2, 32, 5, 31, 25, 27, 28))
  expect_equal(e$pscs_bottom, c(89, 59, 52, 62, 55, 106, 100, 102, 103))
})

test_that("estimatePSCS() organic profiles",{
  sp1@horizons$name[1:60] <- "Oi"
  sp1@horizons$name[48:51] <- "W"
  sp1@horizons$name[57:60] <- "Oif"
  horizons(sp1)$lieutex <- NA
  oo <- estimatePSCS(sp1, hzdesgn = "name", clay.attr = "prop", texcl.attr = "texture", lieutex = "texture")
  expect_equal(oo$pscs_top, rep(0, 9))
  expect_equal(oo$pscs_bottom, c(89, 59, 67, 62, 68, 160, 160, 68, 133))
})

# AGB (2020/05/27)
#'
#' Eliminate duplicate instances of profile IDs in a list of SoilProfileCollections
#'
#'  @description Experimental function to "clean" list input where duplicates exist (that would otherwise prevent \code{pbindlist}). Useful for queries that may have overlapping instances of the same data, for instance a list of SoilProfileCollections where each list element contains profiles gathered from a set of (potentially overlapping) extents.
#'
#' @param l A list of SoilProfileCollections.
#'
#' @return A list of SoilProfileCollections, with duplicate profile IDs removed.
#'
#' @author Andrew G. Brown
#'
#' @export lunique
#'
#' @examples
#'
#' data(sp5)
#'
#' # EXAMPLE #1 -- resolving overlap
#'
#' # 6 profiles in four sets, and 5,6,7 are missing
#' input <- lapply(list(c(1,3,4), c(2,2,3), NA, c(8,9,1)), function(idx) {
#'       if(!all(is.na(idx)))
#'        sp5[idx,]
#' })
#'
#' output <- lunique(input)
#'
#' # 6 profiles are in final SPC; 5,6,7 are missing
#' match(profile_id(pbindlist(output)), profile_id(sp5))
#'
#' # EXAMPLE #2 -- exact duplicates
#'
#' # deliberately duplicate an SPC
#' sp5_2 <- sp5
#' res <- lunique(list(sp5, sp5_2))
#'
#' # the number of profiles in first element is equal to number in sp5
#' length(res[[1]]) == length(sp5)
#'
#' # second list element contains NA b/c all uniques are in #1
#' res[[2]]
#'
lunique <- function(l) {

  # calculate profile IDs for each SPC element in l
  l.pid <- lapply(l, function(x) {
    if(!inherits(x, 'SoilProfileCollection')) {
      return(NA)
    }
    profile_id(x)
  })

  # keep track of all-NA and length (number of profiles per set)
  l.na <- as.logical(unlist(lapply(l.pid, function(x) all(is.na(x)))))
  l.n <- lapply(l.pid, length)

  # make data frame of profile ID and input list index
  df <- data.frame(pid = do.call('c', l.pid),
                   idx = do.call('c', lapply(1:length(l.n),
                                             function(n) rep(n, l.n[n]))))

  # squash non-uniques in combined data.frame, based on profile ID
  df <- df[order(df$pid),]
  d.out <- do.call('rbind', lapply(split(df, df$pid), function(d) {
    # note diagnostics on d can be helpful to ID type of non-uniqueness
    return(d[1,])
  }))

  # split back into "original" list elements -- sans empty ones
  d.split <- split(d.out, d.out$idx)
  l.out <- lapply(1:length(l.na), function(k) {
    if(!l.na[k]) {
      needle <- l.pid[[k]]
      haystack <- d.split[[as.character(k)]]
      ids <- which(needle %in% haystack$pid)
      if(length(ids) == 0)
        return(NA)
      return(l[[k]][ids, ])
    } else {
      return(NA)
    }
  })

  # put NAs back in that were lost during concatenation
  return(l.out)
}

context("unique concatenation of lists of SPCs (lunique)")

data(sp5)

# TODO: rebuild SPC datasets
metadata(sp5)$aqp_df_class <- "data.frame"

test_that("basic functionality", {

  # ensure that first 9 profiles, in order, are returned
  input <- lapply(list(c(1, 2, 3),
                c(2, 3, 4),
                c(5, 6, 7),
                c(8, 9, 1)),
           function(idx) {
            sp5[idx, ]
           })


  output <- lunique(input)

  expect_equal(do.call('c', lapply(output, function(x) {
    match(profile_id(x), profile_id(sp5))
  })), 1:9)
  expect_equal(match(profile_id(pbindlist(output)), profile_id(sp5)), 1:9)
})


test_that("NAs and error handling", {
  # check NA handling
  input <- lapply(list(c(1, NA, 3), c(2, 3, 4), NA, c(8, 9, 1)),
                  function(idx) {
                    if (all(is.na(idx)))
                      return(NA)
                    sp5[idx[which(!is.na(idx))], ]
                  })
  output <- lunique(input)

  # profiles 1 and 3 from first set
  expect_true(all(c("soil1", "soil100") %in% profile_id(output[[1]])))

  # profiles 2 and 4 from second set
  expect_true(all(c("soil10", "soil101") %in% profile_id(output[[2]])))

  # profiles 8 and 9 from fourth set
  expect_true(all(c("soil105", "soil106") %in% profile_id(output[[4]])))

  # total of four sets
  expect_equal(length(output), 4)

  # of which the third contains NA
  expect_equal(which(is.na(output)), 3)

  # we get a warning when combining, but it is successful
  expect_message(res <- length(profile_id(pbindlist(output))),
                 regexp = "pbindlist: one or more input list elements is NA")

  # the resulting SoilProfileCollection has 6 profiles
  expect_equal(res, 6)

})

# finds all horizons that are possibly part of a cambic horizon
# excluding those that are part of an argillic horizon
# (defined either by depth interval or getArgillicBounds)

#' Find all intervals that are potentially part of a Cambic horizon
#' @description Find all intervals that are potentially part of a Cambic horizon excluding those that are part of an argillic horizon (defined either by depth interval or \code{getArgillicBounds()}).
#'
#' There may be multiple cambic horizons (indexes) in a profile. Each cambic index has a top and bottom depth associated: cambic_top and cambic_bottom. This result is designed to be used for single profiles, or with \code{profileApply(..., frameify = TRUE)}
#'
#' @param p A single-profile SoilProfileCollection
#' @param hzdesgn Column name containing horizon designations.
#' @param texcl.attr Arguments to \code{getArgillicBounds()}
#' @param clay.attr Arguments to \code{getArgillicBounds()}
#' @param ... Arguments to \code{getArgillicBounds()}
#' @param argi_bounds Optional: numeric vector of length 2 with top and bottom of argillic; (Default: NULL)
#' @param d_value Column name containing dry value. Default: d_value
#' @param m_value Column name containing moist value. Default: m_value
#' @param m_chroma Column name containing moist chroma Default: m_chroma
#' @param sandy.texture.pattern this is a pattern for matching sandy textural classes: `-S$|^S$|COS$|L[^V]FS$|[^L]VFS$|LS$|LFS$`
#' 
#' @return A \code{data.frame} containing profile, cambic indexes, along with top and bottom depths.
#'
#' @author Andrew G. Brown
#'
#' @export
#'
#' @examples
#' # construct a fake profile
#' spc <- data.frame(id=1, taxsubgrp = "Lithic Haploxerepts",
#'                   hzname   = c("A","AB","Bw","BC","R"),
#'                   hzdept   = c(0,  20, 32, 42,  49),
#'                   hzdepb   = c(20, 32, 42, 49, 200),
#'                   clay     = c(19, 22, 22, 21,  NA),
#'                   texcl    = c("l","l","l", "l","br"),
#'                   d_value  = c(5,   5,  5,  6,  NA),
#'                   m_value  = c(2.5, 3,  3,  4,  NA),
#'                   m_chroma = c(2,   3,  4,  4,  NA))
#'
#' # promote to SoilProfileCollection
#' depths(spc) <- id ~ hzdept + hzdepb
#' 
#' # set required metadata
#' hzdesgnname(spc) <- 'hzname'
#' hztexclname(spc) <- 'texcl'
#' hzmetaname(spc, 'clay') <- 'clay'
#' 
#' # print results in table
#' getCambicBounds(spc)
#'
getCambicBounds <- function(p,
                            hzdesgn = hzdesgnname(p, required = TRUE),
                            texcl.attr = hztexclname(p, required = TRUE),
                            clay.attr = hzmetaname(p, "clay", required = TRUE),
                            argi_bounds = NULL,
                            d_value = "d_value",
                            m_value = "m_value",
                            m_chroma = "m_chroma", 
                            sandy.texture.pattern = "-S$|^S$|COS$|L[^V]FS$|[^L]VFS$|LS$|LFS$",
                            ...) {

  # sacrifice to CRAN gods in the name of NSE
  id <- NULL
  
  if (is.null(hzdesgn) || !hzdesgn %in% horizonNames(p)) {
    stop("Horizon designation column (", hzdesgn, ") does not exist.")
  }
  
  # construct data.frame result for no-cambic-found (NA)
  empty_frame <- data.frame(id = character(0),
                            cambic_id = numeric(0),
                            cambic_top = numeric(0),
                            cambic_bottom = numeric(0))

  empty_frame_names <- names(empty_frame)
  empty_frame_names[1] <- idname(p)
  names(empty_frame) <- empty_frame_names

  hzd <- horizonDepths(p)

  if (is.null(argi_bounds)) {
    argi_bounds <- getArgillicBounds(p, hzdesgn = hzdesgn, 
                                     clay.attr = clay.attr, texcl.attr = texcl.attr, 
                                     ..., 
                                     simplify = FALSE)
  } else if (is.vector(argi_bounds)) {
    if (length(p) > 1) {
      stop("`argi_bounds` should be a data.frame if `p` has more than one profile", call. = FALSE)
    }
    argi_bounds <- data.frame(id = profile_id(p), ubound = argi_bounds[1], lbound = argi_bounds[2])
    colnames(argi_bounds)[1] <- idname(p)
  } 

  cambictop <- minDepthOf(p, pattern = "B", hzdesgn = hzdesgn, no.contact.assigned = NA, simplify = FALSE)
  cambicbottom <- maxDepthOf(p, pattern = "B", hzdesgn = hzdesgn, top = FALSE, no.contact.assigned = NA, simplify = FALSE)

  cambic <- glom(p, cambictop[[hzd[1]]], cambicbottom[[hzd[2]]], truncate = TRUE, fill = TRUE)

  if (any(!is.na(argi_bounds$ubound), !is.na(argi_bounds$lbound))) {
    # if an argillic is present, remove with glom truncate+invert
    cab <- argi_bounds[complete.cases(argi_bounds),]
    non.argillic <- glom(cambic, 
                         argi_bounds$ubound, argi_bounds$lbound, 
                         truncate = TRUE, invert = TRUE, fill = TRUE)
  } else {
    non.argillic <- cambic
  }

  dark.colors <- hasDarkColors(non.argillic)
  non.argillic$w <- rep(1, nrow(non.argillic))

  textures <- non.argillic[[texcl.attr]]
  sandy.textures <- grepl(sandy.texture.pattern, textures, ignore.case = TRUE)

  nhz <- horizons(non.argillic)

  # remove horizons that are sandy or have dark colors
  if (any(sandy.textures | dark.colors, na.rm = TRUE)) {
    nhz <- nhz[-which(sandy.textures | dark.colors),]
  }

  final <- data.frame(id = NA_character_, cambic_top = NA_real_, cambic_bottom = NA_real_)[0,]

  # iterate through combinations of horizons, check topology and thickness
  # finds multiple occurrences of cambic horizons, excluding argillics
  nhzs <- split(nhz, nhz[[idname(p)]])
  for (k in seq_along(nhzs)){
    for (j in 1:nrow(nhzs[[k]])) {
      for (i in j:nrow(nhzs[[k]])) {
  
        ftop <- nhzs[[k]][j:i, hzd[1]]
        fbot <- nhzs[[k]][j:i, hzd[2]]

        if (any(suppressWarnings(hzDepthTests(ftop, fbot)))) {
          i <- i - 1
          break;
        }
      }
  
      if (is.numeric(fbot) & is.numeric(ftop)) {
        pcamb.thickness <- fbot - ftop
        if (length(pcamb.thickness) > 0 & sum(pcamb.thickness, na.rm = TRUE) >= 15) {
    
          final <- rbind(final, data.frame(id = as.character(nhzs[[k]][[idname(p)]][1]),
                                           cambic_top = min(ftop, na.rm = TRUE),
                                           cambic_bottom = max(fbot, na.rm = TRUE)))
          if (i == nrow(nhzs[[k]])) {
              k = k + 1
              break
          }
        }
      }
      if (k > length(nhzs)) { 
        break
      }
    }
  }

  .N <- NULL 
  iddf <- data.table::data.table(id = as.character(final[["id"]]))[, list(cambic_index = 1:.N), by = "id"]
  nadf <- data.frame(id = profile_id(p)[!profile_id(p) %in% final[["id"]]])
  if (nrow(nadf) > 0) {
    nadf$cambic_index <- rep(NA_real_, nrow(nadf))
    nadf$cambic_top <- rep(NA_real_, nrow(nadf))
    nadf$cambic_bottom <- rep(NA_real_, nrow(nadf))
  }
  iddf <- iddf[,-1]
  res <- cbind(final, iddf)
  
  # note: NSE used to specifiy ordering condition for data.table object
  # much faster than ordering on the data.frame representation used next
  res <- data.table::rbindlist(list(res, nadf), fill = TRUE)[order(id)]
  colnames(res)[1] <- idname(p)
  .as.data.frame.aqp(res, aqp_df_class(p))
}

#' @title Probabilistic Estimation of Soil Depth within Groups
#'
#' @description Estimate the most-likely depth to contact within a collection of soil profiles. Consider `getSoilDepthClass` followed by group-wise percentile estimation as a faster alternative.
#'
#' @param x a \code{SoilProfileCollection} object
#' @param groups the name of a site-level attribute that defines groups of profiles within a collection
#' @param crit.prob probability cutoff used to determine where the most likely depth to contact will be, e.g. 0.9 translates to 90% of profiles are shallower than this depth
#' @param name horizon-level attribute where horizon designation is stored, defaults to `hzdesgnname(x)`
#' @param p a REGEX pattern that matches non-soil genetic horizons
#' @param ... additional arguments to \code{slab}
#'
#' @details This function computes a probability-based estimate of soil depth by group. If no grouping variable exists, a dummy value can be used to compute a single estimate. The \code{crit.prob} argument sets the critical probability (e.g. 0.9) at which soil depth within a group of profiles is determined. For example, a \code{crit.prob} of 0.95 might result in an estimated soil depth (e.g. 120cm) where 95% of the profiles (by group) had depths that were less than or equal to 120cm.
#'
#' @return A `data.frame` is returned, with as many rows as there are unique group labels, as specified in `groups`.
#'
#' @author D.E. Beaudette
#'
#' @seealso [estimateSoilDepth()] [slab()]
#'
#' @export
#'
#' @examples
#'
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' # set horizon designation in SPC
#' hzdesgnname(sp1) <- 'name'
#'
#' aggregateSoilDepth(sp1, 'group', crit.prob = 0.9)
#'
aggregateSoilDepth <- function(x, groups, crit.prob = 0.9, name = hzdesgnname(x), p = 'Cr|R|Cd', ...) {

  # sanity checks:
  # * does the group label variable exist in @site?
  if(! groups %in% siteNames(x))
    stop('`groups` must specify a site-level attribute')

  # * does the horizon name variable exist in @horizons?
  if(! name %in% horizonNames(x))
    stop('`name` must specify a horizon-level attribute, containing the horizon designation')

  # mark soil vs. non-soil horizons
  x$soil.flag <- rep('soil', times = nrow(x))
  x$soil.flag[grep(p, horizons(x)[[name]])] <- 'not-soil'

  # convert to factor
  x$soil.flag <- factor(x$soil.flag, levels = c('soil', 'not-soil'))

  ## TODO: this is the slowest step
  # compute slice-wise probabilities that sum to contributing fraction
  fm <- as.formula(paste0(groups, ' ~ soil.flag'))
  a.ml.soil <- slab(x, fm, cpm = 2, ...)

  ## note: this is important when aggregating over several depth classes
  # compute adjusted soil flag probability: cf * Pr(soil)
  a.ml.soil$adj.soil.flag <- with(a.ml.soil, contributing_fraction * soil)

  # convert probability of contact into probability of soil
  crit.prob <- 1 - crit.prob

  # extract depth at specific probability of contact
  ss <- split(a.ml.soil, a.ml.soil[[groups]])
  depth.prob <- lapply(ss, function(i) {

    soil.top <- 0
    soil.bottom <- max(i$top[which(i$adj.soil.flag >= crit.prob)])

    res <- data.frame(
      .id = i[[groups]][1],
      soil.top,
      soil.bottom,
      stringsAsFactors = FALSE
    )

    names(res)[1] <- groups
    return(res)
  })

  # flatten list -> data.frame, reset rownames
  depth.prob <- do.call('rbind', depth.prob)
  row.names(depth.prob) <- NULL

  return(depth.prob)
}

context("aggregateSoilDepth")

data(sp1)
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# set horizon designation in SPC
hzdesgnname(sp1) <- 'name'


## tests

test_that("works as expected", {
  
  a <- aggregateSoilDepth(sp1, groups = 'group', crit.prob = 0.9)
  
  # structure
  expect_true(inherits(a, 'data.frame'))
  expect_true(ncol(a) == 3)
  expect_true(nrow(a) == length(unique(sp1$group)))
  
  # values
  expect_true(all(a$group == c(1, 2)))
  
  
  ## TODO: cross-reference with getSoilDepthClass
  ## these don't currently give the same results
  
  # site(sp1) <- getSoilDepthClass(sp1)
  # tapply(sp1$depth, sp1$group, quantile, probs = c(0.9))
  
})

## TODO: add verbose = FALSE argument cut down on chatter


#' @title Allocate soil properties within various classification systems.
#' 
#' @description Generic function to allocate soil properties to different classification schemes.
#'
#' @param ... arguments to specific allocation functions, see details and examples
#' 
#' @param to character specifying the classification scheme: FAO Salt Severity, FAO Black Soil (see details for the required \code{...})
#' 
#' @param droplevels logical indicating whether to drop unused levels in factors. This is useful when the results have a large number of unused classes, which can waste space in tables and figures.
#' 
#' 
#' @details
#' This function is intended to allocate a set of soil properties to an established soil classification scheme, such as Salt Severity or Black Soil. Allocation is semantically different from classification. While classification is the 'act' of developing a grouping scheme, allocation is the assignment or identification of measurements to a established class (Powell, 2008). 
#'  
#' ## Usage Details
#' 
#' Each classification scheme (\code{to} argument) uses a different set of arguments.
#' 
#' - `FAO Salt Severity`
#'   + **EC:** electrical conductivity column name, dS/m
#'   + **pH:** pH column name, saturated paste extract
#'   + **ESP:** exchangeable sodium percentage column name, percent
#'   
#' - `FAO Black Soils`
#'   + **object:** a `data.frame` or `SoilProfileCollection`
#'   + **pedonid:** pedon ID column name, required when \code{object} is a \code{data.frame}
#'   + **hztop:** horizon top depth column name, required when \code{object} is a \code{data.frame}
#'   + **hzbot:** horizon bottom depth column name, required when \code{object} is a \code{data.frame}
#'   + **OC**: organic carbon column name, percent
#'   + **m_chroma:** moist Munsell chroma column name
#'   + **m_value:** moist Munsell value column name
#'   + **d_value:** dry Munsell value column name
#'   + **CEC:** cation exchange capacity column name (NH4OAc at pH 7), units of cmol(+)/kg soil
#'   + **BS:** base saturation column name (NH4OAc at pH 7), percent
#'   + **tropical:** logical, data are associated with "tropical soils"
#'   
#' - `ST Diagnostic Features`
#'   + **object:** a `data.frame` or `SoilProfileCollection`
#'   + **pedonid:** pedon ID column name, required when \code{object} is a \code{data.frame}
#'   + **hzname:** horizon name column, required when \code{object} is a \code{data.frame}
#'   + **hztop:** horizon top depth column name, required when \code{object} is a \code{data.frame}
#'   + **hzbot:** horizon bottom depth column name, required when \code{object} is a \code{data.frame}
#'   + **texcl:** soil texture class (USDA) column name
#'   + **rupresblkcem:** rupture resistance column name
#'   + **m_value:** moist Munsell value column name
#'   + **m_chroma:** moist Munsell chroma column name
#'   + **d_value:** dry Munsell value column name
#'   + **BS:** base saturation column name (method ??), percent
#'   + **OC**: organic carbon column name, percent
#'   + **n_value:** ??
#'   + **featkind:** ??
#' 
#' @note The results returned by \code{allocate(to = "ST Diagnostic Features")} currently return a limited set of diagnostic features that are easily defined. Also, the logic implemented for some features does not include all the criteria defined in the Keys to Soil Taxonomy.
#' 
#' 
#'
#' @return A vector or \code{data.frame} object.
#' 
#' @author Stephen Roecker
#' 
#' @references 
#' Abrol, I., Yadav, J. & Massoud, F. 1988. [Salt-affected soils and their management](https://www.fao.org/4/x5871e/x5871e00.htm). No. Bulletin 39. Rome, FAO Soils.
#' 
#' FAO. 2006. [Guidelines for soil description](https://www.fao.org/4/a0541e/a0541e.pdf). Rome, Food and Agriculture Organization of the United Nations.
#' 
#' FAO. 2020. DEFINITION | What is a black soil? (online). (Cited 28 December 2020). http://www.fao.org/global-soil-partnership/intergovernmental-technical-panel-soils/gsoc17-implementation/internationalnetworkblacksoils/more-on-black-soils/definition-what-is-a-black-soil/es/
#'   
#'   Powell, B., 2008. Classifying soil and land, in: McKenzie, N.J., Grundy, M.J., Webster, R., Ringrose-Voase, A.J. (Eds.), Guidelines for Survey Soil and Land Resources, Australian Soil and Land Survey Handbook Series. CSIRO, Melbourne, p. 572.
#' 
#' Richards, L.A. 1954. \href{https://www.ars.usda.gov/ARSUserFiles/20360500/hb60_pdf/hb60complete.pdf}{Diagnosis and Improvement of Saline and Alkali Soils}. U. S. Government Printing Office. 166 pp.
#' 
#' Soil Survey Staff, 2014. Keys to Soil Taxonomy, 12th ed. USDA-Natural Resources Conservation Service, Washington, D.C.
#' 
#' 
#' @export
#'
#' @examples
#' 
#' # Salt Severity
#' test <- expand.grid(
#'   EC  = sort(sapply(c(0, 0.75, 2, 4, 8, 15, 30), function(x) x + c(0, -0.05, 0.05))),
#'   pH  = c(8.1, 8.2, 8.3, 8.4, 8.5, 8.6),
#'   ESP = sort(sapply(c(0, 15, 30, 50, 70, 100), function(x) x + c(0, 0.1, -0.1)))
#' )
#' test$ss      <- with(test, allocate(EC = EC, pH = pH, ESP = ESP, to = "FAO Salt Severity"))
#' table(test$ss)
#' 
#' # Black Soil Category 1 (BS1)
#' test <- expand.grid(
#'   dept = seq(0, 50, 10),
#'   OC   = sort(sapply(c(0, 0.6, 1.2, 20, 40), function(x) x + c(0, -0.05, 0.05))),
#'   chroma_moist  = 2:4,
#'   value_moist   = 2:4,
#'   value_dry     = 4:6,
#'   thickness     = 24:26,
#'   CEC           = 24:26,
#'   BS            = 49:51,
#'   tropical      = c(TRUE, FALSE)
#' )
#' test$pedon_id <- rep(1:21870, each = 6)
#' test$depb     <- test$dept + 10
#' 
#' bs1 <- allocate(test, pedonid = "pedon_id", hztop = "dept", hzbot = "depb", 
#'                 OC = "OC", m_chroma = "chroma_moist", m_value = "value_moist", 
#'                 d_value = "value_dry", CEC = "CEC", BS = "BS", 
#'                 to = "FAO Black Soil"
#' )
#' 
#' table(BS1 = bs1$BS1, BS2 = bs1$BS2)
#' 
#' 
#' # SoilProfileCollection interface
#' 
#' data(sp3)
#' depths(sp3) <- id ~ top + bottom
#' hzdesgnname(sp3) <- 'name'
#' 
#' # fake base saturation
#' horizons(sp3)$bs <- 75
#' 
#' plotSPC(sp3)
#' 
#' allocate(
#'   sp3, 
#'   to = 'FAO Black Soil', 
#'   OC = 'tc', 
#'   m_chroma = 'chroma', 
#'   m_value = 'value', 
#'   d_value = 'value',
#'   CEC = 'cec',
#'   BS = 'bs'
#' )
#' 
#' # make a copy and edit horizon values
#' x <- sp3
#' x$value <- 2
#' x$chroma <- 2
#' x$cec <- 26
#' x$tc <- 2
#' 
#' x$soil_color <- munsell2rgb(x$hue, x$value, x$chroma)
#' 
#' plotSPC(x)
#' 
#' allocate(
#'   x, 
#'   to = 'FAO Black Soil', 
#'   OC = 'tc', 
#'   m_chroma = 'chroma', 
#'   m_value = 'value', 
#'   d_value = 'value',
#'   CEC = 'cec',
#'   BS = 'bs'
#' )
#' 
#' 
#' # Soil Taxonomy Diagnostic Features
#' data(sp1)
#' sp1$texcl = gsub("gr|grv|cbv", "", sp1$texture)
#' df <- allocate(object = sp1, pedonid = "id", hzname = "name", 
#'                hzdept = "top", hzdepb = "bottom", texcl = "texcl", 
#'                to = "ST Diagnostic Features"
#' )
#' aggregate(featdept ~ id, data = df, summary)
#' 
allocate <- function(..., to = c("FAO Salt Severity", "FAO Black Soil", "ST Diagnostic Features"), droplevels = FALSE) {
  
  # sanity check
  to <- match.arg(to, several.ok = FALSE)
  
  # select the appropriate system
  a <- switch(
    to,
    "FAO Salt Severity" = {
      .rank_salts(..., system = to, droplevels = droplevels) 
    },
    "FAO Black Soil" = {
      .black_soil(...)  
    },
    "ST Diagnostic Features" = {
      # object = object
      # pedonid = "peiid"
      # hzname = "hzname"
      # hzdept = "hzdept"
      # hzdepb = "hzdepb"
      # texcl = "texcl"
      # hz_pat = ""
      # tex_pat = "br"
      # featkind = "argillic horizon"
      
      featkind <- c("lithic contact", "paralithic contact", "densic contact", "petrocalcic horizon", "calcic horizon", "secondary carbonates", "mollic epipedon") #, "reduced matrix")
      
      a <- lapply(featkind, function(x) {
        # a <- .guess_df(pedonid = "peiid", hzname = "hzname", hzdept = "hzdept", hzdepb = "hzdepb", texcl = "texcl", featkind = "lithic contact")
        a <- .guess_df(..., featkind = x)
      })
      a <- do.call("rbind", a)
    }
  )
  
  return(a)  
}
    

# To do add USDA and other salt classes
## TODO consider optional object = NULL
## TODO safe handling of NA
.rank_salts <- function(EC = NULL, pH = NULL, ESP = NULL, SAR = NULL, system = "FAO Salt Severity", droplevels = FALSE) {
  
  # EC = 1; pH = 3; ESP = 50
  l <- list(EC = EC, pH = pH, ESP = ESP, SAR = SAR)
  
  # tests ----
  # ESP vs SAR
  if (!is.null(SAR)) {
    warning("SAR will be converted to ESP via Richards (1958) conversion formula.")
  }
  if (!is.null(ESP) & !is.null(SAR)) {
    warning("Both ESP & SAR are present, SAR will only be used where ESP is missing.")
  }
  # minimum dataset
  if (any(sapply(l[c(1, 3)], is.null)) & any(sapply(l[c(1, 4)], is.null))) {
    warning("the minimum dataset of soil properites for allocating to the Salt Severity classes are: EC (aka Electrial Conductivity), and ESP (aka Exchangable Sodium Percentage) or SAR (aka Sodium Adsorption Ratio)")
  }
  # pH rule
  if (any(!complete.cases(EC, ESP)) | any(!complete.cases(EC, SAR))) {
    warning("pH is used in where both ESP and SAR are missing")
  }
  # length
  n <- sapply(l, length)
  if (! all(max(n) == n[1:3]) & ! all(max(n) == n[c(1:2, 4)])) {
    stop("all arguments must have the same length")
  }
  
  
  # levels ----
  fao_lev <- c(
    c("none", "slightly saline", "moderately saline", "strongly saline", "very strongly saline", "extremely saline"),
    c("none", "slightly sodic", "moderately sodic", "strongly sodic", "very strongly sodic")
    )
  
  
  ## TODO: why?
  sc <- rep("none", times = length(EC))
  
  ## TODO: consider separate saline / sodic classification
  # estimate ESP from SAR ----
  if (is.null(ESP)) ESP <- rep(NA_real_, times = length(EC))
  if (is.null(SAR)) SAR <- rep(NA_real_, times = length(EC))
  
  .esp <- function(SAR) {
    (100 * (-0.0126 + 0.01475 * SAR)) / 
      (1   + (-0.0126 + 0.01475 * SAR))
  }
  ESPx <- .esp(SAR)
  ESP <- ifelse(is.na(ESP) & !is.na(SAR), ESPx, ESP)
  
  
  # rank ----
  # saline soils
  sc <- ifelse(EC > -1 & (ESP <= 15 | (is.na(ESP) & pH <= 8.2)), # & EC > 4 & pH <= 8.5, 
               as.character(
                 cut(EC,
                     breaks = c(-1, 0.75, 2, 4, 8, 15, 1500), 
                     labels = fao_lev[1:6],
                     right  = FALSE
                     )),
               sc
               )
  # sodic soils
  # ESP
  sc <- ifelse(EC <= 4 & ESP > 15,  # | pH > 8.2
               as.character(
                 cut(ESP,
                     # breaks = c(0, 15, 30, 50, 70, 100),
                     breaks = c(-2, 30, 50, 70, 102),
                     # labels = fao_lev[7:11],
                     labels = fao_lev[8:11],
                     right  = FALSE
                     )),
               sc
               )
  # saline-sodic soils
  sc <- ifelse(EC > 4 & (ESP > 15 | (is.na(ESP) & pH > 8.2)), "saline-sodic", sc)
  
  
  # convert to factor
  sc <- factor(sc, levels = c(fao_lev[6:1], fao_lev[8:11], "saline-sodic"))

  
  # droplevels
  if (droplevels == TRUE) {
    sc <- droplevels(sc)
  }
  
  return(sc)
}



.codify <- function(x, system = "salt severity", droplevels = FALSE) {

  if (system == "salt severity") {

    .codify_salt_severity(x, droplevels = droplevels)
    }
  }


.codify_salt_severity <- function(x, droplevels = FALSE) {
  
  # set levels
  fao_lev <- c(
    c("none", "slightly saline", "moderately saline", "strongly saline", "very strongly saline", "extremely saline"),
    c("none", "slightly sodic", "moderately sodic", "strongly sodic", "very strongly sodic")
  )
  
  # test
  if (!is.integer(x)) stop("x is not an integer")
  if (!all(unique(x) %in% c(1:11, NA))) warning("some x values do not match the lookup table")

  sc <- factor(
    x, 
    levels = 1:11, 
    labels = c(fao_lev[6:1], fao_lev[8:11], "saline-sodic")
    )
  
  if (droplevels == TRUE) {
    sc <- droplevels(sc)
  }
  
  return(sc)
}
  

## TODO: useful error message, this isn't helpful:
# Error in .black_soil(...) : 
# column names in object must match the other character vector input arguments

## TODO: there is currently no way to document these arguments, critical because each has an expected unit of measure


.black_soil <- function(object, pedonid = "peiid", hztop = "hzdept", hzbot = "hzdepb", OC = NULL, m_chroma = "m_chroma", m_value = "m_value", d_value = "d_value", CEC = NULL, BS = NULL, tropical = FALSE) { # thickness = NULL, horizon = TRUE 
  
  # OC = 1; chroma_moist = 3; value_moist = 3; value_dry = 5; thickness = 25; CEC = 20; BS = 50
  # pedonid = "123"; hztop = 0; hzbot = 25; OC = 1.6; m_chroma = 3; m_value = 3; d_value = 5; CEC = 25; BS = 50; tropical = FALSE
  # object <- data.frame(pedonid, hztop, hzbot, OC, m_chroma, m_value, d_value, CEC, BS)
  # pedonid = "pedonid"; hztop = "hztop"; hzbot = "hzbot"; OC = "OC"; m_chroma = "m_chroma"; m_value = "m_value"; d_value = "d_value"; CEC = "CEC"; BS = "BS"
  # pedonid = "idp"; hztop = "top"; hzbot = "bot"; OC = "oc"; m_chroma = "w_chroma"; m_value = "w_value"; d_value = "d_value"; CEC = "cec"; BS = NULL
  
  # check object type
  # SoilProfileCollection objects have a number of required arguments defined internally
  if (inherits(object, 'SoilProfileCollection')) {
    # extract relevant metadata from the SPC
    pID <- idname(object)
    hztb <- horizonDepths(object)
    
    # horizons as data.frame-like obj
    df <- horizons(object)
    
    # setup variables used later, using SPC metadata if possible
    vars <- list(pedonid = pID, hztop = hztb[1], hzbot = hztb[2], OC = OC, m_chroma = m_chroma, m_value = m_value, d_value = d_value, CEC = CEC, BS = BS)
    
  } else {
    # this is likely a data.frame
    df <- object
    
    # setup variables used later, all from provided arguments
    vars <- list(pedonid = pedonid, hztop = hztop, hzbot = hzbot, OC = OC, m_chroma = m_chroma, m_value = m_value, d_value = d_value, CEC = CEC, BS = BS)
  }
  
  # check length of arguments
  if (any(sapply(vars, length) > 1)) {
    stop("the length of all arguments must be 1, except for object")
    # this will drop NULL arguments, which is ok for CEC & BS
  } else vars <- unlist(vars)
  
  # check arguments match df colnames & subset
  # no vars should be NA, but this will catch them if they are
  idx <- !is.na(vars) 
  if (! all(vars[idx] %in% names(df))) {
    stop("column names in object must match the other character vector input arguments")
  } else {
    df <- df[vars[idx]]
    vars2 <- names(vars[idx])
    names(df) <- vars2
    }
  
  # criteria
  # 2nd category of Black Soils
  # minimum dataset
  if (any(sapply(df[vars2][1:7], function(x) all(is.na(x))))) {
    stop("the minimum dataset of soil properites for allocating to the 2nd category of Black Soils are: OC (aka Organic Carbon), m_chroma, m_value, and d_value") # and thickness
  }
  bs2 <- with(df,
              (OC <= 20 & (OC >= 1.2 | (tropical == TRUE & OC >= 0.6))) 
              & m_chroma <= 3 
              & (m_value <= 3 & d_value <= 5)
  )
  
  # 1st category of Black Soils
  # minimum dataset
  if (!is.null(CEC) & !is.null(BS) & all(c("CEC", "BS") %in% names(df))) {
    
    bs1 <- bs2 & df$CEC >= 25 & df$BS >= 50
    
    } else {
    message("the minimum dataset of soil properites for allocating to the 1nd category of Black Soils, in addition to the 2nd category, are: CEC (aka Cation Exchange Capacity), BS (aka Base Saturation)")
    bs1 <- NA[1:nrow(df)]
  }
  
  # combine results and subset to 0-25cm
  df_bs  <- cbind(df[vars2[1:3]], BS1 = bs1, BS2 = bs2)
  df_bs  <- hz_segment(df_bs, intervals = c(0, 25), depthcols = c("hztop", "hzbot"))
  df_bs  <- df_bs[df_bs$segment_id == "00-25", -6]
  
  # aggregate the horizons
  df_bs2 <- aggregate(cbind(BS1, BS2) ~ pedonid, data = df_bs, FUN = all, na.action = na.pass)
  df_bot <- aggregate(hzbot ~ pedonid, data = df_bs, FUN = function(x) max(x, na.rm = TRUE))
  
  # filter thickness > 25cm
  df_bs  <- merge(df_bs2, df_bot, by = "pedonid", all.x = TRUE)
  df_bs  <- within(df_bs, {
    BS1 = BS1 & as.integer(hzbot) == 25L
    BS2 = BS2 & as.integer(hzbot) == 25L
    hzbot = NULL
  })
  
  names(df_bs)[1] <- pedonid
  
  return(df_bs)
}



# guess diagnostic features
.guess_df <- function(object = NULL, pedonid = "peiid", hzname = "hzname", hzdept = "hzdept", hzdepb = "hzdepb", texcl = "texcl", rupresblkcem = "rupresblkcem", m_value = "m_value", d_value = "d_value", m_chroma = "m_chroma", BS = "BS", OC = "OC", n_value = "n_value", featkind = NULL) {
  
  # pedonid = "peiid"; hzname = "hzname"; hzdept = "hzdept"; hzdepb = "hzdepb"; texcl = "texcl"; hz_pat = ""; tex_pat = "br"; featkind = "mollic epipedon"; rupresblkcem = "rupresblkcem"; m_value = "m_value"; d_value = "d_value"; m_chroma = "m_chroma"; BS = NA; OC = NA; n_value = "n_value"
  # object = sp1; pedonid = "id"; hzname = "name"; hzdept = "top"; hzdepb = "bot"; texcl = "texture"
  # vars <- list(pedonid = "peiid", hzname = "hzname", hzdept = "hzdept", hzdepb = "hzdepb", texcl = "texture", rupresblkcem = "rupresblkcem", m_value = "m_value", d_value = "d_value", m_chroma = "m_chroma", OC = "OC", BS = "BS", n_value = "n_value")
  
  # standardize inputs
  vars <- list(pedonid = pedonid, hzname = hzname, hzdept = hzdept, hzdepb = hzdepb, texcl = texcl, rupresblkcem = rupresblkcem, m_value = m_value, d_value = d_value, m_chroma = m_chroma, OC = OC, BS = BS, n_value = n_value)
  
  # standardize inputs
  if (class(object)[1] == "SoilProfileCollection") {
    df <- horizons(object)
  } else df <- object
  
  # check length of arguments
  if (any(sapply(vars, length) > 1)) {
    stop("the length of all arguments must be 1, except for object")
    # this will drop NULL arguments, which is ok for CEC & BS
  } else vars <- unlist(vars)
  
  # check arguments match df colnames & subset
  # no vars should be NA, but this will catch them if they are
  if (! all(vars %in% names(df))) {
    warning("the minimum dataset includes: pedonid, hzdept, hzdepb, and hzname; if texcl or rupreblkcem are missing the resulting diagnostic features are inferred from the available information")
    idx <- vars %in% names(df)
    mis <- vars[! idx]
    df <- df[vars[idx]]
    df[names(mis)] <- NA
    vars2 <- names(vars)
    names(df) <- vars2
  } else {
    df <- df[vars]
    vars2 <- names(vars)
    names(df) <- vars2
  }
  df$texcl      <- tolower(df$texcl)
  df$rupresblkcem <- tolower(df$rupresblkcem)
  
  # match pattern
  
  # lithic contact ----
  if (featkind == "lithic contact") {
    message(paste("guessing", featkind))
    idx_hzn <-  grepl("R|Dr", df$hzname) & !grepl("\\/", df$hzname)
    idx_tex <- !grepl("Cr|CR", df$hzname) & (df$texcl %in% c("br", "wb", "uwb") | is.na(df$texcl))
    
    lev <- c("strongly cemented", "very strongly cemented", "indurated", "strongly", "extremely strongly", "H", "moderately coherent", "strongly coherent", "very strongly coherent") 
    idx_cem <- df$rupresblkcem %in% lev | is.na(df$rupresblkcem)
    
    # error
    idx_err <- idx_hzn  & (!idx_tex | !idx_cem) 
    if (any(idx_err)) {
      message(paste("the following pedonid have R horizons that do not meeting the texcl or rupture resistance cementation criteria and will be excluded: ", paste0(df$pedonid[idx_err], collapse = ", ")))
    }
    
    df$featkind <- ifelse(idx_hzn & idx_tex & idx_cem, featkind, NA)
  }
  
  
  # paralithic contact ----
  if (featkind == "paralithic contact") {
    message(paste("guessing", featkind))
    
    idx_hzn <-  grepl("Cr|CR", df$hzname)
    idx_tex <- !grepl("R|Dr", df$hzname) & (df$texcl %in% c("br", "wb", "uwb") | is.na(df$texcl))
    
    lev <- c("extremely weakly cememented", "very weakly cemented", "weakly cemented", "moderately cemented", "weakly", "moderately", "S") 
    idx_cem <- df$rupresblkcem %in% lev | is.na(df$rupresblkcem)
    
    # error
    idx_err <- idx_hzn  & (!idx_tex | !idx_cem) 
    if (any(idx_err)) {
      message(paste("the following pedonid have Cr horizons that do not meeting the texcl or rupture resistance cementation criteria and will be excluded: ", paste0(df$pedonid[idx_err], collapse = ", ")))
    }

    df$featkind <- ifelse(idx_hzn & idx_tex & idx_cem, featkind, NA)
  }
  
  
  # densic contact ----
  if (featkind == "densic contact") {
    message(paste("guessing", featkind))
    idx <- grepl("d$|D$|d[1:9]|D[1:9]", df$hzname)
    df$featkind <- ifelse(idx == TRUE, featkind, NA)
  }
  
  
  # petrocalcic horizon ----
  if (featkind == "petrocalcic horizon") {
    message(paste("guessing", featkind))
    
    idx_hzn  <- grepl("kkm|kkqm", df$hzname)
    idx_tex  <- ((grepl("cem", df$texcl) & !grepl("-br", df$texcl)) | is.na(df$texcl))
    
    lev <- "noncemented" 
    idx_cem <- !df$rupresblkcem %in% lev | is.na(df$rupresblkcem)
    
    # error
    idx_err <- idx_hzn  & (!idx_tex | !idx_cem) 
    if (any(idx_err)) {
      message(paste("the following pedonid have Bkkm|Bkkqm horizons that do not meeting the texcl or rupture resistance cementation criteria and will be excluded: ", paste0(df$pedonid[idx_err], collapse = ", ")))
    }
    
    df$featkind <- ifelse(idx_hzn & idx_tex & idx_cem, featkind, NA)
  }
  
  
  # calcic horizon ----
  if (featkind == "calcic horizon") {
    message(paste("guessing", featkind))
    
    idx_hzn <-  grepl("kk$|kk[1:9]|kkq$|kkq[1:9]|kkb$|kkb[1:9]", df$hzname)
    idx_tex <- (!grepl("cem-", df$texcl) | is.na(df$texcl))
    
    lev <- "noncemented" 
    idx_cem <- df$rupresblkcem %in% lev | is.na(df$rupresblkcem)
    
    # error
    idx_err <- idx_hzn  & (!idx_tex | !idx_cem) 
    if (any(idx_err)) {
      message(paste("the following pedonid have Bkk horizons that do not meeting the texcl or rupture resistance cementation criteria and will be excluded: ", paste0(df$pedonid[idx_err], collapse = ", ")))
    }
    
    df$featkind <- ifelse(idx_hzn & idx_tex & idx_cem, featkind, NA)
  }
  
  
  # secondary carbonates ----
  if (featkind == "secondary carbonates") {
    message(paste("guessing", featkind))
    
    idx_hzn <- grepl("k", df$hzname) & !grepl("kk", df$hzname)
    
    df$featkind <- ifelse(idx_hzn, featkind, NA)
  }
  
  
  # mollic epipedon ----
   if (featkind == "mollic epipedon") {
     message(paste("guessing", featkind))
     
     idx_hzn <- !grepl("O|Ao|R|W|M|C|\\/", df$hzname)
     idx_tex <- df$texcl %in% levels(SoilTextureLevels()) | is.na(df$texcl)
     # need to add structure to fetchNASIS
     idx_col <- 
       (df$m_value  <= 3 | is.na(df$m_value))  & 
       (df$d_value  <= 5 | is.na(df$d_value))  &
        df$m_chroma <= 3                        &
       (!is.na(df$m_value) | !is.na(df$d_value))
     idx_bs  <- (df$BS >= 50 | is.na(df$BS)) # & (df$ph1to1 > 6 | is.na(df$ph1to1))
     idx_oc  <- 
       ((df$OC >= 2.5 | is.na(df$OC)) & (df$m_value %in% 4:5 | is.na(df$m_value))) |
       ((df$OC >= 0.6 | is.na(df$OC)) & (df$m_value < 4      | is.na(df$m_value)))
     idx_nv <- (df$n_value < 0.7 | is.na(df$n_value))
     
     df$featkind <- ifelse(idx_hzn & idx_tex & idx_col & idx_bs & idx_oc & idx_nv, featkind, NA)
   }
  
  
  # subset features
  idx <- "featkind" %in% names(df)
  if (idx) {
    
    df_sub <- df[!is.na(df$featkind), ]
    
    # aggregate depths ----
    idx <- !is.na(df_sub$featkind)
    if (any(idx) & sum(idx, na.rm = TRUE) > 1) {
      sp <- aggregate(hzdept ~ pedonid + featkind, data = df_sub, FUN = function(x) min(x, na.rm = TRUE))
      sp$featdepb <- aggregate(hzdepb ~ pedonid + featkind, data = df_sub, FUN = function(x) max(x, na.rm = TRUE))$hzdepb
      names(sp)[3] <- "featdept"
    } else {
      sp <- df_sub[c("pedonid", "featkind", "hzdept", "hzdepb")]
      names(sp)[3:4] <- c("featdept", "featdepb")
    }
  
    names(sp)[1] <- vars[1]
    
    
    if (featkind == "petrocalcic horizon") {
      sp <- sp[(sp$featdepb - sp$featdept) >= 10, ]
    }
    
    # need to add more logic to capture when < 10cm is cemented 
    if (featkind == "calcic horizon") {
      sp <- sp[(sp$featdepb - sp$featdept) >= 15, ]
    }
    
    
    # needs additional criteria to get a variable depth or use Andrew's function
    if (featkind == "mollic epipedon") {
      sp <- sp[(sp$featdepb - sp$featdept) >= 18, ]
    }
  
  } else sp <- NULL 
  
  return(sp)
}


#' @title Allocate Particle Size Class for the Control Section.
#' 
#' @description This function aggregates information in the horizon table and allocates it to the particle size class for the control section.
#'
#' @param x a \code{data.frame} containing the original horizon table.
#' @param y a \code{data.frame} containing the particle size control section depths for each idcol.
#' @param taxpartsize \code{character} column name for taxonomic family particle size class.
#' @param clay \code{character} column name for clay percent.
# #' @param frags \code{character} column name for total rock fragments percent.
#' @param idcol character: column name of the pedon ID within the object.
#' @param depthcols a character vector of length 2 specifying the names of the horizon depths (e.g. `c("top", "bottom")`).
#' 
#' 
#' @details
#' This function differs from \code{\link{texture_to_taxpartsize}} in that is aggregates the results of \code{\link{texture_to_taxpartsize}}, and accounts for strongly contrasting particle size classes.
#'  
#'
#' @return A \code{data.frame} object containing the original idcol, the aggregated particle size control section allocation, and an aniso column to indicate more than one contrasting class.
#' 
#' @author Stephen Roecker
#' 
#' @seealso [texture_to_taxpartsize()], [lookup_taxpartsize()]
#' 
#' @export
#' @examples
#' 
#' h <- data.frame(
#'   id = 1,
#'   hzname = c("A", "BA", "Bw", "BC", "C"),
#'   top    = c(0, 10, 45, 60,  90),
#'   bottom = c(10, 45, 60, 90, 150),
#'   clay   = c(15, 16, 45, 20,  10),
#'   sand   = c(10, 35, 40, 50,  90),
#'   frags  = c(0,  5, 10, 38,  40)
#' )
#' 
#' h <- cbind(h,
#'            texcl = ssc_to_texcl(clay = h$clay, sand = h$sand))
#' 
#' pscs <- data.frame(id = 1,
#'                    top = 25,
#'                    bottom = 100)
#' 
#' h <- cbind(h,
#'            taxpartsize = texture_to_taxpartsize(
#'              texcl = h$texcl,
#'              clay  = h$clay,
#'              sand  = h$sand,
#'              fragvoltot = h$frags
#'            ))
#' 
#' depths(h) <- id ~ top + bottom
#' 
#' # set required metadata for estimatePSCS()
#' hzdesgnname(h) <- "hzname"
#' hztexclname(h) <- "texcl"
#' hzmetaname(h, "clay") <- "clay"
#' 
#' pscs <- data.frame(id = h$id, rbind(estimatePSCS(h)))
#' names(pscs)[2:3] <- c("top", "bottom")
#' 
#' hz_to_taxpartsize(horizons(h), pscs)
#'  
#' 
hz_to_taxpartsize <- function(x, y, taxpartsize = "taxpartsize", clay = "clay", idcol = "id", depthcols = c("top", "bottom")) {
  # need to incorporate fine sand for special cases of strongly contrasting classes and rock fragments (?)
  # frags = "frags", 
  
  # strongly contrasting
  
  x$rn <- 1:nrow(x)
  # xy <- hz_intersect(x, y, idcol = idcol, depthcols = depthcols)
  # x_sub <- x[x$rn %in% xy$rn, ]
  
  # check segment_id ----
  ## if it exists, overwrite it
  x_nm <- names(x)
  y_nm <- names(y)
  if (any(x_nm == "segment_id") | any(y_nm == "segment_id")) {
    x[x_nm == "segment_id"] <- NULL
    y[y_nm == "segment_id"] <- NULL
  }
  
  # check dissolve_id ----
  ## if it exists, overwrite it
  x_nm <- names(x)
  y_nm <- names(y)
  if (any(x_nm == "dissolve_id") | any(y_nm == "dissolve_id")) {
    x[x_nm == "dissolve_id"] <- NULL
    y[y_nm == "dissolve_id"] <- NULL
  }
  
  # standardize inputs ----
  vars <- c(idcol, depthcols, clay, taxpartsize)
  x <- x[vars]
  x_std <- .standardize_inputs(x, idcol = idcol, depthcols = depthcols, clay = clay, taxpartsize = taxpartsize)
  x <- x_std$x; x_conv <- x_std$x_conversion
  x_std <- NULL
  
  y <- y[c(idcol, depthcols)]
  y <- .standardize_inputs(y, idcol = idcol, depthcols = depthcols)$x
  
  # dissolve on pscs ----
  # calculate non-trimmed horizon thickness
  x_dis <- transform(hz_dissolve(x, 
                                 by = "taxpartsize", 
                                 idcol = "idcol", 
                                 depthcols = c("top", "bot")), 
                     thk_o = bot - top)
  
  # trim depths ----
  # calculate trimmed horizon thickness
  xy_dis <- transform(hz_intersect(x_dis, y, 
                                   idcol = "idcol", 
                                   depthcols = c("top", "bot")),
                      thk_t = bot - top)
  
  # rejoin dissolved pscs to the original horizon table ----
  xy <- suppressWarnings(hz_intersect(x, xy_dis, 
                                      idcol = "idcol", 
                                      depthcols = c("top", "bot")))
  x_dis  <- NULL
  xy_dis <- NULL

  # aggregate clay values within dissolved pscs ----
  top       <- NULL
  bot       <- NULL
  thk_o     <- NULL
  thk_t     <- NULL
  clay_wt   <- NULL
  # sandvf_wt <- NULL
  
  xy_agg <- data.table::as.data.table(xy)[,
    list(
      top       = min(top,                          na.rm = TRUE),
      bot       = max(bot,                          na.rm = TRUE),
      clay_wt   = weighted.mean(clay,   w = thk_t,  na.rm = TRUE), 
      # sandvf_wt = weighted.mean(sandvf, w = thk_t,  na.rm = TRUE),
      # need to impute frags
      # frag_wt = weighted.mean(total_frags_pct_nopf, w = thk_t), na.rm = TRUE,
      thk_o   = sum(thk_o, na.rm = TRUE),
      thk_t   = sum(thk_t, na.rm = TRUE)
    ), by = c("idcol", "taxpartsize", "dissolve_id")
    ]
  data.table::setorder(xy_agg, idcol, top)
  xy_agg <- as.data.frame(xy_agg)
  
  # find adjacent horizons ----
  xy_lag <- hz_lag(xy_agg, idcol = "idcol", depthcols = c("top", "bot"))
  
  # address special cases of strongly contrasting classes ----
  clay_wt_bot.1     <- NULL
  sandvf_wt_bot.1   <- NULL
  taxpartsize_bot.1 <- NULL
  
  # still needs special cases for very fine sand
  xy_agg <- within(
    cbind(xy_agg, xy_lag), 
    {
      clay_dif = clay_wt_bot.1 - clay_wt
      sc       = paste0(taxpartsize, " over ", taxpartsize_bot.1)
      sc       = gsub(" over NA$", "", sc)
      
      sc = gsub("^fine over|^very-fine over", "clayey over", sc)
      sc = gsub("over fine$|over very-fine$", "over clayey", sc)
      sc = gsub("over fine over|over very-fine over", "over clayey over", sc)
      sc = gsub("over sandy|over sandy-skeletal", "over sandy or sandy-skeletal", sc)
      # clay over loamy
      sc = ifelse( 
        abs(clay_dif) >= 25 & sc %in% c("clayey over fine-loamy", "clayey over coarse-loamy"),
        gsub("clayey over fine-loamy|clayey over coarse-loamy", "clayey over loamy", sc),
        sc
      )
      # clay over loamy-skeletal
      sc = ifelse( 
        sc == "clayey over loamy-skeletal" & abs(clay_dif) < 25, 
        taxpartsize,
        sc
      )
      # fine-silty over clayey
      sc = ifelse( 
        sc == "fine-silty over clayey" & abs(clay_dif) < 25, 
        taxpartsize,
        sc
      )
      # loamy material contains less than 50 percent, by weight
      # need to include a vfs percent in the function arguments, which is only present in the lab data, and otherwise you typically wouldn't assume the vfs percent is high enough to qualify for these special cases
      sc = ifelse( 
        sc %in% c("coarse-loamy over sandy or sandy-skeletal", "loamy over sandy or sandy-skeletal", "loamy-skeletal over sandy or sandy-skeletal", "sandy over loamy", "sandy-skeletal over loamy"),
        taxpartsize,
        sc
      )
      idx_sc = sc %in% .pscs_sc
      # # sandy over loamy
      # sc = ifelse(
      #   sc %in% c("sandy over coarse-loamy", "sandy over fine-loamy") & taxpartsize_bot.1 %in% c("coarse-loamy", "fine-loamy") & sandvf_wt_bot.1 > 50,
      #   "sandy over loamy",
      #   sc
      #   )
      # # sandy-skeletal over loamy
      # sc = ifelse(
      #   sc %in% c("sandy-skeletal over coarse-loamy", "sandy over fine-loamy") & taxpartsize_bot.1 %in% c("coarse-loamy", "fine-loamy") & sandvf_wt_bot.1 > 50,
      #   "sandy-skeletal over loamy",
      #   sc
      # )
      # idx_sc = grepl("over", sc)
      sc = ifelse(idx_sc, sc, taxpartsize)
    }
  )
  xy_lag <- NULL
  
  # find multiple strongly contrasting ps classes within the control section
  n_sc    <- NULL
  n_peiid <- NULL
  
  test <- as.data.frame(data.table::as.data.table(xy_agg)[, 
                        list(n_sc    = sum(idx_sc, na.rm = TRUE),
                             # sum(grepl(" over ", sc), na.rm = TRUE),
                             n_peiid = length(idx_sc)),
                        by = "idcol"])
  
  # pick the sc pscs with the largest contrast or pscs with the greatest thickness
  xy_res <- transform(
      merge(
        xy_agg,
        test,
        by = "idcol",
        all.x = TRUE,
        sort = FALSE
      ),
      idx_sc = sc %in% .pscs_sc,
      # idx_sc = grepl(" over ", sc),
      idx_c_ov_l = sc %in% c("clayey over fine-loamy")
    )
  
  xy_res <- within(as.data.frame(
    data.table::as.data.table(xy_res)[ , 
      list(
        pscs1 = sc[n_sc == 0 & n_peiid == 1],
        pscs2 = sc[n_sc == 1 & n_peiid  > 1 & idx_sc],
        pscs3 = sc[which.max(thk_t[n_sc == 0 & n_peiid > 1])],
        pscs4 = sc[n_sc == 1 & idx_sc],
        pscs5 = sc[which.max(abs(clay_dif[n_sc > 1 & !is.na(sc)]))],
        taxpartsizemod = ifelse(max(n_sc) > 1, "aniso", "not used")
      ), 
      by =  "idcol"]),
      {
        # need to add fix for special case of sandy over loamy which requires fine sand percent
        taxpartsize = paste(pscs1, pscs3, pscs4, pscs5, sep = "")
        taxpartsize = gsub("NA", "", taxpartsize)
        pscs1 = NULL
        pscs2 = NULL
        pscs3 = NULL
        pscs4 = NULL
        pscs5 = NULL
      })
  
  # reset inputs
  xy_res <- .reset_inputs(xy_res, x_conv[1])
  
  return(xy_res)
}

## TODO: large gaps in data result in some strange-looking polygons

## TODO: 'id' isn't a very safe argument for defining profile IDs... (e.g. sp1)

## TODO: iterate over profile IDs instead of groups
## TODO: confidence bands not defined when NA is present


#' @title Lattice Panel Function for Soil Profiles
#'
#' @description Panel function for plotting grouped soil property data, along with upper and
#' lower estimates of uncertainty.
#'
#' This function can be used to replace \code{panel.superpose} when plotting
#' depth function data. When requested, contributing fraction data are printed
#' using colors the same color as corresponding depth function lines unless a
#' single color value is given via \code{cf.col}.
#' 
#' This function is not able to apply transformations (typically `log = 10`) applied in the `scales` argument to `xyplot` to  upper/lower bounds. These will have to be manually applied. See examples.
#'
#' @aliases panel.depth_function prepanel.depth_function
#' 
#' @param x x values (generated by calling lattice function)
#' 
#' @param y y values (generated by calling lattice function)
#' 
#' @param id vector of id labels, same length as x and y--only required when
#' plotting segments (see Details section)
#' 
#' @param upper vector of upper confidence envelope values
#' 
#' @param lower vector of lower confidence envelope values
#' 
#' @param subscripts paneling indices (generated by calling lattice function)
#' 
#' @param groups grouping data (generated by calling lattice function)
#' 
#' @param sync.colors optionally sync the fill color within the region bounded
#' by (lower--upper) with the line colors
#' 
#' @param cf optionally annotate contributing fraction data at regular depth
#' intervals see \code{\link{slab}}
#' 
#' @param cf.col optional color for contributing fraction values, typically
#' used to override the line color
#' 
#' @param cf.interval number of depth units to space printed contributing
#' fraction values
#' 
#' @param \dots further arguments to lower-level lattice plotting functions,
#' see below
#' 
#' @author D.E. Beaudette
#' 
#' @seealso \code{\link{sp1}}, \code{\link{slice}}, \code{\link{slab}}
#' @export
#' @keywords hplot
#' @examples
#'
#' library(lattice)
#' data(sp1)
#' 
#' # 1. plotting mechanism for step-functions derived from soil profile data
#' xyplot(
#'   cbind(top, bottom) ~ prop,
#'   data = sp1,
#'   id = sp1$id,
#'   panel = panel.depth_function,
#'   ylim = c(250, -10),
#'   scales = list(y = list(tick.number = 10)),
#'   xlab = 'Property',
#'   ylab = 'Depth (cm)',
#'   main = 'panel.depth_function() demo'
#' )
#' 
#' # 1.1 include groups argument to leverage lattice styling framework
#' sp1$group <- factor(sp1$group, labels = c('Group 1', 'Group2'))
#' 
#' xyplot(
#'   cbind(top, bottom) ~ prop,
#'   groups = group,
#'   data = sp1,
#'   id = sp1$id,
#'   panel = panel.depth_function,
#'   ylim = c(250, -10),
#'   scales = list(y = list(tick.number = 10)),
#'   xlab = 'Property',
#'   ylab = 'Depth (cm)',
#'   main = 'panel.depth_function() demo',
#'   auto.key = list(
#'     columns = 2,
#'     points = FALSE,
#'     lines = TRUE
#'   ),
#'   par.settings = list(superpose.line = list(col = c(
#'     'Orange', 'RoyalBlue'
#'   )))
#' )
#' 
#' 
#' # more complex examples, using step functions with grouped data
#' # better looking figures with less customization via tactile package
#' if(requireNamespace('tactile')) {
#'   
#'   library(data.table)
#'   library(lattice)
#'   library(tactile)
#'   
#'   # example data
#'   data(sp6)
#'   
#'   # a single profile
#'   x <- sp6[1:5, ]
#'   
#'   # wide -> long format
#'   x.long <- data.table::melt(
#'     data.table::data.table(x), 
#'     id.vars = c('id', 'top', 'bottom'), 
#'     measure.vars = c('sand', 'silt', 'clay')
#'   )
#'   
#'   # (optional) convert back to data.frame
#'   x.long <- as.data.frame(x.long)
#'   
#'   # three variables sharing a common axis
#'   # factor levels set by melt()
#'   xyplot(
#'     cbind(top, bottom) ~ value | id,
#'     groups = variable,
#'     data = x.long,
#'     id = x.long$id,
#'     ylim = c(200, -5), xlim = c(10, 60),
#'     scales = list(alternating = 1, y = list(tick.number = 10)),
#'     par.settings = tactile.theme(superpose.line = list(lwd = 2)),
#'     xlab = 'Sand, Silt, Clay (%)',
#'     ylab = 'Depth (cm)',
#'     panel = panel.depth_function,
#'     auto.key = list(columns = 3, lines = TRUE, points = FALSE),
#'     asp = 1.5
#'   )
#'   
#'   
#'   # all profiles
#'   x <- sp6
#'   
#'   # wide -> long format
#'   x.long <- data.table::melt(
#'     data.table::data.table(x), 
#'     id.vars = c('id', 'top', 'bottom'), 
#'     measure.vars = c('sand', 'silt', 'clay')
#'   )
#'   
#'   # (optional) convert back to data.frame
#'   x.long <- as.data.frame(x.long)
#'   
#'   # three variables sharing a common axis
#'   # factor levels set by melt()
#'   xyplot(
#'     cbind(top, bottom) ~ value | id,
#'     groups = variable,
#'     data = x.long,
#'     id = x.long$id,
#'     ylim = c(200, -5), xlim = c(0, 70),
#'     scales = list(alternating = 1, y = list(tick.number = 10)),
#'     par.settings = tactile.theme(superpose.line = list(lwd = 2)),
#'     xlab = 'Sand, Silt, Clay (%)',
#'     ylab = 'Depth (cm)',
#'     panel = panel.depth_function,
#'     auto.key = list(columns = 3, lines = TRUE, points = FALSE),
#'     as.table = TRUE
#'   )
#'   
#'   xyplot(
#'     cbind(top, bottom) ~ value,
#'     groups = variable,
#'     data = x.long,
#'     id = x.long$id,
#'     ylim = c(200, -5), xlim = c(0, 70),
#'     scales = list(alternating = 1, y = list(tick.number = 10)),
#'     par.settings = tactile.theme(superpose.line = list(lwd = 2)),
#'     xlab = 'Sand, Silt, Clay (%)',
#'     ylab = 'Depth (cm)',
#'     panel = panel.depth_function,
#'     auto.key = list(columns = 3, lines = TRUE, points = FALSE),
#'     as.table = TRUE
#'   )
#'   
#'   xyplot(
#'     cbind(top, bottom) ~ value | variable,
#'     groups = variable,
#'     data = x.long,
#'     id = x.long$id,
#'     ylim = c(200, -5), xlim = c(0, 70),
#'     scales = list(alternating = 1, y = list(tick.number = 10)),
#'     par.settings = tactile.theme(superpose.line = list(lwd = 2)),
#'     xlab = 'Sand, Silt, Clay (%)',
#'     ylab = 'Depth (cm)',
#'     panel = panel.depth_function,
#'     auto.key = list(columns = 3, lines = TRUE, points = FALSE),
#'     as.table = TRUE
#'   )
#'   
#'   
#'   xyplot(
#'     cbind(top, bottom) ~ value | variable,
#'     data = x.long,
#'     id = x.long$id,
#'     ylim = c(200, -5), xlim = c(0, 70),
#'     scales = list(alternating = 1, y = list(tick.number = 10)),
#'     par.settings = tactile.theme(superpose.line = list(lwd = 2)),
#'     xlab = 'Sand, Silt, Clay (%)',
#'     ylab = 'Depth (cm)',
#'     panel = panel.depth_function,
#'     auto.key = list(columns = 3, lines = TRUE, points = FALSE),
#'     as.table = TRUE
#'   )
#'   
#'   
#' }
panel.depth_function <- function(x, y, id, upper=NA, lower=NA, subscripts=NULL, groups=NULL, sync.colors=FALSE, cf=NA, cf.col=NA, cf.interval=20, ...) {
  
  # add grid
  panel.grid(h=-1, v=-1, lty=3, col=1)
  
  # load current line style
  superpose.line <- trellis.par.get("superpose.line")
  
  # when the length of 'y' is > 'x', we are plotting a step function
  if(length(y) > length(x)) {
    
    # sanity check
    if(missing(id)) {
      stop('must provide a profile id')
    }
      
    # re-make a nice data.frame, assuming that we have 'groups' defined
    if(!missing(groups)) {
      d <- data.frame(
        prop = x, 
        bnd = y, 
        upper = upper[subscripts], 
        lower = lower[subscripts], 
        groups = groups[subscripts], 
        id = id[subscripts]
      )
    } else {
      # 'groups' is missing, add a fake 'groups' column
      d <- data.frame(
        prop = x, 
        bnd = y, 
        upper = upper[subscripts], 
        lower = lower[subscripts], 
        groups = factor(1), 
        id = id[subscripts]
      )
    }
    
    
    # iterate over IDs creating step functions, by group
    by(d, d$id, .make.segments, ...)
  }
  
  # normal plot -- not a step function
  else {
    # if we have an upper and lower bound defined, plot them
    if(!missing(upper) & !missing(lower)) {
      # working with grouped data and paneled data
      if(!missing(groups) & !missing(subscripts)) {
        d <- data.frame(
          yhat = x, 
          top = y, 
          upper = upper[subscripts], 
          lower = lower[subscripts], 
          groups = groups[subscripts]
        )
        
        # levels in the groups, for color matching
        ll <- levels(d$groups)
        n_groups <- length(ll)
      }
      
      # no grouping, add a fake group for compatibility
      if(missing(groups)) {
        fake.groups <- factor(1)
        d <- data.frame(
          yhat = x, 
          top = y, 
          upper = upper[subscripts], 
          lower = lower[subscripts], 
          groups = fake.groups
        )
        
        # levels in the groups, for color matching
        ll <- levels(d$groups)
        n_groups <- length(ll)
      }
      
      # optionally sync region + main line colors
      if(sync.colors) {
        region.col <- rep(superpose.line$col, length.out = n_groups)
      } else {
        region.col <- rep(grey(0.7), length.out = n_groups)
      }
        
      
      # add conf. intervals / aggregation uncertainty
      by(d, d$groups, function(d_i) {
        
        # lookup color
        m <- match(as.character(d_i$group[1]), ll)
        
        # cannot have NA in values that define polygon boundaries
        d_i <- d_i[which(!is.na(d_i$upper) & !is.na(d_i$lower)), ]
        
        # make conf.int polygon
        x.coords <- c(d_i$lower, rev(d_i$upper))
        y.coords <- c(d_i$top, rev(d_i$top))
        poly.col <- region.col[m]
        
        # note: using tactile interfers with alpha
        panel.polygon(x = x.coords, y = y.coords, col = poly.col, border = NA, ...)
      })
    }
    
    # no upper/lower polygon boundaries defined
    else {
      if(missing(groups)) {
        fake.groups <- factor(1)
        d <- data.frame(yhat=x, top=y, groups=fake.groups)
      } else {
        d <- data.frame(yhat=x, top=y, groups=groups[subscripts])
      }
      
      # levels in the groups, for color matching
      ll <- levels(d$groups)
      n_groups <- length(ll)
    }
    
    # setup style parameters for main lines
    # repeat enough times for the current number of groups
    line.col <- rep(superpose.line$col, length.out=n_groups)
    line.lty <- rep(superpose.line$lty, length.out=n_groups)
    line.lwd <- rep(superpose.line$lwd, length.out=n_groups)
    
    # add main lines
    by(d, d$groups, function(d_i){
      # lookup color
      m <- match(unique(d_i$group), ll)
      
      # add line
      panel.lines(d_i$yhat, d_i$top, lwd=line.lwd[m], col=line.col[m], lty=line.lty[m])
    })
  }
  
  
  # annotate with contributing fraction
  if(!missing(cf)) {
    
    # add CF data to panel's worth of data
    d$cf <- cf[subscripts]
    
    # annotate with contributing fraction by group
    by(d, d$groups, function(d_i){
      
      # lookup linestyle by group
      m <- match(unique(d_i$group), ll)
      
      # if there is no user-specified CF color, then use the same as line style
      if(is.na(cf.col)) {
        cf.col <- line.col[m]
      }
      
      # make a function for linear interpolation of CF values based on depth
      cf.approx.fun <- approxfun(d_i$top, d_i$cf, method='linear')
      
      # generate annotated depths: 5 cm to 95th percentile of max depth
      y.q95 <- quantile(d_i$top, probs=c(0.95), na.rm=TRUE)
      a.seq <- seq(from=2, to=y.q95, by=cf.interval)
      
      # offset CF sequence according to group index
      a.seq <- a.seq + ((m-1) * cf.interval/4)
      
      # interpolate CF at annotated depths
      a.CF <- cf.approx.fun(a.seq)
      a.text <- paste(round(a.CF * 100), '%')
      
      # remove any NAs in CF sequence and text
      not.na.idx <- which(! is.na(a.CF))
      a.seq <- a.seq[not.na.idx]
      a.text <- a.text[not.na.idx]
      
      # add to right-hand side of the panel
      unit <- gpar <- NULL
      grid::grid.text(
        a.text,
        x = grid::unit(0.99, 'npc'),
        y = grid::unit(a.seq, 'native'),
        just = 'right',
        gp = grid::gpar(font = 3, cex = 0.8, col = cf.col)
      )
    })
    
  }
  
}

# internally used function, for panel.depth_function() step-function mode
.make.segments <- function(df) {
  
  # group data
  ll <- levels(df$groups)	
  n_groups = length(ll)
  
  # get the current group index
  m <- match(unique(df$group), ll)
  
  # load current line style
  superpose.line <- trellis.par.get("superpose.line")
  
  # repeat enough times for the current number of groups
  line.col <- rep(superpose.line$col, length.out = n_groups)
  line.lty <- rep(superpose.line$lty, length.out = n_groups)
  line.lwd <- rep(superpose.line$lwd, length.out = n_groups)
  
  for(i in m) {
    
    this.group <- df[df$groups == ll[i], ]
    
    # get the number of depth slices
    n_hz <- length(this.group$prop) / 2
    
    # need at least 2 horizons
    if(n_hz > 1) {
      
      .top <- this.group$bnd[1:n_hz]
      .bottom <- this.group$bnd[(n_hz+1):length(this.group$prop)]
      .prop <- this.group$prop[1:n_hz]
      
      # vertical segments
      panel.segments(x0 = .prop, y0 = .top, x1 = .prop, y1 = .bottom, 
                     lwd=line.lwd[i], col=line.col[i], lty=line.lty[i]) 
      
      # horizontal segments
      panel.segments(x0 = .prop[-n_hz], y0 = .bottom[-n_hz], x1 = .prop[-1], y1 = .top[-1], 
                     lwd=line.lwd[i], col=line.col[i], lty=line.lty[i])
    }
    
    else {
      message(paste('only 1 horizon, skipping!', df$groups[1]))
    }
    
  }
  
}

## TODO: this may not scale to > 5 colors
## TODO: this should have more options for title / subtitle


#' @title Soil Color Palette
#' 
#' @description A very simple function for generating labeled swatches of soil colors. Largely based on `colorspace::swatchplot`.
#' 
#' @param colors vector of hex colors (e.g. #A66E46FF)
#' 
#' @param lab vector of labels
#' 
#' @param lab.cex character scaling for labels
#' 
#' @param dynamic.labels logical, adjust label colors for maximum contrast via `invertLabelColor`
#' 
#' @param x.inset horizontal adjustment for labels
#' 
#' @param y.inset vertical adjustment for labels
#' 
#' @param \dots further arguments to `colorspace::swatchplot`
#' 
#' @author D.E. Beaudette
#' 
#' @note The result is a simple figure on the active plotting device.
#' 
#' @return nothing is returned, function called for graphic output
#' 
#' @keywords hplots
#' @export
#' @examples
#' 
#' # maybe useful for teaching about soil color
#' 
#' par(mfrow=c(2,1), mar=c(1,1,1,1))
#' 
#' # demonstrate range of Munsell value
#' m <- sprintf('10YR %s/4', 2:8)
#' # convert to hex representation
#' cols <- parseMunsell(m)
#' # plot
#' soilPalette(cols, m)
#' 
#' # demonstrate range of Munsell chroma
#' m <- sprintf('10YR 4/%s', 2:8)
#' # convert to hex representation
#' cols <- parseMunsell(m)
#' # plot
#' soilPalette(cols, m)
#' 

soilPalette <- function(colors, lab = colors, lab.cex=0.75, dynamic.labels = TRUE, x.inset = 0.01, y.inset = 0.01, ...) {
  
  ## TODO: consider extending swatchplot beyond the original scope, more flexibility
  # basic plot
  colorspace::swatchplot(colors, ...)
  
  ## TODO: this doesn't scale to more than a single row of colors
  ## use code from colorContrastPlot to dynamically adjust for multiple rows
  ## possibly share code
  
  # annotation
  nx <- length(colors)
  x.pos <- seq(from = 0, to = 1, by = 1/nx)[1:nx] + x.inset
  y.pos <- rep(y.inset, times = nx)
  
  # label color varies as a function of value
  if(dynamic.labels) {
    label.color <- invertLabelColor(colors)
  } else {
    # use device background color
    label.color <- par()$bg
  }
  
  # annotate
  text(x.pos, y.pos, labels = lab, cex=lab.cex, col = label.color, font = 2, adj = c(0, 0))
}

# 
#' DEPRECATED Simulate Soil Profiles
#' 
#' @description Simulate a collection of soil profiles based on the horizonation of a single soil profile. Now deprecated: use [perturb()] for perturbations of horizon thicknesses or boundaries.
#' 
#' @param x a SoilProfileCollection object containing a single profile from which to draw simulated data
#' @param n the number of requested simulations
#' @param iterations sampling iterations used to determine each horizon thickness
#' @param hz.sd standard deviation used to simulate horizon thickness, can be a vector but must divide evenly into the number of horizons found in \code{x}
#' @param min.thick minimum horizon thickness allowed in simulation results
#'
#' @return A SoilProfileCollection object with \code{n} simulated profiles, each containing the same number of horizons and same data as \code{x}
#' 
#' @details This function generates a collection of simulated soil profiles based on the horizon thickness data associated with a single "template" profile. Simulation is based on sampling from a family of Gaussian distribution with means defined by the "template" profile and standard deviation defined by the user.
#' 
#' @export
#' 
#' @seealso \code{\link{random_profile}}  \code{\link{perturb}}
#' 
#' @author D. E. Beaudette
#' 
#' @examples
#' 
#' # please see documentation for perturb() for examples
#' #  the sim() function calls perturb() internally
#' 
sim <- function(x,
                n = 1,
                iterations = 25,
                hz.sd = 2,
                min.thick = 2) {	
  
  .Deprecated("perturb")
  
  horizons(x)$.perturb_internalHzSD <- hz.sd

  res <- perturb(x, 
                 n = n, 
                 thickness.attr = ".perturb_internalHzSD",
                 min.thickness = min.thick,
                 max.depth = NULL)
  
  res$.perturb_internalHzSD <- NULL
  return(res)
#   hd <- horizonDepths(x)
# 	h <- horizons(x)
# 	thick <- h[[hd[2]]] - h[[hd[1]]]
# 	
# 	# remove original depth columns
# 	h[[hd[1]]] <- NULL
# 	h[[hd[2]]] <- NULL
# 	
# 	# remove horizon ID so as not to create conflicts later on
# 	h[[hzidname(x)]] <- NULL
# 	
# 	# keep track of old id
# 	old.id.name <- idname(x)
# 	
# 	# sanity checks
# 	if(length(x) > 1)
# 		stop('this function can only simulate data from a SoilProfileCollection containing a single profile')
# 	
# 	if(length(thick) %% length(hz.sd) != 0)
# 		stop('the length of hz.sd must divide evenly into the number of horizons', call.=FALSE)
# 	
# 	# define function to wrap rnorm for use with outer
# 	rnorm.vect <- function(mean, sd, n) {
# 		rnorm(n=n, mean=mean, sd=sd)
# 	}
# 	rnorm.vect <- Vectorize(rnorm.vect, vectorize.args=c('mean', 'sd'))
# 	
# 	# allocate storage for simulated horizon depths
# 	l <- list()
# 	
# 	# generate n-simulated horizon depths
# 	for(i in 1:n) {
# 		# simulate
# 		s <- mapply(rnorm.vect, thick, hz.sd, n=iterations)
# 		
# 		# sample a single value per horizon, and use as the representative value
# 		s <- round(apply(s, 2, sample, size=1))
# 		
# 		# convert thickness values that are below min.thick -> min.thick
# 		s <- pmax(s, min.thick)
# 		
# 		# convert thickness -> depths
# 		s <- cumsum(s)
# 		# use a ID column name that isn't likely to conflict with existing column names
# 		d <- data.frame(.new_id=i, top=c(0, s[-length(s)]), bottom=s)
# 		
# 		# combine with original horizon data, and save to list element
# 		l[[i]] <- cbind(d, h)
# 	}
# 	
# 	# convert list -> data.frame
# 	x.s <- do.call(rbind, l)
# 	
# 	# upgrade to SoilProfileCollection
# 	depths(x.s) <- .new_id ~ top + bottom
# 	
#   # copy over depth units
# 	depth_units(x.s) <- depth_units(x)
# 	
# 	# move old ID into @site
# 	fm <- as.formula(paste0('~ ', old.id.name))
# 	site(x.s) <- fm
# 	
# 	# done
# 	return(x.s)
}

#' @title Annotate Diagnostic Features
#' @description Annotate diagnostic features within a sketch of soil profiles.
#' 
#' @param s `SoilProfileCollection` object
#' @param kind character, filter applied to `feature` column of diagnostic horizons registered within `s`
#' @param feature column name containing feature kind
#' @param top column name containing feature top depth
#' @param bottom column name containing feature top depth
#' @param ... additional arguments passed to [addBracket()]
#'
#' @details Additional examples can be found in [this tutorial](http://ncss-tech.github.io/AQP/aqp/SPC-plotting-ideas.html).
#' 
#' @note This is a `low-level` plotting function: you must first plot a `SoilProfileCollection` object before using this function.
#' 
#' @author D.E. Beaudette
#' 
#' @seealso [addBracket()], [plotSPC()]
#'
#' @export
#'
#' @examples
#' 
#'  # example data
#' x <- c(
#'   'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
#'   'P2:Ap|AA|E|BhsBhs|Bw1Bw1|CCCCC',
#'   'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR',
#'   'P4:AA|EEE|BhsBhsBhsBhs|BwBw|CCCCC',
#'   'P5:AAAA|ACACACACAC|CCCCCCCCCCC|CdCdCd'
#' )
#' 
#' s <- quickSPC(x)
#' 
#' diagnostic_hz(s) <- data.frame(
#'   id = c('P1', 'P4'),
#'   t = c(12, 25), 
#'   b = c(70, 100),
#'   kind = c('Best', 'Best')
#' )
#' 
#' op <- par(no.readonly = TRUE)
#' par(mar = c(0, 0, 3, 2))
#' 
#' # sketches
#' plotSPC(
#'   s, name = 'name', name.style = 'center-center', cex.names = 0.75, max.depth = 210
#' )
#' 
#' # note that custom top/bottom depths must be supplied
#' addDiagnosticBracket(
#'   s, feature = 'kind', kind = 'Best', top = 't', bottom = 'b',
#'   labcol = 'kind',
#'   offset = -0.35, col = 'firebrick', tick.length = 0.04, lwd = 2
#' )
#' 
#' par(op)
#' 
#'
addDiagnosticBracket <- function(s, kind, feature = 'featkind', top = 'featdept', bottom = 'featdepb', ...) {

  # extract diagnostic horizon information
  # note: the idname is already present in `d`
  d <- diagnostic_hz(s)
  d <- d[which(d[[feature]] == kind), ]

  # there may be no matching features, in that case issue a message and do nothing
  if(nrow(d) < 1) {
    message('no matching features found')
  } else {
    # add backets
    # sorting is done via matching idname to plot order of idname
    addBracket(d, hzDepths = c(top, bottom), ...)
  }

}



#' @title Add Depth Brackets
#' 
#' @description Add depth brackets to soil profile sketches.
#'
#' @param x `data.frame` containing at least: `idname(x)` (profile ID) and `horizonDepths(x)` (horizon top and bottom depths)
#' 
#' @param labcol character, optional name of a column in `x` used for labeling
#' 
#' @param agg logical, aggregate multiple brackets per profile into a single depth range?
#' 
#' @param hzDepths character vector of length 2, optional column names in `x` that define bracket top and bottom depths. When `NULL`, use horizon top/bottom column names from the `SoilProfileCollection` object used by the last call to `plotSPC()`.
#' 
#' @param label.cex numeric, scaling factor for label font
#' 
#' @param tick.length numeric, length of bracket "tick" mark
#' 
#' @param arrow.length numeric, length of arrowhead (see `arrows()`)
#' 
#' @param offset numeric, left-hand offset from each profile
#' 
#' @param missing.bottom.depth numeric, distance (in depth units) to extend brackets that are missing a lower depth (defaults to max depth of collection)
#' 
#' @param ... further arguments passed on to `segments()` or `arrows()`
#' 
#' @details When `x` contains multiple records per profile a bracket will be created for each record. Setting `agg = TRUE` will first aggregate all records per profile, then add a single bracket spanning the depth range of those records. Additional examples can be found in \href{http://ncss-tech.github.io/AQP/aqp/SPC-plotting-ideas.html}{this tutorial}.
#' 
#' @note This is a `low-level` plotting function: you must first plot a `SoilProfileCollection` object before using this function. Details about the last plotted `SoilProfileCollection` are available using `get('last_spc_plot', envir = aqp.env)`. 
#' 
#' @author D.E. Beaudette
#' 
#' @export
#' 
#' @seealso [addDiagnosticBracket()], [plotSPC()]
#' 
#' @examples
#' 
#' # example data
#' x <- c(
#'   'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
#'   'P2:Ap|AA|E|BhsBhs|Bw1Bw1|CCCCC',
#'   'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR',
#'   'P4:AA|EEE|BhsBhsBhsBhs|BwBw|CCCCC',
#'   'P5:AAAA|ACACACACAC|CCCCCCC|CdCdCd'
#' )
#' 
#' s <- quickSPC(x)
#' 
#' # change horizon depth names
#' # ensure that plotSPC() -> addBracket() can find them
#' horizonDepths(s) <- c('tt', 'bb')
#' 
#' # expression defines a single reference horizon in most profiles
#' .ex <- grepl('Bt3|Bw', s$name)
#' # encode for thematic profile sketches
#' s$e <- factor(as.character(.ex), levels = c('FALSE', 'TRUE'), labels = c('Horizons', 'Reference'))
#' 
#' # get horizon row indices to horizons above reference
#' a <- hzAbove(s, .ex, SPC = FALSE, simplify = TRUE)
#' 
#' # create bracket data.frame
#' b <- depths(s, hzID = FALSE)[a, ]
#' 
#' # add labels
#' b$label <- c('S')
#' 
#' op <- par(no.readonly = TRUE)
#' par(mar = c(0, 0, 3, 2), mfcol = c(1, 2))
#' 
#' # sketches
#' plotSPC(
#'   s, color = 'e', col.label = 'Original', col.palette = c('grey', 'royalblue'), 
#'   name = 'name', name.style = 'center-center', cex.names = 0.75,
#'   max.depth = 180
#' )
#' 
#' # plot individual brackets, no labels
#' addBracket(
#'   b, 
#'   agg = FALSE, labcol = 'label',
#'   offset = -0.35, col = 'black', tick.length = 0.04, lwd = 1
#' )
#' 
#' # sketches
#' plotSPC(
#'   s, color = 'e', col.label = 'Aggregate', col.palette = c('grey', 'royalblue'), 
#'   name = 'name', name.style = 'center-center', cex.names = 0.75,
#'   max.depth = 180
#' )
#' 
#' # aggregate multiple brackets into single depth span
#' # include first label from each group
#' addBracket(
#'   b, 
#'   agg = TRUE, labcol = 'label',
#'   offset = -0.35, col = 'firebrick', tick.length = 0.04, lwd = 2
#' ) 
#' 
#' par(op)
#' 
addBracket <- function(x, labcol = NULL, agg = FALSE, hzDepths = NULL, label.cex = 0.75, tick.length = 0.05, arrow.length = 0.05, offset = -0.3, missing.bottom.depth = NULL, ...) {

  # get plotting details from aqp environment
  lsp <- get('last_spc_plot', envir = aqp.env)
  
  # y.offset is a vector length(x)
  depth.offset <- lsp$y.offset
  sf <- lsp$scaling.factor
  
  
  # setup short names for horizon top and bottom depths
  
  # bracket depth column names inherited from the last plotSPC()
  if(is.null(hzDepths)) {
    .top <-lsp$horizonDepths[1]
    .bottom <- lsp$horizonDepths[2]
  } else {
    # bracket depth column names provided
    .top <- hzDepths[1]
    .bottom <- hzDepths[2]
  }

  
  # line ending "2" is usually what we want
  # however, when using very large line widths and tick.length = 0, the interval will extend
  # beyond the top/bottom depths
  if(tick.length == 0) {
    seg.lend <- 1
  } else {
    seg.lend <- 2
  }
  
  
  # test for required columns:
  # profile ID, top, bottom <-- these are defined in lsp
  if(is.null(x[[.top]]) | is.null(x[[.bottom]]) | is.null(x[[lsp$idname]])) {
    stop('required columns missing', call. = FALSE)
  }
  
  # test for < 0 rows
  if(nrow(x) < 1) {
    warning('no rows')
    return(FALSE)
  }

  # test for all missing top depths
  if(all(is.na(x[[.top]]))) {
    stop('no top depths supplied')
  }
    

  # test for missing label column
  do.label <- TRUE
  if(is.null(labcol)) {
    do.label <- FALSE
  }
  
  ## TODO: this must also be done a profile at a time
  # test for all bottom depths missing
  no.bottom <- FALSE
  if(all(is.na(x[[.bottom]]))) {
    no.bottom <- TRUE
  }
  
  # if not specified, use the max depth in the last plot
  if(is.null(missing.bottom.depth)) {
    missing.bottom.depth <- lsp$max.depth
  }

  # apply scale and offset to missing bottom depth
  # depth.offset is a vector length(x)
  missing.bottom.depth <- (missing.bottom.depth * sf) + depth.offset

  # apply scaling factor and offset
  # depth_prime = (depth * scaling factor) + y.offset
  # in case x is a subset of the last plotSPC call, must index
  idx <- match(x[[lsp$idname]], lsp$pIDs)
  x[[.top]] <- (x[[.top]] * sf) + depth.offset[idx]
  x[[.bottom]] <- (x[[.bottom]] * sf) + depth.offset[idx]
  
  ## x-coordinates
  # 2019-07-15: using relative position
  x.base <- lsp$x0
  
  # there may be more than 1 bracket per ID
  x.list <- split(x, x[[lsp$idname]])
  
  ## TODO: match logic seems backwards
  # re-order list elements, according to plot order
  # there may be some profiles without brackets to add
  # resulting in NA in re-ordering index
  idx <- match(lsp$pIDs, names(x.list))
  # re-order in the presence of NA
  x.list <- x.list[idx]
  
  # iterate over list of profile IDs
  for(i in seq_along(x.list)) {
    
    # current set of brackets (single profile)
    brackets <- x.list[[i]]
    
    # skipping profiles with missing brackets
    if(! is.null(brackets)) {
      
      
      # optional aggregation over multiple spans / profile
      # resulting label will be the first row
      if(agg) {
        .atop <- aggregate(brackets[[.top]], by = list(brackets[[lsp$idname]]), FUN = min)
        .abottom <- aggregate(brackets[[.bottom]], by = list(brackets[[lsp$idname]]), FUN = max)
        
        if(do.label) {
          .res <- cbind(.atop, .abottom[, 2], brackets[[labcol]][1])
          names(.res) <- c(lsp$idname, lsp$horizonDepths, labcol)
        } else {
          .res <- cbind(.atop, .abottom[, 2])
          names(.res) <- c(lsp$idname, lsp$horizonDepths)
        }
        
        brackets <- .res
      }
      
      # note: these may contain NA
      # variables recycled within loop
      top <- brackets[[.top]]
      bottom <- brackets[[.bottom]]
      
      # missing bottom: replace bottom tick with arrow head
      if(no.bottom) {
        
        # x-positions
        # x.base is a vector
        x.1 <- x.base[i] + offset
        x.2 <- x.1 + tick.length
        # top tick
        segments(x.1, top, x.2, top, lend = seg.lend, ...)
        # vertical bar is now an arrow
        arrows(x.1, top, x.1, top + missing.bottom.depth[i], length = arrow.length, lend = seg.lend, ...)
      } else {
        # normal usage
        
        # x-positions
        # x.base is a vector
        x.1 <- x.base[i] + offset
        x.2 <- x.1 + tick.length
        # top tick
        segments(x.1, top, x.2, top, lend = seg.lend, ...)
        # bottom tick
        segments(x.1, bottom, x.2, bottom, lend = seg.lend, ...)
        # vertical bar
        segments(x.1, top, x.1, bottom, lend = seg.lend, ...)
      }
      
      # optionally plot label
      if(do.label) {
        label <- brackets[[labcol]]
        if(no.bottom) {
          bottom <- rep(missing.bottom.depth[i], times = length(bottom))
        }
        # add labels at mid-points
        # requires a small horizontal nudge
        text(x.1 - 0.1, (top + bottom)/2, label, srt = 90, cex = label.cex, adj = c(0.5, 0.5))
      }
    }
  }
}

context("plotSPC")


## TODO:
# * handling of degenerate profiles


data(sp1)
# convert colors from Munsell to hex-encoded RGB
sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))

# promote to SoilProfileCollection
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# additional example data
set.seed(101010)
p <- lapply(letters[1:10], random_profile, method = 'LPP', SPC = TRUE)
p <- combine(p)

# make a factor style hz attr
p$p.factor <- cut(p$p1, quantile(p$p1), include.lowest = TRUE)

# note: hzdesgnname() returns "" when metadata are missing

## tests

test_that("plotSPC: aqp.env settings", {

  # explainer
  explainPlotSPC(sp1)

  # get plotting details from aqp environment
  lsp <- get('last_spc_plot', envir = aqp.env)

  # should be a list
  expect_true(is.list(lsp))

  # check for required components
  expect_true(
    all(
      c("width", "plot.order", "x0", "pIDs", "idname", "y.offset", "scaling.factor","max.depth", "n", "extra_x_space", "extra_y_space", "hz.depth.LAI") %in% names(lsp)
    )
  )
    
  # basic integrity checks
  expect_equal(profile_id(sp1), lsp$pIDs)
  expect_equal(idname(sp1), lsp$idname)
  expect_true(length(sp1) == lsp$n)
  expect_equal(1:length(sp1), lsp$plot.order)
})


test_that("plotSPC: figure settings", {

  # explainer returns `lsp`
  lsp <- explainPlotSPC(sp1, scaling.factor=0.5, width=0.8,
                        y.offset=8, n=15, max.depth = 100)

  # check adjustments
  expect_equal(lsp$scaling.factor, 0.5)
  expect_equal(lsp$width, 0.8)
  expect_equal(lsp$y.offset, rep(8, times = length(sp1)))
  expect_equal(lsp$n, 15)
  expect_equal(lsp$max.depth, 100)
})


test_that("plotSPC: re-ordering of profiles", {

  # re-order
  new.order <- c(1,3,5,7,9,2,4,6,8)
  # explainer returns `lsp`
  lsp <- explainPlotSPC(sp1, plot.order=new.order)

  # profile IDs should be sorted according to plot.order
  expect_equal(profile_id(sp1)[new.order], lsp$pIDs)

  # plotting order should be saved
  expect_equal(new.order, lsp$plot.order)
})

# output is purely graphical, testing to make sure no errors / warnings are generated
test_that("plotSPC: shrinking horizon names", {
  
  # make sure these work
  plotSPC(sp1, name.style = 'center-center', shrink.thin = NULL, shrink = TRUE, cex.names = 0.8)
  plotSPC(sp1, name.style = 'center-center', shrink.thin = 5, shrink = TRUE, cex.names = 0.8)
  # test
  lsp <- explainPlotSPC(sp1, name.style = 'center-center', shrink.thin = 15, shrink = TRUE, cex.names = 0.8)
  expect_true(is.list(lsp))
})



test_that("plotSPC: relative spacing", {

  # relative positions
  x.pos <- c(1+0, 2+0, 3+0, 4+0.1, 5+0.1, 6+0.1, 7+0.2, 8+0.2, 9+0.2)
  # explainer returns `lsp`
  lsp <- explainPlotSPC(sp1, relative.pos = x.pos)

  # plot order not affected
  expect_equal(lsp$plot.order, 1:length(sp1))

  # x0 adjusted as expected
  # 1:length(x) + offsets
  expect_equal(lsp$x0, x.pos)
})



test_that("plotSPC: re-ordering of profiles and relative spacing", {

  # new order
  new.order <- c(1,3,5,7,9,2,4,6,8)
  # relative positions: in the new ordering
  x.pos <- c(1+0, 2+0, 3+0, 4+0.1, 5+0.1, 6+0.1, 7+0.2, 8+0.2, 9+0.2)

  # explainer returns `lsp`
  lsp <- explainPlotSPC(sp1, plot.order=new.order, relative.pos = x.pos)

  # profile IDs should be sorted according to plot.order
  expect_equal(profile_id(sp1)[new.order], lsp$pIDs)

  # plotting order should be saved
  expect_equal(new.order, lsp$plot.order)

  # x0 adjusted as expected
  expect_equal(lsp$x0, x.pos)
})


test_that("plotSPC: re-ordering via relative spacing", {

  # re-order by adjusting the relative positions
  x.pos <- length(sp1):1

  # explainer returns `lsp`
  lsp <- explainPlotSPC(sp1, relative.pos = x.pos)

  # plotting order / IDs are not modified!
  expect_equal(profile_id(sp1), lsp$pIDs)
  expect_equal(lsp$plot.order, 1:length(sp1))

  # x0 adjusted as expected
  expect_equal(lsp$x0, x.pos)
})


test_that("plotSPC: y-offset reordered by plot.order", {

  data("jacobs2000")
  x <- jacobs2000
  hzdesgnname(x) <- 'name'
  
  # y-offset + reverse order
  lsp <- explainPlotSPC(x, y.offset = (1:7) * 10, plot.order = 7:1)
  
  # check that y-offset is re-ordered
  expect_true(
    all(lsp$y.offset == rev((1:7) * 10))
  )
  
  # check IDs are reordered
  expect_equal(rev(profile_id(x)), lsp$pIDs)
  
})





test_that("addBracket works", {

  expect_silent(addBracket(data.frame(id = profile_id(sp1), label="bar", top = 0, bottom = 25)))

  expect_silent(addBracket(data.frame(id = profile_id(sp1), top = 0, bottom = NA)))

  diagnostic_hz(sp1) <- data.frame(id = profile_id(sp1), featkind = "foo", featdept = 0, featdepb = 50)

  expect_silent(addDiagnosticBracket(sp1, kind =  "foo"))

})

test_that("addVolumeFraction works", {

  # does it work with default arguments?
  plotSPC(sp1, name = 'name')

  expect_silent(addVolumeFraction(sp1, 'prop'))

  # additional arguments
  expect_silent(addVolumeFraction(sp1, 'prop', res = 5))

  expect_silent(addVolumeFraction(sp1, 'prop', cex.min = 0.5, cex.max = 2))

  expect_silent(addVolumeFraction(sp1, 'prop', pch = 15))

  # color specification is important:
  # single color
  expect_silent(addVolumeFraction(sp1, 'prop', col = 'red'))

  # or vector of colors, must be same length as nrow(x)
  expect_silent(addVolumeFraction(sp1, 'prop', col = rep('green', times=nrow(sp1))))

  sp1$prop1k <- sp1$prop / 1000
  # message due to values < 1
  expect_message(addVolumeFraction(sp1, 'prop1k'),
                 "all prop1k values are < 0.5, likely a fraction vs. percent")

  sp1$prop1k <- sp1$prop * 1000
  # warnings due to values >100
  expect_warning(addVolumeFraction(sp1, 'prop1k'))

})


test_that("addVolumeFraction expected errors", {

  plotSPC(sp1, name='name')


  # bad column name
  expect_error(addVolumeFraction(sp1, 'prop1'))

  # incorrectly specified colors
  expect_error(addVolumeFraction(sp1, 'prop', col = c('red', 'green')))

})

# https://github.com/ncss-tech/aqp/issues/8
test_that("addVolumeFraction fractional horizon depths", {

  plotSPC(sp1, name='name')


  # modify depths
  sp1$top[4] <- sp1$top[4] + 0.5
  sp1$bottom[3] <- sp1$top[4]

  # fractional horizon depths
  expect_message(addVolumeFraction(sp1, 'prop'), regexp = 'truncating')

})


test_that("horizon color specification interpreted correctly", {
  
  # this function works with contents of @horizons
  h <- horizons(p)
  
  # colors to use
  cols <- c("#5E4FA2", "#3288BD", "#66C2A5",
            "#ABDDA4", "#E6F598", "#FEE08B",
            "#FDAE61", "#F46D43", "#D53E4F",
            "#9E0142")
  
  # attempt to interpret
  x <- .interpretHorizonColor(
    h,
    color = 'p1',
    default.color = 'grey',
    col.palette = cols,
    col.palette.bias = 1,
    n.legend = 8
  )
  
  # reasonable object?
  expect_true(inherits(x, 'list'))
  expect_true(length(x) == 2)
  expect_true(length(x$color.legend.data) == 4)
  expect_false(x$color.legend.data$multi.row.legend)
  expect_null(x$color.legend.data$leg.row.indices)
  
  # another try, no colors specified
  x <- .interpretHorizonColor(
    h, 
    color = NA, 
    default.color = 'grey', 
    col.palette = cols,
    col.palette.bias = 1,
    n.legend = 8
  )
  
  # horizon colors should match default.color
  expect_true(all(x$colors == 'grey'))
  
  # there is no legend
  expect_true(is.null(x$color.legend.data))
  
  # factor variable + multi-line legend
  x <- .interpretHorizonColor(
    h, 
    color = 'p.factor', 
    default.color = 'grey', 
    col.palette = cols,
    col.palette.bias = 1,
    n.legend = 2
  )
  
  # multi-line legend details
  expect_true(x$color.legend.data$multi.row.legend)
  # row indices are stored in a list
  expect_true(inherits(x$color.legend.data$leg.row.indices, 'list'))
  # there should be two rows
  expect_true(length(x$color.legend.data$leg.row.indices) == 2)
  # legend item indices are stored by row 
  expect_equal(x$color.legend.data$leg.row.indices[[1]], c(1,2))
  expect_equal(x$color.legend.data$leg.row.indices[[2]], c(3,4))
  
})

.Gmedian.chunk <- function(i, id = '.chunk') {
  
  # iterate over chunks
  depth.list <- split(i, i[[id]])
  
  res <- lapply(depth.list, function(j) {
    
    # structure: group, chunk, top, bottom, var1, var2, var3, ...
    v <- 5:ncol(j)
    
    # NA not allowed, keep track and filter
    idx <- which(complete.cases(j[, v, drop = FALSE]))
    
    # catch conditions where there are no data
    # TODO: what is the min number of required records?
    if(length(idx) < 1)
      return(NULL)
    
    ## TODO: how do arguments to Gmedian() affect the results?
    # L1 median for non-NA records, variables of interest only
    G <- data.frame(Gmedian::Gmedian(X = j[idx, v, drop = FALSE]))
    
    # retain original names
    names(G) <- names(j)[v]
    
    # package group + top + bottom + L1 data
    d <- data.frame(group = j[1, 1], top = j[1, 3], bottom = j[1, 4], G, stringsAsFactors = FALSE)
    
    return(d)
  })
  
  # list -> DF
  res <- do.call('rbind', res)
  
  return(res)
}


## TODO:
# would it make more sense to select a single profile based on NCSP?
# how does variable scale affect Gmedian? 
# best practices for variable name compatibility (slab-style vs. original names)
# slab-style support for arbitrary depth intervals via basis argument
# most-likely horizon designation by chunk
# parallel Gmedian computation
# add principal components for viz
# document!

#' @title Create Representative Soil Profiles via L1 Estimator
#' 
#' @description The L1 estimator, or \href{https://en.wikipedia.org/wiki/Geometric_median}{geometric median}, is a multivariate generalization of the (univariate) \href{https://en.wikipedia.org/wiki/Median}{median} concept. This function performs a multivariate aggregation (via L1 estimator) according to a suite of ratio-scale soil properties. The L1 estimator is applied to soil profile data that have been sliced to a 1-depth-unit basis. Data should be well stratified by groups defined in `fm`, otherwise the L1 median may not make any sense.
#' 
#' See the \href{https://ncss-tech.github.io/AQP/aqp/L1-profiles.html}{L1 Profiles Tutorial} for additional examples.
#'
#' @note This function requires the `Gmedian` package.
#' 
#' @references Cardot, H., Cenac, P. and Zitt, P-A. (2013). Efficient and fast estimation of the geometric median in Hilbert spaces with an averaged stochastic gradient algorithm. Bernoulli, 19, 18-43.
#'
#' @param x \code{SoilProfileCollection} object
#' 
#' @param fm formula, for example: `group ~ p1 + p2 + p3`, where "group" is a site-level grouping variable, and "p1", "p2", and "p3" are horizon level variables
#' 
#' @param basis positive integer, aggregation basis (e.g. 1 for 1-depth-unit intervals). Values other than 1 are not currently supported.
#' 
#' @param method soil depth evaluation method: "regex" for regular expression, "simple", or "constant". See details.
#' 
#' @param maxDepthRule maximum depth rule: "max" or "min" See details.
#' 
#' @param maxDepthConstant positive integer, maximum depth when \code{maxDepthRule = 'constant'}
#' 
#' @details See [this related tutorial](https://ncss-tech.github.io/AQP/aqp/L1-profiles.html) for additional examples. The `method`, `maxDepthRule`, and `maxDepthConstant` arguments set the maximum depth (over the entire collection) of analysis used to build "L1 profiles". The following rules are available:
#'   
#'   * `method = 'regex'` uses pattern matching on horizon designations (note that `hzdesgnname` metadata must be set with `hzdesgnname(x) <- 'columnname'`)
#'   
#'    * `method = 'simple'` uses `min` or `max` as applied to `x`, no accounting for non-soil horizons (e.g. Cr or R)
#'    
#'    * `method = 'constant'` uses a fixed depth value supplied by `maxDepthConstant`
#'
#' The `maxDepthRule` argument sets depth calculation constraint, applied to soil depths computed according to `method` (`min` or `max`).
#' 
#' @return a \code{SoilProfileCollection} object
#' @export
#'
L1_profiles <- function(x, fm, basis = 1, method = c('regex', 'simple', 'constant'), maxDepthRule = c('max', 'min'), maxDepthConstant = NULL) {
  
  # sanity check, need this for L1 median
  if(!requireNamespace('Gmedian')) {
    stop('package `Gmedian` is required', call. = FALSE)
  }
  
  # sanity checks: is this an SPC?
  if(! inherits(x, 'SoilProfileCollection')) {
    stop('`object` should be a SoilProfileCollection', call. = FALSE)
  }
  
  # extract components of the formula:
  g <- all.vars(update(fm, . ~ 0)) # left-hand side
  vars <- all.vars(update(fm, 0 ~ .)) # right-hand side
  
  # sanity check: do the variables specified in fm exist in the correct site / hz slots?
  if(
    ! all(vars %in% horizonNames(x)) |
    ! g %in% siteNames(x)
  ) {
    stop('`fm` must reference one or more horizon-level attributes and a single site-level attribute', call. = FALSE)
  }
  
  # arguments
  method <- match.arg(method)
  maxDepthRule <- match.arg(maxDepthRule)
  
  # multi-argument sanity
  if(method == 'constant' & !is.numeric(maxDepthConstant)) {
    stop('contant max depth must be specified by single numeric value', call. = FALSE)
  }
  
  # ensure that there is a horizon name defined when method = 'regex'
  if(method == 'regex'){
    if(hzdesgnname(x) == '') {
      stop('`x` must have a horizon designation defined to use `method = "regex"`, see `?hzdesgnname`', call. = FALSE)
    }
  }
  
  # SPC metadata
  hztb <- horizonDepths(x)
  
  # upper depth logic, usually 0
  collection.top <- min(x[[hztb[1]]], na.rm = TRUE)
  
  # lower depth logic
  collection.bottom <- switch(method,
         regex = {
           x.depths <- profileApply(x, estimateSoilDepth, name = hzdesgnname(x))
           switch(maxDepthRule, 
                  min = {
                    min(x.depths, na.rm = TRUE)
                  },
                  max = {
                    max(x.depths, na.rm = TRUE)
                  })
         },
         simple = {
           switch(maxDepthRule,
                  min = {
                    min(x)
                  },
                  max = {
                    max(x)
                  })
         },
         constant = {
           maxDepthConstant
         }
         
  )
   
  
  # create formula
  # this only needs the top/bottom depths
  slice.fm <- as.formula(sprintf("%s:%s ~ %s", collection.top, collection.bottom, paste(vars, collapse = ' + ')))
  
  # re-format for slice-wise L1 median
  s <- dice(x, fm = slice.fm, strict = TRUE, SPC = TRUE)
  
  # work on de-normalized data as a DF
  h <- as(s, 'data.frame')
  
  ## TODO: interpret basis as either slab thickness, or like slab.structure
  
  # determine which via length(basis)
  
  # assign slab IDs
  
  
  # simplest case: working with raw slices, use top depth
  if(basis == 1) {
    h$.chunk <- h[[hztb[1]]]
  } else {
    stop('sorry, depth basis other than `1` are not currently supported', call. = FALSE)
  }
  
  
  # retain only those columns we need
  h <- h[, c(g, '.chunk', hztb, vars)] 
  
  # iterate over groups
  group.list <- split(h, h[[g]])
  
  ## TODO: parallel
  agg <- lapply(group.list, .Gmedian.chunk)
  agg <- do.call('rbind', agg)
  
  # init SPC with new ID, top, bottom names
  depths(agg) <- group ~ top + bottom
  
  # transfer metadata
  agg <- .transfer.metadata.aqp(x, agg)
  
  # reset horizon designation for now
  hzdesgnname(agg) <- NULL
  
  return(agg)
}

context("L1 profiles")

## TODO: think about a good test dataset


test_that("works as expected", {
  skip_if_not_installed('Gmedian')
  
  # overly simplistic example data
  p <- lapply(as.character(1:20), random_profile, method = 'LPP', SPC = TRUE)
  p <- combine(p)
  
  # init site data with groups
  site(p)$group <- rep(letters[1:2], times = 10)
  p$group <- factor(p$group)
  
  # set horizon designation
  horizons(p)$name <- 'XX'
  hzdesgnname(p) <- 'name'
  
  # try it
  z <- L1_profiles(p, fm = group ~ p1 + p2 + p3, method = 'constant', maxDepthRule = 'max', maxDepthConstant = 100)
  
  # plotSPC(z, color = 'p1', divide.hz = FALSE, name = NA, width = 0.2)
  
  
  # result is an SPC
  expect_true(inherits(z, 'SoilProfileCollection'))
  
  # conservation of group IDs
  expect_true(all(profile_id(z) == c('a', 'b')))
  
})

#' @title Interpret a color signature containing color groups of CIELAB coordinates using perceptual distance via CIE dE00
#' 
#' @param pig `data.frame` results from `soilColorSignature(..., method = c('pam', 'depthSlices))`
#' @return dist object
.colorSig2PerceptualDistMat <- function(pig) {
  
  # first column contains profile IDs
  .ids <- pig[, 1]
  row.names(pig) <- .ids
  
  # remaining columns are color groups of L,A,B coordinates
  .n <- ncol(pig) - 1
  .ngroup <- .n / 3
  
  # iterate over color groups, result is a distance matrix (CIE dE00)
  dE00 <- lapply(1:.ngroup, function(i) {
    
    # LAB coordinates are named L.i, A.i, B.i, ...
    v.names <- paste(c('L', 'A', 'B'), i, sep = '.')
    
    # pair-wise delta-E00
    # IDs are preserved via row.names in pig
    # more efficient than compare_colour(x, x, ...)
    # output is transposed relative to `dist` object
    d.i <- farver::compare_colour(
      pig[, v.names], 
      from_space = 'lab', 
      white_from = 'D65', 
      method = 'cie2000'
    )
    
    # convert to dist object, note transpose
    d.i <- as.dist(t(d.i))
    
    return(d.i)
  })
  
  # sum distance matrices over color groups
  d <- Reduce('+', dE00)
  
  return(d)
}


## TODO: profile this function, it is much slower than depth slices
# compute LAB coordinates from clusters of slices
.pigments.pam <- function(x, k) {
  
  # extract just horizons that have color data
  h <- horizons(x)
  idx <- which(!is.na(h$L))
  h <- h[idx, ]
  
  # can't go on without some data
  ## TODO: this will cause an error in the calling function if all results are NULL
  if(nrow(h) == 0) {
    return(NULL)
  }
  
  
  # cluster colors using frequency weights
  # https://stat.ethz.ch/pipermail/r-help/2008-February/153501.html
  hd <- horizonDepths(x)
  top <- min(h[[hd[1]]], na.rm = TRUE)
  bottom <- max(h[[hd[2]]], na.rm = TRUE)
  
  # slice to unwind weighted data into un-weighted data (weights are thickness)
  fm <- as.formula(paste0(top, ':', bottom, ' ~ L + A + B'))
  x.slices <- suppressMessages(dice(x, fm = fm, SPC = FALSE))
  x.slices <- na.omit(x.slices[, c(idname(x), 'L', 'A', 'B')])
  
  # have to have at least 1 more row than k
  ## TODO: this will cause an error in the calling function if all results are NULL
  if(nrow(x.slices) < k + 1) {
    return(NULL)
  }
  
  # compute perceptually based distance matrix on sliced colors, CIELAB
  # IDs are preserved via row.names in pig
  # more efficient than compare_colour(x, x, ...)
  # output is transposed relative to `dist` object
  dE00 <- farver::compare_colour(
    x.slices[, -1], 
    from_space = 'lab',
    method = 'CIE2000', 
    white_from = 'D65'
  )
  
  # convert to dist object
  # note transpose
  dE00 <- as.dist(
    t(dE00)
  )
  
  # partitioning around medoids
  # optimization notes: 
  #       * pamonce = 5, 6 will hang on some soils (AMELAR OSD)
  #       * pamonce = 3 seems more stable
  cl <- cluster::pam(dE00, k = k, diss = TRUE, pamonce = 3)
  
  # subset medoids
  x.medoids <- x.slices[cl$id.med, c(idname(x), 'L', 'A', 'B')]
  
  # 2025-12-18
  # medoid order depends on source data
  # approximate a standardization scheme
  # by ordering medoids along L, A, B axes
  x.medoids <- x.medoids[order(x.medoids$L, x.medoids$A, x.medoids$B), ]
  
  # make IDs
  x.medoids$.ids <- paste0('.', 1:k)
  
  # convert wide -> long and create new variable names
  # using data.table::melt()
  m <- data.table::melt(
    data.table::as.data.table(x.medoids), 
    id.var = c(idname(x), '.ids'), 
    measure.vars = c('L', 'A', 'B')
  )
  
  # leave as data.table for dcast()
  
  # new ID
  m$variable <- paste0(m$variable, m$.ids)
  
  # convert long -> wide format
  # using data.table::dcast
  fm <- as.formula(paste0(idname(x), ' ~ variable'))
  res <- dcast(m, formula = fm, value.var = 'value')
  
  # convert back to data.frame
  res <- as.data.frame(res)
  
  # result is a data.frame with profile ID
  return(res)
}


# compute LAB coordinates at select percentiles of depth
.pigments.depths <- function(x, p) {
  
  # index for naming columns in the results
  .seq <- seq_along(p)
  
  # extract just horizons that have color data
  h <- horizons(x)
  idx <- which(!is.na(h$L))
  h <- h[idx, ]
  
  # TODO finish this
  if(nrow(h) == 0) {
    return(NULL)
  }
  
  # determine sampling depths via quantiles
  hd <- horizonDepths(x)
  # get approximate horizon mid-points
  h$.mid <- round(((h[[hd[2]]] + h[[hd[1]]]) / 2))
  
  # use discontinuous style quantiles, to force quantiles to "fit" original data
  sample.depths <- round(quantile(h$.mid, probs = p, na.rm = TRUE, type = 1))
  
  # locate the horizons that are as close as possible to these depths
  sample.idx <- sapply(sample.depths, function(i) which.min(abs(i - h$.mid)))
  
  # get data
  x.slices <- h[sample.idx, c(idname(x), 'L', 'A', 'B')]
  
  # make depth IDs
  x.slices$depth.id <- paste0('.', p)
  
  # convert wide -> long and create new variable names
  # using data.table::melt()
  m <- data.table::melt(
    data.table::as.data.table(x.slices),
    id.var = c(idname(x), 'depth.id'), 
    measure.vars = c('L', 'A', 'B')
  )
  
  # leave as data.table for re-shape
  
  # new ID
  m$variable <- sprintf("%s.%s", m$variable, .seq)
  
  # convert long -> wide format
  # using data.table::dcast
  fm <- as.formula(paste0(idname(x), ' ~ variable'))
  res <- dcast(m, formula = fm, value.var = 'value')
  
  # convert back to data.frame
  res <- as.data.frame(res)
  
  # percentiles used
  attr(res, 'p') <- p
  
  # result is a data.frame with profile ID
  return(res)
}



# CIELAB axes interpretation
# L - brightness
# A - green | red
# B - blue | yellow
# x: single profile
# requires L, pos.A, neg.A, pos.B, neg.B in @horizon
.pigments.proportions <- function(x, useProportions, pigmentNames) {
  
  h <- horizons(x)
  dc <- horizonDepths(x)
  hz.thick <- h[[dc[2]]] - h[[dc[1]]]
  h <- h[, c('L', 'pos.A', 'neg.A', 'pos.B', 'neg.B')]
  
  # TODO: this may need to be normalized
  # TODO: this will create NA
  hz.pigments <- sweep(h, MARGIN = 1, STATS = hz.thick, FUN = '*')
  pigment <- colSums(hz.pigments, na.rm = TRUE)
  names(pigment) <- pigmentNames
  
  # convert to proportions: remove the effect of soil depth
  if(useProportions) {
    pigment <- pigment / sum(pigment, na.rm = TRUE)
  }
  
  # results
  res <- data.frame(
    .id = profile_id(x)[1],
    t(pigment),
    stringsAsFactors = FALSE
  )
  
  # reset ID
  names(res)[1] <- idname(x)
  
  return(res)
}

## TODO: 
#   * data.table optimization
#   * demonstrate convergence with NCSP()

#' @title Soil Profile Color Signatures
#' @description Generate a color signature for each soil profile in a collection.
#'
#' @param spc a `SoilProfileCollection` object
#' 
#' @param color horizon-level attributes, either character of length 1 specifying a column containing Munsell or sRGB in hex notation, or character vector of three column names containing either sRGB or CIELAB color coordinates. sRGB color coordinates should be within the range of 0 to 1.
#' @param space character, either 'sRGB' or 'LAB', specifying color space

#' @param method algorithm used to compute color signature, `colorBucket`, `depthSlices`, or `pam`
#' @param perceptualDistMat logical, optionally return a distance matrix based on perceptual color distances, when ``method` is one of 'depthSlices' or 'pam', see Details

#' @param prob numeric vector, requested percentiles for `method = 'depthSlices'`
#' @param pam.k number of color classes for `method = 'pam'`

#' @param useProportions use proportions or quantities, see details
#' @param pigmentNames names for resulting pigment proportions or quantities
#' @param apply.fun function passed to `aqp::profileApply(APPLY.FUN)` argument, can be used to add progress bars via `pbapply::pblapply()`, or parallel processing with `furrr::future_map()` 
#' 
#' @param r deprecated, use `color` argument
#' @param g deprecated, use `color` argument
#' @param b deprecated, use `color` argument
#' @param RescaleLightnessBy deprecated, scaling factor for CIELAB L-coordinate
#'
#'
#' @details 
#' 
#' Missing colors (NA) are ignored in the creation of soil color signatures.
#' 
#' 
#' 
#' Interpreation of color signature.
#' 
#' Choices related to weighting, scaling, and distance metric.
#' 
#' Perceptual distances (dE00), summed over color groups.
#' 
#' See the [related tutorial](http://ncss-tech.github.io/AQP/aqp/soil-color-signatures.html).
#' 
#' @return
#' 
#' For the `colorBucket` method, a `data.frame`:
#' 
#'  * id column: set according to `idname(spc)`
#'  * `.white.pigment`: proportion or quantity of CIELAB L-values
#'  * `.red.pigment`: proportion or quantity of CIELAB positive A-values
#'  * `.green.pigment`: proportion or quantity of CIELAB negative A-values
#'  * `.yellow.pigment`: proportion or quantity of CIELAB positive B-values
#'  * `.blue.pigment`: proportion or quantity of CIELAB negative B-values
#' 
#' Column names can be adjusted with the `pigmentNames` argument.
#' 
#' 
#' For the `depthSlices` method, a `data.frame`:
#' 
#'  * id column: set according to `idname(spc)`
#'  * `L.1`, `A.1`, `B.1`: CIELAB color coordinates associated with the first depth slice, at depth percentile given in `prob[1]`
#'  * ...
#'  * `L.n`, `A.n`, `B.n`: CIELAB color coordinates associated with the `n` depth slice, at depth percentile given in `prob[n]`
#' 
#' 
#' For the `pam` method, a `data.frame`:
#' 
#'  * id column: set according to `idname(spc)`
#'  * `L.1`, `A.1`, `B.1`: CIELAB color coordinates associated with the first color cluster, after sorting all clusters in ascending order along L, A, B axes.
#'  * ...
#'  * `L.n`, `A.n`, `B.n`: CIELAB color coordinates associated with the `nth` color cluster, after sorting all clusters in ascending order along L, A, B axes.
#'  
#'  When `perceptualDistMat = TRUE` and `method` is one of 'depthSlices' or 'pam', a distance matrix is returned.
#'  
#' 
#' @references https://en.wikipedia.org/wiki/Lab_color_space
#' 
#' @author D.E. Beaudette
#' 
#' @seealso [plotProfileDendrogram()]
#' 
#' 
#' @export
#'
#' @examples
#' 
#' # trivial example, not very interesting
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#' 
#' # Munsell notation
#' sp1$m <- sprintf("%s %s/%s", sp1$hue, sp1$value, sp1$chroma)
#' 
#' # extract color signature
#' pig <- soilColorSignature(sp1, color = 'm')
#' 
soilColorSignature <- function(
    spc, 
    color,
    space = c('sRGB', 'CIELAB'), 
    method = c('colorBucket', 'depthSlices', 'pam'), 
    perceptualDistMat = FALSE,
    pam.k = 3, 
    prob = c(0.1, 0.5, 0.9),
    useProportions = TRUE, 
    pigmentNames = c('.white.pigment', '.red.pigment', '.green.pigment', '.yellow.pigment', '.blue.pigment'), 
    apply.fun = lapply,
    r = NULL, 
    g = NULL, 
    b = NULL, 
    RescaleLightnessBy = NULL
) {
  
  # 2025-12-16: deprecated arguments r, g, b
  if(!is.null(r) || !is.null(g) || !is.null(b)) {
    warning('arguments `r`, `g`, and `b` have been deprecated, please use the `color` argument instead')
    
    # temporarily cobble together the new specification
    color <- c(r, g, b)
  }
  
  # 2025-12-18: deprecated argument RescaleLightnessBy
  if(!is.null(RescaleLightnessBy)) {
    warning('argument `RescaleLightnessBy` has been deprecated, consider weighting or standardizing color signature matrix before computing distances')
  }
  
  
  # sanity checks fixed choice arguments
  method <- match.arg(method)
  space <- match.arg(space)
  
  ## TODO: think about how to avoid making another copy of hz data
  ##       iterate over these data vs. SPC
  # extract horizons
  h <- horizons(spc)
  
  # detect color specification
  # complicated because color could be length 1 or 3
  if(length(color) == 1) {
    # single column
    .spec <- .detectColorSpec(h[[color]])
  } else if(length(color) == 3) {
    # three columns
    .spec <- .detectColorSpec(h[, color])
  } else{
    # error condition
    stop('`color` should be of length 1 or 3')
  }
  
  
  # conditionally convert colors to CIELAB
  # specific NA handling required
  lab.colors <- switch(
    .spec,
    
    # hex encoded sRGB color coordinates
    `hex-sRGB` = {
      # NOTES: 
      #  * rescale to [0,1]
      #  * NA -> LAB [100, 0, 0] 
      
      # keep track of NA
      na.idx <- which(is.na(h[[color]]))
      
      # convert
      .conv <- convertColor(
        t(col2rgb(h[[color]]) / 255), 
        from = 'sRGB', 
        to = 'Lab', 
        from.ref.white = 'D65', 
        to.ref.white = 'D65'
      )
      
      # pad with NA if present
      if(length(na.idx > 0)) {
        .conv[na.idx, ] <- cbind(NA_real_, NA_real_, NA_real_)
      }
      
      .conv
      
    },
    
    # 3 columns subset from a data.frame
    `color-coordinate-data.frame` = {
      .m <- as.matrix(h[, color])
      
      # convert to CIELAB
      # data.frame -> matrix
      if(space == 'sRGB') {
        
        # check for mis-specification of color space
        if(any(range(.m, na.rm = TRUE) > 5)) {
          stop('color space coordinates do not appear to be sRGB [0,1]: check `space` argument')
        }
        
        # NA are left in place
        .res <- convertColor(.m, from = 'sRGB', to = 'Lab', from.ref.white = 'D65', to.ref.white = 'D65')
      }
      
      # no colorspace conversion
      if(space == 'CIELAB') {
        .res <- .m
      }
      
      .res
    },
    
    # plain Munsell notation
    `munsell` = {
      # NA are left in place
      parseMunsell(h[[color]], returnLAB = TRUE)
    },
    
    # all others
    stop('unknown color specifcation')
  )
  
  
  # choose method
  if(method == 'colorBucket') {
    ## L is always positive
    ## split A/B axes into positive / negative pigments
    pos.A <- ifelse(lab.colors[, 2] > 0, lab.colors[, 2], 0)
    neg.A <- ifelse(lab.colors[, 2] < 0, -lab.colors[, 2], 0)
    
    pos.B <- ifelse(lab.colors[, 3] > 0, lab.colors[, 3], 0)
    neg.B <- ifelse(lab.colors[, 3] < 0, -lab.colors[, 3], 0)
    
    ## TODO: this is sloppy
    # assign back to original SPC
    spc$L <- lab.colors[, 1]
    spc$pos.A <- pos.A
    spc$neg.A <- neg.A
    spc$pos.B <- pos.B
    spc$neg.B <- neg.B
    
    # result is a list
    col.data <- profileApply(
      spc, 
      FUN = .pigments.proportions, 
      useProportions = useProportions, 
      pigmentNames = pigmentNames, 
      simplify = FALSE,
      APPLY.FUN = apply.fun
    )
    
  }
  
  # use horizons at depths proportional to percentiles: 0.1, 0.5, 0.9
  if(method == 'depthSlices') {
    spc$L <- lab.colors[, 1]
    spc$A <- lab.colors[, 2]
    spc$B <- lab.colors[, 3]
    
    ## TODO: optimize
    # result is a list of data.frames
    col.data <- profileApply(
      spc, 
      FUN = .pigments.depths, 
      p = prob,
      simplify = FALSE,
      APPLY.FUN = apply.fun
    )
    
  }
  
  # PAM
  if(method == 'pam') {
    spc$L <- lab.colors[, 1]
    spc$A <- lab.colors[, 2]
    spc$B <- lab.colors[, 3]
    
    ## TODO: optimize
    # result is a list
    col.data <- profileApply(
      spc, 
      FUN = .pigments.pam, 
      k = pam.k, 
      simplify = FALSE,
      APPLY.FUN = apply.fun
    )
    
  }
  
  # back to data.frame
  col.data <- do.call('rbind', col.data)
  
  # reset rownames
  row.names(col.data) <- NULL
  
  # set attributes as metadata
  attr(col.data, 'colorspec') <- .spec
  
  # optionally convert color signature -> perceptual distance matrix
  if(perceptualDistMat) {
    return(.colorSig2PerceptualDistMat(col.data))
  } else {
    # just the signature
    return(col.data)
  }
  
}

# generate depth axis labels
# m: max depth
# i: interval
.depthAxisSeq <- function(m, i = NULL) {
  
  # reasonable heuristics if an interval is not specified
  if(is.null(i)) {
    if(m > 800) {
      i <- 100
    } else if(m <= 800 & m > 200) {
      i <- 50
    } else if(m <= 200 & m > 100) {
      i <- 25
    } else if(m <= 100 & m > 50) {
      i <- 10
    } else if(m <= 50) {
      i <- 5
    }
  }
  
  # convenient rounding-to function
  # e.g. round 73, to nearest 10s place
  # thanks SO: https://stackoverflow.com/a/32508105
  .roundTo <- function(x, y) {
    if((y - x %% y) <= x %% y) { x + (y - x %% y)}
    else { x - (x %% y)}
  }
  
  # clamp axis max via customized rounding
  .max <- .roundTo(m, i)
  
  # axis sequence
  .res <- seq(0, .max, by = i)
  return(.res)
}



#' @title Draw a depth axis along side soil profile sketches
#'
#' @param style character, depth axis style
#' @param .at numeric vector, depth axis annotation coordinates
#' @param .labels character vector, tick mark labels
#' @param .line numeric, line offset for horizontal placement of axis
#' @param .cex numeric, scaling factor for axis
#' 
#' @noRd
#' @return nothing, low-level plotting function
.drawDepthAxis <- function(style = c('compact', 'traditional', 'tape'), .at, .labels, .line, .cex) {
  
  # launder / check style
  style <- tolower(style)
  style <- match.arg(style)
  
  switch(style,
         'compact' = {
           # minimal axis
           axis(
             side = 4, 
             col = NA, 
             col.axis = par('fg'), 
             col.ticks = par('fg'), 
             las = 1, 
             lend = 3, 
             mgp = c(3, 0.25, 0),
             font = 1, lwd.ticks = 1, tck = 0.02,
             line = .line, 
             at = .at, 
             labels = .labels, 
             cex.axis = .cex
           ) 
         },
         
         'traditional' = {
           # traditional axis
           axis(
             side = 4, 
             col = par('fg'), 
             col.axis = par('fg'), 
             col.ticks = par('fg'), 
             las = 1, 
             lend = 3, 
             line = .line, 
             at = .at, 
             labels = .labels, 
             cex.axis = .cex
           )
         },
         
         'tape' = {
           # something like the fabric, graduated tapes
           # used by NCSS and elsewhere
           
           # alternating colors
           .cols <- c('white', grey(0.3))
           
           # remove depth units by splitting on space
           .labels <- sapply(strsplit(.labels, split = ' ', fixed = TRUE), function(i) i[1])
           
           # width of tape based on widest depth annotation
           .w <- strwidth(as.character(max(.labels)), units = 'user', cex = .cex) * 0.9
           
           ## not currently used, "line" depends on aspect ratio of figure
           # convert line notation to user coordinates
           # .dx <- grconvertX(.line, from = 'line', to = 'user')
           
           ## TODO: figure out line-style adjustment that ignores figure aspect ratio
           # shift using max x value in user coordinates of figure
           .x <- par('usr')[2] - (2 * .w)
           
           # number of depth labels
           .nat <- length(.at)
           
           # temporarily disable figure clipping
           # this is important because rect() adds to the figure area vs. margin by axis()
           .op <- par(xpd = NA)
           
           # alternating colors of tape
           rect(
             xleft = .x - .w,
             xright = .x + .w, 
             ybottom = .at[-1], 
             ytop = .at[-.nat], 
             col = .cols, 
             border = NA
           )
           
           # outline
           rect(
             xleft = .x - .w, 
             xright = .x + .w, 
             ybottom = .at[.nat], 
             ytop = .at[1],
             col = NA,
             border = par('fg')
           )
           
           # consider computing an exact coordinate, vs pos / offset
           # use at - (text height)
           # .labY <- .at[-1] - (diff(.at) / 2)
           .labY <- .at
           
           # label tape sections
           text(
             x = .x, 
             y = .labY, 
             labels = .labels, 
             cex = .cex, 
             col = invertLabelColor(.cols), 
             offset = 0.25,
             pos = 1
           )
           
           # return to original graphics state
           par(.op)

         }
  )
  
  # done
  
}


# draft replacement for scales::col_factor (without looking at it)
.aqp_color_map <- function(palette, domain, na.color, ordered) {
  .FUN <- function(x) {
    if (is.numeric(x)) {
      cuts <- c(gsub("^\\[(.*),.*", "\\1", domain[1]),
                gsub(".*,(.*)\\]$", "\\1", domain))
      x <- cut(x, breaks = cuts)
    } else {
      x <- factor(x, levels = domain, ordered = ordered)
    }
    idx <- as.integer(x)
    res <- palette[idx]
    res[is.na(res)] <- na.color
    res
  }
}

## TODO: switch to .detectColorSpec()
## TODO: allow direct interpretation of Munsell notation

# generalize and make into an exported function
.interpretHorizonColor <- function(h, color, default.color, col.palette, col.palette.bias, n.legend) {
  
  # this is optionally replaced with real data when using thematic colors
  color.legend.data <- NULL
  
  # toggle as needed for more room
  multi.row.legend <- FALSE
  # multi-row legend indices
  leg.row.indices <- NULL
  
  ## TODO: manually setting color=NULL will throw an error
  # think about how to best handle this
  # if(is.null(color)) {
  #
  # }
  
  # short-circuit: if all h[[color]] are NA the following logic will not reliably work
  # this is because sometimes all NA are interpreted as logical vectors
  if (all(is.na(h[[color]]))) {
    h[[".color"]] <- rep(NA_character_, nrow(h))
  } else {
    
    # there is at least 1 non-NA color to work with
    
    # 1. numeric vector, re-scale and apply color ramp
    if(is.numeric(h[[color]])) {
      
      # generate color ramp function
      cr <- grDevices::colorRamp(col.palette, bias = col.palette.bias, space = 'Lab', interpolate = 'spline')
      
      # re-scale to [0,1]
      # may contain NAs
      c.rgb <- cr(.rescaleRange(h[[color]], x0 = 0, x1 = 1))
      cc <- which(complete.cases(c.rgb))
      h$.color <- NA
      
      # convert non-NA values into colors
      h$.color[cc] <- rgb(c.rgb[cc, , drop = FALSE], maxColorValue = 255)
      
      # generate range / colors for legend
      pretty.vals <- pretty(h[[color]], n = n.legend)
      
      ## TODO: think about a smarter way to do this
      ##       -> sometimes the colors used in the legend are misleading
      
      # constrain legend to min/max
      # pretty.vals[1] <- min(h[[color]], na.rm = TRUE)
      # pretty.vals[length(pretty.vals)] <- max(h[[color]], na.rm = TRUE)
      
      # truncate to 3 significant digits and convert to character for correct interpretation of floating point values
      leg.pretty.vals <- as.character(signif(pretty.vals, 3))
      
      # special case: there are < 3 unique values -> convert to factor
      # previous calculations are ignored
      low.n.test.vals <- as.character(signif(h[[color]], digits = 3))
      if(length(unique(na.omit(low.n.test.vals))) < 3) {
        # replace with character representation with 3 significant digits
        h[[color]] <- low.n.test.vals
        message('less than 3 unique values, converting to factor')
      }
      
      # put into a list for later
      color.legend.data <- list(
        legend = leg.pretty.vals,
        col = rgb(cr(.rescaleRange(pretty.vals, x0 = 0, x1 = 1)), maxColorValue=255),
        multi.row.legend = multi.row.legend,
        leg.row.indices = leg.row.indices
      )
    }
    
    # 2. vector of categorical data
    if(is.character(h[[color]]) | is.factor(h[[color]])) {
      
      # testing if ALL valid colors
      if( all(.isColorValid(na.omit(h[[color]])))) {
        # YES: interpret values directly as colors
        h$.color <- h[[color]]
      } else {
        # NO: this is or can be converted into a factor
        if(!is.factor(h[[color]]))
          h[[color]] <- factor(h[[color]])
        
        # get color mapping levels after dropping missing levels
        h[[color]] <- droplevels(h[[color]])
        color.levels <- levels(h[[color]])
        
        crp <- grDevices::colorRampPalette(col.palette, bias = col.palette.bias, space = 'Lab', interpolate = 'spline')
        
        # make a color mapping function
        if (requireNamespace("scales", quietly = TRUE)) {
          ## TODO: replace with native function
          FUN <- scales::col_factor
        } else {
          FUN <- .aqp_color_map
        }
        
        color.mapper <- FUN(
          palette = crp(length(color.levels)),
          domain = color.levels,
          na.color = default.color,
          ordered = TRUE
        )
        
        # apply color mapping
        h$.color <- color.mapper(h[[color]])
        
        # generate colors and labels for legend
        pretty.vals <- color.levels
        
        # interpret n.legend as max(items) / row
        n.leg.classes <- length(pretty.vals)
        
        # create more room via multiple calls to legend
        if(n.legend < n.leg.classes) {
          
          # make indices to two rows of legends
          # safely accounts for even / odd n.leg.classes
          leg.row.indices <- .splitLegend(n.leg.classes)
          
          # set flag for later
          multi.row.legend <- TRUE
        }
        
        # pack into a list for later use
        color.legend.data <- list(
          legend = pretty.vals,
          col = color.mapper(pretty.vals),
          multi.row.legend = multi.row.legend,
          leg.row.indices = leg.row.indices
        )
        
      }
    }
    
  }
  
  
  # if the color column doesn't exist, fill with NA
  if(is.null(h[[color]]))
    h[[".color"]] <- NA
  
  # fill missing colors with a reasonable default
  h[['.color']] <- ifelse(is.na(h[['.color']]), default.color, h[['.color']])
  
  # assemble results
  res <- list(
    colors = h[['.color']],
    color.legend.data = color.legend.data
  )
  
  return(res)
}


# split legend into two rows, and create indices
# any more classes than that and things become impossible to read
# n: total number of classes
.splitLegend <- function(n) {
  
  #  make enough room for even division of odd numbers
  n.per.row <- ceiling(n / 2)
  
  # make indices for first row
  row.1.idx <- seq(from=1, to=n.per.row)
  row.2.idx <- seq(from=n.per.row + 1, to=n)
  
  res <- list(
    row.1=row.1.idx,
    row.2=row.2.idx
  )
  
  return(res)
}


## TODO: switch to .detectColorSpec()
# Test for valid colors in vector `x`:
#   1. named colors from colors()
#   2. RGB / RGBA encoding of colors
.isColorValid <- function(x) {
  # check for named colors
  test.1 <- x %in% colors()
  
  # check for valid RGB
  test.2 <- grepl('^#[a-f0-9]{6}', x, ignore.case = TRUE)
  
  # check for valid RGBA colors
  test.3 <- grepl('^#[a-f0-9]{8}', x, ignore.case = TRUE)
  
  # must pass at least 1 test
  res <- test.1 | test.2 | test.3
  return(res)
}


# x1: left-side x coordinate
# x2: right-side x coordinate
# y:  common y coordinate
# n: number of vertices
# o: vertical offsets (must be pre-scaled by plotSPC's scaling.factor argument)
# f: jittering factor
.raggedLines <- function(x1, x2, y, n = 16, o = c(-0.25, 0.75), f = 1) {
  
  # initial vertices, includes left and right points
  .x <- seq(from = x1, to = x2, length.out = n)
  .y <- rep(y, times = n)
  
  # add offsets + jitter to all but outermost vertices
  .idx <- 2:(n-1)
  .y[.idx] <- jitter(.y[.idx] + o, factor = f)
  
  # TODO: convert to step function, would look better on screen
  
  # final coordinates
  return(cbind(.x, .y))
}

# profileApply

# if (!isGeneric("profileApply"))
 	setGeneric("profileApply", function(object, FUN,
 	                                    simplify = TRUE,
 	                                    frameify = FALSE,
 	                                    chunk.size = 100,
 	                                    column.names = NULL,
 	                                    APPLY.FUN = lapply,
 	                                    ...) standardGeneric("profileApply"))
# made internal by AGB 2020/01/19
#
# analog to apply():
# operates by profile, returns vector of length(object) or nrow(object)
.profileApply <- function(object, FUN, simplify=TRUE, ...) {

		# get profile IDs
		pIDs <- profile_id(object)

		# iterate over SPC, spliting into a list of SPC_i ... SPC_n
		l <- lapply(seq_along(pIDs), function(i) do.call(FUN, list(object[i, ], ...)))
    names(l) <- pIDs

		# optionally simplify
		if (simplify)
			return(unlist(l))

		return(l)
}
#' Iterate over profiles in a SoilProfileCollection
#'
#' @name profileApply
#'
#' @description Iterate over all profiles in a SoilProfileCollection, calling \code{FUN} on a single-profile SoilProfileCollection for each step.
#'
#' @param object a SoilProfileCollection
#'
#' @param FUN a function to be applied to each profile within the collection
#'
#' @param simplify logical, should the result be simplified to a vector? default: TRUE; see examples
#'
#' @param frameify logical, should the result be collapsed into a data.frame? default: FALSE; overrides simplify argument; see examples
#'
#' @param chunk.size numeric, size of "chunks" for faster processing of large SoilProfileCollection objects; default: 100
#'
#' @param column.names character, optional character vector to replace frameify-derived column names; should match length of colnames() from FUN result; default: NULL
#'
#' @param ... additional arguments passed to FUN
#' 
#' @param APPLY.FUN function, optional alternate `lapply()`-like function for processing chunks. For example `future.apply::future_lapply()` for processing chunks in parallel. Default `base::lapply()`.
#' 
#' @return When simplify is TRUE, a vector of length nrow(object) (horizon data) or of length length(object) (site data). When simplify is FALSE, a list is returned. When frameify is TRUE, a data.frame is returned. An attempt is made to identify idname and/or hzidname in the data.frame result, safely ensuring that IDs are preserved to facilitate merging profileApply result downstream.
#'
#' @aliases profileApply,SoilProfileCollection-method
#' @docType methods
#' @rdname profileApply
#' @export
#' @examples
#'
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#'
#' # estimate soil depth using horizon designations
#' profileApply(sp1, estimateSoilDepth, name='name')
#'
#' # scale a single property 'prop' in horizon table
#' # scaled = (x - mean(x)) / sd(x)
#' sp1$d <- profileApply(sp1, FUN=function(x) round(scale(x$prop), 2))
#' plot(sp1, name='d')
#'
#' # compute depth-wise differencing by profile
#' # note that our function expects that the column 'prop' exists
#' f <- function(x) { c(x$prop[1], diff(x$prop)) }
#' sp1$d <- profileApply(sp1, FUN=f)
#' plot(sp1, name='d')
#'
#' # compute depth-wise cumulative sum by profile
#' # note the use of an anonymous function
#' sp1$d <- profileApply(sp1, FUN=function(x) cumsum(x$prop))
#' plot(sp1, name='d')
#'
#' # compute profile-means, and save to @site
#' # there must be some data in @site for this to work
#' site(sp1) <- ~ group
#' sp1$mean_prop <- profileApply(sp1, FUN=function(x) mean(x$prop, na.rm=TRUE))
#'
#' # re-plot using ranks defined by computed summaries (in @site)
#' plot(sp1, plot.order=rank(sp1$mean_prop))
#'
#' ## iterate over profiles, calculate on each horizon, merge into original SPC
#'
#' # example data
#' data(sp1)
#'
#' # promote to SoilProfileCollection
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' # calculate horizon thickness and proportional thickness
#' # returns a data.frame result with multiple attributes per horizon
#' thicknessFunction <- function(p) {
#'   hz <- horizons(p)
#'   depthnames <- horizonDepths(p)
#'   res <- data.frame(profile_id(p), hzID(p),
#'                     thk=(hz[[depthnames[[2]]]] - hz[[depthnames[1]]]))
#'   res$hz_prop <- res$thk / sum(res$thk)
#'   colnames(res) <- c(idname(p), hzidname(p), 'hz_thickness', 'hz_prop')
#'   return(res)
#' }
#'
#' # list output option with simplify=F, list names are profile_id(sp1)
#' list.output <- profileApply(sp1, thicknessFunction, simplify = FALSE)
#' head(list.output)
#'
#' # data.frame output option with frameify=TRUE
#' df.output <- profileApply(sp1, thicknessFunction, frameify = TRUE)
#' head(df.output)
#'
#' # since df.output contains idname(sp1) and hzidname(sp1),
#' # it can safely be merged by a left-join via horizons<- setter
#' horizons(sp1) <- df.output
#'
#' plot(density(sp1$hz_thickness, na.rm=TRUE), main="Density plot of Horizon Thickness")
#'
#' ## iterate over profiles, subsetting horizon data
#'
#' # example data
#' data(sp1)
#'
#' # promote to SoilProfileCollection
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' # make some fake site data related to a depth of some importance
#' sp1$dep <- profileApply(sp1, function(i) {round(rnorm(n=1, mean=mean(i$top)))})
#'
#' # custom function for subsetting horizon data, by profile
#' # keep horizons with lower boundary < site-level attribute 'dep'
#' fun <- function(i) {
#'   # extract horizons
#'   h <- horizons(i)
#'   # make an expression to subset horizons
#'   exp <- paste('bottom < ', i$dep, sep='')
#'   # subset horizons, and write-back into current SPC
#'   slot(i, 'horizons') <- subset(h, subset=eval(parse(text=exp)))
#'   # return modified SPC
#'   return(i)
#' }
#'
#' # list of modified SoilProfileCollection objects
#' l <- profileApply(sp1, fun, simplify=FALSE)
#'
#' # re-combine list of SoilProfileCollection objects into a single SoilProfileCollection
#' sp1.sub <- pbindlist(l)
#'
#' # graphically check
#' par(mfrow=c(2,1), mar=c(0,0,1,0))
#' plot(sp1)
#' points(1:length(sp1), sp1$dep, col='red', pch=7)
#' plot(sp1.sub)
#'
setMethod('profileApply', signature(object = 'SoilProfileCollection'), function(object, FUN,
                        simplify = TRUE,
                        frameify = FALSE,
                        chunk.size = 100,
                        column.names = NULL,
                        APPLY.FUN = lapply,
                        ...) {
  if (simplify & frameify) {
    # simplify and frameify are both TRUE -- ignore simplify argument
    simplify <- FALSE
  }

  # total number of profiles we have to iterate over
  n <- length(object)

  # split the SPC of size n into chunk.size chunks
  chunk <- sort(1:n %% max(1, round(n / chunk.size))) + 1

  # we first iterate over list of chunks of size chunk.size, keeping lists smallish
  # by dividing by a tunable factor -- set as 100 by default
  # then we iterate through each chunk, calling FUN on each element (profile)
  # then, concatenate the result into a list (or vector if simplify == TRUE)
  res <- do.call('c', APPLY.FUN(split(1:n, chunk), function(idx) {
    .profileApply(object[idx,], FUN, simplify, ...)
  }))

  # return profile IDs if it makes sense for result
  if (length(res) == length(object))
    names(res) <- profile_id(object)

  # combine a list (one element per profile) into data.frame result
  if (!simplify & frameify) {

    # make sure the first result is a data.frame (i.e. FUN returns a data.frame)
    if (inherits(res[[1]], 'data.frame')) {

      # make a big data.frame
      if (requireNamespace("data.table")) {
        res <- .as.data.frame.aqp(data.table::rbindlist(res, fill = TRUE), aqp_df_class(object))
      } else {
        res <- .as.data.frame.aqp(do.call('rbind', res), aqp_df_class(object))
      }

      if (!is.null(column.names))
        colnames(res) <- column.names

    } else {
      warning("first result does not inherit `data.frame` and frameify is TRUE. defaulting to list output.", call. = FALSE)
    }
  }

  if (simplify)
   return(unlist(res))

  return(res)
})

#' @title Check a SoilProfileCollection object for errors in horizon depths.
#'
#' @description This function inspects a `SoilProfileCollection` object, looking for four common errors in horizon depths:
#'
#'   1. bottom depth shallower than top depth
#'   2. equal top and bottom depth
#'   3. missing top or bottom depth (e.g. `NA`)
#'   4. gap or overlap between adjacent horizons (only if `byhz = FALSE`)
#'
#' @param x `SoilProfileCollection` or `data.frame` object to check
#' 
#' @param hzdepths character vector, describing top and bottom depths in a `SoilProfileCollection` or `data.frame`. `horizonDepths(x)` is used when `x` is a `SoilProfileCollection`.
#' 
#' @param idname character, describing the column containing profile IDs in a `SoilProfileCollection` or `data.frame`. `idname(x)` is used when `x` is a `SoilProfileCollection`.
#' 
#' @param fast logical, When `TRUE`, details about specific test results are not needed, the operation can allocate less memory and run approximately 5x faster.
#' 
#' @param byhz logical, apply logic tests to profiles (`FALSE`) or individual horizons (`TRUE`)?
#'
#' @return A `data.frame` containing profile IDs, validity boolean (`valid`) and test results if `fast = FALSE`.
#'
#' The `data.frame` will have as many rows as profiles in `x` (`length(x)`).
#'
#'  - `id` : Profile IDs, named according to `idname(x)`
#'  - `valid` : boolean, profile passes all of the following tests
#'  - `depthLogic` : boolean, errors related to depth logic
#'  - `sameDepth` : boolean, errors related to same top/bottom depths
#'  - `missingDepth` : boolean, NA in top / bottom depths
#'  - `overlapOrGap` : boolean, gaps or overlap in adjacent horizons (`NA` when `byhz = TRUE`)
#'
#' @export
#' @author D.E. Beaudette, A.G. Brown, S.M. Roecker
#' @examples
#'
#' ## sample data
#'
#' data(sp3)
#' depths(sp3) <- id ~ top + bottom
#'
#' # these data should be clean
#' res <- checkHzDepthLogic(sp3)
#'
#' head(res)
#'
#' # less memory if only concerned about net validity
#' res <- checkHzDepthLogic(sp3, fast = TRUE)
#'
#' head(res)
#'
checkHzDepthLogic <- function(x,
                              hzdepths = NULL,
                              idname = NULL,
                              fast = FALSE,
                              byhz = FALSE) {

  stopifnot(inherits(x, 'SoilProfileCollection') |
              inherits(x, 'data.frame'))

  if (inherits(x, 'SoilProfileCollection')) {
    h <- data.table::as.data.table(horizons(x))
    hzd <- horizonDepths(x)
    idn <- idname(x)
  } else {
    h <- data.table::as.data.table(x)

    # must have horizon top and bottom depth column
    stopifnot(length(hzdepths) == 2 &
                is.character(hzdepths) &
                all(hzdepths %in% colnames(h)))
    hzd <- hzdepths

    # must have id name column
    stopifnot(length(idname) == 1 &
                is.character(idname) &
                all(idname %in% colnames(h)))
    idn <- idname
  }

  hby <- substitute(idn)
  res <- NULL

  # data.table R CMD check
  tests <- NULL

  if (!fast) {
    if (!byhz) {
      res <- h[, list(tests = list(tests = data.frame(t(hzDepthTests(.SD))))), .SDcols = hzd, by = c(eval(hby))][,
                 list(tests = tests, valid = all(!tests[[1]])), by = c(eval(hby))]
    } else {
      res <- h[, list(tests = list(tests = data.frame(t(hzDepthTests(.SD))))), .SDcols = hzd, by = list(.hzID = seq_len(nrow(h)))][,
                 list(tests = tests, valid = all(!tests[[1]])), by = list(.hzID = seq_len(nrow(h)))]
      if (inherits(x, 'SoilProfileCollection')) {
        res[[hzidname(x)]] <- hzID(x)
      } else {
        res$hzID <- res$.hzID
      }
      res$.hzID <- NULL
    }
    res <- cbind(res, data.table::rbindlist(res$tests))
    res$tests <- NULL

  } else {

    if (!byhz) {
      res <- h[, all(!hzDepthTests(.SD)), .SDcols = hzd, by = c(eval(hby))]
      colnames(res) <- c(idn, "valid")
    } else {
      res <- h[, all(!hzDepthTests(.SD)), .SDcols = hzd, by = list(.hzID = seq_len(nrow(h)))]
      if (inherits(x, 'SoilProfileCollection')) {
        res[[hzidname(x)]] <- hzID(x)
        colnames(res) <- c(".hzID", "valid", hzidname(x))
      } else {
        res$hzID <- res$.hzID
        colnames(res) <- c(".hzID", "valid", "hzID")
      }
      res$.hzID <- NULL
    }
  }
  
  # add profile ID and top/bottom depth for byhz==TRUE
  if (byhz) {
    res <- cbind(h[, .SD, .SDcols = c(idn, hzd)], res)
    
    # fill overlapOrGap with NA as that test isn't possible on horizon-basis
    res[['overlapOrGap']] <- NA
  }
  
  return(as.data.frame(res))

}

#' @title Tests of horizon depth logic
#'
#' @description Function used internally by `checkHzDepthLogic()`, `glom()` and various other functions that operate on horizon data from single soil profiles and require a priori depth logic checks. Checks for bottom depths less than top depth / bad top depth order ("depthLogic"), bottom depths equal to top depth ("sameDepth"), overlaps/gaps ("overlapOrGap") and missing depths ("missingDepth"). Use `names(res)[res]` on result `res` of `hzDepthTest()` to to determine type of logic error(s) found -- see examples below.
#'
#' @param top A numeric vector containing horizon top depths. Or a `data.frame` with two columns (first containing top depths, second containing bottom)
#' @param bottom A numeric vector containing horizon bottom depths.
#'
#' @return A named logical vector containing TRUE for each type of horizon logic error found in the given data.
#' @author Andrew G. Brown & Dylan E. Beaudette
#' @examples
#'
#' # no logic errors
#' res <- hzDepthTests(top = c(0,10,20,30), bottom = c(10,20,30,50))
#' names(res)[res]
#'
#' # bottom < top
#' hzDepthTests(top = c(10,20,30,50), bottom = c(0,10,20,30))
#' names(res)[res]
#'
#' # bottom == top
#' hzDepthTests(top = c(10,20,30,50), bottom = c(0,20,20,30))
#' names(res)[res]
#'
#' # overlap
#' hzDepthTests(top = c(0,5,20,30), bottom = c(10,20,30,50))
#' names(res)[res]
#'
#' # gap
#' hzDepthTests(top = c(0,15,20,30), bottom = c(10,20,30,50))
#' names(res)[res]
#'
#' # missing
#' hzDepthTests(c(0,15,NA,30),c(10,NA,30,50))
#' names(res)[res]
#'
#' @rdname hzDepthTests
#' @export hzDepthTests
hzDepthTests <- function(top, bottom = NULL) {

  if(inherits(top, 'data.frame') && ncol(top) >= 2) {
    bottom <- top[[2]]
    top <- top[[1]]
  }

  top <- as.numeric(top)
  bottom <- as.numeric(bottom)

  stopifnot(is.numeric(top) && is.numeric(bottom))

  n <- length(top)

  # sanity checks, since this will be exported provide a little checking
  #   for most internal usesF these errors will never trigger...
  # but in case of corrupted hz data or bad inputs... anything can happen
  if (length(top) != length(bottom)) {
    stop("cannot evaluate horizon depth logic: vectors do not have same length")
  }

  # bottom depth < top depth? or horizons not in top-depth order?
  test.1 <- any(bottom < top, na.rm = TRUE) | any(suppressWarnings(sort(top) != top))

  if (is.na(test.1)) {
    # test.1 is NA if test.3 is true for both top and bottom depth
    test.1 <- TRUE
  }

  # bottom depth == top depth
  test.2 <- any(top == bottom, na.rm = TRUE)

  # NA depths
  test.3 <- any(is.na(top) | is.na(bottom), na.rm = TRUE)

  # bottom != next top
  test.4 <- any(bottom[-n] != top[-1], na.rm = TRUE)

  res <- as.logical(c(test.1, test.2, test.3, test.4))

  names(res) <- c("depthLogic","sameDepth","missingDepth","overlapOrGap")
  return(res)
}

test_that("accumulateDepths works", {
  
  # sp4 is a data.frame
  data(sp4)
  
  # example using hzdatum argument data(sp4) data.frame
  hz <- accumulateDepths(sp4, id = "id", hzdepths = c("top", "bottom"), hzname = "name", hzdatum = 15)
  depths(hz) <- id ~ top + bottom
  expect_equal(min(hz$top), 15)
  
  # use the SPC interface by promoting sp4
  depths(sp4) <- id ~ top + bottom
  
  hzdesgnname(sp4) <- "name"
  
  # and a custom hzdatum for each profile
  hz <- accumulateDepths(sp4, hzdatum = 5 * 1:length(sp4))
  
  # promote the result
  depths(hz) <- id ~ top + bottom
  
  # deepest horizon will be in second to last profile with a linearly increasing datum
  expect_true(max(hz$bottom) == 85 && sp4$id[which.max(hz$bottom)] == "shasta-trinity")
  # plot(hz) 
  
  # example using old-style O horizons
  hz <- read.table(text = "peiidref hzdept hzdepb hzname seqnum phiid
                                11      0      5      A      2   295
                                11      3      1     Oi      1   293
                                11      1      0     Oe      1   294
                                11      5     13     C1      3   296
                                11     13     58     C2      4   297
                                11     58    152     C3      5   298
                                13      0      5      A      2   303
                                13      1      0     Oe      1   302
                                13      5     25     Bw      3   304
                                13     25     61      C      4   305
                                13     61     NA      R      5   306
                               136      0     13     A1      3   695
                               136      1      0     Oe      2   694
                               136      2      1     Oi      1   693
                               136     13     61     C1      4   696
                               136     61     76     C2      5   697",
    header = TRUE, row.names = NULL)

  expect_warning({depths(hz) <- peiidref ~ hzdept + hzdepb})

  hzdesgnname(hz) <- "hzname"
  
  hz_fixed <- accumulateDepths(hz)

  is_valid <- checkHzDepthLogic(hz_fixed)$valid

  test1 <- subset(hz_fixed, is_valid)
  origO <- subset(hz, grepl("O", hzname))
  fixedO <- subset(hz_fixed, grepl("O", hzname))
  
  # no O horizons are lost by the fixing
  expect_true(length(origO) == 3 && length(fixedO) == 3)
  
  # all original profiles that were invalid now have valid geometry
  expect_length(test1, 3)
  
  # and the correct O horizons have 0 top depth and appropriate bottom depths
  expect_true(all(test1[,1]$hzname == c("Oi", "Oe", "Oi") &
                  test1[,1]$hzdept == rep(0, 3) &
                  test1[,1]$hzdepb == c(2, 1, 1)))
})

context("SoilProfileCollection init, methods, coercion")

## make sample data
data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# add real coordinates
sp1$x <- seq(-119, -120, length.out = length(sp1))
sp1$y <- seq(38, 39, length.out = length(sp1))

## tests

test_that("SPC construction from a data.frame", {

  # did it work?
  expect_true(inherits(sp1, 'SoilProfileCollection'))

  # ID correctly initialized?
  expect_equal(idname(sp1), 'id')
  expect_true(length(profile_id(sp1)) == length(sp1))

  # ID in the correct order?
  expect_identical(profile_id(sp1), site(sp1)[[idname(sp1)]])

  # depth names?
  expect_equal(horizonDepths(sp1), c('top', 'bottom'))

  # site-level attributes correctly initialized?
  expect_true(length(sp1$group) == length(sp1))

  # correct number of profiles and horizons?
  expect_equal(length(sp1), 9)
  expect_equal(nrow(sp1), 60)

  # test construction with disordered ID and top depths
  daf <- data.frame(id = c(2,2,2,1,1,1), top = c(4,3,2,4,3,2), bottom = c(5,4,3,5,4,3))

  # the input data profiles both have bad "top depth logic" (reversed order of horizons)
  expect_true(hzDepthTests(daf$top[1:3], daf$bottom[1:3])["depthLogic"])

  expect_silent({depths(daf) <- id ~ top + bottom})
  # inspect

  # plot(df) # plot "works" even with invalid depth logic

  # whole SPC is valid, regardless of whether order is corrected
  expect_true(spc_in_sync(daf)$valid)

  # however, after promotion, the depth logic from input data has been corrected
  expect_true(all(checkHzDepthLogic(daf)$valid))

  # the numeric IDs from the input data are in order
  expect_true(all(profile_id(daf) == as.character(1:2)))

})

test_that("SPC diagnostics and restrictions", {
  # diagnostic & restriction slot should be initialized as an empty data.frame
  sp1.dh <- diagnostic_hz(sp1)
  expect_true(inherits(sp1.dh, 'data.frame'))
  expect_equal(nrow(sp1.dh), 0)

  sp1.rh <- restrictions(sp1)
  expect_true(inherits(sp1.rh, 'data.frame'))
  expect_equal(nrow(sp1.rh), 0)
})

test_that("SPC data.frame interface", {

  # init site-level data
  sp1$x <- seq(-119, -120, length.out = length(sp1))
  sp1$y <- seq(38, 39, length.out = length(sp1))

  # init hz-level data
  sp1$z <- runif(n = nrow(sp1))

  expect_equal(length(sp1$x), length(sp1))
  expect_equal(length(sp1$z), nrow(sp1))
})

test_that("SPC deconstruction into a data.frame", {

  # do it here
  h <- horizons(sp1)
  s <- site(sp1)
  d <- as(sp1, 'data.frame')

  expect_true(inherits(h, 'data.frame'))
  expect_true(inherits(s, 'data.frame'))
  expect_true(inherits(d, 'data.frame'))
})


test_that("SPC deconstruction into a list", {

  # do it here
  l <- as(sp1, 'list')

  # result should be a list
  expect_true(inherits(l, 'list'))

  # there should be no NULL data, e.g. missing slots
  res <- sapply(l, is.null)
  expect_false(any(res))

  # check internals
  expect_equivalent(l$idcol, idname(sp1))
  expect_equivalent(l$hzidcol, hzidname(sp1))
  expect_equivalent(l$depthcols, horizonDepths(sp1))
  expect_equivalent(names(l$metadata), names(metadata(sp1)))

  expect_equivalent(l$horizons, horizons(sp1))
  expect_equivalent(l$site, site(sp1))
  expect_equivalent(l$diagnostic, diagnostic_hz(sp1))
  expect_equivalent(l$restrictions, restrictions(sp1))

  # check internals after [-subsetting
  sp1.sub <- sp1[1:2,]
  # none of these slots should change, the others will be subset
  # verifying these are transferred ensures key info slots are handled
  # by the SPC subset method
  expect_equivalent(l$idcol, idname(sp1.sub))
  expect_equivalent(l$hzidcol, hzidname(sp1.sub))
  expect_equivalent(l$depthcols, horizonDepths(sp1.sub))
  expect_equivalent(names(l$metadata), names(metadata(sp1.sub)))
})


test_that("SPC subsetting ", {

  # profile subsets
  expect_true(inherits(sp1[1, ], 'SoilProfileCollection'))
  expect_true(inherits(sp1[1:5, ], 'SoilProfileCollection'))

  # profile and horizon subsets
  expect_true(inherits(sp1[1, 1], 'SoilProfileCollection'))

  # horizon subsets
  expect_true(inherits(sp1[, 2], 'SoilProfileCollection'))
  expect_true(inherits(sp1[, 6], 'SoilProfileCollection'))
  
  # j-index subset with drop=FALSE argument
  site(sp1)$foo <- "bar"
  sp1d <- sp1[, 6, drop = FALSE]
  expect_true(inherits(sp1d, 'SoilProfileCollection'))
  expect_equal(length(sp1d), length(sp1)) # no profiles removed 
                                          # (4 with less than 6 horizons)
  expect_equal(sum(is.na(sp1d$foo)), 4)   # 4 profiles with NA $foo data
  
  # check empty profiles
  expect_true(all(!isEmpty(sp1)))
  expect_equal(isEmpty(sp1d), c(FALSE, TRUE, TRUE, TRUE, TRUE,
                                FALSE, FALSE, FALSE, FALSE))
  
  # there should only be 1 profile and 1 horizon
  expect_equal(length(sp1[1, 1]), 1)
  expect_equal(nrow(sp1[1, 1]), 1)

  # there should be 5 profiles and 1 horizon / profile
  expect_equal(length(sp1[1:5, 1]), 5)
  expect_equal(nrow(sp1[1:5, 1]), 5)

})

test_that("SPC subsetting with tidy verbs ", {

  # filter works as expected
  expect_equal(length(subset(sp1, structure_type == "PL")), 1)

  # ensure multiple expressions yields same result as single expression
  l1 <- subset(sp1, !is.na(texture), prop > mean(prop, na.rm=TRUE))
  l2 <- subset(sp1, !is.na(texture) & prop > mean(prop, na.rm=TRUE))
  expect_equivalent(length(l1), length(l2))

  # mixing of site and horizon level expressions is the intersection
  l1 <- subset(sp1, group == 2, prop > mean(prop, na.rm=TRUE))
  expect_equivalent(length(l1), 4)

  # grepSPC works as expected
  expect_equal(length(grepSPC(sp1, texture, "SCL")), 1)

  # subApply works as expected
  expect_equal(length(subApply(sp1, function(p) TRUE)), length(sp1))
})

test_that("SPC graceful failure of spatial operations when data are missing", {

  expect_false(validSpatialData(sp1))

  # coercion should not work
  expect_error(as(sp1, 'SpatialPoints'))
  expect_error(as(sp1, 'SpatialPointsDataFrame'))

  # square-bracket indexing should work with n = 1
  # https://github.com/ncss-tech/aqp/issues/85
  s <- sp1[1, 1]
  expect_true(inherits(s, 'SoilProfileCollection'))

})


test_that("SPC spatial operations ", {
  
  skip_if_not_installed("sf")
  
  # init / extract coordinates
  initSpatial(sp1) <- ~ x + y
  
  # "coordinates" getter is getSpatial
  co <- getSpatial(sp1)
  
  # these are valid coordinates
  expect_true(validSpatialData(sp1))
  
  # coordinates should be a matrix
  expect_true(inherits(co, 'matrix'))
  
  # as many rows as length and 2 columns
  expect_equal(dim(co), c(length(sp1), 2))
  
  # coordinate columns should be removed from @site
  # expect_true(all(!dimnames(co)[[2]] %in% siteNames(sp1)))

  # set CRS
  expect_silent(prj(sp1) <- "OGC:CRS84")
  
  # get CRS (via crs(<SPC>) method)
  expect_true(nchar(prj(sp1)) > 0)
  
  # # basic coercion
  expect_true(inherits(as(sp1, 'SpatialPoints'), 'SpatialPoints'))

  # # down-grade to {site + sp} = SpatialPointsDataFrame
  expect_message(as(sp1, 'SpatialPointsDataFrame'), 'only site data are extracted')

  # retain SPC object when using unit-length j index
  sp1.spc <- suppressMessages(sp1[, 1])
  expect_true(inherits(sp1.spc, 'SoilProfileCollection'))

  # again, with profile indexing
  sp1.spc <- suppressMessages(sp1[1, 1])
  expect_true(inherits(sp1.spc, 'SoilProfileCollection'))

})


test_that("SPC misc. ", {

  # units
  depth_units(sp1) <- 'inches'
  expect_equal(depth_units(sp1), 'inches')

  # metadata
  m <- metadata(sp1)
  m$citation <- 'this is a citation'
  metadata(sp1) <- m
  expect_true(is.list(metadata(sp1)))


  expect_equal(length(metadata(sp1)$citation), 1)

})


test_that("SPC depth columns get/set ", {

  # getting
  hd <- horizonDepths(sp1)
  expect_equal(hd, c('top', 'bottom'))

  # setting
  hd.new <- c('t', 'b')
  horizonDepths(sp1) <- hd.new
  expect_equal(horizonDepths(sp1), hd.new)

  # error conditions
  expect_error(horizonDepths(sp1) <- NA)
  expect_error(horizonDepths(sp1) <- NULL)
  expect_error(horizonDepths(sp1) <- c(1,2,3))
  expect_error(horizonDepths(sp1) <- c('t'))
  expect_error(horizonDepths(sp1) <- c('t', NA))

  # warnings
  expect_warning(horizonDepths(sp1) <- c('t', '2342sdrse'))

})

test_that("SPC min/max overrides work as expected", {

  set.seed(20202)
  df <- lapply(1:10, random_profile, SPC = TRUE)
  df <- pbindlist(df)

  ## visually inspect output
  # profileApply(df, min)
  # profileApply(df, max)

  # both min and max should return 10cm
  expect_equal(min(df), 44)
  expect_equal(max(df), 134)

  expect_equal(min(df, v = "p2"), 44)
  expect_equal(max(df, v = "p2"), 134)
})

test_that("SPC horizonNames get/set ", {

  # getting
  hn <- horizonNames(sp1)
  expect_equal(hn, c("id", "top", "bottom", "bound_distinct", "bound_topography",
                     "name", "texture", "prop", "structure_grade", "structure_size",
                     "structure_type", "stickiness", "plasticity", "field_ph", "hue",
                     "value", "chroma", "hzID"))

  # setting
  idx <- match('chroma', hn)
  hn[idx] <- 'g'
  horizonNames(sp1) <- hn
  expect_equal(horizonNames(sp1), hn)

  # error conditions
  expect_error(horizonNames(sp1) <- NA)
  expect_error(horizonNames(sp1) <- NULL)
  expect_error(horizonNames(sp1) <- c(1,2,3))
  expect_error(horizonNames(sp1) <- c('t'))
  expect_error(horizonNames(sp1) <- c('t', NA))
  expect_error(horizonNames(sp1) <- hn[-1])

  # warnings
  hn[idx] <- '   g'
  expect_warning(horizonNames(sp1) <- hn)

})


test_that("SPC horizon ID get/set ", {

  # automatically generated horizon IDs
  auto.hz.ids <- hzID(sp1)

  # should be 1:nrow(sp1)
  expect_equivalent(auto.hz.ids, as.character(seq_len(nrow(sp1))))

  # try replacing with reasonable IDs
  hzID(sp1) <- rev(hzID(sp1))
  expect_equivalent(hzID(sp1), as.character(rev(seq_len(nrow(sp1)))))

  # try replacing with bogus values
  expect_error(hzID(sp1) <- 1)
  # non-unique
  expect_error(hzID(sp1) <- sample(hzID(sp1), replace = TRUE))

})


test_that("SPC horizon ID name get/set ", {

  # check default
  expect_equivalent(hzidname(sp1), 'hzID')

  # make a new horizon ID
  sp1$junk <- 1:nrow(sp1)
  hzidname(sp1) <- 'junk'
  expect_equivalent(hzidname(sp1), 'junk')

  # error conditions:
  # no column
  expect_error(hzidname(sp1) <- 'xxx')

  # warning conditions:
  # not unique, keep default
  expect_warning(hzidname(sp1) <- 'top')

})

test_that("SPC horizon designation/texcl name get/set ", {

  # check intended behavior of setters
  hzdesgnname(sp1) <- 'name'
  hztexclname(sp1) <- 'texture'

  expect_equivalent(hzdesgnname(sp1), 'name')
  expect_equivalent(hztexclname(sp1), 'texture')

  # check handy accessor for hz designations
  designations <- hzDesgn(sp1)
  expect_type(designations, 'character')
  expect_equal(length(designations), 60)

  # make a new horizon column
  sp1$junk <- rep("foo", nrow(sp1))
  hzdesgnname(sp1) <- 'junk'
  hztexclname(sp1) <- 'junk'
  expect_equivalent(hzdesgnname(sp1), 'junk')
  expect_equivalent(hztexclname(sp1), 'junk')

  # error conditions:
  # no column in horizon table 'xxx'
  expect_error(hzdesgnname(sp1) <- 'xxx')

  # set to empty
  expect_silent(hzdesgnname(sp1) <- '')

  # null when cannot find the column name
  expect_silent(designations <- hzDesgn(sp1))
  expect_true(is.null(designations))
})

test_that("SPC horizon ID get/set ", {

  # automatically generated horizon IDs
  auto.hz.ids <- hzID(sp1)

  # should be 1:nrow(sp1)
  expect_equivalent(auto.hz.ids, as.character(seq_len(nrow(sp1))))

  # try replacing with reasonable IDs
  hzID(sp1) <- rev(hzID(sp1))
  expect_equivalent(hzID(sp1), as.character(rev(seq_len(nrow(sp1)))))

  # try replacing with bogus values
  expect_error(hzID(sp1) <- 1)

  # non-unique
  expect_error(hzID(sp1) <- sample(hzID(sp1), replace = TRUE))

})


test_that("SPC profile ID get/set ", {

  # original
  pIDs <- profile_id(sp1)

  # new
  pIDs.new <- sprintf("%s-copy", pIDs)

  # try re-setting
  profile_id(sp1) <- pIDs.new

  # were the IDs altered?
  expect_equivalent(profile_id(sp1), pIDs.new)

  # bogus edits
  expect_error(profile_id(sp1) <- 1)
  expect_error(profile_id(sp1) <- sample(pIDs, replace = TRUE))
  expect_error(profile_id(sp1) <- c(NA, pIDs[-1]))

})

context("SoilProfileCollection integrity")

test_that("SPC profile ID reset integrity: site", {

  # test site
  data(sp4)

  # message due to unordered site IDs
  expect_silent(depths(sp4) <- id ~ top + bottom)

  # save old ID and replace with known pattern
  sp4$old_id <- profile_id(sp4)
  profile_id(sp4) <- sprintf("%s-zzz", profile_id(sp4))

  # stripping the pattern should return original labels, in order
  expect_equal(sp4$old_id, gsub('-zzz', '', profile_id(sp4)))

})


test_that("SPC profile ID reset integrity: horizon", {

  # test hz
  data(sp4)

  # message due to unordered site IDs
  expect_silent(depths(sp4) <- id ~ top + bottom)

  # save old ID and replace with known pattern
  sp4$old_id <- as.vector(unlist(horizons(sp4)[idname(sp4)]))
  profile_id(sp4) <- sprintf("%s-zzz", profile_id(sp4))

  # stripping the pattern should return original labels, in order
  new.ids <- as.vector(unlist(horizons(sp4)[idname(sp4)]))
  new.ids <- gsub(pattern='-zzz', replacement = '', x = new.ids)
  expect_equal(sp4$old_id, new.ids)

})

test_that("SPC horizon ID init conflicts", {

  # decompose, re-init and test for message
  x <- sp1
  x <- as(x, 'data.frame')
  expect_message(depths(x) <- id ~ top + bottom, "^using")
  expect_equivalent(hzidname(x), 'hzID')
  expect_true(checkSPC(x))

  # decompose, add non-unique column conflicing with hzID
  x <- sp1
  x <- as(x, 'data.frame')
  x$hzID <- 1
  expect_warning(depths(x) <- id ~ top + bottom, "not a unique horizon ID, using")
  # test backup name
  expect_equivalent(hzidname(x), 'hzID_')

  # special case: IDs resulting from dice()
  s <- dice(sp1, 0:100 ~ ., SPC = TRUE)
  expect_equivalent(hzidname(s), 'sliceID')
  # check to make sure hzID and sliceID are present
  expect_equal(length(grep('hzID|sliceID', horizonNames(s))), 2)

})

test_that("horizons<- left-join", {
  x <- sp1

  # take unique site ID, horizon ID, and corresponding property (clay)
  hnew <- horizons(x)[, c(idname(x), hzidname(x), 'prop')]

  # do some calculation, create a few new variables
  hnew$prop100 <- hnew$prop / 100
  hnew$prop200 <- hnew$prop / 200
  hnew$prop300 <- hnew$prop / 300

  # change a value of existing variable
  hnew$prop[1] <- 50

  # utilize horizons<- left join
  expect_silent(horizons(x) <- hnew)

  # verify old columns have same names
  # (i.e. no issues with duplication of column names in merge)
  expect_true(all(c(idname(x), hzidname(x), 'prop') %in% names(horizons(x))))

  # verify old columns have same value
  clay_prop <- horizons(sp1)[2,'prop']
  expect_equivalent(horizons(x)[2, c('prop')], clay_prop)

  # verify new columns have been added
  # now with proper sorting; first profile, first horizon
  expect_equivalent(horizons(x)[2, c('prop100','prop200','prop300')],
                    c(clay_prop /  100, clay_prop / 200, clay_prop / 300))


})

test_that("ordering of profiles and horizons is retained after left-join", {
  # IDs that when sorted will not be in this order
  s <- c('a', "1188707", "1188710", "120786", "1207894", 'z')

  l <- lapply(s, random_profile)

  d <- do.call('rbind', l)

  # init SPC
  # message due to unordered site IDs
  expect_silent({depths(d) <- id ~ top + bottom})

  ## former bug on set of a new horizon-level attr
  d$zzz <- rep(NA, times = nrow(d))

  # previously mysterious warning message
  z <- d[1:5, ]

  # ordering of profile IDs (unique, from @horizons) != ordering of IDs in @site
  expect_true(all(profile_id(d) == site(d)[[idname(d)]]))
})

test_that("replaceHorizons<- works as expected", {
  x <- sp1

  # replacement with existing value -- works
  hz.before <- horizons(x)
  replaceHorizons(x) <- hz.before
  expect_equal(hz.before, horizons(x))

  # works when hzidname is missing, defaults to hzID
  expect_message(replaceHorizons(x) <- horizons(x)[,c(idname(x),
                                                      horizonDepths(x))])
  expect_equal(x$hzID, as.character(1:nrow(x)))

  # missing idname = error
  expect_error(replaceHorizons(x) <- horizons(x)[,c(horizonDepths(x))])

  # missing depths = error
  expect_error(replaceHorizons(x) <- horizons(x)[,c(idname(x))])
})

spc <- data.frame(id = do.call('c', as.list((lapply(1:4, function(i) rep(i, 10))))),
                  top = rep(0:9, 4), bottom = rep(1:10, 4))

depths(spc) <- id ~ top+bottom

rev.ord <- rev(1:nrow(spc@horizons))

test_that("basic integrity checks", {

  # a new SPC is valid
  expect_true(spc_in_sync(spc)$valid)

  spc@horizons <- spc@horizons[rev.ord,]

  # inverting the horizon order makes it invalid
  expect_true(!spc_in_sync(spc)$valid)

  # an empty spc derived from invalid spc is valid
  expect_true(spc_in_sync(spc[0,])$valid)

  # reordering the horizons with reorderHorizons resolves integrity issues
  expect_true(spc_in_sync(reorderHorizons(spc, seq(nrow(spc))))$valid)

  # override and reverse it
  expect_false(spc_in_sync(reorderHorizons(spc, rev(seq(nrow(spc)))))$valid)

  # removing the metadata works because target order matches sequential order
  #  this cannot be guaranteed to be the case in general but is a reasonable default
  spc@metadata$target.order <- NULL
  expect_true(spc_in_sync(reorderHorizons(spc, seq(nrow(spc))))$valid)

  # reordering horizons with any order works, even if invalid
  spc <- reorderHorizons(spc, target.order = c(20:40,1:19))
  expect_true(!spc_in_sync(spc)$valid)

  # inspect the hzids -- in this case we know they should be 20:40 then 1:19
  expect_true(all(spc[[hzidname(spc)]] == c(20:40,1:19)))

  # subset the broken SPC to get the 4th profile
  spc4 <- subset(spc, id == "4")

  # the target order reflects a reasonable result for the single profile SPC
  expect_true(all(metadata(spc4)$target.order == 1:10))

  # the subset of the broken SPC is valid
  expect_true(spc_in_sync(spc4)$valid)

})

context("SoilProfileCollection depth error checking")

## sample data
data(sp3)

expect_silent({depths(sp3) <- id ~ top + bottom})

test_that("hzDepthTests works as expected", {
  hdep <- horizonDepths(sp3)

  # vector of top and bottom depths -> 4 logical test results
  res <- hzDepthTests(sp3[[hdep[1]]], sp3[[hdep[2]]])
  expect_equal(names(res)[res], c("depthLogic","overlapOrGap")) # depthLogic & overlapOrGap errors
  expect_equal(length(res), 4)

  # mismatched lengths (top and bottom must have same number of values)
  expect_error(hzDepthTests(sp3[[hdep[1]]], sp3[[hdep[2]]][1]))
})

test_that("checkHzDepthLogic() works as expected", {

  # these data should be clean
  res <- checkHzDepthLogic(sp3)

  # result is an data.frame
  expect_true(inherits(res, 'data.frame'))

  # number of rows should match length(SPC)
  expect_true(nrow(res) == length(sp3))

  # all clear
  expect_true(all(!res$depthLogic))
  expect_true(all(!res$sameDepth))
  expect_true(all(!res$missingDepth))
  expect_true(all(!res$overlapOrGap))
  expect_true(all(res$valid))
  
  # works on a horizons()-like data.frame
  h <- horizons(sp3)
  h$foo <- h$id
  h$id <- NULL
  h$hzdept <- h$top
  h$top <- NULL
  h$hzdepb <- h$bottom
  h$bottom <- NULL
  expect_true(all(checkHzDepthLogic(h, c("hzdept","hzdepb"), "foo", fast = TRUE)$valid))
  expect_true(all(checkHzDepthLogic(h, c("hzdept","hzdepb"), "foo", fast = TRUE, byhz = TRUE)$valid))
})


test_that("preservation of hzID", {
  
  # modify hzID in-place
  hzID(sp3) <- sprintf("%04d", as.integer(sp3$hzID))
  
  res <- checkHzDepthLogic(sp3, byhz = TRUE, fast = FALSE)
  
  expect_true(
    all(hzID(sp3) == res$hzID)
  )
  
  res <- checkHzDepthLogic(sp3, byhz = TRUE, fast = TRUE)
  
  expect_true(
    all(hzID(sp3) == res$hzID)
  )
  
  # data.frame interface
  
  res <- checkHzDepthLogic(horizons(sp3), idname = idname(sp3),
                           hzdepths = horizonDepths(sp3),
                           byhz = TRUE, fast = FALSE)
  
  # data.frame interface always returns calculated integer hzID
  expect_true(
    all(1:nrow(sp3) == res$hzID)
  )
  
  res <- checkHzDepthLogic(horizons(sp3), idname = idname(sp3), 
                           hzdepths = horizonDepths(sp3),
                           byhz = TRUE, fast = TRUE)
  
  expect_true(
    all(1:nrow(sp3) == res$hzID)
  )
})

test_that("checkHzDepthLogic() depth logic errors", {

  # local copy
  x <- sp3[1, ]
  x$top[1] <- 10
  res <- checkHzDepthLogic(x)

  # errors only affect the first profile in this set
  expect_true(res$depthLogic[1])
  expect_false(res$valid[1])
  
  res2 <- checkHzDepthLogic(x, byhz = TRUE)
  expect_true(res2$depthLogic[1])
  expect_false(res2$valid[1])
})

test_that("checkHzDepthLogic() same top / bottom depths", {

  # local copy
  x <- sp3[7, ]
  x$bottom[3] <- x$top[3]
  res <- checkHzDepthLogic(x)

  # errors only affect the first profile in this set
  expect_true(res$sameDepth[1])
  expect_false(res$valid[1])
  
  res2 <- checkHzDepthLogic(x, byhz = TRUE)
  expect_true(res2$sameDepth[3])
  expect_false(res2$valid[3])
})


test_that("checkHzDepthLogic() NA in depths", {

  # local copy
  x <- sp3[4, ]
  x$bottom[3] <- NA
  res <- checkHzDepthLogic(x)

  # errors only affect the first profile in this set
  expect_true(res$missingDepth[1])
  expect_false(res$valid[1])
  
  res2 <- checkHzDepthLogic(x, byhz = TRUE)
  expect_true(res2$missingDepth[3])
  expect_false(res2$valid[3])
})


test_that("checkHzDepthLogic() gap", {

  # local copy
  x <- sp3[8, ]
  # create a gap
  x$top[4] <- 82
  res <- checkHzDepthLogic(x)

  # errors only affect the first profile in this set
  expect_true(res$overlapOrGap[1])
  expect_false(res$valid[1])
  
  # NOTE: OVERLAP OR GAP NOT MEANINGFUL FOR byhz=TRUE
  # result is NA
  res2 <- checkHzDepthLogic(x, byhz = TRUE)
  expect_true(is.na(res2$overlapOrGap[1]))
  expect_true(res2$valid[1])
})


test_that("checkHzDepthLogic() overlap", {

  # local copy
  x <- sp3[8, ]
  # create overlap
  x$top[4] <- 75
  res <- checkHzDepthLogic(x)

  # errors only affect the first profile in this set
  expect_true(res$overlapOrGap[1])
  expect_false(res$valid[1])
  
  # NOTE: OVERLAP OR GAP NOT MEANINGFUL FOR byhz=TRUE
  # result is NA
  res2 <- checkHzDepthLogic(x, byhz = TRUE)
  expect_true(is.na(res2$overlapOrGap[1]))
  expect_true(res2$valid[1])
})


test_that("splitLogicErrors", {
  data(sp4)
  depths(sp4) <- id ~ top + bottom

  # no errors (all list elements return NULL)
  expect_equal(unlist(splitLogicErrors(sp4)), c(NULL, NULL, NULL, NULL))

  # NA in top depth triggers depth logic and missing depth errors
  data(sp4)
  sp4$top[1] <- NA
  expect_message(depths(sp4) <- id ~ top + bottom)

  res <- splitLogicErrors(sp4)
  
  # the same profile occurs in two groups, since NA causes depth logic and missingDepth errors
  expect_true(profile_id(res$depthLogic) == profile_id(res$missingDepth))

  # interact = TRUE gets these in the same (interaction) group
  #  each SPC profile occurs once, name/number elements varies with your data
  #  (and whether or not you use split.default(..., drop = TRUE))
  res2 <- splitLogicErrors(sp4, interact = TRUE, sep = "_", drop = TRUE)
  expect_true(length(res2$depthLogic__missingDepth_) == 1)
})

## 2019-07-16: moved util functions to `sketch-utils.R`

# TODO: behavior not defined for horizons with an indefinite lower boundary
# TODO: move some of the processing outside of the main loop: column names, etc.
# TODO: use .detectColorSpec() for interpreting colors, then add Munsell notation as option

#' @title Create Soil Profile Sketches
#' @name plotSPC
#' @rdname SoilProfileCollection-plotting-methods
#' @docType methods
#' @aliases plot,SoilProfileCollection,ANY-method,plot.SoilProfileCollection
#'
#' @description Generate a diagram of soil profile sketches from a `SoilProfileCollection` object. The [Introduction to SoilProfileCollection Objects Vignette](http://ncss-tech.github.io/aqp/articles/Introduction-to-SoilProfileCollection-Objects.html) contains many examples and discussion of the large number of arguments to this function. The [Soil Profile Sketches](https://ncss-tech.github.io/AQP/aqp/sketches.html) tutorial has longer-form discussion and examples pertaining to suites of related arguments.
#'
#' Options can be used to conveniently specify sets of arguments that will be used in several calls to `plotSPC()` within a single R session. For example, arguments can be specified in a named list (`.a`) and set using: `options(.aqp.plotSPC.args = .a)`. Reset these options via `options(.aqp.plotSPC.args = NULL)`. Arguments explicitly passed to `plotSPC()` will override arguments set via [options()].
#'
#'
#' @param x a `SoilProfileCollection` object
#'
#' @param color quoted column name containing R-compatible color descriptions, or numeric / categorical data to be displayed thematically; see details
#'
#' @param width scaling of profile widths (typically 0.1 - 0.4)
#'
#' @param name quoted column name of the (horizon-level) attribute containing horizon designations or labels, if missing `hzdesgnname(x)` is used. Suppress horizon name printing by setting `name = NA` or `name = ''`.
#'
#' @param name.style one of several possible horizon designations labeling styles: `c('right-center', 'left-center', 'left-top', 'center-center', 'center-top')`
#'
#' @param label quoted column name of the (site-level) attribute used to identify profile sketches
#' 
#' @param raggedBottom either quoted column name of the (site-level) attribute (logical) used to mark profiles with a truncated lower boundary, or `FALSE` suppress ragged bottom depths when `max.depth < max(x)`
#'
#' @param hz.depths logical, annotate horizon top depths to the right of each sketch (`FALSE`)
#' 
#' @param hz.depths.offset numeric, user coordinates for left-right adjustment for horizon depth annotation; reasonable values are usually within 0.01-0.05 (default: 0)
#' 
#' @param hz.depths.lines logical, draw segments between horizon depth labels and actual horizon depth; this is useful when including horizon boundary distinctness and/or `fixLabelCollisions = TRUE`
#'
#' @param depth.axis logical or list. Use a logical to suppress (`FALSE`) or add depth axis using defaults (`TRUE`). Use a list to specify one or more of: 
#'  - `style`: character, one of 'traditional', 'compact', or 'tape'
#'  - `line`: numeric, negative values move axis to the left (does not apply to `style = 'tape'`)
#'  - `cex`: numeric, scaling applied to entire depth axis
#'  - `interval`: numeric, axis interval
#' See examples.
#'
#' @param alt.label quoted column name of the (site-level) attribute used for secondary annotation
#'
#' @param alt.label.col color used for secondary annotation text
#'
#' @param cex.names baseline character scaling applied to all text labels
#'
#' @param cex.id character scaling applied to `label`
#'
#' @param font.id font style applied to `label`, default is 2 (bold)
#' 
#' @param srt.id rotation applied to `label`, only when `id.style = 'top'`
#' 
#' @param offset.id vertical offset applied to `label`, only when `id.style = 'top'`
#'
#' @param print.id logical, print `label` above/beside each profile? (`TRUE`)
#'
#' @param id.style `label` printing style: 'auto' (default) = simple heuristic used to select from: 'top' = centered above each profile, 'side' = 'along the top-left edge of profiles'
#'
#' @param plot.order integer vector describing the order in which individual soil profiles should be plotted
#'
#' @param relative.pos vector of relative positions along the x-axis, within \{1, n\}, ignores `plot.order` see details
#'
#' @param add logical, add to an existing figure
#'
#' @param scaling.factor vertical scaling of profile depths, useful for adding profiles to an existing figure
#'
#' @param y.offset numeric vector of vertical offset for top of profiles in depth units of `x`, can either be a single numeric value or vector of length = `length(x)`. A vector of y-offsets will be automatically re-ordered according to `plot.order`.
#'
#' @param x.idx.offset integer specifying horizontal offset from 0 (left-hand edge)
#'
#' @param n integer describing amount of space along x-axis to allocate, defaults to `length(x)`
#'
#' @param max.depth numeric. The lower depth for all sketches, deeper profiles are truncated at this depth. Use larger values to arbitrarily extend the vertical dimension, convenient for leaving extract space for annotation.
#'
#' @param n.depth.ticks suggested number of ticks in depth scale
#'
#' @param shrink logical, reduce character scaling for 'long' horizon by 80%
#'
#' @param shrink.cutoff character length defining 'long' horizon names
#' 
#' @param shrink.thin integer, horizon thickness threshold for shrinking horizon names by 80%, only activated when `shrink = TRUE` (`NULL` = no shrinkage)
#'
#' @param abbr logical, abbreviate `label`
#'
#' @param abbr.cutoff suggested minimum length for abbreviated `label`
#'
#' @param divide.hz logical, divide horizons with line segment? (`TRUE`), see details
#'
#' @param hz.distinctness.offset `NULL`, or quoted column name (horizon-level attribute) containing vertical offsets used to depict horizon boundary distinctness (same units as profiles), see details and [hzDistinctnessCodeToOffset()]; consider setting `hz.depths.lines = TRUE` when used in conjunction with `hz.depths = TRUE`
#' 
#' @param hz.topography.offset `NULL`, or quoted column name (horizon-level attribute) containing offsets used to depict horizon boundary topography (same units as profiles), see details and [hzTopographyCodeToOffset()]
#'
#' @param hz.boundary.lty quoted column name (horizon-level attribute) containing line style (integers) used to encode horizon topography
#'
#'
#' @param density fill density used for horizon color shading, either a single integer or a quoted column name (horizon-level attribute) containing integer values (default is `NULL`, no shading)
#'
#' @param show.legend logical, show legend? (default is `TRUE`)
#'
#' @param col.label thematic legend title
#'
#' @param col.palette color palette used for thematic sketches (default is `rev(brewer.pal(10, 'Spectral'))`)
#'
#' @param col.palette.bias color ramp bias (skew), see [colorRamp()]
#'
#' @param col.legend.cex scaling of thematic legend
#'
#' @param n.legend approximate number of classes used in numeric legend, max number of items per row in categorical legend
#'
#' @param lwd line width multiplier used for sketches
#'
#' @param lty line style used for sketches
#'
#' @param default.color default horizon fill color used when `color` attribute is `NA`
#' 
#' @param fixLabelCollisions use [fixOverlap()] to attempt fixing hz depth labeling collisions, will slow plotting of large collections; enabling also sets `hz.depths.lines = TRUE`. Additional arguments to [fixOverlap()] can be passed via `fixOverlapArgs`. Overlap collisions cannot be fixed within profiles containing degenerate or missing horizon depths (e.g. top == bottom).
#' 
#' @param fixOverlapArgs a named list of arguments to [fixOverlap()]. Overlap adjustments are attempted using electrostatic simulation with arguments: `list(method = 'E', q = 1)`. Alternatively, select adjustment by simulated annealing via `list(method = 'S')`. See [electroStatics_1D()] and [SANN_1D()] for details.
#'
#' @param \dots other arguments passed into lower level plotting functions
#'
#'
#' @param axis.line.offset (deprecated, use `depth.axis` instead) horizontal offset applied to depth axis (default is -2, larger numbers move the axis to the right)
#'
#' @param plot.depth.axis (deprecated, use `depth.axis` instead) logical, plot depth axis?
#'
#' @param cex.depth.axis (deprecated, use `depth.axis` instead) character scaling applied to depth scale
#' 
#'
#' @details
#' Depth limits (`max.depth`) and number of depth ticks (`n.depth.ticks`) are *suggestions* to the [pretty()] function. You may have to tinker with both parameters to get what you want.
#'
#' The 'side' `id.style` is useful when plotting a large collection of profiles, and/or, when profile IDs are long.
#'
#' If the column containing horizon designations is not specified (the `name` argument), a column (presumed to contain horizon designation labels) is guessed based on regular expression matching of the pattern 'name'--this usually works, but it is best to manual specify the name of the column containing horizon designations.
#'
#' The `color` argument can either name a column containing R-compatible colors, possibly created via [munsell2rgb()], or column containing either numeric or categorical (either factor or character) values. In the second case, values are converted into colors and displayed along with a simple legend above the plot. Note that this functionality makes several assumptions about plot geometry and is most useful in an interactive setting.
#'
#' Adjustments to the legend can be specified via `col.label` (legend title), `col.palette` (palette of colors, automatically expanded), `col.legend.cex` (legend scaling), and `n.legend` (approximate number of classes for numeric variables, or, maximum number of legend items per row for categorical variables). Currently, `plotSPC` will only generate two rows of legend items. Consider reducing the number of classes if two rows isn't enough room.
#'
#' Profile sketches can be added according to relative positions along the x-axis (vs. integer sequence) via `relative.pos` argument. This should be a vector of positions within \{1,n\} that are used for horizontal placement. Default values are `1:length(x)`. Care must be taken when both `plot.order` and `relative.pos` are used simultaneously: `relative.pos` specifies horizontal placement after sorting. [addDiagnosticBracket()] and [addVolumeFraction()] use the `relative.pos` values for subsequent annotation.
#'
#' Relative positions that are too close will result in overplotting of sketches. Adjustments to relative positions such that overlap is minimized can be performed with `fixOverlap(pos)`, where `pos` is the original vector of relative positions.
#'
#' The `x.idx.offset` argument can be used to shift a collection of pedons from left to right in the figure. This can be useful when plotting several different `SoilProfileCollection` objects within the same figure. Space must be pre-allocated in the first plotting call, with an offset specified in the second call. See examples below.
#' 
#' Horizon depths (e.g. cm) are converted to figure y-coordinates via: y = (depth * scaling.factor) + y.offset. 
#'
#'
#' @note A new plot of soil profiles is generated, or optionally added to an existing plot.
#'
#' @author D.E. Beaudette
#'
#' @references Beaudette, D.E., Roudier P., and A.T. O'Geen. 2013. Algorithms for Quantitative Pedology: A Toolkit for
#' Soil Scientists. Computers & Geosciences. 52:258 - 268.
#'
#' @keywords hplots
#' @export
#'
#' @seealso [fixOverlap()], [explainPlotSPC()], [SoilProfileCollection-class], [pretty()], [hzDistinctnessCodeToOffset()], [addBracket()], [profileGroupLabels()]
#'
#' @examples
#'
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' 
#' # example data
#' data(sp1)
#' # usually best to adjust margins
#' par(mar = c(0,0,3,0))
#'
#' # add color vector
#' sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))
#'
#' # promote to SoilProfileCollection
#' depths(sp1) <- id ~ top + bottom
#' 
#' # init horizon designation 
#' hzdesgnname(sp1) <- 'name'
#' 
#' # plot profiles
#' plotSPC(sp1, id.style = 'side')
#'
#' # title, note line argument:
#' title('Sample Data 1', line = 1, cex.main = 0.75)
#'
#' # plot profiles without horizon-line divisions
#' plotSPC(sp1, divide.hz = FALSE)
#'
#' # diagonal lines encode horizon boundary distinctness
#' sp1$hzD <- hzDistinctnessCodeToOffset(sp1$bound_distinct)
#' plotSPC(sp1, hz.distinctness.offset = 'hzD', name.style = 'center-center')
#'
#' # plot horizon color according to some property
#' data(sp4)
#' depths(sp4) <- id ~ top + bottom
#' hzdesgnname(sp4) <- 'name'
#' plotSPC(sp4, color = 'clay')
#'
#' # another example
#' data(sp2)
#' depths(sp2) <- id ~ top + bottom
#' hzdesgnname(sp2) <- 'name'
#' site(sp2) <- ~ surface
#'
#' # some of these profiles are very deep, truncate plot at 400cm
#' # label / re-order with site-level attribute: `surface`
#' plotSPC(sp2, label = 'surface', plot.order = order(sp2$surface), 
#' max.depth = 400)
#'
#' # example using a categorical attribute
#' plotSPC(sp2, color = "plasticity", 
#' max.depth = 400)
#'
#' # plot two SPC objects in the same figure
#' par(mar = c(1,1,1,1))
#' 
#' # plot the first SPC object and
#' # allocate space for the second SPC object
#' plotSPC(sp1, n = length(sp1) + length(sp2))
#' 
#' # plot the second SPC, starting from the first empty space
#' plotSPC(sp2, x.idx.offset = length(sp1), add = TRUE)
#'
#'
#' ##
#' ## demonstrate horizon designation shrinkage
#' ##
#' 
#' data("jacobs2000")
#' 
#' # shrink "long" horizon names
#' plotSPC(
#'   jacobs2000, 
#'   name = 'name',
#'   name.style = 'center-center', 
#'   shrink = TRUE, 
#'   cex.names = 0.8
#' )
#' 
#' # shrink horizon names in "thin" horizons
#' plotSPC(
#'   jacobs2000, 
#'   name = 'name',
#'   name.style = 'center-center', 
#'   shrink = TRUE, 
#'   shrink.thin = 15,
#'   cex.names = 0.8,
#' )
#' 
#'
#' ##
#' ## demonstrate adaptive legend
#' ##
#'
#' data(sp3)
#' depths(sp3) <- id ~ top + bottom
#'
#' # make some fake categorical data
#' horizons(sp3)$fake.data <- sample(letters[1:15], size = nrow(sp3), replace=TRUE)
#'
#' # better margins
#' par(mar=c(0,0,3,1))
#'
#' # note that there are enough colors for 15 classes (vs. previous limit of 10)
#' # note that the legend is split into 2 rows when length(classes) > n.legend argument
#' plotSPC(sp3, color='fake.data', name='fake.data', cex.names=0.8)
#'
#' # make enough room in a single legend row
#' plotSPC(sp3, color='fake.data', name='fake.data', cex.names=0.8, n.legend=15)
#' 
#' 
#' ##
#' ## demonstrate y.offset argument
#' ## must be of length 1 or length(x)
#' ##
#' 
#' # example data and local copy
#' data("jacobs2000")
#' x <- jacobs2000
#' hzdesgnname(x) <- 'name'
#' 
#' # y-axis offsets, simulating a elevation along a hillslope sequence
#' # same units as horizon depths in `x`
#' # same order as profiles in `x`
#' y.offset <- c(-5, -10, 22, 65, 35, 15, 12)
#' 
#' par(mar = c(0, 0, 2, 2))
#' 
#' # y-offset at 0
#' plotSPC(x, color = 'matrix_color', cex.names = 0.66)
#' 
#' # constant adjustment to y-offset
#' plotSPC(x, color = 'matrix_color', cex.names = 0.66, y.offset = 50)
#' 
#' # attempt using invalid y.offset
#' # warning issued and default value of '0' used
#' # plotSPC(x, color = 'matrix_color', cex.names = 0.66, y.offset = 1:2)
#' 
#' # variable y-offset
#' # fix overlapping horizon depth labels
#' par(mar = c(0, 0, 1, 0))
#' plotSPC(
#'   x, 
#'   y.offset = y.offset, 
#'   color = 'matrix_color', 
#'   cex.names = 0.75, 
#'   shrink = TRUE,
#'   hz.depths = TRUE, 
#'   hz.depths.offset = 0.05,
#'   fixLabelCollisions = TRUE,
#'   name.style = 'center-center'
#' )
#' 
#' 
#' 
#' # random y-axis offsets
#' yoff <- runif(n = length(x), min = 1, max = 100)
#' 
#' # random gradient of x-positions
#' xoff <- runif(n = length(x), min = 1, max = length(x))
#' 
#' # note profiles overlap
#' plotSPC(x, 
#'         relative.pos = xoff, 
#'         y.offset = yoff, 
#'         color = 'matrix_color', 
#'         cex.names = 0.66, 
#'         hz.depths = TRUE, 
#'         name.style = 'center-center'
#' )
#' 
#' 
#' # align / adjust relative x positions
#' set.seed(111)
#' pos <- alignTransect(xoff, x.min = 1, x.max = length(x), thresh = 0.65)
#' 
#' # y-offset is automatically re-ordered according to
#' # plot.order
#' 
#' par(mar = c(0.5, 0.5, 0.5, 0.5))
#' plotSPC(x, 
#'         plot.order = pos$order, 
#'         relative.pos = pos$relative.pos, 
#'         y.offset = yoff, 
#'         color = 'matrix_color', 
#'         cex.names = 0.66, 
#'         hz.depths = TRUE, 
#'         name.style = 'center-center'
#' )
#' 
#' box()
#' 
plotSPC <- function(
    x,
    color = 'soil_color',
    width = ifelse(length(x) < 2, 0.15, 0.25),
    name = hzdesgnname(x),
    name.style = 'right-center',
    label = idname(x),
    raggedBottom = NULL,
    hz.depths = FALSE,
    hz.depths.offset = ifelse(fixLabelCollisions, 0.03, 0),
    hz.depths.lines = fixLabelCollisions,
    depth.axis = list(style = 'traditional', cex = cex.names * 1.15),
    alt.label = NULL,
    alt.label.col = 'black',
    cex.names = 0.5,
    cex.id = cex.names + (0.2 * cex.names),
    font.id = 2,
    srt.id = 0, 
    offset.id = 0.5, 
    print.id = TRUE,
    id.style = 'auto',
    plot.order = 1:length(x),
    relative.pos = 1:length(x),
    add = FALSE,
    scaling.factor = 1,
    y.offset = rep(0, times = length(x)),
    x.idx.offset = 0,
    n = length(x),
    max.depth = ifelse(is.infinite(max(x)), 200, max(x)),
    n.depth.ticks = 10,
    shrink = FALSE,
    shrink.cutoff = 3,
    shrink.thin = NULL,
    abbr = FALSE,
    abbr.cutoff = 5,
    divide.hz = TRUE,
    hz.distinctness.offset = NULL,
    hz.topography.offset = NULL,
    hz.boundary.lty = NULL,
    density = NULL,
    show.legend = TRUE,
    col.label = color,
    col.palette = c("#5E4FA2", "#3288BD", "#66C2A5","#ABDDA4", "#E6F598",
                             "#FEE08B","#FDAE61", "#F46D43", "#D53E4F","#9E0142"),
    col.palette.bias = 1,
    col.legend.cex = 1,
    n.legend = 8,
    lwd = 1,
    lty = 1,
    default.color = grey(0.95),
    fixLabelCollisions = hz.depths,
    fixOverlapArgs = list(method = 'E', q = 1),
    
    cex.depth.axis = cex.names,
    axis.line.offset = -2,
    plot.depth.axis = TRUE,
    ...
) {
  
  
  # make R CMD check happy
  .LAST <- NULL
  .BOTTOM <- NULL
  
  # +++++++++++++++++++++++++++
  # arguments from options ----
  # +++++++++++++++++++++++++++
  
  # specified as: options(.aqp.plotSPC.args = list())
  
  # get any options set, result is a list
  # NULL otherwise
  .opArgs <- getOption(".aqp.plotSPC.args", default = NULL)
  
  # process options
  if(!is.null(.opArgs)) {
    
    # e.g. some of: names(formals('plotSPC'))
    # not all arguments can be set via options
    .argSet <- c("color", "width", "name", "name.style", "label", "raggedBottom", "hz.depths", 
                 "hz.depths.offset", "hz.depths.lines", "depth.axis", "alt.label", "alt.label.col", 
                 "cex.names", "cex.id", "font.id", "srt.id", 
                 "print.id", "id.style", "plot.order", "relative.pos", "add", 
                 "scaling.factor", "y.offset", "x.idx.offset", "n", "max.depth", 
                 "n.depth.ticks", "shrink", "shrink.cutoff", "shrink.thin", "abbr", 
                 "abbr.cutoff", "divide.hz", "hz.distinctness.offset", "hz.topography.offset", 
                 "hz.boundary.lty", "density", 
                 "show.legend", "col.label", "col.palette", "col.palette.bias", 
                 "col.legend.cex", "n.legend", "lwd", "lty", "default.color", 
                 "fixLabelCollisions", "fixOverlapArgs", "axis.line.offset", "plot.depth.axis", "cex.depth.axis"
    )
    
    # iterate over all possible arguments that can be modified in this way
    for(.arg in .argSet) {
      
      # test to see if the current argument has been explicitly provided
      .missing <- do.call(missing, list(.arg))
      
      # consider only:
      # specified via options AND NOT (XOR) given by function arguments
      if(!is.null(.opArgs[[.arg]]) && .missing) {
        # debug
        # print(sprintf("using options: %s", .arg))
        # set to local variable in place of default argument
        assign(.arg, .opArgs[[.arg]])
      }
    }
    
  }
  
  
  # ++++++++++++++++++
  # sanity checks ----
  # ++++++++++++++++++
  
  # `x` should be a SoilProfileCollection
  if (!inherits(x, 'SoilProfileCollection')) {
    stop("Object `x` must be a SoilProfileCollection", call. = FALSE) 
  }
  
  ## TODO: update plot.order and relative.pos when n != length(x)
  
  # don't plot an empty collection
  if (nrow(x) == 0) {
    stop("SoilProfileCollection `x` contains no horizon data", call. = FALSE)
  }
  
  # horizon name style
  if(! name.style %in% c('right-center', 'left-center', 'left-top', 'center-center', 'center-top')) {
    warning('invalid `name.style`', call. = FALSE)
    name.style <- 'right-center'
  }
  
  # horizon distinctness offset column name must be valid
  if(!missing(hz.distinctness.offset)) {
    # valid name
    if(! hz.distinctness.offset %in% horizonNames(x))
      stop('invalid `hz.distinctness.offset` column name', call. = FALSE)
    
    # must be numeric
    if(! is.numeric(x[[hz.distinctness.offset]]))
      stop('`hz.distinctness.offset` must be numeric', call. = FALSE)
    
  } else {
    # no horizon distinctness offset specified, use 0
    hz.distinctness.offset <- '.hdo'
    horizons(x)[[hz.distinctness.offset]] <- 0
  }
  
  # horizon topography offset column name must be valid
  if(!missing(hz.topography.offset)) {
    # valid name
    if(! hz.topography.offset %in% horizonNames(x))
      stop('invalid `hz.topography.offset` column name', call. = FALSE)
    
    # must be numeric
    if(! is.numeric(x[[hz.topography.offset]]))
      stop('`hz.topography.offset` must be numeric', call. = FALSE)
    
  } else {
    # no horizon topography offset specified, use 0
    hz.topography.offset <- '.hto'
    horizons(x)[[hz.topography.offset]] <- 0
  }
  
  # horizon line type
  if(!missing(hz.boundary.lty)) {
    # valid name
    if(! hz.boundary.lty %in% horizonNames(x))
      stop('invalid `hz.boundary.lty` column name', call. = FALSE)
    
    # must be numeric
    if(! is.numeric(x[[hz.boundary.lty]]))
      stop('`hz.boundary.lty` must be numeric', call. = FALSE)
  }
  
  # ragged bottom flag
  if(!missing(raggedBottom)) {
    
    # valid type
    if(! inherits(raggedBottom, c('logical', 'character'))) {
      stop('`raggedBottom` must be logical or character', call. = FALSE)
    }
    
    # valid name
    if((is.character(raggedBottom)) & ! raggedBottom %in% siteNames(x)) {
      stop('invalid `raggedBottom` column name', call. = FALSE)
    }
    
  }
  
  # length of y.offset must length(x) or 1
  if(length(y.offset) == 1) {
    y.offset <- rep(y.offset, times = length(x))
  }
  
  if(length(y.offset) != length(x)) {
    warning('length of `y.offset` must be `length(x)` or 1, using `0`', call. = FALSE)
    y.offset <- rep(0, times = length(x))
  }
  
  # fixOverlap arguments
  if(is.null(fixOverlapArgs)) {
    fixOverlapArgs <- list()
  }
  
  
  # ++++++++++++++++++++++++++++++++
  # SPC truncation and flagging ----
  # ++++++++++++++++++++++++++++++++
  
  # keep track of truncation by max.depth via special site-level attr
  site(x)$.isTruncated <- FALSE
  
  if(!missing(max.depth)) {
    
    # keep track of original profile bottom depths
    .pb0 <- x[, , .LAST, .BOTTOM]
    
    # truncate, keeping degenerate profiles
    x <- trunc(x, 0, max.depth, drop = FALSE)
    
    # truncated bottom depths
    .pb1 <- x[, , .LAST, .BOTTOM]
    
    # flag
    x$.isTruncated[which(.pb0 != .pb1)] <- TRUE
  }
  
  # accommodate other flag specified in raggedBottom
  if(!missing(raggedBottom)) {
    truncation_flag <- x[['.isTruncated']] | x[[raggedBottom]]
  } else {
    truncation_flag <- x[['.isTruncated']]
  }
  
  # ragged line characteristics depend on vertical scale
  .raggedOffsets <- scaling.factor * max.depth * c(-0.01,  0.03) / 2
  
  
  # +++++++++++++++++
  # SPC cleanup -----
  # +++++++++++++++++
  
  # tempting, but I don't like the auto-magical nature of this
  # easy to miss important horizon logic errors
  # creates the possibility for new errors based on unexpected results
  # x <- repairMissingHzDepths(x)
  
  
  # ++++++++++++++++++
  # fudge factors ----
  # ++++++++++++++++++
  
  # TODO: base calculations on strwidth() AFTER plot() has been called
  
  # padding along x-axis, prevents crowding
  # dynamic adjustment must also taking into account figure size
  # roughly 10% of length(x)
  extra_x_space <- n * 0.1
  
  # multiplier (width * x_left_space_mult) used to set left-side space along x-axis
  x_left_space_mult <- 2
  
  # add a little extra x-space when n <= 5
  if(n <= 5 & n > 2) {
    extra_x_space <- extra_x_space + 0.25
    x_left_space_mult <- 2
    
  }
  
  # special cases
  if(n == 2) {
    extra_x_space <- extra_x_space + 0.5
    x_left_space_mult <- 2.75
    
  }
  
  # special cases
  if (n == 1) {
    extra_x_space <- extra_x_space + 0.5
    x_left_space_mult <- 3.5
  }
  
  # padding above profiles, ~ 15 is about right for n in {1,25} and max depth near 150cm
  # a sketch of shalllow profiles could benefit from ~ 5
  if(max.depth <= 50)
    extra_y_space <- 5
  if(max.depth > 50 & max.depth <= 100)
    extra_y_space <- 10
  if(max.depth > 100)
    extra_y_space <- 15
  
  # get profile IDs
  pIDs <- profile_id(x)
  
  # re-order y-offset according to plot.order
  y.offset <- y.offset[plot.order]
  
  # get horizons
  h <- horizons(x)
  
  # get column names from horizon data
  nm <- names(h)
  
  
  # ++++++++++++++++++++++++
  # horizon designation ----
  # ++++++++++++++++++++++++
  
  # hzdesgnname() does most of the work here, with sensible defaults when metadata are missing
  
  # code can safely run with all input except NULL
  if (is.null(name)) {
    # set to empty string
    name <- ''
  }
  
  
  
  # +++++++++++++++++++
  # horizon colors ----
  # +++++++++++++++++++
  
  ## TODO: allow use of Munsell notation here
  
  # .interpretHorizonColors() expects numeric | categorical data
  # convert logical to factor
  if (is.logical(h[[color]])) {
    h[[color]] <- factor(h[[color]])
  }
  
  # results contain:
  # vector of horizon colors, row-order preserved
  # legend data if relevant, otherwise NULL
  hz.color.interpretation <- .interpretHorizonColor(h, color, default.color, col.palette, col.palette.bias, n.legend)
  h[['.color']] <- hz.color.interpretation$colors
  
  
  
  
  # +++++++++++++++++++
  # horizon depths ----
  # +++++++++++++++++++
  
  # get top/bottom column names
  IDcol <- idname(x)
  hzDepthCols <- horizonDepths(x)
  tcol <- hzDepthCols[1]
  bcol <- hzDepthCols[2]
  
  if (is.null(label)) {
    label <- NA
  }
  
  # get profile labels from @site
  pLabels <- site(x)[[label]]
  
  if (is.null(pLabels) && !is.na(label) && nchar(label) > 0) {
    warning("column name '", label, "' not found in site names", call. = FALSE)
  }
  
  # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  # init plotting region, unless we are appending to an existing plot ----
  # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  # y-limits also include y.offset range
  ylim.range <- c(
    max.depth + max(y.offset), 
    -extra_y_space
  )
  
  # x-limits
  xlim.range <- c(width * x_left_space_mult, n + extra_x_space)
  
  # note that we are using some fudge-factors to get the plotting region just right
  if(!add) {
    # margins are set outside of this function
    
    # make a new, empty plot
    plot(x = 0, y = 0, type = 'n', 
         xlim = xlim.range,
         ylim = ylim.range,
         axes = FALSE, xlab = '', ylab = ''
    )
  }
  
  
  
  # +++++++++++++++++++++++
  # device information ----
  # +++++++++++++++++++++++
  
  # note: all of this has to happen after plot(...), or if `add = TRUE`
  
  .par_usr <- par('usr')
  .par_dev <- par('pin')
  .par_xWidth <- diff(.par_usr[1:2])
  .par_devWidth <- .par_dev[1]
  
  # profiles / device width (inches)
  .dev_sketch_density <- (.par_xWidth / .par_devWidth)
  
  # calculate width of a single character on current plot device
  one.char.width <- strwidth('W')
  
  
  # +++++++++++++++++++++++++++++++
  # profile ID style selection ----
  # +++++++++++++++++++++++++++++++
  
  # if profile style is auto, determine style based on font metrics
  if(id.style == 'auto' & !is.null(pLabels)) {
    sum.ID.str.width <- sum(sapply(pLabels, strwidth, units = 'inches', cex = cex.id, font = 2))
    ID.width.ratio <- sum.ID.str.width  / .par_devWidth
    
    # debug
    # print(ID.width.ratio)
    
    if(ID.width.ratio > 0.7) {
      id.style <- 'side'
    }
    
    else {
      id.style <- 'top'
    }
    
  }
  
  
  # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  # sketch parameters for follow-up overlay / inspection ----
  # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  lsp <- list('width' = width,
              'plot.order' = plot.order,
              'x0' = relative.pos + x.idx.offset,
              'pIDs' = pIDs[plot.order],
              'idname' = idname(x),
              'horizonDepths' = horizonDepths(x),
              'y.offset' = y.offset,
              'scaling.factor' = scaling.factor,
              'max.depth' = max.depth,
              'n' = n,
              'xlim' = xlim.range,
              'ylim' = ylim.range,
              'extra_x_space' = extra_x_space,
              'extra_y_space' = extra_y_space,
              'hz.depth.LAI' = rep(NA_real_, n),
              'legend.colors' = hz.color.interpretation$colors,
              'legend.data' = hz.color.interpretation$color.legend.data
  )
  
  
  
  ## TODO dynamically adjust `width` based on strwidth(longest.hz.name)
  ## TODO abstract single profile sketch code into a single function
  ## TODO skip sketch rendering when i == `n` outside of length(SPC) (depths are NA)
  
  ## TODO: NO DATA template for empty profiles (now possible)
  
  # extract horizon column names
  cn <- names(h)
  
  # extract / generate horizon name
  m <- match(name, cn)
  
  if(!is.null(name) && is.na(m) && !is.na(name) && nchar(name) > 0) {
    warning("column name '", name, "' not found in horizon names", call. = FALSE)
  }
  
  
  # ++++++++++++++++++++++++++
  # iterate over profiles ----
  # ++++++++++++++++++++++++++
  
  # iterate over profile index from 1 -> n
  # note: there may not be `n` profiles
  for(i in 1:n) {
    # convert linear sequence into plotting order
    # this is NA when accessing beyond length(SPC)
    profile_i <- plot.order[i]
    
    # get truncation flag
    truncation_flag_i <- truncation_flag[profile_i]
    
    # fill NA with FALSE:
    #  - if n > length(x)
    #  - NA present in x$raggedBottom
    truncation_flag_i[is.na(truncation_flag_i)] <- FALSE
    
    # extract the current profile's horizon data
    this_profile_label <- pLabels[profile_i]
    this_profile_id <- pIDs[profile_i]
    this_profile_data <- h[which(h[[IDcol]] == this_profile_id), ]
    
    ## TODO: use empty / template SPC
    # this is an "empty" profile, site data only
    # skip this iteration
    if(nrow(this_profile_data) < 1){
      next
    }
    
    # extract / generate horizon color
    # note: the ".color" horizon attribute is auto-generated above
    # missing and NA colors have already been dealt with above
    this_profile_colors <- this_profile_data$.color
    
    # extract / generate horizon fill density
    if(! missing(density)) {
      # if a single number was given, then recylce it over all horizons
      if(is.numeric(density))
        this_profile_density <- density
      # otherwise we have a column name
      else {
        m <- match(density, cn)
        if(! is.na(m))
          this_profile_density <- this_profile_data[[m]]
        else # user-defined column is missing
          this_profile_density <- NULL
      }
    } else {
      # no user-defined density column
      this_profile_density <- NULL
    }
    
    
    if(!is.na(m)) {
      this_profile_names <- as.character(this_profile_data[[m]])
    } else {
      # otherwise use an empty string
      this_profile_names <- ''
    }
    
    
    
    # ++++++++++++++++++++++++++++++++++++++++
    ## generate baseline horizon geometry ----
    # ++++++++++++++++++++++++++++++++++++++++
    
    ## center of each sketch
    # 2019-07-15: added relative position feature, could use some more testing
    #             y.offset is indexed to plot.order
    #
    # x0 <- x.idx.offset + i
    x0 <- x.idx.offset + relative.pos[i]
    
    # get vectors of horizon boundaries, and scale
    y0 <- (this_profile_data[[bcol]] * scaling.factor) + y.offset[i]
    y1 <- (this_profile_data[[tcol]] * scaling.factor) + y.offset[i]
    
    
    
    # ++++++++++++++++++++++++++++++
    ## create horizons + colors ----
    # ++++++++++++++++++++++++++++++
    
    # horizons are parallelograms
    # offset described by hz.distinctness.offset
    # horizontal mid-points described by hz.topography.offset
    
    # iterate over horizons
    nh <- length(y0)
    
    # extract current distinctness offset
    hdo <- this_profile_data[[hz.distinctness.offset]]
    
    # apply sketch-wide scaling factor
    hdo <- hdo * scaling.factor
    
    # extract current topography offset
    hto <- this_profile_data[[hz.topography.offset]]
    
    # apply sketch-wide scaling factor
    hto <- hto * scaling.factor
    
    # extract current set of line types if provided
    if(! is.null(hz.boundary.lty)) {
      ht.lty <- this_profile_data[, hz.boundary.lty]
    } else {
      # use constant value
      ht.lty <- rep(lty, times = nh)
    }
    
    
    
    # empty list for storing y-coordinates
    coords.list <- vector(mode = 'list', length = nh)
    
    # iterate over horizons
    for(j in 1:nh) {
      
      ## rectangle for reference
      # xx <- c(x0 - width, x0 + width, x0 + width, x0 - width)
      # yy <- c(y0[j], y0[j], y1[j], y1[j])
      
      # parallelogram geometry: x-coords are fixed, y-coords vary based on horizon sequence
      # y0 are bottom depths
      # y1 are top depths
      #
      # vertex order: ll, lc, lr, ur, uc, ul
      x.ll <- x0 - width
      x.lr <- x0 + width
      x.ur <- x0 + width
      x.ul <- x0 - width
      # mid-points for horizon topography
      x.lc <- x0
      x.uc <- x0
      xx <- c(x.ll, x.lc, x.lr, x.ur, x.uc, x.ul)
      
      # make polygons based on 1st, 2nd to j-1, last horizon
      
      # yikes! https://github.com/ncss-tech/aqp/issues/189
      # this doesn't work when: 
      # *  there is only a single horizon [now fixed below]
      # *  there is a "gap" between adjacent horizons (overlapOrGap = TRUE)
      
      if(j == 1 & nh == 1){
        ## first horizon of a single-horizon profile ##
        
        # ragged bottom place-holder
        .r <- NULL
        
        # 
        y.ll <- pmin(y0[j] + hdo[j], y0[j]) # cannot exceed y.ll of next horizon
        y.lr <- pmax(y0[j] - hdo[j], y1[j]) # cannot exceed y.ur of this horizon
        y.ur <- y1[j] # use upper-right verbatim
        y.ul <- y1[j] # use upper-left verbatim
        # mid-points for horizon topography
        y.lc <- pmax(y0[j] - hto[j], y1[j]) # cannot exceed top depth of current horizon
        y.uc <- y1[j] # use upper-center verbatim
        
        ## truncated bottom
        # this is only possible when horizon geometry is available
        # e.g. NOT:
        #  - horizon-less SPC
        #  - n > length(SPC)
        if(truncation_flag_i & !all(is.na(xx))) {
          
          # must be an even number of oscillations
          # computed as function of (ideal oscillations / 2) / sketch density
          # adjusted to width (n.osc increases with width)
          # min value of 6
          # max value of 32
          .raggedN <- round((2.5 * width) * (8 / .dev_sketch_density)) * 2
          .raggedN <- pmax(6, .raggedN)
          .raggedN <- pmin(32, .raggedN)
          
          # ragged bottom line segment: lr -> ll ordering
          .r <- .raggedLines(x1 = x.ll, x2 = x.lr, y = y0[j], o = .raggedOffsets, n = .raggedN)
          
          # modify ragged x and y to include upper right, center, left coords
          xx <- c(.r[, 1], x.ur, x.uc, x.ul)
          yy <- c(.r[, 2], y.ur, y.uc, y.ul)
        } else {
          # place-holder for ragged bottom
          .r <- NULL
          
          # assemble standard y-coords
          yy <- c(y.ll, y.lc, y.lr, y.ur, y.uc, y.ul)
        }
        
        # plot first horizon polygon, without borders
        polygon(x = xx, y = yy, col = this_profile_colors[j], border = NA, density = this_profile_density[j], lend = 1)
        
      } else if(j == 1 & nh > 1){
        ## first horizon, of several ##
        
        # ragged bottom place-holder
        .r <- NULL
        
        # 
        y.ll <- pmin(y0[j] + hdo[j], y0[j+1]) # cannot exceed y.ll of next horizon
        y.lr <- pmax(y0[j] - hdo[j], y1[j]) # cannot exceed y.ur of this horizon
        y.ur <- y1[j] # use upper-right verbatim
        y.ul <- y1[j] # use upper-left verbatim
        # mid-points for horizon topography
        y.lc <- pmax(y0[j] - hto[j], y1[j]) # cannot exceed top depth of current horizon
        y.uc <- y1[j] # use upper-center verbatim
        
        # assemble y-coords and plot first horizon polygon, without borders
        yy <- c(y.ll, y.lc, y.lr, y.ur, y.uc, y.ul)
        polygon(x = xx, y = yy, col = this_profile_colors[j], border = NA, density = this_profile_density[j], lend = 1)
        
      } else if(j < nh) {
        ## next horizons, except bottom-most horizon ##
        
        # ragged bottom place-holder
        .r <- NULL
        
        # 
        y.ll <- pmin(y0[j] + hdo[j], y0[j+1], na.rm = TRUE) # cannot exceed y.ll of next horizon
        y.lr <- pmax(y0[j] - hdo[j], y0[j-1]) # cannot exceed y.lr of previous horizon
        y.ur <- pmax(y1[j] - hdo[j-1], y1[j-1]) # cannot exceed y.ur of previous horizon
        y.ul <- pmin(y1[j] + hdo[j-1], y0[j]) # cannot exceed y.ul of previous horizon
        # mid-points for horizon topography
        y.lc <- pmax(y0[j] - hto[j], y1[j]) # cannot exceed top depth of current horizon
        y.uc <- pmax(y1[j] - hto[j-1], y1[j-1]) # cannot exceed top depth of previous horizon
        
        # assemble y-coords and plot next n horizon's polygon, without borders
        yy <- c(y.ll, y.lc, y.lr, y.ur, y.uc, y.ul)
        polygon(x = xx, y = yy, col=this_profile_colors[j], border=NA, density=this_profile_density[j], lend=1)
        ## debugging
        # polygon(x = xx, y = yy, col=NA, border='red', density=this_profile_density[j], lwd=lwd, lty=lty)
        
      } else {
        ## last horizon ##
        
        # ragged bottom place-holder
        .r <- NULL
        
        #
        y.ll <- y0[j] # user lower-left verbatim
        y.lr <- y0[j] # use lower-right verbatim
        y.ur <- pmax(y1[j] - hdo[j-1], y1[j-1]) # cannot exceed y.ur of previous horizon
        y.ul <- pmin(y1[j] + hdo[j-1], y0[j]) # cannot exceed lower depth of profile
        # mid-points for horizon topography
        y.lc <- y0[j] # truncate to lower depth of last horizon
        y.uc <- pmax(y1[j] - hto[j-1], y1[j-1]) # cannot exceed top of previous horizon
        
        # https://github.com/ncss-tech/aqp/issues/189
        # NA lower horizon depths ll, lc, lr to all be NA
        # this will break divide.hz functionality
        
        ## truncated bottom
        # this is only possible when horizon geometry is available
        # e.g. NOT:
        #  - horizon-less SPC
        #  - n > length(SPC)
        if(truncation_flag_i & !all(is.na(xx))) {
          
          # must be an even number of oscillations
          # computed as function of (ideal oscillations / 2) / sketch density
          # adjusted to width (n.osc increases with width)
          # min value of 6
          # max value of 32
          .raggedN <- round((2.5 * width) * (8 / .dev_sketch_density)) * 2
          .raggedN <- pmax(6, .raggedN)
          .raggedN <- pmin(32, .raggedN)
          
          ## TODO: allow user adjustments via argument
          
          # ragged bottom line segment: lr -> ll ordering
          .r <- .raggedLines(x1 = x.ll, x2 = x.lr, y = y0[j], o = .raggedOffsets, n = .raggedN)
          
          # modify ragged x and y to include upper right, center, left coords
          xx <- c(.r[, 1], x.ur, x.uc, x.ul)
          yy <- c(.r[, 2], y.ur, y.uc, y.ul)
        } else {
          # place-holder for ragged bottom
          .r <- NULL
          
          # assemble standard y-coords
          yy <- c(y.ll, y.lc, y.lr, y.ur, y.uc, y.ul)
        }
        
        # plot last horizon polygon, without borders
        polygon(x = xx, y = yy, col = this_profile_colors[j], border = NA, density = this_profile_density[j], lend = 1)
        
        ## debugging
        # polygon(x = xx, y = yy, col = NA, border = 'red', density=this_profile_density[j], lwd = 2)
      } # end selecting horizon sequence class
      
      # save current iteration of coordinates and line type
      # also save ragged bottom coordinates, usually NULL
      coords.list[[j]] <- list(xx = xx, yy = yy, lty = ht.lty[j], ragged = .r)
      
    } # end looping over horizons
    
    
    ## note: have to do this after the polygons, otherwise the lines are over-plotted
    # optionally divide horizons with line segment
    if(divide.hz) {
      
      # iterate over coordinates
      # coordinate logic: ll, lc, lr, ur, uc, ul
      # line style included
      lapply(coords.list, function(seg) {
        
        # lower left -> lower center
        # segments(x0 = seg$xx[1], y0 = seg$yy[1], x1 = seg$xx[2], y1 = seg$yy[2], lwd = lwd, lty =  seg$lty, lend = 1)
        # lower center -> lower right
        # segments(x0 = seg$xx[2], y0 = seg$yy[2], x1 = seg$xx[3], y1 = seg$yy[3], lwd = lwd, lty = seg$lty, lend = 1)
        
        # add line segments in order
        lines(seg$xx, seg$yy, lwd = lwd, lty = seg$lty, lend = 1)
        
      })
    }
    
    
    ## final border left-side, top, right-side
    # note: when manually specifying n > length(SPC)
    # x0,x1,y0,y1 are NA
    # using `na.rm = TRUE` in the following calls to min() or max() will generate warnings
    
    # left-side, top -> bottom
    suppressWarnings(segments(
      x0 = x0 - width, 
      x1 =  x0 - width, 
      y0 = min(y1, na.rm = TRUE), 
      y1 = max(y0, na.rm = TRUE), 
      lwd = lwd,
      lty = lty,
      lend = 2
    ))
    
    # top, left -> right
    suppressWarnings(segments(
      x0 = x0 - width, 
      x1 =  x0 + width, 
      y0 = min(y1, na.rm = TRUE), 
      y1 = min(y1, na.rm = TRUE), 
      lwd = lwd,
      lty = lty,
      lend = 2
    ))
    
    # right-side, top -> bottom
    suppressWarnings(segments(
      x0 = x0 + width, 
      x1 =  x0 + width, 
      y0 = min(y1, na.rm = TRUE), 
      y1 = max(y0, na.rm = TRUE), 
      lwd = lwd,
      lty = lty,
      lend = 2
    ))
    
    
    ## final bottom-most line or ragged edge
    # no line when divide.hz = TRUE
    # simple line when divide.hz = FALSE
    # ragged edge when profiles are truncated
    
    # bottom, left -> right
    if(!divide.hz) {
      
      # use ragged bottom for truncated profiles
      if(truncation_flag_i) {
        
        # derive from last element of coords.list
        lines(
          x = coords.list[[nh]]$ragged[, 1], 
          y = coords.list[[nh]]$ragged[, 2], 
          lwd = lwd, 
          lty = coords.list[[nh]]$lty, 
          lend = 1
        )
        
      } else {
        
        suppressWarnings(segments(
          x0 = x0 - width,
          x1 =  x0 + width,
          y0 = max(y0, na.rm = TRUE),
          y1 = max(y0, na.rm = TRUE),
          lwd = lwd,
          lty = lty,
          lend = 2
        ))  
        
      }
      
    }
    
    
    
    # ++++++++++++++++++++++++++++++++++
    ## horizon designations (names) ----
    # ++++++++++++++++++++++++++++++++++
    switch(
      name.style,
      'right-center' = {
        # standard annotation
        # offset from right-hand side
        hzname.x0 <- x0 + width + (one.char.width * 0.1)
        # horizon depth mid-point
        hzname.y0 <- ( y1 + y0 ) / 2
        # left-hand / vertical center justification
        hzname.adj <- c(0, 0.5)
        hzname.col <- par('fg') # use whatever the foreground color is
      },
      'left-center' = {
        # inset from left-hand side
        hzname.x0 <- (x0 - width) + (one.char.width * 0.1)
        # horizon depth mid-point
        hzname.y0 <- ( y1 + y0 ) / 2
        # left-hand / vertical center justification
        hzname.adj <- c(0, 0.5)
        # high-contrast labels
        hzname.col <- invertLabelColor(this_profile_colors)
      },
      'left-top' = {
        # soilweb style
        # inset from upper-left corner
        hzname.x0 <- (x0 - width) + (one.char.width * 0.1)
        hzname.y0 <- y1
        # left-hand / vertical top justification
        hzname.adj <- c(0, 1)
        # high-contrast labels
        hzname.col <- invertLabelColor(this_profile_colors)
      },
      'center-center' = {
        # inset from upper-left corner
        hzname.x0 <- x0
        hzname.y0 <- (y1 + y0) / 2
        # center just
        hzname.adj <- c(0.5, 0.5)
        # high-contrast labels
        hzname.col <- invertLabelColor(this_profile_colors)
      },
      'center-top' = {
        # inset from upper-left corner
        hzname.x0 <- x0
        hzname.y0 <- y1
        # center just
        hzname.adj <- c(0.5, 1)
        # high-contrast labels
        hzname.col <- invertLabelColor(this_profile_colors)
      }
      
    )
    
    
    
    ## TODO: use find/fixOverlap() to adjust horizon designations in the presence of collisions (PARDEE)    
    ## See hz depth annotation code below
    
    # optionally shrink the size of names if they are longer than a given thresh
    if(shrink) {
      
      # identify affected horizon names
      names.wide <- which(nchar(this_profile_names) > shrink.cutoff)
      cex.names.shrunk <- rep(cex.names, length(this_profile_data[[tcol]]))
      cex.names.shrunk[names.wide] <- cex.names.shrunk[names.wide] * 0.8
      
      # optionally shrink based on horizon thickness threshold
      if(!is.null(shrink.thin)) {
        
        # identify affected horizon names
        names.thin <- which((y0 - y1) < shrink.thin)
        # only apply shrinkage to those not already shrunk above
        names.thin <- names.thin[which(! names.thin %in% names.wide)]
        # apply shrinkage
        cex.names.shrunk[names.thin] <- cex.names.shrunk[names.thin] * 0.8
      }
      
      # add horizon names
      text(
        x = hzname.x0, 
        y = hzname.y0, 
        labels = this_profile_names, 
        cex = cex.names.shrunk, 
        adj = hzname.adj, 
        col = hzname.col
      )
      
    } else {
      # standard printing of names, all at the same size
      text(
        x = hzname.x0, 
        y = hzname.y0, 
        labels = this_profile_names, 
        cex = cex.names, 
        adj = hzname.adj, 
        col = hzname.col
      )
      
    }
    
    
    # ++++++++++++++++++++++++++++++++++
    ## horizon depth annotation     ----
    # ++++++++++++++++++++++++++++++++++
    if(hz.depths) {
      
      # scaling factor for hz depths
      hz.depths.cex <- cex.names * 0.9
      
      # extra space between profile sketch and left-justified label
      # seems to scale with cex.names
      hz.depths.xfuzz <- strwidth('5', cex = hz.depths.cex) / 2
      
      # device coordinates: scaling / offset applied
      # along with user adjustment
      hzd.txt.x <- x0 + width + hz.depths.xfuzz + hz.depths.offset
      
      ## top-most horizon
      # top depth: vertical alignment is close to "top"
      hzd.txt <- this_profile_data[[tcol]][1]
      text(
        x = hzd.txt.x, 
        y = y1[1], 
        labels = hzd.txt, 
        cex = hz.depths.cex, 
        font = 1, 
        adj = c(0, 0.8)
      )
      
      
      ## in-between horizons, if present: vertical align is "center"
      if(nh > 1) {
        # horizon depth annotation text (no scaling / offset applied here)
        hzd.txt <- this_profile_data[[tcol]][-1]
        
        # device coordinates: scaling / offset applied
        hzd.txt.y <- y1[-1]
        
        
        # sanity checks before attempting label collision repair:
        #  - no NA
        #  - no top == bottom
        if(any(is.na(c(y0, y1)))) {
          .canFixLabelCollision <- FALSE
          message(sprintf('[%s:%s] horizon with top == bottom, cannot fix horizon depth overlap\n consider using repairMissingHzDepths()', this_profile_id, nh))
          
        } else {
          
          # test for top == bottom, cannot perform label collision if this is the case
          if(any(y0 == y1)) {
            .canFixLabelCollision <- FALSE
            message(sprintf('[%s:%s] horizon with top == bottom, cannot fix horizon depth overlap\n consider using repairMissingHzDepths()', this_profile_id, nh))
          } else {
            .canFixLabelCollision <- TRUE
          }
        }
        
        ## collision detection / fix
        if(fixLabelCollisions && .canFixLabelCollision) {
          
          ## TODO: only need to compute this 1 time, move outside of loop
          ## TODO: consider adjusting by scaling.factor
          
          # reasonable threshold for label collision detection
          # depends on aesthetic weighting / graphics device / hz.depths.cex
          y.thresh <- 1.125 * abs(strheight('0', cex = hz.depths.cex))
          # print(y.thresh)
          
          # must include top + bottom depths for collision detection
          # account for the fact that top-most and bottom-most horizon depths are inset
          # based on native units (e.g. cm)
          
          # top vertical buffer should not exceed bottom of first hz
          .vbf <- 0.8
          .vertical_buffer_top <- y1[1] + pmin((.vbf * scaling.factor), min(y1[-1]))
          
          # bottom vertical buffer must be truncated by max.depth
          # this can result in bottom anchor < previous elements in pos
          # remove in next steps
          .vertical_buffer_bottom <- pmin(
            ((max.depth - .vbf) * scaling.factor) + y.offset[i], 
            y0[nh] - (.vbf * scaling.factor)
          )
          
          # add synthetic top / bottom anchors
          .pos <- c(
            .vertical_buffer_top,
            y1[-1],
            .vertical_buffer_bottom
          )
          
          # keep only positions that are < max.depth, after y.offset and scaling
          # these are scaled positions
          .pos <- .pos[which(.pos < ((max.depth * scaling.factor) + y.offset[i]))]
          
          # setup arguments to fixOverlap()
          fixOverlapArgs[['x']] <- .pos
          fixOverlapArgs[['thresh']] <- y.thresh
          
          ## TODO: increase top / bottom q-values
          # https://github.com/ncss-tech/aqp/issues/293
          
          
          # find / fix overlap using electrostatic simulation
          # this includes top/bottom anchor points
          hzd.txt.y.fixed <- suppressMessages(
            do.call('fixOverlap', fixOverlapArgs)
          )
          
          # remove top AND bottom anchors
          hzd.txt.y.fixed <- hzd.txt.y.fixed[-c(1, length(hzd.txt.y.fixed))]
          
          ## this is the Label Adjustment Index (LAI)
          # how much shuffling was performed?
          .LAI <- (hzd.txt.y - hzd.txt.y.fixed)
          # normalize
          .LAI <- sum(abs(.LAI), na.rm = TRUE) / y0[nh]
          
          # keep track of overlap metric for each profile
          lsp[['hz.depth.LAI']][i] <- .LAI
          
          # make changes to annotation label locations
          hzd.txt.y <- hzd.txt.y.fixed
        }
        
        
        # annotate
        text(
          x = hzd.txt.x, 
          y = hzd.txt.y, 
          labels = hzd.txt, 
          cex = hz.depths.cex, 
          font = 1, 
          adj = c(0, 0.5)
        ) 
      }
      
      ## bottom-most horizon
      # bottom depth: vertical alignment is "bottom"
      hzd.txt <- this_profile_data[[bcol]][nh]
      text(
        x = hzd.txt.x, 
        y = y0[nh], 
        labels = hzd.txt, 
        cex = hz.depths.cex, 
        font = 1, 
        adj = c(0, 0)
      )
      
      
      # +++++++++++++++++++++++++++++
      ## hz depth connector lines ---
      # +++++++++++++++++++++++++++++
      if(hz.depths.lines) {
        
        # .x0, .y0: staring coordinates of spike
        # .x1, .y1: ending coordinates of spike + staring coordinates for diagonal member
        # .xfuzz: horizontal offset for labels
        # .txt.x: annotation x-coordinate
        # .txt.y: annotation y-coordinate
        .connectorLines <- function(.x0, .y0, .x1, .y1, .xfuzz, .txt.x, .txt.y) {
          
          # horizontal spike (skipping top + bottom labels)
          # right-edge, at depth -> padded, left-edge/center of annotation (no collision fix)
          segments(
            x0 = .x0, 
            y0 = .y0, 
            x1 = .x1 + (.xfuzz / 2), 
            y1 = .y1
          )
          
          # connector member (skipping top + bottom labels)
          # diagonal when collision fix has been applied
          #
          # right-end of spike member ->
          # padded, left-edge/center of annotation (after collision fix)
          segments(
            x0 = .x0 + (.xfuzz / 2), 
            y0 = .y1,
            x1 = .txt.x - (.xfuzz / 3), 
            y1 = .txt.y, 
          ) 
          
        }
        
        
        ## middle horizon labels, if present
        if(nh > 1) {
          
          .connectorLines(
            .x0 = x0 + width,
            .y0 = y1[-1],
            .x1 = x0 + width,
            .y1 = y1[-1],
            .xfuzz = hz.depths.xfuzz,
            .txt.x = hzd.txt.x,
            .txt.y = hzd.txt.y
          )
        }
        
        
        ## top horizon label
        .connectorLines(
          .x0 = x0[1] + width,
          .y0 = y1[1],
          .x1 = x0[1] + width,
          .y1 = y1[1],
          .xfuzz = hz.depths.xfuzz,
          .txt.x = hzd.txt.x,
          .txt.y = y1[1] - (strheight('0', cex = hz.depths.cex) / 4)
        )
        
        
        ## bottom horizon label
        .connectorLines(
          .x0 = x0[1] + width,
          .y0 = y0[nh],
          .x1 = x0[1] + width,
          .y1 = y0[nh],
          .xfuzz = hz.depths.xfuzz,
          .txt.x = hzd.txt.x,
          .txt.y = y0[nh] + (strheight('0', cex = hz.depths.cex) / 3)
        )
        
      } # end depth annotation lines
      
    } # end horizon depth annotation
    
    
    
    
    ## TODO: move this to the top of profile-process loop, in case there is no hz data
    
    
    # +++++++++++++++++++++
    ## add profile IDs ----
    # +++++++++++++++++++++
    
    # profile is placed relative to the y-offset vector
    
    if(print.id) {
      # optionally abbreviate
      if(abbr)
        id.text <- abbreviate(as.character(this_profile_label), abbr.cutoff)
      
      # no abbreviations of th ID
      else
        id.text <- as.character(this_profile_label)
      
      # add the text: according to style
      if(id.style == 'top') {
        text(x = x0, y = y.offset[i], id.text, pos = 3, font = font.id, cex = cex.id, srt = srt.id, offset = offset.id)
      }
      
      if(id.style == 'side') {
        text(x0 - (width + 0.025), y.offset[i], id.text, adj = c(1, -width), font = font.id, cex = cex.id, srt = 90)
      }
      
    }
    
  } # end looping over profiles
  
  
  # ++++++++++++++++++++
  # add depth axis ----
  # ++++++++++++++++++++
  
  ## 
  ## retain for 2 (?) minor versions
  ##
  
  # handling of aqp 1.x style arguments
  if(!missing(plot.depth.axis)) {
    if(missing(depth.axis)) {
      depth.axis <- plot.depth.axis
    }
    message('`plot.depth.axis` is now deprecated, please use `depth.axis` argument')
  }
  
  if(!missing(cex.depth.axis)) {
    if(inherits(depth.axis, 'list')) {
      depth.axis[['cex']] <- cex.depth.axis
    } else {
      depth.axis <- list(cex = cex.depth.axis)
    }
    message('`cex.depth.axis` is now deprecated, please use `depth.axis` argument')
  }
  
  if(!missing(axis.line.offset)) {
    if(inherits(depth.axis, 'list')) {
      depth.axis[['line']] <- axis.line.offset
    } else {
      depth.axis <- list(line = axis.line.offset)  
    }
    message('`axis.line.offset` is now deprecated, please use `depth.axis` argument')
  }
  
  ##
  ## end backwards compatibility
  ##
  
  # suppress when there are multiple y.offsets
  if(length(unique(y.offset)) > 1) {
    message('depth axis is disabled when more than 1 unique y offsets are supplied')
    depth.axis <- FALSE 
  }
  
  # add depth axis
  if(isTRUE(depth.axis) || is.list(depth.axis)) {
    
    # compose list with defaults if not already
    # NULL `interval` will use sensible defaults
    if(!is.list(depth.axis)) {
      depth.axis <- list(style = 'traditional', line = -2, cex = cex.names * 1.15)
    }
    
    # default style if missing
    if(is.null(depth.axis[['style']])) {
      depth.axis[['style']] <- 'traditional'
    }
    
    # default line if missing
    if(is.null(depth.axis[['line']])) {
      depth.axis[['line']] <- switch(
        depth.axis[['style']],
        'compact' = -1.75,
        'traditional' = -2,
        'tape' = 4
      )
    }
    
    if(is.null(depth.axis[['cex']])) {
      depth.axis[['cex']] <- cex.names * 1.15
    }
    
    # compute nice range for depth axis with sensible interval and max value
    depth_axis_intervals <- .depthAxisSeq(max.depth, i = depth.axis[['interval']])
    
    # convert to plot scale/offset
    depth_axis_tick_locations <- (depth_axis_intervals * scaling.factor) + y.offset[1]
    
    # add depth units
    depth_axis_labels <- paste(depth_axis_intervals, depth_units(x))
    
    # draw axis
    .drawDepthAxis(
      style = depth.axis[['style']],
      .at = depth_axis_tick_locations, 
      .labels = depth_axis_labels,
      .line = depth.axis[['line']], 
      .cex = depth.axis[['cex']]
    )
    
  }
  
  
  
  # +++++++++++++++++++++
  # alternate labels ----
  # +++++++++++++++++++++
  
  if(!missing(alt.label)) {
    al <- site(x)[[alt.label]]
    al <- al[plot.order]
    text(x = 1:length(x), y = y.offset + 3, labels = al, srt = 90, adj = c(1, 0.5), font = 2, cex = cex.id * 1.5, col = alt.label.col)
  }
  
  
  # +++++++++++++++++++++++++++++++++++++++
  # legend for thematic profile sketch ----
  # +++++++++++++++++++++++++++++++++++++++
  
  # optionally show legend
  if(show.legend) {
    # extract from interpretation of horizon colors
    cld <- hz.color.interpretation$color.legend.data
    
    if(! is.null(cld)) {
      # if no title given, set col.label to name of column containing thematic information
      mtext(side=3, text=col.label, font=2, line=1.6)
      
      # gracefully handle all-NA in thematic variable
      if(length(cld$legend) > 0) {
        
        # possibly split legend across multiple rows
        if(cld$multi.row.legend) {
          
          # compute max space required for legend items
          # better formatting
          # note: must be called AFTER high level plot()
          leg.text.width <- (max(strwidth(cld$legend, cex = col.legend.cex)))
          
          # row 1
          legend('bottom', inset = c(0, 0.99),
                 legend = cld$legend[cld$leg.row.indices$row.1],
                 col = cld$col[cld$leg.row.indices$row.1],
                 text.width = leg.text.width,
                 bty = 'n', pch = 15, horiz = TRUE, xpd = TRUE, cex = col.legend.cex, x.intersp = 1
          )
          
          # row 2
          legend('bottom', inset = c(0, 0.94),
                 legend = cld$legend[cld$leg.row.indices$row.2],
                 col = cld$col[cld$leg.row.indices$row.2],
                 text.width = leg.text.width,
                 bty = 'n', pch = 15, horiz = TRUE, xpd = TRUE, cex = col.legend.cex, x.intersp = 1
          )
          
        } else {
          # standard invocation
          legend(
            'bottom', 
            legend = cld$legend, 
            col = cld$col,
            bty = 'n',
            pch = 15,
            horiz = TRUE,
            xpd = TRUE,
            inset = c(0, 0.99),
            cex = col.legend.cex,
            x.intersp = 1
          )
        }
      }
      
    }
  }
  
  # save to aqp env
  assign('last_spc_plot', lsp, envir = aqp.env)
  
  
}

#' generic plot method for \code{SoilProfileCollection} objects
#' @name plot
#' @param y (not used)
#' @aliases plot,SoilProfileCollection-method plot,SoilProfileCollection,ANY-method
#' @rdname SoilProfileCollection-plotting-methods
#' @export
setMethod("plot", signature(x = "SoilProfileCollection"),
          definition = {function(x, y, ...) plotSPC(x, ...)})

context("getLastHorizonID")



test_that("works as expected", {
  
  # IDs sort in non-intuitive alpha-order
  set.seed(10101)
  s <- lapply(1:20, random_profile, SPC = TRUE)
  s <- combine(s)
  res <- getLastHorizonID(s)
  
  # output
  expect_true(inherits(res, 'character'))
  expect_true(length(res) == length(s))
  
  # spot checks
  # plotSPC(s, name = 'hzID')
  expect_true(res[1] == "6")
  expect_true(res[2] == "11")
  expect_true(res[3] == "17")
  
  # ID ordering
  expect_true(all(profile_id(s) == names(res)))
  
  
  # IDs naturally sort in alpha-order
  s <- lapply(sample(letters, size = 26), random_profile, SPC = TRUE)
  s <- combine(s)
  res <- getLastHorizonID(s)
  
  # output with different sorting
  expect_true(inherits(res, 'character'))
  expect_true(length(res) == length(s))
  
  expect_true(all(profile_id(s) == names(res)))
  
})

context("plotProfileDendrogram")

## sample data
suppressWarnings(library(cluster, quietly = TRUE))
suppressWarnings(library(ape, quietly = TRUE))

set.seed(10101)
s <- aqp::combine(lapply(c('Z', LETTERS[1:5]), random_profile, SPC = TRUE))

## tests

test_that("plotProfileDendrogram works as expected", {
  
  # suppressing warnings until we switch to NCSP() and aqp 2.0 is on CRAN
  d <- suppressWarnings(NCSP(s, vars = c('p1','p2'), maxDepth = 100, k = 0))
  dd <- diana(d)
  
  # par(mfrow = c(2,1), mar = c(0,0,0,1))
  # plotSPC(s, color='p1')
  res <- plotProfileDendrogram(s, dd, scaling.factor = 0.8, y.offset = 10, color='p1', debug = TRUE, width = 0.3)
  
  # debug info
  expect_true(inherits(res, 'data.frame'))
  
  # profile ID ordering preserved
  expect_equal(res$profileID, c('A', 'B', 'C', 'D', 'E', 'Z'))
  
  # clustering ID should be the same
  expect_equal(res$clustID, c('A', 'B', 'C', 'D', 'E', 'Z'))
  
  # left -> right dendrogram node labels
  expect_equal(res$clustID.ordered, c('A', 'Z', 'B', 'E', 'D', 'C'))
  
  # profile plotting order
  expect_equal(res$profile.plot.order, c(1, 6, 2, 5, 4, 3))
  
})

context("dice (update to slice)")

data(sp4, package = 'aqp')
depths(sp4) <- id ~ top + bottom

test_that("basic functionality", {
  
  # SPC
  s <- dice(sp4)
  
  # as a data.frame
  s.d <- dice(sp4, SPC = FALSE)
  
  # did it work?
  expect_true(inherits(s, 'SoilProfileCollection'))
  expect_true(inherits(s.d, 'data.frame'))
  
  # new metadata columns
  expect_true(all(c('sliceID', '.oldTop', '.oldBottom') %in% horizonNames(s)))
  expect_true(all(c('sliceID', '.oldTop', '.oldBottom') %in% names(s.d)))
  
  # there should be as many slices as original profiles are deep
  # names should be identical
  expect_equal(
    profileApply(sp4, max),
    profileApply(s, nrow)
  )
  
})

test_that("data integrity, default arguments", {
  
  # SPC
  s <- dice(sp4, SPC = FALSE)
  
  # values match
  # reference horizons
  .ref <- horizons(sp4)
  
  # inner join on original hz ID
  x <- merge(.ref[, c('hzID', 'Mg')], s[, c('hzID', 'Mg')], by = 'hzID', sort = FALSE, all.x = FALSE)
  
  # sliced == original
  expect_equal(x$Mg.x, x$Mg.y)
})

test_that("formula interface", {
  
  # reference
  s <- dice(sp4)
  
  # equivalent to `fm = NULL`
  s1 <- dice(sp4, fm = ~ .)
  expect_equivalent(s, s1)
  
  # LHS
  s2 <- dice(sp4, fm = 0:20 ~ .)
  
  # all should be <= 21 slices
  expect_true(all(profileApply(s2, nrow) <= 21))
  
  # RHS
  s3 <- dice(sp4, fm = ~ Ca)
  
  # all other hz attr should be missing
  hz.diff <- setdiff(horizonNames(sp4), horizonNames(s3))
  expect_false(
    all(
      horizonNames(s3) %in% c("name", "K", "Mg", "CEC_7", "ex_Ca_to_Mg", "sand", "silt", "clay", "CF")
    )
  )

  # LHS + RHS
  s4 <- dice(sp4, fm = 0:30 ~ Ca + K)
  
  # hz names
  expect_false(
    all(
      horizonNames(s4) %in% c("name", "Mg", "CEC_7", "ex_Ca_to_Mg", "sand", "silt", "clay", "CF")
    )
  )
  
  expect_true(all(profileApply(s4, nrow) <= 31))
})

.slices <<- 0:30
test_that("formula interface (with non-standard evaluation)", {
  d5 <- dice(sp4, .slices ~ Ca + K)
  expect_true(inherits(d5, 'SoilProfileCollection'))
})

test_that("discrete slices entirely within SPC", {
  
  # single slice
  s <- dice(sp4, fm = 5 ~ Mg, SPC = FALSE)
  
  # NA should be returned for slices within gaps / below profile bottoms
  expect_true(nrow(s) == length(sp4))
  
  # reference horizons
  .ref <- horizons(sp4)
  
  # inner join on original hz ID
  x <- merge(.ref[, c('hzID', 'Mg')], s[, c('hzID', 'Mg')], by = 'hzID', sort = FALSE, all.x = FALSE)
  
  # sliced == original
  expect_equal(x$Mg.x, x$Mg.y)

  # multiple slices, all within SPC depth interval
  .slices <- c(5, 10, 15)
  s <- dice(sp4, fm = c(5, 10, 15) ~ Mg, SPC = FALSE)
  
  # NA should be returned for slices within gaps / below profile bottoms
  expect_true(nrow(s) == length(sp4) * length(.slices))
  
  # reference horizons
  .ref <- horizons(sp4)
  
  # inner join on original hz ID
  x <- merge(.ref[, c('hzID', 'Mg')], s[, c('hzID', 'Mg')], by = 'hzID', sort = FALSE, all.x = FALSE)
  
  # sliced == original
  expect_equal(x$Mg.x, x$Mg.y)
  
})

test_that("slices below bottom of profiles or entire collection", {
  
  # single slice, deeper than some profiles
  s <- dice(sp4, fm = 25 ~ Mg, SPC = FALSE)
  
  # NA should be returned for slices within gaps / below profile bottoms
  # filled horizons will create hzID beyond original sequence
  expect_true(nrow(s) == length(sp4))
  
  # there should be 3 NA
  expect_true(length(which(is.na(s$Mg))) == 3)
  
  # reference horizons
  .ref <- horizons(sp4)
  
  # inner join on original hz ID
  x <- merge(s[, c('hzID', 'Mg')], .ref[, c('hzID', 'Mg')], by = 'hzID', sort = FALSE, all.x = TRUE)
  
  # after NA-removal
  x <- na.omit(x)
  # sliced == original
  expect_equal(x$Mg.x, x$Mg.y)
  
  # single slice, deeper than all profiles
  s <- dice(sp4, fm = 75 ~ Mg, SPC = FALSE)
  
  # NA should be returned for slices within gaps / below profile bottoms
  expect_true(nrow(s) == length(sp4))
  
  # there should be as many NA as profiles in sp4
  expect_true(length(which(is.na(s$Mg))) == length(sp4))
  
  # multiple slices, some beyond profile depths
  .slices <- c(5, 10, 15, 50, 100)
  s <- dice(sp4, fm = c(5, 10, 15, 50, 100) ~ Mg, SPC = FALSE)
  
  # NA should be returned for slices within gaps / below profile bottoms
  expect_true(nrow(s) == length(sp4) * length(.slices))
  
  # reference horizons
  .ref <- horizons(sp4)
  
  # inner join on original hz ID
  x <- merge(s[, c('hzID', 'Mg')], .ref[, c('hzID', 'Mg')], by = 'hzID', sort = FALSE, all.x = TRUE)
  
  # after NA-removal
  x <- na.omit(x)
  # sliced == original
  expect_equal(x$Mg.x, x$Mg.y)
  
})

test_that("percent missing calculation", {
  
  sp4$K[c(1, 5, 10, 30)] <- NA
  sp4$Ca[c(1, 5, 20, 30)] <- NA
  sp4$CEC_7[c(30)] <- NA
  
  s <- dice(sp4, fm = ~ Ca + K + CEC_7, pctMissing = TRUE)
  
  # visual check
  plotSPC(s, color = 'Ca')
  plotSPC(s, color = 'K')
  plotSPC(s, color = '.pctMissing')
  
  # new column
  expect_true('.pctMissing' %in% horizonNames(s))
  
  # some should be non-zero
  expect_true(any(s$.pctMissing > 0) & !all(s$.pctMissing > 0))
  
  # check exact values
  # 1st horizon, 2/3 missing
  expect_true(all(s[1, ]$.pctMissing[1:3] == 2/3))
  # all horizons, 1/3 missing
  expect_true(all(s[7, ]$.pctMissing == 1/3))
  # last horizon, 100% missing
  expect_true(all(s[10, ]$.pctMissing[8:16] == 1))
})


test_that("padding with NA, backwards-compatible with slice", {
  
  # fill = TRUE is implied with formula interface
  s <- dice(sp4, fm = 0:80 ~ .)
  
  # all profiles should be the same "depth", including empty (NA) horizons
  expect_true(all(profileApply(s, max) == 81))
  
})


test_that("testing exact values", {
  
  # typical soil profile, no problem horizons
  x <- data.frame(
    id = 'A',
    top = c(0, 10, 15, 20, 40, 50, 100),
    bottom = c(10, 15, 20, 40, 50, 100, 165),
    p = 1:7
  )
  
  # init SPC
  depths(x) <- id ~ top + bottom
  
  # dice with defaults
  s <- dice(x)
  
  # horizon thickness
  thick <- x$bottom - x$top
  
  # check no. contiguous sliced values = horizon thickness
  expect_true(all(rle(s$p)$lengths == thick))
  
  # check values within chunks
  expect_true(all(rle(s$p)$values == x$p))
  
  # do it again with a formula
  s <- dice(x, fm = 0:106 ~ .)
  
  # the last horizon is truncated to 107cm
  expect_true(rle(s$p)$lengths[7] == 7)
  
})

test_that("overlapping horizons", {
  # overlapping horizons results in increased number of slices (according to overlapping thickness)
  x1 <- horizons(dice(sp4, ~ .))
  
  # create overlap
  sp4@horizons[2,]$bottom <- sp4@horizons[2,]$bottom + 12
  
  x2 <- horizons(dice(sp4, ~ .))
  
  # evaluate logic by profile, not horizon
  expect_message({ x3 <- horizons(dice(sp4, ~ ., byhz = FALSE)) })
  
  # default case--nothing removed, nothing added
  expect_equal(nrow(x1), 331)
  
  # 12cm of overlap
  expect_equal(nrow(x2), 331 + 12)
  
  # 1 profile removed due to overlap
  expect_equal(nrow(x3), 289)
})

test_that("perfectly overlapping horizons", {
  
  # two overlapping horizons
  z <- data.frame(
    id = 'SPC',
    top = c(0, 25, 25, 50, 75, 100, 100),
    bottom = c(25, 50, 50, 75, 100, 125, 125),
    property = 1:7
  )
  
  depths(z) <- id ~ top + bottom
  
  # formula specified
  # single slice within zone of overlapping horizons
  d <- dice(z, fm = 30 ~ property , SPC = FALSE)
  
  # two rows, one for each overlapping horizon
  expect_equal(nrow(d), 2)
  
  # slices within and beyond zone of overlap
  d <- dice(z, fm = 40:54 ~ property , SPC = FALSE)
  
  # 10 slices of hzID 2/3
  # 5 slices of hzID 4
  tab <- table(d$hzID)
  expect_equal(as.vector(tab), c(10, 10, 5))
  
})

test_that("dropped profile IDs", {
  
  # corrupt depth
  sp4$top[5] <- sp4$bottom[5]
  
  # offending horizons removed
  expect_message(s <- dice(sp4, byhz = TRUE))
  
  expect_equal(
    setdiff(profile_id(sp4), profile_id(s)),
    character(0)
  )
  
  # offending profiles removed
  expect_message(s <- dice(sp4, byhz = FALSE))
  
  # single dropped ID should be present in metadata
  expect_equal(
    setdiff(profile_id(sp4), profile_id(s)),
    metadata(s)$removed.profiles
  )
    
})

test_that("safely handle hzdesgnname", {
  
  # hzdesgnname is not set
  d <- dice(sp4, ~ K)
  expect_equal(hzdesgnname(d), '')
  
  # init hzname
  hzdesgnname(sp4) <- 'name'
  
  # hzdesgnname is not set
  d <- dice(sp4, ~ K)
  expect_equal(hzdesgnname(d), 'name')
  
  # hzdesgnname retained
  d <- dice(sp4, ~ .)
  expect_equal(hzdesgnname(d), 'name')
  
  # hzdesgnname retained
  d <- dice(sp4, ~ name + K)
  expect_equal(hzdesgnname(d), 'name')
  
})

# NOTE: split,SoilProfileCollection method is redundant aside from the ability to enter a site var by name
#
# data("loafercreek", package = "soilDB")
#
# loafercreek <- subset(loafercreek, !is.na(pmorigin))
# loafercreek$pmorigin <- factor(loafercreek$pmorigin)
#
# # split.default(SPC) and the aqp method are identical due to overloading of `[`
# microbenchmark::microbenchmark( default = { split.default(loafercreek, site(loafercreek)$pmorigin) },
#                                 aqp =     { split(loafercreek, "pmorigin") })
#

#' @title Split a SoilProfileCollection object into a list of SoilProfileCollection objects.
#'
#' @description This function splits a `SoilProfileCollection` into a list of `SoilProfileCollection` objects using a site-level attribute to define groups or profile ID (`idname(x)`).
#'
#' @param x `SoilProfileCollection` object
#' @param f character vector naming a single site-level attribute that defines groups, a ‘factor’ in the sense that `as.factor(f)` defines the grouping, or a list of such factors in which case their interaction is used for the grouping.
#' @param drop logical indicating if levels that do not occur should be dropped (if f is a factor or a list). When `drop=FALSE` and `f` contains missing values an additional group "missing" is returned.
#' @param ...	additional arguments are ignored
#'
#' @details As of aqp 1.25, omission of `f` argument is no longer possible, as the base R generic is overloaded by this `SoilProfileCollection` method. This used to result in an "identity" split, according to `idname(x)`, e.g. a list as long as `length(x)`, with a single-profile `SoilProfileCollection` per list element. Replicate this behavior using `f = idname(x)` or `f = profile_id(x)`.
#'
#' @author D.E. Beaudette and A.G. Brown
#'
#' @return A list of `SoilProfileCollection` or `NULL` for empty result.
#' @export
#'
#' @examples
#'
#' data(sp2)
#' depths(sp2) <- id ~ top + bottom
#'
#' # add a more interesting site-level attribute
#' site(sp2) <- ~ surface
#'
#' # using identity site-level attribute (profile ID)
#' p1 <- split(sp2, f = idname(sp2))
#' names(p1)
#' length(p1)
#'
#' # using vector equal in length to number of profiles (profile ID, again)
#' p2 <- split(sp2, f = profile_id(sp2))
#' names(p2)
#' length(p2)
#'
#' # which are both equivalent to setting `f` to NULL
#' p3 <- split(sp2, f = NULL)
#' names(p3)
#' length(p3)
#'
#' # split on surface (age) site-level var
#' p4 <- split(sp2, f = "surface")
#' names(p4)
#' length(p4) # 5 unique "surfaces", 5 SPCs in result list
#'
setMethod("split",
          signature(x = "SoilProfileCollection", f = "ANY"),
          function(x, f, drop = TRUE, ...) {

  # identity split, use idname()
  if (is.null(f)) {

    # grouping factor, make sure to use original ordering
    fg <- site(x)[[idname(x)]]
    fg <- factor(fg, levels = fg)

  } else {
    # standard, site-level group split
    # using an existing site-level attribute
    if (is.character(f) && length(f) == 1) {
      if (!f %in% siteNames(x)) {
        stop(sprintf('%s must be site-level attribute', f), call. = FALSE)
      }
      
      # extract to local variable, so as not to modify original data
      fg <- x[[f]]
      
      # all NA, empty list is the result
      if(all(is.na(fg)) && drop) {
        message('all groups have been dropped due to NA')
        # nothing left to do
        return(list())
      }
      
      # no NA allowed
      if (any(is.na(fg)) && !drop) {
        fg[is.na(fg)] <- "<missing>"
      }
   
      # splitting variable should be a factor
      if (!inherits(fg, 'factor')) {
        fg <- factor(fg)
        message(sprintf('converting %s to factor', f))
      }
  
    # using a vector coercible to factor (like base::split)
    } else if (length(f) == length(x) && !is.list(f)) {

      # preserve NA levels as missing
      if (any(is.na(f)) && !drop) {
        if (is.factor(f)) {
         levels(f) <- c(levels(f), "<missing>")
        }
        f[is.na(f)] <- "<missing>"
      }

      if (!inherits(f, 'factor')) {
        fg <- factor(f)
      } else {
        fg <- f
      }
    } else if (is.list(f)) {

      # preserve NA levels as missing
      f <- lapply(f, function(ff) {
        if (any(is.na(ff)) && !drop) {
          if (is.factor(ff)) {
            levels(ff) <- c(levels(ff), "<missing>")
          }
          ff[is.na(ff)] <- "<missing>"
        }
        ff
      })
      
      # using a interaction of a list coercible to factor (like base::split)
      fg <- interaction(f, drop = drop, sep = ".", lex.order = FALSE)

    } else {
      stop(sprintf('invalid argument `f`'), call. = FALSE)
    }
  }

  # iterate over levels
  lv <- levels(fg)

  # index and split
  res <- lapply(lv, function(i) {

    # simple indexing on site-level data only
    rr <- x[which(fg == i), ]

    return(rr)
  })
  
  # save names
  names(res) <- lv

  # result is a list
  return(res)
})


# #note: RE: identity split with f not specified
#        unfortunately this S4 method for SPC,missing does not work
# #      f is required in the generic (no default)
# #      neither base nor data.table provide a default for f
#
# setMethod(f = "split",
#           signature(x = "SoilProfileCollection", f = "missing"),
#           function(x, f = NULL, drop = TRUE, ...) {
#               aqp::split(x, f = NULL, drop, ...)
#           })
#
# # sadly this doesnt work either :( (S3 definition where f has default)
# # alas I tried...
#
# split.SoilProfileCollection <- function(x, f = NULL, drop = TRUE, ...) {
#   aqp::split(x = x, f = f, drop = drop, ...)
# }

## S4 magic
# already exists, but we are modifying it... good idea or bad idea?
# setGeneric("split", function(x, f=NULL, drop=TRUE, ...) standardGeneric("split"))
#
#  we should not over-write generics if they exist. we use same method signature
#  but dispatch to x = SoilProfileCollection so we dont mask.
#  note that this causes problems for the identity split with no default `f` agb 2020/09/28

# what is the most likely sequence, given a markovchain and initial state
# the result isn't likely correct when there are non-zero ties in t
#' @param mc Passed to `markovchain` `conditionalDistribution()`
#' @param t0 Passed to `markovchain` `conditionalDistribution()`
#' @param maxIterations Maximum number of iterations. Default: `10`
#'
#' @export
#' @rdname hzTransitionProbabilities
mostLikelyHzSequence <- function(mc, t0, maxIterations = 10) {

  if(!requireNamespace('markovchain', quietly = TRUE))
    stop('please install the `markovchain` package.', call. = FALSE)

  # check for ties
  if(!is.null(attr(as(mc, 'matrix'), 'ties')))
    if(attr(as(mc, 'matrix'), 'ties'))
      warning('ties in transition probability matrix, results may not be reliable', call. = FALSE)

  # store sequence here
  s <- vector(mode = 'character')
  # save the intial state
  i <- 1
  s[i] <- t0
  # compute probabilities for the next state
  cd <- markovchain::conditionalDistribution(mc, t0)

  # search for all but the current state
  not.this.state <- which(!names(cd) == t0)

  ## TODO: this doesn't work when there are ties
  next.state <- names(which.max(cd[not.this.state]))
  i <- i + 1
  s[i] <- next.state
  # continue searching for the most likely next state
  # until reaching the second absorbing state
  # the first absorbing state is retained in 's'
  while(! next.state %in% markovchain::absorbingStates(mc)) {
    cd <- markovchain::conditionalDistribution(mc, next.state)
    # search for all but the current state
    not.this.state <- which(!names(cd) == next.state)
    next.state <- names(which.max(cd[not.this.state]))
    i <- i + 1
    s[i] <- next.state
    # trap run-away iteration
    if(i > maxIterations)
      break
  }
  
  return(s)
}

#' @title Visualize soil colors in Munsell notation according to within-group frequency.
#' @param m character vector of color in Munsell notation ('10YR 4/6')
#' 
#' @param g factor describing group membership, typically a generalization of horizon designation, default value will generate a fake grouping that covers all of the colors in `m`
#' 
#' @param size logical, encode group-wise frequency with chip size
#' 
#' @param annotate logical, annotate color chip frequency
#' 
#' @param launderColors logical, fix common errors with encoding neutral colors, see [launderMunsell()]
#' 
#' @param chip.cex scaling factor applied to each color chip
#' 
#' @param chip.cex.min lower limit for color chip frequency depiction
#' 
#' @param chip.cex.max lower limit for color chip frequency depiction
#' 
#' @param chip.border.col color for chip borders (outline)
#' 
#' @param annotate.cex scaling factor for chip frequency annotation
#' 
#' @param annotate.type character, within-group `count` or `percentage`
#'
#' @param threshold numeric within 0-1, color chips with proportion `< threshold` are removed
#'
#' @return a `trellis` object
#' 
#' @export
#' 
#' @examplesIf requireNamespace('latticeExtra') && requireNamespace('scales')
#' 
#' library(lattice)
#' 
#' # two hue pages
#' ric <- expand.grid(
#'   hue = c('5YR', '7.5YR'),
#'   value = 2:8,
#'   chroma = 2:8
#' )
#' 
#' # combine hue, value, chroma into standard Munsell notation
#' ric <- sprintf("%s %s/%s", ric$hue, ric$value, ric$chroma)
#' 
#' # note that chip frequency-based size is disabled
#' # because all chips have equal frequency
#' colorChart(ric, chip.cex = 4, size = TRUE)
#' 
#' # annotation of frequency
#' colorChart(ric, chip.cex = 4, annotate = TRUE)
#' 
#' # bootstrap to larger size
#' ric.big <- sample(ric, size = 100, replace = TRUE)
#' 
#' # frequency can be encoded in size
#' colorChart(ric.big, chip.cex = 3)
#' colorChart(ric.big, chip.cex = 5, annotate = TRUE)
#' 
#' # constant size
#' colorChart(ric.big, chip.cex = 3, size = FALSE)
#' colorChart(ric.big, chip.cex = 3, size = FALSE, chip.border.col = 'NA')
#' 
#' # simulate colors based dE00 thresholding
#' p <- list(
#'   list(m = '10YR 4/4', thresh = 10, hues = c('10YR', '7.5YR'))
#' )
#' 
#' # perform 500 simulations
#' s <- simulateColor(method = 'dE00', n = 500, parameters = p)
#' 
#' # result is a list, use the first element
#' colorChart(s[[1]], chip.cex = 4)
#' 
#' # increase the possible range of color chip sizes
#' colorChart(s[[1]], chip.cex = 4, chip.cex.min = 0.01, chip.cex.max = 2)
#' 
#' # demonstrate support for neutral chips
#' N <- sprintf('N %s/', 2:8)
#' cols <- c(rep(N, times = 5), ric.big)
#' 
#' # note special panel used to show neutral chips
#' colorChart(cols, size = FALSE, annotate = TRUE)
#' 
#' # filter proportions below given threshold
#' colorChart(cols, size = FALSE, annotate = TRUE, threshold = 0.01,
#'            chip.cex = 4, annotate.type = 'percentage')
#'            
#' # extreme values
#' colorChart(c('N 8/', '10YR 4/12', '10YR 3/1', '10YR 10/4'))
#' 
#' # unusual notation, 10YR 4/0 --> effectively N 4/
#' colorChart(c('N 8/', 'N 4/', '10YR 4/0', '5G 6/6', '5Y 10/4'))
#' 
#' # demonstrate groups
#' m <- c('10YR 4/4', 'N 2/', 'N 3/', '5Y 6/6', '5P 4/4')
#' g <- factor(c(1, 1, 2, 2, 2))
#' colorChart(m, g)
#' 
colorChart <- function(m, g = factor('All'), size = TRUE, annotate = FALSE, launderColors = TRUE, chip.cex = 3, chip.cex.min = 0.1, chip.cex.max = 1.5, chip.border.col = 'black', annotate.cex = chip.cex * 0.25, annotate.type = c('count', 'percentage'), threshold = NULL) {
  
  # requires latticeExtra and scales
  if(!requireNamespace('latticeExtra', quietly = TRUE) || !requireNamespace('scales', quietly = TRUE)) {
    stop('please install the `latticeExtra` and `scales` packages.', call. = FALSE)
  }
  
  # custom panel function defined here, so that it can "find" data within the colorChart function scope
  panel.colorChart <- function(x, y, subscripts = subscripts, ...) {
    
    # panel-wise subset of required data
    p.data <- tab[subscripts, ]
    p.data$x <- x
    p.data$y <- y
    
    # grid lines: use original data
    if(nrow(p.data) > 0) {
      panel.abline(
        v = unique(p.data$x), 
        h = unique(p.data$y), 
        col = grey(0.85), 
        lty = 3
      )
    }
    
    
    # plot single instances of each chip, transparency weighted by frequency
    panel.points(
      p.data$x, 
      p.data$y, 
      pch = 22, 
      col = chip.border.col,
      fill = p.data$transformed.col, 
      cex =  p.data$chip.size
    )
    
    # annotate chips
    if(annotate) {
      
      # count or percentage
      # these are computed by group NOT by panel
      ann.txt <- switch(
        annotate.type,
        'count' = round(p.data$count),
        'percentage' = signif(100 * p.data$prop, 2)
      )
      
      # adjust based on lightness
      anno.col <- invertLabelColor(p.data$.color)
      
      # within-group frequency
      panel.text(
        x = p.data$x, 
        y = p.data$y, 
        labels = ann.txt, 
        cex = annotate.cex, 
        col = anno.col
      )
    }
    
  }
  
  
  # annotation type
  if(annotate) {
    annotate.type <- match.arg(annotate.type)
  }
  
  # optionally launder Munsell notation:
  #  - interpret poorly specified N chips
  m <- launderMunsell(m)
  
  
  # extract pieces / convert colors
  z <- data.frame(
    .munsell = m,
    .groups = g
  )
  
  # remove obvious NA in colors or groups
  z <- na.omit(z)
  
  # bogus colors can creep-in when composing Munsell notation from pieces
  idx <- which(validateMunsell(z$.munsell))
  z <- z[idx, ]
  
  # within-group counts + proportions
  tab <- as.data.frame(table(z$.groups, z$.munsell))
  names(tab) <- c('.groups', '.munsell', 'count')
  
  # group totals
  group.totals <- as.data.frame(table(z$.groups))
  names(group.totals) <- c('.groups', 'total')
  
  # combine
  tab <- merge(tab, group.totals, by = '.groups', all.x = TRUE, sort = FALSE)
  
  # within-group proportions
  tab$prop <- tab$count / tab$total
  
  # filter 0s
  tab$prop <- ifelse(is.na(tab$prop), 0, tab$prop)
  tab <- tab[which(tab$prop > 0), ]
  
  # apply optional threshold on proportions
  if( !is.null(threshold) ) {
    # sanity check
    if(threshold < 0 && threshold > 1) {
      message('ignoring out of bounds threshold')
    } else {
     # apply threshold
      tab <- tab[which(tab$prop > threshold), ]
    }
  }
  
  # check for 0 rows (too much filtering)
  if(nrow(tab) < 1) {
    stop('all rows have been removed, consider setting `threshold` higher', call. = FALSE)
  }
  
  # convert colors
  tab$.color <- parseMunsell(tab$.munsell)
  
  ## TODO: probably a cleaner way
  pm <- parseMunsell(munsellColor = tab$.munsell, convertColors = FALSE)
  tab$hue <- pm$hue
  tab$value <- pm$value
  tab$chroma <- pm$chroma
  
  # remove missing colors / groups
  idx <- which(complete.cases(tab[, c('hue', 'value', 'chroma', '.groups')]))
  tab <- tab[idx, ]
  
  # encode hue as factor using standard hue order
  # second call to factor() drops unused levels
  # note special argument to include neutral hues
  # N is in position 41, we need it in position 1
  hp <- huePosition(returnHues = TRUE, includeNeutral = TRUE)
  N.idx <- match('N', hp)
  hp <- c(hp[N.idx], hp[-N.idx])
  
  tab$hue <- factor(
    factor(tab$hue, levels = hp)
  )
  
  # disable variable size when all frequencies are the same (no useful information)
  if(length(unique(tab$count)) == 1) {
    no.differeneces <- TRUE
  } else {
    no.differeneces <- FALSE
  }
  

  # variable chip size
  if(size) {
    if(no.differeneces) {
      tab$chip.size <- chip.cex
    } else {
      tab$chip.size <- sqrt(scales::rescale(tab$prop, to = c(chip.cex.min, chip.cex.max))) * chip.cex
    }
  } else {
    tab$chip.size <- chip.cex
  }
  
  ## TODO: thematic coloring based on density
  ## needs to be adjusted a group at a time
  # .f <- scales::col_numeric(viridis::viridis(100), domain = c(0, max(tab$prop)))
  # tab$transformed.col <- .f(tab$prop)
  
  # chip color is not modified for now
  tab$transformed.col <- tab$.color
  
  ## TODO: 
  # * consider reporting Shannon entropy / group
  #   print(shannonEntropy(tab$prop))
  
  
  # set reasonable [x, y] limits: 
  # value / chroma: 2-8 unless data extend beyond those limits
  # exception for N hues (chroma 0), typical min value is chroma of 1 (not always)
  x.lim <- c(
    pmin(min(tab$chroma[which(!tab$hue == 'N')]), 2) - 0.5,
    pmax(max(tab$chroma), 8) + 0.5
  )
  
  y.lim <- c(
    pmin(min(tab$value), 2) - 0.5,
    pmax(max(tab$value), 8) + 0.5
  )
  
  # assemble basic plot
  pp <- lattice::xyplot(
    value ~ chroma | hue + .groups, 
    data = tab,
    drop = TRUE, 
    as.table = TRUE,
    subscripts = TRUE, 
    # this works until we attempt to use neutral hues
    xlim = x.lim,
    ylim = y.lim,
    # simple version when no neutral hues
    scales = list(
      x = list(alternating = 1), 
      y = list(rot = 0, alternating = 3)
    ), 
    main = '', 
    xlab = 'Chroma', 
    ylab = 'Value', 
    panel = panel.colorChart
  )
  
  
  ## get hues to iterate over
  ll <- levels(tab$hue)
  
  ## in the presence of N chips, custom x-axis and limits, and panel size required
  if('N' %in% ll) {
    
    ## NOTE: with complex hue x group structure, in the presence of N chips
    #        we need to compute panel-specific limits and axes labels
    #        applying these custom parameters won't work if any panel has no data
    #        therefore, generate fake data for every combination of hue x groups
    # 
    #        see:
    #        https://stackoverflow.com/questions/11161739/how-to-specify-different-ylim-values-for-each-panel-in-xyplot
    #
    #
    # side-effect: unused levels must be removed prior to calling colorChart()
    #
    # solution:    keep track of panels with data, then assigning scales / limits accordingly
    
    # (hue x group) combinations before adding fake obs
    int.b <- levels(interaction(tab$hue, tab$.groups, drop = TRUE))
    
    ## HACK: add a single bogus point for every hue x group combination
    #        this ensures that limits are correctly established
    #        only required when there are neutral colors
    .fake <- expand.grid(hue = levels(tab$hue), .groups = levels(tab$.groups))
    
    .fake <- data.frame(
      .groups = .fake$.groups,
      .munsell = '', 
      count = 1, 
      total = 1, 
      prop = 0,
      .color = '#FFFFFFFF',
      hue = .fake$hue,
      value = 100, 
      chroma = 100,
      chip.size = 0.1,
      transformed.col = '#FFFFFFFF'
    )
    
    tab <- rbind(tab, .fake)
    
    ## end HACK
    
    # (hue x group) combinations after adding fake obs
    int.a <- levels(interaction(tab$hue, tab$.groups))
    
    
    # standards x-limits
    x.at <- seq(x.lim[1] + 0.5, x.lim[2] - 0.5, by = 1)
    
    # replicate as many times are there are columns
    # first column (N) gets no axes
    x.at.list <- c(
      list(''), 
      lapply(1:(length(ll)-1), FUN = function(i) {x.at})
    )
    
    # limits for as many columns
    x.limits.list <- c(
      list(c(0, 0)), 
      lapply(1:(length(ll)-1), FUN = function(i) {x.lim})
    )
    
    
    # replicate scales to cover all columns * rows
    n.groups <- length(levels(tab$.groups))
    x.at.list <- rep(x.at.list, times = n.groups)
    x.limits.list <- rep(x.limits.list, times = n.groups)
    
    # index just those panels (hue x group) with no data (before adding fake obs)
    nd.idx <- which(is.na(match(int.a, int.b, nomatch = NA)))
    
    # suppress scales in these panels
    x.at.list[nd.idx] <- list('')
    
    pp <- lattice::xyplot(
      value ~ chroma | hue + .groups, 
      data = tab,
      as.table = TRUE,
      subscripts = TRUE, 
      # panel-specific limits gets complicated
      scales = list(
        alternating = 3, 
        y = list(rot = 0, limits = y.lim), 
        x = list(
          relation = 'free', 
          at = x.at.list,
          limits = x.limits.list
        )
      ), 
      main = '', 
      xlab = 'Chroma', 
      ylab = 'Value', 
      panel = panel.colorChart
    )
    
    # proportional panel sizes, neutral gets 20% of normal space
    pp <- latticeExtra::resizePanels(pp, w = c(0.2, rep(1, times = length(ll) - 1)))
  }
  
  # move paneling by groups to outer panels
  pp <- latticeExtra::useOuterStrips(
    pp, 
    strip = strip.custom(bg = grey(0.85), par.strip.text = list(cex=0.85)), 
    strip.left = strip.custom(bg = grey(0.85), par.strip.text = list(rot=0))
  )
  
  if(annotate) {
    sub.txt <- sprintf('chip labels represent %ss', annotate.type)
    pp <- update(pp, sub = list(sub.txt, font = 1))
  }
  
  return(pp)  
}

#' @title Plot aggregate soil color data
#' 
#' @description Generate a plot from summaries generated by `aggregateColor()`.
#'
#' @param x a `list`, results from `aggregateColor()`
#' @param print.label logical, print Munsell color labels inside of rectangles, only if they fit 
#' @param label.font font specification for color labels
#' @param label.cex font size for color labels
#' @param label.orientation label orientation, `v` for vertical or `h` for horizontal
#' @param buffer.pct extra space between labels and color rectangles
#' @param print.n.hz optionally print the number of horizons below Munsell color labels
#' @param rect.border color for rectangle border
#' @param horizontal.borders optionally add horizontal borders between bands of color
#' @param horizontal.border.lwd line width for horizontal borders
#' @param x.axis logical, add a scale and label to x-axis?
#' @param y.axis logical, add group labels to y-axis?
#' @param ... additional arguments passed to `plot`
#'
#' @details See the [related tutorial for additional examples](http://ncss-tech.github.io/AQP/sharpshootR/aggregate-soil-color.html).
#' 
#' @author D.E. Beaudette
#'
#' @return no data are returned, function is called for graphical output
#' @export
#' 
#' @keywords hplots
#'
#' @examples
#' 
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' # load some example data
#' data(sp1, package = 'aqp')
#' 
#' # upgrade to SoilProfileCollection and convert Munsell colors
#' sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#' 
#' # generalize horizon names
#' n <- c('O', 'A', 'B', 'C')
#' p <- c('O', 'A', 'B', 'C')
#' sp1$genhz <- generalize.hz(sp1$name, n, p)
#' 
#' # aggregate colors over horizon-level attribute: 'genhz'
#' a <- aggregateColor(sp1, groups = 'genhz', col = 'soil_color')
#' 
#' # check results
#' str(a)
#' 
#' # simple visualization
#' aggregateColorPlot(a)
#' 
aggregateColorPlot <- function(x, print.label=TRUE, label.font=1, label.cex=0.65, label.orientation = c('v', 'h'), buffer.pct=0.02, print.n.hz=FALSE, rect.border='black', horizontal.borders=FALSE, horizontal.border.lwd=2, x.axis=TRUE, y.axis=TRUE, ...) {
  
  # sanity check
  label.orientation <- match.arg(label.orientation)
  
  # extract just the scaled data from the results of aggregateColor()
  s.scaled <- x$scaled.data
  
  # get max re-scaled summation for xlim
  max.plot <- max(sapply(s.scaled, function(i) sum(i$weight)))
  
  # setup plot
  plot(1, 1, type = 'n', xlim = c(0, max.plot), ylim = c(length(names(s.scaled))+0.5, 0.5), axes = FALSE, ylab = '', xlab = 'Cumulative Proportion', col.main = par('fg'), col.lab = par('fg'), ...)
  
  # iterate over horizons
  for(i in seq_along(names(s.scaled))) {
    
    # current iteration
    s.i <- s.scaled[[i]]
    n.colors <- nrow(s.i)
    
    if(n.colors > 0) {
      # get an index to the last weight
      last.weight <- length(s.i$weight)
      # compute cumulative left / right rectangle boundaries
      x.left <- cumsum(c(0, s.i$weight[-last.weight]))
      x.right <- c(x.left[-1], x.left[last.weight] + s.i$weight[last.weight])
      
      # plot rectangles from vectorized coordinates / colors
      # first column in each chunk is the R color
      rect(xleft=x.left, ybottom=i-0.5, xright=x.right, ytop=i+0.5, col=s.i[, 1], border=rect.border)
      
      # compute center point for color labels
      centers <- (x.right + x.left) / 2
      
      # create label
      if(print.n.hz) {
        # with number of horizons
        color.labels <- paste0(s.i$munsell, '\n', '(', s.i$n.hz, ')')
      } else {
        # just colors
        color.labels <- s.i$munsell
      }
      
      
      # determine if there is enough room to label colors: some % of visible space on plot
      # first, get the plot aspect ratio
      plot.w <- par("pin")[1]/diff(par("usr")[1:2])
      plot.h <- par("pin")[2]/diff(par("usr")[3:4])
      plot.asp <- abs(plot.w / plot.h)
      
      # determine if label will fit within a reasonable buffer percentage
      if(label.orientation == 'v') {
        # vertical calculation
        
        # get text heights, as we will be printing labels at 90 deg
        text.heights <- abs(strheight(color.labels, cex = label.cex, font = label.font))
        # convert text heights into equivalent widths
        text.heights <- text.heights / plot.asp
        # compare re-scaled text heights with rectangle widths (weights) + some buffer
        label.fits <- which(text.heights < (s.i$weight - buffer.pct) )
        
      } else {
        # horizontal calculation
        
        # get text heights, as we will be printing labels at 90 deg
        text.widths <- abs(strwidth(color.labels, cex = label.cex, font = label.font))
        
        # compare re-scaled text widths with rectangle widths (weights) + some buffer
        label.fits <- which(text.widths < (s.i$weight - buffer.pct) )
      }
      
      
      # print labels
      if(print.label & (length(label.fits) > 0)) {
        
        # adjust label color based on background
        # from aqp
        label.col <- invertLabelColor(s.i[, 1])
        
        # adjust label angle based argument
        label.srt <- switch(label.orientation, v = 90, h = 0)
        
        text(
          x = centers[label.fits], 
          y = i, 
          labels = color.labels[label.fits], 
          col = label.col[label.fits], 
          font = label.font, 
          cex = label.cex, 
          srt = label.srt
        )
      }
      
    }
  }
  
  # add horizontal separator lines, typically used when rectange borders are not drawn
  if(horizontal.borders){
    hz.line.y <- 1:(length(names(s.scaled))-1) + 0.5
    segments(x0 = 0, y0 = hz.line.y, x1 = 1, y1 = hz.line.y, lwd=horizontal.border.lwd)
  }
  
  # label x-axis with a scale
  if(x.axis) {
    axis(1, at=round(seq(0, 1, length.out = 11), 2), col=par('fg'), col.axis=par('fg'))
  }
  
  
  # label x-axis with group names
  if(y.axis) {
    axis(2, at = seq_along(names(s.scaled)), labels = names(s.scaled), las=2, tick=FALSE, font=2, hadj=1, line=-2.125, cex.axis=1, col=par('fg'), col.axis=par('fg'))
  }
  
}

setGeneric("prj", function(object, ...)
  standardGeneric("prj"))

#' Get or Set Coordinate Reference System for SoilProfileCollection
#'
#' @description `prj()`: Get Coordinate Reference System (Projection) metadata
#' @param object A SoilProfileCollection
#' @param value character. Representation of Coordinate Reference System as `"authority:code"`, integer EPSG code, WKT2019 / PROJ4 string, an sf `crs` or sp `CRS` object.
#' @param ... Additional arguments (not used)
#' @param obj A SoilProfileCollection
#' @export
#' @seealso [initSpatial<-()]
#' @rdname SoilProfileCollection-crs
#' @aliases prj proj4string
setMethod("prj", 'SoilProfileCollection',
          function(object, ...) {
            value <- metadata(object)$crs
            if (length(value) == 0 || (!is.na(value) && nchar(value) == 0)) {
              value <- NA_character_
            }
            return(value)
          }
)

setGeneric("prj<-", function(object, ..., value)
  standardGeneric("prj<-"))

#' @description `prj()<-`: Set Coordinate Reference System metadata for the SoilProfileCollection
#' @aliases prj<- proj4string<-
#' @rdname SoilProfileCollection-crs
#' @export
setReplaceMethod("prj", 'SoilProfileCollection',
                 function(object, ..., value) {
                   
                   # backward compatibility for sp::CRS object
                   if (inherits(value, 'CRS')) {
                     if (!is.null(attr(value, 'comment'))) {
                       # use WKT2019 if available
                       value <- attr(value, 'comment')
                     } else {
                       # otherwise use proj arg string
                       value <- value@projargs
                     }
                   }
                   
                   # added compatibility for sf::crs object
                   if (inherits(value, 'crs')) {
                     if (!is.null(value$wkt)) {
                       # use WKT2019 if available
                       value <- value$wkt
                     } else {
                       # otherwise use input proj arg string
                       value <- value$input
                     }
                   }
                   
                   # 0-length or empty character are equivalent to NA CRS
                   if (length(value) == 0 || nchar(value) == 0) {
                     value <- NA_character_
                   }
                   
                   # "crs" metadata stores the WKT string in the SPC
                   metadata(object)$crs <- value
                   return(object)
                 }
)

setGeneric("proj4string", function(obj)
  standardGeneric("proj4string"))

#' @description `proj4string()`: (Deprecated) Get Coordinate Reference System as PROJ4 String
#' @rdname SoilProfileCollection-crs
#' @export
setMethod("proj4string", signature(obj = 'SoilProfileCollection'),
          function(obj) {
            .Deprecated("prj", package = "aqp", 
                        msg = "Methods based on sp class definitions have been deprecated and will be removed by October 1, 2023. See ?prj")
            return(prj(obj))
          }
)

setGeneric("proj4string<-", function(obj, value)
  standardGeneric("proj4string<-"))

#' @description `proj4string()<-`: (Deprecated) Set Coordinate Reference System metadata for the SoilProfileCollection
#' @rdname SoilProfileCollection-crs
#' @export
setReplaceMethod("proj4string", signature(obj = "SoilProfileCollection",
                                          value = "ANY"),
                 function(obj, value) {
                   .Deprecated("prj<-", package = "aqp", 
                               msg = "Methods based on sp class definitions have been deprecated and will be removed by October 1, 2023. See ?`prj<-`")
                   prj(obj) <- value
                   return(obj)
                 })
    
setGeneric("initSpatial<-", function(object, crs = NULL, value)
  standardGeneric("initSpatial<-"))

#' Initialize Spatial Data in a SoilProfileCollection 
#' 
#' `initSpatial()<-`: Set the column names containing spatial data and the corresponding coordinate reference system for a SoilProfileCollection.
#' 
#' @param object A SoilProfileCollection
#' @param value A formula specifying names of columns containing geometry (x and y coordinates), or character with the column names
#' @param crs Optional: character. Representation of Coordinate Reference System as `"authority:code"`, integer EPSG code, WKT2019 or PROJ4 string, an sf `crs` or sp `CRS` object.
#' @aliases initSpatial<-
#' @rdname initSpatial
#' @export
#' @seealso [prj()]
#' @examples
#'
#' data(sp5)
#'
#' # coordinates are stored in x and y column of site
#' sp5$x <- rnorm(length(sp5))
#' sp5$y <- rnorm(length(sp5))
#'
#' # coordinates takes a formula object as input
#' initSpatial(sp5) <- ~ x + y
#' 
#' # optionally specify Coordinate Reference System (crs) on left-hand side
#' initSpatial(sp5, crs = "OGC:CRS84") <- ~ x + y
setReplaceMethod("initSpatial", signature(object = "SoilProfileCollection", value = "ANY"),
                 function(object, crs = NULL, value) {
                   # basic sanity check
                   if (!inherits(value, "formula"))
                     stop('invalid formula: ', quote(value), call. = FALSE)
                   
                   fterms <- attr(terms(value), "term.labels")
                   
                   if (all(fterms %in% siteNames(object))) {
                     mf <- data.matrix(model.frame(value, site(object), na.action = na.pass))
                   } else if (all(fterms %in% horizonNames(object))) {
                     mf <- unique(data.matrix(model.frame(value, horizons(object), na.action = na.pass)))
                   } else {
                     stop("formula terms not found in site or horizon table: ",
                          paste0(fterms[!fterms %in% names(object)], collapse = ","))
                   }
                   
                   # make sure that "normalization" worked
                   if (nrow(mf) != length(object)) {
                     stop("coordinates in horizon data are not unique within site: ", quote(value), call. = FALSE)
                   }
                   
                   # set coordinates metadata entry
                   metadata(object)$coordinates <- fterms
                   
                   # set CRS if specified
                   if (!is.null(crs)) {
                     prj(object) <- crs
                   }
                   
                   return(object)
                 })

#' @rdname initSpatial
#' @export
setReplaceMethod("initSpatial", signature(object = "SoilProfileCollection", value = "character"),
                 function(object, crs = NULL, value) {
                   initSpatial(object, crs = crs) <- as.formula(paste0("~", paste0(value, collapse = "+")))
                   object
                  })


setGeneric("getSpatial", function(object, ...)
  standardGeneric("getSpatial"))

#' Get Soil Profile Spatial Data
#'
#' @description `getSpatial()`: Get spatial data associated with a SoilProfileCollection
#'
#' @param object A SoilProfileCollection
#' @docType methods
#' 
#' @export
#' @aliases getSpatial
#' @rdname initSpatial
setMethod("getSpatial", signature(object = "SoilProfileCollection"),
          function(object) {
            cn <- metadata(object)$coordinates
            
            # if not specified or not valid, return a 0x2 matrix
            if (length(cn) == 0 || !validSpatialData(object))
              return(matrix(numeric(0), ncol = 2, dimnames = list(NULL, c("x", "y"))))
            
            return(as.matrix(sapply(cn, function(x) object[[x]])))
          })

setGeneric("coordinates", function(obj)
  standardGeneric("coordinates"))

#' @param obj A SoilProfileCollection
#' @export
#' @rdname initSpatial
#' @aliases coordinates
#' @export
setMethod("coordinates", signature(obj = "SoilProfileCollection"),
          function(obj) {
            .Deprecated("initSpatial<-", package = "aqp", 
                        msg = "Methods based on sp class definitions have been deprecated and will be removed by October 1, 2023. See ?getSpatial")
            getSpatial(obj)
          })

setGeneric("coordinates<-", function(object, value)
  standardGeneric("coordinates<-"))

#' @rdname initSpatial
#' @export
#' @aliases coordinates<-
#' @export
setReplaceMethod("coordinates", c("SoilProfileCollection", "ANY"),
                 function(object, value) {
                   .Deprecated("initSpatial<-", package = "aqp", 
                               msg = "Methods based on sp class definitions have been deprecated and will be removed by October 1, 2023. See ?`initSpatial<-`")
                   initSpatial(object) <- value
                   object
                 })

#' @rdname initSpatial
#' @export
setReplaceMethod("coordinates", c("SoilProfileCollection", "character"),
                function(object, value) {
                  .Deprecated("initSpatial<-", package = "aqp", 
                              msg = "Methods based on sp class definitions have been deprecated and will be removed by October 1, 2023. See ?`initSpatial<-`")
                  initSpatial(object) <-  as.formula(paste0("~", paste0(value, collapse = "+")))
                  object
                })

#' @title Fix Common Errors in Munsell Notation
#' 
#' @description
#' This function will "launder" colors in Munsell notation, fixing (and optionally flagging) the following:
#' 
#'   * colors specified with 0-chroma will be converted to neutral hue (N) and original value
#'   * for neutral colors, any chroma >0 will be set to 0
#'   * non-standard hues will be converted to NA (unless `standardHues = FALSE`)
#' 
#' Examples:
#'   * '10YR 2/0' -> 'N 2/0'
#'   * 'N 4/1 -> 'N 4/0'
#'   * '5Z 3/3' -> NA
#'   * '5YR 3/' -> NA (hues other than N must have a valid chroma)
#' 
#' See [formatMunsell()] for additional details. 
#'
#' @param m character vector of Munsell colors
#' @param verbose logical, optionally return a `data.frame` comparing modifications
#'
#' @param \dots additional arguments to [formatMunsell()]
#'
#' @returns either character vector, or when `verbose = TRUE` a `data.frame`
#'
#' @seealso [formatMunsell()]
#' @export
#'
#' @examples
#' 
#' # => 'N 2/0'
#' launderMunsell('10YR 2/0')
#' 
#' # => 'N 4/0'
#' launderMunsell('N 4/1')
#' 
#' # alternative neutral convention => 'N 4/'
#' launderMunsell('N 4/0', neutralConvention = 'empty')
#'
#' # => 'N 4/0'
#' launderMunsell('N 4/NA')
#' launderMunsell('N 4/')
#' 
#' # => 'N 4/0'
#' launderMunsell('5GY 4/0')
#' 
#' # not a standard hue => NA
#' launderMunsell('4ZR 4/6')
#' 
#' # missing chroma, not N => NA
#' launderMunsell('2.5Y 4/')
#' 
#' # invalid chroma => NA
#' launderMunsell('2.5Y 4/A')
#' 
launderMunsell <- function(m, verbose = FALSE, ...) {
  
  # split into components
  # missing chroma for N chips converted to 0
  .p <- parseMunsell(m, convertColors = FALSE)
  
  ## now handled by formatMunsell()
  # # N chips must have unspecified or 0 chroma
  # .idx <- which(.p$hue == 'N' & .p$chroma > 0)
  # 
  # if(length(.idx) > 0) {
  #   .p$chroma[.idx] <- 0
  # }
  
  # any hue other than N with chroma of 0 => N
  .idx <- which(.p$hue != 'N' & .p$chroma == 0)
  
  if(length(.idx) > 0) {
    .p$hue[.idx] <- 'N'
  }
  
  # combine back into standard notation
  # following conventions
  .res <- formatMunsell(hue = .p$hue, value = .p$value, chroma = .p$chroma, ...)
  
  # optionally combine into a data.frame for inspection
  if(verbose) {
    .res <- data.frame(
      original = m, 
      fixed = .res, 
      modified = !m == .res
    )
  }
  
  return(.res)
  
}

## TODO: consider always returning an ordered factor, why would this ever not be the intended output?

#' @title Generalize Horizon Names
#'
#' @description Generalize a vector of horizon names, based on new classes, and REGEX
#' patterns. Or create a new column `ghl` in a `SoilProfileCollection` (requires a horizon designation name to be defined for the collection, see details)
#'
#' @param x character vector of horizon names or a `SoilProfileCollection` object
#' @param new character vector of generalized horizon labels (GHL)
#' @param pattern character vector of REGEX patterns, same length as `new`
#' @param non.matching.code character, label used for any horizon not matched by `pattern`
#' @param hzdepm numeric vector of horizon mid-points; `NA` values in `hzdepm` will result in `non.matching.code` (or `NA` if not defined) in result
#' @param ordered logical, default `TRUE` when `hzdepm` argument is specified
#' @param na.rm logical, default `TRUE` will ignore missing depths in calculating sort order when `hzdepm` is specified and `ordered=TRUE`
#' @param ... additional arguments passed to `grep()` such as `perl = TRUE` for advanced REGEX
#'
#' @return factor (an ordered factor when `ordered=TRUE`) of the same length as `x` (if character) or as number of horizons in `x` (if `SoilProfileCollection`)
#' 
#' @details When `x` is a `SoilProfileCollection` the `ghl` column will be updated with the factor results. This requires that the "horizon designation name" metadata be defined for the collection to set the column for input designations.
#' 
#' @seealso [hzdesgnname()]
#' @author D.E. Beaudette
#' @keywords manip
#' 
#' @references 
#' Beaudette, D.E., Roudier, P., Skovlin, J. (2016). Probabilistic Representation of Genetic Soil Horizons. In: Hartemink, A., Minasny, B. (eds) Digital Soil Morphometrics. Progress in Soil Science. Springer, Cham. https://doi.org/10.1007/978-3-319-28295-4_18
#' 
#' @export
#' @examples
#' 
#' data(sp1)
#' 
#' # check original distribution of hz designations
#' table(sp1$name)
#' 
#' # generalized horizon labels
#' # character vector input
#' sp1$genhz <- generalizeHz(
#'   sp1$name,
#'   new = c('O','A','B','C','R'),
#'   pattern = c('O', '^A','^B','C','R'),
#'   ordered = TRUE
#' )
#' 
#' # see how we did / what we missed
#' table(sp1$genhz, sp1$name)
#' 
#' 
#' ## a more advanced example, requries `perl = TRUE`
#' # example data
#' x <- c('A', 'AC', 'Bt1', '^AC', 'C', 'BC', 'CB')
#' 
#' # new labels
#' n <- c('A', '^AC', 'C')
#' 
#' # patterns:
#' # "A anywhere in the name"
#' # "literal '^A' anywhere in the name"
#' # "C anywhere in name, but without preceding A"
#' p <- c('A', '^A', '(?<!A)C')
#' 
#' # note additional argument
#' res <- generalizeHz(
#'   x, 
#'   new = n, 
#'   pattern = p, 
#'   perl = TRUE
#' )
#' 
#' # double-check: OK
#' table(res, x)
#' 
#' ## apply to a SoilProfileCollection
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#' 
#' # must set horizon designation metadata
#' hzdesgnname(sp1) <- 'name'
#' 
#' # result is a SoilProfileCollection
#' x <- generalizeHz(
#'   sp1,
#'   new = c('O','A','B','C','R'),
#'   pattern = c('O', '^A','^B','C','R'),
#'   ordered = TRUE
#' )
#' 
#' # GHL stored in 'genhz' column
#' x$genhz
#' 
#' # GHL metadata is set
#' GHL(x)
#'
generalize.hz <- function(x, new, pattern, non.matching.code = 'not-used', hzdepm = NULL, ordered = !missing(hzdepm), na.rm = TRUE, ...) {

	# init vector of 'other', same length as original horizon name vector
	g <- rep(non.matching.code, times = length(x))

	# iterate over new new names, match with patterns, assign new hz names
	for (i in seq_along(new)) {
		g[grep(pattern[i], x, ...)] <- new[i]
	}

	# convert to (ordered) factor with levels set by median depth
	if (!missing(hzdepm) && !is.null(hzdepm) && !all(is.na(hzdepm))) {
	  # more stringent NA handling:
	  # stopifnot(`Number of non-NA \`hzdepm\` does match number of new levels` =
	  #           { sum(!is.na(hzdepm)) == length(new) })
	  
	  # less stringent:
	  # default na.rm (TRUE) ignores depths that are NA
	  new_sort <- names(sort(tapply(hzdepm, g, median, na.rm = na.rm), na.last = ifelse(na.rm, na.rm, NA)))
	  new_sort <- new_sort[new_sort != non.matching.code]
	  
	  # use an ordered factor (may be overridden w/ ordered = FALSE)
	  g <- factor(g, levels = c(new_sort, non.matching.code), ordered = ordered)
	  
	  # if any are not matched (i.e. hzdepm is NA), replace with non.matching.code (if defined)
	  if (!is.null(non.matching.code)) {
	    g[is.na(g) | is.na(hzdepm)] <- non.matching.code
	  }
	  g
	} else {
  	# no adjustment of factor levels based on hz mid-points
  	# no order implied (by default; override with ordered=TRUE)
  	factor(g, levels = c(new, non.matching.code), ordered = ordered)
	}
}

setGeneric("generalizeHz",  function(x, new, pattern, non.matching.code = 'not-used', hzdepm = NULL, ordered = !missing(hzdepm), ...) standardGeneric("generalizeHz"))

#' @export
#' @aliases generalizeHz
#' @rdname generalize.hz
setMethod("generalizeHz", signature(x = "character"), function(x, new, pattern, non.matching.code = 'not-used', hzdepm = NULL, ordered = !missing(hzdepm), ...) {
 generalize.hz(
    x,
    new = new,
    pattern = pattern,
    non.matching.code = 'not-used',
    hzdepm = hzdepm,
    ordered = ordered, 
    ...
  )
})

# SoilProfileCollection method
#' @param ghl Generalized Horizon Designation column name (to be created/updated when `x` is a `SoilProfileCollection`)
#' @export
#' @rdname generalize.hz
setMethod("generalizeHz", signature(x = "SoilProfileCollection"), function(x, new, pattern, non.matching.code = 'not-used', hzdepm = NULL, ordered = !missing(hzdepm), ghl = "genhz", ...) {
  hzdesgn <- hzdesgnname(x, required = TRUE)
  
  x[[ghl]] <- generalize.hz(
    x[[hzdesgn]],
    new = new,
    pattern = pattern,
    non.matching.code = 'not-used',
    hzdepm = hzdepm,
    ordered = ordered, 
    ...
  )
  
  # set label in metadata
  GHL(x) <- ghl
  
  x
})

#' @title Convert cross-tabulation to adjacency matrix.
#' @description Convert a cross-tabulation of e.g. original horizon designations vs. generalized horizon labels to adjacency matrix form.
#' 
#' @param tab table, cross-tabulation of original and generalized horizon labels e.g. `table(original, genhz)`
#' @return matrix of numeric values
#' @author D.E. Beaudette
#' @export
genhzTableToAdjMat <- function(tab) {
  tab <- as.matrix(tab)
  # extract unique set of names
  nm <- sort(unique(unlist(dimnames(tab))))
  # generate full matrix with named dimensions
  m <- matrix(nrow = length(nm), ncol = length(nm), data = 0)
  dimnames(m) <- list(nm, nm)
  # set adjacency information via sub-matrix containing the original cross-tab
  m[rownames(tab), colnames(tab)] <- tab
  return(m)
}

## TODO: consider using Munsell notation for `col`, this would save an entire round-trip through the conversion process
## TODO: consider data.frame interface

#' @title Summarize Soil Colors
#' 
#' @description Summarize soil color data, weighted by occurrence and horizon thickness.
#'
#' @param x a `SoilProfileCollection` object
#' @param groups the name of a horizon or site attribute used to group horizons, see examples
#' @param col the name of a horizon-level attribute with soil color specified in hexadecimal (i.e. "#rrggbb")
#' @param k single integer specifying the number of colors discretized via PAM ([cluster::pam()]), see details
#' @param profile_wt the name of a site-level attribute used to modify weighting, e.g. area
#' 
#' @param mixingMethod method used to estimate "aggregate" soil colors, see [mixMunsell()]
#'
#' @return A list with the following components:
#' 
#'  * `scaled.data`: a `list` of colors and associated weights, one item for each generalized horizon label with at least one color specified in the source data
#'  * `aggregate.data`: a `data.frame` of weighted-mean colors, one row for each generalized horizon label with at least one color specified in the source data
#' 
#' @export
#' 
#' @details Weights are computed by:
#' `w_i = sqrt(sum(thickness_i)) * n_i`
#' where `w_i` is the weight associated with color `i`, `thickness_i` is the total thickness of all horizons associated with the color `i`, and `n_i` is the number of horizons associated with color `i`. Weights are computed within groups specified by `groups`.
#' 
#' See the [related tutorial for additional examples](http://ncss-tech.github.io/AQP/sharpshootR/aggregate-soil-color.html).
#' 
#' @author D.E. Beaudette
#' 
#' @seealso [generalize.hz()], [aggregateColorPlot()]
#' 
#' @examples
#' 
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' # load some example data
#' data(sp1, package = 'aqp')
#' 
#' # upgrade to SoilProfileCollection and convert Munsell colors
#' sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#' 
#' # generalize horizon names
#' n <- c('O', 'A', 'B', 'C')
#' p <- c('O', 'A', 'B', 'C')
#' sp1$genhz <- generalize.hz(sp1$name, n, p)
#' 
#' # aggregate colors over horizon-level attribute: 'genhz'
#' a <- aggregateColor(sp1, groups = 'genhz', col = 'soil_color')
#' 
#' # check results
#' str(a)
#' 
#' # simple visualization
#' aggregateColorPlot(a)
#' 
aggregateColor <- function(x, groups = 'genhz', col = 'soil_color', k = NULL, profile_wt = NULL, mixingMethod = c('estimate', 'exact')) {
  
  # sanity check
  mixingMethod <- match.arg(mixingMethod)
  
  # sanity check
  if(!is.null(k)) {
    k <- round(k)
    
    if(is.na(k)) {
      stop('k must be a single integer > 0')
    }
  }
  
  # sanity check: profile weight must be a valid site-level attribute
  if(!is.null(profile_wt)) {
    if(! profile_wt %in% siteNames(x)) {
      stop('`profile_wt` must be a site-level attribute of `x`', call. = FALSE)
    }
  }
  
  # sanity check: groups must be a horizon-level attribute
  if(! groups %in% names(x))
    stop('`groups` must be a site or horizon attribute of `x`', call. = FALSE)
  
  # sanity check: col must be a horizon-level attribute
  if(! col %in% horizonNames(x))
    stop('`col` must be a horizon-level attribute of `x`', call. = FALSE)
  
  ## hack to make R CMD check --as-cran happy
  top <- bottom <- thick <- NULL
  
  # extract pieces
  h <- as(x, 'data.frame')
  
  # keep track of just those variables we are using
  # profile_wt is NULL by default
  vars <- c(groups, horizonDepths(x), col, profile_wt)
  
  # remove missing data
  # note: missing color data will result in 'white' when converting to sRGB, why?
  h.no.na <- na.omit(h[, vars])
  
  # re-name depths
  names(h.no.na)[2:3] <- c('top', 'bottom')
  
  # safely compute thickness
  # abs() used in rare cases where horizon logic is wrong: e.g. old-style O horizons
  h.no.na$thick <- abs(h.no.na$bottom - h.no.na$top)
  
  # 0-thickness will result in NA weights
  # replace with a 1-unit slice
  idx <- which(h.no.na$thick == 0)
  if(length(idx) > 0) {
    h.no.na$thick[idx] <- 1
  }
  
  # apply site-level weighting here, if present
  if(!is.null(profile_wt)) {
    # multiply thickness by site-level weight
    h.no.na$thick <- h.no.na$thick * h.no.na[[profile_wt]]
  }
  
  # drop empty group levels
  h.no.na[[groups]] <- factor(h.no.na[[groups]])
  
  # split by genhz
  ss <- split(h.no.na, h.no.na[[groups]])
  
  # iterate over groups
  # note that some genhz will have 0 records
  s <- lapply(ss, function(i) {
    
    # current group label
    this.group <- i[[groups]][1]
    
    # optionally reduce to `k` colors via PAM, by group
    # this will only work if there are >= 1 unique colors
    n.cols <- length(unique(i[[col]]))
    
    if(is.numeric(k) & n.cols > 1) {
      
      # condition number of classes on available data
      k.adj <- pmin(k, n.cols - 1)
      
      # work with a unique subset, there is no need to compute distances / cluster all colors
      lut <- data.frame(col = unique(i[[col]]))
      
      # same order as LUT
      # convert to sRGB
      # NA will create bogus colors, filtered above
      v <- t(col2rgb(lut$col) / 255)
      
      # convert to LAB
      v <- convertColor(
        v, 
        from = 'sRGB', 
        to = 'Lab', 
        from.ref.white = 'D65', 
        to.ref.white = 'D65', 
        clip = FALSE
      )
      
      # compute perceptually based distance matrix on unique colors
      # output is transposed relative to `dist` object
      dE00 <- farver::compare_colour(
        from = v, 
        from_space = 'lab', 
        method = 'CIE2000', 
        white_from = 'D65'
      )
      
      # note transpose
      dE00 <- as.dist(t(dE00))
      
      # clustering from distance matrix
      # TODO: save clustering results for later
      v.pam <- cluster::pam(dE00, k = k.adj, diss = TRUE, pamonce = 3)
      
      # put clustering vector into LUT
      lut$cluster <- v.pam$clustering
      
      # get medoid colors
      med.col <- lut$col[v.pam$medoids]
      
      # expand original colors to medoid colors based on LUT
      idx <- base::match(i[[col]], lut$col)
      
      # replace original colors with discretized colors
      i[[col]] <- med.col[lut$cluster[idx]]
    }
    
    # aggregate depth by unique soil color
    # this assumes that thickness > 0, otherwise NaN is returned
    # convert to data.table for summary
    i <- data.table::as.data.table(i)
    
    # not sure about most readable style
    res <- i[, 
             list(
               weight = sqrt(sum(thick)) * length(thick),
               n.hz = length(thick)
             ), 
             by = col
    ]
    
    # convert back to data.frame
    res <- as.data.frame(res)
    
    # sort by thickness-- this is our metric for relevance
    res <- res[order(res$weight, decreasing=TRUE), ]
    
    ## TODO: this is wasteful, as we likely "knew" the munsell notation before-hand
    # back-calculate the closest Munsell color
    m <- col2Munsell(res[[col]])
    
    
    # format as text
    res$munsell <- paste0(m[, 1], ' ', m[, 2], '/', m[, 3])
    
    # add group ID
    res[['.id']] <- this.group
    
    return(res)
  })
  
  
  
  # rescale using the sum of the weights within the current horizon
  s.scaled <- lapply(s, function(i) {
    i$weight <- i$weight / sum(i$weight)
    return(i)
  })
  
  
  # iteration over groups, estimation of:
  # aggregate colors via mixing
  # number of associated colors
  # Shannon H, base 2
  s.agg <- lapply(s.scaled, function(i) {
    
    # estimation of mixture via wt. mean in CIELAB coordinates
    mix <- mixMunsell(x = i$munsell, w = i$weight, mixingMethod = mixingMethod)
    
    # split into pieces for backwards-compatibility
    m.pieces <- parseMunsell(mix$munsell, convertColors = FALSE)
    col <- parseMunsell(mix$munsell)
    
    # consolidate and return
    res <- data.frame(
      .id = i[['.id']][1],
      m.pieces,
      munsell = mix$munsell,
      distance = mix$distance,
      col = col, 
      n = nrow(i),
      H = shannonEntropy(i$weight)
    )
    
    return(res)
  })
  
  s.agg <- do.call('rbind', s.agg)
  names(s.agg)[1] <- groups
  row.names(s.agg) <- NULL
  
  .res <- list(scaled.data = s.scaled, aggregate.data = s.agg)
  
  # return scaled color data
  return(.res)
}

# https://rip94550.wordpress.com/2012/05/21/color-from-spectrum-to-xyz-and-beyond/



## TODO: allow for more flexibility in spectra min/max/res wavelength

#' @title Convert reflectance spectra to closest Munsell chip
#' 
#' @details See the [expanded tutorial](https://ncss-tech.github.io/AQP/aqp/mix-colors.html) for additional examples.
#' 
#' @param x reflectance spectra, must range from 380nm to 730nm with resolution specified in `res`
#' 
#' @param res spectra resolution in nm, typically 5nm or 10nm
#' 
#' @param convert logical, convert sRGB coordinates to closest Munsell chip (see `?munsell`)
#' 
#' @param SO CIE standard observer: these are the color matching functions defined by CIE and used to represent "average" human color perception. CIE1931 is the 2 degree standard observer more useful for describing color perception over very small areas or at distance. CIE1964 is the 10 degree standard observer, used for most industrial color matching applications.
#' 
#' @param illuminant CIE standard illuminants: 
#'    * D65 represents average daylight
#'    * F2 represents typical fluorescent lighting
#' 
#' @param ... further arguments to [col2Munsell()]
#'
#' @return output from [col2Munsell()]
#' @export
#'
#' @references 
#' Marcus, R.T. (1998). The Measurement of Color. In K. Nassau (Ed.), Color for Science, Art, and Technology (pp. 32-96). North-Holland.
#' 
#' CIE Colorimetry – Part 1: CIE standard colorimetric observers. CIES014-1/E:2006 – ISO 11664-1:2007(E)
#' 
#' CIE. (n.d.). CIE 15:2004 Tables Data. Retrieved from https://law.resource.org/pub/us/cfr/ibr/003/cie.15.2004.tables.xls
#' 
#' 
spec2Munsell <- function(x, res = 10, convert = TRUE, SO = c('CIE1931', 'CIE1964'), illuminant = c('D65', 'F2'), ...) {
  
  # D65 and CIE1931 reference data at 5nm
  spectral.reference <- NULL
  load(system.file("data/spectral.reference.rda", package="aqp")[1])
  
  # select standard observer
  SO <- match.arg(SO)
  
  # select illuminant
  illuminant <- match.arg(illuminant)
  
  # this is the range of wavelengths we have to work with
  # in the Munsell spectra libraries 
  #   * munsell.spectra
  #   * munsell.spectra.wide
  .wl <- seq(from = 380, to = 730, by = res)
  
  
  ## 10nm res: length(x) == 36
  ## 5nm res: length(x) == 71
  # sanity check
  if(length(.wl) != length(x)) {
    stop('inconsistent spectral limits / resolution', call. = FALSE)
  }
  
  # spline interpolator: reflectance ~ wavelength
  .sf <- splinefun(.wl, x)
  
  ## TODO: interpolate spectral.reference to match resolution of x
  
  # spectral reference is 5nm resolution
  # interpolate to 5nm res
  w.5 <- seq(from = 380, to = 730, by = 5)
  R <- data.frame(
    w = w.5, 
    x = .sf(w.5)
  )
  
  # reflectance spectra * illuminant
  S <- R$x * spectral.reference[[illuminant]]
  
  
  # select the appropriate standard observer
  SO.vars <- switch(SO,
    'CIE1931' = c('xbar_2', 'ybar_2', 'zbar_2'),
    'CIE1964' = c('xbar_10', 'ybar_10', 'zbar_10')
  )
  
  # A-matrix: these are the color-matching functions of the standard observer
  A <- spectral.reference[, SO.vars]
  
  # apply A-matrix to spectra
  XYZ <- t(A) %*% S
  # convert to vector
  XYZ <- t(XYZ)[1, ]
  
  # apply A-matrix to illuminant
  k <- t(A) %*% spectral.reference[[illuminant]]
  # convert to vector
  k <- t(k)[1, ]
  
  # re-scale by "photopic response"
  col.XYZ <- XYZ / k[2]
  
  # convert to sRGB
  col.srgb <- convertColor(
    rbind(col.XYZ), 
    from = 'XYZ', 
    to = 'sRGB', 
    from.ref.white = 'D65', 
    to.ref.white = 'D65'
  )
  
  # convert sRGB to closest Munsell chip
  if(convert) {
    m <- col2Munsell(col.srgb, ...)
    return(m)
  } else {
    # return sRGB coordinates
    return(col.srgb)
  }
  
  
  
}

context("mixing Munsell colors")

# reference spectra
data("munsell.spectra.wide")


test_that("spec2Munsell works as expected", {
  
  # try a color, result should be identical chip
  # depends on SO and illuminant
  chip <- '10YR 3/3'
  m <- spec2Munsell(munsell.spectra.wide[, chip], SO = 'CIE1931', illuminant = 'D65')  
  
  # object structure / contents
  expect_true(inherits(m, 'data.frame'))
  
  # same output as col2Munsell
  expect_true(
    all(
      names(m) == c('hue', 'value', 'chroma', 'sigma')
        )
  )

  # input == output
  expect_true(
    sprintf("%s %s/%s", m$hue, m$value, m$chroma) == chip
    )
  
  # expected dE00
  expect_equal(m$sigma, 0.64, tolerance = 0.01)
      
})

test_that("neutral chips", {
  
  # try a color, result should be identical chip
  # depends on SO and illuminant
  chip <- 'N 4/'
  m <- spec2Munsell(munsell.spectra.wide[, chip], SO = 'CIE1931', illuminant = 'D65')  
  
  # object structure / contents
  expect_true(inherits(m, 'data.frame'))
  
  # same output as col2Munsell
  expect_true(
    all(
      names(m) == c('hue', 'value', 'chroma', 'sigma')
    )
  )
  
  # input == output
  expect_true(
    sprintf("%s %s/", m$hue, m$value) == chip
  )
  
  # expected dE00
  expect_equal(m$sigma, 0.36, tolerance = 0.01)
  
})



test_that("interpolated / extrapolated spectra", {
  
  # 1-chroma chips are extrapolated for some hues
  # depends on SO and illuminant
  chip <- '7.5YR 2/1'
  m <- spec2Munsell(munsell.spectra.wide[, chip], SO = 'CIE1931', illuminant = 'D65')  
  
  # object structure / contents
  expect_true(inherits(m, 'data.frame'))
  
  # same output as col2Munsell
  expect_true(
    all(
      names(m) == c('hue', 'value', 'chroma', 'sigma')
    )
  )
  
  # input == output
  expect_true(
    sprintf("%s %s/%s", m$hue, m$value, m$chroma) == chip
  )
  
  # expected dE00
  expect_equal(m$sigma, 3, tolerance = 0.1)
  
  
  # 8.5-value chips are interpolated for some hues
  chip <- '7.5YR 8.5/6'
  m <- spec2Munsell(munsell.spectra.wide[, chip], SO = 'CIE1931', illuminant = 'D65') 
  
  # input == output
  expect_true(
    sprintf("%s %s/%s", m$hue, m$value, m$chroma) == chip
  )
  
  # expected dE00
  expect_equal(m$sigma, 0.13, tolerance = 0.01)
  
})


test_that("supporting data match range of Munsell spectra", {
  
  # standard illuminants + observers
  data("spectral.reference")
  
  expect_true(
    all(
      range(spectral.reference$w) == range(munsell.spectra.wide$wavelength)
    )
  )
  
})

## TODO: allow standard Munsell notation vs. split into HVC
## TODO: allow CIELAB specification

## simulate via multivariate normal distribution
# n: number of simulations (typically horizons)
# parameters: list of parameters
#  data.frame with Munsell [hue, value, chroma]
.simulateColorFromMV <- function(n, parameters) {
  
  # sanity check, need this for rmvnorm()
  if(!requireNamespace('mvtnorm')) {
    stop('package `mvtnorm` is required for multivariate simulation', call. = FALSE)
  }
  
  ## TODO: consider pre-estimated mean vector + covariance matrix
  
  # extract parameters
  .hvc <- parameters[['hvc']]
  
  # convert Munsell -> CIELAB
  .lab <- munsell2rgb(the_hue = .hvc$hue, the_value = .hvc$value, the_chroma = .hvc$chroma, returnLAB = TRUE)
  
  # removing missing values which interfere with mean and covariance
  .lab <- na.omit(.lab)
  
  ## TODO: stop if nrow(.lab) < 3
  if(nrow(.lab) < 3) {
    return(NULL)
  }
  
  # multivariate simulation
  # assuming approx. joint normal distribution over L, A, B coordinates
  s <- mvtnorm::rmvnorm(
    n = n, 
    mean = colMeans(.lab), 
    sigma = cov(.lab),
  )
  
  
  ## TODO: consider returning CIELAB coordinates
  # .cols <- convertColor(s, from = 'Lab', to = 'sRGB', from.ref.white = 'D65', to.ref.white = 'D65')
  # previewColors(rgb(.cols, maxColorValue = 1), method = 'MDS')
  
  # this is slow
  # CIELAB -> Munsell hue, value, chroma
  m <- col2Munsell(s, space = 'CIELAB')
  
  ## TODO: consider including only hues in reference set
  
  # flatten to standard notation
  m <- sprintf('%s %s/%s', m$hue, m$value, m$chroma)
  
  return(m)
}




## simulate color via sampling with replacement and estimated proportions
# n: number of simulations (typically horizons)
# parameters: output from aqp::aggregateColor()
.simulateColorFromProportions <- function(n, parameters) {
  x <- parameters[['scaled.data']]
  
  res <- lapply(x, function(i) {
    sample(i[['munsell']], size = n, replace = TRUE, prob = i[['weight']])
  })
  
  return(res)
}


## simulate color from an RV color in Munsell notation, dE00 threshold, and vector of possible Munsell hues
# n: number of simulations (typically horizons)
# parameters: list of parameters
.simulateColorFromDE00 <- function(n, parameters) {
  # load Munsell LUT
  # safe for CRAN check
  munsell <- NULL
  load(system.file("data/munsell.rda", package = "aqp")[1])
  
  # extract parameters
  thresh <- parameters[['thresh']]
  m <- parameters[['m']]
  hues <- parameters[['hues']]
  
  ## borrowed / adapted from contrastChart()
  
  # extract just requested hues
  x <- munsell[which(munsell$hue %in% hues), ]
  
  # convert into hex notation for plotting
  x$color <- munsell2rgb(x$hue, x$value, x$chroma)
  x$munsell <- sprintf("%s %s/%s", x$hue, x$value, x$chroma)
  
  # re-level hues according to color contrast guidance
  hh <- unique(x$hue)
  ll <- hh[order(huePosition(hh))]
  x$hue <- factor(x$hue, levels=ll)
  
  # setup query color table
  m <- data.frame(
    queryColor = m, 
    parseMunsell(m, convertColors = FALSE),
    stringsAsFactors = FALSE
  )
  m$value <- as.integer(m$value)
  m$chroma <- as.integer(m$chroma)
  
  # compute all pair-wise contrast classes and dE00
  cc <- colorContrast(x$munsell, rep(m$queryColor, times = nrow(x)))
  
  # join for plotting
  z <- merge(x, cc, by.x = 'munsell', by.y = 'm1', all.x = TRUE, sort = FALSE)
  
  # dE00 threshold
  idx <- which(z$dE00 < thresh)
  
  # catch cases where the threshold is too low ... and ?
  if(length(idx) == 0) {
    message('Threshold too low.')
    return(rep(m, times = n))
  } else {
    # apply filter
    z <- z[idx, ]  
  }
  
  
  ## TODO: think about alternatives:
  #
  #  * ? --> perform conversion without RV, then re-add just before sampling 
  #  * z <- z[z$munsell != m$queryColor, ]
  
  ## convert distances -> similarities, interpret as sampling weights
  
  # standard conversion
  # too fast of a drop off between RV and simulated values
  s <- 1 / (1 + (z$dE00))
    
  # linear re-mapping of dE00 -> similarity
  # simulated values too close to RV
  # s <- 1 - (z$dE00 / max(z$dE00))
  
  
  ## diagnostics for dE00 -> probability
  # plot(s, z$dE00, type = 'n', las = 1)
  # points(s, z$dE00, col = z$color, pch = 15)
  # text(s, 0, z$munsell, cex = 0.5, srt = 90)

  
  # sample with replacement
  # according to ?sample, there is no need to convert weights -> probabilities
  # using translated dE00 as prior probabilities
  res <- sample(z$munsell, replace = TRUE, size = n, prob = s)
  
  return(res)
  
}



#' @title Simulate Soil Colors
#' 
#' @description Simulate plausible soil colors based on several possible parameterization of a "range in characteristics" (RIC). Soil color RIC can be specified by a list of parameters:
#'  * soil color proportions, as output from [aggregateColor()] -- `method = 'proportions'`
#'  * most likely Munsell color, CIE2000 threshold, and vector of acceptable hues -- `method = 'dE00'`
#'  * `data.frame` of Munsell hue, value, and chroma representing observed soil colors -- `method = 'mvnorm'`
#' 
#' 
#' 
#' @author D.E. Beaudette
#'
#' @param method simulation method, see details
#' 
#' @param n number of simulated colors per group
#' 
#' @param parameters a `list`, format depends on `method`:
#'   * `proportions`: output from [aggregateColor()] 
#'   * `dE00`: formatted as `list(m = '7.5YR 3/3', thresh = 5, hues = c('7.5YR'))`
#'   * `mvnorm`: formatted as `list(hvc = x)`
#'   
#'   Where `m` is a single representative Munsell chip, `thresh` is a threshold specified in CIE2000 color contrast (dE00), `hues` is a vector of allowed Munsell hues, and `x` is a `data.frame` representing columns of Munsell hue, value, and chroma having at least 3 rows.
#' 
#' @param SPC `SoilProfileCollection`, attempt to modify `SPC` with simulated colors
#'
#' @return a `list`, unless `SPC` is specified, then a `SoilProfileCollection` object
#' 
#' @export
#' 
#' @examples 
#' 
#' # restrict examples to 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' # m: representative or most likely color
#' # thresh: dE00 threshold
#' # hues: allowed Munsell hues
#' p <- list(
#'   'A' = list(m = '7.5YR 3/3', thresh = 5, hues = c('7.5YR')),
#'   'BA' = list(m = '7.5YR 4/4', thresh = 8, hues = c('7.5YR')),
#'   'Bt1' = list(m = '7.5YR 4/4', thresh = 8, hues = c('5YR', '7.5YR')),
#'   'Bt2' = list(m = '5YR 4/5', thresh = 8, hues = c('5YR', '7.5YR')),
#'   'Bt3' = list(m = '10YR 4/6', thresh = 10, hues = c('10YR', '7.5YR')),
#'   'Cr' = list(m = '2.5G 6/2', thresh = 15, hues = c('2.5G', '2.5GY', '2.5BG'))
#' )
#' 
#' # simulate
#' (cols <- simulateColor(method = 'dE00', n = 10, parameters = p))
#' 
#' # preview
#' previewColors(parseMunsell(unlist(cols)), method = 'MDS')
#'
#' # another example, this time using a larger dE00 threshold
#' p <- list(
#'   'A' = list(m = '7.5YR 3/3', thresh = 20, hues = c('10YR', '7.5YR', '5YR'))
#' )
#' 
#' # simulate
#' set.seed(54654)
#' cols <- simulateColor(method = 'dE00', n = 200, parameters = p)
#' 
#' # flatten
#' cols <- unlist(cols)
#' 
#' # tabulate, sort: most frequent color should be 7.5YR 3/3
#' sort(table(cols), decreasing = TRUE)
#' 
#' # review colors
#' previewColors(parseMunsell(cols))
#' 
#' # what does a dE00 threshold look like on 3 pages of hue?
#' contrastChart('7.5YR 3/3', hues = c('10YR', '7.5YR', '5YR'), thresh = 10)
#'
simulateColor <- function(method = c('dE00', 'proportions', 'mvnorm'), n, parameters, SPC = NULL) {
  
  # safely select method
  method <- match.arg(method)
  
  # if parameters is a single-depth list, add one more level
  if(!inherits(parameters[[1]], 'list')) {
    parameters <- list(parameters)
  }
  
  ## TODO: basic error checking, depends on method
  
  # select method
  res <- switch(
    method,
    'dE00' = {
      # manual iteration over parameters
      lapply(parameters, function(i) {
        .simulateColorFromDE00(n = n, parameters = i)
      })
    },
    # automatic iteration over output from aggregateColor()
    'proportions' = {
      .simulateColorFromProportions(n = n, parameters = parameters)
    },
    # manual iteration over parameters
    'mvnorm' = {
      lapply(parameters, function(i) {
        .simulateColorFromMV(n = n, parameters = i)
      })
    }
  )
  
  # result is a list
  if(is.null(SPC)) {
    return(res)
  } else {
    
    ## TODO: make this more efficient
    
    # result is a modified SPC
    
    # basic error checking: number of horizons
    
    # copy colors in horizon-order, profile order doesn't matter
    l <- list()
    for(i in 1:length(SPC)) {
      SPC.i <- SPC[i, ]
      
      ## TODO: hard-coded soil color column
      horizons(SPC.i)$soil_color <- parseMunsell(sapply(res, '[[', i))
      l[[i]] <- SPC.i
    }
    
    # flatten and done
    return(combine(l))
  }
  
}

## TODO:
##      * accept colors in any format via aqp:::.detectColorSpec()
##      * stratified model by major hz types
##      * add `sameHue` argument to enforce no change in hue


#' @title Estimate dry soil colors from moist soil colors and vice versa.
#' 
#' @description All else equal, soil color will predictably shift in perceived lightness (change in Munsell value) as moisture content changes. Field-described soil colors are typically collected at approximately air dry ("dry") and field capacity ("moist") states. This function estimates "dry" soil colors from "moist" soil colors and vice versa. Two methods are available for estimation, both developed from a national collection of field-described soil colors (approx. 800k horizons). Estimates are only valid for mineral soil material, having Munsell values and chroma < 10. Estimation has a median error rate of approximately (CIE dE2000) 5. 
#' 
#' Available Methods:
#'  * "procrustes": soil colors are converted using scale, rotation, and translation parameters in CIELAB color space
#'  * "ols": soil colors are converted using 3 multiple linear regression models (CIELAB coordinates)
#' 
#' This is still a work in progress.
#' 
#' @details 
#' 
#' For both methods, estimation proceeds as:
#'   * convert Munsell notation to CIELAB color coordinates via `munsell2rgb()`
#'   * apply rotation or regression model to color coordinates in CIELAB space
#'   * locate closest Munsell chip to resulting CIELAB coordinates via `col2munsell()`
#'   
#' Estimation of dry from moist soil color state is not guaranteed to be symmetric with estimation of moist from dry.
#' 
#' Scaling, rotation, and translation parameters for shifting between dry <--> moist CIELAB coordinates were determined using `vegan::procrustes()`. Multiple linear regression models were fit using `rms::ols()`.
#' 
#' Estimation error for both methods, converting from either moisture states is, approximately 5 (CIE dE 2000).

#' 
#' @author D.E. Beaudette
#' 
#' @param hue vector of Munsell hue ('10YR', '2.5Y', etc.)
#' @param value vector of Munsell value (2,2.5 2.5, 3, 5, 6, etc.)
#' @param chroma vector of Munsell chroma (2, 3, 4, etc.)
#' @param method character, one of 'procrustes' or 'ols', see details
#' @param sourceMoistureState character, source colors are either 'dry' or 'moist' 
#' @param returnMunsell logical, `TRUE`: return closest Munsell chip, `FALSE`: return estimated CIELAB coordinates
#'
#' @return `data.frame` of estimated colors in Munsell notation. The `sigma` column contains CIE2000 color contrast metric values describing the perceptual distance between estimated color in CIELAB coordinates and closest Munsell chip.
#' 
#' @references
#' 
#' J. A. Shields, E. A. Paul, R. J. St. Arnaud, and W. K. Head. 1968. SPECTROPHOTOMETRY MEASUREMENT OF SOIL COLOR AND ITS RELATIONSHIP TO MOISTURE AND ORGANIC MATTER. Canadian Journal of Soil Science. 48(3): 271-280. https://doi.org/10.4141/cjss68-037
#' 
#' 
#' @export
#'
#' @examples
#' 
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' estimateSoilColor(hue = '10YR', value = 3, chroma = 3, sourceMoistureState = 'moist')
#' 
#' # estimation is not always symmetric
#' estimateSoilColor(hue = '10YR', value = 4, chroma = 3, sourceMoistureState = 'dry')
#' 
#' # more examples
#' estimateSoilColor(hue = '2.5Y', value = 8, chroma = 2, sourceMoistureState = 'dry')
#' estimateSoilColor(hue = '2.5YR', value = 3, chroma = 4, sourceMoistureState = 'moist')
#' 
#' estimateSoilColor(hue = 'N', value = 2, chroma = 0, sourceMoistureState = 'moist')
#' 
#' estimateSoilColor(hue = '7.5YR', value = 2, chroma = 2, sourceMoistureState = 'moist')
#' 
#' # resulting hue is not always the same
#' estimateSoilColor(hue = '5G', value = 6, chroma = 6, sourceMoistureState = 'dry')
#' 
#' # return estimated CIELAB coordinates
#' estimateSoilColor(hue = '5G', value = 6, chroma = 6, sourceMoistureState = 'dry',
#'  returnMunsell = FALSE)
#' 
estimateSoilColor <- function(hue, value, chroma, method = c('procrustes', 'ols'), sourceMoistureState = c('dry', 'moist'), returnMunsell = TRUE) {
  
  # sanity checks
  sourceMoistureState <- match.arg(sourceMoistureState)
  method <- match.arg(method)
  
  
  # detect color spec
  # .spec <- .detectColorSpec(color)
  
  # convert input to CIELAB
  .lab <- munsell2rgb(hue, value, chroma, returnLAB = TRUE)
  
  
  ## TODO: add logic to enforce same hue, not at important with latest model
  # # optionally keep track of original hues
  # if(sameHue) {
  #   o.hue <- hue
  # }
  
  
  # manuscript in progress
  # latest models: soil-color/moist-dry-model/
  
  if(method == 'procrustes') {
    res <- .ESC_procrustes(.lab, .state = sourceMoistureState, .m = returnMunsell)
  }
  
  if(method == 'ols') {
    res <- .ESC_OLS(.lab, .state = sourceMoistureState, .m = returnMunsell)
  }
  
  
  
  ## TODO: post-processing or additional diagnostics?
  
  return(res)
}



.ESC_procrustes <- function(.lab, .state, .m) {
  
  # select transformation
  # transformation parameters via vegan::procrustes()
  # updated 2026-02-13 (NASIS pedon model)
  params <- switch(.state,
                   dry = {
                     # dry -> moist
                     list(
                       scale = 0.7728589,
                       
                       rotation = structure(
                         c(0.997992548998439, 0.0262814715398672, -0.0576207983049428, 
                           -0.0234753940832354, 0.998530378505848, 0.0488465871234181, 
                           0.0588198777305472, -0.047395859045575, 0.997142845598912), 
                         dim = c(3L, 3L)
                       ),
                       
                       translation = structure(
                         c(-1.36373229346336, 1.77648038381813, 0.205785719295584), 
                         dim = c(1L, 3L)
                       )
                     )
                   },
                   
                   moist = {
                     # moist -> dry  
                     list(
                       scale = 0.835203,
                       
                       rotation = structure(
                         c(0.997992548998438, -0.0234753940832355, 0.058819877730547, 
                           0.0262814715398672, 0.998530378505848, -0.0473958590455751, 
                           -0.0576207983049427, 0.048846587123418, 0.997142845598912), 
                         dim = c(3L, 3L)),
                       
                       translation = structure(
                         c(20.194756551203, -0.0896637372378102, 4.07182351403834), 
                         dim = c(1L, 3L)
                       )
                     )
                   }
  )
  
  
  # apply transformation
  Y <- as.matrix(.lab)
  Y <- params$scale * Y %*% params$rotation
  Y <- sweep(Y, MARGIN = 2, STATS = params$translation, FUN = "+")
  
  # CIELAB -> closest Munsell
  if(.m) {
    .res <- col2Munsell(Y, space = 'CIELAB', nClosest = 1)  
  } else {
    # CIELAB coordinate
    .res <- as.data.frame(Y)
    names(.res) <- c('L', 'A', 'B')
  }
 
  return(.res)
   
}


.ESC_OLS <- function(.lab, .state, .m) {
  
  # updated from NASIS pedons
  # ols(m_L ~ rcs(d_L, 3) + d_A + d_B, data = z)
  # 
  
  .dmL <- function(d_L = 51.396738, d_A = 2.8906427, d_B = 12.657859) {
    14.458147+0.34626849* d_L+0.00079258202*pmax(d_L-40.900674,0)^3-0.0012058348*pmax(d_L-51.396738,0)^3+0.00041325275*pmax(d_L-71.527256,0)^3+0.14117082*d_A+0.17977648*d_B 
  }
  
  .dmA <- function(d_A = 2.8906427,d_L = 51.396738,d_B = 12.657859) {
    -2.0541836+1.1614388* d_A-0.0028361881*pmax(d_A,0)^3+0.0043178943*pmax(d_A-2.8906427,0)^3-0.0014817062*pmax(d_A-8.423728,0)^3+0.052411824*d_L-0.05196947*d_B 
  }
  
  .dmB <- function(d_B = 12.657859,d_L = 51.396738,d_A = 2.8906427) {
    -5.5112278+0.73963288* d_B-0.00022874794*pmax(d_B-5.7294897,0)^3+0.00045814458*pmax(d_B-12.657859,0)^3-0.00022939663*pmax(d_B-19.566637,0)^3+0.1445493*d_L+0.21897497*d_A 
  }
  
  .mdL <- function(m_L = 40.90522,m_A = 3.2269026,m_B = 12.070734) {
    14.107465+1.030697* m_L-0.00062374023*pmax(m_L-30.240189,0)^3+0.0012570867*pmax(m_L-40.90522,0)^3-0.00063334643*pmax(m_L-51.40849,0)^3-0.55323226*m_A+0.23956231*m_B 
  }
  
  .mdA <- function(m_A = 3.2269026,m_L = 40.90522,m_B = 12.070734) {
    0.67518171+1.0529881* m_A-0.00046032631*pmax(m_A,0)^3+0.00071535947*pmax(m_A-3.2269026,0)^3-0.00025503316*pmax(m_A-9.0513538,0)^3-0.0037767466*m_L-0.089514303*m_B 
  }
  
  .mdB <- function(m_B = 12.070734,m_L = 40.90522,m_A = 3.2269026) {
    3.0242045+0.86434908* m_B-0.00088833302*pmax(m_B-5.5707716,0)^3+0.001658638*pmax(m_B-12.070734,0)^3-0.00077030501*pmax(m_B-19.566637,0)^3-0.038454218*m_L+0.28100782*m_A 
  }
  
  
  .P <- switch(
    .state, 
    'dry' = {
      data.frame(
        L = .dmL(d_L = .lab$L, d_A = .lab$A, d_B = .lab$B),
        A = .dmA(d_A = .lab$A, d_L = .lab$L, d_B = .lab$B),
        B = .dmB(d_B = .lab$B, d_L = .lab$L, d_A = .lab$A)
      )   
    }, 
    'moist' = {
      data.frame(
        L = .mdL(m_L = .lab$L, m_A = .lab$A, m_B = .lab$B),
        A = .mdA(m_A = .lab$A, m_L = .lab$L, m_B = .lab$B),
        B = .mdB(m_B = .lab$B, m_L = .lab$L, m_A = .lab$A)
      )
    })
  
  
  # CIELAB -> closest Munsell
  if(.m) {
    .res <- col2Munsell(.P, space = 'CIELAB', nClosest = 1)  
  } else {
    # CIELAB coordinate
    .res <- .P
  }
  
  return(.res)
}

#' Quickly assess relative state of site and horizon slots
#'
#' @description
#' 
#' Determine "state" of SoilProfileCollection before or after major modifications of site or horizon slot contents. 
#' 
#' Two logical checks are performed on the site and horizon tables, and a third element \code{valid} returns \code{TRUE} when both checks are \code{TRUE}.
#'
#' Check 1: Same number of sites in site as number of sites in horizons. No intermingling of IDs, no orphan horizons, no sites without horizons (for now)
#' 
#' Check 2: Site IDs match coalesced profile ID from horizons. Ensures the same *relative* ordering, but horizons still may be out of order within profiles
#' 
#' @param object A SoilProfileCollection 
#' @return data.frame
#' @docType methods
#' @author Andrew G. Brown
#' @export
#' @examples 
#' data(sp5)
#' 
#' spc_in_sync(sp5)
#' 
spc_in_sync <- function(object) {
  
  # get site and horizon slot contents
  s <- object@site
  h <- object@horizons
  
  # object profile ID name
  oid <- idname(object)
  hzidnm <- hzidname(object)
  
  # profile ID from site
  sid <- as.character(s[[oid]])
  
  # profile ID from horizon
  hid <- as.character(h[[oid]])
  hzid <- as.character(h[[hzidnm]])

  if (length(hid) == 0 & length(hzid)  == 0)  {
    # this is an empty soil profile collection
    return(data.frame(siteDepth = TRUE,
                    relativeOrder = TRUE,
                    depthOrder = TRUE,
                    valid = TRUE))
  }
  
  # coalesced horizon IDs 
  # identifies intermingling of profiles within horizon
  cohid <- rle(hid)$values
  
  # if cohid is longer than sid, horizons from different profiles
  # are mixed or IDs have been corrupted (by e.g. direct edit)
  # in less-common case, a site has been removed or duplicated
  
  # check 1: site order is reflected in horizons 
  #          no intermingling of sites within horizon table
  #          though, they _may_ be out of order within profiles
  one <- (length(sid) == length(cohid))
  
  # check 2: site IDs match coalesced profile ID from horizon
  #          this ensures the same _relative_ ordering, but
  #          horizons still may be out of order within profiles
  #          use checkHzDepthLogic -- as only one problematic pedon will break SPC depth order
  two <- ifelse(one, all(sid == cohid), FALSE)
  
  return(data.frame(nSites = one,
                    relativeOrder = two,
                    valid = all(one, two)))
}

# if (!isGeneric('reorderHorizons'))
  setGeneric('reorderHorizons', 
             function(object, target.order = NULL)
    standardGeneric('reorderHorizons'))

#' Re-order corrupted horizon data
#' 
#' @name reorderHorizons
#' @description This is a method made available primarily to repair horizon data that have been corrupted relative to their order at time of SoilProfileCollection construction. 
#' 
#' There is an option to specify the target order, but this will not update the corresponding metadata entry tracking the original order. Use this functionality at your own risk.
#' 
#' @param object A SoilProfileCollection
#' @param target.order A numeric vector of equal length to \code{object}. Default value is \code{NULL} which restores the internal order of the collection. 
#' @aliases reorderHorizons,SoilProfileCollection-method
#' @return SoilProfileCollection
#' @docType methods
#' @rdname reorderHorizons
#' @export
setMethod('reorderHorizons',
          signature(object = 'SoilProfileCollection', target.order = "ANY"),
          function(object, target.order = NULL) {
            
            h <- object@horizons
            
            stopifnot(!is.null(target.order))
            
            current.order <- match(target.order,
                                   order(as.character(h[[idname(object)]]),
                                         h[[horizonDepths(object)[1]]]))
            
            h <- .as.data.frame.aqp(h[current.order,], 
                                          aqp_df_class(object))
            object@horizons <- h
            return(object)
          })

#' @title Shannon Entropy
#' @description A very simple implementation of Shannon entropy.
#'
#' @param x vector of probabilities (0,1), must sum to 1, should not contain NA
#' @param b logarithm base
#' 
#' @details `0`s are automatically removed by `na.rm = TRUE`, as `(0 * log(0) = Nan)`
#' 
#' @note When `b = length(x)` the result is the normalized Shannon entropy of (Kempen et al, 2009).
#'
#' @return A single numeric value.
#' 
#' @references 
#' Kempen, Bas, Dick J. Brus, Gerard B.M. Heuvelink, and Jetse J. Stoorvogel. 2009. "Updating the 1:50,000 Dutch Soil Map Using Legacy Soil Data: A Multinominal Logistic Regression Approach." Geoderma 151: 311-26. doi:10.1016/j.geoderma.2009.04.023
#' 
#' Shannon, Claude E. (July-October 1948). "A Mathematical Theory of Communication". Bell System Technical Journal. 27 (3): 379-423. doi:10.1002/j.1538-7305.1948.tb01338.x
#' 
#' 
#' @export
#'
#' @examples
#' 
#' # a very simple example
#' p <- c(0.25, 0.25, 0.4, 0.05, 0.05)
#' 
#' shannonEntropy(p)
#' 
#' 
#' 
## TODO: test that sum(x) == 1
shannonEntropy <- function(x, b = 2) {
  # 0s automatically removed by na.rm=TRUE (0 * log(0) = Nan)
  # consider: ifelse(x == 0, 0, x * log(x))
  
  res <- -1 * sum(x * log(x, base = b), na.rm = TRUE)
  return(res)
}



#' @title Confusion Index
#' 
#' @description Calculate the confusion index of Burrough et al., 1997.
#'
#' @param x vector of probabilities (0,1), should not contain NA
#' 
#' @author D.E. Beaudette
#' 
#' @references Burrough, P.A., P.F.M. van Gaans, and R. Hootsmans. 1997. "Continuous Classification in Soil Survey: Spatial Correlation, Confusion and Boundaries." Geoderma 77: 115-35. doi:10.1016/S0016-7061(97)00018-9.
#' 
#' @return A single numeric value.
#' @export
#'
#' @examples
#' 
#' # a very simple example
#' p <- c(0.25, 0.25, 0.4, 0.05, 0.05)
#' confusionIndex(p)
#' 
#' # for comparison
#' shannonEntropy(p)
#' 
confusionIndex <- function(x) {
  x <- sort(x, decreasing = TRUE)
  res <- 1 - (x[1] - x[2])
  return(res)
}



#' @title Multinominal Brier Score
#' 
#' @description Compute a multinominal Brier score from predicted class probabilities and observed class label. Lower values are associated with a more accurate classifier.
#'
#' @param x `data.frame` of class probabilities (numeric) and observed class label (character), see examples
#' 
#' @param classLabels vector of predicted class labels (probabilities), corresponding to column names in `x`
#' @param actual name of column containing the observed class, should be character vector not factor
#' 
#' @references Brier, Glenn W. 1950. "Verification of Forecasts Expressed in Terms of Probability." Monthly Weather Review 78 (1): 1-3. doi:10.1175/1520-0493(1950)078<0001:VOFEIT>2.0.CO;2.
#' 
#' @author D.E. Beaudette
#' 
#' @return a single Brier score, representative of data in `x`
#' @export
#'
#' @examples
#' 
#' # columns 'a', 'b', 'c' contain predicted probabilities
#' # column 'actual' contains observed class label
#' 
#' # a good classifier
#' d.good <- data.frame(
#'   a = c(0.05, 0.05, 0.10),
#'   b = c(0.90, 0.85, 0.75),
#'   c = c(0.05, 0.10, 0.15),
#'   actual = c('b', 'b', 'b'),
#'   stringsAsFactors = FALSE
#' )
#' 
#' # a rather bad classifier
#' d.bad <- data.frame(
#'   a = c(0.05, 0.05, 0.10),
#'   b = c(0.90, 0.85, 0.75),
#'   c = c(0.05, 0.10, 0.15),
#'   actual = c('c', 'c', 'c'),
#'   stringsAsFactors = FALSE
#' )
#' 
#' # class labels are factors
#' d.factors <- data.frame(
#'   a = c(0.05, 0.05, 0.10),
#'   b = c(0.90, 0.85, 0.75),
#'   c = c(0.05, 0.10, 0.15),
#'   actual = c('b', 'b', 'b'),
#'   stringsAsFactors = TRUE
#' )
#' 
#' # relatively low value = accurate
#' brierScore(x = d.good, classLabels = c('a', 'b', 'c'), actual = 'actual')
#' 
#' # high values = not accuate
#' brierScore(x = d.bad, classLabels = c('a', 'b', 'c'), actual = 'actual')
#' 
#' # message related to conversion of factor -> character
#' brierScore(x = d.factors, classLabels = c('a', 'b', 'c'), actual = 'actual')
#' 
brierScore <- function(x, classLabels, actual = 'actual') {
  if (inherits(x, 'data.frame')) {
    x <- data.frame(x)
  } else stop("`x` should be a data.frame", call. = FALSE)
  
  # number of observations
  n <- nrow(x)
  
  # sanity check: no factors allowed in class labels
  if(inherits(x[[actual]], 'factor')) {
    message('converting `actual` from factor to character')
    x[[actual]] <- as.character(x[[actual]])
  }
  
  # extract vector of observed classes
  x.actual <- x[[actual]]
  
  # keep only probabilities as matrix
  x <- as.matrix(x[, classLabels, drop = FALSE])
  
  # init new matrix to store most-likely class
  m <- matrix(0, ncol = ncol(x), nrow = n)
  # same structure as x.pr
  dimnames(m)[[2]] <- classLabels
  
  # set cells of actual observed outcome to 1
  for(i in 1:n) {
    x.i <- x.actual[i]
    m[i, x.i] <- 1
  }
  
  # compute multinominal brier score
  # 1/n * sum((x - m)^2)
  # x: matrix of predictions
  # m: indicator matrix of outcomes
  bs <- (1/n) * sum((x - m)^2, na.rm = TRUE)
  return(bs)
}


#' Map unit confusion matrix and other classification measures
#' 
#' This function reverse engineers a confusion matrix and other classification measures from soil map unit component percentages (i.e. composition) and area (i.e. acres).
#' @param x `data.frame`
#' @param mapunit `character` column name containing the mapunit identifier (e.g. nationalmusym)
#' @param cophase `character` column name containing the soil component phase identifier (e.g. coiid or paste(compname, localphase))
#' @param comppct `character` column name containing the component percent (e.g. comppct_r)
#' @param muacres `character` column name containing the total area of the mapunit (e.g. muacres)
#' 
#' @details `mu_confusion_matrix` There are several common statistical measures used to gauge the accuracy of categorical maps. These measures are typically not estimated for soil surveys but can be inferred from a map unit’s soil component composition percentages and size (i.e. acres). In general, overall purity or accuracy (OP) is related to map unit kind (e.g. consociations vs complexes). For several reasons, the “true” accuracies are unknown, and these values should be interpreted as Bayesian prior estimates. However, it is likely that the estimates are optimistic if they haven't been derived from an external validation. Existing and future digital soil mapping products could be used to validate how optimistic the current OA estimates are.
#'
#' @return `list` a confusion matrix, overall purity (OP) (i.e. overall accuracy), map unit purity (MP) (i.e. user accuracy), class representation (CR) (i.e. producer accuracy), and Shannon entropy (E). The measure names were selected to be consistent with the alternative terms proposed by Lark (1995) and Brus (2011). 
#' 
#' @author Stephen Roecker
#' 
#' @references
#'
#'    - Brus DJ, Kempen B, Heuvelink GBM. 2011. Sampling for validation of digital soil maps. European Journal of Soil Science. 62(3):394–407. \doi{10.1111/j.1365-2389.2011.01364.x}
#'    - Lark RM. 1995. Components of accuracy of maps with special reference to discriminant analysis on remote sensor data. International Journal of Remote Sensing. 16(8):1461–1480. \doi{10.1080/01431169508954488}
#' 
#' @seealso [tauW()], [shannonEntropy()], [confusionIndex()]
#' 
#' @export
#'
#' @examples
#' 
#' # example data
#' mu <- rbind(
#'     data.frame(mapunit = "A", cophase = c("Alpha", "Beta"), comppct = c(90, 10), muacres = 100),
#'     data.frame(mapunit = "B", cophase = c("Beta", "Alpha"), comppct = c(70, 30), muacres = 1000)
#'     )
#' 
#' mu_confusion_matrix(mu, mapunit = "mapunit", cophase = "cophase", comppct = "comppct")
#' 

mu_confusion_matrix <- function(x, mapunit = "nationalmusym", cophase = "coiid", comppct = "comppct_r", muacres = "muacres") {
  
  vars <- c(mapunit = mapunit, cophase = cophase, comppct = comppct, muacres = muacres)
  idx  <- sapply(vars, function(y) which(names(x) %in% y))
  names(x)[idx] <- names(vars)
  x <- x[names(vars)]
  
  x$coacres <- x$muacres * x$comppct/100
  
  CM <- xtabs(coacres ~ mapunit + cophase, data = x)
  rowmax <- apply(CM, 1, max)
  colmax <- apply(CM, 2, max)
  MP <- rowmax / rowSums(CM)
  CR <- colmax / colSums(CM)
  OP <- sum(rowmax) / sum(CM)
  H  <- apply(CM, 1, function(x) shannonEntropy(x/sum(x)))
  # CM <- CM |> cbind(MP, H) |>  rbind(CR = c(CR, c(NA, NA)))
  names(dimnames(CM)) <- c("map unit", "component phase")
  
  return(
    list(
      confusion_matrix     = round(CM, 2), 
      overall_purity       = OP, 
      map_purity           = MP, 
      class_representation = CR, 
      shannon_entropy      = H
    )
  )
}

context("classification accuracy and uncertainty")

# a good classifier
d.good <- data.frame(
  a = c(0.05, 0.05, 0.10),
  b = c(0.90, 0.85, 0.75),
  c = c(0.05, 0.10, 0.15),
  actual = c('b', 'b', 'b'),
  stringsAsFactors = FALSE
)

# a rather bad classifier
d.bad <- data.frame(
  a = c(0.05, 0.05, 0.10),
  b = c(0.90, 0.85, 0.75),
  c = c(0.05, 0.10, 0.15),
  actual = c('c', 'c', 'c'),
  stringsAsFactors = FALSE
)

# class labels are factors
d.factors <- data.frame(
  a = c(0.05, 0.05, 0.10),
  b = c(0.90, 0.85, 0.75),
  c = c(0.05, 0.10, 0.15),
  actual = c('b', 'b', 'b'),
  stringsAsFactors = TRUE
)

# hypothetical three class probabilitites: 0.5 0.3 0.2
probs <- c(5,3,2) / 10

test_that("shannonEntropy", {
  expect_equal(round(shannonEntropy(probs), 4), 1.4855)
  expect_equal(round(shannonEntropy(probs, 3), 4), 0.9372)
})

test_that("confusionIndex", {
  expect_equal(confusionIndex(probs), 0.8)
})

test_that("brierScore", {
  dff <- data.frame(diag(3))
  expect_equal(brierScore(dff, c("X1","X2","X3")), 1)

  set.seed(1)
  dff2 <- data.frame(matrix(runif(9), 3, 3))
  expect_equal(round(brierScore(dff2, c("X1","X2","X3")), 2), 1.31)
  
  # factor -> character conversion emits a message
  expect_message(brierScore(x = d.factors, classLabels = c('a', 'b', 'c'), actual = 'actual'))
  
  # results should be the same
  bs1 <- brierScore(x = d.good, classLabels = c('a', 'b', 'c'), actual = 'actual')
  bs2 <- brierScore(x = d.factors, classLabels = c('a', 'b', 'c'), actual = 'actual')
  
  expect_equal(bs1, bs2)
})

#' @title Transform a SPC with expressions based on site or horizon level attributes
#' 
#' @rdname transform
#' 
#' @aliases mutate mutate,SoilProfileCollection-method
#' @description \code{transform()} is a function used for modifying columns in SoilProfileCollections. 
#' 
#' It allows the user to specify an arbitrary number of expressions that resolve to the (re-)calculation of one or more site or horizon level attributes. For instance: \code{mutate(spc, thickness = hzdepb - hzdept)}. The expressions may depend on one another, and are evaluated from left to right.
#' 
#' @param _data A SoilProfileCollection
#' 
#' @param ... Comma-separated set of R expressions e.g. \code{thickness = hzdepb - hzdept, hzdepm = hzdept + round(thk / 2)}
#' 
#' @return A SoilProfileCollection
#' @author Andrew G. Brown.
#'
#' @export 
setMethod("transform", signature(`_data` = "SoilProfileCollection"), 
          function(`_data`, ...)  {

  # create composite object to facilitate eval
  .data <- compositeSPC(`_data`)

  # capture expression(s) at function
  .dots <- substitute(list(...))
  .dots <- .dots[2:length(.dots)]
  .names <- names(.dots)

  if (is.null(.names))
    .names <- as.character(.dots)

  for(n in 1:length(.dots)) {
    foo <- .data_dots(.data, eval(.dots[[n]]))[[1]]
    `_data`[[.names[n]]] <- foo
    .data[[.names[n]]] <- foo
  }

  return(`_data`)
})

#' Slicing of SoilProfileCollection Objects
#'
#' A method for "slicing" of SoilProfileCollection objects into constant depth intervals. Now deprecated, see `[dice()]`.
#'
#' @name slice-methods
#' 
#' @aliases get.slice slice.fast slice slice,SoilProfileCollection-method
#' 
#' @docType methods
#' 
#' @param object a SoilProfileCollection
#' 
#' @param fm A formula: either `integer.vector ~ var1 + var2 + var3` where
#' named variables are sliced according to `integer.vector` OR where all
#' variables are sliced according to `integer.vector`: `integer.vector ~ .`.
#' 
#' @param top.down logical, slices are defined from the top-down: `0:10` implies 0-11 depth units.
#' 
#' @param just.the.data Logical, return just the sliced data or a new `SoilProfileCollection` object.
#' 
#' @param strict Logical, should the horizonation be strictly checked for self-consistency?
#' 
#' @return Either a new `SoilProfileCollection` with data sliced according to `fm`, or a `data.frame`.
#' 
#' @section Details: By default, slices are defined from the top-down:
#' \code{0:10} implies 0-11 depth units.
#' 
#' @author D.E. Beaudette
#' 
#' @seealso \code{\link{slab}}
#' 
#' @references D.E. Beaudette, P. Roudier, A.T. O'Geen, Algorithms for
#' quantitative pedology: A toolkit for soil scientists, Computers &
#' Geosciences, Volume 52, March 2013, Pages 258-268,
#' 10.1016/j.cageo.2012.10.020.
#' 
#' @keywords methods manip
#' @export
#' @rdname slice
#' @examples
#'
#' library(aqp)
#'
#' # simulate some data, IDs are 1:20
#' d <- lapply(1:20, random_profile)
#' d <- do.call('rbind', d)
#'
#' # init SoilProfileCollection object
#' depths(d) <- id ~ top + bottom
#' head(horizons(d))
#'
#' # generate single slice at 10 cm
#' # output is a SoilProfileCollection object
#' s <- dice(d, fm = 10 ~ name + p1 + p2 + p3)
#'
#' # generate single slice at 10 cm, output data.frame
#' s <- dice(d, 10 ~ name + p1 + p2 + p3, SPC = FALSE)
#'
#' # generate integer slices from 0 - 26 cm
#' # note that slices are specified by default as "top-down"
#' # result is a SoilProfileCollection
#' # e.g. the lower depth will always by top + 1
#' s <- dice(d, fm = 0:25 ~ name + p1 + p2 + p3)
#' par(mar=c(0,1,0,1))
#' plotSPC(s)
#'
#' # generate slices from 0 - 11 cm, for all variables
#' s <- dice(d, fm = 0:10 ~ .)
#' print(s)
#'
#' # compute percent missing, for each slice,
#' # if all vars are missing, then NA is returned
#' d$p1[1:10] <- NA
#' s <- dice(d, 10 ~ ., SPC = FALSE, pctMissing = TRUE)
#' head(s)
#'
#' \dontrun{
#' ##
#' ## check sliced data
#' ##
#'
#' # test that mean of 1 cm slices property is equal to the
#' # hz-thickness weighted mean value of that property
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#'
#' # get the first profile
#' sp1.sub <- sp1[which(profile_id(sp1) == 'P009'), ]
#'
#' # compute hz-thickness wt. mean
#' hz.wt.mean <- with(
#'   horizons(sp1.sub),
#'   sum((bottom - top) * prop) / sum(bottom - top)
#' )
#'
#' # hopefully the same value, calculated via slice()
#' s <- dice(sp1.sub, fm = 0:max(sp1.sub) ~ prop)
#' hz.slice.mean <- mean(s$prop, na.rm = TRUE)
#'
#' # they are the same
#' all.equal(hz.slice.mean, hz.wt.mean)
#' }
#'
slice.fast <- function(object, fm, top.down=TRUE, just.the.data=FALSE, strict=TRUE){

  ## Plan:
  #   1. message about future deprecation -> dice()
  # message('Note: aqp::slice() will be deprecated in aqp version 2.0\n--> Please consider using the more efficient aqp::dice()')
  
  #   2. deprecation -> dice() in aqp 2.0
  .Deprecated(new = 'dice', msg = 'slice() has been deprecated, please use the more efficient aqp::dice()')
  
  #   3. masking / removal of slice(), shortly there after
  
  
  ## important: change the default behavior of data.frame and melt
  opt.original <- options(stringsAsFactors = FALSE)

  # test for logical input
  if(! inherits(fm, "formula")) {
    stop('must provide a valid formula: ~ var1 + var2 + ...', call.=FALSE)
  }


  # extract components of the formula:
  formula <- paste(deparse(fm, 500), collapse = "")
  elements <- strsplit(x = fm, split = '~', fixed = )[[1]]
  formula <- lapply(strsplit(x = elements, split = '[+*]'), trimws)

  # TODO: this will have to be changed when we implement no LHS = all slices
  if (length(formula) > 2)
    stop("please provide a valid formula", call.=FALSE)

  # extract parsed formula components
  vars <- formula[[2]] # RHS, simple enough
  # LHS: could be either single integer or vector of slices
  z <- as.numeric(eval(parse(text=formula[[1]])))

  # get horizons + depth column names + ID column name
  h <- horizons(object)
  hd <- horizonDepths(object)
  top <- hd[1] ; bottom <- hd[2] # convenience vars
  id <- idname(object)
  id.order <- profile_id(object) # this is the original ordering of profiles

  # check for bogus left/right side problems with the formula
  if(any(z < 0) | any(is.na(z)))
    stop('z-slice must be >= 1', call.=FALSE)

  ## TODO: this will have to be updated for z-slices defined by data in @site
  if(! inherits(z,  c('numeric','integer'))) # bogus z-slice
		stop('z-slice must be either numeric or integer', call.=FALSE)

  # check for '.' --> all variables, minus ID/depths
  if(any(vars == '.')) {
  	nh <- names(h)
  	cols.to.remove.idx <- match(c(id, top, bottom), nh)
  	vars <- nh[-cols.to.remove.idx]
  }


  # check for column names that don't exist
  if(any(vars %in% names(h)) == FALSE) # bogus column names in right-hand side
		stop('column names in formula do not match any horizon data', call.=FALSE)


  ## extract all vars by slice_i
  # pre-allocate storage as list
  hd.slices <- vector(mode='list', length=length(z))
  # prepare an index for the list
  slice.idx <- seq_along(z)

  # convert h into an imutable data.frame for speed
  # h <- idata.frame(h)

  ## TODO: list-based processing faster
  # iterate over this index
  for(slice.i in slice.idx) {

    # extract all vars for current slice
    # TODO: this is wasteful as the entire pile of horizons is passed to get.slice in each iteration of the loop
    m.i.sub <- get.slice(h, id=id, top=top, bottom=bottom, vars=vars, z=z[slice.i], strict=strict)

    # join with original IDs in order to account for NA, or bad horizonation
    d <- data.frame(temp_id = id.order)
    names(d) <- id

    m.i <- merge(d, m.i.sub[!duplicated(m.i.sub[[id]]),], by = id, all.x = TRUE, sort = FALSE)

    # add depth range:
    # top-down, means that the slice starts from the user-defined depths (default)
    if(top.down) {
      m.i[[top]] <- z[slice.i]          # "top"
      m.i[[bottom]] <- z[slice.i] + 1   # "bottom"
    }
    # otherwise, the slice starts at the bottom (why would someone do this?)
    else {
      m.i[[top]] <- z[slice.i] - 1      # "top"
      m.i[[bottom]] <- z[slice.i]       # "bottom"
    }

    # save to the list
    hd.slices[[slice.i]] <- m.i
  }

  # convert list into DF
  hd.slices <- data.frame(data.table::rbindlist(hd.slices))

  # re-order by id, then top
  # keep only data we care about
  # note that we have a new column in there used to store pct not NA
  hd.slices <- hd.slices[order(match(hd.slices[[id]], id.order), hd.slices[[top]]), c(id, top, bottom, vars, '.pctMissing')]

  # if we just want the data:
  if(just.the.data)
    return(hd.slices)

  ## otherwise return an SPC, be sure to copy over the spatial data

  # init new SPC
  # if all horizon attr are requested then a warning will be issued
  # `hzID` is not a unique horizon ID, using `hzID_`
  suppressWarnings(depths(hd.slices) <- as.formula(paste(id, '~', top, '+', bottom)))

  # reset auto-generated horizon ID so that we know it is now the slice ID
  idx <- match(hzidname(hd.slices), horizonNames(hd.slices))
  horizonNames(hd.slices)[idx] <- 'sliceID'
  hzidname(hd.slices) <- 'sliceID'

  # # copy spatial data
  # hd.slices@sp <- object@sp

  # safely copy site data via JOIN
  site(hd.slices) <- site(object)

  # copy over any diagnostic features/restrictions
  hd.slices@diagnostic <- object@diagnostic
  hd.slices@restrictions <- object@restrictions

  # copy over metadata
  hd.slices <- .transfer.metadata.aqp(object, hd.slices)

  ## un-set horizon designation and texture class metadata if they aren't in the sliced data
  hn <- horizonNames(hd.slices)

  # hz designation
  # this will be TRUE if hzdesgnname(object) is unset ('')
  if(! hzdesgnname(object) %in% hn) {
    hzdesgnname(hd.slices) <- ''
  }

  # texture class
  # this will be TRUE if hztexclname(object) is unset ('')
  if(! hztexclname(object) %in% hn) {
    hztexclname(hd.slices) <- ''
  }

  # reset options:
  options(opt.original)

  # done
  return(hd.slices)
}

setGeneric("slice", function(object, fm, top.down = TRUE, just.the.data = FALSE, strict = TRUE) 
  standardGeneric("slice"))

#' @export
#' @rdname slice
setMethod(f = 'slice', signature(object = 'SoilProfileCollection'), slice.fast)


# this function is run on the horizon data, once for each depth slice
#' @param h Horizon data.frame
#' @param id Profile ID
#' @param top Top Depth Column Name
#' @param bottom Bottom Depth Column Name
#' @param vars Variables of Interest
#' @param z Slice Depth (index).
#' @param include Either `'top'` or `'bottom'`. Boundary to include in slice. Default: `'top'`
#' @param strict Check for logic errors? Default: `TRUE`
#'
#' @export
#' @rdname slice
get.slice <- function(h, id, top, bottom, vars, z, include='top', strict=TRUE) {
  
  ## TODO: this is likely very slow
  # 1. get indices to rows matchings current depth slice (z)
  # this is the default method
  if(include == 'top')
    idx <- which(z >= h[[top]] & z < h[[bottom]])
  # not sure why someone would use this approach, but include it anyways
  if(include == 'bottom')
    idx <- which(z > h[[top]] & z <= h[[bottom]])
  
  ## TODO: split -> list -> process -> combine
  # 2. extract data.frame along slice, and named vars + id
  h <- h[idx, c(id, vars)]
  
  # 2.5 compute fraction missing
  # if there is only 1 variable, don't try to compute this value
  # if all data are missing NA is returned
  h$.pctMissing <- apply(as.matrix(h[, vars]), 1, function(i, n=length(vars)) length(which(is.na(i))) / n)
  
  # 3. QA/QC
  # how many unique IDs?
  l.ids <- length(unique(h[[id]]))
  # how many rows in the result?
  n.res <- nrow(h)
  
  # more rows than IDs --> bad horizonation
  if(l.ids != n.res) {
    if(strict == TRUE) {
      # get offending IDs
      id.tab <- table(h[[id]])
      bad.ids <- paste(names(id.tab)[which(id.tab > 1)], collapse=', ')
      stop(paste('bad horizonation in IDs:', bad.ids), call.=FALSE)
    }
    
    # looser interp of the data... issue warning and return multuple rows/ID
    # join(..., match='first') will correct the problem
    else
      warning('Bad horizonation detected, first matching horizon selected. Use strict=TRUE to enforce QA/QC.')
  }
  
  # done: return subset of original data + pct not NA
  return(h)
}

#' Calculate the minimum thickness requirement for Mollic epipedon
#'
#' @description Utilize horizon depths, designations and textures in a profile to estimate the thickness requirement for the Mollic or Umbric epipedon, per criterion 6 in the U.S. Keys to Soil Taxonomy (12th Edition).
#'
#' @param p A single-profile SoilProfileCollection.
#' @param hzdesgn Column in horizon table containing designations. Default: `guessHzDesgnName(p)`
#' @param texcl.attr Column in horizon table containing texture classes. Default: \code{guessHzTexClName(p)}
#' @param clay.attr Column in horizon table containing clay contents. Default: \code{guessHzAttrName(p, 'clay', c('total','_r'))}
#' @param truncate Should sliding scale (Criterion 6C) results be truncated to 18 to 25cm interval? (Experimental; Default: TRUE)
#'
#' @return A unit length numeric vector containing Mollic or Umbric epipedon minimum thickness requirement.
#'
#' @author Andrew G. Brown
#' @export mollic.thickness.requirement
#'
#' @examples
#' # construct a fake profile
#' spc <- data.frame(id=1, taxsubgrp = "Lithic Haploxeralfs",
#'                   hzname   = c("A","AB","Bt","BCt","R"),
#'                   hzdept   = c(0,  20, 32, 42,  49),
#'                   hzdepb   = c(20, 32, 42, 49, 200),
#'                   prop     = c(18, 22, 28, 24,  NA),
#'                   texcl    = c("l","l","cl", "l","br"),
#'                   d_value  = c(5,   5,  5,  6,  NA),
#'                   m_value  = c(2.5, 3,  3,  4,  NA),
#'                   m_chroma = c(2,   3,  4,  4,  NA))
#'
#' # promote to SoilProfileCollection
#' depths(spc) <- id ~ hzdept + hzdepb
#' hzdesgnname(spc) <- 'hzname'
#' hztexclname(spc) <- 'texcl'
#'
#' # print results in table
#' data.frame(id = spc[[idname(spc)]],
#'            thickness_req = mollic.thickness.requirement(spc, clay.attr='prop'),
#'            thickness_req_nobound = mollic.thickness.requirement(spc,
#'                                         clay.attr='prop', truncate=FALSE))
#'
mollic.thickness.requirement <- function(p, 
                                         hzdesgn = hzdesgnname(p, required = TRUE),
                                         texcl.attr = hztexclname(p, required = TRUE),
                                         clay.attr = hzmetaname(p, "clay", required = TRUE),
                                         truncate = TRUE) {

  hzd <- horizonDepths(p)
  buf <- rep(NA_real_, length(p))
  
  if (nrow(p) == 0) {
    return(numeric(0))
  }

  # determine boundaries
  # For purposes of identification of minimum thickness of mollic for field descriptions
  #   technically it is not applying the true taxonomic rules b/c it is based on hz desgn
  mss <- getMineralSoilSurfaceDepth(p, hzdesgn = hzdesgn, simplify = FALSE)

  soil_depth <- minDepthOf(p, pattern = "Cr|R|Cd|m",
                           hzdesgn = hzdesgn,
                           no.contact.depth = 200,
                           no.contact.assigned = NA,
                           simplify = FALSE)
  .LAST <- NULL
  idx1 <- which(soil_depth[[hzd[1]]] == 0)
  buf[idx1] <- 10

  #  if criteria aren't met in at least some part of upper 25 of mineral soil material, they won't be met deeper
  #  it is possible there are criteria greater than 25cm that affect total mollic thickness, but here we
  #  are only calculating the thickness _requirement_

  # get horizon data within mineral soil surface "critical zone"
  cztop <- mss[[hzd[2]]]
  czbot <- cztop + pmin(soil_depth[[hzd[1]]] - cztop, 25, na.rm = TRUE)
  deepest.bot.depth <- p[, , .LAST][[hzd[2]]]
  
  # soils with no mineral soil material have NA minimum thickness
  buf[which(cztop >= deepest.bot.depth)] <- NA
  
  epi <- trunc(p, cztop, czbot)

  # get horizon data from mineral soil surface to bedrock, physical root restriction,
  #  or pedogenic cementation >90% OR bottom of profile OR 250cm (~beyond SCS)
  soil.bottom <- pmin(soil_depth[[hzd[1]]], deepest.bot.depth, 250, na.rm = TRUE)
  sol <- trunc(p, mss[[hzd[2]]], soil.bottom)

  # 6C1 - TODO: create functions for testing basic field criteria for these diagnostics
  #       and lab data extension to support carbonates/calcic
  # cemented, carbonates or fragipan depth
  pan_depth <- minDepthOf(p,
                          pattern = "m|k|x",
                          hzdesgn = hzdesgn,
                          no.contact.assigned = NA,
                          simplify = FALSE)[[hzd[1]]]
  
  # calculate "shallowest of secondary carbonates/calcic, petrocalcic, duripan, fragipan"
  crit6c1 <- pan_depth 
  
  # 6C2 - identify argillic boundaries via aqp::getArgillicBounds
  # TODO: create functions for testing basic field criteria for these diagnostics (color structure)
  #       and lab data extension for spodic materials; can use logic from "sandy cambic" demo -- but more is needed
  argi_bounds <- getArgillicBounds(p, hzdesgn = hzdesgn, 
                                   clay.attr = clay.attr, 
                                   texcl.attr = texcl.attr, 
                                   simplify = FALSE)
  argillic_bottom <- argi_bounds$lbound

  # AGAIN, for purposes of identification of minimum thickness of mollic this is fast and probably fine
  #   but technically it is not applying the true taxonomic rules b/c it is based on hz desgn

  natric_oxic_spodic_cambic_bottom <- maxDepthOf(p,
                                                 pattern = "n|o|h|s[^s]",
                                                 top = FALSE,
                                                 no.contact.assigned = NA,
                                                 simplify = FALSE)
  
  #suppressWarnings(max(getCambicBounds(p, argi_bounds = argi_bounds)$cambic_bottom, na.rm=TRUE))
  
  .I <- NULL; cambic_bottom <- NULL
  
  .maxcmb <- function(x) {
    res <- suppressWarnings(max(x, na.rm = TRUE))
    if (!is.finite(res) || length(x) == 0) return(NA_real_)
    res
  }
  cmb <- data.table::data.table(getCambicBounds(p, 
                                                hzdesgn = hzdesgn,
                                                clay.attr = clay.attr,
                                                texcl.attr = texcl.attr,
                                                argi_bounds = argi_bounds))[,
                                .maxcmb(cambic_bottom), by = c(idname(p))]$V1

  # calculate "deepest of lower boundary of argillic/natric, cambic, oxic, spodic"
  crit6c2 <- pmax(
    argillic_bottom,
    natric_oxic_spodic_cambic_bottom[[hzd[2]]],
    cmb,
    na.rm = TRUE
  )

  # SHORT CIRCUITS
  depi <- data.table::data.table(id = horizons(epi)[[idname(p)]], texcl = epi[[texcl.attr]])
  
  sandy.textures <- depi[, list(all(
       grepl("S$", .SD[["texcl"]], ignore.case = TRUE) 
       & !grepl("LVFS$|LFS$", .SD[["texcl"]], ignore.case = TRUE)
    )), by = "id"]$V1
  
  # 6A - assumes you must check for sandy textures throughout 0-25 to trigger minimum thickness of 25cm
  #      there is an implicit assumption that after mixing any non-sandy texture into sandy texture
  #      you would have non-sandy. this logic could be altered to be more "restrictive" by changing
  #      all() to any() -- forcing the 25cm requirement.
  #
  #      in that case, _any_ sandy texture would imply sandy textures, importantly: precluding
  #      the 10cm requirement.
  buf[sandy.textures] <- 25

  maxdepth <- sol[ , , .LAST][[hzd[2]]]    
  # 6B - if all horizons above a contact are non-sandy and meet all mollic characteristics then
  #      the minimum thickness could be only 10cm, we have filtered out sandy textures
  #
  #      technically, the 10cm requirement requires knowledge of whether other mollic requirements are met,
  #      so gives the epipedon a somewhat circular definition. This case applies in deep soils with mollic materials
  #      all the way to contact -- but only practically matters where a contact is less than 25cm depth,
  #      so that is when we return it.
  buf[maxdepth < 25] <- 10
  
  # no diagnostics present
  crit6c <- !is.na(crit6c1) & !is.na(crit6c2) 
  
  # TODO: "fluventic" soils with no other diagnostics have minimum thickness of 25cm  
  #     # for now, hope that the assigned "taxsubgrp" reflects floodplain condition where it matters
  #     ## for example:
  #     # tolerance <- 0.1 # set a threshold for "different" carbon contents of 0.1%
  #     #                  # carbon-depth "fluctuations" less or equal to tolerance will be ignored
  #     #                  # for different data sources this number may need to be higher
  # 
  #     # irregular.decrease <- diff(epi$estimated_organic_carbon) < 0
  #     # if(length(irregular.decrease[irregular.decrease >= tolerance]) {
  #     #  return(25)
  #     # }
  #     # TODO: most pedons don't have OC measured, develop color-carbon-depth-spatial surrogate model?
  buf[crit6c & grepl("fluv|cumulic", site(p)$taxsubgrp)] <- 25

  # calculate the most restrictive requirement from 6a, 6b, 6c
  #   which contain several restatements of the fundamental criteria for a sliding scale thickness
  sixcdepths <- pmax(crit6c1 / 3, crit6c2 / 3, na.rm = TRUE)

  # the thickness requirement is based on the most restrictive condition
  sixcdepths.t <- sixcdepths
  
  # sliding scale depths based on diagnostics are truncated to [18, 25]
  if (truncate) {
    sixcdepths.t[sixcdepths > 25] <- 25
    sixcdepths.t[sixcdepths < 18] <- 18
  }
  buf[is.na(buf)] <- sixcdepths.t[is.na(buf)]
  
  # this is criterion 6d (if "none of above" apply)
  buf[is.na(buf)] <- 18
  
  buf
}

#' Find horizons with colors darker than a Munsell hue, value, chroma threshold
#'
#' @description \code{hasDarkColors} returns a boolean value by horizon representing whether darkness thresholds are met. The code is fully vectorized and deals with missing data and optional thresholds.
#'
#' Default arguments are set up for "5-3-3 colors" -- the basic criteria for Mollic/Umbric epipedon/mineral soil darkness. Any of the thresholds or column names can be altered. Any thresholds that are set equal to \code{NA} will be ignored.
#'
#' @param p A SoilProfileCollection.
#' @param d_hue Optional: character vector of dry hues to match (default: NA)
#' @param m_hue Optional: character vector of moist hues to match (default: NA)
#' @param d_value Maximum value of dry value (default: 5)
#' @param d_chroma Optional: Maximum value of dry chroma (default: NA)
#' @param m_value Maximum value of moist value (default: 3)
#' @param m_chroma Maximum value of moist chroma (default: 3)
#' @param dhuenm Column name containing dry hue.
#' @param dvalnm Column name containing dry value.
#' @param dchrnm Column name containing dry chroma.
#' @param mhuenm Column name containing moist hue.
#' @param mvalnm Column name containing moist value.
#' @param mchrnm Column name containing moist chroma.
#'
#' @return Boolean value (for each horizon in \code{p}) reflecting whether "darkness" criteria are met.
#'
#' @author Andrew G. Brown
#'
#' @export hasDarkColors
#'
#' @examples
#' # construct a fake profile
#' spc <- data.frame(id=1, taxsubgrp = "Lithic Haploxeralfs",
#'                   hzdesgn  = c("A","AB","Bt","BCt","R"),
#'                   hzdept   = c(0, 20, 32, 42,  49),
#'                   hzdepb   = c(20, 32, 42, 49, 200),
#'                   d_value  = c(5,   5,  5,  6,  NA),
#'                   m_value  = c(2.5, 3,  3,  4,  NA),
#'                   m_chroma = c(2,   3,  4,  4,  NA))
#'
#' # promote to SoilProfileCollection
#' depths(spc) <- id ~ hzdept + hzdepb
#'
#' # print results in table
#' data.frame(id = spc[[idname(spc)]],
#'            hz_desgn = spc$hzdesgn,
#'            has_dark_colors = hasDarkColors(spc))
#'
hasDarkColors <-  function(p, d_hue=NA, m_hue=NA, d_value=5, d_chroma=NA, m_value=3, m_chroma=3,
                           dhuenm='d_hue', dvalnm = "d_value", dchrnm="d_chroma",
                           mhuenm='m_hue', mvalnm = "m_value", mchrnm = "m_chroma") {
  hz <- horizons(p)
  r <- matrix(nrow = nrow(hz), ncol = 6)

  r1 <- hz[[dhuenm]] %in% d_hue
  r2 <- hz[[mhuenm]] %in% m_hue
  r3 <- hz[[dvalnm]] <= d_value
  r4 <- hz[[dchrnm]] <= d_chroma
  r5 <- hz[[mvalnm]] <= m_value
  r6 <- hz[[mchrnm]] <= m_chroma

  if(length(r1) > 0) r[,1] <- r1
  if(length(r2) > 0) r[,2] <- r2
  if(length(r3) > 0) r[,3] <- r3
  if(length(r4) > 0) r[,4] <- r4
  if(length(r5) > 0) r[,5] <- r5
  if(length(r6) > 0) r[,6] <- r6

  required <- !is.na(c(d_hue, m_hue, d_value, d_chroma, m_value, m_chroma))
  risna <- apply(r, 2, function(x) all(is.na(x)))

  # if all data are missing, return NA for each horizon
  if(all(required == risna)) {
    return(rep(NA, nrow(p)))
  }

  # return only required results
  r <- r[,required]

  # result may contain NA for individual horizons without color data
  if(is.null(dim(r))) {
    # single horizon result
    return(all(r))
  } else {
    # multi horizon result
    return(apply(r, 1, all))
  }
}

context("estimate mollic epipedon bounds")
# construct a fake profile
spc <- data.frame(
  id = 1,
  taxsubgrp = "Lithic Haploxeralfs",
  hzname   = c("A", "AB", "Bt", "BCt", "R"),
  hzdept   = c(0,  20, 32, 42,  49),
  hzdepb   = c(20, 32, 42, 49, 200),
  prop     = c(18, 22, 28, 24,  NA),
  texcl    = c("l", "l", "cl", "l", "br"),
  d_value  = c(5,   5,  5,  6,  NA),
  m_value  = c(2.5, 3,  3,  4,  NA),
  m_chroma = c(2,   3,  4,  4,  NA)
)

spc2 <- data.frame(
  id = c(1, 1, 1, 1),
  hzname = c("A", "Bt1", "Bt2", "Bk"),
  hzdept = c(0L, 5L, 16L, 36L),
  hzdepb = c(5L, 16L, 36L, 66L),
  claytotest = c(11L, 30L, 32L, 11L),
  texcl = c("sl", "scl", "scl", "sl"),
  d_value = c(6L, 4L, 4L, 4L),
  m_chroma = c(2L, 2L,  3L, 3L)
)

# promote to SoilProfileCollection
depths(spc) <- id ~ hzdept + hzdepb
hzdesgnname(spc) <- 'hzname'
hztexclname(spc) <- 'texcl'
hzmetaname(spc, "clay") <- 'prop'

depths(spc2) <- id ~ hzdept + hzdepb
hzdesgnname(spc2) <- 'hzname'
hztexclname(spc2) <- 'texcl'
hzmetaname(spc2, "clay") <- 'claytotest'

spc3 <- data.frame(pedon_key = c("10016", "10016", "10016", "10047",  "10047",
                                   "10047", "10047", "10047", "10047", "10047"), 
                   hzn_top = c(0, 18, 51, 0, 26, 45, 55, 90, 102, 130), 
                   hzn_bot = c(18, 51, 107, 26, 45, 55, 90, 102, 130, 185), 
                   hzn_desgn = c("A", "C1",  "C2", "Ap", "E", "Btn", "Bt", 
                                 "Eb", "2Btk1", "2Btk2"), 
                   texture_lab = c("sl", "sil", "sil", "sil", "sil", 
                                   "sil", "sil", "sil", "sil", "sil"), 
                   clay_total = c(4.4, 7.7, 11.9, 16.7, 9.9, 
                                  19.5, 14.2, 7.3, 11.3, 7.6))
depths(spc3) <-pedon_key ~ hzn_top + hzn_bot
hzdesgnname(spc3) <- "hzn_desgn"
hztexclname(spc3) <- "texture_lab"
hzmetaname(spc3, "clay") <- "clay_total"

test_that("mollic.thickness.requirement", {
  expect_equal(mollic.thickness.requirement(spc, clay.attr = 'prop'), 18)
  expect_equal(mollic.thickness.requirement(spc, clay.attr = 'prop', truncate = FALSE), 49 / 3)
  expect_equal(mollic.thickness.requirement(trunc(spc, 0, 9), clay.attr = 'prop'), 10)

  # this is a controversial/undefined case:
  # 
  #  the Bk is identified as a cambic below an argillic 
  #  
  #  so most limiting crit is the bottom depth of cambic: 66/3 = 22 
  #  rather than 36/3 = 12 truncated to 18 which comes from argillic bottom depth/2nd carbonate upperbound
  expect_equal(mollic.thickness.requirement(spc2, clay.attr = 'claytotest'), 22)
  
  expect_equal(mollic.thickness.requirement(spc3), c(18, 25))
})

#' @title Compare Site Level Attributes of a SoilProfileCollection
#' 
#' @description Compare site level attributes of a `SoilProfileCollection` object, returning a distance matrix conformal with the output from [NCSP()]. Values are within the range of 0-1.
#' 
#' @details This function is typically used in conjunction with the output from [NCSP()].
#' 
#' @param x `SoilProfileCollection` object
#' @param vars character vector listing one or more site level attributes of `x`
#' @param weights numeric vector, same length as `vars`, variable weighting
#' @param ... additional arguments to [cluster::daisy()]
#'
#' @return `dissimilarity` / `dist` class object containing pair-wise distances, row/column names derived from `profile_id(x)`
#' 
#' @export
#' 
#' @seealso [NCSP()] [cluster::daisy()]
#' 
compareSites <- function(x, vars, weights = rep(1, times = length(vars)), ...) {
  
  # extract 1 or more site variables
  .s <- site(x)[, vars, drop = FALSE]
  
  # transfer profile IDs -> row names
  row.names(.s) <- profile_id(x)
  
  # wrap cluster::daisy
  .d <- cluster::daisy(.s, metric = 'gower', weights = weights, ...)
  
  # convert to base class
  .d <- as.dist(.d)
  
  # standardize attributes to match NCSP()
  attr(.d, 'Metric') <- 'Gower'
  
  return(.d)  
}

## TODO: 
# * data.table optimization -> no iteration over profiles

#' @title Generate Soil Depth Class Matrix
#'
#' @description Generate a boolean matrix of soil depth classes, actual soil depth class, and estimate of soil depth from a `SoilProfileCollection` object. Soil depths are estimated using pattern matching applied to horizon designations, by [`estimateSoilDepth()`]. The default REGEX pattern (`p = 'Cr|R|Cd'`) will match most "contacts" described using the USDA / Soil Taxonomy horizon designation conventions.
#'
#'
#' @param f a SoilProfileCollection object
#' @param depth.classes a named vector of classes and depth breaks
#' @param \dots arguments passed to \code{\link{estimateSoilDepth}}
#' @return a \code{data.frame} containing soil depth and depth class for each
#' profile, see examples
#' @author D.E. Beaudette and J.M. Skovlin
#' @seealso \code{\link{estimateSoilDepth}}
#' @keywords manip
#' @export
#' @examples
#'
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#'
#' # generate depth-class matrix
#' sdc <- getSoilDepthClass(sp1, name = 'name')
#'
#' # inspect
#' head(sdc)
#'
#' # join back into sp1 as site-level data
#' site(sp1) <- sdc
#'
#' \dontrun{
#' # sample data
#' data(gopheridge, package='soilDB')
#'
#' getSoilDepthClass(gopheridge, name = 'hzname')
#' }
#'
getSoilDepthClass <- function(f, depth.classes=c('very.shallow'=25, 'shallow'=50, 'mod.deep'=100, 'deep'=150, 'very.deep'=10000), ...) {

  # apply soil-depth finding function
  # horizon top/bottom depths automatically extracted
  soil.depth <- profileApply(f, estimateSoilDepth, ...)

  # convert to 1-column matrix
  soil.depth <- matrix(soil.depth)

  # evaluate depth-class rules
  depth.class.matrix <- t(sapply(soil.depth, function(i) i < depth.classes))

  # determine shallowest matching rule by-row
  # safely deal with NA or inf
  best.match <- apply(
    depth.class.matrix, 
    MARGIN = 1, FUN = function(i) {
      if(all(is.na(i)) | all(is.infinite(i))) {
        NA
      } else {
        names(which.min(which(i)))      
      }
    }
  )

  # zero-out the depth class matrix
  depth.class.matrix[] <- FALSE

  # load best matching depth class by row
  for(i in 1:nrow(depth.class.matrix)) {
    # safely account for NA
    if(is.na(best.match[i])) {
      depth.class.matrix[i, ] <- NA
    } else {
      depth.class.matrix[i, best.match[i]] <- TRUE
    }
  }

  
  # add-in ID and actual depth
  d <- data.frame(
    profile_id(f), 
    depth = soil.depth, 
    depth.class.matrix, 
    depth.class = best.match, 
    stringsAsFactors=FALSE
  )

  # fix ID name
  names(d)[1] <- idname(f)

  # set factor levels
  d$depth.class <- factor(d$depth.class, levels = names(depth.classes))

  return(d)
}

#' @title Visualize Spectral Mixing of Munsell Colors
#' 
#' @description Lattice visualization demonstrating subtractive mixtures of colors in Munsell notation and associated spectra.
#' 
#' @details If present, `names` attribute of `x` is used for the figure legend. See the [expanded tutorial](https://ncss-tech.github.io/AQP/aqp/mix-colors.html) for examples.
#' 
#' @author D.E. Beaudette
#' 
#' @param x vector of colors in Munsell notation, should not contain duplicates
#' 
#' @param w vector of weights, can sum to any number
#' 
#' @param mixingMethod approach used to simulate a mixture: 
#'    
#'    * `exact`: simulate a subtractive mixture of pigments, color conversion via CIE1931 color-matching functions (see [mixMunsell()])
#'    
#'    * `reference`  : simulate a subtractive mixture of pigments, selecting `n` closest reference spectra, requires `gower` package
#' 
#' @param n number of closest mixture candidates when `mixingMethod = 'reference'` (see [mixMunsell()]), results can be hard to interpret when `n > 2`
#' 
#' @param swatch.cex scaling factor for color swatch rectangle width and height, relative to `label.cex`, typically between 1 and 3
#' 
#' @param label.cex scaling factor for swatch labels
#' 
#' @param showMixedSpec show weighted geometric mean (mixed) spectra as dotted line (only when `mixingMethod = 'reference'`)
#' 
#' @param overlapFix attempt to "fix" overlapping chip labels via [fixOverlap()], using `method = 'E'`
#' 
#' @return a `lattice` graphics object
#' 
#' @seealso [mixMunsell()]
#' @export
#' 
plotColorMixture <- function(x, w = rep(1, times = length(x)) / length(x), mixingMethod = c('exact', 'reference'), n = 1, swatch.cex = 1.5, label.cex = 0.85, showMixedSpec = FALSE, overlapFix = TRUE) {
  
  # TODO plot will be incorrect if duplicate Munsell chips are specified
  
  # TODO: feedback on spectral distance is required
  
  # TODO: ideas on styling legend (size, placement, etc.)
  
  # mixture method sanity checks
  mixingMethod <- match.arg(mixingMethod)
  
  # 'reference' mixing method requires gower package
  if(mixingMethod == 'reference' & !requireNamespace('gower')) {
    stop('package `gower` is required for `reference` mixingMethod', call. = FALSE)
  }
  
  # can't use n > 1 with mixingMethod = 'exact'
  if(mixingMethod == 'exact') {
    
    if(n > 1 ) {
      stop('cannot request multiple matches with `mixingMethod = "exact"`', call. = FALSE)
    }
    
    # must retain mixed spectra
    showMixedSpec <- TRUE
  }
  
  # mix colors
  mx <- suppressMessages(
    mixMunsell(x = x, w = w, n = n, mixingMethod = mixingMethod, keepMixedSpec = showMixedSpec)
  )
  
  # make local copy of the mixed colors when asking for the mixed spectra too
  if(showMixedSpec) {
    m <- mx$mixed
  } else {
    m <- mx
  }
  
  # sanity check: it could be that reference spectra aren't available for requested colors
  if(all(is.na(m$munsell))) {
    stop('reference spectra not available', call. = FALSE)
  }
  
  # satisfy R CMD check
  munsell.spectra <- NULL
  ID <- NULL
  
  # safely load reference spectra
  load(system.file("data/munsell.spectra.rda", package="aqp")[1])
  
  # vector of colors to mix and result
  colors <- c(x, m$munsell)
  
  ## TODO: this will fail if a color is not in the spectral library
  #        -> possibly when mixingMethod = 'exact'
  #        solution: provide a template data.frame
  
  # select spectra from reference library and assign an ID
  # IDs should use names(x) if !NULL
  # otherwise generate an ID
  nm <- names(x)
  if(is.null(nm)) {
    IDs <- sprintf('color %s', 1:length(x))
  } else {
    IDs <- nm
  }
  
  # iteration over colors to-mix + mixture(s)
  s <- lapply(seq_along(colors), function(i) {
    # select current color + spectra
    if(mixingMethod == 'reference') {
      # all colors selected from library
      z <- munsell.spectra[which(munsell.spectra$munsell == colors[i]), ]
      
    } else {
      # exact mixing, last color is mixed spectrum
      z <- munsell.spectra[which(munsell.spectra$munsell == colors[i]), ]
      
      # last color is the mixture, 
      # replace reference spectra / munsell chip with actual mixture
      if(i == length(colors)) {
        
        # avoid trying to append to a 0-length DF 
        # if the last color has no reference spectra
        if(nrow(z) < 1) {
          # pad with NA
          z <- z[1:length(mx$spec), ]
          
          # fill columns with pieces of the closest Munsell
          .pm <- parseMunsell(mx$mixed$munsell, convertColors = FALSE)
          z$hue <- .pm$hue
          z$value <- .pm$value
          z$chroma <- .pm$chroma
          
          ## NOTE: hard-coded spectra wavelengths!!
          # fill wavelength column
          z$wavelength <- seq(from = 380, to = 730, by = 10)
        }
        
        # save mixed spectra and closest Munsell color
        z$reflectance <- mx$spec
        z$munsell <- mx$mixed$munsell
      }
      
    }
    
    
    # assign an ID for plotting
    if( i <= length(x)) {
      # current ID
      z$ID <- IDs[i]
    } else {
      # reset counter to mix color ranks
      z$ID <- sprintf('mix #%s', i - length(x))
    }
    
    return(z)
  })
  
  s <- do.call('rbind', s)
  row.names(s) <- NULL
  
  # create sensible levels for plotting / legend
  # names + mixtures
  all.names <- unique(s$ID)
  
  # mixtures, sorted
  mix.names <- sort(setdiff(all.names, nm))
  
  # original chip names + mixtures
  color.chip.levels <- c(nm, mix.names)
  
  # encode chips names + mixture names as into factor
  s$ID <- factor(s$ID, levels = color.chip.levels)
  
  # convert into colors for plotting
  s$color <- parseMunsell(s$munsell)
  
  # plotting style, colors sorted by mixing logic
  cols <- parseMunsell(colors)
  
  # line style
  #  1: colors-to-mix
  #  4: mixture results (n)
  col.lty <- c(
    rep(1, times = length(x)), 
    rep(4, times = nrow(m))
  )
  
  # compile line styles
  tps <- list(
    superpose.line = list(
      col = cols, 
      lwd = 5, 
      lty = col.lty
    ))
  
  # labels for figure
  #  all colors
  munsell.labels <- colors
  # weights
  wt.labels <- sprintf('%s%%', round((w / sum(w)) * 100))
  # ranked matches
  match.rank <- sprintf("#%s", seq(from = 1, to = nrow(m)))
  # combined labels
  lab.text <- sprintf('%s\n%s', munsell.labels, c(wt.labels, match.rank))
  
  ## TODO: use grid::grid.layout() / grid viewports to manage a multi-panel figure
  
  # final figure
  pp <- xyplot(
    reflectance ~ wavelength, groups = ID, data = s, 
    type = c('l', 'g'),
    ylab = 'Reflectance',
    xlab = 'Wavelength (nm)',
    scales = list(tick.number = 12),
    auto.key = list(lines = TRUE, points = FALSE, cex = 1, space='top', columns = length(colors)),
    par.settings = tps,
    xlim = c(370, 780),
    subscripts = TRUE,
    panel = function(x, y, groups, ...) {
      # setup plot
      lattice::panel.xyplot(x = x, y = y, groups = groups, ...)
      
      # split spectra by groups
      d <- split(y, groups)
      
      # get last y-coordinate by group
      last.y.coords <- sapply(d, function(i) {
        i[length(i)]
      })
      
      # get width / height of labels
      # dummy label is not printed
      .dummyLabel <- grid::grid.text(
        label = '10GY 4/4\n100%',
        x = grid::unit(750, units = 'native'), 
        y = grid::unit(max(y, na.rm = TRUE), units = 'native'),
        gp = grid::gpar(
          cex = label.cex,
          font = 2
        ), draw = FALSE
      )
      
      .labelHeight <- grid::convertHeight(
        grid::unit(1, 'grobheight', .dummyLabel),
        unitTo = 'native'
      )
      
      .labelWidth <- grid::convertWidth(
        grid::unit(1, 'grobwidth', .dummyLabel),, 
        unitTo = 'native'
      )
      
      ## debugging grob geom
      # print(cbind(.labelWidth, .labelHeight))
      
      
      # attempt to fix overlap
      if(overlapFix) {
        
        # save original indexing information via name
        .original_order <- names(last.y.coords)
        
        # sort according to y-coordinates
        last.y.coords <- sort(last.y.coords)
        
        # new ordering
        .new_order <- names(last.y.coords)
        
        # vertical threshold is the grob height of color swatch
        ov.thresh <- as.vector(.labelHeight) * swatch.cex
        
        # attempt to find overlap using the above threshold
        ov <- overlapMetrics(last.y.coords, thresh = ov.thresh)
        
        # if present, iteratively find suitable alternatives
        # search is bounded to the min/max of all spectra
        # consider set.seed() for consistent output
        if(length(ov$idx) > 0) {
          message('fixing overlap')
          
          # init position vector 
          # set boundary conditions for overlap adjustment
          # adding fake min / max values, based on spectra reflectance values
          .pos <- c(min(y, na.rm = TRUE), last.y.coords, max(y, na.rm = TRUE))
          
          last.y.coords <- fixOverlap(
            .pos, method = 'E',
            thresh = ov.thresh * 0.95,
            q = 1
          )
          
          # ignore first and last positions
          # those are the boundary conditions
          last.y.coords <- last.y.coords[2:(length(.pos) - 1)]
        }
        
        # revert to original order
        last.y.coords <- last.y.coords[match(.original_order, .new_order)]
      }
      
      
      
      # iterate over groups and put munsell chip next to last data point
      for(i in seq_along(d)) {
        # last spectral value
        yy <- d[[i]]
        
        # last y-value in this group
        last.y <- last.y.coords[i]
        
        # current color
        this.col <- tps$superpose.line$col[i]
        # current label
        
        # rectangle color swatch surrounds label
        # label widths and heights are used to size t he rectangle (native units)
        grid::grid.rect(
          x = grid::unit(750, units = 'native'), 
          y = grid::unit(last.y, units = 'native'), 
          width = .labelWidth * swatch.cex,
          height = .labelHeight * swatch.cex,
          default.units = 'native',
          gp = grid::gpar(
            col = this.col,
            fill = this.col
          )
        )
        
        # place label
        grid::grid.text(
          label = lab.text[i],
          x = grid::unit(750, units = 'native'), 
          y = grid::unit(last.y, units = 'native'),
          gp = grid::gpar(
            cex = label.cex,
            col = invertLabelColor(this.col),
            font = 2
          )
          
        )
      }
      
      # the mixed spectra is only shown as a dotted line when mixingMethod = 'reference'
      if(showMixedSpec & mixingMethod != 'exact'){
        panel.lines(x = unique(s$wavelength), y = mx$spec, lty = 3, col = 'black')
      }
      
      
    }
  )
  
  return(pp)
  
}

## TODO: encode some kind of ID into the dissimilarity matrix for later use
#' Evaluate Generalized Horizon Labels
#'
#' Data-driven evaluation of generalized horizon labels using nMDS and
#' silhouette width.
#'
#' Classic multidimensional scaling is performed via [stats::cmdscale()].
#' The input distance matrix is generated by [cluster::daisy()] using
#' (complete cases of) horizon-level attributes from `obj` as named in
#' `vars`.
#'
#' Silhouette widths are computed via [cluster::silhouette()]. The input
#' distance matrix is generated by [cluster::daisy()] using (complete cases
#' of) horizon-level attributes from `obj` as named in `vars`. Note
#' that observations with genhz labels specified in `non.matching.code`
#' are removed filtered before calculation of the distance matrix.
#'
#' @param obj a `SoilProfileCollection` object
#' 
#' @param genhz name of horizon-level attribute containing generalized horizon labels
#' 
#' @param vars character vector of horizon-level attributes to include in the evaluation
#' 
#' @param non.matching.code code used to represent horizons not assigned a generalized horizon label
#' 
#' @param stand standardize variables before computing distance matrix, passed to [cluster::daisy()]
#' 
#' @param metric distance metric, passed to [cluster::daisy()]
#' 
#' @return a list is returned containing: 
#'   * horizons: `c('mds.1', mds.2', 'sil.width', 'neighbor')` 
#'   * stats: mean and standard deviation `vars`, computed by generalized horizon label
#'   * dist: the distance matrix as passed to [stats::cmdscale()]
#'   
#' @author D.E. Beaudette
#' 
#' @seealso [get.ml.hz()]
#' 
#' @keywords manip
#' 
#' @export
evalGenHZ <- function(obj, genhz = GHL(obj, required = TRUE), vars, non.matching.code = 'not-used', stand = TRUE, metric = 'euclidean') {

  # hack to make R CMD check happy
  value <- summarize <- NULL

  # extract site / horizons as DF
  h <- as(obj, 'data.frame')

  # genhz may have its own levels set, but if not a factor, make one
  if (!is.factor(h[[genhz]]))
    h[[genhz]] <- factor(h[[genhz]])

  # make an index to complete data
  no.na.idx <- which(complete.cases(h[, vars, drop = FALSE]))

  ## TODO: all of vars should be numeric or convertable to numeric
  # numeric.test <- sapply(vars, function(i) is.numeric(h[[i]]))

  # test for duplicate data
  # unique IDs are based on a concatenation of variables used... digest would be safer
  h.to.test <- h[no.na.idx, c(idname(obj), vars)]
  h.to.test$id <- apply(h.to.test[, vars, drop = FALSE], 1, function(i) paste0(i, collapse = '|'))
  dupe.names <- names(which(table(h.to.test$id) > 1))
  dupe.rows <- h.to.test[which(h.to.test$id %in% dupe.names), ]
  dupe.ids <- paste0(unique(dupe.rows[[idname(obj)]]), collapse = ', ')
  if (dupe.ids != "")
    warning(paste0('duplicate data associated with pedons: ', dupe.ids), call. = FALSE)

  # compute pair-wise dissimilarities using our variables of interest
  d <- cluster::daisy(h[no.na.idx, vars, drop = FALSE], stand = stand, metric = metric)

  # fudge-factor in case of duplicate data (0s in the dissimilarity matrix)
  dupe.idx <- which(d < 1e-8)
  if (length(dupe.idx) > 0) {
    fudge <- min(d[which(d > 0)]) / 100
    d[dupe.idx] <- fudge
  }

  ## TODO: allow user-defined ordination function
  # perform metric MDS of dissimilarity matrix
  mds <- stats::cmdscale(d, k = 2)

  # compute silhouette widths after removing not-used genhz class
  sil.idx <-  which(complete.cases(h[, vars, drop = FALSE]) & h[[genhz]] != non.matching.code)
  d.sil <- cluster::daisy(h[sil.idx, vars, drop = FALSE], stand=stand)
  sil <- cluster::silhouette(as.numeric(h[[genhz]])[sil.idx], d.sil)

  # add new columns
  h$mds.1 <- NA
  h$mds.2 <- NA
  h$sil.width <- NA
  h$neighbor <- NA

  ## TODO: access results from user-defined ordination function
  # copy values
  h$mds.1[no.na.idx] <- mds[, 1]
  h$mds.2[no.na.idx] <- mds[, 2]
  h$sil.width[sil.idx] <- sil[, 3]
  h$neighbor[sil.idx] <- levels(h[[genhz]])[sil[, 2]]

  ## TODO: enforce / check above
  ## important note: all 'vars' should be numeric


  # convert wide -> long form
  # using data.table::melt
  # suppressing warnings related to mixture of int / numeric
  m <- suppressWarnings(
    data.table::melt(
      data.table::as.data.table(h),
      id.vars = genhz,
      measure.vars = c(vars, 'sil.width')
    )
  )

  # leave as data.table for aggregation
  # compute group-wise summaries
  m.summary <- m[, list(mean = mean(value, na.rm = TRUE), sd = sd(value, na.rm = TRUE)), by = c((genhz), 'variable')]

  # format text
  m.summary$stats <- sprintf(
    "%s (%s)",
    round(m.summary$mean, 2),
    round(m.summary$sd, 2)
  )

  # using data.table::dcast
  fm <- paste0(genhz, ' ~ variable')
  genhz.stats <- dcast(data = m.summary, formula = fm, value.var = 'stats')

  # convert back to data.frame
  genhz.stats <- as.data.frame(genhz.stats)

  # composite into a list
  res <- list(horizons = h[, c('mds.1', 'mds.2', 'sil.width', 'neighbor')],
              stats = genhz.stats, dist = d)
  return(res)
}

## TODO: https://github.com/ncss-tech/aqp/issues/68
# * plot in layout similar to Munsell Color Book, alpha ~ frequency
# * distance matrix calculation takes time, careful with n > 1,000
# * gower's distance of {L,A,B} looks pretty good too
# * more control over graphical elements and par()
# * grid size estimation needs some work
# * return dE00 stats for method = MDS or grid
 



#' @title Preview Colors
#' 
#' @description Preview colors arranged according to CIE2000 distances or manual specification.
#'
#' @param cols vector of R colors
#' @param method either "grid", "MDS", or "manual", see details
#' @param labels optional vector of labels, disabled when `length(cols) > 5000`
#' @param labels.cex scaling factor for labels
#' @param col.order integer vector used to order colors
#' @param nrow number of rows used by "grid" method
#' @param ncol number of columns used by "grid" method
#' @param border.col border color used by "grid" method
#' @param pt.cex point scaling factor used by "MDS" method
#' @param pt.pch point symbol used by "MDS" method
#' 
#' @details Color sorting is based on CIE2000 distances as calculated by `farver::compare_colour()`. The "grid" method arranges colors in a rectangular grid with ordering based on divisive hierarchical clustering of the pair-wise distances. Unique colors are used when `cols` contains more than 5,000 colors.
#' 
#' The "MDS" method arranges unique colors via classical multidimensional scaling (principal coordinates) via `cmdscale()`.
#' 
#' Colors can be manually arranged by supplying a vector of integers to `col.order` and setting `method='manual'`.
#' 
#' @author D.E. Beaudette
#' 
#'
#' @return When `method` = "grid" or "manual", a vector of color order is returned. When `method = "MDS"`, the output from [stats::cmdscale()].
#' @export
#'
#' @examples
#' 
#' # example data
#' data(sp2)
#' 
#' # convert into SoilProfileCollection object
#' depths(sp2) <- id ~ top + bottom
#' 
#' previewColors(sp2$soil_color)
#' previewColors(sp2$soil_color, method = 'MDS', pt.cex = 3)
#' 
#' # create colors using HCL space
#' cols.hcl <- hcl(h = 0:360, c = 100, l = 50)
#' 
#' # grid, colors sorted by dE00
#' previewColors(cols.hcl)
#' 
#' # manual specification
#' previewColors(cols.hcl, method = 'manual', col.order = 1:361)
#' 
#' # MDS
#' previewColors(cols.hcl, method = 'MDS', pt.cex = 1)
#' 
previewColors <- function(cols, method = c('grid', 'MDS', 'manual'), labels = NULL, labels.cex = 1, col.order = NULL, nrow = ceiling(sqrt(length(cols))), ncol = nrow, border.col = 'black', pt.cex = 2, pt.pch = 15) {

  # remove NA
  cols <- na.omit(cols)
  
  # expected method argument
  method <- match.arg(method)

  # safety catch: a 5000 x 5000 element distance matrix is about all we really need for a preview
  if(length(cols) > 5000) {
    cols <- unique(cols)
    pt.cex <- pt.cex[1]
    pt.pch <- pt.pch[1]
    labels <- NULL
    warning('using unique colors, suppressing labels, and first `pt.cex` / `pt.pch')
  }

  # use unique colors when using MDS
  if(method == 'MDS') {
    cols <- unique(cols)
  }

  # manual ordering, defined by col.order
  if(method == 'manual') {
    # sanity check
    if(is.null(col.order)) {
      stop('must specify color ordering vector', call. = FALSE)
    }

    # re-order colors and convert into a matrix
    m <- matrix(NA, nrow=nrow, ncol=ncol)
    m[1:length(cols)] <- cols[col.order]
    
    par(mar=c(1,0,3,0))
    plot(1, 1, type='n', axes=FALSE, xlab='', ylab='', xlim=c(0.5, ncol+0.5), ylim=c(nrow+0.5, 0.5))
    rect(xleft = col(m) - 0.5, ybottom = row(m) -0.5, xright = col(m) + 0.5, ytop = row(m) + 0.5, col = m, border = border.col, lwd=0.5)
    
    # add labels
    if(!is.null(labels)) {
      m.l <- m
      m.l[1:length(cols)] <- labels[col.order]
      text(x = col(m.l), y = row(m.l), labels = m.l, col = invertLabelColor(m), cex = labels.cex)
    }
    
    invisible(col.order)
  }

  # hex representation -> sRGB
  cols.srgb <- t(col2rgb(cols)) / 255
  # sRGB -> CIE LAB
  cols.lab <- grDevices::convertColor(cols.srgb, from = 'sRGB', to = 'Lab', from.ref.white='D65', to.ref.white='D65', clip=FALSE)

  # distances are based on CIE2000 color comparison
  # note: single argument -> all pair-wise distances
  # output is transposed relative to `dist` object
  d <- farver::compare_colour(cols.lab, from_space = 'lab', to_space = 'lab', method = 'CIE2000')
  d <- as.dist(t(d))

  ## TODO: return / report group-wise summary (select quantiles) of dE00
  
  if(method == 'grid') {
    # divisive hierarchical clustering for order
    col.order <- cluster::diana(d)$order

    # re-order colors and convert into a matrix
    m <- matrix(NA, nrow = nrow, ncol = ncol)
    m[1:length(cols)] <- cols[col.order]

    par(mar=c(1,0,3,0))
    plot(1, 1, type='n', axes=FALSE, xlab='', ylab='', xlim=c(0.5, ncol+0.5), ylim=c(nrow+0.5, 0.5))
    rect(xleft = col(m) - 0.5, ybottom = row(m) -0.5, xright = col(m) + 0.5, ytop = row(m) + 0.5, col = m, border = border.col, lwd=0.5)

    # add labels
    if(!is.null(labels)) {
      m.l <- m
      m.l[1:length(cols)] <- labels[col.order]
      text(x = col(m.l), y = row(m.l), labels = m.l, col = invertLabelColor(m), cex = labels.cex)
    }
    
    invisible(col.order)
  }


  # MDS display of unique colors
  if(method == 'MDS') {

    # consider other methods
    # this is fairly robust to 0 distances
    mds <- cmdscale(d)

    # simple plot, no indication of density
    par(mar=c(1,1,3,1))
    plot(mds, type='n', axes=FALSE)
    grid(nx=10, ny=10, col=par('fg'))
    points(mds, col = cols, cex = pt.cex, pch = pt.pch)
    box()
    
    # add labels
    if(!is.null(labels)) {
      text(x = mds[, 1], y =  mds[, 2], labels = labels, col = par('fg'), cex = labels.cex, pos = 1)
    }
    
    invisible(mds)
  }

}

# guessColumnNames.R

#' @title Guess Horizon Slot Column Names
#'
#' @description `guessHzAttrName()`: Guess the horizon column name where possible/preferred formative elements are known. There is a preference for records where more optional requirements are met to handle cases where there will be many matches. For example, working with soil data one might have "low, RV and high" total clay, as well as clay fractions. One could distinguish between these different measurements using standard formative elements for column names from the database of interest. Result is the first match in \code{horizonNames(x)} with the most required plus optional patterns matched.
#'
#' e.g. \code{guessHzAttrName(x, attr="clay", optional=c("total", "_r"))} matches (\code{claytotal_r == totalclay_r}) over (\code{clay_r == claytotal == totalclay}) over \code{clay}.
#'
#' @param x A _SoilProfileCollection_
#' @param attr _character_. A regular expression containing required formative element of attribute name.
#' @param optional _character_. Vector of regular expression(s) containing optional formative elements of attribute name.
#' @param verbose _logical_. Produce message output about guesses? Default: `TRUE`
#'
#' @return Character containing horizon attribute column name. Result is the first match in \code{horizonNames(x)} with the most required plus optional patterns matched.
#'
#' @author Andrew G. Brown
#' @rdname guessHzAttrName
#' @export
#'
#' @examples
#'
#' # a has the required attr pattern, but none of the optional
#' a <- data.frame(id = 1, top = c(0,10), bottom=c(10,40),
#'                 clay=c(18,19))
#' depths(a) <- id ~ top + bottom
#'
#' guessHzAttrName(a, attr="clay", optional=c("total", "_r"))
#'
#' # b has requried attr pattern, and one of the opional patterns
#' #   notice that it also contains "clay" but preferentially matches more optional patterns
#' b <- data.frame(id = 1, top = c(0,10), bottom=c(10,40),
#'                 clay=c(0.18,0.19), clay_r=c(18,19))
#' depths(b) <- id ~ top + bottom
#'
#' guessHzAttrName(b, attr="clay", optional=c("total", "_r"))
#'
#' # c has total and _r (both optional) on either side of clay
#' # having all of the optional patterns plus required is best evidence, and first
#' # column containing that combination will be returned
#' c <- data.frame(id = 1, top = c(0,10), bottom=c(10,40),
#'                 totalclay_r=c(18,19), claytotal_r=c(0.18,0.19))
#' depths(c) <- id ~ top + bottom
#'
#' guessHzAttrName(c, attr="clay", optional=c("total", "_r"))
#'
guessHzAttrName <- function(x, attr, optional = NULL, verbose = TRUE, required = FALSE) {
  nm <- horizonNames(x)
  
  if (!inherits(x, 'SoilProfileCollection')) {
    stop("x must be a SoilProfileCollection")
  }
  
  # possible names include column names with name in the name
  req <- grepl(attr, nm, ignore.case = TRUE)
  
  opt <- lapply(as.list(optional), function(i) grepl(i, nm, ignore.case = TRUE))
  if (is.null(optional) | length(optional) == 0)
    opt <- as.list(req)
  opt <- rowSums(do.call('cbind', opt))
  
  # all optional requirements met
  idx1 <- which(req & opt == length(optional))
  
  # one or more optional requirements met
  idx2 <- which(req & opt > 0)
  
  # basic requirement met
  idx3 <- which(req)
  
  # return first index matching in decreasing precedence
  #  all optional met, some optional met, basic requirement met, no requirement met
  res <- NA
  if (length(idx1)) {
    res <- nm[idx1[1]]
  } else if (length(idx2)) {
    res <- nm[idx2[1]]
  } else if (length(idx3)) {
    res <- nm[idx3[1]]
  }
  if (!is.na(res)) {
    if (verbose)
      message(sprintf('guessing horizon attribute \'%s\' is stored in `%s`', attr, res))
  } else {
    msg <- sprintf('unable to guess column containing horizon attribute \'%s\'', attr)
    if (required)
      stop(msg, .call = FALSE)
    if (verbose)
      message(msg)
  }
  return(res)
}

#' @description `guessHzDesgnName()`: **DEPRECATED** This follows the historic convention used by \code{aqp::plotSPC()} looking for "hzname" or other column names containing the regular expression "name". If the pattern "name" is not found, the pattern "desgn" is searched as a fallback, as "hzdesgn" or "hz_desgn" are other common column naming schemes for horizon designation name.
#'
#' @param x A SoilProfileCollection
#' @param required logical Default: `FALSE`. Is this attribute required? If it is, set to `TRUE` to trigger error on invalid value.
#' 
#' @rdname guessHzAttrName
#'
#' @export
guessHzDesgnName <- function(x, required = FALSE) {
  
  .Deprecated(msg = "`guessHzDesgnName()` is deprecated. Use `hzdesgnname()` (with `required=TRUE` argument if needed), or use `guessHzAttrName()` with appropriate values for your use case")
  
  nm <- horizonNames(x)
  name <- NA

  if (!inherits(x, 'SoilProfileCollection')) {
    stop("x must be a SoilProfileCollection")
  }

  hzd <- hzdesgnname(x)
  if (length(hzd) == 1) {
    if (hzd != "")
      return(hzd)
  }

  # possible names include column names with name in the name
  possible.name <- nm[grep('name', nm, ignore.case = TRUE)]

  # use the first valid guess
  if (length(possible.name) > 0) {
    possible.name <- possible.name[1]
    name <- possible.name
  } else {
    # hail mary
    try.again <- guessHzAttrName(x, "desgn", c("hz"), verbose = FALSE)
    if (!is.na(try.again)) {
      name <- try.again
    } else {
      message('unable to guess column containing horizon designations')
    }
  }
  
  # this implements the required argument 
  # (even though this method does not return an SPC)
  hzdesgnname(x, required = required) <- name
  
  return(name)
}

#' @description `guessHzTexClName()`: **DEPRECATED** This function is used to provide a texture class attribute column name to functions. It will use regular expressions to match "texcl" which is typically the texture of the fine earth fraction, without modifiers or in-lieu textures. Alternately, it will match "texture" for cases where "texcl" is absent (e.g. in NASIS Component Horizon).
#'
#' @rdname guessHzAttrName
#' 
#' @export guessHzTexClName
guessHzTexClName <- function(x, required = FALSE) {
  
  .Deprecated(msg = "`guessHzTexClName()` is deprecated. Use `hztexclname()` (with `required=TRUE` argument if needed), or use `guessHzAttrName()` with appropriate values for your use case")
  
  nm <- horizonNames(x)

  if (!inherits(x, 'SoilProfileCollection')) {
    stop("x must be a SoilProfileCollection")
  }

  if (length(hztexclname(x)) == 1) {
    # ideally use metadata if it contains a value
    name <- hztexclname(x)
    if (name != "")
      return(name)
  }

  # possible names include column names with name in the name
  possible.name1 <- nm[grep('texcl', nm, ignore.case = TRUE)]

  # use the first valid guess matching texcl
  if (length(possible.name1) > 0) {
    possible.name1 <- possible.name1[1]
    
    # this implements the required argument 
    # (even though this method does not return an SPC)
    hztexclname(x, required = required) <- possible.name1
    return(possible.name1)
  }
  
  
  # alternately, try for something called "texture"
  possible.name2 <- nm[grep('texture', nm, ignore.case = TRUE)]
  if (length(possible.name2) > 0) {
    possible.name2 <- possible.name2[1]
    hztexclname(x, required = required) <- possible.name2
    return(possible.name2)
  } else {
    message('unable to guess column containing horizon texture classes')
  }
  return("")
}

context("SoilProfileCollection attribute guessing functions")

## sample data
data(sp3, package = 'aqp')
depths(sp3) <- id ~ top + bottom

## tests
test_that("basic functionality", {

  # historic horizon designation name (e.g. used by plotSPC)
  expect_warning(expect_equal(guessHzDesgnName(sp3), "name"))

  # basic attribute name guessing
  expect_message(expect_equal(guessHzAttrName(sp3, "clay", ""), "clay"),
                "guessing horizon attribute 'clay' is stored in `clay`")

  # more complex attribute name guessing
  sp3$clay_r <- sp3$clay
  sp3$claytotal_r <- sp3$clay

  expect_message(expect_equal(guessHzAttrName(sp3, "clay", ""), "clay"),
                 "guessing horizon attribute 'clay' is stored in `clay`")

  expect_message(expect_equal(guessHzAttrName(sp3, "clay", c("_r", "total")), "claytotal_r"),
                 "guessing horizon attribute 'clay' is stored in `claytotal_r`")

  # basic attribute name guessing
  expect_warning(expect_equal(guessHzTexClName(sp3), ""))

  # texcl
  horizons(sp3)$texcl <- "l"
  expect_warning(expect_equal(guessHzTexClName(sp3), "texcl"))

  # texture
  sp3$texture <-  sp3$texcl
  sp3$texcl <- NULL
  expect_warning(expect_equal(guessHzTexClName(sp3), "texture"))

  # descriptive name
  sp3$hzdesgn <- sp3$name
  sp3$name <- NULL
  sp3$desgn <- 1:nrow(sp3)
  expect_warning(expect_equal(guessHzDesgnName(sp3), "hzdesgn"))

  # unable to guess name
  sp3$foo <- sp3$hzdesgn
  sp3$hzdesgn <- NULL
  sp3$desgn <- NULL
  expect_warning(expect_equal(guessHzDesgnName(sp3), NA))

  # custom name
  hzdesgnname(sp3) <- "foo"
  expect_warning(expect_equal(guessHzDesgnName(sp3), "foo"))
})

## TODO: check for rowSums != 1
##       this will cause Shannon H to fail

## TODO: conversion from original <-> safe names is clunky


#' @title Determine ML Horizon Boundaries
#'
#' @description This function accepts input from [slab()] (a `data.frame`) along with a vector of
#' horizon names, and returns a `data.frame` of the most likely horizon
#' boundaries.
#'
#' This function expects that `x` is a `data.frame` generated by
#' [slab()]. If `x` was not generated by [slab()], then `o.names` is required.
#'
#' @param x `data.frame`, output from [slab()]
#' 
#' @param o.names an optional character vector of horizon designations that will be used in the final table
#' 
#' @return A `data.frame` with the following columns: 
#'   * `hz`: horizon names
#'   * `top`: horizon top depth
#'   * `bottom`: horizon bottom depth
#'   * `confidence`: integrated probability over thickness of each ML horizon, rounded to the nearest integer
#'   * `pseudo.brier`: A "pseudo"" Brier Score for a multi-class prediction, where the most-likely horizon label is treated as the "correct" outcome. Details on the calculation for traditional Brier Scores here: \url{https://en.wikipedia.org/wiki/Brier_score}. Lower values suggest better agreement between ML horizon label and class-wise probabilities.
#'   * `mean.H`: mean Shannon entropy (bits), derived from probabilities within each most-likely horizon. Larger values suggest more confusion within each ML.
#' 
#' @author D.E. Beaudette
#' @seealso [slab()]
#' @keywords manip
#' 
#' @references 
#' Beaudette, D. E., Roudier, P., & Skovlin, J. (2016). Probabilistic representation of genetic soil horizons. Digital soil morphometrics, 281-293.
#' 
#' @export
#' @examples
#'
#' # init SPC
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#' 
#' # set horizon designation metadata
#' hzdesgnname(sp1) <- 'name'
#' 
#' # generalize horizon designations from character vector
#' # result is an ordered factor
#' sp1$genhz <- generalizeHz(
#'   sp1$name,
#'   new = c('O','A','B','C'),
#'   pat = c('O', '^A','^B','C'),
#'   ordered = TRUE
#' )
#' 
#' # compute slice-wise GHL probability
#' # so that it sums to contributing fraction
#' # from 0-150cm
#' a <- slab(sp1, fm = ~ genhz, cpm = 1, slab.structure = 0:150)
#' 
#' # note original GHL names are set by slab()
#' attr(a, 'original.levels')
#' 
#' # generate table of ML horizonation
#' get.ml.hz(a)
#'
get.ml.hz <- function(x, o.names = attr(x, which = 'original.levels')) {

  # trick R CMD check
  .SD = H = top = bottom = NULL
  
  ## TODO: cleanup use of this column name
  # internally used, temporary column for storing ML hz name
  name <- '.name'
  
  # extract horizon levels, set by slab()
  o.levels <- attr(x, which = 'original.levels')
  
  # temporary conversion to data.table
  x <- data.table::data.table(x)
  
  # sanity check
  if(missing(o.names) && is.null(o.levels)) {
    stop('x not derived from slab() or o.names is missing')
  } else if (!is.null(o.names)) {
    o.levels <- o.names
  }
  
  ## this should accommodate DF-safe names returned by slab()
  # make DF-safe names for hz that violate DF constraints
  safe.names <- make.names(o.levels)

  # LUT for names
  names.LUT <- data.frame(
    original = o.levels,
    safe = safe.names,
    stringsAsFactors = FALSE
    )

	# get index to max probability,
	# but only when there is at least one value > 0 and all are not NA
	.f.ML.hz <- function(i) {
		if(any(i > 0) & !all(is.na(i)))
			which.max(i)
		else
			NA
	}


	# get most probable, original, horizon designation by slice
	x[[name]] <- safe.names[apply(x[, .SD, .SDcols = safe.names], 1, .f.ML.hz)]

	# extract ML hz sequences
	x.rle <- rle(as.vector(na.omit(x[[name]])))
	x.hz.bounds <- cumsum(x.rle$lengths)

	# composite into a data.frame
	# note: we always start from 0
	x.ml <- data.table::data.table(
	  hz = x.rle$value,
	  top = c(0, x.hz.bounds[-length(x.hz.bounds)]),
	  bottom = x.hz.bounds,
	  stringsAsFactors = FALSE
	)

	# in cases where probability depth-functions cross more than once,
	# it is necessary to account for overlaps
	x.ml <- x.ml[, list(top = min(top), bottom = max(bottom)), by = "hz"]

	# re-order using vector of original horizon names-- this will result in NAs if a named horizon was not the most likely
	x.ml <- x.ml[match(safe.names, x.ml$hz), ]
	x.ml <- na.omit(x.ml)

	# integrate probability density function over ML bounds
	x.ml$confidence <- NA
	for(i in seq_along(x.ml$hz)) {
		slice.seq <- seq(from = x.ml$top[i], to = x.ml$bottom[i])
		x.i <- x[slice.seq, .SD, .SDcols = x.ml$hz[i]][[1]]
		hz.int.prob.pct <- round( (sum(x.i) / length(slice.seq)) * 100)
		x.ml$confidence[i] <- hz.int.prob.pct
	}

	# compute a pseudo-brier score using ML hz as the "true" outcome
  # Brier's multi-class score : http://en.wikipedia.org/wiki/Brier_score#Original_definition_by_Brier
	# filter NA: why would this happen?
	idx <- which(!is.na(x[[name]]))
	
	.internalName <- name
	x.bs <- x[idx, ][, list(brierScore(.SD, classLabels = safe.names, actual = .internalName)), 
	                by = c(name), .SDcols = c(name, safe.names)]
	
	# Shannon entropy, (log base 2) bits
  x$H <- apply(x[, .SD, .SDcols = safe.names], 1, shannonEntropy, b = 2)
  x.H <- x[idx, ][, list(H = mean(H)), by = c(name)]

  # fix names for joining
  names(x.bs) <- c('hz', 'pseudo.brier')
  names(x.H) <- c('hz', 'mean.H')

  # join brier scores to ML hz table
  x.ml <- merge(x.ml, x.bs, by = 'hz', sort = FALSE)

  # join shannon H to ML hz table
  x.ml <- merge(x.ml, x.H, by = 'hz', sort = FALSE)

  # convert safe names -> original names
  x.ml$hz <- names.LUT$original[match(x.ml$hz, names.LUT$safe)]
  
  # back to data.frame
  x.ml <- as.data.frame(x.ml)

	return(x.ml)
}

#' Calculate Thickness of Horizons Matching Logical Criteria
#'
#' This function calculates the cumulative (`method="cumulative"`, default) or maximum difference between (`method="minmax"`) horizons within a profile that match a defined pattern (`pattern`) or, more generally, any set of horizon-level logical expressions encoded in a function (`FUN`).
#' 
#' @param x A _SoilProfileCollection_
#' @param pattern _character_. A pattern to match in `hzdesgn`; used with the default `FUN` definition for regular expression pattern matching on horizons.
#' @param hzdesgn _character_. A column containing horizon designations or other horizon-level character label used to identify matches; used with the default `FUN` definition.
#' @param method _character_. Either `"cumulative"` (default) or `"minmax"`. See details.
#' @param prefix _character_. Column prefix for calculated `thickvar` (and `depthvar` for `method="minmax"`) column results. Default: `""`.
#' @param thickvar _character_ Length `1`. Column name to use for calculated thickness column. Default: `"thickness"`
#' @param depthvars _character_. Length `2`. Column names to use for calculated minimum top depth and maximum bottom depth in `method="minmax"`. Default: `horizonDepths(x)`
#' @param FUN _function_. A function that returns a _logical_ vector equal in length to the number of horizons in `x`. See details.
#' @param na.rm _logical_. Omit `NA` values in summaries of thickness and in matching? Default: `FALSE`
#' @param ... Additional arguments passed to the matching function `FUN`.
#'
#' @return A _data.frame_-like object (corresponding to the `aqp_df_class()` of `x`) with one row per profile in `x`. First column is always the profile ID which is followed by `"thickness"`. In `method="minmax"` the upper and lower boundaries used to calculate `"thickness"` are also returned as `"tmin"` and `"tmax"` columns, respectively.
#' 
#' @details
#' The two thickness methods currently available are:
#'  - `method="cumulative"` (default): cumulative thickness of horizons where `FUN` returns true
#'  - `method="minmax"`: maximum bottom depth minus minimum top depth of horizons where `FUN` returns true
#' 
#' If a custom function (`FUN`) is used, it should accept arbitrary additional arguments via an ellipsis (`...`).
#' It is not necessary to do anything with arguments, but the result should match the number of horizons found in the input SoilProfileCollection `x`.
#' 
#' @export
#'
#' @examples
#' data("jacobs2000")
#' 
#' # cumulative thickness of horizon designations matching "Bt"
#' thicknessOf(jacobs2000, "Bt")
#' 
#' # maximum bottom depth minus minimum top depth of horizon designations matching "Bt"
#' thicknessOf(jacobs2000, "Bt", prefix = "Bt_", method = "minmax")
#' 
#' # cumulative thickness of horizon designations matching "A|B"
#' thicknessOf(jacobs2000, "A|B", prefix = "AorB_")
#' 
#' # maximum bottom depth minus minimum top depth of horizon designations matching "A|B"
#' thicknessOf(jacobs2000, "A|B", method = "minmax", prefix = "AorB_")
#' # note that the latter includes the thickness of E horizons between the A and the B
#' 
#' # when using a custom function (be sure to accept ... and consider the effect of NA values)
#' 
#' # calculate cumulative thickness of horizons containing >18% clay
#' thicknessOf(jacobs2000, prefix = "claygt18_", 
#'             FUN = function(x, ...) !is.na(x[["clay"]]) & x[["clay"]] > 18)
#' 
thicknessOf <- function(x, 
                        pattern = NULL, 
                        hzdesgn = hzdesgnname(x, required = TRUE), 
                        method = "cumulative",
                        prefix = "",
                        thickvar = "thickness",
                        depthvars = horizonDepths(x),
                        FUN = function(x, pattern, hzdesgn, ...) grepl(pattern, x[[hzdesgn]]),
                        na.rm = FALSE,
                        ...) {
  
  .internalTHK <- NULL
  .internalHZM <- NULL
  
  if (is.null(hzdesgn) || !hzdesgn %in% horizonNames(x)) {
    stop("Horizon designation column (", hzdesgn, ") does not exist.")
  }
  
  # check inputs
  method <- match.arg(tolower(gsub("/", "", method)), c("cumulative", "minmax"))
  stopifnot(length(thickvar) == 1)
  stopifnot(length(depthvars) == 2)
  
  # extract SPC column names and horizon data
  hzd <- horizonDepths(x)
  idn <- idname(x)
  h <- data.table::data.table(horizons(x))
  
  # determine horizons matching criteria of interest
  h$.internalHZM <- FUN(x, pattern = pattern, hzdesgn = hzdesgn, na.rm = na.rm, ...)
 
  # create a named list for data.table aggregation 
  lid <- list(h[[idn]])
  names(lid) <- idn
  
  if (method == "cumulative") {
    
    # sum thicknesses of all matching horizons
    h$.internalTHK <- x[[hzd[2]]] - x[[hzd[1]]]
    res <- h[, list(thickness = sum(.internalTHK[.internalHZM], na.rm = na.rm)), by = lid]
    colnames(res)[2] <- thickvar
    
  } else if (method == "minmax") {
    
    # determine minimum top depth and maximum bottom depth of matching horizons
    res <- h[, list(`min` = suppressWarnings(min(.SD[[1]][.internalHZM], na.rm = na.rm)),
                    `max` = suppressWarnings(max(.SD[[2]][.internalHZM], na.rm = na.rm))),
              by = lid, .SDcols = c(hzd, ".internalHZM")]
    
    # calculate thickness as MAX(bottom) - MIN(top)
    res$thickness <- res$`max` - res$`min`

    # if a profile has no matching horizons min/max results will be +/- infinity 
    # this means the profile has 0 thickness of matching horizons
    res$thickness[!is.finite(res$thickness)] <- 0L
    
    # use user-defined labels
    colnames(res)[2:3] <- depthvars
    colnames(res)[4] <- thickvar
  } else stop("unknown thickness method: ", shQuote(method), call. = FALSE)
  
  # add prefix
  colnames(res)[2:ncol(res)] <- paste0(prefix, colnames(res)[2:ncol(res)])
  
  # return as SPC data.frame class type
  .as.data.frame.aqp(res, aqp_df_class(x))
}

test_that("thicknessOf works", {
  data("jacobs2000")

  ## method="cumulative"
  
  # cumulative thickness of horizon designations matching "A|B"
  x1 <- thicknessOf(jacobs2000, "A|B", prefix = "AorB_")
  expect_equal(nrow(x1), length(jacobs2000))
  expect_equal(x1$AorB_thickness, c(131, 117, 136, 20, 54, 110, 43))

  ## method="minmax"
  
  # maximum bottom depth minus minimum top depth of horizon designations matching "A|B"
  x2 <- thicknessOf(jacobs2000, "A|B", method = "minmax", prefix = "AorB_")
  expect_equal(ncol(x2), 4)
  expect_equal(x2$AorB_top, rep(0, nrow(x2)))
  expect_equal(x2$AorB_thickness, c(156, 145, 175, 20, 135, 168, 140))
  expect_true(all(x2$AorB_thickness >= x1$AorB_thickness))
  
  ## custom logical function
  
  # calculate cumulative thickness of horizons containing >18% clay
  x3 <- thicknessOf(jacobs2000, FUN = function(x, ...) !is.na(x[["clay"]]) & x[["clay"]] > 18)
  expect_equal(x3$thickness, c(170, 167, 81, 0, 0, 49, 0))
  
  ## missing property and or depth data
  
  # function without NA handling, and na.rm=FALSE
  x4 <- thicknessOf(jacobs2000, FUN = function(x, ...) x[["clay"]] > 18)
  expect_equal(x4$thickness, c(170, 167, 81, 0, NA_integer_, 49, 0))
  
  # function without NA handling, and na.rm=TRUE
  x5 <- thicknessOf(jacobs2000, FUN = function(x, ...) x[["clay"]] > 18, na.rm = TRUE)
  expect_equal(x5$thickness, c(170, 167, 81, 0, 0, 49, 0))
  
  # missing horizon depths, and na.rm=FALSE
  jacobs2000@horizons$top[1] <- NA_integer_
  x6 <- thicknessOf(jacobs2000, "A|B")
  expect_equal(x6$thickness, c(NA_integer_, 117, 136, 20, 54, 110, 43))
  
  # missing horizond depths, and na.rm = TRUE
  x7 <- thicknessOf(jacobs2000, "A|B", na.rm = TRUE)
  expect_equal(x7$thickness, c(113, 117, 136, 20, 54, 110, 43))
})

# x: vector of any type, NA removed
# d: number of significant digits to retain for numeric x
.prepareVector <- function(x, d) {
  
  if(!inherits(x, 'character')) {
    
    # format numeric values to a fixed precision
    if(inherits(x, 'numeric')) {
      x <- format(round(x, digits = d), digits = d)
    }
    
    # convert logical to integers
    if(inherits(x, 'logical')) {
      x <- as.integer(x)
    }
    
    # convert factors to integers
    if(inherits(x, 'factor')) {
      x <- as.integer(x)
    }
    
    # to character
    x <- as.character(x)
  }
  
  # collapse to single string
  # otherwise, memCompress will add '\n' between vector elements
  x <- paste0(x, collapse = '')
  
  return(x)
}

# x: vector of any type, ignoring NA
# d: number of significant digits to retain for numeric x
.prepareBaseline <- function(x, d, type) {
  
  if(type == 'numeric') {
    # replicate simple mean over all horizons
    # weighted by 1cm slices
    b <- format(
      round(
        rep(
          mean(x, na.rm = TRUE), 
          times = length(x)
        ), 
        digits = d
      ),
      digits = d
    )
  } else {
    # logical and factor -> convert to integer codes
    if(!inherits(x, 'character')) {
      x <- as.integer(x)
    }
    
    # baseline is the most frequent, 
    # integer-coded (factor, logical)
    # frequency weighted by hz thickness, 
    # non-NA value
    b <- names(sort(table(x, useNA = 'no'), decreasing = TRUE))[1]
    b <- rep(b, times = length(x))
  }
  
  
  # collapse to single string
  # otherwise, memCompress will add '\n' between vector elements
  b <- paste0(b, collapse = '')
  
  return(b)
}



# x: character, length 1 output from .prepareVector()
# d: number of significant digits to retain for numeric x
# n: compression method
.compressedLength <- function(x, d = 10, m = 'gzip') {
  
  # ideally, numeric data have mean of 0 and sd of 1
  # otherwise, 1, 10, 100 will have different compressed lengths
  
  # memCompress() requires character data
  # this algorithm is based on character vectors flattened to single string
  # output from .prepareVector() is ideal
  if(!inherits(x, 'character') || length(x) > 1) {
    message('flattening character vector')
    x <- .prepareVector(x, d = d)
  }
  
  # otherwise using a single string, with NA removed
  
  # length of compressed vector is the metric
  res <- length(memCompress(x, type = m))
  
  # remove length of empty string ('')
  res <- res - switch(
    m,
    'gzip' = 8, 
    'bzip2' = 14,
    'xz' = 32,
    'none' = 0
  )
  
  return(res)
}



# i: vector, any type
# numericDigits: requested precision for numeric data
.prepareVariable <- function(i, numericDigits) {
  
  # remove all NA
  i <- as.vector(na.omit(i))
  
  # short-circuit for all NA
  if(all(is.na(i))) {
    return(list(v = NULL, b = NULL))
  }
  
  # baseline is mean(i)
  if(is.numeric(i)) {
    
    # baseline for numeric data: 1cm slice wt. mean
    b <- .prepareBaseline(i, d = numericDigits, type = 'numeric')
    
    # prepare for compression
    v <- .prepareVector(i, d = numericDigits)
    
  } else {
    # categorical or logical
    # baseline is most-frequent, non-NA value
    
    # treating all categorical variables as nominal for now
    
    # baseline is the most frequent, weighted by hz thickness, non-NA value
    b <- .prepareBaseline(i, d = numericDigits, type = 'other')
    
    
    # prepare for compression
    v <- .prepareVector(i, d = numericDigits)
  }
  
  return(list(v = v, b = b))
}


.PII_by_profile <- function(x, vars, baseline, method, numericDigits, compression) {
  
  # diced SPC -> DF of horizon data
  # just variables of interest
  h <- as(x, 'data.frame')[, vars, drop = FALSE]
  
  # iterate over columns and prepare data + baseline for PII
  h <- lapply(
    h[, vars, drop = FALSE], 
    FUN = .prepareVariable, 
    numericDigits = numericDigits
  )
  
  # extract variables
  v <- lapply(h, '[[', 'v')
  
  # extract baseline
  b <- lapply(h, '[[', 'b')
  
  res <- switch(
    method,
    'joint' = {
      
      # joint complexity
      
      # concatenate prepared variables, by flattening character vectors
      v <- as.vector(do.call('paste0', list(v, collapse = '')))
      # concatenate prepared baseline, by flattening character vectors
      b <- as.vector(do.call('paste0', list(b, collapse = '')))
      
      # compress entire collection of data
      v <- .compressedLength(v, m = compression)
      
      # this is the joint baseline complexity
      b <- .compressedLength(b, m = compression)
      
      # ratio complexity vs. baseline complexity
      if(baseline) {
        v <- v / b 
      }
      v
    },
    'individual' = {
      
      ## TODO: flattening by .compressedLength() or here?
      # sum of individual complexities
      v <- suppressMessages(sum(sapply(v, .compressedLength, m = compression), na.rm = TRUE))
      b <- suppressMessages(sum(sapply(b, .compressedLength, m = compression), na.rm = TRUE))
      
      # ratio complexity vs. baseline complexity
      if(baseline) {
        v <- v / b 
      }
      v
    }
  )
  
  return(res)
}


#' @title Soil Profile Information Index
#' 
#' @description A simple index of "information" content associated with individuals in a `SoilProfileCollection` object. Information content is quantified by number of bytes after compression via `memCompress()`.
#'
#' @param x `SoilProfileCollection` object
#' 
#' @param vars character vector of site or horizon level attributes to consider
#' 
#' @param method character, 'individual' or 'joint' complexity
#' 
#' @param baseline logical, compute ratio to "baseline" information content, see details
#' 
#' @param numericDigits integer, number of significant digits to retain in numeric -> character conversion
#' 
#' @param padNA logical, pad depths to `max(x)`, supplied to `dice(fill = padNA)`
#' 
#' @param scaleNumeric logical, `scale()` each numeric variable, causing "profile information" to vary based on other profiles in the collection
#' 
#' @param compression character, compression method as used by [memCompress()]: 'gzip', 'bzip2', 'xz', 'none'
#'
#' @return a numeric vector of the same length as `length(x)` and in the same order, suitable for direct assignment to a new site-level attribute 
#' 
#' @export
#'
#' @author D.E. Beaudette
#' 
#' @details Information content via compression (gzip) is the central assumption behind this function: the values associated with a simple soil profile having few horizons and little variation between horizons (isotropic depth-functions) will compress to a much smaller size than a complex profile (many horizons, strong anisotropy). Information content is evaluated a profile at a time, over each site or horizon level attribute specified in `vars`. The `baseline` argument invokes a comparison to the simplest possible representation of each depth-function:
#' 
#'    * `numeric`: replication of the mean value to match the number of horizons with non-NA values
#'    * `character` or `factor`: replication of the most frequent value to match the number of horizons with non-NA values
#'
#' The ratios computed against a "simple" baseline represent something like "information gain". Larger baseline ratios suggest more complexity (more information) associated with a soil profile's depth-functions. Alternatively, the total quantity of information (in bytes) can be determined by setting `baseline = FALSE`.
#'
#'
#'
#' @examples 
#' 
#' # single horizon, constant value
#' p1 <- data.frame(id = 1, top = 0, bottom = 100, p = 5, name = 'H')
#' 
#' # multiple horizons, constant value
#' p2 <- data.frame(
#'   id = 2, top = c(0, 10, 20, 30, 40, 50),
#'   bottom = c(10, 20, 30, 40, 50, 100),
#'   p = rep(5, times = 6),
#'   name = c('A1', 'A2', 'Bw', 'Bt1', 'Bt2', 'C')
#' )
#' 
#' # multiple horizons, random values
#' p3 <- data.frame(
#'   id = 3, top = c(0, 10, 20, 30, 40, 50),
#'   bottom = c(10, 20, 30, 40, 50, 100),
#'   p = c(1, 5, 10, 35, 6, 2),
#'   name = c('A1', 'A2', 'Bw', 'Bt1', 'Bt2', 'C')
#' )
#' 
#' # multiple horizons, mostly NA
#' p4 <- data.frame(
#'   id = 4, top = c(0, 10, 20, 30, 40, 50),
#'   bottom = c(10, 20, 30, 40, 50, 100),
#'   p = c(1, NA, NA, NA, NA, NA),
#'   name = c('A1', 'A2', 'Bw', 'Bt1', 'Bt2', 'C')
#' )
#' 
#' # shallower version of p1
#' p5 <- data.frame(id = 5, top = 0, bottom = 50, p = 5, name = 'H')
#' 
#' # combine and upgrade to SPC
#' z <- rbind(p1, p2, p3, p4, p5)
#' depths(z) <- id ~ top + bottom
#' hzdesgnname(z) <- 'name'
#' 
#' z <- fillHzGaps(z)
#' 
#' # visual check
#' par(mar = c(1, 0, 3, 3))
#' plotSPC(z, color = 'p', name.style = 'center-center', cex.names = 0.8, max.depth = 110)
#' 
#' # factor version of horizon name
#' z$fname <- factor(z$name)
#' 
#' vars <- c('p', 'name')
#' # result is total bytes
#' pi <- profileInformationIndex(z, vars = vars, method = 'joint', baseline = FALSE)
#' 
#' text(x = 1:5, y = 105, labels = pi, cex = 0.85)
#' mtext('Profile Information Index (bytes)', side = 1, line = -1)
#'
profileInformationIndex <- function(x, vars, method = c('joint', 'individual'), baseline = FALSE, numericDigits = 8, padNA = FALSE, scaleNumeric = FALSE, compression = 'gzip') {
  
  # sanity check
  method <- match.arg(method)
  
  ## TODO: think more about this
  # scaling means that PII depends on collection and is not absolute
  if(scaleNumeric) {
    # scale numeric variables to mean of 0 and SD of 1
    for(i in vars) {
      if(inherits(x[[i]], 'numeric')) {
        
        # scale() will return NaN when SD() is NA or 0
        .sd_i <- sd(x[[i]], na.rm = TRUE)
        
        # threshold for very small numbers 
        if(!is.na(.sd_i)) {
          if(.sd_i > 0.0001) {
            x[[i]] <- scale(x[[i]])
          }
        }
        
        # otherwise, use as-is
      }
    }
  }
  
  
  ## note: this will error / drop profiles in the presence of bad horizonation
  
  ## dice() to 1cm intervals for common baseline
  #  -> 10 horizons of the same data NOT more informative than 1 horizon 
  #  -> causes data corruption when bad hz depths present (lots of messages)
  x <- dice(x, fill = padNA)
  
  ## iterate over profiles
  # result is a vector suitable for site-level attribute
  
  ## slow / but simple to understand and test
  # res <- profileApply(
  #   x,
  #   simplify = TRUE,
  #   FUN = .PII_by_profile,
  #   vars = vars,
  #   baseline = baseline,
  #   method = method,
  #   numericDigits = numericDigits,
  #   compression = compression
  # )
  # return(res)
  
  ## fast via data.table
  .idn <- idname(x)
  .pIDs <- profile_id(x)
  
  # result is data.table: id + PII
  # work on horizon data as data.frame
  x <- data.table(horizons(x))
  res <- x[ , 
            .PII_by_profile(
              x = .SD, 
              vars = vars, 
              baseline = baseline, 
              method = method, 
              numericDigits = numericDigits, 
              compression = compression
            ), 
            by = .idn
  ]
  
  # double check order is preserved
  stopifnot(all(res$id == .pIDs))
  
  # down-grade data.table to named numeric vector
  res <- as.vector(res$V1)
  names(res) <- .pIDs
  
  # done
  return(res)
}

context("profile information metrics")

## TODO: finish these
test_that("helper functions", {
  
  v <- aqp:::.prepareVector(1:10, d = 10)
  expect_equal(length(v), 1)
  expect_equal(nchar(v), 11)
  
  # v <- aqp:::.prepareVector(letters, d = 10)
  # v <- aqp:::.prepareVector(TRUE, d = 10)
  # v <- aqp:::.prepareVector(factor(c('A', 'Z')), d = 10)
  # 
  # b <- aqp:::.prepareBaseline(1:10, d = 10, type = 'numeric')
  # 
  # x <- 1:10
  # i <- aqp:::.prepareVariable(x, removeNA = TRUE, numericDigits = 4)
  # 
  # x <- c('A', 'C', 'C', 'D')
  # i <- aqp:::.prepareVariable(x, removeNA = TRUE, numericDigits = 4)
  # 
  # x <- factor(c('A', 'C', 'C', 'D'))
  # i <- aqp:::.prepareVariable(x, removeNA = TRUE, numericDigits = 4)
  # 
  # x <- c(TRUE, TRUE, FALSE)
  # i <- aqp:::.prepareVariable(x, removeNA = TRUE, numericDigits = 4)
  
})


test_that("basic operation", {
  
  a <- data.frame(id = 'A', top = 0, bottom = 100, p = 5)
  b <- data.frame(
    id = 'B', 
    top = c(0, 10, 20, 30, 40, 50), 
    bottom = c(10, 20, 30, 40, 50, 100), 
    p = rep(5, times = 6)
  )
  
  x <- rbind(a, b)
  depths(x) <- id ~ top + bottom
  
  
  # memCompress() give slightly different results across platforms
  # adapt test after this is sorted out
  
  p <- profileInformationIndex(x, vars = 'p', method = 'joint', baseline = FALSE)
  expect_true(inherits(p, c('numeric', 'integer')))
  expect_true(length(p) == 2)
  
  # expect_equal(as.vector(p), c(3, 4.3), tolerance = 0.1)
  
  p <- profileInformationIndex(x, vars = c('p'), method = 'joint')
  expect_true(inherits(p, c('numeric', 'integer')))
  expect_true(length(p) == 2)
  
  # expect_equal(as.vector(p), c(0, 0.458), tolerance = 0.1)
  
  p <- profileInformationIndex(x, vars = c('p'), method = 'joint')
  expect_true(inherits(p, c('numeric', 'integer')))
  expect_true(length(p) == 2)
  
  # expect_equal(as.vector(p), c(0, 0.529), tolerance = 0.1)
  
  p <- profileInformationIndex(x, vars = c('p'), method = 'joint', baseline = FALSE)
  expect_true(inherits(p, c('numeric', 'integer')))
  expect_true(length(p) == 2)
  
  # expect_equal(as.vector(p), c(29, 62))
  
  p <- profileInformationIndex(x, vars = c('p'), method = 'joint', baseline = FALSE)
  expect_true(inherits(p, c('numeric', 'integer')))
  expect_true(length(p) == 2)
  
  # expect_equal(as.vector(p), c(8.666, 19.666), tolerance = 0.1)
  
  p <- profileInformationIndex(x, vars = c('p'), method = 'joint', baseline = FALSE)
  expect_true(inherits(p, c('numeric', 'integer')))
  expect_true(length(p) == 2)
  
  # expect_equal(as.vector(p), c(8, 23))
  
})

#' @name contrastChart
#' @author D.E. Beaudette
#'
#' @title Color Contrast Chart
#'
#' @description Compare one or more pages from a simulated Munsell book of soil colors to a reference color.
#'
#' @param m Munsell representation of a single color for comparison e.g. '10YR 4/3'
#'
#' @param hues vector of one or more Munsell hue pages to display
#'
#' @param ccAbbreviate length of abbreviated contrast classes, use 0 to suppress labels
#'
#' @param style 'hue' or 'CC', see details
#' 
#' @param gridLines logical, add grid lines to the color contrast chart
#' 
#' @param de00.cex character scaling applied to dE00 annotation
#' 
#' @param cc.cex character scaling applied to contrast class annotation
#'
#' @param thresh threshold (<) applied to pair-wise comparisons and resulting color chips
#'
#' @param returnData logical, return lattice figure + data used to generate the figure
#'
#' @details
#' A simulated Munsell color book page or pages are used to demonstrate color contrast between all chips and the reference color \code{m} (highlighted in red). NCSS color contrast class and CIE delta-E00 values are printed below all other color chips. Munsell color chips for chroma 5 and 7 are omitted, but axis labels are retained as a reminder of this fact.
#'
#' Setting \code{style='hue'} emphasizes the contrast classes and CIE delta-E00 of chips adjacent to \code{m}. Setting \code{style='CC'} emphasizes adjacent chips according to respective contrast class via lattice panels.
#'
#' Two-way panels are used when multiple hues are provided and \code{style='CC'}. The default output can be greatly enhanced via:
#'
#'   \code{
#'   latticeExtra::useOuterStrips(...,
#'                                strip = strip.custom(bg=grey(0.85)),
#'                                strip.left = strip.custom(bg=grey(0.85))
#'                               )
#'   }
#'
#' @keywords hplots manip
#' @export
#'
#' @examples
#' 
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' # single hue page
#' contrastChart(m = '10YR 3/3', hues = '10YR')
#'
#' # multiple hue pages
#' contrastChart(m = '10YR 3/3', hues = c('10YR', '2.5Y'))
#'
#' # contrast class, single hue
#' contrastChart(m = '10YR 3/3', hues = '10YR', style='CC')
#'
#' # contrast class, multiple hues
#' # consider latticeExtra::useOuterStrips()
#' contrastChart(m = '10YR 5/6', hues = c('10YR', '2.5Y'), style='CC')
#'

contrastChart <- function(m, hues, ccAbbreviate = 1, style = 'hue', gridLines = FALSE, de00.cex = 0.6, cc.cex = 0.6, thresh = NULL, returnData = FALSE) {

  # load Munsell LUT
  # safe for CRAN check
  munsell <- NULL
  load(system.file("data/munsell.rda", package = "aqp")[1])

  # plotting details
  y.offset.CC <- 0.08
  y.offset.dE00 <- 0.25

  # safety checks on arguments
  # check style
  if(! style %in% c('hue', 'CC', 'dE00')) {
    stop("style must be one of: 'hue', 'CC'", call. = FALSE)
  }

  # multiHue required when length(hues) > 1
  if(length(hues) > 1 & style == 'hue') {
    style <- 'multiHue'
    
    # move dE00 up if CC is disabled via 0-length abbreviation
    if(ccAbbreviate == 0) {
      y.offset.dE00 <- y.offset.dE00 - 0.1
    }
  }

  # 2. disable CC label when style = CC
  if(style == 'CC') {
    # hack to disable CC label
    ccAbbreviate <- 0
    # move dE00 up
    y.offset.dE00 <- y.offset.dE00 - 0.1

    # multiple hues
    if(length(hues) > 1) {
      style <- 'multiCC'
    }
  }

  # extract just requested hues
  # along with standard value/chroma pairs found on a typical color book page
  chroma.subset <- c(1, 2, 3, 4, 6, 8)
  x <- munsell[which(munsell$value %in% 2:8 & munsell$chroma %in% chroma.subset & munsell$hue %in% hues), ]

  # convert into hex notation for plotting
  x$color <- munsell2rgb(x$hue, x$value, x$chroma)
  x$munsell <- sprintf("%s %s/%s", x$hue, x$value, x$chroma)

  # re-level hues according to color contrast guidance
  hh <- unique(x$hue)
  ll <- hh[order(huePosition(hh))]
  x$hue <- factor(x$hue, levels=ll)

  # setup query color table
  m <- data.frame(
    queryColor = m, 
    parseMunsell(m, convertColors = FALSE),
    stringsAsFactors = FALSE
  )
  m$value <- as.integer(m$value)
  m$chroma <- as.integer(m$chroma)

  # compute all pair-wise contrast classes and dE00
  cc <- colorContrast(x$munsell, rep(m$queryColor, times = nrow(x)))

  # join for plotting
  z <- merge(x, cc, by.x='munsell', by.y='m1', all.x=TRUE, sort=FALSE)

  # dE00 thresholding
  if( ! is.null(thresh) ) {
    z <- z[which(z$dE00 < thresh), ]

  }

  # convert chroma into a factor, in this way we can skip chroma 5/7
  # note that the full levels of chroma have to be set 
  z$chroma <- factor(z$chroma, levels = chroma.subset)
  
  # alternative modes
  fm <- switch(style,
               'hue' = as.formula('value ~ chroma'),
               'multiHue' = as.formula('value ~ chroma | hue'),
               'CC' = as.formula('value ~ chroma | cc'),
               'multiCC' = as.formula('value ~ chroma | cc + hue')
  )


  # custom x-axis labeling based on factors vs. integers
  chroma.axis.at <- c(seq_along(chroma.subset), 4.5, 5.5)
  chroma.subset.labels <- as.character(c(chroma.subset, 5, 7))

  # make plot
  pp <- lattice::xyplot(fm, data = z,
               main = sprintf('Color Contrast Chart: %s', m$queryColor),
               asp = 1.2, xlab='Chroma', ylab='Value',
               xlim = c(0.75, 6.25), 
               ylim = c(1.75, 8.25),
               scales = list(
                 alternating = 1, tick.number = 8, relation = 'free', y = list(
                   rot = 0
                 )
                 ,
                 x = list(
                   at = chroma.axis.at,
                   labels = chroma.subset.labels
                 )
               ),
               as.table = TRUE, strip = strip.custom(bg='grey'),
               subscripts = TRUE,
               panel = function(xx, yy, subscripts, ...) {
                 
                 # note: source `xx` and `yy` aren't actually used for anything
                 #       -> chroma factor levels reset when full levels aren't present
                 
                 # prep data for this panel
                 d <- z[subscripts, ]
                 d$cc <- as.character(d$cc)
                 d$dE00 <- format(d$dE00, digits = 2)
                 
                 # extract from d vs. function arguments
                 xx <- as.integer(d$chroma)
                 yy <- d$value
                 
                 # remove query color contrast and dE00
                 idx <- which(d$munsell == m$queryColor)
                 d$cc[idx] <- ''
                 d$dE00[idx] <- ''
                 
                 # grid system
                 if(gridLines) {
                   panel.abline(h = 2:8, v=1:8, col=grey(0.85), lty=1) 
                 }
                 
                 # offsets, may require additional tinkering
                 bd.side <- 0.3
                 bd.bottom <- 0.2
                 bd.top <- 0.4
                 bd.annot <- 0.05
                 
                 # color chips
                 # border encodes query chip
                 chip.border <- rep('black', times=nrow(d))
                 chip.lwd <- rep(1, times=nrow(d))
                 
                 # update border colors and thickness for the query color
                 border.idx <- which(d$hue == m$hue & d$value == m$value & d$chroma == m$chroma)
                 chip.border[border.idx] <- 'red'
                 chip.lwd[border.idx] <- 3
                 
                 panel.rect(
                   xleft = xx - bd.side,
                   ybottom = yy - bd.bottom,
                   xright = xx + bd.side,
                   ytop = yy + bd.top,
                   col = d$color,
                   border = chip.border,
                   lwd = chip.lwd
                 )
                 
                 # optionally annotate contrast class and abbreviate
                 if(ccAbbreviate >= 1) {
                   panel.text(
                     x = xx,
                     y = yy - (bd.bottom + y.offset.CC),
                     as.character(abbreviate(d$cc, minlength = ccAbbreviate)),
                     cex = cc.cex,
                     font = 4
                   )
                 }
                 
                 
                 # annotate dE00
                 panel.text(
                   x = xx,
                   y = yy - (bd.bottom + y.offset.dE00),
                   d$dE00,
                   cex = de00.cex
                 )
               }
  )

  # return list containing figure and data
  # figure is first so that it is printed to the graphics device
  if(returnData) {
    return(list(fig = pp, data = z))
  } else {
    # just the figure
    return(pp)
  }

}

#' @title Grouped Soil Profile Plot
#'
#' @description Plot a collection of soil profiles, sorted by group.
#'
#' The left-right ordering of groups can be adjusted by converting
#' \code{groups} into a factor and explicitly setting factor levels.
#' Alpha-numeric ordering is used for all other types.
#'
#' @param x a `SoilProfileCollection` object
#' 
#' @param groups the name of a site-level attribute that defines groups, factor
#' levels will influence plotting order
#' 
#' @param group.name.offset vertical offset for group names, single numeric
#' value or vector of offsets
#' 
#' @param group.name.cex font size for group names
#' 
#' @param group.line.col color for line that splits groups
#' 
#' @param group.line.lwd width of line that splits groups
#' 
#' @param group.line.lty style of line that splits groups
#' 
#' @param break.style style of group boundaries: "line", "arrow", "both"
#' 
#' @param break.offset horizontal offset used to place vertical breaks and/or arrows, shifted slightly to the right of default when `hz.depths=TRUE` is passed to `plotSPC()`
#' 
#' @param arrow.offset vertical offset for "arrow" style boundaries, single
#' numeric value or vector of offsets
#' 
#' @param arrow.length value passed to `arrows` to define arrow head size
#' 
#' @param \dots further arguments to `plotSPC`
#' 
#' @author D.E. Beaudette
#' 
#' @seealso \code{\link{plotSPC}}
#' 
#' @keywords hplots
#' @export
#' @examples
#'
#' # sample data
#' data(sp1)
#' # convert colors from Munsell to hex-encoded RGB
#' sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))
#'
#' # promote to SoilProfileCollection
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' # add a groups
#' sp1$group.2 <- sprintf("%s-%s", rev(LETTERS[1:3]), sp1$group)
#'
#' # convert fake groupt to factor with new levels
#' sp1$group.3 <- factor(sp1$group.2, levels=c('C-2', 'B-2', 'A-2', 'C-1', 'B-1', 'A-1'))
#'
#' # plot profiles, sorted and annotated by 'group' (integers)
#' par(mar=c(1,1,1,1))
#' groupedProfilePlot(sp1, groups='group', max.depth=150, group.name.offset = -5, id.style='side')
#'
#' # plot profiles, sorted and annotated by 'group.2' (characters)
#' par(mar=c(1,1,1,1))
#' groupedProfilePlot(sp1, groups='group.2', max.depth=150, group.name.offset = -5, id.style='side')
#'
#' # plot profiles, sorted and annotated by 'group.3' (characters)
#' par(mar=c(1,1,1,1))
#' groupedProfilePlot(sp1, groups='group.3', max.depth=150, group.name.offset = -5, id.style='side')
#'
#'
#' # make fake site-level attribute and adjust levels
#' sp1$new.group <- sample(letters[1:3], size=length(sp1), replace=TRUE)
#'
#' # tabulate pedons / group
#' tab <- table(sp1$new.group)
#'
#' # sort large -> small
#' tab <- sort(tab, decreasing = TRUE)
#'
#' # set levels based on sorted tabulation
#' # assign custom labels
#' sp1$new.group <- factor(sp1$new.group, levels=names(tab),
#' labels=paste0(names(tab), ' (', tab, ')'))
#'
#' groupedProfilePlot(sp1, groups='new.group', max.depth=150,
#' group.name.offset = -10, id.style='side')
#'
#' # offsets can be set using a vector of values, recycled as needed
#' groupedProfilePlot(sp1, groups='new.group', max.depth=150,
#' group.name.offset=c(-10, -5), id.style='side')
#'
#' # annotate with arrows instead of vertical lines
#' groupedProfilePlot(sp1, groups='new.group', max.depth=150,
#' group.name.offset=c(-10, -12), break.style='arrow', arrow.offset=-3,
#' group.line.lty = 1, group.line.lwd = 1, id.style='side')
#'
#'
#' \dontrun{
#' # more complete example using data from soilDB package
#' data(loafercreek, package='soilDB')
#' par(mar=c(1,1,1,1))
#' # lines
#' groupedProfilePlot(loafercreek, groups='hillslopeprof', group.name.cex = 0.5,
#' group.name.offset = -10)
#'
#' # arrows
#' groupedProfilePlot(loafercreek, groups='hillslopeprof', group.name.cex = 0.5,
#' group.name.offset = -10, break.style ='arrow', group.line.lty = 1,
#' group.line.lwd = 1)
#'
#' # both
#' groupedProfilePlot(loafercreek, groups='hillslopeprof', group.name.cex = 0.5,
#' group.name.offset = -10, break.style ='both', group.line.lty = 1,
#' group.line.lwd = 1)
#' }
#'
groupedProfilePlot <- function(x, groups, group.name.offset = -5, group.name.cex = 0.75, group.line.col = 'RoyalBlue', group.line.lwd = 2, group.line.lty = 2, break.style = c('line', 'arrow', 'both'), break.offset = 0.5, arrow.offset = group.name.offset + 5, arrow.length = 0.1, ...) {

  # sanity check, `groups` must be a site-level attribute
  if(! any(groups %in% siteNames(x))) {
    stop(sprintf("%s is not a site-level attribute", groups), call. = FALSE)
  }

  # sanity check
  break.style <- match.arg(break.style)
  
  # extract site data, used for ordering and labeling
  s <- site(x)

  # break.offset needs to be shifted to the right a little bit, when hz.depths = TRUE is specified
  if('hz.depths' %in% names(list(...)) && list(...)[['hz.depths']]) {
    # only make a change if break.offset isn't specified
    if(missing(break.offset)){
      # shift to the right slightly
      # message('`hz.depths=TRUE`, shifting `break.offset` to the right')
      break.offset <- 0.62
    }
  }
  
  # if groups are already a factor, keep existing levels
  # note that an ordered factor has multiple classes: "ordered" , "factor"
  if(! any(inherits(s[[groups]], 'factor'))) {
    s[[groups]] <- factor(as.character(s[[groups]]))
  }

  # derive ordering based on `groups`
  # alpha sorting when `grpups` is numeric or character
  # factor levels when `groups` is a factor
  new.order <- order(s[[groups]])
  lab <- s[[groups]][new.order]

  # test for NA
  NA.lab <- which(is.na(lab))

  # replace with missing label with '<missing>'
  # this requires conversion: factor -> character -> replace NA -> factor with new levels
  if(length(NA.lab) > 0) {
    message('NA in grouping label, filling with `<missing>`')
    o.levels <- levels(lab)
    lab <- as.character(lab)
    lab[NA.lab] <- '<missing>'
    lab <- factor(lab, levels=c(o.levels, '<missing>'))
  }

  # setup plot with plot.SoilProfileCollection
  plotSPC(x, plot.order = new.order, ...)

  # get last plot parameters
  lsp <- get('last_spc_plot', envir=aqp.env)

  # get just those levels that are in our data, preserving order of original levels
  unique.lab <- levels(lab)[which(levels(lab) %in% unique(lab))]
  group.lengths <- rle(as.numeric(lab))$lengths

  # label positions
  lab.positions <- (cumsum(group.lengths) - (group.lengths / 2)) + break.offset

  # group boundaries on x-axis
  boundary.positions <-  cumsum(group.lengths)[-length(group.lengths)] + break.offset

  # reasonable upper / lower boundaries on y-axis
  # these are informed by plotting parameters sent to plotSPC()
  # note: this is a vector -> reduce to scalar
  .yoffset <- mean(lsp$y.offset)
  
  upper.position <- (.yoffset) + (group.name.offset/2 * lsp$scaling.factor)
  lower.position <- (.yoffset) + (lsp$max.depth * lsp$scaling.factor)
  
  # only add grouping symbols if number of groups is > 1
  if(length(boundary.positions)) {
    # add group boundaries
    if(break.style == 'line')
      segments(y0 = upper.position, y1=lower.position, x0=boundary.positions, x1=boundary.positions, lty=group.line.lty, lwd=group.line.lwd, col=group.line.col)

    if(break.style == 'arrow')
      arrows(x0=c(0.5, boundary.positions), x1=c(boundary.positions, length(x)+0.5), y0=arrow.offset, code=3, length=arrow.length, lty=group.line.lty, lwd=group.line.lwd, col=group.line.col)

    if(break.style == 'both') {
      segments(y0 = upper.position, y1=lower.position, x0=boundary.positions, x1=boundary.positions, lty=group.line.lty, lwd=group.line.lwd, col=group.line.col)
      arrows(x0=c(0.5, boundary.positions), x1=c(boundary.positions, length(x)+0.5), y0=arrow.offset, code=3, length=arrow.length, lty=group.line.lty, lwd=group.line.lwd, col=group.line.col)
    }
  }

  # annotate with group labels at group centers
  text(lab.positions, group.name.offset, unique.lab, cex = group.name.cex, adj = c(0.5, 0), font = 4)
}

# GENERIC FUNCTION for dealing with old-style O horizons and other? abnormalities in datum

#' @title Accumulate horizon depths, and reflect reversed depths, relative to new datum
#' 
#' @description
#' Fix old-style organic horizon depths or depths with a non-standard datum by the "depth accumulation" method.
#' 
#' @details The "depth accumulation" method calculates thicknesses of individual horizons and then cumulative sums them after putting them in `id` + top depth order. The routine tries to determine context based on `hzname` and `pattern`. The main transformation is if a top depth is deeper than the bottom depth, the depths are reflected on the Z-axis (made negative). The data are then `id` + top depth sorted again, the thickness calculated and accumulated to replace the old depths.
#' 
#' This function uses several heuristics to adjust data before transformation and thickness calculation:
#'   
#' #### Regex matching of horizon designation patterns and similar
#'   - matches of `pattern` where both top and bottom depth `NA` -> `[0,1]` `[top,bottom]` depth
#' - REMOVE horizons that do not match `pattern` where both top and bottom depths `NA`
#' 
#' #### Over-ride `hzname` handling with the sequence column argument `seqnum`
#' - if `seqnum` column specified "first record with `NA` `hzname`" is considered a `pattern` match if `seqnum == 1`
#' 
#' #### Trigger "fixing" with the `fix` argument:
#' - Add 1 cm to bottom-most horizons with `NA` bottom depth
#' - Add 1 cm thickness to horizons with top and bottom depth equal
#' - Add 1 cm thickness to horizons with `NA` top depth and bottom depth `0`
#' 
#' @param x A `data.frame` or `SoilProfileCollection`
#' 
#' @param id unique profile ID. Default: `NULL`, if `x` is a SoilProfileCollection `idname(x)`
#' 
#' @param hzdepths character vector containing horizon top and bottom depth column names. Default: `NULL`, if `x` is
#'  a SoilProfileCollection `horizonDepths(x)`
#'  
#' @param hzname character vector containing horizon designation or other label column names. Default: `NULL`, if `x`
#'  is a SoilProfileCollection `hzdesgnname(x)`
#'  
#' @param hzdatum a numeric vector to add to accumulated depths. Default: `0`. Can be equal in length to number of profiles if `x` is a `SoilProfileCollection` or number of (unique) IDs if `x` is a `data.frame`.
#' 
#' @param seqnum Optional: character vector containing record "sequence number" column name; used in-lieu of `hzname` (when `NA`) to identify "first" record in a profile
#' 
#' @param pattern pattern to search for in `hzname` to identify matching horizons to append the profile to
#' 
#' @param fix apply adjustments to missing (`NA`) depths and expand 0-thickness horizons? Default: `TRUE` 
#'
#' @return A horizon-level `data.frame`, suitable for promoting to SPC with `depths<-`, or a `SoilProfileCollection`, depending on the class of `x`.
#' @export
#'
#' @examples
#' 
#' # example using hzdatum argument
#' data(sp4)
#' depths(sp4) <- id ~ top + bottom
#' hz <- accumulateDepths(sp4,
#'                        id = "id",
#'                        hzdepths = c("top", "bottom"),
#'                        hzname = "name",
#'                        hzdatum = 5 * 1:length(sp4))
#' plot(hz)
#'   
#' # example using old-style O horizons
#' hz <- read.table(text = "peiidref hzdept hzdepb hzname seqnum phiid
#'                     1        11      0      5      A      2   295
#'                     2        11      1      0     Oe      1   294
#'                     3        11      5     13     C1      3   296
#'                     4        11     13     58     C2      4   297
#'                     5        11     58    152     C3      5   298
#'                     6        13      0      5      A      2   303
#'                     7        13      1      0     Oe      1   302
#'                     8        13      5     25     Bw      3   304
#'                     9        13     25     61      C      4   305
#'                     10       13     61     NA      R      5   306
#'                     11      136      0     13     A1      3   695
#'                     12      136      1      0     Oe      2   694
#'                     13      136      2      1     Oi      1   693
#'                     14      136     13     61     C1      4   696
#'                     15      136     61     76     C2      5   697")
#'                     
#' depths(hz) <- peiidref ~ hzdept + hzdepb
#'
#' hz_fixed <- accumulateDepths(hz,
#'                              id = "peiidref",
#'                              hzdepths = c("hzdept", "hzdepb"),
#'                              hzname = "hzname")
#' is_valid <- checkHzDepthLogic(hz_fixed)$valid
#' 
#' test0 <- subset(hz_fixed, !is_valid)
#' test1 <- subset(hz_fixed, is_valid)
#' 
#' origO <- subset(hz, grepl("O", hzname))
#' fixedO <- subset(hz_fixed, grepl("O", hzname))
#' 
#' par(mfrow = c(2, 1), mar = c(0, 0, 3, 2))
#' 
#' plotSPC(origO, max.depth = 25)
#' plotSPC(fixedO, max.depth = 25)
#' 
accumulateDepths <- function(x,
                             id = NULL,
                             hzdepths = NULL,
                             hzname = NULL,
                             hzdatum = 0,
                             seqnum = NULL,
                             pattern = "O",
                             fix = TRUE) {
  
  # x (becomes) a data.frame of horizon-level data
  if (inherits(x, 'SoilProfileCollection')) {
    dat <- horizons(x)
    
    if (is.null(id)) {
      id <- idname(x)
    }
    
    if (is.null(hzname)) {
      hzname <- hzdesgnname(x, required = TRUE)
    }
    
    if (is.null(hzdepths)) {
      hzdepths <- horizonDepths(x)
    }
    
  } else if (inherits(x, 'data.frame')) {
    
    dat <- x    
    
    if (is.null(id) || is.null(hzname) || is.null(hzdepths)) 
      stop("If x is a data.frame `id`, `hzdepths` and `hzname` must be specified!", call. = FALSE)
    
  } else {
    stop("x must be a SoilProfileCollection or data.frame", call. = FALSE)
  }
  
  # convert to data.table, set up some safe globals to use
  .internalID <- id
  .I <- NULL
  .N <- NULL
  .SD <- NULL
  .TOP <- hzdepths[1]
  .BOTTOM <- hzdepths[2]
  thk <- NULL
  .hzdatum <- NULL
  dat <- data.table::as.data.table(dat)
  nunique <- length(unique(dat[[.internalID]]))

  # hzdatum must be constant (unit length) or unique to each profile ID
  if (is.numeric(hzdatum) && length(hzdatum) == 1) {
    hzdatum <- as.numeric(rep(hzdatum, nunique))
  }
  if (!is.numeric(hzdatum) || length(hzdatum) != nunique) {
    stop("hzdatum must be a numeric vector equal in length to number of unique `id`, or constant (with length 1)", call. = FALSE)
  }
  
  # set up for fix of matches where both top and bottom NA
  #  the rest of profile is "appended" to them if they match
  match.na.idx <- which(is.na(dat[[.TOP]]) & 
                          is.na(dat[[.BOTTOM]]) &
                          grepl(pattern, dat[[hzname]]))
  
  if (length(match.na.idx) > 0 && fix)
    dat[[.BOTTOM]][match.na.idx] <- 0
  
  # seqnum specified triggers "first record with NA hzname" fixing
  if (!is.null(seqnum)) {
    bad.seq.idx <- which(is.na(dat[[hzname]]) & dat[[seqnum]] == 1)
    
    if (length(bad.seq.idx) > 0 && fix)
      dat[[.BOTTOM]][bad.seq.idx] <- 0
  }
  
  # remove rows where both top and bottom depths NA (nontarget horizons)
  not.na.idx <- !(is.na(dat[[.TOP]]) & is.na(dat[[.BOTTOM]])) 
  dat <- dat[not.na.idx,]
  
  .BYLIST1 <- list(dat[[.internalID]])
  names(.BYLIST1) <- .internalID
  
  # add 1cm to bottom-most horizons with NA bottom depth
  bad.bottom.idx <- na.omit(dat[, .I[.SD[[.TOP]] == suppressWarnings(max(.SD[[.TOP]], 
                                                                         na.rm = TRUE)) &
                                       is.na(.SD[[.BOTTOM]])], by = .BYLIST1]$V1)
  
  # 0-thickness O horizon
  
  # matches pattern or no name
  thk1 <- (grepl(pattern, dat[[hzname]]) | is.na(dat[[hzname]]))
  
  # top and bottom depth equal
  thk2 <- dat[[.TOP]] == dat[[.BOTTOM]]
  
  # top depth NA and bottom depth 0
  thk3 <- is.na(dat[[.TOP]]) & dat[[.BOTTOM]] == 0
  bad.thk.idx <- which(thk1 & thk2) 
  bad.thk.idx2 <- which(thk1 & thk3)  
  
  if (fix) {
    
    # add 1 to bottom most horizons with na bottom depth
    dat[bad.bottom.idx, ][[.BOTTOM]] <- dat[bad.bottom.idx,][[.TOP]] + 1
    
    # add 1 to 0 thickness
    dat[bad.thk.idx, ][[.TOP]] <- dat[[.TOP]][bad.thk.idx] + 1
    
    # this creates "old style" o horizon for missing TOP depth with bottom==0
    dat[bad.thk.idx2, ][[.TOP]] <- dat[[.BOTTOM]][bad.thk.idx2] + 1
  }
  
  ## BUG: if any top depth is NA
  ## sorting doesn't work
  ## subsequent logic corrupts depths
  
  # ORDER
  dat <- dat[order(as.character(dat[[.internalID]]), dat[[.TOP]]),]
  
  # calculate index
  bad.ldx <- grepl(pattern, dat[[hzname]]) & dat[[.BOTTOM]] < dat[[.TOP]]
  bad.ldx[is.na(bad.ldx)] <- TRUE
  
  bad.idx <- which(bad.ldx)
  
  # bad.peiids <- unique(dat[bad.idx, id])
  
  # make depths that are top/bottom reversed NEGATIVE
  dat[bad.idx, hzdepths] <- -dat[bad.idx, .SD, .SDcols = hzdepths]
  
  # re-order using SPC-style ID+top depth sorting (again!)
  dat <- dat[order(as.character(dat[[.internalID]]), dat[[.TOP]]),]
  
  # calculate thickness using ordered horizons
  dat$thk <- dat[[.BOTTOM]] - dat[[.TOP]]
  dat$thk[is.na(dat$thk)] <- 0
  
  ## previously below code needed more than 1 record
  # test1 <- dat[, .N, by = .BYLIST1]
  # dat <- dat[!dat[[.internalID]] %in% test1[[.internalID]][test1$N == 1],]
  
  .BYLIST2 <- list(dat[[.internalID]])
  names(.BYLIST2) <- .internalID
  
  # cumulative sums of thickness to make new top/bottom depths
  dat$.hzdatum <- hzdatum[match(dat[[.internalID]], unique(dat[[.internalID]]))]
           
  dat_after <- dat[, list(top = c(.hzdatum[1] + min(abs(.SD[[.TOP]])), .hzdatum[1] + min(abs(.SD[[.TOP]])) + cumsum(thk[1:(.N - 1)]))[seq_along(thk)],
                              bottom = .hzdatum[1] + min(abs(.SD[[.TOP]])) + cumsum(thk)), 
                       by = .BYLIST2]
  
  .BYLIST3 <- list(dat_after[[.internalID]])
  names(.BYLIST3) <- .internalID
  
  ## TODO: this doesn't appear to be used
  # calculate number of rows in resulting horizon data
  test2 <- dat_after[, .N, by = .BYLIST3]
  
  # insert into working subset
  dat[[.TOP]] <- dat_after$top
  dat[[.BOTTOM]] <- dat_after$bottom
  
  ndrop <- sum(!dat[[.internalID]] %in% dat[[.internalID]])
  
  if (ndrop > 0)
    message("profile IDs (n=", ndrop, ") dropped from set")
  
  ## this shouldn't happen any more
  # if (nrow(dat) != nrow(dat_sub))
  #   message("horizons (n=",(nrow(dat) - nrow(dat_sub)),") dropped from set")
  
  # return SPC if input was SPC
  if (inherits(x, 'SoilProfileCollection')) {
    
    # subset with [,
    x <- x[which(profile_id(x) %in% dat[[idname(x)]]), ]
    
    # replace horizons
    replaceHorizons(x) <- .as.data.frame.aqp(dat, aqp_df_class(x))
    
    return(x)
  } else {
    # regular data.frame result
    return(data.frame(dat))
  }
}

#
#' @title Get IDs of Deepest Horizons by Profile
#'
#' @description Return horizon IDs of the deepest horizon within each profile of a `SoilProfileCollection`. IDs are returned in the same order as `profile_id(x)`. Horizon top depths are used because there are cases where bottom depths may be missing.
#'
#' @param x a `SoilProfileCollection`
#' @export
getLastHorizonID <- function(x) {

  # with data.table
  hztb <- horizonDepths(x)
  hzidn <- hzidname(x)
  idn <- idname(x)

  .N <- NULL
  .I <- NULL

  # ensure site and horizons are "in sync"
  stopifnot(spc_in_sync(x)$valid)

  h <- data.table::as.data.table(horizons(x))

  # get row index of last horizon by group
  idx <- h[, .I[.N], by = list(h[[idn]])]$V1

  # get hzID for just last horizon
  res <- hzID(x)[idx]

  # add profile IDs as names
  names(res) <- profile_id(x)

  # a named vector, in the same order as profile_id(x)
  return(res)
}

## R functions to compute the weighted naive and tau statistics and display the results
## tauW : compute
## summaryTauW : display

## R function to compute the weighted naive and tau statistics
##
## Arguments:
##   CM: a square confusion (cross-classification) matrix
##       (rows: allocation, columns: reference)
##   W : weights; 1 on diagonals, [-1..1] off giving partial credit to this error
##       0 = no credit; 1 = full credit; -1 = maximum penalty
##       If absent, default is no partial credit, i.e., unweighted
##   P : prior probability vector, length = number of rows/columns in CM and W
##	 If absent, P are equal priors for each class
##       Special value P = 0 is interpreted as P = column marginals
##
## Input matrices may be in data.frame format and will be converted
##
## Computes: (1) unweighted naive, (2) weighted naive, (3) unweighted tau, (4) weighted tau
##
## Results are returned in a list with obvious R names
## Error checks: CM must be square; P must have correct number of classes
##	and sum to 1 +/- 0.0001; W & CM must be same size
#
## Author: D G Rossiter, 26-Dec-2016, Ithaca, NY (USA)

#' Compute weighted naive and \emph{tau} statistics for a cross-classification matrix
#'
#' \code{tauW}: Computes: (1) unweighted naive, (2) weighted naive, (3) unweighted \emph{tau}, (4) weighted \emph{tau} accuracy statistics
#'
#' \code{summaryTauW}: prints a summary of the results from \emph{tauW}
#'
#' \code{xtableTauW}: formats a LaTeX table with results from \emph{tauW} and saves it as a \code{.tex} file for import into a LaTeX document.
#'
#' Input matrices \code{CM} and \code{W} may be in \code{data.frame} format and will be converted
#'
#' Weights matrix \code{W}: 0 = no credit; 1 = full credit; -1 = maximum penalty
#'
#' If absent, default is no partial credit, i.e., unweighted.
#'
#' Prior probabilities vector \code{P}: If absent, \code{P} are equal priors for each class. Special value \code{P = 0} is interpreted as \code{P} = column marginals.
#'
#' Error checks: \code{CM} must be square; \code{P} must have correct number of classes and sum to 1 +/- 0.0001; \code{W} & \code{CM} must be conformable
#'
#' @param CM a square confusion (cross-classification) matrix (rows: allocation, columns: reference)
#' @param W weights: 1 on diagonals, \[-1..1] off giving partial credit to this error
#' @param P prior probability vector, length = number of rows/columns in \code{CM} and \code{W}
#' @return Results are returned in a list with obvious R names
#' @author D.G. Rossiter
#'
#' @references
#'
#'    - Rossiter, D. G., Zeng, R., & Zhang, G.-L. (2017). Accounting for taxonomic distance in accuracy assessment of soil class predictions. Geoderma, 292, 118–127. \doi{10.1016/j.geoderma.2017.01.012}
#'
#'    - Ma, Z. K., & Redmond, R. L. (1995). Tau-coefficients for accuracy assessment of classification of remote-sensing data. Photogrammetric Engineering and Remote Sensing, 61(4), 435–439.
#'
#'    - Naesset, E. (1996). Conditional tau coefficient for assessment of producer’s accuracy of classified remotely sensed data. ISPRS Journal of Photogrammetry and Remote Sensing, 51(2), 91–98. \doi{10.1016/0924-2716(69)00007-4}
#'
#' @keywords array
#' @export
#' @examples
#'
#' # example confusion matrix
#' # rows: allocation (user's counts)
#' # columns: reference (producer's counts)
#' crossclass <- matrix(data=c(2,1,0,5,0,0,
#'                             1,74,2,1,3,6,
#'                             0,5,8,6,1,3,
#'                             6,1,3,91,0,0,
#'                             0,4,0,0,0,4,
#'                             0,6,2,2,4,38),
#'                      nrow=6, byrow=TRUE)
#' row.names(crossclass) <- c("OP", "SA", "UA", "UC", "AV", "AC")
#' colnames(crossclass) <- row.names(crossclass)
#'
#' # build the weights matrix
#' # how much credit for a mis-allocation
#' weights <- matrix(data=c(1.00,0.05,0.05,0.15,0.05,0.15,
#'                          0.05,1.00,0.05,0.05,0.05,0.35,
#'                          0.05,0.05,1.00,0.20,0.15,0.15,
#'                          0.15,0.05,0.25,1.00,0.10,0.25,
#'                          0.05,0.10,0.15,0.10,1.00,0.15,
#'                          0.20,0.30,0.10,0.25,0.20,1.00),
#'                   nrow=6, byrow=TRUE)
#'
#' # unweighted accuracy
#' summaryTauW(nnaive <- tauW(crossclass))
#'
#' # unweighted tau with equal priors, equivalent to Foody (1992) modified Kappa
#' tauW(crossclass)$tau
#'
#' # unweighted tau with user's = producer's marginals, equivalent to original kappa
#' (priors <-  apply(crossclass, 2, sum)/sum(crossclass))
#' tauW(crossclass, P=priors)$tau
#'
#' # weighted accuracy; tau with equal priors
#' summaryTauW(weighted <- tauW(crossclass, W=weights))
#'
#' # weighted accuracy; tau with user's = producer's marginals
#' summaryTauW(tauW(crossclass, W=weights, P=priors))
#'
#' # change in accuracy statistics weighted vs. non-weighted
#' (weighted$overall.weighted - weighted$overall.naive)
#' (weighted$user.weighted - weighted$user.naive)
#' (weighted$prod.weighted - weighted$prod.naive)
#'
tauW <- function(CM,
                W = diag(sqrt(length(as.matrix(CM)))),
                P = rep(1/nrow(as.matrix(CM)), nrow(as.matrix(CM)))) {

    ## convert both data frames and vectors to matrices
  cmx <- as.matrix(CM); wx <- as.matrix(W); pv <- as.vector(P)

  ## convert a 1-D vector to a square matrix
  if (ncol(cmx) == 1)
    cmx <- matrix(cmx, byrow=TRUE, nrow=sqrt(nrow(cmx)))
  if (ncol(wx) == 1)
    wx <- matrix(wx, byrow=TRUE, nrow=sqrt(nrow(wx)))
  nr <- nrow(cmx); nc <- ncol(cmx)

  ## check CM
  if (nr != nc) {
    print("Error: cross-classification matrix is not square")
    return(NULL)
  }

    ## check W
  if (dim(wx)[1] != dim(wx)[2]) {
    print("Error: weights matrix is not square")
    return(NULL)
  }
  if (dim(wx)[1] != dim(cmx)[1]) {
    print("Weight and Confusion Matrices are not the same size")
    return(NULL)
  }
  if ((any(wx > 1)) || (any(wx < -1))) {
    print("All weights must be on [-1..1]")
    return(NULL)
  }

  ## totals of matrix, rows and columns
  n <- sum(cmx); rsum = apply(cmx, 1, sum); csum = apply(cmx, 2, sum)

  ## if P is special value 0, set to reference distribution
  if ((length(pv))==1 && (pv == 0)) {
    print("Setting prior probabilities to reference class distribution")
    pv <- csum/n
  }

  ## check P
  if (length(pv) != nc) {
    print("Error: number of prior probabilities not equal to number of classes")
    return(NULL)
  }
  if (abs(1-sum(pv)) > 0.0001) {
    print("Error: prior probabilities must sum to 1")
    return(NULL)
  }

  ## make class names consistent among objects
  row.names(cmx) <- names(cmx)
  row.names(wx) <- names(wx)
  names(pv) <- names(cmx)

  ## compute unweighted naive statistics
  d <- diag(cmx); dsum <- sum(d); oa <- dsum/n
  ua <- d/rsum; pa <- d/csum
  names(ua) <- names(pa)

  ## compute weighted naive statistics
  ##

  ## confusion matrix and marginals as proportions
  cmxp <- cmx/n; cp <- csum/n; rp <- rsum/n;

  ## expected proportions
  pp<- rp %o% cp

  ## weighted weights
  ## wr <- wx %*% cp; wc <- t(t(wx) %*% rp)
  ## overall weighted accuracy
  oaw <- sum(wx * cmxp)

  ## marginal weighted accuracy
  uaw <- apply(wx * cmxp, 1, sum)/rp
  paw <- apply(wx * cmxp, 2, sum)/cp
  names(uaw) <- names(paw)

  ## compute unweighted tau
  ##
  th1 <- sum(diag(cmxp))
  th2 <- sum(pv %*% (csum/n))
  tau <- (th1-th2)/(1-th2);

  ## compute weighted tau
  ##
  thw1 <- sum(wx * cmxp)
  thw2 <- sum(wx * (pv %o% (csum/n)))
  tau.w <- (thw1-thw2)/(1-thw2);


  ## return list of results
  return(list(crossclass = cmx, weights=wx,
              obs=rsum, ref=csum, n=n,
              overall.naive=oa,
              user.naive=ua, prod.naive=pa,
              overall.weighted=oaw,
              user.weighted=uaw, prod.weighted=paw,
              tau.priors=pv,
              tau=tau,
              tau.w=tau.w))
}

#' @param result.tau `tauW()` result
#'
#' @export
#' @rdname tauW
summaryTauW <- function(result.tau) {
  print("Cross-classification matrix:", quote=F)
  print(result.tau$crossclass)
  print(paste("Number of observations:", result.tau$n), quote=F)
  print("Weights:", quote=F)
  print(result.tau$weights, quote=F)

  print(paste("Overall accuracy (unweighted):",
              round(result.tau$overall.naive,4)), quote=F)
  print(paste("Overall accuracy (weighted):",
              round(result.tau$overall.weighted,4)), quote=F)

  print("User's accuracy (unweighted):", quote=F)
  print(round(result.tau$user.naive,4))
  print("User's accuracy (weighted):", quote=F)
  print(round(result.tau$user.weighted,4))
  print("Producer's reliability (unweighted):", quote=F)
  print(round(result.tau$prod.naive,4))
  print("Producer's reliability (weighted):", quote=F)
  print(round(result.tau$prod.weighted,4))

  print("Reference class proportions:", quote=F)
  print(round(result.tau$ref/result.tau$n,4), quote=F)
  print("Observed class proportions:", quote=F)
  print(round(result.tau$obs/result.tau$n,4), quote=F)
  print("Prior class probabilities:", quote=F)
  print(result.tau$tau.priors, quote=F)
  print(paste("Tau (unweighted):", round(result.tau$tau,4)), quote=F)
  print(paste("Tau (weighted):", round(result.tau$tau.w,4)), quote=F)
}

#' Format a LaTeX table with results
#'
#' @param result.tau results returned by \code{tauW}
#' @param file.name name of file to write output TeX file; Default: `file.name="tau_results_table.tex"`
#'
#' @return NULL
#' @export
xtableTauW <- function(result.tau, file.name="tau_results_table.tex") {

  # safely check for required packages
  if(!requireNamespace('xtable', quietly = TRUE))
    stop('this function requires the `xtable` package.', call.=FALSE)

  options(xtable.floating = FALSE)
  options(xtable.timestamp = "")
  tab <- data.frame(result.tau$user.naive, result.tau$user.weighted,
                    result.tau$prod.naive, result.tau$prod.weighted)
  names(tab) <- c("UA", "UA(W)", "PA", "PA(W)")
  x <- xtable::xtable(tab)
  xtable::autoformat(x)
  sink(file=file.name)
  cat("\\begin{tabular}{rl}\n")
  cat(paste("Unweighted accuracy $A_o$: &", sprintf("%1.4f", result.tau$overall.naive), "\\\\"))
  cat(paste("Weighted accuracy ${A_o}_w$: &", sprintf("%1.4f", result.tau$overall.weighted), "\\\\"))
  cat(paste("Unweighted \\emph{tau} $\\tau$: &", sprintf("%1.4f", result.tau$tau), "\\\\"))
  cat(paste("Weighted \\emph{tau} $\\tau_w$: &", sprintf("%1.4f", result.tau$tauW), "\\\\"))
  cat("\\end{tabular}\n")
  cat("\\par\n")
  print(x)
  sink(file=NULL)
}

setGeneric("glom", function(p, z1, z2 = NULL,
                            ids = FALSE, df = FALSE,
                            truncate = FALSE, invert = FALSE, fill = FALSE,
                            modality = "all", drop = !fill, ...)
  standardGeneric("glom"))

#' Subset soil horizon data using a depth or depth interval
#' @param p A SoilProfileCollection
#' @param x A SoilProfileCollection
#' @param z1 numeric vector of top depth to intersect horizon (required). Can be an expression involving `siteNames(p)` or quoted column name. Should evaluate to numeric length `1` or length equal to `length(p)`
#' @param z2 numeric vector bottom depth of intersection interval (optional). Can also be an expression involving `siteNames(p)` or quoted column name. Should evaluate to numeric length `1`, length equal to `length(p)` or `NULL`. Default: `NULL` is "point" intersection 
#' @param ids return only horizon IDs? default: `FALSE`
#' @param df return a data.frame, by intersection with `horizons(p)`? default: `FALSE`
#' @param truncate truncate horizon top and bottom depths to `z1` and \code{z2}? default: `FALSE`
#' @param invert get horizons _outside_ the interval `[z1,z2]`? default: `FALSE`
#' @param fill keep sites and preserve order for profiles that do not have horizons in interval by filling with a single horizon with `NA` top and bottom depth. default: `FALSE`
#' @param modality default: `"all"` return all horizons; or `modality = "thickest"`) to return the _thickest_ horizon in interval. If multiple horizons have equal thickness, the first (shallowest) is returned.
#' @param drop Inverted alias of fill for consistency with other methods. When `drop=FALSE`, filling occurs.
#' @param ... Additional arguments (not used)
#'
#' @description Make a "clod" of horizons from a SoilProfileCollection given a point or a depth interval to intersect. The interval `[z1,z2]` may be profile-specific (equal in length to `p`), or may be recycled over all profiles (if boundaries are length 1). For "point" intersection, `z2` may be left as the default value `NULL`. 
#'
#' @details "To glom" is "to steal" or to "become stuck or attached to". The word is related to the compound "glomalin", which is a glycoprotein produced by mycorrhizal fungi in soil.
#' 
#' The full depth range of horizons included within the interval are returned (a "ragged" SoilProfileCollection) unless the `truncate` argument is set as `TRUE`. Horizon intersection is based on unique ID \code{hzidname(spc)} and depth range of interest. Profiles that lack data in the range of interest will be dropped from the resulting SoilProfileCollection. 
#'
#' If inverting results with `invert`, it is possible that thick horizons (whose boundaries span wider than the specified interval) will be split into _two_ horizons, where previously they were one. This may make the results from `ids = TRUE` different from what you expect, as they will be based on a profile with an "extra" horizon and re-calculated unique horizon ID (`hzidname(spc)`) `"hzID"`.
#'
#' @seealso \code{\link{glomApply}} \code{\link{trunc}}
#'
#' @author Andrew G. Brown
#'
#' @return a SoilProfileCollection, data.frame, or a vector of horizon IDs. \code{NULL} if no result.
#'
#' @export glom
#' @aliases glom
#' @rdname glom
#' @examples
#' 
#' # keep examples from using more than 2 cores
#' data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))
#' 
#' data(sp1, package = 'aqp')
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' p <- glom(sp1, 25, 150)
#'
#' # 28 horizons
#' nrow(p) 
#' 
#' # inspect graphically
#' par(mar = c(1,1,3,1))
#' plot(p, color = "prop", max.depth = 200)
#' abline(h = c(25, 100), lty = 2)
#' 
#' ## glom(..., truncate = TRUE)
#' 
#' p2 <- glom(sp1, 25, 150, truncate = TRUE)
#'
#' # 28 horizons
#' nrow(p2) 
#' 
#' # inspect graphically
#' par(mar = c(1,1,3,1))
#' plot(p2, color = "prop", max.depth = 200)
#' abline(h = c(25, 100), lty = 2)
#' 
#' ## glom(..., truncate = TRUE, invert = TRUE)
#' 
#' p3 <- glom(sp1, 25, 150, truncate = TRUE, invert = TRUE)
#'
#' # 45 horizons
#' nrow(p3) 
#' 
#' # inspect graphically
#' par(mar = c(1,1,3,1))
#' plot(p3, color = "prop", max.depth = 200)
#' abline(h = c(25, 100), lty = 2)
#' 
#' ## profile-specific interval, using expressions evaluated within sp1@site
#' 
#' # calculate some new site-level variables containing target interval
#' sp1$glom_top <- (1:9) * 10
#' sp1$glom_bottom <- 10 + sp1$glom_top
#' 
#' # glom evaluates non-standard expressions using siteNames(sp1) column names
#' p4 <- glom(sp1, glom_top / 2, glom_bottom * 1.2, truncate = TRUE)
#' 
#' # inspect graphically
#' par(mar = c(1,1,3,1))
#' plot(p4, color = "prop", max.depth = 200)
#' 
setMethod("glom", signature(p = "SoilProfileCollection"),
          function(p,
                   z1,
                   z2 = NULL,
                   ids = FALSE,
                   df = FALSE,
                   truncate = FALSE,
                   invert = FALSE,
                   fill = FALSE,
                   modality = "all",
                   drop = !fill, 
                   ...) {
            
  
  # handle unquoted symbols from @site column names (or expressions using them)
  zz <- eval(substitute(list(z1, z2)),
             envir = site(p), 
             enclos = parent.frame())
  
  z1 <- zz[[1]]
  z2 <- zz[[2]]
  
  # handle _quoted_ character column names from @site for z1/z2
  if (is.character(z1) && length(z1) == 1) {
    if (z1 %in% siteNames(p)) {
      sitez1 <- p[[z1]]
      if (!is.numeric(sitez1))
        stop("If `z1` is a column name, must refer to a numeric column in `siteNames(p)`")
      z1 <- sitez1
    }
  }
  if (is.character(z2) && length(z2) == 1) {
    if (z2 %in% siteNames(p)) {
      sitez2 <- p[[z2]]
      if (!is.numeric(sitez2))
        stop("If `z2` is a column name, must refer to a numeric column in `siteNames(p)`")
      z2 <- sitez2
    }
  }
  
  depthn <- horizonDepths(p) 
  fillids <- profile_id(p)
  
  # recycle z1 to size of SPC if length is 1
  if (length(z1) == 1) {
    z1 <- rep(z1, length(p))
  
  # error if z1 does not match number of profiles
  } else if (length(z1) != length(p)) {
    stop("`z1` should have length 1 or length equal to number of profiles in `p`", call. = FALSE)
  }
  
  if (is.null(z2)) {
    z2 <- z1
  } else if (length(z2) == 1) {
    z2 <- rep(z2, length(p))
  }
  
  if (length(z2) != length(z1)) {
    stop("`z2`, if specified, should have same length as `z1`", call. = FALSE)
  }
  
  # handle bad z logic
  bad.idx <- which(z2 < z1)
  if (length(bad.idx) > 0) {
    warning("`z2` smaller than `z1`; setting top and bottom to `NA`")
    z1[bad.idx] <- NA
    z2[bad.idx] <- NA
  }
  
  if (!invert) {
    idx <- .multiglomDT(p, z1 = z1, z2 = z2, modality = modality)
  } else {
    idx <- c(.multiglomDT(p, min(p[[depthn[1]]], na.rm = TRUE), z1, modality = "all"),
             .multiglomDT(p, z2, max(p[[depthn[2]]], na.rm = TRUE), modality = "all"))
  }

  h <- horizons(p)[which(hzID(p) %in% idx), ]
  id <- idname(p)
  
  # truncate ragged edges
  if (!invert & truncate & !is.null(z2)) {
    
    pdepths <- data.table::data.table(id = as.character(h[[id]]),
                                      tdep = h[[depthn[1]]],
                                      bdep = h[[depthn[2]]])
    
    glomdepths <- data.table::data.table(id = profile_id(p),
                                         z1 = z1,
                                         z2 = z2)
    
    # join and remove horizons that are not part of interval
    newhz <- pdepths[glomdepths, on = "id"]
    newhz <- newhz[complete.cases(newhz),]
    
    # truncate upper bounds to z1
    z1tidx <- newhz[,.I[newhz$tdep < z1]]
    z1bidx <- newhz[,.I[newhz$bdep < z1]]
    newhz[z1tidx, 'tdep'] <- newhz[z1tidx, 'z1']
    newhz[z1bidx, 'bdep'] <- newhz[z1bidx, 'z1']
    
    # truncate lower bounds to z2
    z2tidx <- newhz[,.I[newhz$tdep > z2]]
    z2bidx <- newhz[,.I[newhz$bdep > z2]]
    newhz[z2tidx, 'tdep'] <- newhz[z2tidx, 'z2']
    newhz[z2bidx, 'bdep'] <- newhz[z2bidx, 'z2']

    # replace original values with truncated 
    h[[depthn[1]]][!is.na(h[[depthn[1]]])] <- newhz$tdep[!is.na(newhz$tdep)]
    h[[depthn[2]]][!is.na(h[[depthn[2]]])] <- newhz$bdep[!is.na(newhz$bdep)]

  # or "invert" and truncate
  } else if (invert & truncate & !is.null(z2)) {
    p0 <- p

    .top <- h[[depthn[1]]]
    .bottom <- h[[depthn[2]]]

    # create an upper and lower SPC
    z0 <- rep(0, length(z1))
    z0[is.na(z1)] <- NA
    p1 <- glom(p, z0, z1, truncate = TRUE)
    p2 <- glom(p, z2, max(p[[depthn[2]]]), truncate = TRUE)

    # combine horizon data
    if (inherits(p1, 'SoilProfileCollection') && 
        inherits(p2, 'SoilProfileCollection')) {
      
      if (nrow(p1) > 0) {
        p <- p1
        if (inherits(p2, 'SoilProfileCollection') && nrow(p2) > 0) {
          h <- rbind(horizons(p1), horizons(p2))
        } else {
          h <- horizons(p1)
        }
      } 
    } else {
      h <- horizons(p2)
    }

    if (modality == "thickest") {
      first.thickest.idx <- .thickestHzID(p, h)
      h <- h[first.thickest.idx,]
    }
    
    # enforce ID+top depth order after combining upper and lower SPCs
    h <- h[order(h[[idname(p)]], h[[depthn[1]]]), ]          
    
    # replace horizon IDs because we split horizons possibly
    h$hzID <- as.character(seq_len(nrow(h)))
    
    # force autocalculated ID
    hzidname(p) <- "hzID"
    
    if (fill) {
      p <- p0
    }
  }
  
  if (!missing(drop)) {
    fill <- !drop
  }
  
  # update slots for new h
  if (!fill) {
    p <- .updateSite(p, h)
  }
  
  if (fill) {
    p <- .insert_dropped_horizons(p, horizons = h)
  } else {
    if (nrow(h) > 0) {
      # replace @horizons with h
      replaceHorizons(p) <- h
      
      # if pre-existing "filled" horizons are present, drop them
      p <- p[!isEmpty(p),]
    } else {
      p <- p[0, ]
    }
  }
    
  # short circuit to get hzIDs of result
  if (ids) {
    if (invert)
      warning("invert = TRUE may split horizons, which affects resulting ids when ids=TRUE!",
              call. = FALSE)
    return(idx)
  }

  if (!df) {
    # verify updating of non-horizons slots worked
    if (!spc_in_sync(p)$valid) {
      warning("invalid SPC created by glom result!", call. = FALSE)
    }
    # return SPC
    return(p)
  } else {
    # return data.frame
    return(h)
  }
})

# helper functions for creating new horizon and paired SPC data
.thickestHzID <- function(p, newhz = horizons(p)) {
  
  # appease R CMD check
  .thk <- NULL
  
  htb <- horizonDepths(p)
  # make a data.table with id and thickness (.thk)
  res <- data.table::data.table(id = newhz[[idname(p)]], .thk = newhz[[htb[2]]] - newhz[[htb[1]]])
  # get the row index where the thickness is maximized (by profile ID)
  res[, .I[.thk >= suppressWarnings(max(.thk))], by = "id"]$V1
}

.updateSite <- function(p, newhz = horizons(p)) {
  # profile_id relies on p@horizons. we are testing new horizons data.frame
  keep <- unique(newhz[[idname(p)]])
  
  # keep only relevant site and other slots using SPC[i, ]
  p <- p[which(site(p)[[idname(p)]] %in% keep), ]
  return(p)
}

# new workhorse function replacing .glom and .multiglom
.multiglomDT <- function (p, z1, z2 = NULL, modality = "all", as.list = FALSE) {
  
  # access SPC slots to get important info about p
  hz <- data.table::as.data.table(horizons(p))
  dz <- horizonDepths(p)
  id <- idname(p)
  pid <- profile_id(p)
  hzid <- hzidname(p)
  
  # get top and bottom horizon depth vectors
  tdep <- hz[[dz[1]]]
  bdep <- hz[[dz[2]]]
  
  pdepths <- data.table::data.table(id = as.character(hz[[id]]),
                                    tdep = tdep,
                                    bdep = bdep)
  
  glomdepths <- data.table::data.table(id = profile_id(p),
                                       z1 = z1,
                                       z2 = z2)
  
  if (is.null(z2))
    z1 <- z2
  
  idx <- pdepths[glomdepths, on = "id"][,.I[((tdep >= z1) | 
                                              (bdep > z1)) & 
                                              ((bdep < z2) |
                                              (bdep == z2) | 
                                              (tdep < z2))]]
  if (modality == "thickest") {
    # get index of thickest horizon by profile
    idx <- .thickestHzID(p, hz[idx, ])
  }
  
  # get the ID values out of horizon table
  idval <- hz[idx, ][[hzidname(p)]]
  
  # just the horizon IDs
  if (!as.list) {
    return(idval)
  }
  
  # # list result.
  return(list(
    hzid = hzid,
    hz.idx = idx,
    value = idval
  ))
}


#' @description `trunc()` is a wrapper method (using S4 generic) for `glom()` where `truncate=TRUE`
#' @param ... `trunc()`: additional arguments passed to `glom()`
#' @export 
#' @rdname glom
#' @examples
#'
#' # load sample data
#' data("sp3")
#'
#' # promote to SPC
#' depths(sp3) <- id ~ top + bottom
#'
#' ### TRUNCATE all profiles in sp3 to [0,25]
#'
#' # set up plot parameters
#' par(mfrow=c(2,1), mar=c(0,0,0,0))
#'
#' # full profiles
#' plot(sp3)
#'
#' # trunc'd profiles
#' plot(trunc(sp3, 0, 25))
setMethod("trunc", signature(x = "SoilProfileCollection"),
          function(x, z1, z2, ...) {
            return(glom(x, z1, z2, truncate = TRUE, ...))
          })

## TODO: consider use of alternative distance metrics that DO NOT rescale always to 0,1
## TODO: interpret various forms of soil color via .detectColorSpec()

## Note: sanity checking on w is performed outside of this function


#' @title Internal wrapper for distance calculations by NCSP
#'
#' @param m `data.frame`, rows are individuals columns are characteristics, IDs should be saved in rownames
#' 
#' @param sm vector, a single row from the soil/non-soil matrix, must be equal to `nrow(m)`
#' 
#' @param w numeric vector of length `ncol(m)`, optionally specified for weighted distance calc
#' 
#' @param isColor logical, `m` contains CIELAB color coordinates, CIE2000 color contrast metric is used
#' 
#' @note Sanity checking on inputs is performed outside of this function.
#'
#' @return `dissimilarity`, `dist` object
#' 
#' @keywords internal
#' @noRd
#'
.NCSP_distanceCalc <- function(m, sm, w = NULL, isColor) {
  
  ## TODO: enable other metrics
  
  # maximum distance used to replace soil + non-soil distances
  # set to 1 for metric = 'gower'
  d.max <- 1
  
  ## TODO: consider adding as an argument
  # CIE2000 will requires an alternative max distance
  # use something reasonable, as this function is applied slice-wise
  # ~ Munsell value 1 -> 8 : dE00 ~ 70
  if(isColor) {
    d.max <- 70
  }
  
  # distance used to replace (NA) non-soil + non-soil distances
  d.notsoil <- 0
  
  # distance used to replace NA (NA in m, not related to non-soil evaluation)
  d.NA <- 0
  
  # number of individuals
  n <- ncol(m)
  
  if(isColor) {
    # CIE2000 color contrast
    # weights are ignored
    
    ## TODO: verify result when NA are present in data matrix
    # result is the full form of a distance matrix
    # requesting the full matrix form is slower
    d <- farver::compare_colour(
      from = m, 
      to = m, 
      from_space = 'lab', 
      to_space = 'lab',
      method = 'CIE2000', 
      white_from = 'D65'
    )
    
  } else {
    # Gower distances, these are always within 0-1
    
    # suppressing warnings issued when <2 unique values causes
    # WARNING: binary variable(s) 1, 2 treated as interval scaled
    if(!is.null(w)) {
      # weighted distances
      d <- suppressWarnings(cluster::daisy(m, metric = 'gower', weights = w))
    } else {
      # standard, un-weighted distances
      d <- suppressWarnings(cluster::daisy(m, metric = 'gower'))
    }
    
    # convert to full matrix for manipulation by row/col index
    d <- as.matrix(d)
  }
  
  
  # locate soil / non-soil pairs
  # TRUE = soil | FALSE = non-soil
  idx.notsoil <- which(!sm)
  
  # of soil / non-soil pairs exist, process accordingly
  if(length(idx.notsoil) > 0) {
    
    ## create logical matrix of soil + non-soil co-occurrences
    ## --> set to MAX distance
    #
    # soil XOR non-soil -> TRUE
    # soil XOR soil -> FALSE
    # non-soil XOR non-soil -> FALSE
    sm.mat <- outer(sm, sm, FUN = xor)
    d[which(sm.mat)] <- d.max
    
    
    ## create logical matrix of non-soil + non-soil co-occurrences
    ## --> set to d.notsoil
    #
    # !soil AND !non-soil -> FALSE
    # !soil AND !soil -> FALSE
    # !non-soil AND !non-soil -> FALSE
    sm.mat <- outer(!sm, !sm, FUN = '&')
    d[which(sm.mat)] <- d.notsoil
    
    ## any remaining NA are related to missing data in m
    ## --> set to distance given to missing data
    d[which(is.na(d))] <- d.NA
    
    # reset diagonal
    diag(d) <- 0
  }
  
  
  ## replace remaining NA (missing data in m) with min distance
  idx.NA <- which(is.na(d))
  d[idx.NA] <- d.NA
  
  # back to reduced format as dist object 
  d <- as.dist(d)
  
  return(d)
}



#' @title Numerical Classification of Soil Profiles
#' @description Replacement for `profile_compare()`.
#' 
#' Performs a numerical comparison of soil profiles using named properties,
#' based on a weighted, summed, depth-segment-aligned dissimilarity calculation.
#'
#' Variability in soil depth can interfere significantly with the calculation
#' of between-profile dissimilarity--what is the numerical ``distance'' (or
#' dissimilarity) between a slice of soil from profile A and the corresponding,
#' but missing, slice from a shallower profile B? Gower's distance metric would
#' yield a NULL distance, despite the fact that intuition suggests otherwise:
#' shallower soils should be more dissimilar from deeper soils. For example,
#' when a 25 cm deep profile is compared with a 50 cm deep profile, numerical
#' distances are only accumulated for the first 25 cm of soil (distances from
#' 26 - 50 cm are NULL). When summed, the total distance between these profiles
#' will generally be less than the distance between two profiles of equal
#' depth. Our algorithm will replace NULL
#' distances with the maximum distance between any pair of profiles for the
#' current depth slice. In this way, the numerical distance between a slice of
#' soil and a corresponding slice of non-soil reflects the fact that these two
#' materials should be treated very differently.
#'
#' This alternative calculation of dissimilarities between soil and non-soil
#' slices solves the problem of comparing shallow profiles with deeper
#' profiles. However, it can result in a new problem: distances calculated
#' between two shallow profiles will be erroneously inflated beyond the extent
#' of either profile's depth. Our algorithm will preserve NULL distances between slices when
#' both slices represent non-soil material. Therefore, shallow
#' profiles will only accumulate mutual dissimilarity to the depth of the
#' deeper profile.
#'
#' Slices are classified as 'soil' down to the maximum depth to which at least one of
#' variables used in the dissimilarity calculation is not NA. This will cause
#' problems when profiles within a collection contain all NAs within the
#' columns used to determine dissimilarity. An approach for identifying and
#' removing these kind of profiles is presented in the examples section below.
#'
#' Our approach builds on the work of (Moore, 1972) and the previously
#' mentioned depth-slicing algorithm. See references below for a detailed explanation of the NCSP algorithm.
#' 
#' @note `NCSP()` will overwrite the `removed.profiles` metadata from `x`.
#' 
#' @param x `SoilProfileColection` object, should be pre-filtered to remove profiles with horizon depth logic, see [`HzDepthLogicSubset`]
#'
#' @param vars character vector, names of horizon attributes to use in the classification
#' 
#' @param fm formula, formula as specified to [`dice()`], not yet implemented
#' 
#' @param weights numeric vector, same length as `vars`: variable importance weights, need not sum to 1
#' 
#' @param maxDepth numeric, maximum depth of analysis
#' 
#' @param k numeric, weighting coefficient, usually between 0-1. A value of 0 results in no depth-weighting. See examples.
#' 
#' @param isColor, logical: variables represent color, should be CIELAB coordinates (D65 illuminant), weights are ignored. Variables should be named `L`, `A`, `B` in specified in that order.
#' 
#' @param rescaleResult logical, distance matrix is rescaled based on max(D)
#' 
#' 
#' @param verbose logical, extra output messages
#' 
#' @param returnDepthDistances logical, return a list of distances by depth slice
#'
#' @author Dylan E. Beaudette and Jon Maynard
#' 
#' @seealso [dice()], [cluster::daisy()], [compareSites()]
#' 
#' @references
#'  - J.J Maynard, S.W. Salley, D.E. Beaudette, J.E Herrick. Numerical soil classification supports soil identification by citizen scientists using limited, simple soil observations. Soil Sci. Soc. Am. J. 2020; 84: 1675-1692. \doi{10.1002/saj2.20119}.
#'  - D.E. Beaudette, P. Roudier, A.T. O'Geen, Algorithms for quantitative pedology: A toolkit for soil scientists, Computers & Geosciences, Volume 52, 2013, Pages 258-268, ISSN 0098-3004, \doi{10.1016/j.cageo.2012.10.020}.
#'  - Moore, A.; Russell, J. & Ward, W. Numerical analysis of soils: A comparison of three soil profile models with field classification. Journal of Soil Science, 1972, 23, 194-209.
#'  
#' @keywords methods manip
#'

#' @export

## TODO:
# * finish testing / comparing
# * example with multiple runs on suites of variables + weighted combination of D1, D2, D3, ...
# * list-output, with diagnostics and other interesting information
# * SS: more examples, vignettes -> interpretation made clear

## long-term, framework for
## SS: need some original research to make suggestions on weights, suites
#
# D = wt.mean(w1 * D1, w2 * D2, w3 * D3, ..., w_site * D_site)
# 
# D_site could be generated / related to adjacency matrix

## Next release:
# * expose full dice() fm argument for simple specification of depths + vars
# * allow pre-diced() or mpspline()-ed input
# * parallel operation -> furrr
# * progress bar for large SPCs -> purrr
# * benchmarking 


## Ideas:
# * see L1_profiles.R for ideas / examples related to selecting logical "bottom depths"
# * write new function for selecting a profile that minimizes distance to all others in SPC
# * NEON data for vignette + paper (SS and JM)

## name suggested by Jon Maynard
## ideas, commentary, updates c/o Maynard et al. 2020


NCSP <- function(
    x, 
    vars, 
    fm = NULL, 
    weights = rep(1, times = length(vars)), 
    maxDepth = max(x), 
    k = 0, 
    isColor = FALSE,
    rescaleResult = FALSE, 
    verbose = TRUE, 
    returnDepthDistances = FALSE
) {
  
  
  ## depreciated arguments
  
  # filter
  # replace_na
  # add_soil_flag
  # strict_hz_eval
  # plot.depth.matrix
  
  
  ## sanity checks
  
  # x must be a SPC
  if(! inherits(x, 'SoilProfileCollection')) {
    stop('`x` must be a SoilProfileCollection', call. = FALSE)
  }
  
  # check for site-level vars
  if(any(vars %in% siteNames(x))) {
    message('consider compareSites() for site level attributes')
    stop('`vars` may only contain horizon level attributes of `x`', call. = FALSE)
  }
  
  # vars must be in horizon names
  if(! all(vars %in% horizonNames(x))) {
    stop('`vars` must specify horizon level attributes of `x`', call. = FALSE)
  }
  
  ## TODO: consider a message and setting maxDepth <- max(x)
  # maxDepth
  if(maxDepth > max(x) | maxDepth < 1) {
    stop('`maxDepth` should be > 0 and <= max(`x`)', call. = FALSE)
  }
  
  # color comparisons require farver pkg for dE00
  if(isColor) {
    
    # for now, color comparisons are based on CIELAB color coordinates
    # must be named "L", "A", "B" and in that order
    if(any(vars != c('L', 'A', 'B'))) {
      stop('CIELAB color coordinates must be specified as `L`, `A`, `B` in `vars`', call. =FALSE)  
    }
  }
  
  
  
  
  ## TODO: reconsider hard-coded top depth
  ## truncate at maxDepth
  x <- trunc(x, 0, maxDepth)
  
  
  ## variables used in NCSP algorithm
  
  # number of variables
  n.vars <- length(vars)
  
  # compute a weighting vector based on k
  # indexed to distance matrix list
  w <- 1 * exp(-k * 1:maxDepth)
  
  ## split horizon vars
  hn <- horizonNames(x)
  h.vars <- intersect(hn, vars)
  
  
  ## TODO: expose LHS of formula to dice()
  ## NOTE: dice() LHS usually starts from 0, sliceSequence and soil.matrix are indexed from 1
  
  ## TODO: consider exposing depth logic subset options in arguments to NCSP
  ##       for now, entire profiles are subset
  
  
  ## dice according to depth sequence and vars
  # preserve SPC for access to all site data
  # it could be better to perform sanity checks on horizonation outside of this function
  .seq <- seq(from = 0, to = maxDepth - 1, by = 1)
  .fm <- as.formula(
    sprintf('c(%s) ~ %s',
            paste0(.seq, collapse = ','),
            paste0(h.vars, collapse = ' + '))
  )
  
  # reset removed.profiles metadata, it may have been set prior to calling NCSP()
  metadata(x)$removed.profiles <- NULL
  
  ## dice
  # pctMissing is used to develop soil/non-soil matrix
  # NOTE: profiles with overlapping horizons will be removed
  s <- suppressMessages(
    dice(x, fm = .fm, SPC = TRUE, fill = TRUE, byhz = FALSE, pctMissing = TRUE, strict = TRUE)
  )
  
  # number of profiles, accounting for subset via dice()
  n.profiles <- length(s)
  # profile IDs
  .ids <- profile_id(s)
  
  # keep track of removed profiles, due to hz logic errors
  .removed.profiles <- metadata(s)$removed.profiles
  if(length(nchar(.removed.profiles)) > 0) {
    warning('hz depth logic subset has removed some profiles')
  }
  
  ## slice sequence
  sliceSequence <- 1:max(s)
  
  ## add soil flag, previously optional now required
  # using same data structure / conventions as profile_compare()
  # logical matrix [sliceSequence, 1:n.profiles]
  #                       [slices, profiles]
  
  ## TODO: ensure that in-line NA are correctly handled
  ## TODO: in-line NA may require another argument to determine assumptions
  ## TODO: define soil / non-soil using pattern matching on horizon designation
  
  # use a vector of horizon (slice) indices to extract all pctMissing values
  # develop a matrix from these
  soil.matrix <- matrix(
    s[, sliceSequence][['.pctMissing']], 
    ncol = n.profiles, 
    nrow = length(sliceSequence), 
    byrow = FALSE
  )
  
  # "soil" has a pctMissing very close to 0
  soil.matrix <- soil.matrix < 0.00001
  # keep track of profile IDs
  dimnames(soil.matrix)[[2]] <- .ids
  
  
  ## evaluate distances by slice
  ## accounting for soil/non-soil comparisons
  ## filling NA due to missing data
  .d <- list()
  
  # distance matrix cache
  dCache <- list()
  
  ## TODO:
  ##  * more efficient data reshaping without calling horizons() in each iteration
  ##  * convert this to parallel evaluation, maybe furrr package
  ##  * if !returnDepthDistances: do not retain full list of dist mat, accumulate in single variable
  
  message(paste('Computing dissimilarity matrices from', n.profiles, 'profiles'), appendLF = FALSE)
  for(i in sliceSequence) {
    
    # horizon data for slice i
    .s <- horizons(s[, i])
    
    # characteristics for slice i
    # NOTE: .data.frame.j() will reset rownames
    .s <- .data.frame.j(.s, vars)
    
    # hash is applied to entire slice of data AND soil matrix slice
    # NOTE: row.names have been reset (those are included in the hash)
    .hash <- digest(
      list(
        .s, soil.matrix[i, ]
      )
    )
    
    # preserve IDs in distance matrix
    row.names(.s) <- .ids
    
    # test cache for equivalent distance matrix via hash
    if(is.element(.hash, names(dCache))) {
      # copy from cache
      .d[[i]] <- dCache[[.hash]]
      
    } else {
      # compute distance, with rules related to soil/non-soil matrix
      .d[[i]] <- .NCSP_distanceCalc(m = .s, sm = soil.matrix[i, ], w = weights, isColor = isColor)
      
      # update cache
      dCache[[.hash]] <- .d[[i]]
    }
    
    # apply depth-weighting
    .d[[i]] <- .d[[i]] * w[i]
  }
  
  ## optionally return list of distance matrices
  if(returnDepthDistances) {
    return(.d)
  }
  
  # print total size of D
  message(paste(" [", signif(object.size(.d) / 1024^2, 1), " Mb]", sep = ''))
  
  .msg <- sprintf("cache: %s | slices: %s", length(dCache), length(sliceSequence))
  message(.msg)
  
  
  ## flatten list of distance matrices
  .d <- Reduce('+', .d)
  
  
  ## optionally normalize by dividing by max(D)
  # this is important when incorporating site data
  # TODO: causes problems for some functions like MASS::sammon()
  # TODO: consider safer re-scaling, which removes possibility of 0-distance
  if(rescaleResult) {
    .d <- .d / max(.d, na.rm = TRUE)
  }
  
  
  ## metadata
  
  # distance metric
  if(isColor) {
    attr(.d, 'Metric') <- 'CIE2000'
  } else {
    attr(.d, 'Metric') <- 'Gower'
  }
  
  # cache ratio, possibly related to SPC complexity as a function of vars and soil depth
  attr(.d, 'cache factor') <- length(sliceSequence) / length(dCache) 
  
  # removed profiles, if any
  attr(.d, 'removed.profiles') <- .removed.profiles
  
  # remove warnings about NA from cluster::daisy()
  attr(.d, 'NA.message') <- NULL
  
  # remove 'call' attribute
  attr(.d, 'call') <- NULL
  
  # full set of profile IDs are stored in attr(.d, 'Labels')
  
  # done
  return(.d)
  
}

context("NCSP algorithm")

data("jacobs2000")
x <- jacobs2000

test_that('error conditions', {

  # x must be a SPC
  expect_error(
    NCSP(3L, vars = c('sand', 'clay'))
  )

  # vars must exist in site / horizon
  expect_error(
    NCSP(x, vars = c('sand', 'tacos'))
  )

  # depthSequence must be reasonable
  expect_error(
    NCSP(x, vars = c('sand', 'clay'), maxDepth = -5)
  )

})


# setup horizon-level data: data are from lab sampled pedons
d <- read.csv(
  textConnection('series,top,bottom,clay,frags,ph
auburn,0,3,21,6,5.6
auburn,3,15,21,13,5.6
auburn,15,25,20,9,5.8
auburn,25,47,21,28,5.8
dunstone,0,5,16,13,6
dunstone,5,17,17,19,6.3
dunstone,17,31,20,6,6.3
dunstone,31,41,21,15,6.3
sobrante,0,5,18,0,5.8
sobrante,5,10,16,2,5.7
sobrante,10,28,15,21,5.8
sobrante,28,51,18,13,6.2
sobrante,51,74,20,12,6.2')
)

# establish site-level data
s <- data.frame(
  series = c('auburn', 'dunstone', 'sobrante'), 
  precip = c(24, 30, 32)
)

# generate fake horizon names with clay / frags / ph
d$name <- with(d, paste(clay, frags, ph, sep='/'))

# upgrade to SoilProfile Collection object
depths(d) <- series ~ top + bottom
site(d) <- s
hzdesgnname(d) <- 'name'

## tests

test_that("NCSP works as expected", {
  
  # compute between-profile dissimilarity, no depth weighting
  d.dis <- suppressMessages(NCSP(d, vars = c('clay', 'ph', 'frags'), k = 0, maxDepth = 61))
  m <- as.matrix(d.dis)
  
  # results should be 3x3 distance matrix
  # this is a cluster package object
  expect_true(inherits(d.dis, c('dissimilarity', 'dist')))
  expect_equal(dim(m), c(3, 3))
  
  ## TODO: manually check these
  # known output
  expect_equivalent(diag(m), c(0, 0, 0))
  expect_equal(m[1,2], 33.03225, tolerance=0.00001)
  expect_equal(m[1,3], 49.02092, tolerance=0.00001)
  expect_equal(m[2,3], 45.94683, tolerance=0.00001)
})



test_that(".NCSP_distanceCalc() with color data", {
  
  # color for 3 profiles
  m <- c('10YR 2/2', '5YR 6/8', '5PB 8/2')
  
  # # manually check dE00
  # mm <- t(combn(m, 2))
  # colorContrast(mm[, 1], mm[, 2])
  
  # convert to CIELAB
  .lab <- parseMunsell(m, returnLAB = TRUE)
  
  # profile IDs
  row.names(.lab) <- letters[1:nrow(.lab)]
  
  # single soil matrix record
  sm <- rep(TRUE, times = nrow(.lab))
  
  # CIE2000 color contrast
  d <- .NCSP_distanceCalc(.lab, sm = sm, isColor = TRUE)
  
  # results should be 3x3 distance matrix
  # not created by cluster package, plain old dist object
  expect_true(inherits(d, 'dist'))
  
  # convert to full matrix representation for eval
  d <- as.matrix(d)
  expect_equal(dim(d), c(3, 3))
  
  # known output
  expect_equivalent(diag(d), c(0, 0, 0))
  expect_equal(d[1,2], 40.458, tolerance = 0.001)
  expect_equal(d[1,3], 63.330, tolerance = 0.001)
  expect_equal(d[2,3], 32.019, tolerance = 0.001)
})




test_that("NCSP() with genhz encoded as factor", {
  
  # simple SPC with deterministic structure
  template <- c(
    'A:AAAA|BBBBBBBB|CCCCC|RRRRRRRRRRR',
    'B:AAAAA|BBBBBBBBBBBBBB|CCCCC|RRRRRR',
    'C:AAAAAAA|BBBBBBBB|CCC|RRRRRRRRRRR',
    'D:A|BBBBBBBB|CCCCC|RRRRRRRRRRRRR',
    'E:AA|BBBBBBBB|CCCCC|RRRRRRRRRRRRR',
    'F:AAAAA|BBB|C|RRRRRRRRRRRRRRRR'
  )
  
  # init SPC and limit depth
  x <- quickSPC(template)
  x <- trunc(x, 0, 250)
  
  # init generalized hz labels
  x$genhz <- factor(x$name, levels = c('A', 'B', 'C', 'R'))
  
  # plotSPC(x, name.style = 'center-center', color = 'genhz', cex.names = 0.8)
  
  d <- NCSP(x, vars = 'genhz')
  
  # par(mar = c(1, 0, 3, 0))
  # plotProfileDendrogram(x, diana(d), name.style = 'center-center', color = 'genhz', cex.names = 0.8, scaling.factor = 1.1, y.offset = 20, width = 0.3, max.depth = 245, hz.depths = TRUE, depth.axis = FALSE)
  
  # should be a dist object
  expect_true(inherits(d, 'dist'))
  
  # check results via digest
  expect_equal(digest::digest(d), 'd603e066ad6790ba53b3f03eecb3800e')
  
  # check cache factor: should be close to 15
  expect_true(attr(d, 'cache factor') - 15 < 1)
  
})

## this will replace soilDB::estimateColorMixture() as an alternative / fallback 

# method for mixMunsell() when reference spectra are missing
# NA not allowed in chips, w
.estimateColorMixture <- function(chips, w) {
  
  # convert to CIELAB
  .lab <- parseMunsell(chips, returnLAB = TRUE)
  
  # weighted mean
  .L <- weighted.mean(.lab[['L']], w = w, na.rm = TRUE)
  .A <- weighted.mean(.lab[['A']], w = w, na.rm = TRUE)
  .B <- weighted.mean(.lab[['B']], w = w, na.rm = TRUE)
  
  # back to Munsell
  m <- col2Munsell(cbind(.L, .A, .B), space = 'CIELAB')
  
  # pack into expected structure
  # scaled distance is only for spectral distance evaluated against the entire library
  #  -> only relevant when combining with 'reference' methods mixing
  res <- data.frame(
    munsell = sprintf('%s %s/%s', m$hue, m$value, m$chroma),
    distance = m$sigma,
    scaledDistance = NA,
    distanceMetric = 'dE00',
    mixingMethod = 'estimate',
    stringsAsFactors = FALSE
  )
  
  return(res)
}



# helper function for printing out value / chroma ranges by hue
.summarizeMunsellSpectraRanges <- function() {
  
  # make R CMD CHECK happy
  munsell.spectra <- NULL
  
  # note: this is incompatible with LazyData: true
  # load look-up table from our package
  load(system.file("data/munsell.spectra.rda", package="aqp")[1])
  
  # set hue position
  munsell.spectra$hue <- factor(munsell.spectra$hue, levels = huePosition(returnHues = TRUE))
  
  # remove non-standard hues (what are they for?)
  munsell.spectra <- na.omit(munsell.spectra)
  
  x <- split(munsell.spectra, munsell.spectra$hue)
  
  x <- lapply(x, function(i) {
    
    data.frame(
      hue = i$hue[1],
      value = sprintf("%s-%s", min(i$value), max(i$value)),
      chroma = sprintf("%s-%s", min(i$chroma), max(i$chroma)),
      stringsAsFactors = FALSE
    )
    
  })
  
  x <- do.call('rbind', x)
  
  return(x)
}


# alternative formulation:
# http://scottburns.us/wp-content/uploads/2015/05/Subtractive-Color-Mixture-Computation.pdf
# 1:1 mixture -> a^(1/2) * b^(1/2)
# 2:1 mixture -> a^(2/3) * b^(1/3)

# weighted geometric mean
# https://en.wikipedia.org/wiki/Weighted_geometric_mean
# note: function will fail if any(v) == 0
# note: NA not allowed in v or w
.wgm <- function(v, w) {
  r <- sum(w * log(v)) / sum(w)
  r <- exp(r)
  return(r)
}






# another possible approach using only sRGB coordinates
# http://scottburns.us/wp-content/uploads/2015/04/ILSS.txt

# related ticket
# https://github.com/ncss-tech/aqp/issues/101

# in response to the commentary here:
# https://soiltxnmyforum.cals.vt.edu/forum/read.php?3,1984,1987#msg-1987

# inspiration / calculations based on:
# https://arxiv.org/ftp/arxiv/papers/1710/1710.06364.pdf

# related discussion here:
# https://stackoverflow.com/questions/10254022/implementing-kubelka-munk-like-krita-to-mix-colours-color-like-paint/29967630#29967630

# base spectral library:
# http://www.munsellcolourscienceforpainters.com/MunsellResources/SpectralReflectancesOf2007MunsellBookOfColorGlossy.txt

# see /misc/util/Munsell for:
# * spectral library prep
# * interpolation of odd chroma
# * reshaping for rapid look-up


#'
#' @title Mix Munsell Colors via Spectral Library
#'
#' @description Simulate mixing of colors in Munsell notation, similar to the way in which mixtures of pigments operate.
#'
#' @param x vector of colors in Munsell notation
#'
#' @param w vector of proportions, can sum to any number
#' 
#' @param mixingMethod approach used to simulate a mixture: 
#'    
#'    * `exact`: simulate a subtractive mixture of pigments, color conversion via CIE1931 color-matching functions (see details)
#'    
#'    * `reference`: simulate a subtractive mixture of pigments, selecting `n` closest reference spectra from [`munsell.spectra.wide`] (requires `gower` package)
#'    
#'    * `estimate`: closest Munsell chip to a weighted mean of CIELAB coordinates (fastest)
#'    
#'    * `adaptive`: use `exact` method when possible, falling-back to `estimate` (weighted mean of CIELAB coordinates) otherwise
#'
#' @param n number of closest matching color chips (`mixingMethod = reference` only)
#'
#' @param keepMixedSpec keep weighted geometric mean spectra, final result is a `list` (`mixingMethod = reference` only)
#' 
#' @param distThreshold spectral distance used to compute `scaledDistance`, default value is based on an analysis of spectral distances associated with adjacent Munsell color chips. This argument is only used with  `mixingMethod = 'reference'`.
#' 
#' @param ... additional arguments to [`spec2Munsell`]
#'
#' @author D.E. Beaudette
#'
#' @references
#'
#' Marcus, R.T. (1998). The Measurement of Color. In K. Nassau (Ed.), Color for Science, Art, and Technology (pp. 32-96). North-Holland.
#' 
#' 
#'  * http://scottburns.us/wp-content/uploads/2015/05/Subtractive-Color-Mixture-Computation.pdf
#'
#'  * [related discussion on Stack Overflow](https://stackoverflow.com/questions/10254022/implementing-kubelka-munk-like-krita-to-mix-colours-color-like-paint/29967630#29967630)
#'
#'  * spectral library source: http://www.munsellcolourscienceforpainters.com/MunsellResources/SpectralReflectancesOf2007MunsellBookOfColorGlossy.txt
#'
#'
#' @details
#' See the [expanded tutorial](https://ncss-tech.github.io/AQP/aqp/mix-colors.html) for examples.
#' 
#' An accurate simulation of pigment mixtures ("subtractive" color mixtures) is incredibly complex due to factors that aren't easily measured or controlled: pigment solubility, pigment particle size distribution, water content, substrate composition, and physical obstruction to name a few. That said, it is possible to simulate reasonable, subtractive color mixtures given a reference spectra library (350-800nm) and some assumptions about pigment qualities and lighting. For the purposes of estimating a mixture of soil colors (these are pigments after all) we can relax these assumptions and assume a standard light source. The only missing piece is the spectral library for all Munsell chips in our color books.
#' 
#' Paul Centore has provided a Munsell color chip reflectance spectra library (http://www.munsellcolourscienceforpainters.com). The estimation of a subtractive mixture of soil colors can proceed as follows:
#'   
#' 1. look up the associated spectra for each color in `x`
#' 2. compute the weighted (`w` argument) geometric mean of the spectra
#' 3. convert the spectral mixture to the closest Munsell color via:
#'   * search for the closest `n` matching spectra in the reference library (`mixtureMethod = 'reference'`)
#'   * direct conversion of spectra to closest Munsell color via [spec2Munsell()] (`mixtureMethod = 'exact'`)
#' 4. suggest resulting Munsell chip(s) as the best candidate for a simulated mixture
#' 
#' Key assumptions include:
#'   
#' * similar particle size distribution
#' * similar mineralogy (i.e. pigmentation qualities)
#' * similar water content. 
#' 
#' For the purposes of estimating (for example) a "mixed soil color within the top 18cm of soil" these assumptions are usually valid. Again, these are estimates that are ultimately "snapped" to the nearest chip and not do not need to approach the accuracy of paint-matching systems.
#' 
#' A message is printed when `scaledDistance` is larger than 1.
#'
#' @return A `data.frame` with the closest matching Munsell color(s):
#' 
#'  * `munsell`: Munsell notation of the n-closest spectra
#'  * `distance`: spectral (Gower) distance to the n-closest spectra
#'  * `scaledDistance`: spectral distance scaled by `distThreshold`
#'  * `mixingMethod`: method used for each mixture
#' 
#' When `keepMixedSpec = TRUE` then a `list`:
#' 
#'  * `mixed`: a `data.frame` containing the same elements as above
#'  * `spec`: spectra for the 1st closest match
#' 
#' 
#'
#' @seealso [munsell.spectra]
#' @export
#' 
mixMunsell <- function(x, w = rep(1, times = length(x)) / length(x), mixingMethod = c('exact', 'reference', 'estimate', 'adaptive'), n = 1, keepMixedSpec = FALSE, distThreshold = 0.025, ...) {
  
  # satisfy R CMD check
  munsell.spectra.wide <- NULL
  
  # enforce mixing method
  mixingMethod <- match.arg(mixingMethod)
  
  # multiple matches only possible when using mixingMethod == 'reference'
  if((n > 1) && mixingMethod != 'reference') {
    stop('`n` is only valid for `mixingMethod = "reference"`', call. = FALSE)
  }
  
  # mixed spectra and multiple matches only possible when using mixingMethod == 'reference'
  if(keepMixedSpec && !mixingMethod %in% c('reference', 'exact')) {
    stop('`keepMixedSpec` is only valid for mixingMethod = "reference" or "exact"', call. = FALSE)
  }
  
  # sanity check, need this for gower::gower_topn()
  if(mixingMethod == 'reference' && !requireNamespace('gower')) {
    stop("package `gower` is required for `mixingMethod='reference'`", call. = FALSE)
  }
  
  
  # can't mix a single color, just give it back at 0 distance
  if (length(unique(x)) == 1) {
    
    res <- data.frame(
      munsell =  x[1],
      distance = 0,
      scaledDistance = NA,
      distanceMetric = NA,
      mixingMethod = NA,
      stringsAsFactors = FALSE
    )
    
    return(res)
  }
  
  # must have as many weights as length of x
  if (length(x) != length(w) && length(w) != 1) {
    
    stop('w should have same length as x or length one')
    
  } else if (length(w) == 1) {
    
    # cannot mix with zero weights
    stopifnot(w > 0)
    
    # a recycled weight is same as function default
    w <- rep(1, times = length(x)) / length(x)
  }
  
  ## TODO: move 0-weight / NA handling up in the logic
  
  # more informative error for colors missing
  if (any(w[is.na(x)] > 0)) {
    stop('cannot mix missing (NA) colors with weight greater than zero')
  }
  
  # more informative error for weights missing
  if (any(is.na(w))) {
    stop('cannot mix colors with missing (NA) weight')
  }
  
  # remove 0-weighted colors
  x <- x[w > 0]
  w <- w[w > 0]
  
  # x with weights > 0 must contain valid Munsell
  if (any(is.na(parseMunsell(x)))) {
    stop('input must be valid Munsell notation, neutral hues and missing not supported')
  }
  
  
  ## main branch: mixing method
  
  # estimate via wtd.mean CIELAB
  if(mixingMethod == 'estimate') {
    
    # simple estimation by weighted mean CIELAB
    res <- .estimateColorMixture(chips = x, w = w)
    
    # stop here
    return(res)
    
  } else {
    # attempt spectral mixing
    
    # wide version for fast searches
    load(system.file("data/munsell.spectra.wide.rda", package="aqp")[1])
    
    # subset reference spectra for colors
    # note that search results are not in the same order as x
    # result are columns of spectra
    munsell.names <- names(munsell.spectra.wide)
    idx <- which(munsell.names %in% x)
    s <- munsell.spectra.wide[, idx, drop = FALSE]
    
    # sanity check: if there aren't sufficient reference spectra then return NA
    # must be at least the same number of spectra (columns) as unique colors specified
    if(ncol(s) < length(unique(x))){
      # helpful message
      missing.chips <- setdiff(x, munsell.names)
      msg <- sprintf(
        'reference spectra not available: %s',
        paste(missing.chips, collapse = ', ')
      )
      message(msg)
      
      # fall-back to wt. mean LAB
      if(mixingMethod == 'adaptive') {
        # assumes cleaned data
        res <- .estimateColorMixture(chips = x, w = w)
      } else {
        
        # otherwise return an empty result
        res <- data.frame(
          munsell = NA,
          distance = NA,
          scaledDistance = NA,
          distanceMetric = NA,
          mixingMethod = NA,
          stringsAsFactors = FALSE
        )
      }
      
      # done
      return(res)
    }
    
    ## proceeding to simulate spectral mixture
    
    # empty vector for mixture
    mixed <- vector(mode = 'numeric', length = nrow(s))
    
    # iterate over wavelength (columns in first spectra)
    for(i in seq_along(mixed)) {
      
      # prepare values:
      # select the i-th wavelength (row)
      # down-grade to a vector
      vals <- unlist(s[i, ])
      
      ## TODO: this wastes some time when weights are obvious, move outside of loop
      ##       -> convert to tapply, order doesn't matter as long as names are preserved
      
      # aggregate weights by "chip" -- in case length(x) != length(unique(x))
      wagg <- aggregate(w, by = list(chip = x), FUN = sum)
      
      # mix via weighted geometric mean
      mixed[i] <- .wgm(v = vals, w = wagg$x[match(names(s), wagg$chip)])
    }
    
    ## "exact" matching
    if(mixingMethod %in% c('exact', 'adaptive')) {
      
      ## 
      # S = R * illuminant (D65 is the default)
      # XYZ = AT %*% standard observer (CIE1964 is the default)
      # XYZ -> sRGB -> Munsell
      mx <- spec2Munsell(mixed, ...)
      
      # NOTE: ... are passed to col2munsell()
      #       convert = TRUE: mx is a data.frame
      #       convert = FALSE: mx is a matrix
      
      if(inherits(mx, 'matrix')) {
        # mx is a matrix
        dimnames(mx)[[2]] <- c('r', 'g', 'b')
        
        # include sRGB coordinates, this is different than what is typically returned by this function
        res <- data.frame(
          mx,
          munsell = NA,
          distance = NA,
          scaledDistance = NA,
          distanceMetric = NA,
          mixingMethod = 'exact',
          stringsAsFactors = FALSE
        )
        
      } else {
        # mx is a data.frame 
        res <- data.frame(
          munsell = sprintf("%s %s/%s", mx$hue, mx$value, mx$chroma),
          distance = mx$sigma,
          scaledDistance = NA,
          distanceMetric = 'dE00',
          mixingMethod = 'exact',
          stringsAsFactors = FALSE
        ) 
      }
      
      # final-cleanup and return is performed outside of if/else
      
    } else {
      ## reference "spectra" method
      ## requires suggested gower package
      
      ## operations on data.table likely faster
      
      # Gower distance: looks good, ~5x faster due to compiled code
      # https://cran.r-project.org/web/packages/gower/vignettes/intro.pdf
      # would make sense to reshape reference data
      
      # NOTE: arguments to col2munsell() are silently ignored
      
      ## TODO: time wasted here
      # reshape reference spectra: wavelength to columns
      z <- t(munsell.spectra.wide[, -1])
      
      # top n matches
      d <- gower::gower_topn(
        data.frame(rbind(mixed)),
        data.frame(z),
        n = n
      )
      
      res <- data.frame(
        munsell = dimnames(z)[[1]][d$index],
        distance = d$distance[, 1],
        scaledDistance = d$distance[, 1] / distThreshold,
        distanceMetric = 'Gower',
        mixingMethod = 'reference',
        stringsAsFactors = FALSE
      )
      
      # report possibly problematic mixtures
      if(any(res$scaledDistance > 1)) {
        message('closest match has a spectral distance that is large, results may be unreliable')
      }
      
    }
    
    ## all done with "exact", "spectra", and "adaptive"
    # clean-up row names
    row.names(res) <- NULL
    
    # optionally return weighted geometric mean (mixed) spectra
    if(keepMixedSpec) {
      return(
        list(
          mixed = res,
          spec = mixed
        )
      )
    } else {
      # not returning the mixed spectra
      return(res)
    }
    
    
  }
  
  
  
  
}

context("mixing Munsell colors")

test_that("mixMunsell works as expected", {
  
  ## error conditions
  expect_error(mixMunsell(c(NA, '10YR 3/4')))

  # invalid Munsell notation
  expect_error(expect_warning(mixMunsell(c('10YR 5/3', '10YR 3/99'))))

  # singleton
  x <- mixMunsell(c('10YR 5/3'))
  expect_true(inherits(x, 'data.frame'))
  expect_true(ncol(x) == 5)
  expect_true(nrow(x) == 1)
  expect_true(x$munsell == '10YR 5/3')

  # standard usage
  x <- mixMunsell(c('10YR 5/3', '10YR 3/2'))
  expect_true(x$munsell == '10YR 4/2')
  
  # correct output, same as above; using recycled weight length 1
  # munsell   distance
  # 1 10YR 4/2 1.955568
  
  x <- mixMunsell(c('10YR 5/3', '10YR 3/2'), w = 1)
  expect_equal(x$munsell, '10YR 4/2')
  expect_equal(x$distance, 1.9555, tolerance = 1e-4)

  # weights when length(x) != length(unique(x))
  expect_equal(mixMunsell(c('10YR 5/3', '10YR 3/2', '10YR 5/3'))$munsell,
               "10YR 4/3")

  # 0 weights to filter NA
  expect_equal(mixMunsell(c(NA, '10YR 3/4'), w = c(0, 1))$munsell,
               "10YR 3/4")

})

# mixingMethod = 'reference'
test_that("multiple matches", {
  skip_if_not_installed('gower')
  
  m <- mixMunsell(c('10YR 6/2', '10YR 2/2'), mixingMethod = 'reference', n = 3)
  
  # verified results
  expect_true(nrow(m) == 3)
  
  # first match
  expect_true(m$munsell[1] == '10Y 4/3')
  # second match
  expect_true(m$munsell[2] == '5Y 4/2')
})

# mixingMethod = 'reference'
test_that("mixed spectra option", {
  skip_if_not_installed('gower')
  
  mx <- mixMunsell(c('10YR 6/2', '10YR 2/2'), n = 3, mixingMethod = 'reference', keepMixedSpec = TRUE)
  
  # results are a list vs. data.frame
  expect_true(inherits(mx, 'list'))
  
  # mixture candidates are here
  m <- mx$mixed
  
  # verified results
  expect_true(nrow(m) == 3)
  # first match
  expect_true(m$munsell[1] == '10Y 4/3')
  # second match
  expect_true(m$munsell[2] == '5Y 4/2')
})



test_that("mixing methods, except reference", {
  
  
  ## all reference spectra available
  
  # each method gives a slightly different result
  
  # exact
  mx <- mixMunsell(c('10YR 6/2', '10YR 2/2'), mixingMethod = 'exact')
  expect_true(mx$munsell[1] == '10YR 4/2')
  
  # wt. mean CIELAB, results are slightly different
  mx <- mixMunsell(c('10YR 6/2', '10YR 2/2'), mixingMethod = 'estimate')
  expect_true(mx$munsell[1] == '10YR 4/2')
  
  # adaptive, should get exact mixture
  mx <- mixMunsell(c('10YR 6/2', '10YR 2/2'), mixingMethod = 'adaptive')
  expect_true(mx$munsell[1] == '10YR 4/2')
  
  ## some reference spectra missing
  
  # fall-back to wt.mean CIELAB
  expect_message(mx <- mixMunsell(c('10YR 6/2', '10YR 1/1'), mixingMethod = 'adaptive'))
  expect_true(mx$munsell[1] == '10YR 3/1')
  
})


test_that("reference mixing methods", {
  
  skip_if_not_installed('gower')
  
  # reference
  # requires gower package
  mx <- mixMunsell(c('10YR 6/2', '10YR 2/2'), mixingMethod = 'reference')
  expect_true(mx$munsell[1] == '10Y 4/3')
  
  ## some reference spectra missing
  
  # message and NA
  expect_message(mx <- mixMunsell(c('10YR 6/2', '10YR 1/1'), mixingMethod = 'reference'))
  expect_true(is.na(mx$munsell[1]))
  
})

test_that("additional arguments via ...", {
  
  # NO conversion to closest Munsell chip via col2munsell()
  m <- mixMunsell(c('10YR 6/2', '10YR 2/2'), mixingMethod = 'exact', convert = FALSE)
  
  # verified results
  expect_true(nrow(m) == 1)
  
  # additional columns
  expect_true(
    all(c('r', 'g', 'b') %in% names(m))
  )
  
})

#' @title Soil Texture Color Palettes
#' 
#' @description
#' Suggested color palettes for USDA soil texture classes, ranked according to average plant-available water holding capacity. The default color mapping schema is based on a palette used by SoilWeb applications.
#' 
#' @param simplify logical, return the base 12 (`TRUE`) or full 21 (`FALSE`) soil texture classes
#' 
#' @param schema select mapping between soil texture classes, and colors, currently limited to 'soilweb'
#'
#' @return `data.frame` from soil texture class codes and colors
#' 
#' @author D.E. Beaudette, Mike Walkinshaw, A.T. O'Geen
#' 
#' @export
#' 
#' @rdname soilTextureColorPal
#' 
#' @examples
#'  
#' # base 12 soil texture classes
#' # ranked by plant available water-holding capacity
#' d <- soilTextureColorPal(simplify = TRUE)
#' soilPalette(d$color, lab = d$class, lab.cex = 1)
#' 
#' # full 21 soil texture classes
#' # ranked by plant available water-holding capacity
#' d <- soilTextureColorPal(simplify = FALSE)
#' soilPalette(d$color, lab = d$class, lab.cex = 1)
#' 
soilTextureColorPal <- function(simplify = FALSE, schema = 'soilweb') {
  
  # SoilWeb soil texture ordering, based on ranking of PAWS of 
  .l <- c("s", "ls", "sl", "scl", "l", "sc", "c", "sic", "cl", "sil", "sicl", "si")
  
  .cols <- c(
    "#BEBEBE", "#FDFD9E", "#ebd834", "#92C158", "#307431", "#4C5323", 
    "#AF4732", "#E93F4A", "#EA6996", "#CD94EA", "#6D94E5", "#546BC3"
  )
  
  # expanded soil texture ordering to include all 21 classes, 
  # from SoilWeb color palette of 12 base classes
  .le <- c("cos", "s", "fs", "vfs", "lcos", "ls", "lfs", "lvfs", "cosl", "sl", "fsl", "vfsl", "scl", "l", "sc", "c", "sic", "cl", "sil", "sicl", "si")
  
  # linear interpolation required to prevent wild "blue" colors on the sandy side
  # bias required as we are increasing the resolution on the coarse side of the scalee
  .cols_e <- colorRampPalette(
    colors = .cols, 
    space = 'Lab', 
    interpolate = 'linear',  
    bias = 0.45
  )(21)
  
  if(simplify) {
    res <- data.frame(class = .l, color = .cols)
  } else {
    res <- data.frame(class = .le, color = .cols_e)
  }
  
  return(res)
  
}

## TODO: this could use an overhaul:
##  * symbol size / resolution should be determined in context
##  * area approximation should work with a larger grid, after scaling to avoid fractional thickness constraints
##  * standardize symbology with rect() vs. points() and cex
##   -> init checkerboard, select squares, fill squares
##   -> number / size of squares is the only parameter

## doesn't work with fractional depths: https://github.com/ncss-tech/aqp/issues/8

# convert volume pct [0, 100] into DF of points along a res x res grid
# v should not be NA
.volume2df <- function(v, depth, res) {
  
  # test for >= 100%
  if(v >= 100) {
    warning(sprintf("%s is >= 100, likely a data error", v), call. = FALSE)
    # truncate at 100%
    v <- 100
  }
  
  # convert volume pct into fraction
  v <- v / 100
  
  # init matrix with NA
  cells <- (depth * res)
  m <- matrix(nrow = depth, ncol = res)
  # set to 1 for debugging output (full grid)
  m[] <- NA 
  
  # determine number of cells required to symbolize volume fraction
  # v must be >0 -> select a single cell when v is <1
  v.n <- pmax(round(v * cells), 1)
  
  # random sample of cells, mark with 1 in matrix
  v.cells <- sample(1:cells, size = v.n)
  m[v.cells] <- 1
  
  # convert matrix into data.frame
  d <- expand.grid(x = (1:res), y = (1:depth))
  d$val <- m[1:cells]
  # keep only those cells with val = 1
  d <- d[which(d$val == 1), ]
  
  # scrub extra columns and return
  d$val <- NULL
  return(d)
}


#' @title Symbolize Volume Fraction within a Soil Profile Collection Plot
#' 
#' @description Symbolize volume fraction on an existing soil profile collection plot.
#' 
#' @note It may be necessary to adjust both `res`, `cex.min`, and `cex.max` for optimal legibility.
#'
#' @param x a `SoilProfileCollection` object
#' @param colname character vector of length 1, naming the column containing volume fraction data (horizon-level attribute). Values should be within 0-100 percent.
#' @param res integer, resolution of the grid used to symbolize volume fraction
#' @param cex.min minimum symbol size
#' @param cex.max maximum symbol size
#' @param pch integer, plotting character code
#' @param col symbol color, either a single color or as many colors as there are horizons in `x`
#' 
#' @details This function can only be called after plotting a `SoilProfileCollection` object. Details associated with a call to `plotSPC()` are automatically accounted for within this function: e.g. `plot.order`, `width`, etc..
#' 
#' @author D.E. Beaudette
#' 
#' @seealso [plotSPC()]
#'
#' @export
#' 
#' 
addVolumeFraction <- function(x, colname, res = 10, cex.min = 0.1, cex.max = 0.5, pch = 1, col = 'black') {
  
  # color should be either:
  # single color name / code
  # vector of colors with length == nrow(x)
  
  # simplest case, single color vector
  if(length(col) == 1) {
    col <- rep(col, times=nrow(x))
  } else {
    # check to make sure that vector of colors is the same length as number of horizons
    if(length(col) != nrow(x))
      stop('length of `col` should be either 1 or nrow(x)', call. = FALSE)
  }
  
  # ensure that `colname` is a horizon-level attribute
  if(! colname %in% horizonNames(x)) {
    stop(sprintf("%s is not a horizon-level attribute", colname), call. = FALSE)
  }
  
  # test for values < 0.5, could be a fraction [0,1] vs. percent [0,100]
  if(all( na.omit(horizons(x)[[colname]]) < 0.5) ) {
    message(sprintf("all %s values are < 0.5, likely a fraction vs. percent", colname))
  }
  
  # get plotting details from aqp environment
  lsp <- get('last_spc_plot', envir = aqp.env)
  w <- lsp$width
  plot.order <- lsp$plot.order
  
  # y.offset is a vector of length(x)
  depth.offset <- lsp$y.offset
  sf <- lsp$scaling.factor
  x0 <- lsp$x0
  
  # horizontal shrinkage factor
  # hard-coded based on aesthetic tests
  w.offset <- w / 7
  
  # get top/bottom colnames
  hd <- horizonDepths(x)
  
  # iterate over profiles
  for(p.i in 1:length(x)) {
    
    # get the current pofile, in plotting order
    h <- horizons(x[plot.order[p.i], ])
    
    ## determine left/right extent of symbols
    # 2019-07-15: using relative position as indexed by current profile
    x.center <- x0[p.i]
    x.left <- x.center - (w - w.offset)
    x.right <- x.center + (w - w.offset)
    
    # iterate over horizons
    for(h.i in 1:nrow(h)) {
      this.hz <- h[h.i, ]
      hz.thick <- this.hz[[hd[2]]] - this.hz[[hd[1]]]
      
      ## hack until #8 is resolved: round the thickness down
      # https://github.com/ncss-tech/aqp/issues/8
      if( hz.thick %% 1 != 0) {
        # truncate at closest integer
        # minimum value of 1 to ensure something is plotted within very thin horizons
        hz.thick <- pmax(floor(hz.thick), 1)
        
        msg <- sprintf(
          "%s%s -> %s%s thick", 
          paste(h[h.i, hd], collapse = ' - '), 
          depth_units(x), 
          hz.thick, 
          depth_units(x)
        )
        message(sprintf('truncating fractional horizon thickness to integer: %s', msg))
        
      }
      
      # proceed only if the current profile/horizon:
      #  * value not NA
      #  * value > 0
      #  * hz.thick not NA
      #  * hz.thick > 0
      
      if((hz.thick > 0) & (!is.na(hz.thick)) & (this.hz[[colname]] > 0) & (! is.na(this.hz[[colname]]))) {
        
        # convert this horizon's data
        v <- .volume2df(v = this.hz[[colname]], depth = hz.thick, res = res)
        
        ## TODO: this can be better solved with dynamic 'res'
        # if (thickness * volume * res) is too small, there may be no samples returned
        # skip
        if(nrow(v) < 1) {
          # print('skipping...')
          next
        }
        
        # get the current color from vector of colors
        # typically the same color repeated, but could be as many colors as hz
        # may be NA if using a horizon column
        v$color <- col[h.i]
        
        # jitter and rescale x-coordinates
        v$x <- .rescaleRange(jitter(v$x), x0 = x.left, x1 = x.right)
        
        # determine horizon depths in current setting
        # depth_prime = (depth * scaling factor) + y.offset[i]
        y.top <- (this.hz[[hd[1]]] * sf) + depth.offset[p.i]
        y.bottom <- (this.hz[[hd[2]]] * sf) + depth.offset[p.i]
        
        # rescale y-coordinates
        v$y <- .rescaleRange(jitter(v$y), x0 = y.top, x1 = y.bottom)
        
        # generate random symbol size
        p.cex <- runif(nrow(v), min = cex.min, max = cex.max)
        
        # add jittered points
        # note that color comes from `v`
        points(v$x, v$y, cex = p.cex, col = v$color, pch = pch)
      }
      
    } # end iteration over horizons
  } # end iteration over profiles
  
}

# if (!isGeneric("spc2mpspline"))
  setGeneric("spc2mpspline", function(object,
                                      var_name = NULL,
                                      method = c("est_1cm", "est_icm", "est_dcm"),
                                      pattern = "R|Cr|Cd|qm",
                                      hzdesgn = NULL, ...)
    standardGeneric("spc2mpspline"))

#' SoilProfileCollection wrapper for `mpspline2::mpspline()`
#'
#' Generate mass-preserving splines for any numeric attribute in a
#' SoilProfileCollection using `mpspline2::mpspline()`. mpspline2 implements the
#' method described in Bishop et al. (1999). 
#' 
#' @details This function now relies on the missing data checks provided by the
#'   mpspline2 package. See `attr(..., 'removed')` to see whole profiles that
#'   were removed from the set. Horizons containing `NA` in the property of
#'   interest are dropped with a message.
#'
#'   Data completeness is assessed and the input SoilProfileCollection is
#'   filtered and truncated to create a container for the results from
#'   `mpspline2::mpspline()`.
#'
#' @param object A SoilProfileCollection
#' @param var_name Column name(s) in `@horizons` slot of `object` containing
#'   numeric values to spline
#' @param pattern Regex pattern to match for bottom of profile (passed to
#'   `minDepthOf()`) default: "R|Cr|Cd|m"; only used if `hzdesgn` is specified
#' @param hzdesgn Column name in `@horizons` slot of `object` containing horizon
#'   designations default: `NULL`
#' @param method Options include "est_1cm" (default; 1cm estimates), "est_icm"
#'   (estimates over original layer boundaries), "est_dcm" (estimates over
#'   constant interval, specified with `d` argument to `mpspline3::mpspline()`).
#'   Default value for `d` is `c(0, 5, 15, 30, 60, 100, 200)`.
#' @param ... Additional arguments to `mpspline2::mpspline()`
#'
#' @author Andrew G. Brown
#'
#' @return A SoilProfileCollection with 1cm slices. Spline variables are in
#'   columns prefixed with "spline_" and RMSE/RMSE_IQR are in columns prefixed
#'   with "rmse_". If any profiles were removed from the collection, their
#'   profile IDs are stored in `attr(result, 'removed')`.
#'
#' @export
#' @importFrom stats setNames
#' @aliases spc2mpspline
#' @references Bishop, T.F.A., McBratney, A.B., Laslett, G.M. (1999) Modelling
#'   soil attribute depth functions with equal-area quadratic smoothing splines.
#'   Geoderma 91(1–2), pp. 27-45.
#'   \doi{https://doi.org/10.1016/S0016-7061(99)00003-8}
#'
#'   O'Brien, Lauren (2025). mpspline2: Mass-Preserving Spline Functions for
#'   Soil Data. R package version 0.1.9.
#'   \url{https://cran.r-project.org/package=mpspline2}
#'
#' @examplesIf requireNamespace("mpspline2")
#' data(sp1)
#' depths(sp1) <- id ~ top + bottom
#' hzdesgnname(sp1) <- "name"
#'   
#' # run on a single variable
#' res <- spc2mpspline(sp1, "prop")
#'   
#' # plot single-variable result
#' plotSPC(res[1:5, ], color = "prop_spline", divide.hz = FALSE)
#' 
#' # add a second continuous numeric variable for demonstration
#' sp1$value2 <- runif(nrow(horizons(sp1)))
#'   
#' # run on multiple variables
#' res2 <- spc2mpspline(sp1, c("prop", "value2"))
#'   
#' # plot multi-variable result
#' plotSPC(res2[1:5, ], color = "value2_spline", divide.hz = FALSE)
#'   
#' # run on multiple variables with custom depth intervals
#' res3 <- spc2mpspline(sp1,
#'                      c("prop", "value2"),
#'                      method = "est_dcm",
#'                      d = c(0, 10, 20, 50, 100))
#' plotSPC(res3[1:5, ], color = "value2_spline", divide.hz = FALSE)
#' 
setMethod("spc2mpspline", signature(object = "SoilProfileCollection"),
          function(object, 
                   var_name = NULL,
                   method = c("est_1cm", "est_icm", "est_dcm"),
                   pattern = "R|Cr|Cd|m",
                   hzdesgn = NULL,
                   ...) {
            .NHZ <- NULL
            .FIRST <- NULL
            .LAST <- NULL 
            .HZID <- NULL
            .TOP <- NULL
            .BOTTOM <- NULL
            
            if (!requireNamespace('mpspline2'))
              stop("package `mpspline2` is required", call. = FALSE)
            
            if (is.null(var_name) || !all(var_name %in% horizonNames(object)))
              stop("all `var_name` must specify horizon-level variables", call. = FALSE)

            method <- match.arg(method[1], c("est_1cm", "est_icm", "est_dcm"))
            
            hztop <- horizonDepths(object)[1]
            hzbot <- horizonDepths(object)[2]
            
            # glom to "available interval" in each profile
            spc.sub <- glom(object, object[, , .FIRST, .TOP], 
                                    object[, , .LAST, .BOTTOM])
            
            # remove any horizons that have 0 or NA thickness (no mass)
            .sameTopBottom <- NULL
            
            # handle any gaps at the surface (e.g. truncated data)
            spc.sub$.sameTopBottom <- spc.sub[[hztop]] == spc.sub[[hzbot]]
            spc.sub$.sameTopBottom <- spc.sub$.sameTopBottom | is.na(spc.sub$.sameTopBottom)
            spc.sub <- subsetHz(spc.sub, !.sameTopBottom)
            
            spc.sub$.mindepth_orig <- spc.sub[, , .FIRST, .TOP]
            
            # handle any gaps at the surface (e.g. truncated data)
            spc.sub <- fillHzGaps(spc.sub, to_top = 0, to_bottom = NULL)
            
            # calculate the top depth and bottom depth for each profile
            spc.sub$.mindepth <- spc.sub[, , .FIRST, .TOP]

            # optionally constrained by some pattern matching
            if (!missing(hzdesgn) && !is.null(hzdesgn)) {
              hzpatdep <- minDepthOf(
                spc.sub,
                pattern = pattern,
                hzdesgn = hzdesgn,
                no.contact.assigned = Inf
              )[[hztop]]
            } else {
              hzpatdep <- rep(Inf, length(spc.sub))
            }
            
            # either the bottom depth of last horizon or the matched pattern top depth
            spc.sub$.maxdepth <- pmin(hzpatdep, spc.sub[, , .LAST, .BOTTOM], na.rm = TRUE)
            
            # truncate using vectors of top and bottom
            spc.sub <- trunc(spc.sub, spc.sub$.mindepth, spc.sub$.maxdepth)

            if (length(profile_id(spc.sub)) == 0) {
              attr(object, "removed") <- unique(profile_id(object))
              return(object[0, ])
            }
            
            res <- mpspline2::mpspline(horizons(spc.sub)[c(idname(spc.sub),
                                                           horizonDepths(spc.sub),
                                                           var_name)],
                                       var_name = var_name, ...)

            # Handle the difference in mpspline2 return structure
            # Single variable returns list with profile IDs as names
            # Multiple variables returns list with variable names as top level
            if (length(var_name) == 1 && inherits(res[[1]], "list") &&
                idname(spc.sub) %in% names(res[[1]])) {
              # Single variable case: convert to multi-variable structure
              res <- list(res)
              names(res) <- var_name
            }

            # Get all profile IDs that were successfully splined for at least one variable
            all.splined.pids <- unique(unlist(lapply(res, function(v_res) {
              if (inherits(v_res, "list")) {
                names(v_res)
              } else {
                NULL # Handle cases where a variable's result might be NA
              }
            })))
            
            # determine profiles that were splined and filter spc.sub
            splined.profiles <- profile_id(spc.sub)[profile_id(spc.sub) %in% all.splined.pids]
            spc.sub <- spc.sub[which(profile_id(spc.sub) %in% splined.profiles), ]
            
            # which profiles were removed?
            removed.profiles <- profile_id(object)[!profile_id(object) %in% splined.profiles]
            
            if (length(splined.profiles) == 0) {
              attr(object, "removed") <- unique(removed.profiles)
              return(object[0, ])
            }
            
            # create slices 1cm thick to insert spline result
            switch(method,
                   "est_1cm" = {
                     spc.spl <- suppressMessages(aqp::dice(spc.sub))
                   },
                   "est_icm" = {
                     spc.spl <- spc.sub
                   },
                   "est_dcm" = {
                     .new_d_horizons <- function(x, ...) {
                       d <- list(...)$d
                       if (is.null(d)) d <- c(0, 5, 15, 30, 60, 100, 200)
                       pids <- profile_id(x)
                       newhzd <- data.frame(
                         id = rep(pids, each = length(d) - 1),
                         top = rep(d[1:(length(d) - 1)], times = length(pids)),
                         bottom = rep(d[2:length(d)], times = length(pids))
                       )
                       colnames(newhzd) <- c(idname(x), horizonDepths(x))
                       depths(newhzd) <- colnames(newhzd)
                       newhzd
                     }
                     spc.spl <- .new_d_horizons(spc.sub, ...)
                   })

            # Create profile ID to index mapping for efficient subsetting
            spl_pids <- profile_id(spc.spl)
            pid_idx <- setNames(seq_along(spl_pids), spl_pids)
            
            # iterate over variables and add results to spc.spl
            for (v in var_name) {
              res.v <- res[[v]]
              current_splined_pids <- profile_id(spc.spl)
              
              res2 <- do.call('c', lapply(current_splined_pids, function(p) {
                if (p %in% names(res.v) && !is.null(res.v[[p]])) {
                  profile_res <- res.v[[p]]
                  switch(method,
                    "est_1cm" = {
                      # Map spline results to 1cm intervals created by dice()
                      p_idx <- pid_idx[p]
                      max_depth <- max(horizons(spc.spl[p_idx, ])$bottom, na.rm = TRUE)
                      vals <- profile_res$est_1cm[1:max_depth]
                      # Pad with NA if needed
                      if (length(vals) < max_depth) vals <- c(vals, rep(NA, max_depth - length(vals)))
                      vals
                    },
                    "est_icm" = profile_res$est_icm,
                    "est_dcm" = profile_res$est_dcm
                  )
                } else {
                  # Return NA vector for profiles not splined for this variable
                  if (p %in% names(pid_idx)) {
                    p_idx <- pid_idx[p]
                    num_hz <- nrow(horizons(spc.spl[p_idx, ]))
                    
                    if (num_hz > 0) {
                      rep(NA, num_hz)
                    } else {
                      NULL
                    }
                  } else {
                    NULL
                  }
                }
              }))
              
              # Extract RMSE values (site-level attributes)
              rmse_vals <- t(sapply(profile_id(spc.spl), function(p) {
                if (p %in% names(res.v) && !is.null(res.v[[p]]) && "est_err" %in% names(res.v[[p]])) {
                  res.v[[p]]$est_err[c("RMSE", "RMSE_IQR")]
                } else {
                  c(RMSE = NA_real_, RMSE_IQR = NA_real_)
                }
              }))
              rmse_values <- rmse_vals[, "RMSE"]
              rmse_iqr_values <- rmse_vals[, "RMSE_IQR"]
              
              # Add splined values and error estimates to spc.spl
              spc.spl[[paste0(v, "_spline")]] <- res2
              site(spc.spl)[[paste0(v, "_rmse")]] <- rmse_values
              site(spc.spl)[[paste0(v, "_rmse_iqr")]] <- rmse_iqr_values
            }
            
            # add an attribute with removed profile IDs
            attr(spc.spl, "removed") <- unique(removed.profiles)
            
            if (method != "est_dcm") {
              spc.spl <- trunc(spc.spl, spc.sub$.mindepth_orig, spc.sub$.maxdepth)
            }
            
            return(spc.spl)
          })

# s: soil profile collection object
# s.fm: slicing formula, including variables requested for missing data test
# cols: vector of colors palette

#' Missing Data Grid
#'
#' Generate a levelplot of missing data from a SoilProfileCollection object.
#'
#' This function evaluates a `missing data fraction` based on slice-wise
#' evaluation of named variables in a \code{SoilProfileCollection} object.
#'
#' @param s a SoilProfileCollection object
#' @param max_depth integer specifying the max depth of analysis
#' @param vars character vector of column names over which to evaluate missing
#' data
#' @param filter.column a character string naming the column to apply the
#' filter REGEX to
#' @param filter.regex a character string with a regular expression used to
#' filter horizon data OUT of the analysis
#' @param cols a vector of colors
#' @param \dots additional arguments passed on to \code{levelplot}
#' @return A \code{data.frame} describing the percentage of missing data by
#' variable.
#' @note A lattice graphic is printed to the active output device.
#' @author D.E. Beaudette
#' @seealso \code{\link{slice}}
#' @keywords hplots
#' @export
#' @examples
#'
#' # 10 random profiles
#' set.seed(10101)
#' s <- lapply(as.character(1:10), random_profile)
#' s <- do.call('rbind', s)
#'
#' # randomly sprinkle some missing data
#' s[sample(nrow(s), 5), 'p1'] <- NA
#' s[sample(nrow(s), 5), 'p2'] <- NA
#' s[sample(nrow(s), 5), 'p3'] <- NA
#'
#' # set all p4 and p5 attributes of `soil 1' to NA
#' s[which(s$id == '1'), 'p5'] <- NA
#' s[which(s$id == '1'), 'p4'] <- NA
#'
#' # upgrade to SPC
#' depths(s) <- id ~ top + bottom
#'
#' # plot missing data via slicing + levelplot
#' missingDataGrid(
#'   s,
#'   max_depth = 100,
#'   vars = c('p1', 'p2', 'p3', 'p4', 'p5'),
#'   main='Missing Data Fraction'
#' )
#'
missingDataGrid <- function(s, max_depth, vars, filter.column = NULL, filter.regex = NULL, cols = NULL, ...) {

  # default color scheme
  if(is.null(cols)) {
    cols <- c("#3288BD", "#66C2A5", "#ABDDA4", "#E6F598",
              "#FEE08B", "#FDAE61", "#F46D43", "#D53E4F")
  }


  # make color pallete and define number of cuts
  cols.palette <- colorRampPalette(cols)
  ncuts <- 20

  ## TODO: use horizon designation from SPC if possible
  # optionally filter horizon data in original object and replace
  if(!is.null(filter.column) & !is.null(filter.regex)) {
    h <- horizons(s)
    idx <- grep(filter.regex, h[, filter.column], invert=TRUE)
    slot(s, 'horizons') <- h[idx, ]
    rm(h)
  }

  ## TODO: this is not efficient
  # get a list of horizon boundary depths for latter annotation of sliced data
  obd <- profileApply(s, simplify=FALSE, FUN=function(i) {
    hd <- horizonDepths(i)
    h <- horizons(i)
    hz.boundaries <- unique(c(h[[hd[1]]], h[[hd[2]]]))
  })


  # compute percent missing data by pedon/variable
  # this is only used as a summary and returned by function

  ## TODO: abstract this, useful in other contexts
  pct_missing <- profileApply(s, frameify = TRUE, FUN = function(i, v = vars) {
    h <- horizons(i)

    frac.missing <- sapply(
      h[, v, drop = FALSE], function(j) {
        length(which(is.na(j)))
      }) / nrow(h)

    res <- data.frame(
      .id = profile_id(i)[1],
      t(round(frac.missing * 100)),
      stringsAsFactors = FALSE
    )

    names(res)[1] <- idname(i)
    return(res)
  })



  # slice according to rules
  s.fm <- as.formula(paste('0:', max_depth, ' ~ ', paste(vars, collapse=' + '), sep=''))
  ss <- dice(s, s.fm, pctMissing = TRUE)

  # get sliced horizon depth names
  hd <- horizonDepths(ss)

  # extract horizons from sliced data
  h <- horizons(ss)

  # get slice mid-points
  h$mid <- (h[[hd[1]]] + h[[hd[2]]]) / 2


  # NOTE: since we are converting profile IDs to a factor,
  # we need to explicitly set levels to match the original ordering of profiles
  forced.levels <- paste("c('", paste(profile_id(ss), collapse="','"), "')", sep='')

  # construct levelplot formula using horizon top boundaries
  f <- as.formula(paste('.pctMissing',  ' ~ ', 'factor(', idname(ss), ', levels=', forced.levels, ') * mid', sep=''))

  # ylab adjustments
  ylab <- paste('Depth ', '(', depth_units(ss), ')', sep='')

  # depth-range adjustments
  ylim <- c(max(h$mid) + 5, -3)

  # plot missing data fraction
  lp <- lattice::levelplot(f, data=h, ylim=ylim, col.regions=cols.palette(ncuts), cuts=ncuts-1, ylab=ylab, xlab='', scales=list(x=list(rot=90), y=list(tick.number=10)), ..., panel=function(...) {
    lattice::panel.levelplot(...)
    lattice::panel.abline(v=1:(length(ss)+1)-0.5)
    lattice::panel.grid(h=-1, v=FALSE, lty=2, col=grey(0.25))
    for(i in 1:length(obd)) {
      lattice::panel.segments(i-0.5, obd[[i]], i+0.5, obd[[i]], col='black')
    }
  })


  # return figure missing data percentages by pedon
  return(list(fig = lp, summary = pct_missing))

}

context("missingDataGrid")

# 10 random profiles
set.seed(10101)
s <- lapply(as.character(1:10), random_profile)
s <- do.call('rbind', s)

# randomly sprinkle some missing data
s[sample(nrow(s), 5), 'p1'] <- NA
s[sample(nrow(s), 5), 'p2'] <- NA
s[sample(nrow(s), 5), 'p3'] <- NA

# set all p4 and p5 attributes of `soil 1' to NA
s[which(s$id == '1'), 'p5'] <- NA
s[which(s$id == '1'), 'p4'] <- NA

# upgrade to SPC
depths(s) <- id ~ top + bottom



## tests


test_that("works as expected", {
  
  v <- c('p1', 'p2', 'p3', 'p4', 'p5')
  
  # run it
  m <- missingDataGrid(
    s, 
    max_depth = 100, 
    vars = v, 
    main='Missing Data Fraction'
  )
  
  # expected output
  expect_true(inherits(m, 'list'))
  expect_true(length(m) == 2)
  
  expect_true(inherits(m$fig, 'trellis'))
  expect_true(inherits(m$summary, 'data.frame'))
  
  expect_true(nrow(m$summary) == length(s))
  expect_true(ncol(m$summary) == length(v) + 1)
  
  # select missing data proportions
  # verified
  expect_true(m$summary$p1[1] == 17)
  expect_true(m$summary$p2[1] == 0)
  expect_true(m$summary$p3[1] == 17)
  expect_true(m$summary$p4[1] == 100)
  expect_true(m$summary$p5[1] == 100)
  
  # verified
  expect_true(m$summary$p1[3] == 0)
  expect_true(m$summary$p2[3] == 40)
  expect_true(m$summary$p3[3] == 20)
  expect_true(m$summary$p4[3] == 0)
  expect_true(m$summary$p5[3] == 0)
  
})

#' @title Sieve the Coarse Fraction of Soil
#' 
#' @description Sieve applies thresholds to a numeric vector of fragment diameter values, returning fragment size classes. Particle diameter thresholds are evaluated as `d < threshold`.
#' 
#' @references 
#' 
#' Soil Science Division Staff. 2017. Soil survey manual. C. Ditzler, K. Scheffe, and H.C. Monger (eds.). USDA Handbook 18. Government Printing Office, Washington, D.C.
#' 
#' @param diameter numeric. Vector of diameters of coarse fragments to "sieve". Default `sieves` are specified in millimeters.
#' 
#' @param sieves leave as `NULL` to use fragment class labels and diameters defined by [fragmentClasses()], or a named vector of fragment diameters. See examples.
#' 
#' @param ordered logical. Return as an ordered factor.
#' 
#' @param prefix character. Add a prefix to result names? Default: `""` adds no prefix. For example `"para"` might be used for size classes of pararock fragments.
#' 
#' @param new_names Optional: apply new labels to result classes. Should match length of `sieves`. 
#' 
#' @param ... additional arguments to [fragmentClasses()], such as `sys`, `flat`, and `rounded`, see examples.
#'
#' @return character. Size class labels based on names of `sieves`, `new_names`, and `prefix` (if specified).
#' 
#' @seealso [fragmentClasses()]
#' 
#' @export
#'
#' @examples
#' 
#' # use a simplified version of the USDA system
#' # common within NRCS/SPSD and NCSS
#' fragmentSieve(c(30, 125, 180, 500, 1000))
#' 
#' # pararock fragments
#' fragmentSieve(c(30, 125, 180, 500, 1000), prefix = 'para')
#' 
#' # result as an ordered factor
#' fragmentSieve(c(30, 125, 180, 500, 1000), ordered = TRUE)
#' 
#' # USDA system, flat size classes
#' fragmentSieve(c(30, 125, 180, 500, 1000), flat = TRUE)
#' 
#' # alternative classification systems
#' fragmentSieve(c(30, 125, 180, 500, 1000), sys = 'usda')
#' fragmentSieve(c(30, 125, 180, 500, 1000), sys = 'international')
#' fragmentSieve(c(30, 125, 180, 500, 1000), sys = 'unified')
#' fragmentSieve(c(30, 125, 180, 500, 1000), sys = 'aashto')
#' fragmentSieve(c(30, 125, 180, 500, 1000), sys = 'mod.wentworth')
#' 
#' # custom fragment labels / diameter
#' fragmentSieve(
#'   c(30, 125, 180, 500, 1000),
#'   sieves = c(clumps = 50, chunks = 300, blocks = 100000)
#' )
#' 
#' # unnamed sieves, generic labels used
#' fragmentSieve(c(10, 50), sieves = c(30, 70))
#' 
#' fragmentSieve(c(10, 50), sieves = c(30, 70), ordered = TRUE)
#'  
fragmentSieve <- function(diameter,
                  sieves = NULL, 
                  ordered = FALSE,
                  prefix = "",
                  new_names = NULL, 
                  ...) {
  
  
  # if not specified as named vector of diameters
  # use fragmentClasses() to lookup one of several systems
  if (is.null(sieves)) {
    sieves <- fragmentClasses(...)
  }
  
  if (!is.null(new_names)) {
    names(sieves) <- new_names
  }
  
  # test for NA, and filter-out
  res <- vector(mode = 'character', length = length(diameter))
  res[which(is.na(diameter))] <- NA
  no.na.idx <- which(!is.na(diameter))
  
  # only assign classes to non-NA diameters
  if (length(no.na.idx) > 0) {
    # pass diameters "through" sieves
    # 2020: latest part 618 uses '<' for all upper values of class range
    # 2022: adjusted gravel upper threshold to 76 mm 
    classes <- t(sapply(diameter[no.na.idx], function(i) i < sieves))
    
    if (length(names(sieves)) == 0) {
      names(sieves) <- paste0("class_", seq_along(sieves))
    }
    
    # determine largest passing sieve name
    res[no.na.idx] <- names(sieves)[apply(classes, 1, which.max)]
    
    # apply prefix if specified, e.g. parafrags
    if (nchar(prefix) > 0) {
      res[no.na.idx] <- paste0(prefix, res[no.na.idx])
    }
  }
  
  # optional conversion to ordered factor
  if (ordered) {
    res <- factor(res, levels = names(sieves), ordered = TRUE)
  }
  
  return(res)
}

context("sieve")


test_that("basic functionality, USDA simplified system", {
  
  # obvious diameters
  expect_equal(fragmentSieve(c(75, 150, 450, 1000)), c('gravel', 'cobbles', 'stones', 'boulders'))
  
  # right on the line
  expect_equal(fragmentSieve(c(76)), 'cobbles')
  
  # NA
  expect_equal(fragmentSieve(c(76, NA)), c('cobbles', NA_character_))
  
  # flat
  expect_equal(fragmentSieve(c(76), flat = TRUE), 'channers')
  
  # prefix
  expect_equal(fragmentSieve(c(76), prefix = 'chewy_'), 'chewy_cobbles')
  
  # custom sieves
  expect_equal(fragmentSieve(c(50, 100), sieves = c('teaspoon' = 76, 'tablespoon' = 250)), c('teaspoon', 'tablespoon'))
  
  # ordered factors
  expect_true(
    inherits(
      fragmentSieve(c(5, 4, 7, 75, 150, 450, 1000), ordered = TRUE),
      'factor'
    )
  )
  
})

test_that("other systems", {
  
  x <- c(4, 35, 150, 400, 650)
  
  # USDA simplified
  expect_equal(fragmentSieve(x), c('gravel', 'gravel', 'cobbles', 'stones', 'boulders'))
  
  # USDA
  expect_equal(fragmentSieve(x, sys = 'USDA'), c('fine_gravel', 'coarse_gravel', 'cobbles', 'stones', 'boulders'))
  
  # International
  expect_equal(fragmentSieve(x, sys = 'international'), c('gravel', 'stones', 'stones', 'stones', 'stones'))
  
  # unified
  expect_equal(fragmentSieve(x, sys = 'unified'), c('fine_gravel', 'coarse_gravel', 'cobbles', 'boulders', 'boulders'))
  
  # AASHTO
  expect_equal(fragmentSieve(x, sys = 'AASHTO'), c('fine_gravel', 'coarse_gravel', 'broken_rock', 'broken_rock', 'broken_rock'))
  expect_equal(fragmentSieve(x, sys = 'AASHTO', rounded = TRUE), c('fine_gravel', 'coarse_gravel', 'boulders', 'boulders', 'boulders'))
  
  # modified wentworth
  expect_equal(fragmentSieve(x, sys = 'mod.wentworth'), c('pebbles', 'pebbles', 'cobbles', 'boulders', 'boulders'))
  
})

# computes proper limits when there is an 'upper' and 'lower' bound
# expands x-axis on either side of upper/lower (if present) by 5%
#' @rdname panel.depth_function
#' @export
prepanel.depth_function <- function(x, y, upper=NA, lower=NA, subscripts, groups = NULL, ...) {

  ## borrowed from prepanel.default.xyplot
  ord <- order(as.numeric(x))
  if (!is.null(groups)) {
    gg <- groups[subscripts]
    dx <- unlist(lapply(split(as.numeric(x)[ord], gg[ord]), 
                        diff))
    dy <- unlist(lapply(split(as.numeric(y)[ord], gg[ord]), 
                        diff))
  }
  else {
    dx <- diff(as.numeric(x[ord]))
    dy <- diff(as.numeric(y[ord]))
  }
  
  
  # composite into a data.frame
  if(!missing(upper) & !missing(lower)) {
    d <- data.frame(
      yhat = x, 
      top = y, 
      upper = upper[subscripts],
      lower = lower[subscripts]
    )
  } else {
    # no upper / lower - use central value as place-holder
    d <- data.frame(
      yhat = x, 
      top = y, 
      upper = x, 
      lower = x
    )
  }
  
  
  # compute better xlim based on range of upper/lower values
  # requires some data in both x and y
  if (any(!is.na(x)) & any(!is.na(y))) {

    x_range <- c(
      min(c(d$lower, d$yhat), na.rm = TRUE), 
      max(c(d$upper, d$yhat), na.rm = TRUE)
    )
    
    y_range <- c(
      min(d$top, na.rm = TRUE), 
      max(d$top, na.rm = TRUE)
    )
    
    # expand by 5%
    x_range[1] <- x_range[1] - (x_range[1] * 0.05)
    x_range[2] <- x_range[2] + (x_range[2] * 0.05)
    
    y_range[1] <- y_range[1] - (y_range[1] * 0.05)
    y_range[2] <- y_range[2] + (y_range[2] * 0.05)
    
    # results as expected by xyplot
    res <- list(
      xlim = x_range, 
      ylim = y_range, 
      dx = dx, 
      dy = dy,
      xat = NULL,
      yat = NULL
    )
    
    return(res)
    
  } else {
    # otherwise simulate lattice:::prepanel.null()
    res <- list(
      xlim = c(NA_real_, NA_real_), 
      ylim = c(NA_real_, NA_real_), 
      dx = NA_real_, 
      dy = NA_real_
      )
    
    return(res)
  }
  
}

.reactionclass <- function() {
  data.frame(
    DescriptiveTerm = c("ultra acid", "extremely acid", "very strongly acid", "strongly acid", 
                        "moderately acid", "slightly acid", "neutral", "slightly alkaline", 
                        "moderately alkaline", "strongly alkaline", "very strongly alkaline"),
    pH_low = c(0, 3.55, 4.45, 5.05, 5.55, 6.05, 6.55, 7.35, 7.85, 8.45, 9.05),
    pH_high = c(3.55, 4.45, 5.05, 5.55, 6.05, 6.55, 7.35, 7.85, 8.45, 9.05, 14)
  )
}

.reactionclass2 <-  function() {
  # rda can be built using reactionclass <- .reactionclass(); save(reactionclass, file = "data/reactionclass.rda")
  reactionclass <- NULL
  load(system.file("data", "reactionclass.rda", package = "aqp"))
  reactionclass
}

.phclasses1 <- function(halfclass = FALSE) {
  lut1 <- .reactionclass()
  if (halfclass) {
    lut2 <- lut1
    lut2$pH_high <- lut2$pH_low + ((lut2$pH_high - lut2$pH_low) / 2)
    lut1$pH_low <- lut2$pH_high
    lut <- rbind(data.frame(id = "low", lut2), data.frame(id = "high", lut1))
    lut <- lut[order(lut$pH_low),]
  } else {
    lut1$id <- ""
    lut <- lut1
  }
  lut
}

.phclass <- function(x, halfclass = FALSE) {
  lut1 <- .phclasses1(halfclass = halfclass)
  idx <- findInterval(x, lut1[["pH_low"]])
  res <- trimws(paste(lut1[["id"]][idx], lut1[["DescriptiveTerm"]][idx]))
  res[res == "NA NA"] <- NA_character_
  res
}

.phrange <- function(x, halfclass = FALSE) {
  if (any(grepl('high|low', x, ignore.case = TRUE))) halfclass <- TRUE
  lut1 <- .phclasses1(halfclass = halfclass)
  res <- lut1[trimws(tolower(paste(lut1[["id"]], lut1[["DescriptiveTerm"]]))) %in% tolower(x), c("pH_low", "pH_high")]
  res2 <- res[0,][1,]
  res2$pH_low <- min(res$pH_low)
  res2$pH_high <- max(res$pH_high)
  rownames(res2) <- NULL
  res2
}

#' Convert pH to/from Reaction Classes
#'
#' @param x input pH values (numeric; `ph_to_rxnclass()`) or reaction classes (character; `rxnclass_to_ph()`)
#' @param halfclass Split the standard classes in half for higher resolution? Default: `FALSE`
#' @param as.is logical. Should character vectors be converted to factors? Default: `FALSE`
#' @param droplevels logical. Drop unused levels in factors? Default: `FALSE`
#' @return `ph_to_rxnclass()`: a vector of reaction classes corresponding to numeric input in `x`; if `as.is=FALSE` an ordered factor using `ReactionClassLevels()`
#' @export
#' @rdname reaction
#' @examples
#' ph_to_rxnclass(6.2)
ph_to_rxnclass <- function(x, 
                           halfclass = FALSE,
                           as.is = FALSE,
                           droplevels = TRUE) {
  res <- .phclass(x, halfclass = halfclass)
  
  if (!as.is) {
    res <- factor(res, levels = ReactionClassLevels(halfclass = halfclass, as.is = TRUE), ordered = TRUE)
    if (droplevels) {
      return(droplevels(res))
    }
  }
  res
}

#' @param digits Number of digits after decimal place; Default: `2`. Used only for `rxnclass_to_ph()`
#' @param simplify Simplify list result to numeric vector when length of result is 1? Default: `TRUE`
#' @export
#' @return `rxnclass_to_ph()`: a list of data.frame objects containing high/low values of reaction class 1:1 with input; if simplify=TRUE and input is  a data.frame.
#' @rdname reaction
#' @examples 
#' rxnclass_to_ph("slightly acid")
#' 
#' rxnclass_to_ph(list(c("Slightly Acid", NA, "Moderately Acid"),
#'                     c("Slightly Acid", NA, "Strongly Acid")), simplify = FALSE)
rxnclass_to_ph <- function(x, halfclass = FALSE, digits = 2, simplify = TRUE) {
  if (!is.list(x))
    x <- list(x)
  res <- lapply(x, function(y) round(.phrange(y, halfclass = halfclass), digits = digits))
  if (length(res) == 1 && simplify) {
    return(res[[1]])
  }
  res
}

#' @export
#' @return `ReactionClassLevels()`: ordered factor containing descriptive terms for reaction classes 
#' @rdname reaction
#' @examples 
#' ReactionClassLevels()
ReactionClassLevels <- function(halfclass = FALSE, as.is = FALSE) {
  rxn <- .reactionclass()
  x <- rxn$DescriptiveTerm
  
  if (halfclass) {
    i <- seq(1, length(x) * 2, 2)
    i2 <- seq(2, length(x) * 2, 2)
    x <- c(paste("low", x), paste("high", x))
    x <- x[order(c(i, i2))]
  }
  
  if (as.is) {
    return(x)
  }
  
  factor(x, levels = x, ordered = TRUE)
}

test_that("ph_to_rxnclass() works", {
  expect_equal(as.character(ph_to_rxnclass(6.2)), "slightly acid")
  expect_equal(ph_to_rxnclass(c(6.2, NA, 6.3), as.is = TRUE), 
               c("slightly acid", NA, "slightly acid"))
  
  expect_equal(ph_to_rxnclass(c(6.2, NA, 6.8)),
               factor(c("slightly acid", NA, "neutral"),
                      levels = c("slightly acid", "neutral"),
                      ordered = TRUE))
  
  expect_equal(ph_to_rxnclass(c(6.2, NA, 6.8), halfclass = TRUE, droplevels = FALSE),
               factor(c("low slightly acid", NA, "low neutral"),
                      levels = ReactionClassLevels(halfclass = TRUE, as.is = TRUE),
                      ordered = TRUE))
  
  expect_equal(length(levels(ph_to_rxnclass(c(6.2, NA, 6.8)))), 2)
})

test_that("rxnclass_to_ph() works", {
  expect_equal(rxnclass_to_ph("slightly acid"), data.frame(pH_low = 6.05, pH_high = 6.55))
  
  expect_equal(rxnclass_to_ph(c("slightly acid", NA, "moderately acid"), simplify = FALSE), 
               list(data.frame(pH_low = 5.55, pH_high = 6.55)))
  
  expect_equal(rxnclass_to_ph(list(c("slightly acid", NA, "moderately acid"),
                                   c("slightly acid", NA, "strongly acid")), simplify = FALSE), 
               list(data.frame(pH_low = 5.55, pH_high = 6.55),
                    data.frame(pH_low = 5.05, pH_high = 6.55)))
})

# x: SoilProfileCollection
# h: horizon-level attribute, typically a GHL
# result: list of levels and median depths
#' Guess Appropriate Ordering for Generalized Horizon Labels
#'
#' This function makes an (educated) guess at an appropriate set of levels for
#' generalized horizon labels using the median of horizon depth mid-points.
#'
#' This function is useful when groups of horizons have been generalized via
#' some method other than \code{generalize.hz}. For example, it may be useful
#' to generalize horizons using labels derived from slice depths. The default
#' sorting of these labels will not follow a logical depth-wise sorting when
#' converted to a factor. \code{guessGenHzLevels} does a good job of "guessing"
#' the proper ordering of these labels based on median horizon depth mid-point.
#'
#' @param x a \code{SoilProfileCollection} object
#' @param hz name of horizon-level attribute containing generalized horizon
#' labels, see details
#' @return a list: \item{levels}{a vector of levels sorted by median horizon
#' depth mid-point} \item{median.depths}{a vector of median horizon mid-points}
#' @author D.E. Beaudette
#' @seealso \code{\link{generalize.hz}}
#' @keywords manip
#' @export
#' @examples
#'
#' # load some example data
#' data(sp1, package='aqp')
#'
#' # upgrade to SoilProfileCollection
#' depths(sp1) <- id ~ top + bottom
#'
#' # generalize horizon names
#' n <- c('O', 'A', 'B', 'C')
#' p <- c('O', 'A', 'B', 'C')
#' sp1$genhz <- generalize.hz(sp1$name, n, p)
#'
#' # note: levels are in the order in which originally defined:
#' levels(sp1$genhz)
#'
#' # generalize horizons by depth slice
#' s <- dice(sp1, c(5, 10, 15, 25, 50, 100, 150) ~ .)
#' s$slice <- paste0(s$top, ' cm')
#' # not a factor
#' levels(s$slice)
#'
#' # the proper ordering of these new labels can be guessed from horizon depths
#' guessGenHzLevels(s, 'slice')
#'
#' # convert to factor, and set proper order
#' s$slice <- factor(s$slice, levels=guessGenHzLevels(s, 'slice')$levels)
#'
#' # that is better
#' levels(s$slice)
#'
guessGenHzLevels <- function(x, hz = GHL(x, required = TRUE)) {
  tb <- horizonDepths(x)
  h <- horizons(x)
  # compute horizon mid-point
  m <- (h[[tb[1]]] + h[[tb[2]]]) / 2
  # median mid-point is probably a good indicator of depth-wise ordering
  m.med <- tapply(m, h[[hz]], median, na.rm=TRUE)
  # sort and return
  s <- sort(m.med)
  return(list(levels=names(s), median.depths=s))
}

## TODO: consider various sorting strategies: WMPD, AWC, {PWP,FC,SAT}
## http://ncss-tech.github.io/AQP/aqp/water-retention-curves.html

## TODO: move ranking system definitions into internal function or environment
## TODO: ensure that creating ranking code is easily repeatable

# .USDA_soil_texture_ranking_systems <- function(sys) {
#   
#   # 1. Field Book (what is the original authority?)
#   
#   # 2. ranked AWC
#   
#   # 3. clustered (SAT, FC, PWP) / sorted (AWC)
#   
# }


#' @title Ranking Systems for USDA Soil Texture Classes
#' 
#' @description Generate a vector of USDA soil texture codes or class names, sorted according to approximate particle size
#'
#' @references \href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}
#' 
#' @param which 'codes' (texture codes) or 'names' (texture class names)
#' 
#' @param simplify Return 12-class factor levels (`TRUE`) or 21-class factor levels (default: `FALSE`)? The 12-class system does not separate sands, loamy sands and sandy loams into sand fraction variants (e.g. "very fine sandy loam" in the 21-class system is "sandy loam" in 12-class system)
#' 
#' @return an ordered factor
#' 
#' @export
#' @examples
#' 
#' # class codes
#' SoilTextureLevels()
#' 
#' # class names
#' SoilTextureLevels(which = 'names')
#' 
#' # simpler class names
#' SoilTextureLevels(which = 'names', simplify = TRUE)
#'  
SoilTextureLevels <- function(which = 'codes', simplify = FALSE) {
  
  # sanity check
  if(! which %in% c('codes', 'names')) {
    stop('must specify `codes` (texture codes) or `names` (texture class names)', call. = FALSE)
  }
  
  # sorted by approximate particle size
  if (!simplify) { # from the Field Book version 3.0
    tx <- data.frame(
      texture = c("coarse sand", "sand", "fine sand", 
                  "very fine sand", "loamy coarse sand", "loamy sand", "loamy fine sandy", 
                  "loamy very fine sand", "coarse sandy loam", "sandy loam", "fine sandy loam", 
                  "very fine sandy loam", "loam", "silt loam", "silt", "sandy clay loam", 
                  "clay loam", "silty clay loam", "sandy clay", "silty clay", "clay"), 
      texcl = c("cos", "s", "fs", "vfs", "lcos", "ls", "lfs", "lvfs", 
                "cosl", "sl", "fsl", "vfsl", "l", "sil", "si", "scl", "cl", "sicl", 
                "sc", "sic", "c")
    ) 
  } else { # From Soil Survey Manual (1951) p. 210
    tx <- data.frame(
      texture = c("sand", "loamy sand", "sandy loam",
                  "loam", "silt loam", "silt", "sandy clay loam",
                  "clay loam", "silty clay loam",
                  "sandy clay", "silty clay", "clay"),
      texcl = c("s", "ls", "sl", "l", "sil", "si",
                "scl", "cl", "sicl", "sc", "sic", "c")
    )
  }
  
  # set levels
  tx$texture <- factor(tx$texture, levels = tx$texture, ordered = TRUE)
  tx$texcl <- factor(tx$texcl, levels = tx$texcl, ordered = TRUE)
  
  # return what was requested
  switch(
    which,
    codes = return(tx$texcl),
    names = return(tx$texture)
  )
  
}

#' @title Find and Fill Horizon Gaps
#'
#' @description This function attempts to find "gaps" in the horizon records of a `SoilProfileCollection` object and fill with placeholder horizons (profile ID, horizon ID, to/bottom depths, all else `NA`). Missing horizon records between the top of each profile and `to_top`, or the bottom of each profile and `to_bottom` are treated as gaps when those arguments are not `NULL`. You can use this function to prepare a potentially messy `SoilProfileCollection` for subsequent analyses that are sensitive to horizon sequence inconsistencies or require a conformal "rectangle" of data spanning known depths.
#'
#' Gaps are defined as:
#'  * within each profile, for horizons `i` to `n_hz`:
#'  * `bottom_i != top_i+1 (but only to i = 1:(n_hz - 1)`
#'
#' @param x `SoilProfileCollection` object
#' 
#' @param flag logical, flag empty horizons that have been added. default: `TRUE`
#' 
#' @param to_top numeric, fill from shallowest top depth in each profile to specified depth? default: `0` 
#' @param to_bottom numeric, fill from deepest bottom depth in each profile to specified depth? default: `aqp::max(x)` 
#' 
#' @return a possibly modified `SoilProfileCollection` object
#' 
#' @author A.G. Brown and D.E. Beaudette
#' 
#' @export
#'
#' @examples
#'
#' data(sp4)
#' depths(sp4) <- id ~ top + bottom
#'
#' # introduce depth logic errors
#' idx <- c(2, 6:7, 8, 12)
#' sp4$top[idx] <- NA
#'
#' # check
#' horizons(sp4)[idx, ]
#'
#' # create gaps by removing logic errors
#' x <- HzDepthLogicSubset(sp4, byhz = TRUE)
#'
#' # check on removed horizons (hzID values)
#' metadata(x)$removed.horizons
#'
#' # inspect
#' par(mar = c(0, 0, 0, 2))
#' plotSPC(x, width = 0.3, default.color = 'royalblue', 
#' name = 'hzID', name.style = 'center-center', cex.names = 0.8,
#' cex.id = 0.66)
#'
#' # fill gaps left by HzDepthLogicSubset()
#' z <- fillHzGaps(x, flag = TRUE)
#' 
#' # graphical check
#' plotSPC(z, width = 0.3, color = '.filledGap', name = 'hzID', 
#' show.legend = FALSE, name.style = 'center-center', cex.names = 0.8,
#' cex.id = 0.66)
#' 
#' # fill top to 0 cm
#' z2 <- fillHzGaps(x, flag = TRUE, to_top = 0)
#' plotSPC(z2, width = 0.3, color = '.filledGap', name = 'hzID', show.legend = FALSE)
#' 
#' # fill bottom to max(SPC)
#' z3 <- fillHzGaps(x, flag = TRUE, to_top = 0, to_bottom = max(x))
#' plotSPC(z3, width = 0.3, color = '.filledGap', name = 'hzID', show.legend = FALSE)
#'
#' ## another example
#' data(sp4)
#' depths(sp4) <- id ~ top + bottom
#' 
#' # remove 1st horizons from profiles 1:4
#' idx <- sp4[,, .FIRST, .HZID]
#' replaceHorizons(sp4) <- horizons(sp4)[-idx[1:4], ]
#' 
#' # prepare for dice()
#' z <- fillHzGaps(sp4, to_top = 0, to_bottom = 50, flag = TRUE) 
#' 
#' # empty-horizon padding is in place for formula interface to dice()
#' d <- dice(z, fm = 0:50 ~ .)
#' plotSPC(d, color = 'Ca', show.legend = FALSE)
#' plotSPC(d, color = '.filledGap', show.legend = FALSE)
#' 
fillHzGaps <- function(x, flag = TRUE, to_top = 0, to_bottom = max(x)) {
  
  # SPC details
  idn <- idname(x)
  htb <- horizonDepths(x)
  hznames <- horizonNames(x)
  hcnames <- c(idn, htb)
  
  # just the horizons, convert to DT
  h <- data.table::as.data.table(horizons(x))
  nh <- nrow(h)
  
  # indices used by gap-detection
  lead.idx <- (2:nh)[seq_len(nh)]
  lag.idx <- (1:(nh - 1))[seq_len(nh)]

  # identify affected horizons
  # this will include the first perfectly overlapping horizons
  # in the case of multiple entries of e.g. E/B
  bad.idx <- which(h[[htb[2]]][lag.idx] != h[[htb[1]]][lead.idx]
                   & h[[idn]][lag.idx] == h[[idn]][lead.idx])
  
  # TODO: data.table more performant than above for large SPC?
  #   bad.idx <- h[, .I[.SD[1:(.N - 1), 3] != .SD[2:.N, 2] &
  #                     .SD[1:(.N - 1), 1] == .SD[2:.N, 1]],
  #                      .SDcols = hcnames]

  ## BUG: need a short-circuit for single-profile / single-horizon objects
  ## https://github.com/ncss-tech/aqp/issues/301
  ## 
  ## NOTE: do not short-circuit, a single profile, single-horizon may still need filling depending on the top/bottom depth and other params
  
  # https://github.com/ncss-tech/aqp/issues/296
  # do not attempt to fill gaps when there are perfectly overlapping horizons
  # likely an intentional data modeling decision
  # vector matches order and length of horizons
  .ov <- flagOverlappingHz(x)
  
  # filter out perfectly overlapping horizons
  bad.idx <- setdiff(bad.idx, which(.ov))
  
  # create template data.frame
  hz.template <- h[bad.idx, ]

  if (nrow(hz.template) > 0) {
    # replace non-ID/depth column values with NA
    hz.template[, hznames[!hznames %in% hcnames]] <- NA

    # fill gaps
    hz.template[[htb[1]]] <- h[[htb[2]]][bad.idx]     # replace top with (overlying) bottom
    hz.template[[htb[2]]] <- h[[htb[1]]][bad.idx + 1] # replace bottom with (underlying) top
  }

  # fill from shallowest top depth to X cm
  surface.template <- hz.template[0, ]
  
  if (!is.null(to_top) && is.numeric(to_top)) {
    .FIRST <- NULL
    .HZID <-  NULL
    surface.template <- h[x[,, .FIRST, .HZID],]
    surface.template <- surface.template[which(surface.template[[htb[1]]] > to_top)]
    if (nrow(surface.template) > 0) {
      surface.template[, hznames[!hznames %in% hcnames]] <- NA

      surface.template[[htb[2]]] <- surface.template[[htb[1]]] # replace bottom with (underlying) top
      surface.template[[htb[1]]] <- to_top                     # replace top with zero
    }
  }
  
  # fill from deepest bottom depth to X cm
  bottom.template <- hz.template[0, ]
  
  if (!is.null(to_bottom) && is.numeric(to_bottom)) {
    .LAST <- NULL
    .HZID <-  NULL
    bottom.template <- h[x[,, .LAST, .HZID],]
    bottom.template <- bottom.template[which(bottom.template[[htb[2]]] < to_bottom)]
    if (nrow(bottom.template) > 0) {
      bottom.template[, hznames[!hznames %in% hcnames]] <- NA
      
      bottom.template[[htb[1]]] <- bottom.template[[htb[2]]] # replace top with (overlying) bottom
      bottom.template[[htb[2]]] <- to_bottom                 # replace bottom with to_bottom
    }
  }
  
  # flag if needed
  if (flag) {
    if (nh > 0) h[['.filledGap']] <- FALSE
    if (nrow(hz.template) > 0) hz.template[['.filledGap']] <- TRUE
    if (nrow(surface.template) > 0) surface.template[['.filledGap']] <- TRUE
    if (nrow(bottom.template) > 0) bottom.template[['.filledGap']] <- TRUE
  }

  # combine original data with filled data
  res <- h
  if (nrow(hz.template) > 0) {
    res <- rbind(res, hz.template)
  }
  
  if (nrow(surface.template) > 0) {
    res <- rbind(res, surface.template)
  }
  
  if (nrow(bottom.template) > 0) {
    res <- rbind(res, bottom.template)
  }
   
  if (nrow(res) != nrow(x)) { 
    # ID + top depth sort
    idx <- order(res[[idn]], res[[htb[1]]])
    res <- res[idx, ]
  
    # re-calculate unique hzID (note: AFTER reorder)
    if (is.null(res$hzID) || !is.numeric(as.numeric(res$hzID))) {
     res$hzID <- as.character(1:nrow(res))
    } else {
     nahz <- is.na(res$hzID)
     res$hzID[nahz] <- max(as.numeric(res$hzID), na.rm = TRUE) + seq_len(sum(nahz))
    }
    # replace horizons (use df class in object x)
    replaceHorizons(x) <- .as.data.frame.aqp(res, aqp_df_class(x))
  } else if (flag) {
    horizons(x)$.filledGap <- FALSE
  }
  
  # use the auto-calculated hzID (in case user had e.g. phiid, chiid set)
  hzidname(x) <- "hzID"

  return(x)
}

#' @title Test for a valid `SoilProfileCollection`
#'
#' @description Test for a valid `SoilProfileCollection`
#'
#' @details Test for valid \code{SoilProfileCollection} by checking for slots defined in
#' the class prototype. Likely only used between major versions of `aqp` where
#' internal structure of \code{SoilProfileCollection} has changed. Use
#' \code{checkHzDepthLogic} to check for common errors in horizon depths.
#'
#' @param x a \code{SoilProfileCollection} object
#' @return TRUE or FALSE. Consider using \code{rebuildSPC()} if FALSE.
#' @author D.E. Beaudette
#' @seealso \code{\link{rebuildSPC}}, \code{\link{checkHzDepthLogic}}
#' @export
checkSPC <- function(x) {

  # get slot names from prototype
  sn <- slotNames(x)

  # test for all slots in the prototype
  s.test <- sapply(sn, function(i) .hasSlot(x, name=i))

  res <- FALSE
  # a valid object will have all slots present
  if(all(s.test)) {
    res <- TRUE
  }

  return(res)
}

#' @title An S4 object representation of a group of soil profiles.
#'
#' @name SoilProfileCollection
#' @export
#'
#' @slot idcol character.
#' @slot hzidcol character.
#' @slot depthcols character.
#' @slot metadata list.
#' @slot horizons data.frame.
#' @slot site data.frame.
#' @slot diagnostic data.frame.
#' @slot restrictions data.frame.
#' 
#' @details
#' After aqp 2.0.2, the `@sp` slot was removed from the SoilProfileCollection object. If you run into errors related to old object definitions, use `rebuildSPC()` on the problematic object.
#' 
#' @aliases SoilProfileCollection-class
#' @rdname SoilProfileCollection-class
setClass(
  Class = 'SoilProfileCollection',
  representation = representation(
    idcol = 'character', # column name containing IDs
    hzidcol = 'character', # column name containing unique horizon IDs
    depthcols = 'character', # 2 element vector with column names for hz top, bottom
    metadata = 'list', # list with key-value mapping
    horizons = 'data.frame', # all horizons sorted by ID & top depth
    site = 'data.frame', # data about the sampling sites
    # sp = 'NULL', # no longer used, formerly 'empty' sp object
    diagnostic = 'data.frame',# (optional) diagnostic horizons are stored here
    restrictions = 'data.frame' # (optional) restrictions are stored here
  ),
  prototype = prototype(
    idcol = 'id',
    hzidcol = 'hzID',
    depthcols = c('top', 'bottom'),
    metadata = list(aqp_df_class = "data.frame", # data.frame subclass
                    aqp_group_by = "",           # grouping variable
                    aqp_hzdesgn = "",            # horizon designation column
                    aqp_texcl = "",              # texture class column
                    stringsAsFactors = FALSE),
    horizons = data.frame(id  = character(0), hzID = character(0),
                          top = numeric(0), bottom = numeric(0),
                          stringsAsFactors = FALSE),
    site = data.frame(id = character(0), stringsAsFactors = FALSE),
    # sp = NULL, 
    diagnostic = data.frame(stringsAsFactors = FALSE),
    restrictions = data.frame(stringsAsFactors = FALSE)
  ),
  validity = function(object) {
    # https://github.com/ncss-tech/aqp/issues/75
    return(spc_in_sync(object)$valid)
  }
)

# 2019-03-15: creating an empty SpatialPoints object requires more effort
# c/o: https://gis.stackexchange.com/questions/291069/creating-empty-spatialpoints-or-spatialpointsdataframe-in-r
# old: new('SpatialPoints')
# new: SpatialPoints(data.frame(x = 0, y = 0))[-1,]
# new new: NULL

#' Constructor for the SoilProfileCollection object
#'
#' @param idcol character. Profile ID Column Name
#' @param hzidcol character. Horizon ID Column Name
#' @param depthcols character. length 2 Top and Bottom Depth Column Names
#' @param metadata list. metadata including data.frame class in use and depth units
#' @param horizons data.frame. An object inheriting from data.frame containing Horizon data.
#' @param site data.frame. An object inheriting from data.frame containing Site data.
#' @param diagnostic data.frame. An object inheriting from data.frame containing diagnostic feature data. Must contain profile ID. See \code{diagnostic_hz()}
#' @param restrictions data.frame. An object inheriting from data.frame containing restrictive feature data. Must contain profile ID. See \code{restrictions()}
#'
#' @description In general, one should use \code{depths()} to initiate a SoilProfileCollection object from data. However, sometimes there are instances where either an empty, or very specific, object is needed. If that is the case, the general constructor \code{SoilProfileCollection} is available.
#'
#' @author Pierre Roudier, Dylan E. Beaudette, Andrew G. Brown
#' @export
#'
#' @rdname SoilProfileCollection-class
#' @examples
#'
#' ## structure of default, empty SoilProfileCollection
#' str(SoilProfileCollection())
#'
#'
#' ## use the depths() formula interface to specify
#' ## profile ID, top and bottom depth and set up
#' ## a SPC that is topologically correct and complete
#'
#' d <- do.call('rbind', lapply(1:10, random_profile))
#'
#' # promote to SoilProfileCollection and plot
#' depths(d) <- id ~ top + bottom
#' plot(d)
#'
#' # split into new SoilProfileCollection objects by index
#' d.1 <- d[1, ]
#' d.2 <- d[2, ]
#' d.345 <- d[3:5, ]
#'
#' # combine profile collections
#' # note that profiles are sorted according to ID
#' d.new <- c(d.345, d.1, d.2)
#' plot(d.new)
#'
#' data(sp1)
#'
#' ## init SoilProfileCollection objects from data.frame
#' depths(sp1) <- id ~ top + bottom
#'
#' ## depth units
#' du <- depth_units(sp1)
#' depth_units(sp1) <- 'in'
#' depth_units(sp1) <- du
#'
#' ## horizon designation column
#' hzdesgnname(sp1) <- "name"
#' hzdesgnname(sp1)
#'
#' ## all designations in an SPC (useful for single profile SPC)
#' hzDesgn(sp1)
#'
#' ## horizon texture class column
#' hztexclname(sp1) <- "texture"
#' hztexclname(sp1)
#'
#' ## get/set metadata on SoilProfileCollection objects
#' # this is a 1-row data.frame
#' m <- metadata(sp1)
#' m$sampler <- 'Dylan'
#' metadata(sp1) <- m
#'
#' ## extract horizon data from SoilProfileCollection objects as data.frame
#' h <- horizons(sp1)
#'
#' # also merge (left-join) of new columns and
#' # replacement of existing columns via horizons<-
#' horizons(sp1) <- h
#'
#' # get number of horizons
#' nrow(sp1)
#'
#'
#' ## getting site-level data
#' site(sp1)
#'
#' ## setting site-level data
#' # site-level data from horizon-level data (stored in @horizons)
#' site(sp1) <- ~ group
#'
#'
#' # make some fake site data, and append from data.frame
#' # a matching ID column must be present in both @site and new data
#' # note that IDs should all be character class
#' d <- data.frame(id=profile_id(sp1), p=runif(n=length(sp1)), stringsAsFactors=FALSE)
#' site(sp1) <- d
#'
#' # edit horizon depths
#' horizonDepths(sp1) <- c('t', 'b')
#' horizonDepths(sp1)
#'
#' # edit profile IDs
#' p <- sprintf("%s-new", profile_id(sp1))
#' profile_id(sp1) <- p
#' profile_id(sp1)
#'
"SoilProfileCollection" <-
  function(idcol = 'id',
           hzidcol = 'hzID',
           depthcols = c('top', 'bottom'),
           metadata = list(aqp_df_class = "data.frame",
                           aqp_group_by = "",
                           aqp_hzdesgn = "",
                           aqp_hztexcl = "",
                           stringsAsFactors = FALSE),
           horizons = data.frame(
             id  = character(0),
             hzID = character(0),
             top = numeric(0),
             bottom = numeric(0),
             stringsAsFactors = FALSE
           ),
           site = data.frame(id = character(0), stringsAsFactors = FALSE),
           # sp = NULL, 
           diagnostic = data.frame(stringsAsFactors = FALSE),
           restrictions = data.frame(stringsAsFactors = FALSE)) {

    # retrieve highest-level data.frame subclass of horizon data
    hzclass <- class(horizons)[1]

    new.metadata <- metadata

    # set metadata (default: data.frame, centimeters)
    metadata <- list(
      aqp_df_class = hzclass,
      aqp_group_by = "",
      aqp_hzdesgn = "",
      aqp_hztexcl = "",
      depth_units = 'cm',
      stringsAsFactors = FALSE
    )

    # add any custom metadata
    metadata <- c(metadata,
                  new.metadata[!names(new.metadata) %in% names(metadata)])

    # add aqp_group_by if not defined
    if (length(metadata$aqp_group_by) == 0)
      metadata$aqp_group_by <- ""

    # "allow" NULL for the optional slots
    if(length(metadata$aqp_hzdesgn) == 0)
      metadata$aqp_hzdesgn <- ""

    if(length(metadata$aqp_hztexcl) == 0)
      metadata$aqp_hztexcl <- ""

    # create object
    new(
      "SoilProfileCollection",
      idcol = idcol,
      hzidcol = hzidcol,
      depthcols = depthcols,
      metadata = metadata,
      horizons = .as.data.frame.aqp(horizons, hzclass),
      site = .as.data.frame.aqp(site, hzclass),
      # sp = sp,
      diagnostic = .as.data.frame.aqp(diagnostic, hzclass),
      restrictions = .as.data.frame.aqp(restrictions, hzclass)
    )
  }

## show
#' SoilProfileCollection show method
#' @name show
#'
#' @description Pretty output method for SoilProfileCollection objects. By default this method limits output to 10 columns and 6 rows from the site and horizon tables respectively.
#'
#' There is an aqp environment option you can set to increase the number of columns shown by default: \code{options(.aqp.show.n.cols = 100)},
#'
#' @param object a SoilProfileCollection
#' @aliases show,SoilProfileCollection-method
#' @docType methods
#' @rdname show
#' @examples
#'
#' # load a SoilProfileCollection
#' data(sp5)
#'
#' # use the show() method
#' show(sp5)
#'
#' # which is same as this (in the console)
#' sp5
#'
setMethod(f = 'show',
          signature(object = 'SoilProfileCollection'),
          function(object) {

            # stats for later
            n.profiles <- length(object)
            n.hz <- nrow(object)

            # count number of hz and site columns for reporting truncated listing
            n.hz.cols <- length(horizonNames(object))
            n.site.cols <- length(siteNames(object))

            # local copies
            s <- object@site
            h <- object@horizons

            # determine number of rows to show
            n.s <- nrow(s)
            n.h <- nrow(h)

            # show up to 6 rows for each slot, or max rows if less
            rows.show.s <- 1:pmin(6, nrow(s))
            rows.show.h <- 1:pmin(6, nrow(h))

            # handle zero row case (so it doesnt fill with NA)
            if (n.h == 0)
              rows.show.h <- 0

            if (n.s == 0)
              rows.show.s <- 0

            # subset rows
            h <- h[rows.show.h, , drop = FALSE]
            s <- s[rows.show.s, , drop = FALSE]

            # move IDs and depths, horizon designation/texture if available
            hzm <- .hzMetadataNames(object, depths = TRUE)
            idx <- match(hzm, names(h))

            # determine number of columns to show, and index to hz / site data
            # user sett-able
            show.cols <- getOption('.aqp.show.n.cols')

            h.n <- names(h)

            # handle empty case
            hz.show <- numeric(0)
            site.show <- numeric(0)

            if (length(h.n) > 0) {

              h <- .data.frame.j(h, c(h.n[idx], h.n[-na.omit(idx)]), aqp_df_class(object))

              # show first n
              hz.show <- seq(from = 1,
                             to = pmin(show.cols, n.hz.cols),
                             by = 1)
              site.show <- seq(from = 1,
                               to = pmin(show.cols, n.site.cols),
                               by = 1)
            }

            # column subseting
            if (length(hz.show) > 0) {
              # generate text explaining truncated summary
              hz.txt <- sprintf(
                "\n----- Horizons (%s / %s rows  |  %s / %s columns) -----\n",
                nrow(h),
                n.h,
                pmin(show.cols, n.hz.cols),
                n.hz.cols
              )
            } else {
              hz.txt <- "[EMPTY]\n"
            }

            if(length(site.show) > 0) {
              site.txt <- sprintf(
                "\n----- Sites (%s / %s rows  |  %s / %s columns) -----\n",
                nrow(s),
                n.s,
                pmin(show.cols, n.site.cols),
                n.site.cols
              )
            } else {
              site.txt <- "[EMPTY]\n"
            }

            # header
            header.txt <-
              sprintf(
                "SoilProfileCollection with %s profiles and %s horizons\nprofile ID: %s  |  horizon ID: %s \nDepth range: %s - %s %s\n",
                n.profiles,
                n.hz,
                idname(object),
                hzidname(object),
                min(object),
                max(object),
                depth_units(object)
              )
            cat(header.txt)

            # make note of additional hz attributes
            cat(hz.txt)

            print(data.frame(h)[, hz.show, drop = FALSE], row.names = FALSE)

            if(n.h > max(rows.show.h))
              cat('[... more horizons ...]\n')

            # make note of additional site attributes
            cat(site.txt)

            print(data.frame(s)[, site.show, drop = FALSE], row.names = FALSE)

            if(n.s > max(rows.show.s))
              cat('[... more sites ...]\n')

            # presence of spatial data
            if (validSpatialData(object)) {
              cat('\nSpatial Data:\n')
              cat("  CRS: ", prj(object), ";", sep = "")
              cat(.spc_bbox(object))
            } else {
              cat('\nSpatial Data:\n[EMPTY]\n')
            }

          })

.spc_bbox <- function(x) {
  crds <- metadata(x)$coordinates
 
  # this bbox outputs for "point" geometries specified in two columns
  # TODO: extend as needed if other geometry types are added (i.e. wrapper around st_bbox())
  if (length(crds) != 2) 
    return("\n")
  
  # if all coordinates in X or Y are NA warnings will be generated & Inf/-Inf result
  suppressWarnings({
    paste0(" ", crds[1], ": ", min(x[[crds[1]]], na.rm = TRUE), " to ", max(x[[crds[1]]], na.rm = TRUE), "; ",
                crds[2], ": ", min(x[[crds[2]]], na.rm = TRUE), " to ", max(x[[crds[2]]], na.rm = TRUE), "\n")
  })
}

#' @description `as.character()`: Character Representation of SoilProfileCollection Object
#' @param x a SoilProfileCollection
#' @param ... additional arguments (not used)
#' @keywords internal
#' @rdname show
#' @export
setMethod('as.character', 'SoilProfileCollection', function(x, ...) {
  paste0('SPC<', length(x), ",", nrow(x), '>')
})

#' @title Wrapper method for data.frame subclass conversion
#' @noRd
#' @param x ANY.
#' @param as.class `"data.frame"`, `"tibble"`, or `"data.table"` default: `"data.frame"`
#' @param ... Additional arguments to coercion function  `as.data.frame`, `as_tibble` or `as.data.table`
#' @return a subclass of `data.frame` corresponding to `as.class`
.as.data.frame.aqp <- function(x, as.class = "data.frame", ...) {
  # 2020-05-30: sub-classes of data.frame have more than one class

  # NULL x -- probably from unusual use cases
  if (class(x)[1] == "NULL")
    stop(sprintf("input object is NULL, expected '%s'", as.class))

  # don't invoke coercion methods if not needed
  if (!inherits(x, 'data.frame')) {
    stop(sprintf(
      "input data class %s does not inherit from `data.frame`",
      class(x)[1]
    ),
    call. = TRUE)
  }

  # note: we handle the possibly NULL/0-length as.class
  #       by letting it fall through default switch EXPR
  #       a warning is generated for non-data.frames
  cond <- class(x)[1] == as.class
  test <- all(length(cond) > 0 & cond)

  # this happens if a SPC has had its metadata entry wiped out or old SPC object in Rda file
  if (is.null(test) | is.na(test)) {
    as.class <- "data.frame"
    message(
      "missing metadata for aqp_df_class -- run aqp::rebuildSPC(object) to fix slots and metadata"
    )
  } else if (test) {
    # rm rownames in slots
    rownames(x) <- NULL

    return(x)
  }

  switch(as.class,
         'data.table' = {
           #print(as.class)
           if (requireNamespace("data.table", quietly = TRUE))
             return(data.table::as.data.table(x, ...))
           message(
             "using data.table class in SoilProfileCollection slots requires the `data.table` package"
           )
         },
         'tbl_df' = {
           #print(as.class)
           if (requireNamespace("tibble", quietly = TRUE))
             return(tibble::as_tibble(x, ...))
           message("using tbl_df class in SoilProfileCollection slots requires the `tibble` package")
         },
         {
           # default/data.frame
           #  if we were supposed to get something else,
           #  make a warning with a stack trace
           if (as.class != "data.frame") {
             message(sprintf("failed to use %s as data.frame class", as.class))
             metadata(x)$aqp_df_class <- "data.frame"
             warning(
               "data.table and tbl_df in SoilProfileCollection data.frame slots are EXPERIMENTAL, defaulting to data.frame",
               call. = FALSE
             )
           }

           # return data.frame no matter what
           res <- as.data.frame(x, ...)

           # rm rownames in slots
           rownames(res) <- NULL

           return(res)
         })
}


# basic wrapper function for multi-j index subsetting of data.frames compatible with data.table

.data.frame.j <- function(x, col.names, use_class = class(x)[1]) {

  # see: https://github.com/ncss-tech/aqp/issues/176
  .SD <- NULL

  if (inherits(x, 'data.table')) {
    res <- x[, .SD, .SDcols = col.names]
  } else {
    res <- x[, col.names, drop = FALSE]
  }

  if (inherits(res, 'data.frame')) {
    h <- .as.data.frame.aqp(res, use_class)
    return(h)
  } else {
    stop(".data.frame.j: result does not inherit from `data.frame`", call. = FALSE)
  }
}

setGeneric('hzID<-', function(object, value)
  standardGeneric('hzID<-'))

#' Set horizon IDs
#'
#' @description Set vector containing horizon IDs
#'
#' @param object a SoilProfileCollection
#' @param value a unique vector of equal length to number of horizons \code{nrow(object)}
#' @aliases hzID<-
#' @docType methods
#' @export
#' @rdname hzID
setReplaceMethod("hzID", signature(object = "SoilProfileCollection"),
                 function(object, value) {
                   if (!inherits(value, 'character')) {
                     message("converting horizon IDs from integer to character")
                     value <- as.character(value)
                   }

                   # can't be missing
                   if (is.null(value)) {
                     stop('horizon IDs cannot be NULL or NA', call. = FALSE)
                   }

                   if (any(is.na(value)) | any(is.null(value))) {
                     stop('horizon IDs cannot be NULL or NA', call. = FALSE)
                   }

                   # length
                   if (length(value) != nrow(object)) {
                     stop('replacement horizon IDs must have same length as original',
                          call. = FALSE)
                   }

                   # unique
                   if (length(value) != length(unique(value))) {
                     stop('replacement horizon IDs must be unique', call. = FALSE)
                   }

                   # extract horizon and replace IDs
                   object@horizons[[hzidname(object)]] <- value

                   return(object)
                 })

setGeneric('profile_id<-', function(object, value)
  standardGeneric('profile_id<-'))

#' Set profile IDs
#'
#' @name profile_id<-
#'
#' @description Set vector containing profile IDs
#'
#' @param object a SoilProfileCollection
#' @param value a unique vector of equal length to number of profiles \code{length(object)}
#' @aliases profile_id<-,SoilProfileCollection-method
#' @docType methods
#' @rdname profile_id
#' @export
setReplaceMethod("profile_id", signature(object = "SoilProfileCollection"),
                 function(object, value) {
                   # can't be missing
                   if (is.null(value)) {
                     stop('profile IDs cannot be NULL or NA', call. = FALSE)
                   }

                   if (any(is.na(value)) | any(is.null(value))) {
                     stop('profile IDs cannot be NULL or NA', call. = FALSE)
                   }

                   # length
                   if (length(value) != length(profile_id(object))) {
                     stop('replacement profile IDs must have same length as original',
                          call. = FALSE)
                   }

                   # unique
                   if (length(value) != length(unique(value))) {
                     stop('replacement profile IDs must be unique', call. = FALSE)
                   }

                   # lookup table for converting old -> new IDs
                   idn <- idname(object)
                   pids <- profile_id(object)
                   lut <- cbind(pids, value)

                   # change @site
                   s <- site(object)
                   s[[idn]] <- value
                   object@site <-
                     .as.data.frame.aqp(s, metadata(object)$aqp_df_class)

                   # change @horizons
                   h <- object@horizons
                   update.idx <- match(h[[idn]], lut[, 1])
                   # apply edits via LUT
                   h[[idn]] <- lut[update.idx, 2]
                   object@horizons <-
                     .as.data.frame.aqp(h, metadata(object)$aqp_df_class)

                   # search in @diagnostic
                   dg <- diagnostic_hz(object)
                   dg.nm <- names(dg)
                   idx <- grep(idn, dg.nm)

                   if (length(idx) > 0) {
                     # apply edits via LUT
                     update.idx <- match(dg[[idx]], lut[, 1])
                     dg[[idx]] <- lut[update.idx, 2]
                     suppressWarnings(diagnostic_hz(object) <- dg)
                   }

                   # search in @restrictions
                   re <- restrictions(object)
                   re.nm <- names(re)
                   idx <- grep(idn, re.nm)

                   if (length(idx) > 0) {
                     # apply edits via LUT
                     update.idx <- match(re[[idx]], lut[, 1])
                     re[[idx]] <- lut[update.idx, 2]
                     suppressWarnings(restrictions(object) <- re)
                   }

                   return(object)
                 }
)

setGeneric('horizonDepths<-', function(object, value)
  standardGeneric('horizonDepths<-'))

#' Set horizon depth column names
#'
#' @name horizonDepths<-
#'
#' @description Set column name containing horizon ID
#'
#' @param object a SoilProfileCollection
#' @param value a character vector of length two with names of columns containing numeric top and bottom depths
#' @aliases horizonDepths<-,SoilProfileCollection-method
#' @docType methods
#' @rdname horizonDepths
#' @export
setReplaceMethod("horizonDepths", signature(object = "SoilProfileCollection"),
                 function(object, value) {
                   # can't be missing
                   if (is.null(value)) {
                     stop('cannot assign NA or NULL depth column names', call. = FALSE)
                   }

                   if (any(is.na(value)) | any(is.null(value))) {
                     stop('cannot assign NA or NULL depth column names', call. = FALSE)
                   }

                   # length
                   if (length(value) != 2) {
                     stop('horizon depth names must be a vector with two items', call. = FALSE)
                   }

                   # warn about changes in names
                   if (any(value != make.names(value))) {
                     warning('names have been modified to legal data.frame column names')
                   }

                   # must be safely convertable to character and safe for DF
                   value <- make.names(value)

                   # save old values
                   hd <- horizonDepths(object)

                   # change @horizons, just the names
                   hn <- horizonNames(object)
                   idx <- match(hd, hn)
                   hn[idx] <- value
                   horizonNames(object) <- hn

                   # change @depthcols
                   object@depthcols <- value

                   return(object)
                 })


setGeneric('horizonNames<-', function(object, value)
  standardGeneric('horizonNames<-'))
  
#' Set horizon column names
#'
#' @name horizonNames<-
#'
#' @description Set horizon column names
#'
#' @param object a SoilProfileCollection
#' @param value a unique vector of equal length to number of columns in horizons \code{length(horizonNames(object))}
#' @aliases horizonNames<-,SoilProfileCollection-method
#' @docType methods
#' @rdname horizonNames
#' @export
setReplaceMethod("horizonNames", signature(object = "SoilProfileCollection"),
                 function(object, value) {
                   # sanity check
                   if (any(is.null(value)))
                     stop('cannot assign NA or NULL column names', call. = FALSE)

                   if (any(is.na(value)))
                     stop('cannot assign NA or NULL column names', call. = FALSE)

                   # must be same length
                   if (length(value) != length(horizonNames(object))) {
                     stop('replacement must have same length as original', call. = FALSE)
                   }

                   # warn about changes in names
                   if (any(value != make.names(value))) {
                     warning('names have been modified to legal data.frame column names')
                   }

                   # assign
                   names(object@horizons) <- make.names(value)
                   return(object)
                 })

setGeneric('siteNames<-', function(object, value)
  standardGeneric('siteNames<-'))

#' Set site column names
#'
#' @name siteNames<-
#'
#' @description Set site column names
#'
#' @param object a SoilProfileCollection
#' @param value a unique vector of equal length to number of columns in site: \code{length(siteNames(object))}
#' @aliases siteNames<-,SoilProfileCollection-method
#' @docType methods
#' @rdname siteNames
#' @export
setReplaceMethod("siteNames",
                signature(object = "SoilProfileCollection"),
                 function(object, value) {

                   # sanity check
                   if (is.na(value) | is.null(value))
                     stop('cannot assign NA or NULL column names', call. = FALSE)

                   names(object@horizons) <- make.names(value)
                   return(object)
                 })


setGeneric('hzidname<-', function(object, value)
  standardGeneric('hzidname<-'))

#' Set horizon ID column name
#'
#' @name hzidname<-
#'
#' @description Set unique horizon ID column name
#'
#' @param object a SoilProfileCollection
#' @param value character, column name containing unique horizon ID values
#' @aliases hzidname<-,SoilProfileCollection-method
#' @docType methods
#' @rdname hzidname
#' @export
#' @examples
#' data(sp1)
#'
#' # promote to SPC
#' depths(sp1) <- id ~ top + bottom
#'
#' # create new horizon ID
#' sp1$hzIDrev <- rev(sp1$hzID)
#'
#' # set horizon designation column
#' hzidname(sp1) <- "hzIDrev"
#'
#' # get horizon designation column
#' hzidname(sp1)
#'
setReplaceMethod("hzidname",
                signature(object = "SoilProfileCollection"),
                 function(object, value) {
                   # quick sanity check
                   if (length(value) != 1)
                     stop("horizon ID name should have length of 1", call. =
                            TRUE)


                   # sanity checks

                   # test: does it exist?
                   if (!value %in% horizonNames(object)) {
                     stop("horizon ID name not in horizon data", call. = TRUE)
                   }

                   # test: unique?
                   x <- object@horizons[[value]]
                   if (length(unique(x)) != nrow(object)) {
                     # convert error to warning,
                     # prevent stoppage from nonunique,
                     # fail gracefully and retain to default
                     warning("horizon ID name (",
                             value,
                             ") not unique. unique ID not changed.",
                             call. = TRUE)
                   } else {
                     # replace
                     object@hzidcol <- value

                     # convert contents to character, if needed
                     if (!is.character(x)) {
                       message(sprintf("converting horizon IDs in column `%s` to character", value))
                       object@horizons[[value]] <-
                         as.character(object@horizons[[value]])
                     }
                   }

                   # done
                   return(object)
                 })

##
## accessors
##


setGeneric("idname", function(object, ...)
  standardGeneric("idname"))

#' Get profile ID column name
#'
#' @description Get column name containing unique profile IDs
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases idname
#' @rdname idname
#' @export
setMethod("idname", signature(object = "SoilProfileCollection"),
          function(object)
            return(object@idcol))

setGeneric("hzidname", function(object, ...)
  standardGeneric("hzidname"))

#' Get horizon ID column name
#'
#' @description Get column name containing unique horizon ID
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases hzidname
#' @rdname hzidname
#' @export
setMethod("hzidname", signature(object = "SoilProfileCollection"),
          function(object)
            return(object@hzidcol))


setGeneric("hzID", function(object)
  standardGeneric("hzID"))

#' Get horizon IDs
#'
#' @description Get vector containing horizon IDs
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases hzID
#' @rdname hzID
#' @export
setMethod("hzID", signature(object = "SoilProfileCollection"),
          function(object) {
            return(object@horizons[[hzidname(object)]])
          })

## distinct profile IDs
setGeneric("profile_id", function(object)
  standardGeneric("profile_id"))

## relies on ordering of profile IDs in horizons matching site

#' Get/set unique (sorted) profile IDs
#'
#' @description Get or set a vector of profile IDs
#'
#' @param object a SoilProfileCollection
#'
#' @docType methods
#' @aliases profile_id
#' @rdname profile_id
#' @export
setMethod("profile_id", signature(object = "SoilProfileCollection"),
          function(object)
            unique(as.character(object@horizons[[idname(object)]])))

setGeneric("horizonDepths", function(object)
  standardGeneric("horizonDepths"))

#' Get horizon depth column names
#'
#' @description Get column names containing horizon depths
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases horizonDepths
#' @rdname horizonDepths
#' @export
setMethod("horizonDepths", signature(object = "SoilProfileCollection"),
          function(object)
            return(object@depthcols))

## site data
setGeneric("site", function(object, ...)
  standardGeneric("site"))

#' Retrieve site data from SoilProfileCollection
#'
#' @description Get site data from SoilProfileCollection. Result is returned in the same \code{data.frame} class used to initially construct the SoilProfileCollection.
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases site
#' @rdname site
#' @export
setMethod("site", signature(object = "SoilProfileCollection"),
          function(object) {
            return(.as.data.frame.aqp(object@site, aqp_df_class(object)))
          })


## diagnostic horizons: stored as a data.frame
setGeneric("diagnostic_hz", function(object, ...)
  standardGeneric("diagnostic_hz"))

#' Get or Set Diagnostic Horizon data in a SoilProfileCollection
#'
#' @description Diagnostic horizons describe features of the soil relevant to taxonomic classification. A single profile may have multiple diagnostic features or horizons, each of which may be comprised of multiple horizons.
#' 
#'  - `diagnostic_hz()` (get method): Get diagnostic feature data from a SoilProfileCollection. 
#'  
#' @param object a SoilProfileCollection
#'
#' @docType methods
#' @aliases diagnostic_hz
#' @rdname diagnostic_hz
#' @export
setMethod(f = 'diagnostic_hz', signature(object = 'SoilProfileCollection'),
          function(object) {
            return(.as.data.frame.aqp(object@diagnostic, aqp_df_class(object)))
          })


## restrictions: stored as data.frame
setGeneric("restrictions", function(object, ...)
  standardGeneric("restrictions"))

#' Get or Set Restriction data in a SoilProfileCollection
#'
#' @description Restrictions describe root-limiting features in the soil. A single profile may have multiple restrictions.
#' 
#'  - `restrictions()` (get method): Get restriction data from a SoilProfileCollection. 
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases restrictions
#' @rdname restrictions
#' @export
setMethod(f = 'restrictions', signature(object = 'SoilProfileCollection'),
          function(object) {
            return(.as.data.frame.aqp(object@restrictions, aqp_df_class(object)))
          })

## horizon data
# returns a data.frame with horizons data
setGeneric("horizons", function(object, ...)
  standardGeneric("horizons"))

#' Retrieve horizon data from SoilProfileCollection
#'
#' @description Get horizon data from SoilProfileCollection. Result is returned in the same \code{data.frame} class used to initially construct the SoilProfileCollection.
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases horizons
#' @rdname horizons
#' @export
setMethod(f = 'horizons', signature(object = 'SoilProfileCollection'),
          function(object) {
            return(.as.data.frame.aqp(object@horizons, metadata(object)$aqp_df_class))
          })

## metadata
# returns a data.frame
setGeneric("metadata", function(object, ...)
  standardGeneric("metadata"))

#' Retrieve metadata from SoilProfileCollection
#'
#' @description Get metadata from SoilProfileCollection. Result is a list. Two entries (aqp_df_class, depth_units) should not be edited in the metadata list directly. There are methods that facilitate changing them -- and propagating their changes throughout the collection. Otherwise, metadata list is a free-form slot used to store arbitrary information about the data, how it was collected, citations, etc.
#'
#' @param object a SoilProfileCollection
#' @export
#' @docType methods
#' @aliases metadata
#' @rdname metadata
#' @export
setMethod(f = 'metadata', signature(object = 'SoilProfileCollection'),
          function(object) {
            return(object@metadata)
          })

setGeneric("aqp_df_class", function(object)
  standardGeneric("aqp_df_class"))

#' Get aqp_df_class entry from metadata or return a safe value.
#'
#' @description This is an accessor and replacement method for the \code{aqp_df_class} entry in the metadata slot. This entry is used internally by methods that interact with \code{data.frame} objects and slots to ensure that the same class used to promote to the SoilProfileCollection initially is used throughout the process.
#'
#' @param object a SoilProfileCollection
#' @aliases aqp_df_class
#' @docType methods
#' @aliases aqp_df_class
#' @rdname aqp_df_class
#' @export
setMethod(f = 'aqp_df_class', signature(object = 'SoilProfileCollection'),
          function(object) {
            u <- as.character(metadata(object)[['aqp_df_class']])

            # all handles the logical(0) for undefined
            if (all(u == '')) {
              message("aqp_df_class metadata entry not found - run aqp::rebuildSPC() to fix")
              return("data.frame")
            }

            return(u)
          })

setGeneric("aqp_df_class<-", function(object, value)
  standardGeneric("aqp_df_class<-"))

#' @param value "data.frame", "data.table" or "tbl_df"
#' @aliases aqp_df_class<-
#' @rdname aqp_df_class
#' @export
setReplaceMethod("aqp_df_class", signature(object = "SoilProfileCollection"),
                 function(object, value) {
                   if (length(value) != 1 | !is.character(value))
                     stop("aqp_df_class metadata entry must be `data.frame`, `data.table` or `tbl_df`",  call.=FALSE)
                   if (value %in% c("tbl","tibble"))
                     value <- "tbl_df"
                   if (!value %in% c('data.frame',"data.table","tbl_df"))
                     stop("aqp_df_class metadata entry must be `data.frame`, `data.table` or `tbl_df`",  call.=FALSE)
                   metadata(object)$aqp_df_class <- value
                   return(object)
                 })

setGeneric("depth_units", function(object, ...)
  standardGeneric("depth_units"))

#' Get depth units from metadata
#'
#' @description Get units of depth measurement from metadata. Default value is centimeters.
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases depth_units
#' @rdname depth_units
#' @export
setMethod(f = 'depth_units', signature(object = 'SoilProfileCollection'),
          function(object) {
            u <- as.character(metadata(object)[['depth_units']])
            # give a warning if not defined
            if (length(u) == 0 | u == '')
              message('depth_units have not been defined')
            return(u)
          })

setGeneric("siteNames", function(object, ...)
  standardGeneric("siteNames"))

#' Get names of columns in site table
#'
#' @description Get names of columns in site table.
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases siteNames
#' @rdname siteNames
#' @export
setMethod("siteNames", signature(object = "SoilProfileCollection"),
          function(object) {
            res <- names(object@site)
            return(res)
          })

setGeneric("horizonNames", function(object, ...)
  standardGeneric("horizonNames"))

#' Get names of columns in horizon table
#'
#' @description Get names of columns in horizon table.
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases horizonNames
#' @rdname horizonNames
#' @export
setMethod("horizonNames", signature(object = "SoilProfileCollection"),
          function(object) {
            res <- names(object@horizons)
            return(res)
          })

.hzMetadataNames <- function(object, depths = FALSE, ...) {
  hzd <- character(0)
  if (depths) {
    hzd <- horizonDepths(object)
  }
  idn <- c(
    idname(object),
    hzidname(object),
    hzd,
    hzdesgnname(object),
    GHL(object),
    hztexclname(object)
  )
  idn[!is.na(idn) & nchar(idn) > 0]
}

.siteMetadataNames <- function(object, idname = TRUE, ...) {
  md <- aqp::metadata(object)
  if (isTRUE(idname)) {
    mdn <- idname(object)
  } else {
    mdn <- NULL
  }
  if (!is.null(md$coordinates)) {
    mdn <- c(mdn, md$coordinates)
  }
  mdn
}

setGeneric("hzMetadata", function(object, ...)
  standardGeneric("hzMetadata"))

#' Get horizon-level metadata
#'
#' Get `idname(object)` and `hzidname(object)`, with `hzdesgnname(object)`, `hztexclname(object)` (if defined)
#'
#' @param object a SoilProfileCollection
#' @docType methods
#' @aliases hzMetadata
#' @rdname hzMetadata
#' @export
setMethod("hzMetadata", signature(object = "SoilProfileCollection"), 
          function(object) {
            idn <- .hzMetadataNames(object)
            .data.frame.j(horizons(object),
                          col.names = idn[idn %in% horizonNames(object)],
                          use_class = aqp_df_class(object))
          })

##
## initialize metadata: object modification in-place
##

setGeneric('metadata<-', function(object, value)
  standardGeneric('metadata<-'))

#' Set Metadata for a SoilProfileCollection
#'
#' @param object A SoilProfileCollection
#' @param value A named list (see examples)
#' @aliases metadata<-
#' @rdname metadata
#' @export
#' @examples
#' 
#' data(sp5)
#'
#' # replace default metadata with itself
#' metadata(sp5) <- metadata(sp5)
#'
#' # set new metadata attribute value
#' metadata(sp5)$newvalue <- 'foo'
#'
#' # get metadata attribute
#' metadata(sp5)$newvalue
#'
setReplaceMethod("metadata", signature(object = "SoilProfileCollection"),
                 function(object, value) {
                   # otherwise assign
                   object@metadata <- value
                   
                   # done
                   return(object)
                 })

##
## initialize depth_units: object modification in-place, depth_units stored in @metadata
##
setGeneric('depth_units<-', function(object, value)
  standardGeneric('depth_units<-'))

#' Set units of measurement for profile depth
#'
#' @param object A SoilProfileCollection
#' @param value character, a value representing units. Default \code{'cm'}.
#' @aliases depth_units<-
#' @rdname depth_units
#' @export
#' @examples
#'
#' data(sp5)
#'
#' ## get depth units
#' du <- depth_units(sp5)
#'
#' # set alternate units; e.g. inches
#' depth_units(sp5) <- 'in'
#'
#' # replace original value (cm)
#' depth_units(sp5) <- du
setReplaceMethod("depth_units", signature(object = "SoilProfileCollection"),
                 function(object, value) {

                   # quick sanity check: character, length 1

                   # keep existing metadata
                   md <- metadata(object)

                   # default depth_units are always in metadata
                   # replace what ever is there
                   md$depth_units <- value

                   # replace metadata
                   metadata(object) <- md

                   # done
                   return(object)
                 }
)

## TODO: tests https://github.com/ncss-tech/aqp/issues/180

#' Segmenting of Soil Horizon Data by Depth Interval
#' 
#' This function segments or subdivides horizon data from a `SoilProfileCollection` or `data.frame` by depth interval (e.g. `c(0, 10)`, `c(0, 50)`, or `25:100`). This results in horizon records being split at the specified depth intervals, which duplicates the original horizon data but also adds new horizon depths. In addition, labels (i.e. `"segment_id"`) are added to each horizon record that correspond with their depth interval (e.g. `025-100`). This function is intended to harmonize horizons to a common support (i.e. depth interval) for further aggregation or summary. See the examples.
#' @param object either a `SoilProfileCollection` or `data.frame`
#' @param intervals a vector of integers over which to slice the horizon data (e.g. `c(25, 100)` or `25:100`)
#' @param trim logical, when `TRUE` horizons in `object` are truncated to the min/max specified in `intervals`. When `FALSE`, those horizons overlapping an interval are marked as such. Care should be taken when specifying more than one depth interval and `trim = FALSE`.
#' @param depthcols a character vector of length 2 specifying the names of the horizon depths (e.g. `c("top", "bottom")`), only necessary if `object` is a 
#' @param hzdepcols deprecated being replaced by depthcols.
#' 
#' @details `hz_segment()` performs no aggregation or resampling of the source data, rather, labels are added to horizon records for subsequent aggregation or summary. This makes it possible to process a very large number of records outside of the constraints associated with e.g. `slice()` or `slab()`.
#'
#' @return Either a `SoilProfileCollection` or `data.frame` with the original horizon data segmented by depth intervals. There are usually more records in the resulting object, one for each time a segment interval partially overlaps with a horizon. A new column called `segment_id` identifying the depth interval is added.
#' 
#' @author Stephen Roecker
#' 
#' @seealso [dice()], [glom()], [hz_dissolve()], [hz_lag()], [hz_intersect()]
#' 
#' @export
#'
#' @examples
#' 
#' # example data
#' data(sp1)
#' 
#' # upgrade to SPC
#' depths(sp1) <- id ~ top + bottom
#' 
#' # segment and trim
#' z <- hz_segment(sp1, intervals = c(0, 10, 20, 30), trim = TRUE)
#' 
#' # display segment labels
#' # note that there are new horizon boundaries at segments
#' par(mar = c(0, 0, 3, 1))
#' plotSPC(z, color = 'segment_id', width = 0.3)
#' 
#' # highlight new horizon records
#' par(mar = c(0, 0, 2, 1))
#' plotSPC(z, color = NA, default.color = NA, width = 0.3, lwd = 1)
#' plotSPC(sp1, color = NA, default.color = NA, 
#' width = 0.3, lwd = 3, add = TRUE, name = NA, print.id = FALSE)
#' legend('top', horiz = TRUE, 
#' legend = c('original', 'segmented'), 
#' lwd = c(1, 3), cex = 0.85, bty = 'n')
#' 
#' \donttest{
#' # same results as slab()
#' # 10 random profiles
#' s <- lapply(1:10, random_profile, n_prop = 1, SPC = TRUE, method = 'random_walk')
#' s <- combine(s)
#' 
#' a.slab <- slab(s, fm = ~ p1, slab.structure = c(0, 10, 20, 30), slab.fun = mean, na.rm = TRUE)
#' 
#' z <- hz_segment(s, intervals = c(0, 10, 20, 30), trim = TRUE)
#' z <- horizons(z)
#' z$thick <- z$bottom - z$top
#' 
#' a.segment <- sapply(split(z, z$segment_id), function(i) {
#'   weighted.mean(i$p1, i$thick)
#' })
#' 
#' 
#' res <- data.frame(
#'   slab = a.slab$value,
#'   segment = a.segment,
#'   diff = a.slab$value - a.segment
#' )
#'
#' print(res)
#' res$diff < 0.001
#'}
#' 
#' 
#' data(sp5)
#' 
#' # segment by upper 25-cm
#' test1 <- hz_segment(sp5, intervals = c(0, 100))
#' print(test1)
#' nrow(test1)
#' print(object.size(test1), units = "Mb")
#' 
#' # segment by 1-cm increments
#' test2 <- hz_segment(sp5, intervals = 0:100)
#' print(test2)
#' nrow(test2)
#' print(object.size(test2), units = "Mb")
#' 
#' 
#' # segment and aggregate
#' test3 <- hz_segment(horizons(sp5), 
#'                  intervals = c(0, 5, 15, 30, 60, 100, 200), 
#'                  depthcols = c("top", "bottom")
#' )
#' test3$hzthk <- test3$bottom - test3$top
#' test3_agg <- by(test3, test3$segment_id, function(x) {
#'   data.frame(
#'     hzID = x$hzID[1],
#'     segment_id = x$segment_id[1],
#'     average = weighted.mean(x$clay, w = x$hzthk)
#'   )
#' })
#' test3_agg <- do.call("rbind", test3_agg)
#' 
#' head(test3_agg)
#' 
hz_segment <- function(object, intervals, trim = TRUE, depthcols = c("top", "bottom")) {

  # depth interval rules
  dep <- data.frame(
    top = intervals[-length(intervals)],
    bot = intervals[-1],
    stringsAsFactors = FALSE
  )
  
  n <- max(nchar(intervals))
  dep$id <- paste0(formatC(dep$top, width = n, flag = 0),
                   "-",
                   formatC(dep$bot, width = n, flag = 0))
  
  # argument sanity check ----
  test_spc <- inherits(object, 'SoilProfileCollection')
  test_df  <- inherits(object, 'data.frame')
  test_dep <- is.numeric(dep$top) & is.numeric(dep$bot) & all(dep$top < dep$bot)
  
  if (!any(test_spc, test_df)) {
    stop("the input must be either a SoilProfileCollection or data.frame")
  }
  
  .check_depthcols_l(depthcols)
  
  if (!test_dep) {
    stop("intervals should be numeric and sequential (e.g. c(0, 1, 2, 3) or 0:100)")
  }
  
  # standardize inputs ----
  if (test_spc) {
    idcol     <- idname(object)
    hzidcol   <- hzidname(object)
    depthcols <- horizonDepths(object)
    h  <- horizons(object)
    names(h)[names(h) %in% c(idcol, hzidcol)] <- c("idcol", "hzidcol")
  } else {
    h <- object
  }
  names(h)[names(h) %in% depthcols] <- c("top", "bot")
  
  ## TODO: consider using dice()
  # filter horizons and trim ----
  .slice <- function(h, top = NULL, bot = NULL) {
    idx <- which(h$top < bot & h$bot > top)
    h <- h[idx, ]
    
    # causing errors when FALSE; fixed?
    if (trim == TRUE) {
      h$top = ifelse(h$top < top, top, h$top)
      h$bot = ifelse(h$bot > bot, bot, h$bot)
      
      # h <- h[(h$bot - h$top) > 0, ]
    }
    
    # h <- h[!is.na(h$peiid), ]
    
    return(h)
  }
  
  # slice spc by intervals ----
  # dep$df <- lapply(1:nrow(dep), function(x) h[0, ]) # pre-allocate memory
  df_str <- cbind(h[0, ], segment_id = NA_character_[0])
  dep$df <- list(df_str)[rep(1, nrow(dep))] # pre-allocate memory faster
  h <- {
    split(dep, dep$id) -> .
    
    lapply(., function(x) {
      temp <- .slice(h, top = x$top, bot = x$bot)
      if (nrow(temp) > 0)
        x$df[[1]] <- cbind(temp, segment_id = x$id)
      return(x$df[[1]])
    }) -> .
    
    do.call("rbind", .) -> .
  }
  names(h)[names(h) %in% c("top", "bot")] <- depthcols
  
  
  if (test_spc) {
    h <- h[order(h$idcol, h[[depthcols[1]]]), ]
    
    # merge to re-add spc with NA
    h_orig <- data.frame(idcol = names(table(.data.frame.j(horizons(object), idcol, use_class = aqp_df_class(object)))), stringsAsFactors = FALSE)
    h <- merge(h_orig, h, by = "idcol", all.x = TRUE, sort = FALSE)
    rm(h_orig)
    
    ## TODO: consider adding a flag to indicate "new" horizon records that have been added
    
    # rebuild SPC ----
    names(h)[names(h) == "idcol"]   <- idcol
    names(h)[names(h) == "hzidcol"] <- hzidcol
    h$hzID <- 1:nrow(h)
    
    replaceHorizons(object) <- h
    suppressMessages(hzidname(object) <- "hzID")
  }
  
  # return
  if(test_spc){
    return(suppressMessages(rebuildSPC(object))) # TODO: this is protection from missing-data/ID offset
  } else {
    return(h)
  }
  
}

#' @export 
#' @rdname hz_segment
segment <- function(object, intervals, trim = TRUE, hzdepcols = c("top", "bottom")) {
  .Deprecated("segment() is deprecated and has been replaced by hz_segment()")
  hz_segment(object, intervals, trim, depthcols = hzdepcols)
}

#' @title Dissolving horizon boundaries by grouping variables
#' 
#' @description This function dissolves or combines horizons that have a common set of grouping variables. It only combines those horizon records that are sequential (e.g. share a horizon boundary). Thus, it can be used to identify discontinuities in the grouping variables along a profile and their unique depths. It is particularly useful for determining the depth to the top or bottom of horizons with a specific category, and should be simpler than previous methods that require aggregating over profiles. 
#'
#' @param object a \code{data.frame}
#' @param by character: column names, to be used as grouping variables, within the object.
#' @param idcol character: column name of the pedon ID within the object.
#' @param depthcols a character vector of length 2 specifying the names of the horizon depths (e.g. `c("top", "bottom")`).
#' @param id deprecated and replaced with idcol.
#' @param hztop deprecated and replaced by depthcols.
#' @param hzbot deprecated and replaced by depthcols.
#' @param collapse logical: indicating whether to not combine grouping variables before dissolving.
#' @param order logical: indicating whether or not to order the object by the id, hztop, and hzbot columns. 
#'
#' @details This function assumes the profiles and horizons within the object follow the logic defined by \code{checkHzDepthLogic} (e.g. records are ordered sequentially by id, hztop, and hzbot and without gaps). If the records are not ordered, set the \code{order = TRUE}.
#'
#' @return A \code{data.frame} with the original idcol, by grouping variables, and non-consecutive horizon depths. 
#' 
#' @author Stephen Roecker
#' 
#' @seealso [hz_lag()], [hz_intersect()], [hz_segment()] , [checkHzDepthLogic()]
#' 
#' @export
#'
#' @examples
#' 
#' # example 1
#' data(jacobs2000)
#' spc <- jacobs2000
#' 
#' spc$dep_5 <- spc$depletion_pct >=5
#' spc$genhz <- generalize.hz(spc$name, c("A", "E", "B", "C"), c("A", "E", "B", "C")) 
#' h <- horizons(spc)
#' 
#' test <- hz_dissolve(h, by = c("genhz", "dep_5"), idcol = "id", depthcols = c("top", "bottom"))
#' 
#' vars <- c("id", "top", "bottom", "genhz", "dep_5")
#' h[h$id == "92-1", vars]
#' test[test$id == "92-1", ]
#' 
#' 
#' # example 2
#' df <- data.frame(
#'     id = 1,
#'     top    = c(0, 5,  10, 15, 25, 50), 
#'     bottom = c(5, 10, 15, 25, 50, 100),
#'     hzname = c("A1",  "A2",  "E/A", "2Bt1", "2Bt2", "2C"),
#'     genhz  = c("A",   "A",   "E",   "2Bt",  "2Bt", "2C"),
#'     texcl  = c("sil", "sil", "sil", "sl",   "sl",   "s")
#'     )
#' 
#' df
#' 
#' hz_dissolve(df, c("genhz", "texcl"))
#' hz_dissolve(df, c("genhz", "texcl"), collapse = TRUE)
#' 
#' test <- hz_dissolve(df, "genhz")
#' subset(test, value == "2Bt")
#' 
hz_dissolve <- function(object, by, idcol = "id", depthcols = c("top", "bottom"), collapse = FALSE, order = FALSE) {
  
  # id = "peiid"; hztop = "hzdept"; hzbot = "hzdepb", collapse = FALSE, order = FALSE
  
  # test inputs ----
  # argument sanity check
  # test_spc <- inherits(object, 'SoilProfileCollection')
  
  # check that object & by are the right class
  test_object <- inherits(object, "data.frame")
  if (!any(test_object)) {
    stop("the object argument must be a data.frame", call. = FALSE)
  }
  
  # check that collapse is a logical of length 1
  if (!inherits(collapse, "logical") || length(collapse) != 1) {
    stop("the collapse argument must be logical and a length of one", call. = FALSE)
  }
  
  # check that by is not NULL
  if (is.null(by)) stop("the by argument must not be NULL")
  
  # check that "by" are characters or convert
  if (any(!"character" %in% sapply(object[by], class))) {
    message("non-character grouping variables are being converted to characters")
    object[by] <- lapply(object[by], as.character)
  }
  
  # check that the column names exist within the object
  .check_names(object, vars = c(idcol = idcol, top = depthcols[1], bot = depthcols[2], by))
  
  # check if previous dissolve_id exists and overwrite
  nm  <- names(object)
  idx <- nm == "dissolve_id"
  if (any(idx)) {
    warning("object contains an existing column named 'dissolve_id', it will be overwritten") 
    object[idx] <- NULL
  }
  
  # standardize inputs ----
  df_std <- .standardize_inputs(object, idcol = idcol, depthcols = depthcols)
  df_conversion <- df_std$x_conversion
  df <- df_std$x; rm(df_std)
  
  
  # valid
  # vd_idx <- validate_depths(df, id = "id", hztop = "hzdept", bot = "hzdepb")
  if (order == TRUE) {
    df <- df[order(df$idcol, df$top, df$bot), ]
  }
  
  if (collapse == TRUE) {
    by_co <- paste(by, collapse = " & ")
    df[by_co] <- apply(df[by], 1, paste, collapse = " & ")
    by    <- by_co
  }
  
  # var thickness ----
  var_dep <- lapply(by, function(x) {
    
    con_bot <- rle(    paste(df$idcol, df[, x]))$length
    con_top <- rle(rev(paste(df$idcol, df[, x])))$length
    
    bot_idx <- cumsum(con_bot)
    top_idx <- cumsum(con_top)
    
    vd <- data.frame(
      idcol = df[bot_idx, "idcol"],
      top   = rev(rev(df$top)[top_idx]),
      bot   = df[bot_idx, "bot"],
      variable = x,
      value = df[bot_idx,    x]
    )
    # vd$dissolve_id = 1:nrow(vd)
    
    return(vd)
  })
  var_dep <- do.call("rbind", var_dep)
  
  
  # this is redundant with collapse = TRUE, and inappropriate unless the grouping by variable matches across all horizon depths, otherwise it'll generate pedons with overlapping depths
  # if (direction == "wide") {
  #   var_dep <- reshape(
  #     var_dep,
  #     direction = "wide",
  #     idvar = c("id", "top", "bot"),
  #     timevar = "variable",
  #     v.names = "value"
  #   )
  # }
  
  # append dissolve_id
  n <- max(nchar(c(var_dep$top, var_dep$bot)), na.rm = TRUE)
  
  var_dep$dissolve_id <- paste0(
    var_dep$idcol,
    "_",
    formatC(var_dep$top, width = n, flag = 0), 
    "-", 
    formatC(var_dep$bot, width = n, flag = 0),
    "_",
    var_dep$value
  )
  
  
  # reset inputs ----
  var_dep <- .reset_inputs(var_dep, df_conversion)
  
  return(var_dep)
}


#' @export 
#' @rdname hz_dissolve

dissolve_hz <- function(object, by, id = "idcol", hztop = "top", hzbot = "bottom", collapse = FALSE, order = FALSE) {
  .Deprecated("dissolve_hz() is deprecated and has been replaced by hz_dissolve()")
  hz_dissolve(object, by, idcol = id, depthcols = c(hztop, hzbot), collapse, order)
}



#' @title Intersecting horizon boundaries by horizon depths
#' 
#' @description This function intersects two horizon tables by harmonizing their depths and merging them where they overlap. This can be useful to rejoin the results of `hz_dissolve()` to it's original horizon table, and then perform an aggregation on the dissolved variables.
#'
#' @param x a \code{data.frame}
#' @param y a \code{data.frame}
#' @param idcol character: column name of the pedon ID within the object.
#' @param depthcols a character vector of length 2 specifying the names of the horizon depths (e.g. `c("top", "bottom")`).
#'
#' @details .
#'
#' @return A \code{data.frame} with harmonized depth intervals (i.e. segment_id) and columns from both of the original \code{data.frame}. If both \code{data.frame} contain the same column names, they will both be returned (with the exception of the idcol and depthcols), and appended with either x or y to indicate which \code{data.frame} they originated from. 
#' 
#' @author Stephen Roecker
#' 
#' @seealso [hz_dissolve()], [hz_lag()], [hz_segment()] 
#' 
#' @export
#'
#' @examples
#' 
#' h <- data.frame(
#'   id = 1,
#'   top    = c(0,  25, 44, 46, 50),
#'   bottom = c(25, 44, 46, 50, 100),
#'   by     = c("Yes", "Yes", "No", "No", "Yes"),
#'   clay   = c(10, 12, 27, 35, 16)
#' )
#' 
#' hz_dissolve(h, "by")
#' 
#' hz_intersect(x = hz_dissolve(h, "by"), y = h)
#' 
#' hi <- hz_intersect(x = h, y = hz_dissolve(h, "by"))
#' aggregate(clay ~ dissolve_id, data = hi, mean)
#' 
hz_intersect <- function(x, y, idcol = "id", depthcols = c("top", "bottom")) {
  
  # test inputs ----
  # argument sanity check
  
  # check that depthcols ----
  ## length == 2
  if (length(depthcols) != 2) stop("depthcols must length must equal 2")
  
  ## check for matching column names
  .check_names(x, c(idcol, depthcols))
  .check_names(y, c(idcol, depthcols))
  
  # check segment_id ----
  ## if it exists, overwrite it
  x_nm <- names(x)
  y_nm <- names(y)
  if (any(x_nm %in% "segment_id")) {
    warning("x includes a column named 'segment_id', it will be overwritten")
    x[x_nm == "segment_id"] <- NULL
  }
  
  if (any(y_nm %in% "segment_id")) {
    warning("y includes a column named 'segment_id', it will be overwritten")
    y[y_nm == "segment_id"] <- NULL
  }
  
  # standardize inputs ----
  x_std <- .standardize_inputs(x, idcol = idcol, depthcols = depthcols)
  x_conversion <- x_std$x_conversion
  x            <- x_std$x; rm(x_std)
  
  y <- .standardize_inputs(y, idcol = idcol, depthcols = depthcols)$x
  
  # intersect x & y ----
  res <- lapply(split(x, x$idcol), function(x) {
    xi <- x
    yi <- y[which(y$idcol == xi$idcol[1]), ]
    
    if (nrow(yi) > 0) {
      
      int <-  unique(sort(c(xi$top, xi$bot, yi$top, yi$bot)))
      
      xi_seg <- hz_segment(xi, intervals = int, depthcols = names(x_conversion[2:3]), trim = TRUE)
      yi_seg <- hz_segment(yi, intervals = int, depthcols = names(x_conversion[2:3]), trim = TRUE)
      
      return(list(x_seg = xi_seg, y_seg = yi_seg))
    }
  })
  
  x_seg <- do.call("rbind", lapply(res, function(x) x[["x_seg"]]))
  y_seg <- do.call("rbind", lapply(res, function(x) x[["y_seg"]]))
  
  xy_int <- merge(x_seg, y_seg, by = c("segment_id", "idcol", "top", "bot"), sort = FALSE)
  
  # reset inputs ----
  xy_int <- .reset_inputs(xy_int, x_conversion)
  
  return(xy_int)
}



#' @title Find lagged horizon values
#' 
#' @description This function finds adjacent values to a horizon values at lagged distances.
#'
#' @param object a \code{data.frame}
#' @param lag integer: number of horizons to lag
#' @param unit character: lag units in index or depth.
#' @param idcol character: column name of the pedon ID within the object.
#' @param depthcols a character vector of length 2 specifying the names of the horizon depths (e.g. `c("top", "bottom")`).
#' @param order logical: indicating whether or not to order the #'
#' @details .
#'
#' @return A \code{data.frame} with lagged values. 
#' 
#' @author Stephen Roecker
#' 
#' @seealso [hz_dissolve()], [hz_intersect()], [hz_segment()] 
#' 
#' @export
#'
#' @examples
#' 
#' h <- data.frame(
#'   id = 1,
#'   top    = c(0,  25, 44, 46, 50),
#'   bottom = c(25, 44, 46, 50, 100),
#'   texcl     = c("SL", "SL", "CL", "CL", "L"),
#'   clay   = c(10, 12, 27, 35, 16)
#' )
#' 
#' hz_lag(h)
#' 
#' hz_lag(h, -1)
#' 
#' hz_lag(h, 10:15, unit = "depth")
#' 
#' transform(cbind(h, lag = hz_lag(h)), 
#'   clay_dif = lag.clay_bot.1 - clay,
#'   texcl_contrast = paste0(texcl, "-", lag.texcl_bot.1)
#' )
#' 
hz_lag <- function(object, lag = 1, unit = "index", idcol = "id", depthcols = c("top", "bottom"), order = FALSE) {
  
  nm <- names(object)
  idx_std <- which(! nm %in% c(idcol, depthcols))
  vars <- nm[idx_std]
  
  # check arguments ----
  .check_depthcols_l(depthcols)
  .check_names(object, vars = c(idcol, depthcols, vars))
  
  # standardize inputs ----
  x_std <- .standardize_inputs(object, idcol = idcol, depthcols = depthcols)
  x_conversion <- x_std$x_conversion
  x <- x_std$x
  rm(x_std)
  
  # check depths ---
  if (unit == "depth" & max(object[[depthcols[2]]] > 1000)) {
    warning("The maximum depth is greater than 1000, which is implausible and will be removed. To avoid this action either remove the offending horizon or convert the depth units to a measure which will not exceed 1000")
    x <- x[x$bot < 1000, ]
  }
  
  test <- max(aggregate(top ~ idcol, data = x, length)$top)
  if (unit == "index") {
    if ((test - 1) < max(lag)) {
    stop("lag can not be greater than the maximum number of horizons")
    }
  }
  
  # order ----
  if (order) {
    x <- x[order(x$idcol, x$top, x$bot), ]
  }
  
  # lag ----
  .lag_ind <- function(x, lag = lag) {
    
    nr  <- nrow(x)
    top <- 1:nr
    if (lag >= 0) bot <- c((1 + lag):nr, rep(NA, lag))
    if (lag <  0) bot <- c(rep(NA, abs(lag)), 1:(nr + lag))
    
    test_idcol <- x$idcol[top] == x$idcol[bot]
    test_idcol <- ifelse(! test_idcol, NA, TRUE)
    x_lag <- x[test_idcol * bot, vars]
    if (lag >= 0) names(x_lag) <- paste0(vars, "_bot.",     lag)
    if (lag <  0) names(x_lag) <- paste0(vars, "_top.", abs(lag))

    return(x_lag)
  }
  
  .lag_dep <- function(x, lag = lag) {
    
    n <- length(x)
    x$.ID <- 1:nrow(x)
    x_seg <- hz_segment(x, intervals = min(x$top):max(x$bot), trim = TRUE, depthcols = c("top", "bot"))
    x_seg <- x_seg[1:(n + 1)]
    
    x_seg <- do.call("cbind", args =  lapply(lag, function(i) {
      x$bot_i <- x$bot + i
      idx <- match(paste(x$idcol, x$bot_i),
                   paste(x_seg$idcol, x_seg$bot))
      xi_seg <- x_seg[idx,]
      xi_seg <- x[xi_seg$.ID, vars, drop = FALSE]
      xi_seg$.ID <- NULL
      
      if (i >= 0)
        names(xi_seg) <- paste0(names(xi_seg), "_bot.",     i)
      
      if (i <  0)
        names(xi_seg) <- paste0(names(xi_seg), "_top.", abs(i))
      
      return(xi_seg)
    }))
    
    return(x_seg)
  }
  
  if (unit == "index") {
    x_lag <- do.call("cbind", lapply(lag, function(i) {
      .lag_ind(x, i)
    }))
    x_lag <- x_lag[sort(names(x_lag))]
  }
  
  if (unit == "depth") {
    x_lag <- .lag_dep(x, lag)
    x_lag <- x_lag[sort(names(x_lag))]
  }
  
  # # reset inputs ----
  x_lag <- .reset_inputs(x_lag, x_conversion)
  
  return(x_lag)
}

# check depthcols length
.check_depthcols_l <- function(x) {
  if (length(x) != 2 & !is.null(x)) stop("depthcols must length must equal 2")
}

## check for matching column names
.check_names <- function(x, vars) {
  x_nm <- names(x)
  
  if (!all(vars %in% x_nm)) {
    stop("x must contain columns with names that match the input arguments")
  }
}

# standardize inputs
.standardize_inputs <- function(
    x,
    idcol = NULL,
    hzidcol = NULL,
    depthcols = NULL,
    texcl = NULL,
    clay = NULL,
    taxpartsize = NULL,
    sand = NULL
) {
  
  # set new names
  var_names <- c(
    idcol   = idcol, 
    hzidcol = hzidcol, 
    top     = depthcols[1], 
    bot     = depthcols[2],
    texcl   = texcl,
    clay    = clay,
    sand    = sand,
    taxpartsize = taxpartsize
  )
  
  # find matches
  idx_x <- sapply(var_names, function(i) which(names(x) == i))
  
  # rename matching column names
  names(x)[idx_x] <- names(var_names)
  
  # remove duplicate names
  nm_x <- names(x)
  idx_dup <- names(which(table(nm_x) > 1))
  idx_dup <- which(nm_x %in% idx_dup)
  if (any(!is.na(idx_dup))) {
    warning("some argument names are duplicated by the function column name harmonization and will be renamed to var_orig (e.g. clay_orig)")
    
    idx_orig <- idx_dup[! idx_dup %in% idx_x]
    names(x)[idx_orig] <- paste0(names(x)[idx_orig], "_orig")
  } else {
    idx_orig <- NULL
  }
    
  return(list(x = x, x_conversion = var_names, x_orig = idx_orig))
}


.reset_inputs <- function(x, conversion) {
  
  # find original names
  idx <- which(names(x) %in% names(conversion))
  
  # reset original names
  names(x)[idx] <- conversion
  
  return(x)
}
  

context("hz_segment")


test_that("data.frame interface works as expected", {
  
  # init local copy of sample data
  data(sp1)
  
  # trimming
  z <- hz_segment(sp1, intervals = c(0, 10, 20, 30), trim = TRUE, depthcols = c('top', 'bottom'))
  
  # correct object type and segment label
  expect_true(inherits(z, 'data.frame'))
  expect_true('segment_id' %in% names(z))
  
  # label class
  expect_true(inherits(z[['segment_id']], 'character'))
  
  # no triming
  z <- hz_segment(sp1, intervals = c(0, 10, 20, 30), trim = FALSE, depthcols = c('top', 'bottom'))
  
  # correct object type and segment label
  expect_true(inherits(z, 'data.frame'))
  expect_true('segment_id' %in% names(z))
  
  # label class
  expect_true(inherits(z[['segment_id']], 'character'))
})


test_that("SPC interface works as expected", {
  
  # init local copy of sample data
  data(sp1)
  depths(sp1) <- id ~ top + bottom
  
  # trimming
  z <- hz_segment(sp1, intervals = c(0, 10, 20, 30), trim = TRUE)
  
  expect_true(inherits(z, 'SoilProfileCollection'))
  expect_true('segment_id' %in% horizonNames(z))
  
  # label class
  expect_true(inherits(z[['segment_id']], 'character'))
  
  # no trimming
  z <- hz_segment(sp1, intervals = c(0, 10, 20, 30), trim = FALSE)
  
  expect_true(inherits(z, 'SoilProfileCollection'))
  expect_true('segment_id' %in% horizonNames(z))
  
  # label class
  expect_true(inherits(z[['segment_id']], 'character'))
  
})



test_that("expected outcome with NA horizon depths", {
  
  # init local copy of sample data
  data(sp1)
  
  # copies
  good <- sp1
  bad <- sp1
  
  # add NA to horizon depths
  bad$top[c(1, 5)] <- NA
  
  # segment
  z.bad <- hz_segment(bad, intervals = c(0, 10, 20, 30), trim = TRUE, depthcols = c('top', 'bottom'))
  z.good <- hz_segment(good, intervals = c(0, 10, 20, 30), trim = TRUE, depthcols = c('top', 'bottom'))
  
  # label class
  expect_true(inherits(z.good[['segment_id']], 'character'))
  expect_true(inherits(z.bad[['segment_id']], 'character'))
  
  ## TODO: is this expected?
  # row count
  expect_false(nrow(z.good) == nrow(z.bad))
  
  # same values
  # expect_false(all(z.good$segment_id == z.bad$segment_id))
  
})


# I think this test needs to be retired or reframed. segment previously exclude results where the thickness of the segment was zero. That feature has been removed. Upon updating segment it was removed to ensure the original data was returned regardless of the horizon errors, which should be dealt with elsewhere.
# test_that("expected outcome with bogus horizon depths", {
#   
#   # init local copy of sample data
#   data(sp1)
#   
#   # copies
#   good <- sp1
#   bad <- sp1
#   
#   # add NA to horizon depths
#   bad$top[c(1, 5)] <- bad$bottom[c(1, 5)]
#   
#   # segment
#   z.bad <- hz_segment(bad, intervals = c(0, 10, 20, 30), trim = TRUE, depthcols = c('top', 'bottom'))
#   z.good <- hz_segment(good, intervals = c(0, 10, 20, 30), trim = TRUE, depthcols = c('top', 'bottom'))
#   
#   # label class
#   expect_true(inherits(z.good[['segment_id']], 'character'))
#   expect_true(inherits(z.bad[['segment_id']], 'character'))
#   
#   ## TODO: is this expected?
#   # row count
#   expect_false(nrow(z.good) == nrow(z.bad))
#   
#   # same values
#   # expect_false(all(z.good$segment_id == z.bad$segment_id))
#   
# })




test_that("same results as weighted mean via slab", {

  # 100 random data
  s <- lapply(1:100, random_profile, n_prop = 1, SPC = TRUE, method = 'random_walk')
  s <- combine(s)
  
  # weighted mean via slab
  a.slab <- slab(s, fm = ~ p1, slab.structure = c(0, 10, 20, 30), slab.fun = mean, na.rm = TRUE)
  
  # segment
  z <- hz_segment(s, intervals = c(0, 10, 20, 30), trim = TRUE)
  
  # compute horizon thickness weights
  z <- horizons(z)
  z$thick <- z$bottom - z$top
  
  # weighted mean from segment output
  a.segment <- sapply(split(z, z$segment_id), function(i) {
    weighted.mean(i$p1, i$thick)
  })
  
  # inspect as needed
  res <- data.frame(
    slab = a.slab$value,
    segment = a.segment,
    diff = a.slab$value - a.segment
  )
  
  expect_true(all(res$diff < 0.001))
  
})

## ideas:

# 2016-03-21
# may not work correctly with compositions
# good idea + code here:
# http://stackoverflow.com/questions/31893559/r-adding-alpha-bags-to-a-2d-or-3d-scatterplot
#

## TODO:
## 1. allow iteration over groups for plotting multiple horizons
## 2. consider embedding into lattice panels

# helper function
.get.ssc.low.rv.high <- function(ssc, p, delta, TT.obj) {

  if(!requireNamespace("soiltexture", quietly = TRUE))
    stop("package `soiltexture` is required", call.=FALSE)

  # sanity checks
  if(delta < 1)
    stop('delta values smaller than 1 may result in memory overflow', call.=FALSE)

  # compute _univariate_ low - rv - high by variable
  ssc.stats <- apply(ssc, 2, Hmisc::hdquantile, probs=p, na.rm=TRUE)

  # re-order for plotting
  sand.stats <- sort(ssc.stats[, 1])
  silt.stats <- sort(ssc.stats[, 2])
  clay.stats <- sort(ssc.stats[, 3])

  # make a grid of sand, silt, clay values within the low--high range of the data
  sand.seq <- round(seq(from=sand.stats[1], to=sand.stats[3], by=delta))
  silt.seq <- round(seq(from=silt.stats[1], to=silt.stats[3], by=delta))
  clay.seq <- round(seq(from=clay.stats[1], to=clay.stats[3], by=delta))
  g <- expand.grid(SAND=sand.seq, SILT=silt.seq, CLAY=clay.seq)

  # subset to only include those sand, silt, clay values that sum to 100% (0.01% tolerance)
  real.textures <- which(abs(apply(g, 1, sum) - 100) < 0.01 )
  g <- g[real.textures, ]

  # plot low and high values with no symbol, so that we can access the {x,y} screen coordinates
  tp.low.high <- soiltexture::TT.points(tri.data = g, geo = TT.obj, pch = NA, tri.sum.tst = FALSE)

  # create range polygon
  poly.order <- chull(tp.low.high$x, tp.low.high$y)

  # package quantiles and bounding polygon geometry
  res <- list(
    stats = ssc.stats, 
    range = list(
      x = tp.low.high$x[poly.order], 
      y = tp.low.high$y[poly.order]
    )
  )
  
  return(res)
}


# compute and plot "low"--"representative value"--"high" soil textures based on:
# ssc: data.frame/matrix of [sand, silt, clay]
# p: requested percentiles


#' 
#' @title Soil Texture Low-RV-High as Defined by Quantiles
#' 
#' @author D.E. Beaudette, J. Nemecek, K. Godsey
#' 
#' @description This function accepts soil texture components (sand, silt, and clay percentages) and plots a soil texture triangle with a "representative value" (point) and low-high region (polygon) defined by quantiles (estimated with \code{Hmisc::hdquantile}). Marginal quantiles of sand, silt, and clay are used to define the boundary of a low-high region. The default settings place the RV symbol at the texture defined by marginal medians of sand, silt, and clay. The default low-high region is defined by the 5th and 95th marginal percentiles of sand, silt, and clay.
#' 
#' @param ssc \code{data.frame} with columns: 'SAND', 'SILT', 'CLAY', values are percentages that should add to 100. No NA allowed.
#' 
#' @param p vector of percentiles (length = 3) defining 'low', 'representative value', and 'high'
#' 
#' @param delta grid size used to form low-high region
#' 
#' @param rv.col color used for representative value (RV) symbol
#' 
#' @param range.border color used for polygon border enclosing the low-high region
#' 
#' @param range.col color used for polygon enclosing the low-high region
#' 
#' @param range.alpha transparency of the low-high range polygon (0-255)
#' 
#' @param range.lty line style for polygon enclosing the low-high region
#' 
#' @param range.lwd line weight polygon enclosing the low-high region
#' 
#' @param main plot title
#' 
#' @param legend.cex scaling factor for legend
#' 
#' @param legend logical, enable/disable automatic legend
#' 
#' @param \dots further arguments passed to \code{soiltexture::TT.points}
#' 
#' @return an invisible \code{matrix} with marginal percentiles of sand, silt, and clay 
#' 
#' @seealso \code{\link{bootstrapSoilTexture}}
#' 
#' @keywords hplots
#' @export
#' @examples
#' 
#' \donttest{
#' 
#' if(
#' requireNamespace("Hmisc") &
#'   requireNamespace("compositions") &
#'   requireNamespace("soiltexture")
#' ) {
#'   
#'   # sample data
#'   data('sp4')
#'   
#'   # subset rows / columns
#'   ssc <- sp4[grep('^Bt', sp4$name), c('sand', 'silt', 'clay')]
#'   names(ssc) <- toupper(names(ssc))
#'   
#'   # make figure, marginal percentiles are silently returned
#'   stats <- textureTriangleSummary(
#'     ssc, pch = 1, cex = 0.5, 
#'     range.alpha = 50, 
#'     range.lwd = 1,
#'     col = grey(0.5), 
#'     legend = FALSE
#'   )
#'   
#'   # check
#'   stats
#'   
#'   # simulate some data and try again
#'   s <- bootstrapSoilTexture(ssc, n = 100)$samples
#'   
#'   # make the figure, ignore results
#'   textureTriangleSummary(
#'     s, pch = 1, cex = 0.5, 
#'     range.alpha = 50, 
#'     range.lwd = 1,
#'     col = grey(0.5), 
#'     legend = FALSE
#'   )
#' }
#' 
#' }
#' 
textureTriangleSummary <- function(ssc, p = c(0.05, 0.5, 0.95), delta = 1, rv.col = 'red', range.border = 'black', range.col = 'RoyalBlue', range.alpha = 80, range.lty = 1, range.lwd = 2, main = 'Soil Textures', legend.cex = 0.75, legend = TRUE, ...) {

  # sanity check
  if(
    ! requireNamespace("soiltexture", quietly = TRUE) | 
    ! requireNamespace("Hmisc", quietly = TRUE)
  ) {
    stop("packages `Hmisc` and `soiltexture` are required", call.=FALSE)
  }
  
  # TODO: filter NA?
  
  # check for appropriate column names
  # all must be present
  # may be other columns as well, ignore those
  name.check <- sapply(c('SAND', 'SILT', 'CLAY'), function(i) {
    any(names(ssc) %in% i)
  })
  
  if(! all(name.check)) {
    stop('`ssc` must contain columns: `SAND`, `SILT`, `CLAY`.')
  }
  
  # subset via column names
  ssc <- ssc[, c('SAND', 'SILT', 'CLAY')]
  
  
	# setup colors
	range.col <- rgb(t(col2rgb(range.col)), maxColorValue=255, alpha=range.alpha)

	# setup legend elements
  rv.text <- paste0('Sample RV (', paste(p[c(2)], collapse='-'), ')')
  low.high.range.text <- paste0('Low-High Range (', paste(p[c(1,3)], collapse='-'), ')')
	legend.text <- c(rv.text, low.high.range.text)
	legend.cols <- c('black', 'black')
	legend.bg <- c(rv.col, range.col)
	legend.pch <- c(22, 22)
  
	# setup plot
	TT <- soiltexture::TT.plot(
	  class.sys = "USDA-NCSS.TT",    # use "our" texture triangle
	  main = main,          # title
	  tri.sum.tst = FALSE,            # do not test for exact sum(sand, silt, clay) == 100
	  cex.lab = 0.75,                 # scaling of label text
	  cex.axis = 0.75,                # scaling of axis
	  frame.bg.col = 'white',         # background color
	  class.lab.col = 'black',        # color for texture class labels
	  lwd.axis = 1.5,
	  lwd.lab = 2,
	  arrows.show=TRUE
	)
	
	# compute RV / range polygon for data
	res <- .get.ssc.low.rv.high(ssc,  p=p, delta=delta, TT.obj = TT)
	
	# add polgon defining range of data
	polygon(res$range$x, res$range$y, col=range.col, border = range.border, lty = range.lty, lwd = range.lwd)
	
	# add original data, passing in additional arguments
	soiltexture::TT.points(tri.data = ssc, geo = TT, tri.sum.tst=FALSE, lwd = 1, ...)
	
	# plot population RV
	soiltexture::TT.points(
	  tri.data = data.frame(t(res$stats[2, ])), 
	  geo = TT, 
	  bg = rv.col, 
	  pch = 22, 
	  cex = 1.25, 
	  lwd = 1, 
	  tri.sum.tst = FALSE
	  )
	
	# legend
	if(legend) {
	  legend(
	    'topleft', 
	    legend = legend.text, 
	    pt.bg = legend.bg, 
	    pch = legend.pch, 
	    col = legend.cols, 
	    bty = 'n', 
	    cex = legend.cex, 
	    pt.cex = 1.25, 
	    horiz = TRUE
	  ) 
	}
	
  # invisibly return stats
	invisible(res$stats)

}

context("soil texture summary via marginal percentiles")

# sample data
data('sp4')

ssc <- sp4[grep('^Bt', sp4$name), c('sand', 'silt', 'clay')]
names(ssc) <- toupper(names(ssc))

test_that("textureTriangleSummary() works as expected", {

  skip_if_not_installed(pkg = "Hmisc")

  # requires Tcl/tk, at least on linux
  # throws warning on GH actions / linux: "no DISPLAY variable so Tk is not available"
  skip_if_not_installed(pkg = "soiltexture")

  # does it work?
  s <- textureTriangleSummary(ssc)

  # result is a matrix
  expect_true(inherits(s, 'matrix'))

  # column order preserved
  expect_true(
    all(dimnames(s)[[2]] == c('SAND', 'SILT', 'CLAY'))
  )

})

## TODO: clarify, new title / description

#' Compute Pair-wise Distances of Soil Properties over Depth
#' 
#' Computes pair-wise distance matrix to determine where an attribute increases within
#' a specified vertical distance threshold.
#'
#' Uses matrix outer product to determine all pair-wise differences in `attr`
#' for the horizons of `p`. Supplies `attr` to `threshold.fun` to determine the
#' minimum value criterion to return TRUE in output matrix for an "increase".
#' Also, computes all pair-wise distances in depth dimension to determine
#' whether the vertical distance criteria have been met simultaneously with
#' `attr` increase.
#'
#' This function assumes that the `threshold.fun` supplied by the user returns
#' either a constant or a vector of equal length to its input.
#'
#' Note that the `threshold.fun` result is allowed to contain NA, but that will
#' result in no output for affected cells.
#'
#' `get.increase.depths()` performs the conversion of the square matrix
#' output of `get.increase.matrix` back to horizon top depth for where
#' criteria were met.
#'
#'
#' @param p a `SoilProfileCollection`, containing a single profile
#' @param attr horizon attribute name to get the "increase" of
#' @param threshold.fun a function that returns the threshold (as a function of
#' attr); may return a constant single value
#' @param vertical.distance the vertical distance (determined from difference
#' SPC top depth variable) within which increase must be met
#' @return Returns a square logical matrix reflecting where the increase
#' criteria were met.
#'
#' `get.increase.depths` converts to horizon top depth by using above
#' matrix output to determine depths where increase is met.
#' 
#' @author Andrew Gene Brown
#' @seealso [getArgillicBounds()], [crit.clay.argillic()]
#' @keywords manip
#' @export
#' @examples
#'
#' data(sp1, package = 'aqp')
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' p <- sp1[1]
#' attr <- 'prop' # clay contents
#' foo <- get.increase.matrix(p, threshold.fun = crit.clay.argillic,
#'                            attr = attr, vertical.distance = 30)
#' foo
#'
get.increase.matrix <- function(p, attr, threshold.fun, vertical.distance) {
  # p - a SoilProfileCollection, containing a single profile
  # attr - attribute name to get the "increase" of
  # threshold.fun - a function that returns the threshold (as a function of attr); may return a constant single value
  # vertical distance - the vertical distance (determined from difference SPC top depth variable) within which increase must be met

  # NOTE: should topdepths or midpoints be used? typically the same horizon index is chosen
  #
  #       with thick horizons/diffuse transitions -- the midpoint is probably more appropriate.
  #
  #       I originally chose top depth because it is more straightforward to validate against known data
  #       whereas the midpoints required some more thought. now that I am pleased with stability of algo,
  #       I think it is worth switching to midpoint given it probably gives a better estimate of
  #       the thickness of transitional zone between horizons when it matters
  #
  h <- horizons(p)
  depthz <- horizonDepths(p)
  middepth <- h[[depthz[1]]] + (h[[depthz[2]]] - h[[depthz[1]]]) / 2 # TODO: evaluate

  increase.var <- horizons(p)[[attr]]

  threshold.vector <- threshold.fun(increase.var)

  if(length(threshold.vector) == 1) {
    # if threshold.fun() returns a constant value, expand it to match the length of the attribute
    threshold.vector <- rep(threshold.vector, length(increase.var))

  } else if(length(threshold.vector) != length(increase.var)) {
    # this function assumes that the threshold.fun() supplied by the user returns either a constant,
    # or a vector of equal length to `increase.var` when called above. otherwise, we cannot calculate the result.
    # note that the threshold.fun() result _is_ allowed to contain NA, but that will result in no output for affected cells
    stop(paste0('profile ID:', profile_id(site(p)),
                " - threshold.fun() result should be length 1 or equal to length of attribute \'",attr,"\' (n=",length(increase.var),")."))
  }

  # repeat attr content by horizon in columns; number of columns = number of horizons
  attr.mat <- outer(increase.var, rep(1, length(increase.var)))

  # create a matrix of corresponding thrsholds
  thresh.mat <- outer(threshold.vector, rep(1, length(increase.var)))

  # calculate a matrix of attr differences (between all horizons, not just adjacent)
  attr.inc.mat <- outer(increase.var, increase.var, `-`)

  # calculate a vertical distance matrix (between all horizons)
  vdist.mat <- outer(middepth, c(0, middepth[-length(middepth)]), `-`)

  # crit1 "an increase of at least [thresh.mat]"
  increase.met <- (attr.mat - thresh.mat) > (attr.inc.mat * upper.tri(attr.inc.mat))
  increase.met[is.na(increase.met)] <- FALSE

  # crit2 "within a vertical distance of [vertical.distance]"
  vdist.met <- abs(vdist.mat) <= vertical.distance

  # are crit 1 and crit 2 met?
  criteria.met <- increase.met & vdist.met
  return(criteria.met)
}

#' Return the horizon top depths from a call to `get.increase.matrix()`
#'
#' \code{get.increase.depths} performs the conversion of the square matrix
#' output of \code{get.increase.matrix} back to horizon top depth for where
#' criteria were met.
#'
#' Note that the `threshold.fun` result is allowed to contain `NA`, but that will
#' result in no output for affected cells.
#'
#'
#' @param p a SoilProfileCollection, containing a single profile
#' @param attr horizon attribute name to get the "increase" of
#' @param threshold.fun a function that returns the threshold (as a function of
#' attr); may return a constant single value
#' @param vertical.distance the vertical distance (determined from difference
#' SPC top depth variable) within which increase must be met
#' @return Returns a numeric vector of depths where the increase requirement is
#' met. For the argillic, the first is the one of interest.
#'
#' `get.increase.depths()` converts to horizon top depth by using above
#' matrix output to determine depths where increase is met.
#' @seealso [getArgillicBounds()] [crit.clay.argillic()]
#' @rdname get.increase.matrix
#' @keywords manip
#' @export
#' @examples
#'
#' data(sp1, package = 'aqp')
#' depths(sp1) <- id ~ top + bottom
#' site(sp1) <- ~ group
#'
#' p <- sp1[1]
#' attr <- 'prop' # clay contents
#' foo <- get.increase.depths(p, threshold.fun = crit.clay.argillic,
#'                            attr = attr, vertical.distance = 30)
#' foo
#'
get.increase.depths <- function(p, attr, threshold.fun, vertical.distance) {
  topdepth <- horizons(p)[[horizonDepths(p)[1]]]
  #TODO: is there interesting pedogenic information that can be derived
  # from the lower triangle of the criteria.met matrix we omit from the
  # "increase" matrix?) -- i.e. pale/haplo clay decrease at depth?
  criteria.met <- get.increase.matrix(p, attr, threshold.fun, vertical.distance)
  # get the index of  _first_ column from matrix criteria.met that has nonzero sum,
  # this column index is the index of horizon where the attr increase is met
  # then return the top depth.
  return(topdepth[which(colSums(criteria.met) > 0)])
}

# 2018-12-17
# D.E. Beaudette
# A.G. Brown
#
#
# This function replaces the previous rbind.SoilProfileCollection function.
#
# TODO:
# * is it possible to implement an S4 interface for a list of SPC? (cleaner code)

# 2020/09/25: yes it is possible [for a list] with the caveat that the y argument _must_ be missing in order to use the base::union generic but dispatch to the aqp method.
# The problem with this is IDEs like Rstudio that probe the generics to identify missing required arguments will erroneously warn about the aqp-usage of union.

# if (!isGeneric("combine"))
setGeneric("combine", function(...)
  standardGeneric("combine"))

#' @title Combine SoilProfileCollection objects
#' 
#' @description Combine `SoilProfileCollection` objects or lists of `SoilProfileCollection` objects. This method provides `...` expansion for the `pbindlist` method. 
#' @param x A `SoilProfileCollection`
#' @param ... `SoilProfileCollection` objects 
#'
#' @return A SoilProfileCollection
#' 
#' @export
#' @aliases combine c
#' @rdname combine-SoilProfileCollection-method
#' @examples
#' 
#' # example data
#' spc1 <- random_profile(1, SPC = TRUE)
#' spc2 <- random_profile(2, SPC = TRUE)
#' spc3 <- random_profile('A', SPC = TRUE)
#' 
#' # combine into a single SPC, ... interface
#' spc <- combine(spc1, spc2, spc3)
#' 
#' # combine into a single SPC, list interface
#' spc <- combine(list(spc1, spc2, spc3))
#'
#' # input are combined into a single SPC
#' spc <- c(spc1, spc2, spc3)
#'
#' # result is a list when a mixture of objects are provided
#' spc <- c(spc1, bar=spc2, baz="foo")
#' 
setMethod("c", signature(x = "SoilProfileCollection"), function(x, ...)  {
  objects <- list(x = x, ...)
  if (!all(sapply(objects, inherits, 'SoilProfileCollection')))
    return(unlist(objects, recursive = FALSE))
  names(objects) <- NULL
  pbindlist(objects)
})

#' @export
#' @rdname combine-SoilProfileCollection-method
setMethod("combine", signature(... = "SoilProfileCollection"), function(...)  {
  pbindlist(list(...))
})

#' @export
#' @rdname combine-SoilProfileCollection-method
setMethod("combine", signature(... = "list"), function(...)  {
  # TODO: should I not be defining this? :X
  pbindlist(do.call('c', list(...)))
})


#' @title Combine a list of SoilProfileCollection objects
#'
#' @description  See \code{combine(...)} for a connotative short-hand method that does not require that `SoilProfileCollection` be in a list. Profiles will be sorted based on character sorting of profile ID.
#'
#' @param l a list of SoilProfileCollection objects
#' @param new.idname Optional: a character referring to a new column name to put unique profile IDs in; default: \code{NULL} to attempt with existing idname in first element
#' @param verbose Produce warnings and messages regarding results? default: \code{TRUE}
#'
#' @details Input data must share a common depth unit, and if spatial data are present, a common CRS and coordinate names. In the case of non-conformal @idname and/or @depthcols, the first SoilProfileCollection is used as a template. If one or more subsequent list elements has non-unique values in a site level attribute of that name, the ID name from the second list element is attempted, and so on. Non-conforming spatial data are dropped from the final result (returns default empty \code{SpatialPoints}).
#'
#' @return a SoilProfileCollection object
#' @author D.E. Beaudette and A.G. Brown
#' @export
#' @examples
#' # example data
#' data(sp2, package = 'aqp')
#' depths(sp2) <- id ~ top + bottom
#' site(sp2) <- ~ surface
#'
#' # copy pieces
#' x <- sp2[1:5, ]
#' y <- sp2[6:10, ]
#'
#' # reset IDs and combine
#' profile_id(y) <- sprintf("%s-copy", profile_id(y))
#'
#' # this should work
#' z <- pbindlist(list(x, y))
#'
#' # check
#' plot(z)
pbindlist <- function(l, new.idname = NULL, verbose = TRUE) {
  spc <- l
  
  # setup some defaults
  options(stringsAsFactors = FALSE)
  
  # short-circuits
  
  # empty list
  if (length(spc) == 0)
    return(NULL)
  
  # singleton
  if (length(spc) == 1)
    return(spc[[1]])
  
  # ALL NULL
  if (all(sapply(spc, is.null)))
    return(NULL)
  
  # ALL NA
  # must suppress warnings because is.na(SoilProfileCollection) throws a warning
  if (suppressWarnings(all(sapply(spc, is.na))))
    return(NULL)
  
  # check/filter for NULL list elements
  idx.null <- suppressWarnings(which(sapply(spc, is.null)))
  if (length(idx.null)) {
    spc <- spc[-idx.null]
    if (verbose)
      message("pbindlist: one or more input list elements is NULL")
  }
  
  # check/filter for NA list elements
  idx.na <- suppressWarnings(which(sapply(spc, is.na)))
  if (length(idx.na)) {
    spc <- spc[-idx.na]
    if (verbose)
      message("pbindlist: one or more input list elements is NA")
  }
  
  idx.notspc <- which(!sapply(spc, inherits, 'SoilProfileCollection'))
  if (length(idx.notspc)) {
    spc <- spc[-idx.notspc]
    if (verbose)
      message("pbindlist: one or more input list elements is not a SoilProfileCollection")
  }
  
  # short circuit:
  # stop here if the removal of NULL | NA | non-SPC has resulted in a singleton list
  if (length(spc) == 1) {
    return(spc[[1]])
  }
  
  # check for non-conformal depth units
  o.depth.units <- unique(lapply(spc, depth_units))
  if (length(o.depth.units) > 1)
    stop('inconsistent depth units', call. = FALSE)
  
  # check for non-conformal df class (revert to data.frame if non-matching)
  o.df.class <- unique(unlist(lapply(spc, function(s) metadata(s)$aqp_df_class)))
  if (length(o.df.class) > 1) {
    if (verbose)
      message("pbindlist: data.frame class type inconsistent, reset to \"data.frame\"")
    o.df.class <- "data.frame"
  }
  
  # check for non-conformal group (revert to "" if non-matching)
  o.group.by <- unique(unlist(lapply(spc, function(s) metadata(s)$aqp_group_by)))
  if (length(o.group.by) != 1) {
    o.group.by <- ""
  }
  
  # check for non-conformal hzdesgn (revert to data.frame if non-matching)
  o.hzdesgn <- unique(unlist(lapply(spc, function(s) metadata(s)$aqp_hzdesgn)))
  if (length(o.hzdesgn) != 1) {
    o.hzdesgn <- ""
  }
  
  # check for non-conformal hztexcl (revert to data.frame if non-matching)
  o.hztexcl <- unique(unlist(lapply(spc, function(s) metadata(s)$aqp_hztexcl)))
  if (length(o.hztexcl) != 1) {
    o.hztexcl <- ""
  }
  
  o.p4s <- unique(lapply(spc, function(x) suppressWarnings(prj(x))))
  
  drop.spatial <- FALSE
  if (length(o.p4s) > 1) {
    if (verbose)
      message('pbindlist: inconsistent CRS, dropping spatial metadata')
    drop.spatial <- TRUE
  }
  
  # decompose to list of lists
  spc.list <- lapply(spc, as, 'list')
  
  # number of SPCs
  n.spc <- length(spc.list)
  
  # total profiles [previously defined, but not used]
  # n.profiles <- sum(sapply(spc, length))
  
  # template for combined data is based on the first element
  new.pID <- spc.list[[1]]$idcol
  
  # allow user to specify new idname a priori
  if (!is.null(new.idname)) {
    if (is.character(new.idname) & length(new.idname) == 1) {
      new.pID <- new.idname
    } else {
      if (verbose)
        message("ignored new.idname argument; must be character vector of length 1")
    }
  }
  new.hzID <- spc.list[[1]]$hzidcol
  
  new.hzd <- spc.list[[1]]$depthcols
  new.metadata <- spc.list[[1]]$metadata
  
  if (drop.spatial) {
    new.metadata$coordinates <- NULL
    new.metadata$crs <- NULL
  }
  
  # get the data.frame class and grouping variable into new metadata
  
  # transfer old slots if they are present
  hzdold <- spc.list[[1]]$hzdesgncol
  hztold <- spc.list[[1]]$hztexclcol
  
  if (length(hzdold) == 1)
    new.metadata$aqp_hzdesgn  <- hzdold
  
  if (length(hztold) == 1)
    new.metadata$aqp_hztexcl  <- hztold
  
  if (!length(new.metadata$aqp_hzdesgn))
    new.metadata$aqp_hzdesgn <- ""
  
  if (!length(new.metadata$aqp_hzdesgn))
    new.metadata$aqp_hztexcl <- ""
  
  # TODO: need a template for coordinate names if spatial data are present in all
  
  # Make aqp::union() robust to profile ID / site attr collisions.
  # https://github.com/ncss-tech/aqp/issues/161
  tries <- 1
  for (i in 1:n.spc) {
    
    # if the template ID exists in ith SPC, check it
    if (new.pID %in% names(spc.list[[i]]$site)) {
      
      # get the values
      potential.pid <- spc.list[[i]]$site[[new.pID]]
      
      # compare length to unique length
      lupid <- length(unique(potential.pid))
      if (lupid != length(potential.pid)) {
        
        # warn on mismatch
        if (verbose)
          warning(sprintf("template profile ID '%s' exists as a non-unique value in SPC element #%s, trying '%s'",
                          new.pID, i, spc.list[[i]]$idcol), call. = FALSE)
        
        # update the template to ith IDcol
        new.pID <- spc.list[[i]]$idcol
        
        # increment counter (present, but non-unique)
        tries <- tries + 1
        
        # reset loop (check everything again)
        i <- 1
      }
    }
    
    if (tries >= n.spc) {
      oldIDs <- sapply(spc.list, function(x) x$idcol)
      # add a dot before the name
      new.pID <- paste0(".", spc.list[[1]]$idcol)
      # if that exists
      if (new.pID %in% oldIDs) {
        # add a number after the name
        new.pIDs <- paste0(new.pID, 1:100)
        # first that does not occur in existing IDs
        new.pID <- new.pIDs[!new.pIDs %in% oldIDs][1]
      } 
      message("using ", new.pID, " as unique profile ID name")
    }
  }
  
  # make sure first element template has "correct" pID name and reasonable values
  splt <- spc.list[[1]]
  if (splt$idcol != new.pID) {
    snames <- colnames(splt$site)
    snames <- c(splt$idcol, snames[!snames %in% splt$idcol])
    hnames <- colnames(splt$horizons)
    hnames <- c(splt$idcol, hnames[!hnames %in% splt$idcol])
    newsite <- .data.frame.j(splt$site, snames, "data.table")
    newhorizons <- .data.frame.j(splt$horizons, hnames, "data.table")
    colnames(newsite)[1] <- new.pID
    colnames(newhorizons)[1] <- new.pID
    newhorizons[[splt$idcol]] <- NULL
    spc.list[[1]]$site <- newsite
    spc.list[[1]]$horizons <- newhorizons
  }
  
  # starting from the second SPC in the list
  # reset profile ID names in all other objects
  # also reset depth names
  for (i in 2:n.spc) {
    # save originals
    old.pID <- spc.list[[i]]$idcol
    old.hzd <- spc.list[[i]]$depthcols
    
    # profile ID in horizons
    idx <- match(old.pID, names(spc.list[[i]]$horizons))
    names(spc.list[[i]]$horizons)[idx] <- new.pID
    
    # profile ID in site
    idx <- match(old.pID, names(spc.list[[i]]$site))
    names(spc.list[[i]]$site)[idx] <- new.pID
    
    # profile ID in diagnostic, may be missing
    if (old.pID %in% names(spc.list[[i]]$diagnostic)) {
      idx <- match(old.pID, names(spc.list[[i]]$diagnostic))
      names(spc.list[[i]]$diagnostic)[idx] <- new.pID
    }
    
    # profile ID in restriction, may be missing
    if (old.pID %in% names(spc.list[[i]]$restriction)) {
      idx <- match(old.pID, names(spc.list[[i]]$restrictions))
      names(spc.list[[i]]$restrictions)[idx] <- new.pID
    }
    
    # hz depth columns
    idx <- match(old.hzd, names(spc.list[[i]]$horizons))
    names(spc.list[[i]]$horizons)[idx] <- new.hzd
    
    # reset id names
    spc.list[[i]]$idcol <- new.pID
    spc.list[[i]]$depthcols <- new.hzd
  }
  
  # extract pieces
  o.h <- lapply(spc.list, '[[', 'horizons')
  o.s <- lapply(spc.list, '[[', 'site')
  o.d <- lapply(spc.list, '[[', 'diagnostic')
  o.r <- lapply(spc.list, '[[', 'restrictions')
  # o.sp <- lapply(spc.list, '[[', 'sp')
  
  # generate new SPC components
  # https://github.com/ncss-tech/aqp/issues/71
  if (requireNamespace("data.table")) {
    # preferentially use data.table if available
    o.h <- data.table::rbindlist(o.h, fill = TRUE)
    # fix nasty duping of names by fill by taking first column of each name
    # enforcing data.frame/tibble-like column-name uniqueness
    ohn <- colnames(o.h)
    o.h <- .data.frame.j(o.h, unique(ohn[match(ohn, ohn)]), "data.table")
    o.s <- data.table::rbindlist(o.s, fill = TRUE)
    osn <- colnames(o.s)
    o.s <- .data.frame.j(o.s, unique(osn[match(osn, osn)]), "data.table")
    o.d <- data.table::rbindlist(o.d, fill = TRUE)
    odn <- colnames(o.d)
    o.d <- .data.frame.j(o.d, unique(odn[match(odn, odn)]), "data.table")
    o.r <- data.table::rbindlist(o.r, fill = TRUE)
    orn <- colnames(o.r)
    o.r <- .data.frame.j(o.r, unique(orn[match(orn, orn)]), "data.table")
  } else {
    stop("package `data.table` is required to combine SoilProfileCollections", call.=FALSE)
  }
  
  # sp slot is deprecated; always empty/uninitialized now
  o.sp <- NULL
  
  ## sanity check: profile IDs should be unique
  if (length(o.s[[new.pID]]) != length(unique(o.s[[new.pID]]))) {
    stop('non-unique profile IDs detected')
  }
  
  # take combined horizon data, convert to target df subclass
  res <- .as.data.frame.aqp(o.h, o.df.class)
  
  ## reset horizon IDs (if the default column hzID is presnt)
  if ('hzID' %in% colnames(res))
    res$hzID <- NULL
  
  # rebuild SPC using safe constructors
  depths(res) <- formula(sprintf("%s ~ %s + %s", new.pID, new.hzd[1], new.hzd[2]))
  
  site(res) <- .as.data.frame.aqp(o.s, o.df.class)
  
  diagnostic_hz(res) <- .as.data.frame.aqp(o.d, o.df.class)
  restrictions(res) <- .as.data.frame.aqp(o.r, o.df.class)
  
  # # append any novel metadata
  # #  note that o.df.class is calculated from all elements, and defaults to data.frame
  # #  this line is to cover metadata like citations etc. that are not part of all SPCs
  # new.names <- !(names(new.metadata) %in% names(res@metadata))
  # 
  # # explicit slot replacement
  # res@metadata <- c(res@metadata, new.metadata[new.names])
  
  res <- .transfer.metadata.aqp(new.metadata, res)
  
  # attempt using a common ID
  suppressWarnings(hzidname(res) <- new.hzID)
  
  # check validity; try to do a bandaid to fix common problems if possible
  if (!spc_in_sync(res)$valid) {
    warning("pbindlist: SoilProfileCollection integrity checks failed. This should not happen! Contact the aqp package developers with your use case on the GitHub Issue Page (https://github.com/ncss-tech/aqp/). Attempting to rebuild object...", call. = FALSE)
    res <- try(rebuildSPC(res))
  }
  
  # if rebuild fails then this will be a try-error
  return(res)
}

test_that("spc2mpspline works as expected", {
  skip_if_not_installed('mpspline2')
  
  data(sp1)
  depths(sp1) <- id ~ top + bottom
  
  # add a second continuous numeric variable for demonstration
  sp1$value2 <- runif(nrow(horizons(sp1)))
  
  # horizons with NA in property of interest are removed (not whole profiles)
  res1 <- spc2mpspline(sp1, "prop", hzdesgn = 'name')
  expect_equal(length(res1), length(sp1))

  # correspond to profiles P002 and P009
  expect_equal(max(res1), 240)
  expect_equal(min(res1), 59)

  # confirm geometry of result is the same with multiple variables
  res2 <- spc2mpspline(sp1, c("prop", "value2"), hzdesgn = 'name')
  expect_equal(length(res2), length(sp1))
  expect_equal(max(res2), 240)
  expect_equal(min(res2), 59)
  expect_equal(nrow(res2), nrow(res1))
  
  # confirm splined results for both variables
  expect_true(all(c("prop_spline", "value2_spline") %in% horizonNames(res2)))
  
  # fix the missing sample data
  sp1fix <- sp1
  
  # profile 1: set bedrock bottom depth to 200cm
  sp1fix@horizons[6, ]$bottom <- 200
  
  # profile 1: set bedrock clay content to zero
  sp1fix@horizons[6, ]$prop <- 0
  
  # pass d= argument for greater max depth
  res3 <- spc2mpspline(sp1fix, "prop", d = c(0, 5, 15, 30, 60, 100, 200, 300))
  
  # pass d= argument (with method="est_dcm")
  res4 <- spc2mpspline(sp1fix, "prop", d = c(0, 5, 15, 30, 60, 100, 200, 300), method = "est_dcm")
  expect_equal(nrow(res4), length(sp1fix)*7) # first profile was fixed
  
  # pass d= argument (with method="est_icm")
  res4 <- spc2mpspline(sp1fix, "prop", d = c(0, 5, 15, 30, 60, 100, 200, 300), method = "est_icm")
  expect_equal(nrow(res4), nrow(sp1fix)) # first profile was fixed

  expect_equal(max(res3), 240)
  expect_equal(min(res1), 59)

  # if you want to show original and spline together, create combined horizon var
  sp1$prop_combined <- sp1$prop
  res3$prop_combined <- res3$prop_spline

  # pbindlist input data works after making unique ID
  profile_id(res3) <- paste0(profile_id(res3), "_spline")
  expect_silent( {sp1union <- pbindlist(list(sp1, res3))} )
  expect_equal(length(sp1union), 2*length(sp1))

  # plot(sp1union, color = "prop_combined", divide.hz = FALSE)
})

test_that("alternate depth output methods", {
  skip_if_not_installed('mpspline2')
  
  data(sp1, package = "aqp")
  depths(sp1) <- id ~ top + bottom
  
  res1 <- spc2mpspline(sp1, 'prop', 
                       method = "est_icm", 
                       hzdesgn = 'name')
  # contains one horizon with equal top and bottom
  expect_equal(nrow(sp1) - 1, nrow(res1))
  
  res2 <- spc2mpspline(sp1, 'prop', 
                       method = "est_dcm", 
                       hzdesgn = 'name')
  # 6 layers per input profile in output
  expect_equal(6*length(sp1), nrow(res2))
  
})

context("profile-level harmonization of horizon properties")

set.seed(1)
spc <- pbindlist(lapply(1:10, random_profile, SPC = TRUE))

# this method should be "immune" to column name ordering -- so shuffle them
spc.nhzcol <- length(horizonNames(spc))
replaceHorizons(spc) <- horizons(spc)[,sample(1:spc.nhzcol, spc.nhzcol)]

test_that("basic profile denormalization of \"range\" for one property", {
  expect_silent({
     h1 <- harmonize(spc, x.names = list(prop = c(q05 = "p1", q50 = "p2", q95 = "p3")), keep.cols = "p5")
    })

  # basic checks on [random] result
  expect_equal(length(h1), 30)
  expect_equal(nrow(h1), 141)
  expect_equal(round(mean(h1$prop),6), -0.390285)

  # original names are gone harmonized name + keep columns are present and in order
  expect_true(all(c("id","top","bottom","prop","p5","hzID") == horizonNames(h1)))
})

#' @title Plot soil profiles below a dendrogram or cladogram
#' 
#' @description Plot soil profiles below a dendrogram, using methods from the ape package.
#'
#' @param x a `SoilProfileCollection` object
#' @param clust a hierarchical clustering object generated by `hclust()`, `cluster::agnes()`, or `cluster::diana()`
#' @param rotateToProfileID logical, attempt rotation of dendrogram according to original profile IDs, requires `dendextend` package
#' @param scaling.factor numeric, vertical scaling of the profile heights (default estimate printed if not specified)
#' @param width numeric, scaling of profile widths
#' @param y.offset numeric, vertical offset for top of profiles (default estimate printed if not specified)
#' @param dend.y.scale numeric, scaling of dendrogram (default estimate printed if not specified)
#' @param dend.color dendrogram line color
#' @param dend.width dendrogram line width
#' @param dend.type dendrogram type, passed to `plot.phylo()`, either "phylogram" or "cladogram"
#' @param debug logical, optionally print debugging data and return a `data.frame` of linking structure
#' @param ... additional arguments to `plotSPC()`
#' 
#' 
#' @note The default values of `scaling.factor`, `y.offset`, and `dend.y.scale` are based on `clust$height` and `max(x)`, and should be close to optimal. Use the defaults as a starting point for finer control.
#' 
#' @author D.E. Beaudette
#' 
#' @keywords hplots
#' 
#' @return This function is typically called to create graphical output, when `debug = TRUE` a `data.frame` of IDs and linking structure used to build the figure.
#' 
#' @export
#'
plotProfileDendrogram <- function(
    x, 
    clust, 
    rotateToProfileID = FALSE, 
    scaling.factor = 5 * mean(clust$height) / max(x), 
    width = 0.25, 
    y.offset = 60 * mean(clust$height) / max(x), 
    dend.y.scale = max(clust$height * 2, na.rm = TRUE), 
    dend.color = par('fg'), 
    dend.width = 1, 
    dend.type = c("phylogram", "cladogram"), 
    debug = FALSE, 
    ...
) {
  
  # limit dendrogram types
  dend.type <- match.arg(tolower(dend.type), c("phylogram", "cladogram"), several.ok = FALSE)
  
  # sanity check: must be either agnes or diana object
  if(! inherits(clust, c('agnes', 'diana', 'hclust'))) {
    stop('clust must be an object generated by hclust(), cluster::diana() or cluster::agnes()')
  }
  
  # convert to hclust for all intermediate steps
  d.hclust <- as.hclust(clust)
  
  
  # IDs from SPC and clustering object: may not be in the same order!
  d.ids <- d.hclust$labels
  x.ids <- profile_id(x)
  
  # sanity check: ID vectors should be the same length
  if( length(d.ids) != length(x.ids) ) {
    print(
      list(
        profileID = profile_id(x), 
        clustID = d.ids, 
        order = d.hclust$order
      )
    )
    stop('inconsistent SoilProfileCollection and clustering object, inconsistent number of IDs', call. = FALSE)
  }
  
  # sanity check: all IDs must be accounted for
  sd.1 <- setdiff(d.ids, x.ids)
  sd.2 <- setdiff(x.ids, d.ids)
  
  if( length(sd.1) > 0 ) {
    print(
      list(
        profileID = profile_id(x), 
        clustID = d.ids, 
        order = d.hclust$order
      )
    )
    msg <- sprintf('IDs missing from SoilProfileCollection: [%s]', paste(sd.1, collapse = ', '))
    stop(msg, call. = FALSE)
  }
  
  if( length(sd.2) > 0 ) {
    print(
      list(
        profileID = profile_id(x), 
        clustID = d.ids, 
        order = d.hclust$order
      )
    )
    msg <- sprintf('IDs missing from cluster object: [%s]', paste(sd.1, collapse = ', '))
    stop(msg, call. = FALSE)
  }
  
  # profile IDs and clustering IDs may not be in the same order
  if(any(x.ids != d.ids)){
    message('profile IDs and clustering IDs are not in the same order')
  }
  
  
  # attempt to rotate dendrogram as close to original profile IDs as possible
  if(rotateToProfileID) {
    
    # rotate dendrogram
    if (requireNamespace('dendextend', quietly = TRUE)) {
      d.hclust <- dendextend::rotate(d.hclust, order = x.ids)
      
    } else {
      message('rotation of dendrogram requires `dendextend` package', call. = FALSE)
      
      ## TODO: think about how to do this via ape methods, 
      ## and without all of the to/from conversion to hclust
    }
    
  }
  
  
  # plotting order
  # create a link between dendrogram-tip-sorted IDS ---> profile IDs
  link.idx <- match(d.ids[d.hclust$order], x.ids)
  
  # allocate extra space
  if(debug){
    # device options are modified locally, reset when done
    # warning: this will reset the device coordinates!
    op <- par(no.readonly = TRUE)
    on.exit(par(op))
    
    par(mar = c(5, 5, 5, 5))
  }
  
  # convert to ape class for plotting
  dend <- as.phylo(d.hclust)
  
  # dendrogram, this also inits plotting coordinates / extent
  plot(
    dend, 
    cex = 0.8, 
    direction = 'up', 
    y.lim = c(dend.y.scale, 0), 
    x.lim = c(0.5, length(x)+1), 
    show.tip.label = debug, 
    edge.color = dend.color, 
    edge.width = dend.width, 
    type = dend.type
  )
  
  # get the last plot geometry
  lastPP <- get("last_plot.phylo", envir = ape::.PlotPhyloEnv)
  
  # plot the profiles, in the ordering defined by the dendrogram
  # with a couple fudge factors to make them fit
  plotSPC(
    x, 
    plot.order = link.idx, 
    add = TRUE, 
    width = width, 
    scaling.factor = scaling.factor, 
    y.offset = max(lastPP$yy) + y.offset, 
    ...
  )
  
  # if not specified, print information about scaling factors etc.
  .msg <- c(
    ifelse(missing(scaling.factor), sprintf("scaling.factor: %s", signif(scaling.factor, 3)), NA),
    ifelse(missing(y.offset), sprintf("y.offset: %s", signif(y.offset, 3)), NA),
    ifelse(missing(dend.y.scale), sprintf("dend.y.scale: %s", signif(dend.y.scale, 3)), NA)
  )
  .msg <- na.omit(.msg)
  
  if(length(.msg) > 0) {
    message(paste(.msg, collapse = '\n'))
  }
    
  
  if(debug) {
    
    # graphical helpers
    axis(1, las = 1, at = 1:length(x))
    axis(2, las = 1)
    
    segments(
      x0 = length(x) + 0.5, 
      x1 = length(x) + 0.5, 
      y0 = max(lastPP$yy), 
      y1 = max(lastPP$yy) + y.offset, 
      lwd = 3, 
      lend = 1,
      col = 'red'
    )
    
    
    # IDs and linking structure
    .res <- data.frame(
      profileID = x.ids, 
      clustID = d.ids, 
      clustID.ordered = d.ids[d.hclust$order],
      profile.plot.order = link.idx
    )
    
    return(invisible(.res))
    
  }
  
}

#' @title Flag perfectly overlapping horizons within a SoilProfileCollection
#'
#' @param x a `SoilProfileCollection` object
#'
#' @return logical vector with length (and order) matching the horizons of `x` 
#' 
#' @author D.E. Beaudette, A.G. Brown
#' 
#' @export
#' @details
#' Horizons with `NA` depths can be flagged as overlapping. Consider finding these horizons with `checkHzDepthLogic(byhz=TRUE)` and removing or fixing them.
#' 
#' @seealso [checkHzDepthLogic()] [fillHzGaps()]
#'
#' @examples
#' 
#' # two overlapping horizons
#' z <- data.frame(
#'   id = 'SPC',
#'   top = c(0, 25, 25, 50, 75, 100, 100),
#'   bottom = c(25, 50, 50, 75, 100, 125, 125)
#' )
#' 
#' # init SPC
#' depths(z) <- id ~ top + bottom
#' 
#' # flag perfectly overlapping horizons
#' z$.overlapFlag <- flagOverlappingHz(z)
#' 
#' # thematic sketches
#' plotSPC(z, color = '.overlapFlag', hz.depths = TRUE, 
#' depth.axis = FALSE, cex.names = 0.85)
#' 
flagOverlappingHz <- function(x) {
  h <- horizons(x)
  idn <- idname(x)
  hzd <- horizonDepths(x)
  
  # extract horizon depths and profile ID
  .id <- h[[idn]]
  .fid <- as.numeric(factor(.id))
  .tops <- paste0(.fid, ":", h[[hzd[1]]])
  .bottoms <- paste0(.fid, ":", h[[hzd[2]]])
  
  # missing depths are recoded so they will be recognized as runs with length >1
  
  .rt <- rle(.tops) 
  .rb <- rle(.bottoms)
  .ot <- .rt$values[which(.rt$lengths > 1)]
  .ob <- .rb$values[which(.rb$lengths > 1)]
  
  # index affected horizons
  .m1 <- outer(.ot, .tops, '==') 
  .m2 <- outer(.ob, .bottoms, '==')
  
  idx1 <- unlist(as.vector(apply(.m1, 1, which)))
  idx2 <- unlist(as.vector(apply(.m2, 1, which)))
  
  # generate flag vector along sequence of horizons 
  .res <- rep(FALSE, times = length(.tops))
  .res[intersect(idx1, idx2)] <- TRUE
  .res
}

context("flagOverlappingHz")

# two overlapping horizons
z <- data.frame(
  id = 'SPC',
  top = c(0, 25, 25, 50, 75, 100, 100),
  bottom = c(25, 50, 50, 75, 100, 125, 125)
)

depths(z) <- id ~ top + bottom

# basic functionality
test_that("flagOverlappingHz: perfect overlap in 2 horizons", {
  
  .overlapFlag <- flagOverlappingHz(z)
  
  # logical vector
  expect_true(length(.overlapFlag) == nrow(z))
  expect_true(inherits(.overlapFlag, 'logical'))
  
  # not overlapping horizons
  expect_true(all( !.overlapFlag[c(1, 4, 5)]))
  
  # overlapping horizons
  expect_true(all(.overlapFlag[c(2, 3, 6, 7)]))
  
})

z2 <- data.frame(
  id = 'SPC',
  top = c(0, 25, 25, 25, 50, 75, 100, 100),
  bottom = c(25, 45, 50, 50, 75, 100, 125, 125)
)

depths(z2) <- id ~ top + bottom

test_that("flagOverlappingHz: imperfect in 1 horizon, perfect in 2 horizons", {
  
  .overlapFlag <- flagOverlappingHz(z2)
  
  # logical vector
  expect_true(length(.overlapFlag) == nrow(z2))
  expect_true(inherits(.overlapFlag, 'logical'))
  
  # not overlapping horizons
  expect_true(all(!.overlapFlag[c(1, 2, 5, 6)]))
  
  # overlapping horizons
  expect_true(all(.overlapFlag[c(3, 4, 7, 8)]))
  
})

# more complex edge case
x <- data.frame(peiid = c("1373969", "1373969", "1373969", "1373969", 
"1373969", "1373969", "1373969", "1373969", "1373969", "1373969", 
"1373969", "1373969"), hzdept = c(0L, 0L, 0L, 0L, 
26L, 45L, 77L, 102L, 102L, 102L, 185L, 205L), hzdepb = c(26L, 
26L, 26L, 26L, 45L, 77L, 102L, 185L, 185L, 185L, 205L, 220L), 
    hzname = c("Ap", "Ap", "Ap", "Ap", "Bt1", "Bt2", "BCt", "Cd", 
    "Cd", "Cd", "2C1", "2C2"), texture = c("L", "L", "L", "L", 
    "CL", "CL", "CL", "L", "L", "L", "SR- VFSL L", "S"))

depths(x) <- peiid ~ hzdept + hzdepb

test_that("edge case", {
  expect_equal(as.logical(flagOverlappingHz(x)), 
               c(TRUE, TRUE, TRUE, TRUE, 
                 FALSE, FALSE, FALSE, 
                 TRUE, TRUE, TRUE, 
                 FALSE, FALSE))
})


z3 <- data.frame(
  id = 'SPC1',
  top = c(0, 25, 25, 25, 50, 75, 100, 100),
  bottom = c(25, 45, 50, 50, 75, 100, 125, 125)
)

z4 <- data.frame(
  id = 'SPC2',
  top = c(0, 25, 50, 75, 100),
  bottom = c(25, 50, 75, 100, 125)
)

z5 <- rbind(z3, z4)
depths(z5) <- id ~ top + bottom

# test multiple profiles with some depths the same between both profiles
# but only the first profile has overlap (1 imperfect, 2 perfect)
test_that("multiple profiles, with and without overlap", {
 .overlapFlag <- flagOverlappingHz(z5)
 expect_false(any(.overlapFlag[9:13]))
})

z6 <- data.frame(
  id = 'SPC1',
  top = 0,
  bottom = 100
)

z7 <- data.frame(
  id = 'SPC2',
  top = c(0, 0, 100),
  bottom = c(100, 100, 200)
)

z8 <- rbind(z6, z7)
depths(z8) <- id ~ top + bottom

test_that("multiple profiles, edge case", {
  # first single horizon profile matches depths that overlap in next profile
  .overlapFlag <- flagOverlappingHz(z8)
  expect_false(.overlapFlag[1])
})

## ----setup, echo=FALSE, results='hide', warning=FALSE---------------------------------------------
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = 96,
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)

## -------------------------------------------------------------------------------------------------
library(aqp)
library(cluster)

# load data and make a copy
data("osd")
x <- osd

## -------------------------------------------------------------------------------------------------
# assume a standard deviation of 10cm for horizon boundary depths
# far too large for most horizons, but helps to make a point
x$hzd <- 10

# generate 4 realizations of each soil profile in `x`
# limit the minimum horizon thickness to 5cm
set.seed(10101)
s <- perturb(x, id = sprintf("sim-%02d", 1:4), boundary.attr = 'hzd', min.thickness = 5)

# combine source + simulated data into a single SoilProfileCollection
z <- combine(x, s)

## ----fig.width=8.5, fig.height=5------------------------------------------------------------------
# set plotSPC argument defaults
options(.aqp.plotSPC.args = list(name.style = 'center-center', depth.axis = list(style = 'compact', line = -2.5), width = 0.33, cex.names = 0.75, cex.id = 0.66, max.depth = 185))

par(mar = c(0, 0, 0, 1))
plotSPC(z)

## ----fig.width=8.5, fig.height=5.25---------------------------------------------------------------
# encode as a factor for distance calculation
z$subgroup <- factor(z$subgroup)

par(mar = c(0, 0, 1, 1))
groupedProfilePlot(z, groups = 'subgroup', group.name.offset = -10, break.style = 'arrow', group.line.lty = 1, group.line.lwd = 1)

## ----fig.width=8.5, fig.height=5.25---------------------------------------------------------------
# assign GHL
z$genhz <- generalize.hz(
  z$hzname, new = c('A', 'E', 'Bt', 'C'), 
  pattern = c('A', 'E', 'Bt', 'C|Bt4')
)

# check GHL
par(mar = c(0, 0, 3, 1))
groupedProfilePlot(z, groups = 'subgroup', group.name.offset = -10, break.style = 'arrow', group.line.lty = 1, group.line.lwd = 1, color = 'genhz')

## -------------------------------------------------------------------------------------------------
# horizon-level distance matrix weight
w1 <- 1
# perform NCSP using only the GHL (ordered factors) to a depth of 185cm
d1 <- NCSP(z, vars = c('genhz'), maxDepth = 185, k = 0, rescaleResult = TRUE)

# site-level distance matrix weight
w2 <- 2
# Gower's distance metric applied to subgroup classification (nominal factor)
d2 <- compareSites(z, 'subgroup')

# perform weighted average of distance matrices
D <- Reduce(
  `+`, 
  list(d1 * w1, d2 * w2)
) / sum(c(w1, w2))

## ----fig.width=8.5, fig.height=6------------------------------------------------------------------
library(ape)

# divisive hierarchical clustering
h <- as.hclust(diana(D))

# hang soil profile sketches from resulting dendrogram
par(mar = c(1, 0, 0, 1))
plotProfileDendrogram(z, clust = h, scaling.factor = 0.0075, y.offset = 0.1, width = 0.33, color = 'genhz', name = NA)

# annotate dendrogram with subgroup classification
# this handy function provided by the ape package
tiplabels(pch = 15, col = c(2, 3)[z$subgroup], cex = 1.5, offset = 0.05)

# helpful legend
legend('topleft', legend = levels(z$subgroup), pch = 15, col = c(2, 3), bty = 'n')

#' @title Validate Standard Munsell Notation
#' 
#' @description
#' This function validates Munsell color notation according to the following requirements:
#' 
#'  1. *hue* must not be absent, and within the set of "standard hues", see [huePosition()]
#'  2. *value* must not be absent, and within `vRange`
#'  3. *chroma* must be within `cRange`, and is only allowed to be absent in neutral colors (e.g. `N 4/`)
#' 
#' @seealso [huePosition()] 
#'
#' @param m character vector of colors in Munsell notation
#' 
#' @param vRange numeric vector of length 2, range of expected Munsell value
#' 
#' @param cRange numeric vector of length 2, range of expected Munsell chroma
#'
#' @returns logical vector,
#'  * `TRUE`: valid Munsell notation
#'  * `FALSE`: invalid notation
#' 
#' @export
#'
#' @examples
#' 
#' # valid
#' validateMunsell('5Y 6/8')
#' validateMunsell('N 4/')
#' validateMunsell('2.5Y 4/0')
#' 
#' # invalid
#' validateMunsell('5G 0/4')
#' validateMunsell(NA)
#' 
#' # mixture
#' validateMunsell(c('5G 4/4', 'N 2/', NA, 'NA', '100R 3/3'))
#' 
validateMunsell <- function(m, vRange = c(1, 12), cRange = c(0, 50)) {
  
  # short-circuits
  if(all(is.na(m))) {
    return(rep(FALSE, times = length(m)))
  }
  
  # split / interpret input ----
  
  # split into [H, V, C] expecting standard notation
  #  * non-standard hues will be split
  #  * Neutral chips, NA chroma converted to 0 chroma
  .p <- suppressWarnings(
    parseMunsell(m, convertColors = FALSE)
  )
  
  ## TODO: maybe consider non-standard hues
  
  # standard hues
  .hp <- huePosition(returnHues = TRUE, includeNeutral = TRUE)
  
  # requirements:
  #  1. must be a standard hue, and not NA
  #  2. value must be within vRange, and not NA
  #  3. chroma must be within cRange, and not NA (unless Neutral chip)
  #
  # --> TRUE: valid, FALSE: invalid
  
  # req. 1 ----
  # NA automatically => FALSE
  .req1 <- .p$hue %in% .hp
  
  # req. 2 ----
  .req2 <- .p$value > vRange[1] & .p$value <= vRange[2]
  
  ## handle NA ----
  .req2.na <- is.na(.p$value)
  
  .idx <- which(is.na(.req2))
  if(length(.idx) > 0) {
    # invert logic -> Munsell notation is invalid
    .req2[.idx] <- ! .req2.na[.idx]
  }
  
  # req 3. ----
  # Neutral chip + NA correctly handled by parseMunsell()
  .req3 <- .p$chroma >= cRange[1] & .p$chroma <= cRange[2]
  
  ## handle NA for non-Neutral chips ----
  .req3.na <- is.na(.p$chroma)
  
  .idx <- which(is.na(.req3))
  if(length(.idx) > 0) {
    # invert logic -> Munsell notation is invalid
    .req3[.idx] <- ! .req3.na[.idx]
  }
  
  # a valid color meets all three requirements
  .res <- .req1 & .req2 & .req3
  
  return(.res)
}

context("Munsell notation validation")


test_that("validateMunsell() works as expected", {
  
  # TODO: split this out into specific tests
  
  ex <- c(
    c('10YR 3/3', '5YR 2/2', '2.5G 8/8'), 
    c('2.6YR 3.3', '6YR 2.2/4.2', '2.5BG 3/3'),
    c('10Y 2.5/4', 'N 4/', 'N 4/0'), 
    c('2.5R /3', '10YR 4/'),
    c(NA, 'NA', 'NA NA/NA', '5YR NA/NA', '5YR 4/NA', '5YR NA/4')
  )
  
  # data.frame(
  #   m = ex,
  #   v = validateMunsell(ex)
  # )
  
  expect_equal(
    validateMunsell(ex), 
    c(TRUE, TRUE, TRUE, FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE)
  )
  
})

## TODO: compute mean, covariance, Dirichlet parameters for all samples 'soil-texture-separate-sim-by-class.R'
##       use those to bootstrap by soil texture class label

## TODO: generalize to ignore letter case of sand, silt, clay columns

## TODO: allow for specification of any two columns



#'
#' @title Bootstrap Soil Texture Data
#' 
#' @description Simulate realistic sand/silt/clay values (a composition) using multivariate Normal distribution or Dirichlet distribution. Simulations from the multivariate Normal distribution are based on the compositional mean and variance-covariance matrix. Simulations from the Dirichlet distribution are based on maximum likelihood estimation of `alpha` parameters.
#' 
#' @author D.E. Beaudette
#' 
#' @param ssc a `data.frame` object with 3 columns: 'SAND', 'SILT', 'CLAY' and at least three rows of data within the range of 0-100 (percent). NA are automatically removed, but care should be taken to ensure that the sand/silt/clay values add to 100 percent. Simulations are based on these examples.
#' 
#' @param method type of simulation: 'dirichlet' or 'normal'. See details.
#' 
#' @param n number of simulated compositions. See details.
#' 
#' @return a `list` containing:
#' 
#'  * `samples` - `data.frame` of simulated sand, silt, clay values
#'  * `mean` - compositional mean
#'  * `var` - compositional variance-covariance matrix
#'  * `D.alpha` - (fitted) alpha parameters of the Dirichlet distribution, `NULL` when `method = 'normal'`
#' 
#' @details Simulations from the multivariate normal distribution will more closely track the marginal distributions of sand, silt, and clay--possibly a better fit for "squished" compositions (TODO elaborate). However, these simulations can result in extreme (unlikely) estimates. 
#' 
#' Simulations from the Dirichlet distribution will usually be a better fit (fewer extreme estimates) but require a fairly large number of records in `ssc` (`n >= 30`?) for a reliable fit.
#' 
#' Additional examples will be added to [this tutorial](http://ncss-tech.github.io/AQP/aqp/soiltexture-vizualization-ideas.html).
#' 
#' @export
#' 
#' 
#' @references 
#' 
#' Aitchison, J. (1986) The Statistical Analysis of Compositional Data Monographs on Statistics and Applied Probability. Chapman & Hall Ltd., London (UK). 416p.
#' 
#' Aitchison, J, C. Barcel'o-Vidal, J.J. Egozcue, V. Pawlowsky-Glahn (2002) A concise guide to the algebraic geometric structure of the simplex, the sample space for compositional data analysis, Terra Nostra, Schriften der Alfred Wegener-Stiftung, 03/2003
#' 
#' Malone Brendan, Searle Ross (2021) Updating the Australian digital soil texture mapping (Part 1*): re-calibration of field soil texture class centroids and description of a field soil texture conversion algorithm. Soil Research. https://www.publish.csiro.au/SR/SR20283
#' 
#' 
#' Malone Brendan, Searle Ross (2021) Updating the Australian digital soil texture mapping (Part 2*): spatial modelling of merged field and lab measurements. Soil Research. https://doi.org/10.1071/SR20284
#'
#' @examples
#' 
#' \donttest{
#' if(
#' requireNamespace("compositions") &
#'   requireNamespace("soiltexture")
#' ) {
#'   
#'   # sample data, data.frame
#'   data('sp4')
#'   
#'   # filter just Bt horizon data
#'   ssc <- sp4[grep('^Bt', sp4$name), c('sand', 'silt', 'clay')]
#'   names(ssc) <- toupper(names(ssc))
#'   
#'   # simulate 100 samples
#'   s <- bootstrapSoilTexture(ssc, n = 100)
#'   s <- s$samples
#'   
#'   # empty soil texture triangle
#'   TT <- soiltexture::TT.plot(
#'     class.sys= "USDA-NCSS.TT",
#'     main= "",
#'     tri.sum.tst=FALSE,
#'     cex.lab=0.75,
#'     cex.axis=0.75,
#'     frame.bg.col='white',
#'     class.lab.col='black',
#'     lwd.axis=1.5,
#'     arrows.show=TRUE,
#'     new.mar = c(3, 0, 0, 0)
#'   )
#'   
#'   # add original data points
#'   soiltexture::TT.points(
#'     tri.data = s, geo = TT, col='firebrick', 
#'     pch = 3, cex = 0.5, lwd = 1, 
#'     tri.sum.tst = FALSE
#'   )
#'   
#'   # add simulated points
#'   soiltexture::TT.points(
#'     tri.data = ssc, geo = TT, bg='royalblue', 
#'     pch = 22, cex = 1, lwd = 1, 
#'     tri.sum.tst = FALSE
#'   )
#'   
#'   # simple legend
#'   legend('top', 
#'          legend = c('Source', 'Simulated'), 
#'          pch = c(22, 3), 
#'          col = c('black', 'firebrick'), 
#'          pt.bg = c('royalblue', NA), 
#'          horiz = TRUE, bty = 'n'
#'   )
#'   
#'   
#' }
#' 
#' }
#' 
bootstrapSoilTexture <- function(ssc, method = c('dirichlet', 'normal'), n = 100) {
  
  if(!requireNamespace("compositions", quietly = TRUE))
    stop("package `compositions` is required", call.=FALSE)
  
  # filter NA
  ssc <- na.omit(ssc)
  
  # method
  method <- match.arg(method)
  
  # sanity check: should have 3 columns and > 3 rows
  if(ncol(ssc) < 3 | nrow(ssc) < 3) {
    stop('insufficient observations or incorrect column specification', call. = FALSE)
  }
  
  # check for appropriate column names
  # all must be present
  # may be other columns as well, ignore those
  name.check <- sapply(c('SAND', 'SILT', 'CLAY'), function(i) {
    any(names(ssc) %in% i)
  })
  
  if(! all(name.check)) {
    stop('`ssc` must contain columns: `SAND`, `SILT`, `CLAY`.')
  }
  
  # subset via column names
  ssc <- ssc[, c('SAND', 'SILT', 'CLAY')]
  
  # sanity check: data should be in the range of 0-100
  range.check <- range(sapply(ssc, range))
  if(! all(range.check >= 0 & range.check < 100)) {
    stop('data should be in the range of 0 to 100 (%)', call. = FALSE)
  }
     
  # convert to a closed / proportional composition object
  # with max value of 100%
  z <- compositions::acomp(ssc, total = 100)
  
  # safely compute compositional mean / variance-covariance
  mean.comp <- compositions::meanCol(z)
  var.comp <- compositions::var(z, robust = FALSE, method = 'pearson')
  
  s <- switch(
    method,
    'normal' = {
      # not used
      D <- NULL
      
      # simulate normal mixture
      compositions::rnorm.acomp(
        n = n, 
        mean = mean.comp, 
        var = var.comp
      )
      
    },
    'dirichlet' = {
      # fit 3-term alpha parameters of Dirichlet distribution
      # note backflips required when not loading entire compositions package
      # the following is the expanded form of default arguments to fitDirichlet()
      el <- compositions::mean.rmult(compositions::ult(z), robust = FALSE)
      D <- compositions::fitDirichlet(z, elog = el)$alpha
      
      # draw simulated values
      compositions::rDirichlet.acomp(n = n, alpha = D)
    }
  )
  
  
  # convert back to format that is suitable for plotting on the TT
  s <- as.data.frame(unclass(s) * 100)
  names(s) <- names(ssc)
  
  # package results into a list
  res <- list(
    samples = s,
    mean = mean.comp,
    var = var.comp,
    D.alpha = D
  )
  
  return(res)
}

context("soil texture simulation")

# sample data
data('sp6')

ssc <- sp6[grep('^Bt', sp6$name), c('sand', 'silt', 'clay')]
names(ssc) <- toupper(names(ssc))

test_that("bootstrapSoilTexture() works as expected", {
  
  skip_if_not_installed('compositions')
  
  # simulated under relatively stable conditions
  set.seed(1010101)
  s <- bootstrapSoilTexture(ssc, n = 50, method = 'dirichlet')
  
  # result is a list
  expect_true(inherits(s, 'list'))
  
  # there should be data.frame with 50 rows and 3 columns
  expect_true(inherits(s$samples, 'data.frame'))
  expect_true(nrow(s$samples) == 50)
  expect_true(ncol(s$samples) == 3)
  
  # mean sand, silt, clay values 
  # should be very close to these
  expect_true(
    all(round(s$mean) - c(31, 42, 27) < 3)
  )
  
  # column order preserved
  expect_true(
    all(names(s$samples) == c('SAND', 'SILT', 'CLAY'))
  )
  
})

## details on slice() -> dice()
# https://github.com/ncss-tech/aqp/issues/115

## TODO: 
##   * DT full outer join ideas
##     https://stackoverflow.com/questions/15170741/how-does-one-do-a-full-join-using-data-table
##   * fully test / document compatibility with overlapping horizons: https://github.com/ncss-tech/aqp/issues/296

setGeneric("dice", function(x,
                            fm = NULL,
                            SPC = TRUE,
                            pctMissing = FALSE,
                            fill = FALSE,
                            strict = TRUE,
                            byhz = TRUE,
                            verbose = FALSE)
  standardGeneric("dice"))


.dice <- function(x,
                  fm = NULL,
                  SPC = TRUE,
                  pctMissing = FALSE,
                  fill = FALSE,
                  strict = TRUE,
                  byhz = TRUE,
                  verbose = FALSE) {
  
  # sacrifice to R CMD check spirits
  .pctMissing <- NULL
  
  # find / flag / remove invalid profiles or horizons
  # this will generate an error if there are no valid profiles remaining
  if (strict) {
    # NOTE: this will not find/remove perfectly overlapping horizons
    x <- HzDepthLogicSubset(x, byhz = byhz)
    
    ## TODO: this could invoke 2x calls to fillHzGaps()
    # removed horizons will trigger an automatic gap-filling
    if (!is.null(metadata(x)$removed.horizons)) {
      message('filling gaps left by HzDepthLogicSubset')
      x <- fillHzGaps(x, flag = TRUE, to_top = NULL, to_bottom = NULL)
    }
  }
  
  # keep track of original object size
  o.size <- object.size(x)
  
  ## extract pieces
  idn <- idname(x)
  hzidn <- hzidname(x)
  htb <- horizonDepths(x)
  hznames <- horizonNames(x)
  
  # IDs + top/bottom
  ids.top.bottom.idx <- match(c(idn, hzidn, htb), hznames)
  
  # interpret optional formula
  if (!is.null(fm)) {
    # something supplied
    # reasonable?
    if (!inherits(fm, "formula")) {
      stop('must provide a valid formula', call. = FALSE)
    }
    
    # extract components of the formula:
    fm <- paste(deparse(fm, 500), collapse = '')
    elements <- strsplit(x = fm, split = '~', fixed = )[[1]]
    fm <- lapply(strsplit(x = elements, split = '[+*]'), trimws)
    
    # test for a multi-part formula A ~ B ~ C ?
    if (length(fm) > 2) {
      stop("please provide a valid formula", call. = FALSE)
    }
    
    # LHS ~ RHS
    # LHS: "", single integer, vector of integers slices
    # z-index
    z <- as.numeric(eval(parse(text = fm[[1]])))
    
    # RHS: variable names
    vars <- fm[[2]] 
    
    # check for bogus left/right side problems with the formula
    if (any(z < 0) | any(is.na(z))) {
      stop('z-index must be >= 0', call. = FALSE)
    }
    
    # check for integer / numeric slices
    if (!inherits(z,  c('numeric','integer'))) {
      stop('z-index must be either numeric or integer', call. = FALSE)
    }
    
    # if z-index is missing set to NULL for later on
    if (length(z) == 0) {
      z <- NULL
    } else {
      # z-index is specified
      # note z-index defines slice tops, lower limit is (z + 1)
      # must fill from min(z) --- [gaps] --- max(z) + 1
      x <- fillHzGaps(x, flag = TRUE, to_top = min(z), to_bottom = max(z) + 1)
      
      # NB: fillHzGaps will recalculate/reset hzID, so update the local var!
      hzidn <- "hzID"
    }
    
    # check for '.' --> all variables, minus ID/depths
    if (any(vars == '.')) {
      # all variables except profile ID, horizon ID, top, bottom
      vars <- hznames[-ids.top.bottom.idx]
    }
    
    # check for column names that don't exist
    if (!any(vars %in% hznames)) {
      stop('names in formula do not match any horizon attributes', call. = FALSE)
    }
    
    
  } else {
    # no formula
    # slice to-depth of collection
    
    # optionally fill all gaps between min(x) --- [gaps] --- max(x)
    if (fill) {
      x <- fillHzGaps(x, flag = TRUE, to_top = min(x), to_bottom = max(x))
    }
    
    # all variables except profile ID, horizon ID, top, bottom
    vars <- hznames[-ids.top.bottom.idx]
    z <- NULL
  }
  
  # time to work with horizons 
  h <- horizons(x)
  
  hmnames <- .hzMetadataNames(x, depths = TRUE)
  newhm <- hmnames[which(!hmnames %in% vars)]
  
  # safely select variables
  h <- .data.frame.j(h, col.names = c(newhm, vars))
  
  # convert to DT, as needed
  if (!inherits(h, 'data.table')) {
    h <- data.table::as.data.table(h)
  }
  
  ## TODO: this needs to be integrated / coordinated with 
  #        strict = TRUE, HzDepthLogicSubset() 
  
  ## TODO: how will this affect the FULL JOIN below if subset?
  
  # flag 0-thickness horizons
  zthk <- h[[htb[1]]] == h[[htb[2]]]
  if (any(zthk, na.rm = TRUE)) {
    message("horizons with zero thickness have been omitted from results")
  }
  
  # remove any 0-thickness horizons
  h.sub <- h[which(!zthk), ]
  
  # expand 1 unit slices to max depth of each profile
  #  calculate sequence of top depths of slices using data.table j index
  tops <- h.sub[, 
                list(
                  # iterate over horizons in each profile
                  sapply(seq_len(.N), function(i) {
                    
                    # calculate the number of slices in source horizon
                    j <- seq_len(abs(.SD[[htb[2]]][i] - .SD[[htb[1]]][i]))
                    
                    # when top and bottom are equal, no slice, return 0-length 
                    if (length(j) == 0) 
                      return(j)
                    
                    # calculate sequence from input top depth, minus 1
                    list(.SD[[htb[1]]][i] + j - 1)
                  })
                ), 
                # iterate over profiles in the collection (by idname)
                by = c(idn), 
                
                # .SD contains only ID + depths (more efficient)
                .SDcols = htb]
  
  tops <- unlist(tops$V1)
  bottoms <- tops + 1
  
  # expand slice IDs (horizon IDs); used to join with horizons
  sliceIDs <- rep(
    h.sub[[hzidn]], 
    times = abs(h.sub[[htb[2]]] - h.sub[[htb[1]]])
  )
  
  # assemble slice LUT for JOIN
  s <- data.table::data.table(
    sliceID = sliceIDs, 
    .sliceTop = tops,
    .sliceBottom = bottoms
  )
  
  # re-name for simpler JOIN
  names(s)[1] <- hzidn
  
  # FULL JOIN via fast data.table compiled code
  res <- merge(h, s, by = hzidn, all = TRUE, sort = FALSE)
  
  # init unique horizon IDs
  res[['sliceID']] <- as.character(1:nrow(res))
  
  # copy old depths
  res[['.oldTop']] <- res[[htb[1]]]
  res[['.oldBottom']] <- res[[htb[2]]]
  
  # replace old depths with sliced depths
  res[[htb[1]]] <- res[['.sliceTop']]
  res[[htb[2]]] <- res[['.sliceBottom']]
  
  # cleanup
  res[['.sliceTop']] <- NULL
  res[['.sliceBottom']] <- NULL
  
  ## TODO: perform z-index subset before slice sequence creation to save a lot of time
  if (!is.null(z)) {
    # selection via top depth within z-index
    res <- res[which(res[[htb[1]]] %in% z), ]
  }
  
  # slice-wise "percent missing" calculation
  ## NOTE: this does not exclude horizon metadata columns like hzname, hztexture
  ##       -> unless formula RHS is specified, but not with ~ .
  if (pctMissing) {
    
    # this is quite slow: DT -> DF -> matrix -> apply (8x longer)
    # res[['.pctMissing']] <- .pctMissing(res, vars)
    
    # hybrid base + DT approach
    res$'.pctMissing' <- NA
    # count number of NA in vars, by row
    data.table::set(res, j = '.pctMissing', value = rowSums(is.na(res[, .SD, .SDcols = vars])))
    
    # convert NA count to percentage
    data.table::set(res, j = '.pctMissing', value = res$.pctMissing / length(vars))
  }
  
  # ensure sliced horizons are in ID/top order
  res <- res[order(res[[idname(x)]], res[[horizonDepths(x)[1]]]), ]
  
  # only returning horizons as a data.frame
  if (!SPC) {
    return(as.data.frame(res))
  }
  
  # re-pack horizon data
  res <- .as.data.frame.aqp(res, aqp_df_class(x))
  
  # this will fail if strict = FALSE, and sub-setting resulted in corrupt SPC
  .condition <- try(replaceHorizons(x) <- res, silent = TRUE)
  
  # gracefully fail
  if (inherits(.condition, 'try-error')) {
    stop('SPC object corruption, please specify `strict = TRUE`', call. = FALSE)
  }
  
  # switch horizon ID to slice ID
  hzidname(x) <- 'sliceID'
  
  # final size
  f.size <- object.size(x)
  
  # object bloat factor
  OBF <- round(f.size / o.size, 1)
  metadata(x)$OBF <- OBF
  
  if (verbose) {
    # turned off by default; possibly informative 
    message(sprintf("Object Bloat Factor: %s", OBF))
  }
  
  return(x)
}

## TODO: allow the use of site data (PSC etc.) to determine the z-slice

#' @title Efficient Slicing of `SoilProfileCollection` Objects
#' 
#' @description Cut ("dice") soil horizons into 1-unit thick slices. This function replaces `aqp::slice()`, which will be deprecated in aqp 2.0.
#'
#'
#' @param x a `SoilProfileCollection` object
#' 
#' @param fm optional `formula` describing top depths and horizon level attributes to include: `integer.vector ~ var1 + var2 + var3` or `integer.vector ~ .` to include all horizon level attributes. Specification of `integer.vector` forces `fill = TRUE`. When `NULL` profiles are "diced" to depth and results will include all horizon level attributes. Note on interpretation of `integer.vector` (slice tops)
#' 
#' @param SPC return the diced `SoilPrfolileCollection`, if `FALSE` a `data.frame` of horizon-level attributes
#' 
#' @param pctMissing compute "percent missing data" by slice (when `TRUE` expect 6-8x longer run time)
#' 
#' @param fill logical, fill with empty placeholder horizons in gaps within profiles, and/or, above/below interval specified in `fm`. Automatically set to `TRUE` when LHS of `fm` is specified. Backwards compatibility with `slice` is maintained by setting `fill = TRUE` with or without `fm`.
#' 
#' @param strict perform horizon depth logic checking / flagging / removal
#' 
#' @param byhz Evaluate horizon depth logic at the horizon level (`TRUE`) or profile level (`FALSE`). Invalid depth logic invokes `HzDepthLogicSubset` which removes offending profiles or horizon records.
#' 
#' @param verbose Print information about object size/memory usage. Default: `FALSE`
#' 
#' @details For large and potentially messy collections that may include missing horizon depth logic errors, consider using `repairMissingHzDepths()` before `dice()`. Consider using `accumulateDepths()` before invoking `dice()` on collections that may contain old-style O horizon notation (e.g. 5-0cm).
#' 
#' @aliases dice
#' 
#' @seealso [repairMissingHzDepths()], [accumulateDepths()], [fillHzGaps()]
#' 
#' @return a `SoilProfileCollection` object, or `data.frame` when `SPC = FALSE`
#' 
#' @author D.E. Beaudette and A.G. Brown 
#' 
#' @export
setMethod(f = 'dice', signature(x = 'SoilProfileCollection'), .dice)

#' Make a data.frame from non-standard expressions evaluated in a data environment
#'
#' @param .data A list, or object coercible to one, describing the data
#' @param ... One or more expressions (preferably named e.g. `foo = "bar"`) to evaluate in `.data`
#'
#' @return A `list` where names are expression "names" from `...` and values are the result of evaluating expressions in context of `.data`
#' @noRd
#' @examples
#' # .data_dots(data.frame(a = 1:10, b = 2:11), cc = a + b, d = cc * 2)
#
#' # data("jacobs2000", package="aqp")
#
#' # .data_dots(compositeSPC(jacobs2000), clayprop = clay / 100)
#'
.data_dots <- function(.data, ...) {
  .dots <- substitute(list(...))
  .dots <- .dots[2:length(.dots)]
  .names <- names(.dots)
  if (is.null(.names))
    .names <- paste0("expr", 1:length(.dots))
  for (i in seq_along(.dots)) {
    .data[[.names[i]]] <- eval(.dots[[i]], .data, parent.frame())
  }
  return(.data[.names])
}

# if (!isGeneric("depths"))
setGeneric("depths", function(x, hzID = FALSE, ...)
  standardGeneric("depths"))

#' @name depths
#' @param x A SoilProfileCollection 
#' @param hzID Include horizon ID? Usually this is calculated from the (sorted) row index unless `hzidname()<-` has been called. Default: `FALSE`
#' @param ... not used
#' @return a `data.frame` containing profile ID, top depth, and bottom depth
#' @export
#' @seealso `horizons()` `idname()` `hzidname()` `horizonDepths()`
#' @rdname depths
#' @aliases depths,SoilProfileCollection-method
#' @examples
#' # load a SoilProfileCollection
#' data(jacobs2000, package = "aqp")
#' 
#' depths(jacobs2000)
setMethod("depths", "SoilProfileCollection", function(x, hzID = FALSE, ...) {
  n <- c(idname(x), hzidname(x)[isTRUE(hzID)], horizonDepths(x))
  .data.frame.j(horizons(x), n, use_class = aqp_df_class(x))
})

# if (!isGeneric("validSpatialData"))
  setGeneric("validSpatialData", function(object, ...)
    standardGeneric("validSpatialData"))

#' Check for valid spatial reference of profiles
#'
#' @description Are coordinate column names defined in metadata and existing in the SoilProfileCollection?
#' @param object a SoilProfileCollection
#' @aliases validSpatialData
#' @docType methods
#' @rdname validSpatialData
#' @return logical `TRUE` if column names are defined and correspond to existing data
#' @export
setMethod("validSpatialData", signature(object = "SoilProfileCollection"),
          function(object) {
            # coordinate column names are defined in metadata
            crds <- metadata(object)$coordinates
            
            # and columns of that name exist in either site or horizon slots
            return(!is.null(crds) && all(crds %in% names(object)))
          })

setGeneric("isEmpty", function(object, ...)
  standardGeneric("isEmpty"))

#' Check for "empty" profiles in a SoilProfileCollection
#' 
#' "Empty" profiles are used as placeholders for positions in a `SoilProfileCollection`
#' These profiles result from operations that remove or extract portions of horizons
#' from source profiles. 
#' 
#' In a `SoilProfileCollection` an empty profile occurs when it has one horizon, 
#' with `NA` top and bottom depths. Generally all non-profile ID site and horizon-level
#' values are all also `NA`, but only the depths are checked by `isEmpty()`.
#' 
#' @param object A SoilProfileCollection
#' @param ... Additional arguments not used.
#' @aliases isEmpty
#' @docType methods
#' @rdname isEmpty
#' @return logical. Vector of length equal to number of profiles in `object`. Returns `TRUE` when a profile has one horizon with `NA` top and bottom depths
#' @export
setMethod("isEmpty", signature(object = "SoilProfileCollection"),
          function(object, ...) {
  .NHZ <- .TOP <- .BOTTOM <- NULL
  d <- object[, 1, .TOP, .BOTTOM]
  object[, , .NHZ] == 1 & is.na(d[[1]]) & is.na(d[[2]])
})

##
## overloads
##

# return a concatenated vector of horizon + site names
# note that we strip out the ID column name from @site

#' Get names of columns in site and horizons table
#'
#' @description Get names of columns in site and horizons table of a SoilProfileCollection.
#' @param x a SoilProfileCollection
#' @aliases names
#' @docType methods
#' @rdname names
#' @export
setMethod("names", signature("SoilProfileCollection"),
          function(x) {
            sn <- siteNames(x)
            res <- c(horizons = horizonNames(x), site = sn[!sn %in% idname(x)])
            return(res)
          })

#' @details
#' The `.DollarNames` SoilProfileCollection method is implemented to support
#' auto-completion after the `$` operator. There is a custom help handler
#' implemented for interactive use in RStudio to integrate help text and data
#' viewer functionality with auto-completion. This custom help handler can be
#' disabled by setting `options(.aqp.rs.rpc.integration=FALSE)`.
#' 
#' @param pattern A regular expression. Only matching names are returned.
#' @docType methods
#' @importFrom utils .DollarNames
#' @rdname names
#' @export
.DollarNames.SoilProfileCollection <- function(x, pattern) {
  n <- names(x)
  res <- n[grep(paste0("^", pattern), n)]
  
  # when running interactively in RStudio, attach a custom SPC help handler
  if (interactive() &&
      getOption(".aqp.rs.rpc.integration", default = TRUE) &&
      !is.null(.Platform$GUI) && .Platform$GUI == "RStudio") {
    attr(res, 'helpHandler') <- "aqp:::.aqp.rs.rpc.spc.HelpHandler"
  }
  
  res
}

#' Custom RStudio Help Handler for SoilProfileCollection objects
#' 
#' SoilProfileCollection Custom Help Handler for `$` auto-completion and data
#' preview in RStudio.
#' 
#' Supports handling for the following internal `tools:rstudio` functions:
#'   - `"completion"` via `.rs.rpc.get_custom_help()`
#'   - `"url"` via `.rs.rpc.show_custom_help_topic()`
#'   - `"parameter"` is not implemented
#'
#' @param object character. Either `"completion"`, `"parameter"`, or `"url"`.
#' @param topic character. The help topic (e.g. the column name in `$` auto-completion).
#' @param source character. The topic source (e.g. the parent object name in `$` auto-completion).
#' @param ... 
#' @importFrom utils capture.output str
#' @returns list or character
#' @noRd
.aqp.rs.rpc.spc.HelpHandler <- function(object, topic, source, ...) {
  if (!exists(".rs.rpc.show_help_topic")) {
    .rs.rpc.show_help_topic <- function(...) NULL
  }
  res <- try({
    obj <- get(source)
    
    switch(
      tolower(object),
      # this supports tab auto-completion for SoilProfileCollection $ S4 method
      "completion" = {
        # handle [[ vs. $ semantics
        if (topic != obj@idcol) {
          val <- obj[[topic]]
          slt <- ifelse(length(val) == length(obj), "site", "horizons")
        } else {
          # obj[[idname(obj)]] returns the site-level ID
          val <- obj@horizons[[topic]]
          slt <- "horizons"
        }
        
        out <- list(
          title = paste0(slt, "(", source, ")"),
          signature = topic,
          returns = class(val),
          description = utils::capture.output(utils::str(val)),
          details = "",
          sections = list()
        )
      },
      
      # this supports "F1 for more information" in help tooltip
      "url" = {
        try(.rs.rpc.show_help_topic("$", "aqp", "method"), silent = TRUE)
        NULL
      }
    )
  }, silent = TRUE)
  
  if (inherits(res, 'try-error')) {
    res <- list(
      title = source,
      signature = topic,
      returns = "",
      description = "",
      details = "",
      sections = ""
    )
  }
  
  res
}

# overload min() to give us the min depth within a collection
#' Get the minimum bottom depth in a SoilProfileCollection
#' @description Get the shallowest depth of description out of all profiles in a SoilProfileCollection. Data missing one or more of: bottom depth, profile ID, or any optional attribute are omitted using \code{complete.cases}.
#' @param x a SoilProfileCollection
#' @param v optional: a vector of horizon attribute names to refine calculation
#' @param na.rm remove `NA`? default: `TRUE`
#' @aliases min
#' @docType methods
#' @rdname min
#' @export
setMethod(
  f = "min",
  signature(x = "SoilProfileCollection"),
  definition = function(x, v = NULL, na.rm = TRUE) {

    htb <- horizonDepths(x)
    target.names <- c(idname(x), hzidname(x), htb)
    
    # optionally use a horizon-level property refine calculation
    if (!missing(v) && !is.null(v) && v %in% horizonNames(x)) {
      target.names <- c(target.names,  v)
    }
    
    # handle empty spc
    bd <- x@horizons[[htb[2]]]
    if (length(bd) == 0 || all(is.na(bd))) {
      return(Inf)
    }
    
    # filter out missing data, accounting for optional `v`
    h <- x@horizons
    idx <- which(complete.cases(.data.frame.j(h, target.names, aqp_df_class(x))))
    x@horizons <- h[idx,]
    
    # return the shallowest (of the deepest depths in each profile)
    # returning Inf with no non-missing arguments to min
    .LAST <- NULL
    .HZID <- NULL
    return(suppressWarnings(min(horizons(x)[x[,, .LAST, .HZID],][[htb[2]]], na.rm = na.rm)))
  }
)

# overload max() to give us the max depth within a collection
#' Get the maximum bottom depth in a SoilProfileCollection
#' @description Get the deepest depth of description out of all profiles in a SoilProfileCollection. Data missing one or more of: bottom depth, profile ID, or any optional attribute are omitted using \code{complete.cases}.
#' @param x a SoilProfileCollection
#' @param v optional: horizon-level column name to refine calculation
#' @param na.rm remove `NA`? default: `TRUE`
#' @aliases max
#' @docType methods
#' @rdname max
#' @export
setMethod(
  f = "max",
  signature(x = "SoilProfileCollection"),
  definition = function(x, v = NULL, na.rm = TRUE) {
    htb <- horizonDepths(x)
    target.names <- c(idname(x), hzidname(x), htb)
    
    # optionally use a horizon-level property refine calculation
    if (!missing(v) && !is.null(v) && v %in% horizonNames(x)) {
      target.names <- c(target.names,  v)
    }
    
    # handle empty spc
    bd <- x@horizons[[htb[2]]]
    if (length(bd) == 0 || all(is.na(bd))) {
      return(-Inf)
    }
    
    # filter out missing data, accounting for optional `v`
    h <- x@horizons
    idx <- which(complete.cases(.data.frame.j(h, target.names, aqp_df_class(x))))
    x@horizons <- h[idx,]
    
    
    # return the deepest (of the deepest depths in each profile)
    # returning -Inf with no non-missing arguments to max
    .LAST <- NULL
    .HZID <- NULL
    return(suppressWarnings(max(horizons(x)[x[,, .LAST, .HZID],][[htb[2]]], na.rm = na.rm)))
  }
)

# overload length() to give us the number of profiles in the collection
#' Get the number of profiles in a SoilProfileCollection
#' @description Get the number of profiles in a SoilProfileCollection
#' @param x a SoilProfileCollection
#' @aliases length
#' @docType methods
#' @rdname length
#' @export
setMethod(
  f = "length",
  signature(x = "SoilProfileCollection"),
  definition = function(x) {
    # faster replacement for profile_id()
    #   which calls unique(horizons(x)[[idname(x)]]) -- expensive
    l <- length(x@site[[idname(x)]])#profile_id(x))
    return(l)
  }
)

# overload nrow() to give us the number of horizons in the collection
# do not need to define a generic at all if we use the base prototype
  
#' Get the number of horizons in a SoilProfileCollection
#' @aliases nrow
#' @description Get the number of horizons in a SoilProfileCollection
#' @param x a SoilProfileCollection
#' @docType methods
#' @rdname nrow
#' @export
setMethod(
  f = "nrow",
  signature(x = "SoilProfileCollection"),
  definition = function(x) {
    nrow(x@horizons)
  }
)
  

#' @title Uniqueness within a `SoilProfileCollection` via MD5 Hash
#' @description Unique profiles within a `SoilProfileCollection` using and MD5 hash of select horizon and / or site level attributes.
#' 
#' @param x a `SoilProfileCollection`
#' @param vars Variables to consider in uniqueness.
#' @param SPC logical return a `SoilProfileCollection` when `TRUE`, otherwise vector of profile indices
#' 
#' @return `SoilProfileCollection` when `SPC = TRUE`, otherwise a vector of integers
#' 
#' @aliases unique
#' @docType methods
#' @rdname unique
#' @export
#' @examples
#'
#'   # an example soil profile
#'   x <- data.frame(
#'     id = 'A',
#'     name = c('A', 'E', 'Bhs', 'Bt1', 'Bt2', 'BC', 'C'),
#'     top = c(0, 10, 20, 30, 40, 50, 100),
#'     bottom = c(10, 20, 30, 40, 50, 100, 125),
#'     z = c(8, 5, 3, 7, 10, 2, 12)
#'   )
#'   
#'   # init SPC
#'   depths(x) <- id ~ top + bottom
#'   
#'   # horizon depth variability for simulation
#'   horizons(x)$.sd <- 2
#'   
#'   # duplicate several times
#'   x.dupes <- duplicate(x, times = 5)
#'   
#'   # simulate some new profiles based on example
#'   x.sim <- perturb(x, n = 5, thickness.attr = '.sd')
#'   
#'   # graphical check
#'   plotSPC(x.dupes, name.style = 'center-center')
#'   plotSPC(x.sim, name.style = 'center-center')
#'   
#'   # inspect unique results
#'   plotSPC(unique(x.dupes, vars = c('top', 'bottom')), name.style = 'center-center')
#'   
#'   # uniqueness is a function of variable selection
#'   plotSPC(unique(x.sim, vars = c('top', 'bottom')), name.style = 'center-center')
#'   plotSPC(unique(x.sim, vars = c('name')), name.style = 'center-center')
#'   
#'
#'
setMethod(f = 'unique',
          signature(x = "SoilProfileCollection"),
          definition = function(x, vars, SPC = TRUE) {
            
            # compute hash by profile, for selected variables
            md5 <- profileApply(x, function(i) {
              # unlist() in order to drop row names
              digest::digest(unlist(as(i, 'data.frame')[, vars]))
            })
            
            # get unique hashes
            u.md5 <- unique(md5)
            
            # list profile idx by hash:
            profiles.by.hash <- sapply(u.md5, function(i) which(md5 == i), simplify = FALSE)
            
            # get an index of the first copy of each profile
            u.profiles <- sapply(profiles.by.hash, function(i) i[1])
            
            # down-grade to un-named vector of indices
            u.profiles <- as.vector(u.profiles)
            
            if(SPC) {
              # return the unique set of profiles as an SPC
              return(x[u.profiles, ])
            } else {
              # return an index of unique profiles
              return(u.profiles)
            }
          }
)
  
#' @title Subset a SoilProfileCollection with logical expressions
#' @description \code{subset()} is a function used for extracting profiles from a SoilProfileCollection based on logical criteria. It allows the user to specify an arbitrary number of logical vectors (equal in length to site or horizon), separated by commas. The function includes some support for non-standard evaluation.
#'
#' @param x A SoilProfileCollection

#' @param ... Comma-separated set of R expressions that evaluate as TRUE or FALSE. Length for individual expressions matches number of sites OR number of horizons, in \code{object}.
#'
#' @param greedy Use "greedy" matching for combination of site and horizon level matches? \code{greedy = TRUE} is the union, whereas \code{greedy = FALSE} (default) is the intersection (of site and horizon matches).
#'
#'
#' @aliases subset
#' @export
#'
#' @details To minimize likelihood of issues with non-standard evaluation context, especially when using `subset()` inside another function, all expressions used in `...` should be in terms of variables that are in the site or horizon data frame.
#'
#' @return A SoilProfileCollection.
#'
#' @author Andrew G. Brown.
#'
setMethod("subset", signature(x = "SoilProfileCollection"),
          function(x, ..., greedy = FALSE) {
            object <- x

            # capture expression(s) at function
              .dots <- substitute(list(...))
              .dots <- .dots[2:length(.dots)]


              # create composite object to facilitate eval
              .data <- compositeSPC(object)

              # loop through list of quosures and evaluate
              res <- vector('list', length(.dots))
              for (i in 1:length(.dots)) {

                # why does n=2 work?!
                res[[i]] <- eval(.dots[[i]], .data, parent.frame(n = 2))

                # print(ls(envir=globalenv()))
                # print(res[[i]])
                # print(.dots[[i]])
              }
              res.l <- lapply(res, length)

              # distinguish site and horizon level attributes
              # in the expression input
              sitematch <- res[res.l == length(object)]
              horizonmatch <- res[res.l == nrow(object)]

              # intersect the multi-prop site constraints
              if (length(sitematch) > 1) {
                sm <- rowSums(do.call('cbind', sitematch))
                sitematch <- (sm == length(sitematch))
              }

              # intersect the multi-prop horizon constraints
              if (length(horizonmatch) > 1) {
                hm <- rowSums(do.call('cbind', horizonmatch))
                horizonmatch <- (hm == length(horizonmatch))
              }

              # empty value to hold site level index
              idx <- numeric()

              # create site level index from site criteria
              if (length(sitematch) == 1 | !is.list(sitematch))
                idx <- which(unlist(sitematch))

              # create site level index from matching horizon criteria
              if (length(horizonmatch) == 1 |
                  !is.list(horizonmatch)) {
                peiid.from.hz <- unique(object@horizons[[idname(object)]][unlist(horizonmatch)])
                hz.idx <- match(peiid.from.hz, profile_id(object))

                # check that we wont be filtering erroneously
                #
                integrity <- spc_in_sync(object)
                if(!integrity$valid) {
                  print(integrity)
                  stop("Unable to filter! SPC integrity checks failed!")
                }

                if (length(idx) & !greedy) {
                  # intersection of site and horizon level matches
                  idx <- idx[idx %in% hz.idx]
                } else if (greedy) {
                  # union of site and horizon level matches
                  idx <- c(idx, hz.idx)
                } else {
                  # if we have only horizon-level, use just horizon level
                  idx <- hz.idx
                }

              }

              # return SPC, subsetted using site level index
              return(object[na.omit(idx),])
          })

setGeneric("subsetHz", function(x, ...)
  standardGeneric("subsetHz"))

#' Subset the horizons in a SoilProfileCollection using logical criteria
#' 
#' \code{subsetHz()} is a function used for extracting horizons from a SoilProfileCollection based on logical criteria.
#' 
#' @param x a SoilProfileCollection
#' @param ... Comma-separated set of R expressions that evaluate as `TRUE` or `FALSE` in context of horizon data frame. Length for individual expressions matches number of horizons, in \code{x}.
#' @param drop Default: `TRUE`. When `drop=FALSE` placeholder horizons (profile ID with all other values `NA`) are created where the specified filter results in removal of all horizons.
#' @details To minimize likelihood of issues with non-standard evaluation context, especially when using `subsetHz()` inside another function, all expressions used in `...` should be in terms of variables that are in the horizon data frame.
#' 
#' @return a SoilProfileCollection with a subset of horizons, possibly with some sites removed
#' @export
#' @aliases subsetHz
#' @examples
#' 
#' data(sp3)
#' 
#' depths(sp3) <- id ~ top + bottom
#' 
#' # show just horizons with 10YR hues
#' plot(subsetHz(sp3, hue == '10YR'))
#' 
setMethod("subsetHz", signature(x = "SoilProfileCollection"), function(x, ..., drop = TRUE) {
  # capture expression(s) at function
  .dots <- substitute(list(...))
  .dots <- .dots[2:length(.dots)]
  
  # create composite object to facilitate eval
  .data <- horizons(x)
  idn <- idname(x)
  pid <- profile_id(x)
  hzd <- horizonDepths(x)
  
  # loop through list of expressions and evaluate
  res <- vector('list', length(.dots))
  for (i in 1:length(.dots)) {
    res[[i]] <- eval(.dots[[i]], .data, parent.frame(n = 2))
  }
  
  subcrit <- Reduce('&', res)
  
  if (!is.logical(subcrit)) {
    badxpr <- paste0("'", paste0(.dots[sapply(.dots, function(x) {
      !is.logical(x)
    })], collapse = ",'"), "'")
    message(sprintf("%s is not logical; returning `x` unchanged", badxpr))
    return(x)
  }
  
  newhz <- .data[which(subcrit),]
  
  if (drop) {
    # subset SPC first to remove sites and other slots
    x <- x[which(pid %in% newhz[[idn]]), ]
    
    # then replace horizons with horizon subset 
    replaceHorizons(x) <- newhz
  } else {
    # reinsert empty horizons and site data for "retained" profiles
    x <- .insert_dropped_horizons(x, newhz)
  }
  
  x
})

# produces an "all NA" data.frame with n rows, based on schema of x
.create_placeholder_df <- function(x, n, as.class = "data.frame") {
  if (n == 0) {
    return(x[0, ])
  }
  template <- x[rep(NA_integer_, n), , drop = FALSE]
  rownames(template) <- NULL
  .as.data.frame.aqp(template, as.class)
}

#' @description  used to implement "drop=FALSE" for various methods that remove horizons from SoilProfileCollection object
#' @noRd
.insert_dropped_horizons <- function(object = SoilProfileCollection(), 
                                     horizons = horizons(object), 
                                     sites = site(object),
                                     pid = idname(object),
                                     depths = horizonDepths(object),
                                     SPC = TRUE) {
  
  s.i <- sites[[pid]]
  i.idx <- unique(horizons[[pid]])
  i.idx2 <- setdiff(s.i, i.idx)
  newid <- sites[[pid]][which(sites[[pid]] %in% i.idx2)]
  
  # create ID-only empty data using original data as templates
  if (length(newid) > 0) {
    h.empty <- .create_placeholder_df(horizons, length(newid), aqp_df_class(object))
    h.empty[[pid]] <- newid
    s.empty <- .create_placeholder_df(sites, length(newid), aqp_df_class(object))
    s.empty[[pid]] <- newid
    # reorder to original id (+ top depth for horizons)
    horizons <- data.table::rbindlist(list(horizons, h.empty))
    horizons <- horizons[order(horizons[[pid]], horizons[[depths[1]]]), ]
    
    sites <- sites[which(!sites[[pid]] %in% h.empty[[pid]]), , drop = FALSE]
    sites <- data.table::rbindlist(list(sites, s.empty))
    sites <- sites[order(sites[[pid]]), , drop = FALSE]
  }
  
  if (inherits(object, 'SoilProfileCollection') && SPC) {
    object@site <- .as.data.frame.aqp(sites, aqp_df_class(object))
    replaceHorizons(object) <- .as.data.frame.aqp(horizons, aqp_df_class(object))
    return(object)
  } else {
    return(list(
      horizons = horizons,
      sites = sites,
      pid = pid,
      depths = depths
    ))
  }
}


# functions tailored for use with magrittr %>% operator / tidyr

# if (!isGeneric("grepSPC"))
setGeneric("grepSPC", function(object, attr, pattern, ...)
  standardGeneric("grepSPC"))

#' @title Subset SPC with pattern-matching for text-based attributes
#' @name grepSPC
#' @aliases grepSPC,SoilProfileCollection-method
#' @description `grepSPC()` is a shorthand function for subsetting `SoilProfileCollection` objects. For example, by `filter(grepl(spc, ...))` or `filter(stringr::str_detect(spc, ...))`. It provides pattern matching for a single text-based site or horizon level attribute.
#' @param object A SoilProfileCollection
#' @param attr A character vector (column in object) for matching patterns against.
#' @param pattern REGEX pattern to match in \code{attr}
#' @param ... Additional arguments are passed to \code{grep()}
#' @return A SoilProfileCollection.
#' @author Andrew G. Brown.
#'
#' @rdname grepSPC
#' @export grepSPC
setMethod("grepSPC", signature(object = "SoilProfileCollection"),
          function(object, attr, pattern, ...) {

            # .Deprecated("subset")

              # capture expression(s) at function
              .dots <- substitute(attr)

              # create composite object to facilitate eval_tidy
              .data <- compositeSPC(object)

              # do eval of attr
              res <- .data_dots(.data, eval(.dots))

              # do the pattern matching
              idx <- grep(res, pattern = pattern, ...)

              # subset the SPC for result
              return(object[idx,])

          })

# if (!isGeneric("subApply"))
setGeneric("subApply", function(object, .fun, ...)
  standardGeneric("subApply"))

#' @title Subset SPC based on result of performing function on each profile
#' @name subApply
#' @aliases subApply,SoilProfileCollection-method
#' @description \code{subApply()} is a function used for subsetting SoilProfileCollections. It currently does NOT support for "tidy" lexical features in the \code{...} arguments passed to \code{profileApply()}. The expectation is that the function \code{.fun} takes a single-profile SoilProfileCollection and returns a logical value of length one. The use case would be for any logical comparisons that cannot be evaluated inline by \code{subSPC()} because they require more than simple logical operations.
#' @param object A SoilProfileCollection
#' @param .fun, A function that takes a single profile, returns _logical_ of length 1.
#' @param ... Additional arguments are passed to \code{.fun}
#' @return A SoilProfileCollection.
#' @author Andrew G. Brown.
#'
#' @rdname subApply
#' @export subApply
setMethod("subApply", signature(object = "SoilProfileCollection"),
          function(object, .fun, ...) {

              # .Deprecated("profileApply")

              # apply .fun to elements of x
              res <- profileApply(object, FUN = .fun, ...)

              # return subset of x where .fun is true
              return(object[which(res), ])
          })

## subset method for SoilProfileCollection objects
## s: site-level subsetting criteria (properly quoted)
## h: horizon-level subsetting criteria (properly quoted)
## result: SoilProfileCollection with all profiles that match _either_ criteria- i.e. greedy matching
# if (!isGeneric("subsetProfiles"))
setGeneric("subsetProfiles", function(object, s, h, ...)
  standardGeneric("subsetProfiles"))

#' DEPRECATED use subset
#'
#' This function is used to subset \code{SoilProfileCollection} objects using
#' either site-level or horizon-level attributes, or both.
#'
#' The \code{s} argument supplies a fully-quoted search criteria for matching
#' via site or horizon-level attributes. The \code{h} argument supplies a fully-quoted
#' search criteria for matching via horizon-level attributes. All horizons
#' associated with a single horizon-level match (i.e. out of several, only a
#' single horizon matches the search criteria) are returned. See examples for
#' usage.
#' @name subsetProfiles
#' @aliases subsetProfiles,SoilProfileCollection-method
#' @param object object
#' @param s fully-quoted search criteria for matching
#' via site-level attributes
#' @param h fully-quoted search criteria for matching
#' via horizon-level attributes
#' @param ... not used
#'
#' @return A \code{SoilProfileCollection} class object.
#' @export
#' @examples
#'
#' # more interesting sample data
#' data(sp2)
#' depths(sp2) <- id ~ top + bottom
#' site(sp2) <- ~ surface
#'
#' # subset by integer index, note that this does not re-order the profiles
#' plot(sp2[1:5, ])
#'
#' # generate an integer index via pattern-matching
#' idx <- grep('modesto', sp2$surface, ignore.case=TRUE)
#' plot(sp2[idx, ])
#'
#' # generate in index via profileApply:
#' # subset those profiles where: min(ph) < 5.6
#' idx <- which(profileApply(sp2, function(i) min(i$field_ph, na.rm=TRUE) < 5.6))
#' plot(sp2[idx, ])
#'
setMethod("subsetProfiles", signature(object = "SoilProfileCollection"),
          function(object, s, h, ...) {
            .Deprecated("subset")

            # sanity checks
            if (missing(s) & missing(h))
              stop('must provide either, site or horizon level subsetting criteria',
                   call. = FALSE)

            # extract parts
            s.d <- site(object)
            h.d <- horizons(object)
            id.col <- idname(object)
            object.ids <- profile_id(object)

            # subset using conventional data.frame methods
            if (!missing(s))
              s.d.sub.IDs <- subset(s.d, select = id.col, subset = eval(parse(text = s)))[, 1] # convert to vector
            else
              s.d.sub.IDs <- NA

            if (!missing(h))
              h.d.sub.IDs <- subset(h.d, select = id.col, subset = eval(parse(text = h)))[, 1] # convert to vector
            else
              h.d.sub.IDs <- NA

            # intersect IDs if s and h were used
            if (!missing(h) & !missing(s))
              matching.IDs <- intersect(s.d.sub.IDs, h.d.sub.IDs)

            # if only h, or only s were used, then
            else
              matching.IDs <- unique(na.omit(c(s.d.sub.IDs, h.d.sub.IDs)))

            # convert IDs into a numerical profile index
            # note: no matches results in idx == 0
            idx <- match(matching.IDs, object.ids)

            # subset SoilProfileCollection
            return(object[idx, ])
          })

context("slab method for SoilProfileCollection objects")

test_that("basic slab functionality", {
  
  data(sp1, package = 'aqp')
  depths(sp1) <- id ~ top + bottom
  
  # aggregate entire collection
  # 1-unit slabs
  a.1 <- slab(sp1, fm = ~ prop, strict=TRUE)
  # 5-unit slabs
  a.2 <- slab(sp1, fm = ~ prop, strict=TRUE, slab.structure=5)
  # custom slabs
  a.3 <- slab(sp1, fm = ~ prop, strict=TRUE, slab.structure=c(0,5,10,25,50))
  
  # did it work?
  expect_true(inherits(a.1, 'data.frame'))
  expect_true(inherits(a.2, 'data.frame'))
  expect_true(inherits(a.3, 'data.frame'))
  
  # number of results, these are in long format
  expect_equal(nrow(a.1), 240)
  expect_equal(nrow(a.2), 48)
  expect_equal(nrow(a.3), 4)
  
  # required column names
  nm <- names(a.1)
  expect_true(any(grepl('variable', nm)))
  expect_true(any(grepl('all.profiles', nm)))
  expect_true(any(grepl('top', nm)))
  expect_true(any(grepl('bottom', nm)))
  expect_true(any(grepl('contributing_fraction', nm)))
  expect_true(any(grepl('p.q', nm)))
})

test_that("slab structure for a single slab", {
  data(sp4, package = 'aqp')
  depths(sp4) <- id ~ top + bottom
  
  sp4$group <- c(rep('A',5), rep('B',5))
  a.1 <- slab(
    sp4,
    fm = ~ sand + silt + clay,
    slab.structure = c(0, 10),
    slab.fun = mean,
    na.rm = TRUE
  )
  expect_equal(nrow(a.1), 3)
  
  # again, this time within groups defined by a site-level attribute:
  a.1 <- slab(
    sp4,
    fm = group ~ sand + silt + clay,
    slab.structure = c(0, 10),
    slab.fun = mean,
    na.rm = TRUE
  )
  expect_equal(nrow(a.1), 6)
})


test_that("extended slab functionality: weighted aggregation", {
  
  data(sp1, package = 'aqp')
  depths(sp1) <- id ~ top + bottom
  
  sp1$weights <- rep(1:2, length(sp1))[1:length(sp1)]
  sp1$wtgrp <- rep(1, length(sp1))
  
  # component weighted averages
  # mukey=461268; Doemill-Jokerst, 3 to 8 percent slopes (615)
  x <- data.frame(cokey = c(21469960L, 21469960L, 21469960L, 21469960L, 
                            21469960L, 21469961L, 21469961L, 21469961L), 
                  comppct_r = c(50L,  50L, 50L, 50L, 50L, 40L, 40L, 40L),
                  hzdept_r = c(0L, 3L, 13L, 23L, 36L, 0L, 3L, 10L), 
                  hzdepb_r = c(3L, 13L, 23L, 36L, 61L, 3L, 10L, 35L), 
                  ph1to1h2o_r = c(6.1, 6.8, 6.7, 6.7, NA, 6.4, 6.5, NA))
  depths(x) <- cokey ~ hzdept_r + hzdepb_r
  site(x) <- ~ comppct_r
  
  # custom function, with actual comppct_r
  a.0 <- slab(x, ~ ph1to1h2o_r,
              slab.structure = c(0, 10),
              weights = "comppct_r",
              slab.fun = weighted.mean,
              na.rm = TRUE
  )
  expect_equal(a.0$value, 6.54, tolerance = 0.005) 
  
  skip_if_not_installed("Hmisc")
  
  # we expect quantile estimates to vary (given weighted v.s. unweighted)
  a.0 <- slab(sp1, fm = ~ prop, weights = "weights")
  a.1 <- slab(sp1, fm = ~ prop, strict = TRUE, weights = "weights")
  a.2 <- slab(sp1, fm = wtgrp ~ prop, strict = TRUE, weights = "weights")
  a.3 <- slab(sp1, fm = wtgrp ~ prop, strict = TRUE)
  
  # expect consistent structure for weighted/unweighted: same column names except for group
  ungroupcols <- colnames(a.3)
  ungroupcols[2] <- "all.profiles"
  expect_equal(colnames(a.0), ungroupcols)
  expect_equal(colnames(a.1), ungroupcols)
  expect_equal(colnames(a.2), colnames(a.3))
  
  # contributing fractions should be identical
  expect_true(all(a.1$contributing_fraction == a.2$contributing_fraction))
  expect_true(all(a.2$contributing_fraction == a.3$contributing_fraction))
  
  # should match this within the tolerance:
  #   soilDB::get_SDA_property(property = 'ph1to1h2o_r', method = "weighted average",
  #                            mukeys = 461268, miscellaneous_areas = FALSE, include_minors = TRUE,
  #                            top_depth = 0, bottom_depth = 10)
  
  # comppct_r adjusted to get higher result more like doemill
  x$comppct_r <- c(90, 10)
  a.1 <- slab(x, ~ ph1to1h2o_r, slab.structure = c(0, 10), weights = "comppct_r")
  expect_equal(a.1$p.q50, 6.8)
  
  # comppct_r adjusted to get lower result more like jokerst
  x$comppct_r <- c(10, 90)
  a.2 <- slab(x, ~ ph1to1h2o_r, slab.structure = c(0, 10), weights = "comppct_r")
  expect_equal(a.2$p.q50, 6.5)
  
  # comppct_r adjusted to get lower result more like jokerst
  x$comppct_r <- c(NA, 90)
  na.1 <- slab(x, ~ ph1to1h2o_r, slab.structure = c(0, 10), weights = "comppct_r", na.rm = TRUE)
  expect_equal(na.1$p.q50, 6.5)
  
  na.2 <- slab(x, ~ ph1to1h2o_r, slab.structure = c(0, 10), weights = "comppct_r", slab.fun = weighted.mean, na.rm = TRUE)
  expect_equal(na.2$value, NA_real_)
  
})

test_that("slab calculations: mean, single profile", {
  
  data(sp1, package = 'aqp')
  depths(sp1) <- id ~ top + bottom
  
  # aggregate single profile
  # custom slabs
  a <- slab(
    sp1[1,],
    fm = ~ prop,
    strict = TRUE,
    slab.structure = c(0, 5, 10, 25, 100),
    slab.fun = mean,
    na.rm = TRUE
  )
  
  # using mean and similar functions will cause slab to store single result / slab into 'value' column
  expect_true(any(grepl('value', names(a))))
  
  # calculations done by hand (DEB)
  # weighted mean, single profile
  # 0-5: 9.400
  expect_equal(a$value[1], 9.4, tolerance = 0.0001)
  # 5-10: 7.000
  expect_equal(a$value[2], 7, tolerance = 0.0001)
  # 10-25: 8.466
  expect_equal(a$value[3], 8.4666, tolerance = 0.0001)
  # 25-100: 15.625
  expect_equal(a$value[4], 15.625, tolerance = 0.0001)
  
})


test_that("slab calculations: mean, several profiles", {
  
  data(sp1, package = 'aqp')
  depths(sp1) <- id ~ top + bottom
  
  # aggregate single profile
  # custom slabs
  a <- slab(sp1, fm = ~ prop, strict = TRUE, slab.structure = c(0, 5, 10, 25, 100), slab.fun = mean, na.rm = TRUE)
  
  # weighted mean calculations done by 1-unit slices
  # note slice formula notation works from horizon "tops"
  s.1 <- dice(sp1, 0:4 ~ prop, SPC = FALSE)$prop
  s.2 <- dice(sp1, 5:9 ~ prop, SPC = FALSE)$prop
  s.3 <- dice(sp1, 10:24 ~ prop, SPC = FALSE)$prop
  s.4 <- dice(sp1, 25:99 ~ prop, SPC = FALSE)$prop
  
  # 0-5
  expect_equal(a$value[1], mean(s.1, na.rm=TRUE), tolerance=0.0001)
  # 5-10
  expect_equal(a$value[2], mean(s.2, na.rm=TRUE), tolerance=0.0001)
  # 10-25
  expect_equal(a$value[3], mean(s.3, na.rm=TRUE), tolerance=0.0001)
  # 25-100
  expect_equal(a$value[4], mean(s.4, na.rm=TRUE), tolerance=0.0001)
  
})


test_that("edge case: non-integer values in slab.structure", {
  
  data(sp1, package = 'aqp')
  depths(sp1) <- id ~ top + bottom
  
  # top == bottom
  expect_error(
    slab(
      sp1, 
      fm = ~ prop, 
      strict = TRUE, 
      slab.structure = c(5.5, 20), 
      slab.fun = mean, 
      na.rm = TRUE
    ), regexp = 'slab.structure may only contain integer depths'
  )
  
  # top > bottom
  expect_error(
    slab(
      sp1, 
      fm = ~ prop, 
      strict = TRUE, 
      slab.structure = 10.3, 
      slab.fun = mean, 
      na.rm = TRUE
    ), regexp = 'slab.structure may only contain integer depths'
  )
  
  
})

test_that("edge case: slab.structure[2] >= slab.structure[1]", {
  
  data(sp1, package = 'aqp')
  depths(sp1) <- id ~ top + bottom
  
  # top == bottom
  expect_error(
    slab(
      sp1, 
      fm = ~ prop, 
      strict = TRUE, 
      slab.structure = c(20, 20), 
      slab.fun = mean, 
      na.rm = TRUE
    ), regexp = 'invalid slab structure'
  )
  
  # top > bottom
  expect_error(
    slab(
      sp1, 
      fm = ~ prop, 
      strict = TRUE, 
      slab.structure = c(30, 20), 
      slab.fun = mean, 
      na.rm = TRUE
    ), regexp = 'invalid slab structure'
  )
  
  
})

test_that("edge case: slab.structure[2] > max(x)", {
  
  data(sp1, package = 'aqp')
  depths(sp1) <- id ~ top + bottom
  
  # single profile, to max(x)
  a.max <- slab(sp1[1, ], fm = ~ prop, strict=TRUE, slab.structure=c(0,max(sp1[1, ])), slab.fun = mean, na.rm=TRUE)
  
  # single profile
  # custom interval, exceeding max(sp1)
  a <- slab(sp1[1, ], fm = ~ prop, strict=TRUE, slab.structure=c(0,300), slab.fun = mean, na.rm=TRUE)
  
  # weighted mean should be the same
  expect_true(a.max$value == a$value)
  
  # contributing fractions will be different
  # smaller value when slab.structure[2] > max.d
  expect_true(a$contributing_fraction < a.max$contributing_fraction)
  
  # compare with hand-calculated value
  expect_equivalent(a$value, 13.58427, tolerance=0.001)
})

test_that("edge case: slab.structure[1] > 0 (w/ custom slab function)", {
  data(sp3, package = "aqp")
  depths(sp3) <- id ~ top + bottom
  
  # custom 'slab' function, returning mean +/- 1SD
  mean.and.sd <- function(values) {
    m <- mean(values, na.rm = TRUE)
    s <- sd(values, na.rm = TRUE)
    upper <- m + s
    lower <- m - s
    res <- c(mean = m,
             lower = lower,
             upper = upper)
    return(res)
  }
  
  ## this time, compute the weighted mean of selected properties, by profile ID
  a <- slab(sp3,
            fm = id ~ L + A + B,
            slab.structure = c(40, 60), 
            slab.fun = mean.and.sd
  )
  
  # convert long -> wide
  res <- data.table::dcast(
    data.table::as.data.table(a),
    formula = id + top + bottom ~ variable,
    value.var = 'mean'
  )
  
  expect_equal(nrow(res), length(sp3))
  expect_equal(ncol(res), 6L)
})

test_that("overlapping horizons", {
  data(sp4, package = 'aqp')
  depths(sp4) <- id ~ top + bottom
  
  # overlapping horizons results in increased number of slices (according to overlapping thickness)
  x1 <- slab(sp4, ~ K + Mg + Ca + CEC_7 + ex_Ca_to_Mg)
  
  # create overlap
  sp4@horizons[2,]$bottom <- sp4@horizons[2,]$bottom + 12
  
  # strict=TRUE byhz=FALSE removes overlap via whole profile
  expect_message({x2 <- slab(sp4, ~ K + Mg + Ca + CEC_7 + ex_Ca_to_Mg, strict = TRUE, byhz = FALSE)})
  expect_equal(nrow(x2), 245)
  
  # strict=TRUE / strict=FALSE (with default byhz) -- we get same number of rows with overlap as without
  expect_silent(x3 <- slab(sp4, ~ K + Mg + Ca + CEC_7 + ex_Ca_to_Mg, strict = FALSE))
  expect_equal(nrow(x3), 245)
  
  # try with a larger constant slab structure
  expect_silent(x3 <- slab(sp4, ~ K + Mg + Ca + CEC_7 + ex_Ca_to_Mg, slab.structure = 10, strict = FALSE))
  expect_equal(nrow(x3), 25)
})

context("split method for SoilProfileCollection objects")

## sample data
data(sp6)
depths(sp6) <- id ~ top + bottom

# fake grouping var
sp6$g <- factor(rep(c('A', 'B'), each = 3))

## tests

test_that("site-level grouping factor", {
  
  # standard, grouping factor
  s <- split(sp6, 'g')
  
  # result should be a list
  expect_true(inherits(s, 'list'))
  
  # expected groups
  expect_true(length(s) == 2)
  expect_equal(names(s), levels(sp6$g))
  
  # three profiles / group
  expect_equivalent(sapply(s, length), c(3, 3))
  
})

test_that("site-level grouping factor, as a vector", {
  
  # standard, grouping factor
  s <- split(sp6, sp6$g)
  
  # result should be a list
  expect_true(inherits(s, 'list'))
  
  # expected groups
  expect_true(length(s) == 2)
  expect_equal(names(s), levels(sp6$g))
  
  # three profiles / group
  expect_equivalent(sapply(s, length), c(3, 3))
  
})

test_that("identity split", {
  
  # idname used
  # s1 <- split(sp6) # doesn't work due to base generic
  s2 <- split(sp6, idname(sp6))
  s3 <- split(sp6, profile_id(sp6))
  s4 <- split(sp6, f = NULL)
  
  # result should be a list
  # expect_true(inherits(s1, 'list'))
  expect_true(inherits(s2, 'list'))
  expect_true(inherits(s3, 'list'))
  expect_true(inherits(s4, 'list'))

  # as many groups as profiles in original
  # expect_true(length(s1) == length(s2))
  expect_true(length(s2) == length(s3))
  expect_true(length(s3) == length(s4))
  expect_true(length(s4) == length(sp6))
  
  # 1 profile / group
  expect_equivalent(sapply(s2, length), rep(1, times = length(sp6)))
})


## errors
test_that("split fails as expected", {
  
  # group variable that is not site-level attribute
  expect_error(split(sp6, f = 'XXX'))
})

test_that("split with NA values in `f`", {
  
  site(sp6)$grp1 <- c(1, 1, 2, 2, NA, NA)
  site(sp6)$grp2 <- c(1, 2, 1, 2,  1,  2)
  site(sp6)$grp3 <- rep(NA_character_, times = length(sp6))
  
  # profiles with NA are dropped
  x <- split(sp6, sp6$grp1)
  # two groups
  expect_equal(length(x), 2)
  # 4/6 profiles remain
  expect_equal(sum(sapply(x, length)), 4)
  
  
  # additional "<missing>" group (with drop=FALSE)
  x <- split(sp6, "grp1", drop = FALSE)
  # three groups
  expect_equal(length(x), 3)
  # check for special NA group
  expect_true('<missing>' %in% names(x))
  # all groups have 2 members
  expect_true(all(sapply(x, length) == 2))
  
  # interaction grp1*grp2
  x <- split(sp6, list(sp6$grp1, sp6$grp2))
  expect_equal(length(x), 4)
  
  # interaction grp1*grp2 (with drop=FALSE)
  x <- split(sp6, list(sp6$grp1, sp6$grp2), drop = FALSE)
  expect_equal(length(x), 6)
  
  ## all NA, not sure why this would happen...
  
  # all NA, empty list is the result
  x <- split(sp6, 'grp3', drop = TRUE)
  expect_true(inherits(x, 'list'))
  expect_equal(length(x), 0)
  
  # single <missing> group 
  x <- split(sp6, 'grp3', drop = FALSE)
  expect_equal(names(x), '<missing>')
  expect_equal(as.vector(sapply(x, length)), 6)
  
}) 

test_that('estimateAWC/correctAWC', {
  
  # test range of organic matter, loam FEF textures
  base.awc <- estimateAWC(texcl = c("l","l","l"), omcl = c(3, 2, 1), na.rm = TRUE)
  expect_equal(base.awc, c(0.21, 0.18, round(0.175, 2)))
  
  # assume base AWC for testing corrections independently
  base.expected <- c(0.21, 0.18, 0.17)
  
  # auto-recycling of singletons, 3 AWCs all corrected with 25% rf
  corrected.awc <- correctAWC(base.expected, total_rf = 25)
  expect_equal(corrected.awc, c(0.18, 0.15, 0.14))
  
  # using gravel-size fragments often gives somewhat more restrictive AWC 
  corrected.awc2 <- correctAWC(base.expected, total_rf = 25, gravel = 25)
  expect_equal(corrected.awc2, c(0.16, 0.15, 0.14))
  
  salty.awc <- correctAWC(base.expected, total_rf = 0, ec = 8)
  expect_equal(salty.awc, c(0.17, 0.14, 0.14))
  
  # NA handling 
  corrected.awc <- correctAWC(base.expected, total_rf = c(25, NA, 25), nullFragsAreZero = TRUE)
  expect_equal(corrected.awc, c(0.18, 0.18, 0.14))
  
  corrected.awc <- correctAWC(base.expected, total_rf = c(25, NA, 25), nullFragsAreZero = FALSE)
  expect_equal(corrected.awc, c(0.18, NA, 0.14))
  
  corrected.awc <- correctAWC(base.expected, ec = c(8, NA, 8), nullFragsAreZero = TRUE)
  expect_equal(corrected.awc, c(0.17, 0.18, 0.14))
  
  corrected.awc <- correctAWC(base.expected, ec = c(8, NA, 8), nullFragsAreZero = FALSE)
  expect_equal(corrected.awc, c(0.17, NA, 0.14))
  
  # multiple inputs (fragments + salts)
  corrected.awc <- correctAWC(base.expected, total_rf = c(25, 25, NA), ec = c(8, NA, 8), nullFragsAreZero = TRUE)
  expect_equal(corrected.awc, c(0.14, 0.15, 0.14))
  
  corrected.awc <- correctAWC(base.expected, total_rf = c(25, 25, NA), ec = c(8, NA, 8), nullFragsAreZero = FALSE)
  expect_equal(corrected.awc, c(0.14, NA, NA))
  
  # more limiting `gravel` is used over NA `total_rf`, only NA is for 2nd texture with NA ec with nullFragsAreZero=F
  corrected.awc <- correctAWC(base.expected, total_rf = c(25, 25, NA), gravel = c(25, 25, 25), ec = c(8, NA, 8), nullFragsAreZero = FALSE)
  expect_equal(corrected.awc, c(0.13, NA, 0.11))
})

context("methods for plotting slab() output")

suppressWarnings(library(lattice, quietly = TRUE))
library(grid, quietly = TRUE)


# load sample data, upgrade to SoilProfileCollection
data(sp1)
depths(sp1) <- id ~ top + bottom

# aggregate entire collection with two different segment sizes
a <- slab(sp1, fm = ~ prop)
b <- slab(sp1, fm = ~ prop, slab.structure=5)

# stack into long format
ab <- make.groups(a, b)
ab$which <- factor(ab$which, levels=c('a','b'),
                   labels=c('1-cm Interval', '5-cm Interval'))


test_that("lattice helper functions work as expected", {
  
  # plot median and IQR
  # custom plotting function for uncertainty viz.
  fig <- xyplot(top ~ p.q50 | which, data=ab, ylab='Depth',
         xlab='median bounded by 25th and 75th percentiles',
         lower=ab$p.q25, upper=ab$p.q75, ylim=c(250,-5),
         panel=panel.depth_function,
         prepanel=prepanel.depth_function,
         cf=ab$contributing_fraction,
         alpha=0.5,
         layout=c(2,1), scales=list(x=list(alternating=1))
  )
  
  expect_true(inherits(fig, 'trellis'))
  
  
})

# base non-standard eval of keyword in ... "k-index": SPC[i, j, ...]
#  support for .LAST, .FIRST, .HZID special keywords
test_that(".LAST, .FIRST, .HZID, .NHZ k-keywords for SoilProfileCollection objects", {
  # define special symbols in global env
  # (not needed for tests, but needed wherever they are used in package)
  .FIRST <- NULL
  .LAST <- NULL
  .HZID <- NULL
  .NHZ <- NULL
  
  data(sp4)
  depths(sp4) <- id ~ top + bottom
  
  # .LAST
  expect_equal(length(sp4[, , .LAST]), 10)
  
  # .NHZ
  expect_equal(sp4[, , .NHZ], c(4, 2, 3, 4, 3,
                                2, 2, 2, 5, 3))
  
  # .HZID
  expect_equal(sp4[, , .HZID], 1:30)
  
  # .LAST .HZID
  expect_equal(sp4[, , .LAST, .HZID], c(4, 6, 9, 13, 16,
                                        18, 20, 22, 27, 30))
  
  # .FIRST sets j <- 1
  expect_equal(sp4[, , .FIRST, .HZID], sp4[, 1, , .HZID])
  
  # .FIRST ignores j input if given
  expect_equal(sp4[, 1000, .FIRST, .HZID], sp4[, 1, , .HZID])
  
  # .LAST ignores j input if given
  expect_equal(sp4[, , .LAST, .HZID], sp4[, 1000, .LAST, .HZID])
  
  # horizon index of horizons in profiles 5 to 10
  expect_equal(sp4[5:10, , .HZID], 14:30)
  
  # horizon index of 2nd horizon in profiles 5 to 10
  expect_equal(sp4[5:10, 2, .HZID], c(15, 18, 20, 22, 24, 29) - 13)
  
  # number of horizons
  expect_equal(sp4[5:10, 2, .NHZ],  c(1, 1, 1, 1, 1, 1))
})

test_that("subsetHz works", {
  data(sp3)

  depths(sp3) <- id ~ top + bottom

  # show just horizons with 10YR hues
  res1 <- subsetHz(sp3, hue == '10YR')
  expect_equal(nrow(res1), sum(sp3$hue == '10YR'))
  
  # test retaining empty profiles with drop=FALSE
  site(sp3)$foo <- "bar"
  res2 <- subsetHz(sp3, hue == '10YR', drop = FALSE)
  expect_equal(length(res2), length(sp3))
  expect_equal(sum(is.na(res2$foo)), 2)
})

---
title: "Munsell Color Conversion"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Munsell Color Conversion}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, results='hide', warning=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = as.integer(Sys.getenv("R_AQP_VIGNETTE_IMAGE_DPI", unset = 32)),
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)

suppressMessages(library(aqp, quietly = TRUE))
```

Soil color is most often described using the [Munsell color system](https://en.wikipedia.org/wiki/Munsell_color_system). The `aqp` package provides several functions for converting to and from Munsell color notation, into other representations suitable for on-screen display ([sRGB](https://en.wikipedia.org/wiki/SRGB)) or numerical analysis ([CIELAB](https://en.wikipedia.org/wiki/CIELAB_color_space)). 


## Color Conversion Functions

Conversion of Munsell notation to "something" that can be displayed on-screen (sRGB color representation) is probably the most common color transformation applied to soil morphology data. The `munsell2rgb()` function is convenient for this operation when Munsell notation has been split into vectors of hue, value, and chroma. These could be columns in a `data.frame` or separate vectors. The `parseMunsell()` function is convenient when a full Munsell notation is given as a character vector. Note that `parseMunsell()` is a wrapper to `munsell2rgb()`. For example, converting `10YR 4/6` with either function can return:

  * hex-notation of a color: `#805921FF`, or
  * sRGB color coordinates: `0.5002233 0.3489249 0.1287694`, or
  * CIELAB color coordinates: `40.95021 10.31088 37.49513`.


Selection of the closest `n` Munsell color "chips", given sRGB or CIELAB colorspace coordinates is performed with the `col2Munsell()` function. Input to this function can be colorspace coordinates, named colors (e.g. `red`), or hex-notation of a color. For example, the selection of the closest Munsell chip for CIELAB coordinates `(51.4337, 9.917916, 38.6889)` results in `10YR 5/6` with a reported sigma (error) of `1.5e-6`. This error is estimated as the [CIE2000 distance](https://en.wikipedia.org/wiki/Color_difference#CIEDE2000) between the source CIELAB coordinates and the CIELAB coordinates of the closest Munsell chip.


A representative Munsell color can be estimated from reflectance spectra in the range of 380nm to 730nm with the `spec2Munsell()` function.

```{r echo = FALSE, fig.width=8, fig.height=4}
.m <- '10YR 6/6'

data("munsell.spectra.wide")
w <- munsell.spectra.wide[, 1]
s <- munsell.spectra.wide[, .m]

par(mar = c(4.5, 4.5, 2, 1), cex.axis = 0.75, lend = 2)
plot(w, s, xlab = 'Wavelength (nm)', ylab = 'Reflectance', type = 'b', main = .m, cex = 0.8, pch = 16, axes = FALSE)
rect(xleft = 400, ybottom = 0.35, xright = 450, ytop = 0.4, col = parseMunsell(.m), border = 1, lwd = 1)
axis(side = 1, at = seq(380, 730, by = 20))
axis(side = 2, las = 1)
```


## Special Cases

Neutral colors are commonly specified two ways in the Munsell system: `N 3/` or `N 3/0`, either format will work with `munsell2rgb()` and `parseMunsell()`.

Non-standard Munsell notation (e.g. `3.6YR 4.4 / 5.6`), possibly collected with a sensor vs. color book, can be approximated with `getClosestMunsellChip()`. A more accurate conversion can be performed with the [`munsellinterpol` package.](https://cran.r-project.org/package=munsellinterpol).


## Examples

```{r}
# Munsell -> hex color
parseMunsell('5PB 4/6')

# Munsell -> sRGB
parseMunsell('5PB 4/6',  return_triplets = TRUE)

# Munsell -> CIELAB
parseMunsell('5PB 4/6',  returnLAB = TRUE)

# hex color -> Munsell
col2Munsell('#476189FF')

# neutral color
parseMunsell('N 5/')

# non-standard notation
getClosestMunsellChip('3.3YR 4.4/6.1', convertColors = FALSE)
```

---
title: "What is new in aqp 2.x?"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{What is new in aqp 2.x?}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, results='hide', warning=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = as.integer(Sys.getenv("R_AQP_VIGNETTE_IMAGE_DPI", unset = 32)),
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)
```


This is a major update to {aqp} that may create some issues for code depending on specific inputs/outputs as of {aqp} <= 1.42, particularly those relying on `slice()`, `slab()`, and `profile_compare()`. As of {aqp} 2.0, `slice()` and `profile_compare()` are deprecated, but will continue to work for the rest of calendar year 2023. There are no plans to maintain these functions beyond {aqp} 2.0. The new version of `slab()` is a drop-in replacement for the previous version.

See the manual pages for `dice()`, `NCSP()`, and `slab()`. Note that these new / updated functions perform stricter tests for horizon depth logic errors. Messages and warnings are issued when invalid depth logic is reported by `checkHzDepthLogic()`. New functions `accumulateDepths()`, `repairMissingHzDepths()`, and `fillHzGaps()` can usually salvage problematic profiles.

The `replace_na` and `add_soil_flag` arguments to `profile_compare()` are not present in `NCSP()`; missing data are always replaced by maximum dissimilarity (1 when Gower's distance metric is used) and the soil/non-soil matrix is always used to compare profiles of different depth.


## Notable Changes

 * Deprecation of `slice()` in favor of the new, faster, more robust implementation via `dice()`. Be sure to update existing code to use `dice()` from now on. Some argument names have changed. 
 * Complete overhaul of `slab()` (thanks to @brownag), with new arguments, faster back-end, and weighted aggregation implemented (finally).
 * Deprecation of `profile_compare()` in favor of the new `NCSP()` function--a complete overhaul based on Maynard et al., 2020. Note that:
   - site level attributes are now handled by `compareSites()`
   - variable weights are specified by argument
 * The functions `perturb()` and `estimatePSCS()` are now vectorized, and optimized for larger `SoilProfileCollection` objects.
 * `mixMunsell()` now uses `mixingMethod = 'exact'` by default for the simulation of subtractive color mixtures
 * `gower` package moved to SUGGESTS
 * `plotColorMixture()` now using grid graphics functions to determine color swatch geometry and setting overlap detection threshold
 * Deprecation of `rgb2munsell()` in favor of the more general `col2Munsell()`
 * Removal of `PMS2Munsell()` and support data
 * Deprecation of `coordinates()<-` and `proj4string()<-` in favor of `initSpatial()<-`
 * Removal of `rruff.sample` example XRD patterns
 * `get.ml.hz()` no longer uses the `name` argument
 
## Major changes to `plotSPC()`

 * The maximum depth range of the figure is now based on `max.depth` or `max(x)`. This means that sketches generated with {aqp} 2.x will generally have less white space at the bottom of the figure. Make more room for additional annotation or visual effect by setting the desired depth range with the `max.depth` argument.
 * Now using the `electroStatics_1D()` method for fixing horizon depth label overlap, solutions are deterministic and almost always better. Adjust label overlap adjustment strategy with `fixOverlapArgs = list(...)`.
 * Better depth axis interval heuristics (if not specified), varying based on figure depth range.
 * Depth axis adjustments via new argument `depth.axis`, logical or list
 * deprecation of arguments:
   - `plot.depth.axis`: set via `depth.axis = TRUE`, `depth.axis = FALSE`, or customize `depth.axis = list(...)`
   - `cex.depth.axis`: set via `depth.axis = list(cex = 1)`
   - `axis.line.offset`: set via `depth.axis = list(line = -2)`

## New Features

 * New example data, `wilson2022`.
 * Fast prototyping of `SoilProfileCollection` objects via `quickSPC()` and list / character templates.
 * Re-use arguments to `plotSPC()` via setting `options(.aqp.plotSPC.args = list(...))`. Set to `options(.aqp.plotSPC.args = NULL)` to disable.
 * Coarse fragment classification via `fragmentSieve()` and `fragmentClasses()`.
 * S4 `as.data.frame(<SPC>)` as shorthand for `as(<SPC>, 'data.frame')`
 * `plotSPC()` now marks truncated profiles with a ragged bottom.
 * `fixOverlap()` now has [two label-placement solvers](https://ncss-tech.github.io/aqp/articles/label-placement.html):
   - electrostatic simulation (`method = "E"`)
   - simulated annealing (`method = "S"`)
 * New [depth axis styles](https://ncss-tech.github.io/AQP/aqp/sketches.html) in `plotSPC()`.
 * New function `flagOverlappingHz()` for identifying horizons with perfect overlap
 * New function `warpHorizons()` for warping horizon thickness (inflate/deflate)
 * `simulateColor()` adds multivariate simulation in CIELAB colorspace
 

## Incremental changes, should have no effect on previous code

 * Bug fix in `plotSPC()` when `fixLabelCollisions = TRUE`, adjustments suggested to `fixOverlap()` are now scaled correctly.
 * `explainPlotSPC()` reports label adjustment index when label collision repair is enabled.
 * Aesthetic cleanup in `explainPlotSPC()`.
 * `soilColorSignature()` gains arguments and perceptual color distances (dE00) via {farver} package.
 * `as(<SPC>, "data.frame")`: Replace `plyr::join()` with `merge()`.
 * `correctAWC()`: NA handling - return NA when frags are NA.
 * `mutate_profile()`: Faster (data.table-based) evaluation of profile-level expressions.
 * `profileApply`: Add support for custom `lapply()`-like function (`APPLY.FUN`) for processing chunks. 
 * Add `.interpretHorizonColor()` outputs to `last_spc_plot` in `aqp.env` for use in custom `legend()`.
 * Add `simplify` argument to `SoilTextureLevels()` and `ssc_to_texcl()` to optionally convert to an ordered factor with maximum of 12 levels (rather than 21). This smaller list of classes excludes sand grain size variants such as fine sand, loamy coarse sand, and very fine sandy loam.
 * Major updates to `profileInformationIndex()` methods and performance boost.
 * Enhancements to `evalMissingData()`.

---
title: "Numerical Classification of Soil Profiles"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Numerical Classification of Soil Profiles}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, results='hide', warning=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = 96,
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)
```

# Introduction

With the release of {aqp} 2.0, the soil profile comparison algorithm implemented in `profile_compare()` (Beaudette et al., 2013) has been completely re-written as `NCSP()` and re-named the "Numerical Comparison of Soil Profiles". A more recent discussion of this algorithm is provided in Maynard et al. (2020).

This short vignette demonstrates how to use the `NCSP()` function from {aqp} 2.x to perform a pair-wise comparison of soil data encoded as a `SoilProfileCollection` object. The pair-wise comparison of site-level attributes, previously available in `profile_compare()` has been removed from `NCSP()` and implemented as a stand-alone function named `compareSites()`. A final distance matrix (combining horizon and site level attributes) is created via weighted average. 
A more detailed version of this vignette can be found in the [Pair-Wise Distances by Generalized Horizon Labels](https://ncss-tech.github.io/AQP/aqp/genhz-distance-eval.html) tutorial.


# A Simple Example

Consider three soil profiles, containing basic morphology associated with the [Appling](https://casoilresource.lawr.ucdavis.edu/sde/?series=appling), [Bonneau](https://casoilresource.lawr.ucdavis.edu/sde/?series=Bonneau), and [Cecil](https://casoilresource.lawr.ucdavis.edu/sde/?series=cecil) soil series. These data are provided in the example data set "osd" as part of the {aqp} package.

```{r}
library(aqp)
library(cluster)

# load data and make a copy
data("osd")
x <- osd
```

Simulation is used below to generate 4 realizations of each soil series, using the `perturb()` function.
```{r}
# assume a standard deviation of 10cm for horizon boundary depths
# far too large for most horizons, but helps to make a point
x$hzd <- 10

# generate 4 realizations of each soil profile in `x`
# limit the minimum horizon thickness to 5cm
set.seed(10101)
s <- perturb(x, id = sprintf("sim-%02d", 1:4), boundary.attr = 'hzd', min.thickness = 5)

# combine source + simulated data into a single SoilProfileCollection
z <- combine(x, s)
```

A quick review of the source and simulated profiles, note patterns in horizon depths, horizon designation, and moist soil colors. The profiles have been visually truncated at 185cm for clarity (note ragged bottoms). The new `.aqp.plotSPC.args` option is used to set default arguments to `plotSPC()` for the remainder of the R session. Simulated profiles are labeled with a numeric suffix (e.g. "-01")
```{r fig.width=8.5, fig.height=5}
# set plotSPC argument defaults
options(.aqp.plotSPC.args = list(name.style = 'center-center', depth.axis = list(style = 'compact', line = -2.5), width = 0.33, cex.names = 0.75, cex.id = 0.66, max.depth = 185))

par(mar = c(0, 0, 0, 1))
plotSPC(z)
```

Subgroup level classification (encoded as an un-ordered factor) will be used as a site-level attribute for computing pair-wise distances. Quickly review the grouping structure with `groupedProfilePlot()`.
```{r fig.width=8.5, fig.height=5.25}
# encode as a factor for distance calculation
z$subgroup <- factor(z$subgroup)

par(mar = c(0, 0, 1, 1))
groupedProfilePlot(z, groups = 'subgroup', group.name.offset = -10, break.style = 'arrow', group.line.lty = 1, group.line.lwd = 1)
```

Horizon designation, grouped into ["generalized horizon labels"](https://ncss-tech.github.io/AQP/aqp/gen-hz-assignment.html) will be used as the horizon-level attribute for computing pair-wise distances. REGEX pattern matching is used to apply generalized horizon labels (GHL) to each horizon, and are encoded as ordered factors. A thematic soil profile sketch (horizon color defined by a property or condition) is a convenient way to graphically check GHL assignment.
```{r fig.width=8.5, fig.height=5.25}
# assign GHL
z$genhz <- generalize.hz(
  z$hzname, new = c('A', 'E', 'Bt', 'C'), 
  pattern = c('A', 'E', 'Bt', 'C|Bt4')
)

# check GHL
par(mar = c(0, 0, 3, 1))
groupedProfilePlot(z, groups = 'subgroup', group.name.offset = -10, break.style = 'arrow', group.line.lty = 1, group.line.lwd = 1, color = 'genhz')
```


Define weights and compute separately horizon and site level distance matrices. In this case, the site-level distances are give double the weight as the horizon-level distances. See the manual pages (`?NCSP` and `?compareSites`) for additional arguments that can be used to further customize the comparison.
```{r}
# horizon-level distance matrix weight
w1 <- 1
# perform NCSP using only the GHL (ordered factors) to a depth of 185cm
d1 <- NCSP(z, vars = c('genhz'), maxDepth = 185, k = 0, rescaleResult = TRUE)

# site-level distance matrix weight
w2 <- 2
# Gower's distance metric applied to subgroup classification (nominal factor)
d2 <- compareSites(z, 'subgroup')

# perform weighted average of distance matrices
D <- Reduce(
  `+`, 
  list(d1 * w1, d2 * w2)
) / sum(c(w1, w2))
```



Investigate the final distance matrix using divisive hierarchical clustering. The `plotProfileDendrogram()` function is a convenient way to combine the resulting dendrogram with soil profile sketches. Note that the site-level distances have a very strong effect on the overall clustering structure.
```{r fig.width=8.5, fig.height=6}
library(ape)

# divisive hierarchical clustering
h <- as.hclust(diana(D))

# hang soil profile sketches from resulting dendrogram
par(mar = c(1, 0, 0, 1))
plotProfileDendrogram(z, clust = h, scaling.factor = 0.0075, y.offset = 0.1, width = 0.33, color = 'genhz', name = NA)

# annotate dendrogram with subgroup classification
# this handy function provided by the ape package
tiplabels(pch = 15, col = c(2, 3)[z$subgroup], cex = 1.5, offset = 0.05)

# helpful legend
legend('topleft', legend = levels(z$subgroup), pch = 15, col = c(2, 3), bty = 'n')
```


# References

  * Beaudette, D.E., P. Roudier, and A.T. O'Geen. 2013. Algorithms for quantitative pedology: A toolkit for soil scientists. Computers and Geoscience 52:258-268.
  * Maynard, J.J., S.W. Salley, D.E. Beaudette, and J.E. Herrick. 2020. Numerical soil classification supports soil identification by citizen scientists using limited, simple soil observations. Soil Science Society of  America Journal 84:1675-1692.
  
  

---
title: "Overlapping Annotation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Overlapping Annotation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, results='hide', warning=FALSE, message=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = as.integer(Sys.getenv("R_AQP_VIGNETTE_IMAGE_DPI", unset = 32)),
  optipng = knitr::hook_optipng,
  comment = "#>"
)
options(width = 100, stringsAsFactors = FALSE, timeout = 600)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

library(aqp, quietly = TRUE)
.bluecolors <- hcl.colors(n = 25, palette = 'Blues')[-25]
```


## Introduction

The creation of [soil profile sketches](https://ncss-tech.github.io/AQP/aqp/sketches.html) has been a cornerstone of the {aqp} package since the first CRAN release in 2010. With the addition of horizon depth annotation in 2017 it became clear that some form of label placement adjustment would be required to avoid overlapping annotation. Thin horizons and / or large font sizes almost always lead to overlapping horizon depth labels (see numbers just to the right of horizon boundaries in the figure below).

```{r fig.width=6, fig.height=5, echo = FALSE}
data("osd")
par(mar = c(0, 0, 0, 0))
plotSPC(osd, name = NA, fixLabelCollisions = FALSE, cex.names = 1.75, hz.depths = TRUE, depth.axis = FALSE, cex.id = 0.66, name.style = 'center-center')
```

Version 2.0 of the {aqp} package provides two algorithms for avoiding overlapping annotation, available in `plotSPC()` for the adjustment of horizon depth annotation, and for general use via `fixOverlap()`. Within the context of `plotSPC()`, overlapping horizon depth labels are flagged according to an overlap threshold computed from label text height on the current graphics device.

Consider the following example soil profile sketches, created from a single soil profile template characterized by multiple thin horizons. Simulation and duplication were used to generate multiple copies of the three variations on the original template.

```{r fig.width=8, fig.height=5}
library(aqp)

# soil profile template
x <- quickSPC(
  "SPC:Oi|Oe|AAAAA|E1|E2|Bhs1Bhs1Bhs1|Bhs2|CCCCCC|RRRRRRRRRR", 
  interval = 3
)

# set variability in horizon thickness
horizons(x)$thick.sd <- 3
# setup horizon designations
horizons(x)$nm <- factor(x$name, levels = c('Oi', 'Oe', 'A', 'E1', 'E2', 'Bhs1', 'Bhs2', 'C', 'R'))

# simulate 3 realizations of original profile
set.seed(10101)
s <- perturb(x, n = 3, thickness.attr = 'thick.sd', min.thickness = 2)

# create 3 copies of each simulation
s <- duplicate(s, times = 3)
```


Overlapping horizon depth labels are clear. Note that there are three copies of each simulated variation on the original template. Colors represent sequence from top to bottom, as a visual aid across perturbed and duplicated profiles.
```{r fig.width=8, fig.height=5, echo = FALSE}
.a <- list(cex.names = 0.8, color = 'nm', name = NA, show.legend = FALSE, hz.depths = TRUE, depth.axis = FALSE, cex.id = 0.66, col.palette = .bluecolors)

options(.aqp.plotSPC.args = .a)

par(mar = c(0, 0, 0, 0))
plotSPC(s, fixLabelCollisions = FALSE)
title('Example Data', line = -1.5)
```


Selecting the optimal label-adjustment methods and parameters will depend on the collection of profiles, horizonation, profile height, font sizes, and graphics device settings (figure height, resolution).

  * *electrostatic simulation:* This is generally the best method, having deterministic solutions that typically move labels as little as possible in fewer iterations. It does not do well with many very thin horizons near the top or bottom of the profile. Very complex profiles and larger font sizes may require setting `q` to values larger than 1. Set `q` to values between 0.25-0.75 to minimize adjustment distances. This method cannot (yet) provide a label adjustment solution in the presence of missing horizon depths or horizons where `top == bottom`.
  
  * *simulated annealing:* This method is the most adaptive, typically capable of providing reasonable solutions for even the most complex profiles. The non-deterministic solutions can be undesirable when preparing figures for publication. Use `set.seed()` to control output between runs.


### Electrostatic Simulation

An electrostatic simulation applies forces of repulsion between labels that are within a given distance threshold and forces of attraction to a uniformly spaced sequence. Affected labels are iteratively perturbed until either no overlap is reported, or a maximum number of iterations has been reached. Label adjustment solutions are deterministic and can be controlled by additional arguments to `electroStatics_1D()`. The most common adjustment is via charge density (`q`). See comparisons below.

```{r fig.width=8, fig.height=5, echo = FALSE}
par(mar = c(0, 0, 0, 0))
# select "Electrostatic" label placement method
# adjust default charge density to 0.5
plotSPC(s, fixOverlapArgs = list(method = 'E', q = 0.5))
title('Label Adjustment by Electrostatic Simulation', line = -1.5)
```



### Simulated Annealing
This approach makes small adjustments to affected label positions until overlap is removed, or until a maximum number of iterations is reached. Rank order and boundary conditions are preserved. The underlying algorithm is based on simulated annealing. Label placement solutions are non-deterministic, use `set.seed()` if repeatability is important. The "cooling schedule" parameters `T0` and `k` can be used to tune the algorithm for specific applications. These additional arguments are passed to `SANN_1D()` via `fixOverlap()`.

Note that each solution, within duplicate profiles, is slightly different.
```{r fig.width=8, fig.height=5, echo = FALSE}
par(mar = c(0, 0, 0, 0))
plotSPC(s, fixOverlapArgs = list(method = 'S'))
title('Label Adjustment by Simulated Annealing', line = -1.5)

options(.aqp.plotSPC.args = NULL)
```


## Quick Comparison

Create a very busy profile with lots of possible overlapping horizon depth labels.
```{r fig.width=8, fig.height=5}
x <- quickSPC(
  "SPC:Oi|Oe|AAA|E1|E2|E3|BhsBhsBhsBhs|Bt1|Bt2|Bt3Bt3|CCCCCC|Ab1|Ab2|2C2C2C2C2C2C|2Cr|2R2R2R2R2R2R2R2R", 
  interval = 1
)

x$z <- as.numeric(x$hzID)
```

Set arguments to `plotSPC()` and define a custom function to demonstrate various label adjustment settings. 
```{r fig.width=8, fig.height=5}
# pretty colors
.bluecolors <- hcl.colors(n = 25, palette = 'Blues')[-25]

# plotSPC arguments
.a <- list(
  width = 0.2, 
  max.depth = 40,
  hz.depths = TRUE, 
  name.style = 'center-center', 
  cex.names = 1.5,
  name = NA,
  depth.axis = FALSE, 
  color = 'z',
  show.legend = FALSE,
  print.id = FALSE,
  col.palette = .bluecolors
)

# set plotSPC default arguments
options(.aqp.plotSPC.args = .a)

# wrapper function to test label collision solutions
testIt <- function(x, ...) {
  
  # make sketches
  plotSPC(x, ...)
  
  # a normalized index of label adjustment
  .LAI <- get('last_spc_plot', envir = aqp.env)$hz.depth.LAI
  .LAI <- ifelse(is.na(.LAI), 0, .LAI)
  
  # annotate with label adjustment index
  .txt <- sprintf("LAI: %0.3f", .LAI)
  mtext(.txt, side = 1, at = 1, line = -1.5, cex = 0.8)
}
```

Compare and contrast. The "LAI" is an index of label adjustment, relative to the original label positions. Larger values imply a higher degree of adjustment. Note that each simulated annealing (SANN) label adjustment is different, unless randomness is "controlled" using `set.seed()`.
```{r fig.width=8, fig.height=5}
par(mar = c(1, 0, 0, 0), mfcol = c(1, 8))

testIt(x, fixLabelCollisions = FALSE)
title('No\nAdjustment', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 1', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 2', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 3', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 1.5))
title('Electrostatic\nq = 1.5', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 1))
title('Electrostatic\nq = 1', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 0.5))
title('Electrostatic\nq = 0.5', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 0.25))
title('Electrostatic\nq = 0.25', line = -3.5, adj = 0.5)
```

```{r eval=requireNamespace("soilDB", quietly=TRUE), fig.width=8, fig.height=5}
library(aqp)
library(soilDB)

s <- c('inks' , 'pardee', 'clarksville', 'palau', 'hao', 'inks', 'eheuiki', 'puaulu', 'zook', 'cecil')
x <- fetchOSD(s)

par(mar = c(0, 0, 0, 3))

.args <- list(width = 0.3, name.style = 'center-center', hz.depths = TRUE, cex.names = 1)
options(.aqp.plotSPC.args = .args)

plotSPC(x, fixOverlapArgs = list(method = 'E', q = 1.25), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 1), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.75), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.5), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.25), max.depth = 151)

plotSPC(x, fixOverlapArgs = list(method = 'S'), max.depth = 151)
```

## General Cases
The label adjustment functionality can be used outside of the context of `plotSPC()`. The function `overlapMetics()` can be used to determine overlap within a vector of positions based on a given distance threshold. 
```{r}
x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)

overlapMetrics(x, thresh = 0.5)
```

The `fixOverlap()` function will attempt adjustment, given distance threshold and a method selection. Additional parameters are passed on to `SANN_1D()` or `electroStatics_1D()`. The `converged` attribute is included in the result to signify that a successful solution was possible. See the manual page for `fixOverlap()` for additional details.
```{r message=TRUE}
# vector of positions, typically labels but could be profile sketch alignment on the x-axis
s <- c(1, 2, 2.3, 4, 5, 5.5, 7)

# simulated annealing, solution is non-deterministic
fixOverlap(s, thresh = 0.5, method = 'S')

# electrostatics-inspired simulation of particles
# solution is deterministic
fixOverlap(s, thresh = 0.5, method = 'E')
```



## Compare Algorithms

Define a custom function for comparing the extended output from `fixOverlap(..., trace = TRUE)` using both electrostatic simulation (`method = 'E'`) and simulated annealing (`method = 'S')`.
```{r fig.width=8.5, fig.height=5}
evalMethods <- function(x, thresh, q, ...) {
  
  cols <- hcl.colors(n = 9, palette = 'Zissou 1', rev = TRUE)
  cols <- colorRampPalette(cols)(length(x))
  
  z <- fixOverlap(x, thresh = thresh, method = 'E', maxIter = 100, trace = TRUE, q = q)
  .n <- nrow(z$states)
  
  op <- par(mar = c(0, 2, 2, 0.5), bg = 'black', fg = 'white')
  layout(matrix(c(1, 2, 3, 4), ncol = 2, nrow = 2), heights = c(0.33, 0.66))
  
  plot(seq_along(z$cost), z$cost, las = 1, type = 'b', axes = FALSE, cex = 0.66, xlim = c(1, .n))
  mtext(text = sprintf("Converged (%s): %s", .n, z$converged), at = 0, side = 3, line = 0, cex = 0.75, font = 3, adj = 0)
  matplot(rbind(x, z$states), type = 'l', lty = 1, las = 1, axes = FALSE, col = cols, lwd = 1)
  
  points(x = rep(1, times = length(x)), y = x, cex = 0.66, pch = 16, col = cols)
  points(x = rep(.n + 1, times = length(x)), y = z$x, cex = 0.66, pch = 16, col = cols)
  
  text(x = 1, y = x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 2)
  text(x = .n + 1, y = z$x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 4)
  
  axis(side = 2, at = unique(x), labels = round(unique(x), 1), col.axis = par('fg'), las = 1, cex.axis = 0.6)
  title(main = 'Electrostatic Simulation', line = 1, col.main = 'white')
  
  ## SANN_1D doesn't always preserve rank ordering
  ##  ->> not designed to use unsorted input
  ##  ->> maybe impossible with ties in x?
  
  z <- fixOverlap(x, thresh = thresh, method = 'S', trace = TRUE, maxIter = 1000)
  .n <- nrow(z$states)
  
  plot(seq_along(z$stats), z$stats, las = 1, type = 'b', axes = FALSE, cex = 0.66, xlim = c(1, .n))
  mtext(text = sprintf("Converged (%s): %s", .n, z$converged), at = 0, side = 3, line = 0, cex = 0.75, font = 3, adj = 0)
  
  matplot(z$states, type = 'l', lty = 1, las = 1, axes = FALSE, col = cols)
  
  points(x = rep(1, times = length(x)), y = z$states[1, ], cex = 0.66, pch = 16, col = cols)
  points(x = rep(.n, times = length(x)), y = z$x, cex = 0.66, pch = 16, col = cols)
  
  text(x = 1, y = z$states[1, ], col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 2)
  text(x = .n, y = z$x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 4)
  
  axis(side = 2, at = unique(x), labels = round(unique(x), 1), col.axis = par('fg'), las = 1, cex.axis = 0.6)
  
  title(main = 'Simulated Annealing', line = 1, col.main = 'white')
  
  # reset graphics state
  par(op)
  layout(1)
  
}
```

The top panels represent the objective function, smaller values are better. Bottom panels illustrate perturbations to elements of the position vector `x`. Larger values of charge density `q` are required applying the electrostatic simulation method to increasingly complex problems (e.g. larger thresholds, more overlap). However, setting `q` too high will result in chaos and failure to converge.
```{r fig.width=8.5, fig.height=5}
# explore effect of charge (q)
# too large -> chaos
x <- c(0, 2, 5, 12, 18, 20, 35, 40, 50, 56, 90, 120, 145, 150)

# just about right, very few perturbations required
evalMethods(x, thresh = 5, q = 1.1)

# ok, but now most label positions are affected
evalMethods(x, thresh = 5, q = 1.8)

# too high, wasting time on more iterations
evalMethods(x, thresh = 5, q = 3)

# far too high, wasting more time with little gain
evalMethods(x, thresh = 5, q = 4)

# chaos and failure to converge
evalMethods(x, thresh = 5, q = 5)
```

Additional examples to tinker with.
```{r, eval = FALSE}
# threshold too large
evalMethods(x, thresh = 10, q = 3)


# large threshold
x <- c(0, 5, 12, 18, 20, 35, 40, 55, 90, 120, 145, 150)
evalMethods(x, thresh = 9, q = 2)

# single iteration enough
x <- c(0, 3, 20, 35, 40, 55, 90, 120, 145, 150)
evalMethods(x, thresh = 6, q = 1)

# clusters
x <- sort(c(0, jitter(rep(10, 3)), jitter(rep(25, 3)), jitter(rep(90, 3)), 150))
evalMethods(x, thresh = 6, q = 3)
evalMethods(x, thresh = 6, q = 2)


## impact of scale / offset
x <- c(0, 5, 12, 18, 20, 35, 40, 50, 120, 145, 150)

# works as expected
evalMethods(x, thresh = 5, q = 1.1)

# works as expected, as long as threshold is scaled
evalMethods(x / 10, thresh = 5 / 10, q = 1.1)

# works as expected, as long as threshold is scaled
evalMethods(x * 10, thresh = 5 * 10, q = 1.1)


# all work as expected, threshold not modified
evalMethods(x + 10, thresh = 5, q = 1.1)
evalMethods(x + 100, thresh = 5, q = 1.1)
evalMethods(x + 1000, thresh = 5, q = 1.1)

# works as expected
x <- c(315, 325, 341, 353, 366, 374, 422)
fixOverlap(x, thresh = 9.7, q = 1, method = 'E')
evalMethods(x, thresh = 9.7, q = 1)


x <- c(1.0075, 1.1200, 1.3450, 1.6450, 1.8700, 1.8825)
fixOverlap(x, thresh = 0.05442329, q = 1)
evalMethods(x, thresh = 0.05442329, q = 1)
```


## Learning More about the SANN Algorithm

Define a custom function for visualizing the extended output from `fixOverlap(..., method = 'S', trace = TRUE)`.
```{r fig.width=8, fig.height=6}
tracePlot <- function(x, z, cex.axis.labels = 0.85) {
  # setup plot device
  op <- par(mar = c(4, 4, 1, 1), bg = 'black', fg = 'white')
  layout(matrix(c(1,2,3)), widths = 1, heights = c(1,1,2))
  
  # order:
  # B: boundary condition violation
  # O: rank (order) violation
  # +: accepted perturbation
  # -: rejected perturbation
  cols <- c(grey(0.5), grey(0.85), 'royalblue', 'firebrick')
  
  cols.lines <- hcl.colors(9, 'Zissou 1', rev = TRUE)
  cols.lines <- colorRampPalette(cols.lines)(length(x))
  
  # total overlap (objective function) progress
  plot(
    seq_along(z$stats), z$stats, 
    type = 'h', las = 1,
    xlab = 'Iteration', ylab = 'Total Overlap',
    axes = FALSE,
    col = cols[as.numeric(z$log)]
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, line = -2)
  mtext('Overlap', side = 2, line = 2, cex = cex.axis.labels, font = 2)
  
  # deviation from original configuration
  plot(
    seq_along(z$stats), z$ssd, 
    type = 'h', las = 1,
    xlab = 'Iteration', ylab = 'Deviation',
    axes = FALSE,
    col = cols[as.numeric(z$log)]
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, line = -2)
  mtext('Deviation', side = 2, line = 2, cex = cex.axis.labels, font = 2)
  legend('top', legend = c('boundary\nviolation', 'rank\nviolation', 'accepted\nperturbation', 'rejected\nperturbation'), col = cols, bty = 'n', horiz = TRUE, inset = -0.5, lty = 1, lwd = 2, xpd = NA)
  
  # adjustments at each iteration
  matplot(
    z$states, type = 'l', 
    lty = 1, las = 1, 
    xlab = 'Iteration', ylab = 'x-position',
    axes = FALSE,
    col = cols.lines
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, at = x, labels = round(x, 1))
  axis(side = 4, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, at = z$x, labels = round(z$x, 1), line = -2)
  mtext('Position', side = 2, line = 2.5, cex = cex.axis.labels, font = 2)
  
  axis(side = 1, cex.axis = 1, col.axis = 'white', line = 0)
  mtext('Iteration', side = 1, line = 2.5, cex = cex.axis.labels, font = 2)
  
  par(op)
  layout(1)
}
```

A relatively challenging example.
```{r fig.width=8, fig.height=6, message=TRUE}
x <- c(0, 1, 2, 2.2, 2.8, 3.5, 6, 8, 10, 10.1, 12.8, 13, 14.8, 15, 15.5)

# fix overlap, return debugging information
set.seed(10101)
z <- fixOverlap(x, thresh = 0.73, method = 'S', trace = TRUE)

# check convergence
z$converged

# inspect algorithm trace
tracePlot(x, z)

# trace log
# B: boundary condition violation
# O: rank (order) violation
# +: accepted perturbation
# -: rejected perturbation
table(z$log)
```

A very challenging example.
```{r fig.width=8, fig.height=6, message=TRUE}
# fix overlap, return debugging information
set.seed(101010)
x <- sort(runif(10, min = 2.5, max = 3.5))

# widen boundary conditions
z <- fixOverlap(x, thresh = 0.2, trace = TRUE, min.x = 0, max.x = 10, maxIter = 2000, adj = 0.05)

# check convergence
z$converged

# inspect algorithm trace
tracePlot(x, z)
```

Cleanup.
```{r}
# reset plotSPC() options
options(.aqp.plotSPC.args = NULL)
```

---
title: "An Overview of the aqp Package"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{An Overview of the aqp Package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, results='hide', warning=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = as.integer(Sys.getenv("R_AQP_VIGNETTE_IMAGE_DPI", unset = 32)),
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)
```


Soil morphology, observed properties, and geomorphic context represent a complex package of interrelated information that can be difficult to analyze and communicate as a whole. Graphical methods such as soil profile sketches and cross-sections represent a few of the possible methods commonly used to report on these kind of data. The Algorithms for Quantitative Pedology (AQP) project encompasses several related R packages tailored to this style of work. A specialized data structure (SoilProfileCollection) maintains linkages between soil horizons, diagnostic features, above-ground data, and geomorphic context. SoilProfileCollection objects can be filtered, subset, resampled (over new depth intervals), and re-ordered; all while preserving links to above and below-ground, linked data. Functions are provided for the conversion of soil colors to and from Munsell notation and several other color space coordinates. Graphical methods for the SoilProfileCollection provide a simple but flexible framework for the design and layout of soil profile sketches, aligned to x and or y axes defined by linked data.


# Related Tutorials by Topic


## Soil Morphology
  * [Soil Profile Sketches](https://ncss-tech.github.io/AQP/aqp/sketches.html)
  * [Visualization of Horizon Boundaries](https://ncss-tech.github.io/AQP/aqp/hz-boundaries.html)
  * [Assigning Generalized Horizon Labels](https://ncss-tech.github.io/AQP/aqp/gen-hz-assignment.html)
  * [Dealing With Troublesome Data](https://ncss-tech.github.io/AQP/aqp/dealing-with-bad-data.html)
  * [Simulation of Soil Morphology by Perturbation](https://ncss-tech.github.io/AQP/aqp/perturb.html)
  * [Soil Texture Visualization Ideas](https://ncss-tech.github.io/AQP/aqp/soiltexture-vizualization-ideas.html)
  * [SPC Sketch Tips](https://ncss-tech.github.io/AQP/aqp/SPC-plotting-ideas.html)
  * [Vertical vs. Perpendicular Horizon Depths](https://ncss-tech.github.io/AQP/aqp/vertical_vs_perpendicular.html)

## Soil Color
  * [Mixing Munsell Colors](https://ncss-tech.github.io/AQP/aqp/mix-colors.html)
  * [Soil Color Contrast](https://ncss-tech.github.io/AQP/aqp/color-contrast.html)
  * [Some ideas on summarizing soil color](https://ncss-tech.github.io/AQP/sharpshootR/aggregate-soil-color.html)
  * [Soil Color Tiles](https://ncss-tech.github.io/AQP/aqp/color-tiles.html)
  * [Investigating Soil Color](https://ncss-tech.github.io/AQP/aqp/investigating-soil-color.html)
  * [Some Ideas on Soil Color Signatures](https://ncss-tech.github.io/AQP/aqp/soil-color-signatures.html)
  * [Soil Color Transition Probability](https://ncss-tech.github.io/AQP/aqp/series-color-TP-graph.html)

## Numerical Classification
  * [Pair-Wise Distances by Generalized Horizon Labels](https://ncss-tech.github.io/AQP/aqp/genhz-distance-eval.html)
  * [Pair-Wise Distance Between Profiles and Mixed Variable Types](https://ncss-tech.github.io/AQP/aqp/profile-compare-factors.html)

## Soil Hydrology  
  * [Theoretical Water Retention Curves](https://ncss-tech.github.io/AQP/aqp/water-retention-curves.html)
  * [Water Retention Curve Development from KSSL Data](https://ncss-tech.github.io/AQP/soilDB/fetchKSSL-VG-demo.html)

## Geomorphology
  * [Hydrologic Ordering of Geomorphic Proportions](https://ncss-tech.github.io/AQP/sharpshootR/geomorphic-summaries-and-ordering.html)
  * [Exploring Geomorphic Summaries](https://ncss-tech.github.io/AQP/soilDB/exploring-geomorph-summary.html)

## Soil Taxonomy
  * [What does a subgroup look like?](https://ncss-tech.github.io/AQP/soilDB/subgroup-series.html)
  * [Querying Soil Series Data](https://ncss-tech.github.io/AQP/soilDB/soil-series-query-functions.html)

## Multivariate Summaries of Soil Property Data
  * [L1 Profiles](https://ncss-tech.github.io/AQP/aqp/L1-profiles.html)
  * [Getting and Comparing KSSL Data](https://ncss-tech.github.io/AQP/soilDB/KSSL-demo.html)

## Soil Spectroscopy
  * [Soil Colors from VNIR Spectra](https://ncss-tech.github.io/AQP/aqp/VNIR-spectra-to-color.html)



# Function Index by Topic


## SoilProfileCollection Objects

### Creation
  * `depths()`: init an SPC from `data.frame`
  * `site()`: set or add site-level attributes of an SPC  
  * `quickSPC()`: quickly build an SPC from simple text templates
  * `random_profile()`: generate random SPC from suite of depth functions
  * `rp()`: convenient wrapper of `random_profile()` for simulating large colections

### Metadata  
  * `hzdesgnname()`: get/set column containing horizon designations
  * `hzDesgn()`: get vector of horizon designations
  * `hztexclname()`: get/set column containing horizon texture class
  * `metadata()`: get/set SPC metadata (list)
  * `hzID()`: get vector of horizon IDs
  * `hzidname()`: get/set column containing horizon IDs
  * `horizonDepths()` get/set columns containing horizon top and bottom depths

### Properties
  * `length()`: number of profiles in a SPC
  * `nrow()`: number of horizons in a SPC
  * `names()`: list of horizon and site names
  * `siteNames()`: site-level column names
  * `horizonNames()`: horizon-level column names

### Subset
  * `glom()`: extract horizons based on overlap criteria defined by point or interval
  * `trunc()`: truncate SPC to given depth interval
  * `subset()`: subset profiles based on logical expressions
  * `subsetHz()`: subset horizons based on logical expressions
  * `[`: `data.frame`-like subsetting of profiles (i-index) and/or horizons (j-index)
  * `[[`: access site or horizon-level columns by name
  * `k-index expressions`: `.FIRST`, `.LAST`, `.HZID`, `.NHZ`, `.BOTTOM`, `.TOP`

### Depth
  * `min()`: minimum bottom depth within a SPC
  * `max()`: maximum bottom depth within a SPC
  * `depthOf()`: generalized "depth to" based on REGEX matching
  * `minDepthOf()`: special case of `depthOf()`
  * `maxDepthOf()`: special case of `depthOf()`
  * `getSoilDepthClass()`: estimate soil depth based on REGEX matching applied to horizon designation and associated depth class
  * `aggregateSoilDepth()`: statistical estimation of soil depth (REGEX matching of horizon designation) within groups of profiles
  
  
### Utility
  * `combine()`, `c()`: combine multiple SPCs into a single SPC
  * `duplicate()`: duplicate profiles within a SPC
  * `perturb()`: randomly adjust horizon thickness or depths to simulate from a template SPC
  * `shuffle()`: randomly shuffle horizon data or entire horizons
  * `warpHorizons()`: expand / contract horizon thickness
  * `harmonize()`: create new profiles within a SPC based sets of related horizon-level data
  * `hzAbove()`, `hzBelow()`: locate horizons above or below some criteria
  * `unique()`: determine uniqueness among profiles of an SPC via MD5 hash
  * `split()`: split SPC into list of SPCs based on grouping factor
  * `site()`: get site data as `data.frame`
  * `horizons()`: get horizon data as `data.frame`
  * `replaceHorizons()`: replace horizon data
  * `diagnostic_hz()`: get/set diagnostic features
  * `restrictions()`: get/set restrictions
  * `denormalize()`: convert site-level data into horizon-level data via replication
  * `compositeSPC()`: downgrade an SPC to list of site and horizon-level data
  
### Iteration
  * `profileApply()`: apply a function to each profile within an SPC (slow but simple interface)
  * `summarizeSPC()`: perform group-wise summaries over profiles within an SPC
  * `transform()`: modify a SPC using expressions that operation on site or horizon-level data

### Change of Support
  * `dice()`: convert SPC to 1 depth-unit intervals by replication
  * `slab()`: apply an aggregate function over groups within a "dice()-ed" SPC
  * `spc2mpspline()`: interface to equal-area spline fitting from mpspline2 package
  * `segment()`: generate segment labels for depth-weighted aggregation 
  * `L1_profiles()`: create representative profiles via multivariate median (L1 estimator)
  * `slicedHSD()`: apply Tukey's HSD over groups within a "dice()-ed" SPC

### Horizon Depth Logic
  * `accumulateDepths()`: fix horizon depths when old-style O horizon notation has been used
  * `fillHzGaps()`: fill topological gaps in horizon depth
  * `repairMissingHzDepths()`: attempt fixing missing or duplicated horizon bottom depths
  * `flagOverlappingHz()`: flag horizons with perfect overlap
  * `checkHzDepthLogic()`: apply battery of horizon depth topological tests
  * `splitLogicErrors()`: split an SPC according to variety of possibly horizon depth errors
  * `HzDepthLogicSubset()`: remove profiles from an SPC if any depth logic errors are present

### Data QC
  * `evalMissingData()`: report metrics of missing data by profile within SPC
  * `missingDataGrid()`: visual indication of missing data
  * `profileInformationIndex()`: experimental indices of "information content" by profile


### Object Coercion
  * `as(SPC, 'list')`: convert SPC to `list`
  * `as(SPC, 'data.frame')`: convert site and horizon data to `data.frame`
  * `as(SPC, 'sf')`: convert site and spatial data to sf object

### Spatial Data
  * `prj()`: get/set coordinate reference system (CRS) metadata
  * `initSpatial()`: set (site-level) column names containing coordinates
  * `getSpatial()`: get spatial data (site + coordinates) from an SPC

### Internal Consistency
  * `checkSPC()`: check SPC for internal consistency 
  * `rebuildSPC()`: re-make an SPC from a previous version of aqp (rarely required)

## Soil Profile Sketches
  * `plotSPC()`: create sketches of soil profiles from an SPC
  * `explainPlotSPC()`: explain various elements of a soil profile sketch
  * `groupedProfilePlot()`: create grouped sketches of soil profiles
  * `plotProfileDendrogram()`: "hang" profile sketches from a dendrogram
  * `plotMultipleSPC()`: create sketches from multiple, possibly non-conformal SPCs
  * `addBracket()`: add vertical brackets beside soil profile sketches
  * `addDiagnosticBracket()`: add vertical brackets beside soil profile sketches based on diagnostic features
  * `addVolumeFraction()`: add visual explanation of coarse fragment volume to soil profile sketches
  * `alignTransect()`: align soil profiles to an external gradient such as topographic sequence


## Soil Color / Color Science
  
### Color Conversion
  * `col2Munsell()`: convert various color notations to Munsell notation
  * `munsell2rgb()`: convert Munsell notation to sRGB or CIELAB color coordinates
  * `parseMunsell()`: parse and optionally convert Munsell color notation
  * `spec2Munsell()`: estimate the closest Munsell color given reflectance spectra in the visible range 
  * `getClosestMunsellChip()`: estimate a reasonably close Munsell color given non-standard notation 
  * `estimateSoilColor()`: estimate moist soil color from dry soil color (and vice versa) 

### Comparison
  * `colorContrast()`: pair-wise color contrast and CIE2000 (dE00) based on colors in Munsell notation 
  * `colorContrastPlot()`: visual explanation of soil color contrast and dE00
  * `contrastChart()`: Munsell color book style explanation of soil color contrast and dE00
  * `soilColorSignature()`: derive soil color signatures for profiles within an SPC

### Aggregation
  * `colorChart()`: Munsell color book representation of color frequency
  * `colorVariation()`: describe color variation using dE00 (based on average human color perception)
  * `aggregateColor()`: estimate color proportions within an SPC according within groups of horizons
  * `aggregateColorPlot()`: graphical explanation of output from `aggregateColor()`
  * `colorQuantiles()`: marginal and L1 quantiles of color in CIELAB coordinates
  

### Utility
  * `validateMunsell()`: test for errors in Munsell notation that prevent correct interpretation
  * `launderMunsell()`: correct minor errors in Munsell notation
  * `huePosition()`: generate an ordered factor of the standard Munsell hues
  * `huePositionCircle()`: graphical representation of the standard Munsell hues, with optional simulation of common color vision deficiency
  * `simulateColor()`: simulate a range of Munsell colors given measures of central tendency and spread
  * `previewColors()`: graphical preview of colors as a grid or via nMDS
  * `soilPalette()`: generate swatch-like arrangements of colors and labels
  * `equivalentMunsellChips()`: for a specified Munsell color, identify other Munsell colors with a very lower CIE2000 color contrast difference

### Simulation of Mixtures
  * `mixMunsell()`: simulate an subtractive mixture of pigments specified in Munsell notation 
  * `plotColorMixture()`: simulate a subtractive mixture of pigments, display reflectance spectra


## Numerical Classification of Soil Profiles
  * `NCSP()`: numerical classification of soil profiles, within a `SoilProfileCollection`
  * `compareSites()`: pair-wise comparison of site-level data from a `SoilProfileCollection`
  
## Pedology
  * `allocate()`: perform one of several classification systems to soil property data
  * `estimateAWC()`: estimate plant-available water holding capacity
  * `correctAWC()`: apply rock fragment or soluble salt corrections to AWC estimates
  * `hzDistinctnessCodeToOffset()`: convert horizon boundary distinctness codes to vertical offset
  * `hzTopographyCodeToLineType()`: convert horizon boundary topography codes to line types
  * `hzTopographyCodeToOffset()`: convert horizon boundary topography codes to offset
  * `ph_to_rxnclass()`: convert pH to reaction class
  * `rxnclass_to_ph()`: convert reaction class to pH range
  * `ReactionClassLevels()`: ordered factor of reaction classes
  * `horizonColorIndices()`: compute various soil color-based indices to horizons within an SPC 

### Soil Texture
  * `textureTriangleSummary()`: graphical summary of sand, silt, clay fractions on a soil texture triangle  
  * `bootstrapSoilTexture()`: simulation of realistic compositions (sand, silt, clay) from a small set of example data
  * `SoilTextureLevels()`: ordered factor of soil texture classes
  * `texcl_to_ssc()`: convert soil texture classes to sand, silt, clay centroids
  * `ssc_to_texcl()`: convert sand, silt, clay values to soil texture class
  * `texture_to_taxpartsize()`: convert soil texture to Soil Taxonomy particle size class
  
### Coarse Fragments
  * `fragmentSieve()`: classify coarse fragments by fragment diameter
  * `texmod_to_fragvoltot()`: estimate ranges in coarse fragment volume based on a soil texture modifier
  * `texture_to_texmod()`
  * `fragvol_to_texmod()`
  * `fragmentClasses()`: coarse fragment diameter thresholds used by USDA-NRCS

### Soil Taxonomy
  * `getArgillicBounds()`: estimate the upper and lower boundaries of an argillic horizon
  * `getCambicBounds()`: estimate the upper and lower boundaries of a cambic horizon
  * `getSurfaceHorizonDepth()`
  * `getMineralSoilSurfaceDepth()`
  * `getPlowLayerDepth()`
  * `hasDarkColors()`
  * `estimatePSCS()`

### Generalized Horizon Labels (GHL)
  * `generalize.hz()`: apply REGEX rules to group horizon designations into a reduced set of "generalized horizon labels"
  * `evalGenHZ()`: evaluate internal consistency of assigned GHL
  * `genhzTableToAdjMat()`: convert a cross-tabulation of GHL vs. original horizon designations to adjacency matrix
  * `get.ml.hz()`: extract most likely horizon boundary depths from probability depth functions 
  * `guessGenHzLevels()`: estimate the correct ordering of GHL given horizon depths
  * `GHL()`: get/set GHL metadata for a `SoilProfileCollection`

## Misc.
  * `invertLabelColor()`: automatic adjustment of label color for maximum contrast, based on specified background color
  * `hzTransitionProbabilities()`: derive transition probability matrix from horizon level data
  * `mostLikelyHzSequence()`: use Markov Chains to predict the most likely sequence of horizons

### Accuracy and Uncertainty
  * `shannonEntropy()`: Shannon entropy
  * `brierScore()`: Brier's score
  * `tauW()`: weighted tau statistic

### Overlapping Annotation  
  * `findOverlap()`: identify overlap within a vector of positions based on a given threshold
  * `overlapMetrics()`: metrics of overlap within a vector or positions based on a given threshold 
  * `fixOverlap()`: attempt the minimum of adjustments to vector of positions such that a given distance threshold is enforced 

## Example Data 
  * `sp1`
  * `sp2`
  * `sp3`
  * `sp4`
  * `sp5`
  * `sp6`
  * `sierraTransect`
  * `wilson2022`
  * `rowley2019`
  * `jacobs2000`
  * `osd`
  * `SPC.with.overlap`
  * `us.state.soils`
  * `soil_minerals`
  * `munsell`
  * `equivalent_munsell`
  * `munsellHuePosition`
  * `munsell.spectra`
  * `spectral.reference`
  * `ROSETTA.centroids`
  * `reactionclass`
  * `soiltexture`

---
title: "Missing Data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Missing Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, results='hide', warning=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = as.integer(Sys.getenv("R_AQP_VIGNETTE_IMAGE_DPI", unset = 32)),
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)
```


```{r}
library(aqp)
```



```{r}
# example data
data("jacobs2000")

# fully populated
plotSPC(jacobs2000, name.style = 'center-center', 
        cex.names = 0.8, color = 'time_saturated')

# missing some data
plotSPC(jacobs2000, name.style = 'center-center', 
        cex.names = 0.8, color = 'concentration_color')

# very nearly complete
plotSPC(jacobs2000, name.style = 'center-center', 
        cex.names = 0.8, color = 'matrix_color')


# variables to consider
v <- c('time_saturated', 'concentration_color', 'matrix_color')

# compute data completeness by profile
# ignore 2C horizons
jacobs2000$data.complete <- evalMissingData(
  jacobs2000, 
  vars = v, 
  method = 'relative',
  p = '2C'
)

jacobs2000$data.complete.abs <- evalMissingData(
  jacobs2000, 
  vars = v, 
  method = 'absolute',
  p = '2C'
)

# compute data completeness by horizon
# ignore 2C horizons
jacobs2000$hz.data.complete <- evalMissingData(
  jacobs2000, 
  vars = v, 
  method = 'horizon',
  p = '2C'
)


# "fraction complete" by horizon
plotSPC(
  jacobs2000, name.style = 'center-center', 
  cex.names = 0.8, color = 'hz.data.complete'
)


# rank on profile completeness
new.order <- order(jacobs2000$data.complete)

# plot along data completeness ranking
plotSPC(
  jacobs2000, name.style = 'center-center', 
  cex.names = 0.8, color = 'hz.data.complete', 
  plot.order = new.order
)

# add relative completeness axis
# note re-ordering of axis labels
axis(
  side = 1, at = 1:length(jacobs2000), 
  labels = round(jacobs2000$data.complete[new.order], 2),
  line = 0, cex.axis = 0.75
)

# add absolute completeness (cm)
axis(
  side = 1, at = 1:length(jacobs2000), 
  labels = jacobs2000$data.complete.abs[new.order],
  line = 2.5, cex.axis=0.75
)

# label axes
mtext('Relative\nCompleteness', side = 1, at = 0.25, line = 0.25, cex = 0.8)
mtext('Absolute\nCompleteness (cm)', side = 1, at = 0.25, line = 2.75, cex = 0.8)
```


```{r, eval=requireNamespace("soilDB", quietly=TRUE)}
library(soilDB)
x <- fetchKSSL(series = 'pierre')
par(mar = c(0, 0, 3, 2))

plotSPC(x, color = 'clay', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')
plotSPC(x, color = 'cec7', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')
plotSPC(x, color = 'estimated_oc', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')
plotSPC(x, color = 'ph_h2o', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')
plotSPC(x, color = 'db_13b', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')


par(mar = c(1, 0, 3, 2))
plotSPC(x, color = 'ph_h2o', width = 0.3, name.style = 'center-center', label = 'pedon_completeness_index')

.b <- x[, , .LAST, .BOTTOM]
text(x = 1:length(x), y = .b, labels = x$pi, cex = 0.85, pos = 1)
mtext('Profile Information Index (bytes)', side = 1, line = -0.5)

v <- c('clay', 'db_13b', 'cec7', 'ph_h2o')
x$rel.not.missing <- evalMissingData(x, vars = v, method = 'relative')
x$abs.not.missing <- evalMissingData(x, vars = v, method = 'absolute')
x$hz.not.missing <- evalMissingData(x, vars = v, method = 'horizon')

o <- order(x$rel.not.missing)
plotSPC(x, color = 'hz.not.missing', width = 0.33, name.style = 'center-center', label = 'pedon_completeness_index', plot.order = o)
text(x = 1:length(x), y = .b[o], labels = round(x$rel.not.missing[o], 2), cex = 0.85, pos = 1)
mtext('Relative Non-Missing Fraction', side = 1, line = -0.5)

o <- order(x$abs.not.missing)
plotSPC(x, color = 'hz.not.missing', width = 0.33, name.style = 'center-center', label = 'pedon_completeness_index', plot.order = o)
text(x = 1:length(x), y = .b[o], labels = x$abs.not.missing[o], cex = 0.85, pos = 1)
mtext('Absolute Non-Missing (cm)', side = 1, line = -0.5)

```

---
title: "Overlapping Annotation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Overlapping Annotation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, results='hide', warning=FALSE, message=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = 96,
  optipng = knitr::hook_optipng,
  comment = "#>"
)
options(width = 100, stringsAsFactors = FALSE, timeout = 600)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

library(aqp, quietly = TRUE)
.bluecolors <- hcl.colors(n = 25, palette = 'Blues')[-25]
```


## Introduction

The creation of [soil profile sketches](https://ncss-tech.github.io/AQP/aqp/sketches.html) has been a cornerstone of the {aqp} package since the first CRAN release in 2010. With the addition of horizon depth annotation in 2017 it became clear that some form of label placement adjustment would be required to avoid overlapping annotation. Thin horizons and / or large font sizes almost always lead to overlapping horizon depth labels (see numbers just to the right of horizon boundaries in the figure below).

```{r fig.width=6, fig.height=5, echo = FALSE}
data("osd")
par(mar = c(0, 0, 0, 0))
plotSPC(osd, name = NA, fixLabelCollisions = FALSE, cex.names = 1.75, hz.depths = TRUE, depth.axis = FALSE, cex.id = 0.66, name.style = 'center-center')
```

Version 2.0 of the {aqp} package provides two algorithms for avoiding overlapping annotation, available in `plotSPC()` for the adjustment of horizon depth annotation, and for general use via `fixOverlap()`. Within the context of `plotSPC()`, overlapping horizon depth labels are flagged according to an overlap threshold computed from label text height on the current graphics device.

Consider the following example soil profile sketches, created from a single soil profile template characterized by multiple thin horizons. Simulation and duplication were used to generate multiple copies of the three variations on the original template.

```{r fig.width=8, fig.height=5}
library(aqp)

# soil profile template
x <- quickSPC(
  "SPC:Oi|Oe|AAAAA|E1|E2|Bhs1Bhs1Bhs1|Bhs2|CCCCCC|RRRRRRRRRR", 
  interval = 3
)

# set variability in horizon thickness
horizons(x)$thick.sd <- 3
# setup horizon designations
horizons(x)$nm <- factor(x$name, levels = c('Oi', 'Oe', 'A', 'E1', 'E2', 'Bhs1', 'Bhs2', 'C', 'R'))

# simulate 3 realizations of original profile
set.seed(10101)
s <- perturb(x, n = 3, thickness.attr = 'thick.sd', min.thickness = 2)

# create 3 copies of each simulation
s <- duplicate(s, times = 3)
```


Overlapping horizon depth labels are clear. Note that there are three copies of each simulated variation on the original template. Colors represent sequence from top to bottom, as a visual aid across perturbed and duplicated profiles.
```{r fig.width=8, fig.height=5, echo = FALSE}
.a <- list(cex.names = 0.8, color = 'nm', name = NA, show.legend = FALSE, hz.depths = TRUE, depth.axis = FALSE, cex.id = 0.66, col.palette = .bluecolors)

options(.aqp.plotSPC.args = .a)

par(mar = c(0, 0, 0, 0))
plotSPC(s, fixLabelCollisions = FALSE)
title('Example Data', line = -1.5)
```


Selecting the optimal label-adjustment methods and parameters will depend on the collection of profiles, horizonation, profile height, font sizes, and graphics device settings (figure height, resolution).

  * *electrostatic simulation:* This is generally the best method, having deterministic solutions that typically move labels as little as possible in fewer iterations. It does not do well with many very thin horizons near the top or bottom of the profile. Very complex profiles and larger font sizes may require setting `q` to values larger than 1. Set `q` to values between 0.25-0.75 to minimize adjustment distances. This method cannot (yet) provide a label adjustment solution in the presence of missing horizon depths or horizons where `top == bottom`.
  
  * *simulated annealing:* This method is the most adaptive, typically capable of providing reasonable solutions for even the most complex profiles. The non-deterministic solutions can be undesirable when preparing figures for publication. Use `set.seed()` to control output between runs.


### Electrostatic Simulation

An electrostatic simulation applies forces of repulsion between labels that are within a given distance threshold and forces of attraction to a uniformly spaced sequence. Affected labels are iteratively perturbed until either no overlap is reported, or a maximum number of iterations has been reached. Label adjustment solutions are deterministic and can be controlled by additional arguments to `electroStatics_1D()`. The most common adjustment is via charge density (`q`). See comparisons below.

```{r fig.width=8, fig.height=5, echo = FALSE}
par(mar = c(0, 0, 0, 0))
# select "Electrostatic" label placement method
# adjust default charge density to 0.5
plotSPC(s, fixOverlapArgs = list(method = 'E', q = 0.5))
title('Label Adjustment by Electrostatic Simulation', line = -1.5)
```



### Simulated Annealing
This approach makes small adjustments to affected label positions until overlap is removed, or until a maximum number of iterations is reached. Rank order and boundary conditions are preserved. The underlying algorithm is based on simulated annealing. Label placement solutions are non-deterministic, use `set.seed()` if repeatability is important. The "cooling schedule" parameters `T0` and `k` can be used to tune the algorithm for specific applications. These additional arguments are passed to `SANN_1D()` via `fixOverlap()`.

Note that each solution, within duplicate profiles, is slightly different.
```{r fig.width=8, fig.height=5, echo = FALSE}
par(mar = c(0, 0, 0, 0))
plotSPC(s, fixOverlapArgs = list(method = 'S'))
title('Label Adjustment by Simulated Annealing', line = -1.5)

options(.aqp.plotSPC.args = NULL)
```


## Quick Comparison

Create a very busy profile with lots of possible overlapping horizon depth labels.
```{r fig.width=8, fig.height=5}
x <- quickSPC(
  "SPC:Oi|Oe|AAA|E1|E2|E3|BhsBhsBhsBhs|Bt1|Bt2|Bt3Bt3|CCCCCC|Ab1|Ab2|2C2C2C2C2C2C|2Cr|2R2R2R2R2R2R2R2R", 
  interval = 1
)

x$z <- as.numeric(x$hzID)
```

Set arguments to `plotSPC()` and define a custom function to demonstrate various label adjustment settings. 
```{r fig.width=8, fig.height=5}
# pretty colors
.bluecolors <- hcl.colors(n = 25, palette = 'Blues')[-25]

# plotSPC arguments
.a <- list(
  width = 0.2, 
  max.depth = 40,
  hz.depths = TRUE, 
  name.style = 'center-center', 
  cex.names = 1.5,
  name = NA,
  depth.axis = FALSE, 
  color = 'z',
  show.legend = FALSE,
  print.id = FALSE,
  col.palette = .bluecolors
)

# set plotSPC default arguments
options(.aqp.plotSPC.args = .a)

# wrapper function to test label collision solutions
testIt <- function(x, ...) {
  
  # make sketches
  plotSPC(x, ...)
  
  # a normalized index of label adjustment
  .LAI <- get('last_spc_plot', envir = aqp.env)$hz.depth.LAI
  .LAI <- ifelse(is.na(.LAI), 0, .LAI)
  
  # annotate with label adjustment index
  .txt <- sprintf("LAI: %0.3f", .LAI)
  mtext(.txt, side = 1, at = 1, line = -1.5, cex = 0.8)
}
```

Compare and contrast. The "LAI" is an index of label adjustment, relative to the original label positions. Larger values imply a higher degree of adjustment. Note that each simulated annealing (SANN) label adjustment is different, unless randomness is "controlled" using `set.seed()`.
```{r fig.width=8, fig.height=5}
par(mar = c(1, 0, 0, 0), mfcol = c(1, 8))

testIt(x, fixLabelCollisions = FALSE)
title('No\nAdjustment', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 1', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 2', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN\nsim 3', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 1.5))
title('Electrostatic\nq = 1.5', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 1))
title('Electrostatic\nq = 1', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 0.5))
title('Electrostatic\nq = 0.5', line = -3.5, adj = 0.5)

testIt(x, fixOverlapArgs = list(method = 'E', q = 0.25))
title('Electrostatic\nq = 0.25', line = -3.5, adj = 0.5)
```

## Compare with Real Soil Series
```{r eval=requireNamespace("soilDB", quietly=TRUE), fig.width=9, fig.height=5}
library(aqp)
library(soilDB)

s <- c('inks' , 'pardee', 'clarksville', 'palau', 'hao', 'inks', 'eheuiki', 'puaulu', 'zook', 'cecil')
x <- fetchOSD(s)

par(mar = c(0, 0, 0, 3))

.args <- list(width = 0.3, name.style = 'center-center', hz.depths = TRUE, cex.names = 0.8)
options(.aqp.plotSPC.args = .args)

plotSPC(x, fixOverlapArgs = list(method = 'E', q = 1.25), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 1), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.75), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.5), max.depth = 151)
plotSPC(x, fixOverlapArgs = list(method = 'E', q = 0.25), max.depth = 151)

plotSPC(x, fixOverlapArgs = list(method = 'S'), max.depth = 151)
```

## General Cases
The label adjustment functionality can be used outside of the context of `plotSPC()`. The function `overlapMetics()` can be used to determine overlap within a vector of positions based on a given distance threshold. 
```{r}
x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)

overlapMetrics(x, thresh = 0.5)
```

The `fixOverlap()` function will attempt adjustment, given distance threshold and a method selection. Additional parameters are passed on to `SANN_1D()` or `electroStatics_1D()`. The `converged` attribute is included in the result to signify that a successful solution was possible. See the manual page for `fixOverlap()` for additional details.
```{r message=TRUE}
# vector of positions, typically labels but could be profile sketch alignment on the x-axis
s <- c(1, 2, 2.3, 4, 5, 5.5, 7)

# simulated annealing, solution is non-deterministic
fixOverlap(s, thresh = 0.5, method = 'S')

# electrostatics-inspired simulation of particles
# solution is deterministic
fixOverlap(s, thresh = 0.5, method = 'E')
```



## Compare Algorithms

Define a custom function for comparing the extended output from `fixOverlap(..., trace = TRUE)` using both electrostatic simulation (`method = 'E'`) and simulated annealing (`method = 'S')`.
```{r fig.width=8.5, fig.height=5}
evalMethods <- function(x, thresh, q, ...) {
  
  cols <- hcl.colors(n = 9, palette = 'Zissou 1', rev = TRUE)
  cols <- colorRampPalette(cols)(length(x))
  
  z <- fixOverlap(x, thresh = thresh, method = 'E', maxIter = 250, trace = TRUE, q = q)
  .n <- nrow(z$states)
  
  op <- par(mar = c(0, 2, 2, 0.5), bg = 'black', fg = 'white')
  layout(matrix(c(1, 2, 3, 4), ncol = 2, nrow = 2), heights = c(0.33, 0.66))
  
  plot(seq_along(z$cost), z$cost, las = 1, type = 'b', axes = FALSE, cex = 0.66, xlim = c(1, .n))
  mtext(text = sprintf("Converged (%s): %s", .n, z$converged), at = 1, side = 3, line = 0, cex = 0.75, font = 3, adj = 0)
  matplot(rbind(x, z$states), type = 'l', lty = 1, las = 1, axes = FALSE, col = cols, lwd = 1)
  
  points(x = rep(1, times = length(x)), y = x, cex = 0.66, pch = 16, col = cols)
  points(x = rep(.n + 1, times = length(x)), y = z$x, cex = 0.66, pch = 16, col = cols)
  
  text(x = 1, y = x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 2)
  text(x = .n + 1, y = z$x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 4)
  
  axis(side = 2, at = unique(x), labels = round(unique(x), 1), col.axis = par('fg'), las = 1, cex.axis = 0.6)
  title(main = 'Electrostatic Simulation', line = 1, col.main = 'white')
  
  ## SANN_1D doesn't always preserve rank ordering
  ##  ->> not designed to use unsorted input
  ##  ->> maybe impossible with ties in x?
  
  z <- fixOverlap(x, thresh = thresh, method = 'S', trace = TRUE, maxIter = 1000)
  .n <- nrow(z$states)
  
  plot(seq_along(z$stats), z$stats, las = 1, type = 'b', axes = FALSE, cex = 0.66, xlim = c(1, .n))
  mtext(text = sprintf("Converged (%s): %s", .n, z$converged), at = 1, side = 3, line = 0, cex = 0.75, font = 3, adj = 0)
  
  matplot(z$states, type = 'l', lty = 1, las = 1, axes = FALSE, col = cols)
  
  points(x = rep(1, times = length(x)), y = z$states[1, ], cex = 0.66, pch = 16, col = cols)
  points(x = rep(.n, times = length(x)), y = z$x, cex = 0.66, pch = 16, col = cols)
  
  text(x = 1, y = z$states[1, ], col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 2)
  text(x = .n, y = z$x, col = cols, labels = seq_along(x), cex = 0.66, font = 2, pos = 4)
  
  axis(side = 2, at = unique(x), labels = round(unique(x), 1), col.axis = par('fg'), las = 1, cex.axis = 0.6)
  
  title(main = 'Simulated Annealing', line = 1, col.main = 'white')
  
  # reset graphics state
  par(op)
  layout(1)
  
}
```

The top panels represent the objective function, smaller values are better. Bottom panels illustrate perturbations to elements of the position vector `x`. Larger values of charge density `q` are required applying the electrostatic simulation method to increasingly complex problems (e.g. larger thresholds, more overlap). However, setting `q` too high will result in chaos and failure to converge.
```{r fig.width=9.5, fig.height=5}
# explore effect of charge (q)
# too large -> chaos
x <- c(0, 2, 5, 12, 18, 20, 35, 40, 50, 56, 90, 120, 145, 150)

# just about right, very few perturbations required
evalMethods(x, thresh = 5, q = 1.1)

# ok, but now most label positions are affected
evalMethods(x, thresh = 5, q = 1.8)

# too high, wasting time on more iterations
evalMethods(x, thresh = 5, q = 3)

# far too high, wasting more time with little gain
evalMethods(x, thresh = 5, q = 4)

# chaos and eventually convergence
evalMethods(x, thresh = 5, q = 5)
```

Additional examples to tinker with.
```{r, eval = FALSE}
# threshold too large
evalMethods(x, thresh = 10, q = 3)

# large threshold
x <- c(0, 5, 12, 18, 20, 35, 40, 55, 90, 120, 145, 150)
evalMethods(x, thresh = 9, q = 2)

# single iteration enough
x <- c(0, 3, 20, 35, 40, 55, 90, 120, 145, 150)
evalMethods(x, thresh = 6, q = 1)

# clusters
x <- sort(c(0, jitter(rep(10, 3)), jitter(rep(25, 3)), jitter(rep(90, 3)), 150))
evalMethods(x, thresh = 6, q = 3)
evalMethods(x, thresh = 6, q = 2)


## impact of scale / offset
x <- c(0, 5, 12, 18, 20, 35, 40, 50, 120, 145, 150)

# works as expected
evalMethods(x, thresh = 5, q = 1.1)

# works as expected, as long as threshold is scaled
evalMethods(x / 10, thresh = 5 / 10, q = 1.1)

# works as expected, as long as threshold is scaled
evalMethods(x * 10, thresh = 5 * 10, q = 1.1)


# all work as expected, threshold not modified
evalMethods(x + 10, thresh = 5, q = 1.1)
evalMethods(x + 100, thresh = 5, q = 1.1)
evalMethods(x + 1000, thresh = 5, q = 1.1)

# works as expected
x <- c(315, 325, 341, 353, 366, 374, 422)
fixOverlap(x, thresh = 9.7, q = 1, method = 'E')
evalMethods(x, thresh = 9.7, q = 1)


x <- c(1.0075, 1.1200, 1.3450, 1.6450, 1.8700, 1.8825)
fixOverlap(x, thresh = 0.05442329, q = 1)
evalMethods(x, thresh = 0.05442329, q = 1)
```


## Learning More about the SANN Algorithm

Define a custom function for visualizing the extended output from `fixOverlap(..., method = 'S', trace = TRUE)`.
```{r fig.width=8, fig.height=6}
tracePlot <- function(x, z, cex.axis.labels = 0.85) {
  # setup plot device
  op <- par(mar = c(4, 4, 1, 1), bg = 'black', fg = 'white')
  layout(matrix(c(1,2,3)), widths = 1, heights = c(1,1,2))
  
  # order:
  # B: boundary condition violation
  # O: rank (order) violation
  # +: accepted perturbation
  # -: rejected perturbation
  cols <- c(grey(0.85), 'yellow', 'royalblue', 'firebrick')
  
  cols.lines <- hcl.colors(9, 'Zissou 1', rev = TRUE)
  cols.lines <- colorRampPalette(cols.lines)(length(x))
  
  # total overlap (objective function) progress
  plot(
    seq_along(z$stats), z$stats, 
    type = 'h', las = 1,
    xlab = 'Iteration', ylab = 'Total Overlap',
    axes = FALSE,
    col = cols[as.numeric(z$log)]
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, line = -2)
  mtext('Overlap', side = 2, line = 2, cex = cex.axis.labels, font = 2)
  
  # deviation from original configuration
  plot(
    seq_along(z$stats), z$ssd, 
    type = 'h', las = 1,
    xlab = 'Iteration', ylab = 'Deviation',
    axes = FALSE,
    col = cols[as.numeric(z$log)]
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, line = -2)
  mtext('Deviation', side = 2, line = 2, cex = cex.axis.labels, font = 2)
  legend('top', legend = c('boundary\nviolation', 'rank\nviolation', 'accepted\nperturbation', 'rejected\nperturbation'), col = cols, bty = 'n', horiz = TRUE, inset = -0.5, lty = 1, lwd = 2, xpd = NA)
  
  # adjustments at each iteration
  matplot(
    z$states, type = 'l', 
    lty = 1, las = 1, 
    xlab = 'Iteration', ylab = 'x-position',
    axes = FALSE,
    col = cols.lines
  )
  
  axis(side = 2, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, at = x, labels = round(x, 1))
  axis(side = 4, cex.axis = cex.axis.labels, col.axis = 'white', las = 1, at = z$x, labels = round(z$x, 1), line = -2)
  mtext('Position', side = 2, line = 2.5, cex = cex.axis.labels, font = 2)
  
  axis(side = 1, cex.axis = 1, col.axis = 'white', line = 0)
  mtext('Iteration', side = 1, line = 2.5, cex = cex.axis.labels, font = 2)
  
  par(op)
  layout(1)
}
```

A relatively challenging example.
```{r fig.width=8, fig.height=6, message=TRUE}
x <- c(0, 1, 2, 2.2, 2.8, 3.5, 6, 8, 10, 10.1, 12.8, 13, 14.8, 15, 15.5)

# fix overlap, return debugging information
set.seed(10101)
z <- fixOverlap(x, thresh = 0.73, method = 'S', trace = TRUE)

# check convergence
z$converged

# inspect algorithm trace
tracePlot(x, z)

# trace log
# B: boundary condition violation
# O: rank (order) violation
# +: accepted perturbation
# -: rejected perturbation
table(z$log)
```

A very challenging example.
```{r fig.width=8, fig.height=6, message=TRUE}
# fix overlap, return debugging information
set.seed(101010)
x <- sort(runif(10, min = 2.5, max = 3.5))

# widen boundary conditions
z <- fixOverlap(x, thresh = 0.2, trace = TRUE, min.x = 0, max.x = 10, maxIter = 2000, adj = 0.05)

# check convergence
z$converged

# inspect algorithm trace
tracePlot(x, z)
```

Cleanup.
```{r}
# reset plotSPC() options
options(.aqp.plotSPC.args = NULL)
```

---
title: "Munsell Color Conversion"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Munsell Color Conversion}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
```{r setup, echo=FALSE, results='hide', warning=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  background = '#F7F7F7',
  fig.align = 'center',
  dev = 'png',
  dpi = 96,
  optipng = knitr::hook_optipng,
  comment = "#>"
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

options(width = 100, stringsAsFactors = FALSE, timeout = 600)

suppressMessages(library(aqp, quietly = TRUE))
```

The `aqp` package provides several functions for working with colors specified in the [Munsell color system](https://en.wikipedia.org/wiki/Munsell_color_system).


## Color Conversion Functions

Conversion of Munsell notation to "something" that can be displayed on-screen (sRGB color representation) is probably the most common color transformation applied to soil morphology data. The `munsell2rgb()` function is convenient for this operation when Munsell notation has been split into vectors of hue, value, and chroma. These could be columns in a `data.frame` or separate vectors. The `parseMunsell()` function is convenient when a full Munsell notation is given as a character vector. Note that `parseMunsell()` is a wrapper to `munsell2rgb()`. For example, converting `10YR 4/6` with either function can return:

* hex-notation of a color: `#805921FF`, or
* [sRGB](https://en.wikipedia.org/wiki/SRGB) color coordinates: `0.5002233 0.3489249 0.1287694`, or
* [CIELAB](https://en.wikipedia.org/wiki/CIELAB_color_space) color coordinates: `40.95021 10.31088 37.49513`.


Selection of the closest `n` Munsell color "chips", given sRGB or CIELAB colorspace coordinates is performed with the `col2Munsell()` function. Input to this function can be colorspace coordinates, named colors (e.g. `red`), or hex-notation of a color. For example, the selection of the closest Munsell chip for CIELAB coordinates `(51.4337, 9.917916, 38.6889)` results in `10YR 5/6` with a reported sigma (error) of `1.5e-6`. This error is estimated as the [CIE2000 distance](https://en.wikipedia.org/wiki/Color_difference#CIEDE2000) between the source CIELAB coordinates and the CIELAB coordinates of the closest Munsell chip.


A representative Munsell color can be estimated from reflectance spectra in the range of 380nm to 730nm with the `spec2Munsell()` function.

```{r echo = FALSE, fig.width=8, fig.height=4}
.m <- '10YR 6/6'

data("munsell.spectra.wide")
w <- munsell.spectra.wide[, 1]
s <- munsell.spectra.wide[, .m]

par(mar = c(4.5, 4.5, 2, 1), cex.axis = 0.75, lend = 2)
plot(w, s, xlab = 'Wavelength (nm)', ylab = 'Reflectance', type = 'b', main = .m, cex = 0.8, pch = 16, axes = FALSE)
rect(xleft = 400, ybottom = 0.35, xright = 450, ytop = 0.4, col = parseMunsell(.m), border = 1, lwd = 1)
axis(side = 1, at = seq(380, 730, by = 20))
axis(side = 2, las = 1)
```


### Special Cases

Neutral colors are commonly specified two ways in the Munsell system: `N 3/` or `N 3/0`, either format will work with `munsell2rgb()` and `parseMunsell()`.

Non-standard Munsell notation (e.g. `3.6YR 4.4 / 5.6`), possibly collected with a sensor vs. color book, can be approximated with `getClosestMunsellChip()`. A more accurate conversion can be performed with the [`munsellinterpol` package.](https://cran.r-project.org/package=munsellinterpol).


### Examples

```{r}
# Munsell -> hex color
parseMunsell('5PB 4/6')

# Munsell -> sRGB
parseMunsell('5PB 4/6',  return_triplets = TRUE)

# Munsell -> CIELAB
parseMunsell('5PB 4/6',  returnLAB = TRUE)

# hex color -> Munsell
col2Munsell('#476189FF')

# neutral color
parseMunsell('N 5/')

# non-standard notation
getClosestMunsellChip('3.3YR 4.4/6.1', convertColors = FALSE)
```


## Estimating Dry ↔ Moist Colors
All else equal, soil color will predictably shift in perceived lightness (change in Munsell value) as moisture content changes. Field-described soil colors are typically collected at approximately air dry ("dry") and field capacity ("moist") states. This function estimates "dry" soil colors from "moist" soil colors and vice versa. Two methods are available for estimation, both developed from a national collection of field-described soil colors (approx. 800k horizons). Estimates are only valid for mineral soil material, having Munsell values and chroma < 10. Estimation has a median error rate of approximately (CIE dE2000) 5. 

Available Methods:
  * "procrustes": soil colors are converted using scale, rotation, and translation parameters in CIELAB color space
  * "ols": soil colors are converted using 3 multiple linear regression models (CIELAB coordinates)


The two methods will give similar results, with `ols` usually closer to observed moist or dry colors. Estimating moist → dry color change is not guaranteed symmetric with estimating dry → moist color.

In the example below, the estimated dry colors are the same for both `method = 'procrustes'` and `method = 'ols'`.
```{r fig.width=10, fig.height=4}
# example moist soil colors from the Musick soil series
m <- c("7.5YR 2.5/1", "10YR 3/2", "7.5YR 4/3", "2.5YR 3/6", "2.5YR 3/6", 
       "2.5YR 4/6", "5YR 4/6", "7.5YR 5/4")

mm <- parseMunsell(m, convertColors = FALSE)

d.p <- estimateSoilColor(
  hue = mm$hue, 
  value = mm$value, 
  chroma = mm$chroma, 
  method = 'procrustes', 
  sourceMoistureState = 'moist'
)

d.ols <- estimateSoilColor(
  hue = mm$hue, 
  value = mm$value, 
  chroma = mm$chroma, 
  method = 'ols', 
  sourceMoistureState = 'moist'
)

d.p <- sprintf("%s %s/%s", d.p$hue, d.p$value, d.p$chroma)
d.ols <- sprintf("%s %s/%s", d.ols$hue, d.ols$value, d.ols$chroma)

colorContrastPlot(m, d.p, labels = c('Moist', 'Estimated\nDry'), d.cex = 0.9)

# it is the same
# colorContrastPlot(m, d.ols, labels = c('Moist', 'Estimated\nDry'), d.cex = 0.9)
```


### Organic Soil Colors
Representative O horizon colors for the most common types of organic soil material, at dry and moist states, have been added to aqp (>=2.3.1) as `Ohz.colors`. Colors were derived from an analysis of soil morphologic data (5245 horizons) within the USDA-NRCS National Soil Information System. Representative colors are the L1 median colors (see `colorVariation()`) within groups generalized to "Oi", "Oe", "Oa", and all "other". These estimates may be useful place-holder values for soil color in collections where O horizon color was not recorded.

```{r fig.width=6, fig.height=3}
data("Ohz.colors")

Ohz.colors$col <- parseMunsell(Ohz.colors$L1.munsell)

op <- par(mfrow = c(2, 1), mar = c(0.5, 0.5, 1.5, 0))

with(
  Ohz.colors[Ohz.colors$state == 'dry', ],
  soilPalette(colors = col, lab = sprintf("%s\n%s", genhz, L1.munsell), lab.cex = 1)
)
title(main = 'Dry Colors')

with(
  Ohz.colors[Ohz.colors$state == 'moist', ],
  soilPalette(colors = col, lab = sprintf("%s\n%s", genhz, L1.munsell), lab.cex = 1)
)
title(main = 'Moist Colors')

# restore original base graphics state
par(op)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-operators.R
\docType{methods}
\name{[[}
\alias{[[}
\alias{[[,SoilProfileCollection-method}
\alias{[[,SoilProfileCollection,ANY-method}
\alias{[[,SoilProfileCollection,ANY,ANY-method}
\title{Get column of horizon or site data in a SoilProfileCollection}
\usage{
\S4method{[[}{SoilProfileCollection}(x, i, j)
}
\arguments{
\item{x}{a SoilProfileCollection}

\item{i}{an expression resolving to a single column name in site or horizon table}

\item{j}{(not used)}
}
\description{
Get the data from a column accessed by name. Column names other than profile ID are not shared between site and horizons.
Bonus: \code{[[} gives access to all site and horizon level variables in tab complete for RStudio using the magrittr pipe operator!
}
\examples{

data(sp2)
depths(sp2) <- id ~ top + bottom
site(sp2) <- ~ surface

# get with [[
sp2[['surface']]

# get using "unknown" expression:
#  "2nd + 3rd horizon column names"
for(i in horizonNames(sp2)[2:3])
 print(sp2[[i]])

data(sp5)

# some column names to work with
rgb.columns <- c("R25","G25","B25")

res <- lapply(rgb.columns, function(x) {

  # [[ allows you to access column names in a loop
  round(sp5[[x]] * 255)

})

# rename scaled results
names(res) <- paste0(rgb.columns,"_scl")

# add horizon ID to results
result <- data.frame(hzID = hzID(sp5), do.call('cbind', res))
head(result)

# join result back into horizons
horizons(sp5) <- result
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatMunsell.R
\name{formatMunsell}
\alias{formatMunsell}
\title{Format Munsell Notation from Hue, Value, and Chroma}
\usage{
formatMunsell(
  hue,
  value,
  chroma,
  neutralConvention = c("zero", "empty"),
  standardHues = TRUE
)
}
\arguments{
\item{hue}{character vector of Munsell hue}

\item{value}{character or numeric vector of Munsell value}

\item{chroma}{character or numeric vector of Munsell chroma}

\item{neutralConvention}{character, neutral color encoding convention

Neutral colors are encoded as:
\itemize{
\item 'empty': N 3/
\item 'zero': N 3/0
}}

\item{standardHues}{logical, when \code{TRUE} non-standard hues are converted to NA}
}
\value{
character vector of same length as \code{hue}, \code{value}, and \code{chroma}
}
\description{
Format hue, value, and chroma values into valid Munsell notation, or NA when not possible. The following rules are applied to the input:
\itemize{
\item for all hues other than N, there must be a valid hue, value, and chroma
\item N hues may specify chroma as 0, NA, or '' (empty string)
\item when \code{standardHues = TRUE}, hue must be a standard Munsell hue, see \code{\link[=huePosition]{huePosition()}}
}
}
\examples{

d <- data.frame(
hue = c('10YR', NA, 'N', 'N', 'N', '5G', '5Z'),
value = c(4, 3, 2, 4, 3, NA, 4),
chroma = c(4, 3, NA, 0, NA_integer_, 6, 3)
)

formatMunsell(d$hue, d$value, d$chroma)
formatMunsell(d$hue, d$value, d$chroma, neutralConvention = 'empty')
formatMunsell(d$hue, d$value, d$chroma, standardHues = FALSE)

# all result in 'N 6/0' (neutralConvention = 'zero')
formatMunsell('N', 6, 0)
formatMunsell('N', 6, '')
formatMunsell('N', 6, NA)

}
\seealso{
\code{\link[=launderMunsell]{launderMunsell()}}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mollicEpipedon.R
\name{hasDarkColors}
\alias{hasDarkColors}
\title{Find horizons with colors darker than a Munsell hue, value, chroma threshold}
\usage{
hasDarkColors(
  p,
  d_hue = NA,
  m_hue = NA,
  d_value = 5,
  d_chroma = NA,
  m_value = 3,
  m_chroma = 3,
  dhuenm = "d_hue",
  dvalnm = "d_value",
  dchrnm = "d_chroma",
  mhuenm = "m_hue",
  mvalnm = "m_value",
  mchrnm = "m_chroma"
)
}
\arguments{
\item{p}{A SoilProfileCollection.}

\item{d_hue}{Optional: character vector of dry hues to match (default: NA)}

\item{m_hue}{Optional: character vector of moist hues to match (default: NA)}

\item{d_value}{Maximum value of dry value (default: 5)}

\item{d_chroma}{Optional: Maximum value of dry chroma (default: NA)}

\item{m_value}{Maximum value of moist value (default: 3)}

\item{m_chroma}{Maximum value of moist chroma (default: 3)}

\item{dhuenm}{Column name containing dry hue.}

\item{dvalnm}{Column name containing dry value.}

\item{dchrnm}{Column name containing dry chroma.}

\item{mhuenm}{Column name containing moist hue.}

\item{mvalnm}{Column name containing moist value.}

\item{mchrnm}{Column name containing moist chroma.}
}
\value{
Boolean value (for each horizon in \code{p}) reflecting whether "darkness" criteria are met.
}
\description{
\code{hasDarkColors} returns a boolean value by horizon representing whether darkness thresholds are met. The code is fully vectorized and deals with missing data and optional thresholds.

Default arguments are set up for "5-3-3 colors" -- the basic criteria for Mollic/Umbric epipedon/mineral soil darkness. Any of the thresholds or column names can be altered. Any thresholds that are set equal to \code{NA} will be ignored.
}
\examples{
# construct a fake profile
spc <- data.frame(id=1, taxsubgrp = "Lithic Haploxeralfs",
                  hzdesgn  = c("A","AB","Bt","BCt","R"),
                  hzdept   = c(0, 20, 32, 42,  49),
                  hzdepb   = c(20, 32, 42, 49, 200),
                  d_value  = c(5,   5,  5,  6,  NA),
                  m_value  = c(2.5, 3,  3,  4,  NA),
                  m_chroma = c(2,   3,  4,  4,  NA))

# promote to SoilProfileCollection
depths(spc) <- id ~ hzdept + hzdepb

# print results in table
data.frame(id = spc[[idname(spc)]],
           hz_desgn = spc$hzdesgn,
           has_dark_colors = hasDarkColors(spc))

}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\name{subset,SoilProfileCollection-method}
\alias{subset,SoilProfileCollection-method}
\alias{subset}
\title{Subset a SoilProfileCollection with logical expressions}
\usage{
\S4method{subset}{SoilProfileCollection}(x, ..., greedy = FALSE)
}
\arguments{
\item{x}{A SoilProfileCollection}

\item{...}{Comma-separated set of R expressions that evaluate as TRUE or FALSE. Length for individual expressions matches number of sites OR number of horizons, in \code{object}.}

\item{greedy}{Use "greedy" matching for combination of site and horizon level matches? \code{greedy = TRUE} is the union, whereas \code{greedy = FALSE} (default) is the intersection (of site and horizon matches).}
}
\value{
A SoilProfileCollection.
}
\description{
\code{subset()} is a function used for extracting profiles from a SoilProfileCollection based on logical criteria. It allows the user to specify an arbitrary number of logical vectors (equal in length to site or horizon), separated by commas. The function includes some support for non-standard evaluation.
}
\details{
To minimize likelihood of issues with non-standard evaluation context, especially when using \code{subset()} inside another function, all expressions used in \code{...} should be in terms of variables that are in the site or horizon data frame.
}
\author{
Andrew G. Brown.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{equivalent_munsell}
\alias{equivalent_munsell}
\title{Indices of "equivalent" Munsell chips in the \code{munsell} data set}
\format{
A named list with 8467 elements, each containing a numeric vector of indices corresponding to the \code{munsell} data set, which has 8467 rows (unique, whole-number chips). Names have the format \code{HUE VALUE/CHROMA}, e.g. \code{"7.5YR 4/4"}
}
\usage{
data(equivalent_munsell)
}
\description{
A pre-calculated lookup list (made with \code{farver::compare_colour}) based on pair-wise color contrast (\code{CIE2000} or \code{dE00}) evaluated over all "chips" in the \code{aqp::munsell} data set.

The intention is to identify Munsell chips that may be "functionally equivalent" to some other given whole chip elsewhere in the Munsell color space -- as discretized in the \code{aqp::munsell} lookup table.

"Equivalent" chips are based (fairly arbitrarily) on the 0.001 probability level of \code{dE00} (default Type 7 \code{quantile}) within the upper triangle of the 8467x8467 contrast matrix. This corresponds to a \code{dE00} threshold of approximately 2.15.

This is a naive (to the subtleties of human color perception, and overall magnitude of contrast between some of the "chips") but computationally consistent approach. Using the lookup list, as opposed to manual contrast via e.g. \code{farver::compare_colour} may have some benefits for efficiency in certain applications where the exact contrast value is not as important as the concept of having some threshold that is non-zero, but very small.
}
\examples{
data(equivalent_munsell)
}
\references{
Gaurav Sharma, Wencheng Wu, Edul N. Dalal. (2005). The CIEDE2000 Color-Difference Formula: Implementation Notes, Supplementary Test Data, and Mathematical Observations. COLOR research and application. 30(1):21-30. http://www2.ece.rochester.edu/~gsharma/ciede2000/ciede2000noteCRNA.pdf

Thomas Lin Pedersen, Berendea Nicolae and Romain Francois (2020). farver: High Performance Colour Space Manipulation. R package version 2.0.3. https://CRAN.R-project.org/package=farver

Dong, C.E., Webb, J.B., Bottrell, M.C., Saginor, I., Lee, B.D. and Stern, L.A. (2020). Strengths, Limitations, and Recommendations for Instrumental Color Measurement in Forensic Soil Characterization. J Forensic Sci, 65: 438-449. https://doi.org/10.1111/1556-4029.14193
}
\seealso{
\code{\link{equivalentMunsellChips}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fragmentSieve.R
\name{fragmentSieve}
\alias{fragmentSieve}
\title{Sieve the Coarse Fraction of Soil}
\usage{
fragmentSieve(
  diameter,
  sieves = NULL,
  ordered = FALSE,
  prefix = "",
  new_names = NULL,
  ...
)
}
\arguments{
\item{diameter}{numeric. Vector of diameters of coarse fragments to "sieve". Default \code{sieves} are specified in millimeters.}

\item{sieves}{leave as \code{NULL} to use fragment class labels and diameters defined by \code{\link[=fragmentClasses]{fragmentClasses()}}, or a named vector of fragment diameters. See examples.}

\item{ordered}{logical. Return as an ordered factor.}

\item{prefix}{character. Add a prefix to result names? Default: \code{""} adds no prefix. For example \code{"para"} might be used for size classes of pararock fragments.}

\item{new_names}{Optional: apply new labels to result classes. Should match length of \code{sieves}.}

\item{...}{additional arguments to \code{\link[=fragmentClasses]{fragmentClasses()}}, such as \code{sys}, \code{flat}, and \code{rounded}, see examples.}
}
\value{
character. Size class labels based on names of \code{sieves}, \code{new_names}, and \code{prefix} (if specified).
}
\description{
Sieve applies thresholds to a numeric vector of fragment diameter values, returning fragment size classes. Particle diameter thresholds are evaluated as \code{d < threshold}.
}
\examples{

# use a simplified version of the USDA system
# common within NRCS/SPSD and NCSS
fragmentSieve(c(30, 125, 180, 500, 1000))

# pararock fragments
fragmentSieve(c(30, 125, 180, 500, 1000), prefix = 'para')

# result as an ordered factor
fragmentSieve(c(30, 125, 180, 500, 1000), ordered = TRUE)

# USDA system, flat size classes
fragmentSieve(c(30, 125, 180, 500, 1000), flat = TRUE)

# alternative classification systems
fragmentSieve(c(30, 125, 180, 500, 1000), sys = 'usda')
fragmentSieve(c(30, 125, 180, 500, 1000), sys = 'international')
fragmentSieve(c(30, 125, 180, 500, 1000), sys = 'unified')
fragmentSieve(c(30, 125, 180, 500, 1000), sys = 'aashto')
fragmentSieve(c(30, 125, 180, 500, 1000), sys = 'mod.wentworth')

# custom fragment labels / diameter
fragmentSieve(
  c(30, 125, 180, 500, 1000),
  sieves = c(clumps = 50, chunks = 300, blocks = 100000)
)

# unnamed sieves, generic labels used
fragmentSieve(c(10, 50), sieves = c(30, 70))

fragmentSieve(c(10, 50), sieves = c(30, 70), ordered = TRUE)
 
}
\references{
Soil Science Division Staff. 2017. Soil survey manual. C. Ditzler, K. Scheffe, and H.C. Monger (eds.). USDA Handbook 18. Government Printing Office, Washington, D.C.
}
\seealso{
\code{\link[=fragmentClasses]{fragmentClasses()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profileInformationIndex.R
\name{profileInformationIndex}
\alias{profileInformationIndex}
\title{Soil Profile Information Index}
\usage{
profileInformationIndex(
  x,
  vars,
  method = c("joint", "individual"),
  baseline = FALSE,
  numericDigits = 8,
  padNA = FALSE,
  scaleNumeric = FALSE,
  compression = "gzip"
)
}
\arguments{
\item{x}{\code{SoilProfileCollection} object}

\item{vars}{character vector of site or horizon level attributes to consider}

\item{method}{character, 'individual' or 'joint' complexity}

\item{baseline}{logical, compute ratio to "baseline" information content, see details}

\item{numericDigits}{integer, number of significant digits to retain in numeric -> character conversion}

\item{padNA}{logical, pad depths to \code{max(x)}, supplied to \code{dice(fill = padNA)}}

\item{scaleNumeric}{logical, \code{scale()} each numeric variable, causing "profile information" to vary based on other profiles in the collection}

\item{compression}{character, compression method as used by \code{\link[=memCompress]{memCompress()}}: 'gzip', 'bzip2', 'xz', 'none'}
}
\value{
a numeric vector of the same length as \code{length(x)} and in the same order, suitable for direct assignment to a new site-level attribute
}
\description{
A simple index of "information" content associated with individuals in a \code{SoilProfileCollection} object. Information content is quantified by number of bytes after compression via \code{memCompress()}.
}
\details{
Information content via compression (gzip) is the central assumption behind this function: the values associated with a simple soil profile having few horizons and little variation between horizons (isotropic depth-functions) will compress to a much smaller size than a complex profile (many horizons, strong anisotropy). Information content is evaluated a profile at a time, over each site or horizon level attribute specified in \code{vars}. The \code{baseline} argument invokes a comparison to the simplest possible representation of each depth-function:
\itemize{
\item \code{numeric}: replication of the mean value to match the number of horizons with non-NA values
\item \code{character} or \code{factor}: replication of the most frequent value to match the number of horizons with non-NA values
}

The ratios computed against a "simple" baseline represent something like "information gain". Larger baseline ratios suggest more complexity (more information) associated with a soil profile's depth-functions. Alternatively, the total quantity of information (in bytes) can be determined by setting \code{baseline = FALSE}.
}
\examples{

# single horizon, constant value
p1 <- data.frame(id = 1, top = 0, bottom = 100, p = 5, name = 'H')

# multiple horizons, constant value
p2 <- data.frame(
  id = 2, top = c(0, 10, 20, 30, 40, 50),
  bottom = c(10, 20, 30, 40, 50, 100),
  p = rep(5, times = 6),
  name = c('A1', 'A2', 'Bw', 'Bt1', 'Bt2', 'C')
)

# multiple horizons, random values
p3 <- data.frame(
  id = 3, top = c(0, 10, 20, 30, 40, 50),
  bottom = c(10, 20, 30, 40, 50, 100),
  p = c(1, 5, 10, 35, 6, 2),
  name = c('A1', 'A2', 'Bw', 'Bt1', 'Bt2', 'C')
)

# multiple horizons, mostly NA
p4 <- data.frame(
  id = 4, top = c(0, 10, 20, 30, 40, 50),
  bottom = c(10, 20, 30, 40, 50, 100),
  p = c(1, NA, NA, NA, NA, NA),
  name = c('A1', 'A2', 'Bw', 'Bt1', 'Bt2', 'C')
)

# shallower version of p1
p5 <- data.frame(id = 5, top = 0, bottom = 50, p = 5, name = 'H')

# combine and upgrade to SPC
z <- rbind(p1, p2, p3, p4, p5)
depths(z) <- id ~ top + bottom
hzdesgnname(z) <- 'name'

z <- fillHzGaps(z)

# visual check
par(mar = c(1, 0, 3, 3))
plotSPC(z, color = 'p', name.style = 'center-center', cex.names = 0.8, max.depth = 110)

# factor version of horizon name
z$fname <- factor(z$name)

vars <- c('p', 'name')
# result is total bytes
pi <- profileInformationIndex(z, vars = vars, method = 'joint', baseline = FALSE)

text(x = 1:5, y = 105, labels = pi, cex = 0.85)
mtext('Profile Information Index (bytes)', side = 1, line = -1)

}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allocate.R
\name{hz_to_taxpartsize}
\alias{hz_to_taxpartsize}
\title{Allocate Particle Size Class for the Control Section.}
\usage{
hz_to_taxpartsize(
  x,
  y,
  taxpartsize = "taxpartsize",
  clay = "clay",
  idcol = "id",
  depthcols = c("top", "bottom")
)
}
\arguments{
\item{x}{a \code{data.frame} containing the original horizon table.}

\item{y}{a \code{data.frame} containing the particle size control section depths for each idcol.}

\item{taxpartsize}{\code{character} column name for taxonomic family particle size class.}

\item{clay}{\code{character} column name for clay percent.}

\item{idcol}{character: column name of the pedon ID within the object.}

\item{depthcols}{a character vector of length 2 specifying the names of the horizon depths (e.g. \code{c("top", "bottom")}).}
}
\value{
A \code{data.frame} object containing the original idcol, the aggregated particle size control section allocation, and an aniso column to indicate more than one contrasting class.
}
\description{
This function aggregates information in the horizon table and allocates it to the particle size class for the control section.
}
\details{
This function differs from \code{\link{texture_to_taxpartsize}} in that is aggregates the results of \code{\link{texture_to_taxpartsize}}, and accounts for strongly contrasting particle size classes.
}
\examples{

h <- data.frame(
  id = 1,
  hzname = c("A", "BA", "Bw", "BC", "C"),
  top    = c(0, 10, 45, 60,  90),
  bottom = c(10, 45, 60, 90, 150),
  clay   = c(15, 16, 45, 20,  10),
  sand   = c(10, 35, 40, 50,  90),
  frags  = c(0,  5, 10, 38,  40)
)

h <- cbind(h,
           texcl = ssc_to_texcl(clay = h$clay, sand = h$sand))

pscs <- data.frame(id = 1,
                   top = 25,
                   bottom = 100)

h <- cbind(h,
           taxpartsize = texture_to_taxpartsize(
             texcl = h$texcl,
             clay  = h$clay,
             sand  = h$sand,
             fragvoltot = h$frags
           ))

depths(h) <- id ~ top + bottom

# set required metadata for estimatePSCS()
hzdesgnname(h) <- "hzname"
hztexclname(h) <- "texcl"
hzmetaname(h, "clay") <- "clay"

pscs <- data.frame(id = h$id, rbind(estimatePSCS(h)))
names(pscs)[2:3] <- c("top", "bottom")

hz_to_taxpartsize(horizons(h), pscs)
 

}
\seealso{
\code{\link[=texture_to_taxpartsize]{texture_to_taxpartsize()}}, \code{\link[=lookup_taxpartsize]{lookup_taxpartsize()}}
}
\author{
Stephen Roecker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{hzID<-,SoilProfileCollection-method}
\alias{hzID<-,SoilProfileCollection-method}
\alias{hzID<-}
\alias{hzID,SoilProfileCollection-method}
\alias{hzID}
\title{Set horizon IDs}
\usage{
\S4method{hzID}{SoilProfileCollection}(object) <- value

\S4method{hzID}{SoilProfileCollection}(object)
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{value}{a unique vector of equal length to number of horizons \code{nrow(object)}}
}
\description{
Set vector containing horizon IDs

Get vector containing horizon IDs
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{us.state.soils}
\alias{us.state.soils}
\title{US State Soils}
\format{
\describe{
\item{state}{state name}
\item{abbreviated}{abbreviated state name}
\item{series}{soil series name}
}
}
\usage{
data(us.state.soils)
}
\description{
A listing of the 50 US state soils, along with Puerto Rico and Virgin Islands.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explainPlotSPC.R
\name{explainPlotSPC}
\alias{explainPlotSPC}
\title{Visual Explanation for \code{plotSPC}}
\usage{
explainPlotSPC(x, ...)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}

\item{\dots}{arguments passed to \code{\link{plotSPC}}}
}
\value{
a list of internally-used ordering vectors and graphical offsets / scaling factors
}
\description{
Create a visual explanation for the many arguments to \code{plotSPC}. Call this function instead of \code{plotSPC}, all objects after \code{x} are passed on to \code{plotSPC}. Nearly all of the figures in the \href{https://ncss-tech.github.io/AQP/aqp/aqp-intro.html}{Introduction to SoilProfileCollection Objects tutorial} are created with this function.
}
\examples{

# sample data
data(sp4)
depths(sp4) <- id ~ top + bottom

# proposed vector of relative positions, overlap likely
pos <- c(1, 1.1, 3, 4, 5, 5.2, 7, 8, 9, 10)

# try it
explainPlotSPC(sp4, name = 'name', relative.pos=pos)

# attempt to fix using an integer sequence, short-circut will prevent adjustments
explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(1:10))

# attempt to adjust using defaults
explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos))

# attempt to adjust and tinker with defaults
explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos, adj = 0.2))

# enforce larger space between
explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos, thresh = 0.7))

# more complex adjustments required
pos <- c(1, 2, 3, 3.3, 5, 5.1, 5.5, 8, 9, 10)

# tinker
explainPlotSPC(sp4, name = 'name', relative.pos = pos)
explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos))

explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos, 
thresh = 0.7))

explainPlotSPC(sp4, name = 'name', relative.pos = fixOverlap(pos, 
thresh = 0.7, adj = 0.2))

# SANN: solution requires many iterations, and will not always converge
explainPlotSPC(sp4, name = 'name', 
relative.pos = fixOverlap(pos, thresh = 0.85, adj = 0.2)
)

# electrostatics: solution requires larger charge (q)
explainPlotSPC(sp4, name = 'name', 
relative.pos = fixOverlap(pos, thresh = 0.85, method = 'E', q = 2)
)

}
\seealso{
\code{\link{plotSPC}}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{spectral.reference}
\alias{spectral.reference}
\title{Standard Illuminants and Observers}
\format{
An object of class \code{data.frame} with 71 rows and 9 columns.
}
\usage{
data(spectral.reference)
}
\description{
D65 and F2 standard illuminant spectral power distributions, CIE1931 Standard Observer (2 degree), and CIE1964 Supplemental Standard Observer (10 degree)
}
\examples{

data("spectral.reference")

matplot(
  x = spectral.reference[, 1],
  y = spectral.reference[, c('xbar_2', 'ybar_2', 'zbar_2')],
  ylim = c(0, 2),
  type = 'l',
  lwd = 2, 
  lty = 1,
  las = 1,
  xlab = 'Wavelength (nm)',
  ylab = 'Weight | Intensity',
  main = "CIE1931 (2\u00B0) and CIE1964 (10\u00B0) Standard Observers
  D65 and F2 Illuminant Power Spectrum (rescaled / offset for clarity)",
  cex.main = 0.9
)

matlines(
  x = spectral.reference[, 1],
  y = spectral.reference[, c('xbar_10', 'ybar_10', 'zbar_10')],
  type = 'l',
  lwd = 2, 
  lty = 2,
  las = 1,
  xlab = 'Wavelength (nm)',
  ylab = 'Weight | Intensity',
  main = 'CIE1931 Standard Observer Weights\nD65 Standard Illuminant'
)

lines(
  x = spectral.reference$w,
  y = (spectral.reference$D65 / 100) + 0.33,
  lty = 1,
  col = 'royalblue'
)

lines(
  x = spectral.reference$w,
  y = (spectral.reference$F2 / 25) + 0.4,
  lty = 1,
  col = 'violet'
)

legend(
  'topright',
  legend = c('X_2', 'Y_2', 'Z_2', 'X_10', 'Y_10', 'Z_10', 'D65', 'F2'),
  col = c(1, 2, 3, 1, 2, 3, 'royalblue', 'violet'),
  lwd = c(2, 2, 2, 2, 2, 2, 1, 1),
  lty = c(1, 1, 1, 2, 2, 2, 1, 1),
  bty = 'n',
  cex = 0.85
)

}
\references{
Marcus, R.T. (1998). The Measurement of Color. In K. Nassau (Ed.), Color for Science, Art, and Technology (pp. 32-96). North-Holland.

CIE Colorimetry – Part 1: CIE standard colorimetric observers. CIES014-1/E:2006 – ISO 11664-1:2007(E)

CIE. (n.d.). CIE 15:2004 Tables Data. Retrieved from https://law.resource.org/pub/us/cfr/ibr/003/cie.15.2004.tables.xls
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getSoilDepthClass.R
\name{getSoilDepthClass}
\alias{getSoilDepthClass}
\title{Generate Soil Depth Class Matrix}
\usage{
getSoilDepthClass(
  f,
  depth.classes = c(very.shallow = 25, shallow = 50, mod.deep = 100, deep = 150,
    very.deep = 10000),
  ...
)
}
\arguments{
\item{f}{a SoilProfileCollection object}

\item{depth.classes}{a named vector of classes and depth breaks}

\item{\dots}{arguments passed to \code{\link{estimateSoilDepth}}}
}
\value{
a \code{data.frame} containing soil depth and depth class for each
profile, see examples
}
\description{
Generate a boolean matrix of soil depth classes, actual soil depth class, and estimate of soil depth from a \code{SoilProfileCollection} object. Soil depths are estimated using pattern matching applied to horizon designations, by \code{\link[=estimateSoilDepth]{estimateSoilDepth()}}. The default REGEX pattern (\code{p = 'Cr|R|Cd'}) will match most "contacts" described using the USDA / Soil Taxonomy horizon designation conventions.
}
\examples{

data(sp1)
depths(sp1) <- id ~ top + bottom

# generate depth-class matrix
sdc <- getSoilDepthClass(sp1, name = 'name')

# inspect
head(sdc)

# join back into sp1 as site-level data
site(sp1) <- sdc

\dontrun{
# sample data
data(gopheridge, package='soilDB')

getSoilDepthClass(gopheridge, name = 'hzname')
}

}
\seealso{
\code{\link{estimateSoilDepth}}
}
\author{
D.E. Beaudette and J.M. Skovlin
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/invertLabelColor.R
\name{invertLabelColor}
\alias{invertLabelColor}
\title{Make High Contrast Label Colors}
\usage{
invertLabelColor(colors, threshold = 0.65)
}
\arguments{
\item{colors}{vector of colors}

\item{threshold}{black | white threshold}
}
\value{
vector of label colors
}
\description{
Generate a vector of white or black label colors conditioned on a vector of colors to maximize label contrast.
}
\examples{

# test with shades of grey
s <- seq(0, 1, by = 0.05)
cols <- grey(s)
soilPalette(cols, lab = as.character(s))

# test with 10YR x/3
m <- sprintf('10YR \%s/3', 1:8)
cols <- parseMunsell(m)
soilPalette(cols, lab = m)


}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{hz_lag}
\alias{hz_lag}
\title{Find lagged horizon values}
\usage{
hz_lag(
  object,
  lag = 1,
  unit = "index",
  idcol = "id",
  depthcols = c("top", "bottom"),
  order = FALSE
)
}
\arguments{
\item{object}{a \code{data.frame}}

\item{lag}{integer: number of horizons to lag}

\item{unit}{character: lag units in index or depth.}

\item{idcol}{character: column name of the pedon ID within the object.}

\item{depthcols}{a character vector of length 2 specifying the names of the horizon depths (e.g. \code{c("top", "bottom")}).}

\item{order}{logical: indicating whether or not to order the #'}
}
\value{
A \code{data.frame} with lagged values.
}
\description{
This function finds adjacent values to a horizon values at lagged distances.
}
\details{
.
}
\examples{

h <- data.frame(
  id = 1,
  top    = c(0,  25, 44, 46, 50),
  bottom = c(25, 44, 46, 50, 100),
  texcl     = c("SL", "SL", "CL", "CL", "L"),
  clay   = c(10, 12, 27, 35, 16)
)

hz_lag(h)

hz_lag(h, -1)

hz_lag(h, 10:15, unit = "depth")

transform(cbind(h, lag = hz_lag(h)), 
  clay_dif = lag.clay_bot.1 - clay,
  texcl_contrast = paste0(texcl, "-", lag.texcl_bot.1)
)

}
\seealso{
\code{\link[=hz_dissolve]{hz_dissolve()}}, \code{\link[=hz_intersect]{hz_intersect()}}, \code{\link[=hz_segment]{hz_segment()}}
}
\author{
Stephen Roecker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotMultipleSPC.R
\name{plotMultipleSPC}
\alias{plotMultipleSPC}
\title{Plot Multiple \code{SoilProfileCollection} Objects}
\usage{
plotMultipleSPC(
  spc.list,
  group.labels,
  args = rep(list(NA), times = length(spc.list)),
  merged.legend = NULL,
  merged.colors = c("#5E4FA2", "#3288BD", "#66C2A5", "#ABDDA4", "#E6F598", "#FEE08B",
    "#FDAE61", "#F46D43", "#D53E4F", "#9E0142"),
  merged.legend.title = merged.legend,
  arrow.offset = 2,
  bracket.base.depth = 95,
  label.offset = 2,
  label.cex = 0.75,
  ...
)
}
\arguments{
\item{spc.list}{a list of \code{SoilProfileCollection} objects}

\item{group.labels}{a vector of group labels, one for each
\code{SoilProfileCollection} object}

\item{args}{a list of arguments passed to \code{plotSPC}, one for each
\code{SoilProfileCollection} object}

\item{merged.legend}{name of a horizon level attribute from which to create thematic sketches and merged legend}

\item{merged.colors}{vector of colors used to create thematic sketches from a shared horizon level attribute}

\item{merged.legend.title}{legend title}

\item{arrow.offset}{vertical offset in depth from base of start / end
profiles and group bracket arrows}

\item{bracket.base.depth}{baseline depth used for group brackets}

\item{label.offset}{vertical offset of group labels from baseline}

\item{label.cex}{label size}

\item{\dots}{additional arguments to the first call to \code{plotSPC}}
}
\description{
Plot Multiple \code{SoilProfileCollection} Objects
}
\details{
Combine multiple \code{SoilProfileCollection} objects into a single profile sketch,
with annotated groups.

See examples below for usage.
}
\note{
For thematic sketches, use the \code{merged.legend} argument instead of \code{color} argument to \code{plotSPC}
}
\examples{
\dontshow{if (requireNamespace("scales", quietly = TRUE)) withAutoprint(\{ # examplesIf}

##
## Simple Example
##

# using default arguments to plotSPC()

# load sample data
data(sp3)
data(sp4)

# promote to SoilProfileCollection
depths(sp3) <- id ~ top + bottom
depths(sp4) <- id ~ top + bottom

# combine into a list
spc.list <- list(sp3, sp4)

# argument list
arg.list <- list(
  list(name='name', id.style='top'),
  list(name='name', id.style='side')
)

# plot multiple SPC objects, 
# with list of named arguments for each call to plotSPC
par(mar=c(1,1,3,3))
plotMultipleSPC(
  spc.list, 
  group.labels = c('Collection 1', 'Collection 2'),
  args = arg.list, 
  bracket.base.depth = 120, label.cex = 1
)

# specify a different max.depth
plotMultipleSPC(
  spc.list, 
  group.labels = c('Collection 1', 'Collection 2'),
  args = arg.list, 
  bracket.base.depth = 120, label.cex = 1,
  max.depth = 250
)



##
## Merged Legend Example
##

# merged legend based on hz attribute 'clay'

# reset sample data
data(sp3)
data(sp4)

# promote to SoilProfileCollection
depths(sp3) <- id ~ top + bottom
depths(sp4) <- id ~ top + bottom

# combine into a list
spc.list <- list(sp3, sp4)

# argument list
arg.list <- list(
  list(name='name', id.style='top'),
  list(name='name', id.style='side')
)


par(mar=c(1,1,3,3))
plotMultipleSPC(
  spc.list, 
  group.labels = c('Collection 1', 'Collection 2'),
  args = arg.list, 
  label.cex = 1, 
  merged.legend = 'clay', merged.legend.title = 'Clay (\%)'
)


##
## Complex Merged Legend Example
##

# create a merged legend from "clay" in sp4 and jacobs2000
# use "soil_color" from sp3

# reset sample data
data(sp3)
data(sp4)
data(jacobs2000)

# promote to SoilProfileCollection
depths(sp3) <- id ~ top + bottom
depths(sp4) <- id ~ top + bottom

# remove 'clay' column from sp3
sp3$clay <- NULL

# combine into a list
spc.list <- list(sp3, sp4, jacobs2000)

# try some variations on the default arguments
# `clay` is missing in the first SPC, safe to specify another column for colors
arg.list <- list(
  list(color = 'soil_color', id.style='top', name = NA, width = 0.3, hz.depths = TRUE),
  list(name='name', id.style='side', name.style = 'center-center'),
  list(name='name', id.style='side', name.style = 'left-center', hz.depths = TRUE)
)

par(mar=c(1,1,3,3))
plotMultipleSPC(
  spc.list, 
  group.labels = c('sp3', 'sp4', 'jacobs2000'),
  label.offset = 3,
  args = arg.list, 
  merged.legend = 'clay', merged.legend.title = 'Clay (\%)',
  depth.axis = list(line = 0)
)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link{profileGroupLabels}}
}
\author{
D.E. Beaudette and Ben Marshall
}
\keyword{hplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R,
%   R/SoilProfileCollection-setters.R
\name{depths}
\alias{depths}
\alias{depths,SoilProfileCollection-method}
\alias{depths<-,SoilProfileCollection-method}
\alias{depths<-}
\alias{depths<-,data.frame-method}
\title{Initialize a SoilProfileCollection from data.frame}
\usage{
\S4method{depths}{SoilProfileCollection}(x, hzID = FALSE, ...)

\S4method{depths}{SoilProfileCollection}(object) <- value

\S4method{depths}{data.frame}(object) <- value
}
\arguments{
\item{x}{A SoilProfileCollection}

\item{hzID}{Include horizon ID? Usually this is calculated from the (sorted) row index unless \verb{hzidname()<-} has been called. Default: \code{FALSE}}

\item{...}{not used}

\item{object}{An object to promote to SoilProfileCollection (inherits from data.frame)}

\item{value}{A formula specifying the unique profile ID, top and bottom depth column names}
}
\value{
a \code{data.frame} containing profile ID, top depth, and bottom depth
}
\description{
\verb{depths(<data.frame>) <- <formula>}: Initialize SoilProfileCollection

\verb{depths(<SoilProfileCollection>)}: Extract profile ID and horizon depths from SoilProfileCollection
}
\details{
The input horizon data, and the resulting profile order, is sorted based on unique profile ID and top depth. ID columns are converted to character, depth columns are converted to integer. If \code{NA} values exist in all of the top depths, a prototype with 1 horizon per profile ID is returned, with \code{NA} in all non-essential columns. If the input \code{object} has 0 rows, a prototype with 0 horizons and 0 rows, but same column names as \code{object}, is returned.
}
\examples{
# load a SoilProfileCollection
data(jacobs2000, package = "aqp")

depths(jacobs2000)
## init SoilProfileCollection objects from data.frame of horizon data

# load demo data
data(sp1)

# promote to SPC
depths(sp1) <- id ~ top + bottom

# plot
plot(sp1)

# number of profiles
length(sp1)

# number of horizons
nrow(sp1)

}
\seealso{
\code{horizons()} \code{idname()} \code{hzidname()} \code{horizonDepths()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_distance_graph.R
\name{plot_distance_graph}
\alias{plot_distance_graph}
\title{Between Individual Distance Plot}
\usage{
plot_distance_graph(D, idx = 1:dim(as.matrix((D)))[1])
}
\arguments{
\item{D}{distance matrix, should be of class 'dist' or compatible class}

\item{idx}{an integer sequence defining which individuals should be compared}
}
\value{
No value is returned.
}
\description{
Plot pair-wise distances between individuals as line segments.
}
\details{
By default all individuals are plotting on the same axis. When there are
more than about 10 individuals, the plot can become quite messy. See
examples below for ideas.
}
\examples{

data(sp2)
depths(sp2) <- id ~ top + bottom
d <- NCSP(
  sp2,
  vars = c('prop', 'field_ph', 'hue', 'value'),
  maxDepth = 100,
  k = 0.01
)

par(mfcol=c(3,1), mar=c(2.5,4.5,1,1))
plot_distance_graph(d, idx=1:6)
plot_distance_graph(d, idx=7:12)
plot_distance_graph(d, idx=12:18)
}
\references{
https://casoilresource.lawr.ucdavis.edu/
}
\seealso{
\link{sp2}
}
\author{
Dylan E Beaudette
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalize.hz.R
\name{genhzTableToAdjMat}
\alias{genhzTableToAdjMat}
\title{Convert cross-tabulation to adjacency matrix.}
\usage{
genhzTableToAdjMat(tab)
}
\arguments{
\item{tab}{table, cross-tabulation of original and generalized horizon labels e.g. \code{table(original, genhz)}}
}
\value{
matrix of numeric values
}
\description{
Convert a cross-tabulation of e.g. original horizon designations vs. generalized horizon labels to adjacency matrix form.
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\name{subsetProfiles}
\alias{subsetProfiles}
\alias{subsetProfiles,SoilProfileCollection-method}
\title{DEPRECATED use subset}
\usage{
\S4method{subsetProfiles}{SoilProfileCollection}(object, s, h, ...)
}
\arguments{
\item{object}{object}

\item{s}{fully-quoted search criteria for matching
via site-level attributes}

\item{h}{fully-quoted search criteria for matching
via horizon-level attributes}

\item{...}{not used}
}
\value{
A \code{SoilProfileCollection} class object.
}
\description{
This function is used to subset \code{SoilProfileCollection} objects using
either site-level or horizon-level attributes, or both.
}
\details{
The \code{s} argument supplies a fully-quoted search criteria for matching
via site or horizon-level attributes. The \code{h} argument supplies a fully-quoted
search criteria for matching via horizon-level attributes. All horizons
associated with a single horizon-level match (i.e. out of several, only a
single horizon matches the search criteria) are returned. See examples for
usage.
}
\examples{

# more interesting sample data
data(sp2)
depths(sp2) <- id ~ top + bottom
site(sp2) <- ~ surface

# subset by integer index, note that this does not re-order the profiles
plot(sp2[1:5, ])

# generate an integer index via pattern-matching
idx <- grep('modesto', sp2$surface, ignore.case=TRUE)
plot(sp2[idx, ])

# generate in index via profileApply:
# subset those profiles where: min(ph) < 5.6
idx <- which(profileApply(sp2, function(i) min(i$field_ph, na.rm=TRUE) < 5.6))
plot(sp2[idx, ])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{sp5}
\alias{sp5}
\title{Sample Soil Database #5}
\format{
\code{SoilProfileCollection} object
}
\source{
296 Soil Profiles from the La Rochelle region of France (F. Carre
and Girard, 2002). These data can be found on the OSACA project page
(\url{http://eusoils.jrc.ec.europa.eu/projects/OSACA/}).
}
\description{
296 Soil Profiles from the La Rochelle region of France (F. Carre and
Girard, 2002)
}
\details{
These data are c/o F. Carre (Florence.CARRE@ineris.fr).
}
\examples{

\dontrun{
library(scales)
data(sp5)
par(mar=c(1,1,1,1))
# plot a random sampling of profiles
s <- sample(1:length(sp5), size=25)
plot(sp5[s, ], divide.hz=FALSE)

# plot the first 100 profiles, as 4 rows of 25, hard-coding the max depth
layout(matrix(c(1,2,3,4), ncol=1), height=c(0.25,0.25,0.25,0.25))
plot(sp5[1:25, ], max.depth=300)
plot(sp5[26:50, ], max.depth=300)
plot(sp5[51:75, ], max.depth=300)
plot(sp5[76:100, ], max.depth=300)


# 4x1 matrix of plotting areas
layout(matrix(c(1,2,3,4), ncol=1), height=c(0.25,0.25,0.25,0.25))

# plot profiles, with points added to the mid-points of randomly selected horizons
sub <- sp5[1:25, ]
plot(sub, max.depth=300) ; mtext('Set 1', 2, line=-0.5, font=2)
y.p <- profileApply(sub, function(x) {
  s <- sample(1:nrow(x), 1)
  h <- horizons(x); with(h[s,], (top+bottom)/2)
  })
points(1:25, y.p, bg='white', pch=21)

# plot profiles, with arrows pointing to profile bottoms
sub <- sp5[26:50, ]
plot(sub, max.depth=300); mtext('Set 2', 2, line=-0.5, font=2)
y.a <- profileApply(sub, function(x) max(x))
arrows(1:25, y.a-50, 1:25, y.a, len=0.1, col='white')

# plot profiles, with points connected by lines: ideally reflecting some kind of measured data
sub <- sp5[51:75, ]
plot(sub, max.depth=300); mtext('Set 3', 2, line=-0.5, font=2)
y.p <- 20*(sin(1:25) + 2*cos(1:25) + 5)
points(1:25, y.p, bg='white', pch=21)
lines(1:25, y.p, lty=2)

# plot profiles, with polygons connecting horizons with max clay content (+/-) 10 cm
sub <- sp5[76:100, ]
y.clay.max <- profileApply(sub, function(x) {
  i <- which.max(x$clay)
  h <- horizons(x)
  with(h[i, ], (top+bottom)/2)
  } )

plot(sub, max.depth=300); mtext('Set 4', 2, line=-0.5, font=2)
polygon(c(1:25, 25:1), c(y.clay.max-10, rev(y.clay.max+10)),
border='black', col=rgb(0,0,0.8, alpha=0.25))
points(1:25, y.clay.max, pch=21, bg='white')

# close plot
dev.off()


# plotting parameters
yo <- 100 # y-offset
sf <- 0.65 # scaling factor
# plot profile sketches
par(mar=c(0,0,0,0))
plot(sp5[1:25, ], max.depth=300, y.offset=yo, scaling.factor=sf)
# optionally add describe plotting area above profiles with lines
# abline(h=c(0,90,100, (300*sf)+yo), lty=2)
# simulate an environmental variable associated with profiles (elevation, etc.)
r <- vector(mode='numeric', length=25)
r[1] <- -50 ; for(i in 2:25) {r[i] <- r[i-1] + rnorm(mean=-1, sd=25, n=1)}
# rescale
r <- rescale(r, to=c(80, 0))
# illustrate gradient with points/lines/arrows
lines(1:25, r)
points(1:25, r, pch=16)
arrows(1:25, r, 1:25, 95, len=0.1)
# add scale for simulated gradient
axis(2, at=pretty(0:80), labels=rev(pretty(0:80)), line=-1, cex.axis=0.75, las=2)
# depict a secondary environmental gradient with polygons (water table depth, etc.)
polygon(c(1:25, 25:1), c((100-r)+150, rep((300*sf)+yo, times=25)),
border='black', col=rgb(0,0,0.8, alpha=0.25))

}


}
\references{
F. Carre, M.C. Girard. 2002. Quantitative mapping of soil types
based on regression kriging of taxonomic distances with landform and land
cover attributes. Geoderma. 110: 241--263.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/soilColorIndices.R
\name{harden.melanization}
\alias{harden.melanization}
\title{Harden (1982) Melanization}
\usage{
harden.melanization(value, value_ref)
}
\arguments{
\item{value}{numeric vector containing Munsell values}

\item{value_ref}{A numeric vector containing Munsell value(s) for reference material}
}
\value{
A numeric vector reflecting horizon darkening relative to a reference (e.g. parent) material.
}
\description{
Calculate "melanization" component of "Profile Development Index" after Harden (1982) "A quantitative index of soil development from field descriptions: Examples from a chronosequence in central California". Accepts vectorized inputs for value and reference value to produce vector output. A convenient use case would be to apply this on a profile-specific basis, where the \code{value_ref} has a single value, and \code{value} is a vector of length equal to the number of horizons within the upper 100 cm.
}
\details{
In Harden (1982), "melanization" is calculated relative to a reference parent material for all horizons within 100cm of the soil surface. In addition, several other non-color components are normalized relative to a maximum value and summed to obtain the overall Profile Development Index.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

library(aqp)
data("jacobs2000", package="aqp")

# LEFT JOIN hue, value, chroma matrix color columns
horizons(jacobs2000) <- cbind(horizons(jacobs2000)[,c(idname(jacobs2000), hzidname(jacobs2000))],
                              parseMunsell(jacobs2000$matrix_color_munsell, convertColors = FALSE))

# calculate a mixed 150-200cm color ~"parent material"

jacobs2000$c_horizon_color <- profileApply(jacobs2000, function(p) {

  # and derive the parent material from the 150-200cm interval
  p150_200 <- glom(p, 150, 200, truncate = TRUE)
  p150_200$thickness <- p150_200$bottom - p150_200$top

  # mix colors
  clrs <- na.omit(horizons(p150_200)[,c('matrix_color_munsell','thickness')])
  mixMunsell(clrs$matrix_color_munsell, w = clrs$thickness)$munsell

})

# segment profile into 1cm slices (for proper depth weighting)
jacobs2000$melan <- profileApply(jacobs2000, function(p) {

  # sum the melanization index over the 0-100cm interval
  p0_100 <- hz_segment(p, 0:100)

  ccol <- parseMunsell(p$c_horizon_color, convertColors = FALSE)

  sum(harden.melanization(
    value = as.numeric(p0_100$value),
    value_ref = as.numeric(ccol$value)), na.rm = TRUE)

})

jacobs2000$melanorder <- order(jacobs2000$melan)

# Plot in order of increasing Melanization index

plotSPC(jacobs2000, 
        color = "matrix_color",
        label = "melan",
        plot.order = jacobs2000$melanorder,
        max.depth = 250
        )

segments(
  x0 = 0.5, 
  x1 = length(jacobs2000) + 0.5, 
  y0 = c(0,100,150,200), 
  y1 = c(0,100,150,200), 
  lty = 2
)

# Add [estimated] parent material color swatches
lapply(seq_along(jacobs2000$c_horizon_color), function(i) {
  rect(i - 0.15, 250, i + 0.15, 225,
       col = parseMunsell(jacobs2000$c_horizon_color[jacobs2000$melanorder[i]]))
})
}
\references{
Harden, J.W. (1982) A quantitative index of soil development from field descriptions: Examples from a chronosequence in central California. Geoderma. 28(1) 1-28. doi:  10.1016/0016-7061(82)90037-4
}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitLogicErrors.R
\name{splitLogicErrors}
\alias{splitLogicErrors}
\title{Split a \code{SoilProfileCollection} into a list based on types of horizon logic errors}
\usage{
splitLogicErrors(object, interact = FALSE, ...)
}
\arguments{
\item{object}{A \code{SoilProfileCollection}}

\item{interact}{Calculate interaction between the four logic errors for groups? Default: \code{FALSE} always returns 4 groups, one for each logic error type.}

\item{...}{Additional arguments to \code{split.default}, called when \code{interact = TRUE}}
}
\value{
A named list of \code{SoilProfileCollections} (or \code{NULL}), with names: "depthLogic", "sameDepth", "missingDepth", "overlapOrGap". If \code{interact = TRUE} then the list elements groups determined by \code{interaction()} of the error types.
}
\description{
Uses \code{checkHzDepthLogic()} to identify presence of depth logic errors, same depths, missing depths, and overlaps/gaps between the horizons of each profile in a \code{SoilProfileCollection.}
}
\examples{

data(sp4)
depths(sp4) <- id ~ top + bottom

# no errors (all four list elements return NULL)
splitLogicErrors(sp4)

# NA in top depth triggers depth logic and missing depth errors
data(sp4)
sp4$top[1] <- NA
depths(sp4) <- id ~ top + bottom

splitLogicErrors(sp4)

# interact = TRUE gets errors for profile 1 in same group
#  and allows you to pass extra arguments to split.default()
splitLogicErrors(sp4, interact = TRUE, sep = "_", drop = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{sp2}
\alias{sp2}
\title{Honcut Creek Soil Profile Data}
\format{
A data frame with 154 observations on the following 21 variables.
\describe{ \item{id}{profile id} \item{surface}{dated
surface} \item{top}{horizon top in cm} \item{bottom}{horizon
bottom in cm} \item{bound_distinct}{horizon lower boundary
distinctness class} \item{bound_topography}{horizon lower boundary
topography class} \item{name}{horizon name}
\item{texture}{USDA soil texture class}
\item{prop}{field-estimated clay content}
\item{structure_grade}{soil structure grade}
\item{structure_size}{soil structure size}
\item{structure_type}{soil structure type}
\item{stickiness}{stickiness} \item{plasticity}{plasticity}
\item{field_ph}{field-measured pH} \item{hue}{Munsell hue}
\item{value}{Munsell value} \item{chroma}{Munsell chroma}
\item{r}{RGB red component} \item{g}{RGB green component}
\item{b}{RGB blue component} \item{soil_color}{R-friendly
encoding of soil color} }
}
\source{
Busacca, Alan J.; Singer, Michael J.; Verosub, Kenneth L. 1989. Late
Cenozoic stratigraphy of the Feather and Yuba rivers area, California, with
a section on soil development in mixed alluvium at Honcut Creek. USGS
Bulletin 1590-G.
}
\description{
A collection of 18 soil profiles, consisting of select soil morphologic
attributes, associated with a stratigraphic study conducted near Honcut
Creek, California.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

data(sp2)

# convert into SoilProfileCollection object
depths(sp2) <- id ~ top + bottom

# transfer site-level data
site(sp2) <- ~ surface

# generate a new plotting order, based on the dated surface each soil was described on
p.order <- order(sp2$surface)

# plot
par(mar=c(1,0,3,0))
plot(sp2, plot.order=p.order)

# setup multi-figure output
par(mfrow=c(2,1), mar=c(0,0,1,0))

# truncate plot to 200 cm depth
plot(sp2, plot.order=p.order, max.depth=200)
abline(h=200, lty=2, lwd=2)

# compute numerical distances between profiles
# based on select horizon-level properties, to a depth of 200 cm
d <- NCSP(sp2, vars=c('prop','field_ph','hue'), maxDepth = 100, k = 0)

# plot dendrogram with ape package:
if(require(ape) & require(cluster)) {
h <- diana(d)
p <- as.phylo(as.hclust(h))
plot(p, cex=0.75, label.offset=0.01, font=1, direct='down', srt=90, adj=0.5, y.lim=c(-0.125, 0.5))

# add in the dated surface type via color
tiplabels(col=as.numeric(sp2$surface), pch=15)

# based on distance matrix values, YMMV
legend('topleft', legend=levels(sp2$surface), col=1:6, pch=15, bty='n', bg='white', cex=0.75)
}


}
\references{
\url{https://casoilresource.lawr.ucdavis.edu/}
}
\author{
Dylan E. Beaudette
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slab.R
\docType{methods}
\name{slab}
\alias{slab}
\alias{slab,SoilProfileCollection-method}
\alias{slab_function}
\title{Slab-Wise Aggregation of SoilProfileCollection Objects}
\usage{
\S4method{slab}{SoilProfileCollection}(
  object,
  fm,
  slab.structure = 1,
  strict = FALSE,
  byhz = TRUE,
  slab.fun = slab_function(method = "numeric"),
  cpm = 1,
  weights = NULL,
  ...
)

slab_function(
  method = c("numeric", "factor", "hd", "weighted.numeric", "weighted.factor", "fast")
)
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{fm}{A formula: either \code{groups ~ var1 + var2 + var3} where named
variables are aggregated within \verb{groups' OR where named variables are aggregated across the entire collection } ~ var1 + var2 + var3\verb{. If }groups\verb{is a factor it must not contain}NA`}

\item{slab.structure}{integer vector: user-defined slab structure. See Details.}

\item{strict}{logical: should horizons be strictly checked for self-consistency?}

\item{byhz}{logical: should horizons or whole profiles be removed by logic checks in \code{strict}? Default \code{TRUE} removes only offending horizons, \code{FALSE} removes whole profiles with one or more illogical horizons.}

\item{slab.fun}{Function used to process each 'slab' of data, ideally returning a vector with names attribute. Defaults to a wrapper function around \code{\link[stats:quantile]{stats::quantile()}}. See details.}

\item{cpm}{Strategy for normalizing slice-wise probabilities, dividing by either:
\itemize{
\item number of profiles with data at the current slice (\code{cpm = 1}), or
\item by the number of profiles in the collection (\code{cpm = 2}).
}

Mode 1 values will always sum to the contributing fraction, while mode 2 values will always sum to 1.}

\item{weights}{Column name containing site-level weights}

\item{\dots}{further arguments passed to \code{slab.fun}}

\item{method}{one of \code{"numeric"}, \code{"factor"}, \code{"hd"}, \code{"weighted.numeric"}, \code{"weighted.factor"}, \code{"fast"}}
}
\value{
Output is returned in long format, such that slice-wise aggregates are returned once for each combination of grouping level (optional), variable described in the \code{fm} argument, and depth-wise 'slab'.

Aggregation of numeric variables, using the default slab function:
\itemize{
\item variable: The names of variables included in \code{fm}
\item groupname: The name of the grouping variable when provided, otherwise a fake grouping variable named 'all.profiles'.
\item p.q5: The slice-wise 5th percentile.
\item p.q25: The slice-wise 25th percentile
\item p.q50: The slice-wise 50th percentile (median)
\item p.q75: The slice-wise 75th percentile
\item p.q95: The slice-wise 95th percentile
\item top: The slab top boundary
\item bottom The slab bottom boundary.
\item contributing_fraction: The fraction of profiles contributing to the aggregate value, ranges from 1/n_profiles to 1
}

When a single factor variable is used, slice-wise probabilities for each
level of that factor are returned as:
\itemize{
\item variable: The names of factor variable included in \code{fm}
\item groupname: The name of the grouping variable when provided, otherwise a fake grouping variable named 'all.profiles'
\item A: The slice-wise probability of level A
\item B: The slice-wise probability of level B
\item n: The slice-wise probability of level n
\item top: The slab top boundary.
\item bottom: The slab bottom boundary
\item contributing_fraction: The fraction of profiles contributing to the aggregate value, ranges from 1/n_profiles to 1
}

\code{slab_function()}: return an aggregation function based on the \code{method} argument
}
\description{
Aggregate soil properties along user-defined \code{slabs}, and optionally within
groups.

Multiple continuous variables OR a single categorical (factor) variable can
be aggregated within a call to \code{slab()}. Basic error checking is
performed to make sure that top and bottom horizon boundaries make sense.
User-defined aggregate functions (\code{slab.fun}) should return a named
vector of results. A new, named column will appear in the results of
\code{slab()} for every named element of a vector returned by \code{slab.fun}.
See examples below for a simple example of a slab function that computes
mean, mean +/- 1 standard deviation. The default slab function wraps
\code{\link[stats:quantile]{stats::quantile()}}. Note that if \code{group} is a factor it must not contain NAs.

\code{slab()} uses \code{dice()} to "resample" profiles to 1cm slices from depth 0 to \code{max(x)} (or \code{slab.structure[2]}, if defined).

Sometimes \code{slab()} is used to conveniently re-arrange data vs. aggregate.
This is performed by specifying \code{identity} in \code{slab.fun}. See
examples below for a demonstration of this functionality.

The default \code{slab.fun} was changed 2019-10-30 from a wrapper around
\code{\link[Hmisc:hdquantile]{Hmisc::hdquantile()}} to a wrapper around \code{\link[stats:quantile]{stats::quantile()}}. See
examples below for a simple way to switch to the HD quantile estimator.

Execution time scales linearly (slower) with the total number of profiles in
\code{object}, and exponentially (faster) as the number of profiles / group
is increased. \code{slab()} and \code{\link[=dice]{dice()}} are much faster and require less
memory if input data are either numeric or character.

There are several possible ways to define slabs, using \code{slab.structure}:
\itemize{
\item a single integer, e.g. 10: data are aggregated over a regular sequence of 10-unit thickness slabs
\item a vector of 2 integers, e.g. c(50, 60): data are aggregated over depths spanning 50-60 depth units
\item a vector of 3 or more integers, e.g. c(0, 5, 10, 50, 100): data are aggregated over the depths spanning 0-5, 5-10, 10-50, 50-100 depth units
}
}
\details{
\code{slab_function()}: The default \code{"numeric"} aggregation method is the \code{"fast"} numeric (quantile) method. Additional methods include \code{"factor"} for categorical data, \code{"hd"} to use the Harrell-Davis Distribution-Free Quantile Estimator from the Hmisc package, and "\code{weighted}" to use a weighted quantile method from the Hmisc package
}
\note{
Arguments to \code{slab()} have changed with aqp 1.5 (2012-12-29) as part of a code clean-up and optimization. Calculation of
weighted-summaries was broken in \verb{aqp 1.2-6 (2012-06-26), and removed as of aqp 1.5 (2012-12-29). }slab()`` replaced the previously defined \code{soil.slot.multiple} function as of aqp 0.98-8.58 (2011-12-21).
}
\section{Methods}{
 \describe{ \item{data = "SoilProfileCollection"}{Typical
usage, where input is a \code{\link{SoilProfileCollection}}.} }
}

\examples{

##
## basic examples
##
library(lattice)
library(grid)
library(data.table)

# load sample data, upgrade to SoilProfileCollection
data(sp1)
depths(sp1) <- id ~ top + bottom

hzdesgnname(sp1) <- "name"

# aggregate entire collection with two different segment sizes
a <- slab(sp1, fm = ~ prop)
b <- slab(sp1, fm = ~ prop, slab.structure=5)

# check output
str(a)

# stack into long format
ab <- make.groups(a, b)
ab$which <- factor(ab$which, levels=c('a','b'),
labels=c('1-cm Interval', '5-cm Interval'))

# plot median and IQR
# custom plotting function for uncertainty viz.
xyplot(top ~ p.q50 | which, data=ab, ylab='Depth',
			 xlab='median bounded by 25th and 75th percentiles',
			 lower=ab$p.q25, upper=ab$p.q75, ylim=c(250,-5),
			 panel=panel.depth_function,
			 prepanel=prepanel.depth_function,
			 cf=ab$contributing_fraction,
			 alpha=0.5,
			 layout=c(2,1), scales=list(x=list(alternating=1))
			 )


###
### re-arrange data / no aggregation
###

# load sample data, upgrade to SoilProfileCollection
data(sp1)
depths(sp1) <- id ~ top + bottom

# arrange data by ID
a <- slab(sp1, fm = id ~ prop, slab.fun=identity)

# convert id to a factor for plotting
a$id <- factor(a$id)

# check output
str(a)

# plot via step function
xyplot(top ~ value | id, data=a, ylab='Depth',
       ylim=c(250, -5), as.table=TRUE,
       panel=panel.depth_function,
       prepanel=prepanel.depth_function,
       scales=list(x=list(alternating=1))
)

##
## categorical variable example
##

data(sp1)
depths(sp1) <- id ~ top + bottom

# normalize horizon names: result is a factor
sp1$name <- generalize.hz(
  sp1$name,
  new = c('O','A','B','C'),
  pat = c('O', '^A','^B','C')
  )

# compute slice-wise probability so that it sums to contributing fraction, from 0-150
a <- slab(sp1, fm= ~ name, cpm=1, slab.structure=0:150)

# convert wide -> long for plotting
# result is a data.table
# genhz factor levels are set by order in `measure.vars`
a.long <- data.table::melt(
  data.table::as.data.table(a),
  id.vars = c('top','bottom'),
  measure.vars = c('O', 'A', 'B', 'C'),
  )


# plot horizon type proportions using panels
xyplot(top ~ value | variable,
       data = a.long, subset=value > 0,
       col = 1, lwd = 2,
       xlab = 'Class Probability',
       ylab = 'Depth (cm)',
       strip = strip.custom(bg = grey(0.85)),
       scales = list(x = list(alternating = FALSE)),
       ylim = c(150, -5), type=c('S','g'),
       horizontal = TRUE, layout = c(4,1)
       )

# again, this time using groups
xyplot(top ~ value,
       data = a.long,
       groups = variable,
       subset = value > 0,
       ylab = 'Depth (cm)',
       ylim = c(150, -5),
       type = c('S','g'),
       horizontal = TRUE,
       asp = 2,
       lwd = 2,
       auto.key = list(
         lines = TRUE,
         points = FALSE,
         cex = 0.8,
         columns = 1,
         space = 'right'
       )
)

# adjust probability to size of collection, from 0-150
a.1 <- slab(sp1, fm= ~ name, cpm = 2, slab.structure = 0:150)

# convert wide -> long for plotting
# result is a data.table
# genhz factor levels are set by order in `measure.vars`
a.1.long <- data.table::melt(
  data.table::as.data.table(a.1),
  id.vars = c('top','bottom'),
  measure.vars = c('O','A','B','C')
)

# combine aggregation from `cpm` modes 1 and 2
g <- make.groups(cmp.mode.1 = a.long, cmp.mode.2 = a.1.long)

# plot horizon type proportions
xyplot(top ~ value | variable,
       groups = which,
       data = g, subset = value > 0,
       ylim = c(240, -5),
       type = c('S','g'),
       horizontal = TRUE,
       layout = c(4,1),
       auto.key = list(lines = TRUE, points = FALSE, columns = 2),
       par.settings = list(superpose.line = list(col = c(1, 2), lwd = 2)),
       scales = list(alternating = 3),
       xlab = 'Class Probability',
       ylab = 'Depth (cm)',
       strip = strip.custom(bg = grey(0.85))
)


# apply slice-wise evaluation of max probability, and assign ML-horizon at each slice
gen.hz.ml <- get.ml.hz(a, c('O','A','B','C'))


\dontrun{
##
## HD quantile estimator
##

library(soilDB)
library(lattice)
library(data.table)

# sample data
data('loafercreek', package = 'soilDB')

# defaul slab.fun wraps stats::quantile()
a <- slab(loafercreek, fm = ~ total_frags_pct + clay)

# use HD quantile estimator from Hmisc package instead
a.HD <- slab(loafercreek, fm = ~ total_frags_pct + clay, slab.fun = aqp:::.slab.fun.numeric.HD)

# combine
g <- make.groups(standard=a, HD=a.HD)

# note differences
densityplot(~ p.q50 | variable, data=g, groups=which,
            scales=list(relation='free', alternating=3, tick.number=10, y=list(rot=0)),
            xlab='50th Percentile', pch=NA, main='Loafercreek',
            auto.key=list(columns=2, points=FALSE, lines=TRUE),
            par.settings=list(superpose.line=list(lwd=2, col=c('RoyalBlue', 'Orange2')))
)

# differences are slight but important
xyplot(
  top ~ p.q50 | variable, data=g, groups=which,
  xlab='Value', ylab='Depth (cm)',
  asp=1.5, main='Loafercreek',
  lower=g$p.q25, upper=g$p.q75,
  sync.colors=TRUE, alpha=0.25, cf=g$contributing_fraction,
  ylim=c(115,-5), layout=c(2,1), scales=list(x=list(relation='free')),
  par.settings=list(superpose.line=list(lwd=2, col=c('RoyalBlue', 'Orange2'))),
  strip=strip.custom(bg=grey(0.85)),
  panel=panel.depth_function,
  prepanel=prepanel.depth_function,
  auto.key=list(columns=2, lines=TRUE, points=FALSE)
)

##
## multivariate examples
##
data(sp3)

# add new grouping factor
sp3$group <- 'group 1'
sp3$group[as.numeric(sp3$id) > 5] <- 'group 2'
sp3$group <- factor(sp3$group)

# upgrade to SPC
depths(sp3) <- id ~ top + bottom
site(sp3) <- ~ group

# custom 'slab' function, returning mean +/- 1SD
mean.and.sd <- function(values) {
  m <- mean(values, na.rm=TRUE)
  s <- sd(values, na.rm=TRUE)
  upper <- m + s
  lower <- m - s
  res <- c(mean=m, lower=lower, upper=upper)
  return(res)
}

# aggregate several variables at once, within 'group'
a <- slab(sp3, fm = group ~ L + A + B, slab.fun = mean.and.sd)

# check the results:
# note that 'group' is the column containing group labels
xyplot(
  top ~ mean | variable, data=a, groups=group,
  lower=a$lower, upper=a$upper,
  sync.colors=TRUE, alpha=0.5,
  cf = a$contributing_fraction,
  xlab = 'Mean Bounded by +/- 1SD',
  ylab = 'Depth (cm)',
  ylim=c(125,-5), layout=c(3,1),
  scales=list(x=list(relation='free')),
  par.settings = list(superpose.line=list(lwd=2, col=c('RoyalBlue', 'Orange2'))),
  panel = panel.depth_function,
  prepanel = prepanel.depth_function,
  strip = strip.custom(bg=grey(0.85)),
  auto.key = list(columns=2, lines=TRUE, points=FALSE)
)


# compare a single profile to the group-level aggregate values
a.1 <- slab(sp3[1, ], fm = group ~ L + A + B, slab.fun = mean.and.sd)

# manually update the group column
a.1$group <- 'profile 1'

# combine into a single data.frame:
g <- rbind(a, a.1)

# plot with customized line styles
xyplot(
  top ~ mean | variable, data=g, groups=group, subscripts=TRUE,
  lower=a$lower, upper=a$upper, ylim=c(125,-5),
  layout=c(3,1), scales=list(x=list(relation='free')),
  xlab = 'Mean Bounded by +/- 1SD',
  ylab = 'Depth (cm)',
  panel=panel.depth_function,
  prepanel=prepanel.depth_function,
  sync.colors = TRUE, alpha = 0.25,
  par.settings = list(
    superpose.line = list(
      col = c('orange', 'royalblue', 'black'),
      lwd = 2, lty = c(1,1,2)
    )
  ),
  strip = strip.custom(bg=grey(0.85)),
  auto.key = list(columns=3, lines=TRUE, points=FALSE)
)




## again, this time for a user-defined slab from 40-60 cm
a <- slab(sp3,
          fm = group ~ L + A + B,
          slab.structure = c(40,60),
          slab.fun = mean.and.sd
)

# now we have weighted average properties (within the defined slab)
# for each variable, and each group
# convert long -> wide
data.table::dcast(
  data.table::as.data.table(a),
  formula = group + top + bottom ~ variable,
  value.var = 'mean'
)

## this time, compute the weighted mean of selected properties, by profile ID
a <- slab(sp3,
          fm = id ~ L + A + B,
          slab.structure = c(40,60),
          slab.fun = mean.and.sd
)

# convert long -> wide
data.table::dcast(
  data.table::as.data.table(a),
  formula = id + top + bottom ~ variable,
  value.var = 'mean'
)


## aggregate the entire collection, using default slab function
## note the missing left-hand side of the formula
a <- slab(sp3, fm = ~ L + A + B)

}

}
\references{
D.E. Beaudette, P. Roudier, A.T. O'Geen, Algorithms for
quantitative pedology: A toolkit for soil scientists, Computers &
Geosciences, Volume 52, March 2013, Pages 258-268,
10.1016/j.cageo.2012.10.020.

Harrell FE, Davis CE (1982): A new distribution-free quantile estimator.
Biometrika 69:635-640.
}
\seealso{
\code{\link[=dice]{dice()}}
}
\author{
D.E. Beaudette
}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getClosestMunsellChip.R
\name{getClosestMunsellChip}
\alias{getClosestMunsellChip}
\title{Get Approximate Munsell Chip}
\usage{
getClosestMunsellChip(munsellColor, convertColors = TRUE, ...)
}
\arguments{
\item{munsellColor}{character vector of strings containing Munsell notation of color, e.g. '10YR 4/3', not NA-safe}

\item{convertColors}{logical, should parsed Munsell colors be converted into sRGB values}

\item{...}{further arguments to \code{munsell2rgb}}
}
\value{
a \code{data.frame} when \code{convertColors=TRUE}, otherwise character vector
}
\description{
Non-standard Munsell notation ('7.9YR 2.7/2.0') can be matched (nearest-neighbor, no interpolation) to the closest color within the \code{munsell} sRGB/CIELAB look-up table via \code{getClosestMunsellChip()}. A more accurate estimate of sRGB values from non-standard notation can be achieved with the \href{https://CRAN.R-project.org/package=munsellinterpol}{munsellinterpol} package. For example, conversion from Munsell to CIELAB, assuming a D65 illuminant via: \code{MunsellToLab('0.1Y 3.3/4.4', white='D65', adapt='Bradford')}.
}
\examples{

# convert a non-standard color to closest "chip" in `munsell` look-up table
getClosestMunsellChip('7.9YR 2.7/2.0', convertColors = FALSE)

# convert directly to R color
getClosestMunsellChip('7.9YR 2.7/2.0')

# special case for 2.5 value -> no rounding, we have these records in the conversion LUT
getClosestMunsellChip('7.5YR 2.5/2', convertColors = FALSE)


getClosestMunsellChip('7.5YR 6.8/4.4', convertColors = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/factor-level-setters.R
\name{SoilTextureLevels}
\alias{SoilTextureLevels}
\title{Ranking Systems for USDA Soil Texture Classes}
\usage{
SoilTextureLevels(which = "codes", simplify = FALSE)
}
\arguments{
\item{which}{'codes' (texture codes) or 'names' (texture class names)}

\item{simplify}{Return 12-class factor levels (\code{TRUE}) or 21-class factor levels (default: \code{FALSE})? The 12-class system does not separate sands, loamy sands and sandy loams into sand fraction variants (e.g. "very fine sandy loam" in the 21-class system is "sandy loam" in 12-class system)}
}
\value{
an ordered factor
}
\description{
Generate a vector of USDA soil texture codes or class names, sorted according to approximate particle size
}
\examples{

# class codes
SoilTextureLevels()

# class names
SoilTextureLevels(which = 'names')

# simpler class names
SoilTextureLevels(which = 'names', simplify = TRUE)
 
}
\references{
\href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorChart.R
\name{colorChart}
\alias{colorChart}
\title{Visualize soil colors in Munsell notation according to within-group frequency.}
\usage{
colorChart(
  m,
  g = factor("All"),
  size = TRUE,
  annotate = FALSE,
  launderColors = TRUE,
  chip.cex = 3,
  chip.cex.min = 0.1,
  chip.cex.max = 1.5,
  chip.border.col = "black",
  annotate.cex = chip.cex * 0.25,
  annotate.type = c("count", "percentage"),
  threshold = NULL
)
}
\arguments{
\item{m}{character vector of color in Munsell notation ('10YR 4/6')}

\item{g}{factor describing group membership, typically a generalization of horizon designation, default value will generate a fake grouping that covers all of the colors in \code{m}}

\item{size}{logical, encode group-wise frequency with chip size}

\item{annotate}{logical, annotate color chip frequency}

\item{launderColors}{logical, fix common errors with encoding neutral colors, see \code{\link[=launderMunsell]{launderMunsell()}}}

\item{chip.cex}{scaling factor applied to each color chip}

\item{chip.cex.min}{lower limit for color chip frequency depiction}

\item{chip.cex.max}{lower limit for color chip frequency depiction}

\item{chip.border.col}{color for chip borders (outline)}

\item{annotate.cex}{scaling factor for chip frequency annotation}

\item{annotate.type}{character, within-group \code{count} or \code{percentage}}

\item{threshold}{numeric within 0-1, color chips with proportion \verb{< threshold} are removed}
}
\value{
a \code{trellis} object
}
\description{
Visualize soil colors in Munsell notation according to within-group frequency.
}
\examples{
\dontshow{if (requireNamespace('latticeExtra') && requireNamespace('scales')) withAutoprint(\{ # examplesIf}

library(lattice)

# two hue pages
ric <- expand.grid(
  hue = c('5YR', '7.5YR'),
  value = 2:8,
  chroma = 2:8
)

# combine hue, value, chroma into standard Munsell notation
ric <- sprintf("\%s \%s/\%s", ric$hue, ric$value, ric$chroma)

# note that chip frequency-based size is disabled
# because all chips have equal frequency
colorChart(ric, chip.cex = 4, size = TRUE)

# annotation of frequency
colorChart(ric, chip.cex = 4, annotate = TRUE)

# bootstrap to larger size
ric.big <- sample(ric, size = 100, replace = TRUE)

# frequency can be encoded in size
colorChart(ric.big, chip.cex = 3)
colorChart(ric.big, chip.cex = 5, annotate = TRUE)

# constant size
colorChart(ric.big, chip.cex = 3, size = FALSE)
colorChart(ric.big, chip.cex = 3, size = FALSE, chip.border.col = 'NA')

# simulate colors based dE00 thresholding
p <- list(
  list(m = '10YR 4/4', thresh = 10, hues = c('10YR', '7.5YR'))
)

# perform 500 simulations
s <- simulateColor(method = 'dE00', n = 500, parameters = p)

# result is a list, use the first element
colorChart(s[[1]], chip.cex = 4)

# increase the possible range of color chip sizes
colorChart(s[[1]], chip.cex = 4, chip.cex.min = 0.01, chip.cex.max = 2)

# demonstrate support for neutral chips
N <- sprintf('N \%s/', 2:8)
cols <- c(rep(N, times = 5), ric.big)

# note special panel used to show neutral chips
colorChart(cols, size = FALSE, annotate = TRUE)

# filter proportions below given threshold
colorChart(cols, size = FALSE, annotate = TRUE, threshold = 0.01,
           chip.cex = 4, annotate.type = 'percentage')
           
# extreme values
colorChart(c('N 8/', '10YR 4/12', '10YR 3/1', '10YR 10/4'))

# unusual notation, 10YR 4/0 --> effectively N 4/
colorChart(c('N 8/', 'N 4/', '10YR 4/0', '5G 6/6', '5Y 10/4'))

# demonstrate groups
m <- c('10YR 4/4', 'N 2/', 'N 3/', '5Y 6/6', '5P 4/4')
g <- factor(c(1, 1, 2, 2, 2))
colorChart(m, g)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkHzDepthLogic.R
\name{hzDepthTests}
\alias{hzDepthTests}
\title{Tests of horizon depth logic}
\usage{
hzDepthTests(top, bottom = NULL)
}
\arguments{
\item{top}{A numeric vector containing horizon top depths. Or a \code{data.frame} with two columns (first containing top depths, second containing bottom)}

\item{bottom}{A numeric vector containing horizon bottom depths.}
}
\value{
A named logical vector containing TRUE for each type of horizon logic error found in the given data.
}
\description{
Function used internally by \code{checkHzDepthLogic()}, \code{glom()} and various other functions that operate on horizon data from single soil profiles and require a priori depth logic checks. Checks for bottom depths less than top depth / bad top depth order ("depthLogic"), bottom depths equal to top depth ("sameDepth"), overlaps/gaps ("overlapOrGap") and missing depths ("missingDepth"). Use \code{names(res)[res]} on result \code{res} of \code{hzDepthTest()} to to determine type of logic error(s) found -- see examples below.
}
\examples{

# no logic errors
res <- hzDepthTests(top = c(0,10,20,30), bottom = c(10,20,30,50))
names(res)[res]

# bottom < top
hzDepthTests(top = c(10,20,30,50), bottom = c(0,10,20,30))
names(res)[res]

# bottom == top
hzDepthTests(top = c(10,20,30,50), bottom = c(0,20,20,30))
names(res)[res]

# overlap
hzDepthTests(top = c(0,5,20,30), bottom = c(10,20,30,50))
names(res)[res]

# gap
hzDepthTests(top = c(0,15,20,30), bottom = c(10,20,30,50))
names(res)[res]

# missing
hzDepthTests(c(0,15,NA,30),c(10,NA,30,50))
names(res)[res]

}
\author{
Andrew G. Brown & Dylan E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{ROSETTA.centroids}
\alias{ROSETTA.centroids}
\title{Average Hydraulic Parameters from the ROSETTA Model by USDA Soil Texture
Class}
\format{
A data frame: \describe{

\item{texture}{soil texture class, ordered from low to high
available water holding capacity} \item{theta_r}{average saturated
water content} \item{theta_s}{average residual water content}
\item{alpha}{average value, related to the inverse of the air entry
suction, log10-transformed values with units of cm}
\item{npar}{average value, index of pore size distribution,
log10-transformed values with units of 1/cm}

\item{theta_r_sd}{1 standard deviation of \code{theta_r}}
\item{theta_s_sd}{1 standard deviation of \code{theta_s}}
\item{alpha_sd}{1 standard deviation of \code{alpha}}
\item{npar_sd}{1 standard deviation of \code{npar}}

\item{sat}{approximate volumetric water content at which soil
material is saturated} \item{fc}{approximate volumetric water
content at which matrix potential = -33kPa} \item{pwp}{approximate
volumetric water content at which matrix potential = -1500kPa}
\item{awc}{approximate available water holding capacity: VWC(-33kPa)
\itemize{
\item VWC(-1500kPa)} }
}
}
\source{
\href{https://www.ars.usda.gov/pacific-west-area/riverside-ca/agricultural-water-efficiency-and-salinity-research-unit/docs/model/rosetta-class-average-hydraulic-parameters/}{ROSETTA Class Average Hydraulic Parameters}
}
\usage{
data(ROSETTA.centroids)
}
\description{
Average soil hydraulic parameters generated by the first stage predictions
of the ROSETTA model by USDA soil texture class. These data were extracted
from \href{https://www.ars.usda.gov/pacific-west-area/riverside-ca/agricultural-water-efficiency-and-salinity-research-unit/docs/model/rosetta-class-average-hydraulic-parameters/}{ROSETTA documentation} and re-formatted for ease of use.
}
\details{
Theoretical water-retention parameters for uniform soil material of each
texture class have been estimated via van Genuchten model.

See \href{https://ncss-tech.github.io/AQP/soilDB/fetchKSSL-VG-demo.html}{the related tutorial}
}
\examples{


\dontrun{

library(aqp)
library(soilDB)
library(latticeExtra)

data("ROSETTA.centroids")

# iterate over horizons and generate VG model curve
res <- lapply(1:nrow(ROSETTA.centroids), function(i) {
  m <- KSSL_VG_model(VG_params = ROSETTA.centroids[i, ], phi_min = 10^-3, phi_max=10^6)$VG_curve
  # copy generalized hz label
  m$hz <- ROSETTA.centroids$hz[i]
  # copy ID
  m$texture_class <- ROSETTA.centroids$texture[i]
  return(m)
})

# copy over lab sample number as ID
res <- do.call('rbind', res)

# check: OK
str(res)


# visual check: OK
xyplot(
  phi ~ theta | texture_class, data=res,
  type=c('l', 'g'),
  scales=list(alternating=3, x=list(tick.number=10), y=list(log=10, tick.number=10)),
  yscale.components=yscale.components.logpower,
  ylab=expression(Suction~~(kPa)),
  xlab=expression(Volumetric~Water~Content~~(cm^3/cm^3)),
  par.settings = list(superpose.line=list(col='RoyalBlue', lwd=2)),
  strip=strip.custom(bg=grey(0.85)),
  as.table=TRUE
)


}


}
\references{
van Genuchten, M.Th. (1980). "A closed-form equation for
predicting the hydraulic conductivity of unsaturated soils". Soil Science
Society of America Journal. 44 (5): 892-898.

Schaap, M.G., F.J. Leij, and M.Th. van Genuchten. 2001. ROSETTA: A computer
program for estimating soil hydraulic parameters with hierarchical
pedotransfer functions. Journal of Hydrology 251(3–4): 163-176.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulateColor.R
\name{simulateColor}
\alias{simulateColor}
\title{Simulate Soil Colors}
\usage{
simulateColor(
  method = c("dE00", "proportions", "mvnorm"),
  n,
  parameters,
  SPC = NULL
)
}
\arguments{
\item{method}{simulation method, see details}

\item{n}{number of simulated colors per group}

\item{parameters}{a \code{list}, format depends on \code{method}:
\itemize{
\item \code{proportions}: output from \code{\link[=aggregateColor]{aggregateColor()}}
\item \code{dE00}: formatted as \code{list(m = '7.5YR 3/3', thresh = 5, hues = c('7.5YR'))}
\item \code{mvnorm}: formatted as \code{list(hvc = x)}
}

Where \code{m} is a single representative Munsell chip, \code{thresh} is a threshold specified in CIE2000 color contrast (dE00), \code{hues} is a vector of allowed Munsell hues, and \code{x} is a \code{data.frame} representing columns of Munsell hue, value, and chroma having at least 3 rows.}

\item{SPC}{\code{SoilProfileCollection}, attempt to modify \code{SPC} with simulated colors}
}
\value{
a \code{list}, unless \code{SPC} is specified, then a \code{SoilProfileCollection} object
}
\description{
Simulate plausible soil colors based on several possible parameterization of a "range in characteristics" (RIC). Soil color RIC can be specified by a list of parameters:
\itemize{
\item soil color proportions, as output from \code{\link[=aggregateColor]{aggregateColor()}} -- \code{method = 'proportions'}
\item most likely Munsell color, CIE2000 threshold, and vector of acceptable hues -- \code{method = 'dE00'}
\item \code{data.frame} of Munsell hue, value, and chroma representing observed soil colors -- \code{method = 'mvnorm'}
}
}
\examples{

# restrict examples to 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# m: representative or most likely color
# thresh: dE00 threshold
# hues: allowed Munsell hues
p <- list(
  'A' = list(m = '7.5YR 3/3', thresh = 5, hues = c('7.5YR')),
  'BA' = list(m = '7.5YR 4/4', thresh = 8, hues = c('7.5YR')),
  'Bt1' = list(m = '7.5YR 4/4', thresh = 8, hues = c('5YR', '7.5YR')),
  'Bt2' = list(m = '5YR 4/5', thresh = 8, hues = c('5YR', '7.5YR')),
  'Bt3' = list(m = '10YR 4/6', thresh = 10, hues = c('10YR', '7.5YR')),
  'Cr' = list(m = '2.5G 6/2', thresh = 15, hues = c('2.5G', '2.5GY', '2.5BG'))
)

# simulate
(cols <- simulateColor(method = 'dE00', n = 10, parameters = p))

# preview
previewColors(parseMunsell(unlist(cols)), method = 'MDS')

# another example, this time using a larger dE00 threshold
p <- list(
  'A' = list(m = '7.5YR 3/3', thresh = 20, hues = c('10YR', '7.5YR', '5YR'))
)

# simulate
set.seed(54654)
cols <- simulateColor(method = 'dE00', n = 200, parameters = p)

# flatten
cols <- unlist(cols)

# tabulate, sort: most frequent color should be 7.5YR 3/3
sort(table(cols), decreasing = TRUE)

# review colors
previewColors(parseMunsell(cols))

# what does a dE00 threshold look like on 3 pages of hue?
contrastChart('7.5YR 3/3', hues = c('10YR', '7.5YR', '5YR'), thresh = 10)

}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spc2mpspline.R
\name{spc2mpspline,SoilProfileCollection-method}
\alias{spc2mpspline,SoilProfileCollection-method}
\alias{spc2mpspline}
\title{SoilProfileCollection wrapper for \code{mpspline2::mpspline()}}
\usage{
\S4method{spc2mpspline}{SoilProfileCollection}(
  object,
  var_name = NULL,
  method = c("est_1cm", "est_icm", "est_dcm"),
  pattern = "R|Cr|Cd|m",
  hzdesgn = NULL,
  ...
)
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{var_name}{Column name(s) in \verb{@horizons} slot of \code{object} containing
numeric values to spline}

\item{method}{Options include "est_1cm" (default; 1cm estimates), "est_icm"
(estimates over original layer boundaries), "est_dcm" (estimates over
constant interval, specified with \code{d} argument to \code{mpspline3::mpspline()}).
Default value for \code{d} is \code{c(0, 5, 15, 30, 60, 100, 200)}.}

\item{pattern}{Regex pattern to match for bottom of profile (passed to
\code{minDepthOf()}) default: "R|Cr|Cd|m"; only used if \code{hzdesgn} is specified}

\item{hzdesgn}{Column name in \verb{@horizons} slot of \code{object} containing horizon
designations default: \code{NULL}}

\item{...}{Additional arguments to \code{mpspline2::mpspline()}}
}
\value{
A SoilProfileCollection with 1cm slices. Spline variables are in
columns prefixed with "spline_" and RMSE/RMSE_IQR are in columns prefixed
with "rmse_". If any profiles were removed from the collection, their
profile IDs are stored in \code{attr(result, 'removed')}.
}
\description{
Generate mass-preserving splines for any numeric attribute in a
SoilProfileCollection using \code{mpspline2::mpspline()}. mpspline2 implements the
method described in Bishop et al. (1999).
}
\details{
This function now relies on the missing data checks provided by the
mpspline2 package. See \code{attr(..., 'removed')} to see whole profiles that
were removed from the set. Horizons containing \code{NA} in the property of
interest are dropped with a message.

Data completeness is assessed and the input SoilProfileCollection is
filtered and truncated to create a container for the results from
\code{mpspline2::mpspline()}.
}
\examples{
\dontshow{if (requireNamespace("mpspline2")) withAutoprint(\{ # examplesIf}
data(sp1)
depths(sp1) <- id ~ top + bottom
hzdesgnname(sp1) <- "name"
  
# run on a single variable
res <- spc2mpspline(sp1, "prop")
  
# plot single-variable result
plotSPC(res[1:5, ], color = "prop_spline", divide.hz = FALSE)

# add a second continuous numeric variable for demonstration
sp1$value2 <- runif(nrow(horizons(sp1)))
  
# run on multiple variables
res2 <- spc2mpspline(sp1, c("prop", "value2"))
  
# plot multi-variable result
plotSPC(res2[1:5, ], color = "value2_spline", divide.hz = FALSE)
  
# run on multiple variables with custom depth intervals
res3 <- spc2mpspline(sp1,
                     c("prop", "value2"),
                     method = "est_dcm",
                     d = c(0, 10, 20, 50, 100))
plotSPC(res3[1:5, ], color = "value2_spline", divide.hz = FALSE)
\dontshow{\}) # examplesIf}
}
\references{
Bishop, T.F.A., McBratney, A.B., Laslett, G.M. (1999) Modelling
soil attribute depth functions with equal-area quadratic smoothing splines.
Geoderma 91(1–2), pp. 27-45.
\doi{https://doi.org/10.1016/S0016-7061(99)00003-8}

O'Brien, Lauren (2025). mpspline2: Mass-Preserving Spline Functions for
Soil Data. R package version 0.1.9.
\url{https://cran.r-project.org/package=mpspline2}
}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hzOffset.R
\name{hzAbove}
\alias{hzAbove}
\alias{hzBelow}
\alias{hzOffset}
\title{Select Horizons Above or Below a Reference Horizon}
\usage{
hzAbove(x, ..., offset = NULL, SPC = TRUE, simplify = SPC, single = FALSE)

hzBelow(x, ..., offset = NULL, SPC = TRUE, simplify = SPC, single = FALSE)

hzOffset(x, hzidx, offset, SPC = FALSE, simplify = TRUE)
}
\arguments{
\item{x}{\code{SoilProfileCollection}}

\item{...}{comma-separated set of logical expressions, in context of horizon data, length for each expression must match number of horizons in \code{x}}

\item{offset}{integer vector or NULL, when specified this is the \verb{[, j]} (horizon/slice) index definining the horizons above or below a reference horizon that should be returned. see details}

\item{SPC}{logical, return a \code{SoilProfileCollection} subset to selected horizons}

\item{simplify}{logical, flatten list of horizon row indices into a numeric vector}

\item{single}{logical, interpret multiple matching reference horizons as a single horizon, defined by max(top) - min(bottom)}

\item{hzidx}{integer vector of target horizon row indices, typically calculated by \code{hzAbove()} and \code{hzBelow()}}
}
\value{
One of:
\itemize{
\item \code{SoilProfileCollection}: when \code{SPC = TRUE}
\item list of horizon row indices when \code{SPC = FALSE} and \code{simplify = FALSE}
\item vector of horizon row indices: when \code{SPC = FALSE} and \code{simplify = TRUE}
}

The return value will not include data from profiles which have no matching reference horizons.
}
\description{
These functions return "selected" horizons offset from "reference" horizon defined by logical expressions which include horizon data elements. Typically, \code{hzAbove()} and \code{hzBelow()} are used as a simpler interface to \code{hzOffset()}. Selected horizons can be returned as a complete \code{SoilProfileCollection} (\code{SPC = TRUE}), a list of horizon data row indices (\verb{SPC = FALSE, simplify = FALSE}), or as a numeric vector of horizon data row indices (\verb{SPC = FALSE, simplify = TRUE}).
}
\details{
Expressions that match multiple reference horizons will result in the inclusion of these reference horizons in the selected horizons. Interpreting multiple reference horizons as a single horizon (\code{single = TRUE}), defined by min(top) - max(bottom), will ensure that only selected horizons are returned.

The \code{offset} argument controls which "horizons above" or "horizons below" reference horizon(s) are selected, using the \code{SoilProfileCollection} \verb{[, j]} (horizon/slice) notation. Possible values and their interpretation are as follows:
\itemize{
\item \code{offset = NULL}: select all horizons above or below reference horizon(s)
\item \code{offset = 1}: select the first horizon above or below reference horizon(s)
\item \code{offset = 1:2}: select the first and second horizons above or below reference horizon(s)
\item \code{offset = 1:10}: safely select the 1st -> 10th horizons, even if there are fewer, above or below reference horizon(s)
}
}
\examples{

# example data
x <- c(
  'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
  'P2:Ap|AA|E|BhsBhs|Bw1Bw1|CCCCC',
  'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR',
  'P4:AA|EEE|BhsBhsBhsBhs|BwBw|CCCCC',
  'P5:AAAA|ACACACACAC|CCCCCCCCCCC|CdCdCd'
)

s <- quickSPC(x)

# single match for most profiles
.ex <- grepl('Bt3|Bw', s$name)
s$e <- .ex

# default: select all horizons above or below
a <- hzAbove(s, .ex, SPC = FALSE, simplify = TRUE)
b <- hzBelow(s, .ex, SPC = FALSE, simplify = TRUE)

op <- par(no.readonly = TRUE)
par(mar = c(0, 0, 3, 0))
plotSPC(
  s, color = 'e', col.label = 'reference', col.palette = c('grey', 'royalblue'), 
  name = 'name', hz.depths = TRUE, depth.axis = FALSE, 
  name.style = 'center-center', cex.names = 0.75
)

# highlight selected horizons above and below with brackets
addBracket(
  depths(s, hzID = FALSE)[a, ], agg = TRUE,
  offset = -0.3, col = 'darkgreen', tick.length = 0, lwd = 3
)

addBracket(
  depths(s, hzID = FALSE)[b, ], agg = TRUE,
  offset = -0.35, col = 'firebrick', tick.length = 0, lwd = 3
)


# select only 1st horizon above / below
a <- hzAbove(s, .ex, SPC = FALSE, simplify = TRUE, offset = 1)
b <- hzBelow(s, .ex, SPC = FALSE, simplify = TRUE, offset = 1)

plotSPC(
  s, color = 'e', col.label = 'reference', col.palette = c('grey', 'royalblue'), 
  name = 'name', hz.depths = TRUE, depth.axis = FALSE, 
  name.style = 'center-center', cex.names = 0.75
)

# highlight selected horizons above and below with brackets
addBracket(
  depths(s, hzID = FALSE)[a, ], agg = TRUE,
  offset = -0.3, col = 'darkgreen', tick.length = 0, lwd = 3
)

addBracket(
  depths(s, hzID = FALSE)[b, ], agg = TRUE,
  offset = -0.35, col = 'firebrick', tick.length = 0, lwd = 3
)


# multiple matches
.ex <- grepl('B', s$name)
s$e <- .ex

# default
a <- hzAbove(s, .ex, SPC = FALSE, simplify = TRUE)
b <- hzBelow(s, .ex, SPC = FALSE, simplify = TRUE)

par(mfcol = c(1, 2))
plotSPC(
  s,  col.label = 'reference', color = 'e', 
  col.palette = c('grey', 'royalblue'), 
  name = 'name', hz.depths = TRUE, depth.axis = FALSE, 
  name.style = 'center-center', cex.names = 0.75
)

addBracket(
  depths(s, hzID = FALSE)[a, ], agg = TRUE,
  offset = -0.3, col = 'darkgreen', tick.length = 0, lwd = 3
)

addBracket(
  depths(s, hzID = FALSE)[b, ], agg = TRUE,
  offset = -0.35, col = 'firebrick', tick.length = 0, lwd = 3
)

mtext('single = FALSE',  side = 1, line = -1.5, at = 0, adj = -0.5)


# interpret multiple reference hz as a single reference hz
a <- hzAbove(s, .ex, SPC = FALSE, simplify = TRUE, single = TRUE)
b <- hzBelow(s, .ex, SPC = FALSE, simplify = TRUE, single = TRUE)


plotSPC(
  s, col.label = 'reference', color = 'e', 
  col.palette = c('grey', 'royalblue'), name = 'name', 
  hz.depths = TRUE, depth.axis = FALSE, 
  name.style = 'center-center', cex.names = 0.75
)

addBracket(
  depths(s, hzID = FALSE)[a, ], agg = TRUE,
  offset = -0.3, col = 'darkgreen', tick.length = 0, lwd = 3
)

addBracket(
  depths(s, hzID = FALSE)[b, ], agg = TRUE,
  offset = -0.35, col = 'firebrick', tick.length = 0, lwd = 3
)

mtext('single = TRUE',  side = 1, line = -1.5, at = 0, adj = -0.5)


# demonstrate SPC = TRUE, single = TRUE
plotSPC(
  s, col.label = 'reference', color = 'e', 
  col.palette = c('grey', 'royalblue'), name = 'name', 
  hz.depths = TRUE, depth.axis = FALSE, 
  name.style = 'center-center', cex.names = 0.75, max.depth = 250
)

a <- hzAbove(s, .ex, SPC = TRUE, single = TRUE)

plotSPC(
  a, name = 'name', hz.depths = TRUE, depth.axis = FALSE, 
  name.style = 'center-center', cex.names = 0.75, max.depth = 250
)
title('selected profiles/horizons')

par(op)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alignTransect.R
\name{alignTransect}
\alias{alignTransect}
\title{Calculate Relative Positions from Transect Data}
\usage{
alignTransect(x, x.min, x.max, fix = TRUE, ...)
}
\arguments{
\item{x}{numeric vector, describing values along a transect: distance, elevation, climatic variables, etc.. Typically sourced from the site level attributes of a \code{SoilProfileCollection} object. Order is not important.}

\item{x.min}{numeric, lower boundary to relative position scale}

\item{x.max}{numeric, upper boundary to relative position scale}

\item{fix}{logical, attempt fixing overlapping positions with \code{fixOverlap}}

\item{...}{additional arguments to \code{fixOverlap}}
}
\value{
\code{list} containing:
\itemize{
\item \code{grad}: values of \code{x} in ascending order
\item \code{order}: ordering vector of \code{x}
\item \code{relative.pos}: elements of \code{x} translated to the new relative scale defined by \code{x.min} and \code{x.max}
}
}
\description{
This function is used to support relative positioning of soil profiles by \code{plotSPC}, based on transect or gradient values typically associated with a site level attribute (e.g. elevation). Gradient values specified in \code{x} are translated to the range used by \code{plotSPC} (usually \verb{1, length(SPC)}) specified in \code{x.min} and \code{x.max}.
}
\details{
See the \href{http://ncss-tech.github.io/AQP/aqp/genhz-distance-eval.html}{Pair-Wise Distances by Generalized Horizon Labels} tutorial for additional examples.
}
\examples{

data("sierraTransect")

# split transects
g <- subset(sierraTransect, transect == 'Granite')
a <- subset(sierraTransect, transect == 'Andesite')

g.p <- alignTransect(g$elev, x.min = 1, x.max = length(g), fix = FALSE)
a.p <- alignTransect(a$elev, x.min = 1, x.max = length(a), fix = FALSE)

op <- par(mar=c(2,0,0,2), mfrow=c(2,1))

plotSPC(g, width=0.25, name.style='center-center', 
        cex.names=0.75, 
        relative.pos = g.p$relative.pos, plot.order = g.p$order)

axis(1, at = g.p$relative.pos, labels = g.p$grad, line = -1.5)

plotSPC(a, width=0.25, name.style='center-center', 
        cex.names=0.75, 
        relative.pos = a.p$relative.pos, plot.order = a.p$order)

axis(1, at = a.p$relative.pos, labels = a.p$grad, line = -1.5)


par(op)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-slice-methods.R
\docType{methods}
\name{slice-methods}
\alias{slice-methods}
\alias{slice.fast}
\alias{get.slice}
\alias{slice}
\alias{slice,SoilProfileCollection-method}
\title{Slicing of SoilProfileCollection Objects}
\usage{
slice.fast(object, fm, top.down = TRUE, just.the.data = FALSE, strict = TRUE)

\S4method{slice}{SoilProfileCollection}(object, fm, top.down = TRUE, just.the.data = FALSE, strict = TRUE)

get.slice(h, id, top, bottom, vars, z, include = "top", strict = TRUE)
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{fm}{A formula: either \code{integer.vector ~ var1 + var2 + var3} where
named variables are sliced according to \code{integer.vector} OR where all
variables are sliced according to \code{integer.vector}: \code{integer.vector ~ .}.}

\item{top.down}{logical, slices are defined from the top-down: \code{0:10} implies 0-11 depth units.}

\item{just.the.data}{Logical, return just the sliced data or a new \code{SoilProfileCollection} object.}

\item{strict}{Check for logic errors? Default: \code{TRUE}}

\item{h}{Horizon data.frame}

\item{id}{Profile ID}

\item{top}{Top Depth Column Name}

\item{bottom}{Bottom Depth Column Name}

\item{vars}{Variables of Interest}

\item{z}{Slice Depth (index).}

\item{include}{Either \code{'top'} or \code{'bottom'}. Boundary to include in slice. Default: \code{'top'}}
}
\value{
Either a new \code{SoilProfileCollection} with data sliced according to \code{fm}, or a \code{data.frame}.
}
\description{
A method for "slicing" of SoilProfileCollection objects into constant depth intervals. Now deprecated, see \verb{[dice()]}.
}
\section{Details}{
 By default, slices are defined from the top-down:
\code{0:10} implies 0-11 depth units.
}

\examples{

library(aqp)

# simulate some data, IDs are 1:20
d <- lapply(1:20, random_profile)
d <- do.call('rbind', d)

# init SoilProfileCollection object
depths(d) <- id ~ top + bottom
head(horizons(d))

# generate single slice at 10 cm
# output is a SoilProfileCollection object
s <- dice(d, fm = 10 ~ name + p1 + p2 + p3)

# generate single slice at 10 cm, output data.frame
s <- dice(d, 10 ~ name + p1 + p2 + p3, SPC = FALSE)

# generate integer slices from 0 - 26 cm
# note that slices are specified by default as "top-down"
# result is a SoilProfileCollection
# e.g. the lower depth will always by top + 1
s <- dice(d, fm = 0:25 ~ name + p1 + p2 + p3)
par(mar=c(0,1,0,1))
plotSPC(s)

# generate slices from 0 - 11 cm, for all variables
s <- dice(d, fm = 0:10 ~ .)
print(s)

# compute percent missing, for each slice,
# if all vars are missing, then NA is returned
d$p1[1:10] <- NA
s <- dice(d, 10 ~ ., SPC = FALSE, pctMissing = TRUE)
head(s)

\dontrun{
##
## check sliced data
##

# test that mean of 1 cm slices property is equal to the
# hz-thickness weighted mean value of that property
data(sp1)
depths(sp1) <- id ~ top + bottom

# get the first profile
sp1.sub <- sp1[which(profile_id(sp1) == 'P009'), ]

# compute hz-thickness wt. mean
hz.wt.mean <- with(
  horizons(sp1.sub),
  sum((bottom - top) * prop) / sum(bottom - top)
)

# hopefully the same value, calculated via slice()
s <- dice(sp1.sub, fm = 0:max(sp1.sub) ~ prop)
hz.slice.mean <- mean(s$prop, na.rm = TRUE)

# they are the same
all.equal(hz.slice.mean, hz.wt.mean)
}

}
\references{
D.E. Beaudette, P. Roudier, A.T. O'Geen, Algorithms for
quantitative pedology: A toolkit for soil scientists, Computers &
Geosciences, Volume 52, March 2013, Pages 258-268,
10.1016/j.cageo.2012.10.020.
}
\seealso{
\code{\link{slab}}
}
\author{
D.E. Beaudette
}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R,
%   R/SoilProfileCollection-setters.R
\docType{methods}
\name{site,SoilProfileCollection-method}
\alias{site,SoilProfileCollection-method}
\alias{site}
\alias{site<-,SoilProfileCollection-method}
\alias{site<-}
\title{Retrieve site data from SoilProfileCollection}
\usage{
\S4method{site}{SoilProfileCollection}(object)

\S4method{site}{SoilProfileCollection}(object) <- value
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{value}{A formula or object inheriting \code{data.frame}}
}
\description{
Get site data from SoilProfileCollection. Result is returned in the same \code{data.frame} class used to initially construct the SoilProfileCollection.

There are two options available via the \code{site<-} setter.

The first is a "normalization" by formula interface, whereby one specifies an attribute that is constant in horizons within profiles to be promoted to a site-level variable: \code{site(spc) <- ~ horizonvariable}

The second is creation of site data from an external \code{data.frame} via merge (LEFT JOIN). There must be one or more same-named columns (with at least some matching data) on the left and right hand side to facilitate the join: \code{site(spc) <- newdata}
}
\examples{

# load test data
data(sp2)

# promote to SPC
depths(sp2) <- id ~ top + bottom

# normalize a horizon-level attribute to site
site(sp2) <- ~ surface

# inspect site table
site(sp2)

# make some data: classify two geomorphic surfaces with numeric value
newdata <- data.frame(surface = c("holocene",
                                  "lower riverbank"),
                      newvalue = c(1,2))

# do left join based on newly-normalized "surface" attribute
site(sp2) <- newdata

# inspect site table: holocene & lower riverbank have values
site(sp2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{depth_units,SoilProfileCollection-method}
\alias{depth_units,SoilProfileCollection-method}
\alias{depth_units}
\alias{depth_units<-,SoilProfileCollection-method}
\alias{depth_units<-}
\title{Get depth units from metadata}
\usage{
\S4method{depth_units}{SoilProfileCollection}(object)

\S4method{depth_units}{SoilProfileCollection}(object) <- value
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{value}{character, a value representing units. Default \code{'cm'}.}
}
\description{
Get units of depth measurement from metadata. Default value is centimeters.
}
\examples{

data(sp5)

## get depth units
du <- depth_units(sp5)

# set alternate units; e.g. inches
depth_units(sp5) <- 'in'

# replace original value (cm)
depth_units(sp5) <- du
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guessGenHzLevels.R
\name{guessGenHzLevels}
\alias{guessGenHzLevels}
\title{Guess Appropriate Ordering for Generalized Horizon Labels}
\usage{
guessGenHzLevels(x, hz = GHL(x, required = TRUE))
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}

\item{hz}{name of horizon-level attribute containing generalized horizon
labels, see details}
}
\value{
a list: \item{levels}{a vector of levels sorted by median horizon
depth mid-point} \item{median.depths}{a vector of median horizon mid-points}
}
\description{
This function makes an (educated) guess at an appropriate set of levels for
generalized horizon labels using the median of horizon depth mid-points.
}
\details{
This function is useful when groups of horizons have been generalized via
some method other than \code{generalize.hz}. For example, it may be useful
to generalize horizons using labels derived from slice depths. The default
sorting of these labels will not follow a logical depth-wise sorting when
converted to a factor. \code{guessGenHzLevels} does a good job of "guessing"
the proper ordering of these labels based on median horizon depth mid-point.
}
\examples{

# load some example data
data(sp1, package='aqp')

# upgrade to SoilProfileCollection
depths(sp1) <- id ~ top + bottom

# generalize horizon names
n <- c('O', 'A', 'B', 'C')
p <- c('O', 'A', 'B', 'C')
sp1$genhz <- generalize.hz(sp1$name, n, p)

# note: levels are in the order in which originally defined:
levels(sp1$genhz)

# generalize horizons by depth slice
s <- dice(sp1, c(5, 10, 15, 25, 50, 100, 150) ~ .)
s$slice <- paste0(s$top, ' cm')
# not a factor
levels(s$slice)

# the proper ordering of these new labels can be guessed from horizon depths
guessGenHzLevels(s, 'slice')

# convert to factor, and set proper order
s$slice <- factor(s$slice, levels=guessGenHzLevels(s, 'slice')$levels)

# that is better
levels(s$slice)

}
\seealso{
\code{\link{generalize.hz}}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform.R
\name{transform,SoilProfileCollection-method}
\alias{transform,SoilProfileCollection-method}
\alias{mutate}
\alias{mutate,SoilProfileCollection-method}
\title{Transform a SPC with expressions based on site or horizon level attributes}
\usage{
\S4method{transform}{SoilProfileCollection}(`_data`, ...)
}
\arguments{
\item{_data}{A SoilProfileCollection}

\item{...}{Comma-separated set of R expressions e.g. \code{thickness = hzdepb - hzdept, hzdepm = hzdept + round(thk / 2)}}
}
\value{
A SoilProfileCollection
}
\description{
\code{transform()} is a function used for modifying columns in SoilProfileCollections.

It allows the user to specify an arbitrary number of expressions that resolve to the (re-)calculation of one or more site or horizon level attributes. For instance: \code{mutate(spc, thickness = hzdepb - hzdept)}. The expressions may depend on one another, and are evaluated from left to right.
}
\author{
Andrew G. Brown.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/soilColorIndices.R
\name{horizonColorIndices}
\alias{horizonColorIndices}
\title{Horizon Color Indices}
\usage{
horizonColorIndices(p, hue = "m_hue", value = "m_value", chroma = "m_chroma")
}
\arguments{
\item{p}{A SoilProfileCollection}

\item{hue}{Column name containing moist hue; default: "m_hue"}

\item{value}{Column name containing moist value; default: "m_value"}

\item{chroma}{Column name containing moist chroma; default: "m_chroma"}
}
\value{
A data.frame containing unique pedon and horizon IDs and horizon-level color indices.
}
\description{
Calculate basic horizon-level color indices for a SoilProfileCollection. Basic indices do not require aggregation over the whole profile or comparison to a "reference" (e.g. parent material) color. Includes Hurst (1977) Redness Index, Barron-Torrent Redness Index (1986) and Buntley-Westin Index (1965). This is a wrapper method around several horizon-level indices. See the individual functions for more details.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

data(sp1)

# promote sp1 data to SoilProfileCollection
depths(sp1) <- id ~ top + bottom

# move site data
site(sp1) <- ~ group

# use Munsell color notation as horizon name
sp1$m <- sprintf("\%s \%s/\%s", sp1$hue, sp1$value, sp1$chroma)

# compute indices
# merged into `sp1` with left-join on hzidname(sp1)
horizons(sp1) <- horizonColorIndices(sp1, hue="hue", value="value", chroma="chroma")

# visualize
par(mar = c(0, 1, 3, 1))
plotSPC(sp1, color='hurst_redness', name = 'm')
plotSPC(sp1, color='barron_torrent_redness', name = 'm')
plotSPC(sp1, color='buntley_westin', name = 'm')

}
\seealso{
\code{\link{hurst.redness}} \code{\link{barron.torrent.redness.LAB}} \code{\link{buntley.westin.index}}
}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregateSoilDepth.R
\name{aggregateSoilDepth}
\alias{aggregateSoilDepth}
\title{Probabilistic Estimation of Soil Depth within Groups}
\usage{
aggregateSoilDepth(
  x,
  groups,
  crit.prob = 0.9,
  name = hzdesgnname(x),
  p = "Cr|R|Cd",
  ...
)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}

\item{groups}{the name of a site-level attribute that defines groups of profiles within a collection}

\item{crit.prob}{probability cutoff used to determine where the most likely depth to contact will be, e.g. 0.9 translates to 90\% of profiles are shallower than this depth}

\item{name}{horizon-level attribute where horizon designation is stored, defaults to \code{hzdesgnname(x)}}

\item{p}{a REGEX pattern that matches non-soil genetic horizons}

\item{...}{additional arguments to \code{slab}}
}
\value{
A \code{data.frame} is returned, with as many rows as there are unique group labels, as specified in \code{groups}.
}
\description{
Estimate the most-likely depth to contact within a collection of soil profiles. Consider \code{getSoilDepthClass} followed by group-wise percentile estimation as a faster alternative.
}
\details{
This function computes a probability-based estimate of soil depth by group. If no grouping variable exists, a dummy value can be used to compute a single estimate. The \code{crit.prob} argument sets the critical probability (e.g. 0.9) at which soil depth within a group of profiles is determined. For example, a \code{crit.prob} of 0.95 might result in an estimated soil depth (e.g. 120cm) where 95\% of the profiles (by group) had depths that were less than or equal to 120cm.
}
\examples{

data(sp1)
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# set horizon designation in SPC
hzdesgnname(sp1) <- 'name'

aggregateSoilDepth(sp1, 'group', crit.prob = 0.9)

}
\seealso{
\code{\link[=estimateSoilDepth]{estimateSoilDepth()}} \code{\link[=slab]{slab()}}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-metadata.R
\docType{methods}
\name{GHL}
\alias{GHL}
\alias{GHL,SoilProfileCollection-method}
\alias{GHL<-}
\alias{GHL<-,SoilProfileCollection-method}
\title{Get or Set Generalized Horizon Label (GHL) Column Name}
\usage{
\S4method{GHL}{SoilProfileCollection}(object, required = FALSE)

\S4method{GHL}{SoilProfileCollection}(object, required = FALSE) <- value
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{required}{logical, is this attribute required? If it is, set to \code{TRUE} to trigger error on invalid \code{value}.}

\item{value}{character, name of column containing generalized horizon labels}
}
\description{
\code{GHL()}: Get column name containing generalized horizon labels

\verb{GHL<-}: Set generalized horizon label column name
}
\details{
Store the column name containing generalized horizon labels in the metadata slot of the SoilProfileCollection.
}
\examples{

data(sp1)

# promote to SPC
depths(sp1) <- id ~ top + bottom

# set horizon designation column
GHL(sp1) <- "name"

# get horizon designation column
GHL(sp1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorContrast.R
\name{colorContrast}
\alias{colorContrast}
\title{Metrics of Contrast Suitable for Comparing Soil Colors}
\usage{
colorContrast(m1, m2 = NULL)
}
\arguments{
\item{m1}{vector of Munsell colors ('10YR 3/3'), must be >1 color if \code{m2} is not specified}

\item{m2}{optional vector of Munsell colors ('10YR 3/6')}
}
\value{
\code{data.frame} with the following columns:
\itemize{
\item m1: Munsell color 1
\item m2: Munsell color 2
\item dH: delta-hue, as computed by \code{huePosition}
\item dV: delta-value, absolute value of difference in Munsell value (m1 vs. m2)
\item dc: delta-chroma, absolute value of difference in Munsell chroma (m1 vs. m2)
\item dE00: delta-E00, e.g. the \href{https://en.wikipedia.org/wiki/Color_difference#CIEDE2000}{CIE delta-E as refined in 2000}
\item cc: soil color contrast class, as specified in Soil Survey Technical Note 2.
}
}
\description{
Pair-wise comparisons of Munsell color specifications, based on the NCSS color contrast classes (Soil Survey Technical Note 2) and CIE delta-E 2000 metric.
}
\details{
This function is fully vectorized over elements of \code{m1} and \code{m2} when both are of the same length. Recycling is not implemented. The minimum set of all pair-wise comparisons are performed within elements of \code{m1} when \code{m2} is not specified. For example, comparisons between A, B, and C are limited to \emph{A}x\emph{B}, \emph{A}x\emph{C}, and \emph{B}x\emph{C}. Use \code{expand.grid()} to generate suitable input from 1:many or many:1 type comparisons. See \href{http://ncss-tech.github.io/AQP/aqp/color-contrast.html}{this tutorial} for an expanded discussion and more examples. Neutral colors are not mentioned in SSTN2: in this function any comparison to a neutral color (e.g. 'N 3/') are assigned a delta-hue of 1. Since SSTN2 expects hues to be counted clock wise from 5R, it possible to get very large delta-hue values for otherwise adjacent colors: '5R' vs. '2.5R'. This will be addressed in an update to the standards.

The most meaningful representation of color contrast is the CIE2000 (dE00) metric.
}
\note{
delta-E00 is computed by the \href{https://CRAN.R-project.org/package=farver}{farver package}.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# two sets of colors to compare
m1 <- c('10YR 6/3', '7.5YR 3/3', '10YR 2/2', '7.5YR 3/4')
m2 <- c('5YR 3/4', '7.5YR 4/4', '2.5YR 2/2', '7.5YR 6/3')

# contrast metrics
colorContrast(m1, m2)

# adjacent chips
colorContrast('10YR 3/3', '10YR 3/4')
colorContrast('10YR 3/3', '7.5YR 3/3')

# highly contrasting colors
# http://colour.granjow.net/fabercastell-polychromos.html
colorContrastPlot('10B 4/13', '10YR 10/15',
labels = c('helioblue-reddish', 'light cadmium yellow')
)


## Note: neutral hues aren't defined in TN2
# approximation / extension of the concept
colorContrast(m1 = 'N 3/', m2 = 'N 6/')

colorContrast(m1 = '10YR 3/3', m2 = 'N 3/')

# pair-wise comparisons
m1 <- c('10YR 6/3', '7.5YR 3/3', '10YR 2/2', 'N 3/')
colorContrast(m1)

}
\references{
\enumerate{
\item https://en.wikipedia.org/wiki/Color_difference
}
}
\seealso{
\link{colorContrastPlot}, \link{huePosition}, \link{huePositionCircle}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/soilColorSignature.R
\name{.colorSig2PerceptualDistMat}
\alias{.colorSig2PerceptualDistMat}
\title{Interpret a color signature containing color groups of CIELAB coordinates using perceptual distance via CIE dE00}
\usage{
.colorSig2PerceptualDistMat(pig)
}
\arguments{
\item{pig}{\code{data.frame} results from \verb{soilColorSignature(..., method = c('pam', 'depthSlices))}}
}
\value{
dist object
}
\description{
Interpret a color signature containing color groups of CIELAB coordinates using perceptual distance via CIE dE00
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy_uncertainty.R
\name{mu_confusion_matrix}
\alias{mu_confusion_matrix}
\title{Map unit confusion matrix and other classification measures}
\usage{
mu_confusion_matrix(
  x,
  mapunit = "nationalmusym",
  cophase = "coiid",
  comppct = "comppct_r",
  muacres = "muacres"
)
}
\arguments{
\item{x}{\code{data.frame}}

\item{mapunit}{\code{character} column name containing the mapunit identifier (e.g. nationalmusym)}

\item{cophase}{\code{character} column name containing the soil component phase identifier (e.g. coiid or paste(compname, localphase))}

\item{comppct}{\code{character} column name containing the component percent (e.g. comppct_r)}

\item{muacres}{\code{character} column name containing the total area of the mapunit (e.g. muacres)}
}
\value{
\code{list} a confusion matrix, overall purity (OP) (i.e. overall accuracy), map unit purity (MP) (i.e. user accuracy), class representation (CR) (i.e. producer accuracy), and Shannon entropy (E). The measure names were selected to be consistent with the alternative terms proposed by Lark (1995) and Brus (2011).
}
\description{
This function reverse engineers a confusion matrix and other classification measures from soil map unit component percentages (i.e. composition) and area (i.e. acres).
}
\details{
\code{mu_confusion_matrix} There are several common statistical measures used to gauge the accuracy of categorical maps. These measures are typically not estimated for soil surveys but can be inferred from a map unit’s soil component composition percentages and size (i.e. acres). In general, overall purity or accuracy (OP) is related to map unit kind (e.g. consociations vs complexes). For several reasons, the “true” accuracies are unknown, and these values should be interpreted as Bayesian prior estimates. However, it is likely that the estimates are optimistic if they haven't been derived from an external validation. Existing and future digital soil mapping products could be used to validate how optimistic the current OA estimates are.
}
\examples{

# example data
mu <- rbind(
    data.frame(mapunit = "A", cophase = c("Alpha", "Beta"), comppct = c(90, 10), muacres = 100),
    data.frame(mapunit = "B", cophase = c("Beta", "Alpha"), comppct = c(70, 30), muacres = 1000)
    )

mu_confusion_matrix(mu, mapunit = "mapunit", cophase = "cophase", comppct = "comppct")

}
\references{
\itemize{
\item Brus DJ, Kempen B, Heuvelink GBM. 2011. Sampling for validation of digital soil maps. European Journal of Soil Science. 62(3):394–407. \doi{10.1111/j.1365-2389.2011.01364.x}
\item Lark RM. 1995. Components of accuracy of maps with special reference to discriminant analysis on remote sensor data. International Journal of Remote Sensing. 16(8):1461–1480. \doi{10.1080/01431169508954488}
}
}
\seealso{
\code{\link[=tauW]{tauW()}}, \code{\link[=shannonEntropy]{shannonEntropy()}}, \code{\link[=confusionIndex]{confusionIndex()}}
}
\author{
Stephen Roecker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thicknessOf.R
\name{thicknessOf}
\alias{thicknessOf}
\title{Calculate Thickness of Horizons Matching Logical Criteria}
\usage{
thicknessOf(
  x,
  pattern = NULL,
  hzdesgn = hzdesgnname(x, required = TRUE),
  method = "cumulative",
  prefix = "",
  thickvar = "thickness",
  depthvars = horizonDepths(x),
  FUN = function(x, pattern, hzdesgn, ...) grepl(pattern, x[[hzdesgn]]),
  na.rm = FALSE,
  ...
)
}
\arguments{
\item{x}{A \emph{SoilProfileCollection}}

\item{pattern}{\emph{character}. A pattern to match in \code{hzdesgn}; used with the default \code{FUN} definition for regular expression pattern matching on horizons.}

\item{hzdesgn}{\emph{character}. A column containing horizon designations or other horizon-level character label used to identify matches; used with the default \code{FUN} definition.}

\item{method}{\emph{character}. Either \code{"cumulative"} (default) or \code{"minmax"}. See details.}

\item{prefix}{\emph{character}. Column prefix for calculated \code{thickvar} (and \code{depthvar} for \code{method="minmax"}) column results. Default: \code{""}.}

\item{thickvar}{\emph{character} Length \code{1}. Column name to use for calculated thickness column. Default: \code{"thickness"}}

\item{depthvars}{\emph{character}. Length \code{2}. Column names to use for calculated minimum top depth and maximum bottom depth in \code{method="minmax"}. Default: \code{horizonDepths(x)}}

\item{FUN}{\emph{function}. A function that returns a \emph{logical} vector equal in length to the number of horizons in \code{x}. See details.}

\item{na.rm}{\emph{logical}. Omit \code{NA} values in summaries of thickness and in matching? Default: \code{FALSE}}

\item{...}{Additional arguments passed to the matching function \code{FUN}.}
}
\value{
A \emph{data.frame}-like object (corresponding to the \code{aqp_df_class()} of \code{x}) with one row per profile in \code{x}. First column is always the profile ID which is followed by \code{"thickness"}. In \code{method="minmax"} the upper and lower boundaries used to calculate \code{"thickness"} are also returned as \code{"tmin"} and \code{"tmax"} columns, respectively.
}
\description{
This function calculates the cumulative (\code{method="cumulative"}, default) or maximum difference between (\code{method="minmax"}) horizons within a profile that match a defined pattern (\code{pattern}) or, more generally, any set of horizon-level logical expressions encoded in a function (\code{FUN}).
}
\details{
The two thickness methods currently available are:
\itemize{
\item \code{method="cumulative"} (default): cumulative thickness of horizons where \code{FUN} returns true
\item \code{method="minmax"}: maximum bottom depth minus minimum top depth of horizons where \code{FUN} returns true
}

If a custom function (\code{FUN}) is used, it should accept arbitrary additional arguments via an ellipsis (\code{...}).
It is not necessary to do anything with arguments, but the result should match the number of horizons found in the input SoilProfileCollection \code{x}.
}
\examples{
data("jacobs2000")

# cumulative thickness of horizon designations matching "Bt"
thicknessOf(jacobs2000, "Bt")

# maximum bottom depth minus minimum top depth of horizon designations matching "Bt"
thicknessOf(jacobs2000, "Bt", prefix = "Bt_", method = "minmax")

# cumulative thickness of horizon designations matching "A|B"
thicknessOf(jacobs2000, "A|B", prefix = "AorB_")

# maximum bottom depth minus minimum top depth of horizon designations matching "A|B"
thicknessOf(jacobs2000, "A|B", method = "minmax", prefix = "AorB_")
# note that the latter includes the thickness of E horizons between the A and the B

# when using a custom function (be sure to accept ... and consider the effect of NA values)

# calculate cumulative thickness of horizons containing >18\% clay
thicknessOf(jacobs2000, prefix = "claygt18_", 
            FUN = function(x, ...) !is.na(x[["clay"]]) & x[["clay"]] > 18)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\docType{methods}
\name{isEmpty,SoilProfileCollection-method}
\alias{isEmpty,SoilProfileCollection-method}
\alias{isEmpty}
\title{Check for "empty" profiles in a SoilProfileCollection}
\usage{
\S4method{isEmpty}{SoilProfileCollection}(object, ...)
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{...}{Additional arguments not used.}
}
\value{
logical. Vector of length equal to number of profiles in \code{object}. Returns \code{TRUE} when a profile has one horizon with \code{NA} top and bottom depths
}
\description{
"Empty" profiles are used as placeholders for positions in a \code{SoilProfileCollection}
These profiles result from operations that remove or extract portions of horizons
from source profiles.
}
\details{
In a \code{SoilProfileCollection} an empty profile occurs when it has one horizon,
with \code{NA} top and bottom depths. Generally all non-profile ID site and horizon-level
values are all also \code{NA}, but only the depths are checked by \code{isEmpty()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotSPC.R
\docType{methods}
\name{plotSPC}
\alias{plotSPC}
\alias{plot,SoilProfileCollection,ANY-method,plot.SoilProfileCollection}
\alias{plot}
\alias{plot,SoilProfileCollection-method}
\alias{plot,SoilProfileCollection,ANY-method}
\title{Create Soil Profile Sketches}
\usage{
plotSPC(
  x,
  color = "soil_color",
  width = ifelse(length(x) < 2, 0.15, 0.25),
  name = hzdesgnname(x),
  name.style = "right-center",
  label = idname(x),
  raggedBottom = NULL,
  hz.depths = FALSE,
  hz.depths.offset = ifelse(fixLabelCollisions, 0.03, 0),
  hz.depths.lines = fixLabelCollisions,
  depth.axis = list(style = "traditional", cex = cex.names * 1.15),
  alt.label = NULL,
  alt.label.col = "black",
  cex.names = 0.5,
  cex.id = cex.names + (0.2 * cex.names),
  font.id = 2,
  srt.id = 0,
  offset.id = 0.5,
  print.id = TRUE,
  id.style = "auto",
  plot.order = 1:length(x),
  relative.pos = 1:length(x),
  add = FALSE,
  scaling.factor = 1,
  y.offset = rep(0, times = length(x)),
  x.idx.offset = 0,
  n = length(x),
  max.depth = ifelse(is.infinite(max(x)), 200, max(x)),
  n.depth.ticks = 10,
  shrink = FALSE,
  shrink.cutoff = 3,
  shrink.thin = NULL,
  abbr = FALSE,
  abbr.cutoff = 5,
  divide.hz = TRUE,
  hz.distinctness.offset = NULL,
  hz.topography.offset = NULL,
  hz.boundary.lty = NULL,
  density = NULL,
  show.legend = TRUE,
  col.label = color,
  col.palette = c("#5E4FA2", "#3288BD", "#66C2A5", "#ABDDA4", "#E6F598", "#FEE08B",
    "#FDAE61", "#F46D43", "#D53E4F", "#9E0142"),
  col.palette.bias = 1,
  col.legend.cex = 1,
  n.legend = 8,
  lwd = 1,
  lty = 1,
  default.color = grey(0.95),
  fixLabelCollisions = hz.depths,
  fixOverlapArgs = list(method = "E", q = 1),
  cex.depth.axis = cex.names,
  axis.line.offset = -2,
  plot.depth.axis = TRUE,
  ...
)

\S4method{plot}{SoilProfileCollection}(x, y, ...)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}

\item{color}{quoted column name containing R-compatible color descriptions, or numeric / categorical data to be displayed thematically; see details}

\item{width}{scaling of profile widths (typically 0.1 - 0.4)}

\item{name}{quoted column name of the (horizon-level) attribute containing horizon designations or labels, if missing \code{hzdesgnname(x)} is used. Suppress horizon name printing by setting \code{name = NA} or \code{name = ''}.}

\item{name.style}{one of several possible horizon designations labeling styles: \code{c('right-center', 'left-center', 'left-top', 'center-center', 'center-top')}}

\item{label}{quoted column name of the (site-level) attribute used to identify profile sketches}

\item{raggedBottom}{either quoted column name of the (site-level) attribute (logical) used to mark profiles with a truncated lower boundary, or \code{FALSE} suppress ragged bottom depths when \code{max.depth < max(x)}}

\item{hz.depths}{logical, annotate horizon top depths to the right of each sketch (\code{FALSE})}

\item{hz.depths.offset}{numeric, user coordinates for left-right adjustment for horizon depth annotation; reasonable values are usually within 0.01-0.05 (default: 0)}

\item{hz.depths.lines}{logical, draw segments between horizon depth labels and actual horizon depth; this is useful when including horizon boundary distinctness and/or \code{fixLabelCollisions = TRUE}}

\item{depth.axis}{logical or list. Use a logical to suppress (\code{FALSE}) or add depth axis using defaults (\code{TRUE}). Use a list to specify one or more of:
\itemize{
\item \code{style}: character, one of 'traditional', 'compact', or 'tape'
\item \code{line}: numeric, negative values move axis to the left (does not apply to \code{style = 'tape'})
\item \code{cex}: numeric, scaling applied to entire depth axis
\item \code{interval}: numeric, axis interval
See examples.
}}

\item{alt.label}{quoted column name of the (site-level) attribute used for secondary annotation}

\item{alt.label.col}{color used for secondary annotation text}

\item{cex.names}{baseline character scaling applied to all text labels}

\item{cex.id}{character scaling applied to \code{label}}

\item{font.id}{font style applied to \code{label}, default is 2 (bold)}

\item{srt.id}{rotation applied to \code{label}, only when \code{id.style = 'top'}}

\item{offset.id}{vertical offset applied to \code{label}, only when \code{id.style = 'top'}}

\item{print.id}{logical, print \code{label} above/beside each profile? (\code{TRUE})}

\item{id.style}{\code{label} printing style: 'auto' (default) = simple heuristic used to select from: 'top' = centered above each profile, 'side' = 'along the top-left edge of profiles'}

\item{plot.order}{integer vector describing the order in which individual soil profiles should be plotted}

\item{relative.pos}{vector of relative positions along the x-axis, within \{1, n\}, ignores \code{plot.order} see details}

\item{add}{logical, add to an existing figure}

\item{scaling.factor}{vertical scaling of profile depths, useful for adding profiles to an existing figure}

\item{y.offset}{numeric vector of vertical offset for top of profiles in depth units of \code{x}, can either be a single numeric value or vector of length = \code{length(x)}. A vector of y-offsets will be automatically re-ordered according to \code{plot.order}.}

\item{x.idx.offset}{integer specifying horizontal offset from 0 (left-hand edge)}

\item{n}{integer describing amount of space along x-axis to allocate, defaults to \code{length(x)}}

\item{max.depth}{numeric. The lower depth for all sketches, deeper profiles are truncated at this depth. Use larger values to arbitrarily extend the vertical dimension, convenient for leaving extract space for annotation.}

\item{n.depth.ticks}{suggested number of ticks in depth scale}

\item{shrink}{logical, reduce character scaling for 'long' horizon by 80\%}

\item{shrink.cutoff}{character length defining 'long' horizon names}

\item{shrink.thin}{integer, horizon thickness threshold for shrinking horizon names by 80\%, only activated when \code{shrink = TRUE} (\code{NULL} = no shrinkage)}

\item{abbr}{logical, abbreviate \code{label}}

\item{abbr.cutoff}{suggested minimum length for abbreviated \code{label}}

\item{divide.hz}{logical, divide horizons with line segment? (\code{TRUE}), see details}

\item{hz.distinctness.offset}{\code{NULL}, or quoted column name (horizon-level attribute) containing vertical offsets used to depict horizon boundary distinctness (same units as profiles), see details and \code{\link[=hzDistinctnessCodeToOffset]{hzDistinctnessCodeToOffset()}}; consider setting \code{hz.depths.lines = TRUE} when used in conjunction with \code{hz.depths = TRUE}}

\item{hz.topography.offset}{\code{NULL}, or quoted column name (horizon-level attribute) containing offsets used to depict horizon boundary topography (same units as profiles), see details and \code{\link[=hzTopographyCodeToOffset]{hzTopographyCodeToOffset()}}}

\item{hz.boundary.lty}{quoted column name (horizon-level attribute) containing line style (integers) used to encode horizon topography}

\item{density}{fill density used for horizon color shading, either a single integer or a quoted column name (horizon-level attribute) containing integer values (default is \code{NULL}, no shading)}

\item{show.legend}{logical, show legend? (default is \code{TRUE})}

\item{col.label}{thematic legend title}

\item{col.palette}{color palette used for thematic sketches (default is \code{rev(brewer.pal(10, 'Spectral'))})}

\item{col.palette.bias}{color ramp bias (skew), see \code{\link[=colorRamp]{colorRamp()}}}

\item{col.legend.cex}{scaling of thematic legend}

\item{n.legend}{approximate number of classes used in numeric legend, max number of items per row in categorical legend}

\item{lwd}{line width multiplier used for sketches}

\item{lty}{line style used for sketches}

\item{default.color}{default horizon fill color used when \code{color} attribute is \code{NA}}

\item{fixLabelCollisions}{use \code{\link[=fixOverlap]{fixOverlap()}} to attempt fixing hz depth labeling collisions, will slow plotting of large collections; enabling also sets \code{hz.depths.lines = TRUE}. Additional arguments to \code{\link[=fixOverlap]{fixOverlap()}} can be passed via \code{fixOverlapArgs}. Overlap collisions cannot be fixed within profiles containing degenerate or missing horizon depths (e.g. top == bottom).}

\item{fixOverlapArgs}{a named list of arguments to \code{\link[=fixOverlap]{fixOverlap()}}. Overlap adjustments are attempted using electrostatic simulation with arguments: \code{list(method = 'E', q = 1)}. Alternatively, select adjustment by simulated annealing via \code{list(method = 'S')}. See \code{\link[=electroStatics_1D]{electroStatics_1D()}} and \code{\link[=SANN_1D]{SANN_1D()}} for details.}

\item{cex.depth.axis}{(deprecated, use \code{depth.axis} instead) character scaling applied to depth scale}

\item{axis.line.offset}{(deprecated, use \code{depth.axis} instead) horizontal offset applied to depth axis (default is -2, larger numbers move the axis to the right)}

\item{plot.depth.axis}{(deprecated, use \code{depth.axis} instead) logical, plot depth axis?}

\item{\dots}{other arguments passed into lower level plotting functions}

\item{y}{(not used)}
}
\description{
Generate a diagram of soil profile sketches from a \code{SoilProfileCollection} object. The \href{http://ncss-tech.github.io/aqp/articles/Introduction-to-SoilProfileCollection-Objects.html}{Introduction to SoilProfileCollection Objects Vignette} contains many examples and discussion of the large number of arguments to this function. The \href{https://ncss-tech.github.io/AQP/aqp/sketches.html}{Soil Profile Sketches} tutorial has longer-form discussion and examples pertaining to suites of related arguments.

Options can be used to conveniently specify sets of arguments that will be used in several calls to \code{plotSPC()} within a single R session. For example, arguments can be specified in a named list (\code{.a}) and set using: \code{options(.aqp.plotSPC.args = .a)}. Reset these options via \code{options(.aqp.plotSPC.args = NULL)}. Arguments explicitly passed to \code{plotSPC()} will override arguments set via \code{\link[=options]{options()}}.
}
\details{
Depth limits (\code{max.depth}) and number of depth ticks (\code{n.depth.ticks}) are \emph{suggestions} to the \code{\link[=pretty]{pretty()}} function. You may have to tinker with both parameters to get what you want.

The 'side' \code{id.style} is useful when plotting a large collection of profiles, and/or, when profile IDs are long.

If the column containing horizon designations is not specified (the \code{name} argument), a column (presumed to contain horizon designation labels) is guessed based on regular expression matching of the pattern 'name'--this usually works, but it is best to manual specify the name of the column containing horizon designations.

The \code{color} argument can either name a column containing R-compatible colors, possibly created via \code{\link[=munsell2rgb]{munsell2rgb()}}, or column containing either numeric or categorical (either factor or character) values. In the second case, values are converted into colors and displayed along with a simple legend above the plot. Note that this functionality makes several assumptions about plot geometry and is most useful in an interactive setting.

Adjustments to the legend can be specified via \code{col.label} (legend title), \code{col.palette} (palette of colors, automatically expanded), \code{col.legend.cex} (legend scaling), and \code{n.legend} (approximate number of classes for numeric variables, or, maximum number of legend items per row for categorical variables). Currently, \code{plotSPC} will only generate two rows of legend items. Consider reducing the number of classes if two rows isn't enough room.

Profile sketches can be added according to relative positions along the x-axis (vs. integer sequence) via \code{relative.pos} argument. This should be a vector of positions within \{1,n\} that are used for horizontal placement. Default values are \code{1:length(x)}. Care must be taken when both \code{plot.order} and \code{relative.pos} are used simultaneously: \code{relative.pos} specifies horizontal placement after sorting. \code{\link[=addDiagnosticBracket]{addDiagnosticBracket()}} and \code{\link[=addVolumeFraction]{addVolumeFraction()}} use the \code{relative.pos} values for subsequent annotation.

Relative positions that are too close will result in overplotting of sketches. Adjustments to relative positions such that overlap is minimized can be performed with \code{fixOverlap(pos)}, where \code{pos} is the original vector of relative positions.

The \code{x.idx.offset} argument can be used to shift a collection of pedons from left to right in the figure. This can be useful when plotting several different \code{SoilProfileCollection} objects within the same figure. Space must be pre-allocated in the first plotting call, with an offset specified in the second call. See examples below.

Horizon depths (e.g. cm) are converted to figure y-coordinates via: y = (depth * scaling.factor) + y.offset.
}
\note{
A new plot of soil profiles is generated, or optionally added to an existing plot.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))


# example data
data(sp1)
# usually best to adjust margins
par(mar = c(0,0,3,0))

# add color vector
sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))

# promote to SoilProfileCollection
depths(sp1) <- id ~ top + bottom

# init horizon designation 
hzdesgnname(sp1) <- 'name'

# plot profiles
plotSPC(sp1, id.style = 'side')

# title, note line argument:
title('Sample Data 1', line = 1, cex.main = 0.75)

# plot profiles without horizon-line divisions
plotSPC(sp1, divide.hz = FALSE)

# diagonal lines encode horizon boundary distinctness
sp1$hzD <- hzDistinctnessCodeToOffset(sp1$bound_distinct)
plotSPC(sp1, hz.distinctness.offset = 'hzD', name.style = 'center-center')

# plot horizon color according to some property
data(sp4)
depths(sp4) <- id ~ top + bottom
hzdesgnname(sp4) <- 'name'
plotSPC(sp4, color = 'clay')

# another example
data(sp2)
depths(sp2) <- id ~ top + bottom
hzdesgnname(sp2) <- 'name'
site(sp2) <- ~ surface

# some of these profiles are very deep, truncate plot at 400cm
# label / re-order with site-level attribute: `surface`
plotSPC(sp2, label = 'surface', plot.order = order(sp2$surface), 
max.depth = 400)

# example using a categorical attribute
plotSPC(sp2, color = "plasticity", 
max.depth = 400)

# plot two SPC objects in the same figure
par(mar = c(1,1,1,1))

# plot the first SPC object and
# allocate space for the second SPC object
plotSPC(sp1, n = length(sp1) + length(sp2))

# plot the second SPC, starting from the first empty space
plotSPC(sp2, x.idx.offset = length(sp1), add = TRUE)


##
## demonstrate horizon designation shrinkage
##

data("jacobs2000")

# shrink "long" horizon names
plotSPC(
  jacobs2000, 
  name = 'name',
  name.style = 'center-center', 
  shrink = TRUE, 
  cex.names = 0.8
)

# shrink horizon names in "thin" horizons
plotSPC(
  jacobs2000, 
  name = 'name',
  name.style = 'center-center', 
  shrink = TRUE, 
  shrink.thin = 15,
  cex.names = 0.8,
)


##
## demonstrate adaptive legend
##

data(sp3)
depths(sp3) <- id ~ top + bottom

# make some fake categorical data
horizons(sp3)$fake.data <- sample(letters[1:15], size = nrow(sp3), replace=TRUE)

# better margins
par(mar=c(0,0,3,1))

# note that there are enough colors for 15 classes (vs. previous limit of 10)
# note that the legend is split into 2 rows when length(classes) > n.legend argument
plotSPC(sp3, color='fake.data', name='fake.data', cex.names=0.8)

# make enough room in a single legend row
plotSPC(sp3, color='fake.data', name='fake.data', cex.names=0.8, n.legend=15)


##
## demonstrate y.offset argument
## must be of length 1 or length(x)
##

# example data and local copy
data("jacobs2000")
x <- jacobs2000
hzdesgnname(x) <- 'name'

# y-axis offsets, simulating a elevation along a hillslope sequence
# same units as horizon depths in `x`
# same order as profiles in `x`
y.offset <- c(-5, -10, 22, 65, 35, 15, 12)

par(mar = c(0, 0, 2, 2))

# y-offset at 0
plotSPC(x, color = 'matrix_color', cex.names = 0.66)

# constant adjustment to y-offset
plotSPC(x, color = 'matrix_color', cex.names = 0.66, y.offset = 50)

# attempt using invalid y.offset
# warning issued and default value of '0' used
# plotSPC(x, color = 'matrix_color', cex.names = 0.66, y.offset = 1:2)

# variable y-offset
# fix overlapping horizon depth labels
par(mar = c(0, 0, 1, 0))
plotSPC(
  x, 
  y.offset = y.offset, 
  color = 'matrix_color', 
  cex.names = 0.75, 
  shrink = TRUE,
  hz.depths = TRUE, 
  hz.depths.offset = 0.05,
  fixLabelCollisions = TRUE,
  name.style = 'center-center'
)



# random y-axis offsets
yoff <- runif(n = length(x), min = 1, max = 100)

# random gradient of x-positions
xoff <- runif(n = length(x), min = 1, max = length(x))

# note profiles overlap
plotSPC(x, 
        relative.pos = xoff, 
        y.offset = yoff, 
        color = 'matrix_color', 
        cex.names = 0.66, 
        hz.depths = TRUE, 
        name.style = 'center-center'
)


# align / adjust relative x positions
set.seed(111)
pos <- alignTransect(xoff, x.min = 1, x.max = length(x), thresh = 0.65)

# y-offset is automatically re-ordered according to
# plot.order

par(mar = c(0.5, 0.5, 0.5, 0.5))
plotSPC(x, 
        plot.order = pos$order, 
        relative.pos = pos$relative.pos, 
        y.offset = yoff, 
        color = 'matrix_color', 
        cex.names = 0.66, 
        hz.depths = TRUE, 
        name.style = 'center-center'
)

box()

}
\references{
Beaudette, D.E., Roudier P., and A.T. O'Geen. 2013. Algorithms for Quantitative Pedology: A Toolkit for
Soil Scientists. Computers & Geosciences. 52:258 - 268.
}
\seealso{
\code{\link[=fixOverlap]{fixOverlap()}}, \code{\link[=explainPlotSPC]{explainPlotSPC()}}, \linkS4class{SoilProfileCollection}, \code{\link[=pretty]{pretty()}}, \code{\link[=hzDistinctnessCodeToOffset]{hzDistinctnessCodeToOffset()}}, \code{\link[=addBracket]{addBracket()}}, \code{\link[=profileGroupLabels]{profileGroupLabels()}}
}
\author{
D.E. Beaudette
}
\keyword{hplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R,
%   R/SoilProfileCollection-setters.R
\docType{methods}
\name{diagnostic_hz,SoilProfileCollection-method}
\alias{diagnostic_hz,SoilProfileCollection-method}
\alias{diagnostic_hz}
\alias{diagnostic_hz<-}
\alias{diagnostic_hz<-,SoilProfileCollection-method}
\title{Get or Set Diagnostic Horizon data in a SoilProfileCollection}
\usage{
\S4method{diagnostic_hz}{SoilProfileCollection}(object)

\S4method{diagnostic_hz}{SoilProfileCollection}(object) <- value
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{value}{An object inheriting from \code{data.frame}}
}
\description{
Diagnostic horizons describe features of the soil relevant to taxonomic classification. A single profile may have multiple diagnostic features or horizons, each of which may be comprised of multiple horizons.
\itemize{
\item \code{diagnostic_hz()} (get method): Get diagnostic feature data from a SoilProfileCollection.
}

\itemize{
\item \verb{diagnostic_hz<-} (set method): Set diagnostic feature data for a SoilProfileCollection. The profile ID column from \code{object} (\code{idname(object)}) must be present in the replacement \code{value} object.
}
}
\examples{

# load test data
data(sp2)

# promote to SPC
depths(sp2) <- id ~ top + bottom

# assign two profiles a zone related to the mollic epipedon
newdata <- data.frame(id = c("hon-1","hon-17"),
                      featkind = "fixed-depth surface sample",
                      featdept = 0,
                      featdepb = 18)

# do left join
diagnostic_hz(sp2) <- newdata

# inspect site table: newvalue TRUE only for horizons
#  with top depth equal to zero
diagnostic_hz(sp2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aqp-label-placement-solvers.R
\name{electroStatics_1D}
\alias{electroStatics_1D}
\title{Label placement based on a simulation of electrostatic forces}
\usage{
electroStatics_1D(
  x,
  thresh,
  q = 1,
  chargeDecayRate = 0.01,
  QkA_GrowthRate = 0.05,
  maxIter = 100,
  tiny = 1e-04,
  const = 0.001,
  trace = FALSE,
  ...
)
}
\arguments{
\item{x}{numeric vector, pre-sorted sorted, without duplication, describing 1D label (particle) configuration}

\item{thresh}{numeric, overlap threshold, same as in \code{\link[=fixOverlap]{fixOverlap()}}}

\item{q}{numeric, electrical charge (typically between 0.1 and 2)}

\item{chargeDecayRate}{numeric, exponential decay rate constant for \code{q} as a function of iteration \code{i}}

\item{QkA_GrowthRate}{numeric, growth rate constant for \code{Qk} applied to attraction to uniform spacing of labels, invoked when rank order is violated during the simulation}

\item{maxIter}{integer, maximum number of iterations before giving up}

\item{tiny}{numeric, 0-values replaced by this number to avoid division by 0 and infinite forces}

\item{const}{numeric, empirical constant added to the 1D electrostatic force equation to dampen oscillation: \code{(Qk * Q1 * Q2) / (d^ex + const)}}

\item{trace}{logical, include diagnostic output}

\item{\dots}{not used, absorbs additional arguments to \code{\link[=fixOverlap]{fixOverlap()}}}
}
\value{
When \code{trace = TRUE} a \code{list}, otherwise numeric vector with \code{converged} attribute.
}
\description{
This function attempts to move labels along a 1D coordinate system such that overlap (as specified by threshold) is minimized. An electrostatic simulation applies forces of repulsion between labels that are within \code{thresh} (e.g. overlapping) and forces of attraction to a uniformly spaced sequence to iteratively perturb affected labels until either no overlap is reported, or a maximum number of iterations (\code{maxIter}) has been reached.
}
\details{
Difficult overlap problems can be addressed by reducing \code{thresh} and increasing \code{q}. Large values of \code{q} can lead to chaotic results.

This function will generate unpredictable output when \code{x} contains duplicate values.

This function requires input to be pre-sorted, although interesting "artistic" simulations will often result from unsorted \code{x}.
}
\examples{

# vector of object locations, with potential overlap
x <- c(1, 2, 3, 3.3, 3.8, 5, 6, 7, 8, 9, 10)

# full diagnostic output
z <- electroStatics_1D(x, thresh = 0.65, trace = TRUE, q = 1)
txt <- sprintf("Converged \%s (\%s iterations)", z$converged, length(z$cost))

plot(
seq_along(z$cost),
z$cost, 
las = 1, 
xlab = 'Iteration', 
ylab = 'Overlap Cost', 
type = 'b', 
main = txt
)

abline(h = 0, lty = 2, col = 2)

# final configuration only
xnew <- electroStatics_1D(x, thresh = 0.65, q = 1)

# check for convergence
attr(xnew, 'converged')

# compare original vs. modified
data.frame(orig = x, new = round(xnew, 2))

}
\seealso{
\code{\link[=fixOverlap]{fixOverlap()}}, \code{\link[=SANN_1D]{SANN_1D()}}
}
\author{
D.E. Beaudette and K.C. Thompson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimateSoilDepth.R
\name{estimateSoilDepth}
\alias{estimateSoilDepth}
\title{Estimate Soil Depth}
\usage{
estimateSoilDepth(
  f,
  name = hzdesgnname(f),
  p = "Cr|R|Cd",
  selection = min,
  no.contact.depth = NULL,
  no.contact.assigned = NULL
)
}
\arguments{
\item{f}{SoilProfileCollection object of length 1, e.g. a single profile}

\item{name}{name of the column that contains horizon designations}

\item{p}{REGEX pattern for determining "contact", or depth to some morphologic feature (e.g. \code{Bt})}

\item{selection}{an R function applied in the presence of multiple matching horizons: \code{min} (default), \code{max}, \code{mean}, etc.}

\item{no.contact.depth}{in the absence of contact matching \code{p}, a depth at which we can assume a standard depth-to-contact}

\item{no.contact.assigned}{value assigned when no contact is encountered at or below \code{no.contact.depth}}
}
\value{
single value representing the depth to \code{contact} or \code{no.contact.assigned}
}
\description{
Estimate the soil depth of a single profile within a \code{SoilProfileCollection} object. This function would typically be called by \code{\link[=profileApply]{profileApply()}}.
}
\details{
The choice of a \code{selection} function usually follows:

\code{min}: the top of the first matching horizon, \code{max}: the top bot the last matching horizon,
or possibly \code{mean}: somewhere in-between.
}
\examples{


## consider a situation where there were multiple candidate
## "contacts": 2 Cd horizons over an R

# init hypothetical profile
d <- data.frame(
  id = '1',
  top = c(0, 10, 20, 30, 40, 50, 60),
  bottom = c(10, 20, 30, 40, 50, 60, 80),
  name = c('A', 'Bt1', 'Bt2', 'BC', 'Cd1', 'Cd2', 'R'),
  stringsAsFactors = FALSE
)

# upgrade to SPC
depths(d) <- id ~ top + bottom

# init horizon designation
hzdesgnname(d) <- 'name'

# visual check
par(mar = c(0, 0, 0, 2))
plotSPC(d, hz.depths = TRUE, name.style = 'center-center', cex.names = 1, width = 0.1)

# top of the first Cd
estimateSoilDepth(d, name = 'name')

# top of the first Cd
estimateSoilDepth(d, name = 'name', selection = min)

# top of the R
estimateSoilDepth(d, name = 'name', selection = max)

# top of the second Cd
estimateSoilDepth(d, name = 'name', selection = max, p = 'Cd')


## another example

data(sp1)
depths(sp1) <- id ~ top + bottom

# init horizon designation
hzdesgnname(d) <- 'name'

# apply to each profile in a collection, and save as site-level attribute
sp1$depth <- profileApply(sp1, estimateSoilDepth, name='name')

# this function can be used to "find" depth to any feature 
# that can be defined via REGEX pattern matching on the horizon name
# for example, locate the depth to the top "Bt" horizon
# returning NA when there is no match
sp1$top_Bt <- profileApply(
  sp1, estimateSoilDepth, 
  name='name', 
  p='Bt', 
  no.contact.depth=0, 
  no.contact.assigned=NA
)

# reduced margins
par(mar=c(1,1,1,2))
# adjust default y-offset and depth scaling for following examples
plotSPC(sp1, y.offset=10, scaling.factor=0.5)

# get plotting parameters for profile widths and depth scaling factors
lsp <- get("last_spc_plot", envir = aqp.env)

# positions on x-axis, same for both depth and top "Bt" horizon
x.positions <- (1:length(sp1)) - lsp$width

# annotate contact with unicode right-arrow
# y-position is adjusted based on plot y-offset and scaling factor
y.positions <- lsp$y.offset + (sp1$depth * lsp$scaling.factor)
text(x.positions, y.positions, '\u2192', col='red', adj=1, cex=1.25, lwd=2)

# annotate top "Bt" depth with unicode right-arrow
# y-position is adjusted based on plot y-offset and scaling factor
y.positions <- lsp$y.offset + (sp1$top_Bt * lsp$scaling.factor)
text(x.positions, y.positions, '\u2192', col='blue', adj=1, cex=1.25, lwd=2)


\dontrun{
  # sample data
  data(gopheridge, package='soilDB')
  
  # run on a single profile
  estimateSoilDepth(gopheridge[1, ], name = 'hzname')
  
  # apply to an entire collection
  profileApply(gopheridge, estimateSoilDepth, name = 'hzname')
}
}
\seealso{
\code{\link{getSoilDepthClass}}, \code{\link{profileApply}}
}
\author{
D.E. Beaudette and J.M. Skovlin
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/soilColorIndices.R
\name{buntley.westin.index}
\alias{buntley.westin.index}
\title{Buntley-Westin (1965) Index}
\usage{
buntley.westin.index(hue, chroma)
}
\arguments{
\item{hue}{A character vector containing Munsell hues (e.g. "7.5YR")}

\item{chroma}{A numeric vector containing Munsell chromas}
}
\value{
A numeric vector reflecting horizon color development.
}
\description{
Calculate "Color Development Equivalent" by the method of Buntley & Westin (1965) "A Comparative Study of Developmental Color in a Chestnut-Chernozem-Brunizem Soil Climosequence" DOI: 10.2136/sssaj1965.03615995002900050029x. Originally developed for Mollisols, the Buntley-Westin index has been used as a tool to separate soils based on depth to particular colors.
}
\references{
Buntley, G.J. and Westin, F.C. (1965), A Comparative Study of Developmental Color in a Chestnut-Chernozem-Brunizem Soil Climosequence. Soil Science Society of America Journal, 29: 579-582. doi:10.2136/sssaj1965.03615995002900050029x
}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allocate.R
\name{allocate}
\alias{allocate}
\title{Allocate soil properties within various classification systems.}
\usage{
allocate(
  ...,
  to = c("FAO Salt Severity", "FAO Black Soil", "ST Diagnostic Features"),
  droplevels = FALSE
)
}
\arguments{
\item{...}{arguments to specific allocation functions, see details and examples}

\item{to}{character specifying the classification scheme: FAO Salt Severity, FAO Black Soil (see details for the required \code{...})}

\item{droplevels}{logical indicating whether to drop unused levels in factors. This is useful when the results have a large number of unused classes, which can waste space in tables and figures.}
}
\value{
A vector or \code{data.frame} object.
}
\description{
Generic function to allocate soil properties to different classification schemes.
}
\details{
This function is intended to allocate a set of soil properties to an established soil classification scheme, such as Salt Severity or Black Soil. Allocation is semantically different from classification. While classification is the 'act' of developing a grouping scheme, allocation is the assignment or identification of measurements to a established class (Powell, 2008).
\subsection{Usage Details}{

Each classification scheme (\code{to} argument) uses a different set of arguments.
\itemize{
\item \verb{FAO Salt Severity}
\itemize{
\item \strong{EC:} electrical conductivity column name, dS/m
\item \strong{pH:} pH column name, saturated paste extract
\item \strong{ESP:} exchangeable sodium percentage column name, percent
}
\item \verb{FAO Black Soils}
\itemize{
\item \strong{object:} a \code{data.frame} or \code{SoilProfileCollection}
\item \strong{pedonid:} pedon ID column name, required when \code{object} is a \code{data.frame}
\item \strong{hztop:} horizon top depth column name, required when \code{object} is a \code{data.frame}
\item \strong{hzbot:} horizon bottom depth column name, required when \code{object} is a \code{data.frame}
\item \strong{OC}: organic carbon column name, percent
\item \strong{m_chroma:} moist Munsell chroma column name
\item \strong{m_value:} moist Munsell value column name
\item \strong{d_value:} dry Munsell value column name
\item \strong{CEC:} cation exchange capacity column name (NH4OAc at pH 7), units of cmol(+)/kg soil
\item \strong{BS:} base saturation column name (NH4OAc at pH 7), percent
\item \strong{tropical:} logical, data are associated with "tropical soils"
}
\item \verb{ST Diagnostic Features}
\itemize{
\item \strong{object:} a \code{data.frame} or \code{SoilProfileCollection}
\item \strong{pedonid:} pedon ID column name, required when \code{object} is a \code{data.frame}
\item \strong{hzname:} horizon name column, required when \code{object} is a \code{data.frame}
\item \strong{hztop:} horizon top depth column name, required when \code{object} is a \code{data.frame}
\item \strong{hzbot:} horizon bottom depth column name, required when \code{object} is a \code{data.frame}
\item \strong{texcl:} soil texture class (USDA) column name
\item \strong{rupresblkcem:} rupture resistance column name
\item \strong{m_value:} moist Munsell value column name
\item \strong{m_chroma:} moist Munsell chroma column name
\item \strong{d_value:} dry Munsell value column name
\item \strong{BS:} base saturation column name (method ??), percent
\item \strong{OC}: organic carbon column name, percent
\item \strong{n_value:} ??
\item \strong{featkind:} ??
}
}
}
}
\note{
The results returned by \code{allocate(to = "ST Diagnostic Features")} currently return a limited set of diagnostic features that are easily defined. Also, the logic implemented for some features does not include all the criteria defined in the Keys to Soil Taxonomy.
}
\examples{

# Salt Severity
test <- expand.grid(
  EC  = sort(sapply(c(0, 0.75, 2, 4, 8, 15, 30), function(x) x + c(0, -0.05, 0.05))),
  pH  = c(8.1, 8.2, 8.3, 8.4, 8.5, 8.6),
  ESP = sort(sapply(c(0, 15, 30, 50, 70, 100), function(x) x + c(0, 0.1, -0.1)))
)
test$ss      <- with(test, allocate(EC = EC, pH = pH, ESP = ESP, to = "FAO Salt Severity"))
table(test$ss)

# Black Soil Category 1 (BS1)
test <- expand.grid(
  dept = seq(0, 50, 10),
  OC   = sort(sapply(c(0, 0.6, 1.2, 20, 40), function(x) x + c(0, -0.05, 0.05))),
  chroma_moist  = 2:4,
  value_moist   = 2:4,
  value_dry     = 4:6,
  thickness     = 24:26,
  CEC           = 24:26,
  BS            = 49:51,
  tropical      = c(TRUE, FALSE)
)
test$pedon_id <- rep(1:21870, each = 6)
test$depb     <- test$dept + 10

bs1 <- allocate(test, pedonid = "pedon_id", hztop = "dept", hzbot = "depb", 
                OC = "OC", m_chroma = "chroma_moist", m_value = "value_moist", 
                d_value = "value_dry", CEC = "CEC", BS = "BS", 
                to = "FAO Black Soil"
)

table(BS1 = bs1$BS1, BS2 = bs1$BS2)


# SoilProfileCollection interface

data(sp3)
depths(sp3) <- id ~ top + bottom
hzdesgnname(sp3) <- 'name'

# fake base saturation
horizons(sp3)$bs <- 75

plotSPC(sp3)

allocate(
  sp3, 
  to = 'FAO Black Soil', 
  OC = 'tc', 
  m_chroma = 'chroma', 
  m_value = 'value', 
  d_value = 'value',
  CEC = 'cec',
  BS = 'bs'
)

# make a copy and edit horizon values
x <- sp3
x$value <- 2
x$chroma <- 2
x$cec <- 26
x$tc <- 2

x$soil_color <- munsell2rgb(x$hue, x$value, x$chroma)

plotSPC(x)

allocate(
  x, 
  to = 'FAO Black Soil', 
  OC = 'tc', 
  m_chroma = 'chroma', 
  m_value = 'value', 
  d_value = 'value',
  CEC = 'cec',
  BS = 'bs'
)


# Soil Taxonomy Diagnostic Features
data(sp1)
sp1$texcl = gsub("gr|grv|cbv", "", sp1$texture)
df <- allocate(object = sp1, pedonid = "id", hzname = "name", 
               hzdept = "top", hzdepb = "bottom", texcl = "texcl", 
               to = "ST Diagnostic Features"
)
aggregate(featdept ~ id, data = df, summary)

}
\references{
Abrol, I., Yadav, J. & Massoud, F. 1988. \href{https://www.fao.org/4/x5871e/x5871e00.htm}{Salt-affected soils and their management}. No. Bulletin 39. Rome, FAO Soils.

FAO. 2006. \href{https://www.fao.org/4/a0541e/a0541e.pdf}{Guidelines for soil description}. Rome, Food and Agriculture Organization of the United Nations.

FAO. 2020. DEFINITION | What is a black soil? (online). (Cited 28 December 2020). http://www.fao.org/global-soil-partnership/intergovernmental-technical-panel-soils/gsoc17-implementation/internationalnetworkblacksoils/more-on-black-soils/definition-what-is-a-black-soil/es/

Powell, B., 2008. Classifying soil and land, in: McKenzie, N.J., Grundy, M.J., Webster, R., Ringrose-Voase, A.J. (Eds.), Guidelines for Survey Soil and Land Resources, Australian Soil and Land Survey Handbook Series. CSIRO, Melbourne, p. 572.

Richards, L.A. 1954. \href{https://www.ars.usda.gov/ARSUserFiles/20360500/hb60_pdf/hb60complete.pdf}{Diagnosis and Improvement of Saline and Alkali Soils}. U. S. Government Printing Office. 166 pp.

Soil Survey Staff, 2014. Keys to Soil Taxonomy, 12th ed. USDA-Natural Resources Conservation Service, Washington, D.C.
}
\author{
Stephen Roecker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{soiltexture}
\alias{soiltexture}
\title{Lookup tables for sand, silt, clay, texture class, and textural modifiers.}
\format{
A list with 3 data frames. The first named values which contains
values for sand, silt and clay by texture class. The second with average
values for sand, silt and clay per texture class. The third has fragvoltot
low, rv and high values for texmod.  \describe{ \item{list("clay")}{clay
percentage of the fine earth fraction, a integer vector}
\item{list("sand")}{sand percentage of the fine earth fraction, a integer
vector} \item{list("silt")}{silt percentage of the fine earth fraction, a
integer vector} \item{list("texcl")}{texture class, a character vector}
\item{list("texmod")}{textural modifiers, a character vector} }
}
\usage{
data(soiltexture)
}
\description{
A list that contains a snapshot of the values generated using the logic from
the particle size estimator calculation in NASIS, the average values per
texture class, and average rock fragment values by textural modifier.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huePositionCircle.R
\name{huePositionCircle}
\alias{huePositionCircle}
\title{Visual Description of Munsell Hue Ordering}
\usage{
huePositionCircle(
  hues = huePosition(returnHues = TRUE),
  value = 6,
  chroma = 10,
  chip.cex = 5.5,
  label.cex = 0.66,
  seg.adj = 0.8,
  seg.col = grey(0.4),
  plot = TRUE,
  simulateCVD = NULL,
  CVDseverity = 1
)
}
\arguments{
\item{hues}{vector of Munsell hues, commonly derived from \code{huePosition()}}

\item{value}{single integer, Munsell value used to create an actual color}

\item{chroma}{single integer, Munsell chroma used to create an actual color}

\item{chip.cex}{numeric, scaling for color chips}

\item{label.cex}{numeric, scaling labels}

\item{seg.adj}{numeric, scaling for line segment cues}

\item{seg.col}{single color, color used for line segment cues}

\item{plot}{logical, generate output on the current graphics device}

\item{simulateCVD}{simulate color vision deficiencies with the colorspace package, should be the character representation of a function name, one of: 'deutan', 'protan', or 'tritan'.}

\item{CVDseverity}{numeric value between 0 (none) and 1 (total), describing the severity of the color vision deficiency}
}
\value{
an invisible \code{data.frame} of data used to create the figure
}
\description{
Munsell hues are arranged on the unit circle with "neutral" at the center.
}
\note{
The best results are obtained when setting margins to zero, and inverting foreground / background colors. For example: \code{par(mar = c(0, 0, 0, 0), fg = 'white', bg = 'black')}.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))


# better graphics defaults
op <- par(
mar = c(0, 0, 0, 0), 
fg = 'white', 
bg = 'black', 
xpd = NA
)

# full set of hues, as generated by huePosition(returnHues = TRUE)
huePositionCircle()

# just a few hues
huePositionCircle(hues = c('5R', '5Y', '5G', '5B', '5P'))

# adjust Munsell value and chroma
huePositionCircle(value = 3, chroma = 6)

# reset graphics state
par(op)

}
\references{
Munsell book of color. 1976. Macbeth, a Division of Kollmorgen Corp., Baltimore, MD.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\docType{methods}
\name{min,SoilProfileCollection-method}
\alias{min,SoilProfileCollection-method}
\alias{min}
\title{Get the minimum bottom depth in a SoilProfileCollection}
\usage{
\S4method{min}{SoilProfileCollection}(x, v = NULL, na.rm = TRUE)
}
\arguments{
\item{x}{a SoilProfileCollection}

\item{v}{optional: a vector of horizon attribute names to refine calculation}

\item{na.rm}{remove \code{NA}? default: \code{TRUE}}
}
\description{
Get the shallowest depth of description out of all profiles in a SoilProfileCollection. Data missing one or more of: bottom depth, profile ID, or any optional attribute are omitted using \code{complete.cases}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorQuantiles.R
\name{plotColorQuantiles}
\alias{plotColorQuantiles}
\title{Visualize Color Quantiles}
\usage{
plotColorQuantiles(res, pt.cex = 7, lab.cex = 0.66)
}
\arguments{
\item{res}{list returned by \code{colorQuantiles}}

\item{pt.cex}{scaling factor for color chips}

\item{lab.cex}{chip label scaling factor}
}
\value{
a \code{lattice} graphics object
}
\description{
This function creates a visualization of the output from \code{colorQuantiles} using lattice graphics.
}
\details{
Marginal percentiles and L1 median CIELAB values from \code{colorQuantiles()} are combined into a single plot, arranged in panels according to L, A, and B coordinates. Munsell "chips" (colors and labels) are based on the closest Munsell color found via \code{\link[=col2Munsell]{col2Munsell()}}.
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{hzidname<-}
\alias{hzidname<-}
\alias{hzidname<-,SoilProfileCollection-method}
\alias{hzidname,SoilProfileCollection-method}
\alias{hzidname}
\title{Set horizon ID column name}
\usage{
\S4method{hzidname}{SoilProfileCollection}(object) <- value

\S4method{hzidname}{SoilProfileCollection}(object)
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{value}{character, column name containing unique horizon ID values}
}
\description{
Set unique horizon ID column name

Get column name containing unique horizon ID
}
\examples{
data(sp1)

# promote to SPC
depths(sp1) <- id ~ top + bottom

# create new horizon ID
sp1$hzIDrev <- rev(sp1$hzID)

# set horizon designation column
hzidname(sp1) <- "hzIDrev"

# get horizon designation column
hzidname(sp1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/textureTriangleSummary.R
\name{textureTriangleSummary}
\alias{textureTriangleSummary}
\title{Soil Texture Low-RV-High as Defined by Quantiles}
\usage{
textureTriangleSummary(
  ssc,
  p = c(0.05, 0.5, 0.95),
  delta = 1,
  rv.col = "red",
  range.border = "black",
  range.col = "RoyalBlue",
  range.alpha = 80,
  range.lty = 1,
  range.lwd = 2,
  main = "Soil Textures",
  legend.cex = 0.75,
  legend = TRUE,
  ...
)
}
\arguments{
\item{ssc}{\code{data.frame} with columns: 'SAND', 'SILT', 'CLAY', values are percentages that should add to 100. No NA allowed.}

\item{p}{vector of percentiles (length = 3) defining 'low', 'representative value', and 'high'}

\item{delta}{grid size used to form low-high region}

\item{rv.col}{color used for representative value (RV) symbol}

\item{range.border}{color used for polygon border enclosing the low-high region}

\item{range.col}{color used for polygon enclosing the low-high region}

\item{range.alpha}{transparency of the low-high range polygon (0-255)}

\item{range.lty}{line style for polygon enclosing the low-high region}

\item{range.lwd}{line weight polygon enclosing the low-high region}

\item{main}{plot title}

\item{legend.cex}{scaling factor for legend}

\item{legend}{logical, enable/disable automatic legend}

\item{\dots}{further arguments passed to \code{soiltexture::TT.points}}
}
\value{
an invisible \code{matrix} with marginal percentiles of sand, silt, and clay
}
\description{
This function accepts soil texture components (sand, silt, and clay percentages) and plots a soil texture triangle with a "representative value" (point) and low-high region (polygon) defined by quantiles (estimated with \code{Hmisc::hdquantile}). Marginal quantiles of sand, silt, and clay are used to define the boundary of a low-high region. The default settings place the RV symbol at the texture defined by marginal medians of sand, silt, and clay. The default low-high region is defined by the 5th and 95th marginal percentiles of sand, silt, and clay.
}
\examples{

\donttest{

if(
requireNamespace("Hmisc") &
  requireNamespace("compositions") &
  requireNamespace("soiltexture")
) {
  
  # sample data
  data('sp4')
  
  # subset rows / columns
  ssc <- sp4[grep('^Bt', sp4$name), c('sand', 'silt', 'clay')]
  names(ssc) <- toupper(names(ssc))
  
  # make figure, marginal percentiles are silently returned
  stats <- textureTriangleSummary(
    ssc, pch = 1, cex = 0.5, 
    range.alpha = 50, 
    range.lwd = 1,
    col = grey(0.5), 
    legend = FALSE
  )
  
  # check
  stats
  
  # simulate some data and try again
  s <- bootstrapSoilTexture(ssc, n = 100)$samples
  
  # make the figure, ignore results
  textureTriangleSummary(
    s, pch = 1, cex = 0.5, 
    range.alpha = 50, 
    range.lwd = 1,
    col = grey(0.5), 
    legend = FALSE
  )
}

}

}
\seealso{
\code{\link{bootstrapSoilTexture}}
}
\author{
D.E. Beaudette, J. Nemecek, K. Godsey
}
\keyword{hplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huePosition.R
\name{huePosition}
\alias{huePosition}
\title{Munsell Hue Reference and Position Searching}
\usage{
huePosition(
  x,
  returnHues = FALSE,
  includeNeutral = FALSE,
  origin = "5R",
  direction = c("cw", "ccw")
)
}
\arguments{
\item{x}{character vector of hues, e.g. c('10YR', '5YR'), optional if \code{returnHues = TRUE}}

\item{returnHues}{logical, should the full set of Munsell hues be returned? See details.}

\item{includeNeutral}{logical, add 'N' to the end of the full set of Munsell hues}

\item{origin}{hue to be used as the starting point for position searches (position 1)}

\item{direction}{indexing direction, should be \code{cw} (clock wise) or \code{ccw} (counter-clock wise)}
}
\value{
A vector of integer hue positions is returned, of the same
length and order as \code{x}. If \code{returnHues = TRUE}, then all hue
names and ordering are returned and \code{x} is ignored.
}
\description{
The 40 Munsell hues are typically arranged from 5R to 2.5R moving clock wise on the unit circle. This function matches a vector of hues to positions on that circle, with options for setting a custom origin or search direction.

This function is fully vectorized.
}
\examples{

# get hue ordering for setting levels of a factor
huePosition(returnHues = TRUE)

# get hue ordering including N (neutral)
huePosition(returnHues = TRUE, includeNeutral = TRUE)

# get position of the '10YR' hue, relative to standard origin of '5R'
# should be 7
huePosition(x = '10YR')

# get position of the '10YR' hue, relative to standard origin of '5YR'
# should be 3
huePosition(x = '10YR', origin = '5YR')

# visualize
op <- par(mar = c(0, 0, 0, 0), fg = 'white', bg = 'black')

huePositionCircle(huePosition(returnHues = TRUE, origin = '5YR'))

par(op)

}
\references{
\itemize{
\item Soil Survey Technical Note 2 \href{https://web.archive.org/web/20220704214918/https://www.nrcs.usda.gov/wps/portal/nrcs/detail/soils/ref/?cid=nrcs142p2_053569}{wayback machine URL}
\item Munsell book of color. 1976. Macbeth, a Division of Kollmorgen Corp., Baltimore, MD.
}
}
\seealso{
\link{colorContrast}, \link{huePositionCircle}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addBracket.R
\name{addBracket}
\alias{addBracket}
\title{Add Depth Brackets}
\usage{
addBracket(
  x,
  labcol = NULL,
  agg = FALSE,
  hzDepths = NULL,
  label.cex = 0.75,
  tick.length = 0.05,
  arrow.length = 0.05,
  offset = -0.3,
  missing.bottom.depth = NULL,
  ...
)
}
\arguments{
\item{x}{\code{data.frame} containing at least: \code{idname(x)} (profile ID) and \code{horizonDepths(x)} (horizon top and bottom depths)}

\item{labcol}{character, optional name of a column in \code{x} used for labeling}

\item{agg}{logical, aggregate multiple brackets per profile into a single depth range?}

\item{hzDepths}{character vector of length 2, optional column names in \code{x} that define bracket top and bottom depths. When \code{NULL}, use horizon top/bottom column names from the \code{SoilProfileCollection} object used by the last call to \code{plotSPC()}.}

\item{label.cex}{numeric, scaling factor for label font}

\item{tick.length}{numeric, length of bracket "tick" mark}

\item{arrow.length}{numeric, length of arrowhead (see \code{arrows()})}

\item{offset}{numeric, left-hand offset from each profile}

\item{missing.bottom.depth}{numeric, distance (in depth units) to extend brackets that are missing a lower depth (defaults to max depth of collection)}

\item{...}{further arguments passed on to \code{segments()} or \code{arrows()}}
}
\description{
Add depth brackets to soil profile sketches.
}
\details{
When \code{x} contains multiple records per profile a bracket will be created for each record. Setting \code{agg = TRUE} will first aggregate all records per profile, then add a single bracket spanning the depth range of those records. Additional examples can be found in \href{http://ncss-tech.github.io/AQP/aqp/SPC-plotting-ideas.html}{this tutorial}.
}
\note{
This is a \code{low-level} plotting function: you must first plot a \code{SoilProfileCollection} object before using this function. Details about the last plotted \code{SoilProfileCollection} are available using \code{get('last_spc_plot', envir = aqp.env)}.
}
\examples{

# example data
x <- c(
  'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
  'P2:Ap|AA|E|BhsBhs|Bw1Bw1|CCCCC',
  'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR',
  'P4:AA|EEE|BhsBhsBhsBhs|BwBw|CCCCC',
  'P5:AAAA|ACACACACAC|CCCCCCC|CdCdCd'
)

s <- quickSPC(x)

# change horizon depth names
# ensure that plotSPC() -> addBracket() can find them
horizonDepths(s) <- c('tt', 'bb')

# expression defines a single reference horizon in most profiles
.ex <- grepl('Bt3|Bw', s$name)
# encode for thematic profile sketches
s$e <- factor(as.character(.ex), levels = c('FALSE', 'TRUE'), labels = c('Horizons', 'Reference'))

# get horizon row indices to horizons above reference
a <- hzAbove(s, .ex, SPC = FALSE, simplify = TRUE)

# create bracket data.frame
b <- depths(s, hzID = FALSE)[a, ]

# add labels
b$label <- c('S')

op <- par(no.readonly = TRUE)
par(mar = c(0, 0, 3, 2), mfcol = c(1, 2))

# sketches
plotSPC(
  s, color = 'e', col.label = 'Original', col.palette = c('grey', 'royalblue'), 
  name = 'name', name.style = 'center-center', cex.names = 0.75,
  max.depth = 180
)

# plot individual brackets, no labels
addBracket(
  b, 
  agg = FALSE, labcol = 'label',
  offset = -0.35, col = 'black', tick.length = 0.04, lwd = 1
)

# sketches
plotSPC(
  s, color = 'e', col.label = 'Aggregate', col.palette = c('grey', 'royalblue'), 
  name = 'name', name.style = 'center-center', cex.names = 0.75,
  max.depth = 180
)

# aggregate multiple brackets into single depth span
# include first label from each group
addBracket(
  b, 
  agg = TRUE, labcol = 'label',
  offset = -0.35, col = 'firebrick', tick.length = 0.04, lwd = 2
) 

par(op)

}
\seealso{
\code{\link[=addDiagnosticBracket]{addDiagnosticBracket()}}, \code{\link[=plotSPC]{plotSPC()}}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorContrastPlot.R
\name{colorContrastPlot}
\alias{colorContrastPlot}
\title{Color Contrast Plot}
\usage{
colorContrastPlot(
  m1,
  m2,
  col.cex = 1,
  col.font = 2,
  d.cex = 1,
  cc.font = 3,
  dE00.font = 1,
  labels = c("m1", "m2"),
  label.cex = 1,
  label.font = 1,
  printMetrics = TRUE,
  ...
)
}
\arguments{
\item{m1}{first set of Munsell colors for comparison (e.g. '5YR 3/2')}

\item{m2}{second set of Munsell colors for comparison}

\item{col.cex}{scaling factor for color labels}

\item{col.font}{font for color labels}

\item{d.cex}{contrast for contrast metric labels}

\item{cc.font}{font for contrast class}

\item{dE00.font}{font for delta-E00}

\item{labels}{labels for compared colors, vector length 2}

\item{label.cex}{scaling factor for labels}

\item{label.font}{font for labels}

\item{printMetrics}{logical, print metrics between color swatches}

\item{\dots}{further arguments to \code{colorspace::swatchplot}}
}
\description{
A simple display of two sets of colors, NCSS color contrast class and CIE
delta-E00.
}
\note{
This function requires the \code{farver} package for calculation of CIE
delta-E00.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# two sets of colors to compare
m1 <- c('10YR 6/3', '7.5YR 3/3', '10YR 2/2', '7.5YR 3/4')
m2 <- c('5YR 3/4', '7.5YR 4/4', '2.5YR 2/2', '7.5YR 6/3')

# contrast metrics
colorContrast(m1, m2)

# graphical display
colorContrastPlot(m1, m2)

}
\seealso{
\code{\link[=colorContrast]{colorContrast()}}
}
\author{
D.E. Beaudette
}
\keyword{hplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/previewColors.R
\name{previewColors}
\alias{previewColors}
\title{Preview Colors}
\usage{
previewColors(
  cols,
  method = c("grid", "MDS", "manual"),
  labels = NULL,
  labels.cex = 1,
  col.order = NULL,
  nrow = ceiling(sqrt(length(cols))),
  ncol = nrow,
  border.col = "black",
  pt.cex = 2,
  pt.pch = 15
)
}
\arguments{
\item{cols}{vector of R colors}

\item{method}{either "grid", "MDS", or "manual", see details}

\item{labels}{optional vector of labels, disabled when \code{length(cols) > 5000}}

\item{labels.cex}{scaling factor for labels}

\item{col.order}{integer vector used to order colors}

\item{nrow}{number of rows used by "grid" method}

\item{ncol}{number of columns used by "grid" method}

\item{border.col}{border color used by "grid" method}

\item{pt.cex}{point scaling factor used by "MDS" method}

\item{pt.pch}{point symbol used by "MDS" method}
}
\value{
When \code{method} = "grid" or "manual", a vector of color order is returned. When \code{method = "MDS"}, the output from \code{\link[stats:cmdscale]{stats::cmdscale()}}.
}
\description{
Preview colors arranged according to CIE2000 distances or manual specification.
}
\details{
Color sorting is based on CIE2000 distances as calculated by \code{farver::compare_colour()}. The "grid" method arranges colors in a rectangular grid with ordering based on divisive hierarchical clustering of the pair-wise distances. Unique colors are used when \code{cols} contains more than 5,000 colors.

The "MDS" method arranges unique colors via classical multidimensional scaling (principal coordinates) via \code{cmdscale()}.

Colors can be manually arranged by supplying a vector of integers to \code{col.order} and setting \code{method='manual'}.
}
\examples{

# example data
data(sp2)

# convert into SoilProfileCollection object
depths(sp2) <- id ~ top + bottom

previewColors(sp2$soil_color)
previewColors(sp2$soil_color, method = 'MDS', pt.cex = 3)

# create colors using HCL space
cols.hcl <- hcl(h = 0:360, c = 100, l = 50)

# grid, colors sorted by dE00
previewColors(cols.hcl)

# manual specification
previewColors(cols.hcl, method = 'manual', col.order = 1:361)

# MDS
previewColors(cols.hcl, method = 'MDS', pt.cex = 1)

}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-spatial.R
\name{prj,SoilProfileCollection-method}
\alias{prj,SoilProfileCollection-method}
\alias{prj}
\alias{proj4string}
\alias{prj<-,SoilProfileCollection-method}
\alias{prj<-}
\alias{proj4string<-}
\alias{proj4string,SoilProfileCollection-method}
\alias{proj4string<-,SoilProfileCollection-method}
\title{Get or Set Coordinate Reference System for SoilProfileCollection}
\usage{
\S4method{prj}{SoilProfileCollection}(object, ...)

\S4method{prj}{SoilProfileCollection}(object, ...) <- value

\S4method{proj4string}{SoilProfileCollection}(obj)

\S4method{proj4string}{SoilProfileCollection}(obj) <- value
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{...}{Additional arguments (not used)}

\item{value}{character. Representation of Coordinate Reference System as \code{"authority:code"}, integer EPSG code, WKT2019 / PROJ4 string, an sf \code{crs} or sp \code{CRS} object.}

\item{obj}{A SoilProfileCollection}
}
\description{
\code{prj()}: Get Coordinate Reference System (Projection) metadata

\verb{prj()<-}: Set Coordinate Reference System metadata for the SoilProfileCollection

\code{proj4string()}: (Deprecated) Get Coordinate Reference System as PROJ4 String

\verb{proj4string()<-}: (Deprecated) Set Coordinate Reference System metadata for the SoilProfileCollection
}
\seealso{
\code{\link[=initSpatial<-]{initSpatial<-()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lunique.R
\name{lunique}
\alias{lunique}
\title{Eliminate duplicate instances of profile IDs in a list of SoilProfileCollections}
\usage{
lunique(l)
}
\arguments{
\item{l}{A list of SoilProfileCollections.}
}
\value{
A list of SoilProfileCollections, with duplicate profile IDs removed.
}
\description{
@description Experimental function to "clean" list input where duplicates exist (that would otherwise prevent \code{pbindlist}). Useful for queries that may have overlapping instances of the same data, for instance a list of SoilProfileCollections where each list element contains profiles gathered from a set of (potentially overlapping) extents.
}
\examples{

data(sp5)

# EXAMPLE #1 -- resolving overlap

# 6 profiles in four sets, and 5,6,7 are missing
input <- lapply(list(c(1,3,4), c(2,2,3), NA, c(8,9,1)), function(idx) {
      if(!all(is.na(idx)))
       sp5[idx,]
})

output <- lunique(input)

# 6 profiles are in final SPC; 5,6,7 are missing
match(profile_id(pbindlist(output)), profile_id(sp5))

# EXAMPLE #2 -- exact duplicates

# deliberately duplicate an SPC
sp5_2 <- sp5
res <- lunique(list(sp5, sp5_2))

# the number of profiles in first element is equal to number in sp5
length(res[[1]]) == length(sp5)

# second list element contains NA b/c all uniques are in #1
res[[2]]

}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{wilson2022}
\alias{wilson2022}
\title{Example Data from Wilson et al. 2022}
\format{
A \code{SoilProfileCollection} with the following elements. Total elemental analysis by lithium borate fusion.

Horizon level attributes:
\itemize{
\item name: horizon designation
\item Al2O3: total Al (wt \%)
\item Fe2O3: total Fe (wt \%)
\item K2O: total K (wt \%)
\item MgO: total Mg (wt \%)
\item Na2O: total Na (wt \%)
\item P2O5: total P (wt \%)
\item SiO2: total Si (wt \%)
\item CaO: total Ca(wt \%)
\item Alo: Oxalate Extractable Al (g/kg)
\item Feo: Oxalate Extractable Fe (g/kg)
\item Fed: Dithionite extractable Fe (g/kg)
\item Fed_minus_Feo: Crystalline Fe (hydr)oxides (g/kg)
\item CIA: Chemical Index of Alteration, see original paper (ratio, unitless)
\item Fed_div_Fet: (ratio, unitless)
\item Fet: Total Fe from lithium borate fusion (g/kg)
\item resin_Pi: Hedley phosphorus fractions (mg/kg)
\item NaHCO3_Pi: Hedley phosphorus fractions (mg/kg)
\item labile_Pi: Sum of resin Pi and NaHCO3 Pi (mg/kg)
\item NaCO3_Po: Hedley phosphorus fractions (mg/kg)
\item NaOH_Pi: Hedley phosphorus fractions (mg/kg)
\item NaOH_Po: Hedley phosphorus fractions (mg/kg)
\item Ca_Pi: Hedley phosphorus fractions (mg/kg)
\item organic_P: Sum of NaHCO3 and NaOH Po fractions (mg/kg)
\item total_P: Total P from lithium borate fusion (mg/kg)
\item occluded_P: Difference between total P and sum of Hedley P fractions (mg/kg)
\item top: horizon top depth (cm)
\item bottom: horizon bottom depth (cm)
\item pedonID: pedon ID (serial number)
}

Site level attributes:
\itemize{
\item pm: parent material group
\item biome: biome
}
}
\usage{
data(wilson2022)
}
\description{
An example \code{SoilProfileCollection}, derived from Wilson et al., 2022. Select data extracted from Appendix tables.
}
\examples{

data(wilson2022)

groupedProfilePlot(wilson2022, groups = 'pm', 
group.name.offset = -15, label = 'biome', 
name.style = 'center-center', color = 'CIA', 
cex.names = 0.66, cex.id = 0.66, width = 0.3)

groupedProfilePlot(wilson2022, groups = 'biome', 
group.name.offset = -15, label = 'pm', 
name.style = 'center-center', color = 'Fet', 
cex.names = 0.66, cex.id = 0.66, width = 0.3)

}
\references{
Stewart G. Wilson, Randy A. Dahlgren, Andrew J. Margenot, Craig Rasmussen, Anthony T. O'Geen. 2022. Expanding the Paradigm: The influence of climate and lithology on soil phosphorus, Geoderma: 421. \doi{10.1016/j.geoderma.2022.115809}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/soilPalette.R
\name{soilPalette}
\alias{soilPalette}
\title{Soil Color Palette}
\usage{
soilPalette(
  colors,
  lab = colors,
  lab.cex = 0.75,
  dynamic.labels = TRUE,
  x.inset = 0.01,
  y.inset = 0.01,
  ...
)
}
\arguments{
\item{colors}{vector of hex colors (e.g. #A66E46FF)}

\item{lab}{vector of labels}

\item{lab.cex}{character scaling for labels}

\item{dynamic.labels}{logical, adjust label colors for maximum contrast via \code{invertLabelColor}}

\item{x.inset}{horizontal adjustment for labels}

\item{y.inset}{vertical adjustment for labels}

\item{\dots}{further arguments to \code{colorspace::swatchplot}}
}
\value{
nothing is returned, function called for graphic output
}
\description{
A very simple function for generating labeled swatches of soil colors. Largely based on \code{colorspace::swatchplot}.
}
\note{
The result is a simple figure on the active plotting device.
}
\examples{

# maybe useful for teaching about soil color

par(mfrow=c(2,1), mar=c(1,1,1,1))

# demonstrate range of Munsell value
m <- sprintf('10YR \%s/4', 2:8)
# convert to hex representation
cols <- parseMunsell(m)
# plot
soilPalette(cols, m)

# demonstrate range of Munsell chroma
m <- sprintf('10YR 4/\%s', 2:8)
# convert to hex representation
cols <- parseMunsell(m)
# plot
soilPalette(cols, m)

}
\author{
D.E. Beaudette
}
\keyword{hplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{horizonDepths<-}
\alias{horizonDepths<-}
\alias{horizonDepths<-,SoilProfileCollection-method}
\alias{horizonDepths,SoilProfileCollection-method}
\alias{horizonDepths}
\title{Set horizon depth column names}
\usage{
\S4method{horizonDepths}{SoilProfileCollection}(object) <- value

\S4method{horizonDepths}{SoilProfileCollection}(object)
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{value}{a character vector of length two with names of columns containing numeric top and bottom depths}
}
\description{
Set column name containing horizon ID

Get column names containing horizon depths
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color-palettes.R
\name{soilTextureColorPal}
\alias{soilTextureColorPal}
\title{Soil Texture Color Palettes}
\usage{
soilTextureColorPal(simplify = FALSE, schema = "soilweb")
}
\arguments{
\item{simplify}{logical, return the base 12 (\code{TRUE}) or full 21 (\code{FALSE}) soil texture classes}

\item{schema}{select mapping between soil texture classes, and colors, currently limited to 'soilweb'}
}
\value{
\code{data.frame} from soil texture class codes and colors
}
\description{
Suggested color palettes for USDA soil texture classes, ranked according to average plant-available water holding capacity. The default color mapping schema is based on a palette used by SoilWeb applications.
}
\examples{
 
# base 12 soil texture classes
# ranked by plant available water-holding capacity
d <- soilTextureColorPal(simplify = TRUE)
soilPalette(d$color, lab = d$class, lab.cex = 1)

# full 21 soil texture classes
# ranked by plant available water-holding capacity
d <- soilTextureColorPal(simplify = FALSE)
soilPalette(d$color, lab = d$class, lab.cex = 1)

}
\author{
D.E. Beaudette, Mike Walkinshaw, A.T. O'Geen
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/panel.depth_function.R,
%   R/prepanel.depth_function.R
\name{panel.depth_function}
\alias{panel.depth_function}
\alias{prepanel.depth_function}
\title{Lattice Panel Function for Soil Profiles}
\usage{
panel.depth_function(
  x,
  y,
  id,
  upper = NA,
  lower = NA,
  subscripts = NULL,
  groups = NULL,
  sync.colors = FALSE,
  cf = NA,
  cf.col = NA,
  cf.interval = 20,
  ...
)

prepanel.depth_function(
  x,
  y,
  upper = NA,
  lower = NA,
  subscripts,
  groups = NULL,
  ...
)
}
\arguments{
\item{x}{x values (generated by calling lattice function)}

\item{y}{y values (generated by calling lattice function)}

\item{id}{vector of id labels, same length as x and y--only required when
plotting segments (see Details section)}

\item{upper}{vector of upper confidence envelope values}

\item{lower}{vector of lower confidence envelope values}

\item{subscripts}{paneling indices (generated by calling lattice function)}

\item{groups}{grouping data (generated by calling lattice function)}

\item{sync.colors}{optionally sync the fill color within the region bounded
by (lower--upper) with the line colors}

\item{cf}{optionally annotate contributing fraction data at regular depth
intervals see \code{\link{slab}}}

\item{cf.col}{optional color for contributing fraction values, typically
used to override the line color}

\item{cf.interval}{number of depth units to space printed contributing
fraction values}

\item{\dots}{further arguments to lower-level lattice plotting functions,
see below}
}
\description{
Panel function for plotting grouped soil property data, along with upper and
lower estimates of uncertainty.

This function can be used to replace \code{panel.superpose} when plotting
depth function data. When requested, contributing fraction data are printed
using colors the same color as corresponding depth function lines unless a
single color value is given via \code{cf.col}.

This function is not able to apply transformations (typically \code{log = 10}) applied in the \code{scales} argument to \code{xyplot} to  upper/lower bounds. These will have to be manually applied. See examples.
}
\examples{

library(lattice)
data(sp1)

# 1. plotting mechanism for step-functions derived from soil profile data
xyplot(
  cbind(top, bottom) ~ prop,
  data = sp1,
  id = sp1$id,
  panel = panel.depth_function,
  ylim = c(250, -10),
  scales = list(y = list(tick.number = 10)),
  xlab = 'Property',
  ylab = 'Depth (cm)',
  main = 'panel.depth_function() demo'
)

# 1.1 include groups argument to leverage lattice styling framework
sp1$group <- factor(sp1$group, labels = c('Group 1', 'Group2'))

xyplot(
  cbind(top, bottom) ~ prop,
  groups = group,
  data = sp1,
  id = sp1$id,
  panel = panel.depth_function,
  ylim = c(250, -10),
  scales = list(y = list(tick.number = 10)),
  xlab = 'Property',
  ylab = 'Depth (cm)',
  main = 'panel.depth_function() demo',
  auto.key = list(
    columns = 2,
    points = FALSE,
    lines = TRUE
  ),
  par.settings = list(superpose.line = list(col = c(
    'Orange', 'RoyalBlue'
  )))
)


# more complex examples, using step functions with grouped data
# better looking figures with less customization via tactile package
if(requireNamespace('tactile')) {
  
  library(data.table)
  library(lattice)
  library(tactile)
  
  # example data
  data(sp6)
  
  # a single profile
  x <- sp6[1:5, ]
  
  # wide -> long format
  x.long <- data.table::melt(
    data.table::data.table(x), 
    id.vars = c('id', 'top', 'bottom'), 
    measure.vars = c('sand', 'silt', 'clay')
  )
  
  # (optional) convert back to data.frame
  x.long <- as.data.frame(x.long)
  
  # three variables sharing a common axis
  # factor levels set by melt()
  xyplot(
    cbind(top, bottom) ~ value | id,
    groups = variable,
    data = x.long,
    id = x.long$id,
    ylim = c(200, -5), xlim = c(10, 60),
    scales = list(alternating = 1, y = list(tick.number = 10)),
    par.settings = tactile.theme(superpose.line = list(lwd = 2)),
    xlab = 'Sand, Silt, Clay (\%)',
    ylab = 'Depth (cm)',
    panel = panel.depth_function,
    auto.key = list(columns = 3, lines = TRUE, points = FALSE),
    asp = 1.5
  )
  
  
  # all profiles
  x <- sp6
  
  # wide -> long format
  x.long <- data.table::melt(
    data.table::data.table(x), 
    id.vars = c('id', 'top', 'bottom'), 
    measure.vars = c('sand', 'silt', 'clay')
  )
  
  # (optional) convert back to data.frame
  x.long <- as.data.frame(x.long)
  
  # three variables sharing a common axis
  # factor levels set by melt()
  xyplot(
    cbind(top, bottom) ~ value | id,
    groups = variable,
    data = x.long,
    id = x.long$id,
    ylim = c(200, -5), xlim = c(0, 70),
    scales = list(alternating = 1, y = list(tick.number = 10)),
    par.settings = tactile.theme(superpose.line = list(lwd = 2)),
    xlab = 'Sand, Silt, Clay (\%)',
    ylab = 'Depth (cm)',
    panel = panel.depth_function,
    auto.key = list(columns = 3, lines = TRUE, points = FALSE),
    as.table = TRUE
  )
  
  xyplot(
    cbind(top, bottom) ~ value,
    groups = variable,
    data = x.long,
    id = x.long$id,
    ylim = c(200, -5), xlim = c(0, 70),
    scales = list(alternating = 1, y = list(tick.number = 10)),
    par.settings = tactile.theme(superpose.line = list(lwd = 2)),
    xlab = 'Sand, Silt, Clay (\%)',
    ylab = 'Depth (cm)',
    panel = panel.depth_function,
    auto.key = list(columns = 3, lines = TRUE, points = FALSE),
    as.table = TRUE
  )
  
  xyplot(
    cbind(top, bottom) ~ value | variable,
    groups = variable,
    data = x.long,
    id = x.long$id,
    ylim = c(200, -5), xlim = c(0, 70),
    scales = list(alternating = 1, y = list(tick.number = 10)),
    par.settings = tactile.theme(superpose.line = list(lwd = 2)),
    xlab = 'Sand, Silt, Clay (\%)',
    ylab = 'Depth (cm)',
    panel = panel.depth_function,
    auto.key = list(columns = 3, lines = TRUE, points = FALSE),
    as.table = TRUE
  )
  
  
  xyplot(
    cbind(top, bottom) ~ value | variable,
    data = x.long,
    id = x.long$id,
    ylim = c(200, -5), xlim = c(0, 70),
    scales = list(alternating = 1, y = list(tick.number = 10)),
    par.settings = tactile.theme(superpose.line = list(lwd = 2)),
    xlab = 'Sand, Silt, Clay (\%)',
    ylab = 'Depth (cm)',
    panel = panel.depth_function,
    auto.key = list(columns = 3, lines = TRUE, points = FALSE),
    as.table = TRUE
  )
  
  
}
}
\seealso{
\code{\link{sp1}}, \code{\link{slice}}, \code{\link{slab}}
}
\author{
D.E. Beaudette
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-operators.R
\name{[,SoilProfileCollection-method}
\alias{[,SoilProfileCollection-method}
\title{Matrix/data.frame-like access to profiles and horizons in a SoilProfileCollection}
\usage{
\S4method{[}{SoilProfileCollection}(x, i, j, ..., drop = TRUE)
}
\arguments{
\item{x}{a SoilProfileCollection}

\item{i}{a numeric or logical value denoting profile indices to select in a subset}

\item{j}{a numeric or logical value denoting horizon indices to select in a subset}

\item{...}{non-standard expressions to evaluate in a subset}

\item{drop}{Default: \code{TRUE}. When \code{drop=FALSE} placeholder horizons (profile ID with all other values \code{NA}) are created where the specified \code{j} index results in removal of all horizons.}
}
\description{
You can access the contents of a SoilProfileCollection by profile and horizon "index", \code{i} and \code{j}, respectively: \code{spc[i, j, ...]}. Subset operations are propagated to other slots (such as diagnostics or spatial) when they result in removal of sites from a collection.
\itemize{
\item \code{i} refers to the profile position within the collection. By default the order is based on the C SORT order of the variable that you specified as your unique profile ID at time of object construction. Note that if your ID variable was numeric, then it has been sorted as a character.
\item \code{j} refers to the horizon or "slice" index. This index is most useful when either a) working with \code{slice}'d SoilProfileCollection or b) working with single-profile collections. \code{j} returns the layer in the specified index positions for all profiles in a collection.
\item \code{...} is an area to specify an expression that is evaluated in the subset. Currently supported
\itemize{
\item \code{.LAST} (last horizon in each profile): return the last horizon from each profile. This uses \code{i} but ignores the regular \code{j} index.
\item \code{.FIRST} (first horizon in each profile): return the last horizon from each profile. This uses \code{i} but ignores the regular \code{j} index.
\item \code{.HZID} (horizon index): return the horizon indices corresponding to \code{i}+\code{j}+\code{...} ("k") constraints
\item \code{.NHZ} (number of horizons): return the number of horizons in the profiles resulting from \code{i}+\code{j}+\code{...} ("k") constraints
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim.R
\name{sim}
\alias{sim}
\title{DEPRECATED Simulate Soil Profiles}
\usage{
sim(x, n = 1, iterations = 25, hz.sd = 2, min.thick = 2)
}
\arguments{
\item{x}{a SoilProfileCollection object containing a single profile from which to draw simulated data}

\item{n}{the number of requested simulations}

\item{iterations}{sampling iterations used to determine each horizon thickness}

\item{hz.sd}{standard deviation used to simulate horizon thickness, can be a vector but must divide evenly into the number of horizons found in \code{x}}

\item{min.thick}{minimum horizon thickness allowed in simulation results}
}
\value{
A SoilProfileCollection object with \code{n} simulated profiles, each containing the same number of horizons and same data as \code{x}
}
\description{
Simulate a collection of soil profiles based on the horizonation of a single soil profile. Now deprecated: use \code{\link[=perturb]{perturb()}} for perturbations of horizon thicknesses or boundaries.
}
\details{
This function generates a collection of simulated soil profiles based on the horizon thickness data associated with a single "template" profile. Simulation is based on sampling from a family of Gaussian distribution with means defined by the "template" profile and standard deviation defined by the user.
}
\examples{

# please see documentation for perturb() for examples
#  the sim() function calls perturb() internally

}
\seealso{
\code{\link{random_profile}}  \code{\link{perturb}}
}
\author{
D. E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy_uncertainty.R
\name{brierScore}
\alias{brierScore}
\title{Multinominal Brier Score}
\usage{
brierScore(x, classLabels, actual = "actual")
}
\arguments{
\item{x}{\code{data.frame} of class probabilities (numeric) and observed class label (character), see examples}

\item{classLabels}{vector of predicted class labels (probabilities), corresponding to column names in \code{x}}

\item{actual}{name of column containing the observed class, should be character vector not factor}
}
\value{
a single Brier score, representative of data in \code{x}
}
\description{
Compute a multinominal Brier score from predicted class probabilities and observed class label. Lower values are associated with a more accurate classifier.
}
\examples{

# columns 'a', 'b', 'c' contain predicted probabilities
# column 'actual' contains observed class label

# a good classifier
d.good <- data.frame(
  a = c(0.05, 0.05, 0.10),
  b = c(0.90, 0.85, 0.75),
  c = c(0.05, 0.10, 0.15),
  actual = c('b', 'b', 'b'),
  stringsAsFactors = FALSE
)

# a rather bad classifier
d.bad <- data.frame(
  a = c(0.05, 0.05, 0.10),
  b = c(0.90, 0.85, 0.75),
  c = c(0.05, 0.10, 0.15),
  actual = c('c', 'c', 'c'),
  stringsAsFactors = FALSE
)

# class labels are factors
d.factors <- data.frame(
  a = c(0.05, 0.05, 0.10),
  b = c(0.90, 0.85, 0.75),
  c = c(0.05, 0.10, 0.15),
  actual = c('b', 'b', 'b'),
  stringsAsFactors = TRUE
)

# relatively low value = accurate
brierScore(x = d.good, classLabels = c('a', 'b', 'c'), actual = 'actual')

# high values = not accuate
brierScore(x = d.bad, classLabels = c('a', 'b', 'c'), actual = 'actual')

# message related to conversion of factor -> character
brierScore(x = d.factors, classLabels = c('a', 'b', 'c'), actual = 'actual')

}
\references{
Brier, Glenn W. 1950. "Verification of Forecasts Expressed in Terms of Probability." Monthly Weather Review 78 (1): 1-3. doi:10.1175/1520-0493(1950)078<0001:VOFEIT>2.0.CO;2.
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/contrastClass.R
\name{contrastClass}
\alias{contrastClass}
\title{Soil Color Contrast}
\usage{
contrastClass(v1, c1, v2, c2, dH, dV, dC, verbose = FALSE)
}
\arguments{
\item{v1}{Munsell value of first color}

\item{c1}{Munsell chroma of first color}

\item{v2}{Munsell value of second color}

\item{c2}{Munsell chroma of second color}

\item{dH}{delta Hue}

\item{dV}{delta Value}

\item{dC}{delta Chroma}

\item{verbose}{return a list for testing rules/cases}
}
\value{
A vector of color contrast classes (ordered factor). A list when
\code{verbose} is TRUE.
}
\description{
Determine soil color contrast class according to methods outlined in the
Soil Survey Manual. This function is typically called from
\code{colorContrast()} which is simpler to use and provides more
information.
}
\details{
This function is fully vectorized but expects all inputs have the same
length.
}
\examples{

## standard use, result is an ordered factor
# 10YR 6/3 vs 5YR 3/4
contrastClass(v1=6, c1=3, v2=3, c2=4, dH=2, dV=3, dC=1)

## verbose output, useful for testing rules/cases
# 10YR 6/3 vs 5YR 3/4
contrastClass(v1=6, c1=3, v2=3, c2=4, dH=2, dV=3, dC=1, verbose = TRUE)

}
\references{
\itemize{
\item Soil Survey Technical Note 2 \href{https://web.archive.org/web/20220704214918/https://www.nrcs.usda.gov/wps/portal/nrcs/detail/soils/ref/?cid=nrcs142p2_053569}{wayback machine URL}
}
}
\seealso{
\code{\link{colorContrast}}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hzTransitionProbabilities.R,
%   R/mostLikelyHzSequence.R
\name{hzTransitionProbabilities}
\alias{hzTransitionProbabilities}
\alias{mostLikelyHzSequence}
\title{Horizon Transition Probabilities}
\usage{
hzTransitionProbabilities(
  x,
  name = GHL(x, required = TRUE),
  loopTerminalStates = FALSE
)

mostLikelyHzSequence(mc, t0, maxIterations = 10)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object.}

\item{name}{A horizon level attribute in \code{x} that names horizons.}

\item{loopTerminalStates}{should terminal states loop back to themselves?

This is useful when the transition probability matrix will be used to
initialize a \code{markovchain} object. See examples below.}

\item{mc}{Passed to \code{markovchain} \code{conditionalDistribution()}}

\item{t0}{Passed to \code{markovchain} \code{conditionalDistribution()}}

\item{maxIterations}{Maximum number of iterations. Default: \code{10}}
}
\value{
A square matrix of transition probabilities. See examples.

The function \code{\link[=genhzTableToAdjMat]{genhzTableToAdjMat()}} returns a square adjacency matrix.
See examples.

The function \code{\link[=mostLikelyHzSequence]{mostLikelyHzSequence()}} returns the most likely sequence of
horizons, given a \code{markovchain} object initialized from horizon
transition probabilities and an initial state, \code{t0}. See examples.
}
\description{
Functions for creating and working with horizon (sequence) transition
probability matrices.

See the following tutorials for some ideas:
\itemize{
\item \href{http://ncss-tech.github.io/AQP/aqp/hz-transition-probabilities.html}{horizon designation TP}
\item \href{http://ncss-tech.github.io/AQP/aqp/series-color-TP-graph.html}{soil color TP}
}
}
\note{
These functions are still experimental and subject to change.
}
\examples{

data(sp4)
depths(sp4) <- id ~ top + bottom

# horizon transition probabilities: row -> col transitions
(tp <- hzTransitionProbabilities(sp4, 'name'))



}
\seealso{
\code{\link[=generalize.hz]{generalize.hz()}}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rebuildSPC.R
\name{rebuildSPC}
\alias{rebuildSPC}
\title{Rebuild a SoilProfileCollection object}
\usage{
rebuildSPC(x)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}
}
\value{
A valid \code{SoilProfileCollection} object.

A valid \code{SoilProfileCollection} object.
}
\description{
Rebuild a SoilProfileCollection object
}
\details{
Attempt rebuilding a \code{SoilProfileCollection} object by splitting into
components and re-assembling. Likely only used to fix outdated
\code{SoilProfileCollection} objects that are missing slots.
}
\seealso{
\code{\link{checkSPC}}
Rebuild a SoilProfileCollection object

Rebuild a SoilProfileCollection object

Attempt rebuilding a \code{SoilProfileCollection} object by splitting into
components and re-assembling. Likely only used to fix outdated
\code{SoilProfileCollection} objects that are missing slots.

\code{\link{checkSPC}}
}
\author{
D.E. Beaudette

D.E. Beaudette, A.G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mixMunsell.R
\name{mixMunsell}
\alias{mixMunsell}
\title{Mix Munsell Colors via Spectral Library}
\usage{
mixMunsell(
  x,
  w = rep(1, times = length(x))/length(x),
  mixingMethod = c("exact", "reference", "estimate", "adaptive"),
  n = 1,
  keepMixedSpec = FALSE,
  distThreshold = 0.025,
  ...
)
}
\arguments{
\item{x}{vector of colors in Munsell notation}

\item{w}{vector of proportions, can sum to any number}

\item{mixingMethod}{approach used to simulate a mixture:
\itemize{
\item \code{exact}: simulate a subtractive mixture of pigments, color conversion via CIE1931 color-matching functions (see details)
\item \code{reference}: simulate a subtractive mixture of pigments, selecting \code{n} closest reference spectra from \code{\link{munsell.spectra.wide}} (requires \code{gower} package)
\item \code{estimate}: closest Munsell chip to a weighted mean of CIELAB coordinates (fastest)
\item \code{adaptive}: use \code{exact} method when possible, falling-back to \code{estimate} (weighted mean of CIELAB coordinates) otherwise
}}

\item{n}{number of closest matching color chips (\code{mixingMethod = reference} only)}

\item{keepMixedSpec}{keep weighted geometric mean spectra, final result is a \code{list} (\code{mixingMethod = reference} only)}

\item{distThreshold}{spectral distance used to compute \code{scaledDistance}, default value is based on an analysis of spectral distances associated with adjacent Munsell color chips. This argument is only used with  \code{mixingMethod = 'reference'}.}

\item{...}{additional arguments to \code{\link{spec2Munsell}}}
}
\value{
A \code{data.frame} with the closest matching Munsell color(s):
\itemize{
\item \code{munsell}: Munsell notation of the n-closest spectra
\item \code{distance}: spectral (Gower) distance to the n-closest spectra
\item \code{scaledDistance}: spectral distance scaled by \code{distThreshold}
\item \code{mixingMethod}: method used for each mixture
}

When \code{keepMixedSpec = TRUE} then a \code{list}:
\itemize{
\item \code{mixed}: a \code{data.frame} containing the same elements as above
\item \code{spec}: spectra for the 1st closest match
}
}
\description{
Simulate mixing of colors in Munsell notation, similar to the way in which mixtures of pigments operate.
}
\details{
See the \href{https://ncss-tech.github.io/AQP/aqp/mix-colors.html}{expanded tutorial} for examples.

An accurate simulation of pigment mixtures ("subtractive" color mixtures) is incredibly complex due to factors that aren't easily measured or controlled: pigment solubility, pigment particle size distribution, water content, substrate composition, and physical obstruction to name a few. That said, it is possible to simulate reasonable, subtractive color mixtures given a reference spectra library (350-800nm) and some assumptions about pigment qualities and lighting. For the purposes of estimating a mixture of soil colors (these are pigments after all) we can relax these assumptions and assume a standard light source. The only missing piece is the spectral library for all Munsell chips in our color books.

Paul Centore has provided a Munsell color chip reflectance spectra library (http://www.munsellcolourscienceforpainters.com). The estimation of a subtractive mixture of soil colors can proceed as follows:
\enumerate{
\item look up the associated spectra for each color in \code{x}
\item compute the weighted (\code{w} argument) geometric mean of the spectra
\item convert the spectral mixture to the closest Munsell color via:
}
\itemize{
\item search for the closest \code{n} matching spectra in the reference library (\code{mixtureMethod = 'reference'})
\item direct conversion of spectra to closest Munsell color via \code{\link[=spec2Munsell]{spec2Munsell()}} (\code{mixtureMethod = 'exact'})
}
\enumerate{
\item suggest resulting Munsell chip(s) as the best candidate for a simulated mixture
}

Key assumptions include:
\itemize{
\item similar particle size distribution
\item similar mineralogy (i.e. pigmentation qualities)
\item similar water content.
}

For the purposes of estimating (for example) a "mixed soil color within the top 18cm of soil" these assumptions are usually valid. Again, these are estimates that are ultimately "snapped" to the nearest chip and not do not need to approach the accuracy of paint-matching systems.

A message is printed when \code{scaledDistance} is larger than 1.
}
\references{
Marcus, R.T. (1998). The Measurement of Color. In K. Nassau (Ed.), Color for Science, Art, and Technology (pp. 32-96). North-Holland.
\itemize{
\item http://scottburns.us/wp-content/uploads/2015/05/Subtractive-Color-Mixture-Computation.pdf
\item \href{https://stackoverflow.com/questions/10254022/implementing-kubelka-munk-like-krita-to-mix-colours-color-like-paint/29967630#29967630}{related discussion on Stack Overflow}
\item spectral library source: http://www.munsellcolourscienceforpainters.com/MunsellResources/SpectralReflectancesOf2007MunsellBookOfColorGlossy.txt
}
}
\seealso{
\link{munsell.spectra}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/L1_profiles.R
\name{L1_profiles}
\alias{L1_profiles}
\title{Create Representative Soil Profiles via L1 Estimator}
\usage{
L1_profiles(
  x,
  fm,
  basis = 1,
  method = c("regex", "simple", "constant"),
  maxDepthRule = c("max", "min"),
  maxDepthConstant = NULL
)
}
\arguments{
\item{x}{\code{SoilProfileCollection} object}

\item{fm}{formula, for example: \code{group ~ p1 + p2 + p3}, where "group" is a site-level grouping variable, and "p1", "p2", and "p3" are horizon level variables}

\item{basis}{positive integer, aggregation basis (e.g. 1 for 1-depth-unit intervals). Values other than 1 are not currently supported.}

\item{method}{soil depth evaluation method: "regex" for regular expression, "simple", or "constant". See details.}

\item{maxDepthRule}{maximum depth rule: "max" or "min" See details.}

\item{maxDepthConstant}{positive integer, maximum depth when \code{maxDepthRule = 'constant'}}
}
\value{
a \code{SoilProfileCollection} object
}
\description{
The L1 estimator, or \href{https://en.wikipedia.org/wiki/Geometric_median}{geometric median}, is a multivariate generalization of the (univariate) \href{https://en.wikipedia.org/wiki/Median}{median} concept. This function performs a multivariate aggregation (via L1 estimator) according to a suite of ratio-scale soil properties. The L1 estimator is applied to soil profile data that have been sliced to a 1-depth-unit basis. Data should be well stratified by groups defined in \code{fm}, otherwise the L1 median may not make any sense.

See the \href{https://ncss-tech.github.io/AQP/aqp/L1-profiles.html}{L1 Profiles Tutorial} for additional examples.
}
\details{
See \href{https://ncss-tech.github.io/AQP/aqp/L1-profiles.html}{this related tutorial} for additional examples. The \code{method}, \code{maxDepthRule}, and \code{maxDepthConstant} arguments set the maximum depth (over the entire collection) of analysis used to build "L1 profiles". The following rules are available:
\itemize{
\item \code{method = 'regex'} uses pattern matching on horizon designations (note that \code{hzdesgnname} metadata must be set with \code{hzdesgnname(x) <- 'columnname'})
\item \code{method = 'simple'} uses \code{min} or \code{max} as applied to \code{x}, no accounting for non-soil horizons (e.g. Cr or R)
\item \code{method = 'constant'} uses a fixed depth value supplied by \code{maxDepthConstant}
}

The \code{maxDepthRule} argument sets depth calculation constraint, applied to soil depths computed according to \code{method} (\code{min} or \code{max}).
}
\note{
This function requires the \code{Gmedian} package.
}
\references{
Cardot, H., Cenac, P. and Zitt, P-A. (2013). Efficient and fast estimation of the geometric median in Hilbert spaces with an averaged stochastic gradient algorithm. Bernoulli, 19, 18-43.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/col2Munsell.R
\name{.detectColorSpec}
\alias{.detectColorSpec}
\title{Detect color specification from a vector of values, or a matrix of color coordinates}
\usage{
.detectColorSpec(col)
}
\arguments{
\item{col}{character vector, numeric matrix with 3 columns, or data.frame with 3 columns}
}
\description{
Detect color specification from a vector of values, or a matrix of color coordinates
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-operators.R
\docType{methods}
\name{$<-}
\alias{$<-}
\alias{$<-,SoilProfileCollection-method}
\title{Set data in column of horizon or site data in a SoilProfileCollection}
\usage{
\S4method{$}{SoilProfileCollection}(x, name) <- value
}
\arguments{
\item{x}{a SoilProfileCollection}

\item{name}{a single column name in site or horizon table}

\item{value}{Replacement values: unit length or equal to number of horizons or sites.}
}
\description{
Set the data in a column accessed by name \code{spc$name}. Column names other than profile ID are not shared between site and horizons.

When using \code{$<-}, the length of input and output matching either the number of sites or number of horizons is used to determine which slot new columns are assigned to. Use \code{site(x)$name <- value} or  \code{horizons(x)$name <- value} to be explicit about which slot is being accessed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/denormalize.R
\name{denormalize}
\alias{denormalize}
\title{Create a (redundant) horizon-level attribute from a site-level attribute}
\usage{
denormalize(object, attr)
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{attr}{Site-level attribute name (character string) to denormalize to horizon.}
}
\value{
A vector of values of equal length to the number of rows in the horizon table of the input SPC.
}
\description{
Create a (redundant) horizon-level attribute from a site-level attribute. Specify a SoilProfileCollection and a site-level attribute from that SPC (by name) to receive a vector of length equal to the number of horizons containing the site-level values. This vector is directly usable with the SoilProfileCollection horizon setter.

\code{denormalize} is the inverse operation for the formula interface that "normalizes" a horizon level variable to site level:

\code{site(object) <- ~ horizonvar}
}
\details{
"Denormalization" is the process of trying to improve the read performance of a database, at the expense of losing some write performance, by adding redundant copies of data or by grouping data. Sometimes it is beneficial to have site-level attributes denormalized for grouping of horizon-level data in analyses. \code{denormalize} achieves this result for SoilProfileCollections.
}
\examples{

data(sp1)

# create a SoilProfileCollection from horizon data
depths(sp1) <- id ~ top + bottom

# create random site-level attribute `sitevar` with a binary (0/1) outcome
sp1$sitevar <- round(runif(length(sp1)))

# use denormalize() to create a mirror of sitevar in the horizon table
# name the attribute something different (e.g. `hz.sitevar`) to 
# prevent collision with the site attribute
# the attributes can have the same name but you will then need 
# site() or horizons() to access explicitly
sp1$hz.sitevar <- denormalize(sp1, 'sitevar')

# compare number of profiles to number of sitevar assignments
length(sp1)
table(sp1$sitevar)

# compare number of horizons to number of horizon-level copies of sitevar `hz.'sitevar`
nrow(sp1)
table(sp1$hz.sitevar)
}
\author{
Andrew G. Brown, Dylan Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{SPC.with.overlap}
\alias{SPC.with.overlap}
\title{Example SoilProfileCollection with Overlapping Horizons}
\format{
A \code{SoilProfileCollection}
}
\usage{
data(SPC.with.overlap)
}
\description{
A \code{SoilProfileCollection} with overlapping horizons, derived from a Dynamic Soil Properties project.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{siteNames<-}
\alias{siteNames<-}
\alias{siteNames<-,SoilProfileCollection-method}
\alias{siteNames,SoilProfileCollection-method}
\alias{siteNames}
\title{Set site column names}
\usage{
\S4method{siteNames}{SoilProfileCollection}(object) <- value

\S4method{siteNames}{SoilProfileCollection}(object)
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{value}{a unique vector of equal length to number of columns in site: \code{length(siteNames(object))}}
}
\description{
Set site column names

Get names of columns in site table.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregateColor.R
\name{aggregateColor}
\alias{aggregateColor}
\title{Summarize Soil Colors}
\usage{
aggregateColor(
  x,
  groups = "genhz",
  col = "soil_color",
  k = NULL,
  profile_wt = NULL,
  mixingMethod = c("estimate", "exact")
)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}

\item{groups}{the name of a horizon or site attribute used to group horizons, see examples}

\item{col}{the name of a horizon-level attribute with soil color specified in hexadecimal (i.e. "#rrggbb")}

\item{k}{single integer specifying the number of colors discretized via PAM (\code{\link[cluster:pam]{cluster::pam()}}), see details}

\item{profile_wt}{the name of a site-level attribute used to modify weighting, e.g. area}

\item{mixingMethod}{method used to estimate "aggregate" soil colors, see \code{\link[=mixMunsell]{mixMunsell()}}}
}
\value{
A list with the following components:
\itemize{
\item \code{scaled.data}: a \code{list} of colors and associated weights, one item for each generalized horizon label with at least one color specified in the source data
\item \code{aggregate.data}: a \code{data.frame} of weighted-mean colors, one row for each generalized horizon label with at least one color specified in the source data
}
}
\description{
Summarize soil color data, weighted by occurrence and horizon thickness.
}
\details{
Weights are computed by:
\code{w_i = sqrt(sum(thickness_i)) * n_i}
where \code{w_i} is the weight associated with color \code{i}, \code{thickness_i} is the total thickness of all horizons associated with the color \code{i}, and \code{n_i} is the number of horizons associated with color \code{i}. Weights are computed within groups specified by \code{groups}.

See the \href{http://ncss-tech.github.io/AQP/sharpshootR/aggregate-soil-color.html}{related tutorial for additional examples}.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# load some example data
data(sp1, package = 'aqp')

# upgrade to SoilProfileCollection and convert Munsell colors
sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# generalize horizon names
n <- c('O', 'A', 'B', 'C')
p <- c('O', 'A', 'B', 'C')
sp1$genhz <- generalize.hz(sp1$name, n, p)

# aggregate colors over horizon-level attribute: 'genhz'
a <- aggregateColor(sp1, groups = 'genhz', col = 'soil_color')

# check results
str(a)

# simple visualization
aggregateColorPlot(a)

}
\seealso{
\code{\link[=generalize.hz]{generalize.hz()}}, \code{\link[=aggregateColorPlot]{aggregateColorPlot()}}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy_uncertainty.R
\name{shannonEntropy}
\alias{shannonEntropy}
\title{Shannon Entropy}
\usage{
shannonEntropy(x, b = 2)
}
\arguments{
\item{x}{vector of probabilities (0,1), must sum to 1, should not contain NA}

\item{b}{logarithm base}
}
\value{
A single numeric value.
}
\description{
A very simple implementation of Shannon entropy.
}
\details{
\code{0}s are automatically removed by \code{na.rm = TRUE}, as \code{(0 * log(0) = Nan)}
}
\note{
When \code{b = length(x)} the result is the normalized Shannon entropy of (Kempen et al, 2009).
}
\examples{

# a very simple example
p <- c(0.25, 0.25, 0.4, 0.05, 0.05)

shannonEntropy(p)



}
\references{
Kempen, Bas, Dick J. Brus, Gerard B.M. Heuvelink, and Jetse J. Stoorvogel. 2009. "Updating the 1:50,000 Dutch Soil Map Using Legacy Soil Data: A Multinominal Logistic Regression Approach." Geoderma 151: 311-26. doi:10.1016/j.geoderma.2009.04.023

Shannon, Claude E. (July-October 1948). "A Mathematical Theory of Communication". Bell System Technical Journal. 27 (3): 379-423. doi:10.1002/j.1538-7305.1948.tb01338.x
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{show}
\alias{show}
\alias{show,SoilProfileCollection-method}
\alias{as.character,SoilProfileCollection-method}
\title{SoilProfileCollection show method}
\usage{
\S4method{show}{SoilProfileCollection}(object)

\S4method{as.character}{SoilProfileCollection}(x, ...)
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{x}{a SoilProfileCollection}

\item{...}{additional arguments (not used)}
}
\description{
Pretty output method for SoilProfileCollection objects. By default this method limits output to 10 columns and 6 rows from the site and horizon tables respectively.

There is an aqp environment option you can set to increase the number of columns shown by default: \code{options(.aqp.show.n.cols = 100)},

\code{as.character()}: Character Representation of SoilProfileCollection Object
}
\examples{

# load a SoilProfileCollection
data(sp5)

# use the show() method
show(sp5)

# which is same as this (in the console)
sp5

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fillHzGaps.R
\name{fillHzGaps}
\alias{fillHzGaps}
\title{Find and Fill Horizon Gaps}
\usage{
fillHzGaps(x, flag = TRUE, to_top = 0, to_bottom = max(x))
}
\arguments{
\item{x}{\code{SoilProfileCollection} object}

\item{flag}{logical, flag empty horizons that have been added. default: \code{TRUE}}

\item{to_top}{numeric, fill from shallowest top depth in each profile to specified depth? default: \code{0}}

\item{to_bottom}{numeric, fill from deepest bottom depth in each profile to specified depth? default: \code{aqp::max(x)}}
}
\value{
a possibly modified \code{SoilProfileCollection} object
}
\description{
This function attempts to find "gaps" in the horizon records of a \code{SoilProfileCollection} object and fill with placeholder horizons (profile ID, horizon ID, to/bottom depths, all else \code{NA}). Missing horizon records between the top of each profile and \code{to_top}, or the bottom of each profile and \code{to_bottom} are treated as gaps when those arguments are not \code{NULL}. You can use this function to prepare a potentially messy \code{SoilProfileCollection} for subsequent analyses that are sensitive to horizon sequence inconsistencies or require a conformal "rectangle" of data spanning known depths.

Gaps are defined as:
\itemize{
\item within each profile, for horizons \code{i} to \code{n_hz}:
\item \verb{bottom_i != top_i+1 (but only to i = 1:(n_hz - 1)}
}
}
\examples{

data(sp4)
depths(sp4) <- id ~ top + bottom

# introduce depth logic errors
idx <- c(2, 6:7, 8, 12)
sp4$top[idx] <- NA

# check
horizons(sp4)[idx, ]

# create gaps by removing logic errors
x <- HzDepthLogicSubset(sp4, byhz = TRUE)

# check on removed horizons (hzID values)
metadata(x)$removed.horizons

# inspect
par(mar = c(0, 0, 0, 2))
plotSPC(x, width = 0.3, default.color = 'royalblue', 
name = 'hzID', name.style = 'center-center', cex.names = 0.8,
cex.id = 0.66)

# fill gaps left by HzDepthLogicSubset()
z <- fillHzGaps(x, flag = TRUE)

# graphical check
plotSPC(z, width = 0.3, color = '.filledGap', name = 'hzID', 
show.legend = FALSE, name.style = 'center-center', cex.names = 0.8,
cex.id = 0.66)

# fill top to 0 cm
z2 <- fillHzGaps(x, flag = TRUE, to_top = 0)
plotSPC(z2, width = 0.3, color = '.filledGap', name = 'hzID', show.legend = FALSE)

# fill bottom to max(SPC)
z3 <- fillHzGaps(x, flag = TRUE, to_top = 0, to_bottom = max(x))
plotSPC(z3, width = 0.3, color = '.filledGap', name = 'hzID', show.legend = FALSE)

## another example
data(sp4)
depths(sp4) <- id ~ top + bottom

# remove 1st horizons from profiles 1:4
idx <- sp4[,, .FIRST, .HZID]
replaceHorizons(sp4) <- horizons(sp4)[-idx[1:4], ]

# prepare for dice()
z <- fillHzGaps(sp4, to_top = 0, to_bottom = 50, flag = TRUE) 

# empty-horizon padding is in place for formula interface to dice()
d <- dice(z, fm = 0:50 ~ .)
plotSPC(d, color = 'Ca', show.legend = FALSE)
plotSPC(d, color = '.filledGap', show.legend = FALSE)

}
\author{
A.G. Brown and D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R,
%   R/SoilProfileCollection-setters.R
\docType{methods}
\name{restrictions,SoilProfileCollection-method}
\alias{restrictions,SoilProfileCollection-method}
\alias{restrictions}
\alias{restrictions<-}
\alias{restrictions<-,SoilProfileCollection-method}
\title{Get or Set Restriction data in a SoilProfileCollection}
\usage{
\S4method{restrictions}{SoilProfileCollection}(object)

\S4method{restrictions}{SoilProfileCollection}(object) <- value
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{value}{An data.frame object containing at least a column with name \code{idname(object)}}
}
\description{
Restrictions describe root-limiting features in the soil. A single profile may have multiple restrictions.
\itemize{
\item \code{restrictions()} (get method): Get restriction data from a SoilProfileCollection.
}

\itemize{
\item \verb{restrictions<-} (set method): Set restriction data for a SoilProfileCollection. The profile ID column from \code{object} (\code{idname(object)}) must be present in the replacement \code{value} object.
}
}
\examples{

# load test data
data(sp2)

# promote to SPC
depths(sp2) <- id ~ top + bottom

# assign abrupt textural change to a profile
newdata <- data.frame(id = c("hon-21"),
                      restrkind = "abrupt textural change",
                      restrdep = 46)

# do left join
restrictions(sp2) <- newdata

# inspect site table: newvalue TRUE only for horizons
#  with top depth equal to zero
restrictions(sp2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getLastHorizonID.R
\name{getLastHorizonID}
\alias{getLastHorizonID}
\title{Get IDs of Deepest Horizons by Profile}
\usage{
getLastHorizonID(x)
}
\arguments{
\item{x}{a \code{SoilProfileCollection}}
}
\description{
Return horizon IDs of the deepest horizon within each profile of a \code{SoilProfileCollection}. IDs are returned in the same order as \code{profile_id(x)}. Horizon top depths are used because there are cases where bottom depths may be missing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-metadata.R
\docType{methods}
\name{hzmetaname}
\alias{hzmetaname}
\alias{hzmetaname,SoilProfileCollection-method}
\alias{hzmetaname<-}
\alias{hzmetaname<-,SoilProfileCollection-method}
\title{Get or Set Horizon Metadata Column Name}
\usage{
\S4method{hzmetaname}{SoilProfileCollection}(object, attr, required = FALSE)

\S4method{hzmetaname}{SoilProfileCollection}(object, attr, required = FALSE) <- value
}
\arguments{
\item{object}{A \emph{SoilProfileCollection}}

\item{attr}{\emph{character}. Base name for attribute to be stored in metadata. This is prefixed with \code{"aqp_hz"} for horizon-level metadata for column attributes. e.g. \code{attr="clay"} results in metadata value retrieved from \code{"aqp_hzclay"}.}

\item{required}{\emph{logical}. Is this attribute required? If it is, set to \code{TRUE} to trigger error on invalid \code{value}.}

\item{value}{\emph{character}. Name of horizon-level column containing  data corresponding to \code{attr}.}
}
\description{
\code{hzmetaname()}: Get column name containing horizon data of interest

\verb{hzmetaname<-}: Set horizon designation column name
}
\details{
Store the column name containing a specific type of horizon data in the metadata slot of the SoilProfileCollection.
}
\examples{

data(sp1)

# promote to SPC
depths(sp1) <- id ~ top + bottom

# set important metadata columns
hzdesgnname(sp1) <- "name"
hztexclname(sp1) <- "texture"

# set custom horizon property (clay content) column
hzmetaname(sp1, "clay") <- "prop"

# inspect metadata list
metadata(sp1)

# get horizon clay content column
hzmetaname(sp1, "clay")

# uses hzdesgname(), hztexclname(), hzmetaname(attr="clay") in function definition
estimatePSCS(sp1)
}
\seealso{
\code{\link[=guessHzAttrName]{guessHzAttrName()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HzDepthLogicSubset.R
\name{HzDepthLogicSubset}
\alias{HzDepthLogicSubset}
\title{Subset \code{SoilProfileCollection} Objects or Horizons via \code{checkHzDepthLogic}}
\usage{
HzDepthLogicSubset(x, byhz = FALSE)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}

\item{byhz}{logical, evaluate horizon depth logic at the horizon level (profile level if \code{FALSE})}
}
\value{
a \code{SoilProfileCollection} object
}
\description{
This function removes profiles or horizons from a \code{SoilProfileCollection} that are flagged as having invalid horizon depth logic by \code{\link{checkHzDepthLogic}}. Invalid profiles may be created when setting \code{byhz = TRUE}; use caution as some functions may not work properly in the presence of gaps. Consider using \code{\link{fillHzGaps}} to fill these gaps.
}
\note{
This function cannot identify (and remove) overlapping horizons when \code{byhz = TRUE}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{osd}
\alias{osd}
\title{Example Output from soilDB::fetchOSD()}
\format{
A \code{SoilProfileCollection}
}
\usage{
data(osd)
}
\description{
An example \code{SoilProfileCollection} object created by \code{soilDB::fetchOSD()}, derived from the Cecil, Appling, and Bonneau Official Series Descriptions.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalize.hz.R
\name{generalize.hz}
\alias{generalize.hz}
\alias{generalizeHz,character-method}
\alias{generalizeHz}
\alias{generalizeHz,SoilProfileCollection-method}
\title{Generalize Horizon Names}
\usage{
generalize.hz(
  x,
  new,
  pattern,
  non.matching.code = "not-used",
  hzdepm = NULL,
  ordered = !missing(hzdepm),
  na.rm = TRUE,
  ...
)

\S4method{generalizeHz}{character}(
  x,
  new,
  pattern,
  non.matching.code = "not-used",
  hzdepm = NULL,
  ordered = !missing(hzdepm),
  ...
)

\S4method{generalizeHz}{SoilProfileCollection}(
  x,
  new,
  pattern,
  non.matching.code = "not-used",
  hzdepm = NULL,
  ordered = !missing(hzdepm),
  ghl = "genhz",
  ...
)
}
\arguments{
\item{x}{character vector of horizon names or a \code{SoilProfileCollection} object}

\item{new}{character vector of generalized horizon labels (GHL)}

\item{pattern}{character vector of REGEX patterns, same length as \code{new}}

\item{non.matching.code}{character, label used for any horizon not matched by \code{pattern}}

\item{hzdepm}{numeric vector of horizon mid-points; \code{NA} values in \code{hzdepm} will result in \code{non.matching.code} (or \code{NA} if not defined) in result}

\item{ordered}{logical, default \code{TRUE} when \code{hzdepm} argument is specified}

\item{na.rm}{logical, default \code{TRUE} will ignore missing depths in calculating sort order when \code{hzdepm} is specified and \code{ordered=TRUE}}

\item{...}{additional arguments passed to \code{grep()} such as \code{perl = TRUE} for advanced REGEX}

\item{ghl}{Generalized Horizon Designation column name (to be created/updated when \code{x} is a \code{SoilProfileCollection})}
}
\value{
factor (an ordered factor when \code{ordered=TRUE}) of the same length as \code{x} (if character) or as number of horizons in \code{x} (if \code{SoilProfileCollection})
}
\description{
Generalize a vector of horizon names, based on new classes, and REGEX
patterns. Or create a new column \code{ghl} in a \code{SoilProfileCollection} (requires a horizon designation name to be defined for the collection, see details)
}
\details{
When \code{x} is a \code{SoilProfileCollection} the \code{ghl} column will be updated with the factor results. This requires that the "horizon designation name" metadata be defined for the collection to set the column for input designations.
}
\examples{

data(sp1)

# check original distribution of hz designations
table(sp1$name)

# generalized horizon labels
# character vector input
sp1$genhz <- generalizeHz(
  sp1$name,
  new = c('O','A','B','C','R'),
  pattern = c('O', '^A','^B','C','R'),
  ordered = TRUE
)

# see how we did / what we missed
table(sp1$genhz, sp1$name)


## a more advanced example, requries `perl = TRUE`
# example data
x <- c('A', 'AC', 'Bt1', '^AC', 'C', 'BC', 'CB')

# new labels
n <- c('A', '^AC', 'C')

# patterns:
# "A anywhere in the name"
# "literal '^A' anywhere in the name"
# "C anywhere in name, but without preceding A"
p <- c('A', '^A', '(?<!A)C')

# note additional argument
res <- generalizeHz(
  x, 
  new = n, 
  pattern = p, 
  perl = TRUE
)

# double-check: OK
table(res, x)

## apply to a SoilProfileCollection
data(sp1)
depths(sp1) <- id ~ top + bottom

# must set horizon designation metadata
hzdesgnname(sp1) <- 'name'

# result is a SoilProfileCollection
x <- generalizeHz(
  sp1,
  new = c('O','A','B','C','R'),
  pattern = c('O', '^A','^B','C','R'),
  ordered = TRUE
)

# GHL stored in 'genhz' column
x$genhz

# GHL metadata is set
GHL(x)

}
\references{
Beaudette, D.E., Roudier, P., Skovlin, J. (2016). Probabilistic Representation of Genetic Soil Horizons. In: Hartemink, A., Minasny, B. (eds) Digital Soil Morphometrics. Progress in Soil Science. Springer, Cham. https://doi.org/10.1007/978-3-319-28295-4_18
}
\seealso{
\code{\link[=hzdesgnname]{hzdesgnname()}}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-operators.R
\docType{methods}
\name{$}
\alias{$}
\alias{$,SoilProfileCollection-method}
\title{Get data from column of horizon or site data in a SoilProfileCollection}
\usage{
\S4method{$}{SoilProfileCollection}(x, name)
}
\arguments{
\item{x}{a SoilProfileCollection}

\item{name}{a single column name in site or horizon table}
}
\description{
Get the data from a column accessed by name \code{x$name}. Column
names other than profile ID are not shared between site and horizons.
Accessing the profile ID column (\code{idname(x)}) with \code{$} returns the
horizon-level data with profile ID repeated for each horizon in a profile.
}
\examples{

data(sp1)

depths(sp1) <- id ~ top + bottom

# get data from a column by name (prop)
sp1$prop

}
\seealso{
\code{\link[=site]{site()}} \code{\link[=horizons]{horizons()}} \code{\link[=idname]{idname()}} \code{\link[=profile_id]{profile_id()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mollicEpipedon.R
\name{mollic.thickness.requirement}
\alias{mollic.thickness.requirement}
\title{Calculate the minimum thickness requirement for Mollic epipedon}
\usage{
mollic.thickness.requirement(
  p,
  hzdesgn = hzdesgnname(p, required = TRUE),
  texcl.attr = hztexclname(p, required = TRUE),
  clay.attr = hzmetaname(p, "clay", required = TRUE),
  truncate = TRUE
)
}
\arguments{
\item{p}{A single-profile SoilProfileCollection.}

\item{hzdesgn}{Column in horizon table containing designations. Default: \code{guessHzDesgnName(p)}}

\item{texcl.attr}{Column in horizon table containing texture classes. Default: \code{guessHzTexClName(p)}}

\item{clay.attr}{Column in horizon table containing clay contents. Default: \code{guessHzAttrName(p, 'clay', c('total','_r'))}}

\item{truncate}{Should sliding scale (Criterion 6C) results be truncated to 18 to 25cm interval? (Experimental; Default: TRUE)}
}
\value{
A unit length numeric vector containing Mollic or Umbric epipedon minimum thickness requirement.
}
\description{
Utilize horizon depths, designations and textures in a profile to estimate the thickness requirement for the Mollic or Umbric epipedon, per criterion 6 in the U.S. Keys to Soil Taxonomy (12th Edition).
}
\examples{
# construct a fake profile
spc <- data.frame(id=1, taxsubgrp = "Lithic Haploxeralfs",
                  hzname   = c("A","AB","Bt","BCt","R"),
                  hzdept   = c(0,  20, 32, 42,  49),
                  hzdepb   = c(20, 32, 42, 49, 200),
                  prop     = c(18, 22, 28, 24,  NA),
                  texcl    = c("l","l","cl", "l","br"),
                  d_value  = c(5,   5,  5,  6,  NA),
                  m_value  = c(2.5, 3,  3,  4,  NA),
                  m_chroma = c(2,   3,  4,  4,  NA))

# promote to SoilProfileCollection
depths(spc) <- id ~ hzdept + hzdepb
hzdesgnname(spc) <- 'hzname'
hztexclname(spc) <- 'texcl'

# print results in table
data.frame(id = spc[[idname(spc)]],
           thickness_req = mollic.thickness.requirement(spc, clay.attr='prop'),
           thickness_req_nobound = mollic.thickness.requirement(spc,
                                        clay.attr='prop', truncate=FALSE))

}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{aqp_df_class,SoilProfileCollection-method}
\alias{aqp_df_class,SoilProfileCollection-method}
\alias{aqp_df_class}
\alias{aqp_df_class<-,SoilProfileCollection-method}
\alias{aqp_df_class<-}
\title{Get aqp_df_class entry from metadata or return a safe value.}
\usage{
\S4method{aqp_df_class}{SoilProfileCollection}(object)

\S4method{aqp_df_class}{SoilProfileCollection}(object) <- value
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{value}{"data.frame", "data.table" or "tbl_df"}
}
\description{
This is an accessor and replacement method for the \code{aqp_df_class} entry in the metadata slot. This entry is used internally by methods that interact with \code{data.frame} objects and slots to ensure that the same class used to promote to the SoilProfileCollection initially is used throughout the process.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{metadata,SoilProfileCollection-method}
\alias{metadata,SoilProfileCollection-method}
\alias{metadata}
\alias{metadata<-,SoilProfileCollection-method}
\alias{metadata<-}
\title{Retrieve metadata from SoilProfileCollection}
\usage{
\S4method{metadata}{SoilProfileCollection}(object)

\S4method{metadata}{SoilProfileCollection}(object) <- value
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{value}{A named list (see examples)}
}
\description{
Get metadata from SoilProfileCollection. Result is a list. Two entries (aqp_df_class, depth_units) should not be edited in the metadata list directly. There are methods that facilitate changing them -- and propagating their changes throughout the collection. Otherwise, metadata list is a free-form slot used to store arbitrary information about the data, how it was collected, citations, etc.
}
\examples{

data(sp5)

# replace default metadata with itself
metadata(sp5) <- metadata(sp5)

# set new metadata attribute value
metadata(sp5)$newvalue <- 'foo'

# get metadata attribute
metadata(sp5)$newvalue

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimateAWC.R
\name{estimateAWC}
\alias{estimateAWC}
\title{Estimate available water capacity for fine-earth fraction}
\usage{
estimateAWC(texcl, omcl, precision = 2, FUN = mean, ...)
}
\arguments{
\item{texcl}{character, USDA textural class fine earth fraction}

\item{omcl}{integer, Organic matter class. 1: less than 1.5 percent, 2: less than 5, 3: greate than 5}

\item{precision}{integer, Number of decimal places in result default: 2}

\item{FUN}{Function for interpolating between table values default: \code{mean}}

\item{...}{Additional arguments to \code{FUN}}
}
\value{
A numeric vector double containing estimated available water capacities for fine-earth fraction.
}
\description{
Estimate available water capacity for fine-earth fraction
}
\examples{

# organic matter, loam texture, low medium and high OM
base.awc <- estimateAWC(c("l","l","l"), c(1, 2, 3), na.rm = TRUE)
base.awc

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_profile.R
\name{rp}
\alias{rp}
\title{Generate a \code{SoilProfileCollection} of random profiles}
\usage{
rp(size, prefix = NULL, ...)
}
\arguments{
\item{size}{integer, number of requested profiles}

\item{prefix}{prefix added to zero-padded, integer IDs}

\item{...}{additional arguments to \code{\link[=random_profile]{random_profile()}}}
}
\value{
a \code{SoilProfileCollection} object
}
\description{
This function provides a convenient abstraction of \code{lapply()}, \code{\link[=random_profile]{random_profile()}}, and \code{\link[=combine]{combine()}} which are typically used together to create a \code{SoilProfileCollection} object with >1 soil profiles. \code{rp()} creates zero-padded integer IDs for logical sorting and indexing of profiles. For more complex IDs or additional flexibility, see \code{\link[=random_profile]{random_profile()}}. See \code{\link[=random_profile]{random_profile()}} for all possible arguments.
}
\examples{

# generate a SoilProfileCollection object with 10 profiles
# using 0-padded, integer IDs for intuitive sorting
spc <- rp(10, method = 'LPP')
plotSPC(spc, color = 'p1')

# apply a prefix to the IDs
spc <- rp(10, prefix = 'A-', method = 'LPP')
plotSPC(spc, color = 'p1')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/groupSPC.R
\name{groupSPC}
\alias{groupSPC}
\title{Store groupings within a profile collection.}
\usage{
groupSPC(object, ...)
}
\arguments{
\item{object}{SoilProfileCollection.}

\item{...}{One or more expressions evaluated within the context of \code{object} that resolve to vectors that can be coerced to factor "groups."}
}
\description{
Store groupings within a profile collection.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/contrastChart.R
\name{contrastChart}
\alias{contrastChart}
\title{Color Contrast Chart}
\usage{
contrastChart(
  m,
  hues,
  ccAbbreviate = 1,
  style = "hue",
  gridLines = FALSE,
  de00.cex = 0.6,
  cc.cex = 0.6,
  thresh = NULL,
  returnData = FALSE
)
}
\arguments{
\item{m}{Munsell representation of a single color for comparison e.g. '10YR 4/3'}

\item{hues}{vector of one or more Munsell hue pages to display}

\item{ccAbbreviate}{length of abbreviated contrast classes, use 0 to suppress labels}

\item{style}{'hue' or 'CC', see details}

\item{gridLines}{logical, add grid lines to the color contrast chart}

\item{de00.cex}{character scaling applied to dE00 annotation}

\item{cc.cex}{character scaling applied to contrast class annotation}

\item{thresh}{threshold (<) applied to pair-wise comparisons and resulting color chips}

\item{returnData}{logical, return lattice figure + data used to generate the figure}
}
\description{
Compare one or more pages from a simulated Munsell book of soil colors to a reference color.
}
\details{
A simulated Munsell color book page or pages are used to demonstrate color contrast between all chips and the reference color \code{m} (highlighted in red). NCSS color contrast class and CIE delta-E00 values are printed below all other color chips. Munsell color chips for chroma 5 and 7 are omitted, but axis labels are retained as a reminder of this fact.

Setting \code{style='hue'} emphasizes the contrast classes and CIE delta-E00 of chips adjacent to \code{m}. Setting \code{style='CC'} emphasizes adjacent chips according to respective contrast class via lattice panels.

Two-way panels are used when multiple hues are provided and \code{style='CC'}. The default output can be greatly enhanced via:

\code{
  latticeExtra::useOuterStrips(...,
                               strip = strip.custom(bg=grey(0.85)),
                               strip.left = strip.custom(bg=grey(0.85))
                              )
  }
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# single hue page
contrastChart(m = '10YR 3/3', hues = '10YR')

# multiple hue pages
contrastChart(m = '10YR 3/3', hues = c('10YR', '2.5Y'))

# contrast class, single hue
contrastChart(m = '10YR 3/3', hues = '10YR', style='CC')

# contrast class, multiple hues
# consider latticeExtra::useOuterStrips()
contrastChart(m = '10YR 5/6', hues = c('10YR', '2.5Y'), style='CC')

}
\author{
D.E. Beaudette
}
\keyword{hplots}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slab-utils.R
\name{genSlabLabels}
\alias{genSlabLabels}
\title{Generate Labels for Slabs}
\usage{
genSlabLabels(
  slab.structure = 1,
  max.d = NULL,
  n.profiles = NULL,
  spc = NULL,
  diced = NULL,
  ...
)
}
\arguments{
\item{slab.structure}{A user-defined slab thickness (defined by an integer), or user-defined structure (numeric vector). See details for \code{slab()}.}

\item{max.d}{Maximum depth}

\item{n.profiles}{Number of profiles}

\item{spc}{Optional: A SoilProfileCollection}

\item{diced}{Optional: The \code{dice()}-ed horizon-level data.frame corresponding to \code{spc}}

\item{...}{Additional arguments passed to \code{dice()} when \code{spc} is specified.}
}
\value{
factor. slab IDs, labels are the segment top and bottom depth separated by \code{"-"}
}
\description{
This method is used by \code{\link[=slab]{slab()}} for generating labels that assign IDs to layers in a SoilProfileCollection
}
\details{
The new routine used in aqp 2.0 requires that, at a minimum, the \code{spc} and \code{slab.structure} arguments be specified.
}
\seealso{
\code{\link[=slab]{slab()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hzBoundary.R
\name{hzTopographyCodeToLineType}
\alias{hzTopographyCodeToLineType}
\title{Convert Horizon Boundary Topography to Line Type}
\usage{
hzTopographyCodeToLineType(
  x,
  codes = c("smooth", "wavy", "irregular", "broken"),
  lty = c(1, 2, 3, 4)
)
}
\arguments{
\item{x}{vector of boundary topography codes to be converted}

\item{codes}{character vector of topography terms ('smooth') or codes ('S'), case insensitive, see details}

\item{lty}{line types}
}
\value{
vector of line types with same length as \code{x}
}
\description{
This function will convert USDA-NCSS horizon boundary topography codes into line types, based on the \href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}.
}
\details{
Visualization of horizon boundary topography can be difficult, line type offers an additional visual cue. See \code{hzTopographyCodeToOffset} for an offset-based approach. Additional examples are available in the \href{https://ncss-tech.github.io/AQP/aqp/hz-boundaries.html}{Visualization of Horizon Boundaries tutorial}. Missing data in \code{x} (NA) or codes that are not defined in \code{codes} are returned as line type 1.

Either format (or mixture) are accepted, case insensitive:
\itemize{
\item terms: \code{c('smooth', 'wavy', 'irregular', 'broken')}
\item coded values: \code{c('s', 'w', 'i', 'b')}
}
}
\references{
\href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}
}
\seealso{
\code{\link{plotSPC}, \link{hzTopographyCodeToOffset}}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/munsell2rgb.R
\name{rgb2munsell}
\alias{rgb2munsell}
\title{sRGB to Munsell Color Conversion}
\usage{
rgb2munsell(color, colorSpace = c("CIE2000", "LAB", "sRGB"), nClosest = 1)
}
\arguments{
\item{color}{a \code{data.frame} or \code{matrix} object containing sRGB coordinates in the range of (0,1)}

\item{colorSpace}{distance metric (colorspace) to use for finding the closest chip: CIE2000 is the most accurate but requires farver >= 2.0.3, Euclidean distance in CIELAB is a close second, while Euclidean distance in sRGB is not at all accurate and should only be used for demonstration purposes.}

\item{nClosest}{number of closest Munsell colors to return (valid range is 1-20)}
}
\value{
an (NA-padded) \code{data.frame} containing \code{hue}, \code{value}, \code{chroma}, and distance (dE00 when \code{colorSpace = 'CIE2000'}, Euclidean distance otherwise) to nearest matching color.
}
\description{
Convert sRGB color coordinates to the closest \code{n} Munsell chips in the \code{munsell} lookup table. This function will be replaced by \code{col2Munsell()} in \strong{aqp 2.1}.
}
\note{
This function is fully vectorized and will pad output with NA-records when NA are present in \code{color}.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# Munsell notation to sRGB triplets [0-1] 
color <- munsell2rgb(
  the_hue = c('10YR', '2.5YR', '5YR'), 
  the_value = c(3, 5, 2.5), 
  the_chroma = c(5, 6, 2), 
  return_triplets = TRUE
)

# result is a data.frame
color

# back-transform sRGB -> closest Munsell color
# sigma is the dE00 color contrast metric
rgb2munsell(color)

}
\references{
\itemize{
\item http://ncss-tech.github.io/AQP/
\item http://www.brucelindbloom.com/index.html?ColorCalcHelp.html
\item http://www.munsellcolourscienceforpainters.com/MunsellAndKubelkaMunkToolbox/MunsellAndKubelkaMunkToolbox.html
\item http://www.cis.rit.edu/mcsl/online/munsell.php
}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/equivalentMunsellChips.R
\name{equivalentMunsellChips}
\alias{equivalentMunsellChips}
\title{Identify "equivalent" (whole number value/chroma) Munsell chips}
\usage{
equivalentMunsellChips(hue = NULL, value = NULL, chroma = NULL)
}
\arguments{
\item{hue}{A character vector containing Munsell hues}

\item{value}{A numeric vector containing Munsell values (integer only)}

\item{chroma}{A numeric vector containing Munsell chromas (integer only)}
}
\value{
A named list; Each list element contains a data.frame with one or more rows of "equivalent" Munsell, RGB and LAB color coordinates from \code{munsell} data set.
}
\description{
Uses a pre-calculated lookup list (\code{\link{equivalent_munsell}}) based on pair-wise CIE2000 contrast (\code{dE00}) of LAB color with D65 illuminant for all whole value/chroma "chips" in the \code{aqp::munsell} data set.

The intention is to identify Munsell chips that may be "functionally equivalent" to some other given whole value/chroma chip elsewhere in the Munsell color space -- as discretized in the \code{aqp::munsell} data table. This basic assumption needs to be validated against your end goal: probably by visual inspection of some or all of the resulting sets. See \code{\link{colorContrast}} and \code{\link{colorContrastPlot}}.

"Equivalent" chips table are based (fairly arbitrarily) on the 0.001 probability level of dE00 (default Type 7 \code{quantile}) within the upper triangle of the 8467x8467 contrast matrix. This corresponds to a \code{dE00} contrast threshold of approximately 2.16.
}
\examples{

# 7.5YR 4/4 (the one and only)

equivalentMunsellChips("7.5YR", 4, 4)
#>
#> $`7.5YR 4/4`
#>        hue value chroma         r        g         b        L       A       B
#> 8330 7.5YR     4      4 0.4923909 0.352334 0.2313328 41.26403 10.8689 23.5914

# 7.5YR 1/1 (two chips are equivalent; 3 row result)

equivalentMunsellChips("7.5YR", 1, 1)
#>
#> $`7.5YR 1/1`
#>        hue value chroma         r         g          b        L        A        B
#> 1983  10YR     1      1 0.1345633 0.1087014 0.07606787 10.64787 1.621323 6.847629
#> 6189   5YR     1      1 0.1330994 0.1076359 0.09450179 10.63901 2.489012 3.515146
#> 8303 7.5YR     1      1 0.1329483 0.1082380 0.08862581 10.64210 2.065514 4.623922

# 10YR 6/8 (two chips are equivalent; 3 row result)

equivalentMunsellChips("10YR", 6, 8)
#>
#> $`10YR 6/8`
#>       hue value chroma         r         g         b        L        A        B
#> 2039 10YR     6      7 0.7382230 0.5512957 0.2680260 61.76795 10.50886 44.78574
#> 2040 10YR     6      8 0.7519872 0.5472116 0.2157209 61.77496 11.83215 51.15496
#> 2041 10YR     6      9 0.7642826 0.5433189 0.1559069 61.78085 13.09599 57.49773

# compare visually a very red color

veryred <- equivalentMunsellChips("10R", 6, 28)[[1]]

par(mar=c(0,0,1,1))

pie(rep(1, nrow(veryred)), col = with(veryred, munsell2rgb(hue, value, chroma)),
    label = with(veryred, sprintf("\%s \%s/\%s", hue, value, chroma)))

table(veryred$hue) # 2 hues
#> 
#>  10R 7.5R 
#>    8   17

table(veryred$value) # 2 values
#> 
#>  5  6 
#> 11 14

table(veryred$chroma) # 10 chromas
#> 
#> 21 22 23 24 25 26 27 28 29 30 
#>  1  2  2  3  3  4  3  3  2  2

}
\references{
Gaurav Sharma, Wencheng Wu, Edul N. Dalal. (2005). The CIEDE2000 Color-Difference Formula: Implementation Notes, Supplementary Test Data, and Mathematical Observations. COLOR research and application. 30(1):21-30. http://www2.ece.rochester.edu/~gsharma/ciede2000/ciede2000noteCRNA.pdf

Thomas Lin Pedersen, Berendea Nicolae and Romain François (2020). farver: High Performance Colour Space Manipulation. R package version 2.0.3. https://CRAN.R-project.org/package=farver

Dong, C.E., Webb, J.B., Bottrell, M.C., Saginor, I., Lee, B.D. and Stern, L.A. (2020). Strengths, Limitations, and Recommendations for Instrumental Color Measurement in Forensic Soil Characterization. J Forensic Sci, 65: 438-449. https://doi.org/10.1111/1556-4029.14193
}
\seealso{
\code{\link{colorContrast}} \code{\link{colorContrastPlot}} \code{\link{equivalent_munsell}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evalMissingData.R
\name{evalMissingData}
\alias{evalMissingData}
\title{Evaluate Missing Data within a SoilProfileCollection}
\usage{
evalMissingData(
  x,
  vars,
  name = hzdesgnname(x),
  p = "Cr|R|Cd",
  method = c("relative", "absolute", "horizon")
)
}
\arguments{
\item{x}{\code{SoilProfileCollection} object}

\item{vars}{character vector, naming horizon-level attributes in \code{x}}

\item{name}{character, the name of a horizon-level attribute where horizon designations are stored, defaults to \code{hzdesgnname(x)}}

\item{p}{character, REGEX pattern used to match non-soil horizons}

\item{method}{character, one of: 'relative' (proportion of total) depth, 'absolute' depth, or 'horizon' (fraction not-missing by horizon)}
}
\value{
A vector values ranging from 0 to 1 (\code{method = 'relative'}) or 0 to maximum depth in specified depth units (\code{method = 'absolute'}) representing the quantity of non-missing data (as specified in \code{vars}) for each profile. When \code{method = 'horizon'} a non-missing data fraction is returned for each horizon.
}
\description{
Evaluate missing data within a \code{SoilProfileCollection} object

Data completeness is evaluated by profile or by horizon. Profile-level evaluation is based on the thickness of horizons (\code{method = absolute}) with complete horizon-level attributes (\code{vars}), optionally divided by the total thickness (\code{method = relative}). The REGEX pattern (\code{p}) is used to filter non-soil horizons from the calculation.
}
\examples{

# example data
data("jacobs2000")

# fully populated
plotSPC(jacobs2000, name.style = 'center-center', 
        cex.names = 0.8, color = 'time_saturated')

# missing some data
plotSPC(jacobs2000, name.style = 'center-center', 
        cex.names = 0.8, color = 'concentration_color')

# very nearly complete
plotSPC(jacobs2000, name.style = 'center-center', 
        cex.names = 0.8, color = 'matrix_color')


# variables to consider
v <- c('time_saturated', 'concentration_color', 'matrix_color')

# compute data completeness by profile
# ignore 2C horizons
jacobs2000$data.complete <- evalMissingData(
  jacobs2000, 
  vars = v, 
  method = 'relative',
  p = '2C'
)

jacobs2000$data.complete.abs <- evalMissingData(
  jacobs2000, 
  vars = v, 
  method = 'absolute',
  p = '2C'
)

# compute data completeness by horizon
# ignore 2C horizons
jacobs2000$hz.data.complete <- evalMissingData(
  jacobs2000, 
  vars = v, 
  method = 'horizon',
  p = '2C'
)


# "fraction complete" by horizon
plotSPC(
  jacobs2000, name.style = 'center-center', 
  cex.names = 0.8, color = 'hz.data.complete'
)


# rank on profile completeness
new.order <- order(jacobs2000$data.complete)

# plot along data completeness ranking
plotSPC(
  jacobs2000, name.style = 'center-center', 
  cex.names = 0.8, color = 'concentration_color', 
  plot.order = new.order
)

# add relative completeness axis
# note re-ordering of axis labels
axis(
  side = 1, at = 1:length(jacobs2000), 
  labels = round(jacobs2000$data.complete[new.order], 2),
  line = 0, cex.axis = 0.75
)

# add absolute completeness (cm)
axis(
  side = 1, at = 1:length(jacobs2000), 
  labels = jacobs2000$data.complete.abs[new.order],
  line = 2.5, cex.axis=0.75
)

}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evalGenHz.R
\name{evalGenHZ}
\alias{evalGenHZ}
\title{Evaluate Generalized Horizon Labels}
\usage{
evalGenHZ(
  obj,
  genhz = GHL(obj, required = TRUE),
  vars,
  non.matching.code = "not-used",
  stand = TRUE,
  metric = "euclidean"
)
}
\arguments{
\item{obj}{a \code{SoilProfileCollection} object}

\item{genhz}{name of horizon-level attribute containing generalized horizon labels}

\item{vars}{character vector of horizon-level attributes to include in the evaluation}

\item{non.matching.code}{code used to represent horizons not assigned a generalized horizon label}

\item{stand}{standardize variables before computing distance matrix, passed to \code{\link[cluster:daisy]{cluster::daisy()}}}

\item{metric}{distance metric, passed to \code{\link[cluster:daisy]{cluster::daisy()}}}
}
\value{
a list is returned containing:
\itemize{
\item horizons: \verb{c('mds.1', mds.2', 'sil.width', 'neighbor')}
\item stats: mean and standard deviation \code{vars}, computed by generalized horizon label
\item dist: the distance matrix as passed to \code{\link[stats:cmdscale]{stats::cmdscale()}}
}
}
\description{
Data-driven evaluation of generalized horizon labels using nMDS and
silhouette width.
}
\details{
Classic multidimensional scaling is performed via \code{\link[stats:cmdscale]{stats::cmdscale()}}.
The input distance matrix is generated by \code{\link[cluster:daisy]{cluster::daisy()}} using
(complete cases of) horizon-level attributes from \code{obj} as named in
\code{vars}.

Silhouette widths are computed via \code{\link[cluster:silhouette]{cluster::silhouette()}}. The input
distance matrix is generated by \code{\link[cluster:daisy]{cluster::daisy()}} using (complete cases
of) horizon-level attributes from \code{obj} as named in \code{vars}. Note
that observations with genhz labels specified in \code{non.matching.code}
are removed filtered before calculation of the distance matrix.
}
\seealso{
\code{\link[=get.ml.hz]{get.ml.hz()}}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-operators.R
\docType{methods}
\name{[[<-}
\alias{[[<-}
\alias{[[<-,SoilProfileCollection-method}
\alias{[[<-,SoilProfileCollection,ANY,ANY-method}
\title{Add or change column of horizon or site data in a SoilProfileCollection}
\usage{
\S4method{[[}{SoilProfileCollection}(x, i) <- value
}
\arguments{
\item{x}{a SoilProfileCollection}

\item{i}{an expression resolving to a single column name in site or horizon
table-}

\item{value}{New value to replace -- unit length or equal in length to number
of sites or horizons in the collection.}
}
\description{
Add or change the data from a column accessed by name. Column names other
than profile ID are not shared between site and horizons. The benefit of
using double bracket setter over \code{$} is that \code{name} can be
calculated, whereas with \code{$}, it must be known a priori and hard coded.

When using the double bracket setter the length of input and output matching
either the number of sites or number of horizons is used to determine which
slot new columns are assigned to.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/texture.R
\name{lookup_taxpartsize}
\alias{lookup_taxpartsize}
\title{Ranking Systems for USDA Taxonomic Particle-Size and Substitute Classes of Mineral Soils}
\usage{
lookup_taxpartsize()
}
\value{
A data.frame with a rank column, taxonomic family particle size class, and a flag for contrasting.
}
\description{
Generate a lookup table of USDA Particle-Size and Substitute Classes  names, ranked according to approximate particle size
}
\examples{

# class codes
lu <- lookup_taxpartsize()

idx <- lu$contrasting == FALSE

lu$taxpartsize[idx]

lu$rank[as.integer(lu$taxpartsize)[idx]]

}
\references{
\href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}
}
\seealso{
\code{\link[=hz_to_taxpartsize]{hz_to_taxpartsize()}}, \code{\link[=texture_to_taxpartsize]{texture_to_taxpartsize()}}, \code{\link[=SoilTextureLevels]{SoilTextureLevels()}}
}
\author{
Stephen Roecker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NCSP.R
\name{NCSP}
\alias{NCSP}
\title{Numerical Classification of Soil Profiles}
\usage{
NCSP(
  x,
  vars,
  fm = NULL,
  weights = rep(1, times = length(vars)),
  maxDepth = max(x),
  k = 0,
  isColor = FALSE,
  rescaleResult = FALSE,
  verbose = TRUE,
  returnDepthDistances = FALSE
)
}
\arguments{
\item{x}{\code{SoilProfileColection} object, should be pre-filtered to remove profiles with horizon depth logic, see \code{\link{HzDepthLogicSubset}}}

\item{vars}{character vector, names of horizon attributes to use in the classification}

\item{fm}{formula, formula as specified to \code{\link[=dice]{dice()}}, not yet implemented}

\item{weights}{numeric vector, same length as \code{vars}: variable importance weights, need not sum to 1}

\item{maxDepth}{numeric, maximum depth of analysis}

\item{k}{numeric, weighting coefficient, usually between 0-1. A value of 0 results in no depth-weighting. See examples.}

\item{isColor, }{logical: variables represent color, should be CIELAB coordinates (D65 illuminant), weights are ignored. Variables should be named \code{L}, \code{A}, \code{B} in specified in that order.}

\item{rescaleResult}{logical, distance matrix is rescaled based on max(D)}

\item{verbose}{logical, extra output messages}

\item{returnDepthDistances}{logical, return a list of distances by depth slice}
}
\description{
Replacement for \code{profile_compare()}.

Performs a numerical comparison of soil profiles using named properties,
based on a weighted, summed, depth-segment-aligned dissimilarity calculation.

Variability in soil depth can interfere significantly with the calculation
of between-profile dissimilarity--what is the numerical ``distance'' (or
dissimilarity) between a slice of soil from profile A and the corresponding,
but missing, slice from a shallower profile B? Gower's distance metric would
yield a NULL distance, despite the fact that intuition suggests otherwise:
shallower soils should be more dissimilar from deeper soils. For example,
when a 25 cm deep profile is compared with a 50 cm deep profile, numerical
distances are only accumulated for the first 25 cm of soil (distances from
26 - 50 cm are NULL). When summed, the total distance between these profiles
will generally be less than the distance between two profiles of equal
depth. Our algorithm will replace NULL
distances with the maximum distance between any pair of profiles for the
current depth slice. In this way, the numerical distance between a slice of
soil and a corresponding slice of non-soil reflects the fact that these two
materials should be treated very differently.

This alternative calculation of dissimilarities between soil and non-soil
slices solves the problem of comparing shallow profiles with deeper
profiles. However, it can result in a new problem: distances calculated
between two shallow profiles will be erroneously inflated beyond the extent
of either profile's depth. Our algorithm will preserve NULL distances between slices when
both slices represent non-soil material. Therefore, shallow
profiles will only accumulate mutual dissimilarity to the depth of the
deeper profile.

Slices are classified as 'soil' down to the maximum depth to which at least one of
variables used in the dissimilarity calculation is not NA. This will cause
problems when profiles within a collection contain all NAs within the
columns used to determine dissimilarity. An approach for identifying and
removing these kind of profiles is presented in the examples section below.

Our approach builds on the work of (Moore, 1972) and the previously
mentioned depth-slicing algorithm. See references below for a detailed explanation of the NCSP algorithm.
}
\note{
\code{NCSP()} will overwrite the \code{removed.profiles} metadata from \code{x}.
}
\references{
\itemize{
\item J.J Maynard, S.W. Salley, D.E. Beaudette, J.E Herrick. Numerical soil classification supports soil identification by citizen scientists using limited, simple soil observations. Soil Sci. Soc. Am. J. 2020; 84: 1675-1692. \doi{10.1002/saj2.20119}.
\item D.E. Beaudette, P. Roudier, A.T. O'Geen, Algorithms for quantitative pedology: A toolkit for soil scientists, Computers & Geosciences, Volume 52, 2013, Pages 258-268, ISSN 0098-3004, \doi{10.1016/j.cageo.2012.10.020}.
\item Moore, A.; Russell, J. & Ward, W. Numerical analysis of soils: A comparison of three soil profile models with field classification. Journal of Soil Science, 1972, 23, 194-209.
}
}
\seealso{
\code{\link[=dice]{dice()}}, \code{\link[cluster:daisy]{cluster::daisy()}}, \code{\link[=compareSites]{compareSites()}}
}
\author{
Dylan E. Beaudette and Jon Maynard
}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compositeSPC.R
\name{compositeSPC}
\alias{compositeSPC}
\alias{compositeSPC,SoilProfileCollection-method}
\title{Return a list representation of site and horizon level data}
\usage{
compositeSPC(object)
}
\arguments{
\item{object}{A SoilProfileCollection}
}
\value{
A list.
}
\description{
\code{compositeSPC()} is a convenience function that returns a named list representation of the columns from the \code{@site} and \code{@horizons} slots.
}
\author{
Andrew G. Brown.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/col2Munsell.R
\name{col2Munsell}
\alias{col2Munsell}
\title{Convert colors into Munsell Notation}
\usage{
col2Munsell(col, space = c("sRGB", "CIELAB"), nClosest = 1)
}
\arguments{
\item{col}{character vector of colors, \code{data.frame} or \code{matrix} of color coordinates in sRGB or CIELAB color space}

\item{space}{character, one of \code{sRGB} or \code{CIELAB}, defines the input color system}

\item{nClosest}{integer, number of closest Munsell colors to return (valid range is 1-20)}
}
\value{
an (NA-padded) \code{data.frame} containing \code{hue}, \code{value}, \code{chroma}, and CIE delta-E 2000 color contrast metric between source and nearest matching color(s).
}
\description{
Lookup the \code{n} closest Munsell chips from the \code{munsell} lookup table from various color notations. This function replaces \code{rgb2munsell()}.
}
\note{
This function is fully vectorized and will pad output with NA-records when NA are present in \code{color}.
}
\examples{

# vector of named R colors
col2Munsell(c('red', 'green', 'blue'))

# sRGB matrix in the range of 0-255
col2Munsell(cbind(255, 0, 0))

# sRGB matrix in the range of 0-1
col2Munsell(cbind(1, 0, 0))

# 10YR 5/6 in CIELAB
col2Munsell(
  cbind(51.4337, 9.917916, 38.6889), 
  space = 'CIELAB'
)

# 2.5YR 6/8 in hex notation
col2Munsell("#D18158FF")

# 7.5YR 8/1 in sRGB {0, 1}
col2Munsell(
  cbind(0.8240707, 0.7856834, 0.7541048)
)

# 7.5YR 8/1 in sRGB {0, 255}
col2Munsell(
  cbind(0.8240707, 0.7856834, 0.7541048) * 255
)

# multple colors in CIELAB
col2Munsell(
  parseMunsell(c('10BG 6/6', '2.5YR 4/6'), returnLAB = TRUE),
  space = 'CIELAB'
)

# data.frame input
col2Munsell(
  data.frame(r = 1, g = 0, b = 0),
  space = 'sRGB'
)

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# Munsell notation to sRGB triplets {0, 1} 
color <- munsell2rgb(
  the_hue = c('10YR', '2.5YR', '5YR'), 
  the_value = c(3, 5, 2.5), 
  the_chroma = c(5, 6, 2), 
  return_triplets = TRUE
)

# result is a data.frame of sRGB {0, 1}
color

# back-transform sRGB -> closest Munsell color
# sigma is the dE00 color contrast metric
col2Munsell(color, space = 'sRGB')

}
\references{
\itemize{
\item http://ncss-tech.github.io/AQP/
\item http://www.brucelindbloom.com/index.html?ColorCalcHelp.html
\item http://www.munsellcolourscienceforpainters.com/MunsellAndKubelkaMunkToolbox/MunsellAndKubelkaMunkToolbox.html
\item http://www.cis.rit.edu/mcsl/online/munsell.php
}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getArgillicBounds.R
\name{crit.clay.argillic}
\alias{crit.clay.argillic}
\title{Determines threshold (minimum) clay content for argillic upper bound}
\usage{
crit.clay.argillic(eluvial_clay_content)
}
\arguments{
\item{eluvial_clay_content}{A numeric vector or matrix containing clay
contents of potential "eluvial" horizons. May contain \code{NA}.}
}
\value{
A vector or matrix (input-dependent) containing minimum "illuvial"
horizon clay contents (thresholds) to be met for argillic horizon clay
increase.
}
\description{
Given a vector or matrix of "eluvial" horizon clay contents (\\%),
\code{crit.clay.argillic()} returns a vector or matrix of minimum clay
contents (thresholds) that must be met for an argillic horizon clay
increase.
}
\details{
Uses the standard equations for clay contents less than 15 \\%, between 15
and 40 \\%, and greater than 40 \\%. Based on the clay increase criteria in
the definition of the argillic horizon from 12th Edition Keys to Soil
Taxonomy (Soil Survey Staff, 2014).
}
\note{
This function is intended for identifying clay content threshold
required for an argillic horizon. These thresholds may not apply depending
on the specifics of your soil. E.g. if the upper part of argillic has been
plowed (has Ap immediately over upper boundary) the clay increase
requirement can be waived (Soil Survey Staff, 2014).
}
\examples{

# crit.clay.argillic uses different equations for clay content
# less than 15 \%, between 15 and 40 \%, and >40 \%

crit.clay.argillic(eluvial_clay_content=c(5, 20, 45))

}
\references{
Soil Survey Staff. 2014. Keys to Soil Taxonomy, 12th ed.
USDA-Natural Resources Conservation Service, Washington, DC.
}
\seealso{
\code{\link{getArgillicBounds}}, \code{\link{get.increase.matrix}}
}
\author{
Andrew Gene Brown
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimateAWC.R
\name{correctAWC}
\alias{correctAWC}
\title{Apply rock fragment or salt correction to available water content}
\usage{
correctAWC(
  awc,
  total_rf = numeric(length(awc)),
  gravel = NULL,
  ec = NULL,
  nullFragsAreZero = TRUE
)
}
\arguments{
\item{awc}{Numeric vector of available water capacities (e.g. from \code{estimateAWC})}

\item{total_rf}{Numeric vector  of rock fragment volume percentage, 0 - 100}

\item{gravel}{Numeric vector of gravel volume percentage, 0 - 100}

\item{ec}{Numeric vector of electrical conductivity, mmhos/cm}

\item{nullFragsAreZero}{Interpret \code{NA} in \code{total_rf}, \code{gravel} or \code{ec} as \code{0}? Default: \code{TRUE}}
}
\value{
A numeric vector (double) containing estimated available water capacities corrected for rock fragments and salts
}
\description{
Apply rock fragment or salt correction to available water content
}
\examples{

# medium organic matter, loam texture 
base.awc <- 0.18 # estimateAWC(texcl = "l", omcl = 2, na.rm = TRUE)

# medium organic matter, loam texture w/ 23\% rock fragments by volume 
corrected.awc <- correctAWC(base.awc, total_rf = 23)
corrected.awc

# medium organic matter, loam texture w/ 0\% frags by volume and 8 mmhos/cm salts
salty.awc <- correctAWC(base.awc, total_rf = 0, ec = 8)
salty.awc
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/soilColorIndices.R
\name{barron.torrent.redness.LAB}
\alias{barron.torrent.redness.LAB}
\title{Barron & Torrent (1986) Redness Index in LAB color space}
\usage{
barron.torrent.redness.LAB(hue, value, chroma)
}
\arguments{
\item{hue}{A character vector containing Munsell hues (e.g. "7.5YR")}

\item{value}{A numeric vector containing Munsell values}

\item{chroma}{A numeric vector containing Munsell chromas}
}
\value{
A numeric vector of horizon redness index (higher values = redder).
}
\description{
Calculate Redness Index after Barron & Torrent (1986) "Use of the Kubelka—Munk Theory to Study the Influence of Iron Oxides on Soil Colour" using Munsell colors converted to LAB. DOI: 10.1111/j.1365-2389.1986.tb00382.x. Accepts vectorized inputs for hue, value and chroma, produces vector output.
}
\references{
Barron, V. and Torrent, J. (1986), Use of the Kubelka-Munk theory to study the influence of iron oxides on soil colour. Journal of Soil Science, 37: 499-510. doi:10.1111/j.1365-2389.1986.tb00382.x
}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collapseHz.R
\name{collapseHz}
\alias{collapseHz}
\title{Collapse Horizons within Profiles Based on Pattern Matching}
\usage{
collapseHz(
  x,
  pattern = NULL,
  by = NULL,
  hzdesgn = hzdesgnname(x, required = TRUE),
  FUN = function(x, pattern, hzdesgn, ...) grepl(pattern, x[[hzdesgn]], ignore.case =
    FALSE),
  ...,
  AGGFUN = NULL,
  ignore_numerics = NULL,
  na.rm = FALSE
)
}
\arguments{
\item{x}{A \emph{SoilProfileCollection}}

\item{pattern}{\emph{character}. A regular expression pattern to match in \code{hzdesgn} column. Default:
\code{NULL}.}

\item{by}{\emph{character}. A column name specifying horizons that should be combined. Aggregation
will be applied to adjacent groups of layers within profiles that have the same value in \code{by}.
Used in lieu of \code{pattern} and \code{hzdesgn}. Default: \code{NULL}.}

\item{hzdesgn}{\emph{character}. Any character column containing horizon-level identifiers. Default:
\code{hzdesgnname(x, required = TRUE)}.}

\item{FUN}{\emph{function}. A function that returns a \emph{logical} vector equal in length to the number
of horizons in \code{x}. Used only when \code{pattern} is specified. See details.}

\item{...}{Additional arguments passed to the matching function \code{FUN}.}

\item{AGGFUN}{\emph{list}. A \emph{named} list containing custom aggregation functions. List element names
should match the column name that they transform. The functions defined should take three
arguments: \code{x} (a vector of horizon property values), \code{top} (a vector of top depths), and
\code{bottom} (a vector of bottom depths). Default: \code{NULL} applies \code{weighted.mean()} to all numeric
columns not listed in \code{ignore_numerics} and takes the dominant condition (value with greatest
aggregate thickness sum) for all other columns. See details.}

\item{ignore_numerics}{\emph{character}. Vector of column names that contain numeric values which
should \emph{not} be aggregated using \code{weighted.mean()}. For example, soil color "value" and
"chroma".}

\item{na.rm}{\emph{logical}. If \code{TRUE} \code{NA} values are ignored when calculating min/max boundaries for
each group and in weighted averages. If \code{FALSE} \code{NA} values are propagated to the result.
Default: \code{FALSE}.}
}
\value{
A \emph{SoilProfileCollection}
}
\description{
Combines layers and aggregates data by grouping adjacent horizons which match \code{pattern} in
\code{hzdesgn} or, alternately, share a common value in \code{by} argument. Numeric properties are combined
using the weighted average, and other properties are derived from the dominant condition based on
thickness of layers and values in each group.
}
\details{
If a custom matching function (\code{FUN}) is used, it should accept arbitrary additional arguments
via an ellipsis (\code{...}). It is not necessary to do anything with arguments, but the result should
match the number of horizons found in the input SoilProfileCollection \code{x}.

Custom aggregation functions defined in the \code{AGGFUN} argument should either return a single
vector value for each group*column combination, or should return a \emph{data.frame} object with named
columns. If the input column name is used as a column name in the result \emph{data.frame}, then the
values of that column name in the result \emph{SoilProfileCollection} will be replaced by the output
of the aggregation function. See examples.
}
\examples{
data(jacobs2000)

# calculate a new SPC with genhz column based on patterns
new_labels <- c("A", "E", "Bt", "Bh", "C")
patterns <- c("A", "E", "B.*t", "B.*h", "C")
jacobs2000_gen <- generalizeHz(jacobs2000, new = new_labels, pattern = patterns)

# use existing generalized horizon labels
i <- collapseHz(jacobs2000_gen, by = "genhz")

profile_id(i) <- paste0(profile_id(i), "_collapse")

plot(
  c(i, jacobs2000),
  color = "genhz",
  name = "name",
  name.style = "center-center",
  cex.names = 1
)
 
# custom pattern argument  
j <- collapseHz(jacobs2000,
                c(
                  `A` = "^A",
                  `E` = "E",
                  `Bt` = "[ABC]+t",
                  `C` = "^C",
                  `foo` = "bar"
                ))
profile_id(j) <- paste0(profile_id(j), "_collapse")
plot(c(j, jacobs2000), color = "clay")

# custom aggregation function for matrix_color_munsell
k <- collapseHz(jacobs2000,
                pattern = c(
                  `A` = "^A",
                  `E` = "E",
                  `Bt` = "[ABC]+t",
                  `C` = "^C",
                  `foo` = "bar"
                ),
                AGGFUN = list(
                  matrix_color_munsell = function(x, top, bottom) {
                    thk <- bottom - top
                    if (length(x) > 1) {
                      xord <- order(thk, decreasing = TRUE)
                      paste0(paste0(x[xord], " (t=", thk[xord], ")"), collapse = ", ")
                    } else
                      x
                  }
                )
              )
profile_id(k) <- paste0(profile_id(k), "_collapse_custom")

unique(k$matrix_color_munsell)

# custom aggregation function for matrix_color_munsell (returns data.frame)
m <- collapseHz(jacobs2000,
                pattern = c(
                  `A` = "^A",
                  `E` = "E",
                  `Bt` = "[ABC]+t",
                  `C` = "^C",
                  `foo` = "bar"
                ),
                AGGFUN = list(
                  matrix_color_munsell = function(x, top, bottom) {
                    thk <- bottom - top
                    if (length(x) > 1) {
                      xord <- order(thk, decreasing = TRUE)
                      data.frame(matrix_color_munsell = paste0(x, collapse = ";"),
                                 n_matrix_color = length(x))
                    } else {
                      data.frame(matrix_color_munsell = x, 
                                 n_matrix_color = length(x))
                    }
                  }
                )
              )
profile_id(m) <- paste0(profile_id(m), "_collapse_custom")

m$matrix_color_munsell.n_matrix_color
}
\seealso{
\code{hz_dissolve()}
}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aqp-label-placement-solvers.R
\name{fixOverlap}
\alias{fixOverlap}
\title{Fix Overlap within a Sequence}
\usage{
fixOverlap(x, thresh = 0.6, method = c("S", "E"), trace = FALSE, ...)
}
\arguments{
\item{x}{vector of initial positions, pre-sorted}

\item{thresh}{numeric, overlap threshold defined on the same scale as \code{x}}

\item{method}{character vector, 'S' for simulated annealing via \code{\link[=SANN_1D]{SANN_1D()}} or 'E' for electrostatic simulation via \code{\link[=electroStatics_1D]{electroStatics_1D()}}}

\item{trace}{logical, return full output}

\item{\dots}{additional arguments to \code{\link[=SANN_1D]{SANN_1D()}} or \code{\link[=electroStatics_1D]{electroStatics_1D()}}}
}
\value{
When \code{trace = FALSE}, a vector of the same length as \code{x}, preserving rank-ordering and boundary conditions. When \code{trace = TRUE} a list containing the new sequence along with information about objective functions and decisions made during adjustment of \code{x}.
}
\description{
Fix Overlap within a Sequence
}
\examples{

s <- c(1, 2, 2.3, 4, 5, 5, 7)

# simulated annealing, solution is non-deterministic
fixOverlap(s, thresh = 0.6, method = 'S')

# electrostatics-inspired simulation of particles
# solution is deterministic
fixOverlap(s, thresh = 0.6, method = 'E')


# create a very busy profile with lots of possible overlapping
# depth annotation
x <- quickSPC(
  "SPC:AAA|BBB|CCC|D|EEEEE|FF|GG|HH|I|I|JJ|KK|LL|M|N|O|P|QQQQ|RR|S|TTTTTT|U", 
  interval = 1
)

# convert horizon ID to numeric
x$z <- as.numeric(x$hzID)

# plotSPC arguments
.a <- list(
  width = 0.2, 
  hz.depths = TRUE, 
  name.style = 'center-center', 
  cex.names = 1.5, 
  depth.axis = FALSE, 
  name = NA,
  color = 'z',
  show.legend = FALSE,
  print.id = FALSE,
  col.palette = hcl.colors(n = 25, palette = 'Spectral', rev = TRUE)
)

# set plotSPC default arguments
options(.aqp.plotSPC.args = .a)

# wrapper function to test label collision solutions
testIt <- function(x, ...) {
  
  plotSPC(x, ...)
  
  # a normalized index of label adjustment
  .txt <- sprintf(
    "LAI: \%0.3f", 
    get('last_spc_plot', envir = aqp.env)$hz.depth.LAI
  )
  mtext(.txt, side = 1, at = 1, line = -2, cex = 0.8)
  
}


# compare and contrast
op <- par(mar = c(0, 0, 0, 0), mfcol = c(1, 6))

testIt(x)
title('ES (defaults)', line = -3)

testIt(x, fixOverlapArgs = list(method = 'S'))
title('SANN (defaults)', line = -3)

testIt(x, fixOverlapArgs = list(method = 'E', q = 1.5))
title('ES (q = 1.5)', line = -3)

testIt(x, fixOverlapArgs = list(method = 'E', q = 1))
title('ES (q = 1)', line = -3)

testIt(x, fixOverlapArgs = list(method = 'E', q = 0.5))
title('ES (q = 0.5)', line = -3)

testIt(x, fixOverlapArgs = list(method = 'E', q = 0.1))
title('ES (q = 0.1)', line = -3)

par(op)

}
\seealso{
\code{\link[=electroStatics_1D]{electroStatics_1D()}}, \code{\link[=SANN_1D]{SANN_1D()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reduceSPC.R
\name{reduceSPC}
\alias{reduceSPC}
\title{Select a subset of columns from a SoilProfileCollection}
\usage{
reduceSPC(p, column_names = NULL)
}
\arguments{
\item{p}{a \code{SoilProfileCollection}}

\item{column_names}{a set of additional columns to include in the result}
}
\value{
a \code{SoilProfileCollection}
}
\description{
Reduce the number of columns in a \code{SoilProfileCollection} to a minimal set plus additional selected columns. Optional metadata columns are included if set. At a minimum the profile ID, horizon top and bottom depths, and horizon ID are included. Horizon designation and horizon texture class column names are included if metadata attributes are set. See details.
}
\details{
Minimum column names included (when \code{column_names = NULL})
\itemize{
\item \code{idname(p)}, \code{horizonDepths(p)}, \code{hzidname(p)}
}

Optional column names included (when metadata are set)
\itemize{
\item \code{hzdesgnname(p)}, \code{hztexclname(p)}, \code{GHL(p)}
}
}
\seealso{
\code{hzdesgnname()} \code{hztexclname()} \code{GHL()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{rowley2019}
\alias{rowley2019}
\title{Soil Morphologic, Geochemical, and Mineralogy Data from Rowley et al. 2019.}
\format{
A \code{SoilProfileCollection} object:

site-level attributes \describe{ \item{id}{profile ID} \item{group}{profile
group} }

horizon-level attributes \describe{ \item{sample_id}{sample ID}
\item{name}{horizon name} \item{pH}{pH} \item{Al_exch}{cmol(+) / kg,
exchangeable Al} \item{Ca_exch}{cmol(+) / kg, exchangeable Ca}
\item{CEC_sum}{cmol(+) / kg, cation exchange capacity calculated as the sum
of exchangeable cations, not including H+}
\item{Ca_exch_saturation}{percent} \item{Al_exch_saturation}{percent}
\item{TON}{percent, total nitrogen} \item{SOC}{percent, soil organic carbon}
\item{C_to_N}{carbon to nitrogen ratio} \item{Alo}{g/kg, oxalate-extractable
Al} \item{Feo}{g/kg, oxalate-extractable Fe} \item{Ald}{g/kg,
dithionite-extractable Al} \item{Fed}{g/kg, dithionite-extractable Fe}
\item{Feo_Fed}{Fe_o to Fe_d ratio} \item{id}{profile ID} \item{top}{horizon
top (cm)} \item{bottom}{horizon bottom (cm)} \item{Al}{g/kg by x-ray
fluorescence} \item{Ca}{g/kg by x-ray fluorescence} \item{Cr}{g/kg by x-ray
fluorescence} \item{Fe}{g/kg by x-ray fluorescence} \item{K}{g/kg by x-ray
fluorescence} \item{Mg}{g/kg by x-ray fluorescence} \item{Mn}{g/kg by x-ray
fluorescence} \item{Na}{g/kg by x-ray fluorescence} \item{Ni}{g/kg by x-ray
fluorescence} \item{P}{g/kg by x-ray fluorescence} \item{Si}{g/kg by x-ray
fluorescence} \item{Ti}{g/kg by x-ray fluorescence}
\item{Phyllosilicates}{percent by x-ray diffraction spectra}
\item{Quartz}{percent by x-ray diffraction spectra}
\item{K_Feldspar}{percent by x-ray diffraction spectra}
\item{Na_Plagioclase}{percent by x-ray diffraction spectra}
\item{Goethite}{percent by x-ray diffraction spectra}
\item{Unidentified}{percent by x-ray diffraction spectra}
\item{CCE_Total}{percent} \item{CCE_Reactive}{percent}
\item{Reactive_carbonate}{percent} \item{Sand}{percent <2um}
\item{Silt}{percent 2-50um} \item{Clay}{percent 50-2000um}

\item{CaH2O}{Milliq ex: grams of Ca per kilogram of dry soil (g kg-1)}
\item{Ca2MKCl}{2M KCl: grams of Ca per kilogram of dry soil (g kg-1)}
\item{CaNa2EDTA}{0.05 M Na2EDTA: grams of Ca per kilogram of dry soil (g
kg-1)} \item{CaCuCl2}{0.5 M CuCl2: grams of Ca per kilogram of dry soil (g
kg-1)}

\item{hzID}{horizon ID} }
}
\usage{
data(rowley2019)
}
\description{
Data from Table 1 and Supplementary Tables 1 and 2 from "A cascading
influence of calcium carbonate on the biogeochemistry and pedogenic
trajectories of subalpine soils, Switzerland".
}
\examples{

library(lattice)

# load data
data('rowley2019')

# check first 5 rows and 10 columns of horizon data
horizons(rowley2019)[1:5, 1:10]

# check site data
site(rowley2019)

# graphical summary
par(mar=c(1,1,3,1))
plotSPC(rowley2019, color='Feo_Fed', name='name', cex.names=0.85)

plotSPC(rowley2019, color='Ca_exch', name='name', cex.names=0.85)

# grouped plot
groupedProfilePlot(rowley2019, groups = 'group', color='Ca_exch',
name='name', cex.names=0.85, group.name.offset = -10)

# aggregate over 1cm slices, for select properties
a <- slab(rowley2019, group ~ Reactive_carbonate + Ca_exch + pH + K_Feldspar + Na_Plagioclase + Al)

# plot styling
tps <- list(superpose.line=list(lwd=2, col=c('royalblue', 'firebrick')))

# make the figure
xyplot(top ~ p.q50 | variable, data=a, ylab='Depth', groups=group,
       main='', as.table=TRUE,
       xlab='median bounded by 25th and 75th percentiles',
       lower=a$p.q25, upper=a$p.q75, ylim=c(55,-5),
       panel=panel.depth_function,
       prepanel=prepanel.depth_function,
       cf=a$contributing_fraction,
       alpha=0.33, sync.colors=TRUE,
       scales=list(x=list(relation='free', alternating=1)),
       par.settings=tps,
       auto.key=list(columns=2, lines=TRUE, points=FALSE),
       strip=strip.custom(bg=grey(0.9))
)


}
\references{
Mike C. Rowley, Stephanie Grand, Thierry Adatte, Eric P.
Verrecchia, Cascading influence of calcium carbonate on the biogeochemistry and
pedogenic trajectories of subalpine soils), Switzerland, Geoderma, 2019,
114065, ISSN 0016-7061, \doi{10.1016/j.geoderma.2019.114065}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getCambicBounds.R
\name{getCambicBounds}
\alias{getCambicBounds}
\title{Find all intervals that are potentially part of a Cambic horizon}
\usage{
getCambicBounds(
  p,
  hzdesgn = hzdesgnname(p, required = TRUE),
  texcl.attr = hztexclname(p, required = TRUE),
  clay.attr = hzmetaname(p, "clay", required = TRUE),
  argi_bounds = NULL,
  d_value = "d_value",
  m_value = "m_value",
  m_chroma = "m_chroma",
  sandy.texture.pattern = "-S$|^S$|COS$|L[^V]FS$|[^L]VFS$|LS$|LFS$",
  ...
)
}
\arguments{
\item{p}{A single-profile SoilProfileCollection}

\item{hzdesgn}{Column name containing horizon designations.}

\item{texcl.attr}{Arguments to \code{getArgillicBounds()}}

\item{clay.attr}{Arguments to \code{getArgillicBounds()}}

\item{argi_bounds}{Optional: numeric vector of length 2 with top and bottom of argillic; (Default: NULL)}

\item{d_value}{Column name containing dry value. Default: d_value}

\item{m_value}{Column name containing moist value. Default: m_value}

\item{m_chroma}{Column name containing moist chroma Default: m_chroma}

\item{sandy.texture.pattern}{this is a pattern for matching sandy textural classes: \verb{-S$|^S$|COS$|L[^V]FS$|[^L]VFS$|LS$|LFS$}}

\item{...}{Arguments to \code{getArgillicBounds()}}
}
\value{
A \code{data.frame} containing profile, cambic indexes, along with top and bottom depths.
}
\description{
Find all intervals that are potentially part of a Cambic horizon excluding those that are part of an argillic horizon (defined either by depth interval or \code{getArgillicBounds()}).

There may be multiple cambic horizons (indexes) in a profile. Each cambic index has a top and bottom depth associated: cambic_top and cambic_bottom. This result is designed to be used for single profiles, or with \code{profileApply(..., frameify = TRUE)}
}
\examples{
# construct a fake profile
spc <- data.frame(id=1, taxsubgrp = "Lithic Haploxerepts",
                  hzname   = c("A","AB","Bw","BC","R"),
                  hzdept   = c(0,  20, 32, 42,  49),
                  hzdepb   = c(20, 32, 42, 49, 200),
                  clay     = c(19, 22, 22, 21,  NA),
                  texcl    = c("l","l","l", "l","br"),
                  d_value  = c(5,   5,  5,  6,  NA),
                  m_value  = c(2.5, 3,  3,  4,  NA),
                  m_chroma = c(2,   3,  4,  4,  NA))

# promote to SoilProfileCollection
depths(spc) <- id ~ hzdept + hzdepb

# set required metadata
hzdesgnname(spc) <- 'hzname'
hztexclname(spc) <- 'texcl'
hzmetaname(spc, 'clay') <- 'clay'

# print results in table
getCambicBounds(spc)

}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine.R
\name{c,SoilProfileCollection-method}
\alias{c,SoilProfileCollection-method}
\alias{combine}
\alias{c}
\alias{combine,SoilProfileCollection-method}
\alias{combine,list-method}
\title{Combine SoilProfileCollection objects}
\usage{
\S4method{c}{SoilProfileCollection}(x, ...)

\S4method{combine}{SoilProfileCollection}(...)

\S4method{combine}{list}(...)
}
\arguments{
\item{x}{A \code{SoilProfileCollection}}

\item{...}{\code{SoilProfileCollection} objects}
}
\value{
A SoilProfileCollection
}
\description{
Combine \code{SoilProfileCollection} objects or lists of \code{SoilProfileCollection} objects. This method provides \code{...} expansion for the \code{pbindlist} method.
}
\examples{

# example data
spc1 <- random_profile(1, SPC = TRUE)
spc2 <- random_profile(2, SPC = TRUE)
spc3 <- random_profile('A', SPC = TRUE)

# combine into a single SPC, ... interface
spc <- combine(spc1, spc2, spc3)

# combine into a single SPC, list interface
spc <- combine(list(spc1, spc2, spc3))

# input are combined into a single SPC
spc <- c(spc1, spc2, spc3)

# result is a list when a mixture of objects are provided
spc <- c(spc1, bar=spc2, baz="foo")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/harmonize.R
\name{harmonize,SoilProfileCollection-method}
\alias{harmonize,SoilProfileCollection-method}
\alias{harmonize}
\title{Harmonize a property by profile-level denormalization for convenient visualization or analysis of ranges}
\usage{
\S4method{harmonize}{SoilProfileCollection}(x, x.names, keep.cols = NULL, grp.name = "hgroup")
}
\arguments{
\item{x}{A \code{SoilProfileCollection}.}

\item{x.names}{a named list of character vectors specifying target names, profile ID suffixes and source attribute names for harmonization}

\item{keep.cols}{a character vector of column names to keep unaltered from the horizon data}

\item{grp.name}{a character vector with column name to store grouping variable in site table (default: "hgroup")}
}
\value{
A (redundant) \code{SoilProfileCollection}, with one profile for each set of harmonizations specified by \code{x.names}.
}
\description{
It is sometimes convenient to be able to "denormalize" to a \code{SoilProfileCollection} with fewer attributes but more profiles. This is helpful wherever calculations are made on a profile basis and ranges or repeated measures are depicted with multiple attributes per soil horizon.

\code{harmonize} is most commonly used for creating "comparison" soil profile sketches with \code{plotSPC}--where the thematic attribute is derived from multiple data sources or summary statistics (such as quantiles of a property for Low-RV-High). However, the method more generally applies wherever one wants to alias between multiple columns containing "similar" data as input to an algorithm.

Data are "harmonized" to a common attribute names specified by the names of list elements in \code{x.names}. Profiles are essentially duplicated. In order to satisfy uniqueness constraints of the \code{SoilProfileCollection}, the label from the sub-elements of \code{x.names} are used to disambiguate profiles. A new column in the site table is calculated to reflect these groupings and facilitate filtering. See examples below.
}
\details{
If attributes reflecting the same or similar property within a soil layer have different names (e.g. \code{socQ05}, \code{socQ50}, \code{socQ95}) it is sometimes inconvenient to work with them as multiple attributes within the same profile. These similar attributes may need to be analyzed together, or in sequence by profile, displayed using the same name or using a common scale. It is also useful to be able to alias different data sources that have the same attributes with different names.

Each list element in \code{x.names} specifies a single "harmonization," which is comprised of one or more mappings from new to old. Each named "sub-element" of \code{x.names} specifies the name and attribute to use for updating the profile ID and site table of  the duplicated profiles.
}
\examples{

### single source "harmonization" of single-profile with range -> single attribute, multi-profile

# make some test data
spc <- combine(lapply(1:10, random_profile, SPC = TRUE))

# assume that p1, p2 and p3 are the low RV and high quantiles for a hypothetical property "foo"
h1 <- harmonize(spc, x.names = list(foo = c(q05 = "p1", q50 = "p2", q95 = "p3")))

# inspect result
plotSPC(h1, color = "foo")

# filter with calculated "harmonized group" to get just RV profiles
plotSPC(subset(h1, hgroup == "q50"), color="foo")

### single source, two properties at once; with common labels: "method1" "method2"

# assume that p1, p2 are measurements by two (!=) methods for a hypothetical property "foo"
#             p3, p4 are measurements by same two methods for a hypothetical property "bar"
h3 <- harmonize(spc, x.names = list(foo = c(method1 = "p1", method2 = "p2"),
                                    bar = c(method1 = "p3", method2 = "p4")))
plotSPC(h3, color = "foo")
plotSPC(h3, color = "bar")
head(horizons(h3))

# a slight modification, "method 1" onlyused for "foo" and "method 3" for "bar"
h3 <- harmonize(spc, x.names = list(foo = c(method1 = "p1", method2 = "p2"),
                                    bar = c(method2 = "p3", method3 = "p4")))
plotSPC(h3, color = "foo") # note the pattern of values missing for foo (*_method3)
plotSPC(h3, color = "bar") #  likewise for bar (*_method1)

#' the new labels need not match across harmonizations -- not sure how useful this is but it works
h3 <- harmonize(spc, x.names = list(foo = c(method1 = "p1", method2 = "p2"),
                                    bar = c(method3 = "p3", method4 = "p4")))
plotSPC(h3, color = "foo") # note the pattern of values missing for foo (*_method 3 + 4)
plotSPC(h3, color = "bar") #  likewise for bar (*_method 1 + 2)

### two-source harmonization

# make test data
spc1 <- combine(lapply(LETTERS[1:5], random_profile, SPC = TRUE))
spc2 <- combine(lapply(letters[1:5], random_profile, SPC = TRUE))

h4 <- combine(list(harmonize(spc1, list(foo = c(transect1 = "p4"))),   # foo is p4 in dataset 1
                      harmonize(spc2, list(foo = c(transect2 = "p2")))))  # foo is p2 in dataset 2

# same property with different name in two different datasets
plotSPC(h4, color = "foo")

### many source harmonization

# make test datasets (n=10); highly redundant IDs (1:3 repeated)
spcs <- lapply(1:10, function(x) pbindlist(lapply(1:3, random_profile, SPC = TRUE)))

# randomly varying column name for demo (in each dataset, foo could could be p1 thru p5)
rcolname <- paste0("p", round(runif(10, 0.5, 5.5)))

# iterate over data sources
bigspc <- combine(lapply(1:length(spcs), function(i) {

  # assume each data source has a unique name for the property "foo"
  xn <- rcolname[i]

  # set names attribute to be equal to index i [creating unique profile IDs]
  #   i.e. 2_10 will be profile ID 2 from 10th dataset
  names(xn) <- i

  # harmonize each data source, using unique column name and target name "foo"
  harmonize(spcs[[i]], x.names = list(foo = xn))
}))

# inspect a subset
plotSPC(bigspc[1:30,], color = "foo")

}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\name{grepSPC}
\alias{grepSPC}
\alias{grepSPC,SoilProfileCollection-method}
\title{Subset SPC with pattern-matching for text-based attributes}
\usage{
\S4method{grepSPC}{SoilProfileCollection}(object, attr, pattern, ...)
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{attr}{A character vector (column in object) for matching patterns against.}

\item{pattern}{REGEX pattern to match in \code{attr}}

\item{...}{Additional arguments are passed to \code{grep()}}
}
\value{
A SoilProfileCollection.
}
\description{
\code{grepSPC()} is a shorthand function for subsetting \code{SoilProfileCollection} objects. For example, by \code{filter(grepl(spc, ...))} or \code{filter(stringr::str_detect(spc, ...))}. It provides pattern matching for a single text-based site or horizon level attribute.
}
\author{
Andrew G. Brown.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-integrity.R
\docType{methods}
\name{spc_in_sync}
\alias{spc_in_sync}
\title{Quickly assess relative state of site and horizon slots}
\usage{
spc_in_sync(object)
}
\arguments{
\item{object}{A SoilProfileCollection}
}
\value{
data.frame
}
\description{
Determine "state" of SoilProfileCollection before or after major modifications of site or horizon slot contents.

Two logical checks are performed on the site and horizon tables, and a third element \code{valid} returns \code{TRUE} when both checks are \code{TRUE}.

Check 1: Same number of sites in site as number of sites in horizons. No intermingling of IDs, no orphan horizons, no sites without horizons (for now)

Check 2: Site IDs match coalesced profile ID from horizons. Ensures the same \emph{relative} ordering, but horizons still may be out of order within profiles
}
\examples{
data(sp5)

spc_in_sync(sp5)

}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-spatial.R
\docType{methods}
\name{initSpatial<-,SoilProfileCollection,ANY,ANY-method}
\alias{initSpatial<-,SoilProfileCollection,ANY,ANY-method}
\alias{initSpatial<-}
\alias{initSpatial<-,SoilProfileCollection,ANY,character-method}
\alias{getSpatial,SoilProfileCollection-method}
\alias{getSpatial}
\alias{coordinates,SoilProfileCollection-method}
\alias{coordinates}
\alias{coordinates<-,SoilProfileCollection,ANY-method}
\alias{coordinates<-}
\alias{coordinates<-,SoilProfileCollection,character-method}
\title{Initialize Spatial Data in a SoilProfileCollection}
\usage{
\S4method{initSpatial}{SoilProfileCollection,ANY,ANY}(object, crs = NULL) <- value

\S4method{initSpatial}{SoilProfileCollection,ANY,character}(object, crs = NULL) <- value

\S4method{getSpatial}{SoilProfileCollection}(object)

\S4method{coordinates}{SoilProfileCollection}(obj)

\S4method{coordinates}{SoilProfileCollection,ANY}(object) <- value

\S4method{coordinates}{SoilProfileCollection,character}(object) <- value
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{crs}{Optional: character. Representation of Coordinate Reference System as \code{"authority:code"}, integer EPSG code, WKT2019 or PROJ4 string, an sf \code{crs} or sp \code{CRS} object.}

\item{value}{A formula specifying names of columns containing geometry (x and y coordinates), or character with the column names}

\item{obj}{A SoilProfileCollection}
}
\description{
\verb{initSpatial()<-}: Set the column names containing spatial data and the corresponding coordinate reference system for a SoilProfileCollection.

\code{getSpatial()}: Get spatial data associated with a SoilProfileCollection
}
\examples{

data(sp5)

# coordinates are stored in x and y column of site
sp5$x <- rnorm(length(sp5))
sp5$y <- rnorm(length(sp5))

# coordinates takes a formula object as input
initSpatial(sp5) <- ~ x + y

# optionally specify Coordinate Reference System (crs) on left-hand side
initSpatial(sp5, crs = "OGC:CRS84") <- ~ x + y
}
\seealso{
\code{\link[=prj]{prj()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spec2Munsell.R
\name{spec2Munsell}
\alias{spec2Munsell}
\title{Convert reflectance spectra to closest Munsell chip}
\usage{
spec2Munsell(
  x,
  res = 10,
  convert = TRUE,
  SO = c("CIE1931", "CIE1964"),
  illuminant = c("D65", "F2"),
  ...
)
}
\arguments{
\item{x}{reflectance spectra, must range from 380nm to 730nm with resolution specified in \code{res}}

\item{res}{spectra resolution in nm, typically 5nm or 10nm}

\item{convert}{logical, convert sRGB coordinates to closest Munsell chip (see \code{?munsell})}

\item{SO}{CIE standard observer: these are the color matching functions defined by CIE and used to represent "average" human color perception. CIE1931 is the 2 degree standard observer more useful for describing color perception over very small areas or at distance. CIE1964 is the 10 degree standard observer, used for most industrial color matching applications.}

\item{illuminant}{CIE standard illuminants:
\itemize{
\item D65 represents average daylight
\item F2 represents typical fluorescent lighting
}}

\item{...}{further arguments to \code{\link[=col2Munsell]{col2Munsell()}}}
}
\value{
output from \code{\link[=col2Munsell]{col2Munsell()}}
}
\description{
Convert reflectance spectra to closest Munsell chip
}
\details{
See the \href{https://ncss-tech.github.io/AQP/aqp/mix-colors.html}{expanded tutorial} for additional examples.
}
\references{
Marcus, R.T. (1998). The Measurement of Color. In K. Nassau (Ed.), Color for Science, Art, and Technology (pp. 32-96). North-Holland.

CIE Colorimetry – Part 1: CIE standard colorimetric observers. CIES014-1/E:2006 – ISO 11664-1:2007(E)

CIE. (n.d.). CIE 15:2004 Tables Data. Retrieved from https://law.resource.org/pub/us/cfr/ibr/003/cie.15.2004.tables.xls
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compareSites.R
\name{compareSites}
\alias{compareSites}
\title{Compare Site Level Attributes of a SoilProfileCollection}
\usage{
compareSites(x, vars, weights = rep(1, times = length(vars)), ...)
}
\arguments{
\item{x}{\code{SoilProfileCollection} object}

\item{vars}{character vector listing one or more site level attributes of \code{x}}

\item{weights}{numeric vector, same length as \code{vars}, variable weighting}

\item{...}{additional arguments to \code{\link[cluster:daisy]{cluster::daisy()}}}
}
\value{
\code{dissimilarity} / \code{dist} class object containing pair-wise distances, row/column names derived from \code{profile_id(x)}
}
\description{
Compare site level attributes of a \code{SoilProfileCollection} object, returning a distance matrix conformal with the output from \code{\link[=NCSP]{NCSP()}}. Values are within the range of 0-1.
}
\details{
This function is typically used in conjunction with the output from \code{\link[=NCSP]{NCSP()}}.
}
\seealso{
\code{\link[=NCSP]{NCSP()}} \code{\link[cluster:daisy]{cluster::daisy()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split.R
\name{split,SoilProfileCollection-method}
\alias{split,SoilProfileCollection-method}
\title{Split a SoilProfileCollection object into a list of SoilProfileCollection objects.}
\usage{
\S4method{split}{SoilProfileCollection}(x, f, drop = TRUE, ...)
}
\arguments{
\item{x}{\code{SoilProfileCollection} object}

\item{f}{character vector naming a single site-level attribute that defines groups, a ‘factor’ in the sense that \code{as.factor(f)} defines the grouping, or a list of such factors in which case their interaction is used for the grouping.}

\item{drop}{logical indicating if levels that do not occur should be dropped (if f is a factor or a list). When \code{drop=FALSE} and \code{f} contains missing values an additional group "missing" is returned.}

\item{...}{additional arguments are ignored}
}
\value{
A list of \code{SoilProfileCollection} or \code{NULL} for empty result.
}
\description{
This function splits a \code{SoilProfileCollection} into a list of \code{SoilProfileCollection} objects using a site-level attribute to define groups or profile ID (\code{idname(x)}).
}
\details{
As of aqp 1.25, omission of \code{f} argument is no longer possible, as the base R generic is overloaded by this \code{SoilProfileCollection} method. This used to result in an "identity" split, according to \code{idname(x)}, e.g. a list as long as \code{length(x)}, with a single-profile \code{SoilProfileCollection} per list element. Replicate this behavior using \code{f = idname(x)} or \code{f = profile_id(x)}.
}
\examples{

data(sp2)
depths(sp2) <- id ~ top + bottom

# add a more interesting site-level attribute
site(sp2) <- ~ surface

# using identity site-level attribute (profile ID)
p1 <- split(sp2, f = idname(sp2))
names(p1)
length(p1)

# using vector equal in length to number of profiles (profile ID, again)
p2 <- split(sp2, f = profile_id(sp2))
names(p2)
length(p2)

# which are both equivalent to setting `f` to NULL
p3 <- split(sp2, f = NULL)
names(p3)
length(p3)

# split on surface (age) site-level var
p4 <- split(sp2, f = "surface")
names(p4)
length(p4) # 5 unique "surfaces", 5 SPCs in result list

}
\author{
D.E. Beaudette and A.G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getArgillicBounds.R
\name{argillic.clay.increase.depth}
\alias{argillic.clay.increase.depth}
\title{Return upper boundary of argillic horizon}
\usage{
argillic.clay.increase.depth(p, clay.attr = "clay")
}
\arguments{
\item{p}{A single-profile \code{SoilProfileCollection} object.}

\item{clay.attr}{OPTIONAL: horizon attribute name referring to clay content.
default: \code{clay}}
}
\value{
A numeric vector containing top depth of argillic horizon, if
present, or NA.
}
\description{
Returns the top depth of the argillic horizon as a numeric vector.
}
\details{
Uses \code{crit.clay.argillic} to determine threshold clay increase, and
\code{get.increase.matrix} to determine where increase is met within a
vertical distance of 30 cm.
}
\examples{

data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

p <- sp1[1]
attr <- 'prop' # clay contents 
foo <- argillic.clay.increase.depth(p, clay.attr = attr)
foo

}
\seealso{
\code{getArgillicBounds}, \code{get.increase.matrix},
\code{crit.clay.argillic}
}
\author{
Andrew Gene Brown
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{reactionclass}
\alias{reactionclass}
\title{pH Reaction Classes}
\format{
An object of class \code{data.frame} with 11 rows and 3 columns.
}
\usage{
data(reactionclass)
}
\description{
Levels of pH (reaction) classes including descriptive name, and range from low to high pH
}
\references{
Soil Science Division Staff. (2017) Soil Survey Manual. C. Ditzler, K. Scheffe, and H.C. Monger (eds.). USDA Handbook 18. Government Printing Office, Washington, D.C.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parseMunsell.R
\name{parseMunsell}
\alias{parseMunsell}
\title{Parse Munsell Color Notation}
\usage{
parseMunsell(munsellColor, convertColors = TRUE, delim = NA, ...)
}
\arguments{
\item{munsellColor}{character vector of Munsell colors (e.g. \code{c('10YR 3/4', '5YR 4/6')})}

\item{convertColors}{logical, convert colors to sRGB hex notation, sRGB coordinates, CIELAB coordinates}

\item{delim}{optional, specify the type of delimiter used between value and chroma parts of the Munsell code. By default ":", ",:, "'", and "/" are supported.}

\item{...}{additional arguments to \code{\link[=munsell2rgb]{munsell2rgb()}}}
}
\value{
a \code{data.frame} object
}
\description{
Split Munsell color notation into "hue", "value", and "chroma", with optional conversion to sRGB hex notation, sRGB coordinates, and CIELAB coordinates. Conversion is performed by \code{\link[=munsell2rgb]{munsell2rgb()}}.
}
\note{
This function should be able to parse and convert 100k Munsell colors in 0.5 seconds (single-threaded, 2.5GHz i7).
}
\examples{

# just sRGB
parseMunsell("10YR 3/5", return_triplets = TRUE)

# sRGB + CIELAB (D65 illuminant)
parseMunsell("10YR 3/5", return_triplets = TRUE, returnLAB = TRUE)

# CIELAB only
parseMunsell("10YR 3/5", return_triplets = FALSE, returnLAB = TRUE)

# neutral hue
# note chroma encoded as '0'
parseMunsell('N 3/', convertColors = FALSE)

}
\author{
P. Roudier and D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-metadata.R
\docType{methods}
\name{hztexclname}
\alias{hztexclname}
\alias{hztexclname,SoilProfileCollection-method}
\alias{hztexclname<-}
\alias{hztexclname<-,SoilProfileCollection-method}
\title{Get or Set Horizon Texture Class Column Name}
\usage{
\S4method{hztexclname}{SoilProfileCollection}(object, required = FALSE)

\S4method{hztexclname}{SoilProfileCollection}(object, required = FALSE) <- value
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{required}{logical, is this attribute required? If it is, set to \code{TRUE} to trigger error on invalid \code{value}.}

\item{value}{character, name of column containing horizon texture classes}
}
\description{
\code{hztexclname()}: Get column name containing horizon designation name

\verb{hztexclname<-}: Set horizon texture class column name for a SoilProfileCollection
}
\details{
Store the column name containing horizon texture classes or other identifiers in the metadata slot of the SoilProfileCollection.
}
\examples{

data(sp1)

# promote to SPC
depths(sp1) <- id ~ top + bottom

# set horizon texture class column
hztexclname(sp1) <- "texture"

# get horizon texture class column
hztexclname(sp1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutate_profile.R
\name{mutate_profile}
\alias{mutate_profile}
\alias{mutate_profile,SoilProfileCollection-method}
\alias{mutate_profile_raw}
\title{Transform a SPC (by profile) with a set of expressions}
\usage{
\S4method{mutate_profile}{SoilProfileCollection}(object, ..., col_names = NULL, horizon_level = NULL)

mutate_profile_raw(object, expr, col_names = NULL, horizon_level = NULL)
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{...}{A set of comma-delimited R expressions that resolve to a transformation to be applied to a single profile e.g \code{mutate_profile(hzdept = max(hzdept) - hzdept)}}

\item{col_names}{character. Optional column names. Should match the number of expressions in \code{...}.}

\item{horizon_level}{logical. If \code{TRUE} results of expressions are added to the SoilProfileCollection's horizon slot, if \code{FALSE} the results are added to the site slot. If \code{NULL} (default) the results are stored in the site or horizon slot based on the number of rows in each slot compared to the length of the result calculated from the \emph{first} and \emph{last} profile in the collection.}

\item{expr}{A list of expressions in terms of column names in site or horizon table of \code{object}}
}
\value{
A SoilProfileCollection.
}
\description{
\code{mutate_profile()} is a function used for transforming SoilProfileCollections. Each expression is applied to site or horizon level attributes of individual profiles. This distinguishes this function from \code{transform}, which is applied to all values in a collection, regardless of which profile they came from.
}
\details{
If the length an expression's result matches the number of horizons, the result is stored as a horizon-level variable. If the result has length 1, it is stored as a site-level variable. In the ambiguous case where the first and last profile have only \emph{one} horizon, the results are stored in the horizon slot by default. To force results into site slot use \code{horizon_level = FALSE}.
}
\examples{

data(sp4)
depths(sp4) <- id ~ top + bottom

mutate_profile(sp4, clay_wtd_average = weighted.mean(clay, bottom - top))

data(jacobs2000)

set.seed(123)

## col_names allows for column names to be calculated
x <- mutate_profile(jacobs2000, bottom - top / sum(bottom - top),
                    col_names = paste0("relthk", floor(runif(1, 0, 100))))
x$relthk28

# mutate_profile_raw allows for lists of expressions to be evaluated
master_desgn <- c("O", "A", "E", "B", "C", "R", "L", "M")
thk_names <- paste0("thk_", master_desgn)

# calculate thickness for each horizon
x$thk <- x$bottom - x$top

## construct an arbitrary number of expressions using variable inputs
ops <- lapply(master_desgn, function(x) {
  substitute(sum(thk[grepl(PATTERN, name)], na.rm = TRUE), list(PATTERN = x))
})
names(ops) <- thk_names

# do mutation
y <- mutate_profile_raw(x, ops)

site(y)[c(idname(y), thk_names)]
}
\author{
Andrew G. Brown.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/soilColorIndices.R
\name{thompson.bell.darkness}
\alias{thompson.bell.darkness}
\title{Thompson-Bell (1996) Index}
\usage{
thompson.bell.darkness(
  p,
  name = hzdesgnname(p, required = TRUE),
  pattern = "^A",
  value = "m_value",
  chroma = "m_chroma"
)
}
\arguments{
\item{p}{A single-profile SoilProfileCollection (e.g. via profileApply())}

\item{name}{Column name containing horizon designations used to find A horizons (default: first column name containing 'name')}

\item{pattern}{Regular expression to match A horizons (default: "^A" which means horizon designation \emph{starts with} A)}

\item{value}{Column name containing horizon color values (default: "m_value")}

\item{chroma}{Column name containing horizon color chromas (default: "m_chroma")}
}
\value{
A numeric vector reflecting horizon darkness (lower values = darker).
}
\description{
Calculate the "Profile Darkness Index" by the method of Thompson & Bell (1996) "Color index for identifying hydric conditions for seasonally saturated mollisols in Minnesota" DOI: 10.2136/sssaj1996.03615995006000060051x. The Thompson-Bell Index has been shown to reflect catenary relationships in some Mollisols of Minnesota (generally: wetter landscape positions = thicker, darker surfaces).
}
\references{
Thompson, J.A. and Bell, J.C. (1996), Color Index for Identifying Hydric Conditions for Seasonally Saturated Mollisols in Minnesota. Soil Science Society of America Journal, 60: 1979-1988. doi:10.2136/sssaj1996.03615995006000060051x
}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{sierraTransect}
\alias{sierraTransect}
\title{Soil Physical and Chemical Data Related to Studies in the Sierra Nevada
Mountains, CA, USA.}
\format{
An object of class \code{SoilProfileCollection} of length 14.
}
\source{
Original manuscripts and personal communication with authors.
}
\usage{
data(sierraTransect)
}
\description{
Soil physical and chemical data associated with two bio-climatic sequences
(granitic and andesitic parent material) from the western flank of the
Sierra Nevada mountains.
}
\details{
These data were assembled from Dahlgren et al. (1997) and Rasmussen et al.
(2007), with permission granted by lead authors, by D.E. Beaudette.
}
\examples{

data(sierraTransect)

# tighter margins
op <- par(mar=c(0,0,0,0))

# quick sketch
plotSPC(sierraTransect, name.style = 'center-center', width=0.3)

# split by transect
par(mar=c(0,0,1,1))
groupedProfilePlot(
sierraTransect, groups='transect',
group.name.offset = -15, width=0.3,
name.style='center-center'
)

# thematic
groupedProfilePlot(
sierraTransect, groups='transect',
group.name.offset = -15, width=0.3,
name.style='center-center', color='Fe_o_to_Fe_d'
)

# horizon boundary viz
sierraTransect$hzd <- hzDistinctnessCodeToOffset(substr(sierraTransect$hz_boundary, 0, 1))

groupedProfilePlot(
sierraTransect, groups='transect', group.name.offset = -15,
width=0.3, name.style='center-center', color='Fe_o_to_Fe_d',
hz.distinctness.offset='hzd')


# split transects
g <- subset(sierraTransect, transect == 'Granite')
a <- subset(sierraTransect, transect == 'Andesite')

g.order <- order(g$elev)
a.order <- order(a$elev)

# order (left -> right) by elevation
par(mar=c(2,0,0,2), mfrow=c(2,1))
plot(g, width=0.3, name.style='center-center', cex.names=0.75, plot.order=g.order)
axis(1, at=1:length(g), labels=g$elev[g.order], line=-1.5)

plot(a, width=0.3, name.style='center-center', cex.names=0.75, plot.order=a.order)
axis(1, at=1:length(a), labels=a$elev[a.order], line=-1.5)


par(op)

}
\references{
R.A. Dahlgren, J.L. Boettinger, G.L. Huntington, R.G. Amundson.
Soil development along an elevational transect in the western Sierra Nevada,
California, Geoderma, Volume 78, Issues 3–4, 1997, Pages 207-236.

Rasmussen, C., Matsuyama, N., Dahlgren, R.A., Southard, R.J. and Brauer, N.
(2007), Soil Genesis and Mineral Transformation Across an Environmental
Gradient on Andesitic Lahar. Soil Sci. Soc. Am. J., 71: 225-237.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\name{subApply}
\alias{subApply}
\alias{subApply,SoilProfileCollection-method}
\title{Subset SPC based on result of performing function on each profile}
\usage{
\S4method{subApply}{SoilProfileCollection}(object, .fun, ...)
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{.fun, }{A function that takes a single profile, returns \emph{logical} of length 1.}

\item{...}{Additional arguments are passed to \code{.fun}}
}
\value{
A SoilProfileCollection.
}
\description{
\code{subApply()} is a function used for subsetting SoilProfileCollections. It currently does NOT support for "tidy" lexical features in the \code{...} arguments passed to \code{profileApply()}. The expectation is that the function \code{.fun} takes a single-profile SoilProfileCollection and returns a logical value of length one. The use case would be for any logical comparisons that cannot be evaluated inline by \code{subSPC()} because they require more than simple logical operations.
}
\author{
Andrew G. Brown.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quickSPC.R
\name{quickSPC}
\alias{quickSPC}
\title{Quickly Assemble a SoilProfileCollection}
\usage{
quickSPC(
  x,
  id = "id",
  d = "depths",
  n = "name",
  m = "soil_color",
  interval = 10
)
}
\arguments{
\item{x}{either a \code{list} or \code{character} vector, see Details and Examples}

\item{id}{character, specified when \code{x} is a \code{list}, name of ID list element}

\item{d}{character, specified when \code{x} is a \code{list}, name of depths list element}

\item{n}{character, specified when \code{x} is a \code{list}, name of horizon name list element}

\item{m}{character, specified when \code{x} is a \code{list}, name of list element containing Munsell color notation}

\item{interval, }{numeric, typically an integer and only specified when using character templates in mode 2. See Details.}
}
\value{
\code{SoilProfileCollection} object
}
\description{
Quickly assemble a single-profile, \code{SoilProfileCollection} object from two possible templates. This function is a useful shortcut for creating theoretical \code{SoilProfileCollection} objects for testing or demonstrative purposes.
}
\details{
The list template for a single SPC allows for full specification of ID, horizon designation, bottom depths, and an arbitrary number of horizon-level attributes. A compact notation is used for profile ID (single value) and horizon depths (bottom depths, assuming datum of 0). Horizon designation and additional data (e.g. clay content) are specified as vectors all of equal length, matching the number of horizons in the profile.

The character template can be provided in one of several formats:
\enumerate{
\item 'A-Bt1-Bt2-Bt3-Cr-R'
\item 'ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC'
}

Format 1 is interpreted as a horizon sequence delimited by '-' or newline character (\\n). Random integer thickness are assigned to horizons, and profile ID created via \code{digest::digest(..., algo = 'xxhash32')}. Iteration over templates in this format is automatic when \code{x} is a character vector of \code{length > 1}.

Format 2 is interpreted as a horizon sequence delimited by '|'. Horizon thickness is proportional to replication of horizon designation and scaled by the \code{interval} argument. Profile ID is created via \code{digest::digest(..., algo = 'xxhash32')}. Iteration over templates in this format is automatic when \code{x} is a character vector of \code{length > 1}.

Explicit naming of profile IDs can be accomplished by specifying an ID via prefix, as in "ID:A-Bt1-Bt2-Cr-R" or "ID:ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC". Labels specified before a ":" will be interpreted as a profile ID. These labels are optional but if specified must be unique within \code{x}.

Single-horizon profile templates must include a trailing horizon delimiter: '-', '\\n', or '|' depending on the format.
}
\examples{

# list-based template
x <- list(
id = 'P1',
depths = c(25, 33, 100, 150),
name = c('A', 'Bw', 'Bt', 'Cr'),
clay = c(12, 15, 22, 25),
soil_color = c('10YR 3/3', '10YR 4/4', '10YR 4/6', '5G 6/2')
)

s <- quickSPC(x)
plotSPC(s, name.style = 'center-center', cex.names = 1)

# character template, mode 1
# horizon thickness is generated at random (uniform [5,20])
x <- 'A-Bt1-Bt2-Bt3-Cr-R'

s <- quickSPC(x)
plotSPC(s, name.style = 'center-center', cex.names = 1)


# multiple templates
x <- c(
'A-Bt1-Bt2-Bt3-Cr-R', 
'A-C1-C2-C3-C4-Ab', 
'Ap-A-A/E-E-Bhs-Cr'
)

# this interface is vectorized 
s <- quickSPC(x)
plotSPC(s, name.style = 'center-center', cex.names = 1)


# optionally specify profile IDs using "ID:" prefix
x <- c(
'P1:A-Bt1-Bt2-Bt3-Cr-R',
'P2:A-C1-C2-C3-C4-Ab',
'P3:Ap-A-A/E-E-Bhs-Cr'
)

s <- quickSPC(x)
plotSPC(s, name.style = 'center-center', cex.names = 1)


# optionally specify:
# horizon bottom depths in cm
# soil color in Munsell notation
x <- c(
'1. simple:Oe-A-E-Bhs',
'2. full:Oe,10,10YR 2/2-A,20,10YR 3/3-E,30,2.5Y 8/2-Bhs,60,7.5YR 4/6'
)

s <- quickSPC(x)
plotSPC(s, name.style = 'center-center', cex.names = 1)

# use newline character as delimiter, more compact
x <- 'Oe,10,10YR 2/2
A,20,10YR 3/3
E,30,2.5Y 8/2
Bhs,60,7.5YR 4/6
BC,125,7.5YR 6/4
C,150,10YR 6/2'

plotSPC(quickSPC(x), name.style = 'center-center', cex.names = 1)


# character template, mode 2
# horizon thickness is proportional to replication of 
# horizon designation and scaled by 'interval' argument
# default of 10 depth units
# e.g. A horizon is 3 * 10 = 30 depth units thick.
x <- c(
  'AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
  'ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC',
  'A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR'
  )

# each horizon label is '10' depth-units (default)
s <- quickSPC(x)
plotSPC(s, name.style = 'center-center', 
        cex.names = 1, depth.axis = FALSE, 
        hz.depths = TRUE
)

# each horizon label is '5' depth-units
s <- quickSPC(x, interval = 5)
plotSPC(s, name.style = 'center-center', 
        cex.names = 1, depth.axis = FALSE, 
        hz.depths = TRUE
)

# optionally specify some / all profile IDs with "ID:" prefix
x <- c(
  'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
  'P2:ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC',
  'A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR'
  )

s <- quickSPC(x)
plotSPC(s, name.style = 'center-center', 
        cex.names = 1, depth.axis = FALSE, 
        hz.depths = TRUE
)


# make a NODATA profile, with a random hash ID
#  note the use of trailing horizon delimiter
#  note the use of NA soil color field
x <- 'NODATA,150,NA-'
s <- quickSPC(x)
plotSPC(s, name.style = 'center-center', 
        cex.names = 1, depth.axis = FALSE, 
        hz.depths = TRUE)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotColorMixture.R
\name{plotColorMixture}
\alias{plotColorMixture}
\title{Visualize Spectral Mixing of Munsell Colors}
\usage{
plotColorMixture(
  x,
  w = rep(1, times = length(x))/length(x),
  mixingMethod = c("exact", "reference"),
  n = 1,
  swatch.cex = 1.5,
  label.cex = 0.85,
  showMixedSpec = FALSE,
  overlapFix = TRUE
)
}
\arguments{
\item{x}{vector of colors in Munsell notation, should not contain duplicates}

\item{w}{vector of weights, can sum to any number}

\item{mixingMethod}{approach used to simulate a mixture:
\itemize{
\item \code{exact}: simulate a subtractive mixture of pigments, color conversion via CIE1931 color-matching functions (see \code{\link[=mixMunsell]{mixMunsell()}})
\item \code{reference}  : simulate a subtractive mixture of pigments, selecting \code{n} closest reference spectra, requires \code{gower} package
}}

\item{n}{number of closest mixture candidates when \code{mixingMethod = 'reference'} (see \code{\link[=mixMunsell]{mixMunsell()}}), results can be hard to interpret when \code{n > 2}}

\item{swatch.cex}{scaling factor for color swatch rectangle width and height, relative to \code{label.cex}, typically between 1 and 3}

\item{label.cex}{scaling factor for swatch labels}

\item{showMixedSpec}{show weighted geometric mean (mixed) spectra as dotted line (only when \code{mixingMethod = 'reference'})}

\item{overlapFix}{attempt to "fix" overlapping chip labels via \code{\link[=fixOverlap]{fixOverlap()}}, using \code{method = 'E'}}
}
\value{
a \code{lattice} graphics object
}
\description{
Lattice visualization demonstrating subtractive mixtures of colors in Munsell notation and associated spectra.
}
\details{
If present, \code{names} attribute of \code{x} is used for the figure legend. See the \href{https://ncss-tech.github.io/AQP/aqp/mix-colors.html}{expanded tutorial} for examples.
}
\seealso{
\code{\link[=mixMunsell]{mixMunsell()}}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addBracket.R
\name{addDiagnosticBracket}
\alias{addDiagnosticBracket}
\title{Annotate Diagnostic Features}
\usage{
addDiagnosticBracket(
  s,
  kind,
  feature = "featkind",
  top = "featdept",
  bottom = "featdepb",
  ...
)
}
\arguments{
\item{s}{\code{SoilProfileCollection} object}

\item{kind}{character, filter applied to \code{feature} column of diagnostic horizons registered within \code{s}}

\item{feature}{column name containing feature kind}

\item{top}{column name containing feature top depth}

\item{bottom}{column name containing feature top depth}

\item{...}{additional arguments passed to \code{\link[=addBracket]{addBracket()}}}
}
\description{
Annotate diagnostic features within a sketch of soil profiles.
}
\details{
Additional examples can be found in \href{http://ncss-tech.github.io/AQP/aqp/SPC-plotting-ideas.html}{this tutorial}.
}
\note{
This is a \code{low-level} plotting function: you must first plot a \code{SoilProfileCollection} object before using this function.
}
\examples{

 # example data
x <- c(
  'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
  'P2:Ap|AA|E|BhsBhs|Bw1Bw1|CCCCC',
  'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR',
  'P4:AA|EEE|BhsBhsBhsBhs|BwBw|CCCCC',
  'P5:AAAA|ACACACACAC|CCCCCCCCCCC|CdCdCd'
)

s <- quickSPC(x)

diagnostic_hz(s) <- data.frame(
  id = c('P1', 'P4'),
  t = c(12, 25), 
  b = c(70, 100),
  kind = c('Best', 'Best')
)

op <- par(no.readonly = TRUE)
par(mar = c(0, 0, 3, 2))

# sketches
plotSPC(
  s, name = 'name', name.style = 'center-center', cex.names = 0.75, max.depth = 210
)

# note that custom top/bottom depths must be supplied
addDiagnosticBracket(
  s, feature = 'kind', kind = 'Best', top = 't', bottom = 'b',
  labcol = 'kind',
  offset = -0.35, col = 'firebrick', tick.length = 0.04, lwd = 2
)

par(op)


}
\seealso{
\code{\link[=addBracket]{addBracket()}}, \code{\link[=plotSPC]{plotSPC()}}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotProfileDendrogram.R
\name{plotProfileDendrogram}
\alias{plotProfileDendrogram}
\title{Plot soil profiles below a dendrogram or cladogram}
\usage{
plotProfileDendrogram(
  x,
  clust,
  rotateToProfileID = FALSE,
  scaling.factor = 5 * mean(clust$height)/max(x),
  width = 0.25,
  y.offset = 60 * mean(clust$height)/max(x),
  dend.y.scale = max(clust$height * 2, na.rm = TRUE),
  dend.color = par("fg"),
  dend.width = 1,
  dend.type = c("phylogram", "cladogram"),
  debug = FALSE,
  ...
)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}

\item{clust}{a hierarchical clustering object generated by \code{hclust()}, \code{cluster::agnes()}, or \code{cluster::diana()}}

\item{rotateToProfileID}{logical, attempt rotation of dendrogram according to original profile IDs, requires \code{dendextend} package}

\item{scaling.factor}{numeric, vertical scaling of the profile heights (default estimate printed if not specified)}

\item{width}{numeric, scaling of profile widths}

\item{y.offset}{numeric, vertical offset for top of profiles (default estimate printed if not specified)}

\item{dend.y.scale}{numeric, scaling of dendrogram (default estimate printed if not specified)}

\item{dend.color}{dendrogram line color}

\item{dend.width}{dendrogram line width}

\item{dend.type}{dendrogram type, passed to \code{plot.phylo()}, either "phylogram" or "cladogram"}

\item{debug}{logical, optionally print debugging data and return a \code{data.frame} of linking structure}

\item{...}{additional arguments to \code{plotSPC()}}
}
\value{
This function is typically called to create graphical output, when \code{debug = TRUE} a \code{data.frame} of IDs and linking structure used to build the figure.
}
\description{
Plot soil profiles below a dendrogram, using methods from the ape package.
}
\note{
The default values of \code{scaling.factor}, \code{y.offset}, and \code{dend.y.scale} are based on \code{clust$height} and \code{max(x)}, and should be close to optimal. Use the defaults as a starting point for finer control.
}
\author{
D.E. Beaudette
}
\keyword{hplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{Ohz.colors}
\alias{Ohz.colors}
\title{Representative O Horizon Colors}
\format{
A \code{data.frame} with the following elements:
\itemize{
\item state: soil moisture state, "dry" or "moist"
\item genhz: horizon designation generalized into one of "Oi", "Oe", "Oa", or "other"
\item L1.munsell: representative color in Munsell notation for each moisture state / generalized horizon designation
\item group.dE00: weighted-mean CIE dE 2000 color contrast metric between all soil colors in this group and representative color
}
}
\usage{
data(Ohz.colors)
}
\description{
Representative O horizon colors for the most common types of organic soil material, at dry and moist states. Colors were derived from an analysis of soil morphologic data (5245 horizons) within the USDA-NRCS National Soil Information System. Representative colors are the L1 median colors (by \code{\link[=colorVariation]{colorVariation()}}) within groups generalized to "Oi", "Oe", "Oa", and all "other". These estimates may be useful place-holder values for soil color in collections where O horizon color was not recorded.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarize.R
\name{summarizeSPC}
\alias{summarizeSPC}
\alias{summarizeSPC,SoilProfileCollection-method,}
\alias{summarize}
\title{Perform summaries on groups (from \code{group_by}) and create new site or horizon level attributes}
\usage{
summarizeSPC(object, ...)
}
\arguments{
\item{object}{A \code{SoilProfileCollection}}

\item{...}{A set of (named) comma-delimited R expressions that resolve to a summary value. e.g \code{groupmean = mean(clay, na.rm = TRUE)}}
}
\value{
A \code{data.frame} with one row per level in the grouping variable, and one column for each summary
}
\description{
\code{summarize()} is a function used for summarizing a \code{SoilProfileCollection} object. Specify the groups using the group_by verb, and then (named) expressions to evaluate on each group. The result is a data.frame with one row per categorical level in the grouping variable and one column for each summary variable.
}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{hzMetadata,SoilProfileCollection-method}
\alias{hzMetadata,SoilProfileCollection-method}
\alias{hzMetadata}
\title{Get horizon-level metadata}
\usage{
\S4method{hzMetadata}{SoilProfileCollection}(object)
}
\arguments{
\item{object}{a SoilProfileCollection}
}
\description{
Get \code{idname(object)} and \code{hzidname(object)}, with \code{hzdesgnname(object)}, \code{hztexclname(object)} (if defined)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorQuantiles.R
\name{colorQuantiles}
\alias{colorQuantiles}
\title{Soil Color Range via Quantiles}
\usage{
colorQuantiles(soilColors, p = c(0.05, 0.5, 0.95))
}
\arguments{
\item{soilColors}{vector of R colors (sRGB colorspace)}

\item{p}{marginal quantiles of interest}
}
\value{
A List containing the following elements:
\itemize{
\item \code{marginal}: \code{data.frame} containing marginal quantiles in CIELAB (D65), closest Munsell chips, and dE00
\item \code{L1}: L1 median CIELAB (D65) values, closest Munsell chip, and dE00
}
}
\description{
Estimate central tendency and spread of soil color using marginal quantiles and L1 median of CIELAB coordinates.
}
\details{
Colors are converted from sRGB to CIELAB (D65 illuminant), marginal quantiles of (L,A,B) coordinates are estimated, and L1 median (L,A,B) is estimates. The closest Munsell chips (via Munsell/CIELAB lookup table provided by \code{munsell}) and R colors are determined by locating chips closest to the marginal quantiles and L1 median.

The results can be conveniently inspected using \code{\link[=plotColorQuantiles]{plotColorQuantiles()}}.
}
\examples{

\dontrun{
# example data, see manual page for details
data(sp5)

# slice top 25 cm
# 24-25cm is the last slice
s <- dice(sp5, 0:24 ~ .)

# check some of the data
par(mar=c(0,0,0,0))
plotSPC(sample(s, 25), divide.hz = FALSE, name = '', print.id = FALSE, width = 0.5)

# colors
previewColors(unique(s$soil_color))

# compute marginal quantiles and L1 median
cq <- colorQuantiles(s$soil_color)

# simple graphical display of results
plotColorQuantiles(cq)
}

}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accumulateDepths.R
\name{accumulateDepths}
\alias{accumulateDepths}
\title{Accumulate horizon depths, and reflect reversed depths, relative to new datum}
\usage{
accumulateDepths(
  x,
  id = NULL,
  hzdepths = NULL,
  hzname = NULL,
  hzdatum = 0,
  seqnum = NULL,
  pattern = "O",
  fix = TRUE
)
}
\arguments{
\item{x}{A \code{data.frame} or \code{SoilProfileCollection}}

\item{id}{unique profile ID. Default: \code{NULL}, if \code{x} is a SoilProfileCollection \code{idname(x)}}

\item{hzdepths}{character vector containing horizon top and bottom depth column names. Default: \code{NULL}, if \code{x} is
a SoilProfileCollection \code{horizonDepths(x)}}

\item{hzname}{character vector containing horizon designation or other label column names. Default: \code{NULL}, if \code{x}
is a SoilProfileCollection \code{hzdesgnname(x)}}

\item{hzdatum}{a numeric vector to add to accumulated depths. Default: \code{0}. Can be equal in length to number of profiles if \code{x} is a \code{SoilProfileCollection} or number of (unique) IDs if \code{x} is a \code{data.frame}.}

\item{seqnum}{Optional: character vector containing record "sequence number" column name; used in-lieu of \code{hzname} (when \code{NA}) to identify "first" record in a profile}

\item{pattern}{pattern to search for in \code{hzname} to identify matching horizons to append the profile to}

\item{fix}{apply adjustments to missing (\code{NA}) depths and expand 0-thickness horizons? Default: \code{TRUE}}
}
\value{
A horizon-level \code{data.frame}, suitable for promoting to SPC with \verb{depths<-}, or a \code{SoilProfileCollection}, depending on the class of \code{x}.
}
\description{
Fix old-style organic horizon depths or depths with a non-standard datum by the "depth accumulation" method.
}
\details{
The "depth accumulation" method calculates thicknesses of individual horizons and then cumulative sums them after putting them in \code{id} + top depth order. The routine tries to determine context based on \code{hzname} and \code{pattern}. The main transformation is if a top depth is deeper than the bottom depth, the depths are reflected on the Z-axis (made negative). The data are then \code{id} + top depth sorted again, the thickness calculated and accumulated to replace the old depths.

This function uses several heuristics to adjust data before transformation and thickness calculation:
\subsection{Regex matching of horizon designation patterns and similar}{
\itemize{
\item matches of \code{pattern} where both top and bottom depth \code{NA} -> \verb{[0,1]} \verb{[top,bottom]} depth
\item REMOVE horizons that do not match \code{pattern} where both top and bottom depths \code{NA}
}
}

\subsection{Over-ride \code{hzname} handling with the sequence column argument \code{seqnum}}{
\itemize{
\item if \code{seqnum} column specified "first record with \code{NA} \code{hzname}" is considered a \code{pattern} match if \code{seqnum == 1}
}
}

\subsection{Trigger "fixing" with the \code{fix} argument:}{
\itemize{
\item Add 1 cm to bottom-most horizons with \code{NA} bottom depth
\item Add 1 cm thickness to horizons with top and bottom depth equal
\item Add 1 cm thickness to horizons with \code{NA} top depth and bottom depth \code{0}
}
}
}
\examples{

# example using hzdatum argument
data(sp4)
depths(sp4) <- id ~ top + bottom
hz <- accumulateDepths(sp4,
                       id = "id",
                       hzdepths = c("top", "bottom"),
                       hzname = "name",
                       hzdatum = 5 * 1:length(sp4))
plot(hz)
  
# example using old-style O horizons
hz <- read.table(text = "peiidref hzdept hzdepb hzname seqnum phiid
                    1        11      0      5      A      2   295
                    2        11      1      0     Oe      1   294
                    3        11      5     13     C1      3   296
                    4        11     13     58     C2      4   297
                    5        11     58    152     C3      5   298
                    6        13      0      5      A      2   303
                    7        13      1      0     Oe      1   302
                    8        13      5     25     Bw      3   304
                    9        13     25     61      C      4   305
                    10       13     61     NA      R      5   306
                    11      136      0     13     A1      3   695
                    12      136      1      0     Oe      2   694
                    13      136      2      1     Oi      1   693
                    14      136     13     61     C1      4   696
                    15      136     61     76     C2      5   697")
                    
depths(hz) <- peiidref ~ hzdept + hzdepb

hz_fixed <- accumulateDepths(hz,
                             id = "peiidref",
                             hzdepths = c("hzdept", "hzdepb"),
                             hzname = "hzname")
is_valid <- checkHzDepthLogic(hz_fixed)$valid

test0 <- subset(hz_fixed, !is_valid)
test1 <- subset(hz_fixed, is_valid)

origO <- subset(hz, grepl("O", hzname))
fixedO <- subset(hz_fixed, grepl("O", hzname))

par(mfrow = c(2, 1), mar = c(0, 0, 3, 2))

plotSPC(origO, max.depth = 25)
plotSPC(fixedO, max.depth = 25)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/perturb.R
\name{perturb}
\alias{perturb}
\alias{permute_profile}
\title{Perturb soil horizon depths using boundary distinctness}
\usage{
perturb(
  p,
  n = 100,
  id = NULL,
  thickness.attr = NULL,
  boundary.attr = NULL,
  min.thickness = 1,
  max.depth = NULL,
  new.idname = "pID"
)
}
\arguments{
\item{p}{A SoilProfileCollection}

\item{n}{Number of new profiles to generate (default: \code{100}) per profile in \code{p}}

\item{id}{a vector of profile IDs with length equal to (\code{n}). Overrides use of \code{seq_len(n)} as default profile ID values.}

\item{thickness.attr}{Horizon variance attribute containing numeric "standard deviations" reflecting horizon thickness}

\item{boundary.attr}{Horizon variance attribute containing numeric "standard deviations" reflecting boundary transition distinctness}

\item{min.thickness}{Minimum thickness of permuted horizons (default: \code{1})}

\item{max.depth}{Depth below which horizon depths are not perturbed (default: \code{NULL})}

\item{new.idname}{New column name to contain unique profile ID (default: \code{pID})}
}
\value{
a SoilProfileCollection with \code{n} realizations of each profile in \code{p}
}
\description{
"Perturbs" the \strong{boundary between horizons} or the \strong{thickness of horizons} using a standard deviation specified as a horizon-level attribute. This is selected using either \code{boundary.attr} or \code{thickness.attr} to specify the column name.

The boundary standard deviation corresponds roughly to the concept of "horizon boundary distinctness." In contrast, the \emph{horizon thickness} standard deviation corresponds roughly to the "variation in horizon thickness" so it may be determined from several similar profiles that have a particular layer "in common."
}
\details{
Imagine a Normal curve with mean centered on the vertical (depth axis) at a representative value (RV) horizon bottom depth or thickness. By the Empirical Rule for Normal distribution, two "standard deviations" above or below that "central" mean value represent 95\% of the "typical volume" of that horizon or boundary.

\code{perturb()} can leverage semi-quantitative (ordered factor) levels of boundary distinctness/topography for the upper and lower boundary of individual horizons. A handy function for this is \code{\link[=hzDistinctnessCodeToOffset]{hzDistinctnessCodeToOffset()}}. The \code{boundary.attr} is arguably easier to parameterize from a single profile description or "Form 232" where \emph{horizon boundary distinctness} classes (based on vertical distance of transition) are conventionally recorded for each layer.

Alternately, \code{perturb()} can be parameterized using standard deviation in thickness of layers derived from a group. Say, the variance parameters are defined from a set of pedons correlated to a particular series or component, and the template "seed" profile is, for example, the Official Series Description or the Representative Component Pedon.
}
\examples{

### THICKNESS

# load sample data and convert into SoilProfileCollection
data(sp3)
depths(sp3) <- id ~ top + bottom

# select a profile to use as the basis for simulation
s <- sp3[3,]

# reset horizon names
s$name <- paste('H', seq_along(s$name), sep = '')

# simulate 25 new profiles
horizons(s)$hz.sd <- 2 # constant standard deviation
sim.1 <- perturb(s, n = 25, thickness.attr = "hz.sd")

# simulate 25 new profiles using different SD for each horizon
horizons(s)$hz.sd <- c(1, 2, 5, 5, 5, 10, 3)
sim.2 <- perturb(s, n = 25, thickness.attr = "hz.sd")

# plot
par(mfrow = c(2, 1), mar = c(0, 0, 0, 0))
plot(sim.1)
mtext(
  'SD = 2',
  side = 2,
  line = -1.5,
  font = 2,
  cex = 0.75
)
plot(sim.2)
mtext(
  'SD = c(1, 2, 5, 5, 5, 10, 3)',
  side = 2,
  line = -1.5,
  font = 2,
  cex = 0.75
)

# aggregate horizonation of simulated data
# note: set class_prob_mode=2 as profiles were not defined to a constant depth
sim.2$name <- factor(sim.2$name)
a <- slab(sim.2, ~ name, cpm=2)

# convert to long format for plotting simplicity
library(data.table)
a.long <- data.table::melt(data.table::as.data.table(a),
               id.vars = c('top', 'bottom'),
                measure.vars = levels(sim.2$name))

# plot horizon probabilities derived from simulated data
# dashed lines are the original horizon boundaries
library(lattice)

xyplot(
  top ~ value,
  groups = variable,
  data = a.long,
  subset = value > 0,
  ylim = c(100,-5),
  type = c('l', 'g'),
  asp = 1.5,
  ylab = 'Depth (cm)',
  xlab = 'Probability',
  auto.key = list(
    columns = 4,
    lines = TRUE,
    points = FALSE
  ),
  panel = function(...) {
    panel.xyplot(...)
    panel.abline(h = s$top, lty = 2, lwd = 2)
  }
)

### BOUNDARIES

# example with sp1 (using boundary distinctness)
data("sp1")
depths(sp1) <- id ~ top + bottom

# specify "standard deviation" for boundary thickness
#   consider a normal curve centered at boundary RV depth
# lookup table: ~maximum thickness of boundary distinctness classes, divided by 3
bound.lut <- c('V'=0.5,'A'=2,'C'=5,'G'=15,'D'=45) / 3

## V          A          C          G          D
## 0.1666667  0.6666667  1.6666667  5.0000000 15.0000000

sp1$bound_sd <- bound.lut[sp1$bound_distinct]

# hold any NA boundary distinctness constant
sp1$bound_sd[is.na(sp1$bound_sd)] <- 0

quantile(sp1$bound_sd, na.rm = TRUE)
p <- sp1[3]

# assume boundary sd is 1/12 midpoint of horizon depth
# (i.e. general relationship: SD increases (less well known) with depth)
sp1 <- transform(sp1, midpt = (bottom - top) / 2 + top, bound_sd = midpt / 12)
quantile(sp1$bound_sd)

perturb(p, boundary.attr = "bound_sd", n = 10)


### Custom IDs

ids <- sprintf("\%s-\%03d", profile_id(p), 1:10) 
perturb(p, boundary.attr = "bound_sd", id = ids)


}
\seealso{
\code{\link[=random_profile]{random_profile()}} \code{\link[=hzDistinctnessCodeToOffset]{hzDistinctnessCodeToOffset()}}
}
\author{
D.E. Beaudette, A.G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/depthOf.R
\name{depthOf}
\alias{depthOf}
\alias{maxDepthOf}
\alias{minDepthOf}
\title{Get top or bottom depths of horizons matching a regular expression pattern}
\usage{
depthOf(
  p,
  pattern,
  FUN = NULL,
  top = TRUE,
  hzdesgn = hzdesgnname(p, required = TRUE),
  no.contact.depth = NULL,
  no.contact.assigned = NA_real_,
  na.rm = TRUE,
  simplify = TRUE
)

maxDepthOf(
  p,
  pattern,
  top = TRUE,
  hzdesgn = hzdesgnname(p, required = TRUE),
  no.contact.depth = NULL,
  no.contact.assigned = NA,
  na.rm = TRUE,
  simplify = TRUE
)

minDepthOf(
  p,
  pattern,
  top = TRUE,
  hzdesgn = hzdesgnname(p, required = TRUE),
  no.contact.depth = NULL,
  no.contact.assigned = NA,
  na.rm = TRUE,
  simplify = TRUE
)
}
\arguments{
\item{p}{a SoilProfileCollection}

\item{pattern}{a regular expression to match in the horizon designation column. See:\code{hzdesgn}}

\item{FUN}{a function that returns a single value, and takes argument \code{na.rm}}

\item{top}{should the top (TRUE) or bottom (FALSE) depth be returned for matching horizons? Default: \code{TRUE}.}

\item{hzdesgn}{column name containing horizon designations. Default: \code{guessHzDesgnName(p)}}

\item{no.contact.depth}{depth to assume that contact did not occur.}

\item{no.contact.assigned}{depth to assign when a contact did not occur.}

\item{na.rm}{logical. Remove \code{NA}? (default: \code{TRUE})}

\item{simplify}{logical. Return single profile results as vector (default: \code{TRUE}) or \code{data.frame} (\code{FALSE})}
}
\value{
a numeric vector containing specified depth(s) of horizons matching a pattern. If \code{length(p) > 1} then a \emph{data.frame} containing profile ID, horizon ID, top or bottom depths, horizon designation and pattern.
}
\description{
The \code{depthOf} family of functions calculate depth of occurrence of a horizon designation pattern, or any other value that can be coerced to character and matched with a regular expression.

If you need all depths of occurrence for a particular pattern, \code{depthOf} is what you are looking for. \code{minDepthOf} and \code{maxDepthOf} are wrappers around \code{depthOf} that return the minimum and maximum depth. They are all set up to handle missing values and missing "contacts" with the target pattern.
}
\examples{
# construct a fake profile
spc <- data.frame(id=1, taxsubgrp = "Lithic Haploxerepts",
                  hzname   = c("A","AB","Bw","BC","R"),
                  hzdept   = c(0,  20, 32, 42,  49),
                  hzdepb   = c(20, 32, 42, 49, 200),
                  clay     = c(19, 22, 22, 21,  NA),
                  texcl    = c("l","l","l", "l","br"),
                  d_value  = c(5,   5,  5,  6,  NA),
                  m_value  = c(2.5, 3,  3,  4,  NA),
                  m_chroma = c(2,   3,  4,  4,  NA))

# promote to SoilProfileCollection
depths(spc) <- id ~ hzdept + hzdepb
hzdesgnname(spc) <- 'hzname'
hztexclname(spc) <- 'texcl'

# multiple horizons contain B
depthOf(spc, "B")

# deepest top depth of horizon containing B
maxDepthOf(spc, "B")

# shallowest top depth
minDepthOf(spc, "B")

# deepest bottom depth
maxDepthOf(spc, "B", top = FALSE)

# deepest bottom depth above 35cm
maxDepthOf(spc, "B", top = FALSE, no.contact.depth = 35)

# assign infinity (Inf) if B horizon does not start within 10cm
minDepthOf(spc, "B", no.contact.depth = 10, no.contact.assigned = Inf)

}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tauW.R
\name{xtableTauW}
\alias{xtableTauW}
\title{Format a LaTeX table with results}
\usage{
xtableTauW(result.tau, file.name = "tau_results_table.tex")
}
\arguments{
\item{result.tau}{results returned by \code{tauW}}

\item{file.name}{name of file to write output TeX file; Default: \code{file.name="tau_results_table.tex"}}
}
\description{
Format a LaTeX table with results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{munsell.spectra}
\alias{munsell.spectra}
\alias{munsell.spectra.wide}
\title{Spectral Library of Munsell Colors}
\format{
A data frame with 89496 rows and 10 variables:
\describe{
\item{munsell}{munsell color}
\item{hue}{hue component}
\item{value}{value component}
\item{chroma}{chroma component}
\item{wavelength}{wavelength (nm)}
\item{reflectance}{reflectance}
}
}
\usage{
data(munsell.spectra)
}
\description{
The original database "SpectralReflectancesOf2007MunsellBookOfColorGlossy.txt" was provided by Paul Centore and downloaded July, 2020. Reflectance values for odd chroma and 8.5-value chips have been interpolated from adjacent chips. See \code{aqp/misc/utils/Munsell/} for the entire set of processing steps.

Munsell value typically ranges from 2-9, and chroma from 1-12. Ranges vary by hue. Run \code{aqp:::.summarizeMunsellSpectraRanges()} for a detailed listing by hue.

The original database contains the following description:

This file contains spectral reflectance measurements of X-Rite's 2007 Munsell Book of Color (Glossy Finish).  The measurements were made in 2012 with a ColorMunki spectrophotometer.  The first column is the Munsell name.  The remaining columns give reflectance values for 380 nm to 730 nm, in steps of 10 nm.  The reflectance is a value between 0 (indicating that no light at that wavelength is reflected) and 1 (indicating that all the light at that wavelength is reflected).  Occasionally an entry is slightly greater than 1.  The likely cause is random variability, and those entries can be adjusted to 1 with negligible loss. In all, 1485 colour samples were measured.  Researchers are invited to analyze the data in this file.
}
\references{
Centore, Paul. Colour Tools for Painters. http://www.munsellcolourscienceforpainters.com/.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/warpHorizons.R
\name{warpHorizons}
\alias{warpHorizons}
\title{Inflate / Deflate Horizon Thickness}
\usage{
warpHorizons(
  x,
  fact = NULL,
  scaleTo = NULL,
  soilDepthFun = estimateSoilDepth,
  updateProfileID = TRUE,
  suffix = "-w"
)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object with a single soil profile}

\item{fact}{numeric or character; warping factor specified as a single numeric value, vector of numeric values (length = \code{nrow(x)}), or column name of a horizon level attribute containing numeric values}

\item{scaleTo}{numeric, target depth resulting from rescaling all horizon thickness values, due to rounding the actual depth may be within +/- 1 depth unit}

\item{soilDepthFun}{function for computing soil depth, either:
\itemize{
\item \code{estimateSoilDepth()}: "soil depth" is defined as depth to contact, requires that \code{hzdesgnname()} be set, see \code{\link[=estimateSoilDepth]{estimateSoilDepth()}} for details
\item \code{max()}: "soil depth" is defined as the bottom depth of the soil profile
}}

\item{updateProfileID}{logical; modify profile IDs}

\item{suffix}{character; suffix added to profile IDs when \code{updateProfileID = TRUE}}
}
\value{
a modified version of \code{x}, \code{SoilProfileCollection} object
}
\description{
This function applies a warping factor to the horizons of a single-profile \code{SoilProfileCollection} object. Warping values >1 will inflate horizon thickness, values <1 will deflate horizon thickness.
}
\examples{

# create an example profile
s <- quickSPC('p1:AA|Bt1Bt1Bt1|Bt2Bt2B|Bt3|Cr|RRRRR')

# warp each horizon
# values > 1: inflation
# values < 1: deflation (erosion / compaction)
s.w <- warpHorizons(s, fact = c(1.3, 0.7, 0.8, 1, 1, 1))

# combine original + warped
x <- combine(s, s.w)

# compute profile bottom depths
.bottoms <- x[, , .LAST, .BOTTOM]

# change in total depth after warping
# used to vertically offset the warped profile
.yoff <- c(0, .bottoms[1] - .bottoms[2])

# depths for line segments connecting horizon tops
.y1 <- x[1, , .TOP]
.y2 <- x[2, , .TOP] + .yoff[2]

# sketches
# can't automatically add a depth axis
par(mar = c(0.5, 0, 0, 2))
plotSPC(
  x, 
  name.style = 'center-center', 
  cex.names = 0.8, 
  width = 0.2, 
  max.depth = 150, 
  depth.axis = FALSE, 
  y.offset = .yoff
)

# illustrate warping with arrows
arrows(x0 = 1 + 0.25, y0 = .y1, x1 = 2 - 0.25, y1 = .y2, len = 0.1, col = 2)

# manually add depth axis
axis(side = 4, line = -3.5, las = 1, at = seq(from = 0, to = 150, by = 25))


# apply to multiple profiles
# text-based template
.template <- c(
'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
'P2:ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC',
'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR'
)

# each horizon label is '10' depth-units (default)
s <- quickSPC(.template)

# random warping factor, by horizon
s$f <- runif(n = nrow(s), min = 0.8, max = 1.2)

# warp horizons by profile, result is a list of SPCs
s.w <- profileApply(s, FUN = warpHorizons, fact = 'f')

# flatten list -> SoilProfileCollection
s.w <- combine(s.w)

# combine with original SPC
x <- combine(s, s.w)

# sketches 
par(mar = c(0.5, 0, 0, 2.5))
plotSPC(
  x,
  name.style = 'center-center',
  cex.names = 0.8,
  width = 0.3,
  max.depth = 165,
  depth.axis = list(line = -2)
)

# rescale all profiles to 100cm soil depth (depth to contact)
.template <- c(
  'P1:AAA|BwBwBwBw|CCCCCCC|CdCdCdCd',
  'P2:ApAp|AA|E|BhsBhs|Bw1Bw1|CCCCC',
  'P3:A|Bt1Bt1Bt1|Bt2Bt2Bt2|Bt3|Cr|RRRRR',
  'P4:AAAAA|CCC|RRRRRR'
)

# each horizon label is '10' depth-units (default)
s <- quickSPC(.template)

# warp horizons by profile, result is a list of SPCs
w <- profileApply(s, FUN = function(i) {
  warpHorizons(i, scaleTo = 100, soilDepthFun = estimateSoilDepth)
})

# flatten list -> SoilProfileCollection
w <- combine(w)

# combine with original SPC
x <- combine(s, w)

# highlight "contact"
x$color <- rep(grey(0.9), times = nrow(x))
x$color[grep('R|Cr|Cd', x$name)] <- 'royalblue'

# sketches
par(mar = c(0.5, 0, 0, 2.5))
plotSPC(
  x,
  color = 'color',
  name.style = 'center-center',
  cex.names = 0.8,
  width = 0.3,
  max.depth = 165,
  depth.axis = list(line = -2)
)

abline(h = 100, lty = 3)

}
\author{
D.E. Beaudette and S.W. Salley
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkHzDepthLogic.R
\name{checkHzDepthLogic}
\alias{checkHzDepthLogic}
\title{Check a SoilProfileCollection object for errors in horizon depths.}
\usage{
checkHzDepthLogic(
  x,
  hzdepths = NULL,
  idname = NULL,
  fast = FALSE,
  byhz = FALSE
)
}
\arguments{
\item{x}{\code{SoilProfileCollection} or \code{data.frame} object to check}

\item{hzdepths}{character vector, describing top and bottom depths in a \code{SoilProfileCollection} or \code{data.frame}. \code{horizonDepths(x)} is used when \code{x} is a \code{SoilProfileCollection}.}

\item{idname}{character, describing the column containing profile IDs in a \code{SoilProfileCollection} or \code{data.frame}. \code{idname(x)} is used when \code{x} is a \code{SoilProfileCollection}.}

\item{fast}{logical, When \code{TRUE}, details about specific test results are not needed, the operation can allocate less memory and run approximately 5x faster.}

\item{byhz}{logical, apply logic tests to profiles (\code{FALSE}) or individual horizons (\code{TRUE})?}
}
\value{
A \code{data.frame} containing profile IDs, validity boolean (\code{valid}) and test results if \code{fast = FALSE}.

The \code{data.frame} will have as many rows as profiles in \code{x} (\code{length(x)}).
\itemize{
\item \code{id} : Profile IDs, named according to \code{idname(x)}
\item \code{valid} : boolean, profile passes all of the following tests
\item \code{depthLogic} : boolean, errors related to depth logic
\item \code{sameDepth} : boolean, errors related to same top/bottom depths
\item \code{missingDepth} : boolean, NA in top / bottom depths
\item \code{overlapOrGap} : boolean, gaps or overlap in adjacent horizons (\code{NA} when \code{byhz = TRUE})
}
}
\description{
This function inspects a \code{SoilProfileCollection} object, looking for four common errors in horizon depths:
\enumerate{
\item bottom depth shallower than top depth
\item equal top and bottom depth
\item missing top or bottom depth (e.g. \code{NA})
\item gap or overlap between adjacent horizons (only if \code{byhz = FALSE})
}
}
\examples{

## sample data

data(sp3)
depths(sp3) <- id ~ top + bottom

# these data should be clean
res <- checkHzDepthLogic(sp3)

head(res)

# less memory if only concerned about net validity
res <- checkHzDepthLogic(sp3, fast = TRUE)

head(res)

}
\author{
D.E. Beaudette, A.G. Brown, S.M. Roecker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shuffle.R
\name{shuffle}
\alias{shuffle}
\title{Shuffle Horizons of a SoilProfileCollection}
\usage{
shuffle(x, mode = c("data", "horizon"), replace = FALSE, size = NULL)
}
\arguments{
\item{x}{\code{SoilProfileCollection}}

\item{mode}{character, one of 'data' or 'horizon'
\itemize{
\item 'data': shuffle the \emph{data} associated with physical horizons, making no change to the original horizon thickness and horizon IDs
\item 'horizon': shuffle physical horizons, horizon top and bottom depths are re-calculated
}}

\item{replace}{logical, replacement argument to \code{base::sample()}, only used when \code{mode = 'horizon'}}

\item{size}{integer, size argument to \code{base::sample()}, only used when \code{mode = 'horizon'}}
}
\value{
\code{SoilProfileCollection}
}
\description{
This function shuffles the horizon data or physical ordering of horizons within profiles of a \code{SoilProfileCollection} object.
}
\examples{

data('osd', package = 'aqp')
o <- osd

# shuffling of data only
o.d <- shuffle(o, mode = 'data')

# shuffling of horizons
o.h <- shuffle(o, mode = 'horizon')

# shuffling / sampling with replacement
o.h2 <- shuffle(o, mode = 'horizon', replace = TRUE)

# add method to IDs
profile_id(o.d) <- sprintf("\%s\ndata", profile_id(o.d))
profile_id(o.h) <- sprintf("\%s\nhz", profile_id(o.h))
profile_id(o.h2) <- sprintf("\%s\nhz R", profile_id(o.h2))

# combine into single SPC
g <- combine(o, o.d, o.h, o.h2)

# graphical comparison
op <- par(mar = c(0, 0, 0.5, 2.5))
plotSPC(g, name.style = 'center-center', cex.names = 0.66, width = 0.3, cex.id = 0.75)

par(op)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glomApply.R
\name{glomApply}
\alias{glomApply}
\alias{glomApply,SoilProfileCollection-method}
\title{Subset an SPC by applying glom to each profile}
\usage{
glomApply(
  object,
  .fun = NULL,
  truncate = FALSE,
  invert = FALSE,
  modality = "all",
  ...,
  chunk.size = 100
)
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{.fun}{A function that returns vector with top and bottom depth (\code{z1} and \code{z2} arguments to \code{glom}) for a single profile \code{p} (as passed by \code{profileApply})}

\item{truncate}{Truncate horizon top and bottom depths to \code{[z1, z2]}}

\item{invert}{Truncate horizon top and bottom depths to \code{[z1, z2]} and then invert result?}

\item{modality}{Aggregation method for glom result. Default \code{"all"}: return all horizons; \code{"thickest"}: return (shallowest) thickest horizon}

\item{...}{A set of comma-delimited R expressions that resolve to a transformation to be applied to a single profile e.g \code{glomApply(hzdept = max(hzdept) - hzdept)} like \code{aqp::mutate}}

\item{chunk.size}{Chunk size parameter for \code{profileApply}}
}
\value{
A SoilProfileCollection.
}
\description{
\code{glomApply()} is a function used for subsetting SoilProfileCollection objects by depth. It is a wrapper around \code{glom} which is intended to subset single-profile SPCs based on depth intervals/intersection.

\code{glomApply} works by accepting a function \code{.fun} as argument. This function is used on each profile to process a multi-profile SPC for input to \code{glom} (via \code{profileApply}). For each profile, \code{.fun} returns a 2-length numeric vector of top and bottom boundaries \code{glom} arguments: \code{z1}, \code{z2}.

\code{glomApply} provides the option to generate profile-specific glom depths for a large SPC and handles iteration and rebuilding of a subset SPC object. Optional arguments include: \code{truncate} to cut the boundaries to specified \code{[z1, z2]}; \code{invert} to the portion outside \code{[z1, z2]}, \code{modality} to either \code{"all"} horizons or \code{"thickest"} horizon in the \code{glom} interval. \code{...} are various expressions you can run on the individual profiles using NSE, similar to \code{mutate}.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

data(sp3)
depths(sp3) <- id ~ top + bottom

# init horizon designation column in metadata, used by estimateSoilDepth
hzdesgnname(sp3) <- 'name'

# constant depths, whole horizon returns by default
plot(glomApply(sp3, function(p) c(25,100)))

# constant depths, truncated
#(see aqp::trunc for helper function)
plot(glomApply(sp3, function(p) c(25,30), truncate = TRUE))

# constant depths, inverted
plot(glomApply(sp3, function(p) c(25,100), invert = TRUE))

# constant depths, inverted + truncated (same as above)
plot(glomApply(sp3, function(p) c(25,30), invert = TRUE, truncate=TRUE))

# random boundaries in each profile
plot(glomApply(sp3, function(p) round(sort(runif(2, 0, max(sp3))))))

# random boundaries in each profile (truncated)
plot(glomApply(sp3, function(p) round(sort(runif(2, 0, max(sp3)))), truncate = TRUE))

# calculate some boundaries as site level attribtes
sp3$glom_top <- profileApply(sp3, getMineralSoilSurfaceDepth)
sp3$glom_bottom <- profileApply(sp3, estimateSoilDepth)

# use site level attributes for glom intervals for each profile
plot(glomApply(sp3, function(p) return(c(p$glom_top, p$glom_bottom))))

}
\seealso{
\code{\link{glom}}  \code{\link{trunc}}

\code{\link{glom}}  \code{\link{glomApply}}
}
\author{
Andrew G. Brown.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{horizonNames<-}
\alias{horizonNames<-}
\alias{horizonNames<-,SoilProfileCollection-method}
\alias{horizonNames,SoilProfileCollection-method}
\alias{horizonNames}
\title{Set horizon column names}
\usage{
\S4method{horizonNames}{SoilProfileCollection}(object) <- value

\S4method{horizonNames}{SoilProfileCollection}(object)
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{value}{a unique vector of equal length to number of columns in horizons \code{length(horizonNames(object))}}
}
\description{
Set horizon column names

Get names of columns in horizon table.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glom.R
\name{glom,SoilProfileCollection-method}
\alias{glom,SoilProfileCollection-method}
\alias{glom}
\alias{trunc,SoilProfileCollection-method}
\title{Subset soil horizon data using a depth or depth interval}
\usage{
\S4method{glom}{SoilProfileCollection}(
  p,
  z1,
  z2 = NULL,
  ids = FALSE,
  df = FALSE,
  truncate = FALSE,
  invert = FALSE,
  fill = FALSE,
  modality = "all",
  drop = !fill,
  ...
)

\S4method{trunc}{SoilProfileCollection}(x, z1, z2, ...)
}
\arguments{
\item{p}{A SoilProfileCollection}

\item{z1}{numeric vector of top depth to intersect horizon (required). Can be an expression involving \code{siteNames(p)} or quoted column name. Should evaluate to numeric length \code{1} or length equal to \code{length(p)}}

\item{z2}{numeric vector bottom depth of intersection interval (optional). Can also be an expression involving \code{siteNames(p)} or quoted column name. Should evaluate to numeric length \code{1}, length equal to \code{length(p)} or \code{NULL}. Default: \code{NULL} is "point" intersection}

\item{ids}{return only horizon IDs? default: \code{FALSE}}

\item{df}{return a data.frame, by intersection with \code{horizons(p)}? default: \code{FALSE}}

\item{truncate}{truncate horizon top and bottom depths to \code{z1} and \code{z2}? default: \code{FALSE}}

\item{invert}{get horizons \emph{outside} the interval \verb{[z1,z2]}? default: \code{FALSE}}

\item{fill}{keep sites and preserve order for profiles that do not have horizons in interval by filling with a single horizon with \code{NA} top and bottom depth. default: \code{FALSE}}

\item{modality}{default: \code{"all"} return all horizons; or \code{modality = "thickest"}) to return the \emph{thickest} horizon in interval. If multiple horizons have equal thickness, the first (shallowest) is returned.}

\item{drop}{Inverted alias of fill for consistency with other methods. When \code{drop=FALSE}, filling occurs.}

\item{...}{\code{trunc()}: additional arguments passed to \code{glom()}}

\item{x}{A SoilProfileCollection}
}
\value{
a SoilProfileCollection, data.frame, or a vector of horizon IDs. \code{NULL} if no result.
}
\description{
Make a "clod" of horizons from a SoilProfileCollection given a point or a depth interval to intersect. The interval \verb{[z1,z2]} may be profile-specific (equal in length to \code{p}), or may be recycled over all profiles (if boundaries are length 1). For "point" intersection, \code{z2} may be left as the default value \code{NULL}.

\code{trunc()} is a wrapper method (using S4 generic) for \code{glom()} where \code{truncate=TRUE}
}
\details{
"To glom" is "to steal" or to "become stuck or attached to". The word is related to the compound "glomalin", which is a glycoprotein produced by mycorrhizal fungi in soil.

The full depth range of horizons included within the interval are returned (a "ragged" SoilProfileCollection) unless the \code{truncate} argument is set as \code{TRUE}. Horizon intersection is based on unique ID \code{hzidname(spc)} and depth range of interest. Profiles that lack data in the range of interest will be dropped from the resulting SoilProfileCollection.

If inverting results with \code{invert}, it is possible that thick horizons (whose boundaries span wider than the specified interval) will be split into \emph{two} horizons, where previously they were one. This may make the results from \code{ids = TRUE} different from what you expect, as they will be based on a profile with an "extra" horizon and re-calculated unique horizon ID (\code{hzidname(spc)}) \code{"hzID"}.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

p <- glom(sp1, 25, 150)

# 28 horizons
nrow(p) 

# inspect graphically
par(mar = c(1,1,3,1))
plot(p, color = "prop", max.depth = 200)
abline(h = c(25, 100), lty = 2)

## glom(..., truncate = TRUE)

p2 <- glom(sp1, 25, 150, truncate = TRUE)

# 28 horizons
nrow(p2) 

# inspect graphically
par(mar = c(1,1,3,1))
plot(p2, color = "prop", max.depth = 200)
abline(h = c(25, 100), lty = 2)

## glom(..., truncate = TRUE, invert = TRUE)

p3 <- glom(sp1, 25, 150, truncate = TRUE, invert = TRUE)

# 45 horizons
nrow(p3) 

# inspect graphically
par(mar = c(1,1,3,1))
plot(p3, color = "prop", max.depth = 200)
abline(h = c(25, 100), lty = 2)

## profile-specific interval, using expressions evaluated within sp1@site

# calculate some new site-level variables containing target interval
sp1$glom_top <- (1:9) * 10
sp1$glom_bottom <- 10 + sp1$glom_top

# glom evaluates non-standard expressions using siteNames(sp1) column names
p4 <- glom(sp1, glom_top / 2, glom_bottom * 1.2, truncate = TRUE)

# inspect graphically
par(mar = c(1,1,3,1))
plot(p4, color = "prop", max.depth = 200)


# load sample data
data("sp3")

# promote to SPC
depths(sp3) <- id ~ top + bottom

### TRUNCATE all profiles in sp3 to [0,25]

# set up plot parameters
par(mfrow=c(2,1), mar=c(0,0,0,0))

# full profiles
plot(sp3)

# trunc'd profiles
plot(trunc(sp3, 0, 25))
}
\seealso{
\code{\link{glomApply}} \code{\link{trunc}}
}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkSPC.R
\name{checkSPC}
\alias{checkSPC}
\title{Test for a valid \code{SoilProfileCollection}}
\usage{
checkSPC(x)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}
}
\value{
TRUE or FALSE. Consider using \code{rebuildSPC()} if FALSE.
}
\description{
Test for a valid \code{SoilProfileCollection}
}
\details{
Test for valid \code{SoilProfileCollection} by checking for slots defined in
the class prototype. Likely only used between major versions of \code{aqp} where
internal structure of \code{SoilProfileCollection} has changed. Use
\code{checkHzDepthLogic} to check for common errors in horizon depths.
}
\seealso{
\code{\link{rebuildSPC}}, \code{\link{checkHzDepthLogic}}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R,
%   R/SoilProfileCollection-setters.R
\docType{methods}
\name{horizons,SoilProfileCollection-method}
\alias{horizons,SoilProfileCollection-method}
\alias{horizons}
\alias{horizons<-}
\alias{horizons<-,SoilProfileCollection-method}
\title{Retrieve horizon data from SoilProfileCollection}
\usage{
\S4method{horizons}{SoilProfileCollection}(object)

\S4method{horizons}{SoilProfileCollection}(object) <- value
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{value}{An object inheriting \code{data.frame}}
}
\description{
Get horizon data from SoilProfileCollection. Result is returned in the same \code{data.frame} class used to initially construct the SoilProfileCollection.

Horizon data in an object inheriting from \code{data.frame} can easily be added via merge (LEFT JOIN). There must be one or more same-named columns (with at least some matching data) on the left and right hand side to facilitate the join: \code{horizons(spc) <- newdata}
}
\examples{

# load test data
data(sp2)

# promote to SPC
depths(sp2) <- id ~ top + bottom

# assign true to surface horizon
newdata <- data.frame(top = 0,
                      newvalue = TRUE)

# do left join
horizons(sp2) <- newdata

# inspect site table: newvalue TRUE only for horizons
#  with top depth equal to zero
horizons(sp2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{jacobs2000}
\alias{jacobs2000}
\title{Soil Morphologic Data from Jacobs et al. 2002.}
\format{
A \code{SoilProfileCollection} object.
}
\usage{
data(jacobs2000)
}
\description{
Select soil morphologic data from "Redoximorphic Features as Indicators of
Seasonal Saturation, Lowndes County, Georgia". This is a useful sample
dataset for testing the analysis and visualization of redoximorphic
features.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# load
data(jacobs2000)

# basic plot
par(mar = c(0, 1, 3, 1.5))
plotSPC(jacobs2000, name='name', color='matrix_color', width=0.3)

# add concentrations
addVolumeFraction(jacobs2000, 'concentration_pct',
                  col = jacobs2000$concentration_color, pch = 16, cex.max = 0.5)

# add depletions
plotSPC(jacobs2000, name='name', color='matrix_color', width=0.3)
addVolumeFraction(jacobs2000, 'depletion_pct',
                  col = jacobs2000$depletion_color, pch = 16, cex.max = 0.5)

# time saturated
plotSPC(jacobs2000, color='time_saturated', cex.names=0.8, col.label = 'Time Saturated')


# color contrast: matrix vs. concentrations
cc <- colorContrast(jacobs2000$matrix_color_munsell, jacobs2000$concentration_munsell)
cc <- na.omit(cc)

cc <- cc[order(cc$dE00), ]
cc <- unique(cc)

par(bg = 'black', fg = 'white')
colorContrastPlot(cc$m1[1:10], cc$m2[1:10], labels = c('matrix', 'concentration'))
colorContrastPlot(cc$m1[11:21], cc$m2[11:21], labels = c('matrix', 'concentration'))


# color contrast: depletion vs. concentrations
cc <- colorContrast(jacobs2000$depletion_munsell, jacobs2000$concentration_munsell)
cc <- na.omit(cc)

cc <- cc[order(cc$dE00), ]
cc <- unique(cc)

par(bg = 'black', fg = 'white')
colorContrastPlot(cc$m1, cc$m2, labels = c('depletion', 'concentration'))

}
\references{
Jacobs, P. M., L. T. West, and J. N. Shaw. 2002. Redoximorphic
Features as Indicators of Seasonal Saturation, Lowndes County, Georgia. Soil
Sci. Soc. Am. J. 66:315-323. doi:\doi{10.2136/sssaj2002.3150}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aqp-label-placement-solvers.R
\name{SANN_1D}
\alias{SANN_1D}
\title{Fix Overlap within a Sequence via Simulated Annealing}
\usage{
SANN_1D(
  x,
  thresh = 0.6,
  adj = thresh * 2/3,
  min.x = min(x) - 0.2,
  max.x = max(x) + 0.2,
  maxIter = 1000,
  trace = FALSE,
  tiny = 1e-04,
  T0 = 500,
  k = 10,
  ...
)
}
\arguments{
\item{x}{vector of horizontal positions, pre-sorted}

\item{thresh}{horizontal threshold defining "overlap" or distance between elements of \code{x}. For adjusting soil profile sketches values are typically < 1 and likely in (0.3, 0.8).}

\item{adj}{specifies the size of perturbations within \code{runif(min = adj * -1, max = adj)}. Larger values will sometimes reduce the number of iterations required to solve particularly difficult overlap conditions. See \code{coolingRate} argument when \code{adj} is large}

\item{min.x}{left-side boundary condition, consider expanding if a solution cannot be found within \code{maxIter}.}

\item{max.x}{right-side boundary condition, consider expanding if a solution cannot be found within \code{maxIter}.}

\item{maxIter}{maximum number of iterations to attempt before giving up and returning a regularly-spaced sequence}

\item{trace}{print diagnostics, result is a \code{list} vs \code{vector}}

\item{tiny}{the smallest allowable overlap}

\item{T0}{starting temperature}

\item{k}{cooling constant}

\item{\dots}{not used, absorbs additional arguments to \code{\link[=fixOverlap]{fixOverlap()}}}
}
\value{
When \code{trace = FALSE}, a vector of the same length as \code{x}, preserving rank-ordering and boundary conditions. When \code{trace = TRUE} a list containing the new sequence along with information about objective functions and decisions made during iteration.
}
\description{
This function makes small adjustments to elements of \code{x} until overlap defined by \code{thresh} is removed, or until \code{maxIter} is reached. Rank order and boundary conditions (defined by \code{min.x} and \code{max.x}) are preserved. The underlying algorithm is based on simulated annealing. The "cooling schedule" parameters \code{T0} and \code{k} can be used to tune the algorithm for specific applications.
}
\details{
Ideas for solving difficult overlap scenarios:
\itemize{
\item widen the boundary conditions by adjusting \code{min.x} and \code{max.x} beyond the original scale of \code{x}
\item reduce the allowable overlap threshold \code{thresh}
\item reduce the magnitude of perturbations (\code{adj}) and increase \code{maxIter}
\item increase \code{k}
}
}
\examples{

x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)

# easy
z <- fixOverlap(x, thresh = 0.2, trace = TRUE)

# harder
z <- fixOverlap(x, thresh = 0.6, trace = TRUE)

# much harder
z <- fixOverlap(x, thresh = 0.9, trace = TRUE)


# interpret `trace` output

# relatively challenging
x <- c(1, 2, 3.4, 3.4, 3.4, 3.4, 6, 8, 10, 12, 13, 13, 15, 15.5)

# fix overlap, return debugging information
set.seed(10101)
z <- fixOverlap(x, thresh = 0.8, trace = TRUE)

# setup plot device
par(mar = c(4, 4, 1, 1))
layout(matrix(c(1,2,3)), widths = 1, heights = c(1,1,2))

# objective function = overlap + SSD
plot(
  seq_along(z$stats), z$stats, 
  type = 'h', las = 1,
  xlab = 'Iteration', ylab = 'Overlap',
  cex.axis = 0.8
)

# SSD: deviation from original configuration 
plot(
  seq_along(z$ssd), z$ssd, 
  type = 'h', las = 1,
  xlab = 'Iteration', ylab = 'Deviation',
  cex.axis = 0.8
)
# adjustments at each iteration
matplot(
  z$states, type = 'l', 
  lty = 1, las = 1, 
  xlab = 'Iteration', ylab = 'x-position'
)

# trace log
# B: boundary condition violation
# O: rank (order) violation
# +: accepted perturbation
# -: rejected perturbation
table(z$log)

}
\seealso{
\code{\link[=electroStatics_1D]{electroStatics_1D()}}, \code{\link[=fixOverlap]{fixOverlap()}}
}
\author{
D.E. Beaudette and K.C. Thompson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/munsell2rgb.R
\name{munsell2rgb}
\alias{munsell2rgb}
\title{Convert Munsell Color Notation to other Color Space Coordinates (sRGB and CIELAB)}
\usage{
munsell2rgb(
  the_hue,
  the_value,
  the_chroma,
  alpha = 1,
  maxColorValue = 1,
  return_triplets = FALSE,
  returnLAB = FALSE
)
}
\arguments{
\item{the_hue}{a vector of one or more more hues, upper-case (e.g. '10YR')}

\item{the_value}{a vector of one or more values (e.g. '4')}

\item{the_chroma}{a vector of one or more chromas (e.g. '6'), may be NA for neutral hues}

\item{alpha}{numeric, transparency setting used when \code{return_triplets = FALSE} and \code{returnLAB = FALSE}}

\item{maxColorValue}{maximum sRGB color value, typically \code{1} (see \code{\link{rgb})}}

\item{return_triplets}{logical, return sRGB coordinates (range 0-1) instead of standard hex notation of sRGB (e.g. '#8080B')}

\item{returnLAB}{logical, return CIELAB coordinates (D65 illuminant)}
}
\value{
A vector of R colors is returned that is the same length as the input data. When \code{return_triplets = TRUE} and/or \code{returnLAB = TRUE}, then a \code{data.frame} (of sample length as input) is returned.
}
\description{
Color conversion based on a look-up table of common soil colors.
}
\details{
This function is vectorized without recycling: i.e. the length of each argument must be the same. Both functions will pad output with NA if there are any NA present in the inputs.

Neutral hues are approximated by greyscale shades ranging from 20\\% (darker) to 80\\% (lighter). No chroma is required for neutral hues.

Gley soil colors that are missing a chroma will not be correctly interpreted. Consider using a chroma of 1.
Non-standard Munsell notation (e.g. '7.9YR 2.7/2.0') can be matched (nearest-neighbor, no interpolation) to the closest color within the \code{munsell} sRGB/CIELAB look-up table via \code{getClosestMunsellChip()}. A more accurate estimate of sRGB values from non-standard notation can be achieved with the \href{https://CRAN.R-project.org/package=munsellinterpol}{munsellinterpol} package.

See examples below.
}
\note{
Care should be taken when using the resulting sRGB values; they are close to their Munsell counterparts, but will vary based on your monitor and ambient lighting conditions. Also, the value used for \code{maxColorValue} will affect the brightness of the colors. Th default value (1) will usually give acceptable results, but can be adjusted to force the colors closer to what the user thinks they should look like.
}
\examples{

# neutral hues (N) can be defined with chroma of 0 or NA 
g <- expand.grid(hue = 'N', value = 2:8, chroma = 0, stringsAsFactors = FALSE)
(m <- munsell2rgb(g$hue, g$value, g$chroma))
soilPalette(m)

# back-transform to Munsell notation
col2Munsell(t(col2rgb(m)) / 255)


# basic example
d <- expand.grid(hue = '10YR', value = 2:8, chroma = 1:8, stringsAsFactors = FALSE)
d$color <- with(d, munsell2rgb(hue, value, chroma))

# similar to the 10YR color book page
plot(value ~ chroma, data = d, col = d$color, pch = 15, cex = 3, las = 1)

# multiple pages of hue:
hues <- c('2.5YR', '5YR', '7.5YR', '10YR')
d <- expand.grid(
  hue = hues, 
  value = c(2, 2.5, 3:8), 
  chroma = seq(2, 8, by = 2), 
  stringsAsFactors = FALSE
)
# convert Munsell -> sRGB
d$color <- with(d, munsell2rgb(hue, value, chroma))

# extract CIELAB coordinates
with(d, munsell2rgb(hue, value, chroma, returnLAB = TRUE))

# plot: note that we are setting panel order from red --> yellow
library(lattice)

xyplot(
  value ~ factor(chroma) | factor(hue, levels = hues),
  main = "Common Soil Colors", layout = c(4, 1), scales = list(alternating = 1),
  strip = strip.custom(bg = grey(0.85)),
  data = d, as.table = TRUE, subscripts = TRUE,
  xlab = 'Chroma', ylab = 'Value',
  panel = function(x, y, subscripts, ...) {
    panel.xyplot(x, y, pch = 15, cex = 4, col = d$color[subscripts])
  }
)


# convert a non-standard color to closest "chip" in `munsell` look-up table
getClosestMunsellChip('7.9YR 2.7/2.0', convertColors = FALSE)

# convert directly to hex notation of sRGB
getClosestMunsellChip('7.9YR 2.7/2.0')

}
\references{
\itemize{
\item http://www.brucelindbloom.com/index.html?ColorCalcHelp.html
\item https://www.munsellcolourscienceforpainters.com/MunsellAndKubelkaMunkToolbox/MunsellAndKubelkaMunkToolbox.html
\item https://www.rit.edu/science/munsell-color-lab
}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missingDataGrid.R
\name{missingDataGrid}
\alias{missingDataGrid}
\title{Missing Data Grid}
\usage{
missingDataGrid(
  s,
  max_depth,
  vars,
  filter.column = NULL,
  filter.regex = NULL,
  cols = NULL,
  ...
)
}
\arguments{
\item{s}{a SoilProfileCollection object}

\item{max_depth}{integer specifying the max depth of analysis}

\item{vars}{character vector of column names over which to evaluate missing
data}

\item{filter.column}{a character string naming the column to apply the
filter REGEX to}

\item{filter.regex}{a character string with a regular expression used to
filter horizon data OUT of the analysis}

\item{cols}{a vector of colors}

\item{\dots}{additional arguments passed on to \code{levelplot}}
}
\value{
A \code{data.frame} describing the percentage of missing data by
variable.
}
\description{
Generate a levelplot of missing data from a SoilProfileCollection object.
}
\details{
This function evaluates a \verb{missing data fraction} based on slice-wise
evaluation of named variables in a \code{SoilProfileCollection} object.
}
\note{
A lattice graphic is printed to the active output device.
}
\examples{

# 10 random profiles
set.seed(10101)
s <- lapply(as.character(1:10), random_profile)
s <- do.call('rbind', s)

# randomly sprinkle some missing data
s[sample(nrow(s), 5), 'p1'] <- NA
s[sample(nrow(s), 5), 'p2'] <- NA
s[sample(nrow(s), 5), 'p3'] <- NA

# set all p4 and p5 attributes of `soil 1' to NA
s[which(s$id == '1'), 'p5'] <- NA
s[which(s$id == '1'), 'p4'] <- NA

# upgrade to SPC
depths(s) <- id ~ top + bottom

# plot missing data via slicing + levelplot
missingDataGrid(
  s,
  max_depth = 100,
  vars = c('p1', 'p2', 'p3', 'p4', 'p5'),
  main='Missing Data Fraction'
)

}
\seealso{
\code{\link{slice}}
}
\author{
D.E. Beaudette
}
\keyword{hplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getArgillicBounds.R
\name{getArgillicBounds}
\alias{getArgillicBounds}
\title{Estimate upper and lower boundary of argillic diagnostic subsurface horizon}
\usage{
getArgillicBounds(
  p,
  hzdesgn = hzdesgnname(p, required = TRUE),
  clay.attr = hzmetaname(p, "clay", required = TRUE),
  texcl.attr = hztexclname(p, required = TRUE),
  require_t = TRUE,
  bottom.pattern = "Cr|R|Cd",
  lower.grad.pattern = "^[2-9]*B*CB*[^rtd]*[1-9]*$",
  sandy.texture.pattern = "-S$|^S$|COS$|L[^V]FS$|[^L]VFS$|LS$|LFS$",
  vertical.distance = 30,
  simplify = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{p}{A SoilProfileCollection}

\item{hzdesgn}{the name of the column/attribute containing the horizon designation; default="hzname"}

\item{clay.attr}{the name of the column/attribute containing the clay content; default="clay"}

\item{texcl.attr}{the name of the column/attribute containing the textural class (used for finding sandy horizons); default="texcl"}

\item{require_t}{require a "t" subscript for positive identification of upper and lower bound of argillic? default: TRUE}

\item{bottom.pattern}{regular expression passed to \code{estimateSoilDepth} to match the lower boundary of the soil. default is "Cr|R|Cd" which approximately matches paralithic, lithic and densic contacts.}

\item{lower.grad.pattern}{this is a pattern for adjusting the bottom depth of the argillic horizon upwards from the bottom depth of the soil. The absence of illuviation is used as a final control on horizon pattern matching.}

\item{sandy.texture.pattern}{this is a pattern for matching sandy textural classes: \verb{-S$|^S$|COS$|L[^V]FS$|[^L]VFS$|LS$|LFS$}}

\item{vertical.distance}{Vertical distance in which clay increase must be met. Default \code{30} cm}

\item{simplify}{Return a length 2 vector with upper and lower boundary when \code{p} has length 1? Default \code{TRUE}.}

\item{verbose}{Print out information about 't' subscripts, sandy textures, plow layers and lower gradational horizons?}
}
\value{
Returns a numeric vector; first value is top depth, second value is bottom depth. If as.list is TRUE, returns a list with top depth named "ubound" and bottom depth named "lbound". If \code{p} has more than one profile or if \code{simplify = FALSE} the result is a data.frame containing profile ID, upper and lower boundary columns.
}
\description{
\code{getArgillicBounds} estimates the upper and lower boundary of argillic diagnostic subsurface horizon for a profile in a single-profile SoilProfileCollection object (\code{p}).

The upper boundary is where the clay increase threshold is met. The function uses \code{crit.clay.argillic} as the threshold function for determining whether a clay increase occurs and \code{get.increase.matrix} to determine whether the increase is met, whether vertical distance of increase is sufficiently small, and in which horizon.
}
\details{
The lower boundary is first approximated as the depth to a lithic/paralithic/densic contact, or some other horizon matchable by a custom regular expression pattern. Subsequently, that boundary is extended upwards to the end of "evidence of illuviation."

The depth to contact is estimated using 'bottom.pattern' "Cr|R|Cd" by default. It matches anything containing Cr, R or Cd.

The lower gradational horizon regular expression ‘lower.grad.pattern' default is \verb{^[2-9]*B*CB*[^rtd]*[1-9]*$\}}. It matches anything that starts with a lithologic discontinuity (or none) and a C master horizon designation. May contain B as second horizon designation in transitional horizon. May not contain 'r' or 't' subscript.

The minimum thickness of the argillic horizon is dependent on whether all subhorizons are "sandy" or not. The \code{sandy.texture.pattern} default \verb{-S$|^S$|COS$|L[^V]FS$|[^L]VFS$|LS$|LFS$} captures USDA textural class fine earth fractions that meet "sandy" particle size class criteria.

There also is an option ‘require_t' to omit the requirement for evidence of eluviation in form of 't' subscript in 'hzdesgn'. Even if "t" subscript is not required for positive identification, the presence of lower gradational C horizons lacking 't' will still be used to modify the lower boundary upward from a detected contact, if needed. If this behavior is not desired, just set 'lower.grad.pattern' to something that will not match any horizons in your data.
}
\examples{
data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# set required metadata
hzdesgnname(sp1) <- 'name'
hztexclname(sp1) <- 'texture'
hzmetaname(sp1, 'clay') <- 'prop'

x <- getArgillicBounds(sp1)
x

}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{sp6}
\alias{sp6}
\title{Soil Physical and Chemical Data from Manganiferous Soils}
\format{
A data frame with 30 observations on the following 13 variables.
\describe{ \item{id}{pedon name} \item{name}{horizon
designation} \item{top}{horizon top boundary in cm}
\item{bottom}{horizon bottom boundary in cm}
\item{color}{moist soil color in Munsell notation}
\item{texture}{USDA soil texture class} \item{sand}{sand
content by weight percentage} \item{silt}{silt content by weight
percentage} \item{clay}{clay content by weight percentage}
\item{Fe}{DCB-extracted Fe in g/kg (see citation)}
\item{Mn}{DCB-extracted Mn in g/kg (see citation)}
\item{C}{total organic carbon as g/kg} \item{pH}{measured in
1:1 H20 slurry} \item{Db}{bulk density (g/cc), clod method} }
}
\source{
http://www.sciencedirect.com/science/article/pii/S0016706111001972
}
\description{
Soil Physical and Chemical Data from Manganiferous Soils (Bourgault and
Rabenhorst, 2011)
}
\details{
Selected soil physical and chemical data from (Bourgault and Rabenhorst,
2011).
}
\examples{

  # setup environment
  library(aqp)
  data(sp6)

  # init SPC
  depths(sp6) <- id ~ top + bottom
  # convert non-standard Munsell colors
  sp6$soil_color <- getClosestMunsellChip(sp6$color)

  # profile sketches
  par(mar=c(0,0,3,0))
  plot(sp6, color='soil_color')
  plot(sp6, color='Mn')
  plot(sp6, color='Fe')
  plot(sp6, color='pH')
  plot(sp6, color='texture')


}
\references{
Rebecca R. Bourgault, Martin C. Rabenhorst. 2011.  Genesis and
characterization of manganiferous soils in the Eastern Piedmont, USA.
Geoderma. 165:84-94.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reaction.R
\name{ph_to_rxnclass}
\alias{ph_to_rxnclass}
\alias{rxnclass_to_ph}
\alias{ReactionClassLevels}
\title{Convert pH to/from Reaction Classes}
\usage{
ph_to_rxnclass(x, halfclass = FALSE, as.is = FALSE, droplevels = TRUE)

rxnclass_to_ph(x, halfclass = FALSE, digits = 2, simplify = TRUE)

ReactionClassLevels(halfclass = FALSE, as.is = FALSE)
}
\arguments{
\item{x}{input pH values (numeric; \code{ph_to_rxnclass()}) or reaction classes (character; \code{rxnclass_to_ph()})}

\item{halfclass}{Split the standard classes in half for higher resolution? Default: \code{FALSE}}

\item{as.is}{logical. Should character vectors be converted to factors? Default: \code{FALSE}}

\item{droplevels}{logical. Drop unused levels in factors? Default: \code{FALSE}}

\item{digits}{Number of digits after decimal place; Default: \code{2}. Used only for \code{rxnclass_to_ph()}}

\item{simplify}{Simplify list result to numeric vector when length of result is 1? Default: \code{TRUE}}
}
\value{
\code{ph_to_rxnclass()}: a vector of reaction classes corresponding to numeric input in \code{x}; if \code{as.is=FALSE} an ordered factor using \code{ReactionClassLevels()}

\code{rxnclass_to_ph()}: a list of data.frame objects containing high/low values of reaction class 1:1 with input; if simplify=TRUE and input is  a data.frame.

\code{ReactionClassLevels()}: ordered factor containing descriptive terms for reaction classes
}
\description{
Convert pH to/from Reaction Classes
}
\examples{
ph_to_rxnclass(6.2)
rxnclass_to_ph("slightly acid")

rxnclass_to_ph(list(c("Slightly Acid", NA, "Moderately Acid"),
                    c("Slightly Acid", NA, "Strongly Acid")), simplify = FALSE)
ReactionClassLevels()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{ca630}
\alias{ca630}
\title{Soil Data from the Central Sierra Nevada Region of California}
\format{
List containing:

$site : A data frame containing site information.  \describe{
\item{user_site_id}{national user site id} \item{mlra}{the
MLRA} \item{county}{the county} \item{ssa}{soil survey area}
\item{lon}{longitude, WGS84} \item{lat}{latitude, WGS84}
\item{pedon_key}{national soil profile id}
\item{user_pedon_id}{local soil profile id}
\item{cntrl_depth_to_top}{control section top depth (cm)}
\item{cntrl_depth_to_bot}{control section bottom depth (cm)}
\item{sampled_taxon_name}{soil series name} }

$lab : A data frame containing horizon information.  \describe{
\item{pedon_key}{national soil profile id}
\item{layer_key}{national horizon id}
\item{layer_sequence}{horizon sequence number}
\item{hzn_top}{horizon top (cm)} \item{hzn_bot}{horizon
bottom (cm)} \item{hzn_desgn}{horizon name}
\item{texture_description}{USDA soil texture}
\item{nh4_sum_bases}{sum of bases extracted by ammonium acetate (pH
7)} \item{ex_acid}{exchangeable acidity [method ?]}
\item{CEC8.2}{cation exchange capacity by sum of cations method (pH
8.2)} \item{CEC7}{cation exchange capacity by ammonium acetate (pH
7)} \item{bs_8.2}{base saturation by sum of cations method (pH 8.2)}
\item{bs_7}{base saturation by ammonium acetate (pH 7)} }
}
\source{
\url{https://ncsslabdatamart.sc.egov.usda.gov/}
}
\usage{
data(ca630)
}
\description{
Site and laboratory data from soils sampled in the central Sierra Nevada
Region of California.
}
\details{
These data were extracted from the NSSL database. \code{ca630} is a list composed
of site and lab data, each stored as \code{data.frame} objects. These data are modeled by a
1:many (site:lab) relation, with the \code{pedon_id} acting as the primary key in
the \code{site} table and as the foreign key in the \code{lab} table.
}
\note{
These data are out of date. Pending some new data + documentation. Use
with caution
}
\examples{

\dontrun{
library(tactile)
library(lattice)
library(Hmisc)
library(sp)

# check the data out:
data(ca630)
str(ca630)

# note that pedon_key is the link between the two tables

# make a copy of the horizon data
ca <- ca630$lab

# promote to a SoilProfileCollection class object
depths(ca) <- pedon_key ~ hzn_top + hzn_bot

# add site data, based on pedon_key
site(ca) <- ca630$site

# ID data missing coordinates: '|' is a logical OR
(missing.coords.idx <- which(is.na(ca$lat) | is.na(ca$lon)))

# remove missing coordinates by safely subsetting
if(length(missing.coords.idx) > 0)
	ca <- ca[-missing.coords.idx, ]

# register spatial data
initSpatial(ca) <- ~ lon + lat

# assign a coordinate reference system
prj(ca) <- 'EPSG:4269'

# check the result
print(ca)

# aggregate \%BS 7 for all profiles into 1 cm slices
a <- slab(ca, fm= ~ bs_7)

# plot median & IQR by 1 cm slice
xyplot(
top ~ p.q50, 
data = a, 
lower=a$p.q25, 
upper=a$p.q75,
alpha=0.5,
ylim=c(160,-5), 
scales = list(alternating = 1, y = list(tick.num = 7)),
panel = panel.depth_function, 
prepanel = prepanel.depth_function,
ylab='Depth (cm)', xlab='Base Saturation at pH 7',
par.settings = tactile.theme(superpose.line = list(col = 'black', lwd = 2))
)

# aggregate \%BS at pH 8.2 for all profiles by MLRA, along 1 cm slices
# note that mlra is stored in @site
a <- slab(ca, mlra ~ bs_8.2)

# keep only MLRA 18 and 22
a <- subset(a, subset=mlra \%in\% c('18', '22'))

# plot median & IQR by 1 cm slice, using different colors for each MLRA
xyplot(
top ~ p.q50, 
groups = factor(mlra), 
data = a,
lower=a$p.q25, 
upper=a$p.q75,
alpha=0.25,
sync.colors = TRUE,
ylim=c(160,-5), 
scales = list(alternating = 1, y = list(tick.num = 7)),
panel = panel.depth_function, 
prepanel = prepanel.depth_function,
ylab='Depth (cm)', xlab='Base Saturation at pH 7',
par.settings = tactile.theme(superpose.line = list(lwd = 2)),
auto.key = list(lines = TRUE, points = FALSE, columns = 2)
)



# Extract the 2nd horizon from all profiles as SPDF
ca.2 <- ca[, 2]

# subset profiles 1 through 10
ca.1.to.10 <- ca[1:10, ]

# basic plot method: profile plot
par(mar = c(0, 0, 3, 1))
plotSPC(ca.1.to.10, name='hzn_desgn', color = 'CEC7')
}

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-metadata.R
\docType{methods}
\name{hzDesgn,SoilProfileCollection-method}
\alias{hzDesgn,SoilProfileCollection-method}
\alias{hzDesgn}
\title{Get horizon designation column name}
\usage{
\S4method{hzDesgn}{SoilProfileCollection}(object)
}
\arguments{
\item{object}{a SoilProfileCollection}
}
\description{
Get horizon designation names
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getSurfaceHorizonDepth.R
\name{getSurfaceHorizonDepth}
\alias{getSurfaceHorizonDepth}
\alias{getMineralSoilSurfaceDepth}
\alias{getPlowLayerDepth}
\title{Determine thickness of horizons (continuous from surface) matching a pattern}
\usage{
getSurfaceHorizonDepth(
  p,
  pattern,
  hzdesgn = hzdesgnname(p, required = TRUE),
  simplify = TRUE
)

getMineralSoilSurfaceDepth(
  p,
  hzdesgn = hzdesgnname(p, required = TRUE),
  pattern = "O",
  simplify = TRUE
)

getPlowLayerDepth(
  p,
  hzdesgn = hzdesgnname(p, required = TRUE),
  pattern = "^Ap[^b]*",
  simplify = TRUE
)
}
\arguments{
\item{p}{a SoilProfileCollection}

\item{pattern}{a regular expression pattern to match for all horizons to be considered part of the "surface".}

\item{hzdesgn}{column name containing horizon designation. Default: \code{hzdesgnname(p, required = TRUE)}.}

\item{simplify}{logical. Return single profile results as vector (default: \code{TRUE}) or \code{data.frame} (\code{FALSE})}
}
\value{
a numeric value corresponding to the bottom depth of the last horizon matching 'pattern' that is contiguous with other matching horizons up to the soil surface. If \code{length(p) > 1} then a \emph{data.frame} containing profile ID, horizon ID, top or bottom depths, horizon designation and pattern.
}
\description{
Find the thickness of horizon designations, or any other character patterns, that are continuous from the soil surface (depth = 0 or shallowest depth in profile).
}
\details{
The horizon designation to match is specified with the regular expression pattern 'pattern'. All horizons matching that pattern, that are continuous from the soil surface, count towards the depth / thickness value that is ultimately returned. For instance: horizon designations: A1-A2-A3-C-Ab , would return A3 bottom depth given \code{pattern = "^A[1-9]*$"}.

\code{getSurfaceHorizonDepth} is used by \code{getPlowLayerDepth} for matching Ap horizons; and, it is used by \code{getMineralSoilSurfaceDepth} to find the thickness of O horizons in lieu of lab data.
}
\examples{
library(aqp)
data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

p <- sp1[1]
q <- sp1[2]

# look at horizon designations in p and q
p$name
q$name

# thickness of all surface horizons containing A
getSurfaceHorizonDepth(p, pattern = 'A', hzdesgn = 'name')

# thickness of all surface horizons that start with A
getSurfaceHorizonDepth(p, pattern = '^A', hzdesgn = 'name')

# thickness of all surface horizons that start with A, and the A is not followed by B
getSurfaceHorizonDepth(p, pattern = '^A[^B]*', hzdesgn = 'name')

# thickness of all surface horizons that start with A
#  followed by a number from _2_ to 9 (returns ZERO)
getSurfaceHorizonDepth(p, pattern = '^A[2-9]*', hzdesgn = 'name')

# getPlowLayerDepth matches first two horizons in fake Ap horizon data with "buried Ap"
p$aphorizons <- c("Ap1","Ap2","AB", rep('C', nrow(p) - 4), "Apb")
getPlowLayerDepth(p, hzdesgn = 'aphorizons')

# getMineralSoilSurfaceDepthmatches first 3 horizons in fake O horizon data
p$ohorizons <- c("Oi1","Oi2","Oe", rep('C', nrow(p) - 4), "2C")
getMineralSoilSurfaceDepth(p, hzdesgn='ohorizons')

# matches first Oi horizon with original horizon designations of pedon 2
getMineralSoilSurfaceDepth(q, hzdesgn='name')

}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/texture.R
\name{texcl_to_ssc}
\alias{texcl_to_ssc}
\alias{ssc_to_texcl}
\alias{texmod_to_fragvoltot}
\alias{texture_to_taxpartsize}
\alias{texture_to_texmod}
\alias{fragvol_to_texmod}
\title{Textural conversions}
\usage{
texcl_to_ssc(texcl = NULL, clay = NULL, sample = FALSE)

ssc_to_texcl(
  sand = NULL,
  clay = NULL,
  simplify = FALSE,
  as.is = FALSE,
  droplevels = TRUE
)

texmod_to_fragvoltot(texmod = NULL, lieutex = NULL)

texture_to_taxpartsize(
  texcl = NULL,
  clay = NULL,
  sand = NULL,
  sandvf = NULL,
  fragvoltot = NULL
)

texture_to_texmod(texture, duplicates = "combine")

fragvol_to_texmod(
  object = NULL,
  gravel = NULL,
  cobbles = NULL,
  stones = NULL,
  boulders = NULL,
  channers = NULL,
  flagstones = NULL,
  paragravel = NULL,
  paracobbles = NULL,
  parastones = NULL,
  paraboulders = NULL,
  parachanners = NULL,
  paraflagstones = NULL,
  as.is = TRUE,
  droplevels = TRUE
)
}
\arguments{
\item{texcl}{vector of texture classes than conform to the USDA code
conventions (e.g. c|C, sil|SIL, sl|SL, cos|COS)}

\item{clay}{vector of clay percentages}

\item{sample}{logical: should ssc be random sampled from the lookup table?
(default: FALSE)}

\item{sand}{vector of sand percentages}

\item{simplify}{Passed to \code{SoilTextureLevels()} to set the number of possible texture classes. If \code{TRUE}, the ordered factor has a maximum of 12 levels, if \code{FALSE} (default) the ordered factor has a maximum of 21 levels (including e.g. very fine/fine/coarse variants)}

\item{as.is}{logical: should character vectors be converted to factors?
(default: TRUE)}

\item{droplevels}{logical: indicating whether to drop unused levels in
factors. This is useful when the results have a large number of unused
classes, which can waste space in tables and figures.}

\item{texmod}{vector of textural modifiers that conform to the USDA code
conventions (e.g. gr|GR, grv|GRV)}

\item{lieutex}{vector of in lieu of texture terms that conform to the USDA
code conventions (e.g. gr|GR, pg|PG), only used when fragments or artifacts
are > 90 percent by volume (default: NULL))}

\item{sandvf}{vector of very fine sand percentages}

\item{fragvoltot}{vector of total rock fragment percentages}

\item{texture}{vector of combinations of texcl, texmod, and lieutex (e.g. CL, GR-CL, CBV-S, GR)}

\item{duplicates}{character: specifying how multiple values should be handled, options are \code{"combined"} (e.g. 'GR & GRV) or \code{"max"}(e.g. 'GRV')}

\item{object}{data.frame: containing the following column names: gravel, cobbles,
stones, boulders, channers, flagstones, paragravel, paracobbles, parastones,
paraboulders, parachanners, paraflagstones}

\item{gravel}{numeric: gravel volume \%}

\item{cobbles}{numeric: cobble volume \%}

\item{stones}{numeric: stone volume \%}

\item{boulders}{numeric: boulder volume \%}

\item{channers}{numeric: channer volume \%}

\item{flagstones}{numeric: flagstone volume \%}

\item{paragravel}{numeric: para gravel volume \%}

\item{paracobbles}{numeric: para cobble volume \%}

\item{parastones}{numeric: para stone volume \%}

\item{paraboulders}{numeric: para boulder volume \%}

\item{parachanners}{numeric: para channer volume \%}

\item{paraflagstones}{numeric: para flagstone volume \%}
}
\value{
\itemize{
\item \code{texcl_to_ssc}: A \code{data.frame} containing columns \code{"sand"},\code{"silt"}, \code{"clay"}
}

\itemize{
\item \code{ssc_to_texcl}: A \code{character} vector containing texture class
}

\itemize{
\item \code{texmod_to_fragvoltot}: A \code{data.frame} containing columns \code{"fragvoltot_l"},
\code{"fragvoltot_r"}, \code{"fragvoltot_h"}, \code{"fragvoltot_l_nopf"}, \code{"fragvoltot_r_nopf"}, \code{"fragvoltot_h_nopf"}
}

\itemize{
\item \code{texture_to_taxpartsize}: a character vector containing \code{"taxpartsize"} classes
}

\itemize{
\item \code{texture_to_texmod}: a character vector containing \code{"texmod"} classes
}

\itemize{
\item \code{texmod_to_fragvol}: a data.frame containing \code{"texmod"} and \code{"lieutex"} classes
}
}
\description{
These functions consist of several conversions between sand, silt and clay
to texture class and visa versa, textural modifiers to rock fragments, and
grain size composition to the family particle size class.
}
\details{
These functions are intended to estimate missing values or allocate particle
size fractions to classes. The \code{ssc_to_texcl()} function uses the same
logic as the particle size estimator calculation in NASIS to classify sand
and clay into texture class. The results are stored in \code{soiltexture}
and used by \code{texcl_to_ssc()} as a lookup table to convert texture class
to sand, silt and clay. The function \code{texcl_to_ssc()} replicates the
functionality described by Levi (2017). The \code{texmod_to_fragvol()}
function similarly uses the logical from the
Exhibit618-11_texture_modifier.xls spreadsheet to determine the textural
modifier from the various combinations of rock and pararock fragments
(e.g. GR and PGR).

When \code{sample = TRUE}, the results can be used to estimate within-class,
marginal distributions of sand, silt, and clay fractions. It is recommended
that at least 10 samples be drawn for reasonable estimates.

The function \code{texmod_to_fragvoltot} returns a data.frame with multiple
fragvoltot columns differentiated by tailing abbreviations (e.g. _r) which
refer to the following:
\enumerate{
\item l = low
\item r = representative
\item h = high
\item nopf = no pararock fragments (i.e. total fragments - pararock fragments)
}

The function \code{texture_to_texmod()} parses texture (e.g. GR-CL) to extract the texmod values from it in the scenario where it is missing from
texmod column. If multiple texmod values are present (for example in the
case of stratified textures) and \code{duplicates = "combine"} they will be combined in the output (e.g. GR & CBV). Otherwise if \code{duplicates = "max"}
the texmod with the highest rock fragment (e.g. CBV) will be returned.

Unlike the other functions, \code{texture_to_taxpartsize()} is intended to
be computed on weighted averages within the family particle size control
section. Also recall from the criteria that carbonate clay should be
subtracted from clay content and added to silt content. Similarly, if the
percent of very fine sand is known it should be subtracted from the sand,
and added to the silt content. Unlike the other functions,
\code{texture_to_taxpartsize()} is intended to be computed on weighted
averages within the family particle size control section. Also recall from
the criteria that carbonate clay should be subtracted from clay content and
added to silt content. Similarly, if the percent of very fine sand is known
it should be subtracted from the sand, and added to the silt content.
}
\examples{
\donttest{
# example of ssc_to_texcl()
tex <- expand.grid(sand = 0:100, clay = 0:100)
tex <- subset(tex, (sand + clay) < 101)
tex$texcl <- ssc_to_texcl(sand = tex$sand, clay = tex$clay)
head(tex)

# example of texcl_to_ssc(texcl)
texcl <- c("cos", "s", "fs", "vfs", "lcos", "ls",
          "lfs", "lvfs", "cosl", "sl", "fsl", "vfsl", "l",
          "sil", "si", "scl", "cl", "sicl", "sc", "sic", "c"
          )
test <- texcl_to_ssc(texcl)
head(test <- cbind(texcl, test), 10)


# example of texcl_to_ssc(texcl, clay)
data(soiltexture)
st <- soiltexture$values
idx <- sample(1:length(st$texcl), 10)
st <- st[idx, ]
ssc <- texcl_to_ssc(texcl = st$texcl)
head(cbind(texcl = st$texcl, clay = ssc$clay))


# example of texmod_to_fragvoltol
frags <- c("gr", "grv", "grx", "pgr", "pgrv", "pgrx")
head(texmod_to_fragvoltot(frags))


# example of texture_to_taxpartsize()
tex <- data.frame(texcl = c("c", "cl", "l", "ls", "s"),
                  clay  = c(55, 33, 18, 6, 3),
                  sand  = c(20, 33, 42, 82, 93),
                  fragvoltot = c(35, 15, 34, 60, 91))
tex$fpsc <- texture_to_taxpartsize(texcl = tex$texcl,
                                   clay = tex$clay,
                                   sand = tex$sand,
                                   fragvoltot = tex$fragvoltot)
head(tex)


# example of texture_to_taxpartsize() with carbonate clay and very fine sand
carbclay <- rnorm(5, 2, 3)
vfs <- rnorm(5, 10, 3)
st$fpsc <- texture_to_taxpartsize(texcl = tex$texcl,
                                  clay = tex$clay - carbclay,
                                  sand = tex$sand - vfs,
                                  fragvoltot = tex$fragvoltot)
head(tex)


# example of sample = TRUE
texcl <- rep(c("cl", "sil", "sl"), 10)
ssc1 <- cbind(texcl, texcl_to_ssc(texcl = texcl, sample = FALSE))
ssc2 <- cbind(texcl, texcl_to_ssc(texcl = texcl, sample = TRUE))
ssc1$sample <- FALSE
ssc2$sample <- TRUE
ssc  <- rbind(ssc1, ssc2)
aggregate(clay ~ sample + texcl, data = ssc, summary)
}
\donttest{
# example of texture_to_texmod()
tex <- c("SL", "GR-SL", "CBV-L", "SR- GR-FS GRX-COS")
texture_to_texmod(tex)
texture_to_texmod(tex, duplicates = "max")
}
\donttest{
# example of fragvol_to_texmod()
df <- expand.grid(
  gravel  = seq(0, 100, 5), 
  cobbles = seq(0, 100, 5), 
  stones  = seq(0, 100, 5), 
  boulders = seq(0, 100, 5)
  )
df <- df[rowSums(df) < 100, ]

# data.frame input
test <- fragvol_to_texmod(df)
table(test$texmod)
table(test$lieutex)

# vector inputs
fragvol_to_texmod(gravel = 10, cobbles = 10)

}
}
\references{
Matthew R. Levi, Modified Centroid for Estimating Sand, Silt, and Clay from Soil Texture Class, Soil Science Society of America Journal, 2017, 81(3):578-588, ISSN 1435-0661, \doi{10.2136/sssaj2016.09.0301}.
}
\seealso{
\code{\link{SoilTextureLevels}}

\code{\link[=hz_to_taxpartsize]{hz_to_taxpartsize()}}, \code{\link[=lookup_taxpartsize]{lookup_taxpartsize()}}
}
\author{
Stephen Roecker
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\name{subsetHz,SoilProfileCollection-method}
\alias{subsetHz,SoilProfileCollection-method}
\alias{subsetHz}
\title{Subset the horizons in a SoilProfileCollection using logical criteria}
\usage{
\S4method{subsetHz}{SoilProfileCollection}(x, ..., drop = TRUE)
}
\arguments{
\item{x}{a SoilProfileCollection}

\item{...}{Comma-separated set of R expressions that evaluate as \code{TRUE} or \code{FALSE} in context of horizon data frame. Length for individual expressions matches number of horizons, in \code{x}.}

\item{drop}{Default: \code{TRUE}. When \code{drop=FALSE} placeholder horizons (profile ID with all other values \code{NA}) are created where the specified filter results in removal of all horizons.}
}
\value{
a SoilProfileCollection with a subset of horizons, possibly with some sites removed
}
\description{
\code{subsetHz()} is a function used for extracting horizons from a SoilProfileCollection based on logical criteria.
}
\details{
To minimize likelihood of issues with non-standard evaluation context, especially when using \code{subsetHz()} inside another function, all expressions used in \code{...} should be in terms of variables that are in the horizon data frame.
}
\examples{

data(sp3)

depths(sp3) <- id ~ top + bottom

# show just horizons with 10YR hues
plot(subsetHz(sp3, hue == '10YR'))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hzBoundary.R
\name{hzDistinctnessCodeToOffset}
\alias{hzDistinctnessCodeToOffset}
\title{Convert Horizon Boundary Distinctness to Vertical Offset}
\usage{
hzDistinctnessCodeToOffset(
  x,
  codes = c("very abrupt", "abrupt", "clear", "gradual", "diffuse"),
  offset = c(0.5, 2, 5, 15, 20)/2
)
}
\arguments{
\item{x}{vector of boundary distinctness codes to be converted}

\item{codes}{character vector of distinctness terms ('clear') or codes ('C'), case insensitive, see details}

\item{offset}{vertical offset factors (cm), approximating 1/2 of the transitional zone thickness, see details}
}
\value{
vector of offsets with same length as \code{x}
}
\description{
This function will convert USDA-NCSS horizon boundary distinctness codes into vertical (+/-) offsets in cm, based on the \href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}.
}
\details{
The default offsets are based on the high-end of ranges presented in "transitional zone thickness criteria" from the Field Book version 3.0 (page 2-6). Offsets are returned as 1/2 of the transitional zone thickness so that horizon boundaries can be adjusted up/down from horizon depths. See \code{\link{plotSPC}}, specifically the \code{hz.distinctness.offset} argument for visualization ideas. Missing data in \code{x} (NA) or codes that are not defined in \code{codes} are returned as 0 offsets.

Either format (or mixture) are accepted, case insensitive:
\itemize{
\item terms: \code{c('very abrupt', 'abrupt', 'clear', 'gradual', 'diffuse')}
\item coded values: \verb{c('v', 'a', 'c', 'g', d')}
}

Additional examples are available in the \href{https://ncss-tech.github.io/AQP/aqp/hz-boundaries.html}{Visualization of Horizon Boundaries tutorial}.
}
\examples{
# example data
data(sp1)

# compute 1/2 transitional zone thickness from distinctness codes
sp1$hzdo <- hzDistinctnessCodeToOffset(sp1$bound_distinct)

# convert colors from Munsell to hex-encoded RGB
sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))

# promote to SoilProfileCollection
depths(sp1) <- id ~ top + bottom
hzdesgnname(sp1) <- 'name'

# adjust margins
op <- par(mar=c(0,0,0,1.5))

# sketches, adjust width, adjust text size, include coded hz distinctness offsets
plotSPC(sp1, width=0.3, cex.names=0.75, hz.distinctness.offset = 'hzdo')

# clean-up
par(op)
}
\references{
\href{https://nrcspad.sc.egov.usda.gov/DistributionCenter/product.aspx?ProductID=991}{Field Book for Describing and Sampling Soils, version 3.0}
}
\seealso{
\code{\link{plotSPC}}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addVolumeFraction.R
\name{addVolumeFraction}
\alias{addVolumeFraction}
\title{Symbolize Volume Fraction within a Soil Profile Collection Plot}
\usage{
addVolumeFraction(
  x,
  colname,
  res = 10,
  cex.min = 0.1,
  cex.max = 0.5,
  pch = 1,
  col = "black"
)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}

\item{colname}{character vector of length 1, naming the column containing volume fraction data (horizon-level attribute). Values should be within 0-100 percent.}

\item{res}{integer, resolution of the grid used to symbolize volume fraction}

\item{cex.min}{minimum symbol size}

\item{cex.max}{maximum symbol size}

\item{pch}{integer, plotting character code}

\item{col}{symbol color, either a single color or as many colors as there are horizons in \code{x}}
}
\description{
Symbolize volume fraction on an existing soil profile collection plot.
}
\details{
This function can only be called after plotting a \code{SoilProfileCollection} object. Details associated with a call to \code{plotSPC()} are automatically accounted for within this function: e.g. \code{plot.order}, \code{width}, etc..
}
\note{
It may be necessary to adjust both \code{res}, \code{cex.min}, and \code{cex.max} for optimal legibility.
}
\seealso{
\code{\link[=plotSPC]{plotSPC()}}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fragmentClasses.R
\name{fragmentClasses}
\alias{fragmentClasses}
\title{Coarse Fragment Class Labels and Diameter}
\usage{
fragmentClasses(
  sys = c("usda_simplified", "usda", "international", "unified", "aashto",
    "mod.wentworth"),
  flat = FALSE,
  rounded = FALSE
)
}
\arguments{
\item{sys}{character, length 1. This is an abbreviated name used to select class labels and fragment diameter.}

\item{flat}{logical. Fragments are flat, only used by USDA systems.}

\item{rounded}{logical. Fragments are rounded, only used by AASHTO system.}
}
\value{
named vector of fragment diameter in mm
}
\description{
This is a convenience function for accessing coarse fragment class labels and associated diameter (mm), as defined in various classification systems such as USDA, Unified, and AASHTO.
}
\examples{

# use default system: "usda_simplified"
fragmentClasses()
fragmentClasses(flat = TRUE)

fragmentClasses(sys = 'usda')
fragmentClasses(sys = 'USDA', flat = TRUE)

fragmentClasses(sys = 'international')

fragmentClasses(sys = 'unified')

fragmentClasses(sys = 'aashto')
fragmentClasses(sys = 'aashto', rounded = TRUE)

fragmentClasses(sys = 'mod.wentworth')

}
\references{
Schoeneberger, P.J., D.A. Wysocki, E.C. Benham, and Soil Survey Staff. 2012. Field book for describing and sampling soils, Version 3.0. Natural Resources Conservation Service, National Soil Survey Center, Lincoln, NE.
}
\seealso{
\code{\link[=fragmentSieve]{fragmentSieve()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.increase.matrix.R
\name{get.increase.matrix}
\alias{get.increase.matrix}
\alias{get.increase.depths}
\title{Compute Pair-wise Distances of Soil Properties over Depth}
\usage{
get.increase.matrix(p, attr, threshold.fun, vertical.distance)

get.increase.depths(p, attr, threshold.fun, vertical.distance)
}
\arguments{
\item{p}{a SoilProfileCollection, containing a single profile}

\item{attr}{horizon attribute name to get the "increase" of}

\item{threshold.fun}{a function that returns the threshold (as a function of
attr); may return a constant single value}

\item{vertical.distance}{the vertical distance (determined from difference
SPC top depth variable) within which increase must be met}
}
\value{
Returns a square logical matrix reflecting where the increase
criteria were met.

\code{get.increase.depths} converts to horizon top depth by using above
matrix output to determine depths where increase is met.

Returns a numeric vector of depths where the increase requirement is
met. For the argillic, the first is the one of interest.

\code{get.increase.depths()} converts to horizon top depth by using above
matrix output to determine depths where increase is met.
}
\description{
Computes pair-wise distance matrix to determine where an attribute increases within
a specified vertical distance threshold.

\code{get.increase.depths} performs the conversion of the square matrix
output of \code{get.increase.matrix} back to horizon top depth for where
criteria were met.
}
\details{
Uses matrix outer product to determine all pair-wise differences in \code{attr}
for the horizons of \code{p}. Supplies \code{attr} to \code{threshold.fun} to determine the
minimum value criterion to return TRUE in output matrix for an "increase".
Also, computes all pair-wise distances in depth dimension to determine
whether the vertical distance criteria have been met simultaneously with
\code{attr} increase.

This function assumes that the \code{threshold.fun} supplied by the user returns
either a constant or a vector of equal length to its input.

Note that the \code{threshold.fun} result is allowed to contain NA, but that will
result in no output for affected cells.

\code{get.increase.depths()} performs the conversion of the square matrix
output of \code{get.increase.matrix} back to horizon top depth for where
criteria were met.

Note that the \code{threshold.fun} result is allowed to contain \code{NA}, but that will
result in no output for affected cells.
}
\examples{

data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

p <- sp1[1]
attr <- 'prop' # clay contents
foo <- get.increase.matrix(p, threshold.fun = crit.clay.argillic,
                           attr = attr, vertical.distance = 30)
foo


data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

p <- sp1[1]
attr <- 'prop' # clay contents
foo <- get.increase.depths(p, threshold.fun = crit.clay.argillic,
                           attr = attr, vertical.distance = 30)
foo

}
\seealso{
\code{\link[=getArgillicBounds]{getArgillicBounds()}}, \code{\link[=crit.clay.argillic]{crit.clay.argillic()}}

\code{\link[=getArgillicBounds]{getArgillicBounds()}} \code{\link[=crit.clay.argillic]{crit.clay.argillic()}}
}
\author{
Andrew Gene Brown
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimateSoilColor.R
\name{estimateSoilColor}
\alias{estimateSoilColor}
\title{Estimate dry soil colors from moist soil colors and vice versa.}
\usage{
estimateSoilColor(
  hue,
  value,
  chroma,
  method = c("procrustes", "ols"),
  sourceMoistureState = c("dry", "moist"),
  returnMunsell = TRUE
)
}
\arguments{
\item{hue}{vector of Munsell hue ('10YR', '2.5Y', etc.)}

\item{value}{vector of Munsell value (2,2.5 2.5, 3, 5, 6, etc.)}

\item{chroma}{vector of Munsell chroma (2, 3, 4, etc.)}

\item{method}{character, one of 'procrustes' or 'ols', see details}

\item{sourceMoistureState}{character, source colors are either 'dry' or 'moist'}

\item{returnMunsell}{logical, \code{TRUE}: return closest Munsell chip, \code{FALSE}: return estimated CIELAB coordinates}
}
\value{
\code{data.frame} of estimated colors in Munsell notation. The \code{sigma} column contains CIE2000 color contrast metric values describing the perceptual distance between estimated color in CIELAB coordinates and closest Munsell chip.
}
\description{
All else equal, soil color will predictably shift in perceived lightness (change in Munsell value) as moisture content changes. Field-described soil colors are typically collected at approximately air dry ("dry") and field capacity ("moist") states. This function estimates "dry" soil colors from "moist" soil colors and vice versa. Two methods are available for estimation, both developed from a national collection of field-described soil colors (approx. 800k horizons). Estimates are only valid for mineral soil material, having Munsell values and chroma < 10. Estimation has a median error rate of approximately (CIE dE2000) 5.

Available Methods:
\itemize{
\item "procrustes": soil colors are converted using scale, rotation, and translation parameters in CIELAB color space
\item "ols": soil colors are converted using 3 multiple linear regression models (CIELAB coordinates)
}

This is still a work in progress.
}
\details{
For both methods, estimation proceeds as:
\itemize{
\item convert Munsell notation to CIELAB color coordinates via \code{munsell2rgb()}
\item apply rotation or regression model to color coordinates in CIELAB space
\item locate closest Munsell chip to resulting CIELAB coordinates via \code{col2munsell()}
}

Estimation of dry from moist soil color state is not guaranteed to be symmetric with estimation of moist from dry.

Scaling, rotation, and translation parameters for shifting between dry <--> moist CIELAB coordinates were determined using \code{vegan::procrustes()}. Multiple linear regression models were fit using \code{rms::ols()}.

Estimation error for both methods, converting from either moisture states is, approximately 5 (CIE dE 2000).
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

estimateSoilColor(hue = '10YR', value = 3, chroma = 3, sourceMoistureState = 'moist')

# estimation is not always symmetric
estimateSoilColor(hue = '10YR', value = 4, chroma = 3, sourceMoistureState = 'dry')

# more examples
estimateSoilColor(hue = '2.5Y', value = 8, chroma = 2, sourceMoistureState = 'dry')
estimateSoilColor(hue = '2.5YR', value = 3, chroma = 4, sourceMoistureState = 'moist')

estimateSoilColor(hue = 'N', value = 2, chroma = 0, sourceMoistureState = 'moist')

estimateSoilColor(hue = '7.5YR', value = 2, chroma = 2, sourceMoistureState = 'moist')

# resulting hue is not always the same
estimateSoilColor(hue = '5G', value = 6, chroma = 6, sourceMoistureState = 'dry')

# return estimated CIELAB coordinates
estimateSoilColor(hue = '5G', value = 6, chroma = 6, sourceMoistureState = 'dry',
 returnMunsell = FALSE)

}
\references{
J. A. Shields, E. A. Paul, R. J. St. Arnaud, and W. K. Head. 1968. SPECTROPHOTOMETRY MEASUREMENT OF SOIL COLOR AND ITS RELATIONSHIP TO MOISTURE AND ORGANIC MATTER. Canadian Journal of Soil Science. 48(3): 271-280. https://doi.org/10.4141/cjss68-037
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorVariation.R
\name{colorVariation}
\alias{colorVariation}
\title{Quantitative Description of Color Variation}
\usage{
colorVariation(
  m,
  method = c("frequency", "centroid", "L1", "reference"),
  ref = NULL
)
}
\arguments{
\item{m}{character vector of colors, described using the Munsell system e.g. \code{c('10YR 3/3', '5YR 4/6')}}

\item{method}{character, one of \code{c('frequency', 'centroid', 'reference')}, see Details}

\item{ref}{character, a reference color specified in the Munsell system when \code{method = 'reference'}}
}
\value{
numeric, dE00 summary of color variation along with group centroid for \code{method = c('frequency', 'centroid', 'L1')}
}
\description{
This function computes several measures of "color variation", typically associated with soil colors described in the Munsell system, using the CIE2000 dE (dE00) color contrast metric. The resulting dE00 summaries map closely to color differences as perceived by "average human vision".
}
\details{
dE00 values are computed according to \code{method}:
\itemize{
\item 'frequency': relative to most frequent color in \code{m}
\item 'centroid': relative to centroid (CIELAB coordinates) of colors specified in \code{m}
\item 'L1': relative to L1-median (geometric median) CIELAB coordinates of colors specified in \code{m}, via \code{Gmedian::Gmedian()}
\item 'reference': relative to color specified in \code{ref}
}

The \code{L1} method is more robust to outliers in \code{m} as compared to other methods.
}
\examples{

# some brownish colors with a wild outlier
m <- c('10YR 3/3', '10YR 4/4', '10YR 4/4', '5GY 6/8')

# useful when there may be a lot of duplicates
# error when there is not a single, most-frequent color
colorVariation(m, method = 'frequency')

# statistical "centroid" of colors, not robust to outliers
# result may not match any color in `m`
colorVariation(m, method = 'centroid')

# deviation from a known reference
colorVariation(m, method = 'reference', ref = '10YR 2/1')

# L1-median (requires Gmedian package) like 'centroid'
# more robust to outliers
# result will usually be very close to a color in `m`
if(requireNamespace('Gmedian')) {
  colorVariation(m, method = 'L1')
}


# compare methods using a range of colors on a 
# single hue page
x <- expand.grid(
  hue = '10YR', 
  value = 2:7, 
  chroma = 2:7
)

x$m <- sprintf("\%s \%s/\%s", x$hue, x$value, x$chroma)

colorChart(x$m)

(v <- colorVariation(x$m, method = 'centroid'))
contrastChart(attr(v, 'centroid'), hues = x$hue[1], thresh = v)

if(requireNamespace('Gmedian')) {
  (v <- colorVariation(x$m, method = 'L1'))
  contrastChart(attr(v, 'L1'), hues = x$hue[1], thresh = v)
}


# attempt to simulate colors from a group centroid
\dontrun{
v <- colorVariation(x$m, method = 'centroid')

s <- simulateColor(
method = 'dE00', 
n = 200, 
parameters = list(m = attr(v, 'centroid'), thresh = v * 1.96, hues = x$hue[1])
)
colorChart(s[[1]])
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-coercion.R
\docType{methods}
\name{as}
\alias{as}
\alias{as,SoilProfileCollection-method}
\alias{as.data.frame,SoilProfileCollection-method}
\title{Coerce SoilProfileCollection with \code{as()}}
\usage{
\S4method{as.data.frame}{SoilProfileCollection}(x)
}
\arguments{
\item{x}{a SoilProfileCollection}
}
\value{
list

data.frame

tbl_df

data.table

SpatialPointsDataFrame

sf

SpatialPoints
}
\description{
SoilProfileCollections can be coerced to other R object types using \code{as(spc, 'type')}.

Possible endpoints include: \code{list}, \code{data.frame}, \code{SpatialPointsDataFrame} and \code{SpatialPoints}.
}
\examples{
# load example data stored as SoilProfileCollection
data(sp5)

# sp5
str(sp5)

# list output
str(as(sp5, 'list'))

# data.frame output
str(as(sp5, 'data.frame'))

# Spatial Objects
# make some random coordinate data for each profile
sp5$x <- sp5$y <- rnorm(length(sp5))
initSpatial(sp5, crs = "OGC:CRS84") <- ~ x + y

if (requireNamespace("sf")) {

  # sf output
  str(as(sp5, 'sf'))

  # SpatialPointsDataFrame output
  str(as(sp5, 'SpatialPointsDataFrame'))

  # SpatialPoints output
  str(as(sp5, 'SpatialPoints'))
  
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tauW.R
\name{tauW}
\alias{tauW}
\alias{summaryTauW}
\title{Compute weighted naive and \emph{tau} statistics for a cross-classification matrix}
\usage{
tauW(
  CM,
  W = diag(sqrt(length(as.matrix(CM)))),
  P = rep(1/nrow(as.matrix(CM)), nrow(as.matrix(CM)))
)

summaryTauW(result.tau)
}
\arguments{
\item{CM}{a square confusion (cross-classification) matrix (rows: allocation, columns: reference)}

\item{W}{weights: 1 on diagonals, [-1..1] off giving partial credit to this error}

\item{P}{prior probability vector, length = number of rows/columns in \code{CM} and \code{W}}

\item{result.tau}{\code{tauW()} result}
}
\value{
Results are returned in a list with obvious R names
}
\description{
\code{tauW}: Computes: (1) unweighted naive, (2) weighted naive, (3) unweighted \emph{tau}, (4) weighted \emph{tau} accuracy statistics
}
\details{
\code{summaryTauW}: prints a summary of the results from \emph{tauW}

\code{xtableTauW}: formats a LaTeX table with results from \emph{tauW} and saves it as a \code{.tex} file for import into a LaTeX document.

Input matrices \code{CM} and \code{W} may be in \code{data.frame} format and will be converted

Weights matrix \code{W}: 0 = no credit; 1 = full credit; -1 = maximum penalty

If absent, default is no partial credit, i.e., unweighted.

Prior probabilities vector \code{P}: If absent, \code{P} are equal priors for each class. Special value \code{P = 0} is interpreted as \code{P} = column marginals.

Error checks: \code{CM} must be square; \code{P} must have correct number of classes and sum to 1 +/- 0.0001; \code{W} & \code{CM} must be conformable
}
\examples{

# example confusion matrix
# rows: allocation (user's counts)
# columns: reference (producer's counts)
crossclass <- matrix(data=c(2,1,0,5,0,0,
                            1,74,2,1,3,6,
                            0,5,8,6,1,3,
                            6,1,3,91,0,0,
                            0,4,0,0,0,4,
                            0,6,2,2,4,38),
                     nrow=6, byrow=TRUE)
row.names(crossclass) <- c("OP", "SA", "UA", "UC", "AV", "AC")
colnames(crossclass) <- row.names(crossclass)

# build the weights matrix
# how much credit for a mis-allocation
weights <- matrix(data=c(1.00,0.05,0.05,0.15,0.05,0.15,
                         0.05,1.00,0.05,0.05,0.05,0.35,
                         0.05,0.05,1.00,0.20,0.15,0.15,
                         0.15,0.05,0.25,1.00,0.10,0.25,
                         0.05,0.10,0.15,0.10,1.00,0.15,
                         0.20,0.30,0.10,0.25,0.20,1.00),
                  nrow=6, byrow=TRUE)

# unweighted accuracy
summaryTauW(nnaive <- tauW(crossclass))

# unweighted tau with equal priors, equivalent to Foody (1992) modified Kappa
tauW(crossclass)$tau

# unweighted tau with user's = producer's marginals, equivalent to original kappa
(priors <-  apply(crossclass, 2, sum)/sum(crossclass))
tauW(crossclass, P=priors)$tau

# weighted accuracy; tau with equal priors
summaryTauW(weighted <- tauW(crossclass, W=weights))

# weighted accuracy; tau with user's = producer's marginals
summaryTauW(tauW(crossclass, W=weights, P=priors))

# change in accuracy statistics weighted vs. non-weighted
(weighted$overall.weighted - weighted$overall.naive)
(weighted$user.weighted - weighted$user.naive)
(weighted$prod.weighted - weighted$prod.naive)

}
\references{
\itemize{
\item Rossiter, D. G., Zeng, R., & Zhang, G.-L. (2017). Accounting for taxonomic distance in accuracy assessment of soil class predictions. Geoderma, 292, 118–127. \doi{10.1016/j.geoderma.2017.01.012}
\item Ma, Z. K., & Redmond, R. L. (1995). Tau-coefficients for accuracy assessment of classification of remote-sensing data. Photogrammetric Engineering and Remote Sensing, 61(4), 435–439.
\item Naesset, E. (1996). Conditional tau coefficient for assessment of producer’s accuracy of classified remotely sensed data. ISPRS Journal of Photogrammetry and Remote Sensing, 51(2), 91–98. \doi{10.1016/0924-2716(69)00007-4}
}
}
\author{
D.G. Rossiter
}
\keyword{array}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\docType{methods}
\name{length,SoilProfileCollection-method}
\alias{length,SoilProfileCollection-method}
\alias{length}
\title{Get the number of profiles in a SoilProfileCollection}
\usage{
\S4method{length}{SoilProfileCollection}(x)
}
\arguments{
\item{x}{a SoilProfileCollection}
}
\description{
Get the number of profiles in a SoilProfileCollection
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotMultipleSPC.R
\name{profileGroupLabels}
\alias{profileGroupLabels}
\title{Soil Profile Group Labels}
\usage{
profileGroupLabels(
  x0,
  x1,
  labels,
  y0 = 100,
  y1 = 98,
  label.offset = 2,
  label.cex = 0.75
)
}
\arguments{
\item{x0}{integer indices to the first profile within each group}

\item{x1}{integer indices to the last profile within each group}

\item{labels}{vector of group labels}

\item{y0}{baseline depth used for group brackets}

\item{y1}{depth used for start and end markers for group brackets (see
examples)}

\item{label.offset}{vertical offset of group labels from baseline}

\item{label.cex}{label size}
}
\description{
Labels groups of soil profiles within soil profile sketches.

See examples below for ideas.
}
\note{
This function is typically called by some other convenience function
such as \code{\link{plotMultipleSPC}}.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# load sample data
data(sp3)
data(sp4)

# convert soil colors
sp3$h <- NA ; sp3$s <- NA ; sp3$v <- NA
sp3.rgb <- with(sp3, munsell2rgb(hue, value, chroma, return_triplets=TRUE))
sp3[, c('h','s','v')] <- t(with(sp3.rgb, rgb2hsv(r, g, b, maxColorValue=1)))

# promote to SoilProfileCollection
depths(sp3) <- id ~ top + bottom
depths(sp4) <- id ~ top + bottom

# combine into a list
spc.list <- list(sp3, sp4)

# compute group lengths and start/stop locations
n.groups <- length(spc.list)
spc.lengths <- sapply(spc.list, length)
n.pedons <- sum(spc.lengths)
group.starts <- c(1, 1 + cumsum(spc.lengths[-n.groups]))
group.ends <- cumsum(spc.lengths)

# determine depths of first / last profile in each group
yy <- unlist(sapply(spc.list, function(i) profileApply(i, max)))
tick.heights <- yy[c(group.starts, group.ends)] + 2

# plot 2 SoilProfileCollection objects on the same axis
par(mar=c(1,1,1,1))
plotSPC(sp3, n = n.pedons)
plotSPC(sp4, add = TRUE, x.idx.offset = group.ends[1], 
depth.axis = FALSE, id.style = 'side')

# annotate groups
profileGroupLabels(x0 = group.starts, x1 = group.ends,
labels=c('Collection 1', 'Collection 2'), y0=120, y1=tick.heights)

}
\seealso{
\code{\link{plotMultipleSPC}}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/launderMunsell.R
\name{launderMunsell}
\alias{launderMunsell}
\title{Fix Common Errors in Munsell Notation}
\usage{
launderMunsell(m, verbose = FALSE, ...)
}
\arguments{
\item{m}{character vector of Munsell colors}

\item{verbose}{logical, optionally return a \code{data.frame} comparing modifications}

\item{\dots}{additional arguments to \code{\link[=formatMunsell]{formatMunsell()}}}
}
\value{
either character vector, or when \code{verbose = TRUE} a \code{data.frame}
}
\description{
This function will "launder" colors in Munsell notation, fixing (and optionally flagging) the following:
\itemize{
\item colors specified with 0-chroma will be converted to neutral hue (N) and original value
\item for neutral colors, any chroma >0 will be set to 0
\item non-standard hues will be converted to NA (unless \code{standardHues = FALSE})
}

Examples:
\itemize{
\item '10YR 2/0' -> 'N 2/0'
\item 'N 4/1 -> 'N 4/0'
\item '5Z 3/3' -> NA
\item '5YR 3/' -> NA (hues other than N must have a valid chroma)
}

See \code{\link[=formatMunsell]{formatMunsell()}} for additional details.
}
\examples{

# => 'N 2/0'
launderMunsell('10YR 2/0')

# => 'N 4/0'
launderMunsell('N 4/1')

# alternative neutral convention => 'N 4/'
launderMunsell('N 4/0', neutralConvention = 'empty')

# => 'N 4/0'
launderMunsell('N 4/NA')
launderMunsell('N 4/')

# => 'N 4/0'
launderMunsell('5GY 4/0')

# not a standard hue => NA
launderMunsell('4ZR 4/6')

# missing chroma, not N => NA
launderMunsell('2.5Y 4/')

# invalid chroma => NA
launderMunsell('2.5Y 4/A')

}
\seealso{
\code{\link[=formatMunsell]{formatMunsell()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/soilColorSignature.R
\name{soilColorSignature}
\alias{soilColorSignature}
\title{Soil Profile Color Signatures}
\usage{
soilColorSignature(
  spc,
  color,
  space = c("sRGB", "CIELAB"),
  method = c("colorBucket", "depthSlices", "pam"),
  perceptualDistMat = FALSE,
  pam.k = 3,
  prob = c(0.1, 0.5, 0.9),
  useProportions = TRUE,
  pigmentNames = c(".white.pigment", ".red.pigment", ".green.pigment", ".yellow.pigment",
    ".blue.pigment"),
  apply.fun = lapply,
  r = NULL,
  g = NULL,
  b = NULL,
  RescaleLightnessBy = NULL
)
}
\arguments{
\item{spc}{a \code{SoilProfileCollection} object}

\item{color}{horizon-level attributes, either character of length 1 specifying a column containing Munsell or sRGB in hex notation, or character vector of three column names containing either sRGB or CIELAB color coordinates. sRGB color coordinates should be within the range of 0 to 1.}

\item{space}{character, either 'sRGB' or 'LAB', specifying color space}

\item{method}{algorithm used to compute color signature, \code{colorBucket}, \code{depthSlices}, or \code{pam}}

\item{perceptualDistMat}{logical, optionally return a distance matrix based on perceptual color distances, when ``method` is one of 'depthSlices' or 'pam', see Details}

\item{pam.k}{number of color classes for \code{method = 'pam'}}

\item{prob}{numeric vector, requested percentiles for \code{method = 'depthSlices'}}

\item{useProportions}{use proportions or quantities, see details}

\item{pigmentNames}{names for resulting pigment proportions or quantities}

\item{apply.fun}{function passed to \code{aqp::profileApply(APPLY.FUN)} argument, can be used to add progress bars via \code{pbapply::pblapply()}, or parallel processing with \code{furrr::future_map()}}

\item{r}{deprecated, use \code{color} argument}

\item{g}{deprecated, use \code{color} argument}

\item{b}{deprecated, use \code{color} argument}

\item{RescaleLightnessBy}{deprecated, scaling factor for CIELAB L-coordinate}
}
\value{
For the \code{colorBucket} method, a \code{data.frame}:
\itemize{
\item id column: set according to \code{idname(spc)}
\item \code{.white.pigment}: proportion or quantity of CIELAB L-values
\item \code{.red.pigment}: proportion or quantity of CIELAB positive A-values
\item \code{.green.pigment}: proportion or quantity of CIELAB negative A-values
\item \code{.yellow.pigment}: proportion or quantity of CIELAB positive B-values
\item \code{.blue.pigment}: proportion or quantity of CIELAB negative B-values
}

Column names can be adjusted with the \code{pigmentNames} argument.

For the \code{depthSlices} method, a \code{data.frame}:
\itemize{
\item id column: set according to \code{idname(spc)}
\item \code{L.1}, \code{A.1}, \code{B.1}: CIELAB color coordinates associated with the first depth slice, at depth percentile given in \code{prob[1]}
\item ...
\item \code{L.n}, \code{A.n}, \code{B.n}: CIELAB color coordinates associated with the \code{n} depth slice, at depth percentile given in \code{prob[n]}
}

For the \code{pam} method, a \code{data.frame}:
\itemize{
\item id column: set according to \code{idname(spc)}
\item \code{L.1}, \code{A.1}, \code{B.1}: CIELAB color coordinates associated with the first color cluster, after sorting all clusters in ascending order along L, A, B axes.
\item ...
\item \code{L.n}, \code{A.n}, \code{B.n}: CIELAB color coordinates associated with the \code{nth} color cluster, after sorting all clusters in ascending order along L, A, B axes.
}

When \code{perceptualDistMat = TRUE} and \code{method} is one of 'depthSlices' or 'pam', a distance matrix is returned.
}
\description{
Generate a color signature for each soil profile in a collection.
}
\details{
Missing colors (NA) are ignored in the creation of soil color signatures.

Interpreation of color signature.

Choices related to weighting, scaling, and distance metric.

Perceptual distances (dE00), summed over color groups.

See the \href{http://ncss-tech.github.io/AQP/aqp/soil-color-signatures.html}{related tutorial}.
}
\examples{

# trivial example, not very interesting
data(sp1)
depths(sp1) <- id ~ top + bottom

# Munsell notation
sp1$m <- sprintf("\%s \%s/\%s", sp1$hue, sp1$value, sp1$chroma)

# extract color signature
pig <- soilColorSignature(sp1, color = 'm')

}
\references{
https://en.wikipedia.org/wiki/Lab_color_space
}
\seealso{
\code{\link[=plotProfileDendrogram]{plotProfileDendrogram()}}
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-metadata.R
\docType{methods}
\name{hzdesgnname}
\alias{hzdesgnname}
\alias{hzdesgnname,SoilProfileCollection-method}
\alias{hzdesgnname<-}
\alias{hzdesgnname<-,SoilProfileCollection-method}
\title{Get or Set Horizon Designation Column Name}
\usage{
\S4method{hzdesgnname}{SoilProfileCollection}(object, required = FALSE)

\S4method{hzdesgnname}{SoilProfileCollection}(object, required = FALSE) <- value
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{required}{logical, is this attribute required? If it is, set to \code{TRUE} to trigger error on invalid \code{value}.}

\item{value}{character, name of column containing horizon designations}
}
\description{
\code{hzdesgnname()}: Get column name containing horizon designations

\verb{hzdesgnname<-}: Set horizon designation column name
}
\details{
Store the column name containing horizon designations or other identifiers in the metadata slot of the SoilProfileCollection.
}
\examples{

data(sp1)

# promote to SPC
depths(sp1) <- id ~ top + bottom

# set horizon designation column
hzdesgnname(sp1) <- "name"

# get horizon designation column
hzdesgnname(sp1)
}
\seealso{
\code{\link[=hzDesgn]{hzDesgn()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guessColumnNames.R
\name{guessHzAttrName}
\alias{guessHzAttrName}
\alias{guessHzDesgnName}
\alias{guessHzTexClName}
\title{Guess Horizon Slot Column Names}
\usage{
guessHzAttrName(x, attr, optional = NULL, verbose = TRUE, required = FALSE)

guessHzDesgnName(x, required = FALSE)

guessHzTexClName(x, required = FALSE)
}
\arguments{
\item{x}{A SoilProfileCollection}

\item{attr}{\emph{character}. A regular expression containing required formative element of attribute name.}

\item{optional}{\emph{character}. Vector of regular expression(s) containing optional formative elements of attribute name.}

\item{verbose}{\emph{logical}. Produce message output about guesses? Default: \code{TRUE}}

\item{required}{logical Default: \code{FALSE}. Is this attribute required? If it is, set to \code{TRUE} to trigger error on invalid value.}
}
\value{
Character containing horizon attribute column name. Result is the first match in \code{horizonNames(x)} with the most required plus optional patterns matched.
}
\description{
\code{guessHzAttrName()}: Guess the horizon column name where possible/preferred formative elements are known. There is a preference for records where more optional requirements are met to handle cases where there will be many matches. For example, working with soil data one might have "low, RV and high" total clay, as well as clay fractions. One could distinguish between these different measurements using standard formative elements for column names from the database of interest. Result is the first match in \code{horizonNames(x)} with the most required plus optional patterns matched.

e.g. \code{guessHzAttrName(x, attr="clay", optional=c("total", "_r"))} matches (\code{claytotal_r == totalclay_r}) over (\code{clay_r == claytotal == totalclay}) over \code{clay}.

\code{guessHzDesgnName()}: \strong{DEPRECATED} This follows the historic convention used by \code{aqp::plotSPC()} looking for "hzname" or other column names containing the regular expression "name". If the pattern "name" is not found, the pattern "desgn" is searched as a fallback, as "hzdesgn" or "hz_desgn" are other common column naming schemes for horizon designation name.

\code{guessHzTexClName()}: \strong{DEPRECATED} This function is used to provide a texture class attribute column name to functions. It will use regular expressions to match "texcl" which is typically the texture of the fine earth fraction, without modifiers or in-lieu textures. Alternately, it will match "texture" for cases where "texcl" is absent (e.g. in NASIS Component Horizon).
}
\examples{

# a has the required attr pattern, but none of the optional
a <- data.frame(id = 1, top = c(0,10), bottom=c(10,40),
                clay=c(18,19))
depths(a) <- id ~ top + bottom

guessHzAttrName(a, attr="clay", optional=c("total", "_r"))

# b has requried attr pattern, and one of the opional patterns
#   notice that it also contains "clay" but preferentially matches more optional patterns
b <- data.frame(id = 1, top = c(0,10), bottom=c(10,40),
                clay=c(0.18,0.19), clay_r=c(18,19))
depths(b) <- id ~ top + bottom

guessHzAttrName(b, attr="clay", optional=c("total", "_r"))

# c has total and _r (both optional) on either side of clay
# having all of the optional patterns plus required is best evidence, and first
# column containing that combination will be returned
c <- data.frame(id = 1, top = c(0,10), bottom=c(10,40),
                totalclay_r=c(18,19), claytotal_r=c(0.18,0.19))
depths(c) <- id ~ top + bottom

guessHzAttrName(c, attr="clay", optional=c("total", "_r"))

}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{hz_dissolve}
\alias{hz_dissolve}
\alias{dissolve_hz}
\title{Dissolving horizon boundaries by grouping variables}
\usage{
hz_dissolve(
  object,
  by,
  idcol = "id",
  depthcols = c("top", "bottom"),
  collapse = FALSE,
  order = FALSE
)

dissolve_hz(
  object,
  by,
  id = "idcol",
  hztop = "top",
  hzbot = "bottom",
  collapse = FALSE,
  order = FALSE
)
}
\arguments{
\item{object}{a \code{data.frame}}

\item{by}{character: column names, to be used as grouping variables, within the object.}

\item{idcol}{character: column name of the pedon ID within the object.}

\item{depthcols}{a character vector of length 2 specifying the names of the horizon depths (e.g. \code{c("top", "bottom")}).}

\item{collapse}{logical: indicating whether to not combine grouping variables before dissolving.}

\item{order}{logical: indicating whether or not to order the object by the id, hztop, and hzbot columns.}

\item{id}{deprecated and replaced with idcol.}

\item{hztop}{deprecated and replaced by depthcols.}

\item{hzbot}{deprecated and replaced by depthcols.}
}
\value{
A \code{data.frame} with the original idcol, by grouping variables, and non-consecutive horizon depths.
}
\description{
This function dissolves or combines horizons that have a common set of grouping variables. It only combines those horizon records that are sequential (e.g. share a horizon boundary). Thus, it can be used to identify discontinuities in the grouping variables along a profile and their unique depths. It is particularly useful for determining the depth to the top or bottom of horizons with a specific category, and should be simpler than previous methods that require aggregating over profiles.
}
\details{
This function assumes the profiles and horizons within the object follow the logic defined by \code{checkHzDepthLogic} (e.g. records are ordered sequentially by id, hztop, and hzbot and without gaps). If the records are not ordered, set the \code{order = TRUE}.
}
\examples{

# example 1
data(jacobs2000)
spc <- jacobs2000

spc$dep_5 <- spc$depletion_pct >=5
spc$genhz <- generalize.hz(spc$name, c("A", "E", "B", "C"), c("A", "E", "B", "C")) 
h <- horizons(spc)

test <- hz_dissolve(h, by = c("genhz", "dep_5"), idcol = "id", depthcols = c("top", "bottom"))

vars <- c("id", "top", "bottom", "genhz", "dep_5")
h[h$id == "92-1", vars]
test[test$id == "92-1", ]


# example 2
df <- data.frame(
    id = 1,
    top    = c(0, 5,  10, 15, 25, 50), 
    bottom = c(5, 10, 15, 25, 50, 100),
    hzname = c("A1",  "A2",  "E/A", "2Bt1", "2Bt2", "2C"),
    genhz  = c("A",   "A",   "E",   "2Bt",  "2Bt", "2C"),
    texcl  = c("sil", "sil", "sil", "sl",   "sl",   "s")
    )

df

hz_dissolve(df, c("genhz", "texcl"))
hz_dissolve(df, c("genhz", "texcl"), collapse = TRUE)

test <- hz_dissolve(df, "genhz")
subset(test, value == "2Bt")

}
\seealso{
\code{\link[=hz_lag]{hz_lag()}}, \code{\link[=hz_intersect]{hz_intersect()}}, \code{\link[=hz_segment]{hz_segment()}} , \code{\link[=checkHzDepthLogic]{checkHzDepthLogic()}}
}
\author{
Stephen Roecker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{sp4}
\alias{sp4}
\title{Soil Chemical Data from Serpentinitic Soils of California}
\format{
A data frame with 30 observations on the following 13 variables.
\describe{ \item{id}{site name} \item{name}{horizon
designation} \item{top}{horizon top boundary in cm}
\item{bottom}{horizon bottom boundary in cm}
\item{K}{exchangeable K in c mol/kg} \item{Mg}{exchangeable
Mg in cmol/kg} \item{Ca}{exchangeable Ca in cmol/kg}
\item{CEC_7}{cation exchange capacity (NH4OAc at pH 7)}
\item{ex_Ca_to_Mg}{extractable Ca:Mg ratio} \item{sand}{sand
content by weight percentage} \item{silt}{silt content by weight
percentage} \item{clay}{clay content by weight percentage}
\item{CF}{>2mm fraction by volume percentage} }
}
\source{
https://www.soils.org/publications/sssaj/articles/73/6/2087
}
\description{
Soil Chemical Data from Serpentinitic Soils of California
}
\details{
Selected soil physical and chemical data from (McGahan et al., 2009).
}
\examples{

# load sample data set, a simple data.frame object with horizon-level data from 10 profiles
library(aqp)
data(sp4)
str(sp4)
sp4$idbak <- sp4$id

# upgrade to SoilProfileCollection
# 'id' is the name of the column containing the profile ID
# 'top' is the name of the column containing horizon upper boundaries
# 'bottom' is the name of the column containing horizon lower boundaries
depths(sp4) <- id ~ top + bottom

# check it out
class(sp4) # class name
str(sp4) # internal structure

# check integrity of site:horizon linkage
spc_in_sync(sp4)

# check horizon depth logic
checkHzDepthLogic(sp4)

# inspect object properties
idname(sp4) # self-explanitory
horizonDepths(sp4) # self-explanitory

# you can change these:
depth_units(sp4) # defaults to 'cm'
metadata(sp4) # not much to start with

# alter the depth unit metadata
depth_units(sp4) <- 'inches' # units are really 'cm'

# more generic interface for adjusting metadata

# add attributes to metadata list
metadata(sp4)$describer <- 'DGM'
metadata(sp4)$date <- as.Date('2009-01-01')
metadata(sp4)$citation <- 'McGahan, D.G., Southard, R.J, Claassen, V.P.
2009. Plant-Available Calcium Varies Widely in Soils
on Serpentinite Landscapes. Soil Sci. Soc. Am. J. 73: 2087-2095.'

depth_units(sp4) <- 'cm' # fix depth units, back to 'cm'

# further inspection with common function overloads
length(sp4) # number of profiles in the collection
nrow(sp4) # number of horizons in the collection
names(sp4) # column names
min(sp4) # shallowest profile depth in collection
max(sp4) # deepest profile depth in collection

# extraction of soil profile components
profile_id(sp4) # vector of profile IDs
horizons(sp4) # horizon data

# extraction of specific horizon attributes
sp4$clay # vector of clay content

# subsetting SoilProfileCollection objects
sp4[1, ] # first profile in the collection
sp4[, 1] # first horizon from each profile

# basic plot method, highly customizable: see manual page ?plotSPC
plot(sp4)
# inspect plotting area, very simple to overlay graphical elements
abline(v=1:length(sp4), lty=3, col='blue')
# profiles are centered at integers, from 1 to length(obj)
axis(1, line=-1.5, at=1:10, cex.axis=0.75, font=4, col='blue', lwd=2)
# y-axis is based on profile depths
axis(2, line=-1, at=pretty(1:max(sp4)), cex.axis=0.75, font=4, las=1, col='blue', lwd=2)


# symbolize soil properties via color
par(mar=c(0,0,4,0))
plot(sp4, color='clay')
plot(sp4, color='CF')

# apply a function to each profile, returning a single value per profile,
# in the same order as profile_id(sp4)
soil.depths <- profileApply(sp4, max) # recall that max() gives the depth of a soil profile

# check that the order is correct
all.equal(names(soil.depths), profile_id(sp4))

# a vector of values that is the same length as the number of profiles
# can be stored into site-level data
sp4$depth <- soil.depths
# check: looks good
max(sp4[1, ]) == sp4$depth[1]

# extract site-level data
site(sp4) # as a data.frame
sp4$depth # specific columns as a vector

# use site-level data to alter plotting order
new.order <- order(sp4$depth) # the result is an index of rank
par(mar=c(0,0,0,0))
plot(sp4, plot.order=new.order)

# deconstruct SoilProfileCollection into a data.frame, with horizon+site data
as(sp4, 'data.frame')

}
\references{
McGahan, D.G., Southard, R.J, Claassen, V.P. 2009.
Plant-Available Calcium Varies Widely in Soils on Serpentinite Landscapes.
Soil Sci. Soc. Am. J. 73: 2087-2095.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repairMissingHzDepths.R
\name{repairMissingHzDepths}
\alias{repairMissingHzDepths}
\title{Repair Problematic Lower Horizon Depths}
\usage{
repairMissingHzDepths(x, adj = 10, max.depth = 200)
}
\arguments{
\item{x}{\code{SoilProfileCollection}}

\item{adj}{vertical offset applied to "repair" missing bottom depths when top and bottom depth are equal or bottom depth is missing. (\code{NA} to use \code{max.depth})}

\item{max.depth}{If adj is \code{NA}, or the resulting offset sum exceeds \code{max.depth}, \code{max.depth} is used.}
}
\value{
\code{SoilProfileCollection} with a new (logical) horizon-level attribute \code{.repaired} marking affected horizons
}
\description{
Attempt a simple repair of horizon bottom depths in the presence of \code{NA}, or in cases where the horizon shares a common top and bottom depth. Both situations are common in pedon description where "contact" (Cd, Cr, R, etc.) was described without a lower depth.
}
\details{
This repair is applied to the deepest horizon within a profile as identified by \code{\link{getLastHorizonID}}, as well as to bottom depths of any horizon that has a horizon below it. Horizon bottom depths are adjusted by adding \code{adj} (if non-NA). If the resulting value exceeds \code{max.depth}, the \code{max.depth} value is returned (if not \code{NA}).
}
\examples{

h <- data.frame(
id = c(1, 1, 1, 2, 2, 2, 2, 3, 3),
top = c(0:2, 0:3, 0:1) * 10,
bottom = c(rep(NA_integer_, 7), c(10, 99))
)

# NA depths result in warnings
suppressWarnings({
  depths(h) <- id ~ top + bottom
})

# inspect data before repairs
plotSPC(h)

g <- repairMissingHzDepths(h)

# all depth logic now valid
all(checkHzDepthLogic(g)$valid)

# inspect
plotSPC(g)

# no adj, max.depth only
f <- repairMissingHzDepths(h, adj = NA, max.depth = 200)
all(checkHzDepthLogic(f)$valid)
plotSPC(f)

# max.depth defaults to max(x) if too small
f$bottom[c(3,7)] <- NA
d <- repairMissingHzDepths(f, adj = NA, max.depth = 20)
all(checkHzDepthLogic(d)$valid)
plotSPC(d)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimatePSCS.R
\name{estimatePSCS}
\alias{estimatePSCS}
\title{Estimate boundaries of the U.S Soil Taxonomy Particle Size Control Section}
\usage{
estimatePSCS(
  p,
  hzdesgn = hzdesgnname(p, required = TRUE),
  clay.attr = hzmetaname(p, "clay", required = TRUE),
  texcl.attr = hztexclname(p, required = TRUE),
  lieutex = hzmetaname(p, "lieutex"),
  tax_order_field = "tax_order",
  bottom.pattern = "Cr|R|Cd|m",
  simplify = TRUE,
  ...
)
}
\arguments{
\item{p}{A SoilProfileCollection}

\item{hzdesgn}{Name of the horizon attribute containing the horizon
designation. Default \code{hzdesgnname(p, required = TRUE)}}

\item{clay.attr}{Name of the horizon attribute containing clay contents.
Default \code{hzmetaname(p, "clay", required = TRUE)}}

\item{texcl.attr}{Name of the horizon attribute containing textural class
(used for finding sandy textures). Default \code{hztexclname(p, required = TRUE)}}

\item{lieutex}{Optional data element used in addition to the horizon
designation to identify kinds of organic soil material for soils with organic
surfaces. Default: \code{hzmetaname(p, "lieutex")}}

\item{tax_order_field}{Name of the site attribute containing taxonomic
order; for handling PSCS rules for Andisols in lieu of lab data. May be NA
or column missing altogether, in which case Andisol PSC possibility is
ignored.}

\item{bottom.pattern}{Regular expression pattern to match a root-restrictive
contact. Default matches Cr, R, Cd or m. This argument is passed to both
\code{minDepthOf()} and \code{getArgillicBounds()}.}

\item{simplify}{Return a length 2 vector with upper and lower boundary when
p has length 1? Default TRUE.}

\item{...}{additional arguments are passed to getArgillicBounds()}
}
\value{
A numeric vector (when \code{simplify=TRUE}) containing the top and bottom
depth of the particle
size control section. First value is top, second value is bottom.
If \code{p} contains more than one profile, the result is a data.frame with
profile ID plus PSCS top and bottom depths.
}
\description{
Estimates the upper and lower boundary of the particle size control section
for Mineral or Organic soilsby applying a programmatic version of the
particle size control section key from the Keys to Soil Taxonomy (13th edition).
See details for assumptions and required data elements.
}
\details{
Requires information to identify argillic horizons (clay contents, horizon
designations) with \code{getArgillicBounds()} as well as the presence of
plow layers and surface organic soil material. Any
\code{getArgillicBounds()} arguments may be passed to \code{estimatePSCS}.

Also, requires information on taxonomic order to handle Andisols.

In aqp 2.1, particle size control sections of organic soils Histosols and
Histels are supported. This requires setting the \code{"lieutex"} horizon metadata
column using \verb{hzmetaname<-()} Horizon designations containing \code{"f"} or \code{"W"}
are  recognized as permafrost and water layers, respectively, for application
of the organic soils key to control sections. In lieu textures \code{"SPM"} and
\code{"PEAT"} are used to identify low density organic materials used for surface
tier thickness. To avoid using the 160 cm surface tier, set the \code{"lieutex"}
column to any column that does not contain \code{"SPM"} or \code{"PEAT"} values.

WARNING: Soils in arenic or grossarenic subgroups, with fragipans, or with
strongly contrasting PSCs may not be classified correctly. The author would
welcome a dataset to develop this functionality for.
}
\examples{

data(sp1, package = 'aqp')
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# set required metadata
hzdesgnname(sp1) <- 'name'
hztexclname(sp1) <- 'texture'
hzmetaname(sp1, 'clay') <- 'prop'

x <- estimatePSCS(sp1)
x

# change horizon texture and set inlieu texture column to turn
# first profile into an organic soil
sp1$name[1:6] <- c("Oi1", "Oi2", "Oi3", "Oaf", "Cf1", "Cf2")
sp1$texture <- as.character(sp1$texture)
sp1$texture[1:6] <- c("PEAT", "PEAT", "PEAT", "MUCK", "GRVLS", "GRVLS")
sp1$bottom[6] <- 200
hzmetaname(sp1, 'lieutex') <- 'texture'

y <- estimatePSCS(sp1[1, ], simplify = FALSE)

# 74cm lower boundary is 25cm past the upper boundary of permafrost (49cm)
# but minimum depth is 100cm unless there is a root-limiting layer
y

}
\references{
Soil Survey Staff. 2014. Keys to Soil Taxonomy, 12th ed.
USDA-Natural Resources Conservation Service, Washington, DC.
}
\seealso{
\code{\link[=getArgillicBounds]{getArgillicBounds()}}, \code{\link[=getSurfaceHorizonDepth]{getSurfaceHorizonDepth()}}
}
\author{
Andrew Gene Brown
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{munsell}
\alias{munsell}
\title{Munsell to sRGB Lookup Table for Common Soil Colors}
\format{
A data.frame with 8825 rows.
\itemize{
\item \code{hue}: Munsell Hue, upper case
\item \code{value}: Munsell Value
\item \code{chroma}: Munsell Chroma
\item \code{r}: sRGB "red" value (0-1)
\item \code{g}: sRGB "green" value (0-1)
\item \code{b}: sRGB "blue" value (0-1)
\item \code{L}: CIELAB "L" coordinate
\item \code{A}: CIELAB "A" coordinate
\item \code{B}: CIELAB "B" coordinate
}
}
\source{
Color chip XYZ values: \url{https://www.rit.edu/science/munsell-color-science-lab-educational-resources#munsell-renotation-data}
}
\usage{
data(munsell)
}
\description{
A lookup table of interpolated Munsell color chips for common soil colors.
}
\details{
See \code{munsell2rgb} for conversion examples. Values
are referenced to the D65 standard illuminant.
}
\examples{

data(munsell)

}
\references{
\itemize{
\item Color conversion equations
\itemize{
\item http://www.brucelindbloom.com/index.html?ColorCalcHelp.html
}
\item Methods used to generate this table
\itemize{
\item http://dx.doi.org/10.1016/j.cageo.2012.10.020
}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aqp-package.R
\docType{package}
\name{aqp-package}
\alias{aqp-package}
\alias{aqp}
\alias{aqp.env}
\title{Algorithms for Quantitative Pedology}
\description{
The aqp (Algorithms for Quantitative Pedology) package for R was developed
to address some of the difficulties associated with processing soils
information, specifically related to visualization, aggregation, and
classification of soil profile data. This package is based on a mix of S3/S4
functions and classes, and most functions use basic dataframes as input,
where rows represent soil horizons and columns define properties of those
horizons. Common to most functions are the requirements that horizon
boundaries are defined as depth from 0, and that profiles are uniquely
defined by an id column. The aqp package defines an S4 class,
"SoilProfileCollection", for storage of profile-level metadata, as well as
summary, print, and plotting methods that have been customized for common
tasks related to soils data.
}
\details{
Demos: \code{demo(aqp)}

\href{http://ncss-tech.github.io/AQP/}{Project homepage}
}
\seealso{
\verb{depths<-()}, \code{SoilProfileCollection()}, \code{\link{sp1}}, \code{\link{sp2}}, \code{\link{sp3}}, \code{\link{sp4}}, \code{\link{sp5}}, \code{\link{sp6}}
}
\author{
Dylan E. Beaudette \href{mailto:debeaudette@ucdavis.edu}{debeaudette@ucdavis.edu}, Pierre Roudier, Andrew G. Brown
}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bootstrapSoilTexture.R
\name{bootstrapSoilTexture}
\alias{bootstrapSoilTexture}
\title{Bootstrap Soil Texture Data}
\usage{
bootstrapSoilTexture(ssc, method = c("dirichlet", "normal"), n = 100)
}
\arguments{
\item{ssc}{a \code{data.frame} object with 3 columns: 'SAND', 'SILT', 'CLAY' and at least three rows of data within the range of 0-100 (percent). NA are automatically removed, but care should be taken to ensure that the sand/silt/clay values add to 100 percent. Simulations are based on these examples.}

\item{method}{type of simulation: 'dirichlet' or 'normal'. See details.}

\item{n}{number of simulated compositions. See details.}
}
\value{
a \code{list} containing:
\itemize{
\item \code{samples} - \code{data.frame} of simulated sand, silt, clay values
\item \code{mean} - compositional mean
\item \code{var} - compositional variance-covariance matrix
\item \code{D.alpha} - (fitted) alpha parameters of the Dirichlet distribution, \code{NULL} when \code{method = 'normal'}
}
}
\description{
Simulate realistic sand/silt/clay values (a composition) using multivariate Normal distribution or Dirichlet distribution. Simulations from the multivariate Normal distribution are based on the compositional mean and variance-covariance matrix. Simulations from the Dirichlet distribution are based on maximum likelihood estimation of \code{alpha} parameters.
}
\details{
Simulations from the multivariate normal distribution will more closely track the marginal distributions of sand, silt, and clay--possibly a better fit for "squished" compositions (TODO elaborate). However, these simulations can result in extreme (unlikely) estimates.

Simulations from the Dirichlet distribution will usually be a better fit (fewer extreme estimates) but require a fairly large number of records in \code{ssc} (\code{n >= 30}?) for a reliable fit.

Additional examples will be added to \href{http://ncss-tech.github.io/AQP/aqp/soiltexture-vizualization-ideas.html}{this tutorial}.
}
\examples{

\donttest{
if(
requireNamespace("compositions") &
  requireNamespace("soiltexture")
) {
  
  # sample data, data.frame
  data('sp4')
  
  # filter just Bt horizon data
  ssc <- sp4[grep('^Bt', sp4$name), c('sand', 'silt', 'clay')]
  names(ssc) <- toupper(names(ssc))
  
  # simulate 100 samples
  s <- bootstrapSoilTexture(ssc, n = 100)
  s <- s$samples
  
  # empty soil texture triangle
  TT <- soiltexture::TT.plot(
    class.sys= "USDA-NCSS.TT",
    main= "",
    tri.sum.tst=FALSE,
    cex.lab=0.75,
    cex.axis=0.75,
    frame.bg.col='white',
    class.lab.col='black',
    lwd.axis=1.5,
    arrows.show=TRUE,
    new.mar = c(3, 0, 0, 0)
  )
  
  # add original data points
  soiltexture::TT.points(
    tri.data = s, geo = TT, col='firebrick', 
    pch = 3, cex = 0.5, lwd = 1, 
    tri.sum.tst = FALSE
  )
  
  # add simulated points
  soiltexture::TT.points(
    tri.data = ssc, geo = TT, bg='royalblue', 
    pch = 22, cex = 1, lwd = 1, 
    tri.sum.tst = FALSE
  )
  
  # simple legend
  legend('top', 
         legend = c('Source', 'Simulated'), 
         pch = c(22, 3), 
         col = c('black', 'firebrick'), 
         pt.bg = c('royalblue', NA), 
         horiz = TRUE, bty = 'n'
  )
  
  
}

}

}
\references{
Aitchison, J. (1986) The Statistical Analysis of Compositional Data Monographs on Statistics and Applied Probability. Chapman & Hall Ltd., London (UK). 416p.

Aitchison, J, C. Barcel'o-Vidal, J.J. Egozcue, V. Pawlowsky-Glahn (2002) A concise guide to the algebraic geometric structure of the simplex, the sample space for compositional data analysis, Terra Nostra, Schriften der Alfred Wegener-Stiftung, 03/2003

Malone Brendan, Searle Ross (2021) Updating the Australian digital soil texture mapping (Part 1*): re-calibration of field soil texture class centroids and description of a field soil texture conversion algorithm. Soil Research. https://www.publish.csiro.au/SR/SR20283

Malone Brendan, Searle Ross (2021) Updating the Australian digital soil texture mapping (Part 2*): spatial modelling of merged field and lab measurements. Soil Research. https://doi.org/10.1071/SR20284
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{hz_intersect}
\alias{hz_intersect}
\title{Intersecting horizon boundaries by horizon depths}
\usage{
hz_intersect(x, y, idcol = "id", depthcols = c("top", "bottom"))
}
\arguments{
\item{x}{a \code{data.frame}}

\item{y}{a \code{data.frame}}

\item{idcol}{character: column name of the pedon ID within the object.}

\item{depthcols}{a character vector of length 2 specifying the names of the horizon depths (e.g. \code{c("top", "bottom")}).}
}
\value{
A \code{data.frame} with harmonized depth intervals (i.e. segment_id) and columns from both of the original \code{data.frame}. If both \code{data.frame} contain the same column names, they will both be returned (with the exception of the idcol and depthcols), and appended with either x or y to indicate which \code{data.frame} they originated from.
}
\description{
This function intersects two horizon tables by harmonizing their depths and merging them where they overlap. This can be useful to rejoin the results of \code{hz_dissolve()} to it's original horizon table, and then perform an aggregation on the dissolved variables.
}
\details{
.
}
\examples{

h <- data.frame(
  id = 1,
  top    = c(0,  25, 44, 46, 50),
  bottom = c(25, 44, 46, 50, 100),
  by     = c("Yes", "Yes", "No", "No", "Yes"),
  clay   = c(10, 12, 27, 35, 16)
)

hz_dissolve(h, "by")

hz_intersect(x = hz_dissolve(h, "by"), y = h)

hi <- hz_intersect(x = h, y = hz_dissolve(h, "by"))
aggregate(clay ~ dissolve_id, data = hi, mean)

}
\seealso{
\code{\link[=hz_dissolve]{hz_dissolve()}}, \code{\link[=hz_lag]{hz_lag()}}, \code{\link[=hz_segment]{hz_segment()}}
}
\author{
Stephen Roecker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validateMunsell.R
\name{validateMunsell}
\alias{validateMunsell}
\title{Validate Standard Munsell Notation}
\usage{
validateMunsell(m, vRange = c(1, 12), cRange = c(0, 50))
}
\arguments{
\item{m}{character vector of colors in Munsell notation}

\item{vRange}{numeric vector of length 2, range of expected Munsell value}

\item{cRange}{numeric vector of length 2, range of expected Munsell chroma}
}
\value{
logical vector,
\itemize{
\item \code{TRUE}: valid Munsell notation
\item \code{FALSE}: invalid notation
}
}
\description{
This function validates Munsell color notation according to the following requirements:
\enumerate{
\item \emph{hue} must not be absent, and within the set of "standard hues", see \code{\link[=huePosition]{huePosition()}}
\item \emph{value} must not be absent, and within \code{vRange}
\item \emph{chroma} must be within \code{cRange}, and is only allowed to be absent in neutral colors (e.g. \verb{N 4/})
}
}
\examples{

# valid
validateMunsell('5Y 6/8')
validateMunsell('N 4/')
validateMunsell('2.5Y 4/0')

# invalid
validateMunsell('5G 0/4')
validateMunsell(NA)

# mixture
validateMunsell(c('5G 4/4', 'N 2/', NA, 'NA', '100R 3/3'))

}
\seealso{
\code{\link[=huePosition]{huePosition()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{class}
\name{SoilProfileCollection}
\alias{SoilProfileCollection}
\alias{SoilProfileCollection-class}
\title{An S4 object representation of a group of soil profiles.}
\usage{
SoilProfileCollection(
  idcol = "id",
  hzidcol = "hzID",
  depthcols = c("top", "bottom"),
  metadata = list(aqp_df_class = "data.frame", aqp_group_by = "", aqp_hzdesgn = "",
    aqp_hztexcl = "", stringsAsFactors = FALSE),
  horizons = data.frame(id = character(0), hzID = character(0), top = numeric(0), bottom
    = numeric(0), stringsAsFactors = FALSE),
  site = data.frame(id = character(0), stringsAsFactors = FALSE),
  diagnostic = data.frame(stringsAsFactors = FALSE),
  restrictions = data.frame(stringsAsFactors = FALSE)
)
}
\arguments{
\item{idcol}{character. Profile ID Column Name}

\item{hzidcol}{character. Horizon ID Column Name}

\item{depthcols}{character. length 2 Top and Bottom Depth Column Names}

\item{metadata}{list. metadata including data.frame class in use and depth units}

\item{horizons}{data.frame. An object inheriting from data.frame containing Horizon data.}

\item{site}{data.frame. An object inheriting from data.frame containing Site data.}

\item{diagnostic}{data.frame. An object inheriting from data.frame containing diagnostic feature data. Must contain profile ID. See \code{diagnostic_hz()}}

\item{restrictions}{data.frame. An object inheriting from data.frame containing restrictive feature data. Must contain profile ID. See \code{restrictions()}}
}
\description{
In general, one should use \code{depths()} to initiate a SoilProfileCollection object from data. However, sometimes there are instances where either an empty, or very specific, object is needed. If that is the case, the general constructor \code{SoilProfileCollection} is available.
}
\details{
After aqp 2.0.2, the \verb{@sp} slot was removed from the SoilProfileCollection object. If you run into errors related to old object definitions, use \code{rebuildSPC()} on the problematic object.
}
\section{Slots}{

\describe{
\item{\code{idcol}}{character.}

\item{\code{hzidcol}}{character.}

\item{\code{depthcols}}{character.}

\item{\code{metadata}}{list.}

\item{\code{horizons}}{data.frame.}

\item{\code{site}}{data.frame.}

\item{\code{diagnostic}}{data.frame.}

\item{\code{restrictions}}{data.frame.}
}}

\examples{

## structure of default, empty SoilProfileCollection
str(SoilProfileCollection())


## use the depths() formula interface to specify
## profile ID, top and bottom depth and set up
## a SPC that is topologically correct and complete

d <- do.call('rbind', lapply(1:10, random_profile))

# promote to SoilProfileCollection and plot
depths(d) <- id ~ top + bottom
plot(d)

# split into new SoilProfileCollection objects by index
d.1 <- d[1, ]
d.2 <- d[2, ]
d.345 <- d[3:5, ]

# combine profile collections
# note that profiles are sorted according to ID
d.new <- c(d.345, d.1, d.2)
plot(d.new)

data(sp1)

## init SoilProfileCollection objects from data.frame
depths(sp1) <- id ~ top + bottom

## depth units
du <- depth_units(sp1)
depth_units(sp1) <- 'in'
depth_units(sp1) <- du

## horizon designation column
hzdesgnname(sp1) <- "name"
hzdesgnname(sp1)

## all designations in an SPC (useful for single profile SPC)
hzDesgn(sp1)

## horizon texture class column
hztexclname(sp1) <- "texture"
hztexclname(sp1)

## get/set metadata on SoilProfileCollection objects
# this is a 1-row data.frame
m <- metadata(sp1)
m$sampler <- 'Dylan'
metadata(sp1) <- m

## extract horizon data from SoilProfileCollection objects as data.frame
h <- horizons(sp1)

# also merge (left-join) of new columns and
# replacement of existing columns via horizons<-
horizons(sp1) <- h

# get number of horizons
nrow(sp1)


## getting site-level data
site(sp1)

## setting site-level data
# site-level data from horizon-level data (stored in @horizons)
site(sp1) <- ~ group


# make some fake site data, and append from data.frame
# a matching ID column must be present in both @site and new data
# note that IDs should all be character class
d <- data.frame(id=profile_id(sp1), p=runif(n=length(sp1)), stringsAsFactors=FALSE)
site(sp1) <- d

# edit horizon depths
horizonDepths(sp1) <- c('t', 'b')
horizonDepths(sp1)

# edit profile IDs
p <- sprintf("\%s-new", profile_id(sp1))
profile_id(sp1) <- p
profile_id(sp1)

}
\author{
Pierre Roudier, Dylan E. Beaudette, Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_profile.R
\name{random_profile}
\alias{random_profile}
\alias{.lpp}
\title{Random Profile}
\usage{
random_profile(
  id,
  n = c(3, 4, 5, 6),
  min_thick = 5,
  max_thick = 30,
  n_prop = 5,
  exact = FALSE,
  method = "random_walk",
  HzDistinctSim = FALSE,
  SPC = FALSE,
  ...
)
}
\arguments{
\item{id}{a character or numeric id used for this profile}

\item{n}{vector of possible number of horizons, or the exact number of
horizons (see below)}

\item{min_thick}{minimum thickness criteria for a simulated horizon}

\item{max_thick}{maximum thickness criteria for a simulated horizon}

\item{n_prop}{number of simulated soil properties (columns in the returned
dataframe)}

\item{exact}{should the exact number of requested horizons be generated?
(defaults to FALSE)}

\item{method}{named method used to synthesize depth function ('random_walk'
or 'LPP'), see details}

\item{HzDistinctSim}{optionally simulate horizon boundary distinctness codes}

\item{SPC}{result is a \code{SoilProfileCollection} object, otherwise a \code{data.frame} object}

\item{\dots}{additional parameters passed-in to the LPP \code{.lpp})
function}
}
\value{
A \code{data.frame} or \code{SoilProfileCollection} object.
}
\description{
Generate a random soil profile according to set criteria, with correlated
depth trends.

The random walk method produces profiles with considerable variation between
horizons and is based on values from the normal distribution seeded with
means and standard deviations drawn from the uniform distribution of [0,
10].

The logistic power peak (LPP) function can be used to generate random soil
property depth functions that are sharply peaked. LPP parameters can be
hard-coded using the optional arguments: "lpp.a", "lpp.b", "lpp.u", "lpp.d",
"lpp.e". Amplitude of the peak is controlled by ("lpp.a + "lpp.b"), depth of
the peak by "lpp.u", and abruptness by "lpp.d" and "lpp.e". Further
description of the method is outlined in (Brenton et al, 2011). Simulated
horizon distinctness codes are based on the USDA-NCSS field description
methods.

Simulated distinctness codes are constrained according to horizon thickness,
i.e. a gradual boundary (+/- 5cm) will not be simulated for horizons that
are thinner than 3x this vertical distance

The \code{\link[=rp]{rp()}} function is a convenient wrapper to this function, when requesting a simulated \code{SoilProfileCollection} of specified size.
}
\note{
See examples for ideas on simulating several profiles at once.
}
\examples{


# generate 10 random profiles, result is a list of SoilProfileCollection objects
d <- lapply(1:10, random_profile, SPC=TRUE)

# combine
d <- combine(d)

# plot
opar <- par(mar=c(0,0,3,2))
plotSPC(d, color='p1', name='name', cex.names=0.75)
par(opar)

# simulate horizon boundary distinctness codes:
d <- lapply(1:10, random_profile, SPC=TRUE, HzDistinctSim=TRUE)
d <- combine(d)

d$HzD <- hzDistinctnessCodeToOffset(d$HzDistinctCode)

opar <- par(mar=c(0,0,3,2))
plotSPC(d, name='name', color='p1', hz.distinctness.offset='HzD')
par(opar)


# depth functions are generated using the LPP function
opar <- par(mfrow=c(2,1), mar=c(0,0,3,0))

# generate data
d.1 <- lapply(1:10, random_profile, SPC=TRUE, n=c(6, 7, 8), n_prop=1, method='LPP')
d.1 <- combine(d.1)

# plot
plotSPC(d.1, name='name', color='p1', col.label = 'LPP Defaults')


# do this again, this time set all of the LPP parameters
d.2 <- lapply(1:10, random_profile, SPC=TRUE, n=c(6, 7, 8), n_prop=1, method='LPP',
           lpp.a=5, lpp.b=10, lpp.d=5, lpp.e=5, lpp.u=25)
d.2 <- combine(d.2)

# plot
plotSPC(d.2, name='name', color='p1', col.label = 'Custom LPP Parameters')


# reset plotting defaults
par(opar)



# try plotting the LPP-derived simulated data
# aggregated over all profiles
a <- slab(d.2, fm= ~ p1)
a$mid <- with(a, (top + bottom) / 2)

library(lattice)
(p1 <- xyplot(mid ~ p.q50, data=a,
              lower=a$p.q25, upper=a$p.q75, ylim=c(150,-5), alpha=0.5,
              panel=panel.depth_function, prepanel=prepanel.depth_function,
              cf=a$contributing_fraction, xlab='Simulated Data', ylab='Depth',
              main='LPP(a=5, b=10, d=5, e=5, u=25)',
              par.settings=list(superpose.line=list(col='black', lwd=2))
))

# optionally add original data as step-functions
if(require(latticeExtra)) {
  h <- horizons(d.2)
  p1 + as.layer(xyplot(top ~ p1, groups=id, data=h,
                       horizontal=TRUE, type='S',
                       par.settings=list(superpose.line=list(col='blue', lwd=1, lty=2))))
}



}
\references{
Myers, D. B.; Kitchen, N. R.; Sudduth, K. A.; Miles, R. J.;
Sadler, E. J. & Grunwald, S. Peak functions for modeling high resolution
soil profile data Geoderma, 2011, 166, 74-83.
}
\seealso{
\code{\link[=rp]{rp()}}, \code{\link[=hzDistinctnessCodeToOffset]{hzDistinctnessCodeToOffset()}}
}
\author{
Dylan E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregateColorPlot.R
\name{aggregateColorPlot}
\alias{aggregateColorPlot}
\title{Plot aggregate soil color data}
\usage{
aggregateColorPlot(
  x,
  print.label = TRUE,
  label.font = 1,
  label.cex = 0.65,
  label.orientation = c("v", "h"),
  buffer.pct = 0.02,
  print.n.hz = FALSE,
  rect.border = "black",
  horizontal.borders = FALSE,
  horizontal.border.lwd = 2,
  x.axis = TRUE,
  y.axis = TRUE,
  ...
)
}
\arguments{
\item{x}{a \code{list}, results from \code{aggregateColor()}}

\item{print.label}{logical, print Munsell color labels inside of rectangles, only if they fit}

\item{label.font}{font specification for color labels}

\item{label.cex}{font size for color labels}

\item{label.orientation}{label orientation, \code{v} for vertical or \code{h} for horizontal}

\item{buffer.pct}{extra space between labels and color rectangles}

\item{print.n.hz}{optionally print the number of horizons below Munsell color labels}

\item{rect.border}{color for rectangle border}

\item{horizontal.borders}{optionally add horizontal borders between bands of color}

\item{horizontal.border.lwd}{line width for horizontal borders}

\item{x.axis}{logical, add a scale and label to x-axis?}

\item{y.axis}{logical, add group labels to y-axis?}

\item{...}{additional arguments passed to \code{plot}}
}
\value{
no data are returned, function is called for graphical output
}
\description{
Generate a plot from summaries generated by \code{aggregateColor()}.
}
\details{
See the \href{http://ncss-tech.github.io/AQP/sharpshootR/aggregate-soil-color.html}{related tutorial for additional examples}.
}
\examples{

# keep examples from using more than 2 cores
data.table::setDTthreads(Sys.getenv("OMP_THREAD_LIMIT", unset = 2))

# load some example data
data(sp1, package = 'aqp')

# upgrade to SoilProfileCollection and convert Munsell colors
sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# generalize horizon names
n <- c('O', 'A', 'B', 'C')
p <- c('O', 'A', 'B', 'C')
sp1$genhz <- generalize.hz(sp1$name, n, p)

# aggregate colors over horizon-level attribute: 'genhz'
a <- aggregateColor(sp1, groups = 'genhz', col = 'soil_color')

# check results
str(a)

# simple visualization
aggregateColorPlot(a)

}
\author{
D.E. Beaudette
}
\keyword{hplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine.R
\name{pbindlist}
\alias{pbindlist}
\title{Combine a list of SoilProfileCollection objects}
\usage{
pbindlist(l, new.idname = NULL, verbose = TRUE)
}
\arguments{
\item{l}{a list of SoilProfileCollection objects}

\item{new.idname}{Optional: a character referring to a new column name to put unique profile IDs in; default: \code{NULL} to attempt with existing idname in first element}

\item{verbose}{Produce warnings and messages regarding results? default: \code{TRUE}}
}
\value{
a SoilProfileCollection object
}
\description{
See \code{combine(...)} for a connotative short-hand method that does not require that \code{SoilProfileCollection} be in a list. Profiles will be sorted based on character sorting of profile ID.
}
\details{
Input data must share a common depth unit, and if spatial data are present, a common CRS and coordinate names. In the case of non-conformal @idname and/or @depthcols, the first SoilProfileCollection is used as a template. If one or more subsequent list elements has non-unique values in a site level attribute of that name, the ID name from the second list element is attempted, and so on. Non-conforming spatial data are dropped from the final result (returns default empty \code{SpatialPoints}).
}
\examples{
# example data
data(sp2, package = 'aqp')
depths(sp2) <- id ~ top + bottom
site(sp2) <- ~ surface

# copy pieces
x <- sp2[1:5, ]
y <- sp2[6:10, ]

# reset IDs and combine
profile_id(y) <- sprintf("\%s-copy", profile_id(y))

# this should work
z <- pbindlist(list(x, y))

# check
plot(z)
}
\author{
D.E. Beaudette and A.G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy_uncertainty.R
\name{confusionIndex}
\alias{confusionIndex}
\title{Confusion Index}
\usage{
confusionIndex(x)
}
\arguments{
\item{x}{vector of probabilities (0,1), should not contain NA}
}
\value{
A single numeric value.
}
\description{
Calculate the confusion index of Burrough et al., 1997.
}
\examples{

# a very simple example
p <- c(0.25, 0.25, 0.4, 0.05, 0.05)
confusionIndex(p)

# for comparison
shannonEntropy(p)

}
\references{
Burrough, P.A., P.F.M. van Gaans, and R. Hootsmans. 1997. "Continuous Classification in Soil Survey: Spatial Correlation, Confusion and Boundaries." Geoderma 77: 115-35. doi:10.1016/S0016-7061(97)00018-9.
}
\author{
D.E. Beaudette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/groupedProfilePlot.R
\name{groupedProfilePlot}
\alias{groupedProfilePlot}
\title{Grouped Soil Profile Plot}
\usage{
groupedProfilePlot(
  x,
  groups,
  group.name.offset = -5,
  group.name.cex = 0.75,
  group.line.col = "RoyalBlue",
  group.line.lwd = 2,
  group.line.lty = 2,
  break.style = c("line", "arrow", "both"),
  break.offset = 0.5,
  arrow.offset = group.name.offset + 5,
  arrow.length = 0.1,
  ...
)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}

\item{groups}{the name of a site-level attribute that defines groups, factor
levels will influence plotting order}

\item{group.name.offset}{vertical offset for group names, single numeric
value or vector of offsets}

\item{group.name.cex}{font size for group names}

\item{group.line.col}{color for line that splits groups}

\item{group.line.lwd}{width of line that splits groups}

\item{group.line.lty}{style of line that splits groups}

\item{break.style}{style of group boundaries: "line", "arrow", "both"}

\item{break.offset}{horizontal offset used to place vertical breaks and/or arrows, shifted slightly to the right of default when \code{hz.depths=TRUE} is passed to \code{plotSPC()}}

\item{arrow.offset}{vertical offset for "arrow" style boundaries, single
numeric value or vector of offsets}

\item{arrow.length}{value passed to \code{arrows} to define arrow head size}

\item{\dots}{further arguments to \code{plotSPC}}
}
\description{
Plot a collection of soil profiles, sorted by group.

The left-right ordering of groups can be adjusted by converting
\code{groups} into a factor and explicitly setting factor levels.
Alpha-numeric ordering is used for all other types.
}
\examples{

# sample data
data(sp1)
# convert colors from Munsell to hex-encoded RGB
sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))

# promote to SoilProfileCollection
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# add a groups
sp1$group.2 <- sprintf("\%s-\%s", rev(LETTERS[1:3]), sp1$group)

# convert fake groupt to factor with new levels
sp1$group.3 <- factor(sp1$group.2, levels=c('C-2', 'B-2', 'A-2', 'C-1', 'B-1', 'A-1'))

# plot profiles, sorted and annotated by 'group' (integers)
par(mar=c(1,1,1,1))
groupedProfilePlot(sp1, groups='group', max.depth=150, group.name.offset = -5, id.style='side')

# plot profiles, sorted and annotated by 'group.2' (characters)
par(mar=c(1,1,1,1))
groupedProfilePlot(sp1, groups='group.2', max.depth=150, group.name.offset = -5, id.style='side')

# plot profiles, sorted and annotated by 'group.3' (characters)
par(mar=c(1,1,1,1))
groupedProfilePlot(sp1, groups='group.3', max.depth=150, group.name.offset = -5, id.style='side')


# make fake site-level attribute and adjust levels
sp1$new.group <- sample(letters[1:3], size=length(sp1), replace=TRUE)

# tabulate pedons / group
tab <- table(sp1$new.group)

# sort large -> small
tab <- sort(tab, decreasing = TRUE)

# set levels based on sorted tabulation
# assign custom labels
sp1$new.group <- factor(sp1$new.group, levels=names(tab),
labels=paste0(names(tab), ' (', tab, ')'))

groupedProfilePlot(sp1, groups='new.group', max.depth=150,
group.name.offset = -10, id.style='side')

# offsets can be set using a vector of values, recycled as needed
groupedProfilePlot(sp1, groups='new.group', max.depth=150,
group.name.offset=c(-10, -5), id.style='side')

# annotate with arrows instead of vertical lines
groupedProfilePlot(sp1, groups='new.group', max.depth=150,
group.name.offset=c(-10, -12), break.style='arrow', arrow.offset=-3,
group.line.lty = 1, group.line.lwd = 1, id.style='side')


\dontrun{
# more complete example using data from soilDB package
data(loafercreek, package='soilDB')
par(mar=c(1,1,1,1))
# lines
groupedProfilePlot(loafercreek, groups='hillslopeprof', group.name.cex = 0.5,
group.name.offset = -10)

# arrows
groupedProfilePlot(loafercreek, groups='hillslopeprof', group.name.cex = 0.5,
group.name.offset = -10, break.style ='arrow', group.line.lty = 1,
group.line.lwd = 1)

# both
groupedProfilePlot(loafercreek, groups='hillslopeprof', group.name.cex = 0.5,
group.name.offset = -10, break.style ='both', group.line.lty = 1,
group.line.lwd = 1)
}

}
\seealso{
\code{\link{plotSPC}}
}
\author{
D.E. Beaudette
}
\keyword{hplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\docType{methods}
\name{names,SoilProfileCollection-method}
\alias{names,SoilProfileCollection-method}
\alias{names}
\alias{.DollarNames.SoilProfileCollection}
\title{Get names of columns in site and horizons table}
\usage{
\S4method{names}{SoilProfileCollection}(x)

\method{.DollarNames}{SoilProfileCollection}(x, pattern)
}
\arguments{
\item{x}{a SoilProfileCollection}

\item{pattern}{A regular expression. Only matching names are returned.}
}
\description{
Get names of columns in site and horizons table of a SoilProfileCollection.
}
\details{
The \code{.DollarNames} SoilProfileCollection method is implemented to support
auto-completion after the \code{$} operator. There is a custom help handler
implemented for interactive use in RStudio to integrate help text and data
viewer functionality with auto-completion. This custom help handler can be
disabled by setting \code{options(.aqp.rs.rpc.integration=FALSE)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\docType{methods}
\name{validSpatialData,SoilProfileCollection-method}
\alias{validSpatialData,SoilProfileCollection-method}
\alias{validSpatialData}
\title{Check for valid spatial reference of profiles}
\usage{
\S4method{validSpatialData}{SoilProfileCollection}(object)
}
\arguments{
\item{object}{a SoilProfileCollection}
}
\value{
logical \code{TRUE} if column names are defined and correspond to existing data
}
\description{
Are coordinate column names defined in metadata and existing in the SoilProfileCollection?
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{hz_segment}
\alias{hz_segment}
\alias{segment}
\title{Segmenting of Soil Horizon Data by Depth Interval}
\usage{
hz_segment(object, intervals, trim = TRUE, depthcols = c("top", "bottom"))

segment(object, intervals, trim = TRUE, hzdepcols = c("top", "bottom"))
}
\arguments{
\item{object}{either a \code{SoilProfileCollection} or \code{data.frame}}

\item{intervals}{a vector of integers over which to slice the horizon data (e.g. \code{c(25, 100)} or \code{25:100})}

\item{trim}{logical, when \code{TRUE} horizons in \code{object} are truncated to the min/max specified in \code{intervals}. When \code{FALSE}, those horizons overlapping an interval are marked as such. Care should be taken when specifying more than one depth interval and \code{trim = FALSE}.}

\item{depthcols}{a character vector of length 2 specifying the names of the horizon depths (e.g. \code{c("top", "bottom")}), only necessary if \code{object} is a}

\item{hzdepcols}{deprecated being replaced by depthcols.}
}
\value{
Either a \code{SoilProfileCollection} or \code{data.frame} with the original horizon data segmented by depth intervals. There are usually more records in the resulting object, one for each time a segment interval partially overlaps with a horizon. A new column called \code{segment_id} identifying the depth interval is added.
}
\description{
This function segments or subdivides horizon data from a \code{SoilProfileCollection} or \code{data.frame} by depth interval (e.g. \code{c(0, 10)}, \code{c(0, 50)}, or \code{25:100}). This results in horizon records being split at the specified depth intervals, which duplicates the original horizon data but also adds new horizon depths. In addition, labels (i.e. \code{"segment_id"}) are added to each horizon record that correspond with their depth interval (e.g. \code{025-100}). This function is intended to harmonize horizons to a common support (i.e. depth interval) for further aggregation or summary. See the examples.
}
\details{
\code{hz_segment()} performs no aggregation or resampling of the source data, rather, labels are added to horizon records for subsequent aggregation or summary. This makes it possible to process a very large number of records outside of the constraints associated with e.g. \code{slice()} or \code{slab()}.
}
\examples{

# example data
data(sp1)

# upgrade to SPC
depths(sp1) <- id ~ top + bottom

# segment and trim
z <- hz_segment(sp1, intervals = c(0, 10, 20, 30), trim = TRUE)

# display segment labels
# note that there are new horizon boundaries at segments
par(mar = c(0, 0, 3, 1))
plotSPC(z, color = 'segment_id', width = 0.3)

# highlight new horizon records
par(mar = c(0, 0, 2, 1))
plotSPC(z, color = NA, default.color = NA, width = 0.3, lwd = 1)
plotSPC(sp1, color = NA, default.color = NA, 
width = 0.3, lwd = 3, add = TRUE, name = NA, print.id = FALSE)
legend('top', horiz = TRUE, 
legend = c('original', 'segmented'), 
lwd = c(1, 3), cex = 0.85, bty = 'n')

\donttest{
# same results as slab()
# 10 random profiles
s <- lapply(1:10, random_profile, n_prop = 1, SPC = TRUE, method = 'random_walk')
s <- combine(s)

a.slab <- slab(s, fm = ~ p1, slab.structure = c(0, 10, 20, 30), slab.fun = mean, na.rm = TRUE)

z <- hz_segment(s, intervals = c(0, 10, 20, 30), trim = TRUE)
z <- horizons(z)
z$thick <- z$bottom - z$top

a.segment <- sapply(split(z, z$segment_id), function(i) {
  weighted.mean(i$p1, i$thick)
})


res <- data.frame(
  slab = a.slab$value,
  segment = a.segment,
  diff = a.slab$value - a.segment
)

print(res)
res$diff < 0.001
}


data(sp5)

# segment by upper 25-cm
test1 <- hz_segment(sp5, intervals = c(0, 100))
print(test1)
nrow(test1)
print(object.size(test1), units = "Mb")

# segment by 1-cm increments
test2 <- hz_segment(sp5, intervals = 0:100)
print(test2)
nrow(test2)
print(object.size(test2), units = "Mb")


# segment and aggregate
test3 <- hz_segment(horizons(sp5), 
                 intervals = c(0, 5, 15, 30, 60, 100, 200), 
                 depthcols = c("top", "bottom")
)
test3$hzthk <- test3$bottom - test3$top
test3_agg <- by(test3, test3$segment_id, function(x) {
  data.frame(
    hzID = x$hzID[1],
    segment_id = x$segment_id[1],
    average = weighted.mean(x$clay, w = x$hzthk)
  )
})
test3_agg <- do.call("rbind", test3_agg)

head(test3_agg)

}
\seealso{
\code{\link[=dice]{dice()}}, \code{\link[=glom]{glom()}}, \code{\link[=hz_dissolve]{hz_dissolve()}}, \code{\link[=hz_lag]{hz_lag()}}, \code{\link[=hz_intersect]{hz_intersect()}}
}
\author{
Stephen Roecker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{sp3}
\alias{sp3}
\title{Soil Profile Data Example 3}
\format{
A data frame with 46 observations on the following 15 variables.
\describe{ \item{id}{soil id} \item{top}{horizon upper
boundary (cm)} \item{bottom}{horizon lower boundary (cm)}
\item{clay}{clay content} \item{cec}{CEC by amonium acetate
at pH 7} \item{ph}{pH in 1:1 water-soil mixture}
\item{tc}{total carbon percent} \item{hue}{Munsell hue
(dry)} \item{value}{Munsell value (dry)}
\item{chroma}{Munsell chroma (dry)} \item{mid}{horizon
midpoint (cm)} \item{ln_tc}{natural log of total carbon percent}
\item{L}{color: l-coordinate, CIE-LAB colorspace (dry)}
\item{A}{color: a-coordinate, CIE-LAB colorspace (dry)}
\item{B}{color: b-coordinate, CIE-LAB colorspace (dry)}
\item{name}{horizon name} \item{soil_color}{horizon color} }
}
\description{
Soil samples from 10 soil profiles, taken from the Sierra Foothill Region of
California.
}
\details{
These data were collected to support research funded by the Kearney
Foundation of Soil Science.
}
\examples{

## this example investigates the concept of a "median profile"

# required packages
if (require(ape) & require(cluster)) {
  data(sp3)

  # generate a RGB version of soil colors
  # and convert to HSV for aggregation
  sp3$h <- NA
  sp3$s <- NA
  sp3$v <- NA
  sp3.rgb <- with(sp3, munsell2rgb(hue, value, chroma, return_triplets = TRUE))

  sp3[, c('h', 's', 'v')] <- t(with(sp3.rgb, rgb2hsv(r, g, b, maxColorValue = 1)))

  # promote to SoilProfileCollection
  depths(sp3) <- id ~ top + bottom

  # aggregate across entire collection
  a <- slab(sp3, fm = ~ clay + cec + ph + h + s + v, slab.structure = 10)

  # check
  str(a)

  # convert back to wide format
  library(data.table)

  a.wide.q25 <- dcast(as.data.table(a), top + bottom ~ variable, value.var = c('p.q25'))
  a.wide.q50 <- dcast(as.data.table(a), top + bottom ~ variable, value.var = c('p.q50'))
  a.wide.q75 <- dcast(as.data.table(a), top + bottom ~ variable, value.var = c('p.q75'))

  # add a new id for the 25th, 50th, and 75th percentile pedons
  a.wide.q25$id <- 'Q25'
  a.wide.q50$id <- 'Q50'
  a.wide.q75$id <- 'Q75'

  # combine original data with "mean profile"
  vars <- c('top', 'bottom', 'id', 'clay', 'cec', 'ph', 'h', 's', 'v')
  # make data.frame version of sp3
  sp3.df <- as(sp3, 'data.frame')

  sp3.grouped <- as.data.frame(rbind(as.data.table(horizons(sp3))[, .SD, .SDcol = vars], 
                                     a.wide.q25[, .SD, .SDcol = vars],
                                     a.wide.q50[, .SD, .SDcol = vars], 
                                     a.wide.q75[, .SD, .SDcol = vars]))
                                     
  # re-constitute the soil color from HSV triplets
  # convert HSV back to standard R colors
  sp3.grouped$soil_color <- with(sp3.grouped, hsv(h, s, v))

  # give each horizon a name
  sp3.grouped$name <- paste(
    round(sp3.grouped$clay),
    '/' ,
    round(sp3.grouped$cec),
    '/',
    round(sp3.grouped$ph, 1)
  )


# first promote to SoilProfileCollection
depths(sp3.grouped) <- id ~ top + bottom

plot(sp3.grouped)

## perform comparison, and convert to phylo class object
## D is rescaled to [0,]
d <- NCSP(
  sp3.grouped,
  vars = c('clay', 'cec', 'ph'),
  maxDepth = 100,
  k = 0.01
)

h <- agnes(d, method = 'ward')
p <- ladderize(as.phylo(as.hclust(h)))

# look at distance plot-- just the median profile
plot_distance_graph(d, 12)

# similarity relative to median profile (profile #12)
round(1 - (as.matrix(d)[12, ] / max(as.matrix(d)[12, ])), 2)

## make dendrogram + soil profiles

# setup plot: note that D has a scale of [0,1]
par(mar = c(1, 1, 1, 1))
p.plot <- plot(p,
               cex = 0.8,
               label.offset = 3,
               direction = 'up',
               y.lim = c(200, 0),
               x.lim = c(1.25, length(sp3.grouped) + 1),
               show.tip.label = FALSE)

# get the last plot geometry
lastPP <- get("last_plot.phylo", envir = ape::.PlotPhyloEnv)

# the original labels, and new (indexed) order of pedons in dendrogram
d.labels <- attr(d, 'Labels')

new_order <- sapply(1:lastPP$Ntip,
                    function(i)
                      which(as.integer(lastPP$xx[1:lastPP$Ntip]) == i))

# plot the profiles, in the ordering defined by the dendrogram
# with a couple fudge factors to make them fit
plotSPC(
  sp3.grouped,
  color = "soil_color",
  plot.order = new_order,
  y.offset = max(lastPP$yy) + 10,
  width = 0.1,
  cex.names = 0.5,
  add = TRUE
)
}
}
\references{
\url{https://casoilresource.lawr.ucdavis.edu/}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-setters.R
\docType{methods}
\name{replaceHorizons<-}
\alias{replaceHorizons<-}
\alias{replaceHorizons<-,SoilProfileCollection-method}
\title{Replace Data in Horizon Slot}
\usage{
\S4method{replaceHorizons}{SoilProfileCollection}(object) <- value
}
\arguments{
\item{object}{A \code{SoilProfileCollection}}

\item{value}{An object inheriting \code{data.frame}}
}
\description{
Replaces horizon data with new \code{data.frame} object.
}
\examples{

# load test data
data(sp2)

# promote to SPC
depths(sp2) <- id ~ top + bottom

# one profile
p <- sp2[1,]

# 23 variables in horizon data
length(horizonNames(sp2))

# remove all but essential ones
replaceHorizons(p) <- horizons(p)[,c(idname(p), hzidname(p), horizonDepths(p))]

# inspect result (a clean slate)
horizons(p)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flagOverlappingHz.R
\name{flagOverlappingHz}
\alias{flagOverlappingHz}
\title{Flag perfectly overlapping horizons within a SoilProfileCollection}
\usage{
flagOverlappingHz(x)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}
}
\value{
logical vector with length (and order) matching the horizons of \code{x}
}
\description{
Flag perfectly overlapping horizons within a SoilProfileCollection
}
\details{
Horizons with \code{NA} depths can be flagged as overlapping. Consider finding these horizons with \code{checkHzDepthLogic(byhz=TRUE)} and removing or fixing them.
}
\examples{

# two overlapping horizons
z <- data.frame(
  id = 'SPC',
  top = c(0, 25, 25, 50, 75, 100, 100),
  bottom = c(25, 50, 50, 75, 100, 125, 125)
)

# init SPC
depths(z) <- id ~ top + bottom

# flag perfectly overlapping horizons
z$.overlapFlag <- flagOverlappingHz(z)

# thematic sketches
plotSPC(z, color = '.overlapFlag', hz.depths = TRUE, 
depth.axis = FALSE, cex.names = 0.85)

}
\seealso{
\code{\link[=checkHzDepthLogic]{checkHzDepthLogic()}} \code{\link[=fillHzGaps]{fillHzGaps()}}
}
\author{
D.E. Beaudette, A.G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/soilColorIndices.R
\name{hurst.redness}
\alias{hurst.redness}
\title{Hurst (1977) Redness Index}
\usage{
hurst.redness(hue, value, chroma)
}
\arguments{
\item{hue}{A character vector containing Munsell hues (e.g. "7.5YR")}

\item{value}{A numeric vector containing Munsell values}

\item{chroma}{A numeric vector containing Munsell chromas}
}
\value{
A numeric vector of horizon redness index (lower values = redder).
}
\description{
Calculate Redness Index after Hurst (1977) "Visual estimation of iron in saprolite" DOI: 10.1130/0016-7606(1977)88<174:VEOIIS>2.0.CO;2. Accepts vectorized inputs for hue, value and chroma, produces vector output.
}
\references{
Hurst, V.J. (1977) Visual estimation of iron in saprolite. GSA Bulletin. 88(2): 174–176. doi: https://doi.org/10.1130/0016-7606(1977)88<174:VEOIIS>2.0.CO;2
}
\author{
Andrew G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-methods.R
\docType{methods}
\name{unique,SoilProfileCollection-method}
\alias{unique,SoilProfileCollection-method}
\alias{unique}
\title{Uniqueness within a \code{SoilProfileCollection} via MD5 Hash}
\usage{
\S4method{unique}{SoilProfileCollection}(x, vars, SPC = TRUE)
}
\arguments{
\item{x}{a \code{SoilProfileCollection}}

\item{vars}{Variables to consider in uniqueness.}

\item{SPC}{logical return a \code{SoilProfileCollection} when \code{TRUE}, otherwise vector of profile indices}
}
\value{
\code{SoilProfileCollection} when \code{SPC = TRUE}, otherwise a vector of integers
}
\description{
Unique profiles within a \code{SoilProfileCollection} using and MD5 hash of select horizon and / or site level attributes.
}
\examples{

  # an example soil profile
  x <- data.frame(
    id = 'A',
    name = c('A', 'E', 'Bhs', 'Bt1', 'Bt2', 'BC', 'C'),
    top = c(0, 10, 20, 30, 40, 50, 100),
    bottom = c(10, 20, 30, 40, 50, 100, 125),
    z = c(8, 5, 3, 7, 10, 2, 12)
  )
  
  # init SPC
  depths(x) <- id ~ top + bottom
  
  # horizon depth variability for simulation
  horizons(x)$.sd <- 2
  
  # duplicate several times
  x.dupes <- duplicate(x, times = 5)
  
  # simulate some new profiles based on example
  x.sim <- perturb(x, n = 5, thickness.attr = '.sd')
  
  # graphical check
  plotSPC(x.dupes, name.style = 'center-center')
  plotSPC(x.sim, name.style = 'center-center')
  
  # inspect unique results
  plotSPC(unique(x.dupes, vars = c('top', 'bottom')), name.style = 'center-center')
  
  # uniqueness is a function of variable selection
  plotSPC(unique(x.sim, vars = c('top', 'bottom')), name.style = 'center-center')
  plotSPC(unique(x.sim, vars = c('name')), name.style = 'center-center')
  


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/munsell2rgb.R
\name{munsell2spc,SoilProfileCollection-method}
\alias{munsell2spc,SoilProfileCollection-method}
\alias{munsell2spc}
\title{Merge Munsell Hue, Value, Chroma converted to sRGB & CIELAB into a SoilProfileCollection}
\usage{
\S4method{munsell2spc}{SoilProfileCollection}(
  object,
  hue = "hue",
  value = "value",
  chroma = "chroma",
  .data = NULL,
  as.spc = TRUE
)
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{hue}{Column name containing numeric hue values. Default: \code{"hue"}}

\item{value}{Column name containing numeric value values. Default: \code{"value"}}

\item{chroma}{Column name containing numeric chroma values. Default: \code{"chroma"}}

\item{.data}{Optional: a character vector of equal length to number of horizons (containing Munsell notation), or a column name in the horizon data OR a data.frame containing three columns (names specified in \code{hue}, \code{value}, \code{chroma})}

\item{as.spc}{Return a data.frame-like object with ID columns?}
}
\value{
A SoilProfileCollection or \code{data.frame}-like object
}
\description{
Convert Munsell hue, value and chroma into sRGB (\code{rgb_R, rgb_G, rgb_B}) and CIELAB (lab_L, lab_A, lab_B) color coordinates using \code{munsell2rgb}. The converted values are stored in the \code{horizons()} slot unless \code{as.spc} is \code{FALSE}, in which case the results are combined with profile and horizon ID columns and returned as the \code{data.frame} subclass used by the SPC.
}
\examples{

data(sp3)
depths(sp3) <- id ~ top + bottom

# inspect input data
horizons(sp3)[,c("hue","value","chroma")]

# do color conversions to sRGB and LAB, join into horizon data
sp3 <- munsell2spc(sp3)

# plot rgb "R" coordinate by horizon
plot(sp3, color = "rgb_R")

# plot lab "A" coordinate by horizon
plot(sp3, color = "lab_A")

# note that `lab_A` values do not exactly match the original `A` values
# this is because `lab_A` was computed from the (field determined) Munsell color notation,
# while `A` was directly measured in the lab by colorimeter
plot(sp3$A, sp3$lab_A, xlab = 'Measured', ylab = 'Converted from Field Observed Munsell')

}
\seealso{
\code{\link{parseMunsell}} \code{\link{rgb2munsell}} \code{\link{munsell2rgb}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{munsellHuePosition}
\alias{munsellHuePosition}
\title{Munsell Hue Position Reference}
\format{
An object of class \code{data.frame} with 41 rows and 4 columns.
}
\usage{
data(munsellHuePosition)
}
\description{
Position data for the 40 standard Munsell hues (and neutral). Data include angular positions (compass-style, origin at \verb{[x = 0, y = 1]}, CW rotation) and Cartesian coordinates on the unit circle.
}
\references{
Munsell book of color. 1976. Macbeth, a Division of Kollmorgen Corp., Baltimore, MD.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{profile_id<-}
\alias{profile_id<-}
\alias{profile_id<-,SoilProfileCollection-method}
\alias{profile_id,SoilProfileCollection-method}
\alias{profile_id}
\title{Set profile IDs}
\usage{
\S4method{profile_id}{SoilProfileCollection}(object) <- value

\S4method{profile_id}{SoilProfileCollection}(object)
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{value}{a unique vector of equal length to number of profiles \code{length(object)}}
}
\description{
Set vector containing profile IDs

Get or set a vector of profile IDs
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{sp1}
\alias{sp1}
\title{Soil Profile Data Example 1}
\format{
A data frame with 60 observations on the following 21 variables.
\describe{ \item{group}{a numeric vector} \item{id}{a
character vector} \item{top}{a numeric vector}
\item{bottom}{a numeric vector} \item{bound_distinct}{a
character vector} \item{bound_topography}{a character vector}
\item{name}{a character vector} \item{texture}{a character
vector} \item{prop}{a numeric vector}
\item{structure_grade}{a character vector}
\item{structure_size}{a character vector}
\item{structure_type}{a character vector}
\item{stickiness}{a character vector} \item{plasticity}{a
character vector} \item{field_ph}{a numeric vector}
\item{hue}{a character vector} \item{value}{a numeric
vector} \item{chroma}{a numeric vector} }
}
\description{
Soil profile data from Pinnacles National Monument, CA.
}
\examples{

data(sp1)
# convert colors from Munsell to hex-encoded RGB
sp1$soil_color <- with(sp1, munsell2rgb(hue, value, chroma))

# promote to SoilProfileCollection
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# re-sample each profile into 1 cm (thick) depth slices
# for the variables 'prop', 'name', 'soil_color'
# result is a SoilProfileCollection object
s <- dice(sp1, 0:25 ~ prop + name + soil_color)

# plot, note slices
plot(s)


# aggregate all profiles along 1 cm depth slices,
# using data from column 'prop'
s1 <- slab(sp1, fm= ~ prop)

# check median & IQR
library(lattice)
xyplot(top ~ p.q50 + p.q25 + p.q75,
data=s1, type='S', horizontal=TRUE, col=1, lty=c(1,2,2),
panel=panel.superpose, ylim=c(110,-5), asp=2)


}
\references{
\url{https://casoilresource.lawr.ucdavis.edu/}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{traditionalColorNames}
\alias{traditionalColorNames}
\title{Traditional Soil Color Names}
\format{
An object of class \code{data.frame} with 482 rows and 2 columns.
}
\usage{
data(traditionalColorNames)
}
\description{
Traditional soil color names associated with select Munsell colors.
}
\references{
Sourced from the "colorconverter" NASIS property script.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-documentation.R
\docType{data}
\name{soil_minerals}
\alias{soil_minerals}
\title{Munsell Colors of Common Soil Minerals}
\format{
A data frame with 20 observations on the following 5 variables.
\describe{ \item{mineral}{mineral name} \item{color}{Munsell
color} \item{hue}{Munsell hue} \item{value}{Munsell value}
\item{chroma}{Munsell chroma} }
}
\usage{
data(soil_minerals)
}
\description{
Munsell colors for some common soil minerals.
}
\details{
Soil color and other properties including texture, structure, and
consistence are used to distinguish and identify soil horizons (layers) and
to group soils according to the soil classification system called Soil
Taxonomy. Color development and distribution of color within a soil profile
are part of weathering. As rocks containing iron or manganese weather, the
elements oxidize. Iron forms small crystals with a yellow or red color,
organic matter decomposes into black humus, and manganese forms black
mineral deposits. These pigments paint the soil (Michigan State Soil). Color
is also affected by the environment: aerobic environments produce sweeping
vistas of uniform or subtly changing color, and anaerobic (lacking oxygen),
wet environments disrupt color flow with complex, often intriguing patterns
and points of accent. With depth below the soil surface, colors usually
become lighter, yellower, or redder.
}
\examples{

\dontrun{

library(aqp)
library(ape)
library(cluster)
library(farver)

# load common soil mineral colors
data(soil_minerals)

# convert Munsell to R colors
soil_minerals$col <- munsell2rgb(
  soil_minerals$hue, 
  soil_minerals$value,
  soil_minerals$chroma
)

# make a grid for plotting
n <- ceiling(sqrt(nrow(soil_minerals)))

# read from top-left to bottom-right
g <- expand.grid(x = 1:n, y = n:1)[1:nrow(soil_minerals),]

# convert Munsell -> CIELAB
col.lab <- parseMunsell(soil_minerals$color, returnLAB = TRUE)

# keep track of soil mineral names
# in a way that will persist in a dist obj
row.names(col.lab) <- soil_minerals$mineral

# pair-wise perceptual distances via CIE dE00
d <- compare_colour(
  from = col.lab,
  from_space = 'lab',
  to_space = 'lab',
  white_from = 'D65',
  method = 'CIE2000'
)

# matrix -> dist
# note transpose, 
# required when specifying only `from` to compare_colour()
d <- as.dist(t(d))

# divisive hierarchical clustering of LAB coordinates
h <- as.hclust(diana(d))
p <- as.phylo(h)

# colors, in order based on clustering
# starting from top-left
min.cols <- rev(soil_minerals$col[h$order])

# mineral names, in order based on clustering
# starting from top-left
min.names <- rev(soil_minerals$mineral[h$order])

min.munsell <- rev(soil_minerals$color[h$order])

# plot grid of mineral names / colors
layout(matrix(c(1, 2), nrow = 1), widths = c(1.25, 1))

par(mar = c(1, 0, 0, 1))
plot(g$x, g$y, pch = 15, cex = 12, axes = FALSE, xlab = '', ylab = '',
     col = min.cols, 
     xlim = c(0.5, 5.5), ylim = c(1.5, 5.5)
)
text(g$x, g$y, min.names, adj = c(0.45, 5.5), cex = 0.75, font = 2)
text(g$x, g$y, min.munsell, col = invertLabelColor(min.cols), cex = 0.85, font = 2)

title(main = 'Common Soil Pigments', line = -1.75, cex.main = 2)
mtext('U. Schwertmann, 1993. SSSA Special Publication no. 31, pages 51--69', side = 1,
      cex = 0.75, line = -1.5)

# dendrogram + tip labels with mineral colors
plot(p, cex = 0.85, label.offset = 5, font = 1)
tiplabels(pch = 15, cex = 3, offset = 2, col = soil_minerals$col)

}


}
\references{
\enumerate{
\item Lynn, W.C. and Pearson, M.J., The Color of Soil, The Science
Teacher, May 2000. 2. Schwertmann, U. 1993. Relations Between Iron Oxides,
Soil Color, and Soil Formation. "Soil Color". SSSA Special Publication no.
31, pages 51--69.
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-iterators.R
\docType{methods}
\name{profileApply}
\alias{profileApply}
\alias{profileApply,SoilProfileCollection-method}
\title{Iterate over profiles in a SoilProfileCollection}
\usage{
\S4method{profileApply}{SoilProfileCollection}(
  object,
  FUN,
  simplify = TRUE,
  frameify = FALSE,
  chunk.size = 100,
  column.names = NULL,
  APPLY.FUN = lapply,
  ...
)
}
\arguments{
\item{object}{a SoilProfileCollection}

\item{FUN}{a function to be applied to each profile within the collection}

\item{simplify}{logical, should the result be simplified to a vector? default: TRUE; see examples}

\item{frameify}{logical, should the result be collapsed into a data.frame? default: FALSE; overrides simplify argument; see examples}

\item{chunk.size}{numeric, size of "chunks" for faster processing of large SoilProfileCollection objects; default: 100}

\item{column.names}{character, optional character vector to replace frameify-derived column names; should match length of colnames() from FUN result; default: NULL}

\item{APPLY.FUN}{function, optional alternate \code{lapply()}-like function for processing chunks. For example \code{future.apply::future_lapply()} for processing chunks in parallel. Default \code{base::lapply()}.}

\item{...}{additional arguments passed to FUN}
}
\value{
When simplify is TRUE, a vector of length nrow(object) (horizon data) or of length length(object) (site data). When simplify is FALSE, a list is returned. When frameify is TRUE, a data.frame is returned. An attempt is made to identify idname and/or hzidname in the data.frame result, safely ensuring that IDs are preserved to facilitate merging profileApply result downstream.
}
\description{
Iterate over all profiles in a SoilProfileCollection, calling \code{FUN} on a single-profile SoilProfileCollection for each step.
}
\examples{

data(sp1)
depths(sp1) <- id ~ top + bottom

# estimate soil depth using horizon designations
profileApply(sp1, estimateSoilDepth, name='name')

# scale a single property 'prop' in horizon table
# scaled = (x - mean(x)) / sd(x)
sp1$d <- profileApply(sp1, FUN=function(x) round(scale(x$prop), 2))
plot(sp1, name='d')

# compute depth-wise differencing by profile
# note that our function expects that the column 'prop' exists
f <- function(x) { c(x$prop[1], diff(x$prop)) }
sp1$d <- profileApply(sp1, FUN=f)
plot(sp1, name='d')

# compute depth-wise cumulative sum by profile
# note the use of an anonymous function
sp1$d <- profileApply(sp1, FUN=function(x) cumsum(x$prop))
plot(sp1, name='d')

# compute profile-means, and save to @site
# there must be some data in @site for this to work
site(sp1) <- ~ group
sp1$mean_prop <- profileApply(sp1, FUN=function(x) mean(x$prop, na.rm=TRUE))

# re-plot using ranks defined by computed summaries (in @site)
plot(sp1, plot.order=rank(sp1$mean_prop))

## iterate over profiles, calculate on each horizon, merge into original SPC

# example data
data(sp1)

# promote to SoilProfileCollection
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# calculate horizon thickness and proportional thickness
# returns a data.frame result with multiple attributes per horizon
thicknessFunction <- function(p) {
  hz <- horizons(p)
  depthnames <- horizonDepths(p)
  res <- data.frame(profile_id(p), hzID(p),
                    thk=(hz[[depthnames[[2]]]] - hz[[depthnames[1]]]))
  res$hz_prop <- res$thk / sum(res$thk)
  colnames(res) <- c(idname(p), hzidname(p), 'hz_thickness', 'hz_prop')
  return(res)
}

# list output option with simplify=F, list names are profile_id(sp1)
list.output <- profileApply(sp1, thicknessFunction, simplify = FALSE)
head(list.output)

# data.frame output option with frameify=TRUE
df.output <- profileApply(sp1, thicknessFunction, frameify = TRUE)
head(df.output)

# since df.output contains idname(sp1) and hzidname(sp1),
# it can safely be merged by a left-join via horizons<- setter
horizons(sp1) <- df.output

plot(density(sp1$hz_thickness, na.rm=TRUE), main="Density plot of Horizon Thickness")

## iterate over profiles, subsetting horizon data

# example data
data(sp1)

# promote to SoilProfileCollection
depths(sp1) <- id ~ top + bottom
site(sp1) <- ~ group

# make some fake site data related to a depth of some importance
sp1$dep <- profileApply(sp1, function(i) {round(rnorm(n=1, mean=mean(i$top)))})

# custom function for subsetting horizon data, by profile
# keep horizons with lower boundary < site-level attribute 'dep'
fun <- function(i) {
  # extract horizons
  h <- horizons(i)
  # make an expression to subset horizons
  exp <- paste('bottom < ', i$dep, sep='')
  # subset horizons, and write-back into current SPC
  slot(i, 'horizons') <- subset(h, subset=eval(parse(text=exp)))
  # return modified SPC
  return(i)
}

# list of modified SoilProfileCollection objects
l <- profileApply(sp1, fun, simplify=FALSE)

# re-combine list of SoilProfileCollection objects into a single SoilProfileCollection
sp1.sub <- pbindlist(l)

# graphically check
par(mfrow=c(2,1), mar=c(0,0,1,0))
plot(sp1)
points(1:length(sp1), sp1$dep, col='red', pch=7)
plot(sp1.sub)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SoilProfileCollection-integrity.R
\docType{methods}
\name{reorderHorizons}
\alias{reorderHorizons}
\alias{reorderHorizons,SoilProfileCollection-method}
\title{Re-order corrupted horizon data}
\usage{
\S4method{reorderHorizons}{SoilProfileCollection}(object, target.order = NULL)
}
\arguments{
\item{object}{A SoilProfileCollection}

\item{target.order}{A numeric vector of equal length to \code{object}. Default value is \code{NULL} which restores the internal order of the collection.}
}
\value{
SoilProfileCollection
}
\description{
This is a method made available primarily to repair horizon data that have been corrupted relative to their order at time of SoilProfileCollection construction.

There is an option to specify the target order, but this will not update the corresponding metadata entry tracking the original order. Use this functionality at your own risk.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slicedHSD.R
\name{slicedHSD}
\alias{slicedHSD}
\title{Tukey's HSD Over Slices}
\usage{
slicedHSD(object, fm, conf = 0.95)
}
\arguments{
\item{object}{\code{SoilProfileCollection} object}

\item{fm}{a formula describing depth sequence, horizon attribute, and site (grouping) attribute.
For example 0:100 ~ estimated_oc | taxonname}

\item{conf}{confidence applied in \code{TukeyHSD}}
}
\description{
Apply Tukey's HSD over 1-unit depth slices.
}
\author{
D.E. Beaudette and Sharon Perrone
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aqp-label-placement-solvers.R
\name{findOverlap}
\alias{findOverlap}
\alias{overlapMetrics}
\title{Find Overlap within a Sequence}
\usage{
findOverlap(x, thresh)

overlapMetrics(x, thresh)
}
\arguments{
\item{x}{vector of relative horizontal positions, one for each profile}

\item{thresh}{threshold defining "overlap", typically < 1

@return a \code{list}:
\itemize{
\item \code{idx}: unique index to overlapping elements in \code{x}
\item \code{ov}: normalized overlap (see details)
}}
}
\value{
unique index to affected (overlapping) elements in \code{x}
}
\description{
Establish which elements within a vector of horizontal positions overlap beyond a given threshold

Desc.
}
\examples{

x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)

findOverlap(x, thresh = 0.5)


x <- c(1, 2, 3, 3.4, 3.5, 5, 6, 10)

overlapMetrics(x, thresh = 0.5)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.ml.hz.R
\name{get.ml.hz}
\alias{get.ml.hz}
\title{Determine ML Horizon Boundaries}
\usage{
get.ml.hz(x, o.names = attr(x, which = "original.levels"))
}
\arguments{
\item{x}{\code{data.frame}, output from \code{\link[=slab]{slab()}}}

\item{o.names}{an optional character vector of horizon designations that will be used in the final table}
}
\value{
A \code{data.frame} with the following columns:
\itemize{
\item \code{hz}: horizon names
\item \code{top}: horizon top depth
\item \code{bottom}: horizon bottom depth
\item \code{confidence}: integrated probability over thickness of each ML horizon, rounded to the nearest integer
\item \code{pseudo.brier}: A "pseudo"" Brier Score for a multi-class prediction, where the most-likely horizon label is treated as the "correct" outcome. Details on the calculation for traditional Brier Scores here: \url{https://en.wikipedia.org/wiki/Brier_score}. Lower values suggest better agreement between ML horizon label and class-wise probabilities.
\item \code{mean.H}: mean Shannon entropy (bits), derived from probabilities within each most-likely horizon. Larger values suggest more confusion within each ML.
}
}
\description{
This function accepts input from \code{\link[=slab]{slab()}} (a \code{data.frame}) along with a vector of
horizon names, and returns a \code{data.frame} of the most likely horizon
boundaries.

This function expects that \code{x} is a \code{data.frame} generated by
\code{\link[=slab]{slab()}}. If \code{x} was not generated by \code{\link[=slab]{slab()}}, then \code{o.names} is required.
}
\examples{

# init SPC
data(sp1)
depths(sp1) <- id ~ top + bottom

# set horizon designation metadata
hzdesgnname(sp1) <- 'name'

# generalize horizon designations from character vector
# result is an ordered factor
sp1$genhz <- generalizeHz(
  sp1$name,
  new = c('O','A','B','C'),
  pat = c('O', '^A','^B','C'),
  ordered = TRUE
)

# compute slice-wise GHL probability
# so that it sums to contributing fraction
# from 0-150cm
a <- slab(sp1, fm = ~ genhz, cpm = 1, slab.structure = 0:150)

# note original GHL names are set by slab()
attr(a, 'original.levels')

# generate table of ML horizonation
get.ml.hz(a)

}
\references{
Beaudette, D. E., Roudier, P., & Skovlin, J. (2016). Probabilistic representation of genetic soil horizons. Digital soil morphometrics, 281-293.
}
\seealso{
\code{\link[=slab]{slab()}}
}
\author{
D.E. Beaudette
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dice.R
\name{dice,SoilProfileCollection-method}
\alias{dice,SoilProfileCollection-method}
\alias{dice}
\title{Efficient Slicing of \code{SoilProfileCollection} Objects}
\usage{
\S4method{dice}{SoilProfileCollection}(
  x,
  fm = NULL,
  SPC = TRUE,
  pctMissing = FALSE,
  fill = FALSE,
  strict = TRUE,
  byhz = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object}

\item{fm}{optional \code{formula} describing top depths and horizon level attributes to include: \code{integer.vector ~ var1 + var2 + var3} or \code{integer.vector ~ .} to include all horizon level attributes. Specification of \code{integer.vector} forces \code{fill = TRUE}. When \code{NULL} profiles are "diced" to depth and results will include all horizon level attributes. Note on interpretation of \code{integer.vector} (slice tops)}

\item{SPC}{return the diced \code{SoilPrfolileCollection}, if \code{FALSE} a \code{data.frame} of horizon-level attributes}

\item{pctMissing}{compute "percent missing data" by slice (when \code{TRUE} expect 6-8x longer run time)}

\item{fill}{logical, fill with empty placeholder horizons in gaps within profiles, and/or, above/below interval specified in \code{fm}. Automatically set to \code{TRUE} when LHS of \code{fm} is specified. Backwards compatibility with \code{slice} is maintained by setting \code{fill = TRUE} with or without \code{fm}.}

\item{strict}{perform horizon depth logic checking / flagging / removal}

\item{byhz}{Evaluate horizon depth logic at the horizon level (\code{TRUE}) or profile level (\code{FALSE}). Invalid depth logic invokes \code{HzDepthLogicSubset} which removes offending profiles or horizon records.}

\item{verbose}{Print information about object size/memory usage. Default: \code{FALSE}}
}
\value{
a \code{SoilProfileCollection} object, or \code{data.frame} when \code{SPC = FALSE}
}
\description{
Cut ("dice") soil horizons into 1-unit thick slices. This function replaces \code{aqp::slice()}, which will be deprecated in aqp 2.0.
}
\details{
For large and potentially messy collections that may include missing horizon depth logic errors, consider using \code{repairMissingHzDepths()} before \code{dice()}. Consider using \code{accumulateDepths()} before invoking \code{dice()} on collections that may contain old-style O horizon notation (e.g. 5-0cm).
}
\seealso{
\code{\link[=repairMissingHzDepths]{repairMissingHzDepths()}}, \code{\link[=accumulateDepths]{accumulateDepths()}}, \code{\link[=fillHzGaps]{fillHzGaps()}}
}
\author{
D.E. Beaudette and A.G. Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-SoilProfileCollection.R
\docType{methods}
\name{idname,SoilProfileCollection-method}
\alias{idname,SoilProfileCollection-method}
\alias{idname}
\title{Get profile ID column name}
\usage{
\S4method{idname}{SoilProfileCollection}(object)
}
\arguments{
\item{object}{a SoilProfileCollection}
}
\description{
Get column name containing unique profile IDs
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/depthWeights.R
\name{depthWeights}
\alias{depthWeights}
\alias{depthWeights,SoilProfileCollection-method}
\title{Return a vector of contributing fractions over a depth interval}
\usage{
depthWeights(top, bottom, upper, lower)
}
\arguments{
\item{top}{A numeric vector of horizon top depths.}

\item{bottom}{A numeric vector of horizon bottom depths.}

\item{upper}{A unit length numeric vector with upper boundary.}

\item{lower}{A unit length numeric vector with lower boundary.}
}
\value{
A named list.
}
\description{
\code{depthWeights()} calculates the contributing fraction for each pair of horizon top and bottom depths, given an upper and lower boundary.
}
\author{
Andrew G. Brown.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duplicate.R
\name{duplicate}
\alias{duplicate}
\title{Duplicate Profiles of a SoilProfileCollection}
\usage{
duplicate(x, times = 3, oldID = ".oldID")
}
\arguments{
\item{x}{a \code{SoilProfileCollection} object with 1 or more profiles}

\item{times}{requested number of copies}

\item{oldID}{site-level attribute used to store the original profile IDs}
}
\value{
a \code{SoilProfileCollection} object
}
\description{
A simple function to duplicate the contents of a \code{SoilProfileCollection} object. Old profile IDs are saved as a site-level attribute (\code{oldID}) and new IDs are generated using a numeric serial number.
}
\examples{

# sample data
data('sp4')

# promote to SPC
depths(sp4) <- id ~ top + bottom

# duplicate each profile 2 times
d <- duplicate(sp4, times = 2)

# graphical check
par(mar = c(0, 0, 3, 1))
plotSPC(d, color = 'Ca', width = 0.25)

}
\author{
D.E. Beaudette
}
\keyword{manip}
